suite suiteName
   defstatus suspended
   family familyName0
      defstatus complete
      task taskName0
      	 defstatus complete  
      	 defstatus complete  
  		 edit VAR2 value2
   endfamily
endsuite
