suite cron

   clock real  # Need real clock otherwise crons would be complete
   task cron0
      cron 00:25 23:59 00:30
      
   task cron1
      cron -w 0 , 1, 2 -m 5, 6, 7 , 8,9,10      10:00  20:00  01:00  # run every sunday, between May-Aug, every hour between 10am and 8pm

   task cron1_1
      cron -w 0,1,2 -m 9,10  10:00  14:00  01:00  
      
   task cron2
         cron 23:00                              # run every day at 23:00
         
   task cron3
      cron 10:00 20:00 01:00                  # run every hour between 10am and 8pm
      
   task cron4
      cron -w 0,1 10:00                       # run every sunday and monday at 10am
      
   task cron5
      cron -d 10,11,12 12:00                  # run 10th, 11th and 12th of each month at noon
      
   task cron6
      cron -m 1,2,3 12:00                     # run on Jan,Feb and March every day at noon.
      
   task cron7
      cron -w 0,6 12:00
      
   task cron8
      cron -w 0,1,2,3,4,5,6 12:00
      
   task cron9
      cron -w 0 -m 5,6,7,8 10:00 20:00 01:00  # run every sunday, between May-Aug, every hour between 10am and 8pm
   
   task cron10
      cron -w 0,1,2 -d 1,20,24 -m 5,6,7,8 10:00 20:00 01:00  #  


   task cron11
      cron 00:00 23:59 00:01  #  run every minute
      
   family f1
      cron -d 22 12:00
      task t1
      task t2
   endfamily
   family f2
      cron -d 8 12:00
      task t1
      task t2
   endfamily
   
   family t1
      task t1
         cron 00:01
         cron 23:59
         
     task t2
         cron 09:00
         cron 12:00

     task t3
         cron 12:00
         cron 18:00

     task t4
         cron 18:00
         cron 23:00
         
     task t5
         cron 10:00
    task t6
         cron 11:00
    task t7
         cron 12:00
    task t8
         cron 13:00
    task t9
         cron 14:00
    task t10
         cron 15:00
    task t11
         cron 16:00
    task t12
         cron 17:00
    task t13
         cron 18:00
    task t14
         cron 19:00
    task t15
         cron 20:00
    task t16
         cron 21:00
    task t17
         cron 22:00
    task t18
         cron 23:00
    task t19
         cron 23:00
    task t20
         cron 00:00   
   endfamily          
endsuite