suite repeat_enumerated1
   
  repeat enumerated quotes "first" "second" "third"  # sdsdsd 
  task t1

endsuite

suite repeat_enumerated2
   
 repeat enumerated quotes 'first' 'second' "third"  # sdsdsd 
  task t1
   
endsuite