suite date
   task a
      date *.*.2009
   endtask
   family familyName0
   	  date 1.*.2009
      task taskName0
   	  		date 1.12.2009
      task fred
   	  		date 01.*.*
      endtask
   endfamily
endsuite

suite date_s1;
   family familyName0
   	  date 1.*.2009
      task taskName0
   	  		date 1.12.2009
      task fred
   	  		date 01.*.*
      endtask
   endfamily
endsuite

