suite csim_repeat_day
  clock real 04.08.2019            # Sunday
  endclock   13.08.2019            # added endclock to avoid simulation for a year due to repeat
  family f
    repeat integer rep 0 1
    family f1
       day monday                   # This acts like a guard over time attributes. They are not freed until day is free
       task t1
            time 10:00   
            time 12:00   
            verify complete:4      
    endfamily
  endfamily
endsuite

suite csim_repeat_day3
  clock real 04.08.2019   # Sunday
  endclock   31.08.2019   # added endclock to avoid simulation for a year due to repeat
  family f
    repeat integer rep 0 1
    family f1
       repeat integer rep 0 1
       task t1
            time 11:00   
            day monday          
            verify complete:4     
    endfamily
  endfamily
endsuite
