suite repeat_date
  repeat date YMD 20090331 20121212 1 # status 0
  task t1
endsuite

suite repeat_date_s2
  repeat date YMD 20121212 20090331  -1 # status 0
  task t1
endsuite

suite repeat_date_s3; repeat date YMD 20090331 20121212 1 # status 0
  task t1
endsuite