suite suite
    task pertinic
        event 1
    task eda_pert
    task rot
    family pert_ic
             complete pertinic:1
             trigger eda_pert
             trigger -a rot==complete
             trigger -a pertinic==complete and not pertinic:1
    endfamily
endsuite
