# 3.1.7
defs_state MIGRATE state>:aborted flag:late,message state_change:983039 modify_change:1018
  edit SLEEP '10'
suite suite #  begun:1 state:aborted dur:01:29:57 flag:message,no_reque
  defstatus suspended
  edit ECF_EXTN '.sms'
  clock real 
  calendar initTime:2013-Sep-10 12:37:03 suiteTime:2013-Sep-10 15:56:00 duration:03:18:57 initLocalTime:2013-Sep-10 12:37:03 lastTime:2013-Sep-10 15:56:00 calendarIncrement:00:01:00
  family family # state:aborted dur:02:18:57 flag:message
    task task # passwd:jxX0gIbR abort<:   Script for /suite/family/task cannot be found:
   ECF_SCRIPT(/tmp/map/work/p4/merlin/workspace/MyProject/Pyext/test/data/CUSTOMER/ECF_HOME/suite/family/task.sms) does not exist:
   Variable ECF_FETCH not defined:
   Variable ECF_FILES not defined:
   Search of directory ECF_HOME(/tmp/map/work/p4/merlin/workspace/MyProject/Pyext/test/data/CUSTOMER/ECF_HOME) failed:
>abort try:2 state:aborted dur:02:19:57 flag:task_aborted,no_script
      edit VARIABLE 'VALUE'
    task rescue # passwd:Bs7hroqB abort<:   Script for /suite/family/rescue cannot be found:
   ECF_SCRIPT(/tmp/map/work/p4/merlin/workspace/MyProject/Pyext/test/data/CUSTOMER/ECF_HOME/suite/family/rescue.sms) does not exist:
   Variable ECF_FETCH not defined:
   Variable ECF_FILES not defined:
   Search of directory ECF_HOME(/tmp/map/work/p4/merlin/workspace/MyProject/Pyext/test/data/CUSTOMER/ECF_HOME) failed:
>abort try:2 state:aborted dur:02:19:57 flag:task_aborted,no_script
      late -s +00:05 -c 00:05
      complete task eq complete
      trigger task == aborted
    task step # passwd:4N9cOzVA abort<:   Script for /suite/family/step cannot be found:
   ECF_SCRIPT(/tmp/map/work/p4/merlin/workspace/MyProject/Pyext/test/data/CUSTOMER/ECF_HOME/suite/family/step.sms) does not exist:
   Variable ECF_FETCH not defined:
   Variable ECF_FILES not defined:
   Search of directory ECF_HOME(/tmp/map/work/p4/merlin/workspace/MyProject/Pyext/test/data/CUSTOMER/ECF_HOME) failed:
>abort try:2 state:aborted dur:02:19:57 flag:task_aborted,no_script
      trigger task:step le step:STEP
      trigger -o task eq complete
      repeat integer STEP 0 100
    family loop_in_time # state:complete
      defstatus complete
      time 12:00
      task hold # state:complete
        complete 1==1
        trigger 1==0
    endfamily
  endfamily
  task household # passwd:31NZt9lV abort<:   Script for /suite/household cannot be found:
   ECF_SCRIPT(/tmp/map/work/p4/merlin/workspace/MyProject/Pyext/test/data/CUSTOMER/ECF_HOME/suite/household.sms) does not exist:
   Variable ECF_FETCH not defined:
   Variable ECF_FILES not defined:
   Search of directory ECF_HOME(/tmp/map/work/p4/merlin/workspace/MyProject/Pyext/test/data/CUSTOMER/ECF_HOME) failed:
>abort try:2 state:aborted dur:02:22:57 flag:task_aborted,no_script
    cron 00:00 23:59 01:00 # nextTimeSlot/13:00
  task monthly # state:queued
    date 1.*.*
  family remote # state:aborted dur:01:29:57 flag:message,no_reque suspended:1
    edit ECF_TRIES '1'
    family c2a # state:aborted dur:01:29:57 flag:no_reque
      edit LOGDIR '/s2o1/emos_esuite/emos_data/log'
      task sub # passwd:gyBC8q.O abort<: PID(28027)  path(/suite/remote/c2a/sub)  exited with status 255 [ /home/ma/emos/bin/smssubmit.cray emos c2a /vol/emos_nc/output/suite/remote/c2a/sub.job1 /s2o1/emos_esuite/emos_data/log/suite/remote/c2a/sub.1 ]>abort try:1 state:aborted dur:01:29:57 flag:task_aborted,ecfcmd_failed,message,no_reque
        edit SCHOST 'c2a'
    endfamily
    family ecgate # state:aborted dur:02:02:57 flag:no_reque
      task sub # passwd:6pIZXvrl abort<: PID(32480)  path(/suite/remote/ecgate/sub)  exited with status 255 [ /home/ma/emos/bin/smssubmit.cray emos ecgate /vol/emos_nc/output/suite/remote/ecgate/sub.job1 /vol/emos_nc/output/suite/remote/ecgate/sub.1 ]>abort try:1 state:aborted dur:02:02:57 flag:task_aborted,ecfcmd_failed,message,no_reque
        edit SCHOST 'ecgate'
    endfamily
    family ibis # state:aborted dur:02:02:57 flag:no_reque
      task sub # passwd:AEX5NtyU abort<: PID(32532)  path(/suite/remote/ibis/sub)  exited with status 1 [ /home/ma/emos/bin/smssubmit.cray emos ibis /vol/emos_nc/output/suite/remote/ibis/sub.job1 /vol/emos_nc/output/suite/remote/ibis/sub.1 ]>abort try:1 state:aborted dur:02:02:57 flag:task_aborted,ecfcmd_failed,message,no_reque
        edit SCHOST 'ibis'
    endfamily
    family bilbo # state:aborted dur:02:02:57 flag:no_reque
      task sub # passwd:30IPtlxu abort<: PID(32543)  path(/suite/remote/bilbo/sub)  exited with status 1 [ /home/ma/emos/bin/smssubmit.cray emos bilbo /vol/emos_nc/output/suite/remote/bilbo/sub.job1 /vol/emos_nc/output/suite/remote/bilbo/sub.1 ]>abort try:1 state:aborted dur:02:02:57 flag:task_aborted,ecfcmd_failed,message,no_reque
        edit SCHOST 'bilbo'
    endfamily
  endfamily
  family catchup2rt # state:aborted dur:03:18:57 flag:message suspended:1
    task doer # passwd:OXQkmELW abort<:   Script for /suite/catchup2rt/doer cannot be found:
   ECF_SCRIPT(/tmp/map/work/p4/merlin/workspace/MyProject/Pyext/test/data/CUSTOMER/ECF_HOME/suite/catchup2rt/doer.sms) does not exist:
   Variable ECF_FETCH not defined:
   Variable ECF_FILES not defined:
   Search of directory ECF_HOME(/tmp/map/work/p4/merlin/workspace/MyProject/Pyext/test/data/CUSTOMER/ECF_HOME) failed:
>abort try:1 state:aborted dur:03:18:57 flag:task_aborted,no_script,message,no_reque
      complete timer eq complete
      trigger 1==0
  endfamily
endsuite
