#!/bin/sh

## Copyright 2009- ECMWF.
## This software is licensed under the terms of the Apache Licence version 2.0 
## which can be obtained at http://www.apache.org/licenses/LICENSE-2.0. 
## In applying this licence, ECMWF does not waive the privileges and immunities 
## granted to it by virtue of its status as an intergovernmental organisation 
## nor does it submit to any jurisdiction. 

cd $WK
 
rm -rf ACore/bin
rm -rf ANattr/bin
rm -rf ANode/bin 
rm -rf ANode/parser/bin 
rm -rf Base/bin
rm -rf CSim/bin
rm -rf Client/bin 
rm -rf Server/bin 
rm -rf Test/bin
rm -rf Pyext/bin 
   