suite suite
  clock hybrid 11.10.2009  # 11 october 2009 was a sunday, task x should complete on begin
  endclock     13.10.2009  # end time
  family family
 	verify complete:1
 	task y
 	  verify complete:1
 	task x
	  trigger ./y == complete
 	  day monday
 	  verify complete:1
  endfamily
endsuite