suite cron6              
  clock real  1.1.2006 # Define a start date for deterministic simulation 
  endclock   17.1.2006 # When to finish. Otherwise we simulate for 1 year due to cron. Ok but wasteful.
  family cronFamily
    task t
  	  cron -d 15,16  -m 1 10:00 11:00 01:00  # run 15,16 January at 10 and 11am	
      verify complete:4	
   endfamily
endsuite
