suite limit_all_inlimit_types
 limit fam 2
 limit task 4
 inlimit task         
 family f1
     limit sub 1
     inlimit -n fam 
     inlimit -s sub         
     task t1
     task t2
     task t3
     task t4
     task t5
     task t6
 endfamily
 family f2
     limit sub 3
     inlimit -s sub         
     inlimit -n fam 
     task t1
     task t2
     task t3
     task t4
     task t5
     task t6
 endfamily
 family f3
     inlimit -n fam 
     task t1
     task t2
     task t3
     task t4
     task t5
     task t6
 endfamily
endsuite