suite suite_hier
	task 4
	task 1
	task 2
	task 3
	family f1
		task 1
		task 2
	endfamily
endsuite
suite suite_hier_s1
	task 4
	task 1
	task 2
	task 3
	family f1
		task f1_t1
		task f1_t2
	endfamily
endsuite