extern /limits:c1a
extern /limits:linux_cluster

suite trigger_limit
    limit sg1  100
    limit mars 100
endsuite

suite trigger_limit_obs
    inlimit /limits:c1a
    family limits
        limit hpcd 20
    endfamily
    
    family anon    
        inlimit limits:hpcd       
        task t1
            inlimit /trigger_limit:sg1        # another suite limit
            trigger /trigger_limit:sg1 < 30
        task t2
            inlimit /trigger_limit_obs/limits:hpcd      # This suite limit
            trigger /trigger_limit_obs/limits:hpcd > 10
        task t3
            inlimit /limits:c1a           # extern suite limit
            trigger /limits:c1a > 10
    endfamily
endsuite
