suite event_spaces
   family f
      task t
      
         event 0 
         
         event name
         
         event 1 eventName
         
         event 2 eventNamea
         
     endfamily
endsuite

