suite trigger_late_flag
    family f1
        task t
        task t1
            event event
    endfamily
    family anon    
        task t1
           trigger /trigger_late_flag/f1/t<flag>late
        task t2
           trigger /trigger_late_flag/f1/t1:event
         task t3
           trigger /<flag>late  # late flag set on Defs. Since defs is alawys there, does not need extern
    endfamily
endsuite
