suite csim_repeat_day3
  clock real 04.08.2019            # Sunday
  endclock   04.09.2019            # added endclock to avoid simulation for a year due to repeat
  family f
    repeat integer rep 0 1
    family f1
       day monday
       task t1
            day tuesday            # cant be free on Monday and Tuesday, 
            time 10:00   
            verify complete:0     
    endfamily
  endfamily
endsuite