/*
 * Copyright 2009- ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef ecflow_viewer_ServerListSyncWidget_HPP
#define ecflow_viewer_ServerListSyncWidget_HPP

#include <vector>

#include <QWidget>

#include "ui_ServerListSyncWidget.h"

class ServerListSyncChangeItem;
class QListWidgetItem;

class ServerListSyncWidget : public QWidget, protected Ui::ServerListSyncWidget {
    Q_OBJECT

public:
    explicit ServerListSyncWidget(QWidget* parent = nullptr);
    ~ServerListSyncWidget() override;
    void reload();

protected Q_SLOTS:
    void slotTypeChanged(QListWidgetItem* item, QListWidgetItem*);

private:
    void build();
    QString buildAddedChange(ServerListSyncChangeItem*);
    QString buildMatchChange(ServerListSyncChangeItem*);
    QString buildSetSysChange(ServerListSyncChangeItem* t);
    QString buildUnsetSysChange(ServerListSyncChangeItem* t);
    QString buildTable(QString name, QString host, QString port) const;
};

#endif /* ecflow_viewer_ServerListSyncWidget_HPP */
