suite event_suite
   # Events added to suite/family cannot be signaled by child commands
   # However alter/force should allow events to set/cleared on Family/suite
   event 0 
   event name
   event 1 eventName
   event 2 eventNamea
   task t
endsuite
