suite suite_ml_empty_x;endsuite
suite suite_ml_obs; endsuite

suite suite_ml_fred; task a; task b; endsuite # a comment suite fred; task a; task b; endsuite

suite suite_ml_s1; task t1; task t2;task t3;family f1;task f1_t1;task f1_t2;endfamily;task t4
endsuite

suite suite_ml_s2;task t1;task t2;task t3;family f1;task f1_t1;task f1_t2;endfamily; task t4 # comment,
endsuite
 