#
# Used to test inclusion with angle brackets  <filename>
# First look in *ALL* the paths specfied in %ECF_INCLUDE%
# If file not there then look at ECF_HOME/filename
#
# THIS ALLOWS USE TO PUT SPECIFIC FILES IN ECF_INCLUDE
# AND COMMON FILE IN ECF_HOME
