from .ecflow import *

# Name        :
# Author      : Avi
# Revision    : $Revision: #10 $ 
#
# Copyright 2009- ECMWF.
# This software is licensed under the terms of the Apache Licence version 2.0 
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0. 
# In applying this licence, ECMWF does not waive the privileges and immunities 
# granted to it by virtue of its status as an intergovernmental organisation 
# nor does it submit to any jurisdiction.  

"""
The ecFlow python module
"""

__version__ = '5.13.8'

# http://stackoverflow.com/questions/13040646/how-do-i-create-documentation-with-pydoc
