suite complete1
    family f
    	task t1
    	task t2
    	task t3
    	task t4
    	task t5
    	task t6
    	task t7
    	task t8
    	task t9
    	task t10
		task a             
         	event eventName 
    	task b  
    	    trigger 1 == 0                  # impossible trigger, keeps task queued           
   			complete t1  == complete        # a comment at end of trigger
   			complete -a t2  == complete     # a comment at end of trigger
   			complete -a t3  == complete     # a comment at end of trigger
   			complete -a t4  == complete     # a comment at end of trigger
   			complete -a t5  == complete     # a comment at end of trigger
   			complete -o t6  == complete      # a comment at end of trigger
   			complete -o t7  == complete      # a comment at end of trigger
   			complete -o t8  == complete      # a comment at end of trigger
   			complete -o t9  == complete      # a comment at end of trigger
   			complete -o t10  == complete      # a comment at end of trigger
    endfamily
endsuite