suite suite     
  clock real 5.1.2017
  family family
    task t1
      date 05.01.2017      # simulator will run for a month, unless endclock used
      time 14:00
      verify complete:1
  endfamily
endsuite