extern /s4
extern /s4/f1
extern /s4/f2

suite autorestore
 family f1
    autorestore ./f2 ./f3 ./f4 # autorestore
    task t1
 endfamily
 family f2
     autorestore /autorestore/f1 /autorestore/f3 /autorestore/f4 # autorestore
     task t1
 endfamily
 family f3
    autorestore /autorestore/f1  # autorestore
    task t1
 endfamily
 family f4
    autorestore /autorestore 
    task t1
 endfamily
 family f5
    autorestore /s4 /s4/f1 /s4/f2 # this is an extern so allow
    task t1
 endfamily
endsuite