# test node referencing in triggers
#      o/ <node> == complete    i.e   a == complete   sibling    
#      o/ ./<node> == complete  i.e  ./a == complete  sibling
#      o/ ../<node> == complete i.e  ../a == complete means trigger nodes parent
suite trigger_s0
  edit FIRE_YMD '19000101'
  task t1
endsuite
suite trigger_s1
  family f1
     trigger f2/f3/f4/a == complete or ../trigger_s0:FIRE_YMD == 19000101  # allow relative suite access
     family f2
        family f3
           family f4
              task a
                 event 1 theEventName
                 meter myMeter 0 100
                 trigger ./b == complete or c == complete or ../f4 == complete or ../../f3 == complete or ../../../f2 == aborted 
               task b
                  trigger ../../../../../trigger_s0 == complete or  ../../../../../trigger_s0/t1<flag>late
               task c
                  trigger ../../../../../trigger_s0:FIRE_YMD == 10 
            endfamily   
         endfamily
      endfamily
   endfamily
   family f0
      task c
         # The initial .. means parents parent( i.e /trigger_s1 in this case)
         # i.e task 'c' parents parent is suite 'trigger_s1'
         trigger  ../f1/f2/f3/f4/a == complete
         complete ../f1/f2/f3/f4/a:myMeter >= 10
      family f1
         family f2
            family f3
               family f4
                  # ..             -> f3 level
                  # ../..          -> f2 level
                  # ../../..       -> f1 level
                  # ../../../..    -> f0 level
                  # ../../../../.. -> suite level
                  trigger ../../../../f1/f2/f3/f4/a == complete or ../../../../../trigger_s1/f1/f2/f3/f4/a == complete or ../../../../../trigger_s0 == complete
            endfamily
         endfamily
      endfamily
   endfamily
endsuite

