suite event_set
   family f1
      task t
         event 0 set
         event name set
         event 1 eventName set
         event 2 eventNamea set
     endfamily
endsuite

suite event_set_s2;family f ; task t ; event 0 set ;event name set;event 1 eventName set
         event 2 eventNamea set; endfamily
endsuite
