/*
 * Copyright 2023- ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef ECFLOW_CORE_FILESYSTEM_HPP
#define ECFLOW_CORE_FILESYSTEM_HPP

#include <cassert>

#include <boost/filesystem.hpp>

namespace fs = boost::filesystem;

#endif
