suite csim_repeat_day2
  clock real 04.08.2019        # Sunday
  endclock   13.08.2019        # added endclock to avoid simulation for a year due to repeat
  family f
    repeat integer rep 0 1
    family f1
       time 08:00 20:00 01:00  # no guard so has no effect
       task t1
            day monday         # run at midnight, 2 times  
            verify complete:2     
    endfamily
  endfamily
endsuite
