extern /limits:c1a
extern /limits:linux_cluster
extern /limits:pdb
extern /limits:q2diss

suite limit
    limit sg1  10
    limit mars 10
endsuite

suite limit_with_node_inlimit
    limit fred 1000
    inlimit -n /limits:c1a
    
    family limits
        limit hpcd 20
    endfamily
    
    family X
        inlimit -n fred 3
        task t1
    endfamily
    
    family anon    
        inlimit -n fred      
        task t1
            inlimit -n /limit:sg1 3          # another suite limit
            inlimit -n /limit_with_node_inlimit/limits:hpcd      # This suite limit
            inlimit -n /limits:c1a           # extern suite limit
        task t2
            inlimit -n /limit_with_node_inlimit/limits:hpcd      # This suite limit
        task t3
            inlimit -n /limits:c1a           # extern suite limit
        task t4
            inlimit -n /limits:c1a  10       # extern suite limit
    endfamily
endsuite