extern /limits:c1a
extern /limits:linux_cluster
extern /limits:pdb
extern /limits:q2diss

suite obs
    limit sg1
    inlimit /limits:c1a
    family limits
    	limit hpcd 20
    endfamily
    
    family anon    
        inlimit -n -s sg1 # cant have -n(limit node only) and -s(limit submission) together       
    	task t1
   		task t2
   		task t3
     endfamily
endsuite
