extern /limits:c1a
extern /limits:linux_cluster
extern /limits:pdb
extern /limits:q2diss
suite suiteName
	limit sg1  10
	limit mars 10
endsuite

suite limit_s1
	limit mars 10
	limit sg1  10
endsuite

suite limit_with_externs
    inlimit /limits:c1a
    family limits
    	limit hpcd 20
    endfamily
    
    family anon    
   		inlimit limits:hpcd       
    	task t1
    		inlimit /suiteName:sg1        # another suite limit
    		inlimit /suiteName:mars       # another suite limit
    		inlimit /limit_s1:mars              # another suite limit
    		inlimit /limit_s1:sg1               # another suite limit
   		task t2
   			inlimit /limit_with_externs/limits:hpcd      # This suite limit
  		task t3
   			inlimit /limits:c1a           # extern suite limit
    endfamily
endsuite
