suite cron5              # use real clock otherwise the date wont change
   clock real 1.10.2009  # need to define a date otherwise log file content varies according to the time 
   endclock   1.11.2009  # When to finish. 
   family cronFamily
		task t
  			 cron -w 0  -m 10   10:00 11:00 01:00   # run every sunday in October at 10 and 11am
  			 verify complete:8		                # 4 sundays in october hence expect 8 task completions
   endfamily
endsuite
