extern /a/b/c/d         
extern /a/b/c/d/e:event        
extern /a/b/c/d/e:meter        

suite suiteName
  family familyName
  		task a
   			trigger /a/b/c/d/e/f == complete  		   # non existing path, not even extern
   endfamily
endsuite

