#
# test auto add extern for triggers and inlimits
#
extern /a/b/c/d         
extern /a/b/c/d/e:event        
extern /a/b/c/d/e:meter       
extern /made/up/redundant/extren     # REDUNDANT extern, should get removed after auto_add_extren(true)
extern /made/up/redundant/extren     # Duplicates should be ignored.

extern /limits:c1a
extern /limits:c1a                   # Duplicates should be ignored
extern fred
extern limits:hpcd
extern /suiteName:sg1
extern /obs/limits:hpcd

extern /extern/f1/parent_variable:var1
extern /extern/f1/parent_variable:var2

# Expect 10 unique externs
# expect:  extern /made/up/redundant/extren  to be removed
# Also because we have defined externs , expect no warnings:

suite extern
   family f1
      task p
            trigger /a/b/c/d == complete          # extern path
      task q
            trigger /a/b/c/d/e:event == set       # extern event path
      task r
            trigger /a/b/c/d/e:meter le 30        # extern meter path   
      task parent_variable
            trigger :var1 == 1 and :var2 == 2
    endfamily
    
    
    inlimit /limits:c1a
    inlimit fred
    family anon    
         inlimit limits:hpcd       
      task t1
         inlimit /suiteName:sg1            
      task t2
            inlimit /obs/limits:hpcd       
      task t3
            inlimit /limits:c1a            
    endfamily
endsuite
