# test node referencing in triggers
#      o/ <node> == complete    i.e   a == complete   means sibling
#      o/ ./<node> == complete  i.e  ./a == complete  means sibling
#      o/ ../<node> == complete i.e  ../a == complete means trigger nodes parent
suite trigger_relative_s
endsuite
suite trigger_relative
  family f1
     trigger f2/f3/f4/a == complete
     family f2
        trigger ../f1 == complete
        family f3
           trigger f4/a == complete
           family f4
              task a
           endfamily   
        endfamily
     endfamily
  endfamily
endsuite
