#
# This defines extern which are defined in second.def
#
extern /a/b/c/d         
extern /a/b/c/d/e:event        
extern /a/b/c/d/e:meter       

extern /a/b/c/d/e/f/g/h/j     # this does not exist in second.def

suite suiteName
   family familyName
   		task p
   			trigger /a/b/c/d == complete  		                                   # extern path
  		task q
   			trigger /a/b/c/d/e:event == set  		                               # extern event path
 		task r
   			trigger /a/b/c/d/e:meter le 30		                                   # extern meter path	
    endfamily
endsuite

