#5.4.0
suite metabuilder
  edit ADDRESS_SANITIZER 'False'
  edit ADD_ACTION_TRIGGERS 'True'
  edit AEC_VERSION '1.0.4'
  edit ALLOW_INSTALL_TO_OVERWRITE 'True'
  edit ASAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
  edit ATLAS_GIT_TAG 'develop'
  edit ATLAS_VERSION '0.20.2'
  edit BANNER_TASK_END 'banner %TASK%-end || true'
  edit BANNER_TASK_START 'banner %TASK%-start || true'
  edit BASEDIR '%WSDIR%/%PACKAGE%'
  edit BASEPREFIX '$PERM/install/%SITE%'
  edit BUFRDC_EMOSLIB_VERSION '4.6.3'
  edit BUFRDC_TABLES_GIT_TAG 'develop'
  edit BUFRDC_TABLES_VERSION '000415'
  edit BUILDIR '%BASEDIR%/builds'
  edit BUILD_ALWAYS_PARALLEL 'False'
  edit CFI_SANITIZER 'False'
  edit CFL_BASE_URL 'https://software-test.ecmwf.int/wiki'
  edit CFL_SPACE_KEY '<TODO>'
  edit CFL_TARBALL_COMMENT ''
  edit CFL_TITLE 'Releases'
  edit CHECK_SYNC_RELEASE 'False'
  edit CLANG_TIDY_ABORT_ON_ERROR 'False'
  edit CMAKE '%ECBUILD%'
  edit CMAKELIST_EXTRACT_PACKAGE_VERSION '$(awk '/^project/ && /%PACKAGE%/ && /VERSION/ {for (I=1;I<=NF;I++) if ($I == "VERSION") {print $(I+1)};}' CMakeLists.txt)'
  edit CMAKE_BUILD_TYPE 'RelWithDebInfo'
  edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE%'
  edit CMAKE_VERSION '3.13.4'
  edit CODES_UI_GIT_TAG 'develop'
  edit CODES_UI_VERSION '1.3.1'
  edit CPACK 'make package'
  edit CPUS '$(lscpu -p | grep -v '#' | wc -l)'
  edit CRAY_FFTW_VERSION '3.3.8.1'
  edit CRAY_NETCDF_VERSION '4.6.1.3'
  edit CRON_BASEPREFIX '%BASEDIR%/install'
  edit CRON_EMAIL ''
  edit CRON_ENABLED 'False'
  edit CRON_GIT_TAG 'develop'
  edit CUSTOM_COMMANDS ''
  edit DATAFLOW_SANITIZER 'False'
  edit DEFAULT_GNU_COMPILER 'gnu.73'
  edit ECBUILD '%BASEPREFIX%/ecbuild/develop/bin/ecbuild'
  edit ECBUILD_CLONE 'False'
  edit ECBUILD_CLONE_COMMIT 'master'
  edit ECBUILD_GIT_TAG 'develop'
  edit ECBUILD_REPO '%GIT_REPO_PREFIX%/ecsdk/ecbuild.git'
  edit ECBUILD_VERSION '3.3.2'
  edit ECCODES_EXTENSIONS_GIT_TAG 'develop'
  edit ECCODES_EXTENSIONS_VERSION '2.17.0'
  edit ECCODES_GIT_TAG 'develop'
  edit ECCODES_PYTHON_VERSION '0.9.6'
  edit ECCODES_VERSION '2.17.1'
  edit ECFLOW5_LATEST_VERSION '5.4.0'
  edit ECFLOW_VERSION '5.4.1'
  edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/metabuilder'
  edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/metabuilder'
  edit ECF_JOB_CMD '%ECF_JOB% > %ECF_JOBOUT% 2>&1'
  edit ECF_KILL_CMD 'kill -15 %ECF_RID%'
  edit ECF_STATUS_CMD 'ps --sid %ECF_RID% -f'
  edit ECF_TRIES '1'
  edit ECKIT_GIT_TAG 'develop'
  edit ECKIT_VERSION '1.10.1'
  edit ECQUIS_VERSION '1.0.0'
  edit EMOS_MOD_VERSION '462'
  edit EMOS_PA_VERSION '000462'
  edit EXPORTS '"export ECACCOUNT=ecodmdma"'
  edit EXPORT_VARIABLES 'False'
  edit FCKIT_GIT_TAG '0.6.3'
  edit FCKIT_VERSION '0.6.3'
  edit FDB_VINTAGE_GIT_TAG '4.10.0'
  edit FDB_VINTAGE_VERSION '4.10.0'
  edit FFTW_VERSION '3.3.8'
  edit FFTW_VERSION_CRAY '3.3.4.11'
  edit GIT 'git'
  edit GITHUB_PACKAGE_NAME '%PACKAGE_NAME%'
  edit GITHUB_REPO 'git@github.com:ecmwf'
  edit GIT_CLONE 'True'
  edit GIT_CLONE_OPTIONS '--depth=1 --branch %GIT_TAG%'
  edit GIT_METABUILDER_TAG 'ci'
  edit GIT_REPO_PREFIX 'ssh://git@git.ecmwf.int:7999'
  edit GIT_TAG 'master'
  edit GIT_VERSION '2.26.0'
  edit GRIB_API_VERSION '1.28.0'
  edit INSTALLABLE_TO_USR_LOCAL_APPS 'True'
  edit IWYU_VERSION '0.11'
  edit KSH_EXTRACT_PACKAGE_VERSION '`grep _VERSION_STR VERSION.cmake | grep -o '[0-9]*\.[0-9]*\.[0-9]*'`'
  edit LAPACK_VERSION '3.5.0'
  edit LIBEMOS_GIT_TAG 'develop'
  edit LIBEMOS_VERSION '4.6.2'
  edit LINTERS_REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}'
  edit LSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/lsan.txt'
  edit MAGICS_GIT_TAG 'develop'
  edit MAGICS_PYTHON_VERSION 'master'
  edit MAGICS_VERSION '4.3.3'
  edit MAKE 'make'
  edit MARS_CLIENT_GIT_TAG '6.27.2.2'
  edit MARS_CLIENT_PACKAGE_GIT_TAG '6.27.6'
  edit MARS_CLIENT_PACKAGE_VERSION '6.27.6'
  edit MARS_CLIENT_VERSION '6.27.2.2'
  edit MEMORY_SANITIZER 'False'
  edit METABUILDER_HOME '/var/tmp/ma0/workspace/metabuilder'
  edit METAB_INSTALL_REMOVE_WRITE_PERMISSIONS 'True'
  edit METAB_PYTHON_VERSION 'python3'
  edit METKIT_GIT_TAG 'develop'
  edit METKIT_VERSION '1.4.6'
  edit METVIEW_GIT_TAG 'develop'
  edit METVIEW_PYTHON_VERSION '1.4.2'
  edit METVIEW_VERSION '5.8.3'
  edit MIR_BUNDLE_VERSION '0.9.0.0'
  edit MIR_GIT_TAG 'develop'
  edit MIR_VERSION '1.5.0'
  edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
  edit MODULE_ACTIONS_FIRST ''
  edit MODULE_ACTIONS_HDR 'module_actions.h'
  edit MODULE_CHECKING_HDR 'module_checking.h'
  edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow"'
  edit MODULE_LOAD_SUPPORTED 'True'
  edit MODULE_NAME 'metabuilder'
  edit MODULE_VERSION '%GIT_TAG%'
  edit NETCDF4_VERSION '4.7.4'
  edit ODB_API_GIT_TAG 'develop'
  edit ODB_API_VERSION '0.19.4'
  edit ODB_GIT_TAG 'develop'
  edit ODB_TOOLS_GIT_TAG 'develop'
  edit ODB_TOOLS_VERSION '0.18.3'
  edit ODB_VERSION '1.0.11'
  edit ODC_GIT_TAG 'develop'
  edit ODC_VERSION '1.0.3'
  edit OPENJPEG_VERSION '2.3.1'
  edit PACKAGE 'metabuilder'
  edit PACKAGE_NAME 'metabuilder'
  edit PACKAGE_OWNERS 'ma0@ecmwf.int'
  edit PACKAGE_VERSION '%GIT_TAG%'
  edit PGEN_VERSION '1.7.1.0'
  edit PKG_BOOST_VERSION '1.71.0'
  edit PREFIX_APPEND ''
  edit PROJ4_VERSION '5.2.0'
  edit PROJ_VERSION '6.3.1'
  edit PYTHON2_VERSION '2.7.15-01'
  edit PYTHON3_VERSION '3.6.10-01'
  edit PYTHON3_VERSION_MAJOR_MINOR '3.6'
  edit PYTHON_GITREPO 'This-needs-tobe-overriden'
  edit PYTHON_PACKAGE 'This-needs-tobe-overriden'
  edit QT_VERSION '5.12.3'
  edit QUEUE 'build'
  edit REMOTE_HOST 'polonius'
  edit REUSE_BUILD_DIR 'False'
  edit SAFE_STACK_SANITIZER 'False'
  edit SANITIZER 'False'
  edit SANITIZER_REFERENCE_DIRECTORY '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
  edit SRCDIR '%BASEDIR%/%PACKAGE%'
  edit STATIC_ANALYSER_ABORT_ON_ERROR 'False'
  edit SWIG_VERSION '3.0.8'
  edit SYNC_RELEASE '2020.04.0'
  edit SYNC_RELEASE_CI_STAGE 'develop'
  edit TARBALL_DIR '/perm/ma/ma0/tmp_tarballs/ci'
  edit TAR_PACKAGE '%PACKAGE%'
  edit TEST 'ctest --output-on-failure -j${CPUS} '
  edit THREAD_SANITIZER 'False'
  edit TRANSI_GIT_TAG 'develop'
  edit TRANSI_VERSION '0.7.1'
  edit TSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
  edit UBSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/ubsan.txt'
  edit UB_SANITIZER 'False'
  edit USER 'ma0'
  edit USE_WORKDIR 'False'
  edit VALGRIND_VERSION '3.15.0'
  edit WORKDIR '%BASEDIR%/workdir/%ECF_NAME%'
  edit WSDIR '/var/tmp/tmpdir/deploy/jtmp.20739'
  task regenerate
    defstatus complete
    edit ECF_PASS 'FREE'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
    label date ""
    label metabuilder_branch "ci"
  task clear_log
    cron -w 6 22:30
  task remove_build_dirs
    defstatus complete
  task add_new_package
    defstatus complete
    edit NEW_PACKAGE_NAME ''
    label msg "Please set new package using variable NEW_PACKAGE_NAME"
    label new_package ""
  task generate_report
    defstatus complete
    edit ONLY_SUITES 'ecbuild eccodes grib_api libemos bufrdc_tables eckit metkit odb odb_api Magics metview'
    edit REPORT_EMAIL ''
    edit SKIP_PLATFORMS 'cron'
    edit SKIP_SUITES 'metabuilder'
    edit WIKI_PARENT_PAGE 'Synchronised Release 2020.04.0'
    edit WIKI_PARENT_PAGE_ID ''
    edit WIKI_SPACE 'DS'
    edit WIKI_TITLE 'Metabuilder deployment report 2020.04.0'
  task remove_lock_file
    defstatus complete
endsuite
suite ci
  edit ADDRESS_SANITIZER 'False'
  edit ADD_ACTION_TRIGGERS 'True'
  edit AEC_VERSION '1.0.4'
  edit ALLOW_INSTALL_TO_OVERWRITE 'True'
  edit ASAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
  edit ATLAS_GIT_TAG 'develop'
  edit ATLAS_VERSION '0.20.2'
  edit BANNER_TASK_END 'banner %TASK%-end || true'
  edit BANNER_TASK_START 'banner %TASK%-start || true'
  edit BASEDIR '%WSDIR%/%PACKAGE%'
  edit BASEPREFIX '$PERM/install/%SITE%'
  edit BUFRDC_EMOSLIB_VERSION '4.6.3'
  edit BUFRDC_TABLES_GIT_TAG 'develop'
  edit BUFRDC_TABLES_VERSION '000415'
  edit BUILDIR '%BASEDIR%/builds'
  edit BUILD_ALWAYS_PARALLEL 'False'
  edit CFI_SANITIZER 'False'
  edit CFL_BASE_URL 'https://software-test.ecmwf.int/wiki'
  edit CFL_SPACE_KEY '<TODO>'
  edit CFL_TARBALL_COMMENT ''
  edit CFL_TITLE 'Releases'
  edit CHECK_SYNC_RELEASE 'False'
  edit CI_EXTERNAL_PACKAGES 'pgen scalability mars_client ifs'
  edit CI_PACKAGES 'metabuilder ci ecbuild eccodes eccodes_python eccodes_extensions bufrdc_tables bufrdc_emoslib libemos eckit metkit transi atlas mir odb odb_api odc odb_tools Magics magics_python metview metview_python codes_ui ecflow'
  edit CI_PACKAGE_OWNERS '{'ecbuild': ['matq@ecmwf.int', 'cgm@ecmwf.int', 'ma0@ecmwf.int', 'olivier.iffrig@ecmwf.int'], 'eccodes': ['masn@ecmwf.int', 'cgm@ecmwf.int', 'ma0@ecmwf.int'], 'eccodes_python': ['masn@ecmwf.int', 'cgm@ecmwf.int', 'ma0@ecmwf.int'], 'eccodes_extensions': ['masn@ecmwf.int', 'ma0@ecmwf.int'], 'bufrdc_tables': ['mapm@ecmwf.int', 'ma0@ecmwf.int'], 'bufrdc_emoslib': ['mapm@ecmwf.int', 'ma0@ecmwf.int'], 'libemos': ['mapm@ecmwf.int', 'ma0@ecmwf.int'], 'eckit': ['matq@ecmwf.int', 'mass@ecmwf.int', 'cgm@ecmwf.int', 'ma0@ecmwf.int', 'olivier.iffrig@ecmwf.int'], 'metkit': ['matq@ecmwf.int', 'cgm@ecmwf.int', 'ma0@ecmwf.int', 'mass@ecmwf.int', 'olivier.iffrig@ecmwf.int'], 'transi': ['nawd@ecmwf.int', 'cgm@ecmwf.int', 'ma0@ecmwf.int'], 'atlas': ['nawd@ecmwf.int', 'cgm@ecmwf.int', 'ma0@ecmwf.int'], 'mir': ['mapm@ecmwf.int', 'matq@ecmwf.int', 'cgm@ecmwf.int', 'ma0@ecmwf.int', 'olivier.iffrig@ecmwf.int'], 'odb': ['datk@ecmwf.int', 'cgm@ecmwf.int', 'ma0@ecmwf.int'], 'odb_api': ['mass@ecmwf.int', 'cgm@ecmwf.int', 'ma0@ecmwf.int'], 'odc': ['mass@ecmwf.int', 'cgm@ecmwf.int', 'ma0@ecmwf.int'], 'odb_tools': ['mass@ecmwf.int', 'cgm@ecmwf.int', 'ma0@ecmwf.int'], 'Magics': ['cgs@ecmwf.int', 'cgm@ecmwf.int', 'ma0@ecmwf.int'], 'magics_python': ['cgs@ecmwf.int', 'cgm@ecmwf.int', 'ma0@ecmwf.int'], 'metview': ['cgi@ecmwf.int', 'cgm@ecmwf.int'], 'metview_python': ['cgi@ecmwf.int', 'cgm@ecmwf.int', 'ma0@ecmwf.int'], 'codes_ui': ['cgr@ecmwf.int', 'cgm@ecmwf.int', 'ma0@ecmwf.int'], 'ecflow': ['ma0@ecmwf.int']}'
  edit CI_TRIGGERS '{'eccodes': ['ecbuild'], 'eccodes_python': ['eccodes'], 'eccodes_extensions': ['ecbuild'], 'bufrdc_tables': ['ecbuild'], 'bufrdc_emoslib': ['ecbuild', 'bufrdc_tables'], 'libemos': ['ecbuild', 'eccodes', 'bufrdc_tables'], 'eckit': ['ecbuild'], 'metkit': ['ecbuild', 'eckit'], 'transi': ['ecbuild'], 'atlas': ['ecbuild', 'eckit', 'transi'], 'mir': ['ecbuild', 'eccodes', 'eckit', 'transi', 'atlas'], 'odb': ['ecbuild'], 'odb_api': ['ecbuild', 'eckit', 'metkit', 'odb'], 'odc': ['ecbuild', 'eckit'], 'odb_tools': ['ecbuild', 'eckit', 'metkit', 'odb_api', 'odc'], 'Magics': ['ecbuild', 'eccodes', 'eckit', 'odb_api'], 'magics_python': ['ecbuild', 'Magics'], 'metview': ['ecbuild', 'eccodes', 'eckit', 'odb', 'odb_api', 'Magics'], 'metview_python': ['metview'], 'codes_ui': ['metview'], 'ecflow': ['ecbuild']}'
  edit CLANG_TIDY_ABORT_ON_ERROR 'False'
  edit CMAKE '%ECBUILD%'
  edit CMAKELIST_EXTRACT_PACKAGE_VERSION '$(awk '/^project/ && /%PACKAGE%/ && /VERSION/ {for (I=1;I<=NF;I++) if ($I == "VERSION") {print $(I+1)};}' CMakeLists.txt)'
  edit CMAKE_BUILD_TYPE 'RelWithDebInfo'
  edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE%'
  edit CMAKE_VERSION '3.13.4'
  edit CODES_UI_GIT_TAG 'develop'
  edit CODES_UI_VERSION '1.3.1'
  edit CPACK 'make package'
  edit CPUS '$(lscpu -p | grep -v '#' | wc -l)'
  edit CRAY_FFTW_VERSION '3.3.8.1'
  edit CRAY_NETCDF_VERSION '4.6.1.3'
  edit CRON_BASEPREFIX '%BASEDIR%/install'
  edit CRON_EMAIL ''
  edit CRON_ENABLED 'False'
  edit CRON_GIT_TAG 'develop'
  edit CUSTOM_COMMANDS ''
  edit DATAFLOW_SANITIZER 'False'
  edit DEFAULT_GNU_COMPILER 'gnu.73'
  edit ECBUILD '%BASEPREFIX%/ecbuild/develop/bin/ecbuild'
  edit ECBUILD_CLONE 'False'
  edit ECBUILD_CLONE_COMMIT 'master'
  edit ECBUILD_GIT_TAG 'develop'
  edit ECBUILD_REPO '%GIT_REPO_PREFIX%/ecsdk/ecbuild.git'
  edit ECBUILD_VERSION '3.3.2'
  edit ECCODES_EXTENSIONS_GIT_TAG 'develop'
  edit ECCODES_EXTENSIONS_VERSION '2.17.0'
  edit ECCODES_GIT_TAG 'develop'
  edit ECCODES_PYTHON_VERSION '0.9.6'
  edit ECCODES_VERSION '2.17.1'
  edit ECFLOW5_LATEST_VERSION '5.4.0'
  edit ECFLOW_VERSION '5.4.1'
  edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/ci'
  edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/ci'
  edit ECF_JOB_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_JOB% %ECF_JOBOUT%'
  edit ECF_KILL_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% kill'
  edit ECF_STATUS_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% stat'
  edit ECF_TRIES '1'
  edit ECKIT_GIT_TAG 'develop'
  edit ECKIT_VERSION '1.10.1'
  edit ECQUIS_VERSION '1.0.0'
  edit EMOS_MOD_VERSION '462'
  edit EMOS_PA_VERSION '000462'
  edit EXPORTS '"export ECACCOUNT=ecodmdma"'
  edit EXPORT_VARIABLES 'False'
  edit FCKIT_GIT_TAG '0.6.3'
  edit FCKIT_VERSION '0.6.3'
  edit FDB_VINTAGE_GIT_TAG '4.10.0'
  edit FDB_VINTAGE_VERSION '4.10.0'
  edit FFTW_VERSION '3.3.8'
  edit FFTW_VERSION_CRAY '3.3.4.11'
  edit GIT 'git'
  edit GITHUB_PACKAGE_NAME '%PACKAGE_NAME%'
  edit GITHUB_REPO 'git@github.com:ecmwf'
  edit GIT_CLONE 'True'
  edit GIT_CLONE_OPTIONS '--depth=1 --branch %GIT_TAG%'
  edit GIT_METABUILDER_TAG 'ci'
  edit GIT_REPO_PREFIX 'ssh://git@git.ecmwf.int:7999'
  edit GIT_TAG 'ci'
  edit GIT_VERSION '2.26.0'
  edit GRIB_API_VERSION '1.28.0'
  edit INSTALLABLE_TO_USR_LOCAL_APPS 'True'
  edit IWYU_VERSION '0.11'
  edit KSH_EXTRACT_PACKAGE_VERSION '`grep _VERSION_STR VERSION.cmake | grep -o '[0-9]*\.[0-9]*\.[0-9]*'`'
  edit LAPACK_VERSION '3.5.0'
  edit LIBEMOS_GIT_TAG 'develop'
  edit LIBEMOS_VERSION '4.6.2'
  edit LINTERS_REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}'
  edit LSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/lsan.txt'
  edit MAGICS_GIT_TAG 'develop'
  edit MAGICS_PYTHON_VERSION 'master'
  edit MAGICS_VERSION '4.3.3'
  edit MAKE 'make'
  edit MARS_CLIENT_GIT_TAG '6.27.2.2'
  edit MARS_CLIENT_PACKAGE_GIT_TAG '6.27.6'
  edit MARS_CLIENT_PACKAGE_VERSION '6.27.6'
  edit MARS_CLIENT_VERSION '6.27.2.2'
  edit MEMORY_SANITIZER 'False'
  edit METABUILDER_HOME '/var/tmp/ma0/workspace/metabuilder'
  edit METAB_INSTALL_REMOVE_WRITE_PERMISSIONS 'True'
  edit METAB_PYTHON_VERSION 'python3'
  edit METKIT_GIT_TAG 'develop'
  edit METKIT_VERSION '1.4.6'
  edit METVIEW_GIT_TAG 'develop'
  edit METVIEW_PYTHON_VERSION '1.4.2'
  edit METVIEW_VERSION '5.8.3'
  edit MIR_BUNDLE_VERSION '0.9.0.0'
  edit MIR_GIT_TAG 'develop'
  edit MIR_VERSION '1.5.0'
  edit MODULE_ACTIONS '"module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
  edit MODULE_ACTIONS_FIRST ''
  edit MODULE_ACTIONS_HDR 'module_actions.h'
  edit MODULE_CHECKING_HDR 'module_checking.h'
  edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow"'
  edit MODULE_LOAD_SUPPORTED 'True'
  edit MODULE_NAME 'ci'
  edit MODULE_VERSION '%GIT_TAG%'
  edit NETCDF4_VERSION '4.7.4'
  edit ODB_API_GIT_TAG 'develop'
  edit ODB_API_VERSION '0.19.4'
  edit ODB_GIT_TAG 'develop'
  edit ODB_TOOLS_GIT_TAG 'develop'
  edit ODB_TOOLS_VERSION '0.18.3'
  edit ODB_VERSION '1.0.11'
  edit ODC_GIT_TAG 'develop'
  edit ODC_VERSION '1.0.3'
  edit OPENJPEG_VERSION '2.3.1'
  edit PACKAGE 'ci'
  edit PACKAGE_NAME 'ci'
  edit PACKAGE_OWNERS 'ma0@ecmwf.int'
  edit PACKAGE_VERSION '%GIT_TAG%'
  edit PGEN_VERSION '1.7.1.0'
  edit PKG_BOOST_VERSION '1.71.0'
  edit PREFIX_APPEND ''
  edit PROJ4_VERSION '5.2.0'
  edit PROJ_VERSION '6.3.1'
  edit PYTHON2_VERSION '2.7.15-01'
  edit PYTHON3_VERSION '3.6.10-01'
  edit PYTHON3_VERSION_MAJOR_MINOR '3.6'
  edit PYTHON_GITREPO 'This-needs-tobe-overriden'
  edit PYTHON_PACKAGE 'This-needs-tobe-overriden'
  edit QT_VERSION '5.12.3'
  edit QUEUE 'build'
  edit REMOTE_HOST 'polonius'
  edit REUSE_BUILD_DIR 'False'
  edit SAFE_STACK_SANITIZER 'False'
  edit SANITIZER 'False'
  edit SANITIZER_REFERENCE_DIRECTORY '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
  edit SRCDIR '%BASEDIR%/%PACKAGE%'
  edit STATIC_ANALYSER_ABORT_ON_ERROR 'False'
  edit SWIG_VERSION '3.0.8'
  edit SYNC_RELEASE '2020.04.0'
  edit SYNC_RELEASE_CI_STAGE 'develop'
  edit SYNC_RELEASE_SITES ''
  edit TARBALL_DIR '/perm/ma/ma0/tmp_tarballs/ci'
  edit TAR_PACKAGE '%PACKAGE%'
  edit TEST 'ctest --output-on-failure -j${CPUS} '
  edit THREAD_SANITIZER 'False'
  edit TRANSI_GIT_TAG 'develop'
  edit TRANSI_VERSION '0.7.1'
  edit TSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
  edit UBSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/ubsan.txt'
  edit UB_SANITIZER 'False'
  edit USER 'ma0'
  edit USE_WORKDIR 'False'
  edit VALGRIND_VERSION '3.15.0'
  edit WORKDIR '%BASEDIR%/workdir/%ECF_NAME%'
  edit WSDIR '/var/tmp/tmpdir/deploy/jtmp.20739'
  label branch "ci"
  task regenerate
    defstatus complete
    edit ECF_JOB_CMD '%ECF_JOB% > %ECF_JOBOUT% 2>&1'
    edit ECF_KILL_CMD 'kill -15 %ECF_RID%'
    edit ECF_PASS 'FREE'
    edit ECF_STATUS_CMD 'ps --sid %ECF_RID% -f'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
    label date ""
    label metabuilder_branch "ci"
  task cron_regenerate
    edit ECF_JOB_CMD '%ECF_JOB% > %ECF_JOBOUT% 2>&1'
    edit ECF_KILL_CMD 'kill -15 %ECF_RID%'
    edit ECF_STATUS_CMD 'ps --sid %ECF_RID% -f'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
    edit REMOTE_HOST 'polonius'
    edit WORKDIR ''
    cron -w 0,1,2,3,4,5,6 19:00
  task start_ci
    edit CI_SITES ''
    edit ECF_JOB_CMD '%ECF_JOB% > %ECF_JOBOUT% 2>&1'
    edit ECF_KILL_CMD 'kill -15 %ECF_RID%'
    edit ECF_STATUS_CMD 'ps --sid %ECF_RID% -f'
    edit REMOTE_HOST 'polonius'
    edit WORKDIR ''
    cron -w 0,1,2,3,4,5,6 19:30
  task monitor
    edit ECF_JOB_CMD '%ECF_JOB% > %ECF_JOBOUT% 2>&1'
    edit ECF_KILL_CMD 'kill -15 %ECF_RID%'
    edit ECF_STATUS_CMD 'ps --sid %ECF_RID% -f'
    edit REMOTE_HOST 'polonius'
    edit WORKDIR ''
    cron -w 1,2,3,4,5 09:00
  task test_mail
    defstatus complete
    edit CI_BODY 'body'
    edit CI_MAIL 'ma0@ecmwf.int'
    edit CI_SUBJECT 'testing email from metabuilder/ci user deploy'
    edit REMOTE_HOST 'polonius'
    edit WORKDIR ''
  task test_add_limits
    defstatus complete
    edit REMOTE_HOST 'polonius'
    edit WORKDIR ''
  task test_add_triggers
    defstatus complete
    edit REMOTE_HOST 'polonius'
    edit WORKDIR ''
  family cca
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DLAPACK_FOUND=TRUE'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/ci/cca'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/ci/cca:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cca-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '24000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cca'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'cca'
    edit SITE 'cca'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    task clean_install
      cron -w 0,1,2,3,4,5,6 18:00
  endfamily
  family ccb
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DLAPACK_FOUND=TRUE'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/ci/ccb'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/ci/ccb:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'ccb-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '24000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'ccb'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'ccb'
    edit SITE 'ccb'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    task clean_install
      cron -w 0,1,2,3,4,5,6 18:00
  endfamily
  family cct
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DLAPACK_FOUND=TRUE'
    edit CPUS '$EC_total_tasks'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/ci/cct'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/ci/cct:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cct-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '4'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '6000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cct'
    edit QSUB_LP 'QSUB -lP %EC_TOTAL_TASKS%'
    edit QUEUE 'nf'
    edit REMOTE_HOST 'cct'
    edit SITE 'cct'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    task clean_install
      cron -w 0,1,2,3,4,5,6 18:00
  endfamily
  family ecflow_metab_vm
    edit BANNER_TASK_END 'echo %TASK%-end'
    edit BANNER_TASK_START 'echo %TASK%-start'
    edit CI_LIMIT '1'
    edit ECF_JOB_CMD '%ECF_JOB% 1> %ECF_JOBOUT% 2>&1'
    edit ECF_KILL_CMD 'kill -15 %ECF_RID%'
    edit ECF_STATUS_CMD 'ps --sid %ECF_RID% -f'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit M_SITE 'vsms1'
    edit REMOTE_HOST 'polonius'
    edit SITE 'vsms1'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/ecflow-metab/%COMPILER:%'
    task clean_install
      cron -w 0,1,2,3,4,5,6 18:00
  endfamily
  family ecgb
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'ecgb'
    edit QUEUE 'express'
    edit REMOTE_HOST 'ecgb'
    edit SITE 'ecgb'
    edit WSDIR '$SCRATCH/metabuilds/%ECF_HOST%_%ECF_PORT%/ecgb/%COMPILER:%'
    task clean_install
      cron -w 0,1,2,3,4,5,6 18:00
  endfamily
  family gorax_dev_000
    edit BANNER_TASK_END 'echo %TASK%-end'
    edit BANNER_TASK_START 'echo %TASK%-start'
    edit BASEPREFIX '/data/webdev/deploy/install/%SITE%'
    edit CI_LIMIT '2'
    edit ECF_CLIENT_EXE_PATH '/home/atls/deploy/install/ecflow/5.3.0/bin/ecflow_client'
    edit ECF_LOGHOST 'gorax-dev-000'
    edit ECF_LOGPORT '38562'
    edit ECF_OUT '/home/atls/deploy/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALLABLE_TO_USR_LOCAL_APPS 'False'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX ''
    edit LOGDIR '%ECF_OUT%'
    edit MODULE_ACTIONS ''
    edit MODULE_ACTIONS_HDR 'empty.h'
    edit MODULE_CLEANUP ''
    edit M_SITE 'gorax_dev_000'
    edit REMOTE_HOST 'gorax-dev-000'
    edit SITE 'gorax_dev_000'
    edit WSDIR '/data/webdev/%ECF_HOST%_%ECF_PORT%/ecflow-metab/%COMPILER:%'
    task clean_install
      cron -w 0,1,2,3,4,5,6 18:00
  endfamily
  family gorax_dev_001
    edit BANNER_TASK_END 'echo %TASK%-end'
    edit BANNER_TASK_START 'echo %TASK%-start'
    edit BASEPREFIX '/data/webdev/deploy/install/%SITE%'
    edit CI_LIMIT '2'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCPACK_PACKAGING_INSTALL_PREFIX=/opt/ecmwf/%CPACK_PACKAGE_NAME% -DENABLE_PYTHON=OFF'
    edit CPACK_PACKAGE_NAME '%MODULE_NAME%'
    edit CPACK_PACKAGING_INSTALL_PREFIX '/opt/ecmwf/%CPACK_PACKAGE_NAME%'
    edit ECBUILD '%BASEDIR%/ecbuild/bin/ecbuild'
    edit ECBUILD_CLONE 'True'
    edit ECBUILD_CLONE_COMMIT '%ECBUILD_GIT_TAG%'
    edit ECF_CLIENT_EXE_PATH '/home/atls/deploy/install/ecflow/5.3.0/bin/ecflow_client'
    edit ECF_LOGHOST 'gorax-dev-001'
    edit ECF_LOGPORT '38562'
    edit ECF_OUT '/home/atls/deploy/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALLABLE_TO_USR_LOCAL_APPS 'False'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX ''
    edit LOGDIR '%ECF_OUT%'
    edit MODULE_ACTIONS ''
    edit MODULE_ACTIONS_HDR 'empty.h'
    edit MODULE_CLEANUP ''
    edit M_SITE 'gorax_dev_001'
    edit REMOTE_HOST 'gorax-dev-001'
    edit RPM_ARCH 'x86_64'
    edit RPM_FILENAME '%CPACK_PACKAGE_NAME%-${PACKAGE_VERSION}-Linux-%RPM_ARCH%.rpm'
    edit RPM_PASSWORD ''
    edit RPM_REPO_DESTINATION 'https://nexus.ecmwf.int/repository/CentOS-staging/7/8/rpms/%RPM_ARCH%/%RPM_FILENAME%'
    edit RPM_USERNAME ''
    edit SITE 'gorax_dev_001'
    edit WSDIR '/data/webdev/%ECF_HOST%_%ECF_PORT%/ecflow-metab/%COMPILER:%'
    task clean_install
      cron -w 0,1,2,3,4,5,6 18:00
  endfamily
  family leap42
    edit CI_LIMIT '7'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'leap42'
    edit REMOTE_HOST 'leap42-deploy'
    edit SITE 'leap42'
    edit WSDIR '/tmp/metabuilds/%ECF_HOST%_%ECF_PORT%/leap42/%COMPILER:%'
    task clean_install
      cron -w 0,1,2,3,4,5,6 18:00
  endfamily
  family lxc
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'lxc'
    edit QUEUE 'express'
    edit REMOTE_HOST 'lxc'
    edit SITE 'lxc'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxc/%COMPILER:%'
    task clean_install
      cron -w 0,1,2,3,4,5,6 18:00
  endfamily
  family lxg
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'lxg'
    edit QUEUE 'normal'
    edit REMOTE_HOST 'lxg'
    edit SITE 'lxg'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxg/%COMPILER:%'
    task clean_install
      cron -w 0,1,2,3,4,5,6 18:00
  endfamily
  family sappb
    edit ASAN_FILE '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
    edit BASEPREFIX '/autofs/vol/deploy/install/%SITE%'
    edit CI_LIMIT '8'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit M_SITE 'sappb'
    edit REMOTE_HOST 'sappb-deploy'
    edit SANITIZER_REFERENCE_DIRECTORY '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
    edit SITE 'sappb'
    edit TSAN_FILE '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
    edit WSDIR '$HOME/metabuilds/%ECF_HOST%_%ECF_PORT%/sappb/%COMPILER:%'
    task clean_install
      cron -w 0,1,2,3,4,5,6 18:00
  endfamily
endsuite
suite ecbuild
  edit ADDRESS_SANITIZER 'False'
  edit ADD_ACTION_TRIGGERS 'True'
  edit AEC_VERSION '1.0.4'
  edit ALLOW_INSTALL_TO_OVERWRITE 'True'
  edit ASAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
  edit ATLAS_GIT_TAG 'develop'
  edit ATLAS_VERSION '0.20.2'
  edit BANNER_TASK_END 'banner %TASK%-end || true'
  edit BANNER_TASK_START 'banner %TASK%-start || true'
  edit BASEDIR '%WSDIR%/%PACKAGE%'
  edit BASEPREFIX '$PERM/install/%SITE%'
  edit BUFRDC_EMOSLIB_VERSION '4.6.3'
  edit BUFRDC_TABLES_GIT_TAG 'develop'
  edit BUFRDC_TABLES_VERSION '000415'
  edit BUILDIR '%BASEDIR%/builds'
  edit BUILD_ALWAYS_PARALLEL 'False'
  edit CFI_SANITIZER 'False'
  edit CFL_BASE_URL 'https://software-test.ecmwf.int/wiki'
  edit CFL_SPACE_KEY '<TODO>'
  edit CFL_TARBALL_COMMENT ''
  edit CFL_TITLE 'Releases'
  edit CHECK_SYNC_RELEASE 'False'
  edit CLANG_TIDY_ABORT_ON_ERROR 'False'
  edit CMAKE '%ECBUILD%'
  edit CMAKELIST_EXTRACT_PACKAGE_VERSION '$(awk '/^project/ && /%PACKAGE%/ && /VERSION/ {for (I=1;I<=NF;I++) if ($I == "VERSION") {print $(I+1)};}' CMakeLists.txt)'
  edit CMAKE_BUILD_TYPE 'RelWithDebInfo'
  edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DENABLE_INSTALL=ON'
  edit CMAKE_VERSION '3.13.4'
  edit CODES_UI_GIT_TAG 'develop'
  edit CODES_UI_VERSION '1.3.1'
  edit CPACK 'make package'
  edit CPUS '$(lscpu -p | grep -v '#' | wc -l)'
  edit CRAY_FFTW_VERSION '3.3.8.1'
  edit CRAY_NETCDF_VERSION '4.6.1.3'
  edit CRON_BASEPREFIX '%BASEDIR%/install'
  edit CRON_EMAIL ''
  edit CRON_ENABLED 'False'
  edit CRON_GIT_TAG 'develop'
  edit CUSTOM_COMMANDS ''
  edit DATAFLOW_SANITIZER 'False'
  edit DEFAULT_GNU_COMPILER 'gnu.73'
  edit ECBUILD 'cmake'
  edit ECBUILD_CLONE 'False'
  edit ECBUILD_CLONE_COMMIT 'master'
  edit ECBUILD_GIT_TAG 'develop'
  edit ECBUILD_REPO '%GIT_REPO_PREFIX%/ecsdk/ecbuild.git'
  edit ECBUILD_VERSION '3.3.2'
  edit ECCODES_EXTENSIONS_GIT_TAG 'develop'
  edit ECCODES_EXTENSIONS_VERSION '2.17.0'
  edit ECCODES_GIT_TAG 'develop'
  edit ECCODES_PYTHON_VERSION '0.9.6'
  edit ECCODES_VERSION '2.17.1'
  edit ECFLOW5_LATEST_VERSION '5.4.0'
  edit ECFLOW_VERSION '5.4.1'
  edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/ecbuild'
  edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/ecbuild'
  edit ECF_JOB_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_JOB% %ECF_JOBOUT%'
  edit ECF_KILL_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% kill'
  edit ECF_STATUS_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% stat'
  edit ECF_TRIES '1'
  edit ECKIT_GIT_TAG 'develop'
  edit ECKIT_VERSION '1.10.1'
  edit ECQUIS_VERSION '1.0.0'
  edit EMOS_MOD_VERSION '462'
  edit EMOS_PA_VERSION '000462'
  edit EXPORTS '"export ECACCOUNT=ecodmdma"'
  edit EXPORT_VARIABLES 'False'
  edit FCKIT_GIT_TAG '0.6.3'
  edit FCKIT_VERSION '0.6.3'
  edit FDB_VINTAGE_GIT_TAG '4.10.0'
  edit FDB_VINTAGE_VERSION '4.10.0'
  edit FFTW_VERSION '3.3.8'
  edit FFTW_VERSION_CRAY '3.3.4.11'
  edit GIT 'git'
  edit GITHUB_PACKAGE_NAME '%PACKAGE_NAME%'
  edit GITHUB_REPO 'git@github.com:ecmwf'
  edit GITREPO '%GIT_REPO_PREFIX%/ecsdk/ecbuild.git'
  edit GIT_CLONE 'True'
  edit GIT_CLONE_OPTIONS '--depth=1 --branch %GIT_TAG%'
  edit GIT_METABUILDER_TAG 'ci'
  edit GIT_REPO_PREFIX 'ssh://git@git.ecmwf.int:7999'
  edit GIT_TAG 'develop'
  edit GIT_VERSION '2.26.0'
  edit GRIB_API_VERSION '1.28.0'
  edit INSTALLABLE_TO_USR_LOCAL_APPS 'True'
  edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop'
  edit IWYU_VERSION '0.11'
  edit KSH_EXTRACT_PACKAGE_VERSION '`grep _VERSION_STR VERSION.cmake | grep -o '[0-9]*\.[0-9]*\.[0-9]*'`'
  edit LAPACK_VERSION '3.5.0'
  edit LIBEMOS_GIT_TAG 'develop'
  edit LIBEMOS_VERSION '4.6.2'
  edit LINTERS_REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}'
  edit LSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/lsan.txt'
  edit MAGICS_GIT_TAG 'develop'
  edit MAGICS_PYTHON_VERSION 'master'
  edit MAGICS_VERSION '4.3.3'
  edit MAKE 'make'
  edit MARS_CLIENT_GIT_TAG '6.27.2.2'
  edit MARS_CLIENT_PACKAGE_GIT_TAG '6.27.6'
  edit MARS_CLIENT_PACKAGE_VERSION '6.27.6'
  edit MARS_CLIENT_VERSION '6.27.2.2'
  edit MEMORY_SANITIZER 'False'
  edit METABUILDER_HOME '/var/tmp/ma0/workspace/metabuilder'
  edit METAB_INSTALL_REMOVE_WRITE_PERMISSIONS 'True'
  edit METAB_PYTHON_VERSION 'python3'
  edit METKIT_GIT_TAG 'develop'
  edit METKIT_VERSION '1.4.6'
  edit METVIEW_GIT_TAG 'develop'
  edit METVIEW_PYTHON_VERSION '1.4.2'
  edit METVIEW_VERSION '5.8.3'
  edit MIR_BUNDLE_VERSION '0.9.0.0'
  edit MIR_GIT_TAG 'develop'
  edit MIR_VERSION '1.5.0'
  edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
  edit MODULE_ACTIONS_FIRST ''
  edit MODULE_ACTIONS_HDR 'module_actions.h'
  edit MODULE_CHECKING_HDR 'module_checking.h'
  edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow"'
  edit MODULE_LOAD_SUPPORTED 'True'
  edit MODULE_NAME 'ecbuild'
  edit MODULE_VERSION '%GIT_TAG%'
  edit NETCDF4_VERSION '4.7.4'
  edit ODB_API_GIT_TAG 'develop'
  edit ODB_API_VERSION '0.19.4'
  edit ODB_GIT_TAG 'develop'
  edit ODB_TOOLS_GIT_TAG 'develop'
  edit ODB_TOOLS_VERSION '0.18.3'
  edit ODB_VERSION '1.0.11'
  edit ODC_GIT_TAG 'develop'
  edit ODC_VERSION '1.0.3'
  edit OPENJPEG_VERSION '2.3.1'
  edit PACKAGE 'ecbuild'
  edit PACKAGE_NAME 'ecbuild'
  edit PACKAGE_OWNERS 'matq@ecmwf.int cgm@ecmwf.int ma0@ecmwf.int olivier.iffrig@ecmwf.int'
  edit PACKAGE_VERSION '%GIT_TAG%'
  edit PGEN_VERSION '1.7.1.0'
  edit PKG_BOOST_VERSION '1.71.0'
  edit PREFIX_APPEND ''
  edit PROJ4_VERSION '5.2.0'
  edit PROJ_VERSION '6.3.1'
  edit PYTHON2_VERSION '2.7.15-01'
  edit PYTHON3_VERSION '3.6.10-01'
  edit PYTHON3_VERSION_MAJOR_MINOR '3.6'
  edit PYTHON_GITREPO 'This-needs-tobe-overriden'
  edit PYTHON_PACKAGE 'This-needs-tobe-overriden'
  edit QT_VERSION '5.12.3'
  edit QUEUE 'build'
  edit REMOTE_HOST 'polonius'
  edit REUSE_BUILD_DIR 'False'
  edit SAFE_STACK_SANITIZER 'False'
  edit SANITIZER 'False'
  edit SANITIZER_REFERENCE_DIRECTORY '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
  edit SRCDIR '%BASEDIR%/%PACKAGE%'
  edit STATIC_ANALYSER_ABORT_ON_ERROR 'False'
  edit SWIG_VERSION '3.0.8'
  edit SYNC_RELEASE '2020.04.0'
  edit SYNC_RELEASE_CI_STAGE 'develop'
  edit TARBALL_DIR '/perm/ma/ma0/tmp_tarballs/ci'
  edit TAR_PACKAGE '%PACKAGE%'
  edit TEST 'ctest --output-on-failure -j${CPUS} '
  edit THREAD_SANITIZER 'False'
  edit TRANSI_GIT_TAG 'develop'
  edit TRANSI_VERSION '0.7.1'
  edit TSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
  edit UBSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/ubsan.txt'
  edit UB_SANITIZER 'False'
  edit USER 'ma0'
  edit USE_WORKDIR 'False'
  edit VALGRIND_VERSION '3.15.0'
  edit WORKDIR '%BASEDIR%/workdir/%ECF_NAME%'
  edit WSDIR '/var/tmp/tmpdir/deploy/jtmp.20739'
  label branch "develop"
  task regenerate
    defstatus complete
    edit ECF_JOB_CMD '%ECF_JOB% > %ECF_JOBOUT% 2>&1'
    edit ECF_KILL_CMD 'kill -15 %ECF_RID%'
    edit ECF_PASS 'FREE'
    edit ECF_STATUS_CMD 'ps --sid %ECF_RID% -f'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
    label date ""
    label metabuilder_branch "ci"
  family cca
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DENABLE_INSTALL=ON -DLAPACK_FOUND=TRUE'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/ecbuild/cca'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/ecbuild/cca:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cca-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '24000'
    edit MODULE_ACTIONS '"module swap PrgEnv-cray PrgEnv-gnu" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cca'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'cca'
    edit SITE 'cca'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task build
      trigger config == complete
      edit ECF_TRIES '2'
      label host ""
      label options ""
    task test
      trigger build == complete
      edit ECF_TRIES '2'
      edit TEST_DIR '%BUILDIR%'
    task install
      trigger test == complete
      edit ECF_TRIES '2'
    task install_log
      trigger install == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install_log == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family ccb
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DENABLE_INSTALL=ON -DLAPACK_FOUND=TRUE'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/ecbuild/ccb'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/ecbuild/ccb:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'ccb-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '24000'
    edit MODULE_ACTIONS '"module swap PrgEnv-cray PrgEnv-gnu" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'ccb'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'ccb'
    edit SITE 'ccb'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task build
      trigger config == complete
      edit ECF_TRIES '2'
      label host ""
      label options ""
    task test
      trigger build == complete
      edit ECF_TRIES '2'
      edit TEST_DIR '%BUILDIR%'
    task install
      trigger test == complete
      edit ECF_TRIES '2'
    task install_log
      trigger install == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install_log == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family cct
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DENABLE_INSTALL=ON -DLAPACK_FOUND=TRUE'
    edit CPUS '$EC_total_tasks'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/ecbuild/cct'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/ecbuild/cct:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cct-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '4'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '6000'
    edit MODULE_ACTIONS '"module swap PrgEnv-cray PrgEnv-gnu" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cct'
    edit QSUB_LP 'QSUB -lP %EC_TOTAL_TASKS%'
    edit QUEUE 'nf'
    edit REMOTE_HOST 'cct'
    edit SITE 'cct'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task build
      trigger config == complete
      edit ECF_TRIES '2'
      label host ""
      label options ""
    task test
      trigger build == complete
      edit ECF_TRIES '2'
      edit TEST_DIR '%BUILDIR%'
    task install
      trigger test == complete
      edit ECF_TRIES '2'
    task install_log
      trigger install == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install_log == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family ecflow_metab_vm
    defstatus suspended
    edit BANNER_TASK_END 'echo %TASK%-end'
    edit BANNER_TASK_START 'echo %TASK%-start'
    edit CI_LIMIT '1'
    edit ECF_JOB_CMD '%ECF_JOB% 1> %ECF_JOBOUT% 2>&1'
    edit ECF_KILL_CMD 'kill -15 %ECF_RID%'
    edit ECF_STATUS_CMD 'ps --sid %ECF_RID% -f'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit M_SITE 'vsms1'
    edit REMOTE_HOST 'polonius'
    edit SITE 'vsms1'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/ecflow-metab/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task build
      trigger config == complete
      edit ECF_TRIES '2'
      label host ""
      label options ""
    task test
      trigger build == complete
      edit ECF_TRIES '2'
      edit TEST_DIR '%BUILDIR%'
    task install
      trigger test == complete
      edit ECF_TRIES '2'
    task install_log
      trigger install == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install_log == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family ecgb
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'ecgb'
    edit QUEUE 'express'
    edit REMOTE_HOST 'ecgb'
    edit SITE 'ecgb'
    edit WSDIR '$SCRATCH/metabuilds/%ECF_HOST%_%ECF_PORT%/ecgb/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task build
      trigger config == complete
      edit ECF_TRIES '2'
      label host ""
      label options ""
    task test
      trigger build == complete
      edit ECF_TRIES '2'
      edit TEST_DIR '%BUILDIR%'
    task install
      trigger test == complete
      edit ECF_TRIES '2'
    task install_log
      trigger install == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install_log == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family gorax_dev_000
    defstatus suspended
    edit BANNER_TASK_END 'echo %TASK%-end'
    edit BANNER_TASK_START 'echo %TASK%-start'
    edit BASEPREFIX '/data/webdev/deploy/install/%SITE%'
    edit CI_LIMIT '2'
    edit ECF_CLIENT_EXE_PATH '/home/atls/deploy/install/ecflow/5.3.0/bin/ecflow_client'
    edit ECF_LOGHOST 'gorax-dev-000'
    edit ECF_LOGPORT '38562'
    edit ECF_OUT '/home/atls/deploy/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALLABLE_TO_USR_LOCAL_APPS 'False'
    edit INSTALL_POSTFIX ''
    edit LOGDIR '%ECF_OUT%'
    edit MODULE_ACTIONS ''
    edit MODULE_ACTIONS_HDR 'empty.h'
    edit MODULE_CLEANUP ''
    edit M_SITE 'gorax_dev_000'
    edit REMOTE_HOST 'gorax-dev-000'
    edit SITE 'gorax_dev_000'
    edit WSDIR '/data/webdev/%ECF_HOST%_%ECF_PORT%/ecflow-metab/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task build
      trigger config == complete
      edit ECF_TRIES '2'
      label host ""
      label options ""
    task install
      trigger build == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family gorax_dev_001
    defstatus suspended
    edit BANNER_TASK_END 'echo %TASK%-end'
    edit BANNER_TASK_START 'echo %TASK%-start'
    edit BASEPREFIX '/data/webdev/deploy/install/%SITE%'
    edit CI_LIMIT '2'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCPACK_PACKAGING_INSTALL_PREFIX=/opt/ecmwf/%CPACK_PACKAGE_NAME% -DENABLE_INSTALL=ON -DENABLE_PYTHON=OFF'
    edit CPACK_PACKAGE_NAME '%MODULE_NAME%'
    edit CPACK_PACKAGING_INSTALL_PREFIX '/opt/ecmwf/%CPACK_PACKAGE_NAME%'
    edit ECBUILD '%BASEDIR%/ecbuild/bin/ecbuild'
    edit ECBUILD_CLONE 'True'
    edit ECBUILD_CLONE_COMMIT '%ECBUILD_GIT_TAG%'
    edit ECF_CLIENT_EXE_PATH '/home/atls/deploy/install/ecflow/5.3.0/bin/ecflow_client'
    edit ECF_LOGHOST 'gorax-dev-001'
    edit ECF_LOGPORT '38562'
    edit ECF_OUT '/home/atls/deploy/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALLABLE_TO_USR_LOCAL_APPS 'False'
    edit INSTALL_POSTFIX ''
    edit LOGDIR '%ECF_OUT%'
    edit MODULE_ACTIONS ''
    edit MODULE_ACTIONS_HDR 'empty.h'
    edit MODULE_CLEANUP ''
    edit M_SITE 'gorax_dev_001'
    edit REMOTE_HOST 'gorax-dev-001'
    edit RPM_ARCH 'x86_64'
    edit RPM_FILENAME '%CPACK_PACKAGE_NAME%-${PACKAGE_VERSION}-Linux-%RPM_ARCH%.rpm'
    edit RPM_PASSWORD ''
    edit RPM_REPO_DESTINATION 'https://nexus.ecmwf.int/repository/CentOS-staging/7/8/rpms/%RPM_ARCH%/%RPM_FILENAME%'
    edit RPM_USERNAME ''
    edit SITE 'gorax_dev_001'
    edit WSDIR '/data/webdev/%ECF_HOST%_%ECF_PORT%/ecflow-metab/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task build
      trigger config == complete
      edit ECF_TRIES '2'
      label host ""
      label options ""
    task make_rpm
      trigger build == complete
    task wipe_build
      trigger make_rpm == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family leap42
    defstatus suspended
    edit CI_LIMIT '7'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit GIT_CLONE_OPTIONS ''
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'leap42'
    edit REMOTE_HOST 'leap42-deploy'
    edit SITE 'leap42'
    edit WSDIR '/tmp/metabuilds/%ECF_HOST%_%ECF_PORT%/leap42/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task pull_from_github
      trigger git_srcs == complete
    task make_build_dir
      trigger pull_from_github == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task build
      trigger config == complete
      edit ECF_TRIES '2'
      label host ""
      label options ""
    task test
      trigger build == complete
      edit ECF_TRIES '2'
      edit TEST_DIR '%BUILDIR%'
    task create_tarball
      trigger test == complete
      edit CMAKE 'cmake'
      edit EXPORTS '"export ECACCOUNT=ecodmdma" "export ecbuild_ROOT=%BASEPREFIX%/ecbuild/develop"'
      label tar ""
      label tardir ""
      label error ""
    task install
      trigger create_tarball == complete
      edit ECF_TRIES '2'
    task push_to_github
      trigger install == complete
    task push_to_bitbucket
      trigger push_to_github == complete
    task install_log
      trigger push_to_bitbucket == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install_log == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family lxc
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'lxc'
    edit QUEUE 'express'
    edit REMOTE_HOST 'lxc'
    edit SITE 'lxc'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxc/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task build
      trigger config == complete
      edit ECF_TRIES '2'
      label host ""
      label options ""
    task test
      trigger build == complete
      edit ECF_TRIES '2'
      edit TEST_DIR '%BUILDIR%'
    task install
      trigger test == complete
      edit ECF_TRIES '2'
    task install_log
      trigger install == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install_log == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family lxg
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'lxg'
    edit QUEUE 'normal'
    edit REMOTE_HOST 'lxg'
    edit SITE 'lxg'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxg/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task build
      trigger config == complete
      edit ECF_TRIES '2'
      label host ""
      label options ""
    task test
      trigger build == complete
      edit ECF_TRIES '2'
      edit TEST_DIR '%BUILDIR%'
    task install
      trigger test == complete
      edit ECF_TRIES '2'
    task install_log
      trigger install == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install_log == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family sappb
    defstatus suspended
    edit ASAN_FILE '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
    edit BASEPREFIX '/autofs/vol/deploy/install/%SITE%'
    edit CI_LIMIT '8'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit M_SITE 'sappb'
    edit REMOTE_HOST 'sappb-deploy'
    edit SANITIZER_REFERENCE_DIRECTORY '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
    edit SITE 'sappb'
    edit TSAN_FILE '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
    edit WSDIR '$HOME/metabuilds/%ECF_HOST%_%ECF_PORT%/sappb/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task build
      trigger config == complete
      edit ECF_TRIES '2'
      label host ""
      label options ""
    task test
      trigger build == complete
      edit ECF_TRIES '2'
      edit TEST_DIR '%BUILDIR%'
    task install
      trigger test == complete
      edit ECF_TRIES '2'
    task install_log
      trigger install == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install_log == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
endsuite
suite eccodes
  edit ADDRESS_SANITIZER 'False'
  edit ADD_ACTION_TRIGGERS 'True'
  edit AEC_VERSION '1.0.4'
  edit ALLOW_INSTALL_TO_OVERWRITE 'True'
  edit ASAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
  edit ATLAS_GIT_TAG 'develop'
  edit ATLAS_VERSION '0.20.2'
  edit BANNER_TASK_END 'banner %TASK%-end || true'
  edit BANNER_TASK_START 'banner %TASK%-start || true'
  edit BASEDIR '%WSDIR%/%PACKAGE%'
  edit BASEPREFIX '$PERM/install/%SITE%'
  edit BUFRDC_EMOSLIB_VERSION '4.6.3'
  edit BUFRDC_TABLES_GIT_TAG 'develop'
  edit BUFRDC_TABLES_VERSION '000415'
  edit BUILDIR '%BASEDIR%/builds'
  edit BUILD_ALWAYS_PARALLEL 'False'
  edit CFI_SANITIZER 'False'
  edit CFL_BASE_URL 'https://software-test.ecmwf.int/wiki'
  edit CFL_SPACE_KEY 'ECC'
  edit CFL_TARBALL_COMMENT ''
  edit CFL_TITLE 'Releases'
  edit CHECK_SYNC_RELEASE 'False'
  edit CLANG_TIDY_ABORT_ON_ERROR 'True'
  edit CMAKE '%ECBUILD%'
  edit CMAKELIST_EXTRACT_PACKAGE_VERSION '$(awk '/^project/ && /%PACKAGE%/ && /VERSION/ {for (I=1;I<=NF;I++) if ($I == "VERSION") {print $(I+1)};}' CMakeLists.txt)'
  edit CMAKE_BUILD_TYPE 'RelWithDebInfo'
  edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DECCODES_IFS_SAMPLES_SUFF=ifs_samples -DECCODES_INSTALL_EXTRA_TOOLS=ON -DENABLE_AEC=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXTRA_TESTS=ON -DOPENJPEG_INCLUDE_DIR=${OPENJPEG_DIR-dummy}/include/openjpeg2'
  edit CMAKE_VERSION '3.13.4'
  edit CODES_UI_GIT_TAG 'develop'
  edit CODES_UI_VERSION '1.3.1'
  edit CPACK 'make package'
  edit CPUS '$(lscpu -p | grep -v '#' | wc -l)'
  edit CRAY_FFTW_VERSION '3.3.8.1'
  edit CRAY_NETCDF_VERSION '4.6.1.3'
  edit CRON_BASEPREFIX '%BASEDIR%/install'
  edit CRON_EMAIL ''
  edit CRON_ENABLED 'False'
  edit CRON_GIT_TAG 'develop'
  edit CUSTOM_COMMANDS ''
  edit DATAFLOW_SANITIZER 'False'
  edit DEFAULT_GNU_COMPILER 'gnu.73'
  edit ECBUILD '%BASEPREFIX%/ecbuild/develop/bin/ecbuild'
  edit ECBUILD_CLONE 'False'
  edit ECBUILD_CLONE_COMMIT 'master'
  edit ECBUILD_GIT_TAG 'develop'
  edit ECBUILD_REPO '%GIT_REPO_PREFIX%/ecsdk/ecbuild.git'
  edit ECBUILD_VERSION '3.3.2'
  edit ECCODES_EXTENSIONS_GIT_TAG 'develop'
  edit ECCODES_EXTENSIONS_VERSION '2.17.0'
  edit ECCODES_GIT_TAG 'develop'
  edit ECCODES_PYTHON_VERSION '0.9.6'
  edit ECCODES_VERSION '2.17.1'
  edit ECFLOW5_LATEST_VERSION '5.4.0'
  edit ECFLOW_VERSION '5.4.1'
  edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/eccodes'
  edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/eccodes'
  edit ECF_JOB_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_JOB% %ECF_JOBOUT%'
  edit ECF_KILL_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% kill'
  edit ECF_STATUS_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% stat'
  edit ECF_TRIES '1'
  edit ECKIT_GIT_TAG 'develop'
  edit ECKIT_VERSION '1.10.1'
  edit ECQUIS_VERSION '1.0.0'
  edit EMOS_MOD_VERSION '462'
  edit EMOS_PA_VERSION '000462'
  edit EXPORTS '"export ECACCOUNT=ecodmdma"'
  edit EXPORT_VARIABLES 'False'
  edit FCKIT_GIT_TAG '0.6.3'
  edit FCKIT_VERSION '0.6.3'
  edit FDB_VINTAGE_GIT_TAG '4.10.0'
  edit FDB_VINTAGE_VERSION '4.10.0'
  edit FFTW_VERSION '3.3.8'
  edit FFTW_VERSION_CRAY '3.3.4.11'
  edit GIT 'git'
  edit GITHUB_PACKAGE_NAME '%PACKAGE_NAME%'
  edit GITHUB_REPO 'git@github.com:ecmwf'
  edit GITREPO '%GIT_REPO_PREFIX%/eccodes/eccodes.git'
  edit GIT_CLONE 'True'
  edit GIT_CLONE_OPTIONS '--depth=1 --branch %GIT_TAG%'
  edit GIT_METABUILDER_TAG 'ci'
  edit GIT_REPO_PREFIX 'ssh://git@git.ecmwf.int:7999'
  edit GIT_TAG 'develop'
  edit GIT_VERSION '2.26.0'
  edit GRIB_API_VERSION '1.28.0'
  edit INSTALLABLE_TO_USR_LOCAL_APPS 'True'
  edit IWYU_VERSION '0.11'
  edit KSH_EXTRACT_PACKAGE_VERSION '`grep _VERSION_STR VERSION.cmake | grep -o '[0-9]*\.[0-9]*\.[0-9]*'`'
  edit LAPACK_VERSION '3.5.0'
  edit LIBEMOS_GIT_TAG 'develop'
  edit LIBEMOS_VERSION '4.6.2'
  edit LINTERS_REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}'
  edit LSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/lsan.txt'
  edit MAGICS_GIT_TAG 'develop'
  edit MAGICS_PYTHON_VERSION 'master'
  edit MAGICS_VERSION '4.3.3'
  edit MAKE 'make'
  edit MARS_CLIENT_GIT_TAG '6.27.2.2'
  edit MARS_CLIENT_PACKAGE_GIT_TAG '6.27.6'
  edit MARS_CLIENT_PACKAGE_VERSION '6.27.6'
  edit MARS_CLIENT_VERSION '6.27.2.2'
  edit MEMORY_SANITIZER 'False'
  edit METABUILDER_HOME '/var/tmp/ma0/workspace/metabuilder'
  edit METAB_INSTALL_REMOVE_WRITE_PERMISSIONS 'True'
  edit METAB_PYTHON_VERSION 'python3'
  edit METKIT_GIT_TAG 'develop'
  edit METKIT_VERSION '1.4.6'
  edit METVIEW_GIT_TAG 'develop'
  edit METVIEW_PYTHON_VERSION '1.4.2'
  edit METVIEW_VERSION '5.8.3'
  edit MIR_BUNDLE_VERSION '0.9.0.0'
  edit MIR_GIT_TAG 'develop'
  edit MIR_VERSION '1.5.0'
  edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load openjpeg/%OPENJPEG_VERSION%" "module load aec/%AEC_VERSION%" "module load ecflow/%ECF_VERSION%"'
  edit MODULE_ACTIONS_FIRST ''
  edit MODULE_ACTIONS_HDR 'module_actions.h'
  edit MODULE_CHECKING_HDR 'module_checking.h'
  edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow"'
  edit MODULE_LOAD_SUPPORTED 'True'
  edit MODULE_NAME 'eccodes'
  edit MODULE_VERSION '%GIT_TAG%'
  edit NETCDF4_VERSION '4.7.4'
  edit ODB_API_GIT_TAG 'develop'
  edit ODB_API_VERSION '0.19.4'
  edit ODB_GIT_TAG 'develop'
  edit ODB_TOOLS_GIT_TAG 'develop'
  edit ODB_TOOLS_VERSION '0.18.3'
  edit ODB_VERSION '1.0.11'
  edit ODC_GIT_TAG 'develop'
  edit ODC_VERSION '1.0.3'
  edit OPENJPEG_VERSION '2.3.1'
  edit PACKAGE 'eccodes'
  edit PACKAGE_NAME 'eccodes'
  edit PACKAGE_OWNERS 'masn@ecmwf.int cgm@ecmwf.int ma0@ecmwf.int'
  edit PACKAGE_VERSION '%GIT_TAG%'
  edit PGEN_VERSION '1.7.1.0'
  edit PKG_BOOST_VERSION '1.71.0'
  edit PREFIX_APPEND ''
  edit PROJ4_VERSION '5.2.0'
  edit PROJ_VERSION '6.3.1'
  edit PYTHON2_VERSION '2.7.15-01'
  edit PYTHON3_VERSION '3.6.10-01'
  edit PYTHON3_VERSION_MAJOR_MINOR '3.6'
  edit PYTHON_GITREPO 'This-needs-tobe-overriden'
  edit PYTHON_PACKAGE 'This-needs-tobe-overriden'
  edit QT_VERSION '5.12.3'
  edit QUEUE 'build'
  edit REMOTE_HOST 'polonius'
  edit REUSE_BUILD_DIR 'False'
  edit SAFE_STACK_SANITIZER 'False'
  edit SANITIZER 'False'
  edit SANITIZER_REFERENCE_DIRECTORY '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
  edit SRCDIR '%BASEDIR%/%PACKAGE%'
  edit STATIC_ANALYSER_ABORT_ON_ERROR 'True'
  edit SWIG_VERSION '3.0.8'
  edit SYNC_RELEASE '2020.04.0'
  edit SYNC_RELEASE_CI_STAGE 'develop'
  edit TARBALL_DIR '/perm/ma/ma0/tmp_tarballs/ci'
  edit TAR_PACKAGE '%PACKAGE%'
  edit TEST 'ctest --output-on-failure -j${CPUS} '
  edit THREAD_SANITIZER 'False'
  edit TRANSI_GIT_TAG 'develop'
  edit TRANSI_VERSION '0.7.1'
  edit TSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
  edit UBSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/ubsan.txt'
  edit UB_SANITIZER 'False'
  edit USER 'ma0'
  edit USE_WORKDIR 'True'
  edit VALGRIND_VERSION '3.15.0'
  edit WORKDIR '%BASEDIR%/workdir/%ECF_NAME%'
  edit WSDIR '/var/tmp/tmpdir/deploy/jtmp.20739'
  label branch "develop"
  task regenerate
    defstatus complete
    edit ECF_JOB_CMD '%ECF_JOB% > %ECF_JOBOUT% 2>&1'
    edit ECF_KILL_CMD 'kill -15 %ECF_RID%'
    edit ECF_PASS 'FREE'
    edit ECF_STATUS_CMD 'ps --sid %ECF_RID% -f'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
    label date ""
    label metabuilder_branch "ci"
  family cca
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECCODES_IFS_SAMPLES_SUFF=ifs_samples -DECCODES_INSTALL_EXTRA_TOOLS=ON -DENABLE_AEC=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXTRA_TESTS=ON -DENABLE_NETCDF=ON -DLAPACK_FOUND=TRUE -DOPENJPEG_INCLUDE_DIR=${OPENJPEG_DIR-dummy}/include/openjpeg2'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/eccodes/cca'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/eccodes/cca:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cca-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '24000'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load openjpeg/%OPENJPEG_VERSION%" "module load aec/%AEC_VERSION%" "module load ecflow/%ECF_VERSION%" "module load perl"'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cca'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'cca'
    edit SITE 'cca'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DECCODES_IFS_SAMPLES_SUFF=ifs_samples -DECCODES_INSTALL_EXTRA_TOOLS=ON -DENABLE_AEC=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXTRA_TESTS=ON -DENABLE_NETCDF=ON -DLAPACK_FOUND=TRUE -DOPENJPEG_INCLUDE_DIR=${OPENJPEG_DIR-dummy}/include/openjpeg2'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load openjpeg/%OPENJPEG_VERSION%" "module load aec/%AEC_VERSION%" "module load ecflow/%ECF_VERSION%" "module load perl"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task post_install
        trigger install == complete
      task install_check
        trigger post_install == complete
      task install_log
        trigger install_check == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DECCODES_IFS_SAMPLES_SUFF=ifs_samples -DECCODES_INSTALL_EXTRA_TOOLS=ON -DENABLE_AEC=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXTRA_TESTS=ON -DENABLE_NETCDF=ON -DLAPACK_FOUND=TRUE -DOPENJPEG_INCLUDE_DIR=${OPENJPEG_DIR-dummy}/include/openjpeg2'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load openjpeg/%OPENJPEG_VERSION%" "module load aec/%AEC_VERSION%" "module load ecflow/%ECF_VERSION%" "module load perl"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task post_install
        trigger install == complete
      task install_check
        trigger post_install == complete
      task install_log
        trigger install_check == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.17
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECCODES_IFS_SAMPLES_SUFF=ifs_samples -DECCODES_INSTALL_EXTRA_TOOLS=ON -DENABLE_AEC=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXTRA_TESTS=ON -DENABLE_NETCDF=ON -DENABLE_PYTHON=OFF -DLAPACK_FOUND=TRUE -DOPENJPEG_INCLUDE_DIR=${OPENJPEG_DIR-dummy}/include/openjpeg2'
      edit COMPILER 'INTEL.17'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap intel/17.0.3.053" "module load gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load openjpeg/%OPENJPEG_VERSION%" "module load aec/%AEC_VERSION%" "module load ecflow/%ECF_VERSION%" "module load perl"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task post_install
        trigger install == complete
      task install_check
        trigger post_install == complete
      task install_log
        trigger install_check == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.85
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_C_FLAGS="-hfp1 -G2 -DNDEBUG" -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_Fortran_FLAGS=-emf -DECCODES_IFS_SAMPLES_SUFF=ifs_samples -DECCODES_INSTALL_EXTRA_TOOLS=ON -DENABLE_AEC=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXTRA_TESTS=ON -DENABLE_NETCDF=ON -DENABLE_PYTHON=OFF -DLAPACK_FOUND=TRUE -DOPENJPEG_INCLUDE_DIR=${OPENJPEG_DIR-dummy}/include/openjpeg2'
      edit COMPILER 'CRAY.85'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load openjpeg/%OPENJPEG_VERSION%" "module load aec/%AEC_VERSION%" "module load ecflow/%ECF_VERSION%" "module load perl"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task post_install
        trigger install == complete
      task install_check
        trigger post_install == complete
      task install_log
        trigger install_check == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.86
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_C_FLAGS="-hfp1 -G2 -DNDEBUG" -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_Fortran_FLAGS=-emf -DECCODES_IFS_SAMPLES_SUFF=ifs_samples -DECCODES_INSTALL_EXTRA_TOOLS=ON -DENABLE_AEC=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXTRA_TESTS=ON -DENABLE_NETCDF=ON -DENABLE_PYTHON=OFF -DLAPACK_FOUND=TRUE -DOPENJPEG_INCLUDE_DIR=${OPENJPEG_DIR-dummy}/include/openjpeg2'
      edit COMPILER 'CRAY.86'
      edit MODULE_ACTIONS '"module swap cdt/17.09" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load openjpeg/%OPENJPEG_VERSION%" "module load aec/%AEC_VERSION%" "module load ecflow/%ECF_VERSION%" "module load perl"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task post_install
        trigger install == complete
      task install_check
        trigger post_install == complete
      task install_log
        trigger install_check == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_C_FLAGS="-hfp1 -G2 -DNDEBUG" -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_Fortran_FLAGS=-emf -DECCODES_IFS_SAMPLES_SUFF=ifs_samples -DECCODES_INSTALL_EXTRA_TOOLS=ON -DENABLE_AEC=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXTRA_TESTS=ON -DENABLE_NETCDF=ON -DENABLE_PYTHON=OFF -DLAPACK_FOUND=TRUE -DOPENJPEG_INCLUDE_DIR=${OPENJPEG_DIR-dummy}/include/openjpeg2'
      edit COMPILER 'CRAY.87'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load openjpeg/%OPENJPEG_VERSION%" "module load aec/%AEC_VERSION%" "module load ecflow/%ECF_VERSION%" "module load perl"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task post_install
        trigger install == complete
      task install_check
        trigger post_install == complete
      task install_log
        trigger install_check == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family ccb
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECCODES_IFS_SAMPLES_SUFF=ifs_samples -DECCODES_INSTALL_EXTRA_TOOLS=ON -DENABLE_AEC=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXTRA_TESTS=ON -DENABLE_NETCDF=ON -DLAPACK_FOUND=TRUE -DOPENJPEG_INCLUDE_DIR=${OPENJPEG_DIR-dummy}/include/openjpeg2'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/eccodes/ccb'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/eccodes/ccb:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'ccb-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '24000'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load openjpeg/%OPENJPEG_VERSION%" "module load aec/%AEC_VERSION%" "module load ecflow/%ECF_VERSION%" "module load perl"'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'ccb'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'ccb'
    edit SITE 'ccb'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DECCODES_IFS_SAMPLES_SUFF=ifs_samples -DECCODES_INSTALL_EXTRA_TOOLS=ON -DENABLE_AEC=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXTRA_TESTS=ON -DENABLE_NETCDF=ON -DLAPACK_FOUND=TRUE -DOPENJPEG_INCLUDE_DIR=${OPENJPEG_DIR-dummy}/include/openjpeg2'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load openjpeg/%OPENJPEG_VERSION%" "module load aec/%AEC_VERSION%" "module load ecflow/%ECF_VERSION%" "module load perl"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task post_install
        trigger install == complete
      task install_check
        trigger post_install == complete
      task install_log
        trigger install_check == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DECCODES_IFS_SAMPLES_SUFF=ifs_samples -DECCODES_INSTALL_EXTRA_TOOLS=ON -DENABLE_AEC=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXTRA_TESTS=ON -DENABLE_NETCDF=ON -DLAPACK_FOUND=TRUE -DOPENJPEG_INCLUDE_DIR=${OPENJPEG_DIR-dummy}/include/openjpeg2'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load openjpeg/%OPENJPEG_VERSION%" "module load aec/%AEC_VERSION%" "module load ecflow/%ECF_VERSION%" "module load perl"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task post_install
        trigger install == complete
      task install_check
        trigger post_install == complete
      task install_log
        trigger install_check == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.17
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECCODES_IFS_SAMPLES_SUFF=ifs_samples -DECCODES_INSTALL_EXTRA_TOOLS=ON -DENABLE_AEC=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXTRA_TESTS=ON -DENABLE_NETCDF=ON -DENABLE_PYTHON=OFF -DLAPACK_FOUND=TRUE -DOPENJPEG_INCLUDE_DIR=${OPENJPEG_DIR-dummy}/include/openjpeg2'
      edit COMPILER 'INTEL.17'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap intel/17.0.3.053" "module load gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load openjpeg/%OPENJPEG_VERSION%" "module load aec/%AEC_VERSION%" "module load ecflow/%ECF_VERSION%" "module load perl"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task post_install
        trigger install == complete
      task install_check
        trigger post_install == complete
      task install_log
        trigger install_check == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.85
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_C_FLAGS="-hfp1 -G2 -DNDEBUG" -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_Fortran_FLAGS=-emf -DECCODES_IFS_SAMPLES_SUFF=ifs_samples -DECCODES_INSTALL_EXTRA_TOOLS=ON -DENABLE_AEC=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXTRA_TESTS=ON -DENABLE_NETCDF=ON -DENABLE_PYTHON=OFF -DLAPACK_FOUND=TRUE -DOPENJPEG_INCLUDE_DIR=${OPENJPEG_DIR-dummy}/include/openjpeg2'
      edit COMPILER 'CRAY.85'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load openjpeg/%OPENJPEG_VERSION%" "module load aec/%AEC_VERSION%" "module load ecflow/%ECF_VERSION%" "module load perl"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task post_install
        trigger install == complete
      task install_check
        trigger post_install == complete
      task install_log
        trigger install_check == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.86
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_C_FLAGS="-hfp1 -G2 -DNDEBUG" -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_Fortran_FLAGS=-emf -DECCODES_IFS_SAMPLES_SUFF=ifs_samples -DECCODES_INSTALL_EXTRA_TOOLS=ON -DENABLE_AEC=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXTRA_TESTS=ON -DENABLE_NETCDF=ON -DENABLE_PYTHON=OFF -DLAPACK_FOUND=TRUE -DOPENJPEG_INCLUDE_DIR=${OPENJPEG_DIR-dummy}/include/openjpeg2'
      edit COMPILER 'CRAY.86'
      edit MODULE_ACTIONS '"module swap cdt/17.09" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load openjpeg/%OPENJPEG_VERSION%" "module load aec/%AEC_VERSION%" "module load ecflow/%ECF_VERSION%" "module load perl"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task post_install
        trigger install == complete
      task install_check
        trigger post_install == complete
      task install_log
        trigger install_check == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_C_FLAGS="-hfp1 -G2 -DNDEBUG" -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_Fortran_FLAGS=-emf -DECCODES_IFS_SAMPLES_SUFF=ifs_samples -DECCODES_INSTALL_EXTRA_TOOLS=ON -DENABLE_AEC=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXTRA_TESTS=ON -DENABLE_NETCDF=ON -DENABLE_PYTHON=OFF -DLAPACK_FOUND=TRUE -DOPENJPEG_INCLUDE_DIR=${OPENJPEG_DIR-dummy}/include/openjpeg2'
      edit COMPILER 'CRAY.87'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load openjpeg/%OPENJPEG_VERSION%" "module load aec/%AEC_VERSION%" "module load ecflow/%ECF_VERSION%" "module load perl"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task post_install
        trigger install == complete
      task install_check
        trigger post_install == complete
      task install_log
        trigger install_check == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family cct
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECCODES_IFS_SAMPLES_SUFF=ifs_samples -DECCODES_INSTALL_EXTRA_TOOLS=ON -DENABLE_AEC=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXTRA_TESTS=ON -DENABLE_NETCDF=ON -DLAPACK_FOUND=TRUE -DOPENJPEG_INCLUDE_DIR=${OPENJPEG_DIR-dummy}/include/openjpeg2'
    edit CPUS '$EC_total_tasks'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/eccodes/cct'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/eccodes/cct:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cct-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '4'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '6000'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load openjpeg/%OPENJPEG_VERSION%" "module load aec/%AEC_VERSION%" "module load ecflow/%ECF_VERSION%" "module load perl"'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cct'
    edit QSUB_LP 'QSUB -lP %EC_TOTAL_TASKS%'
    edit QUEUE 'nf'
    edit REMOTE_HOST 'cct'
    edit SITE 'cct'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DECCODES_IFS_SAMPLES_SUFF=ifs_samples -DECCODES_INSTALL_EXTRA_TOOLS=ON -DENABLE_AEC=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXTRA_TESTS=ON -DENABLE_NETCDF=ON -DLAPACK_FOUND=TRUE -DOPENJPEG_INCLUDE_DIR=${OPENJPEG_DIR-dummy}/include/openjpeg2'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load openjpeg/%OPENJPEG_VERSION%" "module load aec/%AEC_VERSION%" "module load ecflow/%ECF_VERSION%" "module load perl"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task post_install
        trigger install == complete
      task install_check
        trigger post_install == complete
      task install_log
        trigger install_check == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DECCODES_IFS_SAMPLES_SUFF=ifs_samples -DECCODES_INSTALL_EXTRA_TOOLS=ON -DENABLE_AEC=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXTRA_TESTS=ON -DENABLE_NETCDF=ON -DLAPACK_FOUND=TRUE -DOPENJPEG_INCLUDE_DIR=${OPENJPEG_DIR-dummy}/include/openjpeg2'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load openjpeg/%OPENJPEG_VERSION%" "module load aec/%AEC_VERSION%" "module load ecflow/%ECF_VERSION%" "module load perl"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task post_install
        trigger install == complete
      task install_check
        trigger post_install == complete
      task install_log
        trigger install_check == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.17
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECCODES_IFS_SAMPLES_SUFF=ifs_samples -DECCODES_INSTALL_EXTRA_TOOLS=ON -DENABLE_AEC=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXTRA_TESTS=ON -DENABLE_NETCDF=ON -DENABLE_PYTHON=OFF -DLAPACK_FOUND=TRUE -DOPENJPEG_INCLUDE_DIR=${OPENJPEG_DIR-dummy}/include/openjpeg2'
      edit COMPILER 'INTEL.17'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap intel/17.0.3.053" "module load gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load openjpeg/%OPENJPEG_VERSION%" "module load aec/%AEC_VERSION%" "module load ecflow/%ECF_VERSION%" "module load perl"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task post_install
        trigger install == complete
      task install_check
        trigger post_install == complete
      task install_log
        trigger install_check == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.85
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_C_FLAGS="-hfp1 -G2 -DNDEBUG" -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_Fortran_FLAGS=-emf -DECCODES_IFS_SAMPLES_SUFF=ifs_samples -DECCODES_INSTALL_EXTRA_TOOLS=ON -DENABLE_AEC=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXTRA_TESTS=ON -DENABLE_NETCDF=ON -DENABLE_PYTHON=OFF -DLAPACK_FOUND=TRUE -DOPENJPEG_INCLUDE_DIR=${OPENJPEG_DIR-dummy}/include/openjpeg2'
      edit COMPILER 'CRAY.85'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load openjpeg/%OPENJPEG_VERSION%" "module load aec/%AEC_VERSION%" "module load ecflow/%ECF_VERSION%" "module load perl"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task post_install
        trigger install == complete
      task install_check
        trigger post_install == complete
      task install_log
        trigger install_check == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.86
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_C_FLAGS="-hfp1 -G2 -DNDEBUG" -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_Fortran_FLAGS=-emf -DECCODES_IFS_SAMPLES_SUFF=ifs_samples -DECCODES_INSTALL_EXTRA_TOOLS=ON -DENABLE_AEC=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXTRA_TESTS=ON -DENABLE_NETCDF=ON -DENABLE_PYTHON=OFF -DLAPACK_FOUND=TRUE -DOPENJPEG_INCLUDE_DIR=${OPENJPEG_DIR-dummy}/include/openjpeg2'
      edit COMPILER 'CRAY.86'
      edit MODULE_ACTIONS '"module swap cdt/17.09" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load openjpeg/%OPENJPEG_VERSION%" "module load aec/%AEC_VERSION%" "module load ecflow/%ECF_VERSION%" "module load perl"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task post_install
        trigger install == complete
      task install_check
        trigger post_install == complete
      task install_log
        trigger install_check == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_C_FLAGS="-hfp1 -G2 -DNDEBUG" -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_Fortran_FLAGS=-emf -DECCODES_IFS_SAMPLES_SUFF=ifs_samples -DECCODES_INSTALL_EXTRA_TOOLS=ON -DENABLE_AEC=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXTRA_TESTS=ON -DENABLE_NETCDF=ON -DENABLE_PYTHON=OFF -DLAPACK_FOUND=TRUE -DOPENJPEG_INCLUDE_DIR=${OPENJPEG_DIR-dummy}/include/openjpeg2'
      edit COMPILER 'CRAY.87'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load openjpeg/%OPENJPEG_VERSION%" "module load aec/%AEC_VERSION%" "module load ecflow/%ECF_VERSION%" "module load perl"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task post_install
        trigger install == complete
      task install_check
        trigger post_install == complete
      task install_log
        trigger install_check == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family ecgb
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'ecgb'
    edit QUEUE 'express'
    edit REMOTE_HOST 'ecgb'
    edit SITE 'ecgb'
    edit WSDIR '$SCRATCH/metabuilds/%ECF_HOST%_%ECF_PORT%/ecgb/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECCODES_IFS_SAMPLES_SUFF=ifs_samples -DECCODES_INSTALL_EXTRA_TOOLS=ON -DENABLE_AEC=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXTRA_TESTS=ON -DOPENJPEG_INCLUDE_DIR=${OPENJPEG_DIR-dummy}/include/openjpeg2'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load openjpeg/%OPENJPEG_VERSION%" "module load aec/%AEC_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task post_install
        trigger install == complete
      task install_check
        trigger post_install == complete
      task install_log
        trigger install_check == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family gorax_dev_001
    defstatus suspended
    edit BANNER_TASK_END 'echo %TASK%-end'
    edit BANNER_TASK_START 'echo %TASK%-start'
    edit BASEPREFIX '/data/webdev/deploy/install/%SITE%'
    edit CI_LIMIT '2'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCPACK_PACKAGING_INSTALL_PREFIX=/opt/ecmwf/%CPACK_PACKAGE_NAME% -DECCODES_IFS_SAMPLES_SUFF=ifs_samples -DECCODES_INSTALL_EXTRA_TOOLS=ON -DENABLE_AEC=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXTRA_TESTS=ON -DENABLE_PYTHON=OFF -DOPENJPEG_INCLUDE_DIR=${OPENJPEG_DIR-dummy}/include/openjpeg2'
    edit CPACK_PACKAGE_NAME '%MODULE_NAME%'
    edit CPACK_PACKAGING_INSTALL_PREFIX '/opt/ecmwf/%CPACK_PACKAGE_NAME%'
    edit ECBUILD '%BASEDIR%/ecbuild/bin/ecbuild'
    edit ECBUILD_CLONE 'True'
    edit ECBUILD_CLONE_COMMIT '%ECBUILD_GIT_TAG%'
    edit ECF_CLIENT_EXE_PATH '/home/atls/deploy/install/ecflow/5.3.0/bin/ecflow_client'
    edit ECF_LOGHOST 'gorax-dev-001'
    edit ECF_LOGPORT '38562'
    edit ECF_OUT '/home/atls/deploy/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALLABLE_TO_USR_LOCAL_APPS 'False'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX ''
    edit LOGDIR '%ECF_OUT%'
    edit MODULE_ACTIONS ''
    edit MODULE_ACTIONS_HDR 'empty.h'
    edit MODULE_CLEANUP ''
    edit M_SITE 'gorax_dev_001'
    edit REMOTE_HOST 'gorax-dev-001'
    edit RPM_ARCH 'x86_64'
    edit RPM_FILENAME '%CPACK_PACKAGE_NAME%-${PACKAGE_VERSION}-Linux-%RPM_ARCH%.rpm'
    edit RPM_PASSWORD ''
    edit RPM_REPO_DESTINATION 'https://nexus.ecmwf.int/repository/CentOS-staging/7/8/rpms/%RPM_ARCH%/%RPM_FILENAME%'
    edit RPM_USERNAME ''
    edit SITE 'gorax_dev_001'
    edit WSDIR '/data/webdev/%ECF_HOST%_%ECF_PORT%/ecflow-metab/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task build
      trigger config == complete
      edit ECF_TRIES '2'
      label host ""
      label options ""
    task make_rpm
      trigger build == complete
    task wipe_build
      trigger make_rpm == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family leap42
    defstatus suspended
    edit CI_LIMIT '7'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DECCODES_IFS_SAMPLES_SUFF=ifs_samples -DECCODES_INSTALL_EXTRA_TOOLS=ON -DENABLE_AEC=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXTRA_TESTS=ON -DENABLE_PNG=ON -DOPENJPEG_INCLUDE_DIR=${OPENJPEG_DIR-dummy}/include/openjpeg2'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'leap42'
    edit REMOTE_HOST 'leap42-deploy'
    edit SITE 'leap42'
    edit WSDIR '/tmp/metabuilds/%ECF_HOST%_%ECF_PORT%/leap42/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECCODES_IFS_SAMPLES_SUFF=ifs_samples -DECCODES_INSTALL_EXTRA_TOOLS=ON -DENABLE_AEC=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXTRA_TESTS=ON -DENABLE_PNG=ON -DOPENJPEG_INCLUDE_DIR=${OPENJPEG_DIR-dummy}/include/openjpeg2'
      edit COMPILER 'GNU.73'
      edit GIT_CLONE_OPTIONS ''
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load openjpeg/%OPENJPEG_VERSION%" "module load aec/%AEC_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task pull_from_github
        trigger git_srcs == complete
      task make_build_dir
        trigger pull_from_github == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task create_tarball
        trigger test == complete
        edit CMAKE 'cmake'
        edit EXPORTS '"export ECACCOUNT=ecodmdma" "export ecbuild_ROOT=%BASEPREFIX%/ecbuild/develop"'
        label tar ""
        label tardir ""
        label error ""
      task install
        trigger create_tarball == complete
        edit ECF_TRIES '2'
      task post_install
        trigger install == complete
      task install_check
        trigger post_install == complete
      task upload_tarball
        trigger install_check == complete
        edit METAB_PYTHON_VERSION 'python'
        edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
        edit UPLOAD_TARBALL '<package.h>'
        label url ""
        label space_key ""
        label msg ""
      task push_to_github
        trigger upload_tarball == complete
      task push_to_bitbucket
        trigger push_to_github == complete
      task install_log
        trigger push_to_bitbucket == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.93
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECCODES_IFS_SAMPLES_SUFF=ifs_samples -DECCODES_INSTALL_EXTRA_TOOLS=ON -DENABLE_AEC=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXTRA_TESTS=ON -DENABLE_PNG=ON -DOPENJPEG_INCLUDE_DIR=${OPENJPEG_DIR-dummy}/include/openjpeg2'
      edit COMPILER 'GNU.93'
      edit EXPORTS '"export ECACCOUNT=ecodmdma" "export LD_LIBRARY_PATH=/usr/local/apps/gcc/9.3.0/lib64/"'
      edit MODULE_ACTIONS '"module swap gnu/9.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load openjpeg/%OPENJPEG_VERSION%" "module load aec/%AEC_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task post_install
        trigger install == complete
      task install_check
        trigger post_install == complete
      task install_log
        trigger install_check == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.19
      edit COMPILER 'INTEL.19'
      edit MODULE_ACTIONS '"module unload gnu" "module load intel/19.0.4" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load openjpeg/%OPENJPEG_VERSION%" "module load aec/%AEC_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit POST_EXPORTS '"export CXX=icpc\ -cxxlib=/usr/local/apps/gcc/8.3.0" "export LD_RUN_PATH=/usr/local/apps/gcc/8.3.0/lib64" "export PATH=/usr/local/apps/gcc/8.3.0/bin:$PATH"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task post_install
        trigger install == complete
      task install_check
        trigger post_install == complete
      task install_log
        trigger install_check == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECCODES_IFS_SAMPLES_SUFF=ifs_samples -DECCODES_INSTALL_EXTRA_TOOLS=ON -DENABLE_AEC=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXTRA_TESTS=ON -DENABLE_PNG=ON -DOPENJPEG_INCLUDE_DIR=${OPENJPEG_DIR-dummy}/include/openjpeg2'
      edit COMPILER 'CLANG.90'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load openjpeg/%OPENJPEG_VERSION%" "module load aec/%AEC_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task post_install
        trigger install == complete
      task install_check
        trigger post_install == complete
      task install_log
        trigger install_check == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90.linters
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECCODES_IFS_SAMPLES_SUFF=ifs_samples -DECCODES_INSTALL_EXTRA_TOOLS=ON -DENABLE_AEC=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXTRA_TESTS=ON -DENABLE_FORTRAN=OFF -DENABLE_PNG=ON -DENABLE_PYTHON=OFF -DOPENJPEG_INCLUDE_DIR=${OPENJPEG_DIR-dummy}/include/openjpeg2'
      edit COMPILER 'CLANG.90.LINTERS'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load openjpeg/%OPENJPEG_VERSION%" "module load aec/%AEC_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task cppcheck
        trigger config == complete
        edit CPPCHECK_FLAGS '--template='[{file}:{line}]: {severity} : {id} : {message}' --suppress=syntaxError --inline-suppr --platform=unix64 --enable=performance --enable=portability --enable=information --std=c11 --std=c++14 -DDEBUG=1 -q -j8'
        edit CPPCHECK_VERSION '1.89'
        edit REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}/cpp_check'
      task iwyu
        trigger cppcheck == complete
        edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load openjpeg/%OPENJPEG_VERSION%" "module load aec/%AEC_VERSION%" "module load ecflow/%ECF_VERSION%" "module load iwyu/%IWYU_VERSION%"'
      family clang_linters
        trigger iwyu == complete
        limit clang_linters 1
        inlimit clang_linters
        task clang_tidy
          edit REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}/clang_tidy'
        task static_analyser
          edit REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}/static_analyser'
          edit SCAN_BUILD_OPTIONS '--force-analyze-debug-code -o /perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}/static_analyser -k make -j8'
          edit USE_CLANG_CHECK 'False'
          label scan_view ""
      endfamily
      task wipe_build
        trigger clang_linters == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90.address_sanitizer
      edit ADDRESS_SANITIZER 'True'
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_FLAGS="-fsanitize=address -fno-omit-frame-pointer" -DCMAKE_C_FLAGS="-fsanitize=address -fno-omit-frame-pointer" -DCMAKE_EXE_LINKER_FLAGS=-fsanitize=address -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECCODES_IFS_SAMPLES_SUFF=ifs_samples -DECCODES_INSTALL_EXTRA_TOOLS=ON -DENABLE_AEC=ON -DENABLE_ECCODES_THREADS=OFF -DENABLE_EXTRA_TESTS=ON -DENABLE_FORTRAN=OFF -DENABLE_PNG=ON -DENABLE_PYTHON=OFF -DOPENJPEG_INCLUDE_DIR=${OPENJPEG_DIR-dummy}/include/openjpeg2'
      edit COMPILER 'CLANG.90.ADDRESS_SANITIZER'
      edit LSAN_OPTIONS 'suppressions=%SRCDIR%/tests/lsan-leap42.supp'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load openjpeg/%OPENJPEG_VERSION%" "module load aec/%AEC_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit SANITIZER 'True'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit TEST_DIR '%BUILDIR%'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90.ub_sanitizer
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_FLAGS="-fsanitize=undefined -fno-omit-frame-pointer -fno-sanitize-recover=all" -DCMAKE_C_FLAGS="-fsanitize=undefined -fno-omit-frame-pointer" -DCMAKE_EXE_LINKER_FLAGS=-fsanitize=undefined -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECCODES_IFS_SAMPLES_SUFF=ifs_samples -DECCODES_INSTALL_EXTRA_TOOLS=ON -DENABLE_AEC=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXTRA_TESTS=ON -DENABLE_FORTRAN=OFF -DENABLE_PNG=ON -DENABLE_PYTHON=OFF -DOPENJPEG_INCLUDE_DIR=${OPENJPEG_DIR-dummy}/include/openjpeg2'
      edit COMPILER 'CLANG.90.UB_SANITIZER'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load openjpeg/%OPENJPEG_VERSION%" "module load aec/%AEC_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit SANITIZER 'True'
      edit UB_SANITIZER 'True'
      edit UB_SANITIZER_LD_PRELOAD 'export LD_PRELOAD=/usr/local/apps/clang/9.0.1/lib64/libubsan.so'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit TEST_DIR '%BUILDIR%'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family lxc
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load openjpeg/%OPENJPEG_VERSION%" "module load aec/%AEC_VERSION%" "module load ecflow/%ECF_VERSION%" "module load perl"'
    edit M_SITE 'lxc'
    edit QUEUE 'express'
    edit REMOTE_HOST 'lxc'
    edit SITE 'lxc'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxc/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECCODES_IFS_SAMPLES_SUFF=ifs_samples -DECCODES_INSTALL_EXTRA_TOOLS=ON -DENABLE_AEC=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXTRA_TESTS=ON -DOPENJPEG_INCLUDE_DIR=${OPENJPEG_DIR-dummy}/include/openjpeg2'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load openjpeg/%OPENJPEG_VERSION%" "module load aec/%AEC_VERSION%" "module load ecflow/%ECF_VERSION%" "module load perl"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task post_install
        trigger install == complete
      task install_check
        trigger post_install == complete
      task install_log
        trigger install_check == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family lxg
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load openjpeg/%OPENJPEG_VERSION%" "module load aec/%AEC_VERSION%" "module load ecflow/%ECF_VERSION%" "module load perl"'
    edit M_SITE 'lxg'
    edit QUEUE 'normal'
    edit REMOTE_HOST 'lxg'
    edit SITE 'lxg'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxg/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECCODES_IFS_SAMPLES_SUFF=ifs_samples -DECCODES_INSTALL_EXTRA_TOOLS=ON -DENABLE_AEC=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXTRA_TESTS=ON -DOPENJPEG_INCLUDE_DIR=${OPENJPEG_DIR-dummy}/include/openjpeg2'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load openjpeg/%OPENJPEG_VERSION%" "module load aec/%AEC_VERSION%" "module load ecflow/%ECF_VERSION%" "module load perl"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task post_install
        trigger install == complete
      task install_check
        trigger post_install == complete
      task install_log
        trigger install_check == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family sappb
    defstatus suspended
    edit ASAN_FILE '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
    edit BASEPREFIX '/autofs/vol/deploy/install/%SITE%'
    edit CI_LIMIT '8'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load openjpeg/%OPENJPEG_VERSION%" "module load aec/%AEC_VERSION%" "module load ecflow/%ECF_VERSION%" "module load perl"'
    edit M_SITE 'sappb'
    edit REMOTE_HOST 'sappb-deploy'
    edit SANITIZER_REFERENCE_DIRECTORY '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
    edit SITE 'sappb'
    edit TSAN_FILE '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
    edit WSDIR '$HOME/metabuilds/%ECF_HOST%_%ECF_PORT%/sappb/%COMPILER:%'
    family gnu.63
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECCODES_IFS_SAMPLES_SUFF=ifs_samples -DECCODES_INSTALL_EXTRA_TOOLS=ON -DENABLE_AEC=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXTRA_TESTS=ON -DOPENJPEG_INCLUDE_DIR=${OPENJPEG_DIR-dummy}/include/openjpeg2'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module load gnu/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load openjpeg/%OPENJPEG_VERSION%" "module load aec/%AEC_VERSION%" "module load ecflow/%ECF_VERSION%" "module load perl"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task post_install
        trigger install == complete
      task install_check
        trigger post_install == complete
      task install_log
        trigger install_check == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
endsuite
suite eccodes_python
  edit ADDRESS_SANITIZER 'False'
  edit ADD_ACTION_TRIGGERS 'True'
  edit AEC_VERSION '1.0.4'
  edit ALLOW_INSTALL_TO_OVERWRITE 'True'
  edit ASAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
  edit ATLAS_GIT_TAG 'develop'
  edit ATLAS_VERSION '0.20.2'
  edit BANNER_TASK_END 'banner %TASK%-end || true'
  edit BANNER_TASK_START 'banner %TASK%-start || true'
  edit BASEDIR '%WSDIR%/%PACKAGE%'
  edit BASEPREFIX '$PERM/install/%SITE%'
  edit BUFRDC_EMOSLIB_VERSION '4.6.3'
  edit BUFRDC_TABLES_GIT_TAG 'develop'
  edit BUFRDC_TABLES_VERSION '000415'
  edit BUILDIR '%BASEDIR%/builds'
  edit BUILD_ALWAYS_PARALLEL 'False'
  edit CFI_SANITIZER 'False'
  edit CFL_BASE_URL 'https://software-test.ecmwf.int/wiki'
  edit CFL_SPACE_KEY '<TODO>'
  edit CFL_TARBALL_COMMENT ''
  edit CFL_TITLE 'Releases'
  edit CHECK_SYNC_RELEASE 'False'
  edit CLANG_TIDY_ABORT_ON_ERROR 'False'
  edit CMAKE '%ECBUILD%'
  edit CMAKELIST_EXTRACT_PACKAGE_VERSION '$(awk '/^project/ && /%PACKAGE%/ && /VERSION/ {for (I=1;I<=NF;I++) if ($I == "VERSION") {print $(I+1)};}' CMakeLists.txt)'
  edit CMAKE_BUILD_TYPE 'RelWithDebInfo'
  edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE%'
  edit CMAKE_VERSION '3.13.4'
  edit CODES_UI_GIT_TAG 'develop'
  edit CODES_UI_VERSION '1.3.1'
  edit CPACK 'make package'
  edit CPUS '$(lscpu -p | grep -v '#' | wc -l)'
  edit CRAY_FFTW_VERSION '3.3.8.1'
  edit CRAY_NETCDF_VERSION '4.6.1.3'
  edit CRON_BASEPREFIX '%BASEDIR%/install'
  edit CRON_EMAIL ''
  edit CRON_ENABLED 'False'
  edit CRON_GIT_TAG 'develop'
  edit CUSTOM_COMMANDS ''
  edit DATAFLOW_SANITIZER 'False'
  edit DEFAULT_GNU_COMPILER 'gnu.73'
  edit ECBUILD '%BASEPREFIX%/ecbuild/develop/bin/ecbuild'
  edit ECBUILD_CLONE 'False'
  edit ECBUILD_CLONE_COMMIT 'master'
  edit ECBUILD_GIT_TAG 'develop'
  edit ECBUILD_REPO '%GIT_REPO_PREFIX%/ecsdk/ecbuild.git'
  edit ECBUILD_VERSION '3.3.2'
  edit ECCODES_EXTENSIONS_GIT_TAG 'develop'
  edit ECCODES_EXTENSIONS_VERSION '2.17.0'
  edit ECCODES_GIT_TAG 'develop'
  edit ECCODES_PYTHON_VERSION '0.9.6'
  edit ECCODES_VERSION '2.17.1'
  edit ECFLOW5_LATEST_VERSION '5.4.0'
  edit ECFLOW_VERSION '5.4.1'
  edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/eccodes_python'
  edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/eccodes_python'
  edit ECF_JOB_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_JOB% %ECF_JOBOUT%'
  edit ECF_KILL_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% kill'
  edit ECF_STATUS_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% stat'
  edit ECF_TRIES '1'
  edit ECKIT_GIT_TAG 'develop'
  edit ECKIT_VERSION '1.10.1'
  edit ECQUIS_VERSION '1.0.0'
  edit EMOS_MOD_VERSION '462'
  edit EMOS_PA_VERSION '000462'
  edit EXPORTS '"export ECACCOUNT=ecodmdma"'
  edit EXPORT_VARIABLES 'False'
  edit FCKIT_GIT_TAG '0.6.3'
  edit FCKIT_VERSION '0.6.3'
  edit FDB_VINTAGE_GIT_TAG '4.10.0'
  edit FDB_VINTAGE_VERSION '4.10.0'
  edit FFTW_VERSION '3.3.8'
  edit FFTW_VERSION_CRAY '3.3.4.11'
  edit GIT 'git'
  edit GITHUB_PACKAGE_NAME 'eccodes-python'
  edit GITHUB_REPO 'git@github.com:ecmwf'
  edit GITREPO '%GIT_REPO_PREFIX%/eccodes/eccodes-python.git'
  edit GIT_CLONE 'True'
  edit GIT_CLONE_OPTIONS '--depth=1 --branch %GIT_TAG%'
  edit GIT_METABUILDER_TAG 'ci'
  edit GIT_REPO_PREFIX 'ssh://git@git.ecmwf.int:7999'
  edit GIT_TAG 'develop'
  edit GIT_VERSION '2.26.0'
  edit GRIB_API_VERSION '1.28.0'
  edit INSTALLABLE_TO_USR_LOCAL_APPS 'True'
  edit INSTALL_DIR '%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX%'
  edit IWYU_VERSION '0.11'
  edit KSH_EXTRACT_PACKAGE_VERSION '`grep _VERSION_STR VERSION.cmake | grep -o '[0-9]*\.[0-9]*\.[0-9]*'`'
  edit LAPACK_VERSION '3.5.0'
  edit LIBEMOS_GIT_TAG 'develop'
  edit LIBEMOS_VERSION '4.6.2'
  edit LINTERS_REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}'
  edit LSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/lsan.txt'
  edit MAGICS_GIT_TAG 'develop'
  edit MAGICS_PYTHON_VERSION 'master'
  edit MAGICS_VERSION '4.3.3'
  edit MAKE 'make'
  edit MARS_CLIENT_GIT_TAG '6.27.2.2'
  edit MARS_CLIENT_PACKAGE_GIT_TAG '6.27.6'
  edit MARS_CLIENT_PACKAGE_VERSION '6.27.6'
  edit MARS_CLIENT_VERSION '6.27.2.2'
  edit MEMORY_SANITIZER 'False'
  edit METABUILDER_HOME '/var/tmp/ma0/workspace/metabuilder'
  edit METAB_INSTALL_REMOVE_WRITE_PERMISSIONS 'True'
  edit METAB_PYTHON_VERSION 'python3'
  edit METKIT_GIT_TAG 'develop'
  edit METKIT_VERSION '1.4.6'
  edit METVIEW_GIT_TAG 'develop'
  edit METVIEW_PYTHON_VERSION '1.4.2'
  edit METVIEW_VERSION '5.8.3'
  edit MIR_BUNDLE_VERSION '0.9.0.0'
  edit MIR_GIT_TAG 'develop'
  edit MIR_VERSION '1.5.0'
  edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
  edit MODULE_ACTIONS_FIRST ''
  edit MODULE_ACTIONS_HDR 'module_actions.h'
  edit MODULE_CHECKING_HDR 'module_checking.h'
  edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow"'
  edit MODULE_LOAD_SUPPORTED 'True'
  edit MODULE_NAME 'eccodes_python'
  edit MODULE_VERSION '%GIT_TAG%'
  edit NETCDF4_VERSION '4.7.4'
  edit ODB_API_GIT_TAG 'develop'
  edit ODB_API_VERSION '0.19.4'
  edit ODB_GIT_TAG 'develop'
  edit ODB_TOOLS_GIT_TAG 'develop'
  edit ODB_TOOLS_VERSION '0.18.3'
  edit ODB_VERSION '1.0.11'
  edit ODC_GIT_TAG 'develop'
  edit ODC_VERSION '1.0.3'
  edit OPENJPEG_VERSION '2.3.1'
  edit PACKAGE 'eccodes_python'
  edit PACKAGE_NAME 'eccodes_python'
  edit PACKAGE_OWNERS 'masn@ecmwf.int cgm@ecmwf.int ma0@ecmwf.int'
  edit PACKAGE_VERSION '%GIT_TAG%'
  edit PGEN_VERSION '1.7.1.0'
  edit PKG_BOOST_VERSION '1.71.0'
  edit PREFIX_APPEND ''
  edit PROJ4_VERSION '5.2.0'
  edit PROJ_VERSION '6.3.1'
  edit PYTHON2_VERSION '2.7.15-01'
  edit PYTHON3_VERSION '3.6.10-01'
  edit PYTHON3_VERSION_MAJOR_MINOR '3.6'
  edit PYTHON_GITREPO 'This-needs-tobe-overriden'
  edit PYTHON_PACKAGE 'This-needs-tobe-overriden'
  edit QT_VERSION '5.12.3'
  edit QUEUE 'build'
  edit REMOTE_HOST 'polonius'
  edit REUSE_BUILD_DIR 'False'
  edit SAFE_STACK_SANITIZER 'False'
  edit SANITIZER 'False'
  edit SANITIZER_REFERENCE_DIRECTORY '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
  edit SRCDIR '%BASEDIR%/%PACKAGE%'
  edit STATIC_ANALYSER_ABORT_ON_ERROR 'False'
  edit SWIG_VERSION '3.0.8'
  edit SYNC_RELEASE '2020.04.0'
  edit SYNC_RELEASE_CI_STAGE 'develop'
  edit TARBALL_DIR '/perm/ma/ma0/tmp_tarballs/ci'
  edit TAR_PACKAGE '%PACKAGE%'
  edit TEST 'ctest --output-on-failure -j${CPUS} '
  edit THREAD_SANITIZER 'False'
  edit TRANSI_GIT_TAG 'develop'
  edit TRANSI_VERSION '0.7.1'
  edit TSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
  edit UBSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/ubsan.txt'
  edit UB_SANITIZER 'False'
  edit USER 'ma0'
  edit USE_WORKDIR 'False'
  edit VALGRIND_VERSION '3.15.0'
  edit WORKDIR '%BASEDIR%/workdir/%ECF_NAME%'
  edit WSDIR '/var/tmp/tmpdir/deploy/jtmp.20739'
  label branch "develop"
  task regenerate
    defstatus complete
    edit ECF_JOB_CMD '%ECF_JOB% > %ECF_JOBOUT% 2>&1'
    edit ECF_KILL_CMD 'kill -15 %ECF_RID%'
    edit ECF_PASS 'FREE'
    edit ECF_STATUS_CMD 'ps --sid %ECF_RID% -f'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
    label date ""
    label metabuilder_branch "ci"
  family cca
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DLAPACK_FOUND=TRUE'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/eccodes_python/cca'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/eccodes_python/cca:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cca-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '24000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cca'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'cca'
    edit SITE 'cca'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task add_write_perm
        trigger git_srcs == complete
      task install_python
        trigger add_write_perm == complete
        edit ECF_TRIES '2'
        label version ""
        label install ""
      task test
        trigger install_python == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install_log
        trigger test == complete
        edit ECF_TRIES '2'
      task wipe_src
        trigger install_log == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task add_write_perm
        trigger git_srcs == complete
      task install_python
        trigger add_write_perm == complete
        edit ECF_TRIES '2'
        label version ""
        label install ""
      task test
        trigger install_python == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install_log
        trigger test == complete
        edit ECF_TRIES '2'
      task wipe_src
        trigger install_log == complete
    endfamily
    family intel.17
      edit COMPILER 'INTEL.17'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap intel/17.0.3.053" "module load gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task add_write_perm
        trigger git_srcs == complete
      task install_python
        trigger add_write_perm == complete
        edit ECF_TRIES '2'
        label version ""
        label install ""
      task test
        trigger install_python == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install_log
        trigger test == complete
        edit ECF_TRIES '2'
      task wipe_src
        trigger install_log == complete
    endfamily
    family cray.85
      edit COMPILER 'CRAY.85'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task noop
        defstatus complete
        edit ECF_DUMMY_TASK '1'
    endfamily
    family cray.86
      edit COMPILER 'CRAY.86'
      edit MODULE_ACTIONS '"module swap cdt/17.09" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task noop
        defstatus complete
        edit ECF_DUMMY_TASK '1'
    endfamily
    family cray.87
      edit COMPILER 'CRAY.87'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task add_write_perm
        trigger git_srcs == complete
      task install_python
        trigger add_write_perm == complete
        edit ECF_TRIES '2'
        label version ""
        label install ""
      task test
        trigger install_python == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install_log
        trigger test == complete
        edit ECF_TRIES '2'
      task wipe_src
        trigger install_log == complete
    endfamily
  endfamily
  family ccb
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DLAPACK_FOUND=TRUE'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/eccodes_python/ccb'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/eccodes_python/ccb:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'ccb-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '24000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'ccb'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'ccb'
    edit SITE 'ccb'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task add_write_perm
        trigger git_srcs == complete
      task install_python
        trigger add_write_perm == complete
        edit ECF_TRIES '2'
        label version ""
        label install ""
      task test
        trigger install_python == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install_log
        trigger test == complete
        edit ECF_TRIES '2'
      task wipe_src
        trigger install_log == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task add_write_perm
        trigger git_srcs == complete
      task install_python
        trigger add_write_perm == complete
        edit ECF_TRIES '2'
        label version ""
        label install ""
      task test
        trigger install_python == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install_log
        trigger test == complete
        edit ECF_TRIES '2'
      task wipe_src
        trigger install_log == complete
    endfamily
    family intel.17
      edit COMPILER 'INTEL.17'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap intel/17.0.3.053" "module load gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task add_write_perm
        trigger git_srcs == complete
      task install_python
        trigger add_write_perm == complete
        edit ECF_TRIES '2'
        label version ""
        label install ""
      task test
        trigger install_python == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install_log
        trigger test == complete
        edit ECF_TRIES '2'
      task wipe_src
        trigger install_log == complete
    endfamily
    family cray.85
      edit COMPILER 'CRAY.85'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task noop
        defstatus complete
        edit ECF_DUMMY_TASK '1'
    endfamily
    family cray.86
      edit COMPILER 'CRAY.86'
      edit MODULE_ACTIONS '"module swap cdt/17.09" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task noop
        defstatus complete
        edit ECF_DUMMY_TASK '1'
    endfamily
    family cray.87
      edit COMPILER 'CRAY.87'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task add_write_perm
        trigger git_srcs == complete
      task install_python
        trigger add_write_perm == complete
        edit ECF_TRIES '2'
        label version ""
        label install ""
      task test
        trigger install_python == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install_log
        trigger test == complete
        edit ECF_TRIES '2'
      task wipe_src
        trigger install_log == complete
    endfamily
  endfamily
  family cct
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DLAPACK_FOUND=TRUE'
    edit CPUS '$EC_total_tasks'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/eccodes_python/cct'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/eccodes_python/cct:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cct-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '4'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '6000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cct'
    edit QSUB_LP 'QSUB -lP %EC_TOTAL_TASKS%'
    edit QUEUE 'nf'
    edit REMOTE_HOST 'cct'
    edit SITE 'cct'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task add_write_perm
        trigger git_srcs == complete
      task install_python
        trigger add_write_perm == complete
        edit ECF_TRIES '2'
        label version ""
        label install ""
      task test
        trigger install_python == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install_log
        trigger test == complete
        edit ECF_TRIES '2'
      task wipe_src
        trigger install_log == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task add_write_perm
        trigger git_srcs == complete
      task install_python
        trigger add_write_perm == complete
        edit ECF_TRIES '2'
        label version ""
        label install ""
      task test
        trigger install_python == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install_log
        trigger test == complete
        edit ECF_TRIES '2'
      task wipe_src
        trigger install_log == complete
    endfamily
    family intel.17
      edit COMPILER 'INTEL.17'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap intel/17.0.3.053" "module load gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task add_write_perm
        trigger git_srcs == complete
      task install_python
        trigger add_write_perm == complete
        edit ECF_TRIES '2'
        label version ""
        label install ""
      task test
        trigger install_python == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install_log
        trigger test == complete
        edit ECF_TRIES '2'
      task wipe_src
        trigger install_log == complete
    endfamily
    family cray.85
      edit COMPILER 'CRAY.85'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task noop
        defstatus complete
        edit ECF_DUMMY_TASK '1'
    endfamily
    family cray.86
      edit COMPILER 'CRAY.86'
      edit MODULE_ACTIONS '"module swap cdt/17.09" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task noop
        defstatus complete
        edit ECF_DUMMY_TASK '1'
    endfamily
    family cray.87
      edit COMPILER 'CRAY.87'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task add_write_perm
        trigger git_srcs == complete
      task install_python
        trigger add_write_perm == complete
        edit ECF_TRIES '2'
        label version ""
        label install ""
      task test
        trigger install_python == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install_log
        trigger test == complete
        edit ECF_TRIES '2'
      task wipe_src
        trigger install_log == complete
    endfamily
  endfamily
  family ecgb
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'ecgb'
    edit QUEUE 'express'
    edit REMOTE_HOST 'ecgb'
    edit SITE 'ecgb'
    edit WSDIR '$SCRATCH/metabuilds/%ECF_HOST%_%ECF_PORT%/ecgb/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task add_write_perm
        trigger git_srcs == complete
      task install_python
        trigger add_write_perm == complete
        edit ECF_TRIES '2'
        label version ""
        label install ""
      task test
        trigger install_python == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install_log
        trigger test == complete
        edit ECF_TRIES '2'
      task wipe_src
        trigger install_log == complete
    endfamily
  endfamily
  family leap42
    defstatus suspended
    edit CI_LIMIT '7'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'leap42'
    edit REMOTE_HOST 'leap42-deploy'
    edit SITE 'leap42'
    edit WSDIR '/tmp/metabuilds/%ECF_HOST%_%ECF_PORT%/leap42/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON'
      edit COMPILER 'GNU.73'
      edit GIT_CLONE_OPTIONS ''
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task pull_from_github
        trigger git_srcs == complete
      task add_write_perm
        trigger pull_from_github == complete
      task install_python
        trigger add_write_perm == complete
        edit ECF_TRIES '2'
        label version ""
        label install ""
      task test
        trigger install_python == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task push_to_github
        trigger test == complete
      task push_to_bitbucket
        trigger push_to_github == complete
      task install_log
        trigger push_to_bitbucket == complete
        edit ECF_TRIES '2'
      task wipe_src
        trigger install_log == complete
    endfamily
    family gnu.93
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON'
      edit COMPILER 'GNU.93'
      edit EXPORTS '"export ECACCOUNT=ecodmdma" "export LD_LIBRARY_PATH=/usr/local/apps/gcc/9.3.0/lib64/"'
      edit MODULE_ACTIONS '"module swap gnu/9.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task add_write_perm
        trigger git_srcs == complete
      task install_python
        trigger add_write_perm == complete
        edit ECF_TRIES '2'
        label version ""
        label install ""
      task test
        trigger install_python == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install_log
        trigger test == complete
        edit ECF_TRIES '2'
      task wipe_src
        trigger install_log == complete
    endfamily
    family intel.19
      edit COMPILER 'INTEL.19'
      edit MODULE_ACTIONS '"module unload gnu" "module load intel/19.0.4" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit POST_EXPORTS '"export CXX=icpc\ -cxxlib=/usr/local/apps/gcc/8.3.0" "export LD_RUN_PATH=/usr/local/apps/gcc/8.3.0/lib64" "export PATH=/usr/local/apps/gcc/8.3.0/bin:$PATH"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task add_write_perm
        trigger git_srcs == complete
      task install_python
        trigger add_write_perm == complete
        edit ECF_TRIES '2'
        label version ""
        label install ""
      task test
        trigger install_python == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install_log
        trigger test == complete
        edit ECF_TRIES '2'
      task wipe_src
        trigger install_log == complete
    endfamily
    family clang.90
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON'
      edit COMPILER 'CLANG.90'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task add_write_perm
        trigger git_srcs == complete
      task install_python
        trigger add_write_perm == complete
        edit ECF_TRIES '2'
        label version ""
        label install ""
      task test
        trigger install_python == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install_log
        trigger test == complete
        edit ECF_TRIES '2'
      task wipe_src
        trigger install_log == complete
    endfamily
  endfamily
  family lxc
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'lxc'
    edit QUEUE 'express'
    edit REMOTE_HOST 'lxc'
    edit SITE 'lxc'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxc/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task add_write_perm
        trigger git_srcs == complete
      task install_python
        trigger add_write_perm == complete
        edit ECF_TRIES '2'
        label version ""
        label install ""
      task test
        trigger install_python == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install_log
        trigger test == complete
        edit ECF_TRIES '2'
      task wipe_src
        trigger install_log == complete
    endfamily
  endfamily
  family lxg
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'lxg'
    edit QUEUE 'normal'
    edit REMOTE_HOST 'lxg'
    edit SITE 'lxg'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxg/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task add_write_perm
        trigger git_srcs == complete
      task install_python
        trigger add_write_perm == complete
        edit ECF_TRIES '2'
        label version ""
        label install ""
      task test
        trigger install_python == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install_log
        trigger test == complete
        edit ECF_TRIES '2'
      task wipe_src
        trigger install_log == complete
    endfamily
  endfamily
  family sappb
    defstatus suspended
    edit ASAN_FILE '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
    edit BASEPREFIX '/autofs/vol/deploy/install/%SITE%'
    edit CI_LIMIT '8'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit M_SITE 'sappb'
    edit REMOTE_HOST 'sappb-deploy'
    edit SANITIZER_REFERENCE_DIRECTORY '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
    edit SITE 'sappb'
    edit TSAN_FILE '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
    edit WSDIR '$HOME/metabuilds/%ECF_HOST%_%ECF_PORT%/sappb/%COMPILER:%'
    family gnu.63
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module load gnu/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task add_write_perm
        trigger git_srcs == complete
      task install_python
        trigger add_write_perm == complete
        edit ECF_TRIES '2'
        label version ""
        label install ""
      task test
        trigger install_python == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install_log
        trigger test == complete
        edit ECF_TRIES '2'
      task wipe_src
        trigger install_log == complete
    endfamily
  endfamily
endsuite
suite eccodes_extensions
  edit ADDRESS_SANITIZER 'False'
  edit ADD_ACTION_TRIGGERS 'True'
  edit AEC_VERSION '1.0.4'
  edit ALLOW_INSTALL_TO_OVERWRITE 'True'
  edit ASAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
  edit ATLAS_GIT_TAG 'develop'
  edit ATLAS_VERSION '0.20.2'
  edit BANNER_TASK_END 'banner %TASK%-end || true'
  edit BANNER_TASK_START 'banner %TASK%-start || true'
  edit BASEDIR '%WSDIR%/%PACKAGE%'
  edit BASEPREFIX '$PERM/install/%SITE%'
  edit BUFRDC_EMOSLIB_VERSION '4.6.3'
  edit BUFRDC_TABLES_GIT_TAG 'develop'
  edit BUFRDC_TABLES_VERSION '000415'
  edit BUILDIR '%BASEDIR%/builds'
  edit BUILD_ALWAYS_PARALLEL 'False'
  edit CFI_SANITIZER 'False'
  edit CFL_BASE_URL 'https://software-test.ecmwf.int/wiki'
  edit CFL_SPACE_KEY '<TODO>'
  edit CFL_TARBALL_COMMENT ''
  edit CFL_TITLE 'Releases'
  edit CHECK_SYNC_RELEASE 'False'
  edit CLANG_TIDY_ABORT_ON_ERROR 'False'
  edit CMAKE '%ECBUILD%'
  edit CMAKELIST_EXTRACT_PACKAGE_VERSION '$(awk '/^project/ && /%PACKAGE%/ && /VERSION/ {for (I=1;I<=NF;I++) if ($I == "VERSION") {print $(I+1)};}' CMakeLists.txt)'
  edit CMAKE_BUILD_TYPE 'RelWithDebInfo'
  edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DENABLE_TESTS=OFF'
  edit CMAKE_VERSION '3.13.4'
  edit CODES_UI_GIT_TAG 'develop'
  edit CODES_UI_VERSION '1.3.1'
  edit CPACK 'make package'
  edit CPUS '$(lscpu -p | grep -v '#' | wc -l)'
  edit CRAY_FFTW_VERSION '3.3.8.1'
  edit CRAY_NETCDF_VERSION '4.6.1.3'
  edit CRON_BASEPREFIX '%BASEDIR%/install'
  edit CRON_EMAIL ''
  edit CRON_ENABLED 'False'
  edit CRON_GIT_TAG 'develop'
  edit CUSTOM_COMMANDS ''
  edit DATAFLOW_SANITIZER 'False'
  edit DEFAULT_GNU_COMPILER 'gnu.73'
  edit ECBUILD '%BASEPREFIX%/ecbuild/develop/bin/ecbuild'
  edit ECBUILD_CLONE 'False'
  edit ECBUILD_CLONE_COMMIT 'master'
  edit ECBUILD_GIT_TAG 'develop'
  edit ECBUILD_REPO '%GIT_REPO_PREFIX%/ecsdk/ecbuild.git'
  edit ECBUILD_VERSION '3.3.2'
  edit ECCODES_EXTENSIONS_GIT_TAG 'develop'
  edit ECCODES_EXTENSIONS_VERSION '2.17.0'
  edit ECCODES_GIT_TAG 'develop'
  edit ECCODES_PYTHON_VERSION '0.9.6'
  edit ECCODES_VERSION '2.17.1'
  edit ECFLOW5_LATEST_VERSION '5.4.0'
  edit ECFLOW_VERSION '5.4.1'
  edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/eccodes_extensions'
  edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/eccodes_extensions'
  edit ECF_JOB_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_JOB% %ECF_JOBOUT%'
  edit ECF_KILL_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% kill'
  edit ECF_STATUS_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% stat'
  edit ECF_TRIES '1'
  edit ECKIT_GIT_TAG 'develop'
  edit ECKIT_VERSION '1.10.1'
  edit ECQUIS_VERSION '1.0.0'
  edit EMOS_MOD_VERSION '462'
  edit EMOS_PA_VERSION '000462'
  edit EXPORTS '"export ECACCOUNT=ecodmdma"'
  edit EXPORT_VARIABLES 'False'
  edit FCKIT_GIT_TAG '0.6.3'
  edit FCKIT_VERSION '0.6.3'
  edit FDB_VINTAGE_GIT_TAG '4.10.0'
  edit FDB_VINTAGE_VERSION '4.10.0'
  edit FFTW_VERSION '3.3.8'
  edit FFTW_VERSION_CRAY '3.3.4.11'
  edit GIT 'git'
  edit GITHUB_PACKAGE_NAME '%PACKAGE_NAME%'
  edit GITHUB_REPO 'git@github.com:ecmwf'
  edit GITREPO '%GIT_REPO_PREFIX%/eccodes/eccodes-extensions.git'
  edit GIT_CLONE 'True'
  edit GIT_CLONE_OPTIONS '--depth=1 --branch %GIT_TAG%'
  edit GIT_METABUILDER_TAG 'ci'
  edit GIT_REPO_PREFIX 'ssh://git@git.ecmwf.int:7999'
  edit GIT_TAG 'develop'
  edit GIT_VERSION '2.26.0'
  edit GRIB_API_VERSION '1.28.0'
  edit INSTALLABLE_TO_USR_LOCAL_APPS 'True'
  edit INSTALL_DIR '%BASEPREFIX%/eccodes_extensions/%ECCODES_EXTENSIONS_VERSION%'
  edit IWYU_VERSION '0.11'
  edit KSH_EXTRACT_PACKAGE_VERSION '`grep _VERSION_STR VERSION.cmake | grep -o '[0-9]*\.[0-9]*\.[0-9]*'`'
  edit LAPACK_VERSION '3.5.0'
  edit LIBEMOS_GIT_TAG 'develop'
  edit LIBEMOS_VERSION '4.6.2'
  edit LINTERS_REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}'
  edit LSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/lsan.txt'
  edit MAGICS_GIT_TAG 'develop'
  edit MAGICS_PYTHON_VERSION 'master'
  edit MAGICS_VERSION '4.3.3'
  edit MAKE 'make'
  edit MARS_CLIENT_GIT_TAG '6.27.2.2'
  edit MARS_CLIENT_PACKAGE_GIT_TAG '6.27.6'
  edit MARS_CLIENT_PACKAGE_VERSION '6.27.6'
  edit MARS_CLIENT_VERSION '6.27.2.2'
  edit MEMORY_SANITIZER 'False'
  edit METABUILDER_HOME '/var/tmp/ma0/workspace/metabuilder'
  edit METAB_INSTALL_REMOVE_WRITE_PERMISSIONS 'True'
  edit METAB_PYTHON_VERSION 'python3'
  edit METKIT_GIT_TAG 'develop'
  edit METKIT_VERSION '1.4.6'
  edit METVIEW_GIT_TAG 'develop'
  edit METVIEW_PYTHON_VERSION '1.4.2'
  edit METVIEW_VERSION '5.8.3'
  edit MIR_BUNDLE_VERSION '0.9.0.0'
  edit MIR_GIT_TAG 'develop'
  edit MIR_VERSION '1.5.0'
  edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
  edit MODULE_ACTIONS_FIRST ''
  edit MODULE_ACTIONS_HDR 'module_actions.h'
  edit MODULE_CHECKING_HDR 'module_checking.h'
  edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow"'
  edit MODULE_LOAD_SUPPORTED 'False'
  edit MODULE_NAME 'eccodes_extensions'
  edit MODULE_VERSION '%GIT_TAG%'
  edit NETCDF4_VERSION '4.7.4'
  edit ODB_API_GIT_TAG 'develop'
  edit ODB_API_VERSION '0.19.4'
  edit ODB_GIT_TAG 'develop'
  edit ODB_TOOLS_GIT_TAG 'develop'
  edit ODB_TOOLS_VERSION '0.18.3'
  edit ODB_VERSION '1.0.11'
  edit ODC_GIT_TAG 'develop'
  edit ODC_VERSION '1.0.3'
  edit OPENJPEG_VERSION '2.3.1'
  edit PACKAGE 'eccodes_extensions'
  edit PACKAGE_NAME 'eccodes_extensions'
  edit PACKAGE_OWNERS 'masn@ecmwf.int ma0@ecmwf.int'
  edit PACKAGE_VERSION '%GIT_TAG%'
  edit PGEN_VERSION '1.7.1.0'
  edit PKG_BOOST_VERSION '1.71.0'
  edit PREFIX_APPEND ''
  edit PROJ4_VERSION '5.2.0'
  edit PROJ_VERSION '6.3.1'
  edit PYTHON2_VERSION '2.7.15-01'
  edit PYTHON3_VERSION '3.6.10-01'
  edit PYTHON3_VERSION_MAJOR_MINOR '3.6'
  edit PYTHON_GITREPO 'This-needs-tobe-overriden'
  edit PYTHON_PACKAGE 'This-needs-tobe-overriden'
  edit QT_VERSION '5.12.3'
  edit QUEUE 'build'
  edit REMOTE_HOST 'polonius'
  edit REUSE_BUILD_DIR 'False'
  edit SAFE_STACK_SANITIZER 'False'
  edit SANITIZER 'False'
  edit SANITIZER_REFERENCE_DIRECTORY '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
  edit SRCDIR '%BASEDIR%/%PACKAGE%'
  edit STATIC_ANALYSER_ABORT_ON_ERROR 'False'
  edit SWIG_VERSION '3.0.8'
  edit SYNC_RELEASE '2020.04.0'
  edit SYNC_RELEASE_CI_STAGE 'develop'
  edit TARBALL_DIR '/perm/ma/ma0/tmp_tarballs/ci'
  edit TAR_PACKAGE '%PACKAGE%'
  edit TEST 'ctest --output-on-failure -j${CPUS} '
  edit THREAD_SANITIZER 'False'
  edit TRANSI_GIT_TAG 'develop'
  edit TRANSI_VERSION '0.7.1'
  edit TSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
  edit UBSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/ubsan.txt'
  edit UB_SANITIZER 'False'
  edit USER 'ma0'
  edit USE_WORKDIR 'False'
  edit VALGRIND_VERSION '3.15.0'
  edit WORKDIR '%BASEDIR%/workdir/%ECF_NAME%'
  edit WSDIR '/var/tmp/tmpdir/deploy/jtmp.20739'
  label branch "develop"
  task regenerate
    defstatus complete
    edit ECF_JOB_CMD '%ECF_JOB% > %ECF_JOBOUT% 2>&1'
    edit ECF_KILL_CMD 'kill -15 %ECF_RID%'
    edit ECF_PASS 'FREE'
    edit ECF_STATUS_CMD 'ps --sid %ECF_RID% -f'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
    label date ""
    label metabuilder_branch "ci"
  family cca
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DENABLE_TESTS=OFF -DLAPACK_FOUND=TRUE'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/eccodes_extensions/cca'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/eccodes_extensions/cca:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cca-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '24000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cca'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'cca'
    edit SITE 'cca'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task install
      trigger config == complete
      edit ECF_TRIES '2'
    task install_log
      trigger install == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install_log == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family ccb
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DENABLE_TESTS=OFF -DLAPACK_FOUND=TRUE'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/eccodes_extensions/ccb'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/eccodes_extensions/ccb:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'ccb-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '24000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'ccb'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'ccb'
    edit SITE 'ccb'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task install
      trigger config == complete
      edit ECF_TRIES '2'
    task install_log
      trigger install == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install_log == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family cct
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DENABLE_TESTS=OFF -DLAPACK_FOUND=TRUE'
    edit CPUS '$EC_total_tasks'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/eccodes_extensions/cct'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/eccodes_extensions/cct:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cct-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '4'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '6000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cct'
    edit QSUB_LP 'QSUB -lP %EC_TOTAL_TASKS%'
    edit QUEUE 'nf'
    edit REMOTE_HOST 'cct'
    edit SITE 'cct'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task install
      trigger config == complete
      edit ECF_TRIES '2'
    task install_log
      trigger install == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install_log == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family ecgb
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'ecgb'
    edit QUEUE 'express'
    edit REMOTE_HOST 'ecgb'
    edit SITE 'ecgb'
    edit WSDIR '$SCRATCH/metabuilds/%ECF_HOST%_%ECF_PORT%/ecgb/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task install
      trigger config == complete
      edit ECF_TRIES '2'
    task install_log
      trigger install == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install_log == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family leap42
    defstatus suspended
    edit CI_LIMIT '7'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'leap42'
    edit REMOTE_HOST 'leap42-deploy'
    edit SITE 'leap42'
    edit WSDIR '/tmp/metabuilds/%ECF_HOST%_%ECF_PORT%/leap42/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task install
      trigger config == complete
      edit ECF_TRIES '2'
    task install_log
      trigger install == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install_log == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family lxc
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'lxc'
    edit QUEUE 'express'
    edit REMOTE_HOST 'lxc'
    edit SITE 'lxc'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxc/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task install
      trigger config == complete
      edit ECF_TRIES '2'
    task install_log
      trigger install == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install_log == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family lxg
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'lxg'
    edit QUEUE 'normal'
    edit REMOTE_HOST 'lxg'
    edit SITE 'lxg'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxg/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task install
      trigger config == complete
      edit ECF_TRIES '2'
    task install_log
      trigger install == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install_log == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family sappb
    defstatus suspended
    edit ASAN_FILE '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
    edit BASEPREFIX '/autofs/vol/deploy/install/%SITE%'
    edit CI_LIMIT '8'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit M_SITE 'sappb'
    edit REMOTE_HOST 'sappb-deploy'
    edit SANITIZER_REFERENCE_DIRECTORY '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
    edit SITE 'sappb'
    edit TSAN_FILE '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
    edit WSDIR '$HOME/metabuilds/%ECF_HOST%_%ECF_PORT%/sappb/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task install
      trigger config == complete
      edit ECF_TRIES '2'
    task install_log
      trigger install == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install_log == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
endsuite
suite bufrdc_tables
  edit ADDRESS_SANITIZER 'False'
  edit ADD_ACTION_TRIGGERS 'True'
  edit AEC_VERSION '1.0.4'
  edit ALLOW_INSTALL_TO_OVERWRITE 'True'
  edit ASAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
  edit ATLAS_GIT_TAG 'develop'
  edit ATLAS_VERSION '0.20.2'
  edit BANNER_TASK_END 'banner %TASK%-end || true'
  edit BANNER_TASK_START 'banner %TASK%-start || true'
  edit BASEDIR '%WSDIR%/%PACKAGE%'
  edit BASEPREFIX '$PERM/install/%SITE%'
  edit BUFRDC_EMOSLIB_VERSION '4.6.3'
  edit BUFRDC_TABLES_GIT_TAG 'develop'
  edit BUFRDC_TABLES_VERSION '000415'
  edit BUILDIR '%BASEDIR%/builds'
  edit BUILD_ALWAYS_PARALLEL 'False'
  edit CFI_SANITIZER 'False'
  edit CFL_BASE_URL 'https://software-test.ecmwf.int/wiki'
  edit CFL_SPACE_KEY '<TODO>'
  edit CFL_TARBALL_COMMENT ''
  edit CFL_TITLE 'Releases'
  edit CHECK_SYNC_RELEASE 'False'
  edit CLANG_TIDY_ABORT_ON_ERROR 'False'
  edit CMAKE '%ECBUILD%'
  edit CMAKELIST_EXTRACT_PACKAGE_VERSION '$(awk '/^project/ && /%PACKAGE%/ && /VERSION/ {for (I=1;I<=NF;I++) if ($I == "VERSION") {print $(I+1)};}' CMakeLists.txt)'
  edit CMAKE_BUILD_TYPE 'RelWithDebInfo'
  edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE%'
  edit CMAKE_VERSION '3.13.4'
  edit CODES_UI_GIT_TAG 'develop'
  edit CODES_UI_VERSION '1.3.1'
  edit CPACK 'make package'
  edit CPUS '$(lscpu -p | grep -v '#' | wc -l)'
  edit CRAY_FFTW_VERSION '3.3.8.1'
  edit CRAY_NETCDF_VERSION '4.6.1.3'
  edit CRON_BASEPREFIX '%BASEDIR%/install'
  edit CRON_EMAIL ''
  edit CRON_ENABLED 'False'
  edit CRON_GIT_TAG 'develop'
  edit CUSTOM_COMMANDS ''
  edit DATAFLOW_SANITIZER 'False'
  edit DEFAULT_GNU_COMPILER 'gnu.73'
  edit ECBUILD '%BASEPREFIX%/ecbuild/develop/bin/ecbuild'
  edit ECBUILD_CLONE 'False'
  edit ECBUILD_CLONE_COMMIT 'master'
  edit ECBUILD_GIT_TAG 'develop'
  edit ECBUILD_REPO '%GIT_REPO_PREFIX%/ecsdk/ecbuild.git'
  edit ECBUILD_VERSION '3.3.2'
  edit ECCODES_EXTENSIONS_GIT_TAG 'develop'
  edit ECCODES_EXTENSIONS_VERSION '2.17.0'
  edit ECCODES_GIT_TAG 'develop'
  edit ECCODES_PYTHON_VERSION '0.9.6'
  edit ECCODES_VERSION '2.17.1'
  edit ECFLOW5_LATEST_VERSION '5.4.0'
  edit ECFLOW_VERSION '5.4.1'
  edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/bufrdc_tables'
  edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/bufrdc_tables'
  edit ECF_JOB_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_JOB% %ECF_JOBOUT%'
  edit ECF_KILL_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% kill'
  edit ECF_STATUS_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% stat'
  edit ECF_TRIES '1'
  edit ECKIT_GIT_TAG 'develop'
  edit ECKIT_VERSION '1.10.1'
  edit ECQUIS_VERSION '1.0.0'
  edit EMOS_MOD_VERSION '462'
  edit EMOS_PA_VERSION '000462'
  edit EXPORTS '"export ECACCOUNT=ecodmdma"'
  edit EXPORT_VARIABLES 'False'
  edit FCKIT_GIT_TAG '0.6.3'
  edit FCKIT_VERSION '0.6.3'
  edit FDB_VINTAGE_GIT_TAG '4.10.0'
  edit FDB_VINTAGE_VERSION '4.10.0'
  edit FFTW_VERSION '3.3.8'
  edit FFTW_VERSION_CRAY '3.3.4.11'
  edit GIT 'git'
  edit GITHUB_PACKAGE_NAME '%PACKAGE_NAME%'
  edit GITHUB_REPO 'git@github.com:ecmwf'
  edit GITREPO '%GIT_REPO_PREFIX%/mars/bufrdc_tables.git'
  edit GIT_CLONE 'True'
  edit GIT_CLONE_OPTIONS '--depth=1 --branch %GIT_TAG%'
  edit GIT_METABUILDER_TAG 'ci'
  edit GIT_REPO_PREFIX 'ssh://git@git.ecmwf.int:7999'
  edit GIT_TAG 'develop'
  edit GIT_VERSION '2.26.0'
  edit GRIB_API_VERSION '1.28.0'
  edit INSTALLABLE_TO_USR_LOCAL_APPS 'False'
  edit INSTALL_DIR '%BASEPREFIX%/bufrdc/tables/%BUFRDC_TABLES_VERSION%'
  edit IWYU_VERSION '0.11'
  edit KSH_EXTRACT_PACKAGE_VERSION '`grep _VERSION_STR VERSION.cmake | grep -o '[0-9]*\.[0-9]*\.[0-9]*'`'
  edit LAPACK_VERSION '3.5.0'
  edit LIBEMOS_GIT_TAG 'develop'
  edit LIBEMOS_VERSION '4.6.2'
  edit LINTERS_REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}'
  edit LSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/lsan.txt'
  edit MAGICS_GIT_TAG 'develop'
  edit MAGICS_PYTHON_VERSION 'master'
  edit MAGICS_VERSION '4.3.3'
  edit MAKE 'make'
  edit MARS_CLIENT_GIT_TAG '6.27.2.2'
  edit MARS_CLIENT_PACKAGE_GIT_TAG '6.27.6'
  edit MARS_CLIENT_PACKAGE_VERSION '6.27.6'
  edit MARS_CLIENT_VERSION '6.27.2.2'
  edit MEMORY_SANITIZER 'False'
  edit METABUILDER_HOME '/var/tmp/ma0/workspace/metabuilder'
  edit METAB_INSTALL_REMOVE_WRITE_PERMISSIONS 'True'
  edit METAB_PYTHON_VERSION 'python3'
  edit METKIT_GIT_TAG 'develop'
  edit METKIT_VERSION '1.4.6'
  edit METVIEW_GIT_TAG 'develop'
  edit METVIEW_PYTHON_VERSION '1.4.2'
  edit METVIEW_VERSION '5.8.3'
  edit MIR_BUNDLE_VERSION '0.9.0.0'
  edit MIR_GIT_TAG 'develop'
  edit MIR_VERSION '1.5.0'
  edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
  edit MODULE_ACTIONS_FIRST ''
  edit MODULE_ACTIONS_HDR 'module_actions.h'
  edit MODULE_CHECKING_HDR 'module_checking.h'
  edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow"'
  edit MODULE_LOAD_SUPPORTED 'False'
  edit MODULE_NAME 'bufrdc_tables'
  edit MODULE_VERSION '%GIT_TAG%'
  edit NETCDF4_VERSION '4.7.4'
  edit ODB_API_GIT_TAG 'develop'
  edit ODB_API_VERSION '0.19.4'
  edit ODB_GIT_TAG 'develop'
  edit ODB_TOOLS_GIT_TAG 'develop'
  edit ODB_TOOLS_VERSION '0.18.3'
  edit ODB_VERSION '1.0.11'
  edit ODC_GIT_TAG 'develop'
  edit ODC_VERSION '1.0.3'
  edit OPENJPEG_VERSION '2.3.1'
  edit PACKAGE 'bufrdc_tables'
  edit PACKAGE_NAME 'bufrdc_tables'
  edit PACKAGE_OWNERS 'mapm@ecmwf.int ma0@ecmwf.int'
  edit PACKAGE_VERSION '%GIT_TAG%'
  edit PGEN_VERSION '1.7.1.0'
  edit PKG_BOOST_VERSION '1.71.0'
  edit PREFIX_APPEND ''
  edit PROJ4_VERSION '5.2.0'
  edit PROJ_VERSION '6.3.1'
  edit PYTHON2_VERSION '2.7.15-01'
  edit PYTHON3_VERSION '3.6.10-01'
  edit PYTHON3_VERSION_MAJOR_MINOR '3.6'
  edit PYTHON_GITREPO 'This-needs-tobe-overriden'
  edit PYTHON_PACKAGE 'This-needs-tobe-overriden'
  edit QT_VERSION '5.12.3'
  edit QUEUE 'build'
  edit REMOTE_HOST 'polonius'
  edit REUSE_BUILD_DIR 'False'
  edit SAFE_STACK_SANITIZER 'False'
  edit SANITIZER 'False'
  edit SANITIZER_REFERENCE_DIRECTORY '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
  edit SRCDIR '%BASEDIR%/%PACKAGE%'
  edit STATIC_ANALYSER_ABORT_ON_ERROR 'False'
  edit SWIG_VERSION '3.0.8'
  edit SYNC_RELEASE '2020.04.0'
  edit SYNC_RELEASE_CI_STAGE 'develop'
  edit TARBALL_DIR '/perm/ma/ma0/tmp_tarballs/ci'
  edit TAR_PACKAGE '%PACKAGE%'
  edit TEST 'ctest --output-on-failure -j${CPUS} '
  edit THREAD_SANITIZER 'False'
  edit TRANSI_GIT_TAG 'develop'
  edit TRANSI_VERSION '0.7.1'
  edit TSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
  edit UBSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/ubsan.txt'
  edit UB_SANITIZER 'False'
  edit USER 'ma0'
  edit USE_WORKDIR 'False'
  edit VALGRIND_VERSION '3.15.0'
  edit WORKDIR '%BASEDIR%/workdir/%ECF_NAME%'
  edit WSDIR '/var/tmp/tmpdir/deploy/jtmp.20739'
  label branch "develop"
  task regenerate
    defstatus complete
    edit ECF_JOB_CMD '%ECF_JOB% > %ECF_JOBOUT% 2>&1'
    edit ECF_KILL_CMD 'kill -15 %ECF_RID%'
    edit ECF_PASS 'FREE'
    edit ECF_STATUS_CMD 'ps --sid %ECF_RID% -f'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
    label date ""
    label metabuilder_branch "ci"
  family cca
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DLAPACK_FOUND=TRUE'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/bufrdc_tables/cca'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/bufrdc_tables/cca:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cca-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '24000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cca'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'cca'
    edit SITE 'cca'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task install
      trigger config == complete
      edit ECF_TRIES '2'
    task install_log
      trigger install == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install_log == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family ccb
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DLAPACK_FOUND=TRUE'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/bufrdc_tables/ccb'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/bufrdc_tables/ccb:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'ccb-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '24000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'ccb'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'ccb'
    edit SITE 'ccb'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task install
      trigger config == complete
      edit ECF_TRIES '2'
    task install_log
      trigger install == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install_log == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family cct
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DLAPACK_FOUND=TRUE'
    edit CPUS '$EC_total_tasks'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/bufrdc_tables/cct'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/bufrdc_tables/cct:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cct-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '4'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '6000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cct'
    edit QSUB_LP 'QSUB -lP %EC_TOTAL_TASKS%'
    edit QUEUE 'nf'
    edit REMOTE_HOST 'cct'
    edit SITE 'cct'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task install
      trigger config == complete
      edit ECF_TRIES '2'
    task install_log
      trigger install == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install_log == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family ecgb
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'ecgb'
    edit QUEUE 'express'
    edit REMOTE_HOST 'ecgb'
    edit SITE 'ecgb'
    edit WSDIR '$SCRATCH/metabuilds/%ECF_HOST%_%ECF_PORT%/ecgb/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task install
      trigger config == complete
      edit ECF_TRIES '2'
    task install_log
      trigger install == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install_log == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family leap42
    defstatus suspended
    edit CI_LIMIT '7'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'leap42'
    edit REMOTE_HOST 'leap42-deploy'
    edit SITE 'leap42'
    edit WSDIR '/tmp/metabuilds/%ECF_HOST%_%ECF_PORT%/leap42/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task install
      trigger config == complete
      edit ECF_TRIES '2'
    task install_log
      trigger install == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install_log == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family lxc
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'lxc'
    edit QUEUE 'express'
    edit REMOTE_HOST 'lxc'
    edit SITE 'lxc'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxc/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task install
      trigger config == complete
      edit ECF_TRIES '2'
    task install_log
      trigger install == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install_log == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family lxg
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'lxg'
    edit QUEUE 'normal'
    edit REMOTE_HOST 'lxg'
    edit SITE 'lxg'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxg/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task install
      trigger config == complete
      edit ECF_TRIES '2'
    task install_log
      trigger install == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install_log == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family sappb
    defstatus suspended
    edit ASAN_FILE '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
    edit BASEPREFIX '/autofs/vol/deploy/install/%SITE%'
    edit CI_LIMIT '8'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit M_SITE 'sappb'
    edit REMOTE_HOST 'sappb-deploy'
    edit SANITIZER_REFERENCE_DIRECTORY '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
    edit SITE 'sappb'
    edit TSAN_FILE '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
    edit WSDIR '$HOME/metabuilds/%ECF_HOST%_%ECF_PORT%/sappb/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task install
      trigger config == complete
      edit ECF_TRIES '2'
    task install_log
      trigger install == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install_log == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
endsuite
suite bufrdc_emoslib
  edit ADDRESS_SANITIZER 'False'
  edit ADD_ACTION_TRIGGERS 'True'
  edit AEC_VERSION '1.0.4'
  edit ALLOW_INSTALL_TO_OVERWRITE 'True'
  edit ASAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
  edit ATLAS_GIT_TAG 'develop'
  edit ATLAS_VERSION '0.20.2'
  edit BANNER_TASK_END 'banner %TASK%-end || true'
  edit BANNER_TASK_START 'banner %TASK%-start || true'
  edit BASEDIR '%WSDIR%/%PACKAGE%'
  edit BASEPREFIX '$PERM/install/%SITE%'
  edit BUFRDC_EMOSLIB_VERSION '4.6.3'
  edit BUFRDC_TABLES_GIT_TAG 'develop'
  edit BUFRDC_TABLES_VERSION '000415'
  edit BUFR_TABLES '%BASEPREFIX%/bufrdc/tables/%BUFRDC_TABLES_VERSION%'
  edit BUILDIR '%BASEDIR%/builds'
  edit BUILD_ALWAYS_PARALLEL 'False'
  edit CFI_SANITIZER 'False'
  edit CFL_BASE_URL 'https://software-test.ecmwf.int/wiki'
  edit CFL_SPACE_KEY '<TODO>'
  edit CFL_TARBALL_COMMENT ''
  edit CFL_TITLE 'Releases'
  edit CHECK_SYNC_RELEASE 'False'
  edit CLANG_TIDY_ABORT_ON_ERROR 'False'
  edit CMAKE '%ECBUILD%'
  edit CMAKELIST_EXTRACT_PACKAGE_VERSION '$(awk '/^project/ && /%PACKAGE%/ && /VERSION/ {for (I=1;I<=NF;I++) if ($I == "VERSION") {print $(I+1)};}' CMakeLists.txt)'
  edit CMAKE_BUILD_TYPE 'RelWithDebInfo'
  edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE%'
  edit CMAKE_VERSION '3.13.4'
  edit CODES_UI_GIT_TAG 'develop'
  edit CODES_UI_VERSION '1.3.1'
  edit CPACK 'make package'
  edit CPUS '$(lscpu -p | grep -v '#' | wc -l)'
  edit CRAY_FFTW_VERSION '3.3.8.1'
  edit CRAY_NETCDF_VERSION '4.6.1.3'
  edit CRON_BASEPREFIX '%BASEDIR%/install'
  edit CRON_EMAIL ''
  edit CRON_ENABLED 'False'
  edit CRON_GIT_TAG 'develop'
  edit CUSTOM_COMMANDS ''
  edit DATAFLOW_SANITIZER 'False'
  edit DEFAULT_GNU_COMPILER 'gnu.73'
  edit ECBUILD '%BASEPREFIX%/ecbuild/develop/bin/ecbuild'
  edit ECBUILD_CLONE 'False'
  edit ECBUILD_CLONE_COMMIT 'master'
  edit ECBUILD_GIT_TAG 'develop'
  edit ECBUILD_REPO '%GIT_REPO_PREFIX%/ecsdk/ecbuild.git'
  edit ECBUILD_VERSION '3.3.2'
  edit ECCODES_EXTENSIONS_GIT_TAG 'develop'
  edit ECCODES_EXTENSIONS_VERSION '2.17.0'
  edit ECCODES_GIT_TAG 'develop'
  edit ECCODES_PYTHON_VERSION '0.9.6'
  edit ECCODES_VERSION '2.17.1'
  edit ECFLOW5_LATEST_VERSION '5.4.0'
  edit ECFLOW_VERSION '5.4.1'
  edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/bufrdc_emoslib'
  edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/bufrdc_emoslib'
  edit ECF_JOB_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_JOB% %ECF_JOBOUT%'
  edit ECF_KILL_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% kill'
  edit ECF_STATUS_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% stat'
  edit ECF_TRIES '1'
  edit ECKIT_GIT_TAG 'develop'
  edit ECKIT_VERSION '1.10.1'
  edit ECQUIS_VERSION '1.0.0'
  edit EMOS_MOD_VERSION '462'
  edit EMOS_PA_VERSION '000462'
  edit EXPORTS '"export ECACCOUNT=ecodmdma" "export ftp_proxy=http://proxy.ecmwf.int:3333" "export http_proxy=http://proxy.ecmwf.int:3333" "export https_proxy=http://proxy.ecmwf.int:3333"'
  edit EXPORT_VARIABLES 'False'
  edit FCKIT_GIT_TAG '0.6.3'
  edit FCKIT_VERSION '0.6.3'
  edit FDB_VINTAGE_GIT_TAG '4.10.0'
  edit FDB_VINTAGE_VERSION '4.10.0'
  edit FFTW_VERSION '3.3.8'
  edit FFTW_VERSION_CRAY '3.3.4.11'
  edit GIT 'git'
  edit GITHUB_PACKAGE_NAME '%PACKAGE_NAME%'
  edit GITHUB_REPO 'git@github.com:ecmwf'
  edit GITREPO '%GIT_REPO_PREFIX%/mars/bufrdc_emoslib.git'
  edit GIT_CLONE 'True'
  edit GIT_CLONE_OPTIONS '--depth=1 --branch %GIT_TAG%'
  edit GIT_METABUILDER_TAG 'ci'
  edit GIT_REPO_PREFIX 'ssh://git@git.ecmwf.int:7999'
  edit GIT_TAG 'develop'
  edit GIT_VERSION '2.26.0'
  edit GRIB_API_VERSION '1.28.0'
  edit INSTALLABLE_TO_USR_LOCAL_APPS 'True'
  edit IWYU_VERSION '0.11'
  edit KSH_EXTRACT_PACKAGE_VERSION '`grep _VERSION_STR VERSION.cmake | grep -o '[0-9]*\.[0-9]*\.[0-9]*'`'
  edit LAPACK_VERSION '3.5.0'
  edit LIBEMOS_GIT_TAG 'develop'
  edit LIBEMOS_VERSION '4.6.2'
  edit LINTERS_REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}'
  edit LSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/lsan.txt'
  edit MAGICS_GIT_TAG 'develop'
  edit MAGICS_PYTHON_VERSION 'master'
  edit MAGICS_VERSION '4.3.3'
  edit MAKE 'make'
  edit MARS_CLIENT_GIT_TAG '6.27.2.2'
  edit MARS_CLIENT_PACKAGE_GIT_TAG '6.27.6'
  edit MARS_CLIENT_PACKAGE_VERSION '6.27.6'
  edit MARS_CLIENT_VERSION '6.27.2.2'
  edit MEMORY_SANITIZER 'False'
  edit METABUILDER_HOME '/var/tmp/ma0/workspace/metabuilder'
  edit METAB_INSTALL_REMOVE_WRITE_PERMISSIONS 'True'
  edit METAB_PYTHON_VERSION 'python3'
  edit METKIT_GIT_TAG 'develop'
  edit METKIT_VERSION '1.4.6'
  edit METVIEW_GIT_TAG 'develop'
  edit METVIEW_PYTHON_VERSION '1.4.2'
  edit METVIEW_VERSION '5.8.3'
  edit MIR_BUNDLE_VERSION '0.9.0.0'
  edit MIR_GIT_TAG 'develop'
  edit MIR_VERSION '1.5.0'
  edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
  edit MODULE_ACTIONS_FIRST ''
  edit MODULE_ACTIONS_HDR 'module_actions.h'
  edit MODULE_CHECKING_HDR 'module_checking.h'
  edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow"'
  edit MODULE_LOAD_SUPPORTED 'True'
  edit MODULE_NAME 'bufrdc_emoslib'
  edit MODULE_VERSION '%GIT_TAG%'
  edit NETCDF4_VERSION '4.7.4'
  edit ODB_API_GIT_TAG 'develop'
  edit ODB_API_VERSION '0.19.4'
  edit ODB_GIT_TAG 'develop'
  edit ODB_TOOLS_GIT_TAG 'develop'
  edit ODB_TOOLS_VERSION '0.18.3'
  edit ODB_VERSION '1.0.11'
  edit ODC_GIT_TAG 'develop'
  edit ODC_VERSION '1.0.3'
  edit OPENJPEG_VERSION '2.3.1'
  edit PACKAGE 'bufrdc'
  edit PACKAGE_NAME 'bufrdc_emoslib'
  edit PACKAGE_OWNERS 'mapm@ecmwf.int ma0@ecmwf.int'
  edit PACKAGE_VERSION '%GIT_TAG%'
  edit PGEN_VERSION '1.7.1.0'
  edit PKG_BOOST_VERSION '1.71.0'
  edit PREFIX_APPEND ''
  edit PROJ4_VERSION '5.2.0'
  edit PROJ_VERSION '6.3.1'
  edit PYTHON2_VERSION '2.7.15-01'
  edit PYTHON3_VERSION '3.6.10-01'
  edit PYTHON3_VERSION_MAJOR_MINOR '3.6'
  edit PYTHON_GITREPO 'This-needs-tobe-overriden'
  edit PYTHON_PACKAGE 'This-needs-tobe-overriden'
  edit QT_VERSION '5.12.3'
  edit QUEUE 'build'
  edit REMOTE_HOST 'polonius'
  edit REUSE_BUILD_DIR 'False'
  edit SAFE_STACK_SANITIZER 'False'
  edit SANITIZER 'False'
  edit SANITIZER_REFERENCE_DIRECTORY '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
  edit SRCDIR '%BASEDIR%/%PACKAGE%'
  edit STATIC_ANALYSER_ABORT_ON_ERROR 'False'
  edit SWIG_VERSION '3.0.8'
  edit SYNC_RELEASE '2020.04.0'
  edit SYNC_RELEASE_CI_STAGE 'develop'
  edit TARBALL_DIR '/perm/ma/ma0/tmp_tarballs/ci'
  edit TAR_PACKAGE '%PACKAGE%'
  edit TEST 'ctest --output-on-failure -j${CPUS} '
  edit THREAD_SANITIZER 'False'
  edit TRANSI_GIT_TAG 'develop'
  edit TRANSI_VERSION '0.7.1'
  edit TSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
  edit UBSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/ubsan.txt'
  edit UB_SANITIZER 'False'
  edit USER 'ma0'
  edit USE_WORKDIR 'False'
  edit VALGRIND_VERSION '3.15.0'
  edit WORKDIR '%BASEDIR%/workdir/%ECF_NAME%'
  edit WSDIR '/var/tmp/tmpdir/deploy/jtmp.20739'
  label branch "develop"
  task regenerate
    defstatus complete
    edit ECF_JOB_CMD '%ECF_JOB% > %ECF_JOBOUT% 2>&1'
    edit ECF_KILL_CMD 'kill -15 %ECF_RID%'
    edit ECF_PASS 'FREE'
    edit ECF_STATUS_CMD 'ps --sid %ECF_RID% -f'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
    label date ""
    label metabuilder_branch "ci"
  family cca
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DLAPACK_FOUND=TRUE'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/bufrdc_emoslib/cca'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/bufrdc_emoslib/cca:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cca-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export FC=ftn" "export TMPDIR=/tmp" "export ftp_proxy=http://proxy.ecmwf.int:3333" "export http_proxy=http://proxy.ecmwf.int:3333" "export https_proxy=http://proxy.ecmwf.int:3333"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '24000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cca'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'cca'
    edit SITE 'cca'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.17
      edit COMPILER 'INTEL.17'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap intel/17.0.3.053" "module load gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.85
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.85'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.86
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.86'
      edit MODULE_ACTIONS '"module swap cdt/17.09" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.87'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family ccb
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DLAPACK_FOUND=TRUE'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/bufrdc_emoslib/ccb'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/bufrdc_emoslib/ccb:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'ccb-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export FC=ftn" "export TMPDIR=/tmp" "export ftp_proxy=http://proxy.ecmwf.int:3333" "export http_proxy=http://proxy.ecmwf.int:3333" "export https_proxy=http://proxy.ecmwf.int:3333"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '24000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'ccb'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'ccb'
    edit SITE 'ccb'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.17
      edit COMPILER 'INTEL.17'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap intel/17.0.3.053" "module load gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.85
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.85'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.86
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.86'
      edit MODULE_ACTIONS '"module swap cdt/17.09" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.87'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family cct
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DLAPACK_FOUND=TRUE'
    edit CPUS '$EC_total_tasks'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/bufrdc_emoslib/cct'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/bufrdc_emoslib/cct:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cct-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '4'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export FC=ftn" "export TMPDIR=/tmp" "export ftp_proxy=http://proxy.ecmwf.int:3333" "export http_proxy=http://proxy.ecmwf.int:3333" "export https_proxy=http://proxy.ecmwf.int:3333"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '6000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cct'
    edit QSUB_LP 'QSUB -lP %EC_TOTAL_TASKS%'
    edit QUEUE 'nf'
    edit REMOTE_HOST 'cct'
    edit SITE 'cct'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.17
      edit COMPILER 'INTEL.17'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap intel/17.0.3.053" "module load gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.85
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.85'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.86
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.86'
      edit MODULE_ACTIONS '"module swap cdt/17.09" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.87'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family ecgb
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'ecgb'
    edit QUEUE 'express'
    edit REMOTE_HOST 'ecgb'
    edit SITE 'ecgb'
    edit WSDIR '$SCRATCH/metabuilds/%ECF_HOST%_%ECF_PORT%/ecgb/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family leap42
    defstatus suspended
    edit CI_LIMIT '7'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'leap42'
    edit REMOTE_HOST 'leap42-deploy'
    edit SITE 'leap42'
    edit WSDIR '/tmp/metabuilds/%ECF_HOST%_%ECF_PORT%/leap42/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.93
      edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON'
      edit COMPILER 'GNU.93'
      edit EXPORTS '"export ECACCOUNT=ecodmdma" "export LD_LIBRARY_PATH=/usr/local/apps/gcc/9.3.0/lib64/" "export ftp_proxy=http://proxy.ecmwf.int:3333" "export http_proxy=http://proxy.ecmwf.int:3333" "export https_proxy=http://proxy.ecmwf.int:3333"'
      edit MODULE_ACTIONS '"module swap gnu/9.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.19
      edit COMPILER 'INTEL.19'
      edit MODULE_ACTIONS '"module unload gnu" "module load intel/19.0.4" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit POST_EXPORTS '"export CXX=icpc\ -cxxlib=/usr/local/apps/gcc/8.3.0" "export LD_RUN_PATH=/usr/local/apps/gcc/8.3.0/lib64" "export PATH=/usr/local/apps/gcc/8.3.0/bin:$PATH"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90
      edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON'
      edit COMPILER 'CLANG.90'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family lxc
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'lxc'
    edit QUEUE 'express'
    edit REMOTE_HOST 'lxc'
    edit SITE 'lxc'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxc/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family lxg
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'lxg'
    edit QUEUE 'normal'
    edit REMOTE_HOST 'lxg'
    edit SITE 'lxg'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxg/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family sappb
    defstatus suspended
    edit ASAN_FILE '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
    edit BASEPREFIX '/autofs/vol/deploy/install/%SITE%'
    edit CI_LIMIT '8'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit M_SITE 'sappb'
    edit REMOTE_HOST 'sappb-deploy'
    edit SANITIZER_REFERENCE_DIRECTORY '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
    edit SITE 'sappb'
    edit TSAN_FILE '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
    edit WSDIR '$HOME/metabuilds/%ECF_HOST%_%ECF_PORT%/sappb/%COMPILER:%'
    family gnu.63
      edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module load gnu/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
endsuite
suite libemos
  edit ADDRESS_SANITIZER 'False'
  edit ADD_ACTION_TRIGGERS 'True'
  edit AEC_VERSION '1.0.4'
  edit ALLOW_INSTALL_TO_OVERWRITE 'True'
  edit ASAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
  edit ATLAS_GIT_TAG 'develop'
  edit ATLAS_VERSION '0.20.2'
  edit BANNER_TASK_END 'banner %TASK%-end || true'
  edit BANNER_TASK_START 'banner %TASK%-start || true'
  edit BASEDIR '%WSDIR%/%PACKAGE%'
  edit BASEPREFIX '$PERM/install/%SITE%'
  edit BUFRDC_EMOSLIB_VERSION '4.6.3'
  edit BUFRDC_TABLES_GIT_TAG 'develop'
  edit BUFRDC_TABLES_VERSION '000415'
  edit BUFR_TABLES '%BASEPREFIX%/bufrdc/tables/%BUFRDC_TABLES_VERSION%'
  edit BUILDIR '%BASEDIR%/builds'
  edit BUILD_ALWAYS_PARALLEL 'False'
  edit CFI_SANITIZER 'False'
  edit CFL_BASE_URL 'https://software-test.ecmwf.int/wiki'
  edit CFL_SPACE_KEY '<TODO>'
  edit CFL_TARBALL_COMMENT ''
  edit CFL_TITLE 'Releases'
  edit CHECK_SYNC_RELEASE 'False'
  edit CLANG_TIDY_ABORT_ON_ERROR 'False'
  edit CMAKE '%ECBUILD%'
  edit CMAKELIST_EXTRACT_PACKAGE_VERSION '$(awk '/^project/ && /%PACKAGE%/ && /VERSION/ {for (I=1;I<=NF;I++) if ($I == "VERSION") {print $(I+1)};}' CMakeLists.txt)'
  edit CMAKE_BUILD_TYPE 'RelWithDebInfo'
  edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DECBUILD_LOG_LEVEL=DEBUG -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DENABLE_FFTW=ON -DENABLE_LIBEMOS_TESTS_REGRESS=ON'
  edit CMAKE_VERSION '3.13.4'
  edit CODES_UI_GIT_TAG 'develop'
  edit CODES_UI_VERSION '1.3.1'
  edit CPACK 'make package'
  edit CPUS '$(lscpu -p | grep -v '#' | wc -l)'
  edit CRAY_FFTW_VERSION '3.3.8.1'
  edit CRAY_NETCDF_VERSION '4.6.1.3'
  edit CRON_BASEPREFIX '%BASEDIR%/install'
  edit CRON_EMAIL ''
  edit CRON_ENABLED 'False'
  edit CRON_GIT_TAG 'develop'
  edit CUSTOM_COMMANDS ''
  edit DATAFLOW_SANITIZER 'False'
  edit DEFAULT_GNU_COMPILER 'gnu.73'
  edit ECBUILD '%BASEPREFIX%/ecbuild/develop/bin/ecbuild'
  edit ECBUILD_CLONE 'False'
  edit ECBUILD_CLONE_COMMIT 'master'
  edit ECBUILD_GIT_TAG 'develop'
  edit ECBUILD_REPO '%GIT_REPO_PREFIX%/ecsdk/ecbuild.git'
  edit ECBUILD_VERSION '3.3.2'
  edit ECCODES_EXTENSIONS_GIT_TAG 'develop'
  edit ECCODES_EXTENSIONS_VERSION '2.17.0'
  edit ECCODES_GIT_TAG 'develop'
  edit ECCODES_PYTHON_VERSION '0.9.6'
  edit ECCODES_VERSION '2.17.1'
  edit ECFLOW5_LATEST_VERSION '5.4.0'
  edit ECFLOW_VERSION '5.4.1'
  edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/libemos'
  edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/libemos'
  edit ECF_JOB_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_JOB% %ECF_JOBOUT%'
  edit ECF_KILL_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% kill'
  edit ECF_STATUS_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% stat'
  edit ECF_TRIES '1'
  edit ECKIT_GIT_TAG 'develop'
  edit ECKIT_VERSION '1.10.1'
  edit ECQUIS_VERSION '1.0.0'
  edit EMOS_MOD_VERSION '462'
  edit EMOS_PA_VERSION '000462'
  edit EXPORTS '"export ECACCOUNT=ecodmdma" "export ftp_proxy=http://proxy.ecmwf.int:3333" "export http_proxy=http://proxy.ecmwf.int:3333" "export https_proxy=http://proxy.ecmwf.int:3333"'
  edit EXPORT_VARIABLES 'False'
  edit FCKIT_GIT_TAG '0.6.3'
  edit FCKIT_VERSION '0.6.3'
  edit FDB_VINTAGE_GIT_TAG '4.10.0'
  edit FDB_VINTAGE_VERSION '4.10.0'
  edit FFTW_VERSION '3.3.8'
  edit FFTW_VERSION_CRAY '3.3.4.11'
  edit GIT 'git'
  edit GITHUB_PACKAGE_NAME '%PACKAGE_NAME%'
  edit GITHUB_REPO 'git@github.com:ecmwf'
  edit GITREPO '%GIT_REPO_PREFIX%/mars/libemos.git'
  edit GIT_CLONE 'True'
  edit GIT_CLONE_OPTIONS '--depth=1 --branch %GIT_TAG%'
  edit GIT_METABUILDER_TAG 'ci'
  edit GIT_REPO_PREFIX 'ssh://git@git.ecmwf.int:7999'
  edit GIT_TAG 'develop'
  edit GIT_VERSION '2.26.0'
  edit GRIB_API_VERSION '1.28.0'
  edit INSTALLABLE_TO_USR_LOCAL_APPS 'True'
  edit IWYU_VERSION '0.11'
  edit KSH_EXTRACT_PACKAGE_VERSION '`grep _VERSION_STR VERSION.cmake | grep -o '[0-9]*\.[0-9]*\.[0-9]*'`'
  edit LAPACK_VERSION '3.5.0'
  edit LIBEMOS_GIT_TAG 'develop'
  edit LIBEMOS_VERSION '4.6.2'
  edit LINTERS_REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}'
  edit LSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/lsan.txt'
  edit MAGICS_GIT_TAG 'develop'
  edit MAGICS_PYTHON_VERSION 'master'
  edit MAGICS_VERSION '4.3.3'
  edit MAKE 'make'
  edit MARS_CLIENT_GIT_TAG '6.27.2.2'
  edit MARS_CLIENT_PACKAGE_GIT_TAG '6.27.6'
  edit MARS_CLIENT_PACKAGE_VERSION '6.27.6'
  edit MARS_CLIENT_VERSION '6.27.2.2'
  edit MARS_SCRIPT 'mars'
  edit MEMORY_SANITIZER 'False'
  edit METABUILDER_HOME '/var/tmp/ma0/workspace/metabuilder'
  edit METAB_INSTALL_REMOVE_WRITE_PERMISSIONS 'True'
  edit METAB_PYTHON_VERSION 'python3'
  edit METKIT_GIT_TAG 'develop'
  edit METKIT_VERSION '1.4.6'
  edit METVIEW_GIT_TAG 'develop'
  edit METVIEW_PYTHON_VERSION '1.4.2'
  edit METVIEW_VERSION '5.8.3'
  edit MIR_BUNDLE_VERSION '0.9.0.0'
  edit MIR_GIT_TAG 'develop'
  edit MIR_VERSION '1.5.0'
  edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
  edit MODULE_ACTIONS_FIRST ''
  edit MODULE_ACTIONS_HDR 'module_actions.h'
  edit MODULE_CHECKING_HDR 'module_checking.h'
  edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow"'
  edit MODULE_LOAD_SUPPORTED 'True'
  edit MODULE_NAME 'libemos'
  edit MODULE_VERSION '%EMOS_MOD_VERSION%'
  edit NETCDF4_VERSION '4.7.4'
  edit ODB_API_GIT_TAG 'develop'
  edit ODB_API_VERSION '0.19.4'
  edit ODB_GIT_TAG 'develop'
  edit ODB_TOOLS_GIT_TAG 'develop'
  edit ODB_TOOLS_VERSION '0.18.3'
  edit ODB_VERSION '1.0.11'
  edit ODC_GIT_TAG 'develop'
  edit ODC_VERSION '1.0.3'
  edit OPENJPEG_VERSION '2.3.1'
  edit PACKAGE 'libemos'
  edit PACKAGE_NAME 'libemos'
  edit PACKAGE_OWNERS 'mapm@ecmwf.int ma0@ecmwf.int'
  edit PACKAGE_VERSION '%EMOS_PA_VERSION%'
  edit PGEN_VERSION '1.7.1.0'
  edit PKG_BOOST_VERSION '1.71.0'
  edit PREFIX_APPEND ''
  edit PROJ4_VERSION '5.2.0'
  edit PROJ_VERSION '6.3.1'
  edit PYTHON2_VERSION '2.7.15-01'
  edit PYTHON3_VERSION '3.6.10-01'
  edit PYTHON3_VERSION_MAJOR_MINOR '3.6'
  edit PYTHON_GITREPO 'This-needs-tobe-overriden'
  edit PYTHON_PACKAGE 'This-needs-tobe-overriden'
  edit QT_VERSION '5.12.3'
  edit QUEUE 'build'
  edit REMOTE_HOST 'polonius'
  edit REUSE_BUILD_DIR 'False'
  edit SAFE_STACK_SANITIZER 'False'
  edit SANITIZER 'False'
  edit SANITIZER_REFERENCE_DIRECTORY '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
  edit SRCDIR '%BASEDIR%/%PACKAGE%'
  edit STATIC_ANALYSER_ABORT_ON_ERROR 'False'
  edit SWIG_VERSION '3.0.8'
  edit SYNC_RELEASE '2020.04.0'
  edit SYNC_RELEASE_CI_STAGE 'develop'
  edit TARBALL_DIR '/perm/ma/ma0/tmp_tarballs/ci'
  edit TAR_PACKAGE '%PACKAGE%'
  edit TEST 'ctest --output-on-failure -j${CPUS} '
  edit THREAD_SANITIZER 'False'
  edit TRANSI_GIT_TAG 'develop'
  edit TRANSI_VERSION '0.7.1'
  edit TSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
  edit UBSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/ubsan.txt'
  edit UB_SANITIZER 'False'
  edit USER 'ma0'
  edit USE_WORKDIR 'False'
  edit VALGRIND_VERSION '3.15.0'
  edit WORKDIR '%BASEDIR%/workdir/%ECF_NAME%'
  edit WSDIR '/var/tmp/tmpdir/deploy/jtmp.20739'
  label branch "develop"
  task regenerate
    defstatus complete
    edit ECF_JOB_CMD '%ECF_JOB% > %ECF_JOBOUT% 2>&1'
    edit ECF_KILL_CMD 'kill -15 %ECF_RID%'
    edit ECF_PASS 'FREE'
    edit ECF_STATUS_CMD 'ps --sid %ECF_RID% -f'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
    label date ""
    label metabuilder_branch "ci"
  family cca
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECBUILD_LOG_LEVEL=DEBUG -DECBUILD_TRUST_FLAGS=ON -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DENABLE_FFTW=ON -DENABLE_LIBEMOS_TESTS_REGRESS=ON -DLAPACK_FOUND=TRUE'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/libemos/cca'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/libemos/cca:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cca-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export FC=ftn" "export TMPDIR=/tmp" "export ftp_proxy=http://proxy.ecmwf.int:3333" "export http_proxy=http://proxy.ecmwf.int:3333" "export https_proxy=http://proxy.ecmwf.int:3333"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '21000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cca'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'cca'
    edit SITE 'cca'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DECBUILD_LOG_LEVEL=DEBUG -DECBUILD_TRUST_FLAGS=ON -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DENABLE_FFTW=ON -DENABLE_LIBEMOS_TESTS_REGRESS=ON -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DECBUILD_LOG_LEVEL=DEBUG -DECBUILD_TRUST_FLAGS=ON -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DENABLE_FFTW=ON -DENABLE_LIBEMOS_TESTS_REGRESS=ON -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.17
      edit COMPILER 'INTEL.17'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap intel/17.0.3.053" "module load gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.85
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.85'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.86
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.86'
      edit MODULE_ACTIONS '"module swap cdt/17.09" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.87'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family ccb
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECBUILD_LOG_LEVEL=DEBUG -DECBUILD_TRUST_FLAGS=ON -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DENABLE_FFTW=ON -DENABLE_LIBEMOS_TESTS_REGRESS=ON -DLAPACK_FOUND=TRUE'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/libemos/ccb'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/libemos/ccb:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'ccb-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export FC=ftn" "export TMPDIR=/tmp" "export ftp_proxy=http://proxy.ecmwf.int:3333" "export http_proxy=http://proxy.ecmwf.int:3333" "export https_proxy=http://proxy.ecmwf.int:3333"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '21000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'ccb'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'ccb'
    edit SITE 'ccb'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DECBUILD_LOG_LEVEL=DEBUG -DECBUILD_TRUST_FLAGS=ON -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DENABLE_FFTW=ON -DENABLE_LIBEMOS_TESTS_REGRESS=ON -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DECBUILD_LOG_LEVEL=DEBUG -DECBUILD_TRUST_FLAGS=ON -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DENABLE_FFTW=ON -DENABLE_LIBEMOS_TESTS_REGRESS=ON -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.17
      edit COMPILER 'INTEL.17'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap intel/17.0.3.053" "module load gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.85
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.85'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.86
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.86'
      edit MODULE_ACTIONS '"module swap cdt/17.09" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.87'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family cct
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECBUILD_LOG_LEVEL=DEBUG -DECBUILD_TRUST_FLAGS=ON -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DENABLE_FFTW=ON -DENABLE_LIBEMOS_TESTS_REGRESS=ON -DLAPACK_FOUND=TRUE'
    edit CPUS '$EC_total_tasks'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/libemos/cct'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/libemos/cct:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cct-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '4'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export FC=ftn" "export TMPDIR=/tmp" "export ftp_proxy=http://proxy.ecmwf.int:3333" "export http_proxy=http://proxy.ecmwf.int:3333" "export https_proxy=http://proxy.ecmwf.int:3333"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '5250'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cct'
    edit QSUB_LP 'QSUB -lP %EC_TOTAL_TASKS%'
    edit QUEUE 'nf'
    edit REMOTE_HOST 'cct'
    edit SITE 'cct'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DECBUILD_LOG_LEVEL=DEBUG -DECBUILD_TRUST_FLAGS=ON -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DENABLE_FFTW=ON -DENABLE_LIBEMOS_TESTS_REGRESS=ON -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DECBUILD_LOG_LEVEL=DEBUG -DECBUILD_TRUST_FLAGS=ON -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DENABLE_FFTW=ON -DENABLE_LIBEMOS_TESTS_REGRESS=ON -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.17
      edit COMPILER 'INTEL.17'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap intel/17.0.3.053" "module load gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.85
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.85'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.86
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.86'
      edit MODULE_ACTIONS '"module swap cdt/17.09" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.87'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family ecgb
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'ecgb'
    edit QUEUE 'express'
    edit REMOTE_HOST 'ecgb'
    edit SITE 'ecgb'
    edit WSDIR '$SCRATCH/metabuilds/%ECF_HOST%_%ECF_PORT%/ecgb/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECBUILD_LOG_LEVEL=DEBUG -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DENABLE_FFTW=ON -DENABLE_LIBEMOS_TESTS_REGRESS=ON'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family leap42
    defstatus suspended
    edit CI_LIMIT '7'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'leap42'
    edit REMOTE_HOST 'leap42-deploy'
    edit SITE 'leap42'
    edit WSDIR '/tmp/metabuilds/%ECF_HOST%_%ECF_PORT%/leap42/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECBUILD_LOG_LEVEL=DEBUG -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DENABLE_FFTW=ON -DENABLE_LIBEMOS_TESTS_REGRESS=ON'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family lxc
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'lxc'
    edit QUEUE 'express'
    edit REMOTE_HOST 'lxc'
    edit SITE 'lxc'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxc/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECBUILD_LOG_LEVEL=DEBUG -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DENABLE_FFTW=ON -DENABLE_LIBEMOS_TESTS_REGRESS=ON'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family lxg
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'lxg'
    edit QUEUE 'normal'
    edit REMOTE_HOST 'lxg'
    edit SITE 'lxg'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxg/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECBUILD_LOG_LEVEL=DEBUG -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DENABLE_FFTW=ON -DENABLE_LIBEMOS_TESTS_REGRESS=ON'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family sappb
    defstatus suspended
    edit ASAN_FILE '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
    edit BASEPREFIX '/autofs/vol/deploy/install/%SITE%'
    edit CI_LIMIT '8'
    edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DECBUILD_LOG_LEVEL=DEBUG -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DENABLE_FFTW=OFF -DENABLE_LIBEMOS_TESTS_REGRESS=ON -DENABLE_REQUIRE_FFTW=OFF'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
    edit M_SITE 'sappb'
    edit REMOTE_HOST 'sappb-deploy'
    edit SANITIZER_REFERENCE_DIRECTORY '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
    edit SITE 'sappb'
    edit TSAN_FILE '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
    edit WSDIR '$HOME/metabuilds/%ECF_HOST%_%ECF_PORT%/sappb/%COMPILER:%'
    family gnu.63
      edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECBUILD_LOG_LEVEL=DEBUG -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DENABLE_FFTW=OFF -DENABLE_LIBEMOS_TESTS_REGRESS=ON -DENABLE_REQUIRE_FFTW=OFF'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module load gnu/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
endsuite
suite eckit
  edit ADDRESS_SANITIZER 'False'
  edit ADD_ACTION_TRIGGERS 'True'
  edit AEC_VERSION '1.0.4'
  edit ALLOW_INSTALL_TO_OVERWRITE 'True'
  edit ASAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
  edit ATLAS_GIT_TAG 'develop'
  edit ATLAS_VERSION '0.20.2'
  edit BANNER_TASK_END 'banner %TASK%-end || true'
  edit BANNER_TASK_START 'banner %TASK%-start || true'
  edit BASEDIR '%WSDIR%/%PACKAGE%'
  edit BASEPREFIX '$PERM/install/%SITE%'
  edit BUFRDC_EMOSLIB_VERSION '4.6.3'
  edit BUFRDC_TABLES_GIT_TAG 'develop'
  edit BUFRDC_TABLES_VERSION '000415'
  edit BUILDIR '%BASEDIR%/builds'
  edit BUILD_ALWAYS_PARALLEL 'False'
  edit CFI_SANITIZER 'False'
  edit CFL_BASE_URL 'https://software-test.ecmwf.int/wiki'
  edit CFL_SPACE_KEY '<TODO>'
  edit CFL_TARBALL_COMMENT ''
  edit CFL_TITLE 'Releases'
  edit CHECK_SYNC_RELEASE 'False'
  edit CLANG_TIDY_ABORT_ON_ERROR 'False'
  edit CMAKE '%ECBUILD%'
  edit CMAKELIST_EXTRACT_PACKAGE_VERSION '$(awk '/^project/ && /%PACKAGE%/ && /VERSION/ {for (I=1;I<=NF;I++) if ($I == "VERSION") {print $(I+1)};}' CMakeLists.txt)'
  edit CMAKE_BUILD_TYPE 'RelWithDebInfo'
  edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DENABLE_EXPERIMENTAL=OFF -DENABLE_MPI=OFF'
  edit CMAKE_VERSION '3.13.4'
  edit CODES_UI_GIT_TAG 'develop'
  edit CODES_UI_VERSION '1.3.1'
  edit CPACK 'make package'
  edit CPUS '$(lscpu -p | grep -v '#' | wc -l)'
  edit CRAY_FFTW_VERSION '3.3.8.1'
  edit CRAY_NETCDF_VERSION '4.6.1.3'
  edit CRON_BASEPREFIX '%BASEDIR%/install'
  edit CRON_EMAIL ''
  edit CRON_ENABLED 'False'
  edit CRON_GIT_TAG 'develop'
  edit CUSTOM_COMMANDS ''
  edit DATAFLOW_SANITIZER 'False'
  edit DEFAULT_GNU_COMPILER 'gnu.73'
  edit ECBUILD '%BASEPREFIX%/ecbuild/develop/bin/ecbuild'
  edit ECBUILD_CLONE 'False'
  edit ECBUILD_CLONE_COMMIT 'master'
  edit ECBUILD_GIT_TAG 'develop'
  edit ECBUILD_REPO '%GIT_REPO_PREFIX%/ecsdk/ecbuild.git'
  edit ECBUILD_VERSION '3.3.2'
  edit ECCODES_EXTENSIONS_GIT_TAG 'develop'
  edit ECCODES_EXTENSIONS_VERSION '2.17.0'
  edit ECCODES_GIT_TAG 'develop'
  edit ECCODES_PYTHON_VERSION '0.9.6'
  edit ECCODES_VERSION '2.17.1'
  edit ECFLOW5_LATEST_VERSION '5.4.0'
  edit ECFLOW_VERSION '5.4.1'
  edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/eckit'
  edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/eckit'
  edit ECF_JOB_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_JOB% %ECF_JOBOUT%'
  edit ECF_KILL_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% kill'
  edit ECF_STATUS_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% stat'
  edit ECF_TRIES '1'
  edit ECKIT_GIT_TAG 'develop'
  edit ECKIT_VERSION '1.10.1'
  edit ECQUIS_VERSION '1.0.0'
  edit EMOS_MOD_VERSION '462'
  edit EMOS_PA_VERSION '000462'
  edit EXPORTS '"export ECACCOUNT=ecodmdma"'
  edit EXPORT_VARIABLES 'False'
  edit FCKIT_GIT_TAG '0.6.3'
  edit FCKIT_VERSION '0.6.3'
  edit FDB_VINTAGE_GIT_TAG '4.10.0'
  edit FDB_VINTAGE_VERSION '4.10.0'
  edit FFTW_VERSION '3.3.8'
  edit FFTW_VERSION_CRAY '3.3.4.11'
  edit GIT 'git'
  edit GITHUB_PACKAGE_NAME '%PACKAGE_NAME%'
  edit GITHUB_REPO 'git@github.com:ecmwf'
  edit GITREPO '%GIT_REPO_PREFIX%/ecsdk/eckit.git'
  edit GIT_CLONE 'True'
  edit GIT_CLONE_OPTIONS '--depth=1 --branch %GIT_TAG%'
  edit GIT_METABUILDER_TAG 'ci'
  edit GIT_REPO_PREFIX 'ssh://git@git.ecmwf.int:7999'
  edit GIT_TAG 'develop'
  edit GIT_VERSION '2.26.0'
  edit GRIB_API_VERSION '1.28.0'
  edit INSTALLABLE_TO_USR_LOCAL_APPS 'True'
  edit IWYU_VERSION '0.11'
  edit KSH_EXTRACT_PACKAGE_VERSION '`grep _VERSION_STR VERSION.cmake | grep -o '[0-9]*\.[0-9]*\.[0-9]*'`'
  edit LAPACK_VERSION '3.5.0'
  edit LIBEMOS_GIT_TAG 'develop'
  edit LIBEMOS_VERSION '4.6.2'
  edit LINTERS_REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}'
  edit LSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/lsan.txt'
  edit MAGICS_GIT_TAG 'develop'
  edit MAGICS_PYTHON_VERSION 'master'
  edit MAGICS_VERSION '4.3.3'
  edit MAKE 'make'
  edit MARS_CLIENT_GIT_TAG '6.27.2.2'
  edit MARS_CLIENT_PACKAGE_GIT_TAG '6.27.6'
  edit MARS_CLIENT_PACKAGE_VERSION '6.27.6'
  edit MARS_CLIENT_VERSION '6.27.2.2'
  edit MEMORY_SANITIZER 'False'
  edit METABUILDER_HOME '/var/tmp/ma0/workspace/metabuilder'
  edit METAB_INSTALL_REMOVE_WRITE_PERMISSIONS 'True'
  edit METAB_PYTHON_VERSION 'python3'
  edit METKIT_GIT_TAG 'develop'
  edit METKIT_VERSION '1.4.6'
  edit METVIEW_GIT_TAG 'develop'
  edit METVIEW_PYTHON_VERSION '1.4.2'
  edit METVIEW_VERSION '5.8.3'
  edit MIR_BUNDLE_VERSION '0.9.0.0'
  edit MIR_GIT_TAG 'develop'
  edit MIR_VERSION '1.5.0'
  edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
  edit MODULE_ACTIONS_FIRST ''
  edit MODULE_ACTIONS_HDR 'module_actions.h'
  edit MODULE_CHECKING_HDR 'module_checking.h'
  edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow"'
  edit MODULE_LOAD_SUPPORTED 'True'
  edit MODULE_NAME 'eckit'
  edit MODULE_VERSION '%GIT_TAG%'
  edit NETCDF4_VERSION '4.7.4'
  edit ODB_API_GIT_TAG 'develop'
  edit ODB_API_VERSION '0.19.4'
  edit ODB_GIT_TAG 'develop'
  edit ODB_TOOLS_GIT_TAG 'develop'
  edit ODB_TOOLS_VERSION '0.18.3'
  edit ODB_VERSION '1.0.11'
  edit ODC_GIT_TAG 'develop'
  edit ODC_VERSION '1.0.3'
  edit OPENJPEG_VERSION '2.3.1'
  edit PACKAGE 'eckit'
  edit PACKAGE_NAME 'eckit'
  edit PACKAGE_OWNERS 'matq@ecmwf.int mass@ecmwf.int cgm@ecmwf.int ma0@ecmwf.int olivier.iffrig@ecmwf.int'
  edit PACKAGE_VERSION '%GIT_TAG%'
  edit PGEN_VERSION '1.7.1.0'
  edit PKG_BOOST_VERSION '1.71.0'
  edit PREFIX_APPEND ''
  edit PROJ4_VERSION '5.2.0'
  edit PROJ_VERSION '6.3.1'
  edit PYTHON2_VERSION '2.7.15-01'
  edit PYTHON3_VERSION '3.6.10-01'
  edit PYTHON3_VERSION_MAJOR_MINOR '3.6'
  edit PYTHON_GITREPO 'This-needs-tobe-overriden'
  edit PYTHON_PACKAGE 'This-needs-tobe-overriden'
  edit QT_VERSION '5.12.3'
  edit QUEUE 'build'
  edit REMOTE_HOST 'polonius'
  edit REUSE_BUILD_DIR 'False'
  edit SAFE_STACK_SANITIZER 'False'
  edit SANITIZER 'False'
  edit SANITIZER_REFERENCE_DIRECTORY '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
  edit SRCDIR '%BASEDIR%/%PACKAGE%'
  edit STATIC_ANALYSER_ABORT_ON_ERROR 'False'
  edit SWIG_VERSION '3.0.8'
  edit SYNC_RELEASE '2020.04.0'
  edit SYNC_RELEASE_CI_STAGE 'develop'
  edit TARBALL_DIR '/perm/ma/ma0/tmp_tarballs/ci'
  edit TAR_PACKAGE '%PACKAGE%'
  edit TEST 'ctest --output-on-failure -j${CPUS} '
  edit THREAD_SANITIZER 'False'
  edit TRANSI_GIT_TAG 'develop'
  edit TRANSI_VERSION '0.7.1'
  edit TSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
  edit UBSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/ubsan.txt'
  edit UB_SANITIZER 'False'
  edit USER 'ma0'
  edit USE_WORKDIR 'False'
  edit VALGRIND_VERSION '3.15.0'
  edit WORKDIR '%BASEDIR%/workdir/%ECF_NAME%'
  edit WSDIR '/var/tmp/tmpdir/deploy/jtmp.20739'
  label branch "develop"
  task regenerate
    defstatus complete
    edit ECF_JOB_CMD '%ECF_JOB% > %ECF_JOBOUT% 2>&1'
    edit ECF_KILL_CMD 'kill -15 %ECF_RID%'
    edit ECF_PASS 'FREE'
    edit ECF_STATUS_CMD 'ps --sid %ECF_RID% -f'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
    label date ""
    label metabuilder_branch "ci"
  family cca
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DENABLE_EXPERIMENTAL=OFF -DENABLE_MPI=OFF -DLAPACK_FOUND=TRUE'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/eckit/cca'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/eckit/cca:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cca-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '24000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cca'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'cca'
    edit SITE 'cca'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DENABLE_EXPERIMENTAL=OFF -DENABLE_MPI=OFF -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DENABLE_EXPERIMENTAL=OFF -DENABLE_MPI=OFF -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.17
      edit COMPILER 'INTEL.17'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap intel/17.0.3.053" "module load gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.85
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.85'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit TEST 'ctest -E "(eckit_test_types_doublecompare|eckit_test_types_floatcompare)" -j${CPUS} --output-on-failure'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.86
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.86'
      edit MODULE_ACTIONS '"module swap cdt/17.09" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit TEST 'ctest -E "(eckit_test_types_doublecompare|eckit_test_types_floatcompare)" -j${CPUS} --output-on-failure'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.87'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit TEST 'ctest -E "(eckit_test_types_doublecompare|eckit_test_types_floatcompare)" -j${CPUS} --output-on-failure'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family ccb
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DENABLE_EXPERIMENTAL=OFF -DENABLE_MPI=OFF -DLAPACK_FOUND=TRUE'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/eckit/ccb'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/eckit/ccb:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'ccb-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '24000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'ccb'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'ccb'
    edit SITE 'ccb'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DENABLE_EXPERIMENTAL=OFF -DENABLE_MPI=OFF -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DENABLE_EXPERIMENTAL=OFF -DENABLE_MPI=OFF -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.17
      edit COMPILER 'INTEL.17'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap intel/17.0.3.053" "module load gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.85
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.85'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit TEST 'ctest -E "(eckit_test_types_doublecompare|eckit_test_types_floatcompare)" -j${CPUS} --output-on-failure'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.86
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.86'
      edit MODULE_ACTIONS '"module swap cdt/17.09" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit TEST 'ctest -E "(eckit_test_types_doublecompare|eckit_test_types_floatcompare)" -j${CPUS} --output-on-failure'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.87'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit TEST 'ctest -E "(eckit_test_types_doublecompare|eckit_test_types_floatcompare)" -j${CPUS} --output-on-failure'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family cct
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DENABLE_EXPERIMENTAL=OFF -DENABLE_MPI=OFF -DLAPACK_FOUND=TRUE'
    edit CPUS '$EC_total_tasks'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/eckit/cct'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/eckit/cct:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cct-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '4'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '6000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cct'
    edit QSUB_LP 'QSUB -lP %EC_TOTAL_TASKS%'
    edit QUEUE 'nf'
    edit REMOTE_HOST 'cct'
    edit SITE 'cct'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DENABLE_EXPERIMENTAL=OFF -DENABLE_MPI=OFF -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DENABLE_EXPERIMENTAL=OFF -DENABLE_MPI=OFF -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.17
      edit COMPILER 'INTEL.17'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap intel/17.0.3.053" "module load gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.85
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.85'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit TEST 'ctest -E "(eckit_test_types_doublecompare|eckit_test_types_floatcompare)" -j${CPUS} --output-on-failure'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.86
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.86'
      edit MODULE_ACTIONS '"module swap cdt/17.09" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit TEST 'ctest -E "(eckit_test_types_doublecompare|eckit_test_types_floatcompare)" -j${CPUS} --output-on-failure'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.87'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit TEST 'ctest -E "(eckit_test_types_doublecompare|eckit_test_types_floatcompare)" -j${CPUS} --output-on-failure'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family ecgb
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'ecgb'
    edit QUEUE 'express'
    edit REMOTE_HOST 'ecgb'
    edit SITE 'ecgb'
    edit WSDIR '$SCRATCH/metabuilds/%ECF_HOST%_%ECF_PORT%/ecgb/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_MPI=OFF'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family leap42
    defstatus suspended
    edit CI_LIMIT '7'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'leap42'
    edit REMOTE_HOST 'leap42-deploy'
    edit SITE 'leap42'
    edit WSDIR '/tmp/metabuilds/%ECF_HOST%_%ECF_PORT%/leap42/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_MPI=OFF'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task create_tarball
        trigger test == complete
        edit CMAKE 'cmake'
        edit EXPORTS '"export ECACCOUNT=ecodmdma" "export ecbuild_ROOT=%BASEPREFIX%/ecbuild/develop"'
        label tar ""
        label tardir ""
        label error ""
      task install
        trigger create_tarball == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.93
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_MPI=OFF'
      edit COMPILER 'GNU.93'
      edit EXPORTS '"export ECACCOUNT=ecodmdma" "export LD_LIBRARY_PATH=/usr/local/apps/gcc/9.3.0/lib64/"'
      edit MODULE_ACTIONS '"module swap gnu/9.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.19
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_FLAGS=-fPIC -DENABLE_EXPERIMENTAL=OFF -DENABLE_MPI=OFF'
      edit COMPILER 'INTEL.19'
      edit MODULE_ACTIONS '"module unload gnu" "module load intel/19.0.4" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit POST_EXPORTS '"export CXX=icpc\ -cxxlib=/usr/local/apps/gcc/8.3.0" "export LD_RUN_PATH=/usr/local/apps/gcc/8.3.0/lib64" "export PATH=/usr/local/apps/gcc/8.3.0/bin:$PATH"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_MPI=OFF'
      edit COMPILER 'CLANG.90'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90.linters
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_MPI=OFF'
      edit COMPILER 'CLANG.90.LINTERS'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task cppcheck
        trigger config == complete
        edit CPPCHECK_FLAGS '--template='[{file}:{line}]: {severity} : {id} : {message}' --suppress=syntaxError --inline-suppr --platform=unix64 --enable=performance --enable=portability --enable=information --std=c11 --std=c++14 -DDEBUG=1 -q -j8'
        edit CPPCHECK_VERSION '1.89'
        edit REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}/cpp_check'
      task iwyu
        trigger cppcheck == complete
        edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module load iwyu/%IWYU_VERSION%"'
      family clang_linters
        trigger iwyu == complete
        limit clang_linters 1
        inlimit clang_linters
        task clang_tidy
          edit REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}/clang_tidy'
        task static_analyser
          edit REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}/static_analyser'
          edit SCAN_BUILD_OPTIONS '--force-analyze-debug-code -o /perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}/static_analyser -k make -j8'
          edit USE_CLANG_CHECK 'False'
          label scan_view ""
      endfamily
      task wipe_build
        trigger clang_linters == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90.address_sanitizer
      edit ADDRESS_SANITIZER 'True'
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_FLAGS="-fsanitize=address -fno-omit-frame-pointer" -DCMAKE_C_FLAGS="-fsanitize=address -fno-omit-frame-pointer" -DCMAKE_EXE_LINKER_FLAGS=-fsanitize=address -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_MPI=OFF'
      edit COMPILER 'CLANG.90.ADDRESS_SANITIZER'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit SANITIZER 'True'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit TEST_DIR '%BUILDIR%'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90.ub_sanitizer
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_FLAGS="-fsanitize=undefined -fno-omit-frame-pointer -fno-sanitize-recover=all" -DCMAKE_C_FLAGS="-fsanitize=undefined -fno-omit-frame-pointer" -DCMAKE_EXE_LINKER_FLAGS=-fsanitize=undefined -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_MPI=OFF'
      edit COMPILER 'CLANG.90.UB_SANITIZER'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit SANITIZER 'True'
      edit UB_SANITIZER 'True'
      edit UB_SANITIZER_LD_PRELOAD 'export LD_PRELOAD=/usr/local/apps/clang/9.0.1/lib64/libubsan.so'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit TEST_DIR '%BUILDIR%'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family lxc
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'lxc'
    edit QUEUE 'express'
    edit REMOTE_HOST 'lxc'
    edit SITE 'lxc'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxc/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_MPI=OFF'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family lxg
    defstatus suspended
    edit CI_LIMIT '10'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DENABLE_EXPERIMENTAL=OFF -DENABLE_MKL=ON -DENABLE_MPI=OFF -DMKL_ROOT=/usr/local/apps/intel/parallel_studio_xe_2019/compilers_and_libraries_2019.0.117/linux/mkl'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'lxg'
    edit QUEUE 'normal'
    edit REMOTE_HOST 'lxg'
    edit SITE 'lxg'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxg/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_MKL=ON -DENABLE_MPI=OFF -DMKL_ROOT=/usr/local/apps/intel/parallel_studio_xe_2019/compilers_and_libraries_2019.0.117/linux/mkl'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family sappb
    defstatus suspended
    edit ASAN_FILE '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
    edit BASEPREFIX '/autofs/vol/deploy/install/%SITE%'
    edit CI_LIMIT '8'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit M_SITE 'sappb'
    edit REMOTE_HOST 'sappb-deploy'
    edit SANITIZER_REFERENCE_DIRECTORY '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
    edit SITE 'sappb'
    edit TSAN_FILE '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
    edit WSDIR '$HOME/metabuilds/%ECF_HOST%_%ECF_PORT%/sappb/%COMPILER:%'
    family gnu.63
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_MPI=OFF'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module load gnu/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
endsuite
suite metkit
  edit ADDRESS_SANITIZER 'False'
  edit ADD_ACTION_TRIGGERS 'True'
  edit AEC_VERSION '1.0.4'
  edit ALLOW_INSTALL_TO_OVERWRITE 'True'
  edit ASAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
  edit ATLAS_GIT_TAG 'develop'
  edit ATLAS_VERSION '0.20.2'
  edit BANNER_TASK_END 'banner %TASK%-end || true'
  edit BANNER_TASK_START 'banner %TASK%-start || true'
  edit BASEDIR '%WSDIR%/%PACKAGE%'
  edit BASEPREFIX '$PERM/install/%SITE%'
  edit BUFRDC_EMOSLIB_VERSION '4.6.3'
  edit BUFRDC_TABLES_GIT_TAG 'develop'
  edit BUFRDC_TABLES_VERSION '000415'
  edit BUILDIR '%BASEDIR%/builds'
  edit BUILD_ALWAYS_PARALLEL 'False'
  edit CFI_SANITIZER 'False'
  edit CFL_BASE_URL 'https://software-test.ecmwf.int/wiki'
  edit CFL_SPACE_KEY '<TODO>'
  edit CFL_TARBALL_COMMENT ''
  edit CFL_TITLE 'Releases'
  edit CHECK_SYNC_RELEASE 'False'
  edit CLANG_TIDY_ABORT_ON_ERROR 'False'
  edit CMAKE '%ECBUILD%'
  edit CMAKELIST_EXTRACT_PACKAGE_VERSION '$(awk '/^project/ && /%PACKAGE%/ && /VERSION/ {for (I=1;I<=NF;I++) if ($I == "VERSION") {print $(I+1)};}' CMakeLists.txt)'
  edit CMAKE_BUILD_TYPE 'RelWithDebInfo'
  edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_GRIB=OFF'
  edit CMAKE_VERSION '3.13.4'
  edit CODES_UI_GIT_TAG 'develop'
  edit CODES_UI_VERSION '1.3.1'
  edit CPACK 'make package'
  edit CPUS '$(lscpu -p | grep -v '#' | wc -l)'
  edit CRAY_FFTW_VERSION '3.3.8.1'
  edit CRAY_NETCDF_VERSION '4.6.1.3'
  edit CRON_BASEPREFIX '%BASEDIR%/install'
  edit CRON_EMAIL ''
  edit CRON_ENABLED 'False'
  edit CRON_GIT_TAG 'develop'
  edit CUSTOM_COMMANDS ''
  edit DATAFLOW_SANITIZER 'False'
  edit DEFAULT_GNU_COMPILER 'gnu.73'
  edit ECBUILD '%BASEPREFIX%/ecbuild/develop/bin/ecbuild'
  edit ECBUILD_CLONE 'False'
  edit ECBUILD_CLONE_COMMIT 'master'
  edit ECBUILD_GIT_TAG 'develop'
  edit ECBUILD_REPO '%GIT_REPO_PREFIX%/ecsdk/ecbuild.git'
  edit ECBUILD_VERSION '3.3.2'
  edit ECCODES_EXTENSIONS_GIT_TAG 'develop'
  edit ECCODES_EXTENSIONS_VERSION '2.17.0'
  edit ECCODES_GIT_TAG 'develop'
  edit ECCODES_PYTHON_VERSION '0.9.6'
  edit ECCODES_VERSION '2.17.1'
  edit ECFLOW5_LATEST_VERSION '5.4.0'
  edit ECFLOW_VERSION '5.4.1'
  edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/metkit'
  edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/metkit'
  edit ECF_JOB_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_JOB% %ECF_JOBOUT%'
  edit ECF_KILL_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% kill'
  edit ECF_STATUS_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% stat'
  edit ECF_TRIES '1'
  edit ECKIT_GIT_TAG 'develop'
  edit ECKIT_VERSION '1.10.1'
  edit ECQUIS_VERSION '1.0.0'
  edit EMOS_MOD_VERSION '462'
  edit EMOS_PA_VERSION '000462'
  edit EXPORTS '"export ECACCOUNT=ecodmdma"'
  edit EXPORT_VARIABLES 'False'
  edit FCKIT_GIT_TAG '0.6.3'
  edit FCKIT_VERSION '0.6.3'
  edit FDB_VINTAGE_GIT_TAG '4.10.0'
  edit FDB_VINTAGE_VERSION '4.10.0'
  edit FFTW_VERSION '3.3.8'
  edit FFTW_VERSION_CRAY '3.3.4.11'
  edit GIT 'git'
  edit GITHUB_PACKAGE_NAME '%PACKAGE_NAME%'
  edit GITHUB_REPO 'git@github.com:ecmwf'
  edit GITREPO '%GIT_REPO_PREFIX%/ecsdk/metkit.git'
  edit GIT_CLONE 'True'
  edit GIT_CLONE_OPTIONS '--depth=1 --branch %GIT_TAG%'
  edit GIT_METABUILDER_TAG 'ci'
  edit GIT_REPO_PREFIX 'ssh://git@git.ecmwf.int:7999'
  edit GIT_TAG 'develop'
  edit GIT_VERSION '2.26.0'
  edit GRIB_API_VERSION '1.28.0'
  edit INSTALLABLE_TO_USR_LOCAL_APPS 'True'
  edit IWYU_VERSION '0.11'
  edit KSH_EXTRACT_PACKAGE_VERSION '`grep _VERSION_STR VERSION.cmake | grep -o '[0-9]*\.[0-9]*\.[0-9]*'`'
  edit LAPACK_VERSION '3.5.0'
  edit LIBEMOS_GIT_TAG 'develop'
  edit LIBEMOS_VERSION '4.6.2'
  edit LINTERS_REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}'
  edit LSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/lsan.txt'
  edit MAGICS_GIT_TAG 'develop'
  edit MAGICS_PYTHON_VERSION 'master'
  edit MAGICS_VERSION '4.3.3'
  edit MAKE 'make'
  edit MARS_CLIENT_GIT_TAG '6.27.2.2'
  edit MARS_CLIENT_PACKAGE_GIT_TAG '6.27.6'
  edit MARS_CLIENT_PACKAGE_VERSION '6.27.6'
  edit MARS_CLIENT_VERSION '6.27.2.2'
  edit MEMORY_SANITIZER 'False'
  edit METABUILDER_HOME '/var/tmp/ma0/workspace/metabuilder'
  edit METAB_INSTALL_REMOVE_WRITE_PERMISSIONS 'True'
  edit METAB_PYTHON_VERSION 'python3'
  edit METKIT_GIT_TAG 'develop'
  edit METKIT_VERSION '1.4.6'
  edit METVIEW_GIT_TAG 'develop'
  edit METVIEW_PYTHON_VERSION '1.4.2'
  edit METVIEW_VERSION '5.8.3'
  edit MIR_BUNDLE_VERSION '0.9.0.0'
  edit MIR_GIT_TAG 'develop'
  edit MIR_VERSION '1.5.0'
  edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
  edit MODULE_ACTIONS_FIRST ''
  edit MODULE_ACTIONS_HDR 'module_actions.h'
  edit MODULE_CHECKING_HDR 'module_checking.h'
  edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow"'
  edit MODULE_LOAD_SUPPORTED 'True'
  edit MODULE_NAME 'metkit'
  edit MODULE_VERSION '%GIT_TAG%'
  edit NETCDF4_VERSION '4.7.4'
  edit ODB_API_GIT_TAG 'develop'
  edit ODB_API_VERSION '0.19.4'
  edit ODB_GIT_TAG 'develop'
  edit ODB_TOOLS_GIT_TAG 'develop'
  edit ODB_TOOLS_VERSION '0.18.3'
  edit ODB_VERSION '1.0.11'
  edit ODC_GIT_TAG 'develop'
  edit ODC_VERSION '1.0.3'
  edit OPENJPEG_VERSION '2.3.1'
  edit PACKAGE 'metkit'
  edit PACKAGE_NAME 'metkit'
  edit PACKAGE_OWNERS 'matq@ecmwf.int cgm@ecmwf.int ma0@ecmwf.int mass@ecmwf.int olivier.iffrig@ecmwf.int'
  edit PACKAGE_VERSION '%GIT_TAG%'
  edit PGEN_VERSION '1.7.1.0'
  edit PKG_BOOST_VERSION '1.71.0'
  edit PREFIX_APPEND ''
  edit PROJ4_VERSION '5.2.0'
  edit PROJ_VERSION '6.3.1'
  edit PYTHON2_VERSION '2.7.15-01'
  edit PYTHON3_VERSION '3.6.10-01'
  edit PYTHON3_VERSION_MAJOR_MINOR '3.6'
  edit PYTHON_GITREPO 'This-needs-tobe-overriden'
  edit PYTHON_PACKAGE 'This-needs-tobe-overriden'
  edit QT_VERSION '5.12.3'
  edit QUEUE 'build'
  edit REMOTE_HOST 'polonius'
  edit REUSE_BUILD_DIR 'False'
  edit SAFE_STACK_SANITIZER 'False'
  edit SANITIZER 'False'
  edit SANITIZER_REFERENCE_DIRECTORY '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
  edit SRCDIR '%BASEDIR%/%PACKAGE%'
  edit STATIC_ANALYSER_ABORT_ON_ERROR 'False'
  edit SWIG_VERSION '3.0.8'
  edit SYNC_RELEASE '2020.04.0'
  edit SYNC_RELEASE_CI_STAGE 'develop'
  edit TARBALL_DIR '/perm/ma/ma0/tmp_tarballs/ci'
  edit TAR_PACKAGE '%PACKAGE%'
  edit TEST 'ctest --output-on-failure -j${CPUS} '
  edit THREAD_SANITIZER 'False'
  edit TRANSI_GIT_TAG 'develop'
  edit TRANSI_VERSION '0.7.1'
  edit TSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
  edit UBSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/ubsan.txt'
  edit UB_SANITIZER 'False'
  edit USER 'ma0'
  edit USE_WORKDIR 'False'
  edit VALGRIND_VERSION '3.15.0'
  edit WORKDIR '%BASEDIR%/workdir/%ECF_NAME%'
  edit WSDIR '/var/tmp/tmpdir/deploy/jtmp.20739'
  label branch "develop"
  task regenerate
    defstatus complete
    edit ECF_JOB_CMD '%ECF_JOB% > %ECF_JOBOUT% 2>&1'
    edit ECF_KILL_CMD 'kill -15 %ECF_RID%'
    edit ECF_PASS 'FREE'
    edit ECF_STATUS_CMD 'ps --sid %ECF_RID% -f'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
    label date ""
    label metabuilder_branch "ci"
  family cca
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_GRIB=OFF -DLAPACK_FOUND=TRUE'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/metkit/cca'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/metkit/cca:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cca-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '24000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cca'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'cca'
    edit SITE 'cca'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_GRIB=OFF -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_GRIB=OFF -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.17
      edit COMPILER 'INTEL.17'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap intel/17.0.3.053" "module load gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.85
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.85'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit TEST 'ctest -E "(eckit_test_types_doublecompare|eckit_test_types_floatcompare)" -j${CPUS} --output-on-failure'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.86
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.86'
      edit MODULE_ACTIONS '"module swap cdt/17.09" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit TEST 'ctest -E "(eckit_test_types_doublecompare|eckit_test_types_floatcompare)" -j${CPUS} --output-on-failure'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.87'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit TEST 'ctest -E "(eckit_test_types_doublecompare|eckit_test_types_floatcompare)" -j${CPUS} --output-on-failure'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family ccb
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_GRIB=OFF -DLAPACK_FOUND=TRUE'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/metkit/ccb'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/metkit/ccb:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'ccb-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '24000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'ccb'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'ccb'
    edit SITE 'ccb'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_GRIB=OFF -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_GRIB=OFF -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.17
      edit COMPILER 'INTEL.17'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap intel/17.0.3.053" "module load gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.85
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.85'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit TEST 'ctest -E "(eckit_test_types_doublecompare|eckit_test_types_floatcompare)" -j${CPUS} --output-on-failure'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.86
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.86'
      edit MODULE_ACTIONS '"module swap cdt/17.09" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit TEST 'ctest -E "(eckit_test_types_doublecompare|eckit_test_types_floatcompare)" -j${CPUS} --output-on-failure'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.87'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit TEST 'ctest -E "(eckit_test_types_doublecompare|eckit_test_types_floatcompare)" -j${CPUS} --output-on-failure'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family cct
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_GRIB=OFF -DLAPACK_FOUND=TRUE'
    edit CPUS '$EC_total_tasks'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/metkit/cct'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/metkit/cct:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cct-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '4'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '6000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cct'
    edit QSUB_LP 'QSUB -lP %EC_TOTAL_TASKS%'
    edit QUEUE 'nf'
    edit REMOTE_HOST 'cct'
    edit SITE 'cct'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_GRIB=OFF -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_GRIB=OFF -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.17
      edit COMPILER 'INTEL.17'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap intel/17.0.3.053" "module load gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.85
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.85'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit TEST 'ctest -E "(eckit_test_types_doublecompare|eckit_test_types_floatcompare)" -j${CPUS} --output-on-failure'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.86
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.86'
      edit MODULE_ACTIONS '"module swap cdt/17.09" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit TEST 'ctest -E "(eckit_test_types_doublecompare|eckit_test_types_floatcompare)" -j${CPUS} --output-on-failure'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.87'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit TEST 'ctest -E "(eckit_test_types_doublecompare|eckit_test_types_floatcompare)" -j${CPUS} --output-on-failure'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family ecgb
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'ecgb'
    edit QUEUE 'express'
    edit REMOTE_HOST 'ecgb'
    edit SITE 'ecgb'
    edit WSDIR '$SCRATCH/metabuilds/%ECF_HOST%_%ECF_PORT%/ecgb/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_GRIB=OFF'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family leap42
    defstatus suspended
    edit CI_LIMIT '7'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'leap42'
    edit REMOTE_HOST 'leap42-deploy'
    edit SITE 'leap42'
    edit WSDIR '/tmp/metabuilds/%ECF_HOST%_%ECF_PORT%/leap42/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_GRIB=OFF'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.93
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_GRIB=OFF'
      edit COMPILER 'GNU.93'
      edit EXPORTS '"export ECACCOUNT=ecodmdma" "export LD_LIBRARY_PATH=/usr/local/apps/gcc/9.3.0/lib64/"'
      edit MODULE_ACTIONS '"module swap gnu/9.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.19
      edit COMPILER 'INTEL.19'
      edit MODULE_ACTIONS '"module unload gnu" "module load intel/19.0.4" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit POST_EXPORTS '"export CXX=icpc\ -cxxlib=/usr/local/apps/gcc/8.3.0" "export LD_RUN_PATH=/usr/local/apps/gcc/8.3.0/lib64" "export PATH=/usr/local/apps/gcc/8.3.0/bin:$PATH"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_GRIB=OFF'
      edit COMPILER 'CLANG.90'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90.linters
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_GRIB=OFF'
      edit COMPILER 'CLANG.90.LINTERS'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task cppcheck
        trigger config == complete
        edit CPPCHECK_FLAGS '--template='[{file}:{line}]: {severity} : {id} : {message}' --suppress=syntaxError --inline-suppr --platform=unix64 --enable=performance --enable=portability --enable=information --std=c11 --std=c++14 -DDEBUG=1 -q -j8'
        edit CPPCHECK_VERSION '1.89'
        edit REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}/cpp_check'
      task iwyu
        trigger cppcheck == complete
        edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module load iwyu/%IWYU_VERSION%"'
      family clang_linters
        trigger iwyu == complete
        limit clang_linters 1
        inlimit clang_linters
        task clang_tidy
          edit REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}/clang_tidy'
        task static_analyser
          edit REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}/static_analyser'
          edit SCAN_BUILD_OPTIONS '--force-analyze-debug-code -o /perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}/static_analyser -k make -j8'
          edit USE_CLANG_CHECK 'False'
          label scan_view ""
      endfamily
      task wipe_build
        trigger clang_linters == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90.address_sanitizer
      edit ADDRESS_SANITIZER 'True'
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_FLAGS="-fsanitize=address -fno-omit-frame-pointer" -DCMAKE_C_FLAGS="-fsanitize=address -fno-omit-frame-pointer" -DCMAKE_EXE_LINKER_FLAGS=-fsanitize=address -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_GRIB=OFF'
      edit COMPILER 'CLANG.90.ADDRESS_SANITIZER'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit SANITIZER 'True'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit TEST_DIR '%BUILDIR%'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90.ub_sanitizer
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_FLAGS="-fsanitize=undefined -fno-omit-frame-pointer -fno-sanitize-recover=all" -DCMAKE_C_FLAGS="-fsanitize=undefined -fno-omit-frame-pointer" -DCMAKE_EXE_LINKER_FLAGS=-fsanitize=undefined -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_GRIB=OFF'
      edit COMPILER 'CLANG.90.UB_SANITIZER'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit SANITIZER 'True'
      edit UB_SANITIZER 'True'
      edit UB_SANITIZER_LD_PRELOAD 'export LD_PRELOAD=/usr/local/apps/clang/9.0.1/lib64/libubsan.so'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit TEST_DIR '%BUILDIR%'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family lxc
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'lxc'
    edit QUEUE 'express'
    edit REMOTE_HOST 'lxc'
    edit SITE 'lxc'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxc/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_GRIB=OFF'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family lxg
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'lxg'
    edit QUEUE 'normal'
    edit REMOTE_HOST 'lxg'
    edit SITE 'lxg'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxg/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_GRIB=OFF'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family sappb
    defstatus suspended
    edit ASAN_FILE '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
    edit BASEPREFIX '/autofs/vol/deploy/install/%SITE%'
    edit CI_LIMIT '8'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit M_SITE 'sappb'
    edit REMOTE_HOST 'sappb-deploy'
    edit SANITIZER_REFERENCE_DIRECTORY '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
    edit SITE 'sappb'
    edit TSAN_FILE '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
    edit WSDIR '$HOME/metabuilds/%ECF_HOST%_%ECF_PORT%/sappb/%COMPILER:%'
    family gnu.63
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_GRIB=OFF'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module load gnu/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
endsuite
suite transi
  edit ADDRESS_SANITIZER 'False'
  edit ADD_ACTION_TRIGGERS 'True'
  edit AEC_VERSION '1.0.4'
  edit ALLOW_INSTALL_TO_OVERWRITE 'True'
  edit ASAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
  edit ATLAS_GIT_TAG 'develop'
  edit ATLAS_VERSION '0.20.2'
  edit BANNER_TASK_END 'banner %TASK%-end || true'
  edit BANNER_TASK_START 'banner %TASK%-start || true'
  edit BASEDIR '%WSDIR%/%PACKAGE%'
  edit BASEPREFIX '$PERM/install/%SITE%'
  edit BUFRDC_EMOSLIB_VERSION '4.6.3'
  edit BUFRDC_TABLES_GIT_TAG 'develop'
  edit BUFRDC_TABLES_VERSION '000415'
  edit BUILDIR '%BASEDIR%/builds'
  edit BUILD_ALWAYS_PARALLEL 'False'
  edit CFI_SANITIZER 'False'
  edit CFL_BASE_URL 'https://software-test.ecmwf.int/wiki'
  edit CFL_SPACE_KEY '<TODO>'
  edit CFL_TARBALL_COMMENT ''
  edit CFL_TITLE 'Releases'
  edit CHECK_SYNC_RELEASE 'False'
  edit CLANG_TIDY_ABORT_ON_ERROR 'False'
  edit CMAKE '%ECBUILD%'
  edit CMAKELIST_EXTRACT_PACKAGE_VERSION '$(awk '/^project/ && /%PACKAGE%/ && /VERSION/ {for (I=1;I<=NF;I++) if ($I == "VERSION") {print $(I+1)};}' CMakeLists.txt)'
  edit CMAKE_BUILD_TYPE 'Bit'
  edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DENABLE_GRIB=OFF -DENABLE_MPI=OFF'
  edit CMAKE_VERSION '3.13.4'
  edit CODES_UI_GIT_TAG 'develop'
  edit CODES_UI_VERSION '1.3.1'
  edit CPACK 'make package'
  edit CPUS '$(lscpu -p | grep -v '#' | wc -l)'
  edit CRAY_FFTW_VERSION '3.3.8.1'
  edit CRAY_NETCDF_VERSION '4.6.1.3'
  edit CRON_BASEPREFIX '%BASEDIR%/install'
  edit CRON_EMAIL ''
  edit CRON_ENABLED 'False'
  edit CRON_GIT_TAG 'develop'
  edit CUSTOM_COMMANDS ''
  edit DATAFLOW_SANITIZER 'False'
  edit DEFAULT_GNU_COMPILER 'gnu.73'
  edit ECBUILD '%BASEPREFIX%/ecbuild/develop/bin/ecbuild'
  edit ECBUILD_CLONE 'False'
  edit ECBUILD_CLONE_COMMIT 'master'
  edit ECBUILD_GIT_TAG 'develop'
  edit ECBUILD_REPO '%GIT_REPO_PREFIX%/ecsdk/ecbuild.git'
  edit ECBUILD_VERSION '3.3.2'
  edit ECCODES_EXTENSIONS_GIT_TAG 'develop'
  edit ECCODES_EXTENSIONS_VERSION '2.17.0'
  edit ECCODES_GIT_TAG 'develop'
  edit ECCODES_PYTHON_VERSION '0.9.6'
  edit ECCODES_VERSION '2.17.1'
  edit ECFLOW5_LATEST_VERSION '5.4.0'
  edit ECFLOW_VERSION '5.4.1'
  edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/transi'
  edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/transi'
  edit ECF_JOB_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_JOB% %ECF_JOBOUT%'
  edit ECF_KILL_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% kill'
  edit ECF_STATUS_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% stat'
  edit ECF_TRIES '1'
  edit ECKIT_GIT_TAG 'develop'
  edit ECKIT_VERSION '1.10.1'
  edit ECQUIS_VERSION '1.0.0'
  edit EMOS_MOD_VERSION '462'
  edit EMOS_PA_VERSION '000462'
  edit EXPORTS '"export ECACCOUNT=ecodmdma"'
  edit EXPORT_VARIABLES 'False'
  edit FCKIT_GIT_TAG '0.6.3'
  edit FCKIT_VERSION '0.6.3'
  edit FDB_VINTAGE_GIT_TAG '4.10.0'
  edit FDB_VINTAGE_VERSION '4.10.0'
  edit FFTW_VERSION '3.3.8'
  edit FFTW_VERSION_CRAY '3.3.4.11'
  edit GIT 'git'
  edit GITHUB_PACKAGE_NAME '%PACKAGE_NAME%'
  edit GITHUB_REPO 'git@github.com:ecmwf'
  edit GITREPO '%GIT_REPO_PREFIX%/atlas/transi.git'
  edit GIT_CLONE 'True'
  edit GIT_CLONE_OPTIONS '--depth=1 --branch %GIT_TAG%'
  edit GIT_METABUILDER_TAG 'ci'
  edit GIT_REPO_PREFIX 'ssh://git@git.ecmwf.int:7999'
  edit GIT_TAG 'develop'
  edit GIT_VERSION '2.26.0'
  edit GRIB_API_VERSION '1.28.0'
  edit INSTALLABLE_TO_USR_LOCAL_APPS 'False'
  edit IWYU_VERSION '0.11'
  edit KSH_EXTRACT_PACKAGE_VERSION '`grep _VERSION_STR VERSION.cmake | grep -o '[0-9]*\.[0-9]*\.[0-9]*'`'
  edit LAPACK_VERSION '3.5.0'
  edit LIBEMOS_GIT_TAG 'develop'
  edit LIBEMOS_VERSION '4.6.2'
  edit LINTERS_REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}'
  edit LSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/lsan.txt'
  edit MAGICS_GIT_TAG 'develop'
  edit MAGICS_PYTHON_VERSION 'master'
  edit MAGICS_VERSION '4.3.3'
  edit MAKE 'make'
  edit MARS_CLIENT_GIT_TAG '6.27.2.2'
  edit MARS_CLIENT_PACKAGE_GIT_TAG '6.27.6'
  edit MARS_CLIENT_PACKAGE_VERSION '6.27.6'
  edit MARS_CLIENT_VERSION '6.27.2.2'
  edit MEMORY_SANITIZER 'False'
  edit METABUILDER_HOME '/var/tmp/ma0/workspace/metabuilder'
  edit METAB_INSTALL_REMOVE_WRITE_PERMISSIONS 'True'
  edit METAB_PYTHON_VERSION 'python3'
  edit METKIT_GIT_TAG 'develop'
  edit METKIT_VERSION '1.4.6'
  edit METVIEW_GIT_TAG 'develop'
  edit METVIEW_PYTHON_VERSION '1.4.2'
  edit METVIEW_VERSION '5.8.3'
  edit MIR_BUNDLE_VERSION '0.9.0.0'
  edit MIR_GIT_TAG 'develop'
  edit MIR_VERSION '1.5.0'
  edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
  edit MODULE_ACTIONS_FIRST ''
  edit MODULE_ACTIONS_HDR 'module_actions.h'
  edit MODULE_CHECKING_HDR 'module_checking.h'
  edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow"'
  edit MODULE_LOAD_SUPPORTED 'True'
  edit MODULE_NAME 'transi'
  edit MODULE_VERSION '%GIT_TAG%'
  edit NETCDF4_VERSION '4.7.4'
  edit ODB_API_GIT_TAG 'develop'
  edit ODB_API_VERSION '0.19.4'
  edit ODB_GIT_TAG 'develop'
  edit ODB_TOOLS_GIT_TAG 'develop'
  edit ODB_TOOLS_VERSION '0.18.3'
  edit ODB_VERSION '1.0.11'
  edit ODC_GIT_TAG 'develop'
  edit ODC_VERSION '1.0.3'
  edit OPENJPEG_VERSION '2.3.1'
  edit PACKAGE 'transi'
  edit PACKAGE_NAME 'transi'
  edit PACKAGE_OWNERS 'nawd@ecmwf.int cgm@ecmwf.int ma0@ecmwf.int'
  edit PACKAGE_VERSION '%GIT_TAG%'
  edit PGEN_VERSION '1.7.1.0'
  edit PKG_BOOST_VERSION '1.71.0'
  edit PREFIX_APPEND ''
  edit PROJ4_VERSION '5.2.0'
  edit PROJ_VERSION '6.3.1'
  edit PYTHON2_VERSION '2.7.15-01'
  edit PYTHON3_VERSION '3.6.10-01'
  edit PYTHON3_VERSION_MAJOR_MINOR '3.6'
  edit PYTHON_GITREPO 'This-needs-tobe-overriden'
  edit PYTHON_PACKAGE 'This-needs-tobe-overriden'
  edit QT_VERSION '5.12.3'
  edit QUEUE 'build'
  edit REMOTE_HOST 'polonius'
  edit REUSE_BUILD_DIR 'False'
  edit SAFE_STACK_SANITIZER 'False'
  edit SANITIZER 'False'
  edit SANITIZER_REFERENCE_DIRECTORY '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
  edit SRCDIR '%BASEDIR%/%PACKAGE%'
  edit STATIC_ANALYSER_ABORT_ON_ERROR 'False'
  edit SWIG_VERSION '3.0.8'
  edit SYNC_RELEASE '2020.04.0'
  edit SYNC_RELEASE_CI_STAGE 'develop'
  edit TARBALL_DIR '/perm/ma/ma0/tmp_tarballs/ci'
  edit TAR_PACKAGE '%PACKAGE%'
  edit TEST 'ctest --output-on-failure -j${CPUS} '
  edit THREAD_SANITIZER 'False'
  edit TRANSI_GIT_TAG 'develop'
  edit TRANSI_VERSION '0.7.1'
  edit TSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
  edit UBSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/ubsan.txt'
  edit UB_SANITIZER 'False'
  edit USER 'ma0'
  edit USE_WORKDIR 'False'
  edit VALGRIND_VERSION '3.15.0'
  edit WORKDIR '%BASEDIR%/workdir/%ECF_NAME%'
  edit WSDIR '/var/tmp/tmpdir/deploy/jtmp.20739'
  label branch "develop"
  task regenerate
    defstatus complete
    edit ECF_JOB_CMD '%ECF_JOB% > %ECF_JOBOUT% 2>&1'
    edit ECF_KILL_CMD 'kill -15 %ECF_RID%'
    edit ECF_PASS 'FREE'
    edit ECF_STATUS_CMD 'ps --sid %ECF_RID% -f'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
    label date ""
    label metabuilder_branch "ci"
  family cca
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DENABLE_GRIB=OFF -DENABLE_MPI=OFF -DLAPACK_FOUND=TRUE'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/transi/cca'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/transi/cca:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cca-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '24000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cca'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'cca'
    edit SITE 'cca'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DENABLE_GRIB=OFF -DENABLE_MPI=OFF -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DENABLE_GRIB=OFF -DENABLE_MPI=OFF -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.17
      edit COMPILER 'INTEL.17'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap intel/17.0.3.053" "module load gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.85
      edit COMPILER 'CRAY.85'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.86
      edit COMPILER 'CRAY.86'
      edit MODULE_ACTIONS '"module swap cdt/17.09" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit COMPILER 'CRAY.87'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family ccb
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DENABLE_GRIB=OFF -DENABLE_MPI=OFF -DLAPACK_FOUND=TRUE'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/transi/ccb'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/transi/ccb:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'ccb-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '24000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'ccb'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'ccb'
    edit SITE 'ccb'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DENABLE_GRIB=OFF -DENABLE_MPI=OFF -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DENABLE_GRIB=OFF -DENABLE_MPI=OFF -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.17
      edit COMPILER 'INTEL.17'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap intel/17.0.3.053" "module load gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.85
      edit COMPILER 'CRAY.85'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.86
      edit COMPILER 'CRAY.86'
      edit MODULE_ACTIONS '"module swap cdt/17.09" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit COMPILER 'CRAY.87'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family cct
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DENABLE_GRIB=OFF -DENABLE_MPI=OFF -DLAPACK_FOUND=TRUE'
    edit CPUS '$EC_total_tasks'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/transi/cct'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/transi/cct:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cct-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '4'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '6000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cct'
    edit QSUB_LP 'QSUB -lP %EC_TOTAL_TASKS%'
    edit QUEUE 'nf'
    edit REMOTE_HOST 'cct'
    edit SITE 'cct'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DENABLE_GRIB=OFF -DENABLE_MPI=OFF -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DENABLE_GRIB=OFF -DENABLE_MPI=OFF -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.17
      edit COMPILER 'INTEL.17'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap intel/17.0.3.053" "module load gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.85
      edit COMPILER 'CRAY.85'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.86
      edit COMPILER 'CRAY.86'
      edit MODULE_ACTIONS '"module swap cdt/17.09" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit COMPILER 'CRAY.87'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family ecgb
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'ecgb'
    edit QUEUE 'express'
    edit REMOTE_HOST 'ecgb'
    edit SITE 'ecgb'
    edit WSDIR '$SCRATCH/metabuilds/%ECF_HOST%_%ECF_PORT%/ecgb/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DENABLE_GRIB=OFF -DENABLE_MPI=OFF'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family leap42
    defstatus suspended
    edit CI_LIMIT '7'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'leap42'
    edit REMOTE_HOST 'leap42-deploy'
    edit SITE 'leap42'
    edit WSDIR '/tmp/metabuilds/%ECF_HOST%_%ECF_PORT%/leap42/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DENABLE_GRIB=OFF -DENABLE_MPI=OFF'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task create_tarball
        trigger test == complete
        edit CMAKE 'cmake'
        edit EXPORTS '"export ECACCOUNT=ecodmdma" "export ecbuild_ROOT=%BASEPREFIX%/ecbuild/develop"'
        label tar ""
        label tardir ""
        label error ""
      task install
        trigger create_tarball == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.93
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DENABLE_GRIB=OFF -DENABLE_MPI=OFF'
      edit COMPILER 'GNU.93'
      edit EXPORTS '"export ECACCOUNT=ecodmdma" "export LD_LIBRARY_PATH=/usr/local/apps/gcc/9.3.0/lib64/"'
      edit MODULE_ACTIONS '"module swap gnu/9.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.19
      edit COMPILER 'INTEL.19'
      edit MODULE_ACTIONS '"module unload gnu" "module load intel/19.0.4" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit POST_EXPORTS '"export CXX=icpc\ -cxxlib=/usr/local/apps/gcc/8.3.0" "export LD_RUN_PATH=/usr/local/apps/gcc/8.3.0/lib64" "export PATH=/usr/local/apps/gcc/8.3.0/bin:$PATH"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DENABLE_GRIB=OFF -DENABLE_MPI=OFF'
      edit COMPILER 'CLANG.90'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90.linters
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DENABLE_GRIB=OFF -DENABLE_MPI=OFF'
      edit COMPILER 'CLANG.90.LINTERS'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task cppcheck
        trigger config == complete
        edit CPPCHECK_FLAGS '--template='[{file}:{line}]: {severity} : {id} : {message}' --suppress=syntaxError --inline-suppr --platform=unix64 --enable=performance --enable=portability --enable=information --std=c11 --std=c++14 -DDEBUG=1 -q -j8'
        edit CPPCHECK_VERSION '1.89'
        edit REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}/cpp_check'
      task iwyu
        trigger cppcheck == complete
        edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%" "module load iwyu/%IWYU_VERSION%"'
      family clang_linters
        trigger iwyu == complete
        limit clang_linters 1
        inlimit clang_linters
        task clang_tidy
          edit REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}/clang_tidy'
        task static_analyser
          edit REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}/static_analyser'
          edit SCAN_BUILD_OPTIONS '--force-analyze-debug-code -o /perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}/static_analyser -k make -j8'
          edit USE_CLANG_CHECK 'False'
          label scan_view ""
      endfamily
      task wipe_build
        trigger clang_linters == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90.address_sanitizer
      edit ADDRESS_SANITIZER 'True'
      edit CMAKE_BUILD_TYPE 'Debug'
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_FLAGS="-fsanitize=address -fno-omit-frame-pointer" -DCMAKE_C_FLAGS="-fsanitize=address -fno-omit-frame-pointer" -DCMAKE_EXE_LINKER_FLAGS=-fsanitize=address -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DENABLE_GRIB=OFF -DENABLE_MPI=OFF'
      edit COMPILER 'CLANG.90.ADDRESS_SANITIZER'
      edit LSAN_OPTIONS 'suppressions=%SRCDIR%/tools/lsan.supp:fast_unwind_on_malloc=0'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit SANITIZER 'True'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit TEST_DIR '%BUILDIR%'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90.ub_sanitizer
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_FLAGS="-fsanitize=undefined -fno-omit-frame-pointer -fno-sanitize-recover=all" -DCMAKE_C_FLAGS="-fsanitize=undefined -fno-omit-frame-pointer" -DCMAKE_EXE_LINKER_FLAGS=-fsanitize=undefined -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DENABLE_GRIB=OFF -DENABLE_MPI=OFF'
      edit COMPILER 'CLANG.90.UB_SANITIZER'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit SANITIZER 'True'
      edit UB_SANITIZER 'True'
      edit UB_SANITIZER_LD_PRELOAD 'export LD_PRELOAD=/usr/local/apps/clang/9.0.1/lib64/libubsan.so'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit TEST_DIR '%BUILDIR%'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family lxc
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%" "module load lapack/%LAPACK_VERSION%"'
    edit M_SITE 'lxc'
    edit QUEUE 'express'
    edit REMOTE_HOST 'lxc'
    edit SITE 'lxc'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxc/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DENABLE_GRIB=OFF -DENABLE_MPI=OFF'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%" "module load lapack/%LAPACK_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family lxg
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%" "module load lapack/%LAPACK_VERSION%"'
    edit M_SITE 'lxg'
    edit QUEUE 'normal'
    edit REMOTE_HOST 'lxg'
    edit SITE 'lxg'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxg/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DENABLE_GRIB=OFF -DENABLE_MPI=OFF'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%" "module load lapack/%LAPACK_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family sappb
    defstatus suspended
    edit ASAN_FILE '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
    edit BASEPREFIX '/autofs/vol/deploy/install/%SITE%'
    edit CI_LIMIT '8'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit M_SITE 'sappb'
    edit REMOTE_HOST 'sappb-deploy'
    edit SANITIZER_REFERENCE_DIRECTORY '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
    edit SITE 'sappb'
    edit TSAN_FILE '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
    edit WSDIR '$HOME/metabuilds/%ECF_HOST%_%ECF_PORT%/sappb/%COMPILER:%'
    family gnu.63
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DENABLE_GRIB=OFF -DENABLE_MPI=OFF'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module load gnu/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
endsuite
suite atlas
  edit ADDRESS_SANITIZER 'False'
  edit ADD_ACTION_TRIGGERS 'True'
  edit AEC_VERSION '1.0.4'
  edit ALLOW_INSTALL_TO_OVERWRITE 'True'
  edit ASAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
  edit ATLAS_GIT_TAG 'develop'
  edit ATLAS_VERSION '0.20.2'
  edit BANNER_TASK_END 'banner %TASK%-end || true'
  edit BANNER_TASK_START 'banner %TASK%-start || true'
  edit BASEDIR '%WSDIR%/%PACKAGE%'
  edit BASEPREFIX '$PERM/install/%SITE%'
  edit BUFRDC_EMOSLIB_VERSION '4.6.3'
  edit BUFRDC_TABLES_GIT_TAG 'develop'
  edit BUFRDC_TABLES_VERSION '000415'
  edit BUILDIR '%BASEDIR%/builds'
  edit BUILD_ALWAYS_PARALLEL 'False'
  edit CFI_SANITIZER 'False'
  edit CFL_BASE_URL 'https://software-test.ecmwf.int/wiki'
  edit CFL_SPACE_KEY '<TODO>'
  edit CFL_TARBALL_COMMENT ''
  edit CFL_TITLE 'Releases'
  edit CHECK_SYNC_RELEASE 'False'
  edit CLANG_TIDY_ABORT_ON_ERROR 'False'
  edit CMAKE '%ECBUILD%'
  edit CMAKELIST_EXTRACT_PACKAGE_VERSION '$(awk '/^project/ && /%PACKAGE%/ && /VERSION/ {for (I=1;I<=NF;I++) if ($I == "VERSION") {print $(I+1)};}' CMakeLists.txt)'
  edit CMAKE_BUILD_TYPE 'Bit'
  edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_TRANS=ON -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
  edit CMAKE_VERSION '3.13.4'
  edit CODES_UI_GIT_TAG 'develop'
  edit CODES_UI_VERSION '1.3.1'
  edit CPACK 'make package'
  edit CPUS '$(lscpu -p | grep -v '#' | wc -l)'
  edit CRAY_FFTW_VERSION '3.3.8.1'
  edit CRAY_NETCDF_VERSION '4.6.1.3'
  edit CRON_BASEPREFIX '%BASEDIR%/install'
  edit CRON_EMAIL ''
  edit CRON_ENABLED 'False'
  edit CRON_GIT_TAG 'develop'
  edit CUSTOM_COMMANDS ''
  edit DATAFLOW_SANITIZER 'False'
  edit DEFAULT_GNU_COMPILER 'gnu.73'
  edit ECBUILD '%BASEPREFIX%/ecbuild/develop/bin/ecbuild'
  edit ECBUILD_CLONE 'False'
  edit ECBUILD_CLONE_COMMIT 'master'
  edit ECBUILD_GIT_TAG 'develop'
  edit ECBUILD_REPO '%GIT_REPO_PREFIX%/ecsdk/ecbuild.git'
  edit ECBUILD_VERSION '3.3.2'
  edit ECCODES_EXTENSIONS_GIT_TAG 'develop'
  edit ECCODES_EXTENSIONS_VERSION '2.17.0'
  edit ECCODES_GIT_TAG 'develop'
  edit ECCODES_PYTHON_VERSION '0.9.6'
  edit ECCODES_VERSION '2.17.1'
  edit ECFLOW5_LATEST_VERSION '5.4.0'
  edit ECFLOW_VERSION '5.4.1'
  edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/atlas'
  edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/atlas'
  edit ECF_JOB_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_JOB% %ECF_JOBOUT%'
  edit ECF_KILL_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% kill'
  edit ECF_STATUS_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% stat'
  edit ECF_TRIES '1'
  edit ECKIT_GIT_TAG 'develop'
  edit ECKIT_VERSION '1.10.1'
  edit ECQUIS_VERSION '1.0.0'
  edit EMOS_MOD_VERSION '462'
  edit EMOS_PA_VERSION '000462'
  edit EXPORTS '"export ECACCOUNT=ecodmdma"'
  edit EXPORT_VARIABLES 'False'
  edit FCKIT_GIT_TAG '0.6.3'
  edit FCKIT_VERSION '0.6.3'
  edit FDB_VINTAGE_GIT_TAG '4.10.0'
  edit FDB_VINTAGE_VERSION '4.10.0'
  edit FFTW_VERSION '3.3.8'
  edit FFTW_VERSION_CRAY '3.3.4.11'
  edit GIT 'git'
  edit GITHUB_PACKAGE_NAME '%PACKAGE_NAME%'
  edit GITHUB_REPO 'git@github.com:ecmwf'
  edit GITREPO '%GIT_REPO_PREFIX%/atlas/atlas.git'
  edit GIT_CLONE 'True'
  edit GIT_CLONE_OPTIONS '--depth=1 --branch %GIT_TAG%'
  edit GIT_METABUILDER_TAG 'ci'
  edit GIT_REPO_PREFIX 'ssh://git@git.ecmwf.int:7999'
  edit GIT_TAG 'develop'
  edit GIT_VERSION '2.26.0'
  edit GRIB_API_VERSION '1.28.0'
  edit INSTALLABLE_TO_USR_LOCAL_APPS 'False'
  edit IWYU_VERSION '0.11'
  edit KSH_EXTRACT_PACKAGE_VERSION '`grep _VERSION_STR VERSION.cmake | grep -o '[0-9]*\.[0-9]*\.[0-9]*'`'
  edit LAPACK_VERSION '3.5.0'
  edit LIBEMOS_GIT_TAG 'develop'
  edit LIBEMOS_VERSION '4.6.2'
  edit LINTERS_REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}'
  edit LSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/lsan.txt'
  edit MAGICS_GIT_TAG 'develop'
  edit MAGICS_PYTHON_VERSION 'master'
  edit MAGICS_VERSION '4.3.3'
  edit MAKE 'make'
  edit MARS_CLIENT_GIT_TAG '6.27.2.2'
  edit MARS_CLIENT_PACKAGE_GIT_TAG '6.27.6'
  edit MARS_CLIENT_PACKAGE_VERSION '6.27.6'
  edit MARS_CLIENT_VERSION '6.27.2.2'
  edit MEMORY_SANITIZER 'False'
  edit METABUILDER_HOME '/var/tmp/ma0/workspace/metabuilder'
  edit METAB_INSTALL_REMOVE_WRITE_PERMISSIONS 'True'
  edit METAB_PYTHON_VERSION 'python3'
  edit METKIT_GIT_TAG 'develop'
  edit METKIT_VERSION '1.4.6'
  edit METVIEW_GIT_TAG 'develop'
  edit METVIEW_PYTHON_VERSION '1.4.2'
  edit METVIEW_VERSION '5.8.3'
  edit MIR_BUNDLE_VERSION '0.9.0.0'
  edit MIR_GIT_TAG 'develop'
  edit MIR_VERSION '1.5.0'
  edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
  edit MODULE_ACTIONS_FIRST ''
  edit MODULE_ACTIONS_HDR 'module_actions.h'
  edit MODULE_CHECKING_HDR 'module_checking.h'
  edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow"'
  edit MODULE_LOAD_SUPPORTED 'True'
  edit MODULE_NAME 'atlas'
  edit MODULE_VERSION '%GIT_TAG%'
  edit NETCDF4_VERSION '4.7.4'
  edit ODB_API_GIT_TAG 'develop'
  edit ODB_API_VERSION '0.19.4'
  edit ODB_GIT_TAG 'develop'
  edit ODB_TOOLS_GIT_TAG 'develop'
  edit ODB_TOOLS_VERSION '0.18.3'
  edit ODB_VERSION '1.0.11'
  edit ODC_GIT_TAG 'develop'
  edit ODC_VERSION '1.0.3'
  edit OPENJPEG_VERSION '2.3.1'
  edit PACKAGE 'atlas'
  edit PACKAGE_NAME 'atlas'
  edit PACKAGE_OWNERS 'nawd@ecmwf.int cgm@ecmwf.int ma0@ecmwf.int'
  edit PACKAGE_VERSION '%GIT_TAG%'
  edit PGEN_VERSION '1.7.1.0'
  edit PKG_BOOST_VERSION '1.71.0'
  edit PREFIX_APPEND ''
  edit PROJ4_VERSION '5.2.0'
  edit PROJ_VERSION '6.3.1'
  edit PYTHON2_VERSION '2.7.15-01'
  edit PYTHON3_VERSION '3.6.10-01'
  edit PYTHON3_VERSION_MAJOR_MINOR '3.6'
  edit PYTHON_GITREPO 'This-needs-tobe-overriden'
  edit PYTHON_PACKAGE 'This-needs-tobe-overriden'
  edit QT_VERSION '5.12.3'
  edit QUEUE 'build'
  edit REMOTE_HOST 'polonius'
  edit REUSE_BUILD_DIR 'False'
  edit SAFE_STACK_SANITIZER 'False'
  edit SANITIZER 'False'
  edit SANITIZER_REFERENCE_DIRECTORY '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
  edit SRCDIR '%BASEDIR%/%PACKAGE%'
  edit STATIC_ANALYSER_ABORT_ON_ERROR 'False'
  edit SWIG_VERSION '3.0.8'
  edit SYNC_RELEASE '2020.04.0'
  edit SYNC_RELEASE_CI_STAGE 'develop'
  edit TARBALL_DIR '/perm/ma/ma0/tmp_tarballs/ci'
  edit TAR_PACKAGE 'Atlas'
  edit TEST 'ctest --output-on-failure -j${CPUS} '
  edit THREAD_SANITIZER 'False'
  edit TRANSI_GIT_TAG 'develop'
  edit TRANSI_VERSION '0.7.1'
  edit TSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
  edit UBSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/ubsan.txt'
  edit UB_SANITIZER 'False'
  edit USER 'ma0'
  edit USE_WORKDIR 'False'
  edit VALGRIND_VERSION '3.15.0'
  edit WORKDIR '%BASEDIR%/workdir/%ECF_NAME%'
  edit WSDIR '/var/tmp/tmpdir/deploy/jtmp.20739'
  label branch "develop"
  task regenerate
    defstatus complete
    edit ECF_JOB_CMD '%ECF_JOB% > %ECF_JOBOUT% 2>&1'
    edit ECF_KILL_CMD 'kill -15 %ECF_RID%'
    edit ECF_PASS 'FREE'
    edit ECF_STATUS_CMD 'ps --sid %ECF_RID% -f'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
    label date ""
    label metabuilder_branch "ci"
  family cca
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_TRANS=ON -DLAPACK_FOUND=TRUE -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/atlas/cca'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/atlas/cca:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cca-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '17000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cca'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'cca'
    edit SITE 'cca'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_TRANS=ON -DLAPACK_FOUND=TRUE -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_TRANS=ON -DLAPACK_FOUND=TRUE -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.17
      edit COMPILER 'INTEL.17'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap intel/17.0.3.053" "module load gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.85
      edit COMPILER 'CRAY.85'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.86
      edit COMPILER 'CRAY.86'
      edit MODULE_ACTIONS '"module swap cdt/17.09" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit COMPILER 'CRAY.87'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family ccb
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_TRANS=ON -DLAPACK_FOUND=TRUE -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/atlas/ccb'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/atlas/ccb:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'ccb-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '17000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'ccb'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'ccb'
    edit SITE 'ccb'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_TRANS=ON -DLAPACK_FOUND=TRUE -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_TRANS=ON -DLAPACK_FOUND=TRUE -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.17
      edit COMPILER 'INTEL.17'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap intel/17.0.3.053" "module load gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.85
      edit COMPILER 'CRAY.85'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.86
      edit COMPILER 'CRAY.86'
      edit MODULE_ACTIONS '"module swap cdt/17.09" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit COMPILER 'CRAY.87'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family cct
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_TRANS=ON -DLAPACK_FOUND=TRUE -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
    edit CPUS '$EC_total_tasks'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/atlas/cct'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/atlas/cct:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cct-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '4'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '4250'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cct'
    edit QSUB_LP 'QSUB -lP %EC_TOTAL_TASKS%'
    edit QUEUE 'nf'
    edit REMOTE_HOST 'cct'
    edit SITE 'cct'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_TRANS=ON -DLAPACK_FOUND=TRUE -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_TRANS=ON -DLAPACK_FOUND=TRUE -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.17
      edit COMPILER 'INTEL.17'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap intel/17.0.3.053" "module load gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.85
      edit COMPILER 'CRAY.85'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.86
      edit COMPILER 'CRAY.86'
      edit MODULE_ACTIONS '"module swap cdt/17.09" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit COMPILER 'CRAY.87'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family ecgb
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'ecgb'
    edit QUEUE 'express'
    edit REMOTE_HOST 'ecgb'
    edit SITE 'ecgb'
    edit WSDIR '$SCRATCH/metabuilds/%ECF_HOST%_%ECF_PORT%/ecgb/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_TRANS=ON -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family leap42
    defstatus suspended
    edit CI_LIMIT '7'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'leap42'
    edit REMOTE_HOST 'leap42-deploy'
    edit SITE 'leap42'
    edit WSDIR '/tmp/metabuilds/%ECF_HOST%_%ECF_PORT%/leap42/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_TRANS=ON -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task create_tarball
        trigger test == complete
        edit CMAKE 'cmake'
        edit EXPORTS '"export ECACCOUNT=ecodmdma" "export ecbuild_ROOT=%BASEPREFIX%/ecbuild/develop"'
        label tar ""
        label tardir ""
        label error ""
      task install
        trigger create_tarball == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.93
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_TRANS=ON -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.93'
      edit EXPORTS '"export ECACCOUNT=ecodmdma" "export LD_LIBRARY_PATH=/usr/local/apps/gcc/9.3.0/lib64/"'
      edit MODULE_ACTIONS '"module swap gnu/9.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.19
      edit COMPILER 'INTEL.19'
      edit MODULE_ACTIONS '"module unload gnu" "module load intel/19.0.4" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit POST_EXPORTS '"export CXX=icpc\ -cxxlib=/usr/local/apps/gcc/8.3.0" "export LD_RUN_PATH=/usr/local/apps/gcc/8.3.0/lib64" "export PATH=/usr/local/apps/gcc/8.3.0/bin:$PATH"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_TRANS=ON -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'CLANG.90'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90.linters
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_TRANS=ON -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'CLANG.90.LINTERS'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task cppcheck
        trigger config == complete
        edit CPPCHECK_FLAGS '--template='[{file}:{line}]: {severity} : {id} : {message}' --suppress=syntaxError --inline-suppr --platform=unix64 --enable=performance --enable=portability --enable=information --std=c11 --std=c++14 -DDEBUG=1 -q -j8'
        edit CPPCHECK_VERSION '1.89'
        edit REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}/cpp_check'
      task iwyu
        trigger cppcheck == complete
        edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%" "module load iwyu/%IWYU_VERSION%"'
      family clang_linters
        trigger iwyu == complete
        limit clang_linters 1
        inlimit clang_linters
        task clang_tidy
          edit REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}/clang_tidy'
        task static_analyser
          edit REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}/static_analyser'
          edit SCAN_BUILD_OPTIONS '--force-analyze-debug-code -o /perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}/static_analyser -k make -j8'
          edit USE_CLANG_CHECK 'False'
          label scan_view ""
      endfamily
      task wipe_build
        trigger clang_linters == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90.address_sanitizer
      edit ADDRESS_SANITIZER 'True'
      edit CMAKE_BUILD_TYPE 'Debug'
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_FLAGS="-fsanitize=address -fno-omit-frame-pointer" -DCMAKE_C_FLAGS="-fsanitize=address -fno-omit-frame-pointer" -DCMAKE_EXE_LINKER_FLAGS=-fsanitize=address -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_TRANS=ON -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'CLANG.90.ADDRESS_SANITIZER'
      edit LSAN_OPTIONS 'suppressions=%SRCDIR%/tools/lsan.supp:fast_unwind_on_malloc=0'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit SANITIZER 'True'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit TEST_DIR '%BUILDIR%'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90.ub_sanitizer
      edit CMAKE_BUILD_TYPE 'Debug'
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_FLAGS="-fsanitize=undefined -fno-omit-frame-pointer -fno-sanitize-recover=all -fsanitize-blacklist=%SRCDIR%/tools/ubsan.blacklist" -DCMAKE_C_FLAGS="-fsanitize=undefined -fno-omit-frame-pointer" -DCMAKE_EXE_LINKER_FLAGS=-fsanitize=undefined -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_TRANS=ON -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'CLANG.90.UB_SANITIZER'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit SANITIZER 'True'
      edit UB_SANITIZER 'True'
      edit UB_SANITIZER_LD_PRELOAD 'export LD_PRELOAD=/usr/local/apps/clang/9.0.1/lib64/libubsan.so'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit TEST_DIR '%BUILDIR%'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family lxc
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'lxc'
    edit QUEUE 'express'
    edit REMOTE_HOST 'lxc'
    edit SITE 'lxc'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxc/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_TRANS=ON -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family lxg
    defstatus suspended
    edit CI_LIMIT '10'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_MKL=ON -DENABLE_TRANS=ON -DMKL_ROOT=/usr/local/apps/intel/parallel_studio_xe_2019/compilers_and_libraries_2019.0.117/linux/mkl -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'lxg'
    edit QUEUE 'normal'
    edit REMOTE_HOST 'lxg'
    edit SITE 'lxg'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxg/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_MKL=ON -DENABLE_TRANS=ON -DMKL_ROOT=/usr/local/apps/intel/parallel_studio_xe_2019/compilers_and_libraries_2019.0.117/linux/mkl -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family sappb
    defstatus suspended
    edit ASAN_FILE '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
    edit BASEPREFIX '/autofs/vol/deploy/install/%SITE%'
    edit CI_LIMIT '8'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit M_SITE 'sappb'
    edit REMOTE_HOST 'sappb-deploy'
    edit SANITIZER_REFERENCE_DIRECTORY '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
    edit SITE 'sappb'
    edit TSAN_FILE '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
    edit WSDIR '$HOME/metabuilds/%ECF_HOST%_%ECF_PORT%/sappb/%COMPILER:%'
    family gnu.63
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_TRANS=ON -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module load gnu/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
endsuite
suite mir
  edit ADDRESS_SANITIZER 'False'
  edit ADD_ACTION_TRIGGERS 'True'
  edit AEC_VERSION '1.0.4'
  edit ALLOW_INSTALL_TO_OVERWRITE 'True'
  edit ASAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
  edit ATLAS_GIT_TAG 'develop'
  edit ATLAS_VERSION '0.20.2'
  edit BANNER_TASK_END 'banner %TASK%-end || true'
  edit BANNER_TASK_START 'banner %TASK%-start || true'
  edit BASEDIR '%WSDIR%/%PACKAGE%'
  edit BASEPREFIX '$PERM/install/%SITE%'
  edit BUFRDC_EMOSLIB_VERSION '4.6.3'
  edit BUFRDC_TABLES_GIT_TAG 'develop'
  edit BUFRDC_TABLES_VERSION '000415'
  edit BUILDIR '%BASEDIR%/builds'
  edit BUILD_ALWAYS_PARALLEL 'False'
  edit CFI_SANITIZER 'False'
  edit CFL_BASE_URL 'https://software-test.ecmwf.int/wiki'
  edit CFL_SPACE_KEY '<TODO>'
  edit CFL_TARBALL_COMMENT ''
  edit CFL_TITLE 'Releases'
  edit CHECK_SYNC_RELEASE 'False'
  edit CLANG_TIDY_ABORT_ON_ERROR 'False'
  edit CMAKE '%ECBUILD%'
  edit CMAKELIST_EXTRACT_PACKAGE_VERSION '$(awk '/^project/ && /%PACKAGE%/ && /VERSION/ {for (I=1;I<=NF;I++) if ($I == "VERSION") {print $(I+1)};}' CMakeLists.txt)'
  edit CMAKE_BUILD_TYPE 'RelWithDebInfo'
  edit CMAKE_VARS '-DATLAS_PATH=%BASEPREFIX%/atlas/develop%INSTALL_POSTFIX% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DECBUILD_LOG_LEVEL=DEBUG -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_MPI=OFF -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
  edit CMAKE_VERSION '3.13.4'
  edit CODES_UI_GIT_TAG 'develop'
  edit CODES_UI_VERSION '1.3.1'
  edit CPACK 'make package'
  edit CPUS '$(lscpu -p | grep -v '#' | wc -l)'
  edit CRAY_FFTW_VERSION '3.3.8.1'
  edit CRAY_NETCDF_VERSION '4.6.1.3'
  edit CRON_BASEPREFIX '%BASEDIR%/install'
  edit CRON_EMAIL ''
  edit CRON_ENABLED 'False'
  edit CRON_GIT_TAG 'develop'
  edit CUSTOM_COMMANDS ''
  edit DATAFLOW_SANITIZER 'False'
  edit DEFAULT_GNU_COMPILER 'gnu.73'
  edit ECBUILD '%BASEPREFIX%/ecbuild/develop/bin/ecbuild'
  edit ECBUILD_CLONE 'False'
  edit ECBUILD_CLONE_COMMIT 'master'
  edit ECBUILD_GIT_TAG 'develop'
  edit ECBUILD_REPO '%GIT_REPO_PREFIX%/ecsdk/ecbuild.git'
  edit ECBUILD_VERSION '3.3.2'
  edit ECCODES_EXTENSIONS_GIT_TAG 'develop'
  edit ECCODES_EXTENSIONS_VERSION '2.17.0'
  edit ECCODES_GIT_TAG 'develop'
  edit ECCODES_PYTHON_VERSION '0.9.6'
  edit ECCODES_VERSION '2.17.1'
  edit ECFLOW5_LATEST_VERSION '5.4.0'
  edit ECFLOW_VERSION '5.4.1'
  edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/mir'
  edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/mir'
  edit ECF_JOB_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_JOB% %ECF_JOBOUT%'
  edit ECF_KILL_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% kill'
  edit ECF_STATUS_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% stat'
  edit ECF_TRIES '1'
  edit ECKIT_GIT_TAG 'develop'
  edit ECKIT_VERSION '1.10.1'
  edit ECQUIS_VERSION '1.0.0'
  edit EMOS_MOD_VERSION '462'
  edit EMOS_PA_VERSION '000462'
  edit EXPORTS '"export ECACCOUNT=ecodmdma" "export ftp_proxy=http://proxy.ecmwf.int:3333" "export http_proxy=http://proxy.ecmwf.int:3333" "export https_proxy=http://proxy.ecmwf.int:3333"'
  edit EXPORT_VARIABLES 'False'
  edit FCKIT_GIT_TAG '0.6.3'
  edit FCKIT_VERSION '0.6.3'
  edit FDB_VINTAGE_GIT_TAG '4.10.0'
  edit FDB_VINTAGE_VERSION '4.10.0'
  edit FFTW_VERSION '3.3.8'
  edit FFTW_VERSION_CRAY '3.3.4.11'
  edit GIT 'git'
  edit GITHUB_PACKAGE_NAME '%PACKAGE_NAME%'
  edit GITHUB_REPO 'git@github.com:ecmwf'
  edit GITREPO '%GIT_REPO_PREFIX%/mir/mir.git'
  edit GIT_CLONE 'True'
  edit GIT_CLONE_OPTIONS '--depth=1 --branch %GIT_TAG%'
  edit GIT_METABUILDER_TAG 'ci'
  edit GIT_REPO_PREFIX 'ssh://git@git.ecmwf.int:7999'
  edit GIT_TAG 'develop'
  edit GIT_VERSION '2.26.0'
  edit GRIB_API_VERSION '1.28.0'
  edit INSTALLABLE_TO_USR_LOCAL_APPS 'False'
  edit IWYU_VERSION '0.11'
  edit KSH_EXTRACT_PACKAGE_VERSION '`grep _VERSION_STR VERSION.cmake | grep -o '[0-9]*\.[0-9]*\.[0-9]*'`'
  edit LAPACK_VERSION '3.5.0'
  edit LIBEMOS_GIT_TAG 'develop'
  edit LIBEMOS_VERSION '4.6.2'
  edit LINTERS_REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}'
  edit LSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/lsan.txt'
  edit MAGICS_GIT_TAG 'develop'
  edit MAGICS_PYTHON_VERSION 'master'
  edit MAGICS_VERSION '4.3.3'
  edit MAKE 'make'
  edit MARS_CLIENT_GIT_TAG '6.27.2.2'
  edit MARS_CLIENT_PACKAGE_GIT_TAG '6.27.6'
  edit MARS_CLIENT_PACKAGE_VERSION '6.27.6'
  edit MARS_CLIENT_VERSION '6.27.2.2'
  edit MEMORY_SANITIZER 'False'
  edit METABUILDER_HOME '/var/tmp/ma0/workspace/metabuilder'
  edit METAB_INSTALL_REMOVE_WRITE_PERMISSIONS 'True'
  edit METAB_PYTHON_VERSION 'python3'
  edit METKIT_GIT_TAG 'develop'
  edit METKIT_VERSION '1.4.6'
  edit METVIEW_GIT_TAG 'develop'
  edit METVIEW_PYTHON_VERSION '1.4.2'
  edit METVIEW_VERSION '5.8.3'
  edit MIR_BUNDLE_VERSION '0.9.0.0'
  edit MIR_GIT_TAG 'develop'
  edit MIR_VERSION '1.5.0'
  edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
  edit MODULE_ACTIONS_FIRST ''
  edit MODULE_ACTIONS_HDR 'module_actions.h'
  edit MODULE_CHECKING_HDR 'module_checking.h'
  edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow"'
  edit MODULE_LOAD_SUPPORTED 'True'
  edit MODULE_NAME 'mir'
  edit MODULE_VERSION '%GIT_TAG%'
  edit NETCDF4_VERSION '4.7.4'
  edit ODB_API_GIT_TAG 'develop'
  edit ODB_API_VERSION '0.19.4'
  edit ODB_GIT_TAG 'develop'
  edit ODB_TOOLS_GIT_TAG 'develop'
  edit ODB_TOOLS_VERSION '0.18.3'
  edit ODB_VERSION '1.0.11'
  edit ODC_GIT_TAG 'develop'
  edit ODC_VERSION '1.0.3'
  edit OPENJPEG_VERSION '2.3.1'
  edit PACKAGE 'mir'
  edit PACKAGE_NAME 'mir'
  edit PACKAGE_OWNERS 'mapm@ecmwf.int matq@ecmwf.int cgm@ecmwf.int ma0@ecmwf.int olivier.iffrig@ecmwf.int'
  edit PACKAGE_VERSION '%MIR_VERSION%'
  edit PGEN_VERSION '1.7.1.0'
  edit PKG_BOOST_VERSION '1.71.0'
  edit PREFIX_APPEND ''
  edit PROJ4_VERSION '5.2.0'
  edit PROJ_VERSION '6.3.1'
  edit PYTHON2_VERSION '2.7.15-01'
  edit PYTHON3_VERSION '3.6.10-01'
  edit PYTHON3_VERSION_MAJOR_MINOR '3.6'
  edit PYTHON_GITREPO 'This-needs-tobe-overriden'
  edit PYTHON_PACKAGE 'This-needs-tobe-overriden'
  edit QT_VERSION '5.12.3'
  edit QUEUE 'build'
  edit REMOTE_HOST 'polonius'
  edit REUSE_BUILD_DIR 'False'
  edit SAFE_STACK_SANITIZER 'False'
  edit SANITIZER 'False'
  edit SANITIZER_REFERENCE_DIRECTORY '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
  edit SRCDIR '%BASEDIR%/%PACKAGE%'
  edit STATIC_ANALYSER_ABORT_ON_ERROR 'False'
  edit SWIG_VERSION '3.0.8'
  edit SYNC_RELEASE '2020.04.0'
  edit SYNC_RELEASE_CI_STAGE 'develop'
  edit TARBALL_DIR '/perm/ma/ma0/tmp_tarballs/ci'
  edit TAR_PACKAGE '%PACKAGE%'
  edit TEST 'ctest -j${CPUS} --output-on-failure'
  edit THREAD_SANITIZER 'False'
  edit TRANSI_GIT_TAG 'develop'
  edit TRANSI_VERSION '0.7.1'
  edit TSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
  edit UBSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/ubsan.txt'
  edit UB_SANITIZER 'False'
  edit USER 'ma0'
  edit USE_WORKDIR 'False'
  edit VALGRIND_VERSION '3.15.0'
  edit WORKDIR '%BASEDIR%/workdir/%ECF_NAME%'
  edit WSDIR '/var/tmp/tmpdir/deploy/jtmp.20739'
  label branch "develop"
  task regenerate
    defstatus complete
    edit ECF_JOB_CMD '%ECF_JOB% > %ECF_JOBOUT% 2>&1'
    edit ECF_KILL_CMD 'kill -15 %ECF_RID%'
    edit ECF_PASS 'FREE'
    edit ECF_STATUS_CMD 'ps --sid %ECF_RID% -f'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
    label date ""
    label metabuilder_branch "ci"
  family cca
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DATLAS_PATH=%BASEPREFIX%/atlas/develop%INSTALL_POSTFIX% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECBUILD_LOG_LEVEL=DEBUG -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_MPI=OFF -DLAPACK_FOUND=TRUE -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/mir/cca'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/mir/cca:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cca-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp" "export ftp_proxy=http://proxy.ecmwf.int:3333" "export http_proxy=http://proxy.ecmwf.int:3333" "export https_proxy=http://proxy.ecmwf.int:3333"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '67000'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%" "module unload craype-broadwell" "module load craype-ivybridge"'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cca'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'cca'
    edit SITE 'cca'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit TEST 'ctest -E "(0016.test|0038.test)" -j${CPUS} --output-on-failure'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DATLAS_PATH=%BASEPREFIX%/atlas/develop%INSTALL_POSTFIX% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DECBUILD_LOG_LEVEL=DEBUG -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_MPI=OFF -DLAPACK_FOUND=TRUE -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%" "module unload craype-broadwell" "module load craype-ivybridge"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task test_memcheck
        defstatus complete
        edit TEST 'ctest -VV -T memcheck -L mir'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DATLAS_PATH=%BASEPREFIX%/atlas/develop%INSTALL_POSTFIX% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DECBUILD_LOG_LEVEL=DEBUG -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_MPI=OFF -DLAPACK_FOUND=TRUE -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%" "module unload craype-broadwell" "module load craype-ivybridge"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task test_memcheck
        defstatus complete
        edit TEST 'ctest -VV -T memcheck -L mir'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.17
      edit COMPILER 'INTEL.17'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap intel/17.0.3.053" "module load gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%" "module unload craype-broadwell" "module load craype-ivybridge" "module load gcc"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task test_memcheck
        defstatus complete
        edit TEST 'ctest -VV -T memcheck -L mir'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.85
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit CMAKE_VARS '-DATLAS_PATH=%BASEPREFIX%/atlas/develop%INSTALL_POSTFIX% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECBUILD_LOG_LEVEL=DEBUG -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_MPI=OFF -DENABLE_PYTHON=OFF -DLAPACK_FOUND=TRUE -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'CRAY.85'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%" "module unload craype-broadwell" "module load craype-ivybridge"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task test_memcheck
        defstatus complete
        edit TEST 'ctest -VV -T memcheck -L mir'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.86
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit CMAKE_VARS '-DATLAS_PATH=%BASEPREFIX%/atlas/develop%INSTALL_POSTFIX% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECBUILD_LOG_LEVEL=DEBUG -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_MPI=OFF -DENABLE_PYTHON=OFF -DLAPACK_FOUND=TRUE -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'CRAY.86'
      edit MODULE_ACTIONS '"module swap cdt/17.09" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%" "module unload craype-broadwell" "module load craype-ivybridge"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task test_memcheck
        defstatus complete
        edit TEST 'ctest -VV -T memcheck -L mir'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit CMAKE_VARS '-DATLAS_PATH=%BASEPREFIX%/atlas/develop%INSTALL_POSTFIX% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECBUILD_LOG_LEVEL=DEBUG -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_MPI=OFF -DENABLE_PYTHON=OFF -DLAPACK_FOUND=TRUE -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'CRAY.87'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%" "module unload craype-broadwell" "module load craype-ivybridge"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task test_memcheck
        defstatus complete
        edit TEST 'ctest -VV -T memcheck -L mir'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family ccb
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DATLAS_PATH=%BASEPREFIX%/atlas/develop%INSTALL_POSTFIX% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECBUILD_LOG_LEVEL=DEBUG -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_MPI=OFF -DLAPACK_FOUND=TRUE -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/mir/ccb'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/mir/ccb:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'ccb-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp" "export ftp_proxy=http://proxy.ecmwf.int:3333" "export http_proxy=http://proxy.ecmwf.int:3333" "export https_proxy=http://proxy.ecmwf.int:3333"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '67000'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%" "module unload craype-broadwell" "module load craype-ivybridge"'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'ccb'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'ccb'
    edit SITE 'ccb'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit TEST 'ctest -E "(0016.test|0038.test)" -j${CPUS} --output-on-failure'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DATLAS_PATH=%BASEPREFIX%/atlas/develop%INSTALL_POSTFIX% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DECBUILD_LOG_LEVEL=DEBUG -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_MPI=OFF -DLAPACK_FOUND=TRUE -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%" "module unload craype-broadwell" "module load craype-ivybridge"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task test_memcheck
        defstatus complete
        edit TEST 'ctest -VV -T memcheck -L mir'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DATLAS_PATH=%BASEPREFIX%/atlas/develop%INSTALL_POSTFIX% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DECBUILD_LOG_LEVEL=DEBUG -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_MPI=OFF -DLAPACK_FOUND=TRUE -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%" "module unload craype-broadwell" "module load craype-ivybridge"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task test_memcheck
        defstatus complete
        edit TEST 'ctest -VV -T memcheck -L mir'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.17
      edit COMPILER 'INTEL.17'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap intel/17.0.3.053" "module load gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%" "module unload craype-broadwell" "module load craype-ivybridge" "module load gcc"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task test_memcheck
        defstatus complete
        edit TEST 'ctest -VV -T memcheck -L mir'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.85
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit CMAKE_VARS '-DATLAS_PATH=%BASEPREFIX%/atlas/develop%INSTALL_POSTFIX% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECBUILD_LOG_LEVEL=DEBUG -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_MPI=OFF -DENABLE_PYTHON=OFF -DLAPACK_FOUND=TRUE -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'CRAY.85'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%" "module unload craype-broadwell" "module load craype-ivybridge"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task test_memcheck
        defstatus complete
        edit TEST 'ctest -VV -T memcheck -L mir'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.86
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit CMAKE_VARS '-DATLAS_PATH=%BASEPREFIX%/atlas/develop%INSTALL_POSTFIX% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECBUILD_LOG_LEVEL=DEBUG -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_MPI=OFF -DENABLE_PYTHON=OFF -DLAPACK_FOUND=TRUE -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'CRAY.86'
      edit MODULE_ACTIONS '"module swap cdt/17.09" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%" "module unload craype-broadwell" "module load craype-ivybridge"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task test_memcheck
        defstatus complete
        edit TEST 'ctest -VV -T memcheck -L mir'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit CMAKE_VARS '-DATLAS_PATH=%BASEPREFIX%/atlas/develop%INSTALL_POSTFIX% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECBUILD_LOG_LEVEL=DEBUG -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_MPI=OFF -DENABLE_PYTHON=OFF -DLAPACK_FOUND=TRUE -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'CRAY.87'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%" "module unload craype-broadwell" "module load craype-ivybridge"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task test_memcheck
        defstatus complete
        edit TEST 'ctest -VV -T memcheck -L mir'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family cct
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DATLAS_PATH=%BASEPREFIX%/atlas/develop%INSTALL_POSTFIX% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECBUILD_LOG_LEVEL=DEBUG -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_MPI=OFF -DLAPACK_FOUND=TRUE -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
    edit CPUS '$EC_total_tasks'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/mir/cct'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/mir/cct:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cct-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '4'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp" "export ftp_proxy=http://proxy.ecmwf.int:3333" "export http_proxy=http://proxy.ecmwf.int:3333" "export https_proxy=http://proxy.ecmwf.int:3333"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '16750'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%" "module unload craype-broadwell" "module load craype-ivybridge"'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cct'
    edit QSUB_LP 'QSUB -lP %EC_TOTAL_TASKS%'
    edit QUEUE 'nf'
    edit REMOTE_HOST 'cct'
    edit SITE 'cct'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit TEST 'ctest -E "(0016.test|0038.test)" -j${CPUS} --output-on-failure'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DATLAS_PATH=%BASEPREFIX%/atlas/develop%INSTALL_POSTFIX% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DECBUILD_LOG_LEVEL=DEBUG -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_MPI=OFF -DLAPACK_FOUND=TRUE -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%" "module unload craype-broadwell" "module load craype-ivybridge"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task test_memcheck
        defstatus complete
        edit TEST 'ctest -VV -T memcheck -L mir'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DATLAS_PATH=%BASEPREFIX%/atlas/develop%INSTALL_POSTFIX% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DECBUILD_LOG_LEVEL=DEBUG -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_MPI=OFF -DLAPACK_FOUND=TRUE -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%" "module unload craype-broadwell" "module load craype-ivybridge"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task test_memcheck
        defstatus complete
        edit TEST 'ctest -VV -T memcheck -L mir'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.17
      edit COMPILER 'INTEL.17'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap intel/17.0.3.053" "module load gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%" "module unload craype-broadwell" "module load craype-ivybridge" "module load gcc"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task test_memcheck
        defstatus complete
        edit TEST 'ctest -VV -T memcheck -L mir'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.85
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit CMAKE_VARS '-DATLAS_PATH=%BASEPREFIX%/atlas/develop%INSTALL_POSTFIX% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECBUILD_LOG_LEVEL=DEBUG -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_MPI=OFF -DENABLE_PYTHON=OFF -DLAPACK_FOUND=TRUE -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'CRAY.85'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%" "module unload craype-broadwell" "module load craype-ivybridge"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task test_memcheck
        defstatus complete
        edit TEST 'ctest -VV -T memcheck -L mir'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.86
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit CMAKE_VARS '-DATLAS_PATH=%BASEPREFIX%/atlas/develop%INSTALL_POSTFIX% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECBUILD_LOG_LEVEL=DEBUG -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_MPI=OFF -DENABLE_PYTHON=OFF -DLAPACK_FOUND=TRUE -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'CRAY.86'
      edit MODULE_ACTIONS '"module swap cdt/17.09" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%" "module unload craype-broadwell" "module load craype-ivybridge"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task test_memcheck
        defstatus complete
        edit TEST 'ctest -VV -T memcheck -L mir'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit CMAKE_VARS '-DATLAS_PATH=%BASEPREFIX%/atlas/develop%INSTALL_POSTFIX% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECBUILD_LOG_LEVEL=DEBUG -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_MPI=OFF -DENABLE_PYTHON=OFF -DLAPACK_FOUND=TRUE -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'CRAY.87'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%" "module unload craype-broadwell" "module load craype-ivybridge"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task test_memcheck
        defstatus complete
        edit TEST 'ctest -VV -T memcheck -L mir'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family ecgb
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'ecgb'
    edit QUEUE 'express'
    edit REMOTE_HOST 'ecgb'
    edit SITE 'ecgb'
    edit WSDIR '$SCRATCH/metabuilds/%ECF_HOST%_%ECF_PORT%/ecgb/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DATLAS_PATH=%BASEPREFIX%/atlas/develop%INSTALL_POSTFIX% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECBUILD_LOG_LEVEL=DEBUG -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_MPI=OFF -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task test_memcheck
        defstatus complete
        edit TEST 'ctest -VV -T memcheck -L mir'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family leap42
    defstatus suspended
    edit CI_LIMIT '7'
    edit CLANG_TIDY_ABORT_ON_ERROR 'True'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'leap42'
    edit REMOTE_HOST 'leap42-deploy'
    edit SITE 'leap42'
    edit WSDIR '/tmp/metabuilds/%ECF_HOST%_%ECF_PORT%/leap42/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DATLAS_PATH=%BASEPREFIX%/atlas/develop%INSTALL_POSTFIX% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECBUILD_LOG_LEVEL=DEBUG -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_MPI=OFF -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task create_tarball
        trigger build == complete
        edit CMAKE 'cmake'
        edit EXPORTS '"export ECACCOUNT=ecodmdma" "export ecbuild_ROOT=%BASEPREFIX%/ecbuild/develop" "export ftp_proxy=http://proxy.ecmwf.int:3333" "export http_proxy=http://proxy.ecmwf.int:3333" "export https_proxy=http://proxy.ecmwf.int:3333"'
        label tar ""
        label tardir ""
        label error ""
      task test
        trigger create_tarball == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task test_memcheck
        defstatus complete
        edit TEST 'ctest -VV -T memcheck -L mir'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.93
      edit CMAKE_VARS '-DATLAS_PATH=%BASEPREFIX%/atlas/develop%INSTALL_POSTFIX% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECBUILD_LOG_LEVEL=DEBUG -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_MPI=OFF -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.93'
      edit EXPORTS '"export ECACCOUNT=ecodmdma" "export LD_LIBRARY_PATH=/usr/local/apps/gcc/9.3.0/lib64/" "export ftp_proxy=http://proxy.ecmwf.int:3333" "export http_proxy=http://proxy.ecmwf.int:3333" "export https_proxy=http://proxy.ecmwf.int:3333"'
      edit MODULE_ACTIONS '"module swap gnu/9.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task test_memcheck
        defstatus complete
        edit TEST 'ctest -VV -T memcheck -L mir'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.19
      edit COMPILER 'INTEL.19'
      edit MODULE_ACTIONS '"module unload gnu" "module load intel/19.0.4" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit POST_EXPORTS '"export CXX=icpc\ -cxxlib=/usr/local/apps/gcc/8.3.0" "export LD_RUN_PATH=/usr/local/apps/gcc/8.3.0/lib64" "export PATH=/usr/local/apps/gcc/8.3.0/bin:$PATH"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task test_memcheck
        defstatus complete
        edit TEST 'ctest -VV -T memcheck -L mir'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90
      edit CMAKE_VARS '-DATLAS_PATH=%BASEPREFIX%/atlas/develop%INSTALL_POSTFIX% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECBUILD_LOG_LEVEL=DEBUG -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_MPI=OFF -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'CLANG.90'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task test_memcheck
        defstatus complete
        edit TEST 'ctest -VV -T memcheck -L mir'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90.linters
      edit CMAKE_VARS '-DATLAS_PATH=%BASEPREFIX%/atlas/develop%INSTALL_POSTFIX% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECBUILD_LOG_LEVEL=DEBUG -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_MPI=OFF -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'CLANG.90.LINTERS'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task cppcheck
        trigger config == complete
        edit CPPCHECK_FLAGS '--template='[{file}:{line}]: {severity} : {id} : {message}' --suppress=syntaxError --inline-suppr --platform=unix64 --enable=performance --enable=portability --enable=information --std=c11 --std=c++14 -DDEBUG=1 -q -j8'
        edit CPPCHECK_VERSION '1.89'
        edit REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}/cpp_check'
      task iwyu
        trigger cppcheck == complete
        edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%" "module load iwyu/%IWYU_VERSION%"'
      family clang_linters
        trigger iwyu == complete
        limit clang_linters 1
        inlimit clang_linters
        task clang_tidy
          edit REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}/clang_tidy'
        task static_analyser
          edit REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}/static_analyser'
          edit SCAN_BUILD_OPTIONS '--force-analyze-debug-code -o /perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}/static_analyser -k make -j8'
          edit USE_CLANG_CHECK 'False'
          label scan_view ""
      endfamily
      task wipe_build
        trigger clang_linters == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90.address_sanitizer
      edit ADDRESS_SANITIZER 'True'
      edit CMAKE_VARS '-DATLAS_PATH=%BASEPREFIX%/atlas/develop%INSTALL_POSTFIX% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_FLAGS="-fsanitize=address -fno-omit-frame-pointer" -DCMAKE_C_FLAGS="-fsanitize=address -fno-omit-frame-pointer" -DCMAKE_EXE_LINKER_FLAGS=-fsanitize=address -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECBUILD_LOG_LEVEL=DEBUG -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_MPI=OFF -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'CLANG.90.ADDRESS_SANITIZER'
      edit LSAN_OPTIONS 'suppressions=%SRCDIR%/tests/lsan-leap42.supp'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit SANITIZER 'True'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit TEST_DIR '%BUILDIR%'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90.ub_sanitizer
      edit CMAKE_VARS '-DATLAS_PATH=%BASEPREFIX%/atlas/develop%INSTALL_POSTFIX% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_FLAGS="-fsanitize=undefined -fno-omit-frame-pointer -fno-sanitize-recover=all" -DCMAKE_C_FLAGS="-fsanitize=undefined -fno-omit-frame-pointer" -DCMAKE_EXE_LINKER_FLAGS=-fsanitize=undefined -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECBUILD_LOG_LEVEL=DEBUG -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_MPI=OFF -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'CLANG.90.UB_SANITIZER'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit SANITIZER 'True'
      edit UB_SANITIZER 'True'
      edit UB_SANITIZER_LD_PRELOAD 'export LD_PRELOAD=/usr/local/apps/clang/9.0.1/lib64/libubsan.so'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit TEST_DIR '%BUILDIR%'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family lxc
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'lxc'
    edit QUEUE 'express'
    edit REMOTE_HOST 'lxc'
    edit SITE 'lxc'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxc/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DATLAS_PATH=%BASEPREFIX%/atlas/develop%INSTALL_POSTFIX% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECBUILD_LOG_LEVEL=DEBUG -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_MPI=OFF -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task test_memcheck
        defstatus complete
        edit TEST 'ctest -VV -T memcheck -L mir'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family lxg
    defstatus suspended
    edit CI_LIMIT '10'
    edit CMAKE_VARS '-DATLAS_PATH=%BASEPREFIX%/atlas/develop%INSTALL_POSTFIX% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DECBUILD_LOG_LEVEL=DEBUG -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_MPI=OFF -DLAPACK_LIBRARIES=/usr/local/apps/lapack/3.5.0/LP64/lib/liblapack.so -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%" "module load eigen"'
    edit M_SITE 'lxg'
    edit QUEUE 'normal'
    edit REMOTE_HOST 'lxg'
    edit SITE 'lxg'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxg/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DATLAS_PATH=%BASEPREFIX%/atlas/develop%INSTALL_POSTFIX% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECBUILD_LOG_LEVEL=DEBUG -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_MPI=OFF -DLAPACK_LIBRARIES=/usr/local/apps/lapack/3.5.0/LP64/lib/liblapack.so -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%" "module load eigen"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task test_memcheck
        defstatus complete
        edit TEST 'ctest -VV -T memcheck -L mir'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family sappb
    defstatus suspended
    edit ASAN_FILE '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
    edit BASEPREFIX '/autofs/vol/deploy/install/%SITE%'
    edit CI_LIMIT '8'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit M_SITE 'sappb'
    edit REMOTE_HOST 'sappb-deploy'
    edit SANITIZER_REFERENCE_DIRECTORY '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
    edit SITE 'sappb'
    edit TSAN_FILE '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
    edit WSDIR '$HOME/metabuilds/%ECF_HOST%_%ECF_PORT%/sappb/%COMPILER:%'
    family gnu.63
      edit CMAKE_VARS '-DATLAS_PATH=%BASEPREFIX%/atlas/develop%INSTALL_POSTFIX% -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECBUILD_LOG_LEVEL=DEBUG -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_MPI=OFF -DTRANSI_PATH=%BASEPREFIX%/transi/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module load gnu/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task test_memcheck
        defstatus complete
        edit TEST 'ctest -VV -T memcheck -L mir'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
endsuite
suite odb
  edit ADDRESS_SANITIZER 'False'
  edit ADD_ACTION_TRIGGERS 'True'
  edit AEC_VERSION '1.0.4'
  edit ALLOW_INSTALL_TO_OVERWRITE 'True'
  edit ASAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
  edit ATLAS_GIT_TAG 'develop'
  edit ATLAS_VERSION '0.20.2'
  edit BANNER_TASK_END 'banner %TASK%-end || true'
  edit BANNER_TASK_START 'banner %TASK%-start || true'
  edit BASEDIR '%WSDIR%/%PACKAGE%'
  edit BASEPREFIX '$PERM/install/%SITE%'
  edit BUFRDC_EMOSLIB_VERSION '4.6.3'
  edit BUFRDC_TABLES_GIT_TAG 'develop'
  edit BUFRDC_TABLES_VERSION '000415'
  edit BUILDIR '%BASEDIR%/builds'
  edit BUILD_ALWAYS_PARALLEL 'False'
  edit CFI_SANITIZER 'False'
  edit CFL_BASE_URL 'https://software-test.ecmwf.int/wiki'
  edit CFL_SPACE_KEY '<TODO>'
  edit CFL_TARBALL_COMMENT ''
  edit CFL_TITLE 'Releases'
  edit CHECK_SYNC_RELEASE 'False'
  edit CLANG_TIDY_ABORT_ON_ERROR 'False'
  edit CMAKE '%ECBUILD%'
  edit CMAKELIST_EXTRACT_PACKAGE_VERSION '$(awk '/^project/ && /%PACKAGE%/ && /VERSION/ {for (I=1;I<=NF;I++) if ($I == "VERSION") {print $(I+1)};}' CMakeLists.txt)'
  edit CMAKE_BUILD_TYPE 'RelWithDebInfo'
  edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DODB_SCHEMAS=ECMA'
  edit CMAKE_VERSION '3.13.4'
  edit CODES_UI_GIT_TAG 'develop'
  edit CODES_UI_VERSION '1.3.1'
  edit CPACK 'make package'
  edit CPUS '$(lscpu -p | grep -v '#' | wc -l)'
  edit CRAY_FFTW_VERSION '3.3.8.1'
  edit CRAY_NETCDF_VERSION '4.6.1.3'
  edit CRON_BASEPREFIX '%BASEDIR%/install'
  edit CRON_EMAIL ''
  edit CRON_ENABLED 'False'
  edit CRON_GIT_TAG 'develop'
  edit CUSTOM_COMMANDS ''
  edit DATAFLOW_SANITIZER 'False'
  edit DEFAULT_GNU_COMPILER 'gnu.73'
  edit ECBUILD '%BASEPREFIX%/ecbuild/develop/bin/ecbuild'
  edit ECBUILD_CLONE 'False'
  edit ECBUILD_CLONE_COMMIT 'master'
  edit ECBUILD_GIT_TAG 'develop'
  edit ECBUILD_REPO '%GIT_REPO_PREFIX%/ecsdk/ecbuild.git'
  edit ECBUILD_VERSION '3.3.2'
  edit ECCODES_EXTENSIONS_GIT_TAG 'develop'
  edit ECCODES_EXTENSIONS_VERSION '2.17.0'
  edit ECCODES_GIT_TAG 'develop'
  edit ECCODES_PYTHON_VERSION '0.9.6'
  edit ECCODES_VERSION '2.17.1'
  edit ECFLOW5_LATEST_VERSION '5.4.0'
  edit ECFLOW_VERSION '5.4.1'
  edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/odb'
  edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/odb'
  edit ECF_JOB_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_JOB% %ECF_JOBOUT%'
  edit ECF_KILL_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% kill'
  edit ECF_STATUS_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% stat'
  edit ECF_TRIES '1'
  edit ECKIT_GIT_TAG 'develop'
  edit ECKIT_VERSION '1.10.1'
  edit ECQUIS_VERSION '1.0.0'
  edit EMOS_MOD_VERSION '462'
  edit EMOS_PA_VERSION '000462'
  edit EXPORTS '"export ECACCOUNT=ecodmdma"'
  edit EXPORT_VARIABLES 'False'
  edit FCKIT_GIT_TAG '0.6.3'
  edit FCKIT_VERSION '0.6.3'
  edit FDB_VINTAGE_GIT_TAG '4.10.0'
  edit FDB_VINTAGE_VERSION '4.10.0'
  edit FFTW_VERSION '3.3.8'
  edit FFTW_VERSION_CRAY '3.3.4.11'
  edit GIT 'git'
  edit GITHUB_PACKAGE_NAME '%PACKAGE_NAME%'
  edit GITHUB_REPO 'git@github.com:ecmwf'
  edit GITREPO '%GIT_REPO_PREFIX%/odb/odb.git'
  edit GIT_CLONE 'True'
  edit GIT_CLONE_OPTIONS '--depth=1 --branch %GIT_TAG%'
  edit GIT_METABUILDER_TAG 'ci'
  edit GIT_REPO_PREFIX 'ssh://git@git.ecmwf.int:7999'
  edit GIT_TAG 'develop'
  edit GIT_VERSION '2.26.0'
  edit GRIB_API_VERSION '1.28.0'
  edit INSTALLABLE_TO_USR_LOCAL_APPS 'True'
  edit IWYU_VERSION '0.11'
  edit KSH_EXTRACT_PACKAGE_VERSION '`grep _VERSION_STR VERSION.cmake | grep -o '[0-9]*\.[0-9]*\.[0-9]*'`'
  edit LAPACK_VERSION '3.5.0'
  edit LIBEMOS_GIT_TAG 'develop'
  edit LIBEMOS_VERSION '4.6.2'
  edit LINTERS_REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}'
  edit LSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/lsan.txt'
  edit MAGICS_GIT_TAG 'develop'
  edit MAGICS_PYTHON_VERSION 'master'
  edit MAGICS_VERSION '4.3.3'
  edit MAKE 'make'
  edit MARS_CLIENT_GIT_TAG '6.27.2.2'
  edit MARS_CLIENT_PACKAGE_GIT_TAG '6.27.6'
  edit MARS_CLIENT_PACKAGE_VERSION '6.27.6'
  edit MARS_CLIENT_VERSION '6.27.2.2'
  edit MEMORY_SANITIZER 'False'
  edit METABUILDER_HOME '/var/tmp/ma0/workspace/metabuilder'
  edit METAB_INSTALL_REMOVE_WRITE_PERMISSIONS 'True'
  edit METAB_PYTHON_VERSION 'python3'
  edit METKIT_GIT_TAG 'develop'
  edit METKIT_VERSION '1.4.6'
  edit METVIEW_GIT_TAG 'develop'
  edit METVIEW_PYTHON_VERSION '1.4.2'
  edit METVIEW_VERSION '5.8.3'
  edit MIR_BUNDLE_VERSION '0.9.0.0'
  edit MIR_GIT_TAG 'develop'
  edit MIR_VERSION '1.5.0'
  edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
  edit MODULE_ACTIONS_FIRST ''
  edit MODULE_ACTIONS_HDR 'module_actions.h'
  edit MODULE_CHECKING_HDR 'module_checking.h'
  edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow"'
  edit MODULE_LOAD_SUPPORTED 'True'
  edit MODULE_NAME 'odb'
  edit MODULE_VERSION '%GIT_TAG%'
  edit NETCDF4_VERSION '4.7.4'
  edit ODB_API_GIT_TAG 'develop'
  edit ODB_API_VERSION '0.19.4'
  edit ODB_GIT_TAG 'develop'
  edit ODB_TOOLS_GIT_TAG 'develop'
  edit ODB_TOOLS_VERSION '0.18.3'
  edit ODB_VERSION '1.0.11'
  edit ODC_GIT_TAG 'develop'
  edit ODC_VERSION '1.0.3'
  edit OPENJPEG_VERSION '2.3.1'
  edit PACKAGE 'odb'
  edit PACKAGE_NAME 'odb'
  edit PACKAGE_OWNERS 'datk@ecmwf.int cgm@ecmwf.int ma0@ecmwf.int'
  edit PACKAGE_VERSION '%GIT_TAG%'
  edit PGEN_VERSION '1.7.1.0'
  edit PKG_BOOST_VERSION '1.71.0'
  edit PREFIX_APPEND ''
  edit PROJ4_VERSION '5.2.0'
  edit PROJ_VERSION '6.3.1'
  edit PYTHON2_VERSION '2.7.15-01'
  edit PYTHON3_VERSION '3.6.10-01'
  edit PYTHON3_VERSION_MAJOR_MINOR '3.6'
  edit PYTHON_GITREPO 'This-needs-tobe-overriden'
  edit PYTHON_PACKAGE 'This-needs-tobe-overriden'
  edit QT_VERSION '5.12.3'
  edit QUEUE 'build'
  edit REMOTE_HOST 'polonius'
  edit REUSE_BUILD_DIR 'False'
  edit SAFE_STACK_SANITIZER 'False'
  edit SANITIZER 'False'
  edit SANITIZER_REFERENCE_DIRECTORY '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
  edit SRCDIR '%BASEDIR%/%PACKAGE%'
  edit STATIC_ANALYSER_ABORT_ON_ERROR 'False'
  edit SWIG_VERSION '3.0.8'
  edit SYNC_RELEASE '2020.04.0'
  edit SYNC_RELEASE_CI_STAGE 'develop'
  edit TARBALL_DIR '/perm/ma/ma0/tmp_tarballs/ci'
  edit TAR_PACKAGE '%PACKAGE%'
  edit TEST 'ctest --output-on-failure -j${CPUS} '
  edit THREAD_SANITIZER 'False'
  edit TRANSI_GIT_TAG 'develop'
  edit TRANSI_VERSION '0.7.1'
  edit TSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
  edit UBSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/ubsan.txt'
  edit UB_SANITIZER 'False'
  edit USER 'ma0'
  edit USE_WORKDIR 'False'
  edit VALGRIND_VERSION '3.15.0'
  edit WORKDIR '%BASEDIR%/workdir/%ECF_NAME%'
  edit WSDIR '/var/tmp/tmpdir/deploy/jtmp.20739'
  label branch "develop"
  task regenerate
    defstatus complete
    edit ECF_JOB_CMD '%ECF_JOB% > %ECF_JOBOUT% 2>&1'
    edit ECF_KILL_CMD 'kill -15 %ECF_RID%'
    edit ECF_PASS 'FREE'
    edit ECF_STATUS_CMD 'ps --sid %ECF_RID% -f'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
    label date ""
    label metabuilder_branch "ci"
  family cca
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DLAPACK_FOUND=TRUE -DODB_SCHEMAS=ECMA'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/odb/cca'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/odb/cca:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cca-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '24000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cca'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'cca'
    edit SITE 'cca'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DLAPACK_FOUND=TRUE -DODB_SCHEMAS=ECMA'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DLAPACK_FOUND=TRUE -DODB_SCHEMAS=ECMA'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.17
      edit COMPILER 'INTEL.17'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap intel/17.0.3.053" "module load gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.85
      edit COMPILER 'CRAY.85'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.86
      edit COMPILER 'CRAY.86'
      edit MODULE_ACTIONS '"module swap cdt/17.09" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit COMPILER 'CRAY.87'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family ccb
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DLAPACK_FOUND=TRUE -DODB_SCHEMAS=ECMA'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/odb/ccb'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/odb/ccb:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'ccb-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '24000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'ccb'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'ccb'
    edit SITE 'ccb'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DLAPACK_FOUND=TRUE -DODB_SCHEMAS=ECMA'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DLAPACK_FOUND=TRUE -DODB_SCHEMAS=ECMA'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.17
      edit COMPILER 'INTEL.17'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap intel/17.0.3.053" "module load gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.85
      edit COMPILER 'CRAY.85'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.86
      edit COMPILER 'CRAY.86'
      edit MODULE_ACTIONS '"module swap cdt/17.09" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit COMPILER 'CRAY.87'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family cct
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DLAPACK_FOUND=TRUE -DODB_SCHEMAS=ECMA'
    edit CPUS '$EC_total_tasks'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/odb/cct'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/odb/cct:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cct-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '4'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '6000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cct'
    edit QSUB_LP 'QSUB -lP %EC_TOTAL_TASKS%'
    edit QUEUE 'nf'
    edit REMOTE_HOST 'cct'
    edit SITE 'cct'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DLAPACK_FOUND=TRUE -DODB_SCHEMAS=ECMA'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DLAPACK_FOUND=TRUE -DODB_SCHEMAS=ECMA'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.17
      edit COMPILER 'INTEL.17'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap intel/17.0.3.053" "module load gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.85
      edit COMPILER 'CRAY.85'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.86
      edit COMPILER 'CRAY.86'
      edit MODULE_ACTIONS '"module swap cdt/17.09" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit COMPILER 'CRAY.87'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family ecgb
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'ecgb'
    edit QUEUE 'express'
    edit REMOTE_HOST 'ecgb'
    edit SITE 'ecgb'
    edit WSDIR '$SCRATCH/metabuilds/%ECF_HOST%_%ECF_PORT%/ecgb/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DODB_SCHEMAS=ECMA'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family leap42
    defstatus suspended
    edit CI_LIMIT '7'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'leap42'
    edit REMOTE_HOST 'leap42-deploy'
    edit SITE 'leap42'
    edit WSDIR '/tmp/metabuilds/%ECF_HOST%_%ECF_PORT%/leap42/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DODB_SCHEMAS=ECMA'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task create_tarball
        trigger test == complete
        edit CMAKE 'cmake'
        edit EXPORTS '"export ECACCOUNT=ecodmdma" "export ecbuild_ROOT=%BASEPREFIX%/ecbuild/develop"'
        label tar ""
        label tardir ""
        label error ""
      task install
        trigger create_tarball == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.93
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DODB_SCHEMAS=ECMA'
      edit COMPILER 'GNU.93'
      edit EXPORTS '"export ECACCOUNT=ecodmdma" "export LD_LIBRARY_PATH=/usr/local/apps/gcc/9.3.0/lib64/"'
      edit MODULE_ACTIONS '"module swap gnu/9.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DODB_SCHEMAS=ECMA'
      edit COMPILER 'CLANG.90'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family lxc
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'lxc'
    edit QUEUE 'express'
    edit REMOTE_HOST 'lxc'
    edit SITE 'lxc'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxc/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DODB_SCHEMAS=ECMA'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family lxg
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'lxg'
    edit QUEUE 'normal'
    edit REMOTE_HOST 'lxg'
    edit SITE 'lxg'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxg/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DODB_SCHEMAS=ECMA'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
endsuite
suite odb_api
  edit ADDRESS_SANITIZER 'False'
  edit ADD_ACTION_TRIGGERS 'True'
  edit AEC_VERSION '1.0.4'
  edit ALLOW_INSTALL_TO_OVERWRITE 'True'
  edit ASAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
  edit ATLAS_GIT_TAG 'develop'
  edit ATLAS_VERSION '0.20.2'
  edit BANNER_TASK_END 'banner %TASK%-end || true'
  edit BANNER_TASK_START 'banner %TASK%-start || true'
  edit BASEDIR '%WSDIR%/%PACKAGE%'
  edit BASEPREFIX '$PERM/install/%SITE%'
  edit BUFRDC_EMOSLIB_VERSION '4.6.3'
  edit BUFRDC_TABLES_GIT_TAG 'develop'
  edit BUFRDC_TABLES_VERSION '000415'
  edit BUILDIR '%BASEDIR%/builds'
  edit BUILD_ALWAYS_PARALLEL 'False'
  edit CFI_SANITIZER 'False'
  edit CFL_BASE_URL 'https://software-test.ecmwf.int/wiki'
  edit CFL_SPACE_KEY '<TODO>'
  edit CFL_TARBALL_COMMENT ''
  edit CFL_TITLE 'Releases'
  edit CHECK_SYNC_RELEASE 'False'
  edit CLANG_TIDY_ABORT_ON_ERROR 'False'
  edit CMAKE '%ECBUILD%'
  edit CMAKELIST_EXTRACT_PACKAGE_VERSION '$(awk '/^project/ && /%PACKAGE%/ && /VERSION/ {for (I=1;I<=NF;I++) if ($I == "VERSION") {print $(I+1)};}' CMakeLists.txt)'
  edit CMAKE_BUILD_TYPE 'RelWithDebInfo'
  edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_PYTHON=ON -DMETKIT_PATH=%BASEPREFIX%/metkit/develop%INSTALL_POSTFIX%'
  edit CMAKE_VERSION '3.13.4'
  edit CODES_UI_GIT_TAG 'develop'
  edit CODES_UI_VERSION '1.3.1'
  edit CPACK 'make package'
  edit CPUS '$(lscpu -p | grep -v '#' | wc -l)'
  edit CRAY_FFTW_VERSION '3.3.8.1'
  edit CRAY_NETCDF_VERSION '4.6.1.3'
  edit CRON_BASEPREFIX '%BASEDIR%/install'
  edit CRON_EMAIL ''
  edit CRON_ENABLED 'False'
  edit CRON_GIT_TAG 'develop'
  edit CUSTOM_COMMANDS ''
  edit DATAFLOW_SANITIZER 'False'
  edit DEFAULT_GNU_COMPILER 'gnu.73'
  edit ECBUILD '%BASEPREFIX%/ecbuild/develop/bin/ecbuild'
  edit ECBUILD_CLONE 'False'
  edit ECBUILD_CLONE_COMMIT 'master'
  edit ECBUILD_GIT_TAG 'develop'
  edit ECBUILD_REPO '%GIT_REPO_PREFIX%/ecsdk/ecbuild.git'
  edit ECBUILD_VERSION '3.3.2'
  edit ECCODES_EXTENSIONS_GIT_TAG 'develop'
  edit ECCODES_EXTENSIONS_VERSION '2.17.0'
  edit ECCODES_GIT_TAG 'develop'
  edit ECCODES_PYTHON_VERSION '0.9.6'
  edit ECCODES_VERSION '2.17.1'
  edit ECFLOW5_LATEST_VERSION '5.4.0'
  edit ECFLOW_VERSION '5.4.1'
  edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/odb_api'
  edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/odb_api'
  edit ECF_JOB_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_JOB% %ECF_JOBOUT%'
  edit ECF_KILL_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% kill'
  edit ECF_STATUS_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% stat'
  edit ECF_TRIES '1'
  edit ECKIT_GIT_TAG 'develop'
  edit ECKIT_VERSION '1.10.1'
  edit ECQUIS_VERSION '1.0.0'
  edit EMOS_MOD_VERSION '462'
  edit EMOS_PA_VERSION '000462'
  edit EXPORTS '"export ECACCOUNT=ecodmdma"'
  edit EXPORT_VARIABLES 'False'
  edit FCKIT_GIT_TAG '0.6.3'
  edit FCKIT_VERSION '0.6.3'
  edit FDB_VINTAGE_GIT_TAG '4.10.0'
  edit FDB_VINTAGE_VERSION '4.10.0'
  edit FFTW_VERSION '3.3.8'
  edit FFTW_VERSION_CRAY '3.3.4.11'
  edit GIT 'git'
  edit GITHUB_PACKAGE_NAME '%PACKAGE_NAME%'
  edit GITHUB_REPO 'git@github.com:ecmwf'
  edit GITREPO '%GIT_REPO_PREFIX%/odb/odb_api.git'
  edit GIT_CLONE 'True'
  edit GIT_CLONE_OPTIONS '--depth=1 --branch %GIT_TAG%'
  edit GIT_METABUILDER_TAG 'ci'
  edit GIT_REPO_PREFIX 'ssh://git@git.ecmwf.int:7999'
  edit GIT_TAG 'develop'
  edit GIT_VERSION '2.26.0'
  edit GRIB_API_VERSION '1.28.0'
  edit INSTALLABLE_TO_USR_LOCAL_APPS 'True'
  edit IWYU_VERSION '0.11'
  edit KSH_EXTRACT_PACKAGE_VERSION '`grep _VERSION_STR VERSION.cmake | grep -o '[0-9]*\.[0-9]*\.[0-9]*'`'
  edit LAPACK_VERSION '3.5.0'
  edit LIBEMOS_GIT_TAG 'develop'
  edit LIBEMOS_VERSION '4.6.2'
  edit LINTERS_REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}'
  edit LSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/lsan.txt'
  edit MAGICS_GIT_TAG 'develop'
  edit MAGICS_PYTHON_VERSION 'master'
  edit MAGICS_VERSION '4.3.3'
  edit MAKE 'make'
  edit MARS_CLIENT_GIT_TAG '6.27.2.2'
  edit MARS_CLIENT_PACKAGE_GIT_TAG '6.27.6'
  edit MARS_CLIENT_PACKAGE_VERSION '6.27.6'
  edit MARS_CLIENT_VERSION '6.27.2.2'
  edit MEMORY_SANITIZER 'False'
  edit METABUILDER_HOME '/var/tmp/ma0/workspace/metabuilder'
  edit METAB_INSTALL_REMOVE_WRITE_PERMISSIONS 'True'
  edit METAB_PYTHON_VERSION 'python3'
  edit METKIT_GIT_TAG 'develop'
  edit METKIT_VERSION '1.4.6'
  edit METVIEW_GIT_TAG 'develop'
  edit METVIEW_PYTHON_VERSION '1.4.2'
  edit METVIEW_VERSION '5.8.3'
  edit MIR_BUNDLE_VERSION '0.9.0.0'
  edit MIR_GIT_TAG 'develop'
  edit MIR_VERSION '1.5.0'
  edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load ecflow/%ECF_VERSION%"'
  edit MODULE_ACTIONS_FIRST ''
  edit MODULE_ACTIONS_HDR 'module_actions.h'
  edit MODULE_CHECKING_HDR 'module_checking.h'
  edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow"'
  edit MODULE_LOAD_SUPPORTED 'True'
  edit MODULE_NAME 'odb_api'
  edit MODULE_VERSION '%GIT_TAG%'
  edit NETCDF4_VERSION '4.7.4'
  edit ODB_API_GIT_TAG 'develop'
  edit ODB_API_VERSION '0.19.4'
  edit ODB_GIT_TAG 'develop'
  edit ODB_TOOLS_GIT_TAG 'develop'
  edit ODB_TOOLS_VERSION '0.18.3'
  edit ODB_VERSION '1.0.11'
  edit ODC_GIT_TAG 'develop'
  edit ODC_VERSION '1.0.3'
  edit OPENJPEG_VERSION '2.3.1'
  edit PACKAGE 'odb_api'
  edit PACKAGE_NAME 'odb_api'
  edit PACKAGE_OWNERS 'mass@ecmwf.int cgm@ecmwf.int ma0@ecmwf.int'
  edit PACKAGE_VERSION '%GIT_TAG%'
  edit PGEN_VERSION '1.7.1.0'
  edit PKG_BOOST_VERSION '1.71.0'
  edit PREFIX_APPEND ''
  edit PROJ4_VERSION '5.2.0'
  edit PROJ_VERSION '6.3.1'
  edit PYTHON2_VERSION '2.7.15-01'
  edit PYTHON3_VERSION '3.6.10-01'
  edit PYTHON3_VERSION_MAJOR_MINOR '3.6'
  edit PYTHON_GITREPO 'This-needs-tobe-overriden'
  edit PYTHON_PACKAGE 'This-needs-tobe-overriden'
  edit QT_VERSION '5.12.3'
  edit QUEUE 'build'
  edit REMOTE_HOST 'polonius'
  edit REUSE_BUILD_DIR 'False'
  edit SAFE_STACK_SANITIZER 'False'
  edit SANITIZER 'False'
  edit SANITIZER_REFERENCE_DIRECTORY '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
  edit SRCDIR '%BASEDIR%/%PACKAGE%'
  edit STATIC_ANALYSER_ABORT_ON_ERROR 'False'
  edit SWIG_VERSION '3.0.8'
  edit SYNC_RELEASE '2020.04.0'
  edit SYNC_RELEASE_CI_STAGE 'develop'
  edit TARBALL_DIR '/perm/ma/ma0/tmp_tarballs/ci'
  edit TAR_PACKAGE 'OdbAPI'
  edit TEST 'ctest'
  edit THREAD_SANITIZER 'False'
  edit TRANSI_GIT_TAG 'develop'
  edit TRANSI_VERSION '0.7.1'
  edit TSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
  edit UBSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/ubsan.txt'
  edit UB_SANITIZER 'False'
  edit USER 'ma0'
  edit USE_WORKDIR 'False'
  edit VALGRIND_VERSION '3.15.0'
  edit WORKDIR '%BASEDIR%/workdir/%ECF_NAME%'
  edit WSDIR '/var/tmp/tmpdir/deploy/jtmp.20739'
  label branch "develop"
  task regenerate
    defstatus complete
    edit ECF_JOB_CMD '%ECF_JOB% > %ECF_JOBOUT% 2>&1'
    edit ECF_KILL_CMD 'kill -15 %ECF_RID%'
    edit ECF_PASS 'FREE'
    edit ECF_STATUS_CMD 'ps --sid %ECF_RID% -f'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
    label date ""
    label metabuilder_branch "ci"
  family cca
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_MIGRATOR=ON -DENABLE_PYTHON=OFF -DLAPACK_FOUND=TRUE -DMETKIT_PATH=%BASEPREFIX%/metkit/develop%INSTALL_POSTFIX%'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/odb_api/cca'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/odb_api/cca:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cca-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '24000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cca'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'cca'
    edit SITE 'cca'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_CXX_FLAGS=-fopenmp -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_MIGRATOR=ON -DENABLE_PYTHON=OFF -DLAPACK_FOUND=TRUE -DMETKIT_PATH=%BASEPREFIX%/metkit/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_CXX_FLAGS=-fopenmp -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_MIGRATOR=ON -DENABLE_PYTHON=OFF -DLAPACK_FOUND=TRUE -DMETKIT_PATH=%BASEPREFIX%/metkit/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.17
      edit COMPILER 'INTEL.17'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap intel/17.0.3.053" "module load gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.85
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.85'
      edit EXPORTS '"export CRAYLIBS_ARCH_OVERRIDE=ivybridge" "export ECACCOUNT=ecodmdma" "export LIBSCI_ARCH_OVERRIDE=ivybridge" "export TMPDIR=/tmp"'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit TEST 'ctest -E "(odb_api_test_codecs_read)" -j${CPUS} --output-on-failure'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.86
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.86'
      edit EXPORTS '"export CRAYLIBS_ARCH_OVERRIDE=ivybridge" "export ECACCOUNT=ecodmdma" "export LIBSCI_ARCH_OVERRIDE=ivybridge" "export TMPDIR=/tmp"'
      edit MODULE_ACTIONS '"module swap cdt/17.09" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit TEST 'ctest -E "(odb_api_test_codecs_read)" -j${CPUS} --output-on-failure'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.87'
      edit EXPORTS '"export CRAYLIBS_ARCH_OVERRIDE=ivybridge" "export ECACCOUNT=ecodmdma" "export LIBSCI_ARCH_OVERRIDE=ivybridge" "export TMPDIR=/tmp"'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit TEST 'ctest -E "(odb_api_test_codecs_read)" -j${CPUS} --output-on-failure'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family ccb
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_MIGRATOR=ON -DENABLE_PYTHON=OFF -DLAPACK_FOUND=TRUE -DMETKIT_PATH=%BASEPREFIX%/metkit/develop%INSTALL_POSTFIX%'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/odb_api/ccb'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/odb_api/ccb:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'ccb-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '24000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'ccb'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'ccb'
    edit SITE 'ccb'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_CXX_FLAGS=-fopenmp -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_MIGRATOR=ON -DENABLE_PYTHON=OFF -DLAPACK_FOUND=TRUE -DMETKIT_PATH=%BASEPREFIX%/metkit/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_CXX_FLAGS=-fopenmp -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_MIGRATOR=ON -DENABLE_PYTHON=OFF -DLAPACK_FOUND=TRUE -DMETKIT_PATH=%BASEPREFIX%/metkit/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.17
      edit COMPILER 'INTEL.17'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap intel/17.0.3.053" "module load gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.85
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.85'
      edit EXPORTS '"export CRAYLIBS_ARCH_OVERRIDE=ivybridge" "export ECACCOUNT=ecodmdma" "export LIBSCI_ARCH_OVERRIDE=ivybridge" "export TMPDIR=/tmp"'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit TEST 'ctest -E "(odb_api_test_codecs_read)" -j${CPUS} --output-on-failure'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.86
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.86'
      edit EXPORTS '"export CRAYLIBS_ARCH_OVERRIDE=ivybridge" "export ECACCOUNT=ecodmdma" "export LIBSCI_ARCH_OVERRIDE=ivybridge" "export TMPDIR=/tmp"'
      edit MODULE_ACTIONS '"module swap cdt/17.09" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit TEST 'ctest -E "(odb_api_test_codecs_read)" -j${CPUS} --output-on-failure'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.87'
      edit EXPORTS '"export CRAYLIBS_ARCH_OVERRIDE=ivybridge" "export ECACCOUNT=ecodmdma" "export LIBSCI_ARCH_OVERRIDE=ivybridge" "export TMPDIR=/tmp"'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit TEST 'ctest -E "(odb_api_test_codecs_read)" -j${CPUS} --output-on-failure'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family cct
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_MIGRATOR=ON -DENABLE_PYTHON=OFF -DLAPACK_FOUND=TRUE -DMETKIT_PATH=%BASEPREFIX%/metkit/develop%INSTALL_POSTFIX%'
    edit CPUS '$EC_total_tasks'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/odb_api/cct'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/odb_api/cct:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cct-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '4'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '6000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cct'
    edit QSUB_LP 'QSUB -lP %EC_TOTAL_TASKS%'
    edit QUEUE 'nf'
    edit REMOTE_HOST 'cct'
    edit SITE 'cct'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_CXX_FLAGS=-fopenmp -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_MIGRATOR=ON -DENABLE_PYTHON=OFF -DLAPACK_FOUND=TRUE -DMETKIT_PATH=%BASEPREFIX%/metkit/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_CXX_FLAGS=-fopenmp -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_MIGRATOR=ON -DENABLE_PYTHON=OFF -DLAPACK_FOUND=TRUE -DMETKIT_PATH=%BASEPREFIX%/metkit/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.17
      edit COMPILER 'INTEL.17'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap intel/17.0.3.053" "module load gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.85
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.85'
      edit EXPORTS '"export CRAYLIBS_ARCH_OVERRIDE=ivybridge" "export ECACCOUNT=ecodmdma" "export LIBSCI_ARCH_OVERRIDE=ivybridge" "export TMPDIR=/tmp"'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit TEST 'ctest -E "(odb_api_test_codecs_read)" -j${CPUS} --output-on-failure'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.86
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.86'
      edit EXPORTS '"export CRAYLIBS_ARCH_OVERRIDE=ivybridge" "export ECACCOUNT=ecodmdma" "export LIBSCI_ARCH_OVERRIDE=ivybridge" "export TMPDIR=/tmp"'
      edit MODULE_ACTIONS '"module swap cdt/17.09" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit TEST 'ctest -E "(odb_api_test_codecs_read)" -j${CPUS} --output-on-failure'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit CMAKE '%ECBUILD% --toolchain=ecmwf-XC30-Cray'
      edit COMPILER 'CRAY.87'
      edit EXPORTS '"export CRAYLIBS_ARCH_OVERRIDE=ivybridge" "export ECACCOUNT=ecodmdma" "export LIBSCI_ARCH_OVERRIDE=ivybridge" "export TMPDIR=/tmp"'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit TEST 'ctest -E "(odb_api_test_codecs_read)" -j${CPUS} --output-on-failure'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family ecgb
    defstatus suspended
    edit CI_LIMIT '10'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_MIGRATOR=OFF -DENABLE_PYTHON=ON -DMETKIT_PATH=%BASEPREFIX%/metkit/develop%INSTALL_POSTFIX%'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'ecgb'
    edit QUEUE 'express'
    edit REMOTE_HOST 'ecgb'
    edit SITE 'ecgb'
    edit WSDIR '$SCRATCH/metabuilds/%ECF_HOST%_%ECF_PORT%/ecgb/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_MIGRATOR=OFF -DENABLE_PYTHON=ON -DMETKIT_PATH=%BASEPREFIX%/metkit/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family leap42
    defstatus suspended
    edit CI_LIMIT '7'
    edit CMAKE_VARS '-DBUILD_SHARED_LIBS=ON -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_PYTHON=ON -DMETKIT_PATH=%BASEPREFIX%/metkit/develop%INSTALL_POSTFIX%'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'leap42'
    edit REMOTE_HOST 'leap42-deploy'
    edit SITE 'leap42'
    edit WSDIR '/tmp/metabuilds/%ECF_HOST%_%ECF_PORT%/leap42/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DBUILD_SHARED_LIBS=ON -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_PYTHON=ON -DMETKIT_PATH=%BASEPREFIX%/metkit/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task create_tarball
        trigger test == complete
        edit CMAKE 'cmake'
        edit EXPORTS '"export ECACCOUNT=ecodmdma" "export ecbuild_ROOT=%BASEPREFIX%/ecbuild/develop"'
        label tar ""
        label tardir ""
        label error ""
      task install
        trigger create_tarball == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.93
      edit CMAKE_VARS '-DBUILD_SHARED_LIBS=ON -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_PYTHON=ON -DMETKIT_PATH=%BASEPREFIX%/metkit/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.93'
      edit EXPORTS '"export ECACCOUNT=ecodmdma" "export LD_LIBRARY_PATH=/usr/local/apps/gcc/9.3.0/lib64/"'
      edit MODULE_ACTIONS '"module swap gnu/9.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90
      edit CMAKE_VARS '-DBUILD_SHARED_LIBS=ON -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_PYTHON=ON -DMETKIT_PATH=%BASEPREFIX%/metkit/develop%INSTALL_POSTFIX%'
      edit COMPILER 'CLANG.90'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family lxc
    defstatus suspended
    edit CI_LIMIT '10'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_Fortran_FLAGS=-fdefault-real-8 -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_PYTHON=ON -DEXTRA_ODB_MIGRATOR_LIBS=gomp -DMETKIT_PATH=%BASEPREFIX%/metkit/develop%INSTALL_POSTFIX%'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'lxc'
    edit QUEUE 'express'
    edit REMOTE_HOST 'lxc'
    edit SITE 'lxc'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxc/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_FLAGS=-fdefault-real-8 -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_PYTHON=ON -DEXTRA_ODB_MIGRATOR_LIBS=gomp -DMETKIT_PATH=%BASEPREFIX%/metkit/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family lxg
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'lxg'
    edit QUEUE 'normal'
    edit REMOTE_HOST 'lxg'
    edit SITE 'lxg'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxg/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_PYTHON=ON -DMETKIT_PATH=%BASEPREFIX%/metkit/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family sappb
    defstatus suspended
    edit ASAN_FILE '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
    edit BASEPREFIX '/autofs/vol/deploy/install/%SITE%'
    edit CI_LIMIT '8'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_Fortran_COMPILER=gfortran -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_PYTHON=ON -DFORTRAN_LIBRARIES="-lgfortran -fopenmp" -DMETKIT_PATH=%BASEPREFIX%/metkit/develop%INSTALL_POSTFIX%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load ecflow/%ECF_VERSION%"'
    edit M_SITE 'sappb'
    edit REMOTE_HOST 'sappb-deploy'
    edit SANITIZER_REFERENCE_DIRECTORY '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
    edit SITE 'sappb'
    edit TSAN_FILE '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
    edit WSDIR '$HOME/metabuilds/%ECF_HOST%_%ECF_PORT%/sappb/%COMPILER:%'
    family gnu.63
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=gfortran -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_PYTHON=ON -DFORTRAN_LIBRARIES="-lgfortran -fopenmp" -DMETKIT_PATH=%BASEPREFIX%/metkit/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module load gnu/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
endsuite
suite odc
  edit ADDRESS_SANITIZER 'False'
  edit ADD_ACTION_TRIGGERS 'True'
  edit AEC_VERSION '1.0.4'
  edit ALLOW_INSTALL_TO_OVERWRITE 'True'
  edit ASAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
  edit ATLAS_GIT_TAG 'develop'
  edit ATLAS_VERSION '0.20.2'
  edit BANNER_TASK_END 'banner %TASK%-end || true'
  edit BANNER_TASK_START 'banner %TASK%-start || true'
  edit BASEDIR '%WSDIR%/%PACKAGE%'
  edit BASEPREFIX '$PERM/install/%SITE%'
  edit BUFRDC_EMOSLIB_VERSION '4.6.3'
  edit BUFRDC_TABLES_GIT_TAG 'develop'
  edit BUFRDC_TABLES_VERSION '000415'
  edit BUILDIR '%BASEDIR%/builds'
  edit BUILD_ALWAYS_PARALLEL 'False'
  edit CFI_SANITIZER 'False'
  edit CFL_BASE_URL 'https://software-test.ecmwf.int/wiki'
  edit CFL_SPACE_KEY '<TODO>'
  edit CFL_TARBALL_COMMENT ''
  edit CFL_TITLE 'Releases'
  edit CHECK_SYNC_RELEASE 'False'
  edit CLANG_TIDY_ABORT_ON_ERROR 'False'
  edit CMAKE '%ECBUILD%'
  edit CMAKELIST_EXTRACT_PACKAGE_VERSION '$(awk '/^project/ && /%PACKAGE%/ && /VERSION/ {for (I=1;I<=NF;I++) if ($I == "VERSION") {print $(I+1)};}' CMakeLists.txt)'
  edit CMAKE_BUILD_TYPE 'RelWithDebInfo'
  edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_PYTHON=OFF'
  edit CMAKE_VERSION '3.13.4'
  edit CODES_UI_GIT_TAG 'develop'
  edit CODES_UI_VERSION '1.3.1'
  edit CPACK 'make package'
  edit CPUS '$(lscpu -p | grep -v '#' | wc -l)'
  edit CRAY_FFTW_VERSION '3.3.8.1'
  edit CRAY_NETCDF_VERSION '4.6.1.3'
  edit CRON_BASEPREFIX '%BASEDIR%/install'
  edit CRON_EMAIL ''
  edit CRON_ENABLED 'False'
  edit CRON_GIT_TAG 'develop'
  edit CUSTOM_COMMANDS ''
  edit DATAFLOW_SANITIZER 'False'
  edit DEFAULT_GNU_COMPILER 'gnu.73'
  edit ECBUILD '%BASEPREFIX%/ecbuild/develop/bin/ecbuild'
  edit ECBUILD_CLONE 'False'
  edit ECBUILD_CLONE_COMMIT 'master'
  edit ECBUILD_GIT_TAG 'develop'
  edit ECBUILD_REPO '%GIT_REPO_PREFIX%/ecsdk/ecbuild.git'
  edit ECBUILD_VERSION '3.3.2'
  edit ECCODES_EXTENSIONS_GIT_TAG 'develop'
  edit ECCODES_EXTENSIONS_VERSION '2.17.0'
  edit ECCODES_GIT_TAG 'develop'
  edit ECCODES_PYTHON_VERSION '0.9.6'
  edit ECCODES_VERSION '2.17.1'
  edit ECFLOW5_LATEST_VERSION '5.4.0'
  edit ECFLOW_VERSION '5.4.1'
  edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/odc'
  edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/odc'
  edit ECF_JOB_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_JOB% %ECF_JOBOUT%'
  edit ECF_KILL_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% kill'
  edit ECF_STATUS_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% stat'
  edit ECF_TRIES '1'
  edit ECKIT_GIT_TAG 'develop'
  edit ECKIT_VERSION '1.10.1'
  edit ECQUIS_VERSION '1.0.0'
  edit EMOS_MOD_VERSION '462'
  edit EMOS_PA_VERSION '000462'
  edit EXPORTS '"export ECACCOUNT=ecodmdma"'
  edit EXPORT_VARIABLES 'False'
  edit FCKIT_GIT_TAG '0.6.3'
  edit FCKIT_VERSION '0.6.3'
  edit FDB_VINTAGE_GIT_TAG '4.10.0'
  edit FDB_VINTAGE_VERSION '4.10.0'
  edit FFTW_VERSION '3.3.8'
  edit FFTW_VERSION_CRAY '3.3.4.11'
  edit GIT 'git'
  edit GITHUB_PACKAGE_NAME '%PACKAGE_NAME%'
  edit GITHUB_REPO 'git@github.com:ecmwf'
  edit GITREPO '%GIT_REPO_PREFIX%/odb/odc.git'
  edit GIT_CLONE 'True'
  edit GIT_CLONE_OPTIONS '--depth=1 --branch %GIT_TAG%'
  edit GIT_METABUILDER_TAG 'ci'
  edit GIT_REPO_PREFIX 'ssh://git@git.ecmwf.int:7999'
  edit GIT_TAG 'develop'
  edit GIT_VERSION '2.26.0'
  edit GRIB_API_VERSION '1.28.0'
  edit INSTALLABLE_TO_USR_LOCAL_APPS 'True'
  edit IWYU_VERSION '0.11'
  edit KSH_EXTRACT_PACKAGE_VERSION '`grep _VERSION_STR VERSION.cmake | grep -o '[0-9]*\.[0-9]*\.[0-9]*'`'
  edit LAPACK_VERSION '3.5.0'
  edit LIBEMOS_GIT_TAG 'develop'
  edit LIBEMOS_VERSION '4.6.2'
  edit LINTERS_REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}'
  edit LSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/lsan.txt'
  edit MAGICS_GIT_TAG 'develop'
  edit MAGICS_PYTHON_VERSION 'master'
  edit MAGICS_VERSION '4.3.3'
  edit MAKE 'make'
  edit MARS_CLIENT_GIT_TAG '6.27.2.2'
  edit MARS_CLIENT_PACKAGE_GIT_TAG '6.27.6'
  edit MARS_CLIENT_PACKAGE_VERSION '6.27.6'
  edit MARS_CLIENT_VERSION '6.27.2.2'
  edit MEMORY_SANITIZER 'False'
  edit METABUILDER_HOME '/var/tmp/ma0/workspace/metabuilder'
  edit METAB_INSTALL_REMOVE_WRITE_PERMISSIONS 'True'
  edit METAB_PYTHON_VERSION 'python3'
  edit METKIT_GIT_TAG 'develop'
  edit METKIT_VERSION '1.4.6'
  edit METVIEW_GIT_TAG 'develop'
  edit METVIEW_PYTHON_VERSION '1.4.2'
  edit METVIEW_VERSION '5.8.3'
  edit MIR_BUNDLE_VERSION '0.9.0.0'
  edit MIR_GIT_TAG 'develop'
  edit MIR_VERSION '1.5.0'
  edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
  edit MODULE_ACTIONS_FIRST ''
  edit MODULE_ACTIONS_HDR 'module_actions.h'
  edit MODULE_CHECKING_HDR 'module_checking.h'
  edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow"'
  edit MODULE_LOAD_SUPPORTED 'True'
  edit MODULE_NAME 'odc'
  edit MODULE_VERSION '%GIT_TAG%'
  edit NETCDF4_VERSION '4.7.4'
  edit ODB_API_GIT_TAG 'develop'
  edit ODB_API_VERSION '0.19.4'
  edit ODB_GIT_TAG 'develop'
  edit ODB_TOOLS_GIT_TAG 'develop'
  edit ODB_TOOLS_VERSION '0.18.3'
  edit ODB_VERSION '1.0.11'
  edit ODC_GIT_TAG 'develop'
  edit ODC_VERSION '1.0.3'
  edit OPENJPEG_VERSION '2.3.1'
  edit PACKAGE 'odc'
  edit PACKAGE_NAME 'odc'
  edit PACKAGE_OWNERS 'mass@ecmwf.int cgm@ecmwf.int ma0@ecmwf.int'
  edit PACKAGE_VERSION '%GIT_TAG%'
  edit PGEN_VERSION '1.7.1.0'
  edit PKG_BOOST_VERSION '1.71.0'
  edit PREFIX_APPEND ''
  edit PROJ4_VERSION '5.2.0'
  edit PROJ_VERSION '6.3.1'
  edit PYTHON2_VERSION '2.7.15-01'
  edit PYTHON3_VERSION '3.6.10-01'
  edit PYTHON3_VERSION_MAJOR_MINOR '3.6'
  edit PYTHON_GITREPO 'This-needs-tobe-overriden'
  edit PYTHON_PACKAGE 'This-needs-tobe-overriden'
  edit QT_VERSION '5.12.3'
  edit QUEUE 'build'
  edit REMOTE_HOST 'polonius'
  edit REUSE_BUILD_DIR 'False'
  edit SAFE_STACK_SANITIZER 'False'
  edit SANITIZER 'False'
  edit SANITIZER_REFERENCE_DIRECTORY '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
  edit SRCDIR '%BASEDIR%/%PACKAGE%'
  edit STATIC_ANALYSER_ABORT_ON_ERROR 'False'
  edit SWIG_VERSION '3.0.8'
  edit SYNC_RELEASE '2020.04.0'
  edit SYNC_RELEASE_CI_STAGE 'develop'
  edit TARBALL_DIR '/perm/ma/ma0/tmp_tarballs/ci'
  edit TAR_PACKAGE '%PACKAGE%'
  edit TEST 'ctest --output-on-failure -j${CPUS} '
  edit THREAD_SANITIZER 'False'
  edit TRANSI_GIT_TAG 'develop'
  edit TRANSI_VERSION '0.7.1'
  edit TSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
  edit UBSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/ubsan.txt'
  edit UB_SANITIZER 'False'
  edit USER 'ma0'
  edit USE_WORKDIR 'False'
  edit VALGRIND_VERSION '3.15.0'
  edit WORKDIR '%BASEDIR%/workdir/%ECF_NAME%'
  edit WSDIR '/var/tmp/tmpdir/deploy/jtmp.20739'
  label branch "develop"
  task regenerate
    defstatus complete
    edit ECF_JOB_CMD '%ECF_JOB% > %ECF_JOBOUT% 2>&1'
    edit ECF_KILL_CMD 'kill -15 %ECF_RID%'
    edit ECF_PASS 'FREE'
    edit ECF_STATUS_CMD 'ps --sid %ECF_RID% -f'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
    label date ""
    label metabuilder_branch "ci"
  family cca
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_PYTHON=OFF -DLAPACK_FOUND=TRUE'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/odc/cca'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/odc/cca:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cca-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '24000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cca'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'cca'
    edit SITE 'cca'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_PYTHON=OFF -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_PYTHON=OFF -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.17
      edit COMPILER 'INTEL.17'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap intel/17.0.3.053" "module load gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.85
      edit COMPILER 'CRAY.85'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.86
      edit COMPILER 'CRAY.86'
      edit MODULE_ACTIONS '"module swap cdt/17.09" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit COMPILER 'CRAY.87'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family ccb
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_PYTHON=OFF -DLAPACK_FOUND=TRUE'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/odc/ccb'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/odc/ccb:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'ccb-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '24000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'ccb'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'ccb'
    edit SITE 'ccb'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_PYTHON=OFF -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_PYTHON=OFF -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.17
      edit COMPILER 'INTEL.17'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap intel/17.0.3.053" "module load gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.85
      edit COMPILER 'CRAY.85'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.86
      edit COMPILER 'CRAY.86'
      edit MODULE_ACTIONS '"module swap cdt/17.09" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit COMPILER 'CRAY.87'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family cct
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_PYTHON=OFF -DLAPACK_FOUND=TRUE'
    edit CPUS '$EC_total_tasks'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/odc/cct'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/odc/cct:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cct-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '4'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '6000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cct'
    edit QSUB_LP 'QSUB -lP %EC_TOTAL_TASKS%'
    edit QUEUE 'nf'
    edit REMOTE_HOST 'cct'
    edit SITE 'cct'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_PYTHON=OFF -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_PYTHON=OFF -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.17
      edit COMPILER 'INTEL.17'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap intel/17.0.3.053" "module load gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.85
      edit COMPILER 'CRAY.85'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.86
      edit COMPILER 'CRAY.86'
      edit MODULE_ACTIONS '"module swap cdt/17.09" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit COMPILER 'CRAY.87'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family ecgb
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'ecgb'
    edit QUEUE 'express'
    edit REMOTE_HOST 'ecgb'
    edit SITE 'ecgb'
    edit WSDIR '$SCRATCH/metabuilds/%ECF_HOST%_%ECF_PORT%/ecgb/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_PYTHON=OFF'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family leap42
    defstatus suspended
    edit CI_LIMIT '7'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'leap42'
    edit REMOTE_HOST 'leap42-deploy'
    edit SITE 'leap42'
    edit WSDIR '/tmp/metabuilds/%ECF_HOST%_%ECF_PORT%/leap42/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_PYTHON=OFF'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task create_tarball
        trigger test == complete
        edit CMAKE 'cmake'
        edit EXPORTS '"export ECACCOUNT=ecodmdma" "export ecbuild_ROOT=%BASEPREFIX%/ecbuild/develop"'
        label tar ""
        label tardir ""
        label error ""
      task install
        trigger create_tarball == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.93
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_PYTHON=OFF'
      edit COMPILER 'GNU.93'
      edit EXPORTS '"export ECACCOUNT=ecodmdma" "export LD_LIBRARY_PATH=/usr/local/apps/gcc/9.3.0/lib64/"'
      edit MODULE_ACTIONS '"module swap gnu/9.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.19
      edit COMPILER 'INTEL.19'
      edit MODULE_ACTIONS '"module unload gnu" "module load intel/19.0.4" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit POST_EXPORTS '"export CXX=icpc\ -cxxlib=/usr/local/apps/gcc/8.3.0" "export LD_RUN_PATH=/usr/local/apps/gcc/8.3.0/lib64" "export PATH=/usr/local/apps/gcc/8.3.0/bin:$PATH"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_PYTHON=OFF'
      edit COMPILER 'CLANG.90'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90.linters
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_PYTHON=OFF'
      edit COMPILER 'CLANG.90.LINTERS'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task cppcheck
        trigger config == complete
        edit CPPCHECK_FLAGS '--template='[{file}:{line}]: {severity} : {id} : {message}' --suppress=syntaxError --inline-suppr --platform=unix64 --enable=performance --enable=portability --enable=information --std=c11 --std=c++14 -DDEBUG=1 -q -j8'
        edit CPPCHECK_VERSION '1.89'
        edit REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}/cpp_check'
      task iwyu
        trigger cppcheck == complete
        edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module load iwyu/%IWYU_VERSION%"'
      family clang_linters
        trigger iwyu == complete
        limit clang_linters 1
        inlimit clang_linters
        task clang_tidy
          edit REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}/clang_tidy'
        task static_analyser
          edit REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}/static_analyser'
          edit SCAN_BUILD_OPTIONS '--force-analyze-debug-code -o /perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}/static_analyser -k make -j8'
          edit USE_CLANG_CHECK 'False'
          label scan_view ""
      endfamily
      task wipe_build
        trigger clang_linters == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90.address_sanitizer
      edit ADDRESS_SANITIZER 'True'
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_FLAGS="-fsanitize=address -fno-omit-frame-pointer" -DCMAKE_C_FLAGS="-fsanitize=address -fno-omit-frame-pointer" -DCMAKE_EXE_LINKER_FLAGS=-fsanitize=address -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=OFF -DENABLE_PYTHON=OFF'
      edit COMPILER 'CLANG.90.ADDRESS_SANITIZER'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit SANITIZER 'True'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit TEST_DIR '%BUILDIR%'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90.ub_sanitizer
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_FLAGS="-fsanitize=undefined -fno-omit-frame-pointer -fno-sanitize-recover=all" -DCMAKE_C_FLAGS="-fsanitize=undefined -fno-omit-frame-pointer" -DCMAKE_EXE_LINKER_FLAGS=-fsanitize=undefined -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=OFF -DENABLE_PYTHON=OFF'
      edit COMPILER 'CLANG.90.UB_SANITIZER'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit SANITIZER 'True'
      edit UB_SANITIZER 'True'
      edit UB_SANITIZER_LD_PRELOAD 'export LD_PRELOAD=/usr/local/apps/clang/9.0.1/lib64/libubsan.so'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit TEST_DIR '%BUILDIR%'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family lxc
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'lxc'
    edit QUEUE 'express'
    edit REMOTE_HOST 'lxc'
    edit SITE 'lxc'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxc/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_PYTHON=OFF'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family lxg
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'lxg'
    edit QUEUE 'normal'
    edit REMOTE_HOST 'lxg'
    edit SITE 'lxg'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxg/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_PYTHON=OFF'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
endsuite
suite odb_tools
  edit ADDRESS_SANITIZER 'False'
  edit ADD_ACTION_TRIGGERS 'True'
  edit AEC_VERSION '1.0.4'
  edit ALLOW_INSTALL_TO_OVERWRITE 'True'
  edit ASAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
  edit ATLAS_GIT_TAG 'develop'
  edit ATLAS_VERSION '0.20.2'
  edit BANNER_TASK_END 'banner %TASK%-end || true'
  edit BANNER_TASK_START 'banner %TASK%-start || true'
  edit BASEDIR '%WSDIR%/%PACKAGE%'
  edit BASEPREFIX '$PERM/install/%SITE%'
  edit BUFRDC_EMOSLIB_VERSION '4.6.3'
  edit BUFRDC_TABLES_GIT_TAG 'develop'
  edit BUFRDC_TABLES_VERSION '000415'
  edit BUILDIR '%BASEDIR%/builds'
  edit BUILD_ALWAYS_PARALLEL 'False'
  edit CFI_SANITIZER 'False'
  edit CFL_BASE_URL 'https://software-test.ecmwf.int/wiki'
  edit CFL_SPACE_KEY '<TODO>'
  edit CFL_TARBALL_COMMENT ''
  edit CFL_TITLE 'Releases'
  edit CHECK_SYNC_RELEASE 'False'
  edit CLANG_TIDY_ABORT_ON_ERROR 'False'
  edit CMAKE '%ECBUILD%'
  edit CMAKELIST_EXTRACT_PACKAGE_VERSION '$(awk '/^project/ && /%PACKAGE%/ && /VERSION/ {for (I=1;I<=NF;I++) if ($I == "VERSION") {print $(I+1)};}' CMakeLists.txt)'
  edit CMAKE_BUILD_TYPE 'RelWithDebInfo'
  edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_MIGRATOR=ON -DENABLE_NETCDF=ON -DENABLE_PYTHON=ON -DMETKIT_PATH=%BASEPREFIX%/metkit/develop%INSTALL_POSTFIX% -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DODC_PATH=%BASEPREFIX%/odc/develop%INSTALL_POSTFIX%'
  edit CMAKE_VERSION '3.13.4'
  edit CODES_UI_GIT_TAG 'develop'
  edit CODES_UI_VERSION '1.3.1'
  edit CPACK 'make package'
  edit CPUS '$(lscpu -p | grep -v '#' | wc -l)'
  edit CRAY_FFTW_VERSION '3.3.8.1'
  edit CRAY_NETCDF_VERSION '4.6.1.3'
  edit CRON_BASEPREFIX '%BASEDIR%/install'
  edit CRON_EMAIL ''
  edit CRON_ENABLED 'False'
  edit CRON_GIT_TAG 'develop'
  edit CUSTOM_COMMANDS ''
  edit DATAFLOW_SANITIZER 'False'
  edit DEFAULT_GNU_COMPILER 'gnu.73'
  edit ECBUILD '%BASEPREFIX%/ecbuild/develop/bin/ecbuild'
  edit ECBUILD_CLONE 'False'
  edit ECBUILD_CLONE_COMMIT 'master'
  edit ECBUILD_GIT_TAG 'develop'
  edit ECBUILD_REPO '%GIT_REPO_PREFIX%/ecsdk/ecbuild.git'
  edit ECBUILD_VERSION '3.3.2'
  edit ECCODES_EXTENSIONS_GIT_TAG 'develop'
  edit ECCODES_EXTENSIONS_VERSION '2.17.0'
  edit ECCODES_GIT_TAG 'develop'
  edit ECCODES_PYTHON_VERSION '0.9.6'
  edit ECCODES_VERSION '2.17.1'
  edit ECFLOW5_LATEST_VERSION '5.4.0'
  edit ECFLOW_VERSION '5.4.1'
  edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/odb_tools'
  edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/odb_tools'
  edit ECF_JOB_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_JOB% %ECF_JOBOUT%'
  edit ECF_KILL_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% kill'
  edit ECF_STATUS_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% stat'
  edit ECF_TRIES '1'
  edit ECKIT_GIT_TAG 'develop'
  edit ECKIT_VERSION '1.10.1'
  edit ECQUIS_VERSION '1.0.0'
  edit EMOS_MOD_VERSION '462'
  edit EMOS_PA_VERSION '000462'
  edit EXPORTS '"export ECACCOUNT=ecodmdma"'
  edit EXPORT_VARIABLES 'False'
  edit FCKIT_GIT_TAG '0.6.3'
  edit FCKIT_VERSION '0.6.3'
  edit FDB_VINTAGE_GIT_TAG '4.10.0'
  edit FDB_VINTAGE_VERSION '4.10.0'
  edit FFTW_VERSION '3.3.8'
  edit FFTW_VERSION_CRAY '3.3.4.11'
  edit GIT 'git'
  edit GITHUB_PACKAGE_NAME '%PACKAGE_NAME%'
  edit GITHUB_REPO 'git@github.com:ecmwf'
  edit GITREPO '%GIT_REPO_PREFIX%/odb/odb-tools.git'
  edit GIT_CLONE 'True'
  edit GIT_CLONE_OPTIONS '--depth=1 --branch %GIT_TAG%'
  edit GIT_METABUILDER_TAG 'ci'
  edit GIT_REPO_PREFIX 'ssh://git@git.ecmwf.int:7999'
  edit GIT_TAG 'develop'
  edit GIT_VERSION '2.26.0'
  edit GRIB_API_VERSION '1.28.0'
  edit INSTALLABLE_TO_USR_LOCAL_APPS 'False'
  edit IWYU_VERSION '0.11'
  edit KSH_EXTRACT_PACKAGE_VERSION '`grep _VERSION_STR VERSION.cmake | grep -o '[0-9]*\.[0-9]*\.[0-9]*'`'
  edit LAPACK_VERSION '3.5.0'
  edit LIBEMOS_GIT_TAG 'develop'
  edit LIBEMOS_VERSION '4.6.2'
  edit LINTERS_REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}'
  edit LSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/lsan.txt'
  edit MAGICS_GIT_TAG 'develop'
  edit MAGICS_PYTHON_VERSION 'master'
  edit MAGICS_VERSION '4.3.3'
  edit MAKE 'make'
  edit MARS_CLIENT_GIT_TAG '6.27.2.2'
  edit MARS_CLIENT_PACKAGE_GIT_TAG '6.27.6'
  edit MARS_CLIENT_PACKAGE_VERSION '6.27.6'
  edit MARS_CLIENT_VERSION '6.27.2.2'
  edit MEMORY_SANITIZER 'False'
  edit METABUILDER_HOME '/var/tmp/ma0/workspace/metabuilder'
  edit METAB_INSTALL_REMOVE_WRITE_PERMISSIONS 'True'
  edit METAB_PYTHON_VERSION 'python3'
  edit METKIT_GIT_TAG 'develop'
  edit METKIT_VERSION '1.4.6'
  edit METVIEW_GIT_TAG 'develop'
  edit METVIEW_PYTHON_VERSION '1.4.2'
  edit METVIEW_VERSION '5.8.3'
  edit MIR_BUNDLE_VERSION '0.9.0.0'
  edit MIR_GIT_TAG 'develop'
  edit MIR_VERSION '1.5.0'
  edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
  edit MODULE_ACTIONS_FIRST ''
  edit MODULE_ACTIONS_HDR 'module_actions.h'
  edit MODULE_CHECKING_HDR 'module_checking.h'
  edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow"'
  edit MODULE_LOAD_SUPPORTED 'True'
  edit MODULE_NAME 'odb_tools'
  edit MODULE_VERSION '%GIT_TAG%'
  edit NETCDF4_VERSION '4.7.4'
  edit ODB_API_GIT_TAG 'develop'
  edit ODB_API_VERSION '0.19.4'
  edit ODB_GIT_TAG 'develop'
  edit ODB_TOOLS_GIT_TAG 'develop'
  edit ODB_TOOLS_VERSION '0.18.3'
  edit ODB_VERSION '1.0.11'
  edit ODC_GIT_TAG 'develop'
  edit ODC_VERSION '1.0.3'
  edit OPENJPEG_VERSION '2.3.1'
  edit PACKAGE 'odb_tools'
  edit PACKAGE_NAME 'odb_tools'
  edit PACKAGE_OWNERS 'mass@ecmwf.int cgm@ecmwf.int ma0@ecmwf.int'
  edit PACKAGE_VERSION '%GIT_TAG%'
  edit PGEN_VERSION '1.7.1.0'
  edit PKG_BOOST_VERSION '1.71.0'
  edit PREFIX_APPEND ''
  edit PROJ4_VERSION '5.2.0'
  edit PROJ_VERSION '6.3.1'
  edit PYTHON2_VERSION '2.7.15-01'
  edit PYTHON3_VERSION '3.6.10-01'
  edit PYTHON3_VERSION_MAJOR_MINOR '3.6'
  edit PYTHON_GITREPO 'This-needs-tobe-overriden'
  edit PYTHON_PACKAGE 'This-needs-tobe-overriden'
  edit QT_VERSION '5.12.3'
  edit QUEUE 'build'
  edit REMOTE_HOST 'polonius'
  edit REUSE_BUILD_DIR 'False'
  edit SAFE_STACK_SANITIZER 'False'
  edit SANITIZER 'False'
  edit SANITIZER_REFERENCE_DIRECTORY '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
  edit SRCDIR '%BASEDIR%/%PACKAGE%'
  edit STATIC_ANALYSER_ABORT_ON_ERROR 'False'
  edit SWIG_VERSION '3.0.8'
  edit SYNC_RELEASE '2020.04.0'
  edit SYNC_RELEASE_CI_STAGE 'develop'
  edit TARBALL_DIR '/perm/ma/ma0/tmp_tarballs/ci'
  edit TAR_PACKAGE 'OdbTools'
  edit TEST 'ctest'
  edit THREAD_SANITIZER 'False'
  edit TRANSI_GIT_TAG 'develop'
  edit TRANSI_VERSION '0.7.1'
  edit TSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
  edit UBSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/ubsan.txt'
  edit UB_SANITIZER 'False'
  edit USER 'ma0'
  edit USE_WORKDIR 'False'
  edit VALGRIND_VERSION '3.15.0'
  edit WORKDIR '%BASEDIR%/workdir/%ECF_NAME%'
  edit WSDIR '/var/tmp/tmpdir/deploy/jtmp.20739'
  label branch "develop"
  task regenerate
    defstatus complete
    edit ECF_JOB_CMD '%ECF_JOB% > %ECF_JOBOUT% 2>&1'
    edit ECF_KILL_CMD 'kill -15 %ECF_RID%'
    edit ECF_PASS 'FREE'
    edit ECF_STATUS_CMD 'ps --sid %ECF_RID% -f'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
    label date ""
    label metabuilder_branch "ci"
  family cca
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_MIGRATOR=ON -DENABLE_NETCDF=ON -DENABLE_PYTHON=ON -DLAPACK_FOUND=TRUE -DMETKIT_PATH=%BASEPREFIX%/metkit/develop%INSTALL_POSTFIX% -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DODC_PATH=%BASEPREFIX%/odc/develop%INSTALL_POSTFIX%'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/odb_tools/cca'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/odb_tools/cca:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cca-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '24000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cca'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'cca'
    edit SITE 'cca'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_MIGRATOR=ON -DENABLE_NETCDF=ON -DENABLE_PYTHON=ON -DLAPACK_FOUND=TRUE -DMETKIT_PATH=%BASEPREFIX%/metkit/develop%INSTALL_POSTFIX% -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DODC_PATH=%BASEPREFIX%/odc/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task add_write_perm
        defstatus complete
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_MIGRATOR=ON -DENABLE_NETCDF=ON -DENABLE_PYTHON=ON -DLAPACK_FOUND=TRUE -DMETKIT_PATH=%BASEPREFIX%/metkit/develop%INSTALL_POSTFIX% -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DODC_PATH=%BASEPREFIX%/odc/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task add_write_perm
        defstatus complete
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.17
      edit COMPILER 'INTEL.17'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap intel/17.0.3.053" "module load gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task add_write_perm
        defstatus complete
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.85
      edit COMPILER 'CRAY.85'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task add_write_perm
        defstatus complete
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.86
      edit COMPILER 'CRAY.86'
      edit MODULE_ACTIONS '"module swap cdt/17.09" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task add_write_perm
        defstatus complete
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit COMPILER 'CRAY.87'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task add_write_perm
        defstatus complete
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family ccb
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_MIGRATOR=ON -DENABLE_NETCDF=ON -DENABLE_PYTHON=ON -DLAPACK_FOUND=TRUE -DMETKIT_PATH=%BASEPREFIX%/metkit/develop%INSTALL_POSTFIX% -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DODC_PATH=%BASEPREFIX%/odc/develop%INSTALL_POSTFIX%'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/odb_tools/ccb'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/odb_tools/ccb:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'ccb-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '24000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'ccb'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'ccb'
    edit SITE 'ccb'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_MIGRATOR=ON -DENABLE_NETCDF=ON -DENABLE_PYTHON=ON -DLAPACK_FOUND=TRUE -DMETKIT_PATH=%BASEPREFIX%/metkit/develop%INSTALL_POSTFIX% -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DODC_PATH=%BASEPREFIX%/odc/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task add_write_perm
        defstatus complete
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_MIGRATOR=ON -DENABLE_NETCDF=ON -DENABLE_PYTHON=ON -DLAPACK_FOUND=TRUE -DMETKIT_PATH=%BASEPREFIX%/metkit/develop%INSTALL_POSTFIX% -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DODC_PATH=%BASEPREFIX%/odc/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task add_write_perm
        defstatus complete
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.17
      edit COMPILER 'INTEL.17'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap intel/17.0.3.053" "module load gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task add_write_perm
        defstatus complete
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.85
      edit COMPILER 'CRAY.85'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task add_write_perm
        defstatus complete
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.86
      edit COMPILER 'CRAY.86'
      edit MODULE_ACTIONS '"module swap cdt/17.09" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task add_write_perm
        defstatus complete
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit COMPILER 'CRAY.87'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task add_write_perm
        defstatus complete
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family cct
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_MIGRATOR=ON -DENABLE_NETCDF=ON -DENABLE_PYTHON=ON -DLAPACK_FOUND=TRUE -DMETKIT_PATH=%BASEPREFIX%/metkit/develop%INSTALL_POSTFIX% -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DODC_PATH=%BASEPREFIX%/odc/develop%INSTALL_POSTFIX%'
    edit CPUS '$EC_total_tasks'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/odb_tools/cct'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/odb_tools/cct:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cct-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '4'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '6000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cct'
    edit QSUB_LP 'QSUB -lP %EC_TOTAL_TASKS%'
    edit QUEUE 'nf'
    edit REMOTE_HOST 'cct'
    edit SITE 'cct'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_MIGRATOR=ON -DENABLE_NETCDF=ON -DENABLE_PYTHON=ON -DLAPACK_FOUND=TRUE -DMETKIT_PATH=%BASEPREFIX%/metkit/develop%INSTALL_POSTFIX% -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DODC_PATH=%BASEPREFIX%/odc/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task add_write_perm
        defstatus complete
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_MIGRATOR=ON -DENABLE_NETCDF=ON -DENABLE_PYTHON=ON -DLAPACK_FOUND=TRUE -DMETKIT_PATH=%BASEPREFIX%/metkit/develop%INSTALL_POSTFIX% -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DODC_PATH=%BASEPREFIX%/odc/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task add_write_perm
        defstatus complete
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.17
      edit COMPILER 'INTEL.17'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap intel/17.0.3.053" "module load gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task add_write_perm
        defstatus complete
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.85
      edit COMPILER 'CRAY.85'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task add_write_perm
        defstatus complete
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.86
      edit COMPILER 'CRAY.86'
      edit MODULE_ACTIONS '"module swap cdt/17.09" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task add_write_perm
        defstatus complete
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit COMPILER 'CRAY.87'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task add_write_perm
        defstatus complete
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family ecgb
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'ecgb'
    edit QUEUE 'express'
    edit REMOTE_HOST 'ecgb'
    edit SITE 'ecgb'
    edit WSDIR '$SCRATCH/metabuilds/%ECF_HOST%_%ECF_PORT%/ecgb/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_MIGRATOR=ON -DENABLE_NETCDF=ON -DENABLE_PYTHON=ON -DMETKIT_PATH=%BASEPREFIX%/metkit/develop%INSTALL_POSTFIX% -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DODC_PATH=%BASEPREFIX%/odc/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task add_write_perm
        defstatus complete
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family leap42
    defstatus suspended
    edit CI_LIMIT '7'
    edit CMAKE_VARS '-DBUILD_SHARED_LIBS=ON -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_MIGRATOR=ON -DENABLE_NETCDF=ON -DENABLE_PYTHON=ON -DMETKIT_PATH=%BASEPREFIX%/metkit/develop%INSTALL_POSTFIX% -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DODC_PATH=%BASEPREFIX%/odc/develop%INSTALL_POSTFIX%'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'leap42'
    edit REMOTE_HOST 'leap42-deploy'
    edit SITE 'leap42'
    edit WSDIR '/tmp/metabuilds/%ECF_HOST%_%ECF_PORT%/leap42/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DBUILD_SHARED_LIBS=ON -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_MIGRATOR=ON -DENABLE_NETCDF=ON -DENABLE_PYTHON=ON -DMETKIT_PATH=%BASEPREFIX%/metkit/develop%INSTALL_POSTFIX% -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DODC_PATH=%BASEPREFIX%/odc/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task create_tarball
        trigger test == complete
        edit CMAKE 'cmake'
        edit EXPORTS '"export ECACCOUNT=ecodmdma" "export ecbuild_ROOT=%BASEPREFIX%/ecbuild/develop"'
        label tar ""
        label tardir ""
        label error ""
      task add_write_perm
        defstatus complete
      task install
        trigger create_tarball == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.93
      edit CMAKE_VARS '-DBUILD_SHARED_LIBS=ON -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_MIGRATOR=ON -DENABLE_NETCDF=ON -DENABLE_PYTHON=ON -DMETKIT_PATH=%BASEPREFIX%/metkit/develop%INSTALL_POSTFIX% -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DODC_PATH=%BASEPREFIX%/odc/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.93'
      edit EXPORTS '"export ECACCOUNT=ecodmdma" "export LD_LIBRARY_PATH=/usr/local/apps/gcc/9.3.0/lib64/"'
      edit MODULE_ACTIONS '"module swap gnu/9.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task add_write_perm
        defstatus complete
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family lxc
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'lxc'
    edit QUEUE 'express'
    edit REMOTE_HOST 'lxc'
    edit SITE 'lxc'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxc/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_MIGRATOR=ON -DENABLE_NETCDF=ON -DENABLE_PYTHON=ON -DMETKIT_PATH=%BASEPREFIX%/metkit/develop%INSTALL_POSTFIX% -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DODC_PATH=%BASEPREFIX%/odc/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task add_write_perm
        defstatus complete
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family lxg
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'lxg'
    edit QUEUE 'normal'
    edit REMOTE_HOST 'lxg'
    edit SITE 'lxg'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxg/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DENABLE_FORTRAN=ON -DENABLE_MIGRATOR=ON -DENABLE_NETCDF=ON -DENABLE_PYTHON=ON -DMETKIT_PATH=%BASEPREFIX%/metkit/develop%INSTALL_POSTFIX% -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DODC_PATH=%BASEPREFIX%/odc/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load swig/%SWIG_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task add_write_perm
        defstatus complete
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
endsuite
suite Magics
  edit ADDRESS_SANITIZER 'False'
  edit ADD_ACTION_TRIGGERS 'True'
  edit AEC_VERSION '1.0.4'
  edit ALLOW_INSTALL_TO_OVERWRITE 'True'
  edit ALLOW_MERGE_FROM_GITHUB 'True'
  edit ASAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
  edit ATLAS_GIT_TAG 'develop'
  edit ATLAS_VERSION '0.20.2'
  edit BANNER_TASK_END 'banner %TASK%-end || true'
  edit BANNER_TASK_START 'banner %TASK%-start || true'
  edit BASEDIR '%WSDIR%/%PACKAGE%'
  edit BASEPREFIX '$PERM/install/%SITE%'
  edit BUFRDC_EMOSLIB_VERSION '4.6.3'
  edit BUFRDC_TABLES_GIT_TAG 'develop'
  edit BUFRDC_TABLES_VERSION '000415'
  edit BUILDIR '%BASEDIR%/builds'
  edit BUILD_ALWAYS_PARALLEL 'False'
  edit CFI_SANITIZER 'False'
  edit CFL_BASE_URL 'https://software-test.ecmwf.int/wiki'
  edit CFL_SPACE_KEY 'MAGP'
  edit CFL_TARBALL_COMMENT ''
  edit CFL_TITLE 'Releases'
  edit CHECK_SYNC_RELEASE 'False'
  edit CLANG_TIDY_ABORT_ON_ERROR 'False'
  edit CMAKE '%ECBUILD%'
  edit CMAKELIST_EXTRACT_PACKAGE_VERSION '$(awk '/^project/ && /magics/ && /VERSION/' CMakeLists.txt | grep -Po '(?<=VERSION )[^ !\)]+')'
  edit CMAKE_BUILD_TYPE 'Release'
  edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DENABLE_PYTHON=ON -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX%'
  edit CMAKE_VERSION '3.13.4'
  edit CODES_UI_GIT_TAG 'develop'
  edit CODES_UI_VERSION '1.3.1'
  edit CPACK 'make package'
  edit CPUS '$(lscpu -p | grep -v '#' | wc -l)'
  edit CRAY_FFTW_VERSION '3.3.8.1'
  edit CRAY_NETCDF_VERSION '4.6.1.3'
  edit CRON_BASEPREFIX '%BASEDIR%/install'
  edit CRON_EMAIL ''
  edit CRON_ENABLED 'False'
  edit CRON_GIT_TAG 'develop'
  edit CUSTOM_COMMANDS ''
  edit DATAFLOW_SANITIZER 'False'
  edit DEFAULT_GNU_COMPILER 'gnu.73'
  edit ECBUILD '%BASEPREFIX%/ecbuild/develop/bin/ecbuild'
  edit ECBUILD_CLONE 'False'
  edit ECBUILD_CLONE_COMMIT 'master'
  edit ECBUILD_GIT_TAG 'develop'
  edit ECBUILD_REPO '%GIT_REPO_PREFIX%/ecsdk/ecbuild.git'
  edit ECBUILD_VERSION '3.3.2'
  edit ECCODES_EXTENSIONS_GIT_TAG 'develop'
  edit ECCODES_EXTENSIONS_VERSION '2.17.0'
  edit ECCODES_GIT_TAG 'develop'
  edit ECCODES_PYTHON_VERSION '0.9.6'
  edit ECCODES_VERSION '2.17.1'
  edit ECFLOW5_LATEST_VERSION '5.4.0'
  edit ECFLOW_VERSION '5.4.1'
  edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/Magics'
  edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/Magics'
  edit ECF_JOB_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_JOB% %ECF_JOBOUT%'
  edit ECF_KILL_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% kill'
  edit ECF_STATUS_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% stat'
  edit ECF_TRIES '1'
  edit ECKIT_GIT_TAG 'develop'
  edit ECKIT_VERSION '1.10.1'
  edit ECQUIS_VERSION '1.0.0'
  edit EMOS_MOD_VERSION '462'
  edit EMOS_PA_VERSION '000462'
  edit EXPORTS '"export ECACCOUNT=ecodmdma"'
  edit EXPORT_VARIABLES 'False'
  edit FCKIT_GIT_TAG '0.6.3'
  edit FCKIT_VERSION '0.6.3'
  edit FDB_VINTAGE_GIT_TAG '4.10.0'
  edit FDB_VINTAGE_VERSION '4.10.0'
  edit FFTW_VERSION '3.3.8'
  edit FFTW_VERSION_CRAY '3.3.4.11'
  edit GIT 'git'
  edit GITHUB_PACKAGE_NAME '%PACKAGE_NAME%'
  edit GITHUB_REPO 'git@github.com:ecmwf'
  edit GITREPO '%GIT_REPO_PREFIX%/mag/magics.git'
  edit GIT_CLONE 'True'
  edit GIT_CLONE_OPTIONS '--depth=1 --branch %GIT_TAG%'
  edit GIT_METABUILDER_TAG 'ci'
  edit GIT_REPO_PREFIX 'ssh://git@git.ecmwf.int:7999'
  edit GIT_TAG 'develop'
  edit GIT_VERSION '2.26.0'
  edit GRIB_API_VERSION '1.28.0'
  edit INSTALLABLE_TO_USR_LOCAL_APPS 'True'
  edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop'
  edit IWYU_VERSION '0.11'
  edit KSH_EXTRACT_PACKAGE_VERSION '`grep metabuilder_version VERSION.cmake | awk '{print $4}' | sed -e 's/"//g' | sed -e s/\)//g`'
  edit LAPACK_VERSION '3.5.0'
  edit LIBEMOS_GIT_TAG 'develop'
  edit LIBEMOS_VERSION '4.6.2'
  edit LINTERS_REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}'
  edit LSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/lsan.txt'
  edit MAGICS_GIT_TAG 'develop'
  edit MAGICS_PYTHON_VERSION 'master'
  edit MAGICS_VERSION '4.3.3'
  edit MAKE 'make'
  edit MARS_CLIENT_GIT_TAG '6.27.2.2'
  edit MARS_CLIENT_PACKAGE_GIT_TAG '6.27.6'
  edit MARS_CLIENT_PACKAGE_VERSION '6.27.6'
  edit MARS_CLIENT_VERSION '6.27.2.2'
  edit MEMORY_SANITIZER 'False'
  edit METABUILDER_HOME '/var/tmp/ma0/workspace/metabuilder'
  edit METAB_INSTALL_REMOVE_WRITE_PERMISSIONS 'True'
  edit METAB_PYTHON_VERSION 'python3'
  edit METKIT_GIT_TAG 'develop'
  edit METKIT_VERSION '1.4.6'
  edit METVIEW_GIT_TAG 'develop'
  edit METVIEW_PYTHON_VERSION '1.4.2'
  edit METVIEW_VERSION '5.8.3'
  edit MIR_BUNDLE_VERSION '0.9.0.0'
  edit MIR_GIT_TAG 'develop'
  edit MIR_VERSION '1.5.0'
  edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load proj/%PROJ_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
  edit MODULE_ACTIONS_FIRST ''
  edit MODULE_ACTIONS_HDR 'module_actions.h'
  edit MODULE_CHECKING_HDR 'module_checking.h'
  edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload boost" "module unload odb_api" "module unload eckit"'
  edit MODULE_LOAD_SUPPORTED 'True'
  edit MODULE_NAME 'Magics'
  edit MODULE_VERSION '%GIT_TAG%'
  edit NETCDF4_VERSION '4.7.4'
  edit NETCDF_LIB ''
  edit ODB_API_GIT_TAG 'develop'
  edit ODB_API_VERSION '0.19.4'
  edit ODB_GIT_TAG 'develop'
  edit ODB_TOOLS_GIT_TAG 'develop'
  edit ODB_TOOLS_VERSION '0.18.3'
  edit ODB_VERSION '1.0.11'
  edit ODC_GIT_TAG 'develop'
  edit ODC_VERSION '1.0.3'
  edit OPENJPEG_VERSION '2.3.1'
  edit PACKAGE 'Magics'
  edit PACKAGE_NAME 'Magics'
  edit PACKAGE_OWNERS 'cgs@ecmwf.int cgm@ecmwf.int ma0@ecmwf.int'
  edit PACKAGE_VERSION '%GIT_TAG%'
  edit PGEN_VERSION '1.7.1.0'
  edit PGI_LIB '/usr/local/apps/pgi/pgi-10.8/linux86-64/10.8/libso'
  edit PKG_BOOST_VERSION '1.71.0'
  edit PREFIX_APPEND ''
  edit PROJ4_VERSION '5.2.0'
  edit PROJ_VERSION '6.3.1'
  edit PYTHON2_VERSION '2.7.15-01'
  edit PYTHON3_VERSION '3.6.10-01'
  edit PYTHON3_VERSION_MAJOR_MINOR '3.6'
  edit PYTHON_GITREPO 'This-needs-tobe-overriden'
  edit PYTHON_PACKAGE 'This-needs-tobe-overriden'
  edit QT_VERSION '5.12.3'
  edit QUEUE 'build'
  edit REMOTE_HOST 'polonius'
  edit REUSE_BUILD_DIR 'False'
  edit SAFE_STACK_SANITIZER 'False'
  edit SANITIZER 'False'
  edit SANITIZER_REFERENCE_DIRECTORY '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
  edit SPOT_LIB '/usr/local/apps/spot_database/new'
  edit SRCDIR '%BASEDIR%/%PACKAGE%'
  edit STATIC_ANALYSER_ABORT_ON_ERROR 'False'
  edit SWIG_VERSION '3.0.8'
  edit SYNC_RELEASE '2020.04.0'
  edit SYNC_RELEASE_CI_STAGE 'develop'
  edit TARBALL_DIR '/perm/ma/ma0/tmp_tarballs/ci'
  edit TAR_PACKAGE '%PACKAGE%'
  edit TEST 'ctest --output-on-failure -j${CPUS} '
  edit THREAD_SANITIZER 'False'
  edit TRANSI_GIT_TAG 'develop'
  edit TRANSI_VERSION '0.7.1'
  edit TSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
  edit UBSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/ubsan.txt'
  edit UB_SANITIZER 'False'
  edit USER 'ma0'
  edit USE_WORKDIR 'False'
  edit VALGRIND_VERSION '3.15.0'
  edit WORKDIR '%BASEDIR%/workdir/%ECF_NAME%'
  edit WSDIR '/var/tmp/tmpdir/deploy/jtmp.20739'
  label branch "develop"
  task regenerate
    defstatus complete
    edit ECF_JOB_CMD '%ECF_JOB% > %ECF_JOBOUT% 2>&1'
    edit ECF_KILL_CMD 'kill -15 %ECF_RID%'
    edit ECF_PASS 'FREE'
    edit ECF_STATUS_CMD 'ps --sid %ECF_RID% -f'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
    label date ""
    label metabuilder_branch "ci"
  family cca
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DENABLE_NETCDF=ON -DENABLE_PYTHON=ON -DLAPACK_FOUND=TRUE -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX%'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/Magics/cca'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/Magics/cca:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cca-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export CRAY_ADD_RPATH=yes" "export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '24000'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load proj/%PROJ_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload boost" "module unload odb_api" "module unload eckit" "module unload atp"'
    edit M_SITE 'cca'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'cca'
    edit SITE 'cca'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DENABLE_NETCDF=ON -DENABLE_PYTHON=ON -DLAPACK_FOUND=TRUE -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load proj/%PROJ_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_link
        trigger install == complete
      task install_log
        trigger install_link == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family ccb
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DENABLE_NETCDF=ON -DENABLE_PYTHON=ON -DLAPACK_FOUND=TRUE -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX%'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/Magics/ccb'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/Magics/ccb:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'ccb-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export CRAY_ADD_RPATH=yes" "export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '24000'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load proj/%PROJ_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload boost" "module unload odb_api" "module unload eckit" "module unload atp"'
    edit M_SITE 'ccb'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'ccb'
    edit SITE 'ccb'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DENABLE_NETCDF=ON -DENABLE_PYTHON=ON -DLAPACK_FOUND=TRUE -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load proj/%PROJ_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_link
        trigger install == complete
      task install_log
        trigger install_link == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family ecgb
    defstatus suspended
    edit CI_LIMIT '10'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DENABLE_METVIEW=OFF -DENABLE_NETCDF=ON -DENABLE_ODB=ON -DENABLE_PYTHON=ON -DENABLE_REGRESSION=OFF -DENABLE_SPOT=OFF -DMAGICS_BUILD=ecgb -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX%'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'ecgb'
    edit QUEUE 'express'
    edit REMOTE_HOST 'ecgb'
    edit SITE 'ecgb'
    edit WSDIR '$SCRATCH/metabuilds/%ECF_HOST%_%ECF_PORT%/ecgb/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task build
      trigger config == complete
      edit ECF_TRIES '2'
      label host ""
      label options ""
    task test
      trigger build == complete
      edit ECF_TRIES '2'
      edit TEST_DIR '%BUILDIR%'
    task install
      trigger test == complete
      edit ECF_TRIES '2'
    task install_link
      trigger install == complete
    task install_log
      trigger install_link == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install_log == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family leap42
    defstatus suspended
    edit CI_LIMIT '7'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DENABLE_CAIRO=ON -DENABLE_METVIEW=OFF -DENABLE_NETCDF=ON -DENABLE_ODB=ON -DENABLE_PYTHON=ON -DENABLE_REGRESSION=OFF -DENABLE_SPOT=OFF -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX%'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load proj/%PROJ_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
    edit M_SITE 'leap42'
    edit REMOTE_HOST 'leap42-deploy'
    edit SITE 'leap42'
    edit WSDIR '/tmp/metabuilds/%ECF_HOST%_%ECF_PORT%/leap42/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DENABLE_CAIRO=ON -DENABLE_METVIEW=OFF -DENABLE_NETCDF=ON -DENABLE_ODB=ON -DENABLE_PYTHON=ON -DENABLE_REGRESSION=OFF -DENABLE_SPOT=OFF -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.73'
      edit GIT_CLONE_OPTIONS ''
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load proj/%PROJ_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task pull_from_github
        trigger git_srcs == complete
      task make_build_dir
        trigger pull_from_github == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task create_tarball
        trigger install == complete
        edit CMAKE 'cmake'
        edit EXPORTS '"export ECACCOUNT=ecodmdma" "export ecbuild_ROOT=%BASEPREFIX%/ecbuild/develop"'
        label tar ""
        label tardir ""
        label error ""
      task upload_tarball
        trigger create_tarball == complete
        edit METAB_PYTHON_VERSION 'python'
        edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
        edit UPLOAD_TARBALL '<package.h>'
        label url ""
        label space_key ""
        label msg ""
      task install_link
        trigger upload_tarball == complete
      task push_to_github
        trigger install_link == complete
      task push_to_bitbucket
        trigger push_to_github == complete
      task install_log
        trigger push_to_bitbucket == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.93
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DENABLE_CAIRO=ON -DENABLE_METVIEW=OFF -DENABLE_NETCDF=ON -DENABLE_ODB=ON -DENABLE_PYTHON=ON -DENABLE_REGRESSION=OFF -DENABLE_SPOT=OFF -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX%'
      edit COMPILER 'GNU.93'
      edit EXPORTS '"export ECACCOUNT=ecodmdma" "export LD_LIBRARY_PATH=/usr/local/apps/gcc/9.3.0/lib64/"'
      edit MODULE_ACTIONS '"module swap gnu/9.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load proj/%PROJ_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_link
        trigger install == complete
      task install_log
        trigger install_link == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DENABLE_CAIRO=ON -DENABLE_METVIEW=OFF -DENABLE_NETCDF=ON -DENABLE_ODB=ON -DENABLE_PYTHON=ON -DENABLE_REGRESSION=OFF -DENABLE_SPOT=OFF -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX%'
      edit COMPILER 'CLANG.90'
      edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load proj/%PROJ_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_link
        trigger install == complete
      task install_log
        trigger install_link == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90.linters
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DENABLE_CAIRO=ON -DENABLE_METVIEW=OFF -DENABLE_NETCDF=ON -DENABLE_ODB=ON -DENABLE_PYTHON=ON -DENABLE_REGRESSION=OFF -DENABLE_SPOT=OFF -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX%'
      edit COMPILER 'CLANG.90.LINTERS'
      edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load proj/%PROJ_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task cppcheck
        trigger config == complete
        edit CPPCHECK_FLAGS '--template='[{file}:{line}]: {severity} : {id} : {message}' --suppress=syntaxError --inline-suppr --platform=unix64 --enable=performance --enable=portability --enable=information --std=c11 --std=c++14 -DDEBUG=1 -q -j8'
        edit CPPCHECK_VERSION '1.89'
        edit REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}/cpp_check'
      task iwyu
        trigger cppcheck == complete
        edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load proj/%PROJ_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%" "module load iwyu/%IWYU_VERSION%"'
      family clang_linters
        trigger iwyu == complete
        limit clang_linters 1
        inlimit clang_linters
        task clang_tidy
          edit REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}/clang_tidy'
        task static_analyser
          edit REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}/static_analyser'
          edit SCAN_BUILD_OPTIONS '--force-analyze-debug-code -o /perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}/static_analyser -k make -j8'
          edit USE_CLANG_CHECK 'False'
          label scan_view ""
      endfamily
      task wipe_build
        trigger clang_linters == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90.address_sanitizer
      edit ADDRESS_SANITIZER 'True'
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_FLAGS="-fsanitize=address -fno-omit-frame-pointer" -DCMAKE_C_FLAGS="-fsanitize=address -fno-omit-frame-pointer" -DCMAKE_EXE_LINKER_FLAGS=-fsanitize=address -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DENABLE_CAIRO=ON -DENABLE_FORTRAN=OFF -DENABLE_METVIEW=OFF -DENABLE_NETCDF=ON -DENABLE_ODB=ON -DENABLE_PYTHON=OFF -DENABLE_REGRESSION=OFF -DENABLE_SPOT=OFF -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX%'
      edit COMPILER 'CLANG.90.ADDRESS_SANITIZER'
      edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load proj/%PROJ_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit SANITIZER 'True'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit TEST_DIR '%BUILDIR%'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90.ub_sanitizer
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_FLAGS="-fsanitize=undefined -fno-omit-frame-pointer -fno-sanitize-recover=all" -DCMAKE_C_FLAGS="-fsanitize=undefined -fno-omit-frame-pointer" -DCMAKE_EXE_LINKER_FLAGS="-fsanitize=undefined -Wl,-rpath,/usr/local/apps/clang/7.0.1/lib/clang/7.0.1/lib/linux/ -shared-libasan" -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DENABLE_CAIRO=ON -DENABLE_FORTRAN=OFF -DENABLE_METVIEW=OFF -DENABLE_NETCDF=ON -DENABLE_ODB=ON -DENABLE_PYTHON=OFF -DENABLE_REGRESSION=OFF -DENABLE_SPOT=OFF -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX%'
      edit COMPILER 'CLANG.90.UB_SANITIZER'
      edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load proj/%PROJ_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit SANITIZER 'True'
      edit UB_SANITIZER 'True'
      edit UB_SANITIZER_LD_PRELOAD 'export LD_PRELOAD=/usr/local/apps/clang/9.0.1/lib64/libubsan.so'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit TEST_DIR '%BUILDIR%'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family lxc
    defstatus suspended
    edit CI_LIMIT '10'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DENABLE_CAIRO=ON -DENABLE_METVIEW=OFF -DENABLE_NETCDF=ON -DENABLE_ODB=ON -DENABLE_PYTHON=ON -DENABLE_REGRESSION=OFF -DENABLE_SPOT=OFF -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX%'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'lxc'
    edit QUEUE 'express'
    edit REMOTE_HOST 'lxc'
    edit SITE 'lxc'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxc/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task build
      trigger config == complete
      edit ECF_TRIES '2'
      label host ""
      label options ""
    task test
      trigger build == complete
      edit ECF_TRIES '2'
      edit TEST_DIR '%BUILDIR%'
    task install
      trigger test == complete
      edit ECF_TRIES '2'
    task install_link
      trigger install == complete
    task install_log
      trigger install_link == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install_log == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
endsuite
suite magics_python
  edit ADDRESS_SANITIZER 'False'
  edit ADD_ACTION_TRIGGERS 'True'
  edit AEC_VERSION '1.0.4'
  edit ALLOW_INSTALL_TO_OVERWRITE 'True'
  edit ASAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
  edit ATLAS_GIT_TAG 'develop'
  edit ATLAS_VERSION '0.20.2'
  edit BANNER_TASK_END 'banner %TASK%-end || true'
  edit BANNER_TASK_START 'banner %TASK%-start || true'
  edit BASEDIR '%WSDIR%/%PACKAGE%'
  edit BASEPREFIX '$PERM/install/%SITE%'
  edit BUFRDC_EMOSLIB_VERSION '4.6.3'
  edit BUFRDC_TABLES_GIT_TAG 'develop'
  edit BUFRDC_TABLES_VERSION '000415'
  edit BUILDIR '%BASEDIR%/builds'
  edit BUILD_ALWAYS_PARALLEL 'False'
  edit CFI_SANITIZER 'False'
  edit CFL_BASE_URL 'https://software-test.ecmwf.int/wiki'
  edit CFL_SPACE_KEY '<TODO>'
  edit CFL_TARBALL_COMMENT ''
  edit CFL_TITLE 'Releases'
  edit CHECK_SYNC_RELEASE 'False'
  edit CLANG_TIDY_ABORT_ON_ERROR 'False'
  edit CMAKE '%ECBUILD%'
  edit CMAKELIST_EXTRACT_PACKAGE_VERSION '$(awk '/^project/ && /%PACKAGE%/ && /VERSION/ {for (I=1;I<=NF;I++) if ($I == "VERSION") {print $(I+1)};}' CMakeLists.txt)'
  edit CMAKE_BUILD_TYPE 'RelWithDebInfo'
  edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE%'
  edit CMAKE_VERSION '3.13.4'
  edit CODES_UI_GIT_TAG 'develop'
  edit CODES_UI_VERSION '1.3.1'
  edit CPACK 'make package'
  edit CPUS '$(lscpu -p | grep -v '#' | wc -l)'
  edit CRAY_FFTW_VERSION '3.3.8.1'
  edit CRAY_NETCDF_VERSION '4.6.1.3'
  edit CRON_BASEPREFIX '%BASEDIR%/install'
  edit CRON_EMAIL ''
  edit CRON_ENABLED 'False'
  edit CRON_GIT_TAG 'develop'
  edit CUSTOM_COMMANDS ''
  edit DATAFLOW_SANITIZER 'False'
  edit DEFAULT_GNU_COMPILER 'gnu.73'
  edit ECBUILD '%BASEPREFIX%/ecbuild/develop/bin/ecbuild'
  edit ECBUILD_CLONE 'False'
  edit ECBUILD_CLONE_COMMIT 'master'
  edit ECBUILD_GIT_TAG 'develop'
  edit ECBUILD_REPO '%GIT_REPO_PREFIX%/ecsdk/ecbuild.git'
  edit ECBUILD_VERSION '3.3.2'
  edit ECCODES_EXTENSIONS_GIT_TAG 'develop'
  edit ECCODES_EXTENSIONS_VERSION '2.17.0'
  edit ECCODES_GIT_TAG 'develop'
  edit ECCODES_PYTHON_VERSION '0.9.6'
  edit ECCODES_VERSION '2.17.1'
  edit ECFLOW5_LATEST_VERSION '5.4.0'
  edit ECFLOW_VERSION '5.4.1'
  edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/magics_python'
  edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/magics_python'
  edit ECF_JOB_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_JOB% %ECF_JOBOUT%'
  edit ECF_KILL_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% kill'
  edit ECF_STATUS_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% stat'
  edit ECF_TRIES '1'
  edit ECKIT_GIT_TAG 'develop'
  edit ECKIT_VERSION '1.10.1'
  edit ECQUIS_VERSION '1.0.0'
  edit EMOS_MOD_VERSION '462'
  edit EMOS_PA_VERSION '000462'
  edit EXPORTS '"export ECACCOUNT=ecodmdma"'
  edit EXPORT_VARIABLES 'False'
  edit FCKIT_GIT_TAG '0.6.3'
  edit FCKIT_VERSION '0.6.3'
  edit FDB_VINTAGE_GIT_TAG '4.10.0'
  edit FDB_VINTAGE_VERSION '4.10.0'
  edit FFTW_VERSION '3.3.8'
  edit FFTW_VERSION_CRAY '3.3.4.11'
  edit GIT 'git'
  edit GITHUB_PACKAGE_NAME 'magics-python'
  edit GITHUB_REPO 'git@github.com:ecmwf'
  edit GITREPO '%GIT_REPO_PREFIX%/mag/magics-python.git'
  edit GIT_CLONE 'True'
  edit GIT_CLONE_OPTIONS '--depth=1 --branch %GIT_TAG%'
  edit GIT_METABUILDER_TAG 'ci'
  edit GIT_REPO_PREFIX 'ssh://git@git.ecmwf.int:7999'
  edit GIT_TAG 'develop'
  edit GIT_VERSION '2.26.0'
  edit GRIB_API_VERSION '1.28.0'
  edit INSTALLABLE_TO_USR_LOCAL_APPS 'True'
  edit INSTALL_DIR '%BASEPREFIX%/Magics/develop'
  edit IWYU_VERSION '0.11'
  edit KSH_EXTRACT_PACKAGE_VERSION '`grep _VERSION_STR VERSION.cmake | grep -o '[0-9]*\.[0-9]*\.[0-9]*'`'
  edit LAPACK_VERSION '3.5.0'
  edit LIBEMOS_GIT_TAG 'develop'
  edit LIBEMOS_VERSION '4.6.2'
  edit LINTERS_REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}'
  edit LSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/lsan.txt'
  edit MAGICS_GIT_TAG 'develop'
  edit MAGICS_PYTHON_VERSION 'master'
  edit MAGICS_VERSION '4.3.3'
  edit MAKE 'make'
  edit MARS_CLIENT_GIT_TAG '6.27.2.2'
  edit MARS_CLIENT_PACKAGE_GIT_TAG '6.27.6'
  edit MARS_CLIENT_PACKAGE_VERSION '6.27.6'
  edit MARS_CLIENT_VERSION '6.27.2.2'
  edit MEMORY_SANITIZER 'False'
  edit METABUILDER_HOME '/var/tmp/ma0/workspace/metabuilder'
  edit METAB_INSTALL_REMOVE_WRITE_PERMISSIONS 'True'
  edit METAB_PYTHON_VERSION 'python3'
  edit METKIT_GIT_TAG 'develop'
  edit METKIT_VERSION '1.4.6'
  edit METVIEW_GIT_TAG 'develop'
  edit METVIEW_PYTHON_VERSION '1.4.2'
  edit METVIEW_VERSION '5.8.3'
  edit MIR_BUNDLE_VERSION '0.9.0.0'
  edit MIR_GIT_TAG 'develop'
  edit MIR_VERSION '1.5.0'
  edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module swap gnu/7.3.0"'
  edit MODULE_ACTIONS_FIRST ''
  edit MODULE_ACTIONS_HDR 'module_actions.h'
  edit MODULE_CHECKING_HDR 'module_checking.h'
  edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow"'
  edit MODULE_LOAD_SUPPORTED 'True'
  edit MODULE_NAME 'magics_python'
  edit MODULE_VERSION '%GIT_TAG%'
  edit NETCDF4_VERSION '4.7.4'
  edit ODB_API_GIT_TAG 'develop'
  edit ODB_API_VERSION '0.19.4'
  edit ODB_GIT_TAG 'develop'
  edit ODB_TOOLS_GIT_TAG 'develop'
  edit ODB_TOOLS_VERSION '0.18.3'
  edit ODB_VERSION '1.0.11'
  edit ODC_GIT_TAG 'develop'
  edit ODC_VERSION '1.0.3'
  edit OPENJPEG_VERSION '2.3.1'
  edit PACKAGE 'magics_python'
  edit PACKAGE_NAME 'magics_python'
  edit PACKAGE_OWNERS 'cgs@ecmwf.int cgm@ecmwf.int ma0@ecmwf.int'
  edit PACKAGE_VERSION '%GIT_TAG%'
  edit PGEN_VERSION '1.7.1.0'
  edit PKG_BOOST_VERSION '1.71.0'
  edit PREFIX_APPEND ''
  edit PROJ4_VERSION '5.2.0'
  edit PROJ_VERSION '6.3.1'
  edit PYTHON2_INSTALL 'True'
  edit PYTHON2_VERSION '2.7.15-01'
  edit PYTHON3_VERSION '3.6.10-01'
  edit PYTHON3_VERSION_MAJOR_MINOR '3.6'
  edit PYTHON_GITREPO 'This-needs-tobe-overriden'
  edit PYTHON_PACKAGE 'This-needs-tobe-overriden'
  edit QT_VERSION '5.12.3'
  edit QUEUE 'build'
  edit REMOTE_HOST 'polonius'
  edit REUSE_BUILD_DIR 'False'
  edit SAFE_STACK_SANITIZER 'False'
  edit SANITIZER 'False'
  edit SANITIZER_REFERENCE_DIRECTORY '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
  edit SRCDIR '%BASEDIR%/%PACKAGE%'
  edit STATIC_ANALYSER_ABORT_ON_ERROR 'False'
  edit SWIG_VERSION '3.0.8'
  edit SYNC_RELEASE '2020.04.0'
  edit SYNC_RELEASE_CI_STAGE 'develop'
  edit TARBALL_DIR '/perm/ma/ma0/tmp_tarballs/ci'
  edit TAR_PACKAGE '%PACKAGE%'
  edit TEST 'ctest --output-on-failure -j${CPUS} '
  edit THREAD_SANITIZER 'False'
  edit TRANSI_GIT_TAG 'develop'
  edit TRANSI_VERSION '0.7.1'
  edit TSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
  edit UBSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/ubsan.txt'
  edit UB_SANITIZER 'False'
  edit USER 'ma0'
  edit USE_WORKDIR 'False'
  edit VALGRIND_VERSION '3.15.0'
  edit WORKDIR '%BASEDIR%/workdir/%ECF_NAME%'
  edit WSDIR '/var/tmp/tmpdir/deploy/jtmp.20739'
  label branch "develop"
  task regenerate
    defstatus complete
    edit ECF_JOB_CMD '%ECF_JOB% > %ECF_JOBOUT% 2>&1'
    edit ECF_KILL_CMD 'kill -15 %ECF_RID%'
    edit ECF_PASS 'FREE'
    edit ECF_STATUS_CMD 'ps --sid %ECF_RID% -f'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
    label date ""
    label metabuilder_branch "ci"
  family cca
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DLAPACK_FOUND=TRUE'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/magics_python/cca'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/magics_python/cca:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cca-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '24000'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cca'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'cca'
    edit SITE 'cca'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task add_write_perm
        trigger git_srcs == complete
      task install_python
        trigger add_write_perm == complete
        edit ECF_TRIES '2'
        label version ""
        label install ""
      task install_log
        trigger install_python == complete
        edit ECF_TRIES '2'
      task wipe_src
        trigger install_log == complete
    endfamily
  endfamily
  family ccb
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DLAPACK_FOUND=TRUE'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/magics_python/ccb'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/magics_python/ccb:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'ccb-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '24000'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'ccb'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'ccb'
    edit SITE 'ccb'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DLAPACK_FOUND=TRUE'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task add_write_perm
        trigger git_srcs == complete
      task install_python
        trigger add_write_perm == complete
        edit ECF_TRIES '2'
        label version ""
        label install ""
      task install_log
        trigger install_python == complete
        edit ECF_TRIES '2'
      task wipe_src
        trigger install_log == complete
    endfamily
  endfamily
  family ecgb
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'ecgb'
    edit QUEUE 'express'
    edit REMOTE_HOST 'ecgb'
    edit SITE 'ecgb'
    edit WSDIR '$SCRATCH/metabuilds/%ECF_HOST%_%ECF_PORT%/ecgb/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task add_write_perm
      trigger git_srcs == complete
    task install_python
      trigger add_write_perm == complete
      edit ECF_TRIES '2'
      label version ""
      label install ""
    task install_log
      trigger install_python == complete
      edit ECF_TRIES '2'
    task wipe_src
      trigger install_log == complete
  endfamily
  family leap42
    defstatus suspended
    edit CI_LIMIT '7'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
    edit M_SITE 'leap42'
    edit REMOTE_HOST 'leap42-deploy'
    edit SITE 'leap42'
    edit WSDIR '/tmp/metabuilds/%ECF_HOST%_%ECF_PORT%/leap42/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON'
      edit COMPILER 'GNU.73'
      edit GIT_CLONE_OPTIONS ''
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task add_write_perm
        trigger make_build_dir == complete
      task install_python
        trigger add_write_perm == complete
        edit ECF_TRIES '2'
        label version ""
        label install ""
      task post_install_test
        trigger install_python == complete
      task install_log
        trigger post_install_test == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.93
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON'
      edit COMPILER 'GNU.93'
      edit EXPORTS '"export ECACCOUNT=ecodmdma" "export LD_LIBRARY_PATH=/usr/local/apps/gcc/9.3.0/lib64/"'
      edit MODULE_ACTIONS '"module swap gnu/9.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task add_write_perm
        trigger make_build_dir == complete
      task install_python
        trigger add_write_perm == complete
        edit ECF_TRIES '2'
        label version ""
        label install ""
      task post_install_test
        trigger install_python == complete
      task install_log
        trigger post_install_test == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON'
      edit COMPILER 'CLANG.90'
      edit INSTALL_DIR '%BASEPREFIX%/Magics/develop%INSTALL_POSTFIX%'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task add_write_perm
        trigger make_build_dir == complete
      task install_python
        trigger add_write_perm == complete
        edit ECF_TRIES '2'
        label version ""
        label install ""
      task post_install_test
        trigger install_python == complete
      task install_log
        trigger post_install_test == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family lxc
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'lxc'
    edit QUEUE 'express'
    edit REMOTE_HOST 'lxc'
    edit SITE 'lxc'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxc/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task add_write_perm
      trigger git_srcs == complete
    task install_python
      trigger add_write_perm == complete
      edit ECF_TRIES '2'
      label version ""
      label install ""
    task install_log
      trigger install_python == complete
      edit ECF_TRIES '2'
    task wipe_src
      trigger install_log == complete
  endfamily
endsuite
suite metview
  edit ADDRESS_SANITIZER 'False'
  edit ADD_ACTION_TRIGGERS 'True'
  edit AEC_VERSION '1.0.4'
  edit ALLOW_INSTALL_TO_OVERWRITE 'True'
  edit ASAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
  edit ATLAS_GIT_TAG 'develop'
  edit ATLAS_VERSION '0.20.2'
  edit BANNER_TASK_END 'banner %TASK%-end || true'
  edit BANNER_TASK_START 'banner %TASK%-start || true'
  edit BASEDIR '%WSDIR%/%PACKAGE%'
  edit BASEPREFIX '$PERM/install/%SITE%'
  edit BUFRDC_EMOSLIB_VERSION '4.6.3'
  edit BUFRDC_TABLES_GIT_TAG 'develop'
  edit BUFRDC_TABLES_VERSION '000415'
  edit BUFR_TABLES '/usr/local/apps/libemos/tables/bufr/text/%BUFRDC_TABLES_VERSION%'
  edit BUILDIR '%BASEDIR%/builds'
  edit BUILD_ALWAYS_PARALLEL 'False'
  edit CFI_SANITIZER 'False'
  edit CFL_BASE_URL 'https://software-test.ecmwf.int/wiki'
  edit CFL_SPACE_KEY 'METV'
  edit CFL_TARBALL_COMMENT ''
  edit CFL_TITLE 'Releases'
  edit CHECK_SYNC_RELEASE 'False'
  edit CLANG_TIDY_ABORT_ON_ERROR 'False'
  edit CMAKE '%ECBUILD%'
  edit CMAKELIST_EXTRACT_PACKAGE_VERSION '$(awk '/^project/ && /%PACKAGE%/ && /VERSION/ {for (I=1;I<=NF;I++) if ($I == "VERSION") {print $(I+1)};}' CMakeLists.txt)'
  edit CMAKE_BUILD_TYPE 'Release'
  edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DBUNDLE_SKIP_METVIEW-ECMWF-CONFIGS=1 -DCMAKE_BUILD_TYPE=Release -DCMAKE_EXE_LINKER_FLAGS=-s -DCMAKE_Fortran_COMPILER=gfortran -DCMAKE_PREFIX_PATH=/usr/local/apps/qt/5.7.0/5.7/gcc_64/ -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DECMWF=ON -DENABLE_BUFR=ON -DENABLE_CAIRO=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_FORTRAN=ON -DENABLE_MARS=ON -DENABLE_MARS_ODB=ON -DENABLE_METVIEW_FORTRAN=ON -DENABLE_NETCDF=ON -DENABLE_NETWORK_PROXY=ON -DENABLE_ODB=ON -DENABLE_OPERA_RADAR=ON -DENABLE_PYTHON=OFF -DENABLE_SCM_RTTOV=ON -DENABLE_USAGE_LOG=ON -DENABLE_WEATHER_ROOM=OFF -DEXTRA_CONFIG_PATH=%SRCDIR%/metview-ecmwf-configs -DEXTRA_TITLE=%SITE% -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLOG_DIR=/home/graphics/cgx/public -DMARS_LOCAL_HOME=/usr/local/apps/mars/configs/prod -DMETVIEW_BUNDLE_MAGICS=does/not/matter -DMETVIEW_SITE=ecmwf -DNO_PROXY_FOR=OFF -DODB1_MODULE=odb/${ODB_VERSION} -DODB1_VERSION=${ODB_VERSION} -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DPROXY_PORT=3333 -DPROXY_URL=proxy.ecmwf.int'
  edit CMAKE_VERSION '3.13.4'
  edit CODES_UI_GIT_TAG 'develop'
  edit CODES_UI_VERSION '1.3.1'
  edit CPACK 'make package'
  edit CPUS '$(lscpu -p | grep -v '#' | wc -l)'
  edit CRAY_FFTW_VERSION '3.3.8.1'
  edit CRAY_NETCDF_VERSION '4.6.1.3'
  edit CRON_BASEPREFIX '%BASEDIR%/install'
  edit CRON_EMAIL ''
  edit CRON_ENABLED 'False'
  edit CRON_GIT_TAG 'develop'
  edit CUSTOM_COMMANDS ''
  edit DATAFLOW_SANITIZER 'False'
  edit DEFAULT_GNU_COMPILER 'gnu.73'
  edit ECBUILD '%BASEPREFIX%/ecbuild/develop/bin/ecbuild'
  edit ECBUILD_CLONE 'False'
  edit ECBUILD_CLONE_COMMIT 'master'
  edit ECBUILD_GIT_TAG 'develop'
  edit ECBUILD_REPO '%GIT_REPO_PREFIX%/ecsdk/ecbuild.git'
  edit ECBUILD_VERSION '3.3.2'
  edit ECCODES_EXTENSIONS_GIT_TAG 'develop'
  edit ECCODES_EXTENSIONS_VERSION '2.17.0'
  edit ECCODES_GIT_TAG 'develop'
  edit ECCODES_PYTHON_VERSION '0.9.6'
  edit ECCODES_VERSION '2.17.1'
  edit ECFLOW5_LATEST_VERSION '5.4.0'
  edit ECFLOW_VERSION '5.4.1'
  edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/metview'
  edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/metview'
  edit ECF_JOB_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_JOB% %ECF_JOBOUT%'
  edit ECF_KILL_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% kill'
  edit ECF_STATUS_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% stat'
  edit ECF_TRIES '1'
  edit ECKIT_GIT_TAG 'develop'
  edit ECKIT_VERSION '1.10.1'
  edit ECQUIS_VERSION '1.0.0'
  edit EMOS_MOD_VERSION '462'
  edit EMOS_PA_VERSION '000462'
  edit EXPORTS '"export ECACCOUNT=ecodmdma" "export ODB_VERSION=%ODB_VERSION%" "export ftp_proxy=http://proxy.ecmwf.int:3333" "export http_proxy=http://proxy.ecmwf.int:3333" "export https_proxy=http://proxy.ecmwf.int:3333"'
  edit EXPORT_VARIABLES 'False'
  edit FCKIT_GIT_TAG 'develop'
  edit FCKIT_VERSION '0.6.3'
  edit FDB_VINTAGE_GIT_TAG 'develop'
  edit FDB_VINTAGE_VERSION '4.10.0'
  edit FFTW_VERSION '3.3.8'
  edit FFTW_VERSION_CRAY '3.3.4.11'
  edit GIT 'git'
  edit GITHUB_PACKAGE_NAME '%PACKAGE_NAME%'
  edit GITHUB_REPO 'git@github.com:ecmwf'
  edit GITREPO '%GIT_REPO_PREFIX%/metv/metview-magics-bundle.git'
  edit GITREPO_METVIEW_BUNDLE '%GIT_REPO_PREFIX%/metv/metview-bundle.git'
  edit GIT_CLONE 'True'
  edit GIT_CLONE_OPTIONS '--depth=1 --branch %GIT_TAG%'
  edit GIT_METABUILDER_TAG 'ci'
  edit GIT_REPO_PREFIX 'ssh://git@git.ecmwf.int:7999'
  edit GIT_TAG 'develop'
  edit GIT_TAG_METVIEW_BUNDLE 'release'
  edit GIT_VERSION '2.26.0'
  edit GRIB_API_VERSION '1.28.0'
  edit INSTALLABLE_TO_USR_LOCAL_APPS 'True'
  edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop'
  edit IWYU_VERSION '0.11'
  edit KSH_EXTRACT_PACKAGE_VERSION '`cd %SRCDIR%/metview && ./extract_metview_version.sh`'
  edit LAPACK_VERSION '3.5.0'
  edit LIBEMOS_GIT_TAG 'develop'
  edit LIBEMOS_VERSION '4.6.2'
  edit LINTERS_REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}'
  edit LSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/lsan.txt'
  edit MAGICS_GIT_TAG 'develop'
  edit MAGICS_PYTHON_VERSION 'master'
  edit MAGICS_VERSION '4.3.3'
  edit MAKE 'make'
  edit MARS_CLIENT_GIT_TAG 'develop'
  edit MARS_CLIENT_PACKAGE_GIT_TAG '6.27.6'
  edit MARS_CLIENT_PACKAGE_VERSION '6.27.6'
  edit MARS_CLIENT_VERSION '6.27.2.2'
  edit MEMORY_SANITIZER 'False'
  edit METABUILDER_HOME '/var/tmp/ma0/workspace/metabuilder'
  edit METAB_INSTALL_REMOVE_WRITE_PERMISSIONS 'True'
  edit METAB_PYTHON_VERSION 'python3'
  edit METKIT_GIT_TAG 'develop'
  edit METKIT_VERSION '1.4.6'
  edit METVIEW_GIT_TAG 'develop'
  edit METVIEW_PYTHON_VERSION '1.4.2'
  edit METVIEW_VERSION '5.8.3'
  edit MIR_BUNDLE_VERSION '0.9.0.0'
  edit MIR_GIT_TAG 'develop'
  edit MIR_VERSION '1.5.0'
  edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load proj/%PROJ_VERSION%" "module load fftw/%FFTW_VERSION%" "module load hdf5" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
  edit MODULE_ACTIONS_FIRST ''
  edit MODULE_ACTIONS_HDR 'module_actions.h'
  edit MODULE_CHECKING_HDR 'module_checking.h'
  edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow"'
  edit MODULE_LOAD_SUPPORTED 'True'
  edit MODULE_NAME 'metview'
  edit MODULE_VERSION '%GIT_TAG%'
  edit NETCDF4_VERSION '4.7.4'
  edit ODB_API_GIT_TAG 'develop'
  edit ODB_API_VERSION '0.19.4'
  edit ODB_GIT_TAG 'develop'
  edit ODB_TOOLS_GIT_TAG '0.18.3'
  edit ODB_TOOLS_VERSION '0.18.3'
  edit ODB_VERSION '1.0.11'
  edit ODC_GIT_TAG 'develop'
  edit ODC_VERSION '1.0.3'
  edit OPENJPEG_VERSION '2.3.1'
  edit PACKAGE 'metview'
  edit PACKAGE_NAME 'metview'
  edit PACKAGE_OWNERS 'cgi@ecmwf.int cgm@ecmwf.int'
  edit PACKAGE_VERSION '%GIT_TAG%'
  edit PGEN_VERSION '1.7.1.0'
  edit PKG_BOOST_VERSION '1.71.0'
  edit PREFIX_APPEND ''
  edit PREPROCESS_IN 'CMakeLists.in'
  edit PREPROCESS_OUT 'CMakeLists.txt'
  edit PROJ4_VERSION '5.2.0'
  edit PROJ_VERSION '6.3.1'
  edit PYTHON2_VERSION '2.7.15-01'
  edit PYTHON3_VERSION '3.6.10-01'
  edit PYTHON3_VERSION_MAJOR_MINOR '3.6'
  edit PYTHON_GITREPO 'This-needs-tobe-overriden'
  edit PYTHON_PACKAGE 'This-needs-tobe-overriden'
  edit QT_VERSION '5.12.3'
  edit QUEUE 'build'
  edit REMOTE_HOST 'polonius'
  edit REUSE_BUILD_DIR 'False'
  edit SAFE_STACK_SANITIZER 'False'
  edit SANITIZER 'False'
  edit SANITIZER_REFERENCE_DIRECTORY '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
  edit SRCDIR '%BASEDIR%/%PACKAGE%'
  edit STATIC_ANALYSER_ABORT_ON_ERROR 'False'
  edit SWIG_VERSION '3.0.8'
  edit SYNC_RELEASE '2020.04.0'
  edit SYNC_RELEASE_CI_STAGE 'develop'
  edit TARBALL_DIR '/perm/ma/ma0/tmp_tarballs/ci'
  edit TAR_PACKAGE 'Metview'
  edit TEST 'ctest --output-on-failure -j${CPUS} '
  edit THREAD_SANITIZER 'False'
  edit TRANSI_GIT_TAG 'develop'
  edit TRANSI_VERSION '0.7.1'
  edit TSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
  edit UBSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/ubsan.txt'
  edit UB_SANITIZER 'False'
  edit USER 'ma0'
  edit USE_WORKDIR 'False'
  edit VALGRIND_VERSION '3.15.0'
  edit WORKDIR '%BASEDIR%/workdir/%ECF_NAME%'
  edit WSDIR '/var/tmp/tmpdir/deploy/jtmp.20739'
  label branch "develop"
  task regenerate
    defstatus complete
    edit ECF_JOB_CMD '%ECF_JOB% > %ECF_JOBOUT% 2>&1'
    edit ECF_KILL_CMD 'kill -15 %ECF_RID%'
    edit ECF_PASS 'FREE'
    edit ECF_STATUS_CMD 'ps --sid %ECF_RID% -f'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
    label date ""
    label metabuilder_branch "ci"
  family cca
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DBUNDLE_SKIP_METVIEW-ECMWF-CONFIGS=1 -DCMAKE_BUILD_TYPE=Release -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-s -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_PREFIX_PATH=/usr/local/apps/qt/5.7.0/5.7/gcc_64/ -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DECMWF=ON -DENABLE_BUFR=ON -DENABLE_CAIRO=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_FORTRAN=ON -DENABLE_MARS=ON -DENABLE_MARS_ODB=ON -DENABLE_MARS_USER_ENVIRONMENT=ON -DENABLE_METVIEW_FORTRAN=ON -DENABLE_NETCDF=ON -DENABLE_NETWORK_PROXY=ON -DENABLE_ODB=ON -DENABLE_OPERA_RADAR=ON -DENABLE_PYTHON=OFF -DENABLE_SCM_RTTOV=ON -DENABLE_UI=OFF -DENABLE_USAGE_LOG=ON -DENABLE_WEATHER_ROOM=OFF -DEXTRA_CONFIG_PATH=%SRCDIR%/metview-ecmwf-configs -DEXTRA_TITLE=%SITE% -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_FOUND=TRUE -DLOG_DIR=/home/graphics/cgx/public -DMARS_LOCAL_HOME=/usr/local/apps/mars/configs/prod -DMETVIEW_BUNDLE_MAGICS=does/not/matter -DMETVIEW_SITE=ecmwf -DNO_PROXY_FOR=OFF -DODB1_MODULE=odb/${ODB_VERSION} -DODB1_VERSION=${ODB_VERSION} -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DPROXY_PORT=3333 -DPROXY_URL=proxy.ecmwf.int'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/metview/cca'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/metview/cca:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cca-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export ODB_VERSION=%ODB_VERSION%" "export TMPDIR=/tmp" "export ftp_proxy=http://proxy.ecmwf.int:3333" "export http_proxy=http://proxy.ecmwf.int:3333" "export https_proxy=http://proxy.ecmwf.int:3333"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '24000'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load proj/%PROJ_VERSION%" "module load fftw/%FFTW_VERSION%" "module load hdf5" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cca'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'cca'
    edit SITE 'cca'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.73
      edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DBUNDLE_SKIP_METVIEW-ECMWF-CONFIGS=1 -DCMAKE_BUILD_TYPE=Release -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_PREFIX_PATH=/usr/local/apps/qt/5.7.0/5.7/gcc_64/ -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DECMWF=ON -DENABLE_BUFR=ON -DENABLE_CAIRO=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_FORTRAN=ON -DENABLE_MARS=ON -DENABLE_MARS_ODB=ON -DENABLE_MARS_USER_ENVIRONMENT=ON -DENABLE_METVIEW_FORTRAN=ON -DENABLE_NETCDF=ON -DENABLE_NETWORK_PROXY=ON -DENABLE_ODB=ON -DENABLE_OPERA_RADAR=ON -DENABLE_PYTHON=OFF -DENABLE_SCM_RTTOV=ON -DENABLE_UI=OFF -DENABLE_USAGE_LOG=ON -DENABLE_WEATHER_ROOM=OFF -DEXTRA_CONFIG_PATH=%SRCDIR%/metview-ecmwf-configs -DEXTRA_TITLE=%SITE% -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_FOUND=TRUE -DLOG_DIR=/home/graphics/cgx/public -DMARS_LOCAL_HOME=/usr/local/apps/mars/configs/prod -DMETVIEW_BUNDLE_MAGICS=does/not/matter -DMETVIEW_SITE=ecmwf -DNO_PROXY_FOR=OFF -DODB1_MODULE=odb/${ODB_VERSION} -DODB1_VERSION=${ODB_VERSION} -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DPROXY_PORT=3333 -DPROXY_URL=proxy.ecmwf.int'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load proj/%PROJ_VERSION%" "module load fftw/%FFTW_VERSION%" "module load hdf5" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task git_update_repos
        trigger git_srcs == complete
        edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=Release -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_PREFIX_PATH=/usr/local/apps/qt/5.7.0/5.7/gcc_64/ -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DECMWF=ON -DENABLE_BUFR=ON -DENABLE_CAIRO=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_FORTRAN=ON -DENABLE_MARS=ON -DENABLE_MARS_ODB=ON -DENABLE_MARS_USER_ENVIRONMENT=ON -DENABLE_METVIEW_FORTRAN=ON -DENABLE_NETCDF=ON -DENABLE_NETWORK_PROXY=ON -DENABLE_ODB=ON -DENABLE_OPERA_RADAR=ON -DENABLE_PYTHON=OFF -DENABLE_SCM_RTTOV=ON -DENABLE_UI=OFF -DENABLE_USAGE_LOG=ON -DENABLE_WEATHER_ROOM=OFF -DEXTRA_CONFIG_PATH=%SRCDIR%/metview-ecmwf-configs -DEXTRA_TITLE=%SITE% -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_FOUND=TRUE -DLOG_DIR=/home/graphics/cgx/public -DMARS_LOCAL_HOME=/usr/local/apps/mars/configs/prod -DMETVIEW_BUNDLE_MAGICS=does/not/matter -DMETVIEW_SITE=ecmwf -DNO_PROXY_FOR=OFF -DODB1_MODULE=odb/${ODB_VERSION} -DODB1_VERSION=${ODB_VERSION} -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DPROXY_PORT=3333 -DPROXY_URL=proxy.ecmwf.int'
        edit PACKAGE_VERSION 'fake'
        label build ""
      task make_build_dir
        trigger git_update_repos == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%/metview'
      task set_write_permissions
        trigger test == complete
      task install
        trigger set_write_permissions == complete
        edit ECF_TRIES '2'
      task install_test
        trigger install == complete
      task install_log
        trigger install_test == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family ccb
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DBUNDLE_SKIP_METVIEW-ECMWF-CONFIGS=1 -DCMAKE_BUILD_TYPE=Release -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-s -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_PREFIX_PATH=/usr/local/apps/qt/5.7.0/5.7/gcc_64/ -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DECMWF=ON -DENABLE_BUFR=ON -DENABLE_CAIRO=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_FORTRAN=ON -DENABLE_MARS=ON -DENABLE_MARS_ODB=ON -DENABLE_MARS_USER_ENVIRONMENT=ON -DENABLE_METVIEW_FORTRAN=ON -DENABLE_NETCDF=ON -DENABLE_NETWORK_PROXY=ON -DENABLE_ODB=ON -DENABLE_OPERA_RADAR=ON -DENABLE_PYTHON=OFF -DENABLE_SCM_RTTOV=ON -DENABLE_UI=OFF -DENABLE_USAGE_LOG=ON -DENABLE_WEATHER_ROOM=OFF -DEXTRA_CONFIG_PATH=%SRCDIR%/metview-ecmwf-configs -DEXTRA_TITLE=%SITE% -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_FOUND=TRUE -DLOG_DIR=/home/graphics/cgx/public -DMARS_LOCAL_HOME=/usr/local/apps/mars/configs/prod -DMETVIEW_BUNDLE_MAGICS=does/not/matter -DMETVIEW_SITE=ecmwf -DNO_PROXY_FOR=OFF -DODB1_MODULE=odb/${ODB_VERSION} -DODB1_VERSION=${ODB_VERSION} -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DPROXY_PORT=3333 -DPROXY_URL=proxy.ecmwf.int'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/metview/ccb'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/metview/ccb:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'ccb-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export ODB_VERSION=%ODB_VERSION%" "export TMPDIR=/tmp" "export ftp_proxy=http://proxy.ecmwf.int:3333" "export http_proxy=http://proxy.ecmwf.int:3333" "export https_proxy=http://proxy.ecmwf.int:3333"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '24000'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load proj/%PROJ_VERSION%" "module load fftw/%FFTW_VERSION%" "module load hdf5" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'ccb'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'ccb'
    edit SITE 'ccb'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.73
      edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DBUNDLE_SKIP_METVIEW-ECMWF-CONFIGS=1 -DCMAKE_BUILD_TYPE=Release -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_PREFIX_PATH=/usr/local/apps/qt/5.7.0/5.7/gcc_64/ -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DECMWF=ON -DENABLE_BUFR=ON -DENABLE_CAIRO=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_FORTRAN=ON -DENABLE_MARS=ON -DENABLE_MARS_ODB=ON -DENABLE_MARS_USER_ENVIRONMENT=ON -DENABLE_METVIEW_FORTRAN=ON -DENABLE_NETCDF=ON -DENABLE_NETWORK_PROXY=ON -DENABLE_ODB=ON -DENABLE_OPERA_RADAR=ON -DENABLE_PYTHON=OFF -DENABLE_SCM_RTTOV=ON -DENABLE_UI=OFF -DENABLE_USAGE_LOG=ON -DENABLE_WEATHER_ROOM=OFF -DEXTRA_CONFIG_PATH=%SRCDIR%/metview-ecmwf-configs -DEXTRA_TITLE=%SITE% -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_FOUND=TRUE -DLOG_DIR=/home/graphics/cgx/public -DMARS_LOCAL_HOME=/usr/local/apps/mars/configs/prod -DMETVIEW_BUNDLE_MAGICS=does/not/matter -DMETVIEW_SITE=ecmwf -DNO_PROXY_FOR=OFF -DODB1_MODULE=odb/${ODB_VERSION} -DODB1_VERSION=${ODB_VERSION} -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DPROXY_PORT=3333 -DPROXY_URL=proxy.ecmwf.int'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load proj/%PROJ_VERSION%" "module load fftw/%FFTW_VERSION%" "module load hdf5" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task git_update_repos
        trigger git_srcs == complete
        edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=Release -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_PREFIX_PATH=/usr/local/apps/qt/5.7.0/5.7/gcc_64/ -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DECMWF=ON -DENABLE_BUFR=ON -DENABLE_CAIRO=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_FORTRAN=ON -DENABLE_MARS=ON -DENABLE_MARS_ODB=ON -DENABLE_MARS_USER_ENVIRONMENT=ON -DENABLE_METVIEW_FORTRAN=ON -DENABLE_NETCDF=ON -DENABLE_NETWORK_PROXY=ON -DENABLE_ODB=ON -DENABLE_OPERA_RADAR=ON -DENABLE_PYTHON=OFF -DENABLE_SCM_RTTOV=ON -DENABLE_UI=OFF -DENABLE_USAGE_LOG=ON -DENABLE_WEATHER_ROOM=OFF -DEXTRA_CONFIG_PATH=%SRCDIR%/metview-ecmwf-configs -DEXTRA_TITLE=%SITE% -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_FOUND=TRUE -DLOG_DIR=/home/graphics/cgx/public -DMARS_LOCAL_HOME=/usr/local/apps/mars/configs/prod -DMETVIEW_BUNDLE_MAGICS=does/not/matter -DMETVIEW_SITE=ecmwf -DNO_PROXY_FOR=OFF -DODB1_MODULE=odb/${ODB_VERSION} -DODB1_VERSION=${ODB_VERSION} -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DPROXY_PORT=3333 -DPROXY_URL=proxy.ecmwf.int'
        edit PACKAGE_VERSION 'fake'
        label build ""
      task make_build_dir
        trigger git_update_repos == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%/metview'
      task set_write_permissions
        trigger test == complete
      task install
        trigger set_write_permissions == complete
        edit ECF_TRIES '2'
      task install_test
        trigger install == complete
      task install_log
        trigger install_test == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family cct
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DBUNDLE_SKIP_METVIEW-ECMWF-CONFIGS=1 -DCMAKE_BUILD_TYPE=Release -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-s -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_PREFIX_PATH=/usr/local/apps/qt/5.7.0/5.7/gcc_64/ -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DECMWF=ON -DENABLE_BUFR=ON -DENABLE_CAIRO=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_FORTRAN=ON -DENABLE_MARS=ON -DENABLE_MARS_ODB=ON -DENABLE_MARS_USER_ENVIRONMENT=ON -DENABLE_METVIEW_FORTRAN=ON -DENABLE_NETCDF=ON -DENABLE_NETWORK_PROXY=ON -DENABLE_ODB=ON -DENABLE_OPERA_RADAR=ON -DENABLE_PYTHON=OFF -DENABLE_SCM_RTTOV=ON -DENABLE_UI=OFF -DENABLE_USAGE_LOG=ON -DENABLE_WEATHER_ROOM=OFF -DEXTRA_CONFIG_PATH=%SRCDIR%/metview-ecmwf-configs -DEXTRA_TITLE=%SITE% -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_FOUND=TRUE -DLOG_DIR=/home/graphics/cgx/public -DMARS_LOCAL_HOME=/usr/local/apps/mars/configs/prod -DMETVIEW_BUNDLE_MAGICS=does/not/matter -DMETVIEW_SITE=ecmwf -DNO_PROXY_FOR=OFF -DODB1_MODULE=odb/${ODB_VERSION} -DODB1_VERSION=${ODB_VERSION} -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DPROXY_PORT=3333 -DPROXY_URL=proxy.ecmwf.int'
    edit CPUS '$EC_total_tasks'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/metview/cct'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/metview/cct:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cct-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '4'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export ODB_VERSION=%ODB_VERSION%" "export TMPDIR=/tmp" "export ftp_proxy=http://proxy.ecmwf.int:3333" "export http_proxy=http://proxy.ecmwf.int:3333" "export https_proxy=http://proxy.ecmwf.int:3333"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '6000'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load proj/%PROJ_VERSION%" "module load fftw/%FFTW_VERSION%" "module load hdf5" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cct'
    edit QSUB_LP 'QSUB -lP %EC_TOTAL_TASKS%'
    edit QUEUE 'nf'
    edit REMOTE_HOST 'cct'
    edit SITE 'cct'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.73
      edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DBUNDLE_SKIP_METVIEW-ECMWF-CONFIGS=1 -DCMAKE_BUILD_TYPE=Release -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_PREFIX_PATH=/usr/local/apps/qt/5.7.0/5.7/gcc_64/ -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DECMWF=ON -DENABLE_BUFR=ON -DENABLE_CAIRO=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_FORTRAN=ON -DENABLE_MARS=ON -DENABLE_MARS_ODB=ON -DENABLE_MARS_USER_ENVIRONMENT=ON -DENABLE_METVIEW_FORTRAN=ON -DENABLE_NETCDF=ON -DENABLE_NETWORK_PROXY=ON -DENABLE_ODB=ON -DENABLE_OPERA_RADAR=ON -DENABLE_PYTHON=OFF -DENABLE_SCM_RTTOV=ON -DENABLE_UI=OFF -DENABLE_USAGE_LOG=ON -DENABLE_WEATHER_ROOM=OFF -DEXTRA_CONFIG_PATH=%SRCDIR%/metview-ecmwf-configs -DEXTRA_TITLE=%SITE% -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_FOUND=TRUE -DLOG_DIR=/home/graphics/cgx/public -DMARS_LOCAL_HOME=/usr/local/apps/mars/configs/prod -DMETVIEW_BUNDLE_MAGICS=does/not/matter -DMETVIEW_SITE=ecmwf -DNO_PROXY_FOR=OFF -DODB1_MODULE=odb/${ODB_VERSION} -DODB1_VERSION=${ODB_VERSION} -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DPROXY_PORT=3333 -DPROXY_URL=proxy.ecmwf.int'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load proj/%PROJ_VERSION%" "module load fftw/%FFTW_VERSION%" "module load hdf5" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task git_update_repos
        trigger git_srcs == complete
        edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=Release -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_PREFIX_PATH=/usr/local/apps/qt/5.7.0/5.7/gcc_64/ -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DECMWF=ON -DENABLE_BUFR=ON -DENABLE_CAIRO=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_FORTRAN=ON -DENABLE_MARS=ON -DENABLE_MARS_ODB=ON -DENABLE_MARS_USER_ENVIRONMENT=ON -DENABLE_METVIEW_FORTRAN=ON -DENABLE_NETCDF=ON -DENABLE_NETWORK_PROXY=ON -DENABLE_ODB=ON -DENABLE_OPERA_RADAR=ON -DENABLE_PYTHON=OFF -DENABLE_SCM_RTTOV=ON -DENABLE_UI=OFF -DENABLE_USAGE_LOG=ON -DENABLE_WEATHER_ROOM=OFF -DEXTRA_CONFIG_PATH=%SRCDIR%/metview-ecmwf-configs -DEXTRA_TITLE=%SITE% -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_FOUND=TRUE -DLOG_DIR=/home/graphics/cgx/public -DMARS_LOCAL_HOME=/usr/local/apps/mars/configs/prod -DMETVIEW_BUNDLE_MAGICS=does/not/matter -DMETVIEW_SITE=ecmwf -DNO_PROXY_FOR=OFF -DODB1_MODULE=odb/${ODB_VERSION} -DODB1_VERSION=${ODB_VERSION} -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DPROXY_PORT=3333 -DPROXY_URL=proxy.ecmwf.int'
        edit PACKAGE_VERSION 'fake'
        label build ""
      task make_build_dir
        trigger git_update_repos == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%/metview'
      task set_write_permissions
        trigger test == complete
      task install
        trigger set_write_permissions == complete
        edit ECF_TRIES '2'
      task install_test
        trigger install == complete
      task install_log
        trigger install_test == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family ecgb
    defstatus suspended
    edit CI_LIMIT '10'
    edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DBUNDLE_SKIP_METVIEW-ECMWF-CONFIGS=1 -DCMAKE_BUILD_TYPE=Release -DCMAKE_CXX_FLAGS="-march=native -O3" -DCMAKE_EXE_LINKER_FLAGS=-s -DCMAKE_Fortran_COMPILER=gfortran -DCMAKE_PREFIX_PATH=/usr/local/apps/qt/5.5.0/5.5/gcc_64/ -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DECMWF=ON -DENABLE_BUFR=ON -DENABLE_CAIRO=ON -DENABLE_CODE_WEBHELP=OFF -DENABLE_EXPERIMENTAL=OFF -DENABLE_FORTRAN=ON -DENABLE_MARS=ON -DENABLE_MARS_ODB=ON -DENABLE_METVIEW_FORTRAN=ON -DENABLE_NETCDF=ON -DENABLE_NETWORK_PROXY=ON -DENABLE_ODB=ON -DENABLE_OPERA_RADAR=ON -DENABLE_PYTHON=OFF -DENABLE_SCM_RTTOV=ON -DENABLE_USAGE_LOG=ON -DENABLE_WEATHER_ROOM=OFF -DEXTRA_CONFIG_PATH=%SRCDIR%/metview-ecmwf-configs -DEXTRA_TITLE=%SITE% -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DFLEXPART_EXE_PATH=/usr/local/apps/flexpart/902/bin/flexpart -DFLEXPART_RESOURCES_PATH=/usr/local/apps/flexpart/902/options -DFLEXPART_SPECIES_PATH=/usr/local/apps/flexpart/902/options/SPECIES -DFLEXTRA_PATH=/usr/local/apps/flextra/flextra -DLOG_DIR=/home/graphics/cgx/public -DMARS_LOCAL_HOME=/usr/local/apps/mars/configs/prod -DMETVIEW_BUNDLE_MAGICS=does/not/matter -DMETVIEW_SITE=ecmwf -DNO_PROXY_FOR=OFF -DODB1_MODULE=odb/${ODB_VERSION} -DODB1_VERSION=${ODB_VERSION} -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DPROXY_PORT=3333 -DPROXY_URL=proxy.ecmwf.int'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'ecgb'
    edit QUEUE 'express'
    edit REMOTE_HOST 'ecgb'
    edit SITE 'ecgb'
    edit WSDIR '$SCRATCH/metabuilds/%ECF_HOST%_%ECF_PORT%/ecgb/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task git_update_repos
      trigger git_srcs == complete
      edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=Release -DCMAKE_CXX_FLAGS="-march=native -O3" -DCMAKE_EXE_LINKER_FLAGS=-s -DCMAKE_Fortran_COMPILER=gfortran -DCMAKE_PREFIX_PATH=/usr/local/apps/qt/5.5.0/5.5/gcc_64/ -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DECMWF=ON -DENABLE_BUFR=ON -DENABLE_CAIRO=ON -DENABLE_CODE_WEBHELP=OFF -DENABLE_EXPERIMENTAL=OFF -DENABLE_FORTRAN=ON -DENABLE_MARS=ON -DENABLE_MARS_ODB=ON -DENABLE_METVIEW_FORTRAN=ON -DENABLE_NETCDF=ON -DENABLE_NETWORK_PROXY=ON -DENABLE_ODB=ON -DENABLE_OPERA_RADAR=ON -DENABLE_PYTHON=OFF -DENABLE_SCM_RTTOV=ON -DENABLE_USAGE_LOG=ON -DENABLE_WEATHER_ROOM=OFF -DEXTRA_CONFIG_PATH=%SRCDIR%/metview-ecmwf-configs -DEXTRA_TITLE=%SITE% -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DFLEXPART_EXE_PATH=/usr/local/apps/flexpart/902/bin/flexpart -DFLEXPART_RESOURCES_PATH=/usr/local/apps/flexpart/902/options -DFLEXPART_SPECIES_PATH=/usr/local/apps/flexpart/902/options/SPECIES -DFLEXTRA_PATH=/usr/local/apps/flextra/flextra -DLOG_DIR=/home/graphics/cgx/public -DMARS_LOCAL_HOME=/usr/local/apps/mars/configs/prod -DMETVIEW_BUNDLE_MAGICS=does/not/matter -DMETVIEW_SITE=ecmwf -DNO_PROXY_FOR=OFF -DODB1_MODULE=odb/${ODB_VERSION} -DODB1_VERSION=${ODB_VERSION} -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DPROXY_PORT=3333 -DPROXY_URL=proxy.ecmwf.int'
      edit PACKAGE_VERSION 'fake'
      label build ""
    task make_build_dir
      trigger git_update_repos == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task build
      trigger config == complete
      edit ECF_TRIES '2'
      label host ""
      label options ""
    task test
      trigger build == complete
      edit ECF_TRIES '2'
      edit TEST_DIR '%BUILDIR%/metview'
    task set_write_permissions
      trigger test == complete
    task install
      trigger set_write_permissions == complete
      edit ECF_TRIES '2'
    task install_test
      trigger install == complete
    task install_log
      trigger install_test == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install_log == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family leap42
    defstatus suspended
    edit CI_LIMIT '7'
    edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DBUNDLE_SKIP_METVIEW-ECMWF-CONFIGS=1 -DCMAKE_BUILD_TYPE=Release -DCMAKE_EXE_LINKER_FLAGS=-s -DCMAKE_Fortran_COMPILER=gfortran -DCMAKE_PREFIX_PATH=/usr/local/apps/qt/5.7.0/5.7/gcc_64/ -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DECMWF=ON -DENABLE_BUFR=ON -DENABLE_CAIRO=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_FORTRAN=ON -DENABLE_MARS=ON -DENABLE_MARS_ODB=ON -DENABLE_METVIEW_FORTRAN=ON -DENABLE_NETCDF=ON -DENABLE_NETWORK_PROXY=ON -DENABLE_ODB=ON -DENABLE_OPERA_RADAR=ON -DENABLE_PYTHON=OFF -DENABLE_SCM_RTTOV=ON -DENABLE_USAGE_LOG=ON -DENABLE_WEATHER_ROOM=ON -DEXTRA_CONFIG_PATH=%SRCDIR%/metview-ecmwf-configs -DEXTRA_TITLE=%SITE% -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DFLEXPART_EXE_PATH=/usr/local/apps/flexpart/902/bin/flexpart -DFLEXPART_RESOURCES_PATH=/usr/local/apps/flexpart/902/options -DFLEXPART_SPECIES_PATH=/usr/local/apps/flexpart/902/options/SPECIES -DFLEXTRA_PATH=/usr/local/apps/flextra/flextra -DLOG_DIR=/home/graphics/cgx/public -DMARS_LOCAL_HOME=/usr/local/apps/mars/configs/prod -DMETVIEW_BUNDLE_MAGICS=does/not/matter -DMETVIEW_SITE=ecmwf -DNO_PROXY_FOR=OFF -DODB1_MODULE=odb/${ODB_VERSION} -DODB1_VERSION=${ODB_VERSION} -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DPROXY_PORT=3333 -DPROXY_URL=proxy.ecmwf.int'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load proj/%PROJ_VERSION%" "module load fftw/%FFTW_VERSION%" "module load hdf5" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
    edit M_SITE 'leap42'
    edit REMOTE_HOST 'leap42-deploy'
    edit SITE 'leap42'
    edit WSDIR '/tmp/metabuilds/%ECF_HOST%_%ECF_PORT%/leap42/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DBUNDLE_SKIP_METVIEW-ECMWF-CONFIGS=1 -DCMAKE_BUILD_TYPE=Release -DCMAKE_EXE_LINKER_FLAGS=-s -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=gfortran -DCMAKE_PREFIX_PATH=/usr/local/apps/qt/5.7.0/5.7/gcc_64/ -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DECMWF=ON -DENABLE_BUFR=ON -DENABLE_CAIRO=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_FORTRAN=ON -DENABLE_MARS=ON -DENABLE_MARS_ODB=ON -DENABLE_METVIEW_FORTRAN=ON -DENABLE_NETCDF=ON -DENABLE_NETWORK_PROXY=ON -DENABLE_ODB=ON -DENABLE_OPERA_RADAR=ON -DENABLE_PYTHON=OFF -DENABLE_SCM_RTTOV=ON -DENABLE_USAGE_LOG=ON -DENABLE_WEATHER_ROOM=ON -DEXTRA_CONFIG_PATH=%SRCDIR%/metview-ecmwf-configs -DEXTRA_TITLE=%SITE% -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DFLEXPART_EXE_PATH=/usr/local/apps/flexpart/902/bin/flexpart -DFLEXPART_RESOURCES_PATH=/usr/local/apps/flexpart/902/options -DFLEXPART_SPECIES_PATH=/usr/local/apps/flexpart/902/options/SPECIES -DFLEXTRA_PATH=/usr/local/apps/flextra/flextra -DLOG_DIR=/home/graphics/cgx/public -DMARS_LOCAL_HOME=/usr/local/apps/mars/configs/prod -DMETVIEW_BUNDLE_MAGICS=does/not/matter -DMETVIEW_SITE=ecmwf -DNO_PROXY_FOR=OFF -DODB1_MODULE=odb/${ODB_VERSION} -DODB1_VERSION=${ODB_VERSION} -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DPROXY_PORT=3333 -DPROXY_URL=proxy.ecmwf.int'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load proj/%PROJ_VERSION%" "module load fftw/%FFTW_VERSION%" "module load hdf5" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task git_update_repos
        trigger git_srcs == complete
        edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=Release -DCMAKE_EXE_LINKER_FLAGS=-s -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=gfortran -DCMAKE_PREFIX_PATH=/usr/local/apps/qt/5.7.0/5.7/gcc_64/ -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DECMWF=ON -DENABLE_BUFR=ON -DENABLE_CAIRO=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_FORTRAN=ON -DENABLE_MARS=ON -DENABLE_MARS_ODB=ON -DENABLE_METVIEW_FORTRAN=ON -DENABLE_NETCDF=ON -DENABLE_NETWORK_PROXY=ON -DENABLE_ODB=ON -DENABLE_OPERA_RADAR=ON -DENABLE_PYTHON=OFF -DENABLE_SCM_RTTOV=ON -DENABLE_USAGE_LOG=ON -DENABLE_WEATHER_ROOM=ON -DEXTRA_CONFIG_PATH=%SRCDIR%/metview-ecmwf-configs -DEXTRA_TITLE=%SITE% -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DFLEXPART_EXE_PATH=/usr/local/apps/flexpart/902/bin/flexpart -DFLEXPART_RESOURCES_PATH=/usr/local/apps/flexpart/902/options -DFLEXPART_SPECIES_PATH=/usr/local/apps/flexpart/902/options/SPECIES -DFLEXTRA_PATH=/usr/local/apps/flextra/flextra -DLOG_DIR=/home/graphics/cgx/public -DMARS_LOCAL_HOME=/usr/local/apps/mars/configs/prod -DMETVIEW_BUNDLE_MAGICS=does/not/matter -DMETVIEW_SITE=ecmwf -DNO_PROXY_FOR=OFF -DODB1_MODULE=odb/${ODB_VERSION} -DODB1_VERSION=${ODB_VERSION} -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DPROXY_PORT=3333 -DPROXY_URL=proxy.ecmwf.int'
        edit PACKAGE_VERSION 'fake'
        label build ""
      task make_build_dir
        trigger git_update_repos == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%/metview'
      task set_write_permissions
        trigger test == complete
      task install
        trigger set_write_permissions == complete
        edit ECF_TRIES '2'
      task install_test
        trigger install == complete
      task install_log
        trigger install_test == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.93
      edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DBUNDLE_SKIP_METVIEW-ECMWF-CONFIGS=1 -DCMAKE_BUILD_TYPE=Release -DCMAKE_EXE_LINKER_FLAGS=-s -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=gfortran -DCMAKE_PREFIX_PATH=/usr/local/apps/qt/5.7.0/5.7/gcc_64/ -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DECMWF=ON -DENABLE_BUFR=ON -DENABLE_CAIRO=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_FORTRAN=ON -DENABLE_MARS=ON -DENABLE_MARS_ODB=ON -DENABLE_METVIEW_FORTRAN=ON -DENABLE_NETCDF=ON -DENABLE_NETWORK_PROXY=ON -DENABLE_ODB=ON -DENABLE_OPERA_RADAR=ON -DENABLE_PYTHON=OFF -DENABLE_SCM_RTTOV=ON -DENABLE_USAGE_LOG=ON -DENABLE_WEATHER_ROOM=ON -DEXTRA_CONFIG_PATH=%SRCDIR%/metview-ecmwf-configs -DEXTRA_TITLE=%SITE% -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DFLEXPART_EXE_PATH=/usr/local/apps/flexpart/902/bin/flexpart -DFLEXPART_RESOURCES_PATH=/usr/local/apps/flexpart/902/options -DFLEXPART_SPECIES_PATH=/usr/local/apps/flexpart/902/options/SPECIES -DFLEXTRA_PATH=/usr/local/apps/flextra/flextra -DLOG_DIR=/home/graphics/cgx/public -DMARS_LOCAL_HOME=/usr/local/apps/mars/configs/prod -DMETVIEW_BUNDLE_MAGICS=does/not/matter -DMETVIEW_SITE=ecmwf -DNO_PROXY_FOR=OFF -DODB1_MODULE=odb/${ODB_VERSION} -DODB1_VERSION=${ODB_VERSION} -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DPROXY_PORT=3333 -DPROXY_URL=proxy.ecmwf.int'
      edit COMPILER 'GNU.93'
      edit EXPORTS '"export ECACCOUNT=ecodmdma" "export LD_LIBRARY_PATH=/usr/local/apps/gcc/9.3.0/lib64/" "export ODB_VERSION=%ODB_VERSION%" "export ftp_proxy=http://proxy.ecmwf.int:3333" "export http_proxy=http://proxy.ecmwf.int:3333" "export https_proxy=http://proxy.ecmwf.int:3333"'
      edit MODULE_ACTIONS '"module swap gnu/9.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load proj/%PROJ_VERSION%" "module load fftw/%FFTW_VERSION%" "module load hdf5" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task git_update_repos
        trigger git_srcs == complete
        edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=Release -DCMAKE_EXE_LINKER_FLAGS=-s -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=gfortran -DCMAKE_PREFIX_PATH=/usr/local/apps/qt/5.7.0/5.7/gcc_64/ -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DECMWF=ON -DENABLE_BUFR=ON -DENABLE_CAIRO=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_FORTRAN=ON -DENABLE_MARS=ON -DENABLE_MARS_ODB=ON -DENABLE_METVIEW_FORTRAN=ON -DENABLE_NETCDF=ON -DENABLE_NETWORK_PROXY=ON -DENABLE_ODB=ON -DENABLE_OPERA_RADAR=ON -DENABLE_PYTHON=OFF -DENABLE_SCM_RTTOV=ON -DENABLE_USAGE_LOG=ON -DENABLE_WEATHER_ROOM=ON -DEXTRA_CONFIG_PATH=%SRCDIR%/metview-ecmwf-configs -DEXTRA_TITLE=%SITE% -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DFLEXPART_EXE_PATH=/usr/local/apps/flexpart/902/bin/flexpart -DFLEXPART_RESOURCES_PATH=/usr/local/apps/flexpart/902/options -DFLEXPART_SPECIES_PATH=/usr/local/apps/flexpart/902/options/SPECIES -DFLEXTRA_PATH=/usr/local/apps/flextra/flextra -DLOG_DIR=/home/graphics/cgx/public -DMARS_LOCAL_HOME=/usr/local/apps/mars/configs/prod -DMETVIEW_BUNDLE_MAGICS=does/not/matter -DMETVIEW_SITE=ecmwf -DNO_PROXY_FOR=OFF -DODB1_MODULE=odb/${ODB_VERSION} -DODB1_VERSION=${ODB_VERSION} -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DPROXY_PORT=3333 -DPROXY_URL=proxy.ecmwf.int'
        edit PACKAGE_VERSION 'fake'
        label build ""
      task make_build_dir
        trigger git_update_repos == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%/metview'
      task set_write_permissions
        trigger test == complete
      task install
        trigger set_write_permissions == complete
        edit ECF_TRIES '2'
      task install_test
        trigger install == complete
      task install_log
        trigger install_test == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90
      edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DBUNDLE_SKIP_METVIEW-ECMWF-CONFIGS=1 -DCMAKE_BUILD_TYPE=Release -DCMAKE_EXE_LINKER_FLAGS=-s -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=gfortran -DCMAKE_PREFIX_PATH=/usr/local/apps/qt/5.7.0/5.7/gcc_64/ -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DECMWF=ON -DENABLE_BUFR=ON -DENABLE_CAIRO=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_FORTRAN=ON -DENABLE_MARS=ON -DENABLE_MARS_ODB=ON -DENABLE_METVIEW_FORTRAN=ON -DENABLE_NETCDF=ON -DENABLE_NETWORK_PROXY=ON -DENABLE_ODB=ON -DENABLE_OPERA_RADAR=ON -DENABLE_PYTHON=OFF -DENABLE_SCM_RTTOV=ON -DENABLE_USAGE_LOG=ON -DENABLE_WEATHER_ROOM=ON -DEXTRA_CONFIG_PATH=%SRCDIR%/metview-ecmwf-configs -DEXTRA_TITLE=%SITE% -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DFLEXPART_EXE_PATH=/usr/local/apps/flexpart/902/bin/flexpart -DFLEXPART_RESOURCES_PATH=/usr/local/apps/flexpart/902/options -DFLEXPART_SPECIES_PATH=/usr/local/apps/flexpart/902/options/SPECIES -DFLEXTRA_PATH=/usr/local/apps/flextra/flextra -DLOG_DIR=/home/graphics/cgx/public -DMARS_LOCAL_HOME=/usr/local/apps/mars/configs/prod -DMETVIEW_BUNDLE_MAGICS=does/not/matter -DMETVIEW_SITE=ecmwf -DNO_PROXY_FOR=OFF -DODB1_MODULE=odb/${ODB_VERSION} -DODB1_VERSION=${ODB_VERSION} -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DPROXY_PORT=3333 -DPROXY_URL=proxy.ecmwf.int'
      edit COMPILER 'CLANG.90'
      edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load proj/%PROJ_VERSION%" "module load fftw/%FFTW_VERSION%" "module load hdf5" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task git_update_repos
        trigger git_srcs == complete
        edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=Release -DCMAKE_EXE_LINKER_FLAGS=-s -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=gfortran -DCMAKE_PREFIX_PATH=/usr/local/apps/qt/5.7.0/5.7/gcc_64/ -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DECMWF=ON -DENABLE_BUFR=ON -DENABLE_CAIRO=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_FORTRAN=ON -DENABLE_MARS=ON -DENABLE_MARS_ODB=ON -DENABLE_METVIEW_FORTRAN=ON -DENABLE_NETCDF=ON -DENABLE_NETWORK_PROXY=ON -DENABLE_ODB=ON -DENABLE_OPERA_RADAR=ON -DENABLE_PYTHON=OFF -DENABLE_SCM_RTTOV=ON -DENABLE_USAGE_LOG=ON -DENABLE_WEATHER_ROOM=ON -DEXTRA_CONFIG_PATH=%SRCDIR%/metview-ecmwf-configs -DEXTRA_TITLE=%SITE% -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DFLEXPART_EXE_PATH=/usr/local/apps/flexpart/902/bin/flexpart -DFLEXPART_RESOURCES_PATH=/usr/local/apps/flexpart/902/options -DFLEXPART_SPECIES_PATH=/usr/local/apps/flexpart/902/options/SPECIES -DFLEXTRA_PATH=/usr/local/apps/flextra/flextra -DLOG_DIR=/home/graphics/cgx/public -DMARS_LOCAL_HOME=/usr/local/apps/mars/configs/prod -DMETVIEW_BUNDLE_MAGICS=does/not/matter -DMETVIEW_SITE=ecmwf -DNO_PROXY_FOR=OFF -DODB1_MODULE=odb/${ODB_VERSION} -DODB1_VERSION=${ODB_VERSION} -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DPROXY_PORT=3333 -DPROXY_URL=proxy.ecmwf.int'
        edit PACKAGE_VERSION 'fake'
        label build ""
      task make_build_dir
        trigger git_update_repos == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%/metview'
      task set_write_permissions
        trigger test == complete
      task install
        trigger set_write_permissions == complete
        edit ECF_TRIES '2'
      task install_test
        trigger install == complete
      task install_log
        trigger install_test == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90.linters
      edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DBUNDLE_SKIP_METVIEW-ECMWF-CONFIGS=1 -DCMAKE_BUILD_TYPE=Release -DCMAKE_EXE_LINKER_FLAGS=-s -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=gfortran -DCMAKE_PREFIX_PATH=/usr/local/apps/qt/5.7.0/5.7/gcc_64/ -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DECMWF=ON -DENABLE_BUFR=ON -DENABLE_CAIRO=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_FORTRAN=ON -DENABLE_MARS=ON -DENABLE_MARS_ODB=ON -DENABLE_METVIEW_FORTRAN=ON -DENABLE_NETCDF=ON -DENABLE_NETWORK_PROXY=ON -DENABLE_ODB=ON -DENABLE_OPERA_RADAR=ON -DENABLE_PYTHON=OFF -DENABLE_SCM_RTTOV=ON -DENABLE_USAGE_LOG=ON -DENABLE_WEATHER_ROOM=ON -DEXTRA_CONFIG_PATH=%SRCDIR%/metview-ecmwf-configs -DEXTRA_TITLE=%SITE% -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DFLEXPART_EXE_PATH=/usr/local/apps/flexpart/902/bin/flexpart -DFLEXPART_RESOURCES_PATH=/usr/local/apps/flexpart/902/options -DFLEXPART_SPECIES_PATH=/usr/local/apps/flexpart/902/options/SPECIES -DFLEXTRA_PATH=/usr/local/apps/flextra/flextra -DLOG_DIR=/home/graphics/cgx/public -DMARS_LOCAL_HOME=/usr/local/apps/mars/configs/prod -DMETVIEW_BUNDLE_MAGICS=does/not/matter -DMETVIEW_SITE=ecmwf -DNO_PROXY_FOR=OFF -DODB1_MODULE=odb/${ODB_VERSION} -DODB1_VERSION=${ODB_VERSION} -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DPROXY_PORT=3333 -DPROXY_URL=proxy.ecmwf.int'
      edit COMPILER 'CLANG.90.LINTERS'
      edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load proj/%PROJ_VERSION%" "module load fftw/%FFTW_VERSION%" "module load hdf5" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task preprocess
        trigger git_srcs == complete
      task git_update_repos
        trigger preprocess == complete
        edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=Release -DCMAKE_EXE_LINKER_FLAGS=-s -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=gfortran -DCMAKE_PREFIX_PATH=/usr/local/apps/qt/5.7.0/5.7/gcc_64/ -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DECMWF=ON -DENABLE_BUFR=ON -DENABLE_CAIRO=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_FORTRAN=ON -DENABLE_MARS=ON -DENABLE_MARS_ODB=ON -DENABLE_METVIEW_FORTRAN=ON -DENABLE_NETCDF=ON -DENABLE_NETWORK_PROXY=ON -DENABLE_ODB=ON -DENABLE_OPERA_RADAR=ON -DENABLE_PYTHON=OFF -DENABLE_SCM_RTTOV=ON -DENABLE_USAGE_LOG=ON -DENABLE_WEATHER_ROOM=ON -DEXTRA_CONFIG_PATH=%SRCDIR%/metview-ecmwf-configs -DEXTRA_TITLE=%SITE% -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DFLEXPART_EXE_PATH=/usr/local/apps/flexpart/902/bin/flexpart -DFLEXPART_RESOURCES_PATH=/usr/local/apps/flexpart/902/options -DFLEXPART_SPECIES_PATH=/usr/local/apps/flexpart/902/options/SPECIES -DFLEXTRA_PATH=/usr/local/apps/flextra/flextra -DLOG_DIR=/home/graphics/cgx/public -DMARS_LOCAL_HOME=/usr/local/apps/mars/configs/prod -DMETVIEW_BUNDLE_MAGICS=does/not/matter -DMETVIEW_SITE=ecmwf -DNO_PROXY_FOR=OFF -DODB1_MODULE=odb/${ODB_VERSION} -DODB1_VERSION=${ODB_VERSION} -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DPROXY_PORT=3333 -DPROXY_URL=proxy.ecmwf.int'
        edit PACKAGE_VERSION 'fake'
        label build ""
      task make_build_dir
        trigger git_update_repos == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task cppcheck
        trigger config == complete
        edit CPPCHECK_FLAGS '--template='[{file}:{line}]: {severity} : {id} : {message}' --suppress=syntaxError --inline-suppr --platform=unix64 --enable=performance --enable=portability --enable=information --std=c11 --std=c++14 -DDEBUG=1 -q -j8'
        edit CPPCHECK_VERSION '1.89'
        edit REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}/cpp_check'
      task iwyu
        trigger cppcheck == complete
        edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load proj/%PROJ_VERSION%" "module load fftw/%FFTW_VERSION%" "module load hdf5" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%" "module load iwyu/%IWYU_VERSION%"'
      family clang_linters
        trigger iwyu == complete
        limit clang_linters 1
        inlimit clang_linters
        task clang_tidy
          edit REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}/clang_tidy'
        task static_analyser
          edit REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}/static_analyser'
          edit SCAN_BUILD_OPTIONS '--force-analyze-debug-code -o /perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}/static_analyser -k make -j8'
          edit USE_CLANG_CHECK 'False'
          label scan_view ""
      endfamily
      task wipe_build
        trigger clang_linters == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90.address_sanitizer
      edit ADDRESS_SANITIZER 'True'
      edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DBUNDLE_SKIP_LIBEMOS=1 -DBUNDLE_SKIP_METVIEW-ECMWF-CONFIGS=1 -DCMAKE_BUILD_TYPE=Release -DCMAKE_CXX_FLAGS="-fsanitize=address -fno-omit-frame-pointer" -DCMAKE_C_FLAGS="-fsanitize=address -fno-omit-frame-pointer" -DCMAKE_C_LINK_EXECUTABLE="<CMAKE_CXX_COMPILER> <FLAGS> <CMAKE_C_LINK_FLAGS> <LINK_FLAGS> <OBJECTS> -o <TARGET> <LINK_LIBRARIES>" -DCMAKE_EXE_LINKER_FLAGS="-s -fsanitize=address" -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=gfortran -DCMAKE_PREFIX_PATH=/usr/local/apps/qt/5.7.0/5.7/gcc_64/ -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DECMWF=OFF -DENABLE_BUFR=ON -DENABLE_CAIRO=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_FORTRAN=OFF -DENABLE_MARS=ON -DENABLE_MARS_ODB=ON -DENABLE_METVIEW_FORTRAN=OFF -DENABLE_NETCDF=ON -DENABLE_NETWORK_PROXY=ON -DENABLE_ODB=ON -DENABLE_OPERA_RADAR=ON -DENABLE_PYTHON=OFF -DENABLE_SCM_RTTOV=ON -DENABLE_USAGE_LOG=ON -DENABLE_WEATHER_ROOM=ON -DEXTRA_CONFIG_PATH=%SRCDIR%/metview-ecmwf-configs -DEXTRA_TITLE=%SITE% -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DFLEXPART_EXE_PATH=/usr/local/apps/flexpart/902/bin/flexpart -DFLEXPART_RESOURCES_PATH=/usr/local/apps/flexpart/902/options -DFLEXPART_SPECIES_PATH=/usr/local/apps/flexpart/902/options/SPECIES -DFLEXTRA_PATH=/usr/local/apps/flextra/flextra -DLOG_DIR=/home/graphics/cgx/public -DMARS_LOCAL_HOME=/usr/local/apps/mars/configs/prod -DMETVIEW_BUNDLE_MAGICS=does/not/matter -DMETVIEW_SITE=ecmwf -DNO_PROXY_FOR=OFF -DODB1_MODULE=odb/${ODB_VERSION} -DODB1_VERSION=${ODB_VERSION} -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DPROXY_PORT=3333 -DPROXY_URL=proxy.ecmwf.int'
      edit COMPILER 'CLANG.90.ADDRESS_SANITIZER'
      edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load proj/%PROJ_VERSION%" "module load fftw/%FFTW_VERSION%" "module load hdf5" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit SANITIZER 'True'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task git_update_repos
        trigger git_srcs == complete
        edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DBUNDLE_SKIP_LIBEMOS=1 -DCMAKE_BUILD_TYPE=Release -DCMAKE_CXX_FLAGS="-fsanitize=address -fno-omit-frame-pointer" -DCMAKE_C_FLAGS="-fsanitize=address -fno-omit-frame-pointer" -DCMAKE_C_LINK_EXECUTABLE="<CMAKE_CXX_COMPILER> <FLAGS> <CMAKE_C_LINK_FLAGS> <LINK_FLAGS> <OBJECTS> -o <TARGET> <LINK_LIBRARIES>" -DCMAKE_EXE_LINKER_FLAGS="-s -fsanitize=address" -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=gfortran -DCMAKE_PREFIX_PATH=/usr/local/apps/qt/5.7.0/5.7/gcc_64/ -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DECMWF=OFF -DENABLE_BUFR=ON -DENABLE_CAIRO=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_FORTRAN=OFF -DENABLE_MARS=ON -DENABLE_MARS_ODB=ON -DENABLE_METVIEW_FORTRAN=OFF -DENABLE_NETCDF=ON -DENABLE_NETWORK_PROXY=ON -DENABLE_ODB=ON -DENABLE_OPERA_RADAR=ON -DENABLE_PYTHON=OFF -DENABLE_SCM_RTTOV=ON -DENABLE_USAGE_LOG=ON -DENABLE_WEATHER_ROOM=ON -DEXTRA_CONFIG_PATH=%SRCDIR%/metview-ecmwf-configs -DEXTRA_TITLE=%SITE% -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DFLEXPART_EXE_PATH=/usr/local/apps/flexpart/902/bin/flexpart -DFLEXPART_RESOURCES_PATH=/usr/local/apps/flexpart/902/options -DFLEXPART_SPECIES_PATH=/usr/local/apps/flexpart/902/options/SPECIES -DFLEXTRA_PATH=/usr/local/apps/flextra/flextra -DLOG_DIR=/home/graphics/cgx/public -DMARS_LOCAL_HOME=/usr/local/apps/mars/configs/prod -DMETVIEW_BUNDLE_MAGICS=does/not/matter -DMETVIEW_SITE=ecmwf -DNO_PROXY_FOR=OFF -DODB1_MODULE=odb/${ODB_VERSION} -DODB1_VERSION=${ODB_VERSION} -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DPROXY_PORT=3333 -DPROXY_URL=proxy.ecmwf.int'
        edit PACKAGE_VERSION 'fake'
        label build ""
      task make_build_dir
        trigger git_update_repos == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit TEST_DIR '%BUILDIR%/metview'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90.ub_sanitizer
      edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DBUNDLE_SKIP_LIBEMOS=1 -DBUNDLE_SKIP_METVIEW-ECMWF-CONFIGS=1 -DCMAKE_BUILD_TYPE=Release -DCMAKE_CXX_FLAGS="-fsanitize=undefined -fno-omit-frame-pointer -fno-sanitize-recover=all" -DCMAKE_C_FLAGS="-fsanitize=undefined -fno-omit-frame-pointer" -DCMAKE_C_LINK_EXECUTABLE="<CMAKE_CXX_COMPILER> <FLAGS> <CMAKE_C_LINK_FLAGS> <LINK_FLAGS> <OBJECTS> -o <TARGET> <LINK_LIBRARIES>" -DCMAKE_EXE_LINKER_FLAGS="-s -fsanitize=undefined" -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=gfortran -DCMAKE_PREFIX_PATH=/usr/local/apps/qt/5.7.0/5.7/gcc_64/ -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DECMWF=OFF -DENABLE_BUFR=ON -DENABLE_CAIRO=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_FORTRAN=OFF -DENABLE_MARS=ON -DENABLE_MARS_ODB=ON -DENABLE_METVIEW_FORTRAN=OFF -DENABLE_NETCDF=ON -DENABLE_NETWORK_PROXY=ON -DENABLE_ODB=ON -DENABLE_OPERA_RADAR=ON -DENABLE_PYTHON=OFF -DENABLE_SCM_RTTOV=ON -DENABLE_USAGE_LOG=ON -DENABLE_WEATHER_ROOM=ON -DEXTRA_CONFIG_PATH=%SRCDIR%/metview-ecmwf-configs -DEXTRA_TITLE=%SITE% -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DFLEXPART_EXE_PATH=/usr/local/apps/flexpart/902/bin/flexpart -DFLEXPART_RESOURCES_PATH=/usr/local/apps/flexpart/902/options -DFLEXPART_SPECIES_PATH=/usr/local/apps/flexpart/902/options/SPECIES -DFLEXTRA_PATH=/usr/local/apps/flextra/flextra -DLOG_DIR=/home/graphics/cgx/public -DMARS_LOCAL_HOME=/usr/local/apps/mars/configs/prod -DMETVIEW_BUNDLE_MAGICS=does/not/matter -DMETVIEW_SITE=ecmwf -DNO_PROXY_FOR=OFF -DODB1_MODULE=odb/${ODB_VERSION} -DODB1_VERSION=${ODB_VERSION} -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DPROXY_PORT=3333 -DPROXY_URL=proxy.ecmwf.int'
      edit COMPILER 'CLANG.90.UB_SANITIZER'
      edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load proj/%PROJ_VERSION%" "module load fftw/%FFTW_VERSION%" "module load hdf5" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit SANITIZER 'True'
      edit UB_SANITIZER 'True'
      edit UB_SANITIZER_LD_PRELOAD 'export LD_PRELOAD=/usr/local/apps/clang/9.0.1/lib64/libubsan.so'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task git_update_repos
        trigger git_srcs == complete
        edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DBUNDLE_SKIP_LIBEMOS=1 -DCMAKE_BUILD_TYPE=Release -DCMAKE_CXX_FLAGS="-fsanitize=undefined -fno-omit-frame-pointer -fno-sanitize-recover=all" -DCMAKE_C_FLAGS="-fsanitize=undefined -fno-omit-frame-pointer" -DCMAKE_C_LINK_EXECUTABLE="<CMAKE_CXX_COMPILER> <FLAGS> <CMAKE_C_LINK_FLAGS> <LINK_FLAGS> <OBJECTS> -o <TARGET> <LINK_LIBRARIES>" -DCMAKE_EXE_LINKER_FLAGS="-s -fsanitize=undefined" -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=gfortran -DCMAKE_PREFIX_PATH=/usr/local/apps/qt/5.7.0/5.7/gcc_64/ -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DECMWF=OFF -DENABLE_BUFR=ON -DENABLE_CAIRO=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_FORTRAN=OFF -DENABLE_MARS=ON -DENABLE_MARS_ODB=ON -DENABLE_METVIEW_FORTRAN=OFF -DENABLE_NETCDF=ON -DENABLE_NETWORK_PROXY=ON -DENABLE_ODB=ON -DENABLE_OPERA_RADAR=ON -DENABLE_PYTHON=OFF -DENABLE_SCM_RTTOV=ON -DENABLE_USAGE_LOG=ON -DENABLE_WEATHER_ROOM=ON -DEXTRA_CONFIG_PATH=%SRCDIR%/metview-ecmwf-configs -DEXTRA_TITLE=%SITE% -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DFLEXPART_EXE_PATH=/usr/local/apps/flexpart/902/bin/flexpart -DFLEXPART_RESOURCES_PATH=/usr/local/apps/flexpart/902/options -DFLEXPART_SPECIES_PATH=/usr/local/apps/flexpart/902/options/SPECIES -DFLEXTRA_PATH=/usr/local/apps/flextra/flextra -DLOG_DIR=/home/graphics/cgx/public -DMARS_LOCAL_HOME=/usr/local/apps/mars/configs/prod -DMETVIEW_BUNDLE_MAGICS=does/not/matter -DMETVIEW_SITE=ecmwf -DNO_PROXY_FOR=OFF -DODB1_MODULE=odb/${ODB_VERSION} -DODB1_VERSION=${ODB_VERSION} -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DPROXY_PORT=3333 -DPROXY_URL=proxy.ecmwf.int'
        edit PACKAGE_VERSION 'fake'
        label build ""
      task make_build_dir
        trigger git_update_repos == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit TEST_DIR '%BUILDIR%/metview'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family lxc
    defstatus suspended
    edit CI_LIMIT '10'
    edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DBUNDLE_SKIP_METVIEW-ECMWF-CONFIGS=1 -DCMAKE_BUILD_TYPE=Release -DCMAKE_EXE_LINKER_FLAGS=-s -DCMAKE_Fortran_COMPILER=gfortran -DCMAKE_PREFIX_PATH=/usr/local/apps/qt/5.10.1/5.10/gcc_64/ -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DECMWF=ON -DENABLE_BUFR=ON -DENABLE_CAIRO=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_FORTRAN=ON -DENABLE_MARS=ON -DENABLE_MARS_ODB=ON -DENABLE_METVIEW_FORTRAN=ON -DENABLE_NETCDF=ON -DENABLE_NETWORK_PROXY=ON -DENABLE_ODB=ON -DENABLE_OPERA_RADAR=ON -DENABLE_PYTHON=OFF -DENABLE_SCM_RTTOV=ON -DENABLE_USAGE_LOG=ON -DENABLE_WEATHER_ROOM=OFF -DEXTRA_CONFIG_PATH=%SRCDIR%/metview-ecmwf-configs -DEXTRA_TITLE=%SITE% -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_LIBRARIES=/usr/local/apps/lapack/3.5.0/LP64/lib/liblapack.so -DLOG_DIR=/home/graphics/cgx/public -DMARS_LOCAL_HOME=/usr/local/apps/mars/configs/prod -DMETVIEW_BUNDLE_MAGICS=does/not/matter -DMETVIEW_SITE=ecmwf -DNO_PROXY_FOR=OFF -DODB1_MODULE=odb/${ODB_VERSION} -DODB1_VERSION=${ODB_VERSION} -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DPROXY_PORT=3333 -DPROXY_URL=proxy.ecmwf.int'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load proj/%PROJ_VERSION%" "module load fftw/%FFTW_VERSION%" "module load netcdf4/%NETCDF4_VERSION%" "module load ecflow/%ECF_VERSION%" "module load hdf5/1.8.17"'
    edit M_SITE 'lxc'
    edit QUEUE 'express'
    edit REMOTE_HOST 'lxc'
    edit SITE 'lxc'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxc/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task git_update_repos
      trigger git_srcs == complete
      edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=Release -DCMAKE_EXE_LINKER_FLAGS=-s -DCMAKE_Fortran_COMPILER=gfortran -DCMAKE_PREFIX_PATH=/usr/local/apps/qt/5.10.1/5.10/gcc_64/ -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DECMWF=ON -DENABLE_BUFR=ON -DENABLE_CAIRO=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_FORTRAN=ON -DENABLE_MARS=ON -DENABLE_MARS_ODB=ON -DENABLE_METVIEW_FORTRAN=ON -DENABLE_NETCDF=ON -DENABLE_NETWORK_PROXY=ON -DENABLE_ODB=ON -DENABLE_OPERA_RADAR=ON -DENABLE_PYTHON=OFF -DENABLE_SCM_RTTOV=ON -DENABLE_USAGE_LOG=ON -DENABLE_WEATHER_ROOM=OFF -DEXTRA_CONFIG_PATH=%SRCDIR%/metview-ecmwf-configs -DEXTRA_TITLE=%SITE% -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_LIBRARIES=/usr/local/apps/lapack/3.5.0/LP64/lib/liblapack.so -DLOG_DIR=/home/graphics/cgx/public -DMARS_LOCAL_HOME=/usr/local/apps/mars/configs/prod -DMETVIEW_BUNDLE_MAGICS=does/not/matter -DMETVIEW_SITE=ecmwf -DNO_PROXY_FOR=OFF -DODB1_MODULE=odb/${ODB_VERSION} -DODB1_VERSION=${ODB_VERSION} -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DPROXY_PORT=3333 -DPROXY_URL=proxy.ecmwf.int'
      edit PACKAGE_VERSION 'fake'
      label build ""
    task make_build_dir
      trigger git_update_repos == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task build
      trigger config == complete
      edit ECF_TRIES '2'
      label host ""
      label options ""
    task test
      trigger build == complete
      edit ECF_TRIES '2'
      edit TEST_DIR '%BUILDIR%/metview'
    task set_write_permissions
      trigger test == complete
    task install
      trigger set_write_permissions == complete
      edit ECF_TRIES '2'
    task install_test
      trigger install == complete
    task install_log
      trigger install_test == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install_log == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family lxg
    defstatus suspended
    edit CI_LIMIT '10'
    edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DBUNDLE_SKIP_METVIEW-ECMWF-CONFIGS=1 -DCMAKE_BUILD_TYPE=Release -DCMAKE_EXE_LINKER_FLAGS=-s -DCMAKE_Fortran_COMPILER=gfortran -DCMAKE_PREFIX_PATH=/usr/local/apps/qt/5.7.0/5.7/gcc_64/ -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DECMWF=ON -DENABLE_BUFR=ON -DENABLE_CAIRO=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_FORTRAN=ON -DENABLE_MARS=ON -DENABLE_MARS_ODB=ON -DENABLE_METVIEW=OFF -DENABLE_METVIEW_FORTRAN=ON -DENABLE_METVIEW_NO_QT=ON -DENABLE_NETCDF=ON -DENABLE_NETWORK_PROXY=ON -DENABLE_ODB=ON -DENABLE_OPERA_RADAR=ON -DENABLE_PYTHON=OFF -DENABLE_SCM_RTTOV=ON -DENABLE_UI=OFF -DENABLE_USAGE_LOG=ON -DENABLE_WEATHER_ROOM=OFF -DEXTRA_CONFIG_PATH=%SRCDIR%/metview-ecmwf-configs -DEXTRA_TITLE=%SITE% -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_LIBRARIES=/usr/local/apps/lapack/3.5.0/LP64/lib/liblapack.so -DLOG_DIR=/home/graphics/cgx/public -DMARS_LOCAL_HOME=/usr/local/apps/mars/configs/prod -DMETVIEW_BUNDLE_MAGICS=does/not/matter -DMETVIEW_SITE=ecmwf -DNO_PROXY_FOR=OFF -DODB1_MODULE=odb/${ODB_VERSION} -DODB1_VERSION=${ODB_VERSION} -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DPROXY_PORT=3333 -DPROXY_URL=proxy.ecmwf.int'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'lxg'
    edit QUEUE 'normal'
    edit REMOTE_HOST 'lxg'
    edit SITE 'lxg'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxg/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task git_update_repos
      trigger git_srcs == complete
      edit CMAKE_VARS '-DBUFR_TABLES_PATH=%BUFR_TABLES% -DCMAKE_BUILD_TYPE=Release -DCMAKE_EXE_LINKER_FLAGS=-s -DCMAKE_Fortran_COMPILER=gfortran -DCMAKE_PREFIX_PATH=/usr/local/apps/qt/5.7.0/5.7/gcc_64/ -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DECKIT_PATH=%BASEPREFIX%/eckit/develop%INSTALL_POSTFIX% -DECMWF=ON -DENABLE_BUFR=ON -DENABLE_CAIRO=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_FORTRAN=ON -DENABLE_MARS=ON -DENABLE_MARS_ODB=ON -DENABLE_METVIEW=OFF -DENABLE_METVIEW_FORTRAN=ON -DENABLE_METVIEW_NO_QT=ON -DENABLE_NETCDF=ON -DENABLE_NETWORK_PROXY=ON -DENABLE_ODB=ON -DENABLE_OPERA_RADAR=ON -DENABLE_PYTHON=OFF -DENABLE_SCM_RTTOV=ON -DENABLE_UI=OFF -DENABLE_USAGE_LOG=ON -DENABLE_WEATHER_ROOM=OFF -DEXTRA_CONFIG_PATH=%SRCDIR%/metview-ecmwf-configs -DEXTRA_TITLE=%SITE% -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_LIBRARIES=/usr/local/apps/lapack/3.5.0/LP64/lib/liblapack.so -DLOG_DIR=/home/graphics/cgx/public -DMARS_LOCAL_HOME=/usr/local/apps/mars/configs/prod -DMETVIEW_BUNDLE_MAGICS=does/not/matter -DMETVIEW_SITE=ecmwf -DNO_PROXY_FOR=OFF -DODB1_MODULE=odb/${ODB_VERSION} -DODB1_VERSION=${ODB_VERSION} -DODB_API_PATH=%BASEPREFIX%/odb_api/develop%INSTALL_POSTFIX% -DODB_PATH=%BASEPREFIX%/odb/develop%INSTALL_POSTFIX% -DPROXY_PORT=3333 -DPROXY_URL=proxy.ecmwf.int'
      edit PACKAGE_VERSION 'fake'
      label build ""
    task make_build_dir
      trigger git_update_repos == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task build
      trigger config == complete
      edit ECF_TRIES '2'
      label host ""
      label options ""
    task test
      trigger build == complete
      edit ECF_TRIES '2'
      edit TEST_DIR '%BUILDIR%/metview'
    task set_write_permissions
      trigger test == complete
    task install
      trigger set_write_permissions == complete
      edit ECF_TRIES '2'
    task install_test
      trigger install == complete
    task install_log
      trigger install_test == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install_log == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
endsuite
suite metview_python
  edit ADDRESS_SANITIZER 'False'
  edit ADD_ACTION_TRIGGERS 'True'
  edit AEC_VERSION '1.0.4'
  edit ALLOW_INSTALL_TO_OVERWRITE 'True'
  edit ASAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
  edit ATLAS_GIT_TAG 'develop'
  edit ATLAS_VERSION '0.20.2'
  edit BANNER_TASK_END 'banner %TASK%-end || true'
  edit BANNER_TASK_START 'banner %TASK%-start || true'
  edit BASEDIR '%WSDIR%/%PACKAGE%'
  edit BASEPREFIX '$PERM/install/%SITE%'
  edit BUFRDC_EMOSLIB_VERSION '4.6.3'
  edit BUFRDC_TABLES_GIT_TAG 'develop'
  edit BUFRDC_TABLES_VERSION '000415'
  edit BUILDIR '%BASEDIR%/builds'
  edit BUILD_ALWAYS_PARALLEL 'False'
  edit CFI_SANITIZER 'False'
  edit CFL_BASE_URL 'https://software-test.ecmwf.int/wiki'
  edit CFL_SPACE_KEY '<TODO>'
  edit CFL_TARBALL_COMMENT ''
  edit CFL_TITLE 'Releases'
  edit CHECK_SYNC_RELEASE 'False'
  edit CLANG_TIDY_ABORT_ON_ERROR 'False'
  edit CMAKE '%ECBUILD%'
  edit CMAKELIST_EXTRACT_PACKAGE_VERSION '$(awk '/^project/ && /%PACKAGE%/ && /VERSION/ {for (I=1;I<=NF;I++) if ($I == "VERSION") {print $(I+1)};}' CMakeLists.txt)'
  edit CMAKE_BUILD_TYPE 'RelWithDebInfo'
  edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE%'
  edit CMAKE_VERSION '3.13.4'
  edit CODES_UI_GIT_TAG 'develop'
  edit CODES_UI_VERSION '1.3.1'
  edit CPACK 'make package'
  edit CPUS '$(lscpu -p | grep -v '#' | wc -l)'
  edit CRAY_FFTW_VERSION '3.3.8.1'
  edit CRAY_NETCDF_VERSION '4.6.1.3'
  edit CRON_BASEPREFIX '%BASEDIR%/install'
  edit CRON_EMAIL ''
  edit CRON_ENABLED 'False'
  edit CRON_GIT_TAG 'develop'
  edit CUSTOM_COMMANDS ''
  edit DATAFLOW_SANITIZER 'False'
  edit DEFAULT_GNU_COMPILER 'gnu.73'
  edit ECBUILD '%BASEPREFIX%/ecbuild/develop/bin/ecbuild'
  edit ECBUILD_CLONE 'False'
  edit ECBUILD_CLONE_COMMIT 'master'
  edit ECBUILD_GIT_TAG 'develop'
  edit ECBUILD_REPO '%GIT_REPO_PREFIX%/ecsdk/ecbuild.git'
  edit ECBUILD_VERSION '3.3.2'
  edit ECCODES_EXTENSIONS_GIT_TAG 'develop'
  edit ECCODES_EXTENSIONS_VERSION '2.17.0'
  edit ECCODES_GIT_TAG 'develop'
  edit ECCODES_PYTHON_VERSION '0.9.6'
  edit ECCODES_VERSION '2.17.1'
  edit ECFLOW5_LATEST_VERSION '5.4.0'
  edit ECFLOW_VERSION '5.4.1'
  edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/metview_python'
  edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/metview_python'
  edit ECF_JOB_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_JOB% %ECF_JOBOUT%'
  edit ECF_KILL_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% kill'
  edit ECF_STATUS_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% stat'
  edit ECF_TRIES '1'
  edit ECKIT_GIT_TAG 'develop'
  edit ECKIT_VERSION '1.10.1'
  edit ECQUIS_VERSION '1.0.0'
  edit EMOS_MOD_VERSION '462'
  edit EMOS_PA_VERSION '000462'
  edit EXPORTS '"export ECACCOUNT=ecodmdma" "export ftp_proxy=http://proxy.ecmwf.int:3333" "export http_proxy=http://proxy.ecmwf.int:3333" "export https_proxy=http://proxy.ecmwf.int:3333"'
  edit EXPORT_VARIABLES 'False'
  edit FCKIT_GIT_TAG '0.6.3'
  edit FCKIT_VERSION '0.6.3'
  edit FDB_VINTAGE_GIT_TAG '4.10.0'
  edit FDB_VINTAGE_VERSION '4.10.0'
  edit FFTW_VERSION '3.3.8'
  edit FFTW_VERSION_CRAY '3.3.4.11'
  edit GIT 'git'
  edit GITHUB_PACKAGE_NAME 'metview-python'
  edit GITHUB_REPO 'git@github.com:ecmwf'
  edit GITREPO '%GIT_REPO_PREFIX%/mpy/mpy.git'
  edit GIT_CLONE 'True'
  edit GIT_CLONE_OPTIONS '--depth=1 --branch %GIT_TAG%'
  edit GIT_METABUILDER_TAG 'ci'
  edit GIT_REPO_PREFIX 'ssh://git@git.ecmwf.int:7999'
  edit GIT_TAG 'develop'
  edit GIT_VERSION '2.26.0'
  edit GRIB_API_VERSION '1.28.0'
  edit INSTALLABLE_TO_USR_LOCAL_APPS 'True'
  edit INSTALL_DIR '%BASEPREFIX%/metview/develop'
  edit IWYU_VERSION '0.11'
  edit KSH_EXTRACT_PACKAGE_VERSION '`grep _VERSION_STR VERSION.cmake | grep -o '[0-9]*\.[0-9]*\.[0-9]*'`'
  edit LAPACK_VERSION '3.5.0'
  edit LIBEMOS_GIT_TAG 'develop'
  edit LIBEMOS_VERSION '4.6.2'
  edit LINTERS_REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}'
  edit LSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/lsan.txt'
  edit MAGICS_GIT_TAG 'develop'
  edit MAGICS_PYTHON_VERSION 'master'
  edit MAGICS_VERSION '4.3.3'
  edit MAKE 'make'
  edit MARS_CLIENT_GIT_TAG '6.27.2.2'
  edit MARS_CLIENT_PACKAGE_GIT_TAG '6.27.6'
  edit MARS_CLIENT_PACKAGE_VERSION '6.27.6'
  edit MARS_CLIENT_VERSION '6.27.2.2'
  edit MEMORY_SANITIZER 'False'
  edit METABUILDER_HOME '/var/tmp/ma0/workspace/metabuilder'
  edit METAB_INSTALL_REMOVE_WRITE_PERMISSIONS 'True'
  edit METAB_PYTHON_VERSION 'python3'
  edit METKIT_GIT_TAG 'develop'
  edit METKIT_VERSION '1.4.6'
  edit METVIEW_GIT_TAG 'develop'
  edit METVIEW_PYTHON_VERSION '1.4.2'
  edit METVIEW_VERSION '5.8.3'
  edit MIR_BUNDLE_VERSION '0.9.0.0'
  edit MIR_GIT_TAG 'develop'
  edit MIR_VERSION '1.5.0'
  edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module load python3"'
  edit MODULE_ACTIONS_FIRST ''
  edit MODULE_ACTIONS_HDR 'module_actions.h'
  edit MODULE_CHECKING_HDR 'module_checking.h'
  edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow"'
  edit MODULE_LOAD_SUPPORTED 'True'
  edit MODULE_NAME 'metview-python'
  edit MODULE_VERSION '%GIT_TAG%'
  edit NETCDF4_VERSION '4.7.4'
  edit ODB_API_GIT_TAG 'develop'
  edit ODB_API_VERSION '0.19.4'
  edit ODB_GIT_TAG 'develop'
  edit ODB_TOOLS_GIT_TAG 'develop'
  edit ODB_TOOLS_VERSION '0.18.3'
  edit ODB_VERSION '1.0.11'
  edit ODC_GIT_TAG 'develop'
  edit ODC_VERSION '1.0.3'
  edit OPENJPEG_VERSION '2.3.1'
  edit PACKAGE 'metview_python'
  edit PACKAGE_NAME 'metview_python'
  edit PACKAGE_OWNERS 'cgi@ecmwf.int cgm@ecmwf.int ma0@ecmwf.int'
  edit PACKAGE_VERSION '%GIT_TAG%'
  edit PGEN_VERSION '1.7.1.0'
  edit PKG_BOOST_VERSION '1.71.0'
  edit PREFIX_APPEND ''
  edit PROJ4_VERSION '5.2.0'
  edit PROJ_VERSION '6.3.1'
  edit PYTHON2_VERSION '2.7.15-01'
  edit PYTHON3_VERSION '3.6.10-01'
  edit PYTHON3_VERSION_MAJOR_MINOR '3.6'
  edit PYTHON_GITREPO 'This-needs-tobe-overriden'
  edit PYTHON_PACKAGE 'This-needs-tobe-overriden'
  edit QT_VERSION '5.12.3'
  edit QUEUE 'build'
  edit REMOTE_HOST 'polonius'
  edit REUSE_BUILD_DIR 'False'
  edit SAFE_STACK_SANITIZER 'False'
  edit SANITIZER 'False'
  edit SANITIZER_REFERENCE_DIRECTORY '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
  edit SRCDIR '%BASEDIR%/%PACKAGE%'
  edit STATIC_ANALYSER_ABORT_ON_ERROR 'False'
  edit SWIG_VERSION '3.0.8'
  edit SYNC_RELEASE '2020.04.0'
  edit SYNC_RELEASE_CI_STAGE 'develop'
  edit TARBALL_DIR '/perm/ma/ma0/tmp_tarballs/ci'
  edit TAR_PACKAGE '%PACKAGE%'
  edit TEST 'pytest tests/'
  edit TEST_DIR '%SRCDIR%'
  edit THREAD_SANITIZER 'False'
  edit TRANSI_GIT_TAG 'develop'
  edit TRANSI_VERSION '0.7.1'
  edit TSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
  edit UBSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/ubsan.txt'
  edit UB_SANITIZER 'False'
  edit USER 'ma0'
  edit USE_WORKDIR 'False'
  edit VALGRIND_VERSION '3.15.0'
  edit WORKDIR '%BASEDIR%/workdir/%ECF_NAME%'
  edit WSDIR '/var/tmp/tmpdir/deploy/jtmp.20739'
  label branch "develop"
  task regenerate
    defstatus complete
    edit ECF_JOB_CMD '%ECF_JOB% > %ECF_JOBOUT% 2>&1'
    edit ECF_KILL_CMD 'kill -15 %ECF_RID%'
    edit ECF_PASS 'FREE'
    edit ECF_STATUS_CMD 'ps --sid %ECF_RID% -f'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
    label date ""
    label metabuilder_branch "ci"
  family cca
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DLAPACK_FOUND=TRUE'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/metview_python/cca'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/metview_python/cca:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cca-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp" "export ftp_proxy=http://proxy.ecmwf.int:3333" "export http_proxy=http://proxy.ecmwf.int:3333" "export https_proxy=http://proxy.ecmwf.int:3333"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '24000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cca'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'cca'
    edit SITE 'cca'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task create_venv
      trigger git_srcs == complete
    task test
      trigger create_venv == complete
      edit ECF_TRIES '2'
      edit TEST_DIR '%SRCDIR%'
    task add_write_perm
      trigger test == complete
    task install
      trigger add_write_perm == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family ccb
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DLAPACK_FOUND=TRUE'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/metview_python/ccb'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/metview_python/ccb:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'ccb-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp" "export ftp_proxy=http://proxy.ecmwf.int:3333" "export http_proxy=http://proxy.ecmwf.int:3333" "export https_proxy=http://proxy.ecmwf.int:3333"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '24000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'ccb'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'ccb'
    edit SITE 'ccb'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task create_venv
      trigger git_srcs == complete
    task test
      trigger create_venv == complete
      edit ECF_TRIES '2'
      edit TEST_DIR '%SRCDIR%'
    task add_write_perm
      trigger test == complete
    task install
      trigger add_write_perm == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family cct
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DLAPACK_FOUND=TRUE'
    edit CPUS '$EC_total_tasks'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/metview_python/cct'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/metview_python/cct:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cct-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '4'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp" "export ftp_proxy=http://proxy.ecmwf.int:3333" "export http_proxy=http://proxy.ecmwf.int:3333" "export https_proxy=http://proxy.ecmwf.int:3333"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '6000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cct'
    edit QSUB_LP 'QSUB -lP %EC_TOTAL_TASKS%'
    edit QUEUE 'nf'
    edit REMOTE_HOST 'cct'
    edit SITE 'cct'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task create_venv
      trigger git_srcs == complete
    task add_write_perm
      trigger create_venv == complete
    task install
      trigger add_write_perm == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family ecgb
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'ecgb'
    edit QUEUE 'express'
    edit REMOTE_HOST 'ecgb'
    edit SITE 'ecgb'
    edit WSDIR '$SCRATCH/metabuilds/%ECF_HOST%_%ECF_PORT%/ecgb/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task create_venv
      trigger git_srcs == complete
    task test
      trigger create_venv == complete
      edit ECF_TRIES '2'
      edit TEST_DIR '%SRCDIR%'
    task add_write_perm
      trigger test == complete
    task install
      trigger add_write_perm == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family leap42
    defstatus suspended
    edit CI_LIMIT '7'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit GIT_CLONE_OPTIONS ''
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'leap42'
    edit REMOTE_HOST 'leap42-deploy'
    edit SITE 'leap42'
    edit WSDIR '/tmp/metabuilds/%ECF_HOST%_%ECF_PORT%/leap42/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task pull_from_github
      trigger git_srcs == complete
    task create_venv
      trigger pull_from_github == complete
    task test
      trigger create_venv == complete
      edit ECF_TRIES '2'
      edit TEST_DIR '%SRCDIR%'
    task test_uplot
      trigger test == complete
    task push_to_github
      trigger test_uplot == complete
    task push_to_bitbucket
      trigger push_to_github == complete
    task add_write_perm
      trigger push_to_bitbucket == complete
    task install
      trigger add_write_perm == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family lxc
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'lxc'
    edit QUEUE 'express'
    edit REMOTE_HOST 'lxc'
    edit SITE 'lxc'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxc/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task create_venv
      trigger git_srcs == complete
    task test
      trigger create_venv == complete
      edit ECF_TRIES '2'
      edit TEST_DIR '%SRCDIR%'
    task add_write_perm
      trigger test == complete
    task install
      trigger add_write_perm == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family lxg
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'lxg'
    edit QUEUE 'normal'
    edit REMOTE_HOST 'lxg'
    edit SITE 'lxg'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxg/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task create_venv
      trigger git_srcs == complete
    task test
      trigger create_venv == complete
      edit ECF_TRIES '2'
      edit TEST_DIR '%SRCDIR%'
    task add_write_perm
      trigger test == complete
    task install
      trigger add_write_perm == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
endsuite
suite codes_ui
  edit ADDRESS_SANITIZER 'False'
  edit ADD_ACTION_TRIGGERS 'True'
  edit AEC_VERSION '1.0.4'
  edit ALLOW_INSTALL_TO_OVERWRITE 'True'
  edit ASAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
  edit ATLAS_GIT_TAG 'develop'
  edit ATLAS_VERSION '0.20.2'
  edit BANNER_TASK_END 'banner %TASK%-end || true'
  edit BANNER_TASK_START 'banner %TASK%-start || true'
  edit BASEDIR '%WSDIR%/%PACKAGE%'
  edit BASEPREFIX '$PERM/install/%SITE%'
  edit BUFRDC_EMOSLIB_VERSION '4.6.3'
  edit BUFRDC_TABLES_GIT_TAG 'develop'
  edit BUFRDC_TABLES_VERSION '000415'
  edit BUILDIR '%BASEDIR%/builds'
  edit BUILD_ALWAYS_PARALLEL 'False'
  edit CFI_SANITIZER 'False'
  edit CFL_BASE_URL 'https://software-test.ecmwf.int/wiki'
  edit CFL_SPACE_KEY 'METV'
  edit CFL_TARBALL_COMMENT ''
  edit CFL_TITLE 'CodesUI - Releases'
  edit CHECK_SYNC_RELEASE 'False'
  edit CLANG_TIDY_ABORT_ON_ERROR 'False'
  edit CMAKE '%ECBUILD%'
  edit CMAKELIST_EXTRACT_PACKAGE_VERSION '$(awk '/^project/ && /%PACKAGE%/ && /VERSION/ {for (I=1;I<=NF;I++) if ($I == "VERSION") {print $(I+1)};}' CMakeLists.txt)'
  edit CMAKE_BUILD_TYPE 'RelWithDebInfo'
  edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DENABLE_ECCODES=ON -DENABLE_QT5=ON -DEXTRA_TITLE=%SITE%'
  edit CMAKE_VERSION '3.13.4'
  edit CODES_UI_GIT_TAG 'develop'
  edit CODES_UI_VERSION '1.3.1'
  edit CPACK 'make package'
  edit CPUS '$(lscpu -p | grep -v '#' | wc -l)'
  edit CRAY_FFTW_VERSION '3.3.8.1'
  edit CRAY_NETCDF_VERSION '4.6.1.3'
  edit CRON_BASEPREFIX '%BASEDIR%/install'
  edit CRON_EMAIL ''
  edit CRON_ENABLED 'False'
  edit CRON_GIT_TAG 'develop'
  edit CUSTOM_COMMANDS ''
  edit DATAFLOW_SANITIZER 'False'
  edit DEFAULT_GNU_COMPILER 'gnu.73'
  edit ECBUILD '%BASEPREFIX%/ecbuild/develop/bin/ecbuild'
  edit ECBUILD_CLONE 'False'
  edit ECBUILD_CLONE_COMMIT 'master'
  edit ECBUILD_GIT_TAG 'develop'
  edit ECBUILD_REPO '%GIT_REPO_PREFIX%/ecsdk/ecbuild.git'
  edit ECBUILD_VERSION '3.3.2'
  edit ECCODES_EXTENSIONS_GIT_TAG 'develop'
  edit ECCODES_EXTENSIONS_VERSION '2.17.0'
  edit ECCODES_GIT_TAG 'develop'
  edit ECCODES_PYTHON_VERSION '0.9.6'
  edit ECCODES_VERSION '2.17.1'
  edit ECFLOW5_LATEST_VERSION '5.4.0'
  edit ECFLOW_VERSION '5.4.1'
  edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/codes_ui'
  edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/codes_ui'
  edit ECF_JOB_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_JOB% %ECF_JOBOUT%'
  edit ECF_KILL_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% kill'
  edit ECF_STATUS_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% stat'
  edit ECF_TRIES '1'
  edit ECKIT_GIT_TAG 'develop'
  edit ECKIT_VERSION '1.10.1'
  edit ECQUIS_VERSION '1.0.0'
  edit EMOS_MOD_VERSION '462'
  edit EMOS_PA_VERSION '000462'
  edit EXPORTS '"export ECACCOUNT=ecodmdma"'
  edit EXPORT_VARIABLES 'False'
  edit FCKIT_GIT_TAG '0.6.3'
  edit FCKIT_VERSION '0.6.3'
  edit FDB_VINTAGE_GIT_TAG '4.10.0'
  edit FDB_VINTAGE_VERSION '4.10.0'
  edit FFTW_VERSION '3.3.8'
  edit FFTW_VERSION_CRAY '3.3.4.11'
  edit GIT 'git'
  edit GITHUB_PACKAGE_NAME '%PACKAGE_NAME%'
  edit GITHUB_REPO 'git@github.com:ecmwf'
  edit GITREPO '%GIT_REPO_PREFIX%/metv/metview.git'
  edit GIT_CLONE 'True'
  edit GIT_CLONE_OPTIONS '--depth=1 --branch %GIT_TAG%'
  edit GIT_METABUILDER_TAG 'ci'
  edit GIT_REPO_PREFIX 'ssh://git@git.ecmwf.int:7999'
  edit GIT_TAG 'develop'
  edit GIT_VERSION '2.26.0'
  edit GRIB_API_VERSION '1.28.0'
  edit INSTALLABLE_TO_USR_LOCAL_APPS 'True'
  edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop'
  edit IWYU_VERSION '0.11'
  edit KSH_EXTRACT_PACKAGE_VERSION '`cd %SRCDIR%/src/codes_ui && ../../extract_metview_version.sh`'
  edit LAPACK_VERSION '3.5.0'
  edit LIBEMOS_GIT_TAG 'develop'
  edit LIBEMOS_VERSION '4.6.2'
  edit LINTERS_REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}'
  edit LSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/lsan.txt'
  edit MAGICS_GIT_TAG 'develop'
  edit MAGICS_PYTHON_VERSION 'master'
  edit MAGICS_VERSION '4.3.3'
  edit MAKE 'make'
  edit MAKE_BUILD_TYPE 'Release'
  edit MARS_CLIENT_GIT_TAG '6.27.2.2'
  edit MARS_CLIENT_PACKAGE_GIT_TAG '6.27.6'
  edit MARS_CLIENT_PACKAGE_VERSION '6.27.6'
  edit MARS_CLIENT_VERSION '6.27.2.2'
  edit MEMORY_SANITIZER 'False'
  edit METABUILDER_HOME '/var/tmp/ma0/workspace/metabuilder'
  edit METAB_INSTALL_REMOVE_WRITE_PERMISSIONS 'True'
  edit METAB_PYTHON_VERSION 'python3'
  edit METKIT_GIT_TAG 'develop'
  edit METKIT_VERSION '1.4.6'
  edit METVIEW_GIT_TAG 'develop'
  edit METVIEW_PYTHON_VERSION '1.4.2'
  edit METVIEW_VERSION '5.8.3'
  edit MIR_BUNDLE_VERSION '0.9.0.0'
  edit MIR_GIT_TAG 'develop'
  edit MIR_VERSION '1.5.0'
  edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
  edit MODULE_ACTIONS_FIRST ''
  edit MODULE_ACTIONS_HDR 'module_actions.h'
  edit MODULE_CHECKING_HDR 'module_checking.h'
  edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow"'
  edit MODULE_LOAD_SUPPORTED 'True'
  edit MODULE_NAME 'codes_ui'
  edit MODULE_VERSION '%GIT_TAG%'
  edit NETCDF4_VERSION '4.7.4'
  edit ODB_API_GIT_TAG 'develop'
  edit ODB_API_VERSION '0.19.4'
  edit ODB_GIT_TAG 'develop'
  edit ODB_TOOLS_GIT_TAG 'develop'
  edit ODB_TOOLS_VERSION '0.18.3'
  edit ODB_VERSION '1.0.11'
  edit ODC_GIT_TAG 'develop'
  edit ODC_VERSION '1.0.3'
  edit OPENJPEG_VERSION '2.3.1'
  edit PACKAGE 'codes_ui'
  edit PACKAGE_NAME 'codes_ui'
  edit PACKAGE_OWNERS 'cgr@ecmwf.int cgm@ecmwf.int ma0@ecmwf.int'
  edit PACKAGE_VERSION '%CODES_UI_VERSION%'
  edit PGEN_VERSION '1.7.1.0'
  edit PKG_BOOST_VERSION '1.71.0'
  edit PREFIX_APPEND ''
  edit PROJ4_VERSION '5.2.0'
  edit PROJ_VERSION '6.3.1'
  edit PYTHON2_VERSION '2.7.15-01'
  edit PYTHON3_VERSION '3.6.10-01'
  edit PYTHON3_VERSION_MAJOR_MINOR '3.6'
  edit PYTHON_GITREPO 'This-needs-tobe-overriden'
  edit PYTHON_PACKAGE 'This-needs-tobe-overriden'
  edit QT_VERSION '5.12.3'
  edit QUEUE 'build'
  edit REMOTE_HOST 'polonius'
  edit REUSE_BUILD_DIR 'False'
  edit SAFE_STACK_SANITIZER 'False'
  edit SANITIZER 'False'
  edit SANITIZER_REFERENCE_DIRECTORY '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
  edit SRCDIR '%BASEDIR%/%PACKAGE%'
  edit STATIC_ANALYSER_ABORT_ON_ERROR 'False'
  edit SWIG_VERSION '3.0.8'
  edit SYNC_RELEASE '2020.04.0'
  edit SYNC_RELEASE_CI_STAGE 'develop'
  edit TARBALL_DIR '/perm/ma/ma0/tmp_tarballs/ci'
  edit TAR_PACKAGE '%PACKAGE%'
  edit TEST 'ctest --output-on-failure -j${CPUS} '
  edit THREAD_SANITIZER 'False'
  edit TRANSI_GIT_TAG 'develop'
  edit TRANSI_VERSION '0.7.1'
  edit TSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
  edit UBSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/ubsan.txt'
  edit UB_SANITIZER 'False'
  edit USER 'ma0'
  edit USE_WORKDIR 'False'
  edit VALGRIND_VERSION '3.15.0'
  edit WORKDIR '%BASEDIR%/workdir/%ECF_NAME%'
  edit WSDIR '/var/tmp/tmpdir/deploy/jtmp.20739'
  label branch "develop"
  task regenerate
    defstatus complete
    edit ECF_JOB_CMD '%ECF_JOB% > %ECF_JOBOUT% 2>&1'
    edit ECF_KILL_CMD 'kill -15 %ECF_RID%'
    edit ECF_PASS 'FREE'
    edit ECF_STATUS_CMD 'ps --sid %ECF_RID% -f'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
    label date ""
    label metabuilder_branch "ci"
  family ecgb
    defstatus suspended
    edit CI_LIMIT '10'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_PREFIX_PATH=/usr/local/apps/qt/5.5.0/5.5/gcc_64/ -DECCODES_PATH=%BASEPREFIX%/eccodes/develop%INSTALL_POSTFIX% -DENABLE_ECCODES=ON -DENABLE_QT5=ON -DEXTRA_TITLE=%SITE%'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'ecgb'
    edit QUEUE 'express'
    edit REMOTE_HOST 'ecgb'
    edit SITE 'ecgb'
    edit WSDIR '$SCRATCH/metabuilds/%ECF_HOST%_%ECF_PORT%/ecgb/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task build
      trigger config == complete
      edit ECF_TRIES '2'
      label host ""
      label options ""
    task test
      trigger build == complete
      edit ECF_TRIES '2'
      edit TEST_DIR '%BUILDIR%'
    task install
      trigger test == complete
      edit ECF_TRIES '2'
    task install_log
      trigger install == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install_log == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family leap42
    defstatus suspended
    edit CI_LIMIT '7'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'leap42'
    edit REMOTE_HOST 'leap42-deploy'
    edit SITE 'leap42'
    edit WSDIR '/tmp/metabuilds/%ECF_HOST%_%ECF_PORT%/leap42/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task build
      trigger config == complete
      edit ECF_TRIES '2'
      label host ""
      label options ""
    task test
      trigger build == complete
      edit ECF_TRIES '2'
      edit TEST_DIR '%BUILDIR%'
    task create_tarball
      trigger test == complete
      edit CMAKE 'cmake'
      edit EXPORTS '"export ECACCOUNT=ecodmdma" "export ecbuild_ROOT=%BASEPREFIX%/ecbuild/develop"'
      label tar ""
      label tardir ""
      label error ""
    task install
      trigger create_tarball == complete
      edit ECF_TRIES '2'
    task install_log
      trigger install == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install_log == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
endsuite
suite ecflow
  edit ADDRESS_SANITIZER 'False'
  edit ADD_ACTION_TRIGGERS 'True'
  edit AEC_VERSION '1.0.4'
  edit ALLOW_INSTALL_TO_OVERWRITE 'True'
  edit ASAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
  edit ATLAS_GIT_TAG 'develop'
  edit ATLAS_VERSION '0.20.2'
  edit BANNER_TASK_END 'banner %TASK%-end || true'
  edit BANNER_TASK_START 'banner %TASK%-start || true'
  edit BASEDIR '%WSDIR%/%PACKAGE%'
  edit BASEPREFIX '$PERM/install/%SITE%'
  edit BUFRDC_EMOSLIB_VERSION '4.6.3'
  edit BUFRDC_TABLES_GIT_TAG 'develop'
  edit BUFRDC_TABLES_VERSION '000415'
  edit BUILDIR '%BASEDIR%/builds'
  edit BUILD_ALWAYS_PARALLEL 'False'
  edit CFI_SANITIZER 'False'
  edit CFL_BASE_URL 'https://software-test.ecmwf.int/wiki'
  edit CFL_SPACE_KEY 'ECFLOW'
  edit CFL_TARBALL_COMMENT ''
  edit CFL_TITLE 'Releases'
  edit CHECK_SYNC_RELEASE 'False'
  edit CLANG_TIDY_ABORT_ON_ERROR 'False'
  edit CMAKE '%ECBUILD%'
  edit CMAKELIST_EXTRACT_PACKAGE_VERSION '$(awk '/^project/ && /%PACKAGE%/ && /VERSION/ {for (I=1;I<=NF;I++) if ($I == "VERSION") {print $(I+1)};}' CMakeLists.txt)'
  edit CMAKE_BUILD_TYPE 'Release'
  edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_FLAGS=-Wno-deprecated-declarations -DCMAKE_PYTHON_INSTALL_TYPE=local -DENABLE_ALL_TESTS=ON -DENABLE_STATIC_BOOST_LIBS=ON -DENABLE_WARNINGS=ON -DSITE_SPECIFIC_SERVER_SCRIPT=/home/ma/emos/bin/ecflow_site.sh'
  edit CMAKE_VERSION '3.13.4'
  edit CODES_UI_GIT_TAG 'develop'
  edit CODES_UI_VERSION '1.3.1'
  edit CPACK 'make package'
  edit CPUS '$(lscpu -p | grep -v '#' | wc -l)'
  edit CRAY_FFTW_VERSION '3.3.8.1'
  edit CRAY_NETCDF_VERSION '4.6.1.3'
  edit CRON_BASEPREFIX '%BASEDIR%/install'
  edit CRON_EMAIL 'ma0@ecmwf.int'
  edit CRON_ENABLED 'False'
  edit CRON_GIT_TAG 'develop'
  edit CUSTOM_COMMANDS ''
  edit DATAFLOW_SANITIZER 'False'
  edit DEFAULT_GNU_COMPILER 'gnu.73'
  edit ECBUILD '%BASEPREFIX%/ecbuild/develop/bin/ecbuild'
  edit ECBUILD_CLONE 'False'
  edit ECBUILD_CLONE_COMMIT 'master'
  edit ECBUILD_GIT_TAG 'develop'
  edit ECBUILD_REPO '%GIT_REPO_PREFIX%/ecsdk/ecbuild.git'
  edit ECBUILD_VERSION '3.3.2'
  edit ECCODES_EXTENSIONS_GIT_TAG 'develop'
  edit ECCODES_EXTENSIONS_VERSION '2.17.0'
  edit ECCODES_GIT_TAG 'develop'
  edit ECCODES_PYTHON_VERSION '0.9.6'
  edit ECCODES_VERSION '2.17.1'
  edit ECFLOW4_LATEST_VERSION '4.17.2'
  edit ECFLOW5_LATEST_VERSION '5.4.0'
  edit ECFLOW_VERSION '5.4.1'
  edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/ecflow'
  edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/ecflow'
  edit ECF_JOB_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_JOB% %ECF_JOBOUT%'
  edit ECF_KILL_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% kill'
  edit ECF_STATUS_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% stat'
  edit ECF_TRIES '1'
  edit ECKIT_GIT_TAG 'develop'
  edit ECKIT_VERSION '1.10.1'
  edit ECQUIS_VERSION '1.0.0'
  edit EMOS_MOD_VERSION '462'
  edit EMOS_PA_VERSION '000462'
  edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TEST_ECF_HOME_POSTFIX=_python3"'
  edit EXPORT_VARIABLES 'False'
  edit FCKIT_GIT_TAG '0.6.3'
  edit FCKIT_VERSION '0.6.3'
  edit FDB_VINTAGE_GIT_TAG '4.10.0'
  edit FDB_VINTAGE_VERSION '4.10.0'
  edit FFTW_VERSION '3.3.8'
  edit FFTW_VERSION_CRAY '3.3.4.11'
  edit GIT 'git'
  edit GITHUB_PACKAGE_NAME '%PACKAGE_NAME%'
  edit GITHUB_REPO 'git@github.com:ecmwf'
  edit GITREPO '%GIT_REPO_PREFIX%/ecflow/ecflow.git'
  edit GIT_CLONE 'True'
  edit GIT_CLONE_OPTIONS '--depth=1 --branch %GIT_TAG%'
  edit GIT_METABUILDER_TAG 'ci'
  edit GIT_REPO_PREFIX 'ssh://git@git.ecmwf.int:7999'
  edit GIT_TAG 'develop'
  edit GIT_VERSION '2.26.0'
  edit GRIB_API_VERSION '1.28.0'
  edit INSTALLABLE_TO_USR_LOCAL_APPS 'True'
  edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop'
  edit IWYU_VERSION '0.11'
  edit KSH_EXTRACT_PACKAGE_VERSION '`grep _VERSION_STR VERSION.cmake | grep -o '[0-9]*\.[0-9]*\.[0-9]*'`'
  edit LAPACK_VERSION '3.5.0'
  edit LIBEMOS_GIT_TAG 'develop'
  edit LIBEMOS_VERSION '4.6.2'
  edit LINTERS_REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}'
  edit LSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/lsan.txt'
  edit MAGICS_GIT_TAG 'develop'
  edit MAGICS_PYTHON_VERSION 'master'
  edit MAGICS_VERSION '4.3.3'
  edit MAKE 'make'
  edit MARS_CLIENT_GIT_TAG '6.27.2.2'
  edit MARS_CLIENT_PACKAGE_GIT_TAG '6.27.6'
  edit MARS_CLIENT_PACKAGE_VERSION '6.27.6'
  edit MARS_CLIENT_VERSION '6.27.2.2'
  edit MEMORY_SANITIZER 'False'
  edit METABUILDER_HOME '/var/tmp/ma0/workspace/metabuilder'
  edit METAB_INSTALL_REMOVE_WRITE_PERMISSIONS 'True'
  edit METAB_PYTHON_VERSION 'python3'
  edit METKIT_GIT_TAG 'develop'
  edit METKIT_VERSION '1.4.6'
  edit METVIEW_GIT_TAG 'develop'
  edit METVIEW_PYTHON_VERSION '1.4.2'
  edit METVIEW_VERSION '5.8.3'
  edit MIR_BUNDLE_VERSION '0.9.0.0'
  edit MIR_GIT_TAG 'develop'
  edit MIR_VERSION '1.5.0'
  edit MODULE_ACTIONS '"module load gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load ecflow/%ECF_VERSION%"'
  edit MODULE_ACTIONS_FIRST ''
  edit MODULE_ACTIONS_HDR 'module_actions.h'
  edit MODULE_CHECKING_HDR 'module_checking.h'
  edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow"'
  edit MODULE_LOAD_SUPPORTED 'True'
  edit MODULE_NAME 'ecflow'
  edit MODULE_TAG_NAME '5new'
  edit MODULE_VERSION '%GIT_TAG%'
  edit NETCDF4_VERSION '4.7.4'
  edit ODB_API_GIT_TAG 'develop'
  edit ODB_API_VERSION '0.19.4'
  edit ODB_GIT_TAG 'develop'
  edit ODB_TOOLS_GIT_TAG 'develop'
  edit ODB_TOOLS_VERSION '0.18.3'
  edit ODB_VERSION '1.0.11'
  edit ODC_GIT_TAG 'develop'
  edit ODC_VERSION '1.0.3'
  edit OPENJPEG_VERSION '2.3.1'
  edit PACKAGE 'ecflow'
  edit PACKAGE_NAME 'ecflow'
  edit PACKAGE_OWNERS 'ma0@ecmwf.int'
  edit PACKAGE_VERSION '%GIT_TAG%'
  edit PGEN_VERSION '1.7.1.0'
  edit PKG_BOOST_VERSION '1.71.0'
  edit PREFIX_APPEND ''
  edit PROJ4_VERSION '5.2.0'
  edit PROJ_VERSION '6.3.1'
  edit PYTHON 'python3'
  edit PYTHON2 'True'
  edit PYTHON2_VERSION '2.7.15-01'
  edit PYTHON3 'True'
  edit PYTHON3_VERSION '3.6.10-01'
  edit PYTHON3_VERSION_MAJOR_MINOR '3.6'
  edit PYTHON_GITREPO 'This-needs-tobe-overriden'
  edit PYTHON_PACKAGE 'This-needs-tobe-overriden'
  edit QT_VERSION '5.12.3'
  edit QUEUE 'build'
  edit REMOTE_HOST 'polonius'
  edit REUSE_BUILD_DIR 'False'
  edit SAFE_STACK_SANITIZER 'False'
  edit SANITIZER 'False'
  edit SANITIZER_REFERENCE_DIRECTORY '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
  edit SRCDIR '%BASEDIR%/%PACKAGE%'
  edit STATIC_ANALYSER_ABORT_ON_ERROR 'False'
  edit SWIG_VERSION '3.0.8'
  edit SYNC_RELEASE '2020.04.0'
  edit SYNC_RELEASE_CI_STAGE 'develop'
  edit TARBALL_DIR '/perm/ma/ma0/tarballs/ecflow/%PACKAGE_VERSION%'
  edit TAR_PACKAGE 'ecFlow'
  edit TEST 'ctest -E "(perf_|test_migration)" -j${CPUS} -VV --output-on-failure'
  edit THREAD_SANITIZER 'False'
  edit TRANSI_GIT_TAG 'develop'
  edit TRANSI_VERSION '0.7.1'
  edit TSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
  edit UBSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/ubsan.txt'
  edit UB_SANITIZER 'False'
  edit USER 'ma0'
  edit USE_WORKDIR 'False'
  edit VALGRIND_VERSION '3.15.0'
  edit WORKDIR '%BASEDIR%/workdir/%ECF_NAME%'
  edit WSDIR '/var/tmp/tmpdir/deploy/jtmp.20739'
  label branch "develop"
  task regenerate
    defstatus complete
    edit ECF_JOB_CMD '%ECF_JOB% > %ECF_JOBOUT% 2>&1'
    edit ECF_KILL_CMD 'kill -15 %ECF_RID%'
    edit ECF_PASS 'FREE'
    edit ECF_STATUS_CMD 'ps --sid %ECF_RID% -f'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
    label date ""
    label metabuilder_branch "ci"
  family cca
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_CXX_FLAGS=-Wno-deprecated-declarations -DCMAKE_C_COMPILER=cc -DCMAKE_PYTHON_INSTALL_TYPE=local -DENABLE_ALL_TESTS=ON -DENABLE_STATIC_BOOST_LIBS=ON -DENABLE_UI=OFF -DENABLE_WARNINGS=ON -DSITE_SPECIFIC_SERVER_SCRIPT=/home/ma/emos/bin/ecflow_site.sh'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/ecflow/cca'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/ecflow/cca:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cca-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export CRAY_ADD_RPATH=yes" "export ECACCOUNT=ecodmdma" "export ECFLOW_CRAY_BATCH=1" "export TEST_ECF_HOME_POSTFIX=_python3" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '19000'
    edit MODULE_ACTIONS '"module swap PrgEnv-cray PrgEnv-gnu" "module swap craype-broadwell craype-ivybridge" "module swap cdt/17.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load ecflow/%ECF_VERSION%"'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cca'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'cca'
    edit SITE 'cca'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit TEST 'ctest -E "(s_test|perf_|test_migration)" -j${CPUS} -VV'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task build
      trigger config == complete
      edit ECF_TRIES '2'
      label host ""
      label options ""
    task test
      trigger build == complete
      edit ECF_TRIES '1'
      edit TEST_DIR '%BUILDIR%'
    task install
      trigger test == complete
      edit ECF_TRIES '2'
      label info ""
    task install_log
      trigger install == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install_log == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family ccb
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_CXX_FLAGS=-Wno-deprecated-declarations -DCMAKE_C_COMPILER=cc -DCMAKE_PYTHON_INSTALL_TYPE=local -DENABLE_ALL_TESTS=ON -DENABLE_STATIC_BOOST_LIBS=ON -DENABLE_UI=OFF -DENABLE_WARNINGS=ON -DSITE_SPECIFIC_SERVER_SCRIPT=/home/ma/emos/bin/ecflow_site.sh'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/ecflow/ccb'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/ecflow/ccb:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'ccb-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export CRAY_ADD_RPATH=yes" "export ECACCOUNT=ecodmdma" "export ECFLOW_CRAY_BATCH=1" "export TEST_ECF_HOME_POSTFIX=_python3" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '19000'
    edit MODULE_ACTIONS '"module swap PrgEnv-cray PrgEnv-gnu" "module swap craype-broadwell craype-ivybridge" "module swap cdt/17.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load ecflow/%ECF_VERSION%"'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'ccb'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'ccb'
    edit SITE 'ccb'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit TEST 'ctest -E "(s_test|perf_|test_migration)" -j${CPUS} -VV'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task build
      trigger config == complete
      edit ECF_TRIES '2'
      label host ""
      label options ""
    task test
      trigger build == complete
      edit ECF_TRIES '1'
      edit TEST_DIR '%BUILDIR%'
    task install
      trigger test == complete
      edit ECF_TRIES '2'
      label info ""
    task install_log
      trigger install == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install_log == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family cct
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_CXX_FLAGS=-Wno-deprecated-declarations -DCMAKE_C_COMPILER=cc -DCMAKE_PYTHON_INSTALL_TYPE=local -DENABLE_ALL_TESTS=ON -DENABLE_STATIC_BOOST_LIBS=ON -DENABLE_UI=OFF -DENABLE_WARNINGS=ON -DSITE_SPECIFIC_SERVER_SCRIPT=/home/ma/emos/bin/ecflow_site.sh'
    edit CPUS '$EC_total_tasks'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/ecflow/cct'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/ecflow/cct:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cct-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '4'
    edit EXPORTS '"export CRAY_ADD_RPATH=yes" "export ECACCOUNT=ecodmdma" "export ECFLOW_CRAY_BATCH=1" "export TEST_ECF_HOME_POSTFIX=_python3" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '19000'
    edit MODULE_ACTIONS '"module swap PrgEnv-cray PrgEnv-gnu" "module load craype-ivybridge" "module swap cdt/17.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load ecflow/%ECF_VERSION%"'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cct'
    edit QSUB_LP 'QSUB -lP %EC_TOTAL_TASKS%'
    edit QUEUE 'nf'
    edit REMOTE_HOST 'cct'
    edit SITE 'cct'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit TEST 'ctest -E "(s_test|perf_|test_migration)" -j${CPUS} -VV'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task build
      trigger config == complete
      edit ECF_TRIES '2'
      label host ""
      label options ""
    task test
      trigger build == complete
      edit ECF_TRIES '1'
      edit TEST_DIR '%BUILDIR%'
    task install
      trigger test == complete
      edit ECF_TRIES '2'
      label info ""
    task install_log
      trigger install == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install_log == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family ecflow_metab_vm
    defstatus suspended
    edit BANNER_TASK_END 'echo %TASK%-end'
    edit BANNER_TASK_START 'echo %TASK%-start'
    edit CI_LIMIT '1'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_FLAGS=-Wno-deprecated-declarations -DCMAKE_PYTHON_INSTALL_TYPE=local -DENABLE_ALL_TESTS=ON -DENABLE_GUI=OFF -DENABLE_STATIC_BOOST_LIBS=ON -DENABLE_UI=OFF -DENABLE_WARNINGS=ON -DSITE_SPECIFIC_SERVER_SCRIPT=/home/ma/emos/bin/ecflow_site.sh'
    edit ECF_JOB_CMD '%ECF_JOB% 1> %ECF_JOBOUT% 2>&1'
    edit ECF_KILL_CMD 'kill -15 %ECF_RID%'
    edit ECF_STATUS_CMD 'ps --sid %ECF_RID% -f'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit M_SITE 'vsms1'
    edit REMOTE_HOST 'polonius'
    edit SITE 'vsms1'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/ecflow-metab/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task build
      trigger config == complete
      edit ECF_TRIES '2'
      label host ""
      label options ""
    task test
      trigger build == complete
      edit ECF_TRIES '1'
      edit TEST_DIR '%BUILDIR%'
    task install
      trigger test == complete
      edit ECF_TRIES '2'
      label info ""
    task install_log
      trigger install == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install_log == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family ecgb
    defstatus suspended
    edit CI_LIMIT '10'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_FLAGS=-Wno-deprecated-declarations -DCMAKE_PREFIX_PATH=/usr/local/apps/qt/5.5.0/5.5/gcc_64/ -DCMAKE_PYTHON_INSTALL_TYPE=local -DENABLE_ALL_TESTS=ON -DENABLE_STATIC_BOOST_LIBS=ON -DENABLE_UI_BACKTRACE=ON -DENABLE_UI_USAGE_LOG=ON -DENABLE_WARNINGS=ON -DLOGUI_LOG_FILE=/home/graphics/cgx/public/ecflowlogui_log.txt -DSITE_SPECIFIC_SERVER_SCRIPT=/home/ma/emos/bin/ecflow_site.sh -DUI_BACKTRACE_EMAIL_ADDRESS_FILE=/usr/local/apps/ecflow/ui-log-addrs.txt -DUI_LOG_FILE=/home/graphics/cgx/public/ecflowui_log.txt -DUI_LOG_SITE_TAG=%SITE%'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'ecgb'
    edit QUEUE 'express'
    edit REMOTE_HOST 'ecgb'
    edit SITE 'ecgb'
    edit WSDIR '$SCRATCH/metabuilds/%ECF_HOST%_%ECF_PORT%/ecgb/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task build
      trigger config == complete
      edit ECF_TRIES '2'
      label host ""
      label options ""
    task test
      trigger build == complete
      edit ECF_TRIES '1'
      edit TEST_DIR '%BUILDIR%'
    task install
      trigger test == complete
      edit ECF_TRIES '2'
      label info ""
    task install_log
      trigger install == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install_log == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family leap42
    defstatus suspended
    edit CI_LIMIT '7'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_FLAGS=-Wno-deprecated-declarations -DCMAKE_PREFIX_PATH=/usr/local/apps/qt/5.7.0/5.7/gcc_64/ -DCMAKE_PYTHON_INSTALL_TYPE=local -DENABLE_ALL_TESTS=ON -DENABLE_STATIC_BOOST_LIBS=ON -DENABLE_UI_BACKTRACE=ON -DENABLE_UI_USAGE_LOG=ON -DENABLE_WARNINGS=ON -DLOGUI_LOG_FILE=/home/graphics/cgx/public/ecflowlogui_log.txt -DSITE_SPECIFIC_SERVER_SCRIPT=/home/ma/emos/bin/ecflow_site.sh -DUI_BACKTRACE_EMAIL_ADDRESS_FILE=/usr/local/apps/ecflow/ui-log-addrs.txt -DUI_LOG_FILE=/home/graphics/cgx/public/ecflowui_log.txt -DUI_LOG_SITE_TAG=%SITE%'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load ecflow/%ECF_VERSION%"'
    edit M_SITE 'leap42'
    edit REMOTE_HOST 'leap42-deploy'
    edit SITE 'leap42'
    edit WSDIR '/tmp/metabuilds/%ECF_HOST%_%ECF_PORT%/leap42/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_FLAGS=-Wno-deprecated-declarations -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_PREFIX_PATH=/usr/local/apps/qt/5.7.0/5.7/gcc_64/ -DCMAKE_PYTHON_INSTALL_TYPE=local -DENABLE_ALL_TESTS=ON -DENABLE_STATIC_BOOST_LIBS=ON -DENABLE_UI_BACKTRACE=ON -DENABLE_UI_USAGE_LOG=ON -DENABLE_WARNINGS=ON -DLOGUI_LOG_FILE=/home/graphics/cgx/public/ecflowlogui_log.txt -DSITE_SPECIFIC_SERVER_SCRIPT=/home/ma/emos/bin/ecflow_site.sh -DUI_BACKTRACE_EMAIL_ADDRESS_FILE=/usr/local/apps/ecflow/ui-log-addrs.txt -DUI_LOG_FILE=/home/graphics/cgx/public/ecflowui_log.txt -DUI_LOG_SITE_TAG=%SITE%'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '1'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
        label info ""
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.93
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_FLAGS=-Wno-deprecated-declarations -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_PREFIX_PATH=/usr/local/apps/qt/5.7.0/5.7/gcc_64/ -DCMAKE_PYTHON_INSTALL_TYPE=local -DENABLE_ALL_TESTS=ON -DENABLE_STATIC_BOOST_LIBS=ON -DENABLE_UI_BACKTRACE=ON -DENABLE_UI_USAGE_LOG=ON -DENABLE_WARNINGS=ON -DLOGUI_LOG_FILE=/home/graphics/cgx/public/ecflowlogui_log.txt -DSITE_SPECIFIC_SERVER_SCRIPT=/home/ma/emos/bin/ecflow_site.sh -DUI_BACKTRACE_EMAIL_ADDRESS_FILE=/usr/local/apps/ecflow/ui-log-addrs.txt -DUI_LOG_FILE=/home/graphics/cgx/public/ecflowui_log.txt -DUI_LOG_SITE_TAG=%SITE%'
      edit COMPILER 'GNU.93'
      edit EXPORTS '"export ECACCOUNT=ecodmdma" "export LD_LIBRARY_PATH=/usr/local/apps/gcc/9.3.0/lib64/" "export TEST_ECF_HOME_POSTFIX=_python3"'
      edit MODULE_ACTIONS '"module swap gnu/9.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '1'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
        label info ""
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_FLAGS=-Wno-deprecated-declarations -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_PREFIX_PATH=/usr/local/apps/qt/5.7.0/5.7/gcc_64/ -DCMAKE_PYTHON_INSTALL_TYPE=local -DENABLE_ALL_TESTS=ON -DENABLE_STATIC_BOOST_LIBS=ON -DENABLE_UI_BACKTRACE=ON -DENABLE_UI_USAGE_LOG=ON -DENABLE_WARNINGS=ON -DLOGUI_LOG_FILE=/home/graphics/cgx/public/ecflowlogui_log.txt -DSITE_SPECIFIC_SERVER_SCRIPT=/home/ma/emos/bin/ecflow_site.sh -DUI_BACKTRACE_EMAIL_ADDRESS_FILE=/usr/local/apps/ecflow/ui-log-addrs.txt -DUI_LOG_FILE=/home/graphics/cgx/public/ecflowui_log.txt -DUI_LOG_SITE_TAG=%SITE%'
      edit COMPILER 'CLANG.90'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '1'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
        label info ""
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90.linters
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_FLAGS=-Wno-deprecated-declarations -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_PREFIX_PATH=/usr/local/apps/qt/5.7.0/5.7/gcc_64/ -DCMAKE_PYTHON_INSTALL_TYPE=local -DENABLE_ALL_TESTS=ON -DENABLE_STATIC_BOOST_LIBS=ON -DENABLE_UI_BACKTRACE=ON -DENABLE_UI_USAGE_LOG=ON -DENABLE_WARNINGS=ON -DLOGUI_LOG_FILE=/home/graphics/cgx/public/ecflowlogui_log.txt -DSITE_SPECIFIC_SERVER_SCRIPT=/home/ma/emos/bin/ecflow_site.sh -DUI_BACKTRACE_EMAIL_ADDRESS_FILE=/usr/local/apps/ecflow/ui-log-addrs.txt -DUI_LOG_FILE=/home/graphics/cgx/public/ecflowui_log.txt -DUI_LOG_SITE_TAG=%SITE%'
      edit COMPILER 'CLANG.90.LINTERS'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task cppcheck
        trigger config == complete
        edit CPPCHECK_FLAGS '--template='[{file}:{line}]: {severity} : {id} : {message}' --inline-suppr --platform=unix64 --enable=performance --enable=portability --enable=information --std=c11 --std=c++14 -DDEBUG=1 -DUCHAR_MAX=255 -DCEREAL_RAPIDJSON_ENDIAN=CEREAL_RAPIDJSON_LITTLEENDIAN -DNULL=0 -q -j8'
        edit CPPCHECK_VERSION '1.89'
        edit REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}/cpp_check'
      task iwyu
        trigger cppcheck == complete
        edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load ecflow/%ECF_VERSION%" "module load iwyu/%IWYU_VERSION%"'
      family clang_linters
        trigger iwyu == complete
        limit clang_linters 1
        inlimit clang_linters
        task clang_tidy
          edit REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}/clang_tidy'
        task static_analyser
          edit REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}/static_analyser'
          edit SCAN_BUILD_OPTIONS '--force-analyze-debug-code -o /perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}/static_analyser -k make -j8'
          edit USE_CLANG_CHECK 'False'
          label scan_view ""
      endfamily
      task wipe_build
        trigger clang_linters == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90.address_sanitizer
      edit ADDRESS_SANITIZER 'True'
      edit ASAN_OPTIONS 'suppressions=%SRCDIR%/build_scripts/ecflow_asan.supp'
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_FLAGS="-Wno-deprecated-declarations -fsanitize=address -fno-omit-frame-pointer" -DCMAKE_C_FLAGS="-fsanitize=address -fno-omit-frame-pointer" -DCMAKE_EXE_LINKER_FLAGS=-fsanitize=address -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_PREFIX_PATH=/usr/local/apps/qt/5.7.0/5.7/gcc_64/ -DCMAKE_PYTHON_INSTALL_TYPE=local -DENABLE_ALL_TESTS=ON -DENABLE_STATIC_BOOST_LIBS=ON -DENABLE_UI_BACKTRACE=ON -DENABLE_UI_USAGE_LOG=ON -DENABLE_WARNINGS=ON -DLOGUI_LOG_FILE=/home/graphics/cgx/public/ecflowlogui_log.txt -DSITE_SPECIFIC_SERVER_SCRIPT=/home/ma/emos/bin/ecflow_site.sh -DUI_BACKTRACE_EMAIL_ADDRESS_FILE=/usr/local/apps/ecflow/ui-log-addrs.txt -DUI_LOG_FILE=/home/graphics/cgx/public/ecflowui_log.txt -DUI_LOG_SITE_TAG=%SITE%'
      edit COMPILER 'CLANG.90.ADDRESS_SANITIZER'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit PYTHON2 'False'
      edit PYTHON3 'False'
      edit SANITIZER 'True'
      edit TEST 'ctest -E "(perf_|test_migration|s_test|s_test_zombies|py_|py3_)" -j${CPUS} -VV --output-on-failure'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit TEST_DIR '%BUILDIR%'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90.ub_sanitizer
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_FLAGS="-Wno-deprecated-declarations -fsanitize=undefined -fno-omit-frame-pointer -fno-sanitize-recover=all" -DCMAKE_C_FLAGS="-fsanitize=undefined -fno-omit-frame-pointer" -DCMAKE_EXE_LINKER_FLAGS=-fsanitize=undefined -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_PREFIX_PATH=/usr/local/apps/qt/5.7.0/5.7/gcc_64/ -DCMAKE_PYTHON_INSTALL_TYPE=local -DENABLE_ALL_TESTS=ON -DENABLE_STATIC_BOOST_LIBS=ON -DENABLE_UI_BACKTRACE=ON -DENABLE_UI_USAGE_LOG=ON -DENABLE_WARNINGS=ON -DLOGUI_LOG_FILE=/home/graphics/cgx/public/ecflowlogui_log.txt -DSITE_SPECIFIC_SERVER_SCRIPT=/home/ma/emos/bin/ecflow_site.sh -DUI_BACKTRACE_EMAIL_ADDRESS_FILE=/usr/local/apps/ecflow/ui-log-addrs.txt -DUI_LOG_FILE=/home/graphics/cgx/public/ecflowui_log.txt -DUI_LOG_SITE_TAG=%SITE%'
      edit COMPILER 'CLANG.90.UB_SANITIZER'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load boost/%PKG_BOOST_VERSION%" "module load ecflow/%ECF_VERSION%"'
      edit PYTHON2 'False'
      edit PYTHON3 'False'
      edit SANITIZER 'True'
      edit TEST 'ctest -E "(perf_|test_migration|s_test|s_test_zombies|py_|py3_)" -j${CPUS} -VV --output-on-failure'
      edit UB_SANITIZER 'True'
      edit UB_SANITIZER_LD_PRELOAD 'export LD_PRELOAD=/usr/local/apps/clang/9.0.1/lib64/libubsan.so'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit TEST_DIR '%BUILDIR%'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family lxc
    defstatus suspended
    edit CI_LIMIT '10'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_FLAGS=-Wno-deprecated-declarations -DCMAKE_PREFIX_PATH=/usr/local/apps/qt/5.5.0/5.5/gcc_64/ -DCMAKE_PYTHON_INSTALL_TYPE=local -DENABLE_ALL_TESTS=ON -DENABLE_STATIC_BOOST_LIBS=ON -DENABLE_UI_BACKTRACE=ON -DENABLE_UI_USAGE_LOG=ON -DENABLE_WARNINGS=ON -DLOGUI_LOG_FILE=/home/graphics/cgx/public/ecflowlogui_log.txt -DSITE_SPECIFIC_SERVER_SCRIPT=/home/ma/emos/bin/ecflow_site.sh -DUI_BACKTRACE_EMAIL_ADDRESS_FILE=/usr/local/apps/ecflow/ui-log-addrs.txt -DUI_LOG_FILE=/home/graphics/cgx/public/ecflowui_log.txt -DUI_LOG_SITE_TAG=%SITE%'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'lxc'
    edit QUEUE 'express'
    edit REMOTE_HOST 'lxc'
    edit SITE 'lxc'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxc/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task build
      trigger config == complete
      edit ECF_TRIES '2'
      label host ""
      label options ""
    task test
      trigger build == complete
      edit ECF_TRIES '1'
      edit TEST_DIR '%BUILDIR%'
    task install
      trigger test == complete
      edit ECF_TRIES '2'
      label info ""
    task install_log
      trigger install == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install_log == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family lxg
    defstatus suspended
    edit CI_LIMIT '10'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_FLAGS=-Wno-deprecated-declarations -DCMAKE_PREFIX_PATH=/usr/local/apps/qt/5.5.0/5.5/gcc_64/ -DCMAKE_PYTHON_INSTALL_TYPE=local -DENABLE_ALL_TESTS=ON -DENABLE_STATIC_BOOST_LIBS=ON -DENABLE_UI_BACKTRACE=ON -DENABLE_UI_USAGE_LOG=ON -DENABLE_WARNINGS=ON -DLOGUI_LOG_FILE=/home/graphics/cgx/public/ecflowlogui_log.txt -DSITE_SPECIFIC_SERVER_SCRIPT=/home/ma/emos/bin/ecflow_site.sh -DUI_BACKTRACE_EMAIL_ADDRESS_FILE=/usr/local/apps/ecflow/ui-log-addrs.txt -DUI_LOG_FILE=/home/graphics/cgx/public/ecflowui_log.txt -DUI_LOG_SITE_TAG=%SITE%'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'lxg'
    edit QUEUE 'normal'
    edit REMOTE_HOST 'lxg'
    edit SITE 'lxg'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxg/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task build
      trigger config == complete
      edit ECF_TRIES '2'
      label host ""
      label options ""
    task test
      trigger build == complete
      edit ECF_TRIES '1'
      edit TEST_DIR '%BUILDIR%'
    task install
      trigger test == complete
      edit ECF_TRIES '2'
      label info ""
    task install_log
      trigger install == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install_log == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family sappb
    defstatus suspended
    edit ASAN_FILE '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
    edit BASEPREFIX '/autofs/vol/deploy/install/%SITE%'
    edit CI_LIMIT '8'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_FLAGS=-Wno-deprecated-declarations -DCMAKE_PYTHON_INSTALL_TYPE=local -DENABLE_ALL_TESTS=ON -DENABLE_STATIC_BOOST_LIBS=ON -DENABLE_UI=OFF -DENABLE_WARNINGS=ON -DSITE_SPECIFIC_SERVER_SCRIPT=/home/ma/emos/bin/ecflow_site.sh'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit M_SITE 'sappb'
    edit REMOTE_HOST 'sappb-deploy'
    edit SANITIZER_REFERENCE_DIRECTORY '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
    edit SITE 'sappb'
    edit TSAN_FILE '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
    edit WSDIR '$HOME/metabuilds/%ECF_HOST%_%ECF_PORT%/sappb/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task build
      trigger config == complete
      edit ECF_TRIES '2'
      label host ""
      label options ""
    task test
      trigger build == complete
      edit ECF_TRIES '1'
      edit TEST_DIR '%BUILDIR%'
    task install
      trigger test == complete
      edit ECF_TRIES '2'
      label info ""
    task install_log
      trigger install == complete
      edit ECF_TRIES '2'
    task wipe_build
      trigger install_log == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
endsuite
suite pgen
  edit ADDRESS_SANITIZER 'False'
  edit ADD_ACTION_TRIGGERS 'True'
  edit AEC_VERSION '1.0.4'
  edit ALLOW_INSTALL_TO_OVERWRITE 'True'
  edit ASAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
  edit ATLAS_GIT_TAG 'develop'
  edit ATLAS_VERSION '0.20.2'
  edit BANNER_TASK_END 'banner %TASK%-end || true'
  edit BANNER_TASK_START 'banner %TASK%-start || true'
  edit BASEDIR '%WSDIR%/%PACKAGE%'
  edit BASEPREFIX '$PERM/install/%SITE%'
  edit BUFRDC_EMOSLIB_VERSION '4.6.3'
  edit BUFRDC_TABLES_GIT_TAG 'develop'
  edit BUFRDC_TABLES_VERSION '000415'
  edit BUILDIR '%BASEDIR%/builds'
  edit BUILD_ALWAYS_PARALLEL 'False'
  edit CFI_SANITIZER 'False'
  edit CFL_BASE_URL 'https://software-test.ecmwf.int/wiki'
  edit CFL_SPACE_KEY '<TODO>'
  edit CFL_TARBALL_COMMENT ''
  edit CFL_TITLE 'Releases'
  edit CHECK_SYNC_RELEASE 'False'
  edit CLANG_TIDY_ABORT_ON_ERROR 'False'
  edit CMAKE '%ECBUILD%'
  edit CMAKELIST_EXTRACT_PACKAGE_VERSION '$(awk '/^project/ && /%PACKAGE%/ && /VERSION/ {for (I=1;I<=NF;I++) if ($I == "VERSION") {print $(I+1)};}' CMakeLists.txt)'
  edit CMAKE_BUILD_TYPE 'RelWithDebInfo'
  edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DECCODES_IFS_SAMPLES_PATH=ON -DENABLE_ECCODES=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_FDB_TOOLS=OFF -DENABLE_HERMES=OFF -DENABLE_MARS_SERVER=OFF -DENABLE_MEMFS=ON -DENABLE_MIR_TOOLS=ON -DENABLE_MPI=ON -DENABLE_ODB=OFF -DENABLE_OMP=OFF -DENABLE_PYTHON=OFF -DENABLE_SANDBOX=OFF -DENABLE_TESSELATION=OFF -DENABLE_TESTS=ON -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config'
  edit CMAKE_VERSION '3.13.4'
  edit CODES_UI_GIT_TAG 'develop'
  edit CODES_UI_VERSION '1.3.1'
  edit CPACK 'make package'
  edit CPUS '$(lscpu -p | grep -v '#' | wc -l)'
  edit CRAY_FFTW_VERSION '3.3.8.1'
  edit CRAY_NETCDF_VERSION '4.6.1.3'
  edit CRON_BASEPREFIX '%BASEDIR%/install'
  edit CRON_EMAIL ''
  edit CRON_ENABLED 'False'
  edit CRON_GIT_TAG 'develop'
  edit CUSTOM_COMMANDS ''
  edit DATAFLOW_SANITIZER 'False'
  edit DEFAULT_GNU_COMPILER 'gnu.73'
  edit ECBUILD '%BASEPREFIX%/ecbuild/develop/bin/ecbuild'
  edit ECBUILD_CLONE 'False'
  edit ECBUILD_CLONE_COMMIT 'master'
  edit ECBUILD_GIT_TAG 'develop'
  edit ECBUILD_REPO '%GIT_REPO_PREFIX%/ecsdk/ecbuild.git'
  edit ECBUILD_VERSION '3.3.2'
  edit ECCODES_EXTENSIONS_GIT_TAG 'develop'
  edit ECCODES_EXTENSIONS_VERSION '2.17.0'
  edit ECCODES_GIT_TAG 'develop'
  edit ECCODES_PYTHON_VERSION '0.9.6'
  edit ECCODES_VERSION '2.17.1'
  edit ECFLOW5_LATEST_VERSION '5.4.0'
  edit ECFLOW_VERSION '5.4.1'
  edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/pgen'
  edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/pgen'
  edit ECF_JOB_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_JOB% %ECF_JOBOUT%'
  edit ECF_KILL_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% kill'
  edit ECF_STATUS_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% stat'
  edit ECF_TRIES '1'
  edit ECKIT_GIT_TAG 'develop'
  edit ECKIT_VERSION '1.10.1'
  edit ECQUIS_VERSION '1.0.0'
  edit EMOS_MOD_VERSION '462'
  edit EMOS_PA_VERSION '000462'
  edit EXPORTS '"export ECACCOUNT=ecodmdma" "export ftp_proxy=http://proxy.ecmwf.int:3333" "export http_proxy=http://proxy.ecmwf.int:3333" "export https_proxy=http://proxy.ecmwf.int:3333"'
  edit EXPORT_VARIABLES 'False'
  edit FCKIT_GIT_TAG '0.6.3'
  edit FCKIT_VERSION '0.6.3'
  edit FDB_VINTAGE_GIT_TAG '4.10.0'
  edit FDB_VINTAGE_VERSION '4.10.0'
  edit FFTW_VERSION '3.3.8'
  edit FFTW_VERSION_CRAY '3.3.4.11'
  edit GIT 'git'
  edit GITHUB_PACKAGE_NAME '%PACKAGE_NAME%'
  edit GITHUB_REPO 'git@github.com:ecmwf'
  edit GITREPO '%GIT_REPO_PREFIX%/prodgen/pgen-bundle.git'
  edit GIT_CLONE 'True'
  edit GIT_CLONE_OPTIONS '--depth=1 --branch %GIT_TAG%'
  edit GIT_METABUILDER_TAG 'ci'
  edit GIT_REPO_PREFIX 'ssh://git@git.ecmwf.int:7999'
  edit GIT_TAG 'develop'
  edit GIT_VERSION '2.26.0'
  edit GRIB_API_VERSION '1.28.0'
  edit INSTALLABLE_TO_USR_LOCAL_APPS 'True'
  edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop'
  edit IWYU_VERSION '0.11'
  edit KSH_EXTRACT_PACKAGE_VERSION '`grep _VERSION_STR VERSION.cmake | grep -o '[0-9]*\.[0-9]*\.[0-9]*'`'
  edit LAPACK_VERSION '3.5.0'
  edit LIBEMOS_GIT_TAG 'develop'
  edit LIBEMOS_VERSION '4.6.2'
  edit LINTERS_REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}'
  edit LSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/lsan.txt'
  edit MAGICS_GIT_TAG 'develop'
  edit MAGICS_PYTHON_VERSION 'master'
  edit MAGICS_VERSION '4.3.3'
  edit MAKE 'make'
  edit MARS_CLIENT_GIT_TAG '6.27.2.2'
  edit MARS_CLIENT_PACKAGE_GIT_TAG '6.27.6'
  edit MARS_CLIENT_PACKAGE_VERSION '6.27.6'
  edit MARS_CLIENT_VERSION '6.27.2.2'
  edit MEMORY_SANITIZER 'False'
  edit METABUILDER_HOME '/var/tmp/ma0/workspace/metabuilder'
  edit METAB_INSTALL_REMOVE_WRITE_PERMISSIONS 'True'
  edit METAB_PYTHON_VERSION 'python3'
  edit METKIT_GIT_TAG 'develop'
  edit METKIT_VERSION '1.4.6'
  edit METVIEW_GIT_TAG 'develop'
  edit METVIEW_PYTHON_VERSION '1.4.2'
  edit METVIEW_VERSION '5.8.3'
  edit MIR_BUNDLE_VERSION '0.9.0.0'
  edit MIR_GIT_TAG 'develop'
  edit MIR_VERSION '1.5.0'
  edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%"'
  edit MODULE_ACTIONS_FIRST ''
  edit MODULE_ACTIONS_HDR 'module_actions.h'
  edit MODULE_CHECKING_HDR 'module_checking.h'
  edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow"'
  edit MODULE_LOAD_SUPPORTED 'True'
  edit MODULE_NAME 'pgen'
  edit MODULE_VERSION '%GIT_TAG%'
  edit NETCDF4_VERSION '4.7.4'
  edit ODB_API_GIT_TAG 'develop'
  edit ODB_API_VERSION '0.19.4'
  edit ODB_GIT_TAG 'develop'
  edit ODB_TOOLS_GIT_TAG 'develop'
  edit ODB_TOOLS_VERSION '0.18.3'
  edit ODB_VERSION '1.0.11'
  edit ODC_GIT_TAG 'develop'
  edit ODC_VERSION '1.0.3'
  edit OPENJPEG_VERSION '2.3.1'
  edit PACKAGE 'pgen'
  edit PACKAGE_NAME 'pgen'
  edit PACKAGE_OWNERS 'ma0@ecmwf.int'
  edit PACKAGE_VERSION '1.7.1.0'
  edit PGEN_VERSION '1.7.1.0'
  edit PKG_BOOST_VERSION '1.71.0'
  edit PREFIX_APPEND ''
  edit PROJ4_VERSION '5.2.0'
  edit PROJ_VERSION '6.3.1'
  edit PYTHON2_VERSION '2.7.15-01'
  edit PYTHON3_VERSION '3.6.10-01'
  edit PYTHON3_VERSION_MAJOR_MINOR '3.6'
  edit PYTHON_GITREPO 'This-needs-tobe-overriden'
  edit PYTHON_PACKAGE 'This-needs-tobe-overriden'
  edit QT_VERSION '5.12.3'
  edit QUEUE 'build'
  edit REMOTE_HOST 'polonius'
  edit REUSE_BUILD_DIR 'False'
  edit SAFE_STACK_SANITIZER 'False'
  edit SANITIZER 'False'
  edit SANITIZER_REFERENCE_DIRECTORY '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
  edit SRCDIR '%BASEDIR%/%PACKAGE%'
  edit STATIC_ANALYSER_ABORT_ON_ERROR 'False'
  edit SWIG_VERSION '3.0.8'
  edit SYNC_RELEASE '2020.04.0'
  edit SYNC_RELEASE_CI_STAGE 'develop'
  edit TARBALL_DIR '/perm/ma/ma0/tmp_tarballs/ci'
  edit TAR_PACKAGE '%PACKAGE%'
  edit TEST 'ctest --output-on-failure -j${CPUS} '
  edit THREAD_SANITIZER 'False'
  edit TRANSI_GIT_TAG 'develop'
  edit TRANSI_VERSION '0.7.1'
  edit TSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
  edit UBSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/ubsan.txt'
  edit UB_SANITIZER 'False'
  edit USER 'ma0'
  edit USE_WORKDIR 'False'
  edit VALGRIND_VERSION '3.15.0'
  edit WORKDIR '%BASEDIR%/workdir/%ECF_NAME%'
  edit WSDIR '/var/tmp/tmpdir/deploy/jtmp.20739'
  label branch "develop"
  task regenerate
    defstatus complete
    edit ECF_JOB_CMD '%ECF_JOB% > %ECF_JOBOUT% 2>&1'
    edit ECF_KILL_CMD 'kill -15 %ECF_RID%'
    edit ECF_PASS 'FREE'
    edit ECF_STATUS_CMD 'ps --sid %ECF_RID% -f'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
    label date ""
    label metabuilder_branch "ci"
  family cca
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DBLAS_LIBRARIES=/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_blas95_lp64.a -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECCODES_IFS_SAMPLES_PATH=ON -DENABLE_ECCODES=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_FDB_TOOLS=OFF -DENABLE_HERMES=OFF -DENABLE_MARS_SERVER=OFF -DENABLE_MEMFS=ON -DENABLE_MIR_TOOLS=ON -DENABLE_MPI=ON -DENABLE_ODB=OFF -DENABLE_OMP=OFF -DENABLE_PYTHON=OFF -DENABLE_SANDBOX=OFF -DENABLE_TESSELATION=OFF -DENABLE_TESTS=ON -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_FOUND=TRUE -DLAPACK_LIBRARIES=/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_lapack95_lp64.a -DMKL_ROOT=/opt/intel/composer_xe_2018.0.033/mkl'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/pgen/cca'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/pgen/cca:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cca-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export CRAY_ADD_RPATH=yes" "export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp" "export ftp_proxy=http://proxy.ecmwf.int:3333" "export http_proxy=http://proxy.ecmwf.int:3333" "export https_proxy=http://proxy.ecmwf.int:3333"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '14000'
    edit MKL '/opt/intel/composer_xe_2018.0.033/mkl'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%" "module load eigen" "module unload craype-broadwell" "module load craype-ivybridge"'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cca'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'cca'
    edit SITE 'cca'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DBLAS_LIBRARIES=/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_blas95_lp64.a -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DECCODES_IFS_SAMPLES_PATH=ON -DENABLE_ECCODES=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_FDB_TOOLS=OFF -DENABLE_HERMES=OFF -DENABLE_MARS_SERVER=OFF -DENABLE_MEMFS=ON -DENABLE_MIR_TOOLS=ON -DENABLE_MPI=ON -DENABLE_ODB=OFF -DENABLE_OMP=OFF -DENABLE_PYTHON=OFF -DENABLE_SANDBOX=OFF -DENABLE_TESSELATION=OFF -DENABLE_TESTS=ON -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_FOUND=TRUE -DLAPACK_LIBRARIES=/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_lapack95_lp64.a -DMKL_ROOT=/opt/intel/composer_xe_2018.0.033/mkl'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%" "module load eigen" "module unload craype-broadwell" "module load craype-ivybridge"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task install
        trigger build == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family ccb
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DBLAS_LIBRARIES=/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_blas95_lp64.a -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECCODES_IFS_SAMPLES_PATH=ON -DENABLE_ECCODES=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_FDB_TOOLS=OFF -DENABLE_HERMES=OFF -DENABLE_MARS_SERVER=OFF -DENABLE_MEMFS=ON -DENABLE_MIR_TOOLS=ON -DENABLE_MPI=ON -DENABLE_ODB=OFF -DENABLE_OMP=OFF -DENABLE_PYTHON=OFF -DENABLE_SANDBOX=OFF -DENABLE_TESSELATION=OFF -DENABLE_TESTS=ON -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_FOUND=TRUE -DLAPACK_LIBRARIES=/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_lapack95_lp64.a -DMKL_ROOT=/opt/intel/composer_xe_2018.0.033/mkl'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/pgen/ccb'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/pgen/ccb:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'ccb-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export CRAY_ADD_RPATH=yes" "export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp" "export ftp_proxy=http://proxy.ecmwf.int:3333" "export http_proxy=http://proxy.ecmwf.int:3333" "export https_proxy=http://proxy.ecmwf.int:3333"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '14000'
    edit MKL '/opt/intel/composer_xe_2018.0.033/mkl'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%" "module load eigen" "module unload craype-broadwell" "module load craype-ivybridge"'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'ccb'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'ccb'
    edit SITE 'ccb'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DBLAS_LIBRARIES=/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_blas95_lp64.a -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DECCODES_IFS_SAMPLES_PATH=ON -DENABLE_ECCODES=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_FDB_TOOLS=OFF -DENABLE_HERMES=OFF -DENABLE_MARS_SERVER=OFF -DENABLE_MEMFS=ON -DENABLE_MIR_TOOLS=ON -DENABLE_MPI=ON -DENABLE_ODB=OFF -DENABLE_OMP=OFF -DENABLE_PYTHON=OFF -DENABLE_SANDBOX=OFF -DENABLE_TESSELATION=OFF -DENABLE_TESTS=ON -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_FOUND=TRUE -DLAPACK_LIBRARIES=/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_lapack95_lp64.a -DMKL_ROOT=/opt/intel/composer_xe_2018.0.033/mkl'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%" "module load eigen" "module unload craype-broadwell" "module load craype-ivybridge"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task install
        trigger build == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family cct
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DBLAS_LIBRARIES=/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_blas95_lp64.a -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECCODES_IFS_SAMPLES_PATH=ON -DENABLE_ECCODES=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_FDB_TOOLS=OFF -DENABLE_HERMES=OFF -DENABLE_MARS_SERVER=OFF -DENABLE_MEMFS=ON -DENABLE_MIR_TOOLS=ON -DENABLE_MPI=ON -DENABLE_ODB=OFF -DENABLE_OMP=OFF -DENABLE_PYTHON=OFF -DENABLE_SANDBOX=OFF -DENABLE_TESSELATION=OFF -DENABLE_TESTS=ON -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_FOUND=TRUE -DLAPACK_LIBRARIES=/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_lapack95_lp64.a -DMKL_ROOT=/opt/intel/composer_xe_2018.0.033/mkl'
    edit CPUS '$EC_total_tasks'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/pgen/cct'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/pgen/cct:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cct-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '4'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp" "export ftp_proxy=http://proxy.ecmwf.int:3333" "export http_proxy=http://proxy.ecmwf.int:3333" "export https_proxy=http://proxy.ecmwf.int:3333"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '6000'
    edit MKL '/opt/intel/composer_xe_2018.0.033/mkl'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%" "module load eigen" "module unload craype-broadwell" "module load craype-ivybridge"'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cct'
    edit QSUB_LP 'QSUB -lP %EC_TOTAL_TASKS%'
    edit QUEUE 'nf'
    edit REMOTE_HOST 'cct'
    edit SITE 'cct'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DBLAS_LIBRARIES=/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_blas95_lp64.a -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DECCODES_IFS_SAMPLES_PATH=ON -DENABLE_ECCODES=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_FDB_TOOLS=OFF -DENABLE_HERMES=OFF -DENABLE_MARS_SERVER=OFF -DENABLE_MEMFS=ON -DENABLE_MIR_TOOLS=ON -DENABLE_MPI=ON -DENABLE_ODB=OFF -DENABLE_OMP=OFF -DENABLE_PYTHON=OFF -DENABLE_SANDBOX=OFF -DENABLE_TESSELATION=OFF -DENABLE_TESTS=ON -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_FOUND=TRUE -DLAPACK_LIBRARIES=/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_lapack95_lp64.a -DMKL_ROOT=/opt/intel/composer_xe_2018.0.033/mkl'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%" "module load eigen" "module unload craype-broadwell" "module load craype-ivybridge"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task install
        trigger build == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family leap42
    defstatus suspended
    edit CI_LIMIT '7'
    edit CMAKE_VARS '-DBLAS_LIBRARIES=/usr/local/apps/intel/18.0.1/mkl/lib/intel64/libmkl_blas95_lp64.a -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DECCODES_IFS_SAMPLES_PATH=ON -DENABLE_ECCODES=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_FDB_TOOLS=OFF -DENABLE_HERMES=OFF -DENABLE_MARS_SERVER=OFF -DENABLE_MEMFS=ON -DENABLE_MIR_TOOLS=ON -DENABLE_MPI=ON -DENABLE_ODB=OFF -DENABLE_OMP=OFF -DENABLE_PYTHON=OFF -DENABLE_SANDBOX=OFF -DENABLE_TESSELATION=OFF -DENABLE_TESTS=ON -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_LIBRARIES=/usr/local/apps/intel/18.0.1/mkl/lib/intel64/libmkl_lapack95_lp64.a -DMKL_ROOT=/usr/local/apps/intel/18.0.1/mkl'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit MKL '/usr/local/apps/intel/18.0.1/mkl'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%" "module load openmpi"'
    edit M_SITE 'leap42'
    edit REMOTE_HOST 'leap42-deploy'
    edit SITE 'leap42'
    edit WSDIR '/tmp/metabuilds/%ECF_HOST%_%ECF_PORT%/leap42/%COMPILER:%'
    family gnu.63
      edit CMAKE_VARS '-DBLAS_LIBRARIES=/usr/local/apps/intel/18.0.1/mkl/lib/intel64/libmkl_blas95_lp64.a -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECCODES_IFS_SAMPLES_PATH=ON -DENABLE_ECCODES=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_FDB_TOOLS=OFF -DENABLE_HERMES=OFF -DENABLE_MARS_SERVER=OFF -DENABLE_MEMFS=ON -DENABLE_MIR_TOOLS=ON -DENABLE_MPI=ON -DENABLE_ODB=OFF -DENABLE_OMP=OFF -DENABLE_PYTHON=OFF -DENABLE_SANDBOX=OFF -DENABLE_TESSELATION=OFF -DENABLE_TESTS=ON -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_LIBRARIES=/usr/local/apps/intel/18.0.1/mkl/lib/intel64/libmkl_lapack95_lp64.a -DMKL_ROOT=/usr/local/apps/intel/18.0.1/mkl'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap gnu/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%" "module load openmpi"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task install
        trigger build == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family lxc
    defstatus suspended
    edit CI_LIMIT '10'
    edit CMAKE_VARS '-DBLAS_LIBRARIES=/usr/local/apps/intel/17.0.3/mkl/lib/intel64/libmkl_blas95_lp64.a -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DECCODES_IFS_SAMPLES_PATH=ON -DENABLE_ECCODES=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_FDB_TOOLS=OFF -DENABLE_HERMES=OFF -DENABLE_MARS_SERVER=OFF -DENABLE_MEMFS=ON -DENABLE_MIR_TOOLS=ON -DENABLE_MPI=ON -DENABLE_ODB=OFF -DENABLE_OMP=OFF -DENABLE_PYTHON=OFF -DENABLE_SANDBOX=OFF -DENABLE_TESSELATION=OFF -DENABLE_TESTS=ON -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_LIBRARIES=/usr/local/apps/intel/17.0.3/mkl/lib/intel64/libmkl_lapack95_lp64.a -DMKL_ROOT=/usr/local/apps/intel/17.0.3/mkl'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit MKL '/usr/local/apps/intel/17.0.3/mkl'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%" "module load openmpi" "module load lapack"'
    edit M_SITE 'lxc'
    edit QUEUE 'express'
    edit REMOTE_HOST 'lxc'
    edit SITE 'lxc'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxc/%COMPILER:%'
    family gnu.63
      edit CMAKE_VARS '-DBLAS_LIBRARIES=/usr/local/apps/intel/17.0.3/mkl/lib/intel64/libmkl_blas95_lp64.a -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DECCODES_IFS_SAMPLES_PATH=ON -DENABLE_ECCODES=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_EXPERIMENTAL=OFF -DENABLE_FDB_TOOLS=OFF -DENABLE_HERMES=OFF -DENABLE_MARS_SERVER=OFF -DENABLE_MEMFS=ON -DENABLE_MIR_TOOLS=ON -DENABLE_MPI=ON -DENABLE_ODB=OFF -DENABLE_OMP=OFF -DENABLE_PYTHON=OFF -DENABLE_SANDBOX=OFF -DENABLE_TESSELATION=OFF -DENABLE_TESTS=ON -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_LIBRARIES=/usr/local/apps/intel/17.0.3/mkl/lib/intel64/libmkl_lapack95_lp64.a -DMKL_ROOT=/usr/local/apps/intel/17.0.3/mkl'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap gnu/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load fftw/%FFTW_VERSION%" "module load ecflow/%ECF_VERSION%" "module load openmpi" "module load lapack"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task install
        trigger build == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
endsuite
suite scalability
  edit ADDRESS_SANITIZER 'False'
  edit ADD_ACTION_TRIGGERS 'True'
  edit AEC_VERSION '1.0.4'
  edit ALLOW_INSTALL_TO_OVERWRITE 'True'
  edit ASAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
  edit ATLAS_GIT_TAG 'develop'
  edit ATLAS_VERSION '0.20.2'
  edit BANNER_TASK_END 'banner %TASK%-end || true'
  edit BANNER_TASK_START 'banner %TASK%-start || true'
  edit BASEDIR '%WSDIR%/%PACKAGE%'
  edit BASEPREFIX '$PERM/install/%SITE%'
  edit BUFRDC_EMOSLIB_VERSION '4.6.3'
  edit BUFRDC_TABLES_GIT_TAG 'develop'
  edit BUFRDC_TABLES_VERSION '000415'
  edit BUILDIR '%BASEDIR%/builds'
  edit BUILD_ALWAYS_PARALLEL 'False'
  edit CFI_SANITIZER 'False'
  edit CFL_BASE_URL 'https://software-test.ecmwf.int/wiki'
  edit CFL_SPACE_KEY '<TODO>'
  edit CFL_TARBALL_COMMENT ''
  edit CFL_TITLE 'Releases'
  edit CHECK_SYNC_RELEASE 'False'
  edit CLANG_TIDY_ABORT_ON_ERROR 'False'
  edit CMAKE '%ECBUILD%'
  edit CMAKELIST_EXTRACT_PACKAGE_VERSION '$(awk '/^project/ && /%PACKAGE%/ && /VERSION/ {for (I=1;I<=NF;I++) if ($I == "VERSION") {print $(I+1)};}' CMakeLists.txt)'
  edit CMAKE_BUILD_TYPE 'RelWithDebInfo'
  edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DENABLE_AEC=OFF -DENABLE_DUMMY_TAPES=ON -DENABLE_ECCODES=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_ECMWF_DEVELOPMENT=ON -DENABLE_EXPERIMENTAL=ON -DENABLE_FDB=ON -DENABLE_FDB5=ON -DENABLE_FORTRAN=ON -DENABLE_HERMES=OFF -DENABLE_JPG=OFF -DENABLE_MEMFS=ON -DENABLE_MULTIO_SERVER=ON -DENABLE_NETCDF=OFF -DENABLE_ODB=OFF -DENABLE_ODBSVR=OFF -DENABLE_OMP=OFF -DENABLE_PNG=OFF -DENABLE_PYTHON=OFF -DENABLE_SANDBOX=OFF -DENABLE_TESSELATION=OFF -DENABLE_TESTS=ON -DENABLE_WARNINGS=ON -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config'
  edit CMAKE_VERSION '3.13.4'
  edit CODES_UI_GIT_TAG 'develop'
  edit CODES_UI_VERSION '1.3.1'
  edit CPACK 'make package'
  edit CPUS '$(lscpu -p | grep -v '#' | wc -l)'
  edit CRAY_FFTW_VERSION '3.3.8.1'
  edit CRAY_NETCDF_VERSION '4.6.1.3'
  edit CRON_BASEPREFIX '%BASEDIR%/install'
  edit CRON_EMAIL ''
  edit CRON_ENABLED 'False'
  edit CRON_GIT_TAG 'develop'
  edit CUSTOM_COMMANDS ''
  edit DATAFLOW_SANITIZER 'False'
  edit DEFAULT_GNU_COMPILER 'gnu.73'
  edit ECBUILD '%BASEPREFIX%/ecbuild/develop/bin/ecbuild'
  edit ECBUILD_CLONE 'False'
  edit ECBUILD_CLONE_COMMIT 'master'
  edit ECBUILD_GIT_TAG 'develop'
  edit ECBUILD_REPO '%GIT_REPO_PREFIX%/ecsdk/ecbuild.git'
  edit ECBUILD_VERSION '3.3.2'
  edit ECCODES_EXTENSIONS_GIT_TAG 'develop'
  edit ECCODES_EXTENSIONS_VERSION '2.17.0'
  edit ECCODES_GIT_TAG 'develop'
  edit ECCODES_PYTHON_VERSION '0.9.6'
  edit ECCODES_VERSION '2.17.1'
  edit ECFLOW5_LATEST_VERSION '5.4.0'
  edit ECFLOW_VERSION '5.4.1'
  edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/scalability'
  edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/scalability'
  edit ECF_JOB_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_JOB% %ECF_JOBOUT%'
  edit ECF_KILL_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% kill'
  edit ECF_STATUS_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% stat'
  edit ECF_TRIES '1'
  edit ECKIT_GIT_TAG 'develop'
  edit ECKIT_VERSION '1.10.1'
  edit ECQUIS_VERSION '1.0.0'
  edit EMOS_MOD_VERSION '462'
  edit EMOS_PA_VERSION '000462'
  edit EXPORTS '"export ECACCOUNT=ecodmdma" "export ftp_proxy=http://proxy.ecmwf.int:3333" "export http_proxy=http://proxy.ecmwf.int:3333" "export https_proxy=http://proxy.ecmwf.int:3333"'
  edit EXPORT_VARIABLES 'False'
  edit FCKIT_GIT_TAG '0.6.3'
  edit FCKIT_VERSION '0.6.3'
  edit FDB_VINTAGE_GIT_TAG '4.10.0'
  edit FDB_VINTAGE_VERSION '4.10.0'
  edit FFTW_VERSION '3.3.8'
  edit FFTW_VERSION_CRAY '3.3.4.11'
  edit GIT 'git'
  edit GITHUB_PACKAGE_NAME '%PACKAGE_NAME%'
  edit GITHUB_REPO 'git@github.com:ecmwf'
  edit GITREPO '%GIT_REPO_PREFIX%/ecsdk/scalability-bundle.git'
  edit GIT_CLONE 'True'
  edit GIT_CLONE_OPTIONS '--depth=1 --branch %GIT_TAG%'
  edit GIT_METABUILDER_TAG 'ci'
  edit GIT_REPO_PREFIX 'ssh://git@git.ecmwf.int:7999'
  edit GIT_TAG 'develop'
  edit GIT_VERSION '2.26.0'
  edit GRIB_API_VERSION '1.28.0'
  edit INSTALLABLE_TO_USR_LOCAL_APPS 'True'
  edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
  edit IWYU_VERSION '0.11'
  edit KSH_EXTRACT_PACKAGE_VERSION '`grep _VERSION_STR VERSION.cmake | grep -o '[0-9]*\.[0-9]*\.[0-9]*'`'
  edit LAPACK_VERSION '3.5.0'
  edit LIBEMOS_GIT_TAG 'develop'
  edit LIBEMOS_VERSION '4.6.2'
  edit LINTERS_REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}'
  edit LSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/lsan.txt'
  edit MAGICS_GIT_TAG 'develop'
  edit MAGICS_PYTHON_VERSION 'master'
  edit MAGICS_VERSION '4.3.3'
  edit MAKE 'make'
  edit MARS_CLIENT_GIT_TAG '6.27.2.2'
  edit MARS_CLIENT_PACKAGE_GIT_TAG '6.27.6'
  edit MARS_CLIENT_PACKAGE_VERSION '6.27.6'
  edit MARS_CLIENT_VERSION '6.27.2.2'
  edit MEMORY_SANITIZER 'False'
  edit METABUILDER_HOME '/var/tmp/ma0/workspace/metabuilder'
  edit METAB_INSTALL_REMOVE_WRITE_PERMISSIONS 'True'
  edit METAB_PYTHON_VERSION 'python3'
  edit METKIT_GIT_TAG 'develop'
  edit METKIT_VERSION '1.4.6'
  edit METVIEW_GIT_TAG 'develop'
  edit METVIEW_PYTHON_VERSION '1.4.2'
  edit METVIEW_VERSION '5.8.3'
  edit MIR_BUNDLE_VERSION '0.9.0.0'
  edit MIR_GIT_TAG 'develop'
  edit MIR_VERSION '1.5.0'
  edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module load fftw/%FFTW_VERSION%" "module load boost" "module load eigen"'
  edit MODULE_ACTIONS_FIRST ''
  edit MODULE_ACTIONS_HDR 'module_actions.h'
  edit MODULE_CHECKING_HDR 'module_checking.h'
  edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow"'
  edit MODULE_LOAD_SUPPORTED 'True'
  edit MODULE_NAME 'scalability'
  edit MODULE_VERSION '%GIT_TAG%'
  edit NETCDF4_VERSION '4.7.4'
  edit ODB_API_GIT_TAG 'develop'
  edit ODB_API_VERSION '0.19.4'
  edit ODB_GIT_TAG 'develop'
  edit ODB_TOOLS_GIT_TAG 'develop'
  edit ODB_TOOLS_VERSION '0.18.3'
  edit ODB_VERSION '1.0.11'
  edit ODC_GIT_TAG 'develop'
  edit ODC_VERSION '1.0.3'
  edit OPENJPEG_VERSION '2.3.1'
  edit PACKAGE 'scalability'
  edit PACKAGE_NAME 'scalability'
  edit PACKAGE_OWNERS 'ma0@ecmwf.int'
  edit PACKAGE_VERSION ''
  edit PGEN_VERSION '1.7.1.0'
  edit PKG_BOOST_VERSION '1.71.0'
  edit PREFIX_APPEND ''
  edit PROJ4_VERSION '5.2.0'
  edit PROJ_VERSION '6.3.1'
  edit PYTHON2_VERSION '2.7.15-01'
  edit PYTHON3_VERSION '3.6.10-01'
  edit PYTHON3_VERSION_MAJOR_MINOR '3.6'
  edit PYTHON_GITREPO 'This-needs-tobe-overriden'
  edit PYTHON_PACKAGE 'This-needs-tobe-overriden'
  edit QT_VERSION '5.12.3'
  edit QUEUE 'build'
  edit REMOTE_HOST 'polonius'
  edit REUSE_BUILD_DIR 'False'
  edit SAFE_STACK_SANITIZER 'False'
  edit SANITIZER 'False'
  edit SANITIZER_REFERENCE_DIRECTORY '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
  edit SRCDIR '%BASEDIR%/%PACKAGE%'
  edit STATIC_ANALYSER_ABORT_ON_ERROR 'False'
  edit SWIG_VERSION '3.0.8'
  edit SYNC_RELEASE '2020.04.0'
  edit SYNC_RELEASE_CI_STAGE 'develop'
  edit TARBALL_DIR '/perm/ma/ma0/tmp_tarballs/ci'
  edit TAR_PACKAGE '%PACKAGE%'
  edit TEST 'ctest --output-on-failure -j${CPUS} '
  edit THREAD_SANITIZER 'False'
  edit TRANSI_GIT_TAG 'develop'
  edit TRANSI_VERSION '0.7.1'
  edit TSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
  edit UBSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/ubsan.txt'
  edit UB_SANITIZER 'False'
  edit USER 'ma0'
  edit USE_WORKDIR 'False'
  edit VALGRIND_VERSION '3.15.0'
  edit WORKDIR '%BASEDIR%/workdir/%ECF_NAME%'
  edit WSDIR '/var/tmp/tmpdir/deploy/jtmp.20739'
  label branch "develop"
  task regenerate
    defstatus complete
    edit ECF_JOB_CMD '%ECF_JOB% > %ECF_JOBOUT% 2>&1'
    edit ECF_KILL_CMD 'kill -15 %ECF_RID%'
    edit ECF_PASS 'FREE'
    edit ECF_STATUS_CMD 'ps --sid %ECF_RID% -f'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
    label date ""
    label metabuilder_branch "ci"
  family cca
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DBLAS_LIBRARIES=/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_blas95_lp64.a -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DENABLE_AEC=OFF -DENABLE_DUMMY_TAPES=ON -DENABLE_ECCODES=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_ECMWF_DEVELOPMENT=ON -DENABLE_EIGEN=OFF -DENABLE_EXPERIMENTAL=ON -DENABLE_FDB=ON -DENABLE_FDB5=ON -DENABLE_FORTRAN=ON -DENABLE_HERMES=OFF -DENABLE_JPG=OFF -DENABLE_MEMFS=ON -DENABLE_MULTIO_SERVER=ON -DENABLE_NETCDF=OFF -DENABLE_ODB=OFF -DENABLE_ODBSVR=OFF -DENABLE_OMP=OFF -DENABLE_PNG=OFF -DENABLE_PYTHON=OFF -DENABLE_SANDBOX=OFF -DENABLE_TESSELATION=OFF -DENABLE_TESTS=ON -DENABLE_WARNINGS=ON -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_FOUND=TRUE -DLAPACK_LIBRARIES=/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_lapack95_lp64.a -DMKL_ROOT=/opt/intel/composer_xe_2018.0.033/mkl'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/scalability/cca'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/scalability/cca:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cca-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export CRAY_ADD_RPATH=yes" "export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp" "export ftp_proxy=http://proxy.ecmwf.int:3333" "export http_proxy=http://proxy.ecmwf.int:3333" "export https_proxy=http://proxy.ecmwf.int:3333"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '43000'
    edit MKL '/opt/intel/composer_xe_2018.0.033/mkl'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module load fftw/%FFTW_VERSION%" "module load boost" "module load eigen" "module unload craype-broadwell" "module load craype-ivybridge"'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cca'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'cca'
    edit SITE 'cca'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DBLAS_LIBRARIES=/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_blas95_lp64.a -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DENABLE_AEC=OFF -DENABLE_DUMMY_TAPES=ON -DENABLE_ECCODES=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_ECMWF_DEVELOPMENT=ON -DENABLE_EIGEN=OFF -DENABLE_EXPERIMENTAL=ON -DENABLE_FDB=ON -DENABLE_FDB5=ON -DENABLE_FORTRAN=ON -DENABLE_HERMES=OFF -DENABLE_JPG=OFF -DENABLE_MEMFS=ON -DENABLE_MULTIO_SERVER=ON -DENABLE_NETCDF=OFF -DENABLE_ODB=OFF -DENABLE_ODBSVR=OFF -DENABLE_OMP=OFF -DENABLE_PNG=OFF -DENABLE_PYTHON=OFF -DENABLE_SANDBOX=OFF -DENABLE_TESSELATION=OFF -DENABLE_TESTS=ON -DENABLE_WARNINGS=ON -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_FOUND=TRUE -DLAPACK_LIBRARIES=/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_lapack95_lp64.a -DMKL_ROOT=/opt/intel/composer_xe_2018.0.033/mkl'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module load fftw/%FFTW_VERSION%" "module load boost" "module load eigen" "module unload craype-broadwell" "module load craype-ivybridge"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DBLAS_LIBRARIES=/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_blas95_lp64.a -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DENABLE_AEC=OFF -DENABLE_DUMMY_TAPES=ON -DENABLE_ECCODES=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_ECMWF_DEVELOPMENT=ON -DENABLE_EIGEN=OFF -DENABLE_EXPERIMENTAL=ON -DENABLE_FDB=ON -DENABLE_FDB5=ON -DENABLE_FORTRAN=ON -DENABLE_HERMES=OFF -DENABLE_JPG=OFF -DENABLE_MEMFS=ON -DENABLE_MULTIO_SERVER=ON -DENABLE_NETCDF=OFF -DENABLE_ODB=OFF -DENABLE_ODBSVR=OFF -DENABLE_OMP=OFF -DENABLE_PNG=OFF -DENABLE_PYTHON=OFF -DENABLE_SANDBOX=OFF -DENABLE_TESSELATION=OFF -DENABLE_TESTS=ON -DENABLE_WARNINGS=ON -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_FOUND=TRUE -DLAPACK_LIBRARIES=/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_lapack95_lp64.a -DMKL_ROOT=/opt/intel/composer_xe_2018.0.033/mkl'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module load fftw/%FFTW_VERSION%" "module load boost" "module load eigen" "module unload craype-broadwell" "module load craype-ivybridge"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit COMPILER 'CRAY.87'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module load fftw/%FFTW_VERSION%" "module load boost" "module load eigen" "module unload craype-broadwell" "module load craype-ivybridge"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.18
      edit CMAKE_VARS '-DBLAS_LIBRARIES=/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_blas95_lp64.a -DBUNDLE_SKIP_TRANSI=ON -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DENABLE_AEC=OFF -DENABLE_DUMMY_TAPES=ON -DENABLE_ECCODES=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_ECMWF_DEVELOPMENT=ON -DENABLE_EIGEN=OFF -DENABLE_EXPERIMENTAL=ON -DENABLE_FDB=ON -DENABLE_FDB5=ON -DENABLE_FORTRAN=ON -DENABLE_HERMES=OFF -DENABLE_JPG=OFF -DENABLE_MEMFS=ON -DENABLE_MULTIO_SERVER=ON -DENABLE_NETCDF=OFF -DENABLE_ODB=OFF -DENABLE_ODBSVR=OFF -DENABLE_OMP=OFF -DENABLE_PNG=OFF -DENABLE_PYTHON=OFF -DENABLE_SANDBOX=OFF -DENABLE_TESSELATION=OFF -DENABLE_TESTS=ON -DENABLE_TRANS=OFF -DENABLE_WARNINGS=ON -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_FOUND=TRUE -DLAPACK_LIBRARIES=/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_lapack95_lp64.a -DMKL_ROOT=/opt/intel/composer_xe_2018.0.033/mkl'
      edit COMPILER 'INTEL.18'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap intel/18.0.0.033" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module load fftw/%FFTW_VERSION%" "module load boost" "module load eigen" "module unload craype-broadwell" "module load craype-ivybridge"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family ccb
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DBLAS_LIBRARIES=/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_blas95_lp64.a -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DENABLE_AEC=OFF -DENABLE_DUMMY_TAPES=ON -DENABLE_ECCODES=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_ECMWF_DEVELOPMENT=ON -DENABLE_EIGEN=OFF -DENABLE_EXPERIMENTAL=ON -DENABLE_FDB=ON -DENABLE_FDB5=ON -DENABLE_FORTRAN=ON -DENABLE_HERMES=OFF -DENABLE_JPG=OFF -DENABLE_MEMFS=ON -DENABLE_MULTIO_SERVER=ON -DENABLE_NETCDF=OFF -DENABLE_ODB=OFF -DENABLE_ODBSVR=OFF -DENABLE_OMP=OFF -DENABLE_PNG=OFF -DENABLE_PYTHON=OFF -DENABLE_SANDBOX=OFF -DENABLE_TESSELATION=OFF -DENABLE_TESTS=ON -DENABLE_WARNINGS=ON -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_FOUND=TRUE -DLAPACK_LIBRARIES=/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_lapack95_lp64.a -DMKL_ROOT=/opt/intel/composer_xe_2018.0.033/mkl'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/scalability/ccb'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/scalability/ccb:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'ccb-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export CRAY_ADD_RPATH=yes" "export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp" "export ftp_proxy=http://proxy.ecmwf.int:3333" "export http_proxy=http://proxy.ecmwf.int:3333" "export https_proxy=http://proxy.ecmwf.int:3333"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '43000'
    edit MKL '/opt/intel/composer_xe_2018.0.033/mkl'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module load fftw/%FFTW_VERSION%" "module load boost" "module load eigen" "module unload craype-broadwell" "module load craype-ivybridge"'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'ccb'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'ccb'
    edit SITE 'ccb'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DBLAS_LIBRARIES=/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_blas95_lp64.a -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DENABLE_AEC=OFF -DENABLE_DUMMY_TAPES=ON -DENABLE_ECCODES=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_ECMWF_DEVELOPMENT=ON -DENABLE_EIGEN=OFF -DENABLE_EXPERIMENTAL=ON -DENABLE_FDB=ON -DENABLE_FDB5=ON -DENABLE_FORTRAN=ON -DENABLE_HERMES=OFF -DENABLE_JPG=OFF -DENABLE_MEMFS=ON -DENABLE_MULTIO_SERVER=ON -DENABLE_NETCDF=OFF -DENABLE_ODB=OFF -DENABLE_ODBSVR=OFF -DENABLE_OMP=OFF -DENABLE_PNG=OFF -DENABLE_PYTHON=OFF -DENABLE_SANDBOX=OFF -DENABLE_TESSELATION=OFF -DENABLE_TESTS=ON -DENABLE_WARNINGS=ON -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_FOUND=TRUE -DLAPACK_LIBRARIES=/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_lapack95_lp64.a -DMKL_ROOT=/opt/intel/composer_xe_2018.0.033/mkl'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module load fftw/%FFTW_VERSION%" "module load boost" "module load eigen" "module unload craype-broadwell" "module load craype-ivybridge"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DBLAS_LIBRARIES=/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_blas95_lp64.a -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DENABLE_AEC=OFF -DENABLE_DUMMY_TAPES=ON -DENABLE_ECCODES=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_ECMWF_DEVELOPMENT=ON -DENABLE_EIGEN=OFF -DENABLE_EXPERIMENTAL=ON -DENABLE_FDB=ON -DENABLE_FDB5=ON -DENABLE_FORTRAN=ON -DENABLE_HERMES=OFF -DENABLE_JPG=OFF -DENABLE_MEMFS=ON -DENABLE_MULTIO_SERVER=ON -DENABLE_NETCDF=OFF -DENABLE_ODB=OFF -DENABLE_ODBSVR=OFF -DENABLE_OMP=OFF -DENABLE_PNG=OFF -DENABLE_PYTHON=OFF -DENABLE_SANDBOX=OFF -DENABLE_TESSELATION=OFF -DENABLE_TESTS=ON -DENABLE_WARNINGS=ON -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_FOUND=TRUE -DLAPACK_LIBRARIES=/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_lapack95_lp64.a -DMKL_ROOT=/opt/intel/composer_xe_2018.0.033/mkl'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module load fftw/%FFTW_VERSION%" "module load boost" "module load eigen" "module unload craype-broadwell" "module load craype-ivybridge"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit COMPILER 'CRAY.87'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module load fftw/%FFTW_VERSION%" "module load boost" "module load eigen" "module unload craype-broadwell" "module load craype-ivybridge"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.18
      edit CMAKE_VARS '-DBLAS_LIBRARIES=/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_blas95_lp64.a -DBUNDLE_SKIP_TRANSI=ON -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DENABLE_AEC=OFF -DENABLE_DUMMY_TAPES=ON -DENABLE_ECCODES=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_ECMWF_DEVELOPMENT=ON -DENABLE_EIGEN=OFF -DENABLE_EXPERIMENTAL=ON -DENABLE_FDB=ON -DENABLE_FDB5=ON -DENABLE_FORTRAN=ON -DENABLE_HERMES=OFF -DENABLE_JPG=OFF -DENABLE_MEMFS=ON -DENABLE_MULTIO_SERVER=ON -DENABLE_NETCDF=OFF -DENABLE_ODB=OFF -DENABLE_ODBSVR=OFF -DENABLE_OMP=OFF -DENABLE_PNG=OFF -DENABLE_PYTHON=OFF -DENABLE_SANDBOX=OFF -DENABLE_TESSELATION=OFF -DENABLE_TESTS=ON -DENABLE_TRANS=OFF -DENABLE_WARNINGS=ON -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_FOUND=TRUE -DLAPACK_LIBRARIES=/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_lapack95_lp64.a -DMKL_ROOT=/opt/intel/composer_xe_2018.0.033/mkl'
      edit COMPILER 'INTEL.18'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap intel/18.0.0.033" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module load fftw/%FFTW_VERSION%" "module load boost" "module load eigen" "module unload craype-broadwell" "module load craype-ivybridge"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family cct
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DBLAS_LIBRARIES=/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_blas95_lp64.a -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DENABLE_AEC=OFF -DENABLE_DUMMY_TAPES=ON -DENABLE_ECCODES=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_ECMWF_DEVELOPMENT=ON -DENABLE_EIGEN=OFF -DENABLE_EXPERIMENTAL=ON -DENABLE_FDB=ON -DENABLE_FDB5=ON -DENABLE_FORTRAN=ON -DENABLE_HERMES=OFF -DENABLE_JPG=OFF -DENABLE_MEMFS=ON -DENABLE_MULTIO_SERVER=ON -DENABLE_NETCDF=OFF -DENABLE_ODB=OFF -DENABLE_ODBSVR=OFF -DENABLE_OMP=OFF -DENABLE_PNG=OFF -DENABLE_PYTHON=OFF -DENABLE_SANDBOX=OFF -DENABLE_TESSELATION=OFF -DENABLE_TESTS=ON -DENABLE_WARNINGS=ON -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_FOUND=TRUE -DLAPACK_LIBRARIES=/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_lapack95_lp64.a -DMKL_ROOT=/opt/intel/composer_xe_2018.0.033/mkl'
    edit CPUS '$EC_total_tasks'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/scalability/cct'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/scalability/cct:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cct-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '4'
    edit EXPORTS '"export CRAY_ADD_RPATH=yes" "export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp" "export ftp_proxy=http://proxy.ecmwf.int:3333" "export http_proxy=http://proxy.ecmwf.int:3333" "export https_proxy=http://proxy.ecmwf.int:3333"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '10750'
    edit MKL '/opt/intel/composer_xe_2018.0.033/mkl'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module load fftw/%FFTW_VERSION%" "module load boost" "module load eigen" "module unload craype-broadwell" "module load craype-ivybridge"'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cct'
    edit QSUB_LP 'QSUB -lP %EC_TOTAL_TASKS%'
    edit QUEUE 'nf'
    edit REMOTE_HOST 'cct'
    edit SITE 'cct'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.63
      edit CMAKE_VARS '-DBLAS_LIBRARIES=/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_blas95_lp64.a -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DENABLE_AEC=OFF -DENABLE_DUMMY_TAPES=ON -DENABLE_ECCODES=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_ECMWF_DEVELOPMENT=ON -DENABLE_EIGEN=OFF -DENABLE_EXPERIMENTAL=ON -DENABLE_FDB=ON -DENABLE_FDB5=ON -DENABLE_FORTRAN=ON -DENABLE_HERMES=OFF -DENABLE_JPG=OFF -DENABLE_MEMFS=ON -DENABLE_MULTIO_SERVER=ON -DENABLE_NETCDF=OFF -DENABLE_ODB=OFF -DENABLE_ODBSVR=OFF -DENABLE_OMP=OFF -DENABLE_PNG=OFF -DENABLE_PYTHON=OFF -DENABLE_SANDBOX=OFF -DENABLE_TESSELATION=OFF -DENABLE_TESTS=ON -DENABLE_WARNINGS=ON -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_FOUND=TRUE -DLAPACK_LIBRARIES=/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_lapack95_lp64.a -DMKL_ROOT=/opt/intel/composer_xe_2018.0.033/mkl'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module load fftw/%FFTW_VERSION%" "module load boost" "module load eigen" "module unload craype-broadwell" "module load craype-ivybridge"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.73
      edit CMAKE_VARS '-DBLAS_LIBRARIES=/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_blas95_lp64.a -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DENABLE_AEC=OFF -DENABLE_DUMMY_TAPES=ON -DENABLE_ECCODES=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_ECMWF_DEVELOPMENT=ON -DENABLE_EIGEN=OFF -DENABLE_EXPERIMENTAL=ON -DENABLE_FDB=ON -DENABLE_FDB5=ON -DENABLE_FORTRAN=ON -DENABLE_HERMES=OFF -DENABLE_JPG=OFF -DENABLE_MEMFS=ON -DENABLE_MULTIO_SERVER=ON -DENABLE_NETCDF=OFF -DENABLE_ODB=OFF -DENABLE_ODBSVR=OFF -DENABLE_OMP=OFF -DENABLE_PNG=OFF -DENABLE_PYTHON=OFF -DENABLE_SANDBOX=OFF -DENABLE_TESSELATION=OFF -DENABLE_TESTS=ON -DENABLE_WARNINGS=ON -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_FOUND=TRUE -DLAPACK_LIBRARIES=/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_lapack95_lp64.a -DMKL_ROOT=/opt/intel/composer_xe_2018.0.033/mkl'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module load fftw/%FFTW_VERSION%" "module load boost" "module load eigen" "module unload craype-broadwell" "module load craype-ivybridge"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit COMPILER 'CRAY.87'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module load fftw/%FFTW_VERSION%" "module load boost" "module load eigen" "module unload craype-broadwell" "module load craype-ivybridge"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.18
      edit CMAKE_VARS '-DBLAS_LIBRARIES=/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_blas95_lp64.a -DBUNDLE_SKIP_TRANSI=ON -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DENABLE_AEC=OFF -DENABLE_DUMMY_TAPES=ON -DENABLE_ECCODES=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_ECMWF_DEVELOPMENT=ON -DENABLE_EIGEN=OFF -DENABLE_EXPERIMENTAL=ON -DENABLE_FDB=ON -DENABLE_FDB5=ON -DENABLE_FORTRAN=ON -DENABLE_HERMES=OFF -DENABLE_JPG=OFF -DENABLE_MEMFS=ON -DENABLE_MULTIO_SERVER=ON -DENABLE_NETCDF=OFF -DENABLE_ODB=OFF -DENABLE_ODBSVR=OFF -DENABLE_OMP=OFF -DENABLE_PNG=OFF -DENABLE_PYTHON=OFF -DENABLE_SANDBOX=OFF -DENABLE_TESSELATION=OFF -DENABLE_TESTS=ON -DENABLE_TRANS=OFF -DENABLE_WARNINGS=ON -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_FOUND=TRUE -DLAPACK_LIBRARIES=/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_lapack95_lp64.a -DMKL_ROOT=/opt/intel/composer_xe_2018.0.033/mkl'
      edit COMPILER 'INTEL.18'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap intel/18.0.0.033" "module swap gcc/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module load fftw/%FFTW_VERSION%" "module load boost" "module load eigen" "module unload craype-broadwell" "module load craype-ivybridge"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-intel"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family ecgb
    defstatus suspended
    edit CI_LIMIT '10'
    edit CMAKE_VARS '-DBLAS_LIBRARIES=/usr/local/apps/intel/17.0.3/mkl/lib/intel64/libmkl_blas95_lp64.a -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DENABLE_AEC=OFF -DENABLE_DUMMY_TAPES=ON -DENABLE_ECCODES=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_ECMWF_DEVELOPMENT=ON -DENABLE_EXPERIMENTAL=ON -DENABLE_FDB=ON -DENABLE_FDB5=ON -DENABLE_FORTRAN=ON -DENABLE_HERMES=OFF -DENABLE_JPG=OFF -DENABLE_MEMFS=ON -DENABLE_MULTIO_SERVER=ON -DENABLE_NETCDF=OFF -DENABLE_ODB=OFF -DENABLE_ODBSVR=OFF -DENABLE_OMP=OFF -DENABLE_PNG=OFF -DENABLE_PYTHON=OFF -DENABLE_SANDBOX=OFF -DENABLE_TESSELATION=OFF -DENABLE_TESTS=ON -DENABLE_WARNINGS=ON -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_LIBRARIES=/usr/local/apps/intel/17.0.3/mkl/lib/intel64/libmkl_lapack95_lp64.a -DMKL_ROOT=/usr/local/apps/intel/17.0.3/mkl'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit MKL '/usr/local/apps/intel/17.0.3/mkl'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module load fftw/%FFTW_VERSION%" "module load boost" "module load eigen" "module load openmpi"'
    edit M_SITE 'ecgb'
    edit QUEUE 'express'
    edit REMOTE_HOST 'ecgb'
    edit SITE 'ecgb'
    edit WSDIR '$SCRATCH/metabuilds/%ECF_HOST%_%ECF_PORT%/ecgb/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DBLAS_LIBRARIES=/usr/local/apps/intel/17.0.3/mkl/lib/intel64/libmkl_blas95_lp64.a -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DENABLE_AEC=OFF -DENABLE_DUMMY_TAPES=ON -DENABLE_ECCODES=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_ECMWF_DEVELOPMENT=ON -DENABLE_EXPERIMENTAL=ON -DENABLE_FDB=ON -DENABLE_FDB5=ON -DENABLE_FORTRAN=ON -DENABLE_HERMES=OFF -DENABLE_JPG=OFF -DENABLE_MEMFS=ON -DENABLE_MULTIO_SERVER=ON -DENABLE_NETCDF=OFF -DENABLE_ODB=OFF -DENABLE_ODBSVR=OFF -DENABLE_OMP=OFF -DENABLE_PNG=OFF -DENABLE_PYTHON=OFF -DENABLE_SANDBOX=OFF -DENABLE_TESSELATION=OFF -DENABLE_TESTS=ON -DENABLE_WARNINGS=ON -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_LIBRARIES=/usr/local/apps/intel/17.0.3/mkl/lib/intel64/libmkl_lapack95_lp64.a -DMKL_ROOT=/usr/local/apps/intel/17.0.3/mkl'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module load fftw/%FFTW_VERSION%" "module load boost" "module load eigen" "module load openmpi"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family leap42
    defstatus suspended
    edit CI_LIMIT '7'
    edit CMAKE_VARS '-DBLAS_LIBRARIES=/usr/local/apps/intel/18.0.1/mkl/lib/intel64/libmkl_rt.so -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DENABLE_AEC=OFF -DENABLE_DUMMY_TAPES=ON -DENABLE_ECCODES=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_ECMWF_DEVELOPMENT=ON -DENABLE_EXPERIMENTAL=ON -DENABLE_FDB=ON -DENABLE_FDB5=ON -DENABLE_FORTRAN=ON -DENABLE_HERMES=OFF -DENABLE_JPG=OFF -DENABLE_MEMFS=ON -DENABLE_MULTIO_SERVER=ON -DENABLE_NETCDF=OFF -DENABLE_ODB=OFF -DENABLE_ODBSVR=OFF -DENABLE_OMP=OFF -DENABLE_PNG=OFF -DENABLE_PYTHON=OFF -DENABLE_SANDBOX=OFF -DENABLE_TESSELATION=OFF -DENABLE_TESTS=ON -DENABLE_WARNINGS=ON -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_LIBRARIES=/usr/local/apps/intel/18.0.1/mkl/lib/intel64/libmkl_rt.so -DMKL_ROOT=/usr/local/apps/intel/18.0.1/mkl'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit MKL '/usr/local/apps/intel/18.0.1/mkl'
    edit M_SITE 'leap42'
    edit REMOTE_HOST 'leap42-deploy'
    edit SITE 'leap42'
    edit WSDIR '/tmp/metabuilds/%ECF_HOST%_%ECF_PORT%/leap42/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DBLAS_LIBRARIES=/usr/local/apps/intel/18.0.1/mkl/lib/intel64/libmkl_rt.so -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DENABLE_AEC=OFF -DENABLE_DUMMY_TAPES=ON -DENABLE_ECCODES=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_ECMWF_DEVELOPMENT=ON -DENABLE_EXPERIMENTAL=ON -DENABLE_FDB=ON -DENABLE_FDB5=ON -DENABLE_FORTRAN=ON -DENABLE_HERMES=OFF -DENABLE_JPG=OFF -DENABLE_MEMFS=ON -DENABLE_MULTIO_SERVER=ON -DENABLE_NETCDF=OFF -DENABLE_ODB=OFF -DENABLE_ODBSVR=OFF -DENABLE_OMP=OFF -DENABLE_PNG=OFF -DENABLE_PYTHON=OFF -DENABLE_SANDBOX=OFF -DENABLE_TESSELATION=OFF -DENABLE_TESTS=ON -DENABLE_WARNINGS=ON -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_LIBRARIES=/usr/local/apps/intel/18.0.1/mkl/lib/intel64/libmkl_rt.so -DMKL_ROOT=/usr/local/apps/intel/18.0.1/mkl'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module load fftw/%FFTW_VERSION%" "module load boost" "module load eigen"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family gnu.93
      edit CMAKE_VARS '-DBLAS_LIBRARIES=/usr/local/apps/intel/18.0.1/mkl/lib/intel64/libmkl_rt.so -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DENABLE_AEC=OFF -DENABLE_DUMMY_TAPES=ON -DENABLE_ECCODES=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_ECMWF_DEVELOPMENT=ON -DENABLE_EXPERIMENTAL=ON -DENABLE_FDB=ON -DENABLE_FDB5=ON -DENABLE_FORTRAN=ON -DENABLE_HERMES=OFF -DENABLE_JPG=OFF -DENABLE_MEMFS=ON -DENABLE_MULTIO_SERVER=ON -DENABLE_NETCDF=OFF -DENABLE_ODB=OFF -DENABLE_ODBSVR=OFF -DENABLE_OMP=OFF -DENABLE_PNG=OFF -DENABLE_PYTHON=OFF -DENABLE_SANDBOX=OFF -DENABLE_TESSELATION=OFF -DENABLE_TESTS=ON -DENABLE_WARNINGS=ON -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_LIBRARIES=/usr/local/apps/intel/18.0.1/mkl/lib/intel64/libmkl_rt.so -DMKL_ROOT=/usr/local/apps/intel/18.0.1/mkl'
      edit COMPILER 'GNU.93'
      edit EXPORTS '"export ECACCOUNT=ecodmdma" "export LD_LIBRARY_PATH=/usr/local/apps/gcc/9.3.0/lib64/" "export ftp_proxy=http://proxy.ecmwf.int:3333" "export http_proxy=http://proxy.ecmwf.int:3333" "export https_proxy=http://proxy.ecmwf.int:3333"'
      edit MODULE_ACTIONS '"module swap gnu/9.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module load fftw/%FFTW_VERSION%" "module load boost" "module load eigen"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family intel.19
      edit COMPILER 'INTEL.19'
      edit MODULE_ACTIONS '"module unload gnu" "module load intel/19.0.4" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module load fftw/%FFTW_VERSION%" "module load boost" "module load eigen"'
      edit POST_EXPORTS '"export CXX=icpc\ -cxxlib=/usr/local/apps/gcc/8.3.0" "export LD_RUN_PATH=/usr/local/apps/gcc/8.3.0/lib64" "export PATH=/usr/local/apps/gcc/8.3.0/bin:$PATH"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family clang.90
      edit CMAKE_VARS '-DBLAS_LIBRARIES=/usr/local/apps/intel/18.0.1/mkl/lib/intel64/libmkl_rt.so -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DENABLE_AEC=OFF -DENABLE_DUMMY_TAPES=ON -DENABLE_ECCODES=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_ECMWF_DEVELOPMENT=ON -DENABLE_EXPERIMENTAL=ON -DENABLE_FDB=ON -DENABLE_FDB5=ON -DENABLE_FORTRAN=ON -DENABLE_HERMES=OFF -DENABLE_JPG=OFF -DENABLE_MEMFS=ON -DENABLE_MULTIO_SERVER=ON -DENABLE_NETCDF=OFF -DENABLE_ODB=OFF -DENABLE_ODBSVR=OFF -DENABLE_OMP=OFF -DENABLE_PNG=OFF -DENABLE_PYTHON=OFF -DENABLE_SANDBOX=OFF -DENABLE_TESSELATION=OFF -DENABLE_TESTS=ON -DENABLE_WARNINGS=ON -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_LIBRARIES=/usr/local/apps/intel/18.0.1/mkl/lib/intel64/libmkl_rt.so -DMKL_ROOT=/usr/local/apps/intel/18.0.1/mkl'
      edit COMPILER 'CLANG.90'
      edit MODULE_ACTIONS '"module unload gnu" "module load clang/9.0.1" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module load fftw/%FFTW_VERSION%" "module load boost" "module load eigen"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family lxc
    defstatus suspended
    edit CI_LIMIT '10'
    edit CMAKE_VARS '-DBLAS_LIBRARIES=/usr/local/apps/intel/17.0.3/mkl/lib/intel64/libmkl_blas95_lp64.a -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DENABLE_AEC=OFF -DENABLE_DUMMY_TAPES=ON -DENABLE_ECCODES=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_ECMWF_DEVELOPMENT=ON -DENABLE_EXPERIMENTAL=ON -DENABLE_FDB=ON -DENABLE_FDB5=ON -DENABLE_FORTRAN=ON -DENABLE_HERMES=OFF -DENABLE_JPG=OFF -DENABLE_MEMFS=ON -DENABLE_MULTIO_SERVER=ON -DENABLE_NETCDF=OFF -DENABLE_ODB=OFF -DENABLE_ODBSVR=OFF -DENABLE_OMP=OFF -DENABLE_PNG=OFF -DENABLE_PYTHON=OFF -DENABLE_SANDBOX=OFF -DENABLE_TESSELATION=OFF -DENABLE_TESTS=ON -DENABLE_WARNINGS=ON -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_LIBRARIES=/usr/local/apps/intel/17.0.3/mkl/lib/intel64/libmkl_lapack95_lp64.a -DMKL_ROOT=/usr/local/apps/intel/17.0.3/mkl'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit MKL '/usr/local/apps/intel/17.0.3/mkl'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module load fftw/%FFTW_VERSION%" "module load boost" "module load eigen" "module load openmpi"'
    edit M_SITE 'lxc'
    edit QUEUE 'express'
    edit REMOTE_HOST 'lxc'
    edit SITE 'lxc'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxc/%COMPILER:%'
    family gnu.63
      edit CMAKE_VARS '-DBLAS_LIBRARIES=/usr/local/apps/intel/17.0.3/mkl/lib/intel64/libmkl_blas95_lp64.a -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DENABLE_AEC=OFF -DENABLE_DUMMY_TAPES=ON -DENABLE_ECCODES=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_ECMWF_DEVELOPMENT=ON -DENABLE_EXPERIMENTAL=ON -DENABLE_FDB=ON -DENABLE_FDB5=ON -DENABLE_FORTRAN=ON -DENABLE_HERMES=OFF -DENABLE_JPG=OFF -DENABLE_MEMFS=ON -DENABLE_MULTIO_SERVER=ON -DENABLE_NETCDF=OFF -DENABLE_ODB=OFF -DENABLE_ODBSVR=OFF -DENABLE_OMP=OFF -DENABLE_PNG=OFF -DENABLE_PYTHON=OFF -DENABLE_SANDBOX=OFF -DENABLE_TESSELATION=OFF -DENABLE_TESTS=ON -DENABLE_WARNINGS=ON -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_LIBRARIES=/usr/local/apps/intel/17.0.3/mkl/lib/intel64/libmkl_lapack95_lp64.a -DMKL_ROOT=/usr/local/apps/intel/17.0.3/mkl'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module swap gnu/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module load fftw/%FFTW_VERSION%" "module load boost" "module load eigen" "module load openmpi"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family lxg
    defstatus suspended
    edit CI_LIMIT '10'
    edit CMAKE_VARS '-DBLAS_LIBRARIES=/usr/local/apps/intel/19.0.0/mkl/lib/intel64/libmkl_blas95_lp64.a -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DENABLE_AEC=OFF -DENABLE_DUMMY_TAPES=ON -DENABLE_ECCODES=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_ECMWF_DEVELOPMENT=ON -DENABLE_EXPERIMENTAL=ON -DENABLE_FDB=ON -DENABLE_FDB5=ON -DENABLE_FORTRAN=ON -DENABLE_HERMES=OFF -DENABLE_JPG=OFF -DENABLE_MEMFS=ON -DENABLE_MULTIO_SERVER=ON -DENABLE_NETCDF=OFF -DENABLE_ODB=OFF -DENABLE_ODBSVR=OFF -DENABLE_OMP=OFF -DENABLE_PNG=OFF -DENABLE_PYTHON=OFF -DENABLE_SANDBOX=OFF -DENABLE_TESSELATION=OFF -DENABLE_TESTS=ON -DENABLE_WARNINGS=ON -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_LIBRARIES=/usr/local/apps/intel/19.0.0/mkl/lib/intel64/libmkl_lapack95_lp64.a -DMKL_ROOT=/usr/local/apps/intel/19.0.0/mkl'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit MKL '/usr/local/apps/intel/19.0.0/mkl'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module load fftw/%FFTW_VERSION%" "module load boost" "module load eigen" "module load openmpi"'
    edit M_SITE 'lxg'
    edit QUEUE 'normal'
    edit REMOTE_HOST 'lxg'
    edit SITE 'lxg'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxg/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DBLAS_LIBRARIES=/usr/local/apps/intel/19.0.0/mkl/lib/intel64/libmkl_blas95_lp64.a -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DENABLE_AEC=OFF -DENABLE_DUMMY_TAPES=ON -DENABLE_ECCODES=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_ECMWF_DEVELOPMENT=ON -DENABLE_EXPERIMENTAL=ON -DENABLE_FDB=ON -DENABLE_FDB5=ON -DENABLE_FORTRAN=ON -DENABLE_HERMES=OFF -DENABLE_JPG=OFF -DENABLE_MEMFS=ON -DENABLE_MULTIO_SERVER=ON -DENABLE_NETCDF=OFF -DENABLE_ODB=OFF -DENABLE_ODBSVR=OFF -DENABLE_OMP=OFF -DENABLE_PNG=OFF -DENABLE_PYTHON=OFF -DENABLE_SANDBOX=OFF -DENABLE_TESSELATION=OFF -DENABLE_TESTS=ON -DENABLE_WARNINGS=ON -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_LIBRARIES=/usr/local/apps/intel/19.0.0/mkl/lib/intel64/libmkl_lapack95_lp64.a -DMKL_ROOT=/usr/local/apps/intel/19.0.0/mkl'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module load fftw/%FFTW_VERSION%" "module load boost" "module load eigen" "module load openmpi"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family sappb
    defstatus suspended
    edit ASAN_FILE '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
    edit BASEPREFIX '/autofs/vol/deploy/install/%SITE%'
    edit CI_LIMIT '8'
    edit CMAKE_VARS '-DBLAS_LIBRARIES=/usr/local/apps/intel/17.0.3/mkl/lib/intel64/libmkl_blas95_lp64.a -DBUNDLE_SKIP_OOPS=ON -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DENABLE_AEC=OFF -DENABLE_DUMMY_TAPES=ON -DENABLE_ECCODES=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_ECMWF_DEVELOPMENT=ON -DENABLE_EXPERIMENTAL=ON -DENABLE_FDB=ON -DENABLE_FDB5=ON -DENABLE_FORTRAN=ON -DENABLE_HERMES=OFF -DENABLE_JPG=OFF -DENABLE_MEMFS=ON -DENABLE_MULTIO_SERVER=ON -DENABLE_NETCDF=OFF -DENABLE_ODB=OFF -DENABLE_ODBSVR=OFF -DENABLE_OMP=OFF -DENABLE_PNG=OFF -DENABLE_PYTHON=OFF -DENABLE_SANDBOX=OFF -DENABLE_TESSELATION=OFF -DENABLE_TESTS=ON -DENABLE_WARNINGS=ON -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_LIBRARIES=/usr/local/apps/intel/17.0.3/mkl/lib/intel64/libmkl_lapack95_lp64.a -DMKL_ROOT=/usr/local/apps/intel/17.0.3/mkl'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit MKL '/usr/local/apps/intel/17.0.3/mkl'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module load fftw/%FFTW_VERSION%" "module load boost"'
    edit M_SITE 'sappb'
    edit REMOTE_HOST 'sappb-deploy'
    edit SANITIZER_REFERENCE_DIRECTORY '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
    edit SITE 'sappb'
    edit TSAN_FILE '/home/ma/deploy/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
    edit WSDIR '/autofs/vol/deploy/metabuilds/%ECF_HOST%_%ECF_PORT%/sappb/%COMPILER:%'
    family gnu.63
      edit CMAKE_VARS '-DBLAS_LIBRARIES=/usr/local/apps/intel/17.0.3/mkl/lib/intel64/libmkl_blas95_lp64.a -DBUNDLE_SKIP_OOPS=ON -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DENABLE_AEC=OFF -DENABLE_DUMMY_TAPES=ON -DENABLE_ECCODES=ON -DENABLE_ECCODES_THREADS=ON -DENABLE_ECMWF_DEVELOPMENT=ON -DENABLE_EXPERIMENTAL=ON -DENABLE_FDB=ON -DENABLE_FDB5=ON -DENABLE_FORTRAN=ON -DENABLE_HERMES=OFF -DENABLE_JPG=OFF -DENABLE_MEMFS=ON -DENABLE_MULTIO_SERVER=ON -DENABLE_NETCDF=OFF -DENABLE_ODB=OFF -DENABLE_ODBSVR=OFF -DENABLE_OMP=OFF -DENABLE_PNG=OFF -DENABLE_PYTHON=OFF -DENABLE_SANDBOX=OFF -DENABLE_TESSELATION=OFF -DENABLE_TESTS=ON -DENABLE_WARNINGS=ON -DFDB_CONF_FILE=/usr/local/apps/fdb/etc/Config -DLAPACK_LIBRARIES=/usr/local/apps/intel/17.0.3/mkl/lib/intel64/libmkl_lapack95_lp64.a -DMKL_ROOT=/usr/local/apps/intel/17.0.3/mkl'
      edit COMPILER 'GNU.63'
      edit MODULE_ACTIONS '"module load gnu/6.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module load fftw/%FFTW_VERSION%" "module load boost"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
endsuite
suite mars_client
  edit ADDRESS_SANITIZER 'False'
  edit ADD_ACTION_TRIGGERS 'True'
  edit AEC_VERSION '1.0.4'
  edit ALLOW_INSTALL_TO_OVERWRITE 'True'
  edit ASAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
  edit ATLAS_GIT_TAG 'develop'
  edit ATLAS_VERSION '0.20.2'
  edit BANNER_TASK_END 'banner %TASK%-end || true'
  edit BANNER_TASK_START 'banner %TASK%-start || true'
  edit BASEDIR '%WSDIR%/%PACKAGE%'
  edit BASEPREFIX '$PERM/install/%SITE%'
  edit BUFRDC_EMOSLIB_VERSION '4.6.3'
  edit BUFRDC_TABLES_GIT_TAG 'develop'
  edit BUFRDC_TABLES_VERSION '000415'
  edit BUILDIR '%BASEDIR%/builds'
  edit BUILDSTAMP 'DSUITE'
  edit BUILD_ALWAYS_PARALLEL 'False'
  edit CFI_SANITIZER 'False'
  edit CFL_BASE_URL 'https://software-test.ecmwf.int/wiki'
  edit CFL_SPACE_KEY '<TODO>'
  edit CFL_TARBALL_COMMENT ''
  edit CFL_TITLE 'Releases'
  edit CHECK_SYNC_RELEASE 'False'
  edit CLANG_TIDY_ABORT_ON_ERROR 'False'
  edit CMAKE '%ECBUILD%'
  edit CMAKELIST_EXTRACT_PACKAGE_VERSION '$(awk '/^project/ && /%PACKAGE%/ && /VERSION/ {for (I=1;I<=NF;I++) if ($I == "VERSION") {print $(I+1)};}' CMakeLists.txt)'
  edit CMAKE_BUILD_TYPE 'RelWithDebInfo'
  edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DECMWF=ON -DEC_BUILD_TIMESTAMP=%BUILDSTAMP% -DENABLE_TESSELATION=OFF'
  edit CMAKE_VERSION '3.13.4'
  edit CODES_UI_GIT_TAG 'develop'
  edit CODES_UI_VERSION '1.3.1'
  edit CPACK 'make package'
  edit CPACK_PACKAGE_NAME 'mars-client'
  edit CPACK_PACKAGING_INSTALL_PREFIX '/opt/ecmwf/%CPACK_PACKAGE_NAME%'
  edit CPUS '$(lscpu -p | grep -v '#' | wc -l)'
  edit CRAY_FFTW_VERSION '3.3.8.1'
  edit CRAY_NETCDF_VERSION '4.6.1.3'
  edit CRON_BASEPREFIX '%BASEDIR%/install'
  edit CRON_EMAIL ''
  edit CRON_ENABLED 'False'
  edit CRON_GIT_TAG 'develop'
  edit CUSTOM_COMMANDS '"export SKIP_PACKAGE_VERSION_CHECK=True"'
  edit DATAFLOW_SANITIZER 'False'
  edit DEFAULT_GNU_COMPILER 'gnu.73'
  edit ECBUILD '%BASEDIR%/ecbuild/bin/ecbuild'
  edit ECBUILD_CLONE 'True'
  edit ECBUILD_CLONE_COMMIT '3.1.0'
  edit ECBUILD_GIT_TAG 'develop'
  edit ECBUILD_REPO '%GIT_REPO_PREFIX%/ecsdk/ecbuild.git'
  edit ECBUILD_VERSION '3.3.2'
  edit ECCODES_EXTENSIONS_GIT_TAG 'develop'
  edit ECCODES_EXTENSIONS_VERSION '2.17.0'
  edit ECCODES_GIT_TAG 'develop'
  edit ECCODES_PYTHON_VERSION '0.9.6'
  edit ECCODES_VERSION '2.17.1'
  edit ECFLOW5_LATEST_VERSION '5.4.0'
  edit ECFLOW_VERSION '5.4.1'
  edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/mars_client'
  edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/mars_client'
  edit ECF_JOB_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_JOB% %ECF_JOBOUT%'
  edit ECF_KILL_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% kill'
  edit ECF_STATUS_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% stat'
  edit ECF_TRIES '1'
  edit ECKIT_GIT_TAG 'develop'
  edit ECKIT_VERSION '1.10.1'
  edit ECQUIS_VERSION '1.0.0'
  edit EMOS_MOD_VERSION '462'
  edit EMOS_PA_VERSION '000462'
  edit EXPORTS '"export CONFIG_INTERP=ON_FLY" "export ECACCOUNT=ecodmdma" "export ECKIT_LINEAR_ALGEBRA_BACKEND=mkl" "export MARS_PPROC_BACKEND=MIR" "export MIR_CACHING=0" "export ftp_proxy=http://proxy.ecmwf.int:3333" "export http_proxy=http://proxy.ecmwf.int:3333" "export https_proxy=http://proxy.ecmwf.int:3333"'
  edit EXPORT_VARIABLES 'False'
  edit FCKIT_GIT_TAG '0.6.3'
  edit FCKIT_VERSION '0.6.3'
  edit FDB_VINTAGE_GIT_TAG '4.10.0'
  edit FDB_VINTAGE_VERSION '4.10.0'
  edit FFTW_VERSION '3.3.8'
  edit FFTW_VERSION_CRAY '3.3.4.11'
  edit GIT 'git'
  edit GITHUB_PACKAGE_NAME '%PACKAGE_NAME%'
  edit GITHUB_REPO 'git@github.com:ecmwf'
  edit GITREPO '%GIT_REPO_PREFIX%/mars/mars-client-bundle.git'
  edit GIT_CLONE 'True'
  edit GIT_CLONE_OPTIONS '--depth=1 --branch %GIT_TAG%'
  edit GIT_METABUILDER_TAG 'ci'
  edit GIT_REPO_PREFIX 'ssh://git@git.ecmwf.int:7999'
  edit GIT_TAG '6.27.2.2'
  edit GIT_VERSION '2.26.0'
  edit GRIB_API_VERSION '1.28.0'
  edit INSTALLABLE_TO_USR_LOCAL_APPS 'False'
  edit INSTALL_DIR '/usr/local/apps/mars/versions/DSUITE'
  edit IWYU_VERSION '0.11'
  edit KSH_EXTRACT_PACKAGE_VERSION '`grep _VERSION_STR VERSION.cmake | grep -o '[0-9]*\.[0-9]*\.[0-9]*\.[0-9]'`'
  edit LAPACK_VERSION '3.5.0'
  edit LIBEMOS_GIT_TAG 'develop'
  edit LIBEMOS_VERSION '4.6.2'
  edit LINTERS_REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}'
  edit LSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/lsan.txt'
  edit MAGICS_GIT_TAG 'develop'
  edit MAGICS_PYTHON_VERSION 'master'
  edit MAGICS_VERSION '4.3.3'
  edit MAKE 'make'
  edit MARSPROD '$mars_script'
  edit MARSTEST '$mars_script -b $INSTALL_DIR/bin/mars.bin'
  edit MARS_CLIENT_GIT_TAG '6.27.2.2'
  edit MARS_CLIENT_PACKAGE_GIT_TAG '6.27.6'
  edit MARS_CLIENT_PACKAGE_VERSION '6.27.6'
  edit MARS_CLIENT_VERSION '6.27.2.2'
  edit MARS_SCRIPT 'mars'
  edit MEMORY_SANITIZER 'False'
  edit METABUILDER_HOME '/var/tmp/ma0/workspace/metabuilder'
  edit METAB_INSTALL_REMOVE_WRITE_PERMISSIONS 'False'
  edit METAB_PYTHON_VERSION 'python3'
  edit METKIT_GIT_TAG 'develop'
  edit METKIT_VERSION '1.4.6'
  edit METVIEW_GIT_TAG 'develop'
  edit METVIEW_PYTHON_VERSION '1.4.2'
  edit METVIEW_VERSION '5.8.3'
  edit MIR_BUNDLE_VERSION '0.9.0.0'
  edit MIR_GIT_TAG 'develop'
  edit MIR_VERSION '1.5.0'
  edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module load hdf5" "module load netcdf4" "module load fftw/%FFTW_VERSION%"'
  edit MODULE_ACTIONS_FIRST ''
  edit MODULE_ACTIONS_HDR 'module_actions.h'
  edit MODULE_CHECKING_HDR 'module_checking.h'
  edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow"'
  edit MODULE_LOAD_SUPPORTED 'False'
  edit MODULE_NAME 'mars_client'
  edit MODULE_VERSION '%GIT_TAG%'
  edit NETCDF4_VERSION '4.7.4'
  edit ODB_API_GIT_TAG 'develop'
  edit ODB_API_VERSION '0.19.4'
  edit ODB_GIT_TAG 'develop'
  edit ODB_TOOLS_GIT_TAG 'develop'
  edit ODB_TOOLS_VERSION '0.18.3'
  edit ODB_VERSION '1.0.11'
  edit ODC_GIT_TAG 'develop'
  edit ODC_VERSION '1.0.3'
  edit OPENJPEG_VERSION '2.3.1'
  edit OPTIONS '$options -DECBUILD_LOG_LEVEL=DEBUG'
  edit PACKAGE 'mars_client'
  edit PACKAGE_NAME 'mars_client'
  edit PACKAGE_OWNERS 'ma0@ecmwf.int'
  edit PACKAGE_VERSION ''
  edit PGEN_VERSION '1.7.1.0'
  edit PKG_BOOST_VERSION '1.71.0'
  edit PREFIX_APPEND ''
  edit PROJ4_VERSION '5.2.0'
  edit PROJ_VERSION '6.3.1'
  edit PYTHON2_VERSION '2.7.15-01'
  edit PYTHON3_VERSION '3.6.10-01'
  edit PYTHON3_VERSION_MAJOR_MINOR '3.6'
  edit PYTHON_GITREPO 'This-needs-tobe-overriden'
  edit PYTHON_PACKAGE 'This-needs-tobe-overriden'
  edit QT_VERSION '5.12.3'
  edit QUEUE 'build'
  edit REMOTE_HOST 'polonius'
  edit REUSE_BUILD_DIR 'False'
  edit RPM_ARCH 'x86_64'
  edit RPM_FILENAME '%CPACK_PACKAGE_NAME%-${PACKAGE_VERSION}-Linux-%RPM_ARCH%.rpm'
  edit RPM_PASSWORD ''
  edit RPM_REPO_DESTINATION 'https://nexus.ecmwf.int/repository/CentOS/7/6/rpms/%RPM_ARCH%/%RPM_FILENAME%'
  edit RPM_USERNAME ''
  edit SAFE_STACK_SANITIZER 'False'
  edit SANITIZER 'False'
  edit SANITIZER_REFERENCE_DIRECTORY '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
  edit SRCDIR '%BASEDIR%/%PACKAGE%'
  edit STATIC_ANALYSER_ABORT_ON_ERROR 'False'
  edit SWIG_VERSION '3.0.8'
  edit SYNC_RELEASE '2020.04.0'
  edit SYNC_RELEASE_CI_STAGE 'develop'
  edit TARBALL_DIR '/perm/ma/ma0/tmp_tarballs/ci'
  edit TAR_PACKAGE '%PACKAGE%'
  edit TEST 'ctest --output-on-failure -j${CPUS} '
  edit THREAD_SANITIZER 'False'
  edit TRANSI_GIT_TAG 'develop'
  edit TRANSI_VERSION '0.7.1'
  edit TSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
  edit UBSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/ubsan.txt'
  edit UB_SANITIZER 'False'
  edit USER 'ma0'
  edit USE_WORKDIR 'False'
  edit VALGRIND_VERSION '3.15.0'
  edit WORKDIR '%BASEDIR%/workdir/%ECF_NAME%'
  edit WSDIR '/var/tmp/tmpdir/deploy/jtmp.20739'
  label branch "6.27.2.2"
  task regenerate
    defstatus complete
    edit ECF_JOB_CMD '%ECF_JOB% > %ECF_JOBOUT% 2>&1'
    edit ECF_KILL_CMD 'kill -15 %ECF_RID%'
    edit ECF_PASS 'FREE'
    edit ECF_STATUS_CMD 'ps --sid %ECF_RID% -f'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
    label date ""
    label metabuilder_branch "ci"
  family limits
    defstatus complete
    edit REMOTE_HOST 'polonius'
    edit WORKDIR ''
    limit eurobrisa 1
  endfamily
  family cca
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DBLAS_LIBRARIES="/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_intel_lp64.so;/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_sequential.so;/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_core.so" -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECMWF=ON -DECMWF_HPC=ON -DEC_BUILD_TIMESTAMP=%BUILDSTAMP% -DENABLE_TESSELATION=OFF -DLAPACK_FOUND=TRUE -DLAPACK_LIBRARIES="/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_intel_lp64.so;/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_sequential.so;/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_core.so" -DMKL_ROOT=/opt/intel/composer_xe_2018.0.033/mkl'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/mars_client/cca'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/mars_client/cca:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cca-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export CONFIG_INTERP=ON_FLY" "export ECACCOUNT=ecodmdma" "export ECKIT_LINEAR_ALGEBRA_BACKEND=mkl" "export MARS_PPROC_BACKEND=MIR" "export MIR_CACHING=0" "export TMPDIR=/tmp" "export ftp_proxy=http://proxy.ecmwf.int:3333" "export http_proxy=http://proxy.ecmwf.int:3333" "export https_proxy=http://proxy.ecmwf.int:3333"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '28000'
    edit MKL '/opt/intel/composer_xe_2018.0.033/mkl'
    edit MKLLIBS '/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_intel_lp64.so;/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_sequential.so;/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_core.so'
    edit MODULE_ACTIONS '"module swap PrgEnv-cray PrgEnv-gnu" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module load hdf5" "module load netcdf4" "module load fftw/%FFTW_VERSION%" "module switch fftw/%FFTW_VERSION%" "module unload craype-broadwell" "module load craype-ivybridge"'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cca'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'cca'
    edit SITE 'cca'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.73
      edit CMAKE_VARS '-DBLAS_LIBRARIES="/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_intel_lp64.so;/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_sequential.so;/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_core.so" -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DECMWF=ON -DECMWF_HPC=ON -DEC_BUILD_TIMESTAMP=%BUILDSTAMP% -DENABLE_TESSELATION=OFF -DLAPACK_FOUND=TRUE -DLAPACK_LIBRARIES="/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_intel_lp64.so;/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_sequential.so;/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_core.so" -DMKL_ROOT=/opt/intel/composer_xe_2018.0.033/mkl'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module swap PrgEnv-cray PrgEnv-gnu" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module load hdf5" "module load netcdf4" "module load fftw/%FFTW_VERSION%" "module switch fftw/%FFTW_VERSION%" "module unload craype-broadwell" "module load craype-ivybridge"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task install
        trigger build == complete
        edit ECF_TRIES '2'
      task remove_write_perm
        trigger install == complete
      task test
        trigger remove_write_perm == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
      family tests
        defstatus complete
        edit COMPARE 'cmp'
        edit NFSHOME '$HOME/tmp/'
        edit OPERSCHOST 'cca'
        limit tests 3
        inlimit tests
        task archive
          edit DATABASE 'marstest'
          edit GRID ''
          edit PARAM '129/130/135/138/155/157'
        task bad_archive
          edit DATABASE 'marstest'
          edit GRID ''
          edit PARAM '129/130/135/138/155/157'
        task environment
        task grib2fdb
        task grib2fdb5
        task marsgen
          edit DATABASE 'marstest'
        task memberstates
          edit DATABASE 'marstest'
          edit GRID ''
          edit PARAM '129/130/135/138/155/157'
        task reqcheck
        family retrieve
          edit DATABASE ''
          edit GRID ''
          edit PARAM ''
          task L137
          task accuracy
            edit LEVEL 'levtype=pl, level=1000/500,'
            edit PARAM 'T'
            edit SCRIPT 'accuracy.sms'
          task ai
          task coefficients
            edit GRID '0.1/0.1'
            edit PARAM '129'
          task compute
            edit GRID ''
            edit PARAM ''
          task custom
          task datasets
          task dhs_chk
          task dissemination
          task ems
          task enda
          task ensemble
            edit GRID '2.5/2.5'
            edit PARAM '139/141/142/143/144/145/146/147/151/164/165/166'
          task ep
            edit GRID '5/5'
            edit PARAM '228'
          task era5t_embargo
          task eurobrisa
            defstatus complete
            inlimit /mars_client/limits:eurobrisa
          task exact_match
          task expect
          task fdb
            edit DATABASE 'c2a fdb'
            edit TOLERANCE '0.1'
          task fdb5
            edit DATABASE 'fdb5 prod'
            edit TOLERANCE '0.1'
          task fieldset
            edit GRID '1.5/1.5'
            edit PARAM '27/28/29/30/31/32/33/34/35/36/37/38/39/40/41/42/128/129/134/136/137/139/141/148/151/160/161/162/163/164/165/166/167/168/170/172/173/174/183/186/187/188/198/206/234/235/236/238'
          task fileio
          task fourpoints
          task gauss
            edit GRID '160'
            edit PARAM '31/129/134/136/137/139/141'
          task gauss_regular
            edit GRID '160'
            edit PARAM '31/128/129/134/136/137/139/141'
          task gauss_rotation
          task hirl12p
          task interpolation
            edit GRID '2.5/2.5'
          task intuvp
          task issame
            edit GRID '160'
            edit PARAM '31/128/129/134/136/137/139/141'
          task latlon
            edit GRID '2.5/2.5'
          task mail
          task metgram
          task monthly
          task obs
          task obs_restrict
          task ocean
          task odb_archive
            edit ODB '1'
          task odb_filter
            edit ODB '1'
          task olduv
            edit PARAM 'u/v'
            edit TOLERANCE '0.1'
          task padding
            edit COMPARE 'grib_compare'
          task prob
          task resol
          task rotation
            edit LEVEL 'levtype=pl, level=1000/500,'
            edit PARAM 'T'
          task s2s
          task seas
          task seasonal_restrict
          task sfc
            edit GRIB_COMPARE_OPTIONS '-P -T2 -f'
            edit GRID '1.5/1.5'
            edit PARAM '27/28/29/30/31/32/33/34/35/36/37/38/39/40/41/42/128/129/134/136/137/139/141/148/151/160/161/162/163/164/165/166/167/168/170/172/173/174/183/186/187/188/198/206/234/235/236/238'
          task slwet
          task t1279
          task target
          task tc
          task tigge_area
          task tigge_latlon
          task tigge_nointerp
          task uv
            edit PARAM 'u/v'
            edit TOLERANCE '0.1'
          task vareps
          task vector_wind
          task wave
            edit GRID '1.5/1.5'
            edit PARAM '251'
        endfamily
        task size
          edit PERCENT '10'
        task stage
          edit DATABASE 'marsrd'
          edit GRID '2.5/2.5'
          edit PARAM '129'
        family xtra
          edit COMPARE 'grib_compare -f -P -T2'
          limit xtra 1
          task task00
            label id ""
            label oper_fail ""
            label test_fail ""
            label comp_fail ""
            label message ""
          task task01
            label id ""
            label oper_fail ""
            label test_fail ""
            label comp_fail ""
            label message ""
          task task02
            label id ""
            label oper_fail ""
            label test_fail ""
            label comp_fail ""
            label message ""
          task task03
            label id ""
            label oper_fail ""
            label test_fail ""
            label comp_fail ""
            label message ""
          task task04
            label id ""
            label oper_fail ""
            label test_fail ""
            label comp_fail ""
            label message ""
          task task05
            label id ""
            label oper_fail ""
            label test_fail ""
            label comp_fail ""
            label message ""
          task task06
            label id ""
            label oper_fail ""
            label test_fail ""
            label comp_fail ""
            label message ""
          task task07
            label id ""
            label oper_fail ""
            label test_fail ""
            label comp_fail ""
            label message ""
          task task08
            label id ""
            label oper_fail ""
            label test_fail ""
            label comp_fail ""
            label message ""
          task task09
            label id ""
            label oper_fail ""
            label test_fail ""
            label comp_fail ""
            label message ""
          task task10
            label id ""
            label oper_fail ""
            label test_fail ""
            label comp_fail ""
            label message ""
          task task11
            label id ""
            label oper_fail ""
            label test_fail ""
            label comp_fail ""
            label message ""
          task task12
            label id ""
            label oper_fail ""
            label test_fail ""
            label comp_fail ""
            label message ""
          task task13
            label id ""
            label oper_fail ""
            label test_fail ""
            label comp_fail ""
            label message ""
          task task14
            label id ""
            label oper_fail ""
            label test_fail ""
            label comp_fail ""
            label message ""
          task task15
            label id ""
            label oper_fail ""
            label test_fail ""
            label comp_fail ""
            label message ""
          task task16
            label id ""
            label oper_fail ""
            label test_fail ""
            label comp_fail ""
            label message ""
          task task17
            label id ""
            label oper_fail ""
            label test_fail ""
            label comp_fail ""
            label message ""
          task task18
            label id ""
            label oper_fail ""
            label test_fail ""
            label comp_fail ""
            label message ""
          task task19
            label id ""
            label oper_fail ""
            label test_fail ""
            label comp_fail ""
            label message ""
        endfamily
      endfamily
    endfamily
  endfamily
  family ccb
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DBLAS_LIBRARIES="/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_intel_lp64.so;/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_sequential.so;/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_core.so" -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DECMWF=ON -DECMWF_HPC=ON -DEC_BUILD_TIMESTAMP=%BUILDSTAMP% -DENABLE_TESSELATION=OFF -DLAPACK_FOUND=TRUE -DLAPACK_LIBRARIES="/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_intel_lp64.so;/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_sequential.so;/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_core.so" -DMKL_ROOT=/opt/intel/composer_xe_2018.0.033/mkl'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/mars_client/ccb'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/mars_client/ccb:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'ccb-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export CONFIG_INTERP=ON_FLY" "export ECACCOUNT=ecodmdma" "export ECKIT_LINEAR_ALGEBRA_BACKEND=mkl" "export MARS_PPROC_BACKEND=MIR" "export MIR_CACHING=0" "export TMPDIR=/tmp" "export ftp_proxy=http://proxy.ecmwf.int:3333" "export http_proxy=http://proxy.ecmwf.int:3333" "export https_proxy=http://proxy.ecmwf.int:3333"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '28000'
    edit MKL '/opt/intel/composer_xe_2018.0.033/mkl'
    edit MKLLIBS '/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_intel_lp64.so;/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_sequential.so;/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_core.so'
    edit MODULE_ACTIONS '"module swap PrgEnv-cray PrgEnv-gnu" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module load hdf5" "module load netcdf4" "module load fftw/%FFTW_VERSION%" "module switch fftw/%FFTW_VERSION%" "module unload craype-broadwell" "module load craype-ivybridge"'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'ccb'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'ccb'
    edit SITE 'ccb'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family gnu.73
      edit CMAKE_VARS '-DBLAS_LIBRARIES="/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_intel_lp64.so;/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_sequential.so;/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_core.so" -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_EXE_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DCMAKE_EXPORT_COMPILE_COMMANDS=ON -DCMAKE_Fortran_COMPILER=ftn -DCMAKE_SHARED_LINKER_FLAGS=-Wl,-rpath,/opt/cray/gcc-libs -DECMWF=ON -DECMWF_HPC=ON -DEC_BUILD_TIMESTAMP=%BUILDSTAMP% -DENABLE_TESSELATION=OFF -DLAPACK_FOUND=TRUE -DLAPACK_LIBRARIES="/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_intel_lp64.so;/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_sequential.so;/opt/intel/composer_xe_2018.0.033/mkl/lib/intel64/libmkl_core.so" -DMKL_ROOT=/opt/intel/composer_xe_2018.0.033/mkl'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap cdt/18.03" "module swap gcc/7.3.0" "module swap PrgEnv-cray PrgEnv-gnu" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module load hdf5" "module load netcdf4" "module load fftw/%FFTW_VERSION%" "module switch fftw/%FFTW_VERSION%" "module unload craype-broadwell" "module load craype-ivybridge"'
      edit MODULE_ACTIONS_FIRST '"module swap PrgEnv-cray PrgEnv-gnu"'
      task git_srcs
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task install
        trigger build == complete
        edit ECF_TRIES '2'
      task remove_write_perm
        trigger install == complete
      task test
        trigger remove_write_perm == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task wipe_build
        trigger test == complete
      task wipe_src
        trigger wipe_build == complete
      family tests
        defstatus complete
        edit COMPARE 'cmp'
        edit NFSHOME '$HOME/tmp/'
        edit OPERSCHOST 'cca'
        limit tests 3
        inlimit tests
        task archive
          edit DATABASE 'marstest'
          edit GRID ''
          edit PARAM '129/130/135/138/155/157'
        task bad_archive
          edit DATABASE 'marstest'
          edit GRID ''
          edit PARAM '129/130/135/138/155/157'
        task environment
        task grib2fdb
        task grib2fdb5
        task marsgen
          edit DATABASE 'marstest'
        task memberstates
          edit DATABASE 'marstest'
          edit GRID ''
          edit PARAM '129/130/135/138/155/157'
        task reqcheck
        family retrieve
          edit DATABASE ''
          edit GRID ''
          edit PARAM ''
          task L137
          task accuracy
            edit LEVEL 'levtype=pl, level=1000/500,'
            edit PARAM 'T'
            edit SCRIPT 'accuracy.sms'
          task ai
          task coefficients
            edit GRID '0.1/0.1'
            edit PARAM '129'
          task compute
            edit GRID ''
            edit PARAM ''
          task custom
          task datasets
          task dhs_chk
          task dissemination
          task ems
          task enda
          task ensemble
            edit GRID '2.5/2.5'
            edit PARAM '139/141/142/143/144/145/146/147/151/164/165/166'
          task ep
            edit GRID '5/5'
            edit PARAM '228'
          task era5t_embargo
          task eurobrisa
            defstatus complete
            inlimit /mars_client/limits:eurobrisa
          task exact_match
          task expect
          task fdb
            edit DATABASE 'c2a fdb'
            edit TOLERANCE '0.1'
          task fdb5
            edit DATABASE 'fdb5 prod'
            edit TOLERANCE '0.1'
          task fieldset
            edit GRID '1.5/1.5'
            edit PARAM '27/28/29/30/31/32/33/34/35/36/37/38/39/40/41/42/128/129/134/136/137/139/141/148/151/160/161/162/163/164/165/166/167/168/170/172/173/174/183/186/187/188/198/206/234/235/236/238'
          task fileio
          task fourpoints
          task gauss
            edit GRID '160'
            edit PARAM '31/129/134/136/137/139/141'
          task gauss_regular
            edit GRID '160'
            edit PARAM '31/128/129/134/136/137/139/141'
          task gauss_rotation
          task hirl12p
          task interpolation
            edit GRID '2.5/2.5'
          task intuvp
          task issame
            edit GRID '160'
            edit PARAM '31/128/129/134/136/137/139/141'
          task latlon
            edit GRID '2.5/2.5'
          task mail
          task metgram
          task monthly
          task obs
          task obs_restrict
          task ocean
          task odb_archive
            edit ODB '1'
          task odb_filter
            edit ODB '1'
          task olduv
            edit PARAM 'u/v'
            edit TOLERANCE '0.1'
          task padding
            edit COMPARE 'grib_compare'
          task prob
          task resol
          task rotation
            edit LEVEL 'levtype=pl, level=1000/500,'
            edit PARAM 'T'
          task s2s
          task seas
          task seasonal_restrict
          task sfc
            edit GRIB_COMPARE_OPTIONS '-P -T2 -f'
            edit GRID '1.5/1.5'
            edit PARAM '27/28/29/30/31/32/33/34/35/36/37/38/39/40/41/42/128/129/134/136/137/139/141/148/151/160/161/162/163/164/165/166/167/168/170/172/173/174/183/186/187/188/198/206/234/235/236/238'
          task slwet
          task t1279
          task target
          task tc
          task tigge_area
          task tigge_latlon
          task tigge_nointerp
          task uv
            edit PARAM 'u/v'
            edit TOLERANCE '0.1'
          task vareps
          task vector_wind
          task wave
            edit GRID '1.5/1.5'
            edit PARAM '251'
        endfamily
        task size
          edit PERCENT '10'
        task stage
          edit DATABASE 'marsrd'
          edit GRID '2.5/2.5'
          edit PARAM '129'
        family xtra
          edit COMPARE 'grib_compare -f -P -T2'
          limit xtra 1
          task task00
            label id ""
            label oper_fail ""
            label test_fail ""
            label comp_fail ""
            label message ""
          task task01
            label id ""
            label oper_fail ""
            label test_fail ""
            label comp_fail ""
            label message ""
          task task02
            label id ""
            label oper_fail ""
            label test_fail ""
            label comp_fail ""
            label message ""
          task task03
            label id ""
            label oper_fail ""
            label test_fail ""
            label comp_fail ""
            label message ""
          task task04
            label id ""
            label oper_fail ""
            label test_fail ""
            label comp_fail ""
            label message ""
          task task05
            label id ""
            label oper_fail ""
            label test_fail ""
            label comp_fail ""
            label message ""
          task task06
            label id ""
            label oper_fail ""
            label test_fail ""
            label comp_fail ""
            label message ""
          task task07
            label id ""
            label oper_fail ""
            label test_fail ""
            label comp_fail ""
            label message ""
          task task08
            label id ""
            label oper_fail ""
            label test_fail ""
            label comp_fail ""
            label message ""
          task task09
            label id ""
            label oper_fail ""
            label test_fail ""
            label comp_fail ""
            label message ""
          task task10
            label id ""
            label oper_fail ""
            label test_fail ""
            label comp_fail ""
            label message ""
          task task11
            label id ""
            label oper_fail ""
            label test_fail ""
            label comp_fail ""
            label message ""
          task task12
            label id ""
            label oper_fail ""
            label test_fail ""
            label comp_fail ""
            label message ""
          task task13
            label id ""
            label oper_fail ""
            label test_fail ""
            label comp_fail ""
            label message ""
          task task14
            label id ""
            label oper_fail ""
            label test_fail ""
            label comp_fail ""
            label message ""
          task task15
            label id ""
            label oper_fail ""
            label test_fail ""
            label comp_fail ""
            label message ""
          task task16
            label id ""
            label oper_fail ""
            label test_fail ""
            label comp_fail ""
            label message ""
          task task17
            label id ""
            label oper_fail ""
            label test_fail ""
            label comp_fail ""
            label message ""
          task task18
            label id ""
            label oper_fail ""
            label test_fail ""
            label comp_fail ""
            label message ""
          task task19
            label id ""
            label oper_fail ""
            label test_fail ""
            label comp_fail ""
            label message ""
        endfamily
      endfamily
    endfamily
  endfamily
  family ecgb
    defstatus suspended
    edit CI_LIMIT '10'
    edit CMAKE_VARS '-DBLAS_LIBRARIES="/usr/local/apps/intel/17.0.3/mkl/lib/intel64/libmkl_intel_lp64.so;/usr/local/apps/intel/17.0.3/mkl/lib/intel64/libmkl_sequential.so;/usr/local/apps/intel/17.0.3/mkl/lib/intel64/libmkl_core.so" -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DECMWF=ON -DEC_BUILD_TIMESTAMP=%BUILDSTAMP% -DENABLE_TESSELATION=OFF -DLAPACK_LIBRARIES="/usr/local/apps/intel/17.0.3/mkl/lib/intel64/libmkl_intel_lp64.so;/usr/local/apps/intel/17.0.3/mkl/lib/intel64/libmkl_sequential.so;/usr/local/apps/intel/17.0.3/mkl/lib/intel64/libmkl_core.so" -DMKL_ROOT=/usr/local/apps/intel/17.0.3/mkl'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit MKL '/usr/local/apps/intel/17.0.3/mkl'
    edit MKLLIBS '/usr/local/apps/intel/17.0.3/mkl/lib/intel64/libmkl_intel_lp64.so;/usr/local/apps/intel/17.0.3/mkl/lib/intel64/libmkl_sequential.so;/usr/local/apps/intel/17.0.3/mkl/lib/intel64/libmkl_core.so'
    edit M_SITE 'ecgb'
    edit QUEUE 'express'
    edit REMOTE_HOST 'ecgb'
    edit SITE 'ecgb'
    edit WSDIR '$SCRATCH/metabuilds/%ECF_HOST%_%ECF_PORT%/ecgb/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task build
      trigger config == complete
      edit ECF_TRIES '2'
      label host ""
      label options ""
    task install
      trigger build == complete
      edit ECF_TRIES '2'
    task remove_write_perm
      trigger install == complete
    task test
      trigger remove_write_perm == complete
      edit ECF_TRIES '2'
      edit TEST_DIR '%BUILDIR%'
    task wipe_build
      trigger test == complete
    task wipe_src
      trigger wipe_build == complete
    family tests
      defstatus complete
      edit COMPARE 'cmp'
      edit NFSHOME '$HOME/tmp/'
      edit OPERSCHOST 'cca'
      limit tests 3
      inlimit tests
      task archive
        edit DATABASE 'marstest'
        edit GRID ''
        edit PARAM '129/130/135/138/155/157'
      task bad_archive
        edit DATABASE 'marstest'
        edit GRID ''
        edit PARAM '129/130/135/138/155/157'
      task environment
      task grib2fdb
      task grib2fdb5
        defstatus complete
      task marsgen
        edit DATABASE 'marstest'
      task memberstates
        edit DATABASE 'marstest'
        edit GRID ''
        edit PARAM '129/130/135/138/155/157'
      task reqcheck
      family retrieve
        edit DATABASE ''
        edit GRID ''
        edit PARAM ''
        task L137
        task accuracy
          edit LEVEL 'levtype=pl, level=1000/500,'
          edit PARAM 'T'
          edit SCRIPT 'accuracy.sms'
        task ai
        task coefficients
          edit GRID '0.1/0.1'
          edit PARAM '129'
        task compute
          edit GRID ''
          edit PARAM ''
        task custom
        task datasets
        task dhs_chk
        task dissemination
        task ems
        task enda
        task ensemble
          edit GRID '2.5/2.5'
          edit PARAM '139/141/142/143/144/145/146/147/151/164/165/166'
        task ep
          edit GRID '5/5'
          edit PARAM '228'
        task era5t_embargo
        task eurobrisa
          defstatus complete
          inlimit /mars_client/limits:eurobrisa
        task exact_match
        task expect
        task fdb
          edit DATABASE 'c2a fdb'
          edit TOLERANCE '0.1'
        task fdb5
          edit DATABASE 'fdb5 prod'
          edit TOLERANCE '0.1'
        task fieldset
          edit GRID '1.5/1.5'
          edit PARAM '27/28/29/30/31/32/33/34/35/36/37/38/39/40/41/42/128/129/134/136/137/139/141/148/151/160/161/162/163/164/165/166/167/168/170/172/173/174/183/186/187/188/198/206/234/235/236/238'
        task fileio
        task fourpoints
        task gauss
          edit GRID '160'
          edit PARAM '31/129/134/136/137/139/141'
        task gauss_regular
          edit GRID '160'
          edit PARAM '31/128/129/134/136/137/139/141'
        task gauss_rotation
        task hirl12p
        task interpolation
          edit GRID '2.5/2.5'
        task intuvp
        task issame
          edit GRID '160'
          edit PARAM '31/128/129/134/136/137/139/141'
        task latlon
          edit GRID '2.5/2.5'
        task mail
        task metgram
        task monthly
        task obs
        task obs_restrict
        task ocean
        task odb_archive
          edit ODB '1'
        task odb_filter
          edit ODB '1'
        task olduv
          edit PARAM 'u/v'
          edit TOLERANCE '0.1'
        task padding
          edit COMPARE 'grib_compare'
        task prob
        task resol
        task rotation
          edit LEVEL 'levtype=pl, level=1000/500,'
          edit PARAM 'T'
        task s2s
        task seas
        task seasonal_restrict
        task sfc
          edit GRIB_COMPARE_OPTIONS '-P -T2 -f'
          edit GRID '1.5/1.5'
          edit PARAM '27/28/29/30/31/32/33/34/35/36/37/38/39/40/41/42/128/129/134/136/137/139/141/148/151/160/161/162/163/164/165/166/167/168/170/172/173/174/183/186/187/188/198/206/234/235/236/238'
        task slwet
        task t1279
        task target
        task tc
        task tigge_area
        task tigge_latlon
        task tigge_nointerp
        task uv
          edit PARAM 'u/v'
          edit TOLERANCE '0.1'
        task vareps
        task vector_wind
        task wave
          edit GRID '1.5/1.5'
          edit PARAM '251'
      endfamily
      task size
        edit PERCENT '10'
      task stage
        edit DATABASE 'marsrd'
        edit GRID '2.5/2.5'
        edit PARAM '129'
      family xtra
        edit COMPARE 'grib_compare -f -P -T2'
        limit xtra 1
        task task00
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task01
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task02
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task03
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task04
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task05
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task06
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task07
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task08
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task09
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task10
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task11
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task12
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task13
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task14
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task15
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task16
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task17
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task18
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task19
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
      endfamily
    endfamily
  endfamily
  family gorax_dev_000
    defstatus suspended
    edit BANNER_TASK_END 'echo %TASK%-end'
    edit BANNER_TASK_START 'echo %TASK%-start'
    edit BASEPREFIX '/data/webdev/deploy/install/%SITE%'
    edit CI_LIMIT '2'
    edit CMAKE_VARS '-DBUNDLE_SKIP_FDB=ON -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DECMWF=ON -DEC_BUILD_TIMESTAMP=%BUILDSTAMP% -DENABLE_FDB=OFF -DENABLE_TESSELATION=OFF'
    edit ECF_CLIENT_EXE_PATH '/home/atls/deploy/install/ecflow/5.3.0/bin/ecflow_client'
    edit ECF_LOGHOST 'gorax-dev-000'
    edit ECF_LOGPORT '38562'
    edit ECF_OUT '/home/atls/deploy/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX ''
    edit LOGDIR '%ECF_OUT%'
    edit MODULE_ACTIONS ''
    edit MODULE_ACTIONS_HDR 'empty.h'
    edit MODULE_CLEANUP ''
    edit M_SITE 'gorax_dev_000'
    edit REMOTE_HOST 'gorax-dev-000'
    edit SITE 'gorax_dev_000'
    edit WSDIR '/data/webdev/%ECF_HOST%_%ECF_PORT%/ecflow-metab/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task build
      trigger config == complete
      edit ECF_TRIES '2'
      label host ""
      label options ""
    task package_source
      trigger build == complete
    task wipe_build
      trigger package_source == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family gorax_dev_001
    defstatus suspended
    edit BANNER_TASK_END 'echo %TASK%-end'
    edit BANNER_TASK_START 'echo %TASK%-start'
    edit BASEPREFIX '/data/webdev/deploy/install/%SITE%'
    edit CI_LIMIT '2'
    edit CMAKE_VARS '-DBUNDLE_SKIP_FDB=ON -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCPACK_PACKAGING_INSTALL_PREFIX=/opt/ecmwf/%CPACK_PACKAGE_NAME% -DECMWF=ON -DEC_BUILD_TIMESTAMP=%BUILDSTAMP% -DENABLE_FDB=OFF -DENABLE_PYTHON=OFF -DENABLE_TESSELATION=OFF'
    edit ECBUILD_CLONE_COMMIT '%ECBUILD_GIT_TAG%'
    edit ECF_CLIENT_EXE_PATH '/home/atls/deploy/install/ecflow/5.3.0/bin/ecflow_client'
    edit ECF_LOGHOST 'gorax-dev-001'
    edit ECF_LOGPORT '38562'
    edit ECF_OUT '/home/atls/deploy/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX ''
    edit LOGDIR '%ECF_OUT%'
    edit MODULE_ACTIONS ''
    edit MODULE_ACTIONS_HDR 'empty.h'
    edit MODULE_CLEANUP ''
    edit M_SITE 'gorax_dev_001'
    edit REMOTE_HOST 'gorax-dev-001'
    edit RPM_REPO_DESTINATION 'https://nexus.ecmwf.int/repository/CentOS-staging/7/8/rpms/%RPM_ARCH%/%RPM_FILENAME%'
    edit SITE 'gorax_dev_001'
    edit WSDIR '/data/webdev/%ECF_HOST%_%ECF_PORT%/ecflow-metab/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task build
      trigger config == complete
      edit ECF_TRIES '2'
      label host ""
      label options ""
    task make_rpm
      trigger build == complete
    task wipe_build
      trigger make_rpm == complete
    task wipe_src
      trigger wipe_build == complete
  endfamily
  family leap42
    defstatus suspended
    edit CI_LIMIT '7'
    edit CMAKE_VARS '-DBLAS_LIBRARIES="/usr/local/apps/intel/18.0.1/mkl/lib/intel64/libmkl_intel_lp64.so;/usr/local/apps/intel/18.0.1/mkl/lib/intel64/libmkl_sequential.so;/usr/local/apps/intel/18.0.1/mkl/lib/intel64/libmkl_core.so" -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DECMWF=ON -DEC_BUILD_TIMESTAMP=%BUILDSTAMP% -DENABLE_TESSELATION=OFF -DLAPACK_LIBRARIES="/usr/local/apps/intel/18.0.1/mkl/lib/intel64/libmkl_intel_lp64.so;/usr/local/apps/intel/18.0.1/mkl/lib/intel64/libmkl_sequential.so;/usr/local/apps/intel/18.0.1/mkl/lib/intel64/libmkl_core.so" -DMKL_ROOT=/usr/local/apps/intel/18.0.1/mkl'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit MKL '/usr/local/apps/intel/18.0.1/mkl'
    edit MKLLIBS '/usr/local/apps/intel/18.0.1/mkl/lib/intel64/libmkl_intel_lp64.so;/usr/local/apps/intel/18.0.1/mkl/lib/intel64/libmkl_sequential.so;/usr/local/apps/intel/18.0.1/mkl/lib/intel64/libmkl_core.so'
    edit M_SITE 'leap42'
    edit REMOTE_HOST 'leap42-deploy'
    edit SITE 'leap42'
    edit WSDIR '/tmp/metabuilds/%ECF_HOST%_%ECF_PORT%/leap42/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task build
      trigger config == complete
      edit ECF_TRIES '2'
      label host ""
      label options ""
    task install
      trigger build == complete
      edit ECF_TRIES '2'
    task remove_write_perm
      trigger install == complete
    task test
      trigger remove_write_perm == complete
      edit ECF_TRIES '2'
      edit TEST_DIR '%BUILDIR%'
    task wipe_build
      trigger test == complete
    task wipe_src
      trigger wipe_build == complete
    family tests
      defstatus complete
      edit COMPARE 'cmp'
      edit NFSHOME '$HOME/tmp/'
      edit OPERSCHOST 'cca'
      limit tests 3
      inlimit tests
      task archive
        edit DATABASE 'marstest'
        edit GRID ''
        edit PARAM '129/130/135/138/155/157'
      task bad_archive
        edit DATABASE 'marstest'
        edit GRID ''
        edit PARAM '129/130/135/138/155/157'
      task environment
      task grib2fdb
      task grib2fdb5
        defstatus complete
      task marsgen
        edit DATABASE 'marstest'
      task memberstates
        edit DATABASE 'marstest'
        edit GRID ''
        edit PARAM '129/130/135/138/155/157'
      task reqcheck
      family retrieve
        edit DATABASE ''
        edit GRID ''
        edit PARAM ''
        task L137
        task accuracy
          edit LEVEL 'levtype=pl, level=1000/500,'
          edit PARAM 'T'
          edit SCRIPT 'accuracy.sms'
        task ai
        task coefficients
          edit GRID '0.1/0.1'
          edit PARAM '129'
        task compute
          edit GRID ''
          edit PARAM ''
        task custom
        task datasets
        task dhs_chk
        task dissemination
        task ems
        task enda
        task ensemble
          edit GRID '2.5/2.5'
          edit PARAM '139/141/142/143/144/145/146/147/151/164/165/166'
        task ep
          edit GRID '5/5'
          edit PARAM '228'
        task era5t_embargo
        task eurobrisa
          inlimit /mars_client/limits:eurobrisa
        task exact_match
        task expect
        task fdb
          edit DATABASE 'c2a fdb'
          edit TOLERANCE '0.1'
        task fdb5
          edit DATABASE 'fdb5 prod'
          edit TOLERANCE '0.1'
        task fieldset
          edit GRID '1.5/1.5'
          edit PARAM '27/28/29/30/31/32/33/34/35/36/37/38/39/40/41/42/128/129/134/136/137/139/141/148/151/160/161/162/163/164/165/166/167/168/170/172/173/174/183/186/187/188/198/206/234/235/236/238'
        task fileio
        task fourpoints
        task gauss
          edit GRID '160'
          edit PARAM '31/129/134/136/137/139/141'
        task gauss_regular
          edit GRID '160'
          edit PARAM '31/128/129/134/136/137/139/141'
        task gauss_rotation
        task hirl12p
        task interpolation
          edit GRID '2.5/2.5'
        task intuvp
        task issame
          edit GRID '160'
          edit PARAM '31/128/129/134/136/137/139/141'
        task latlon
          edit GRID '2.5/2.5'
        task mail
        task metgram
        task monthly
        task obs
        task obs_restrict
        task ocean
        task odb_archive
          edit ODB '1'
        task odb_filter
          edit ODB '1'
        task olduv
          edit PARAM 'u/v'
          edit TOLERANCE '0.1'
        task padding
          edit COMPARE 'grib_compare'
        task prob
        task resol
        task rotation
          edit LEVEL 'levtype=pl, level=1000/500,'
          edit PARAM 'T'
        task s2s
        task seas
        task seasonal_restrict
        task sfc
          edit GRIB_COMPARE_OPTIONS '-P -T2 -f'
          edit GRID '1.5/1.5'
          edit PARAM '27/28/29/30/31/32/33/34/35/36/37/38/39/40/41/42/128/129/134/136/137/139/141/148/151/160/161/162/163/164/165/166/167/168/170/172/173/174/183/186/187/188/198/206/234/235/236/238'
        task slwet
        task t1279
        task target
        task tc
        task tigge_area
        task tigge_latlon
        task tigge_nointerp
        task uv
          edit PARAM 'u/v'
          edit TOLERANCE '0.1'
        task vareps
        task vector_wind
        task wave
          edit GRID '1.5/1.5'
          edit PARAM '251'
      endfamily
      task size
        edit PERCENT '10'
      task stage
        edit DATABASE 'marsrd'
        edit GRID '2.5/2.5'
        edit PARAM '129'
      family xtra
        edit COMPARE 'grib_compare -f -P -T2'
        limit xtra 1
        task task00
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task01
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task02
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task03
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task04
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task05
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task06
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task07
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task08
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task09
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task10
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task11
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task12
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task13
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task14
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task15
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task16
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task17
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task18
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task19
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
      endfamily
    endfamily
  endfamily
  family lxc
    defstatus suspended
    edit CI_LIMIT '10'
    edit CMAKE_VARS '-DBLAS_LIBRARIES="/usr/local/apps/intel/17.0.3/mkl/lib/intel64/libmkl_intel_lp64.so;/usr/local/apps/intel/17.0.3/mkl/lib/intel64/libmkl_sequential.so;/usr/local/apps/intel/17.0.3/mkl/lib/intel64/libmkl_core.so" -DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DECMWF=ON -DEC_BUILD_TIMESTAMP=%BUILDSTAMP% -DENABLE_TESSELATION=OFF -DLAPACK_LIBRARIES="/usr/local/apps/intel/17.0.3/mkl/lib/intel64/libmkl_intel_lp64.so;/usr/local/apps/intel/17.0.3/mkl/lib/intel64/libmkl_sequential.so;/usr/local/apps/intel/17.0.3/mkl/lib/intel64/libmkl_core.so" -DMKL_ROOT=/usr/local/apps/intel/17.0.3/mkl'
    edit ECF_JOB_CMD '/home/ma/emos/bin/trimurti.41r1 %USER% %REMOTE_HOST% %ECF_JOB% %ECF_JOBOUT%'
    edit ECF_KILL_CMD '/home/ma/emos/bin/trimurti.41r1 %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% kill'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit ECF_STATUS_CMD '/home/ma/emos/bin/trimurti.41r1 %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% stat'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit MKL '/usr/local/apps/intel/17.0.3/mkl'
    edit MKLLIBS '/usr/local/apps/intel/17.0.3/mkl/lib/intel64/libmkl_intel_lp64.so;/usr/local/apps/intel/17.0.3/mkl/lib/intel64/libmkl_sequential.so;/usr/local/apps/intel/17.0.3/mkl/lib/intel64/libmkl_core.so'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module load hdf5" "module load netcdf4" "module load fftw/%FFTW_VERSION%" "module load swig/%SWIG_VERSION%"'
    edit M_SITE 'lxc'
    edit QUEUE 'marstest'
    edit REMOTE_HOST 'lxc'
    edit SITE 'lxc'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxc/%COMPILER:%'
    task git_srcs
      edit ECF_TRIES '3'
      label branch ""
      label source ""
    task make_build_dir
      trigger git_srcs == complete
      label build ""
      label reuse ""
    task config
      trigger make_build_dir == complete
      edit ECF_TRIES '2'
      label ecbuild ""
      label version ""
      label install ""
    task build
      trigger config == complete
      edit ECF_TRIES '2'
      label host ""
      label options ""
    task install
      trigger build == complete
      edit ECF_TRIES '2'
    task remove_write_perm
      trigger install == complete
    task test
      trigger remove_write_perm == complete
      edit ECF_TRIES '2'
      edit TEST_DIR '%BUILDIR%'
    task wipe_build
      trigger test == complete
    task wipe_src
      trigger wipe_build == complete
    family tests
      defstatus complete
      edit COMPARE 'cmp'
      edit NFSHOME '$HOME/tmp/'
      edit OPERSCHOST 'cca'
      limit tests 3
      inlimit tests
      task archive
        edit DATABASE 'marstest'
        edit GRID ''
        edit PARAM '129/130/135/138/155/157'
      task bad_archive
        edit DATABASE 'marstest'
        edit GRID ''
        edit PARAM '129/130/135/138/155/157'
      task environment
      task grib2fdb
      task grib2fdb5
        defstatus complete
      task marsgen
        edit DATABASE 'marstest'
      task memberstates
        edit DATABASE 'marstest'
        edit GRID ''
        edit PARAM '129/130/135/138/155/157'
      task reqcheck
      family retrieve
        edit DATABASE ''
        edit GRID ''
        edit PARAM ''
        task L137
        task accuracy
          edit LEVEL 'levtype=pl, level=1000/500,'
          edit PARAM 'T'
          edit SCRIPT 'accuracy.sms'
        task ai
        task coefficients
          edit GRID '0.1/0.1'
          edit PARAM '129'
        task compute
          edit GRID ''
          edit PARAM ''
        task custom
        task datasets
        task dhs_chk
        task dissemination
        task ems
        task enda
        task ensemble
          edit GRID '2.5/2.5'
          edit PARAM '139/141/142/143/144/145/146/147/151/164/165/166'
        task ep
          edit GRID '5/5'
          edit PARAM '228'
        task era5t_embargo
        task eurobrisa
          defstatus complete
          inlimit /mars_client/limits:eurobrisa
        task exact_match
        task expect
        task fdb
          edit DATABASE 'c2a fdb'
          edit TOLERANCE '0.1'
        task fdb5
          edit DATABASE 'fdb5 prod'
          edit TOLERANCE '0.1'
        task fieldset
          edit GRID '1.5/1.5'
          edit PARAM '27/28/29/30/31/32/33/34/35/36/37/38/39/40/41/42/128/129/134/136/137/139/141/148/151/160/161/162/163/164/165/166/167/168/170/172/173/174/183/186/187/188/198/206/234/235/236/238'
        task fileio
        task fourpoints
        task gauss
          edit GRID '160'
          edit PARAM '31/129/134/136/137/139/141'
        task gauss_regular
          edit GRID '160'
          edit PARAM '31/128/129/134/136/137/139/141'
        task gauss_rotation
        task hirl12p
        task interpolation
          edit GRID '2.5/2.5'
        task intuvp
        task issame
          edit GRID '160'
          edit PARAM '31/128/129/134/136/137/139/141'
        task latlon
          edit GRID '2.5/2.5'
        task mail
        task metgram
        task monthly
        task obs
        task obs_restrict
        task ocean
        task odb_archive
          edit ODB '1'
        task odb_filter
          edit ODB '1'
        task olduv
          edit PARAM 'u/v'
          edit TOLERANCE '0.1'
        task padding
          edit COMPARE 'grib_compare'
        task prob
        task resol
        task rotation
          edit LEVEL 'levtype=pl, level=1000/500,'
          edit PARAM 'T'
        task s2s
        task seas
        task seasonal_restrict
        task sfc
          edit GRIB_COMPARE_OPTIONS '-P -T2 -f'
          edit GRID '1.5/1.5'
          edit PARAM '27/28/29/30/31/32/33/34/35/36/37/38/39/40/41/42/128/129/134/136/137/139/141/148/151/160/161/162/163/164/165/166/167/168/170/172/173/174/183/186/187/188/198/206/234/235/236/238'
        task slwet
        task t1279
        task target
        task tc
        task tigge_area
        task tigge_latlon
        task tigge_nointerp
        task uv
          edit PARAM 'u/v'
          edit TOLERANCE '0.1'
        task vareps
        task vector_wind
        task wave
          edit GRID '1.5/1.5'
          edit PARAM '251'
      endfamily
      task size
        edit PERCENT '10'
      task stage
        edit DATABASE 'marsrd'
        edit GRID '2.5/2.5'
        edit PARAM '129'
      family xtra
        edit COMPARE 'grib_compare -f -P -T2'
        limit xtra 1
        task task00
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task01
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task02
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task03
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task04
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task05
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task06
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task07
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task08
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task09
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task10
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task11
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task12
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task13
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task14
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task15
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task16
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task17
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task18
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
        task task19
          label id ""
          label oper_fail ""
          label test_fail ""
          label comp_fail ""
          label message ""
      endfamily
    endfamily
  endfamily
endsuite
suite ifs
  edit ADDRESS_SANITIZER 'False'
  edit ADD_ACTION_TRIGGERS 'True'
  edit AEC_VERSION '1.0.4'
  edit ALLOW_INSTALL_TO_OVERWRITE 'True'
  edit ASAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/asan.txt'
  edit ATLAS_GIT_TAG 'develop'
  edit ATLAS_VERSION '0.20.2'
  edit BANNER_TASK_END 'banner %TASK%-end || true'
  edit BANNER_TASK_START 'banner %TASK%-start || true'
  edit BASEDIR '%WSDIR%/%PACKAGE%'
  edit BASEPREFIX '$PERM/install/%SITE%'
  edit BUFRDC_EMOSLIB_VERSION '4.6.3'
  edit BUFRDC_TABLES_GIT_TAG 'develop'
  edit BUFRDC_TABLES_VERSION '000415'
  edit BUILDIR '%BASEDIR%/builds'
  edit BUILD_ALWAYS_PARALLEL 'False'
  edit CFI_SANITIZER 'False'
  edit CFL_BASE_URL 'https://software-test.ecmwf.int/wiki'
  edit CFL_SPACE_KEY '<TODO>'
  edit CFL_TARBALL_COMMENT ''
  edit CFL_TITLE 'Releases'
  edit CHECK_SYNC_RELEASE 'False'
  edit CLANG_TIDY_ABORT_ON_ERROR 'False'
  edit CMAKE '%ECBUILD%'
  edit CMAKELIST_EXTRACT_PACKAGE_VERSION '$(awk '/^project/ && /%PACKAGE%/ && /VERSION/ {for (I=1;I<=NF;I++) if ($I == "VERSION") {print $(I+1)};}' CMakeLists.txt)'
  edit CMAKE_BUILD_TYPE 'RelWithDebInfo'
  edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE%'
  edit CMAKE_VERSION '3.13.4'
  edit CODES_UI_GIT_TAG 'develop'
  edit CODES_UI_VERSION '1.3.1'
  edit CONFIG_OPTIONS '--cmake="ENABLE_PYTHON=OFF"'
  edit CPACK 'make package'
  edit CPUS '$(lscpu -p | grep -v '#' | wc -l)'
  edit CRAY_FFTW_VERSION '3.3.8.1'
  edit CRAY_NETCDF_VERSION '4.6.1.3'
  edit CRON_BASEPREFIX '%BASEDIR%/install'
  edit CRON_EMAIL ''
  edit CRON_ENABLED 'False'
  edit CRON_GIT_TAG 'develop'
  edit CUSTOM_COMMANDS ''
  edit DATAFLOW_SANITIZER 'False'
  edit DEFAULT_GNU_COMPILER 'gnu.73'
  edit ECBUILD '%BASEPREFIX%/ecbuild/develop/bin/ecbuild'
  edit ECBUILD_CLONE 'False'
  edit ECBUILD_CLONE_COMMIT 'master'
  edit ECBUILD_GIT_TAG 'develop'
  edit ECBUILD_REPO '%GIT_REPO_PREFIX%/ecsdk/ecbuild.git'
  edit ECBUILD_VERSION '3.3.2'
  edit ECCODES_EXTENSIONS_GIT_TAG 'develop'
  edit ECCODES_EXTENSIONS_VERSION '2.17.0'
  edit ECCODES_GIT_TAG 'develop'
  edit ECCODES_PYTHON_VERSION '0.9.6'
  edit ECCODES_VERSION '2.17.1'
  edit ECFLOW5_LATEST_VERSION '5.4.0'
  edit ECFLOW_VERSION '5.4.1'
  edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/ifs'
  edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/ifs'
  edit ECF_JOB_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_JOB% %ECF_JOBOUT%'
  edit ECF_KILL_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% kill'
  edit ECF_STATUS_CMD '/home/ma/emos/bin/trimurti %USER% %REMOTE_HOST% %ECF_RID% %ECF_JOB% %ECF_JOBOUT% stat'
  edit ECF_TRIES '1'
  edit ECKIT_GIT_TAG 'develop'
  edit ECKIT_VERSION '1.10.1'
  edit ECQUIS_VERSION '1.0.0'
  edit EMOS_MOD_VERSION '462'
  edit EMOS_PA_VERSION '000462'
  edit EXPORTS '"export ECACCOUNT=ecodmdma"'
  edit EXPORT_VARIABLES 'False'
  edit FCKIT_GIT_TAG '0.6.3'
  edit FCKIT_VERSION '0.6.3'
  edit FDB_VINTAGE_GIT_TAG '4.10.0'
  edit FDB_VINTAGE_VERSION '4.10.0'
  edit FFTW_VERSION '3.3.8'
  edit FFTW_VERSION_CRAY '3.3.4.11'
  edit GIT 'git'
  edit GITHUB_PACKAGE_NAME '%PACKAGE_NAME%'
  edit GITHUB_REPO 'git@github.com:ecmwf'
  edit GITREPO '%GIT_REPO_PREFIX%/IFS/ifs-bundle.git'
  edit GIT_CLONE 'True'
  edit GIT_CLONE_OPTIONS '--depth=1 --branch %GIT_TAG%'
  edit GIT_METABUILDER_TAG 'ci'
  edit GIT_REPO_PREFIX 'ssh://git@git.ecmwf.int:7999'
  edit GIT_TAG 'unstable'
  edit GIT_VERSION '2.26.0'
  edit GRIB_API_VERSION '1.28.0'
  edit IFS_COMPILER_FAMILY '`echo %COMPILER% | cut -d. -f1 | tr '[:upper:]' '[:lower:]'`'
  edit IFS_COMPILER_VERSION '$EC_COMPILER_VERSION'
  edit IFS_TEST_EXCLUDES '.*-gpsro-grace-a'
  edit IFS_TEST_LABELS 'ifs_test b2o cope'
  edit INSTALLABLE_TO_USR_LOCAL_APPS 'True'
  edit IWYU_VERSION '0.11'
  edit KSH_EXTRACT_PACKAGE_VERSION '`grep _VERSION_STR VERSION.cmake | grep -o '[0-9]*\.[0-9]*\.[0-9]*'`'
  edit LAPACK_VERSION '3.5.0'
  edit LIBEMOS_GIT_TAG 'develop'
  edit LIBEMOS_VERSION '4.6.2'
  edit LINTERS_REFERENCE_DIRECTORY '/perm/ma/ma0/linters/%PACKAGE%/${PACKAGE_VERSION}'
  edit LSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/lsan.txt'
  edit MAGICS_GIT_TAG 'develop'
  edit MAGICS_PYTHON_VERSION 'master'
  edit MAGICS_VERSION '4.3.3'
  edit MAKE 'make'
  edit MARS_CLIENT_GIT_TAG '6.27.2.2'
  edit MARS_CLIENT_PACKAGE_GIT_TAG '6.27.6'
  edit MARS_CLIENT_PACKAGE_VERSION '6.27.6'
  edit MARS_CLIENT_VERSION '6.27.2.2'
  edit MEMORY_SANITIZER 'False'
  edit METABUILDER_HOME '/var/tmp/ma0/workspace/metabuilder'
  edit METAB_INSTALL_REMOVE_WRITE_PERMISSIONS 'True'
  edit METAB_PYTHON_VERSION 'python3'
  edit METKIT_GIT_TAG 'develop'
  edit METKIT_VERSION '1.4.6'
  edit METVIEW_GIT_TAG 'develop'
  edit METVIEW_PYTHON_VERSION '1.4.2'
  edit METVIEW_VERSION '5.8.3'
  edit MIR_BUNDLE_VERSION '0.9.0.0'
  edit MIR_GIT_TAG 'develop'
  edit MIR_VERSION '1.5.0'
  edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
  edit MODULE_ACTIONS_FIRST ''
  edit MODULE_ACTIONS_HDR 'module_actions.h'
  edit MODULE_CHECKING_HDR 'module_checking.h'
  edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow"'
  edit MODULE_CYCLE '`echo %MODULE_VERSION% | cut -d. -f1`'
  edit MODULE_LOAD_SUPPORTED 'True'
  edit MODULE_NAME 'ifs'
  edit MODULE_VERSION '`echo %GIT_TAG% | tr _ .`'
  edit NETCDF4_VERSION '4.7.4'
  edit ODB_API_GIT_TAG 'develop'
  edit ODB_API_VERSION '0.19.4'
  edit ODB_GIT_TAG 'develop'
  edit ODB_TOOLS_GIT_TAG 'develop'
  edit ODB_TOOLS_VERSION '0.18.3'
  edit ODB_VERSION '1.0.11'
  edit ODC_GIT_TAG 'develop'
  edit ODC_VERSION '1.0.3'
  edit OPENJPEG_VERSION '2.3.1'
  edit PACKAGE 'ifs'
  edit PACKAGE_NAME 'ifs'
  edit PACKAGE_OWNERS 'ma0@ecmwf.int'
  edit PACKAGE_VERSION '%GIT_TAG%'
  edit PGEN_VERSION '1.7.1.0'
  edit PKG_BOOST_VERSION '1.71.0'
  edit PREFIX_APPEND ''
  edit PROJ4_VERSION '5.2.0'
  edit PROJ_VERSION '6.3.1'
  edit PYTHON2_VERSION '2.7.15-01'
  edit PYTHON3_VERSION '3.6.10-01'
  edit PYTHON3_VERSION_MAJOR_MINOR '3.6'
  edit PYTHON_GITREPO '%GIT_REPO_PREFIX%/eccodes/%PYTHON_PACKAGE%.git'
  edit PYTHON_PACKAGE 'eccodes-python'
  edit PYTHON_PACKAGE_VERSION 'develop'
  edit QT_VERSION '5.12.3'
  edit QUEUE 'build'
  edit REMOTE_HOST 'polonius'
  edit REUSE_BUILD_DIR 'False'
  edit SAFE_STACK_SANITIZER 'False'
  edit SANITIZER 'False'
  edit SANITIZER_REFERENCE_DIRECTORY '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%'
  edit SRCDIR '%BASEDIR%/%PACKAGE%'
  edit STATIC_ANALYSER_ABORT_ON_ERROR 'False'
  edit SWIG_VERSION '3.0.8'
  edit SYNC_RELEASE '2020.04.0'
  edit SYNC_RELEASE_CI_STAGE 'develop'
  edit TARBALL_DIR '/perm/ma/ma0/tmp_tarballs/ci'
  edit TAR_PACKAGE '%PACKAGE%'
  edit TEST 'ctest --output-on-failure -j${CPUS} '
  edit THREAD_SANITIZER 'False'
  edit TRANSI_GIT_TAG 'develop'
  edit TRANSI_VERSION '0.7.1'
  edit TSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/tsan.txt'
  edit UBSAN_FILE '/perm/ma/ma0/sanitizer/%PACKAGE%/${PACKAGE_VERSION}/%COMPILER:%/ubsan.txt'
  edit UB_SANITIZER 'False'
  edit USER 'ma0'
  edit USE_WORKDIR 'False'
  edit VALGRIND_VERSION '3.15.0'
  edit WORKDIR '%BASEDIR%/workdir/%ECF_NAME%'
  edit WSDIR '/var/tmp/tmpdir/deploy/jtmp.20739'
  label branch "unstable"
  task regenerate
    defstatus complete
    edit ECF_JOB_CMD '%ECF_JOB% > %ECF_JOBOUT% 2>&1'
    edit ECF_KILL_CMD 'kill -15 %ECF_RID%'
    edit ECF_PASS 'FREE'
    edit ECF_STATUS_CMD 'ps --sid %ECF_RID% -f'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
    label date ""
    label metabuilder_branch "ci"
  task set_module_default
    defstatus suspended
    edit DRY_RUN ''
    edit MODULEMGR_OPTIONS ''
    edit MODULE_ACTIONS '"module load modulemgr"'
    edit M_SITES 'cca,ccb,ecgb,leap42,lxc'
    edit REMOTE_HOST 'polonius'
    edit WORKDIR ''
  task set_module_cycle
    defstatus suspended
    edit DRY_RUN ''
    edit MODULEMGR_OPTIONS ''
    edit MODULE_ACTIONS '"module load modulemgr"'
    edit M_SITES 'cca,ccb,ecgb,leap42,lxc'
    edit REMOTE_HOST 'polonius'
    edit WORKDIR ''
  family cca
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DLAPACK_FOUND=TRUE'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/ifs/cca'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/ifs/cca:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'cca-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '43000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'cca'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'cca'
    edit SITE 'cca'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family cray.85
      edit COMPILER 'CRAY.85'
      edit IFS_COMPILER_VERSION '$CRAY_FTN_VERSION'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task var_summary
        label date ""
        label installed ""
      task git_srcs
        trigger var_summary == complete
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit QUEUE 'np'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit COMPILER 'CRAY.87'
      edit IFS_COMPILER_VERSION '$CRAY_FTN_VERSION'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task var_summary
        label date ""
        label installed ""
      task git_srcs
        trigger var_summary == complete
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit QUEUE 'np'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family ccb
    defstatus suspended
    edit ACCOUNT 'ecodmdma'
    edit CI_LIMIT '20'
    edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_CXX_COMPILER=CC -DCMAKE_C_COMPILER=cc -DCMAKE_Fortran_COMPILER=ftn -DLAPACK_FOUND=TRUE'
    edit CPUS '%EC_TOTAL_TASKS%'
    edit ECF_HOME '/var/tmp/ma0/workspace/metabuilder/suites/ifs/ccb'
    edit ECF_INCLUDE '/var/tmp/ma0/workspace/metabuilder/suites/ifs/ccb:/home/ma/emos/def/o/include'
    edit ECF_LOGHOST 'ccb-log1'
    edit ECF_LOGPORT '9316'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit EC_TOTAL_TASKS '16'
    edit EXPORTS '"export ECACCOUNT=ecodmdma" "export TMPDIR=/tmp"'
    edit FFTW_VERSION '3.3.4.11'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/`prgenvinfo`'
    edit LOGDIR '%ECF_OUT%'
    edit MEM '43000'
    edit MODULE_CLEANUP '"module unload emos" "module unload fftw" "module unload grib_api" "module unload eccodes" "module unload git" "module unload python" "module unload python3" "module unload metview" "module unload ecflow" "module unload atp"'
    edit M_SITE 'ccb'
    edit QSUB_LP ''
    edit QUEUE 'ns'
    edit REMOTE_HOST 'ccb'
    edit SITE 'ccb'
    edit SMSFILES 'ECF_FILES'
    edit SMSHOME ''
    edit SMSINCLUDE ''
    edit SMSJOBOUT '%ECF_JOBOUT%'
    edit SMSNAME '%ECF_NAME%'
    edit SMSNODE '0'
    edit SMSOUT '%ECF_OUT%'
    edit SMSPASS '1'
    edit SMSTRYNO '%ECF_TRYNO%'
    edit SMS_PROG '0'
    edit WSDIR '/scratch/ma/ma0/metabuilds/%ECF_HOST%_%ECF_PORT%/%REMOTE_HOST%/%COMPILER:%'
    edit WSHOST 'polonius'
    family cray.85
      edit COMPILER 'CRAY.85'
      edit IFS_COMPILER_VERSION '$CRAY_FTN_VERSION'
      edit MODULE_ACTIONS '"module swap cdt/17.03" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task var_summary
        label date ""
        label installed ""
      task git_srcs
        trigger var_summary == complete
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit QUEUE 'np'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
    family cray.87
      edit COMPILER 'CRAY.87'
      edit IFS_COMPILER_VERSION '$CRAY_FTN_VERSION'
      edit MODULE_ACTIONS '"module swap cdt/18.12" "module load gcc/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task var_summary
        label date ""
        label installed ""
      task git_srcs
        trigger var_summary == complete
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit QUEUE 'np'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family ecgb
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'ecgb'
    edit QUEUE 'express'
    edit REMOTE_HOST 'ecgb'
    edit SITE 'ecgb'
    edit WSDIR '$SCRATCH/metabuilds/%ECF_HOST%_%ECF_PORT%/ecgb/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task var_summary
        label date ""
        label installed ""
      task git_srcs
        trigger var_summary == complete
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit QUEUE 'parallel'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family leap42
    defstatus suspended
    edit CI_LIMIT '7'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit MODULE_ACTIONS '"module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module load netcdf4/%NETCDF4_VERSION%"'
    edit M_SITE 'leap42'
    edit REMOTE_HOST 'leap42-deploy'
    edit SITE 'leap42'
    edit WSDIR '/tmp/metabuilds/%ECF_HOST%_%ECF_PORT%/leap42/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON'
      edit COMPILER 'GNU.73'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%" "module load netcdf4/%NETCDF4_VERSION%"'
      task var_summary
        label date ""
        label installed ""
      task git_srcs
        trigger var_summary == complete
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
  family lxc
    defstatus suspended
    edit CI_LIMIT '10'
    edit ECF_OUT '/scratch/ma/ma0/ecf_out/%ECF_HOST%_%ECF_PORT%'
    edit INSTALL_DIR '%BASEPREFIX%/%PACKAGE%/develop%INSTALL_POSTFIX%'
    edit INSTALL_POSTFIX '/$EC_COMPILER_FAMILY/$EC_COMPILER_VERSION'
    edit LOGDIR '%ECF_OUT%'
    edit M_SITE 'lxc'
    edit QUEUE 'express'
    edit REMOTE_HOST 'lxc'
    edit SITE 'lxc'
    edit WSDIR '$PERM/metabuilds/%ECF_HOST%_%ECF_PORT%/lxc/%COMPILER:%'
    family gnu.73
      edit CMAKE_VARS '-DCMAKE_BUILD_TYPE=%CMAKE_BUILD_TYPE% -DCMAKE_EXPORT_COMPILE_COMMANDS=ON'
      edit COMPILER 'GNU.73'
      edit CUSTOM_COMMANDS '"export ECBUILD_TOOLCHAIN=%SRCDIR%/arch/ecmwf/lxc/gnu/%IFS_COMPILER_VERSION%/toolchain.cmake"'
      edit MODULE_ACTIONS '"module swap gnu/7.3.0" "module load git/%GIT_VERSION%" "module load cmake/%CMAKE_VERSION%" "module load python/%PYTHON2_VERSION%" "module load python3/%PYTHON3_VERSION%" "module load ecflow/%ECF_VERSION%"'
      task var_summary
        label date ""
        label installed ""
      task git_srcs
        trigger var_summary == complete
        edit ECF_TRIES '3'
        label branch ""
        label source ""
      task make_build_dir
        trigger git_srcs == complete
        label build ""
        label reuse ""
      task config
        trigger make_build_dir == complete
        edit ECF_TRIES '2'
        label ecbuild ""
        label version ""
        label install ""
      task build
        trigger config == complete
        edit ECF_TRIES '2'
        label host ""
        label options ""
      task test
        trigger build == complete
        edit ECF_TRIES '2'
        edit QUEUE 'parallel'
        edit TEST_DIR '%BUILDIR%'
      task install
        trigger test == complete
        edit ECF_TRIES '2'
      task install_log
        trigger install == complete
        edit ECF_TRIES '2'
      task wipe_build
        trigger install_log == complete
      task wipe_src
        trigger wipe_build == complete
    endfamily
  endfamily
endsuite
# enddef
