#============================================================
extern an
extern an
extern fc
extern fc
extern eps_sv
extern eps_sv
extern eps_fc
extern eps_fc
extern oceatm
extern oceatm
extern oce
extern oce
extern fpos
extern fpos
extern sens
extern sens
extern wam
extern wam
extern oc
extern oc
extern climplot
extern climplot
extern coup
extern coup
extern an
extern an
extern fc
extern fc
extern eps_sv
extern eps_sv
extern eps_fc
extern eps_fc
extern oceatm
extern oceatm
extern oce
extern oce
extern fpos
extern fpos
extern sens
extern sens
extern wam
extern wam
extern oc
extern oc
extern climplot
extern climplot
extern coup
extern coup
suite cu9
  clock hybrid 0 # 27.11.2008 16:54
  family a01d
    edit SMSTRIES '1'
    edit SMSFILES '/vol/msverify/sms/cu9/a01d'
    edit SMSINCLUDE '/vol/msverify/sms/cu9/a01d/include'
    edit SMSHOME '/vol/msverify/sms'
    edit SMSOUT '/vol/msverify/sms'
    edit SMSLOGPORT '9314'
    edit SMSLOGHOST 'hpce-batch'
    edit FSFAMILY ''
    edit DELTA_DAY '0'
    edit HOUR ''
    edit MEMBER '0'
    edit NPROCA '1'
    edit NPROCB '1'
    edit THREADS '1'
    edit BRANCH ''
    edit PROJECTS 'none'
    edit PARALLEL '4'
    edit DEBUG '0'
    edit PROFILE '0'
    edit DISPLAY 'define_your_own:0.0'
    edit QUEUE 'normal'
    edit SCHOST 'hpce'
    edit WSHOST 'ecgate'
    edit LOGHOST 'linux_cluster'
    edit EXPVER 'a01d'
    edit OWNER 'cu9'
    edit USER 'xbe'
    edit ACCOUNT 'betro3'
    edit YMD '2008021112'
    edit SECS '43200'
    edit SMSCMD 'smssubmit.x %SMSJOB% %SCHOST% %USER%'
    edit SMSKILL 'smskill %USER% %SCHOST% %SMSRID% %SMSJOB%'
    edit SMSSTATUSCMD 'smscheck %SCHOST% %EXPVER% %SMSJOB%'
    edit LOGDIR '/hpce/rdx_dir/msx'
    family make
      edit YMD '2008021112'
      edit HOUR ''
      edit FSFAMILY 'make'
      task setup
        edit QUEUE 'normal'
      family libs
        defstatus complete
        trigger setup eq complete
        edit CCSUBDIR '.'
        task p4setup_ifs
          edit SMSTRIES '2'
        task p4diff_ifs
          trigger p4setup_ifs eq complete
        family compile
          trigger p4setup_ifs eq complete
          family ifsaux
            defstatus complete
            task libifsaux
          endfamily
          family surf
            defstatus complete
            trigger ifsaux eq complete
            task libsurf
          endfamily
          family trans
            defstatus complete
            trigger ifsaux eq complete
            task libtrans
          endfamily
          family ifs
            defstatus complete
            trigger  ( ifsaux eq complete or ifsaux eq unknown) and  ( trans eq complete or trans eq unknown) and  ( surf eq complete or surf eq unknown)
            task libifs
          endfamily
          family wam
            defstatus complete
            trigger ifsaux eq complete
            task libwam
          endfamily
          family odb
            defstatus complete
            family compiler
              task libodbsqlcompiler
            endfamily
            family databases
              trigger compiler eq complete
              family CCMA
                task libCCMA
              endfamily
              family ECMA
                task libECMA
              endfamily
            endfamily
            family corelibs
              trigger ../ifsaux eq complete
              family odb
                task libodb
              endfamily
              family odbdummy
                trigger odb/libodb eq complete
                task libodbdummy
              endfamily
              family odbmain
                trigger odb/libodb eq complete and odbdummy/libodbdummy eq complete and ../../ifs eq complete
                task libodbmain
              endfamily
              family odbport
                trigger odb/libodb eq complete and odbdummy/libodbdummy eq complete and odbmain/libodbmain eq complete and ../../ifs eq complete
                task libodbport
              endfamily
            endfamily
          endfamily
          family ssa
            defstatus complete
            trigger ifsaux eq complete and ifs eq complete and odb/corelibs/odb/libodb eq complete
            task libssa
          endfamily
          family obstat
            defstatus complete
            trigger ifsaux eq complete and odb/corelibs/odb/libodb eq complete
            task libobstat
          endfamily
          family bl
            defstatus complete
            trigger ifsaux eq complete
            task libbl95
          endfamily
          family scat
            defstatus complete
            trigger ifsaux eq complete
            task libscat
          endfamily
          family satrad
            defstatus complete
            trigger ifsaux eq complete
            task libsatrad
          endfamily
          family aeolus
            defstatus complete
            trigger ifsaux eq complete
            task libaeolus
          endfamily
          family prepdata
            defstatus complete
            trigger  ( ifsaux eq complete or ifsaux eq unknown) and  ( trans eq complete or trans eq unknown) and  ( ifs eq complete or ifs eq unknown)
            task libprepdata
          endfamily
        endfamily
      endfamily
      task links
        trigger  ( setup eq complete or setup eq unknown) and  ( libs eq complete or libs eq unknown)
        edit QUEUE 'normal'
      task logfiles
        trigger  ( links eq complete or links eq unknown)
        edit SMSCMD 'smssubmit.x %SMSJOB% %SCHOST% %USER%'
        edit SMSKILL 'smskill %USER% %SCHOST% %SMSRID% %SMSJOB%'
        edit SMSSTATUSCMD 'smscheck %SCHOST% %EXPVER% %SMSJOB%'
        edit LOGDIR '/hpce/rdx_dir/msx'
    endfamily
    family eps_varfc
      limit SC 6
      limit PIC 6
      edit STREAM 'ENFO'
      edit FSFAMILY 'mc'
      edit HOUR ''
      edit SMSCMD 'smssubmit.x %SMSJOB% %SCHOST% %USER%'
      edit SMSKILL 'smskill %USER% %SCHOST% %SMSRID% %SMSJOB%'
      edit SMSSTATUSCMD 'smscheck %SCHOST% %EXPVER% %SMSJOB%'
      edit LOGDIR '/hpce/rdx_dir/msx'
      family make
        trigger ../make eq complete
        edit YMD '2008021112'
        edit HOUR '12'
        task setup
          edit QUEUE 'normal'
        family bins
          trigger ../../make/links eq complete and ./setup eq complete
          task ifs
          task wamabs
            edit QUEUE 'normal'
          task mc_tools
          task prepdata
          task ma_tools
            defstatus complete
        endfamily
        family const
          task datalinks
            trigger ../setup eq complete
          task wconstA
            trigger ../bins/wamabs eq complete
            edit WAMRESOL 'global100'
            edit WAMNFRE '30'
            edit WAMNANG '24'
          task wconstB
            trigger ./wconstA eq complete
            edit WAMRESOL 'global100'
            edit WAMNFRE '25'
            edit WAMNANG '12'
          task wconstC
            trigger ./wconstB eq complete
            edit WAMRESOL 'global300'
            edit WAMNFRE '25'
            edit WAMNANG '12'
          task getsst
            defstatus complete
            trigger ../setup eq complete and ../../main/fc/inigroup/getini eq complete
        endfamily
        task logfiles
          trigger bins eq complete and const eq complete
          edit SMSCMD 'smssubmit.x %SMSJOB% %SCHOST% %USER%'
          edit SMSKILL 'smskill %USER% %SCHOST% %SMSRID% %SMSJOB%'
          edit SMSSTATUSCMD 'smscheck %SCHOST% %EXPVER% %SMSJOB%'
          edit LOGDIR '/hpce/rdx_dir/msx'
      endfamily
      family main
        family sv
          defstatus complete
          repeat string YMD  "2008021112" # status 0
          edit EPSTYPE 'sv'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          task getini
            trigger ../../make/const/datalinks eq complete
            edit QUEUE 'normal'
            edit INISTREAM 'DCDA'
            edit INISTEP '6'
            edit INITYPE 'fc'
            edit PRIORITY '40'
          task inidata_sv
            trigger ../../make/bins/ifs eq complete and getini eq complete
            edit QUEUE 'parallel'
            edit NPES '16'
            edit THREADS '4'
            edit MEM '960'
            edit CPUTIME '1000'
          task svnh
            trigger inidata_sv eq complete
            edit FCLENGTH '48'
            edit MEMBER '0'
            edit EPSTROP '0'
            edit EPSHEM '1'
            edit MEM '960'
            edit NPES '16'
            edit THREADS '2'
            edit CPUTIME '6000'
            edit QUEUE 'normal'
          task svsh
            trigger inidata_sv eq complete
            edit FCLENGTH '48'
            edit MEMBER '0'
            edit EPSTROP '0'
            edit EPSHEM '2'
            edit MEM '960'
            edit NPES '16'
            edit THREADS '2'
            edit CPUTIME '6000'
            edit QUEUE 'normal'
          task targets
            trigger ../../make/bins/mc_tools eq complete
          task subspace
            trigger svnh eq complete and svsh eq complete and targets eq complete
          task sv1
            trigger targets eq complete and subspace eq complete
            edit FCLENGTH '48'
            edit MEMBER '1'
            edit EPSTROP '1'
            edit MEM '960'
            edit NPES '16'
            edit THREADS '2'
            edit CPUTIME '6000'
            edit QUEUE 'normal'
          task sv2
            trigger targets eq complete and subspace eq complete
            edit FCLENGTH '48'
            edit MEMBER '2'
            edit EPSTROP '1'
            edit MEM '960'
            edit NPES '16'
            edit THREADS '2'
            edit CPUTIME '6000'
            edit QUEUE 'normal'
          task sv3
            trigger targets eq complete and subspace eq complete
            edit FCLENGTH '48'
            edit MEMBER '3'
            edit EPSTROP '1'
            edit MEM '960'
            edit NPES '16'
            edit THREADS '2'
            edit CPUTIME '6000'
            edit QUEUE 'normal'
          task sv4
            trigger targets eq complete and subspace eq complete
            edit FCLENGTH '48'
            edit MEMBER '4'
            edit EPSTROP '1'
            edit MEM '960'
            edit NPES '16'
            edit THREADS '2'
            edit CPUTIME '6000'
            edit QUEUE 'normal'
          task sv5
            trigger targets eq complete and subspace eq complete
            edit FCLENGTH '48'
            edit MEMBER '5'
            edit EPSTROP '1'
            edit MEM '960'
            edit NPES '16'
            edit THREADS '2'
            edit CPUTIME '6000'
            edit QUEUE 'normal'
          task sv6
            trigger targets eq complete and subspace eq complete
            edit FCLENGTH '48'
            edit MEMBER '6'
            edit EPSTROP '1'
            edit MEM '960'
            edit NPES '16'
            edit THREADS '2'
            edit CPUTIME '6000'
            edit QUEUE 'normal'
          task svnhevo
            trigger svnh eq complete
            edit FCLENGTH '48'
            edit EPSHEM '1'
            edit MEM '960'
            edit NPES '16'
            edit THREADS '2'
            edit CPUTIME '6000'
            edit QUEUE 'normal'
          task svshevo
            trigger svsh eq complete
            edit FCLENGTH '48'
            edit EPSHEM '2'
            edit MEM '960'
            edit NPES '16'
            edit THREADS '2'
            edit CPUTIME '6000'
            edit QUEUE 'normal'
          task svsave
            trigger svnh eq complete and svsh eq complete and sv1 eq complete and sv2 eq complete and sv3 eq complete and sv4 eq complete and sv5 eq complete and sv6 eq complete and svnhevo eq complete and svshevo eq complete
          task logfiles
            trigger svsave eq complete
            edit SMSCMD 'smssubmit.x %SMSJOB% %SCHOST% %USER%'
            edit SMSKILL 'smskill %USER% %SCHOST% %SMSRID% %SMSJOB%'
            edit SMSSTATUSCMD 'smscheck %SCHOST% %EXPVER% %SMSJOB%'
            edit LOGDIR '/hpce/rdx_dir/msx'
        endfamily
        family fc
          family inigroup
            repeat string YMD  "2008021112" # status 1
            trigger  ( ../sv:YMD gt inigroup:YMD or ../sv eq complete) and  ( inigroup:YMD lt  ( ensemble:YMD + 2))
            complete  ( inigroup:YMD eq  ~ 1)
            task getini
              trigger ../../../make/setup eq complete
              edit QUEUE 'normal'
              edit INISTREAM 'DA'
              edit INISTEP '0'
              edit INITYPE '4v'
              edit PRIORITY '40'
            task inidata
              trigger ../../../make/bins/ifs eq complete and getini eq complete and ../../../make/const/getsst eq complete
              edit EPSFCRES '399'
              edit EPSFCLEV '62'
              edit QUEUE 'parallel'
              edit NPES '16'
              edit THREADS '4'
              edit MEM '960'
              edit CPUTIME '1000'
            task wavini
              trigger ../../../make/const/wconstA eq complete
              edit QUEUE 'normal'
              edit PRIORITY '40'
              edit MEM '900'
              edit NPES '1'
              edit THREADS '1'
              edit CPUTIME '8000'
            task wcold
              defstatus complete
              trigger inidata eq complete and wavini eq complete
              edit QUEUE 'normal'
              edit MEM '900'
              edit NPES '8'
              edit THREADS '1'
              edit CPUTIME '3000'
            task wavfcdata
              trigger ./wcold eq complete and ../../../make eq complete and wavini eq complete
            task getsvs
              trigger ../../../make/bins/mc_tools eq complete
              edit EPSTROP '1'
            task getae
              trigger ./getini eq complete and ./getsvs eq complete
              event 1
            task rot
              trigger ./inidata eq complete and ./getae eq complete and ../../../make/bins/mc_tools eq complete
              edit EPSTROP '1'
            task geticp
              defstatus complete
              trigger getini eq complete
            family pert_ic
              inlimit PIC
              trigger geticp eq complete and rot eq complete and ./inidata eq complete and ../../../make/bins/mc_tools eq complete
              edit EPSTROP '1'
              edit EPSFCLEV '62'
              edit EPSFCRES '399'
              family 000
                edit MEMBER '0'
                task pertinic
              endfamily
              family 001
                edit MEMBER '1'
                task pertinic
              endfamily
              family 002
                edit MEMBER '2'
                task pertinic
              endfamily
              family 003
                edit MEMBER '3'
                task pertinic
              endfamily
              family 004
                edit MEMBER '4'
                task pertinic
              endfamily
              family 005
                edit MEMBER '5'
                task pertinic
              endfamily
              family 006
                edit MEMBER '6'
                task pertinic
              endfamily
              family 007
                edit MEMBER '7'
                task pertinic
              endfamily
              family 008
                edit MEMBER '8'
                task pertinic
              endfamily
              family 009
                edit MEMBER '9'
                task pertinic
              endfamily
              family 010
                edit MEMBER '10'
                task pertinic
              endfamily
              family 011
                edit MEMBER '11'
                task pertinic
              endfamily
              family 012
                edit MEMBER '12'
                task pertinic
              endfamily
              family 013
                edit MEMBER '13'
                task pertinic
              endfamily
              family 014
                edit MEMBER '14'
                task pertinic
              endfamily
              family 015
                edit MEMBER '15'
                task pertinic
              endfamily
              family 016
                edit MEMBER '16'
                task pertinic
              endfamily
              family 017
                edit MEMBER '17'
                task pertinic
              endfamily
              family 018
                edit MEMBER '18'
                task pertinic
              endfamily
              family 019
                edit MEMBER '19'
                task pertinic
              endfamily
              family 020
                edit MEMBER '20'
                task pertinic
              endfamily
              family 021
                edit MEMBER '21'
                task pertinic
              endfamily
              family 022
                edit MEMBER '22'
                task pertinic
              endfamily
              family 023
                edit MEMBER '23'
                task pertinic
              endfamily
              family 024
                edit MEMBER '24'
                task pertinic
              endfamily
              family 025
                edit MEMBER '25'
                task pertinic
              endfamily
              family 026
                edit MEMBER '26'
                task pertinic
              endfamily
              family 027
                edit MEMBER '27'
                task pertinic
              endfamily
              family 028
                edit MEMBER '28'
                task pertinic
              endfamily
              family 029
                edit MEMBER '29'
                task pertinic
              endfamily
              family 030
                edit MEMBER '30'
                task pertinic
              endfamily
              family 031
                edit MEMBER '31'
                task pertinic
              endfamily
              family 032
                edit MEMBER '32'
                task pertinic
              endfamily
              family 033
                edit MEMBER '33'
                task pertinic
              endfamily
              family 034
                edit MEMBER '34'
                task pertinic
              endfamily
              family 035
                edit MEMBER '35'
                task pertinic
              endfamily
              family 036
                edit MEMBER '36'
                task pertinic
              endfamily
              family 037
                edit MEMBER '37'
                task pertinic
              endfamily
              family 038
                edit MEMBER '38'
                task pertinic
              endfamily
              family 039
                edit MEMBER '39'
                task pertinic
              endfamily
              family 040
                edit MEMBER '40'
                task pertinic
              endfamily
              family 041
                edit MEMBER '41'
                task pertinic
              endfamily
              family 042
                edit MEMBER '42'
                task pertinic
              endfamily
              family 043
                edit MEMBER '43'
                task pertinic
              endfamily
              family 044
                edit MEMBER '44'
                task pertinic
              endfamily
              family 045
                edit MEMBER '45'
                task pertinic
              endfamily
              family 046
                edit MEMBER '46'
                task pertinic
              endfamily
              family 047
                edit MEMBER '47'
                task pertinic
              endfamily
              family 048
                edit MEMBER '48'
                task pertinic
              endfamily
              family 049
                edit MEMBER '49'
                task pertinic
              endfamily
              family 050
                edit MEMBER '50'
                task pertinic
              endfamily
            endfamily
            task cp_pert
              trigger pert_ic eq complete
            task logfiles
              trigger wavfcdata eq complete and wavini eq complete and inidata eq complete and pert_ic eq complete and cp_pert eq complete and rot eq complete
              edit SMSCMD 'smssubmit.x %SMSJOB% %SCHOST% %USER%'
              edit SMSKILL 'smskill %USER% %SCHOST% %SMSRID% %SMSJOB%'
              edit SMSSTATUSCMD 'smscheck %SCHOST% %EXPVER% %SMSJOB%'
              edit LOGDIR '/hpce/rdx_dir/msx'
          endfamily
          family ensemble
            inlimit SC
            repeat string YMD  "2008021112" # status 1
            trigger  (  ( inigroup:YMD gt ensemble:YMD) or  ( inigroup eq complete)) and  ( ensemble:YMD eq lag:YMD)
            complete  ( ensemble:YMD eq  ~ 1)
            edit RUNVARFC '1'
            edit HC_REFDATE '0'
            family cv
              defstatus complete
              trigger  ( ../inigroup:YMD gt ../ensemble:YMD) or  ( ../inigroup/pert_ic/000 eq complete)
              edit EPSTYPE 'cv'
              family control_1
                edit SUBFSFAMILY '/cv001'
                edit EPSMEMBER '1'
                edit EPSWAMRESOL 'global100'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSWAMNSTPW '1'
                task getiniLeg
                  edit EPSFCRES '399'
                  edit EPSFCLEV '62'
                  edit EPSFCGTYPE 'l_2'
                  edit PRIORITY '40'
                task modeleps
                  trigger ./getiniLeg eq complete
                  edit MEM '960'
                  edit NPES '24'
                  edit THREADS '4'
                  edit STACK '550'
                  edit QUEUE 'normal'
                  edit CPUTIME '2700'
                  edit EPSFCRES '399'
                  edit EPSFCLEV '62'
                  edit EPSFCGTYPE 'l_2'
                  edit EPSNLEGS '1'
                  edit EPSLEG '1'
                  edit EPSINISTEP '240'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '216'
                  edit TSTEP '1800'
                  meter step 0 216 216
              endfamily
              family control_2
                edit SUBFSFAMILY '/cv002'
                edit EPSMEMBER '2'
                edit EPSWAMRESOL 'global100'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSWAMNSTPW '1'
                task getiniLeg
                  edit PRIORITY '40'
                  edit EPSFCRES '399'
                  edit EPSFCLEV '62'
                task intHtoLB
                  trigger ./getiniLeg eq complete
                  edit QUEUE 'parallel'
                  edit NPES '8'
                  edit THREADS '2'
                  edit MEM '1599'
                  edit CPUTIME '200'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '255'
                  edit EPSFCLEV '62'
                  edit EPSFCGTYPE 'l_2'
                task modeleps
                  trigger ./intHtoLB eq complete
                  edit MEM '960'
                  edit NPES '16'
                  edit THREADS '2'
                  edit STACK '550'
                  edit QUEUE 'normal'
                  edit CPUTIME 'INF'
                  edit EPSNLEGS '1'
                  edit EPSLEG '1'
                  edit EPSFCRES '255'
                  edit EPSFCLEV '62'
                  edit EPSFCGTYPE 'l_2'
                  edit EPSINISTEP '0'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '216'
                  edit TSTEP '2700'
                  meter step 0 216 216
              endfamily
            endfamily
            family cf
              edit EPSTYPE 'cf'
              family control
                trigger  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/000 eq complete)
                edit SUBFSFAMILY '/cf000'
                edit EPSMEMBER '0'
                family legA
                  edit EPSCOUPLE '0'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                    edit EPSLEG '2'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an0'
                  edit FCGROUP 'fc0'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                    edit EPSLEG '3'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit EPSFCLENGTH_B '0'
                    edit EPSINISTEP_B '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
            endfamily
            family pf
              edit EPSTYPE 'pf'
              family 001
                trigger  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/001 eq complete))
                edit SUBFSFAMILY '/001'
                edit EPSMEMBER '1'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an1'
                  edit FCGROUP 'fc0'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '1'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '1'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an1'
                  edit FCGROUP 'fc0'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '1'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '1'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '1'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an1'
                  edit FCGROUP 'fc0'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '1'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '1'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 002
                trigger  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/002 eq complete))
                edit SUBFSFAMILY '/002'
                edit EPSMEMBER '2'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an2'
                  edit FCGROUP 'fc0'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '2'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '2'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an2'
                  edit FCGROUP 'fc0'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '2'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '2'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '2'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an2'
                  edit FCGROUP 'fc0'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '2'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '2'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 003
                trigger  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/003 eq complete))
                edit SUBFSFAMILY '/003'
                edit EPSMEMBER '3'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an3'
                  edit FCGROUP 'fc0'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '3'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '3'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an3'
                  edit FCGROUP 'fc0'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '3'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '3'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '3'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an3'
                  edit FCGROUP 'fc0'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '3'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '3'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 004
                trigger  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/004 eq complete))
                edit SUBFSFAMILY '/004'
                edit EPSMEMBER '4'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an4'
                  edit FCGROUP 'fc0'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '4'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '4'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an4'
                  edit FCGROUP 'fc0'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '4'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '4'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '4'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an4'
                  edit FCGROUP 'fc0'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '4'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '4'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 005
                trigger  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/005 eq complete))
                edit SUBFSFAMILY '/005'
                edit EPSMEMBER '5'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an0'
                  edit FCGROUP 'fc1'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '5'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '5'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an0'
                  edit FCGROUP 'fc1'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '5'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '5'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '5'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an0'
                  edit FCGROUP 'fc1'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '5'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '5'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 006
                trigger  ( ./001/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/006 eq complete))
                edit SUBFSFAMILY '/006'
                edit EPSMEMBER '6'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an1'
                  edit FCGROUP 'fc1'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '6'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '6'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an1'
                  edit FCGROUP 'fc1'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '6'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '6'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '6'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an1'
                  edit FCGROUP 'fc1'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '6'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '6'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 007
                trigger  ( ./002/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/007 eq complete))
                edit SUBFSFAMILY '/007'
                edit EPSMEMBER '7'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an2'
                  edit FCGROUP 'fc1'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '7'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '7'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an2'
                  edit FCGROUP 'fc1'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '7'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '7'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '7'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an2'
                  edit FCGROUP 'fc1'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '7'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '7'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 008
                trigger  ( ./003/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/008 eq complete))
                edit SUBFSFAMILY '/008'
                edit EPSMEMBER '8'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an3'
                  edit FCGROUP 'fc1'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '8'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '8'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an3'
                  edit FCGROUP 'fc1'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '8'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '8'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '8'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an3'
                  edit FCGROUP 'fc1'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '8'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '8'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 009
                trigger  ( ./004/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/009 eq complete))
                edit SUBFSFAMILY '/009'
                edit EPSMEMBER '9'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an4'
                  edit FCGROUP 'fc1'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '9'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '9'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an4'
                  edit FCGROUP 'fc1'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '9'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '9'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '9'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an4'
                  edit FCGROUP 'fc1'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '9'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '9'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 010
                trigger  ( ./005/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/010 eq complete))
                edit SUBFSFAMILY '/010'
                edit EPSMEMBER '10'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an0'
                  edit FCGROUP 'fc2'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '10'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '10'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an0'
                  edit FCGROUP 'fc2'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '10'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '10'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '10'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an0'
                  edit FCGROUP 'fc2'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '10'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '10'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 011
                trigger  ( ./006/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/011 eq complete))
                edit SUBFSFAMILY '/011'
                edit EPSMEMBER '11'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an1'
                  edit FCGROUP 'fc2'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '11'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '11'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an1'
                  edit FCGROUP 'fc2'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '11'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '11'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '11'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an1'
                  edit FCGROUP 'fc2'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '11'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '11'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 012
                trigger  ( ./007/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/012 eq complete))
                edit SUBFSFAMILY '/012'
                edit EPSMEMBER '12'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an2'
                  edit FCGROUP 'fc2'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '12'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '12'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an2'
                  edit FCGROUP 'fc2'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '12'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '12'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '12'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an2'
                  edit FCGROUP 'fc2'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '12'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '12'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 013
                trigger  ( ./008/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/013 eq complete))
                edit SUBFSFAMILY '/013'
                edit EPSMEMBER '13'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an3'
                  edit FCGROUP 'fc2'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '13'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '13'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an3'
                  edit FCGROUP 'fc2'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '13'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '13'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '13'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an3'
                  edit FCGROUP 'fc2'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '13'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '13'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 014
                trigger  ( ./009/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/014 eq complete))
                edit SUBFSFAMILY '/014'
                edit EPSMEMBER '14'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an4'
                  edit FCGROUP 'fc2'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '14'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '14'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an4'
                  edit FCGROUP 'fc2'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '14'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '14'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '14'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an4'
                  edit FCGROUP 'fc2'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '14'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '14'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 015
                trigger  ( ./010/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/015 eq complete))
                edit SUBFSFAMILY '/015'
                edit EPSMEMBER '15'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an0'
                  edit FCGROUP 'fc3'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '15'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '15'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an0'
                  edit FCGROUP 'fc3'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '15'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '15'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '15'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an0'
                  edit FCGROUP 'fc3'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '15'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '15'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 016
                trigger  ( ./011/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/016 eq complete))
                edit SUBFSFAMILY '/016'
                edit EPSMEMBER '16'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an1'
                  edit FCGROUP 'fc3'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '16'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '16'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an1'
                  edit FCGROUP 'fc3'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '16'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '16'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '16'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an1'
                  edit FCGROUP 'fc3'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '16'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '16'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 017
                trigger  ( ./012/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/017 eq complete))
                edit SUBFSFAMILY '/017'
                edit EPSMEMBER '17'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an2'
                  edit FCGROUP 'fc3'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '17'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '17'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an2'
                  edit FCGROUP 'fc3'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '17'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '17'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '17'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an2'
                  edit FCGROUP 'fc3'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '17'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '17'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 018
                trigger  ( ./013/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/018 eq complete))
                edit SUBFSFAMILY '/018'
                edit EPSMEMBER '18'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an3'
                  edit FCGROUP 'fc3'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '18'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '18'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an3'
                  edit FCGROUP 'fc3'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '18'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '18'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '18'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an3'
                  edit FCGROUP 'fc3'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '18'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '18'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 019
                trigger  ( ./014/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/019 eq complete))
                edit SUBFSFAMILY '/019'
                edit EPSMEMBER '19'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an4'
                  edit FCGROUP 'fc3'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '19'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '19'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an4'
                  edit FCGROUP 'fc3'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '19'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '19'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '19'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an4'
                  edit FCGROUP 'fc3'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '19'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '19'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 020
                trigger  ( ./015/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/020 eq complete))
                edit SUBFSFAMILY '/020'
                edit EPSMEMBER '20'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an0'
                  edit FCGROUP 'fc4'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '20'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '20'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an0'
                  edit FCGROUP 'fc4'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '20'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '20'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '20'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an0'
                  edit FCGROUP 'fc4'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '20'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '20'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 021
                trigger  ( ./016/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/021 eq complete))
                edit SUBFSFAMILY '/021'
                edit EPSMEMBER '21'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an1'
                  edit FCGROUP 'fc4'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '21'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '21'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an1'
                  edit FCGROUP 'fc4'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '21'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '21'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '21'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an1'
                  edit FCGROUP 'fc4'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '21'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '21'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 022
                trigger  ( ./017/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/022 eq complete))
                edit SUBFSFAMILY '/022'
                edit EPSMEMBER '22'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an2'
                  edit FCGROUP 'fc4'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '22'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '22'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an2'
                  edit FCGROUP 'fc4'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '22'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '22'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '22'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an2'
                  edit FCGROUP 'fc4'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '22'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '22'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 023
                trigger  ( ./018/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/023 eq complete))
                edit SUBFSFAMILY '/023'
                edit EPSMEMBER '23'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an3'
                  edit FCGROUP 'fc4'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '23'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '23'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an3'
                  edit FCGROUP 'fc4'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '23'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '23'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '23'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an3'
                  edit FCGROUP 'fc4'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '23'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '23'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 024
                trigger  ( ./019/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/024 eq complete))
                edit SUBFSFAMILY '/024'
                edit EPSMEMBER '24'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an4'
                  edit FCGROUP 'fc4'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '24'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '24'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an4'
                  edit FCGROUP 'fc4'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '24'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '24'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '24'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an4'
                  edit FCGROUP 'fc4'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '24'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '24'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 025
                trigger  ( ./020/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/025 eq complete))
                edit SUBFSFAMILY '/025'
                edit EPSMEMBER '25'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an0'
                  edit FCGROUP 'fc5'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '25'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '25'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an0'
                  edit FCGROUP 'fc5'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '25'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '25'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '25'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an0'
                  edit FCGROUP 'fc5'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '25'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '25'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 026
                trigger  ( ./021/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/026 eq complete))
                edit SUBFSFAMILY '/026'
                edit EPSMEMBER '26'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an1'
                  edit FCGROUP 'fc5'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '26'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '26'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an1'
                  edit FCGROUP 'fc5'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '26'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '26'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '26'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an1'
                  edit FCGROUP 'fc5'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '26'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '26'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 027
                trigger  ( ./022/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/027 eq complete))
                edit SUBFSFAMILY '/027'
                edit EPSMEMBER '27'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an2'
                  edit FCGROUP 'fc5'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '27'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '27'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an2'
                  edit FCGROUP 'fc5'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '27'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '27'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '27'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an2'
                  edit FCGROUP 'fc5'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '27'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '27'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 028
                trigger  ( ./023/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/028 eq complete))
                edit SUBFSFAMILY '/028'
                edit EPSMEMBER '28'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an3'
                  edit FCGROUP 'fc5'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '28'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '28'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an3'
                  edit FCGROUP 'fc5'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '28'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '28'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '28'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an3'
                  edit FCGROUP 'fc5'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '28'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '28'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 029
                trigger  ( ./024/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/029 eq complete))
                edit SUBFSFAMILY '/029'
                edit EPSMEMBER '29'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an4'
                  edit FCGROUP 'fc5'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '29'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '29'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an4'
                  edit FCGROUP 'fc5'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '29'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '29'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '29'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an4'
                  edit FCGROUP 'fc5'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '29'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '29'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 030
                trigger  ( ./025/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/030 eq complete))
                edit SUBFSFAMILY '/030'
                edit EPSMEMBER '30'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an0'
                  edit FCGROUP 'fc6'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '30'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '30'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an0'
                  edit FCGROUP 'fc6'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '30'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '30'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '30'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an0'
                  edit FCGROUP 'fc6'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '30'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '30'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 031
                trigger  ( ./026/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/031 eq complete))
                edit SUBFSFAMILY '/031'
                edit EPSMEMBER '31'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an1'
                  edit FCGROUP 'fc6'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '31'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '31'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an1'
                  edit FCGROUP 'fc6'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '31'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '31'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '31'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an1'
                  edit FCGROUP 'fc6'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '31'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '31'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 032
                trigger  ( ./027/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/032 eq complete))
                edit SUBFSFAMILY '/032'
                edit EPSMEMBER '32'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an2'
                  edit FCGROUP 'fc6'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '32'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '32'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an2'
                  edit FCGROUP 'fc6'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '32'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '32'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '32'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an2'
                  edit FCGROUP 'fc6'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '32'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '32'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 033
                trigger  ( ./028/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/033 eq complete))
                edit SUBFSFAMILY '/033'
                edit EPSMEMBER '33'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an3'
                  edit FCGROUP 'fc6'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '33'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '33'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an3'
                  edit FCGROUP 'fc6'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '33'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '33'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '33'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an3'
                  edit FCGROUP 'fc6'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '33'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '33'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 034
                trigger  ( ./029/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/034 eq complete))
                edit SUBFSFAMILY '/034'
                edit EPSMEMBER '34'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an4'
                  edit FCGROUP 'fc6'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '34'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '34'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an4'
                  edit FCGROUP 'fc6'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '34'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '34'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '34'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an4'
                  edit FCGROUP 'fc6'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '34'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '34'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 035
                trigger  ( ./030/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/035 eq complete))
                edit SUBFSFAMILY '/035'
                edit EPSMEMBER '35'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an0'
                  edit FCGROUP 'fc7'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '35'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '35'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an0'
                  edit FCGROUP 'fc7'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '35'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '35'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '35'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an0'
                  edit FCGROUP 'fc7'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '35'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '35'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 036
                trigger  ( ./031/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/036 eq complete))
                edit SUBFSFAMILY '/036'
                edit EPSMEMBER '36'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an1'
                  edit FCGROUP 'fc7'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '36'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '36'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an1'
                  edit FCGROUP 'fc7'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '36'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '36'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '36'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an1'
                  edit FCGROUP 'fc7'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '36'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '36'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 037
                trigger  ( ./032/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/037 eq complete))
                edit SUBFSFAMILY '/037'
                edit EPSMEMBER '37'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an2'
                  edit FCGROUP 'fc7'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '37'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '37'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an2'
                  edit FCGROUP 'fc7'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '37'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '37'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '37'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an2'
                  edit FCGROUP 'fc7'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '37'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '37'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 038
                trigger  ( ./033/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/038 eq complete))
                edit SUBFSFAMILY '/038'
                edit EPSMEMBER '38'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an3'
                  edit FCGROUP 'fc7'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '38'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '38'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an3'
                  edit FCGROUP 'fc7'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '38'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '38'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '38'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an3'
                  edit FCGROUP 'fc7'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '38'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '38'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 039
                trigger  ( ./034/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/039 eq complete))
                edit SUBFSFAMILY '/039'
                edit EPSMEMBER '39'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an4'
                  edit FCGROUP 'fc7'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '39'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '39'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an4'
                  edit FCGROUP 'fc7'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '39'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '39'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '39'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an4'
                  edit FCGROUP 'fc7'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '39'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '39'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 040
                trigger  ( ./035/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/040 eq complete))
                edit SUBFSFAMILY '/040'
                edit EPSMEMBER '40'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an0'
                  edit FCGROUP 'fc8'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '40'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '40'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an0'
                  edit FCGROUP 'fc8'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '40'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '40'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '40'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an0'
                  edit FCGROUP 'fc8'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '40'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '40'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 041
                trigger  ( ./036/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/041 eq complete))
                edit SUBFSFAMILY '/041'
                edit EPSMEMBER '41'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an1'
                  edit FCGROUP 'fc8'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '41'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '41'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an1'
                  edit FCGROUP 'fc8'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '41'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '41'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '41'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an1'
                  edit FCGROUP 'fc8'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '41'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '41'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 042
                trigger  ( ./037/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/042 eq complete))
                edit SUBFSFAMILY '/042'
                edit EPSMEMBER '42'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an2'
                  edit FCGROUP 'fc8'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '42'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '42'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an2'
                  edit FCGROUP 'fc8'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '42'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '42'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '42'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an2'
                  edit FCGROUP 'fc8'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '42'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '42'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 043
                trigger  ( ./038/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/043 eq complete))
                edit SUBFSFAMILY '/043'
                edit EPSMEMBER '43'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an3'
                  edit FCGROUP 'fc8'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '43'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '43'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an3'
                  edit FCGROUP 'fc8'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '43'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '43'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '43'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an3'
                  edit FCGROUP 'fc8'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '43'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '43'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 044
                trigger  ( ./039/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/044 eq complete))
                edit SUBFSFAMILY '/044'
                edit EPSMEMBER '44'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an4'
                  edit FCGROUP 'fc8'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '44'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '44'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an4'
                  edit FCGROUP 'fc8'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '44'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '44'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '44'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an4'
                  edit FCGROUP 'fc8'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '44'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '44'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 045
                trigger  ( ./040/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/045 eq complete))
                edit SUBFSFAMILY '/045'
                edit EPSMEMBER '45'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an0'
                  edit FCGROUP 'fc9'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '45'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '45'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an0'
                  edit FCGROUP 'fc9'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '45'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '45'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '45'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an0'
                  edit FCGROUP 'fc9'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '45'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '45'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 046
                trigger  ( ./041/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/046 eq complete))
                edit SUBFSFAMILY '/046'
                edit EPSMEMBER '46'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an1'
                  edit FCGROUP 'fc9'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '46'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '46'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an1'
                  edit FCGROUP 'fc9'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '46'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '46'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '46'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an1'
                  edit FCGROUP 'fc9'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '46'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '46'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 047
                trigger  ( ./042/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/047 eq complete))
                edit SUBFSFAMILY '/047'
                edit EPSMEMBER '47'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an2'
                  edit FCGROUP 'fc9'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '47'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '47'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an2'
                  edit FCGROUP 'fc9'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '47'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '47'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '47'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an2'
                  edit FCGROUP 'fc9'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '47'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '47'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 048
                trigger  ( ./043/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/048 eq complete))
                edit SUBFSFAMILY '/048'
                edit EPSMEMBER '48'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an3'
                  edit FCGROUP 'fc9'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '48'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '48'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an3'
                  edit FCGROUP 'fc9'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '48'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '48'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '48'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an3'
                  edit FCGROUP 'fc9'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '48'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '48'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 049
                trigger  ( ./044/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/049 eq complete))
                edit SUBFSFAMILY '/049'
                edit EPSMEMBER '49'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an4'
                  edit FCGROUP 'fc9'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '49'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '49'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an4'
                  edit FCGROUP 'fc9'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '49'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '49'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '49'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an4'
                  edit FCGROUP 'fc9'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '49'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '49'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
              family 050
                trigger  ( ./045/legB/modeleps eq complete) and  (  ( ../../inigroup:YMD gt ../../ensemble:YMD) or  ( ../../inigroup/pert_ic/050 eq complete))
                edit SUBFSFAMILY '/050'
                edit EPSMEMBER '50'
                family legA
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an0'
                  edit FCGROUP 'fc10'
                  task getiniLeg
                    edit PRIORITY '40'
                    edit EPSMEMBER '50'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                  task modeleps
                    trigger ./getiniLeg eq complete
                    edit MEM '960'
                    edit NPES '24'
                    edit CPUTIME '25000'
                    edit THREADS '4'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '1'
                    edit EPSMEMBER '50'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '240'
                    edit EPSINISTEP_PREVIOUS '0'
                    edit FCLENGTH '240'
                    edit TSTEP '1800'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '30'
                    edit EPSWAMNANG '24'
                    edit EPSWAMNSTPW '1'
                    meter step 0 240 240
                endfamily
                family legB
                  trigger legA eq complete
                  complete  ( 2 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an0'
                  edit FCGROUP 'fc10'
                  task getvarepsdata
                    edit EPSLEG '2'
                    edit PRIORITY '40'
                    edit EPSMEMBER '50'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '399'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    complete  ( 2 gt 1)
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSMEMBER '50'
                    edit EPSINISTEP '216'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    complete  ( 2 gt 1)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '2'
                    edit EPSMEMBER '50'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '216'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global100'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    edit EPSFCLENGTH_A '240'
                    meter step 0 0 0
                endfamily
                family legC
                  trigger legB eq complete
                  complete  ( 3 gt 1)
                  edit EPSCOUPLE '0'
                  edit OCGROUP 'an0'
                  edit FCGROUP 'fc10'
                  task getvarepsdata
                    edit EPSLEG '3'
                    edit PRIORITY '40'
                    edit EPSMEMBER '50'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '255'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                  task intHtoL
                    trigger ./getvarepsdata eq complete
                    edit QUEUE 'parallel'
                    edit NPES '8'
                    edit THREADS '2'
                    edit MEM '1599'
                    edit CPUTIME '200'
                    edit EPSINISTEP '168'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit FCLENGTH '0'
                  task modeleps
                    trigger  ( intHtoL eq complete)
                    edit MEM '960'
                    edit NPES '16'
                    edit CPUTIME '4000'
                    edit THREADS '2'
                    edit STACK '550'
                    edit QUEUE 'normal'
                    edit EPSNLEGS '1'
                    edit EPSLEG '3'
                    edit EPSMEMBER '50'
                    edit EPSFCRES '95'
                    edit EPSFCLEV '62'
                    edit EPSFCGTYPE 'l_2'
                    edit EPSINISTEP '0'
                    edit EPSINISTEP_PREVIOUS '168'
                    edit FCLENGTH '0'
                    edit TSTEP '2700'
                    edit EPSWAMRESOL 'global300'
                    edit EPSWAMNFRE '25'
                    edit EPSWAMNANG '12'
                    edit EPSWAMNSTPW '1'
                    meter step 0 0 0
                endfamily
              endfamily
            endfamily
            task logfiles
              trigger cv eq complete and cf eq complete and pf eq complete
              edit SMSCMD 'smssubmit.x %SMSJOB% %SCHOST% %USER%'
              edit SMSKILL 'smskill %USER% %SCHOST% %SMSRID% %SMSJOB%'
              edit SMSSTATUSCMD 'smscheck %SCHOST% %EXPVER% %SMSJOB%'
              edit LOGDIR '/hpce/rdx_dir/msx'
          endfamily
          family lag
            repeat string YMD  "2008021112" # status 3
            trigger  (  ( ensemble:YMD ge lag:YMD) or  ( ensemble eq complete or ensemble eq unknown)) and  (  ( inigroup:YMD gt lag:YMD) or  ( inigroup eq complete or inigroup eq unknown))
            complete  ( lag:YMD eq  ~ 1)
            family archive
              edit SMSTRIES '1'
              edit MEM_ARCH '400'
              edit CPUTIME_ARCH '10000'
              edit PRIORITY '45'
              edit FCLENGTH '240'
              family cv
                defstatus complete
                trigger cf eq complete
                edit ARCHTYPE 'cv'
                family arch000
                  trigger  (  ( ../../../ensemble:YMD eq ../../../lag:YMD) and  ( ../../../ensemble/cv eq complete)) or  ( ../../../ensemble:YMD gt ../../../lag:YMD)
                  edit EPSCHUNKS '5'
                  edit EPSARCHGRP '0'
                  family arch
                    task ml
                      event fdb
                    task pl
                      trigger  ( ml eq complete or ml eq unknown)
                      event fdb
                    task sfc
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown)
                      event fdb
                    task pt
                      defstatus complete
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown) and  ( sfc eq complete or sfc eq unknown)
                      event fdb
                    task pv
                      defstatus complete
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown) and  ( sfc eq complete or sfc eq unknown) and  ( pt eq complete or pt eq unknown)
                      event fdb
                    task hl
                      defstatus complete
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown) and  ( sfc eq complete or sfc eq unknown) and  ( pt eq complete or pt eq unknown) and  ( pv eq complete or pv eq unknown)
                      event fdb
                  endfamily
                  task cleanmc
                    trigger arch eq complete
                endfamily
              endfamily
              family cf
                edit ARCHTYPE 'cf'
                family arch000
                  trigger  (  ( ../../../ensemble:YMD eq ../../../lag:YMD) and  ( ../../../ensemble/cf eq complete)) or  ( ../../../ensemble:YMD gt ../../../lag:YMD)
                  edit EPSCHUNKS '5'
                  edit EPSARCHGRP '0'
                  family arch
                    task ml
                      event fdb
                    task pl
                      trigger  ( ml eq complete or ml eq unknown)
                      event fdb
                    task sfc
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown)
                      event fdb
                    task pt
                      defstatus complete
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown) and  ( sfc eq complete or sfc eq unknown)
                      event fdb
                    task pv
                      defstatus complete
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown) and  ( sfc eq complete or sfc eq unknown) and  ( pt eq complete or pt eq unknown)
                      event fdb
                    task hl
                      defstatus complete
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown) and  ( sfc eq complete or sfc eq unknown) and  ( pt eq complete or pt eq unknown) and  ( pv eq complete or pv eq unknown)
                      event fdb
                    task wm_sfc_arc
                      trigger sfc eq complete
                      complete  (  ( 240 ge 768 and 1 ne 1) or  ( 240 lt 768))
                      edit RUNVARFC '1'
                    task wm_ua_arc
                      trigger wm_sfc_arc eq complete
                      complete  (  ( 240 ge 768 and 1 ne 1) or  ( 240 lt 768))
                      edit RUNVARFC '1'
                  endfamily
                  task cleanmc
                    trigger arch eq complete
                endfamily
              endfamily
              family pf
                trigger cf eq complete and cv eq complete
                edit ARCHTYPE 'pf'
                family arch001
                  trigger  ( ../../../ensemble:YMD gt ../../../lag:YMD or  ( ../../../ensemble/pf/001 eq complete and ../../../ensemble/pf/002 eq complete and ../../../ensemble/pf/003 eq complete and ../../../ensemble/pf/004 eq complete and ../../../ensemble/pf/005 eq complete))
                  edit EPSCHUNKS '5'
                  edit EPSARCHGRP '1'
                  family arch
                    task ml
                      event fdb
                    task pl
                      trigger  ( ml eq complete or ml eq unknown)
                      event fdb
                    task sfc
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown)
                      event fdb
                    task pt
                      defstatus complete
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown) and  ( sfc eq complete or sfc eq unknown)
                      event fdb
                    task pv
                      defstatus complete
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown) and  ( sfc eq complete or sfc eq unknown) and  ( pt eq complete or pt eq unknown)
                      event fdb
                    task hl
                      defstatus complete
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown) and  ( sfc eq complete or sfc eq unknown) and  ( pt eq complete or pt eq unknown) and  ( pv eq complete or pv eq unknown)
                      event fdb
                    task wm_sfc_arc
                      trigger sfc eq complete
                      complete  (  ( 240 ge 768 and 1 ne 1) or  ( 240 lt 768))
                      edit RUNVARFC '1'
                    task wm_ua_arc
                      trigger wm_sfc_arc eq complete
                      complete  (  ( 240 ge 768 and 1 ne 1) or  ( 240 lt 768))
                      edit RUNVARFC '1'
                  endfamily
                  task cleanmc
                    trigger arch eq complete
                endfamily
                family arch002
                  trigger arch001 eq complete and  ( ../../../ensemble:YMD gt ../../../lag:YMD or  ( ../../../ensemble/pf/006 eq complete and ../../../ensemble/pf/007 eq complete and ../../../ensemble/pf/008 eq complete and ../../../ensemble/pf/009 eq complete and ../../../ensemble/pf/010 eq complete))
                  edit EPSCHUNKS '5'
                  edit EPSARCHGRP '2'
                  family arch
                    task ml
                      event fdb
                    task pl
                      trigger  ( ml eq complete or ml eq unknown)
                      event fdb
                    task sfc
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown)
                      event fdb
                    task pt
                      defstatus complete
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown) and  ( sfc eq complete or sfc eq unknown)
                      event fdb
                    task pv
                      defstatus complete
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown) and  ( sfc eq complete or sfc eq unknown) and  ( pt eq complete or pt eq unknown)
                      event fdb
                    task hl
                      defstatus complete
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown) and  ( sfc eq complete or sfc eq unknown) and  ( pt eq complete or pt eq unknown) and  ( pv eq complete or pv eq unknown)
                      event fdb
                    task wm_sfc_arc
                      trigger sfc eq complete
                      complete  (  ( 240 ge 768 and 1 ne 1) or  ( 240 lt 768))
                      edit RUNVARFC '1'
                    task wm_ua_arc
                      trigger wm_sfc_arc eq complete
                      complete  (  ( 240 ge 768 and 1 ne 1) or  ( 240 lt 768))
                      edit RUNVARFC '1'
                  endfamily
                  task cleanmc
                    trigger arch eq complete
                endfamily
                family arch003
                  trigger arch002 eq complete and  ( ../../../ensemble:YMD gt ../../../lag:YMD or  ( ../../../ensemble/pf/011 eq complete and ../../../ensemble/pf/012 eq complete and ../../../ensemble/pf/013 eq complete and ../../../ensemble/pf/014 eq complete and ../../../ensemble/pf/015 eq complete))
                  edit EPSCHUNKS '5'
                  edit EPSARCHGRP '3'
                  family arch
                    task ml
                      event fdb
                    task pl
                      trigger  ( ml eq complete or ml eq unknown)
                      event fdb
                    task sfc
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown)
                      event fdb
                    task pt
                      defstatus complete
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown) and  ( sfc eq complete or sfc eq unknown)
                      event fdb
                    task pv
                      defstatus complete
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown) and  ( sfc eq complete or sfc eq unknown) and  ( pt eq complete or pt eq unknown)
                      event fdb
                    task hl
                      defstatus complete
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown) and  ( sfc eq complete or sfc eq unknown) and  ( pt eq complete or pt eq unknown) and  ( pv eq complete or pv eq unknown)
                      event fdb
                    task wm_sfc_arc
                      trigger sfc eq complete
                      complete  (  ( 240 ge 768 and 1 ne 1) or  ( 240 lt 768))
                      edit RUNVARFC '1'
                    task wm_ua_arc
                      trigger wm_sfc_arc eq complete
                      complete  (  ( 240 ge 768 and 1 ne 1) or  ( 240 lt 768))
                      edit RUNVARFC '1'
                  endfamily
                  task cleanmc
                    trigger arch eq complete
                endfamily
                family arch004
                  trigger arch003 eq complete and  ( ../../../ensemble:YMD gt ../../../lag:YMD or  ( ../../../ensemble/pf/016 eq complete and ../../../ensemble/pf/017 eq complete and ../../../ensemble/pf/018 eq complete and ../../../ensemble/pf/019 eq complete and ../../../ensemble/pf/020 eq complete))
                  edit EPSCHUNKS '5'
                  edit EPSARCHGRP '4'
                  family arch
                    task ml
                      event fdb
                    task pl
                      trigger  ( ml eq complete or ml eq unknown)
                      event fdb
                    task sfc
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown)
                      event fdb
                    task pt
                      defstatus complete
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown) and  ( sfc eq complete or sfc eq unknown)
                      event fdb
                    task pv
                      defstatus complete
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown) and  ( sfc eq complete or sfc eq unknown) and  ( pt eq complete or pt eq unknown)
                      event fdb
                    task hl
                      defstatus complete
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown) and  ( sfc eq complete or sfc eq unknown) and  ( pt eq complete or pt eq unknown) and  ( pv eq complete or pv eq unknown)
                      event fdb
                    task wm_sfc_arc
                      trigger sfc eq complete
                      complete  (  ( 240 ge 768 and 1 ne 1) or  ( 240 lt 768))
                      edit RUNVARFC '1'
                    task wm_ua_arc
                      trigger wm_sfc_arc eq complete
                      complete  (  ( 240 ge 768 and 1 ne 1) or  ( 240 lt 768))
                      edit RUNVARFC '1'
                  endfamily
                  task cleanmc
                    trigger arch eq complete
                endfamily
                family arch005
                  trigger arch004 eq complete and  ( ../../../ensemble:YMD gt ../../../lag:YMD or  ( ../../../ensemble/pf/021 eq complete and ../../../ensemble/pf/022 eq complete and ../../../ensemble/pf/023 eq complete and ../../../ensemble/pf/024 eq complete and ../../../ensemble/pf/025 eq complete))
                  edit EPSCHUNKS '5'
                  edit EPSARCHGRP '5'
                  family arch
                    task ml
                      event fdb
                    task pl
                      trigger  ( ml eq complete or ml eq unknown)
                      event fdb
                    task sfc
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown)
                      event fdb
                    task pt
                      defstatus complete
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown) and  ( sfc eq complete or sfc eq unknown)
                      event fdb
                    task pv
                      defstatus complete
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown) and  ( sfc eq complete or sfc eq unknown) and  ( pt eq complete or pt eq unknown)
                      event fdb
                    task hl
                      defstatus complete
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown) and  ( sfc eq complete or sfc eq unknown) and  ( pt eq complete or pt eq unknown) and  ( pv eq complete or pv eq unknown)
                      event fdb
                    task wm_sfc_arc
                      trigger sfc eq complete
                      complete  (  ( 240 ge 768 and 1 ne 1) or  ( 240 lt 768))
                      edit RUNVARFC '1'
                    task wm_ua_arc
                      trigger wm_sfc_arc eq complete
                      complete  (  ( 240 ge 768 and 1 ne 1) or  ( 240 lt 768))
                      edit RUNVARFC '1'
                  endfamily
                  task cleanmc
                    trigger arch eq complete
                endfamily
                family arch006
                  trigger arch005 eq complete and  ( ../../../ensemble:YMD gt ../../../lag:YMD or  ( ../../../ensemble/pf/026 eq complete and ../../../ensemble/pf/027 eq complete and ../../../ensemble/pf/028 eq complete and ../../../ensemble/pf/029 eq complete and ../../../ensemble/pf/030 eq complete))
                  edit EPSCHUNKS '5'
                  edit EPSARCHGRP '6'
                  family arch
                    task ml
                      event fdb
                    task pl
                      trigger  ( ml eq complete or ml eq unknown)
                      event fdb
                    task sfc
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown)
                      event fdb
                    task pt
                      defstatus complete
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown) and  ( sfc eq complete or sfc eq unknown)
                      event fdb
                    task pv
                      defstatus complete
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown) and  ( sfc eq complete or sfc eq unknown) and  ( pt eq complete or pt eq unknown)
                      event fdb
                    task hl
                      defstatus complete
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown) and  ( sfc eq complete or sfc eq unknown) and  ( pt eq complete or pt eq unknown) and  ( pv eq complete or pv eq unknown)
                      event fdb
                    task wm_sfc_arc
                      trigger sfc eq complete
                      complete  (  ( 240 ge 768 and 1 ne 1) or  ( 240 lt 768))
                      edit RUNVARFC '1'
                    task wm_ua_arc
                      trigger wm_sfc_arc eq complete
                      complete  (  ( 240 ge 768 and 1 ne 1) or  ( 240 lt 768))
                      edit RUNVARFC '1'
                  endfamily
                  task cleanmc
                    trigger arch eq complete
                endfamily
                family arch007
                  trigger arch006 eq complete and  ( ../../../ensemble:YMD gt ../../../lag:YMD or  ( ../../../ensemble/pf/031 eq complete and ../../../ensemble/pf/032 eq complete and ../../../ensemble/pf/033 eq complete and ../../../ensemble/pf/034 eq complete and ../../../ensemble/pf/035 eq complete))
                  edit EPSCHUNKS '5'
                  edit EPSARCHGRP '7'
                  family arch
                    task ml
                      event fdb
                    task pl
                      trigger  ( ml eq complete or ml eq unknown)
                      event fdb
                    task sfc
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown)
                      event fdb
                    task pt
                      defstatus complete
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown) and  ( sfc eq complete or sfc eq unknown)
                      event fdb
                    task pv
                      defstatus complete
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown) and  ( sfc eq complete or sfc eq unknown) and  ( pt eq complete or pt eq unknown)
                      event fdb
                    task hl
                      defstatus complete
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown) and  ( sfc eq complete or sfc eq unknown) and  ( pt eq complete or pt eq unknown) and  ( pv eq complete or pv eq unknown)
                      event fdb
                    task wm_sfc_arc
                      trigger sfc eq complete
                      complete  (  ( 240 ge 768 and 1 ne 1) or  ( 240 lt 768))
                      edit RUNVARFC '1'
                    task wm_ua_arc
                      trigger wm_sfc_arc eq complete
                      complete  (  ( 240 ge 768 and 1 ne 1) or  ( 240 lt 768))
                      edit RUNVARFC '1'
                  endfamily
                  task cleanmc
                    trigger arch eq complete
                endfamily
                family arch008
                  trigger arch007 eq complete and  ( ../../../ensemble:YMD gt ../../../lag:YMD or  ( ../../../ensemble/pf/036 eq complete and ../../../ensemble/pf/037 eq complete and ../../../ensemble/pf/038 eq complete and ../../../ensemble/pf/039 eq complete and ../../../ensemble/pf/040 eq complete))
                  edit EPSCHUNKS '5'
                  edit EPSARCHGRP '8'
                  family arch
                    task ml
                      event fdb
                    task pl
                      trigger  ( ml eq complete or ml eq unknown)
                      event fdb
                    task sfc
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown)
                      event fdb
                    task pt
                      defstatus complete
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown) and  ( sfc eq complete or sfc eq unknown)
                      event fdb
                    task pv
                      defstatus complete
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown) and  ( sfc eq complete or sfc eq unknown) and  ( pt eq complete or pt eq unknown)
                      event fdb
                    task hl
                      defstatus complete
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown) and  ( sfc eq complete or sfc eq unknown) and  ( pt eq complete or pt eq unknown) and  ( pv eq complete or pv eq unknown)
                      event fdb
                    task wm_sfc_arc
                      trigger sfc eq complete
                      complete  (  ( 240 ge 768 and 1 ne 1) or  ( 240 lt 768))
                      edit RUNVARFC '1'
                    task wm_ua_arc
                      trigger wm_sfc_arc eq complete
                      complete  (  ( 240 ge 768 and 1 ne 1) or  ( 240 lt 768))
                      edit RUNVARFC '1'
                  endfamily
                  task cleanmc
                    trigger arch eq complete
                endfamily
                family arch009
                  trigger arch008 eq complete and  ( ../../../ensemble:YMD gt ../../../lag:YMD or  ( ../../../ensemble/pf/041 eq complete and ../../../ensemble/pf/042 eq complete and ../../../ensemble/pf/043 eq complete and ../../../ensemble/pf/044 eq complete and ../../../ensemble/pf/045 eq complete))
                  edit EPSCHUNKS '5'
                  edit EPSARCHGRP '9'
                  family arch
                    task ml
                      event fdb
                    task pl
                      trigger  ( ml eq complete or ml eq unknown)
                      event fdb
                    task sfc
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown)
                      event fdb
                    task pt
                      defstatus complete
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown) and  ( sfc eq complete or sfc eq unknown)
                      event fdb
                    task pv
                      defstatus complete
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown) and  ( sfc eq complete or sfc eq unknown) and  ( pt eq complete or pt eq unknown)
                      event fdb
                    task hl
                      defstatus complete
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown) and  ( sfc eq complete or sfc eq unknown) and  ( pt eq complete or pt eq unknown) and  ( pv eq complete or pv eq unknown)
                      event fdb
                    task wm_sfc_arc
                      trigger sfc eq complete
                      complete  (  ( 240 ge 768 and 1 ne 1) or  ( 240 lt 768))
                      edit RUNVARFC '1'
                    task wm_ua_arc
                      trigger wm_sfc_arc eq complete
                      complete  (  ( 240 ge 768 and 1 ne 1) or  ( 240 lt 768))
                      edit RUNVARFC '1'
                  endfamily
                  task cleanmc
                    trigger arch eq complete
                endfamily
                family arch010
                  trigger arch009 eq complete and  ( ../../../ensemble:YMD gt ../../../lag:YMD or  ( ../../../ensemble/pf/046 eq complete and ../../../ensemble/pf/047 eq complete and ../../../ensemble/pf/048 eq complete and ../../../ensemble/pf/049 eq complete and ../../../ensemble/pf/050 eq complete))
                  edit EPSCHUNKS '5'
                  edit EPSARCHGRP '10'
                  family arch
                    task ml
                      event fdb
                    task pl
                      trigger  ( ml eq complete or ml eq unknown)
                      event fdb
                    task sfc
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown)
                      event fdb
                    task pt
                      defstatus complete
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown) and  ( sfc eq complete or sfc eq unknown)
                      event fdb
                    task pv
                      defstatus complete
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown) and  ( sfc eq complete or sfc eq unknown) and  ( pt eq complete or pt eq unknown)
                      event fdb
                    task hl
                      defstatus complete
                      trigger  ( ml eq complete or ml eq unknown) and  ( pl eq complete or pl eq unknown) and  ( sfc eq complete or sfc eq unknown) and  ( pt eq complete or pt eq unknown) and  ( pv eq complete or pv eq unknown)
                      event fdb
                    task wm_sfc_arc
                      trigger sfc eq complete
                      complete  (  ( 240 ge 768 and 1 ne 1) or  ( 240 lt 768))
                      edit RUNVARFC '1'
                    task wm_ua_arc
                      trigger wm_sfc_arc eq complete
                      complete  (  ( 240 ge 768 and 1 ne 1) or  ( 240 lt 768))
                      edit RUNVARFC '1'
                  endfamily
                  task cleanmc
                    trigger arch eq complete
                endfamily
              endfamily
              family icp
                trigger ./pf eq complete
                edit ARCHTYPE 'icp'
                task ml
                  event fdb
                task cleanmc
                  trigger ml eq complete
              endfamily
            endfamily
            task flush
              trigger archive/pf eq complete and archive/cf eq complete and archive/cv eq complete and archive/icp eq complete
              edit SMSCMD 'smssubmit.x %SMSJOB% %WSHOST% %USER%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit SMSSTATUSCMD 'smscheck %WSHOST% %EXPVER% %SMSJOB%'
              edit LOGDIR '/vol/msexplog/outputs'
              edit QUEUE 'expsms'
            family ecfs
              task save
            endfamily
            family plot
              family prob
                edit SMSCMD 'smssubmit.x %SMSJOB% %WSHOST% %USER%'
                edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
                edit SMSSTATUSCMD 'smscheck %WSHOST% %EXPVER% %SMSJOB%'
                edit LOGDIR '/vol/msexplog/outputs'
                edit QUEUE 'expsms'
                task plumea
                  defstatus complete
                task pb
                  defstatus complete
                  trigger ../../prob eq complete
              endfamily
            endfamily
            family prob
              defstatus complete
              task rain
              task temp
              task wind
            endfamily
            task cleanvarfc
              trigger archive eq complete and flush eq complete and ecfs eq complete and plot eq complete and prob eq complete
            task logfiles
              trigger cleanvarfc eq complete
              edit SMSCMD 'smssubmit.x %SMSJOB% %SCHOST% %USER%'
              edit SMSKILL 'smskill %USER% %SCHOST% %SMSRID% %SMSJOB%'
              edit SMSSTATUSCMD 'smscheck %SCHOST% %EXPVER% %SMSJOB%'
              edit LOGDIR '/hpce/rdx_dir/msx'
          endfamily
        endfamily
      endfamily
    endfamily
    task wipefdb
      trigger  ( an eq complete or an eq unknown) and  ( fc eq complete or fc eq unknown) and  ( eps_sv eq complete or eps_sv eq unknown) and  ( eps_fc eq complete or eps_fc eq unknown) and  ( eps_varfc eq complete or eps_varfc eq unknown) and  ( oceatm eq complete or oceatm eq unknown) and  ( oce eq complete or oce eq unknown) and  ( fpos eq complete or fpos eq unknown) and  ( sens eq complete or sens eq unknown) and  ( wam eq complete or wam eq unknown) and  ( oc eq complete or oc eq unknown) and  ( climplot eq complete or climplot eq unknown) and  ( coup eq complete or coup eq unknown)
      edit YMD '2008021112'
      edit HOUR ''
      edit FSFAMILY 'cancel'
    task cancel
      trigger  ( an eq complete or an eq unknown) and  ( fc eq complete or fc eq unknown) and  ( eps_sv eq complete or eps_sv eq unknown) and  ( eps_fc eq complete or eps_fc eq unknown) and  ( eps_varfc eq complete or eps_varfc eq unknown) and  ( oceatm eq complete or oceatm eq unknown) and  ( oce eq complete or oce eq unknown) and  ( fpos eq complete or fpos eq unknown) and  ( sens eq complete or sens eq unknown) and  ( wam eq complete or wam eq unknown) and  ( oc eq complete or oc eq unknown) and  ( climplot eq complete or climplot eq unknown) and  ( coup eq complete or coup eq unknown) and  ( wipefdb eq complete or wipefdb eq unknown)
      edit YMD '2008021112'
      edit HOUR ''
      edit FSFAMILY 'cancel'
      edit SMSCMD 'smssubmit.x %SMSJOB% %WSHOST% %USER%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit SMSSTATUSCMD 'smscheck %WSHOST% %EXPVER% %SMSJOB%'
      edit LOGDIR '/vol/msexplog/outputs'
      edit QUEUE 'expsms'
  endfamily
endsuite
#============================================================
