# 4.0.6
extern /e_36r1/main/00/an
extern /e_36r1/main/00/an/4dvar
extern /e_36r1/main/00/an/4dvar/ifstraj:finalwave
extern /e_36r1/main/00/an/4dvar/uptraj_0
extern /e_36r1/main/00/an/surf_anal/slwet
extern /e_36r1/main/00/an/surf_anal/sst
extern /e_36r1/main/00/fc/model
extern /e_36r1/main/00/getreq
extern /e_36r1/main/00/obs/get/get_stor
extern /e_36r1/main/00/prod/240/prod_wparam
extern /e_36r1/main/00dc/fc/model
extern /e_36r1/main/00dc/fc/model:step
extern /e_36r1/main/12/an
extern /e_36r1/main/12/an/4dvar
extern /e_36r1/main/12/an/4dvar/ifstraj:finalwave
extern /e_36r1/main/12/an/4dvar/uptraj_0
extern /e_36r1/main/12/an/surf_anal/slwet
extern /e_36r1/main/12/an/surf_anal/sst
extern /e_36r1/main/12/fc/model
extern /e_36r1/main/12/getreq
extern /e_36r1/main/12/obs/get/get_stor
extern /e_36r1/main/12/prod/240/prod_wparam
extern /e_36r1/main/12dc/fc/model
extern /e_36r1/main/12dc/fc/model:step
extern /e_36r1/main:YMD
extern /ecgems/f93i/an/lag/00/archive
extern /ecgems/f93i/an/lag:YMD
extern /emc_36r1/main/00/legB/prod
extern /emc_36r1/main/12/legB/prod
extern /limits:c1a
extern /limits:c1b
extern /limits:hpce
extern /limits:hpcf
extern /limits:linux_cluster
extern /limits:mars
extern /limits:obs_get
extern /limits:pdb
extern /limits:pdb_bc
extern /limits:prod
extern /limits:q2diss
extern /limits:q2diss_bc
extern /limits:swarm
extern /mc/main/00
extern /mc/main/00/legA/fc
extern /mc/main/00/legB/fc
extern /mc/main/00/legB/postproc
extern /mc/main/12/legA/fc
extern /mc/main/12/legB/fc
extern /mc/main/12/legB/postproc
extern /mc/main:YMD
extern /o/lag/00/archive/anml
extern /o/lag/00/archive/anpl
extern /o/lag/00/archive/ansfc
extern /o/lag/00/archive/fc048
extern /o/lag/00/archive/fc240
extern /o/lag/00/archive/obs
extern /o/lag/00/archive/sst
extern /o/lag/00/fb/feedback/matchup
extern /o/lag/00/satim
extern /o/lag/00/verif
extern /o/lag/00dc/archive
extern /o/lag/00dc/fb/feedback/matchup
extern /o/lag/06/archive
extern /o/lag/06/fb/feedback/matchup
extern /o/lag/06bc/archive
extern /o/lag/06bc/fb/feedback/matchup
extern /o/lag/12/archive/fc240
extern /o/lag/12/archive/obs
extern /o/lag/12/fb/feedback/matchup
extern /o/lag/12/satim
extern /o/lag/12/verif
extern /o/lag/12dc/archive
extern /o/lag/12dc/fb/feedback/matchup
extern /o/lag/18/archive
extern /o/lag/18/fb/feedback/matchup
extern /o/lag/18bc/archive
extern /o/lag/18bc/fb/feedback/matchup
extern /o/lag:YMD
extern /o/main/00/an
extern /o/main/00/an/4dvar
extern /o/main/00/an/surf_anal
extern /o/main/00/fc
extern /o/main/00/fc/clean
extern /o/main/00/fc/model
extern /o/main/00/fc/model:step
extern /o/main/00/prod/18an/derived_param
extern /o/main/00/prod2diss/oper
extern /o/main/00/prod2diss/sched/240
extern /o/main/00/prod2diss/wave
extern /o/main/00dc/an/4dvar
extern /o/main/00dc/an/surf_anal
extern /o/main/00dc/fc/clean
extern /o/main/00dc/fc/model
extern /o/main/06bc/an/4dvar
extern /o/main/06bc/an/surf_anal
extern /o/main/06bc/fc/model
extern /o/main/06bc/prod2diss/oper
extern /o/main/06bc/prod2diss/sched/72
extern /o/main/12/an
extern /o/main/12/an/4dvar
extern /o/main/12/an/surf_anal
extern /o/main/12/fc
extern /o/main/12/fc/model
extern /o/main/12/fc/model:step
extern /o/main/12/obs/get/get_stor
extern /o/main/12/prod/06an/derived_param
extern /o/main/12/prod2diss/oper
extern /o/main/12/prod2diss/sched/240
extern /o/main/12/prod2diss/wave
extern /o/main/12dc/an/4dvar
extern /o/main/12dc/an/surf_anal
extern /o/main/12dc/fc/model
extern /o/main/12dc/obs/get/get_stor
extern /o/main/18bc/an/4dvar
extern /o/main/18bc/an/surf_anal
extern /o/main/18bc/fc/model
extern /o/main/18bc/prod2diss/oper
extern /o/main/18bc/prod2diss/sched/72
extern /o/main:YMD
extern /pp1/main/extractions/dc00/airs
extern /pp1/main/extractions/dc00/amsr
extern /pp1/main/extractions/dc00/atap
extern /pp1/main/extractions/dc00/ateu
extern /pp1/main/extractions/dc00/atov
extern /pp1/main/extractions/dc00/bice
extern /pp1/main/extractions/dc00/eas1
extern /pp1/main/extractions/dc00/emer
extern /pp1/main/extractions/dc00/ers1_ra
extern /pp1/main/extractions/dc00/ers1_wa
extern /pp1/main/extractions/dc00/ers1_wi
extern /pp1/main/extractions/dc00/grad
extern /pp1/main/extractions/dc00/grst
extern /pp1/main/extractions/dc00/ice/ice_fdb_c1b
extern /pp1/main/extractions/dc00/nonbin
extern /pp1/main/extractions/dc00/qsca
extern /pp1/main/extractions/dc00/rado
extern /pp1/main/extractions/dc00/rtovs
extern /pp1/main/extractions/dc00/sbuv
extern /pp1/main/extractions/dc00/smin
extern /pp1/main/extractions/dc00/snow
extern /pp1/main/extractions/dc00/ssbt
extern /pp1/main/extractions/dc00/sst/sst_fdb_c1b
extern /pp1/main/extractions/dc00/stor
extern /pp1/main/extractions/dc00/trmm
extern /pp1/main/extractions/dc06/airs
extern /pp1/main/extractions/dc06/amsr
extern /pp1/main/extractions/dc06/atap
extern /pp1/main/extractions/dc06/ateu
extern /pp1/main/extractions/dc06/atov
extern /pp1/main/extractions/dc06/bice
extern /pp1/main/extractions/dc06/eas1
extern /pp1/main/extractions/dc06/emer
extern /pp1/main/extractions/dc06/ers1_ra
extern /pp1/main/extractions/dc06/ers1_wa
extern /pp1/main/extractions/dc06/ers1_wi
extern /pp1/main/extractions/dc06/grad
extern /pp1/main/extractions/dc06/grst
extern /pp1/main/extractions/dc06/ice/ice_fdb_c1b
extern /pp1/main/extractions/dc06/nonbin
extern /pp1/main/extractions/dc06/qsca
extern /pp1/main/extractions/dc06/rado
extern /pp1/main/extractions/dc06/rtovs
extern /pp1/main/extractions/dc06/sbuv
extern /pp1/main/extractions/dc06/smin
extern /pp1/main/extractions/dc06/snow
extern /pp1/main/extractions/dc06/ssbt
extern /pp1/main/extractions/dc06/sst/sst_fdb_c1b
extern /pp1/main/extractions/dc06/stor
extern /pp1/main/extractions/dc06/trmm
extern /pp1/main/extractions/dc12/airs
extern /pp1/main/extractions/dc12/amsr
extern /pp1/main/extractions/dc12/atap
extern /pp1/main/extractions/dc12/ateu
extern /pp1/main/extractions/dc12/atov
extern /pp1/main/extractions/dc12/bice
extern /pp1/main/extractions/dc12/eas1
extern /pp1/main/extractions/dc12/emer
extern /pp1/main/extractions/dc12/ers1_ra
extern /pp1/main/extractions/dc12/ers1_wa
extern /pp1/main/extractions/dc12/ers1_wi
extern /pp1/main/extractions/dc12/grad
extern /pp1/main/extractions/dc12/grst
extern /pp1/main/extractions/dc12/ice/ice_fdb_c1b
extern /pp1/main/extractions/dc12/nonbin
extern /pp1/main/extractions/dc12/qsca
extern /pp1/main/extractions/dc12/rado
extern /pp1/main/extractions/dc12/rtovs
extern /pp1/main/extractions/dc12/sbuv
extern /pp1/main/extractions/dc12/smin
extern /pp1/main/extractions/dc12/snow
extern /pp1/main/extractions/dc12/ssbt
extern /pp1/main/extractions/dc12/sst/sst_fdb_c1b
extern /pp1/main/extractions/dc12/stor
extern /pp1/main/extractions/dc12/trmm
extern /pp1/main/extractions/dc18/airs
extern /pp1/main/extractions/dc18/amsr
extern /pp1/main/extractions/dc18/atap
extern /pp1/main/extractions/dc18/ateu
extern /pp1/main/extractions/dc18/atov
extern /pp1/main/extractions/dc18/bice
extern /pp1/main/extractions/dc18/eas1
extern /pp1/main/extractions/dc18/emer
extern /pp1/main/extractions/dc18/ers1_ra
extern /pp1/main/extractions/dc18/ers1_wa
extern /pp1/main/extractions/dc18/ers1_wi
extern /pp1/main/extractions/dc18/grad
extern /pp1/main/extractions/dc18/grst
extern /pp1/main/extractions/dc18/ice/ice_fdb_c1b
extern /pp1/main/extractions/dc18/nonbin
extern /pp1/main/extractions/dc18/qsca
extern /pp1/main/extractions/dc18/rado
extern /pp1/main/extractions/dc18/rtovs
extern /pp1/main/extractions/dc18/sbuv
extern /pp1/main/extractions/dc18/smin
extern /pp1/main/extractions/dc18/snow
extern /pp1/main/extractions/dc18/ssbt
extern /pp1/main/extractions/dc18/sst/sst_fdb_c1b
extern /pp1/main/extractions/dc18/stor
extern /pp1/main/extractions/dc18/trmm
extern /pp1/main/extractions/ex00/airs
extern /pp1/main/extractions/ex00/amsr
extern /pp1/main/extractions/ex00/atap
extern /pp1/main/extractions/ex00/ateu
extern /pp1/main/extractions/ex00/atov
extern /pp1/main/extractions/ex00/bice
extern /pp1/main/extractions/ex00/eas1
extern /pp1/main/extractions/ex00/emer
extern /pp1/main/extractions/ex00/ers1_ra
extern /pp1/main/extractions/ex00/ers1_wa
extern /pp1/main/extractions/ex00/ers1_wi
extern /pp1/main/extractions/ex00/grad
extern /pp1/main/extractions/ex00/grst
extern /pp1/main/extractions/ex00/ice/ice_fdb_c1b
extern /pp1/main/extractions/ex00/nonbin
extern /pp1/main/extractions/ex00/qsca
extern /pp1/main/extractions/ex00/rado
extern /pp1/main/extractions/ex00/rtovs
extern /pp1/main/extractions/ex00/sbuv
extern /pp1/main/extractions/ex00/smin
extern /pp1/main/extractions/ex00/snow
extern /pp1/main/extractions/ex00/ssbt
extern /pp1/main/extractions/ex00/sst/sst_fdb_c1b
extern /pp1/main/extractions/ex00/stor
extern /pp1/main/extractions/ex00/trmm
extern /pp1/main/extractions/ex06/airs
extern /pp1/main/extractions/ex06/amsr
extern /pp1/main/extractions/ex06/atap
extern /pp1/main/extractions/ex06/ateu
extern /pp1/main/extractions/ex06/atov
extern /pp1/main/extractions/ex06/bice
extern /pp1/main/extractions/ex06/eas1
extern /pp1/main/extractions/ex06/emer
extern /pp1/main/extractions/ex06/ers1_ra
extern /pp1/main/extractions/ex06/ers1_wa
extern /pp1/main/extractions/ex06/ers1_wi
extern /pp1/main/extractions/ex06/grad
extern /pp1/main/extractions/ex06/grst
extern /pp1/main/extractions/ex06/ice/ice_fdb_c1b
extern /pp1/main/extractions/ex06/nonbin
extern /pp1/main/extractions/ex06/qsca
extern /pp1/main/extractions/ex06/rado
extern /pp1/main/extractions/ex06/rtovs
extern /pp1/main/extractions/ex06/sbuv
extern /pp1/main/extractions/ex06/smin
extern /pp1/main/extractions/ex06/snow
extern /pp1/main/extractions/ex06/ssbt
extern /pp1/main/extractions/ex06/sst/sst_fdb_c1b
extern /pp1/main/extractions/ex06/stor
extern /pp1/main/extractions/ex06/trmm
extern /pp1/main/extractions/ex12/airs
extern /pp1/main/extractions/ex12/amsr
extern /pp1/main/extractions/ex12/atap
extern /pp1/main/extractions/ex12/ateu
extern /pp1/main/extractions/ex12/atov
extern /pp1/main/extractions/ex12/bice
extern /pp1/main/extractions/ex12/eas1
extern /pp1/main/extractions/ex12/emer
extern /pp1/main/extractions/ex12/ers1_ra
extern /pp1/main/extractions/ex12/ers1_wa
extern /pp1/main/extractions/ex12/ers1_wi
extern /pp1/main/extractions/ex12/grad
extern /pp1/main/extractions/ex12/grst
extern /pp1/main/extractions/ex12/ice/ice_fdb_c1b
extern /pp1/main/extractions/ex12/nonbin
extern /pp1/main/extractions/ex12/qsca
extern /pp1/main/extractions/ex12/rado
extern /pp1/main/extractions/ex12/rtovs
extern /pp1/main/extractions/ex12/sbuv
extern /pp1/main/extractions/ex12/smin
extern /pp1/main/extractions/ex12/snow
extern /pp1/main/extractions/ex12/ssbt
extern /pp1/main/extractions/ex12/sst/sst_fdb_c1b
extern /pp1/main/extractions/ex12/stor
extern /pp1/main/extractions/ex12/trmm
extern /pp1/main/extractions/ex18/airs
extern /pp1/main/extractions/ex18/amsr
extern /pp1/main/extractions/ex18/atap
extern /pp1/main/extractions/ex18/ateu
extern /pp1/main/extractions/ex18/atov
extern /pp1/main/extractions/ex18/bice
extern /pp1/main/extractions/ex18/eas1
extern /pp1/main/extractions/ex18/emer
extern /pp1/main/extractions/ex18/ers1_ra
extern /pp1/main/extractions/ex18/ers1_wa
extern /pp1/main/extractions/ex18/ers1_wi
extern /pp1/main/extractions/ex18/grad
extern /pp1/main/extractions/ex18/grst
extern /pp1/main/extractions/ex18/ice/ice_fdb_c1b
extern /pp1/main/extractions/ex18/nonbin
extern /pp1/main/extractions/ex18/qsca
extern /pp1/main/extractions/ex18/rado
extern /pp1/main/extractions/ex18/rtovs
extern /pp1/main/extractions/ex18/sbuv
extern /pp1/main/extractions/ex18/smin
extern /pp1/main/extractions/ex18/snow
extern /pp1/main/extractions/ex18/ssbt
extern /pp1/main/extractions/ex18/sst/sst_fdb_c1b
extern /pp1/main/extractions/ex18/stor
extern /pp1/main/extractions/ex18/trmm
extern /pp1/main/extractions/ma
extern /pp2/main/extractions/dc00/amsa
extern /pp2/main/extractions/dc00/asc2
extern /pp2/main/extractions/dc00/ascs
extern /pp2/main/extractions/dc00/asel
extern /pp2/main/extractions/dc00/avhn
extern /pp2/main/extractions/dc00/g2to
extern /pp2/main/extractions/dc00/gras
extern /pp2/main/extractions/dc00/hirs
extern /pp2/main/extractions/dc00/iasi
extern /pp2/main/extractions/dc00/j2eo
extern /pp2/main/extractions/dc00/j2nb
extern /pp2/main/extractions/dc00/mhsa
extern /pp2/main/extractions/dc00/modw
extern /pp2/main/extractions/dc06/amsa
extern /pp2/main/extractions/dc06/asc2
extern /pp2/main/extractions/dc06/ascs
extern /pp2/main/extractions/dc06/asel
extern /pp2/main/extractions/dc06/avhn
extern /pp2/main/extractions/dc06/g2to
extern /pp2/main/extractions/dc06/gras
extern /pp2/main/extractions/dc06/hirs
extern /pp2/main/extractions/dc06/iasi
extern /pp2/main/extractions/dc06/j2eo
extern /pp2/main/extractions/dc06/j2nb
extern /pp2/main/extractions/dc06/mhsa
extern /pp2/main/extractions/dc06/modw
extern /pp2/main/extractions/dc12/amsa
extern /pp2/main/extractions/dc12/asc2
extern /pp2/main/extractions/dc12/ascs
extern /pp2/main/extractions/dc12/asel
extern /pp2/main/extractions/dc12/avhn
extern /pp2/main/extractions/dc12/g2to
extern /pp2/main/extractions/dc12/gras
extern /pp2/main/extractions/dc12/hirs
extern /pp2/main/extractions/dc12/iasi
extern /pp2/main/extractions/dc12/j2eo
extern /pp2/main/extractions/dc12/j2nb
extern /pp2/main/extractions/dc12/mhsa
extern /pp2/main/extractions/dc12/modw
extern /pp2/main/extractions/dc18/amsa
extern /pp2/main/extractions/dc18/asc2
extern /pp2/main/extractions/dc18/ascs
extern /pp2/main/extractions/dc18/asel
extern /pp2/main/extractions/dc18/avhn
extern /pp2/main/extractions/dc18/g2to
extern /pp2/main/extractions/dc18/gras
extern /pp2/main/extractions/dc18/hirs
extern /pp2/main/extractions/dc18/iasi
extern /pp2/main/extractions/dc18/j2eo
extern /pp2/main/extractions/dc18/j2nb
extern /pp2/main/extractions/dc18/mhsa
extern /pp2/main/extractions/dc18/modw
extern /pp2/main/extractions/ex00/amsa
extern /pp2/main/extractions/ex00/asc2
extern /pp2/main/extractions/ex00/ascs
extern /pp2/main/extractions/ex00/asel
extern /pp2/main/extractions/ex00/avhn
extern /pp2/main/extractions/ex00/g2to
extern /pp2/main/extractions/ex00/gras
extern /pp2/main/extractions/ex00/hirs
extern /pp2/main/extractions/ex00/iasi
extern /pp2/main/extractions/ex00/j2eo
extern /pp2/main/extractions/ex00/j2nb
extern /pp2/main/extractions/ex00/mhsa
extern /pp2/main/extractions/ex00/modw
extern /pp2/main/extractions/ex06/amsa
extern /pp2/main/extractions/ex06/asc2
extern /pp2/main/extractions/ex06/ascs
extern /pp2/main/extractions/ex06/asel
extern /pp2/main/extractions/ex06/avhn
extern /pp2/main/extractions/ex06/g2to
extern /pp2/main/extractions/ex06/gras
extern /pp2/main/extractions/ex06/hirs
extern /pp2/main/extractions/ex06/iasi
extern /pp2/main/extractions/ex06/j2eo
extern /pp2/main/extractions/ex06/j2nb
extern /pp2/main/extractions/ex06/mhsa
extern /pp2/main/extractions/ex06/modw
extern /pp2/main/extractions/ex12/amsa
extern /pp2/main/extractions/ex12/asc2
extern /pp2/main/extractions/ex12/ascs
extern /pp2/main/extractions/ex12/asel
extern /pp2/main/extractions/ex12/avhn
extern /pp2/main/extractions/ex12/g2to
extern /pp2/main/extractions/ex12/gras
extern /pp2/main/extractions/ex12/hirs
extern /pp2/main/extractions/ex12/iasi
extern /pp2/main/extractions/ex12/j2eo
extern /pp2/main/extractions/ex12/j2nb
extern /pp2/main/extractions/ex12/mhsa
extern /pp2/main/extractions/ex12/modw
extern /pp2/main/extractions/ex18/amsa
extern /pp2/main/extractions/ex18/asc2
extern /pp2/main/extractions/ex18/ascs
extern /pp2/main/extractions/ex18/asel
extern /pp2/main/extractions/ex18/avhn
extern /pp2/main/extractions/ex18/g2to
extern /pp2/main/extractions/ex18/gras
extern /pp2/main/extractions/ex18/hirs
extern /pp2/main/extractions/ex18/iasi
extern /pp2/main/extractions/ex18/j2eo
extern /pp2/main/extractions/ex18/j2nb
extern /pp2/main/extractions/ex18/mhsa
extern /pp2/main/extractions/ex18/modw
suite limits
  defstatus complete
  edit SLEEPTIME '1'
  limit mars 16
  limit ha_pp 10
  limit ha_pp2 10
  limit swarm 40
  limit linux_cluster 40
  limit obs_get 20
  limit hpcf 150
  limit c1a 150
  limit c1b 150
  limit prod 25
  limit prod2diss 6
  limit prod2ecpds 6
  limit pdb 10
  limit q2diss 10
  limit pdb_bc 6
  limit q2diss_bc 6
  clock hybrid 
endsuite
suite pp1
  edit ECF_DATE '20080711'
  edit SLEEPTIME '1'
  clock hybrid 
  family main
    edit YMD '20100114'
    family extractions
      family ex00
        task nonbin
        task airs
        task amsr
        task at18
        task at19
        task atap
        task ateu
        task atov
        task avhr
        task bice
        task bssh
        task btem
        task buoy
        task cmwi
        task csrj
        task eas1
        task eas2
        task emer
        task erad
        task ers1_at
        task ers1_ra
        task ers1_wa
        task ers1_wi
        task esck
        task goes
        task gomw
        task grad
        task grst
        family ice
          task ice_ex
          task ice_fdb_hpcf
          task ice_fdb_c1a
          task ice_fdb_c1b
        endfamily
        task jaso
        task masr
        family metop_a
          task amsa
          task asca
          task hirs
          task iasi
          task mhsa
        endfamily
        task nomi
        task qsca
        task rado
        task rtovs
        task sb19
        task sbuv
        task smin
        task smis
        task smiu
        task snow
        task ssbt
        family sst
          task sst_ex
          task sst_fdb_hpcf
          task sst_fdb_c1a
          task sst_fdb_c1b
        endfamily
        task stor
        task tozo
        task trmm
        task trmr
        task wavb
        task cmwn
        task ap19
      endfamily
      family ex06
        task nonbin
        task airs
        task amsr
        task at18
        task at19
        task atap
        task ateu
        task atov
        task avhr
        task bice
        task bssh
        task btem
        task cmwi
        task csrj
        task eas1
        task eas2
        task emer
        task erad
        task ers1_at
        task ers1_ra
        task ers1_wa
        task ers1_wi
        task esck
        task goes
        task gomw
        task grad
        task grst
        family ice
          task ice_ex
          task ice_fdb_hpcf
          task ice_fdb_c1a
          task ice_fdb_c1b
        endfamily
        task jaso
        task masr
        family metop_a
          task amsa
          task asca
          task hirs
          task iasi
          task mhsa
        endfamily
        task mtsat
        task nomi
        task qsca
        task rado
        task rtovs
        task sb19
        task sbuv
        task smin
        task smis
        task snow
        task ssbt
        family sst
          task sst_ex
          task sst_fdb_hpcf
          task sst_fdb_c1a
          task sst_fdb_c1b
        endfamily
        task stor
        task tozo
        task trmm
        task trmr
        task wavb
        task smiu
        task cmwn
        task ap19
      endfamily
      family ex12
        task nonbin
        task airs
        task amsr
        task at18
        task at19
        task atap
        task ateu
        task atov
        task avhr
        task bice
        task bssh
        task btem
        task cmwi
        task csrj
        task eas1
        task eas2
        task emer
        task erad
        task ers1_at
        task ers1_ra
        task ers1_wa
        task ers1_wi
        task esck
        task goes
        task gomw
        task grad
        task grst
        family ice
          task ice_ex
          task ice_fdb_hpcf
          task ice_fdb_c1a
          task ice_fdb_c1b
        endfamily
        task jaso
        task masr
        family metop_a
          task amsa
          task asca
          task hirs
          task iasi
          task mhsa
        endfamily
        task mtsat
        task nomi
        task qsca
        task rado
        task rtovs
        task sb19
        task sbuv
        task smin
        task smis
        task snow
        task ssbt
        family sst
          task sst_ex
          task sst_fdb_hpcf
          task sst_fdb_c1a
          task sst_fdb_c1b
        endfamily
        task stor
        task tozo
        task trmm
        task trmr
        task wavb
        task smiu
        task cmwn
        task ap19
      endfamily
      family ex18
        task nonbin
        task airs
        task amsr
        task at18
        task at19
        task atap
        task ateu
        task atov
        task avhr
        task bice
        task bssh
        task btem
        task cmwi
        task csrj
        task eas1
        task eas2
        task emer
        task erad
        task ers1_at
        task ers1_ra
        task ers1_wa
        task ers1_wi
        task esck
        task goes
        task gomw
        task grad
        task grst
        family ice
          task ice_ex
          task ice_fdb_hpcf
          task ice_fdb_c1a
          task ice_fdb_c1b
        endfamily
        task jaso
        task masr
        family metop_a
          task amsa
          task asca
          task hirs
          task iasi
          task mhsa
        endfamily
        task mtsat
        task nomi
        task qsca
        task rado
        task rtovs
        task sb19
        task sbuv
        task smin
        task smis
        task snow
        task ssbt
        family sst
          task sst_ex
          task sst_fdb_hpcf
          task sst_fdb_c1a
          task sst_fdb_c1b
        endfamily
        task stor
        task tozo
        task trmm
        task trmr
        task wavb
        task smiu
        task cmwn
        task ap19
      endfamily
      family dc00
        task nonbin
        task airs
        task amsr
        task at18
        task at19
        task atap
        task ateu
        task atov
        task avhr
        task bice
        task bssh
        task btem
        task cmwi
        task csrj
        task eas1
        task eas2
        task emer
        task erad
        task ers1_at
        task ers1_ra
        task ers1_wa
        task ers1_wi
        task esck
        task goes
        task gomw
        task grad
        task grst
        family ice
          task ice_ex
          task ice_fdb_hpcf
          task ice_fdb_c1a
          task ice_fdb_c1b
        endfamily
        task jaso
        task masr
        family metop_a
          task amsa
          task asca
          task hirs
          task iasi
          task mhsa
        endfamily
        task mtsat
        task nomi
        task qsca
        task rado
        task rtovs
        task sb19
        task sbuv
        task smin
        task smis
        task snow
        task ssbt
        family sst
          task sst_ex
          task sst_fdb_hpcf
          task sst_fdb_c1a
          task sst_fdb_c1b
        endfamily
        task stor
        task tozo
        task trmm
        task trmr
        task wavb
        task smiu
        task cmwn
        task ap19
      endfamily
      family dc06
        task nonbin
        task airs
        task amsr
        task at18
        task at19
        task atap
        task ateu
        task atov
        task avhr
        task bice
        task bssh
        task btem
        task cmwi
        task csrj
        task eas1
        task eas2
        task emer
        task erad
        task ers1_at
        task ers1_ra
        task ers1_wa
        task ers1_wi
        task esck
        task goes
        task gomw
        task grad
        task grst
        family ice
          task ice_ex
          task ice_fdb_hpcf
          task ice_fdb_c1a
          task ice_fdb_c1b
        endfamily
        task jaso
        task masr
        family metop_a
          task amsa
          task asca
          task hirs
          task iasi
          task mhsa
        endfamily
        task mtsat
        task nomi
        task qsca
        task rado
        task rtovs
        task sb19
        task sbuv
        task smin
        task smis
        task snow
        task ssbt
        family sst
          task sst_ex
          task sst_fdb_hpcf
          task sst_fdb_c1a
          task sst_fdb_c1b
        endfamily
        task stor
        task tozo
        task trmm
        task trmr
        task wavb
        task smiu
        task cmwn
        task ap19
      endfamily
      family dc12
        task nonbin
        task airs
        task amsr
        task at18
        task at19
        task atap
        task ateu
        task atov
        task avhr
        task bice
        task bssh
        task btem
        task cmwi
        task csrj
        task eas1
        task eas2
        task emer
        task erad
        task ers1_at
        task ers1_ra
        task ers1_wa
        task ers1_wi
        task esck
        task goes
        task gomw
        task grad
        task grst
        family ice
          task ice_ex
          task ice_fdb_hpcf
          task ice_fdb_c1a
          task ice_fdb_c1b
        endfamily
        task jaso
        task masr
        family metop_a
          task amsa
          task asca
          task hirs
          task iasi
          task mhsa
        endfamily
        task mtsat
        task nomi
        task qsca
        task rado
        task rtovs
        task sb19
        task sbuv
        task smin
        task smis
        task snow
        task ssbt
        family sst
          task sst_ex
          task sst_fdb_hpcf
          task sst_fdb_c1a
          task sst_fdb_c1b
        endfamily
        task stor
        task tozo
        task trmm
        task trmr
        task wavb
        task smiu
        task cmwn
        task ap19
      endfamily
      family dc18
        task nonbin
        task airs
        task amsr
        task at18
        task at19
        task atap
        task ateu
        task atov
        task avhr
        task bice
        task bssh
        task btem
        task cmwi
        task csrj
        task eas1
        task eas2
        task emer
        task erad
        task ers1_at
        task ers1_ra
        task ers1_wa
        task ers1_wi
        task esck
        task goes
        task gomw
        task grad
        task grst
        family ice
          task ice_ex
          task ice_fdb_hpcf
          task ice_fdb_c1a
          task ice_fdb_c1b
        endfamily
        task jaso
        task masr
        family metop_a
          task amsa
          task asca
          task hirs
          task iasi
          task mhsa
        endfamily
        task mtsat
        task nomi
        task qsca
        task rado
        task rtovs
        task sb19
        task sbuv
        task smin
        task smis
        task snow
        task ssbt
        family sst
          task sst_ex
          task sst_fdb_hpcf
          task sst_fdb_c1a
          task sst_fdb_c1b
        endfamily
        task stor
        task tozo
        task trmm
        task trmr
        task wavb
        task smiu
        task cmwn
        task ap19
      endfamily
      family ma
        family egrr
          task extract
          task fdb_c1a
          task fdb_c1b
          task fdb_hpcf
        endfamily
        family edzw
          task extract
          task fdb_c1a
          task fdb_c1b
          task fdb_hpcf
        endfamily
        family lfpw
          task extract
          task fdb_c1a
          task fdb_c1b
          task fdb_hpcf
        endfamily
        family kwbc
          task extract
          task fdb_c1a
          task fdb_c1b
          task fdb_hpcf
        endfamily
      endfamily
    endfamily
  endfamily
endsuite
suite o
  edit SMSFILES '/home/ma/emos/def/sync'
  edit SLEEPTIME '1'
  clock hybrid 
  family main
    edit YMD '20100114'
    family 00
      family an
        task 4dvar
        task surf_anal
      endfamily
      family prod
        family 18an
          task derived_param
        endfamily
      endfamily
      family prod2diss
        task oper
        task wave
        family sched
          task 240
        endfamily
      endfamily
      family fc
        task model
        task clean
      endfamily
    endfamily
    family 12
      family an
        task 4dvar
        task surf_anal
      endfamily
      family prod
        family 06an
          task derived_param
        endfamily
      endfamily
      family prod2diss
        task oper
        task wave
        family sched
          task 240
        endfamily
      endfamily
      family fc
        task model
      endfamily
      family obs
        family get
          task get_stor
        endfamily
      endfamily
    endfamily
    family 00dc
      family an
        task 4dvar
        task surf_anal
      endfamily
      family fc
        task model
        task clean
      endfamily
    endfamily
    family 12dc
      family an
        task 4dvar
        task surf_anal
      endfamily
      family fc
        task model
      endfamily
      family obs
        family get
          task get_stor
        endfamily
      endfamily
    endfamily
    family 06bc
      family an
        task 4dvar
        task surf_anal
      endfamily
      family prod2diss
        task oper
        family sched
          task 72
        endfamily
      endfamily
      family fc
        task model
      endfamily
    endfamily
    family 18bc
      family an
        task 4dvar
        task surf_anal
      endfamily
      family prod2diss
        task oper
        family sched
          task 72
        endfamily
      endfamily
      family fc
        task model
      endfamily
    endfamily
  endfamily
  family lag
    edit YMD '20100114'
    family 00dc
      family fb
        family feedback
          task matchup
        endfamily
      endfamily
      task archive
    endfamily
    family 06
      family fb
        family feedback
          task matchup
        endfamily
      endfamily
      task archive
    endfamily
    family 12
      family fb
        family feedback
          task matchup
        endfamily
      endfamily
      task satim
      task verif
      family archive
        task obs
        task fc240
      endfamily
    endfamily
    family 12dc
      family fb
        family feedback
          task matchup
        endfamily
      endfamily
      task archive
    endfamily
    family 18
      family fb
        family feedback
          task matchup
        endfamily
      endfamily
      task archive
    endfamily
    family 00
      family fb
        family feedback
          task matchup
        endfamily
      endfamily
      task satim
      task verif
      family archive
        task obs
        task anml
        task anpl
        task ansfc
        task sst
        task fc048
        task fc240
      endfamily
    endfamily
    family 18bc
      family fb
        family feedback
          task matchup
        endfamily
      endfamily
      task archive
    endfamily
    family 06bc
      family fb
        family feedback
          task matchup
        endfamily
      endfamily
      task archive
    endfamily
  endfamily
endsuite
suite pp2
  edit ECF_DATE '20091117'
  edit SLEEPTIME '1'
  clock hybrid 
  family main
    edit YMD '20100114'
    family extractions
      family ex00
        task amsa
        task amse
        task asc2
        task asca
        task ascs
        task asch
        task aseh
        task asel
        task ashs
        task avhn
        task cori
        task earn
        task emsg
        task er19
        task g2to
        task gras
        task hirs
        task iasi
        task ias1
        task ias2
        task ias3
        task ias4
        task itwt
        task j2eo
        task j2nb
        task mhsa
        task mhse
        task mloz
        task modw
      endfamily
      family ex06
        task amsa
        task amse
        task asc2
        task asca
        task ascs
        task asch
        task aseh
        task asel
        task ashs
        task avhn
        task cori
        task earn
        task emsg
        task er19
        task g2to
        task gras
        task hirs
        task iasi
        task ias1
        task ias2
        task ias3
        task ias4
        task itwt
        task j2eo
        task j2nb
        task mhsa
        task mhse
        task mloz
        task modw
      endfamily
      family ex12
        task amsa
        task amse
        task asc2
        task asca
        task ascs
        task asch
        task aseh
        task asel
        task ashs
        task avhn
        task cori
        task earn
        task emsg
        task er19
        task g2to
        task gras
        task hirs
        task iasi
        task ias1
        task ias2
        task ias3
        task ias4
        task itwt
        task j2eo
        task j2nb
        task mhsa
        task mhse
        task mloz
        task modw
      endfamily
      family ex18
        task amsa
        task amse
        task asc2
        task asca
        task ascs
        task asch
        task aseh
        task asel
        task ashs
        task avhn
        task cori
        task earn
        task emsg
        task er19
        task g2to
        task gras
        task hirs
        task iasi
        task ias1
        task ias2
        task ias3
        task ias4
        task itwt
        task j2eo
        task j2nb
        task mhsa
        task mhse
        task mloz
        task modw
      endfamily
      family dc00
        task amsa
        task amse
        task asc2
        task asca
        task ascs
        task asch
        task aseh
        task asel
        task ashs
        task avhn
        task cori
        task earn
        task emsg
        task er19
        task g2to
        task gras
        task hirs
        task iasi
        task ias1
        task ias2
        task ias3
        task ias4
        task itwt
        task j2eo
        task j2nb
        task mhsa
        task mhse
        task mloz
        task modw
      endfamily
      family dc06
        task amsa
        task amse
        task asc2
        task asca
        task ascs
        task asch
        task aseh
        task asel
        task ashs
        task avhn
        task cori
        task earn
        task emsg
        task er19
        task g2to
        task gras
        task hirs
        task iasi
        task ias1
        task ias2
        task ias3
        task ias4
        task itwt
        task j2eo
        task j2nb
        task mhsa
        task mhse
        task mloz
        task modw
      endfamily
      family dc12
        task amsa
        task amse
        task asc2
        task asca
        task ascs
        task asch
        task aseh
        task asel
        task ashs
        task avhn
        task cori
        task earn
        task emsg
        task er19
        task g2to
        task gras
        task hirs
        task iasi
        task ias1
        task ias2
        task ias3
        task ias4
        task itwt
        task j2eo
        task j2nb
        task mhsa
        task mhse
        task mloz
        task modw
      endfamily
      family dc18
        task amsa
        task amse
        task asc2
        task asca
        task ascs
        task asch
        task aseh
        task asel
        task ashs
        task avhn
        task cori
        task earn
        task emsg
        task er19
        task g2to
        task gras
        task hirs
        task iasi
        task ias1
        task ias2
        task ias3
        task ias4
        task itwt
        task j2eo
        task j2nb
        task mhsa
        task mhse
        task mloz
        task modw
      endfamily
    endfamily
  endfamily
endsuite
suite pp3
  edit ECF_DATE '20121212'
  edit SLEEPTIME '1'
  clock hybrid 
  family limits
  endfamily
  family first
    task log
      event tar
    task clean
    task clean_stat
    task clean_err_db
    task pp_src
    task chktasks
  endfamily
  family meb
    task restore
    task clean
    task stat
    task feed
  endfamily
  family reb
    task restore
    task clean
      event deqc
    task stat
  endfamily
  family bkup
    family meb
      task strhandle
      task MAER
    endfamily
    family reb
      task strhandle
      task MAER
    endfamily
  endfamily
  family deqc
    task MAER
  endfamily
  family ex00
    task maer
      event 1 
  endfamily
  family ex06
    task maer
      event 1 
  endfamily
  family ex12
    task maer
      event 1 
  endfamily
  family ex18
    task maer
      event 1 
  endfamily
  family dc00
    task maer
      event 1 
  endfamily
  family dc06
    task maer
      event 1 
  endfamily
  family dc12
    task maer
      event 1 
  endfamily
  family dc18
    task maer
      event 1 
  endfamily
  family last
    task stop
    task chksuite
  endfamily
endsuite
suite mc
  edit SMSFILES '/home/ma/emos/def/sync'
  edit SLEEPTIME '1'
  clock hybrid 
  family main
    edit YMD '20100114'
    family 00
      family sv
        task svsh
        task svnh
        task sv1
        task sv2
        task sv3
        task sv4
        task sv5
        task sv6
      endfamily
      family legA
        family fc
          task pert_ic
          task pf
          task cf
        endfamily
        task postproc
      endfamily
      family legB
        task postproc
      endfamily
    endfamily
    family 12
      family legA
        family fc
          task pf
          task cf
        endfamily
        task postproc
      endfamily
      family legB
        task postproc
      endfamily
    endfamily
  endfamily
endsuite
suite esync
  edit USER 'emos'
  edit QUEUE 'emos-exp'
  edit ACCOUNT 'oosu'
  edit WSHOST 'localhost'
  edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
  edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID%'
  edit SMSOUT '/vol/emos/output'
  edit LOGDIR '/vol/emos/output'
  edit SMSHOME '/vol/emos/output'
  edit SMSFILES '/home/ma/emos/def/esync'
  edit SMSINCLUDE '/home/ma/emos/def/esync/include'
  edit SMSTRIES '1'
  edit SYNC_HOSTS 'marhaus percard marhaus percard ibis'
  edit SYNC_PORTS ' 314122 314111 314288 314155 314159 314159'
  edit SYNC_SUITES 'o pp1'
  edit SLEEPTIME '1'
  inlimit /esync/limits:sync
  clock hybrid 
  family limits
    limit sync 3
  endfamily
  family o
    family o
      edit SMSFILES '/home/ma/emos/def/sync'
      family main
        edit YMD '20100114'
        family 00
          family an
            family 4dvar
              task sync
                trigger /o/main/00/an/4dvar eq complete
                edit SYNC_NODE '/o/main/00/an/4dvar'
            endfamily
            family surf_anal
              task sync
                trigger /o/main/00/an/surf_anal eq complete
                edit SYNC_NODE '/o/main/00/an/surf_anal'
            endfamily
          endfamily
          family prod
            family 18an
              family derived_param
                task sync
                  trigger /o/main/00/prod/18an/derived_param eq complete
                  edit SYNC_NODE '/o/main/00/prod/18an/derived_param'
              endfamily
            endfamily
          endfamily
          family prod2diss
            family oper
              task sync
                trigger /o/main/00/prod2diss/oper eq complete
                edit SYNC_NODE '/o/main/00/prod2diss/oper'
            endfamily
            family wave
              task sync
                trigger /o/main/00/prod2diss/wave eq complete
                edit SYNC_NODE '/o/main/00/prod2diss/wave'
            endfamily
            family sched
              family 240
                task sync
                  trigger /o/main/00/prod2diss/sched/240 eq complete
                  edit SYNC_NODE '/o/main/00/prod2diss/sched/240'
              endfamily
            endfamily
          endfamily
          family fc
            family model
              task sync
                trigger /o/main/00/fc/model eq complete
                edit SYNC_NODE '/o/main/00/fc/model'
            endfamily
            family clean
              task sync
                trigger /o/main/00/fc/clean eq complete
                edit SYNC_NODE '/o/main/00/fc/clean'
            endfamily
          endfamily
        endfamily
        family 12
          family an
            family 4dvar
              task sync
                trigger /o/main/12/an/4dvar eq complete
                edit SYNC_NODE '/o/main/12/an/4dvar'
            endfamily
            family surf_anal
              task sync
                trigger /o/main/12/an/surf_anal eq complete
                edit SYNC_NODE '/o/main/12/an/surf_anal'
            endfamily
          endfamily
          family prod
            family 06an
              family derived_param
                task sync
                  trigger /o/main/12/prod/06an/derived_param eq complete
                  edit SYNC_NODE '/o/main/12/prod/06an/derived_param'
              endfamily
            endfamily
          endfamily
          family prod2diss
            family oper
              task sync
                trigger /o/main/12/prod2diss/oper eq complete
                edit SYNC_NODE '/o/main/12/prod2diss/oper'
            endfamily
            family wave
              task sync
                trigger /o/main/12/prod2diss/wave eq complete
                edit SYNC_NODE '/o/main/12/prod2diss/wave'
            endfamily
            family sched
              family 240
                task sync
                  trigger /o/main/12/prod2diss/sched/240 eq complete
                  edit SYNC_NODE '/o/main/12/prod2diss/sched/240'
              endfamily
            endfamily
          endfamily
          family fc
            family model
              task sync
                trigger /o/main/12/fc/model eq complete
                edit SYNC_NODE '/o/main/12/fc/model'
            endfamily
          endfamily
          family obs
            family get
              family get_stor
                task sync
                  trigger /o/main/12/obs/get/get_stor eq complete
                  edit SYNC_NODE '/o/main/12/obs/get/get_stor'
              endfamily
            endfamily
          endfamily
        endfamily
        family 00dc
          family an
            family 4dvar
              task sync
                trigger /o/main/00dc/an/4dvar eq complete
                edit SYNC_NODE '/o/main/00dc/an/4dvar'
            endfamily
            family surf_anal
              task sync
                trigger /o/main/00dc/an/surf_anal eq complete
                edit SYNC_NODE '/o/main/00dc/an/surf_anal'
            endfamily
          endfamily
          family fc
            family model
              task sync
                trigger /o/main/00dc/fc/model eq complete
                edit SYNC_NODE '/o/main/00dc/fc/model'
            endfamily
            family clean
              task sync
                trigger /o/main/00dc/fc/clean eq complete
                edit SYNC_NODE '/o/main/00dc/fc/clean'
            endfamily
          endfamily
        endfamily
        family 12dc
          family an
            family 4dvar
              task sync
                trigger /o/main/12dc/an/4dvar eq complete
                edit SYNC_NODE '/o/main/12dc/an/4dvar'
            endfamily
            family surf_anal
              task sync
                trigger /o/main/12dc/an/surf_anal eq complete
                edit SYNC_NODE '/o/main/12dc/an/surf_anal'
            endfamily
          endfamily
          family fc
            family model
              task sync
                trigger /o/main/12dc/fc/model eq complete
                edit SYNC_NODE '/o/main/12dc/fc/model'
            endfamily
          endfamily
          family obs
            family get
              family get_stor
                task sync
                  trigger /o/main/12dc/obs/get/get_stor eq complete
                  edit SYNC_NODE '/o/main/12dc/obs/get/get_stor'
              endfamily
            endfamily
          endfamily
        endfamily
        family 06bc
          family an
            family 4dvar
              task sync
                trigger /o/main/06bc/an/4dvar eq complete
                edit SYNC_NODE '/o/main/06bc/an/4dvar'
            endfamily
            family surf_anal
              task sync
                trigger /o/main/06bc/an/surf_anal eq complete
                edit SYNC_NODE '/o/main/06bc/an/surf_anal'
            endfamily
          endfamily
          family prod2diss
            family oper
              task sync
                trigger /o/main/06bc/prod2diss/oper eq complete
                edit SYNC_NODE '/o/main/06bc/prod2diss/oper'
            endfamily
            family sched
              family 72
                task sync
                  trigger /o/main/06bc/prod2diss/sched/72 eq complete
                  edit SYNC_NODE '/o/main/06bc/prod2diss/sched/72'
              endfamily
            endfamily
          endfamily
          family fc
            family model
              task sync
                trigger /o/main/06bc/fc/model eq complete
                edit SYNC_NODE '/o/main/06bc/fc/model'
            endfamily
          endfamily
        endfamily
        family 18bc
          family an
            family 4dvar
              task sync
                trigger /o/main/18bc/an/4dvar eq complete
                edit SYNC_NODE '/o/main/18bc/an/4dvar'
            endfamily
            family surf_anal
              task sync
                trigger /o/main/18bc/an/surf_anal eq complete
                edit SYNC_NODE '/o/main/18bc/an/surf_anal'
            endfamily
          endfamily
          family prod2diss
            family oper
              task sync
                trigger /o/main/18bc/prod2diss/oper eq complete
                edit SYNC_NODE '/o/main/18bc/prod2diss/oper'
            endfamily
            family sched
              family 72
                task sync
                  trigger /o/main/18bc/prod2diss/sched/72 eq complete
                  edit SYNC_NODE '/o/main/18bc/prod2diss/sched/72'
              endfamily
            endfamily
          endfamily
          family fc
            family model
              task sync
                trigger /o/main/18bc/fc/model eq complete
                edit SYNC_NODE '/o/main/18bc/fc/model'
            endfamily
          endfamily
        endfamily
      endfamily
      family lag
        edit YMD '20100114'
        family 00dc
          family fb
            family feedback
              family matchup
                task sync
                  trigger /o/lag/00dc/fb/feedback/matchup eq complete
                  edit SYNC_NODE '/o/lag/00dc/fb/feedback/matchup'
              endfamily
            endfamily
          endfamily
          family archive
            task sync
              trigger /o/lag/00dc/archive eq complete
              edit SYNC_NODE '/o/lag/00dc/archive'
          endfamily
        endfamily
        family 06
          family fb
            family feedback
              family matchup
                task sync
                  trigger /o/lag/06/fb/feedback/matchup eq complete
                  edit SYNC_NODE '/o/lag/06/fb/feedback/matchup'
              endfamily
            endfamily
          endfamily
          family archive
            task sync
              trigger /o/lag/06/archive eq complete
              edit SYNC_NODE '/o/lag/06/archive'
          endfamily
        endfamily
        family 12
          family fb
            family feedback
              family matchup
                task sync
                  trigger /o/lag/12/fb/feedback/matchup eq complete
                  edit SYNC_NODE '/o/lag/12/fb/feedback/matchup'
              endfamily
            endfamily
          endfamily
          family satim
            task sync
              trigger /o/lag/12/satim eq complete
              edit SYNC_NODE '/o/lag/12/satim'
          endfamily
          family verif
            task sync
              trigger /o/lag/12/verif eq complete
              edit SYNC_NODE '/o/lag/12/verif'
          endfamily
          family archive
            family obs
              task sync
                trigger /o/lag/12/archive/obs eq complete
                edit SYNC_NODE '/o/lag/12/archive/obs'
            endfamily
            family fc240
              task sync
                trigger /o/lag/12/archive/fc240 eq complete
                edit SYNC_NODE '/o/lag/12/archive/fc240'
            endfamily
          endfamily
        endfamily
        family 12dc
          family fb
            family feedback
              family matchup
                task sync
                  trigger /o/lag/12dc/fb/feedback/matchup eq complete
                  edit SYNC_NODE '/o/lag/12dc/fb/feedback/matchup'
              endfamily
            endfamily
          endfamily
          family archive
            task sync
              trigger /o/lag/12dc/archive eq complete
              edit SYNC_NODE '/o/lag/12dc/archive'
          endfamily
        endfamily
        family 18
          family fb
            family feedback
              family matchup
                task sync
                  trigger /o/lag/18/fb/feedback/matchup eq complete
                  edit SYNC_NODE '/o/lag/18/fb/feedback/matchup'
              endfamily
            endfamily
          endfamily
          family archive
            task sync
              trigger /o/lag/18/archive eq complete
              edit SYNC_NODE '/o/lag/18/archive'
          endfamily
        endfamily
        family 00
          family fb
            family feedback
              family matchup
                task sync
                  trigger /o/lag/00/fb/feedback/matchup eq complete
                  edit SYNC_NODE '/o/lag/00/fb/feedback/matchup'
              endfamily
            endfamily
          endfamily
          family satim
            task sync
              trigger /o/lag/00/satim eq complete
              edit SYNC_NODE '/o/lag/00/satim'
          endfamily
          family verif
            task sync
              trigger /o/lag/00/verif eq complete
              edit SYNC_NODE '/o/lag/00/verif'
          endfamily
          family archive
            family obs
              task sync
                trigger /o/lag/00/archive/obs eq complete
                edit SYNC_NODE '/o/lag/00/archive/obs'
            endfamily
            family anml
              task sync
                trigger /o/lag/00/archive/anml eq complete
                edit SYNC_NODE '/o/lag/00/archive/anml'
            endfamily
            family anpl
              task sync
                trigger /o/lag/00/archive/anpl eq complete
                edit SYNC_NODE '/o/lag/00/archive/anpl'
            endfamily
            family ansfc
              task sync
                trigger /o/lag/00/archive/ansfc eq complete
                edit SYNC_NODE '/o/lag/00/archive/ansfc'
            endfamily
            family sst
              task sync
                trigger /o/lag/00/archive/sst eq complete
                edit SYNC_NODE '/o/lag/00/archive/sst'
            endfamily
            family fc048
              task sync
                trigger /o/lag/00/archive/fc048 eq complete
                edit SYNC_NODE '/o/lag/00/archive/fc048'
            endfamily
            family fc240
              task sync
                trigger /o/lag/00/archive/fc240 eq complete
                edit SYNC_NODE '/o/lag/00/archive/fc240'
            endfamily
          endfamily
        endfamily
        family 18bc
          family fb
            family feedback
              family matchup
                task sync
                  trigger /o/lag/18bc/fb/feedback/matchup eq complete
                  edit SYNC_NODE '/o/lag/18bc/fb/feedback/matchup'
              endfamily
            endfamily
          endfamily
          family archive
            task sync
              trigger /o/lag/18bc/archive eq complete
              edit SYNC_NODE '/o/lag/18bc/archive'
          endfamily
        endfamily
        family 06bc
          family fb
            family feedback
              family matchup
                task sync
                  trigger /o/lag/06bc/fb/feedback/matchup eq complete
                  edit SYNC_NODE '/o/lag/06bc/fb/feedback/matchup'
              endfamily
            endfamily
          endfamily
          family archive
            task sync
              trigger /o/lag/06bc/archive eq complete
              edit SYNC_NODE '/o/lag/06bc/archive'
          endfamily
        endfamily
      endfamily
    endfamily
    task restart
      trigger ( ( /o/lag:YMD gt /esync/o/o/lag:YMD) or ( /o/main:YMD gt /esync/o/o/main:YMD) or 1 eq 0) and /esync/o ne active and /esync/o ne submitted
      edit SMSPASS 'stripped'
  endfamily
endsuite
suite admin3
  repeat day 1
  edit USER 'emos'
  edit ACCOUNT 'oosu'
  edit SCHOST 'hpce'
  edit VERSION '0011'
  edit FIRST_DAY '0'
  edit SMSTRIES '2'
  edit DELTA_DAY '0'
  edit EMOS_BASE '12'
  edit EMOS_TIME_STEP_H '0000'
  edit FSFAMILY ''
  edit STREAM 'TMP'
  edit EPSTYPE 'fc'
  edit MEMBER '0'
  edit INIMA ''
  edit INIMACONS '0'
  edit WSHOST 'swarm'
  edit QUEUE 'emos-e'
  edit QUEUE_EPILOG 'emos-exp'
  edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
  edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID%'
  edit LOGDIR '/vol/emos/output'
  edit SMSOUT '/vol/emos/output'
  edit SMSHOME '/vol/emos/output'
  edit SMSINCLUDE '/home/ma/emos/def/o/include'
  edit SMSFILES '/home/ma/emos/def/o/admin'
  edit SERVER_HOME '/var/tmp/emos/sms'
  edit CLASS 'od'
  edit DATEMASK '*.*.*'
  edit SLEEPTIME '1'
  inlimit /limits:swarm
  clock hybrid 
  family daily
    repeat date YMD 20070115 20121212 1
    edit USE_YMD 'true'
    edit LOGTASK 'daily'
    family check
      edit QUEUE 'emos-e'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID%'
      edit LOGDIR '/vol/emos/output'
      edit SMSOUT '/vol/emos/output'
      task rcp
        time 00:00 23:59 00:10
        time 00:00 06:59 00:10
      task ecfs
        edit SMSTRIES '3'
        time 00:00 23:59 00:30
        time 00:00 06:59 00:30
    endfamily
    family files
      task usr_edt
      task smslog
        trigger usr_edt eq complete or usr_edt eq aborted
        edit QUEUE 'emos-e'
        edit QUEUE_EPILOG 'emos-exp'
        edit WSHOST 'localhost'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID%'
        edit LOGDIR '/vol/emos/output'
        edit SMSOUT '/vol/emos/output'
      task no_jobs
        trigger smslog eq complete or smslog eq aborted
      task tar
        trigger no_jobs eq complete or no_jobs eq aborted
    endfamily
    task logfilestore
      edit QUEUE 'emos-e'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID%'
      edit LOGDIR '/vol/emos/output'
      edit SMSOUT '/vol/emos/output'
      edit SMSFILES '/home/ma/emos/def/admin'
      edit MOVE_TO_ECFS 'yes'
      edit LOGTASK 'daily'
    task logfiles
      trigger ( check eq complete and files eq complete)
      edit QUEUE 'emos-e'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID%'
      edit LOGDIR '/vol/emos/output'
      edit SMSOUT '/vol/emos/output'
      edit KEEPLOGS 'no'
      edit MOVE_TO_ECFS 'no'
      edit SMSFILES '/home/ma/emos/def/o'
      time 06:45
  endfamily
endsuite
suite new_monitor
  edit SMSOUT '/vol/emos_nc/output'
  edit SMSHOME '/vol/emos_nc/output'
  edit SMSFILES '/home/ma/emos/def/new_monitor'
  edit SMSINCLUDE '/home/ma/emos/def/new_monitor/include'
  edit WSHOST 'bilbo'
  edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
  edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
  edit LOGDIR '/vol/emos_nc/output'
  edit EMPDIR '/usr/local/apps/empress/v8.60'
  edit EMPTMPDIR '/acq/tmp/empress'
  edit EDBPATH 'not_used'
  edit EPSTYPE 'not_used'
  edit MEMBER '0'
  edit FIRST_DAY '0'
  edit VERSION '0001'
  edit YMD 'not_used'
  edit DELTA_DAY '0'
  edit WDIR 'not_used'
  edit PACKAGE 'not_used'
  edit SCHOST 'not_used'
  edit HAACQ 'not_used'
  edit AN_TIME 'not_used'
  edit EDBNAME 'not_used'
  edit PP_BASE 'not_used'
  edit EXPVER 'not_used'
  edit STREAM 'not_used'
  edit USEEXTRACTED 'not_used'
  edit SMSHOST 'not_used'
  edit USER 'emos'
  edit QUEUE 'emos'
  edit ACCOUNT 'not_used'
  edit FSFAMILY 'disk_space'
  edit EMOS_BASE '00'
  edit EMOS_TIME_STEP_H '0000'
  edit DATEMASK '*.*.*'
  edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
  edit SLEEPTIME '1'
  clock hybrid 
  family admin
    task display_seaice
      edit WSHOST 'swarm'
      edit SMSLOG '/vol/emos/output/monitor'
      edit MCSLOG '/vol/emos/output'
      edit FSFAMILY 'undef'
      edit EMOS_BASE '12'
      edit SMSURLBASE 'http://w3ec2.ecmwf.int/metapps/preprocessing/Monitoring'
      edit SMSURL 'seaice.html'
      edit YMD '20071025'
      edit METVIEW 'metview'
      edit EMOS_TYPE 'undef'
      edit MARS_DB 'mars_od'
  endfamily
  family disk_space
    edit SMSOUT '/vol/emos_nc/output'
    edit SMSHOME '/vol/emos_nc/output'
    edit SMSLOG '/vol/emos_nc/output'
    edit LOGDIR '/vol/emos_nc/output'
    task chkspace
      edit WSHOST 'swarm'
      edit SMSPASS 'stripped'
      cron 00:00 23:00 00:30
    task chkspace_img
      edit WSHOST 'bilbo'
      edit SMSCMD '/home/ma/emos/bin/smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      cron 00:05 23:05 00:30
    task chkspace_img_copy
      edit WSHOST 'localhost'
      cron 00:10 23:10 00:30
    task rd_chkspace
      edit WSHOST 'swarm'
      cron 00:15 23:15 00:30
    task rd_chkspace_img
      edit WSHOST 'bilbo'
      cron 00:20 23:20 00:30
    task rd_chkspace_img_copy
      edit WSHOST 'localhost'
      cron 00:25 23:25 00:30
    task chkfeed
      cron 00:00 23:00 00:30
  endfamily
  family stats
    edit FSFAMILY 'stats'
    edit ODBCINIFILE '/home/ma/emos/maw/stats2/odbc.ini'
    task chk_stats_meb
      label info ""
      cron 07:00
    task chk_stats_reb
      label info ""
      cron 07:30
    task chk_stats_reb2
      label info ""
      cron 07:35
    task chk_grib_meb
      cron 09:30
    task chk_grib_reb
      cron 09:30
    task copy_stats_to_intra
      edit WSHOST 'localhost'
      cron 09:45
  endfamily
  family acq2
    edit WSHOST 'localhost'
    edit WARNINGMAIL 'true'
    family emc1
      task amsua
        edit WARNINGMAIL 'true'
        cron 00:05 23:05 00:30
      task mhs
        edit WARNINGMAIL 'false'
        cron 00:05 23:05 00:30
      task hirs
        edit WARNINGMAIL 'false'
        cron 00:05 23:05 00:30
    endfamily
    family emc2
      task amsua
        edit WARNINGMAIL 'false'
        cron 00:08 23:08 00:30
      task mhs
        edit WARNINGMAIL 'false'
        cron 00:08 23:08 00:30
      task hirs
        edit WARNINGMAIL 'false'
        cron 00:08 23:08 00:30
    endfamily
  endfamily
  family radar
    task get_radar
      edit WSHOST 'localhost'
      edit SMSLOG '/vol/emos/output/monitor'
      edit MCSLOG '/vol/emos/output'
      edit FSFAMILY 'radar'
      cron 00:00 23:00 00:15
    task get_img
      edit WSHOST 'bilbo'
      edit SMSLOG '/vol/emos/output/monitor'
      edit MCSLOG '/vol/emos/output'
      edit FSFAMILY 'radar'
      cron 00:05 23:05 00:15
  endfamily
  family db_locks
    edit SMSOUT '/vol/emos_nc/output'
    edit SMSHOME '/vol/emos_nc/output'
    edit SMSLOG '/vol/emos_nc/output'
    edit LOGDIR '/vol/emos_nc/output'
    task lockinfo_meb
      edit WSHOST 'pp1'
      cron 09:00 15:00 00:02
    task lockinfo_reb
      edit WSHOST 'pp1'
      cron 09:01 15:01 00:02
    task lockinfo_graph
      edit WSHOST 'bilbo'
      cron 09:05 15:05 00:02
    task lockinfo_img_copy
      edit WSHOST 'localhost'
      cron 09:06 15:06 00:02
  endfamily
endsuite
suite preview
  edit USER 'emos'
  edit SCHOST 'c1a'
  edit SCHOST_BKUP 'c1b'
  edit WSHOST 'swarm'
  edit WSLOGDIR '/vol/emos/output'
  edit SCLOGDIR '/emos_dir/0001/log'
  edit USE_YMD 'true'
  edit YMD '20070712'
  edit DATEMASK '*.*.*'
  edit VERSION '0001'
  edit INIEXPVER '0001'
  edit TCINIEXPVER '0001'
  edit FIRST_DAY '0'
  edit SMSTRIES '2'
  edit DELTA_DAY '0'
  edit EMOS_BASE '00'
  edit EMOS_TIME_STEP_H '0000'
  edit FSFAMILY 'tc'
  edit USER_PRIORITY '70'
  edit REINITIALIZE 'false'
  edit SMSHOME '/vol/emos/output'
  edit SMSFILES '/home/ma/emos/def/preview/smsfiles'
  edit SMSINCLUDE '/home/ma/emos/def/preview/include'
  edit ACCOUNT 'oosu'
  edit STREAM 'DA'
  edit EMOS_STREAM 'EF'
  edit ENSEMBLES '50'
  edit ENS_SIZE '50'
  edit MARS_DB 'marsod'
  edit EPSTYPE 'fc'
  edit MEMBER '0'
  edit IS_REAL_TIME 'is_real_time'
  edit RUN_3HOURS_PP '1'
  edit SMSLOGHOST 'c1a'
  edit SMSLOGPORT '9316'
  edit OD_PROJ 'od'
  edit SUITE_START '20070602'
  edit QUEUE 'emos-e'
  edit QUEUE_EPILOG 'emos-exp'
  edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
  edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
  edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
  edit LOGDIR '/vol/emos/output'
  edit SMSOUT '/vol/emos/output'
  edit METPY 'metpy_new'
  edit SLEEPTIME '1'
  inlimit /limits:swarm
  clock hybrid 
  family pop
    repeat date YMD 20070712 20201215 1
    limit plots 3
    family 00
      edit PLOT_AREAS 'Preview'
      edit EMOS_BASE '00'
      edit UPDATE_WEB_NOW 'yes'
      inlimit plots
      family deterministic
        edit SMSFILES '/home/ma/emos/def/preview/pop/web'
        edit PLOT_STEPS '60,72,84,96,108,120'
        edit STREAM 'DA'
        task delay
          time 07:20
        family data
          trigger delay eq complete
          edit SMSFILES '/home/ma/emos/def/preview/pop/data/fields'
          edit STEP_LIST '60/to/120/by/12'
          task mslp
          task rain
          task rh
          task t2m
          task t850
          task z500
          task w30m
          task wind
          task wr
          task wbpt
        endfamily
        family web
          trigger data eq complete
          task w_msl_uv200
          task w_mslrain
          task w_t2m30mw
          task w_uv_rh
          task w_z500t850
          task w_wr850_700
          task w_wbpt_msl
        endfamily
        task cleanup
          trigger web eq complete
          edit SMSFILES '/home/ma/emos/def/preview/pop'
      endfamily
      family eps
        edit SMSFILES '/home/ma/emos/def/preview/pop/web'
        edit STREAM 'EF'
        edit PLOT_STEPS '60/72/84/96/108/120'
        task delay
          time 08:50
        family data
          trigger delay eq complete
          edit SMSFILES '/home/ma/emos/def/preview/pop/data/fields'
          task efi
          task lsmoro
            edit STEP_LIST '0/to/48/by/12'
          task proba
            edit PLOT_STEPS_RANGE '48-72/60-84/72-96/84-108/96-120/108-132'
            edit PLOT_STEPS '60/72/84/96/108/120'
            edit RANGE 'legA'
        endfamily
        family web
          trigger data eq complete
          edit SMSFILES '/home/ma/emos/def/preview/pop/web'
          edit STEP_RANGE 'legA'
          family probabilities
            task pb_2t
            task pb_ff
            task pb_tp
              edit PLOT_STEPS '48-72/60-84/72-96/84-108/96-120/108-132'
            task w_proba_gusts
              edit PLOT_STEPS '48-72/60-84/72-96/84-108/96-120/108-132'
          endfamily
          family efi
            family efi_tpi
              task w_efi
                edit PARAMETER 'tpi'
            endfamily
            family efi_10wsi
              task w_efi
                edit PARAMETER '10wsi'
            endfamily
            family efi_2ti
              task w_efi
                edit PARAMETER '2ti'
            endfamily
            family efi_10fgi
              task w_efi
                edit PARAMETER '10fgi'
            endfamily
          endfamily
          family others
            edit PLOT_STEPS '60,72,84,96,108,120'
            task w_msl_mean_spread
            task w_z500_mean_spread
            task w_z200_mean_spread
            task w_t850_mean_spread
          endfamily
          family quartiles
            edit AREA_TO_PLOT '-5/-50/60/80'
            edit MAP_VERT_LONG '0'
            edit MAP_PROJECTION 'POLAR_STEREOGRAPHIC'
            edit STEP_LIST '54/TO/126/by/12'
            edit STEP_PLOT '60/TO/120/by/12'
            task w_tp_quartiles
            task w_wind_quartiles
            task w_wgust_quartiles
            task w_wind925_quartiles
            task w_cape_quartiles
          endfamily
        endfamily
        task cleanup
          trigger web eq complete
          edit SMSFILES '/home/ma/emos/def/preview/pop'
      endfamily
    endfamily
    family ukmo
      edit SMSFILES '/home/ma/emos/def/preview/pop/mogreps'
      edit PLOT_AREAS 'Preview'
      edit EMOS_BASE '00'
      edit STREAM 'EF'
      edit UPDATE_WEB_NOW 'yes'
      inlimit plots
      task delay
        time 08:55
      family mogreps
        trigger delay eq complete
        edit PLOT_STEPS '60,72,84,96,108,120'
        family web
          family en_mean_spread
            task m_msl_mean_spread
            task m_z500_mean_spread
            task m_z200_mean_spread
            task m_t850_mean_spread
          endfamily
          family quartiles
            edit AREA_TO_PLOT '-5/-50/60/80'
            edit MAP_VERT_LONG '0'
            edit MAP_PROJECTION 'POLAR_STEREOGRAPHIC'
            edit STEP_LIST '54/TO/126/by/12'
            task m_tp_quartiles
            task m_wind_quartiles
            task m_wind925_quartiles
          endfamily
        endfamily
        task cleanup
          trigger web eq complete
          edit SMSFILES '/home/ma/emos/def/preview/pop'
      endfamily
    endfamily
    task restart
      trigger ./00 eq complete
      time 15:00
  endfamily
endsuite
suite gems
  edit USER 'emos'
  edit SCHOST 'hpce'
  edit SCHOST_BKUP 'hpcf'
  edit WSHOST 'swarm'
  edit WSLOGDIR '/vol/emos_nc/output'
  edit SCLOGDIR '/emos_dir/evnj/log'
  edit USE_YMD 'true'
  edit YMD '20071217'
  edit DATEMASK '*.*.*'
  edit VERSION 'evnj'
  edit INIEXPVER '0001'
  edit TCINIEXPVER '0001'
  edit FIRST_DAY '0'
  edit SMSTRIES '1'
  edit DELTA_DAY '0'
  edit EMOS_BASE '00'
  edit EMOS_TIME_STEP_H '0000'
  edit FSFAMILY 'tc'
  edit USER_PRIORITY '70'
  edit REINITIALIZE 'false'
  edit SMSHOME '/vol/emos_nc/output'
  edit SMSFILES '/home/ma/emos/def/gems/smsfiles'
  edit SMSINCLUDE '/home/ma/emos/def/gems/include'
  edit ACCOUNT 'oosu'
  edit STREAM 'EF'
  edit EMOS_STREAM 'EF'
  edit ENSEMBLES '50'
  edit ENS_SIZE '50'
  edit MARS_DB 'marsod'
  edit EPSTYPE 'fc'
  edit MEMBER '0'
  edit IS_REAL_TIME 'is_real_time'
  edit RUN_3HOURS_PP '1'
  edit SMSLOGHOST 'c1a'
  edit SMSLOGPORT '9316'
  edit OD_PROJ 'od'
  edit SUITE_START '20061126'
  edit QUEUE 'emos-e'
  edit QUEUE_EPILOG 'emos-exp'
  edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
  edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
  edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
  edit LOGDIR '/vol/emos_nc/output'
  edit SMSOUT '/vol/emos_nc/output'
  edit SLEEPTIME '1'
  inlimit /limits:swarm
  clock hybrid 
  family limits
    limit linux_cluster 8
    limit raq_ensemble 3
    limit calchas 4
  endfamily
  family grg
    inlimit /gems/limits:linux_cluster
    family plot
      edit METPY 'metpy_dev'
      edit UPDATE_WEB_NOW 'yes'
      edit DO_PUSH_TO_WEB 'yes'
      family realtime
        repeat date YMD 20081223 20120924 1
        edit VERSION 'f7kn'
        edit METPY 'metpy_new'
        edit PARAMETERS 'co,go3'
        edit LEVELS '850,500,300,50'
        family deps
          task dummy
            time 01:00
        endfamily
        task daily_fields
          trigger deps eq complete
        task daily_totalcolumns
          trigger daily_fields eq complete
          edit PARAMETERS 'tcco,gtco3'
        task daily_zonalmean
          trigger daily_totalcolumns eq complete
        task daily_crosssection
          trigger daily_zonalmean eq complete
      endfamily
      family realtime_da
        repeat date YMD 20081223 20120924 1
        edit METPY 'metpy_new'
        edit VERSION 'f8na'
        edit PARAMETERS 'co,go3'
        edit LEVELS '850,500,300,50'
        family deps
          task dummy
            time 12:00
        endfamily
        task daily_fields_da
          trigger deps eq complete
        task daily_totalcolumns_da
          trigger daily_fields_da eq complete
          edit PARAMETERS 'tcco,gtco3'
        task daily_zonalmean_da
          trigger daily_totalcolumns_da eq complete
        task daily_crosssection_da
          trigger daily_zonalmean_da eq complete
      endfamily
      family uv
        repeat date YMD 20081223 20090924 1
        edit VERSION 'ewsf'
        edit PARAMETERS '82'
        family deps
          task dummy
            time 16:00
        endfamily
        task uv_dose
          trigger deps eq complete
          edit METPY 'metpy_new'
      endfamily
      family reanalysis
        repeat integer month 200301 200312
        edit VERSION 'eyih'
        edit LEVELS '850,500,300,50'
        edit PARAMETERS 'co,go3,hcho,nox'
        task monthly_fields
        task monthly_totalcolumns
          edit PARAMETERS 'tcco,gtco3,tchcho'
        task monthly_zonalmean
          trigger monthly_totalcolumns eq complete
        task monthly_crosssection
          trigger monthly_zonalmean eq complete
      endfamily
      task logfilestore
        trigger realtime eq complete
        edit QUEUE 'emos-e'
        edit QUEUE_EPILOG 'emos-exp'
        edit WSHOST 'localhost'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        edit KEEPLOGS 'no'
        edit LOGSTOECFS 'no'
        edit MOVE_TO_ECFS 'yes'
      task logfiles
        trigger logfilestore eq complete
        edit QUEUE 'emos-e'
        edit QUEUE_EPILOG 'emos-exp'
        edit WSHOST 'localhost'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        edit MOVE_TO_ECFS 'no'
      family polarcat
        repeat date YMD 20081223 20120924 1
        edit VERSION 'f7kn'
        edit OWNER 'nal'
        family deps
          task dummy
            time 11:00
        endfamily
        task polarcat_retrieve_and_process
          trigger deps eq complete
        task polarcat_make_latlon_plots
          trigger polarcat_retrieve_and_process eq complete
        task polarcat_make_polar_plots
          trigger polarcat_make_latlon_plots eq complete
        task polarcat_make_xs_plots
          trigger polarcat_make_polar_plots eq complete
        task polarcat_cleanup
          trigger polarcat_make_xs_plots eq complete
      endfamily
    endfamily
  endfamily
  family ghg
    inlimit /gems/limits:linux_cluster
    family plot
      edit METPY 'metpy_new'
      edit UPDATE_WEB_NOW 'yes'
      edit DO_PUSH_TO_WEB 'yes'
      family reanalysis
        edit PARAMETERS 'tcco2'
        task tcco2_mean
      endfamily
      family ch4_reanalysis
        repeat enumerated month "200301" "200302" "200303" "200304" "200305" "200306" "200307" "200308" "200309" "200310" "200311" "200312" "200401" "200402" "200403" "200404" "200405" "200406" "200407" "200408" "200409" "200410" "200411" "200412" "200501" "200502" "200503" "200504" "200505" "200506" "200507" "200508" "200509" "200510" "200511" "200512" "200601" "200602" "200603" "200604" "200605" "200606" "200607" "200608" "200609" "200610" "200611" "200612" "200701" "200702" "200703" "200704" "200705" "200706" "200707" "200708" "200709" "200710" "200711" "200712"
        edit VERSION 'f46b'
        edit PARAMETERS 'tcch4'
        task ch4_monthly_totalcolumns
      endfamily
      family iasi_plots
        repeat enumerated month "200801" "200802" "200803" "200804" "200805" "200806" "200807" "200808" "200809" "200810" "200811" "200812"
        task iasi
      endfamily
      family ch4_flux_inversion
        repeat enumerated month "200401" "200402" "200403" "200404" "200405" "200406" "200407" "200408" "200409" "200410" "200411" "200412"
        task plot_ch4_flux
      endfamily
    endfamily
  endfamily
  family aer
    edit UPDATE_WEB_NOW 'yes'
    edit DO_PUSH_TO_WEB 'yes'
    edit METPY 'metpy_dev'
    inlimit /gems/limits:linux_cluster
    family realtime
      repeat date YMD 20080507 20081224 1
      edit VERSION 'f0dm'
      edit AERTYPES 'total,natural,anthropogenic'
      edit METPY 'metpy_new'
      family deps
        task dummy
          time 08:30
      endfamily
      task optical_depth
        trigger deps eq complete
    endfamily
    family realtime_da
      repeat date YMD 20080506 20091224 1
      edit VERSION 'f8na'
      edit AERTYPES 'total,natural,anthropogenic'
      edit METPY 'metpy_new'
      family deps
        task dummy
          time 12:00
      endfamily
      task optical_depth_da
        trigger deps eq complete
    endfamily
    family verif
      repeat enumerated month "200301" "200302" "200303" "200304" "200305" "200306" "200307" "200308" "200309" "200310" "200311" "200312" "200401" "200402" "200403" "200404" "200405" "200406" "200407" "200408" "200409" "200410" "200411" "200412"
      task aer_station_plots_exlz_ezub
      task aer_stat_plots_exlz_ezub
        trigger aer_station_plots_exlz_ezub eq complete
    endfamily
    family fire_products
      repeat date YMD 20080506 20091224 1
      edit METPY 'metpy_new'
      family deps
        task dummy
          time 07:00
      endfamily
      task fire_radiative_power
        trigger deps eq complete
    endfamily
    family verif_f1kd
      repeat enumerated month "200807" "200808" "200809" "200810" "200811" "200812" "200901" "200902" "200903" "200904" "200905" "200906" "200907" "200908" "200909" "200910" "200911" "200912"
      task aer_station_plots_f1kd
      task aer_stat_plots_f1kd
        trigger aer_station_plots_f1kd eq complete
    endfamily
  endfamily
  family raq
    edit SMSTRIES '5'
    edit UPDATE_WEB_NOW 'yes'
    edit DO_PUSH_TO_WEB 'yes'
    edit METPY 'metpy_new'
    edit VERSION 'gems'
    inlimit /gems/limits:linux_cluster
    family obs
      repeat date YMD 20081223 20111001 1
      edit METPY 'metpy_new'
      edit FIRST_HOUR '0'
      edit FINAL_HOUR '23'
      edit STEP '1'
      edit PARAMETERS 'o3,no2,so2,co,pm10'
      edit COUNTRIES 'pl,fi,se,nl,fr'
      edit ARCHIVED_COUNTRIES 'uk,es,de'
      family deps
        task dummy
          time 11:35
      endfamily
      task plot_obs
        trigger deps eq complete
    endfamily
    family ensemble
      repeat date YMD 20081223 20111001 1
      edit MODELS 'CHI,CNR,DMI,EMP,FMI,MFM,MOZ,NKU,RIU,SMH,UKN'
      family deps
        task dummy
          time 10:00
      endfamily
      task regrid_models
        trigger deps eq complete
        edit MODELS 'CHI,CNR,DMI,EMP,FMI,MFM,MOZ,NKU,RIU,SMH,UKN'
      family plot_ensemble
        inlimit /gems/limits:raq_ensemble
        task plot_ensemble_o3
          trigger ../regrid_models eq complete
          edit PARAMETER 'o3'
          edit MODELS 'CHI,CNR,DMI,EMP,FMI,MFM,MOZ,NKU,RIU,SMH'
        task plot_ensemble_no2
          trigger ../regrid_models eq complete
          edit MODELS 'CHI,CNR,DMI,EMP,FMI,MFM,MOZ,NKU,RIU,SMH,UKN'
          edit PARAMETER 'no2'
        task plot_ensemble_so2
          trigger ../regrid_models eq complete
          edit PARAMETER 'so2'
          edit MODELS 'CHI,CNR,DMI,EMP,FMI,MFM,MOZ,NKU,RIU,SMH,UKN'
        task plot_ensemble_co
          trigger ../regrid_models eq complete
          edit PARAMETER 'co'
          edit MODELS 'CHI,CNR,DMI,EMP,FMI,MFM,MOZ,NKU,RIU,SMH,UKN'
        task plot_ensemble_pm10
          trigger ../regrid_models eq complete
          edit PARAMETER 'pm10'
          edit MODELS 'CHI,EMP,FMI,NKU,RIU,SMH,UKN'
        task plot_epsgram
          trigger ../regrid_models eq complete
        task plot_ensemble_aqi
          trigger ./plot_ensemble_o3 eq complete and ./plot_ensemble_no2 eq complete and ./plot_ensemble_so2 eq complete and ./plot_ensemble_pm10 eq complete
          edit MODELS 'CHI,CNR,DMI,EMP,FMI,MFM,NKU,RIU,SMH'
      endfamily
      task compute_hourly_ensemble
        trigger plot_ensemble eq complete
        edit PARAMETERS 'o3,no2,so2,co,pm10'
      task cleanup_models
        trigger compute_hourly_ensemble eq complete
        time 15:30
    endfamily
    family verif
      edit METPY 'metpy_new'
      family scores
        repeat date YMD 20080731 20100716 1
        family deps
          task dummy
            time 11:35
        endfamily
        family compute_scores
          edit PARAMETERS 'o3,no2,so2,pm10'
          task verify_RAQ
            trigger ../deps eq complete
            label status ""
        endfamily
        family plot_scores
          edit PARAMETERS 'o3,no2,so2,co,pm10'
          edit COUNTRIES 'pl,fi,se,nl,fr'
          edit ARCHIVED_COUNTRIES 'uk,es,de'
          task plot_RAQ_mean_scores
            trigger ../compute_scores eq complete
          task plot_RAQ_ts_scores
            trigger ../compute_scores eq complete
          task plot_RAQ_taylor_plots
            trigger ../compute_scores eq complete
        endfamily
      endfamily
      family obs_model
        family values
          repeat date YMD 20080731 20100716 1
          edit COUNTRIES 'pl,fi,se,nl,fr'
          edit ARCHIVED_COUNTRIES 'uk,es,de'
          task plot_obs_model_o3
            trigger ( /gems/raq/verif/scores:YMD gt /gems/raq/verif/obs_model/values:YMD) or ../../scores eq complete
            edit PARAMETER 'o3'
            edit MODELS 'CHI,CNR,DMI,EMP,FMI,MFM,MOZ,NKU,RIU,SMH'
          task plot_obs_model_no2
            trigger plot_obs_model_o3 eq complete
            edit MODELS 'CHI,CNR,DMI,EMP,FMI,MFM,MOZ,NKU,RIU,SMH,UKN'
            edit PARAMETER 'no2'
          task plot_obs_model_so2
            trigger plot_obs_model_no2 eq complete
            edit PARAMETER 'so2'
            edit MODELS 'CHI,CNR,DMI,EMP,FMI,MFM,NKU,RIU,SMH,UKN'
          task plot_obs_model_co
            trigger plot_obs_model_so2 eq complete
            edit PARAMETER 'co'
            edit MODELS 'CHI,CNR,DMI,EMP,FMI,MFM,MOZ,NKU,RIU,SMH'
          task plot_obs_model_pm10
            trigger plot_obs_model_co eq complete
            edit PARAMETER 'pm10'
            edit MODELS 'CHI,EMP,FMI,NKU,RIU,SMH,UKN'
        endfamily
      endfamily
      family plot_OMI
        repeat date YMD 20080731 20100716 1
        family deps
          task dummy
            time 11:00
        endfamily
        task plot_OMI_tc
          trigger deps eq complete
          edit MODELS 'CHI,CNR,DMI,EMP,FMI,NKU,RIU,SMH'
      endfamily
      family rescore_fam
        family rescore_1
          repeat date YMD 20080601 20081031 1
          task rescore_RAQ_1
            label status ""
        endfamily
        family rescore_2
          repeat date YMD 20080601 20081031 1
          task rescore_RAQ_2
            label status ""
        endfamily
        family rescore_3
          repeat date YMD 20080601 20081031 1
          task rescore_RAQ_3
            label status ""
        endfamily
        family rescore_4
          repeat date YMD 20080601 20081031 1
          task rescore_RAQ_4
            label status ""
        endfamily
        family rescore_5
          repeat date YMD 20080601 20081031 1
          task rescore_RAQ_5
            label status ""
        endfamily
      endfamily
    endfamily
    family models
      edit FIRST_STEP '0'
      edit FINAL_STEP '72'
      edit STEP '1'
      edit DEADLINE '1500'
      edit LEVELS '500,1000,3000'
      edit PARAMETERS 'o3,no,no2'
      family CHI
        repeat date YMD 20081223 20111001 1
        edit MODEL 'CHI'
        family plot
          complete plot/finish eq complete
          task plot_CHI
            label status "nothing yet"
            event done
            cron 00:00 15:00 00:30
          task finish
            complete plot_CHI:done or plot_CHI eq complete
            trigger 1 eq 0
        endfamily
        task dummy
          trigger plot eq complete
          time 15:15
      endfamily
      family CNR
        repeat date YMD 20081223 20111001 1
        edit MODEL 'CNR'
        family plot
          complete plot/finish eq complete
          task plot_CNR
            label status "nothing yet"
            event done
            cron 00:00 15:00 00:30
          task finish
            complete plot_CNR:done or plot_CNR eq complete
            trigger 1 eq 0
        endfamily
        task dummy
          trigger plot eq complete
          time 15:15
      endfamily
      family DMI
        repeat date YMD 20081223 20111001 1
        edit MODEL 'DMI'
        family plot
          complete plot/finish eq complete
          task plot_DMI
            label status "nothing yet"
            event done
            cron 00:00 15:00 00:30
          task finish
            complete plot_DMI:done or plot_DMI eq complete
            trigger 1 eq 0
        endfamily
        task dummy
          trigger plot eq complete
          time 15:15
      endfamily
      family EMP
        repeat date YMD 20081223 20111001 1
        edit MODEL 'EMP'
        family plot
          complete plot/finish eq complete
          task plot_EMP
            label status "nothing yet"
            event done
            cron 00:00 15:00 00:30
          task finish
            complete plot_EMP:done or plot_EMP eq complete
            trigger 1 eq 0
        endfamily
        task dummy
          trigger plot eq complete
          time 15:15
      endfamily
      family FMI
        repeat date YMD 20081223 20111001 1
        edit MODEL 'FMI'
        edit PARAMETERS 'o3,no,no2'
        family plot
          complete plot/finish eq complete
          task plot_FMI
            label status "nothing yet"
            event done
            cron 00:00 15:00 00:30
          task finish
            complete plot_FMI:done or plot_FMI eq complete
            trigger 1 eq 0
        endfamily
        task dummy
          trigger plot eq complete
          time 15:15
      endfamily
      family KNM
        repeat date YMD 20081223 20111001 1
        edit MODEL 'KNM'
        family plot
          complete plot/finish eq complete
          task plot_KNM
            label status "nothing yet"
            event done
            cron 00:00 15:00 00:30
          task finish
            complete plot_KNM:done or plot_KNM eq complete
            trigger 1 eq 0
        endfamily
        task dummy
          trigger plot eq complete
          time 15:15
      endfamily
      family MFM
        repeat date YMD 20081223 20111001 1
        edit MODEL 'MFM'
        family plot
          complete plot/finish eq complete
          task plot_MFM
            label status "nothing yet"
            event done
            cron 00:00 15:00 00:30
          task finish
            complete plot_MFM:done or plot_MFM eq complete
            trigger 1 eq 0
        endfamily
        task dummy
          trigger plot eq complete
          time 15:15
      endfamily
      family MOZ
        repeat date YMD 20081223 20111001 1
        edit MODEL 'MOZ'
        edit FIRST_STEP '1'
        family plot
          complete plot/finish eq complete
          task plot_MOZ
            label status "nothing yet"
            event done
            cron 10:00 16:00 00:30
          task finish
            complete plot_MOZ:done or plot_MOZ eq complete
            trigger 1 eq 0
        endfamily
        task dummy
          trigger plot eq complete
          time 16:15
      endfamily
      family NKU
        repeat date YMD 20081223 20111001 1
        edit MODEL 'NKU'
        family plot
          complete plot/finish eq complete
          task plot_NKU
            label status "nothing yet"
            event done
            cron 00:00 15:00 00:30
          task finish
            complete plot_NKU:done or plot_NKU eq complete
            trigger 1 eq 0
        endfamily
        task dummy
          trigger plot eq complete
          time 15:15
      endfamily
      family RIU
        repeat date YMD 20081223 20111001 1
        edit MODEL 'RIU'
        family plot
          complete plot/finish eq complete
          task plot_RIU
            label status "nothing yet"
            event done
            cron 00:00 15:00 00:30
          task finish
            complete plot_RIU:done or plot_RIU eq complete
            trigger 1 eq 0
        endfamily
        task dummy
          trigger plot eq complete
          time 15:15
      endfamily
      family SMH
        repeat date YMD 20081223 20111001 1
        edit MODEL 'SMH'
        family plot
          complete plot/finish eq complete
          task plot_SMH
            label status "nothing yet"
            event done
            cron 00:00 15:00 00:30
          task finish
            complete plot_SMH:done or plot_SMH eq complete
            trigger 1 eq 0
        endfamily
        task dummy
          trigger plot eq complete
          time 15:15
      endfamily
      family UKN
        repeat date YMD 20081223 20111001 1
        edit MODEL 'UKN'
        edit FINAL_STEP '71'
        edit PARAMETERS 'no,no2'
        family plot
          complete plot/finish eq complete
          task plot_UKN
            label status "nothing yet"
            event done
            cron 00:00 15:00 00:30
          task finish
            complete plot_UKN:done or plot_UKN eq complete
            trigger 1 eq 0
        endfamily
        task dummy
          trigger plot eq complete
          time 15:15
      endfamily
    endfamily
    family etc
      task cleanup_ftp
        cron 23:55
    endfamily
  endfamily
  family combined
    inlimit /gems/limits:linux_cluster
    family reanalysis
      repeat enumerated month "200301" "200302" "200303" "200304" "200305" "200306" "200307" "200308" "200309" "200310" "200311" "200312" "200401" "200402" "200403" "200404" "200405" "200406" "200407" "200408" "200409" "200410" "200411" "200412" "200501" "200502" "200503" "200504" "200505" "200506" "200507" "200508" "200509" "200510" "200511" "200512" "200601" "200602" "200603" "200604" "200605" "200606" "200607" "200608" "200609" "200610" "200611" "200612" "200701" "200702" "200703" "200704" "200705" "200706" "200707" "200708" "200709" "200710" "200711" "200712" "200801"
      edit VERSION 'f026'
      edit LEVELS '850,500,300,50'
      edit PARAMETERS 'co,go3,hcho,ch4,nox'
      task gems_monthly_fields
      task gems_monthly_zonalmean
      task gems_monthly_totalcolumns
        edit PARAMETERS 'tcco2,tcco,gtco3,tchcho,tcch4'
    endfamily
    family reanalysis_update
      repeat enumerated month "200301" "200302" "200303" "200304" "200305" "200306" "200307" "200308" "200309" "200310" "200311" "200312" "200401" "200402" "200403" "200404" "200405" "200406" "200407" "200408" "200409" "200410" "200411" "200412" "200501" "200502" "200503" "200504" "200505" "200506" "200507" "200508" "200509" "200510" "200511" "200512" "200601" "200602" "200603" "200604" "200605" "200606" "200607" "200608" "200609" "200610" "200611" "200612" "200701" "200702" "200703" "200704" "200705" "200706" "200707" "200708" "200709" "200710" "200711" "200712" "200801"
      edit VERSION 'ey2o'
      task gems_monthly_totalcolumns
        edit PARAMETERS 'tcco2'
    endfamily
  endfamily
endsuite
suite tparc
  edit USER 'emos'
  edit SCHOST 'hpce'
  edit SCHOST_BKUP 'hpcf'
  edit WSHOST 'swarm'
  edit WSLOGDIR '/vol/emos_nc/output'
  edit SCLOGDIR '/emos_dir/0001/log'
  edit USE_YMD 'true'
  edit YMD '20100105'
  edit DATEMASK '*.*.*'
  edit VERSION '0001'
  edit INIEXPVER '0015'
  edit TCINIEXPVER '0001'
  edit FIRST_DAY '0'
  edit SMSTRIES '2'
  edit DELTA_DAY '0'
  edit EMOS_BASE '00'
  edit EMOS_TIME_STEP_H '0000'
  edit FSFAMILY 'tc'
  edit USER_PRIORITY '70'
  edit REINITIALIZE 'false'
  edit SMSHOME '/vol/emos_nc/output'
  edit SMSFILES '/home/ma/emos/def/tparc/smsfiles'
  edit SMSINCLUDE '/home/ma/emos/def/tparc/include'
  edit ACCOUNT 'oosu'
  edit STREAM 'DA'
  edit EMOS_STREAM 'EF'
  edit ENSEMBLES '50'
  edit ENS_SIZE '50'
  edit MARS_DB 'marsod'
  edit EPSTYPE 'fc'
  edit MEMBER '0'
  edit IS_REAL_TIME 'is_real_time'
  edit RUN_3HOURS_PP '1'
  edit SMSLOGHOST 'hpce'
  edit SMSLOGPORT '9316'
  edit OD_PROJ 'od'
  edit SUITE_START '20080611'
  edit QUEUE 'emos-e'
  edit QUEUE_EPILOG 'emos-exp'
  edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
  edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
  edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
  edit LOGDIR '/vol/emos_nc/output'
  edit SMSOUT '/vol/emos_nc/output'
  edit METPY 'metpy_new'
  edit SLEEPTIME '1'
  inlimit /limits:swarm
  clock hybrid 
  family pop
    repeat date YMD 20100105 20120410 1
    limit plots 5
    family 00
      edit DELTA_DAY '1'
      edit PLOT_AREAS 'Tparc_winter'
      edit EMOS_BASE '00'
      edit UPDATE_WEB_NOW 'yes'
      inlimit plots
      family deterministic
        edit SMSFILES '/home/ma/emos/def/tparc/pop/web'
        edit PLOT_STEPS '12,24,36,48,60,72,84,96,108,120,132,144'
        edit STREAM 'DA'
        task delay
          time 07:30
        family data
          trigger delay eq complete
          edit SMSFILES '/home/ma/emos/def/tparc/pop/data/fields'
          edit STEP_LIST '12/to/144/by/12'
          task mslp
          task rain
          task rh
          task t2m
          task t850
          task z500
          task w30m
          task wind
          task wr
          task wbpt
        endfamily
        family web
          trigger data eq complete
          task w_msl_uv200
          task w_mslrain
          task w_t2m30mw
          task w_uv_rh
          task w_z500t850
          task w_wr850_700
          task w_wbpt_msl
        endfamily
        task cleanup
          trigger web eq complete
          edit SMSFILES '/home/ma/emos/def/tparc/pop'
      endfamily
      family interactive_chart
        edit SMSFILES '/home/ma/emos/def/tparc/pop/web'
        family interactive_ghost
          time 00:01
          task w_msl_mean_spread_ghost
            edit PLOT_STEPS '36,48,60,72,84,96,108,120,132,144'
            edit DELTA_VALUE '-1'
        endfamily
        family interactive_real
          time 08:55
          task w_msl_mean_spread_ghost
            edit PLOT_STEPS '36,48,60,72,84,96,108,120,132,144'
            edit DELTA_VALUE '0'
        endfamily
      endfamily
      family eps
        edit SMSFILES '/home/ma/emos/def/tparc/pop/web'
        edit STREAM 'EF'
        edit PLOT_STEPS '12/24/36/48/60/72/84/96/108/120/132/144'
        task delay
          time 08:55
        family data
          trigger delay eq complete
          edit SMSFILES '/home/ma/emos/def/tparc/pop/data/fields'
          task efi
          task lsmoro
            edit STEP_LIST '0/to/48/by/12'
          task proba
            edit PLOT_STEPS_RANGE '0-24/12-36/24-48/36-60/48-72/60-84/72-96/84-108/96-120/108-132/120-144/132-156'
            edit PLOT_STEPS '12/24/36/48/60/72/84/96/108/120/132/144'
            edit RANGE 'legA'
        endfamily
        family web
          trigger data eq complete
          edit SMSFILES '/home/ma/emos/def/tparc/pop/web'
          edit STEP_RANGE 'legA'
          family probabilities
            task pb_2t
            task pb_ff
            task pb_tp
              edit PLOT_STEPS '0-24/12-36/24-48/36-60/48-72/60-84/72-96/84-108/96-120/108-132/120-144/132-156'
            task w_proba_gusts
              edit PLOT_STEPS '0-24/12-36/24-48/36-60/48-72/60-84/72-96/84-108/96-120/108-132/120-144/132-156'
          endfamily
          family efi
            family efi_tpi
              task w_efi
                edit PARAMETER 'tpi'
            endfamily
            family efi_10wsi
              task w_efi
                edit PARAMETER '10wsi'
            endfamily
            family efi_2ti
              task w_efi
                edit PARAMETER '2ti'
            endfamily
            family efi_10fgi
              task w_efi
                edit PARAMETER '10fgi'
            endfamily
          endfamily
          family others
            edit PLOT_STEPS '12,24,36,48,60,72,84,96,108,120,132,144'
            task w_msl_mean_spread
            task w_z500_mean_spread
            task w_z200_mean_spread
            task w_t850_mean_spread
          endfamily
          family quartiles
            edit AREA_TO_PLOT '-10/100/90/300'
            edit AREA_TO_PLOT_ZOOM '20/140/90/280'
            edit MAP_VERT_LONG '160'
            edit MAP_VERT_LONG_ZOOM '160'
            edit MAP_PROJECTION 'CYLINDRICAL'
            edit STEP_LIST '6/TO/150/by/12'
            edit STEP_PLOT '12/TO/144/by/12'
            task w_tp_quartiles
            task w_wind_quartiles
            task w_wgust_quartiles
            task w_wind925_quartiles
            task w_cape_quartiles
          endfamily
        endfamily
        task cleanup
          trigger web eq complete
          edit SMSFILES '/home/ma/emos/def/tparc/pop'
      endfamily
    endfamily
    family 12
      edit PLOT_AREAS 'Tparc_winter'
      edit EMOS_BASE '12'
      edit UPDATE_WEB_NOW 'yes'
      inlimit plots
      family deterministic
        edit SMSFILES '/home/ma/emos/def/tparc/pop/web'
        edit PLOT_STEPS '12,24,36,48,60,72,84,96,108,120,132,144'
        edit STREAM 'DA'
        task delay
          time 19:20
        family data
          trigger delay eq complete
          edit SMSFILES '/home/ma/emos/def/tparc/pop/data/fields'
          edit STEP_LIST '12/to/144/by/12'
          task mslp
          task rain
          task rh
          task t2m
          task t850
          task z500
          task w30m
          task wind
          task wr
          task wbpt
        endfamily
        family web
          trigger data eq complete
          task w_msl_uv200
          task w_mslrain
          task w_t2m30mw
          task w_uv_rh
          task w_z500t850
          task w_wr850_700
          task w_wbpt_msl
        endfamily
        task cleanup
          trigger web eq complete
          edit SMSFILES '/home/ma/emos/def/tparc/pop'
      endfamily
      family eps
        edit SMSFILES '/home/ma/emos/def/tparc/pop/web'
        edit STREAM 'EF'
        edit PLOT_STEPS '12/24/36/48/60/72/84/96/108/120/132/144'
        task delay
          time 20:50
        family data
          trigger delay eq complete
          edit SMSFILES '/home/ma/emos/def/tparc/pop/data/fields'
          task efi
          task lsmoro
            edit STEP_LIST '0/to/48/by/12'
          task proba
            edit PLOT_STEPS_RANGE '0-24/12-36/24-48/36-60/48-72/60-84/72-96/84-108/96-120/108-132/120-144/132-156'
            edit PLOT_STEPS '12/24/36/48/60/72/84/96/108/120/132/144'
            edit RANGE 'legA'
        endfamily
        family web
          trigger data eq complete
          edit SMSFILES '/home/ma/emos/def/tparc/pop/web'
          edit STEP_RANGE 'legA'
          family probabilities
            task pb_2t
            task pb_ff
            task pb_tp
              edit PLOT_STEPS '0-24/12-36/24-48/36-60/48-72/60-84/72-96/84-108/96-120/108-132/120-144/132-156'
            task w_proba_gusts
              edit PLOT_STEPS '0-24/12-36/24-48/36-60/48-72/60-84/72-96/84-108/96-120/108-132/120-144/132-156'
          endfamily
          family efi
            family efi_tpi
              task w_efi
                edit PARAMETER 'tpi'
            endfamily
            family efi_10wsi
              task w_efi
                edit PARAMETER '10wsi'
            endfamily
            family efi_2ti
              task w_efi
                edit PARAMETER '2ti'
            endfamily
            family efi_10fgi
              task w_efi
                edit PARAMETER '10fgi'
            endfamily
          endfamily
          family others
            edit PLOT_STEPS '12,24,36,48,60,72,84,96,108,120,132,144'
            task w_msl_mean_spread
            task w_z500_mean_spread
            task w_z200_mean_spread
            task w_t850_mean_spread
          endfamily
          family quartiles
            edit AREA_TO_PLOT '-10/100/90/300'
            edit AREA_TO_PLOT_ZOOM '20/140/90/280'
            edit MAP_VERT_LONG '160'
            edit MAP_VERT_LONG_ZOOM '160'
            edit MAP_PROJECTION 'CYLINDRICAL'
            edit STEP_LIST '6/TO/150/by/12'
            edit STEP_PLOT '12/TO/144/by/12'
            task w_tp_quartiles
            task w_wind_quartiles
            task w_wgust_quartiles
            task w_wind925_quartiles
            task w_cape_quartiles
          endfamily
        endfamily
        task cleanup
          trigger web eq complete
          edit SMSFILES '/home/ma/emos/def/tparc/pop'
      endfamily
    endfamily
    family ukmo
      edit DELTA_DAY '1'
      edit SMSFILES '/home/ma/emos/def/tparc/pop/mogreps'
      edit PLOT_AREAS 'Tparc_winter'
      edit EMOS_BASE '00'
      edit STREAM 'EF'
      edit UPDATE_WEB_NOW 'yes'
      inlimit plots
      task delay
        time 09:00
      family mogreps
        trigger delay eq complete
        edit PLOT_STEPS '12,24,36,48,60,72,84,96,108,120,132,144'
        family web
          family en_mean_spread
            task m_msl_mean_spread
            task m_z500_mean_spread
            task m_z200_mean_spread
            task m_t850_mean_spread
          endfamily
          family quartiles
            edit AREA_TO_PLOT '-10/100/90/300'
            edit AREA_TO_PLOT_ZOOM '20/140/90/280'
            edit MAP_VERT_LONG '160'
            edit MAP_VERT_LONG_ZOOM '160'
            edit MAP_PROJECTION 'CYLINDRICAL'
            edit STEP_LIST '6/TO/150/by/12'
            edit STEP_PLOT '12/TO/144/by/12'
            task m_tp_quartiles
            task m_wind_quartiles
            task m_wind925_quartiles
          endfamily
        endfamily
        task cleanup
          trigger web eq complete
          edit SMSFILES '/home/ma/emos/def/tparc/pop'
      endfamily
    endfamily
    task restart
      trigger ./00 eq complete and ukmo eq complete and ./12 eq complete
      time 11:00
  endfamily
endsuite
suite tparc_dts
  edit USER 'emos'
  edit SCHOST 'hpcf'
  edit WSHOST 'linux_cluster'
  edit WSLOGDIR '/vol/emos_nc/output'
  edit SCLOGDIR '/emos_esuite/emos_data/0015/log'
  edit EMOS_TYPE '1'
  edit DTS '1'
  edit USE_YMD 'true'
  edit YMD '20090327'
  edit DATEMASK '*.*.*'
  edit VERSION '0015'
  edit INIEXPVER '0001'
  edit FIRST_DAY '0'
  edit SMSTRIES '2'
  edit DELTA_DAY '0'
  edit EMOS_TIME_STEP_H '0000'
  edit FSFAMILY 'mc'
  edit USER_PRIORITY '70'
  edit PRIORITY '70'
  edit ININPES '16'
  edit NPROCA '16'
  edit NPROCB '1'
  edit SVMODE 'dts'
  edit SVPROJVERT 'lower_troposphere'
  edit REINITIALIZE 'false'
  edit SMSHOME '/vol/emos_nc/output'
  edit SMSFILES '/home/ma/emos/def/tparc_33r1/dts'
  edit SMSINCLUDE '/home/ma/emos/def/tparc_33r1/include'
  edit ACCOUNT 'oodts'
  edit STREAM 'EF'
  edit EMOS_STREAM 'EF'
  edit MARS_DB 'marser'
  edit EPSTYPE 'sv'
  edit MEMBER '0'
  edit INILEVELS '91'
  edit IS_REAL_TIME 'is_real_time'
  edit RUN_3HOURS_PP '1'
  edit FCGROUP 'fc0'
  edit SMSLOGHOST 'hpcf'
  edit SMSLOGPORT '9316'
  edit SVTHREADS '2'
  edit QUEUE 'ns'
  edit QUEUE_EPILOG 'ns'
  edit SMSCMD 'smssubmit %USER% %SCHOST% %SMSJOB% %SMSJOBOUT%'
  edit SMSKILL 'smskill %USER% %SCHOST% %SMSRID% %SMSJOB%'
  edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %SCHOST% %SMSRID% %SMSJOB%'
  edit LOGDIR '/emos_esuite/emos_data/0015/log'
  edit SMSOUT '/emos_esuite/emos_data/0015/log'
  edit SLEEPTIME '1'
  inlimit /limits:hpcf
  clock hybrid 
  family limits
    limit hpce 5
    limit hpcf 5
    limit mars 3
    limit webplots 3
    limit linux_cluster 3
  endfamily
  family main
    repeat date YMD 20090327 20190101 1
    family make
      edit EMOS_BASE '00'
      task setup
      task links
        trigger setup eq complete
      task datalinks
        trigger links eq complete and setup eq complete
      family bins
        trigger datalinks eq complete
        task ifs
        task prepdata
        task mc_tools
      endfamily
    endfamily
    family 00
      edit EMOS_BASE '00'
      task concordiasi_control
        edit QUEUE 'emos'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        inlimit /tparc_dts/limits:linux_cluster
        time 10:00
      family case1
        trigger /tparc_dts/main/make eq complete and ./concordiasi_control eq complete
        edit REGIONNR '1'
        edit INISTEP '42'
        edit FCLENGTH '30'
        edit FSFAMILY 'mc/case1'
        edit ALAT1 '20'
        edit ALON1 '0'
        edit ALAT3 '-20'
        edit ALON3 '359'
        edit DELTA_HOUR '42'
        inlimit /tparc_dts/limits:hpcf
        task getini
        task inidata_sv
          trigger getini eq complete
          edit NPES '16'
          edit THREADS '4'
          edit MEM '1300'
          edit MRF '10'
          edit CPUTIME '1000'
          edit STACK '450'
        task sv
          trigger inidata_sv eq complete
          edit MEM '890'
          edit NPES '42'
          edit THREADS '2'
          edit CPUTIME '20000'
        task sv2gg
          trigger sv eq complete
          edit EMOS_TYPE '0'
        task sv_te
          trigger sv2gg eq complete
        task sv2fdb
          trigger sv_te eq complete
      endfamily
      family case2
        trigger /tparc_dts/main/make eq complete and ./concordiasi_control eq complete
        edit REGIONNR '2'
        edit INISTEP '48'
        edit FCLENGTH '24'
        edit FSFAMILY 'mc/case2'
        inlimit /tparc_dts/limits:hpcf
        task getini
        task inidata_sv
          trigger getini eq complete
          edit NPES '16'
          edit THREADS '4'
          edit MEM '1300'
          edit MRF '10'
          edit CPUTIME '1000'
          edit STACK '450'
        task sv
          trigger inidata_sv eq complete
          edit MEM '890'
          edit NPES '42'
          edit THREADS '2'
          edit CPUTIME '20000'
        task sv2gg
          trigger sv eq complete
          edit EMOS_TYPE '0'
        task sv_te
          trigger sv2gg eq complete
        task sv2fdb
          trigger sv_te eq complete
      endfamily
      family case3
        trigger /tparc_dts/main/make eq complete and ./concordiasi_control eq complete
        edit REGIONNR '3'
        edit INISTEP '48'
        edit FCLENGTH '24'
        edit FSFAMILY 'mc/case3'
        inlimit /tparc_dts/limits:hpcf
        task getini
        task inidata_sv
          trigger getini eq complete
          edit NPES '16'
          edit THREADS '4'
          edit MEM '1300'
          edit MRF '10'
          edit CPUTIME '1000'
          edit STACK '450'
        task sv
          trigger inidata_sv eq complete
          edit MEM '890'
          edit NPES '42'
          edit THREADS '2'
          edit CPUTIME '20000'
        task sv2gg
          trigger sv eq complete
          edit EMOS_TYPE '0'
        task sv_te
          trigger sv2gg eq complete
        task sv2fdb
          trigger sv_te eq complete
      endfamily
      family case4
        trigger /tparc_dts/main/make eq complete and ./concordiasi_control eq complete
        edit REGIONNR '4'
        edit INISTEP '48'
        edit FCLENGTH '24'
        edit FSFAMILY 'mc/case4'
        inlimit /tparc_dts/limits:hpcf
        task getini
        task inidata_sv
          trigger getini eq complete
          edit NPES '16'
          edit THREADS '4'
          edit MEM '1300'
          edit MRF '10'
          edit CPUTIME '1000'
          edit STACK '450'
        task sv
          trigger inidata_sv eq complete
          edit MEM '890'
          edit NPES '42'
          edit THREADS '2'
          edit CPUTIME '20000'
        task sv2gg
          trigger sv eq complete
          edit EMOS_TYPE '0'
        task sv_te
          trigger sv2gg eq complete
        task sv2fdb
          trigger sv_te eq complete
      endfamily
      family case5
        trigger /tparc_dts/main/make eq complete and ./concordiasi_control eq complete
        edit REGIONNR '5'
        edit INISTEP '48'
        edit FCLENGTH '24'
        edit FSFAMILY 'mc/case5'
        inlimit /tparc_dts/limits:hpcf
        task getini
        task inidata_sv
          trigger getini eq complete
          edit NPES '16'
          edit THREADS '4'
          edit MEM '1300'
          edit MRF '10'
          edit CPUTIME '1000'
          edit STACK '450'
        task sv
          trigger inidata_sv eq complete
          edit MEM '890'
          edit NPES '42'
          edit THREADS '2'
          edit CPUTIME '20000'
        task sv2gg
          trigger sv eq complete
          edit EMOS_TYPE '0'
        task sv_te
          trigger sv2gg eq complete
        task sv2fdb
          trigger sv_te eq complete
      endfamily
      family case6
        trigger /tparc_dts/main/make eq complete and ./concordiasi_control eq complete
        edit REGIONNR '6'
        edit INISTEP '48'
        edit FCLENGTH '24'
        edit FSFAMILY 'mc/case6'
        inlimit /tparc_dts/limits:hpcf
        task getini
        task inidata_sv
          trigger getini eq complete
          edit NPES '16'
          edit THREADS '4'
          edit MEM '1300'
          edit MRF '10'
          edit CPUTIME '1000'
          edit STACK '450'
        task sv
          trigger inidata_sv eq complete
          edit MEM '890'
          edit NPES '42'
          edit THREADS '2'
          edit CPUTIME '20000'
        task sv2gg
          trigger sv eq complete
          edit EMOS_TYPE '0'
        task sv_te
          trigger sv2gg eq complete
        task sv2fdb
          trigger sv_te eq complete
      endfamily
      family case7
        trigger /tparc_dts/main/make eq complete and ./concordiasi_control eq complete
        edit REGIONNR '7'
        edit INISTEP '48'
        edit FCLENGTH '24'
        edit FSFAMILY 'mc/case7'
        inlimit /tparc_dts/limits:hpcf
        task getini
        task inidata_sv
          trigger getini eq complete
          edit NPES '16'
          edit THREADS '4'
          edit MEM '1300'
          edit MRF '10'
          edit CPUTIME '1000'
          edit STACK '450'
        task sv
          trigger inidata_sv eq complete
          edit MEM '890'
          edit NPES '42'
          edit THREADS '2'
          edit CPUTIME '20000'
        task sv2gg
          trigger sv eq complete
          edit EMOS_TYPE '0'
        task sv_te
          trigger sv2gg eq complete
        task sv2fdb
          trigger sv_te eq complete
      endfamily
      family case8
        trigger /tparc_dts/main/make eq complete and ./concordiasi_control eq complete
        edit REGIONNR '8'
        edit INISTEP '48'
        edit FCLENGTH '24'
        edit FSFAMILY 'mc/case8'
        inlimit /tparc_dts/limits:hpcf
        task getini
        task inidata_sv
          trigger getini eq complete
          edit NPES '16'
          edit THREADS '4'
          edit MEM '1300'
          edit MRF '10'
          edit CPUTIME '1000'
          edit STACK '450'
        task sv
          trigger inidata_sv eq complete
          edit MEM '890'
          edit NPES '42'
          edit THREADS '2'
          edit CPUTIME '20000'
        task sv2gg
          trigger sv eq complete
          edit EMOS_TYPE '0'
        task sv_te
          trigger sv2gg eq complete
        task sv2fdb
          trigger sv_te eq complete
      endfamily
      family case9
        trigger /tparc_dts/main/make eq complete and ./concordiasi_control eq complete
        edit REGIONNR '9'
        edit INISTEP '48'
        edit FCLENGTH '24'
        edit FSFAMILY 'mc/case9'
        inlimit /tparc_dts/limits:hpcf
        task getini
        task inidata_sv
          trigger getini eq complete
          edit NPES '16'
          edit THREADS '4'
          edit MEM '1300'
          edit MRF '10'
          edit CPUTIME '1000'
          edit STACK '450'
        task sv
          trigger inidata_sv eq complete
          edit MEM '890'
          edit NPES '42'
          edit THREADS '2'
          edit CPUTIME '20000'
        task sv2gg
          trigger sv eq complete
          edit EMOS_TYPE '0'
        task sv_te
          trigger sv2gg eq complete
        task sv2fdb
          trigger sv_te eq complete
      endfamily
      family case10
        trigger /tparc_dts/main/make eq complete and ./concordiasi_control eq complete
        edit REGIONNR '10'
        edit INISTEP '48'
        edit FCLENGTH '24'
        edit FSFAMILY 'mc/case10'
        inlimit /tparc_dts/limits:hpcf
        task getini
        task inidata_sv
          trigger getini eq complete
          edit NPES '16'
          edit THREADS '4'
          edit MEM '1300'
          edit MRF '10'
          edit CPUTIME '1000'
          edit STACK '450'
        task sv
          trigger inidata_sv eq complete
          edit MEM '890'
          edit NPES '42'
          edit THREADS '2'
          edit CPUTIME '20000'
        task sv2gg
          trigger sv eq complete
          edit EMOS_TYPE '0'
        task sv_te
          trigger sv2gg eq complete
        task sv2fdb
          trigger sv_te eq complete
      endfamily
    endfamily
    task logfilestore
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit EMOS_BASE '00'
      edit KEEPLOGS 'no'
      edit LOGSTOECFS 'no'
      edit MOVE_TO_ECFS 'yes'
      edit LOGTASK 'main'
    task logfiles
      late -s +00:05 -c +00:15
      trigger ./00 eq complete
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit MOVE_TO_ECFS 'no'
      edit LOGTASK 'main'
      time 19:00
  endfamily
  family lag
    trigger ( main:YMD ge lag:YMD)
    repeat date YMD 20090327 20190101 1
    inlimit /tparc_dts/limits:mars
    family 00
      edit EMOS_BASE '00'
      edit SUBFSFAMILY ''
      task concordiasi_archivecontrol
        trigger /tparc_dts/main/00/concordiasi_control eq complete
        edit QUEUE 'emos'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        inlimit /tparc_dts/limits:linux_cluster
      family case1
        trigger ./concordiasi_archivecontrol eq complete and /tparc_dts/main/00/case1 eq complete
        edit FSFAMILY 'mc/case1'
        edit REGIONNR '1'
        edit INISTEP '42'
        edit FCLENGTH '30'
        edit DELTA_HOUR '42'
        task svsave_concordiasi
        task svsave
        task archive_sv
      endfamily
      family case2
        trigger ./concordiasi_archivecontrol eq complete and /tparc_dts/main/00/case2 eq complete
        edit FSFAMILY 'mc/case2'
        edit REGIONNR '2'
        edit INISTEP '48'
        edit FCLENGTH '24'
        task svsave_concordiasi
        task svsave
        task archive_sv
      endfamily
      family case3
        trigger ./concordiasi_archivecontrol eq complete and /tparc_dts/main/00/case3 eq complete
        edit FSFAMILY 'mc/case3'
        edit REGIONNR '3'
        edit INISTEP '48'
        edit FCLENGTH '24'
        task svsave_concordiasi
        task svsave
        task archive_sv
      endfamily
      family case4
        trigger ./concordiasi_archivecontrol eq complete and /tparc_dts/main/00/case4 eq complete
        edit FSFAMILY 'mc/case4'
        edit REGIONNR '4'
        edit INISTEP '48'
        edit FCLENGTH '24'
        task svsave_concordiasi
        task svsave
        task archive_sv
      endfamily
      family case5
        trigger ./concordiasi_archivecontrol eq complete and /tparc_dts/main/00/case5 eq complete
        edit FSFAMILY 'mc/case5'
        edit REGIONNR '5'
        edit INISTEP '48'
        edit FCLENGTH '24'
        task svsave_concordiasi
        task svsave
        task archive_sv
      endfamily
      family case6
        trigger ./concordiasi_archivecontrol eq complete and /tparc_dts/main/00/case6 eq complete
        edit FSFAMILY 'mc/case6'
        edit REGIONNR '6'
        edit INISTEP '48'
        edit FCLENGTH '24'
        task svsave_concordiasi
        task svsave
        task archive_sv
      endfamily
      family case7
        trigger ./concordiasi_archivecontrol eq complete and /tparc_dts/main/00/case7 eq complete
        edit FSFAMILY 'mc/case7'
        edit REGIONNR '7'
        edit INISTEP '48'
        edit FCLENGTH '24'
        task svsave_concordiasi
        task svsave
        task archive_sv
      endfamily
      family case8
        trigger ./concordiasi_archivecontrol eq complete and /tparc_dts/main/00/case8 eq complete
        edit FSFAMILY 'mc/case8'
        edit REGIONNR '8'
        edit INISTEP '48'
        edit FCLENGTH '24'
        task svsave_concordiasi
        task svsave
        task archive_sv
      endfamily
      family case9
        trigger ./concordiasi_archivecontrol eq complete and /tparc_dts/main/00/case9 eq complete
        edit FSFAMILY 'mc/case9'
        edit REGIONNR '9'
        edit INISTEP '48'
        edit FCLENGTH '24'
        task svsave_concordiasi
        task svsave
        task archive_sv
      endfamily
      family case10
        trigger ./concordiasi_archivecontrol eq complete and /tparc_dts/main/00/case10 eq complete
        edit FSFAMILY 'mc/case10'
        edit REGIONNR '10'
        edit INISTEP '48'
        edit FCLENGTH '24'
        task svsave_concordiasi
        task svsave
        task archive_sv
      endfamily
    endfamily
    task logfilestore
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit LOGTASK 'lag'
      edit KEEPLOGS 'no'
      edit LOGSTOECFS 'no'
      edit MOVE_TO_ECFS 'yes'
    task logfiles
      late -s +00:05 -c +00:15
      trigger ./00 eq complete
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit MOVE_TO_ECFS 'no'
      edit LOGTASK 'lag'
      time 19:00
  endfamily
  family pop
    trigger ( main:YMD ge pop:YMD)
    repeat date YMD 20090327 20190101 1
    edit SMSFILES '/home/ma/emos/def/tparc_33r1/dts'
    edit EMOS_BASE '00'
    inlimit /tparc_dts/limits:webplots
    family 00
      edit SUBFSFAMILY ''
      edit HOUR '0'
      task concordiasi_plotcontrol
        trigger /tparc_dts/main/00/concordiasi_control eq complete
        edit QUEUE 'emos'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        inlimit /tparc_dts/limits:linux_cluster
      task getobslist
        trigger concordiasi_plotcontrol eq complete
        edit QUEUE 'emos'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        inlimit /tparc_dts/limits:linux_cluster
      family case1
        trigger /tparc_dts/main/make eq complete and ./concordiasi_plotcontrol eq complete and /tparc_dts/main/00/case1 eq complete
        edit FSFAMILY 'mc/case1'
        edit INISTEP '42'
        edit FCLENGTH '30'
        edit DELTA_HOUR '42'
        task plotftg
          edit QUEUE 'emos'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /tparc_dts/limits:linux_cluster
      endfamily
      family case2
        trigger /tparc_dts/main/make eq complete and ./concordiasi_plotcontrol eq complete and /tparc_dts/main/00/case2 eq complete
        edit FSFAMILY 'mc/case2'
        edit INISTEP '48'
        edit FCLENGTH '24'
        task plotftg
          edit QUEUE 'emos'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /tparc_dts/limits:linux_cluster
      endfamily
      family case3
        trigger /tparc_dts/main/make eq complete and ./concordiasi_plotcontrol eq complete and /tparc_dts/main/00/case3 eq complete
        edit FSFAMILY 'mc/case3'
        edit INISTEP '48'
        edit FCLENGTH '24'
        task plotftg
          edit QUEUE 'emos'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /tparc_dts/limits:linux_cluster
      endfamily
      family case4
        trigger /tparc_dts/main/make eq complete and ./concordiasi_plotcontrol eq complete and /tparc_dts/main/00/case4 eq complete
        edit FSFAMILY 'mc/case4'
        edit INISTEP '48'
        edit FCLENGTH '24'
        task plotftg
          edit QUEUE 'emos'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /tparc_dts/limits:linux_cluster
      endfamily
      family case5
        trigger /tparc_dts/main/make eq complete and ./concordiasi_plotcontrol eq complete and /tparc_dts/main/00/case5 eq complete
        edit FSFAMILY 'mc/case5'
        edit INISTEP '48'
        edit FCLENGTH '24'
        task plotftg
          edit QUEUE 'emos'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /tparc_dts/limits:linux_cluster
      endfamily
      family case6
        trigger /tparc_dts/main/make eq complete and ./concordiasi_plotcontrol eq complete and /tparc_dts/main/00/case6 eq complete
        edit FSFAMILY 'mc/case6'
        edit INISTEP '48'
        edit FCLENGTH '24'
        task plotftg
          edit QUEUE 'emos'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /tparc_dts/limits:linux_cluster
      endfamily
      family case7
        trigger /tparc_dts/main/make eq complete and ./concordiasi_plotcontrol eq complete and /tparc_dts/main/00/case7 eq complete
        edit FSFAMILY 'mc/case7'
        edit INISTEP '48'
        edit FCLENGTH '24'
        task plotftg
          edit QUEUE 'emos'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /tparc_dts/limits:linux_cluster
      endfamily
      family case8
        trigger /tparc_dts/main/make eq complete and ./concordiasi_plotcontrol eq complete and /tparc_dts/main/00/case8 eq complete
        edit FSFAMILY 'mc/case8'
        edit INISTEP '48'
        edit FCLENGTH '24'
        task plotftg
          edit QUEUE 'emos'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /tparc_dts/limits:linux_cluster
      endfamily
      family case9
        trigger /tparc_dts/main/make eq complete and ./concordiasi_plotcontrol eq complete and /tparc_dts/main/00/case9 eq complete
        edit FSFAMILY 'mc/case9'
        edit INISTEP '48'
        edit FCLENGTH '24'
        task plotftg
          edit QUEUE 'emos'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /tparc_dts/limits:linux_cluster
      endfamily
      family case10
        trigger /tparc_dts/main/make eq complete and ./concordiasi_plotcontrol eq complete and /tparc_dts/main/00/case10 eq complete
        edit FSFAMILY 'mc/case10'
        edit INISTEP '48'
        edit FCLENGTH '24'
        task plotftg
          edit QUEUE 'emos'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /tparc_dts/limits:linux_cluster
      endfamily
    endfamily
    task cases2ftp
      trigger /tparc_dts/main/00/concordiasi_control eq complete
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      inlimit /tparc_dts/limits:linux_cluster
    task cases2web
      trigger ./00 eq complete
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      inlimit /tparc_dts/limits:linux_cluster
    task clean
      trigger ./00 eq complete and cases2web eq complete
    task clean_ws
      trigger ./00 eq complete and cases2web eq complete
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      inlimit /tparc_dts/limits:linux_cluster
    task logfilestore
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit SMSFILES '/home/ma/emos/def/tparc_33r1/dts'
      edit KEEPLOGS 'no'
      edit LOGSTOECFS 'no'
      edit MOVE_TO_ECFS 'yes'
      edit LOGTASK 'pop'
    task logfiles
      late -s +00:05 -c +00:15
      trigger ./00 eq complete and ./clean eq complete
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit SMSFILES '/home/ma/emos/def/tparc_33r1/dts'
      edit MOVE_TO_ECFS 'no'
      edit LOGTASK 'pop'
      time 22:00
  endfamily
  family sap_other_centres
    trigger ( main:YMD ge pop:YMD)
    repeat date YMD 20090327 20190101 1
    edit EMOS_BASE '00'
    family 00
      family ukmo
        edit CENTRE 'metoffice'
        edit SUBFSFAMILY ''
        edit FSFAMILY 'mc/%CENTRE%'
        edit HOUR '0'
        edit INISTEP '48'
        edit FCLENGTH '24'
        edit CENTRE_ORIGIN 'egrr'
        edit SAP_METHOD '0'
        edit SAP_TYPE 'svar'
        edit QUEUE 'emos'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        inlimit /tparc_dts/limits:linux_cluster
        task getdata
          time 17:00
        task plotftg_other
          trigger getdata eq complete
        task sap_other2web
          trigger plotftg_other eq complete and /tparc_dts/pop/cases2web eq complete
        task archive_other
          trigger sap_other2web eq complete
      endfamily
      family ncep
        edit CENTRE 'ncep'
        edit SUBFSFAMILY ''
        edit FSFAMILY 'mc/%CENTRE%'
        edit HOUR '0'
        edit INISTEP '48'
        edit FCLENGTH '24'
        edit CENTRE_ORIGIN 'kwbc'
        edit SAP_METHOD '0'
        edit SAP_TYPE 'svar'
        edit QUEUE 'emos'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        inlimit /tparc_dts/limits:linux_cluster
        task getdata
          time 17:00
        task getfields
          trigger getdata eq complete
        task plotftg_other
          trigger getfields eq complete
        task sap_ncep2web
          trigger plotftg_other eq complete and /tparc_dts/pop/cases2web eq complete
        task others2cfs
          trigger sap_ncep2web eq complete
        task archive_other
          trigger others2cfs eq complete
      endfamily
      family nrl
        edit CENTRE 'nrl'
        edit SUBFSFAMILY ''
        edit FSFAMILY 'mc/%CENTRE%'
        edit HOUR '0'
        edit INISTEP '48'
        edit FCLENGTH '24'
        edit CENTRE_ORIGIN 'fnmoc'
        edit SAP_METHOD '0'
        edit SAP_TYPE 'sv'
        edit QUEUE 'emos'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        inlimit /tparc_dts/limits:linux_cluster
        task get_some_data
          time 18:20
        task get_some_fields
          trigger get_some_data eq complete
        task plotftg_other
          trigger get_some_fields eq complete
        task sap_nrl2web
          trigger plotftg_other eq complete and /tparc_dts/pop/cases2web eq complete
        task others2cfs
          trigger sap_nrl2web eq complete
        task archive_other
          trigger others2cfs eq complete
      endfamily
      family jma
        edit CENTRE 'jma'
        edit SUBFSFAMILY ''
        edit FSFAMILY 'mc/%CENTRE%'
        edit HOUR '0'
        edit INISTEP '48'
        edit FCLENGTH '24'
        edit CENTRE_ORIGIN 'jma'
        edit SAP_METHOD '0'
        edit SAP_TYPE 'svar'
        edit QUEUE 'emos'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        inlimit /tparc_dts/limits:linux_cluster
        task get_some_data
          time 12:00
        task get_some_fields
          trigger get_some_data eq complete
        task grib_mod_jma
          trigger get_some_fields eq complete
        task plotftg_other
          trigger grib_mod_jma eq complete
        task sap_jma2web
          trigger plotftg_other eq complete and /tparc_dts/pop/cases2web eq complete
        task others2cfs
          trigger sap_jma2web eq complete
        task archive_other
          trigger sap_jma2web eq complete
      endfamily
      family ntu
        edit CENTRE 'ntu'
        edit SUBFSFAMILY ''
        edit FSFAMILY 'mc/%CENTRE%'
        edit HOUR '0'
        edit INISTEP '48'
        edit FCLENGTH '24'
        edit CENTRE_ORIGIN 'ntu'
        edit SAP_METHOD '0'
        edit SAP_TYPE 'svar'
        edit QUEUE 'emos'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        inlimit /tparc_dts/limits:linux_cluster
        task getdata
          time 10:00
        task getfields
          trigger getdata eq complete
        task plotftg_other
          trigger getfields eq complete
        task sap_ntu2web
          trigger plotftg_other eq complete and /tparc_dts/pop/cases2web eq complete
        task archive_other
          trigger sap_ntu2web eq complete
      endfamily
      family UYonsei
        edit CENTRE 'UYonsei'
        edit SUBFSFAMILY ''
        edit FSFAMILY 'mc/%CENTRE%'
        edit HOUR '0'
        edit INISTEP '48'
        edit FCLENGTH '24'
        edit CENTRE_ORIGIN 'UYonsei'
        edit SAP_METHOD '0'
        edit SAP_TYPE 'svar'
        edit QUEUE 'emos'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        inlimit /tparc_dts/limits:linux_cluster
        task get_some_data
          time 15:45
        task get_some_fields
          trigger get_some_data eq complete
        task plotftg_other
          trigger get_some_fields eq complete
        task sap_UYonsei2web
          trigger plotftg_other eq complete and /tparc_dts/pop/cases2web eq complete
        task others2cfs
          trigger sap_UYonsei2web eq complete
        task archive_other
          trigger others2cfs eq complete
      endfamily
      family UWash
        edit CENTRE 'UWash'
        edit SUBFSFAMILY ''
        edit FSFAMILY 'mc/%CENTRE%'
        edit HOUR '0'
        edit INISTEP '48'
        edit FCLENGTH '24'
        edit CENTRE_ORIGIN 'UWash'
        edit SAP_METHOD '0'
        edit SAP_TYPE 'svar'
        edit QUEUE 'emos'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        inlimit /tparc_dts/limits:linux_cluster
        task get_some_data
          time 10:00
        task get_some_fields
          trigger get_some_data eq complete
        task plotftg_other
          trigger get_some_fields eq complete
        task sap_UWash2web
          trigger plotftg_other eq complete and /tparc_dts/pop/cases2web eq complete
        task archive_other
          trigger sap_UWash2web eq complete
      endfamily
      family noaa_hrd
        edit CENTRE 'noaa_hrd'
        edit SUBFSFAMILY ''
        edit FSFAMILY 'mc/%CENTRE%'
        edit HOUR '0'
        edit INISTEP '48'
        edit FCLENGTH '24'
        edit CENTRE_ORIGIN 'noaa_hrd'
        edit SAP_METHOD '0'
        edit SAP_TYPE 'svar'
        edit QUEUE 'emos'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        inlimit /tparc_dts/limits:linux_cluster
        task getdata
          time 10:00
        task getfields
          trigger getdata eq complete
        task plotftg_other
          trigger getfields eq complete
        task sap_noaa_hrd2web
          trigger plotftg_other eq complete and /tparc_dts/pop/cases2web eq complete
        task archive_other
          trigger sap_noaa_hrd2web eq complete
      endfamily
    endfamily
    task logfilestore
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit KEEPLOGS 'no'
      edit LOGSTOECFS 'no'
      edit MOVE_TO_ECFS 'yes'
      edit LOGTASK 'sap_other_centres'
    task logfiles
      late -s +00:05 -c +00:15
      trigger ./00 eq complete
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit MOVE_TO_ECFS 'no'
      edit LOGTASK 'sap_other_centres'
      time 22:00
  endfamily
  family post_process
    trigger ( main:YMD ge pop:YMD)
    repeat date YMD 20090327 20190101 1
    edit EMOS_BASE '00'
    edit QUEUE 'emos'
    edit QUEUE_EPILOG 'emos-exp'
    edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
    edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
    edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
    edit LOGDIR '/vol/emos_nc/output'
    edit SMSOUT '/vol/emos_nc/output'
    inlimit /tparc_dts/limits:linux_cluster
    task fixed_cases
      time 10:00
    task logfilestore
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit KEEPLOGS 'no'
      edit LOGSTOECFS 'no'
      edit MOVE_TO_ECFS 'yes'
      edit LOGTASK 'post_process'
    task logfiles
      late -s +00:05 -c +00:15
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit MOVE_TO_ECFS 'no'
      edit LOGTASK 'post_process'
      time 19:00
  endfamily
endsuite
suite dts
  edit USER 'emos'
  edit SCHOST 'c1b'
  edit WSHOST 'swarm'
  edit WSLOGDIR '/vol/emos_nc/output'
  edit SCLOGDIR '/s1b/emos_esuite/emos_data/0014/log'
  edit EMOS_TYPE '1'
  edit USE_YMD 'true'
  edit YMD '20080901'
  edit DATEMASK '*.*.*'
  edit VERSION '0014'
  edit INIEXPVER '0001'
  edit FIRST_DAY '0'
  edit SMSTRIES '2'
  edit DELTA_DAY '0'
  edit EMOS_TIME_STEP_H '0000'
  edit FSFAMILY 'mc'
  edit USER_PRIORITY '70'
  edit PRIORITY '70'
  edit ININPES '16'
  edit NPROCA '16'
  edit NPROCB '1'
  edit SVMODE 'tg'
  edit REINITIALIZE 'false'
  edit SMSHOME '/vol/emos_nc/output'
  edit SMSFILES '/home/ma/emos/def/dts/smsfiles'
  edit SMSINCLUDE '/home/ma/emos/def/dts/include'
  edit ACCOUNT 'oodts'
  edit STREAM 'EF'
  edit EMOS_STREAM 'EF'
  edit MARS_DB 'marser'
  edit EPSTYPE 'sv'
  edit MEMBER '0'
  edit INILEVELS '91'
  edit IS_REAL_TIME 'is_real_time'
  edit RUN_3HOURS_PP '1'
  edit FCGROUP 'fc0'
  edit SMSLOGHOST 'c1b'
  edit SMSLOGPORT '9316'
  edit SVTHREADS '2'
  edit QUEUE 'ns'
  edit QUEUE_EPILOG 'ns'
  edit SMSCMD 'smssubmit %USER% %SCHOST% %SMSJOB% %SMSJOBOUT%'
  edit SMSKILL 'smskill %USER% %SCHOST% %SMSRID% %SMSJOB%'
  edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %SCHOST% %SMSRID% %SMSJOB%'
  edit LOGDIR '/s1b/emos_esuite/emos_data/0014/log'
  edit SMSOUT '/s1b/emos_esuite/emos_data/0014/log'
  edit SLEEPTIME '1'
  inlimit /limits:c1b
  clock hybrid 
  family limits
    limit c1b 5
    limit mars 3
    limit webplots 3
    limit swarm 3
  endfamily
  family main
    repeat date YMD 20080901 20121212 1
    family make
      edit EMOS_BASE '00'
      task setup
      task links
        trigger setup eq complete
      task datalinks
        trigger links eq complete and setup eq complete
      family bins
        trigger datalinks eq complete
        task ifs
        task prepdata
        task mc_tools
      endfamily
    endfamily
    family 00
      edit EMOS_BASE '00'
      task control
        edit QUEUE 'emos'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        inlimit /dts/limits:swarm
        time 10:35
      family case1
        trigger /dts/main/make eq complete and ./control eq complete
        edit FSFAMILY 'mc/case1'
        edit REGIONNR '1'
        edit INISTEP '42'
        edit FCLENGTH '30'
        edit ALAT1 '47.3'
        edit ALON1 '13.5'
        edit ALAT3 '27.8'
        edit ALON3 '37.7'
        edit DELTA_HOUR '42'
        inlimit /dts/limits:c1b
        task getini
        task inidata_sv
          trigger getini eq complete
          edit NPES '16'
          edit THREADS '8'
          edit MEM '1300'
          edit MRF '10'
          edit CPUTIME '1000'
          edit STACK '450'
        task sv
          trigger inidata_sv eq complete
          edit MEM '890'
          edit NPES '42'
          edit THREADS '2'
          edit CPUTIME '20000'
        task sv2gg
          trigger sv eq complete
          edit EMOS_TYPE '0'
        task sv_te
          trigger sv2gg eq complete
        task sv2fdb
          trigger sv_te eq complete
      endfamily
      family case2
        trigger /dts/main/make eq complete and ./control eq complete
        edit FSFAMILY 'mc/case2'
        edit REGIONNR '2'
        edit INISTEP '54'
        edit FCLENGTH '30'
        edit ALAT1 '46.3'
        edit ALON1 '342.9'
        edit ALAT3 '26.9'
        edit ALON3 '7.2'
        edit DELTA_HOUR '54'
        inlimit /dts/limits:c1b
        task getini
        task inidata_sv
          trigger getini eq complete
          edit NPES '16'
          edit THREADS '8'
          edit MEM '1300'
          edit MRF '10'
          edit CPUTIME '1000'
          edit STACK '450'
        task sv
          trigger inidata_sv eq complete
          edit MEM '890'
          edit NPES '42'
          edit THREADS '2'
          edit CPUTIME '20000'
        task sv2gg
          trigger sv eq complete
          edit EMOS_TYPE '0'
        task sv_te
          trigger sv2gg eq complete
        task sv2fdb
          trigger sv_te eq complete
      endfamily
      family case3
        trigger /dts/main/make eq complete and ./control eq complete
        edit FSFAMILY 'mc/case3'
        edit REGIONNR '3'
        edit INISTEP '66'
        edit FCLENGTH '30'
        edit ALAT1 '49.5'
        edit ALON1 '353.7'
        edit ALAT3 '30.1'
        edit ALON3 '18'
        edit DELTA_HOUR '66'
        inlimit /dts/limits:c1b
        task getini
        task inidata_sv
          trigger getini eq complete
          edit NPES '16'
          edit THREADS '8'
          edit MEM '1300'
          edit MRF '10'
          edit CPUTIME '1000'
          edit STACK '450'
        task sv
          trigger inidata_sv eq complete
          edit MEM '890'
          edit NPES '42'
          edit THREADS '2'
          edit CPUTIME '20000'
        task sv2gg
          trigger sv eq complete
          edit EMOS_TYPE '0'
        task sv_te
          trigger sv2gg eq complete
        task sv2fdb
          trigger sv_te eq complete
      endfamily
      family case4
        trigger /dts/main/make eq complete and ./control eq complete
        edit FSFAMILY 'mc/case4'
        edit REGIONNR '4'
        edit INISTEP '42'
        edit FCLENGTH '42'
        edit ALAT1 '48.1'
        edit ALON1 '353.6'
        edit ALAT3 '28.6'
        edit ALON3 '17.8'
        edit DELTA_HOUR '42'
        inlimit /dts/limits:c1b
        task getini
        task inidata_sv
          trigger getini eq complete
          edit NPES '16'
          edit THREADS '8'
          edit MEM '1300'
          edit MRF '10'
          edit CPUTIME '1000'
          edit STACK '450'
        task sv
          trigger inidata_sv eq complete
          edit MEM '890'
          edit NPES '42'
          edit THREADS '2'
          edit CPUTIME '20000'
        task sv2gg
          trigger sv eq complete
          edit EMOS_TYPE '0'
        task sv_te
          trigger sv2gg eq complete
        task sv2fdb
          trigger sv_te eq complete
      endfamily
      family case5
        trigger /dts/main/make eq complete and ./control eq complete
        edit FSFAMILY 'mc/case5'
        edit REGIONNR '5'
        edit INISTEP '54'
        edit FCLENGTH '30'
        edit ALAT1 '48.9'
        edit ALON1 '356.5'
        edit ALAT3 '29.4'
        edit ALON3 '20.8'
        edit DELTA_HOUR '54'
        inlimit /dts/limits:c1b
        task getini
        task inidata_sv
          trigger getini eq complete
          edit NPES '16'
          edit THREADS '8'
          edit MEM '1300'
          edit MRF '10'
          edit CPUTIME '1000'
          edit STACK '450'
        task sv
          trigger inidata_sv eq complete
          edit MEM '890'
          edit NPES '42'
          edit THREADS '2'
          edit CPUTIME '20000'
        task sv2gg
          trigger sv eq complete
          edit EMOS_TYPE '0'
        task sv_te
          trigger sv2gg eq complete
        task sv2fdb
          trigger sv_te eq complete
      endfamily
      family case6
        trigger /dts/main/make eq complete and ./control eq complete
        edit FSFAMILY 'mc/case6'
        edit REGIONNR '6'
        edit INISTEP '48'
        edit FCLENGTH '24'
        inlimit /dts/limits:c1b
        task getini
        task inidata_sv
          trigger getini eq complete
          edit NPES '16'
          edit THREADS '8'
          edit MEM '1300'
          edit MRF '10'
          edit CPUTIME '1000'
          edit STACK '450'
        task sv
          trigger inidata_sv eq complete
          edit MEM '890'
          edit NPES '42'
          edit THREADS '2'
          edit CPUTIME '20000'
        task sv2gg
          trigger sv eq complete
          edit EMOS_TYPE '0'
        task sv_te
          trigger sv2gg eq complete
        task sv2fdb
          trigger sv_te eq complete
      endfamily
      family case7
        trigger /dts/main/make eq complete and ./control eq complete
        edit FSFAMILY 'mc/case7'
        edit REGIONNR '7'
        edit INISTEP '48'
        edit FCLENGTH '24'
        inlimit /dts/limits:c1b
        task getini
        task inidata_sv
          trigger getini eq complete
          edit NPES '16'
          edit THREADS '8'
          edit MEM '1300'
          edit MRF '10'
          edit CPUTIME '1000'
          edit STACK '450'
        task sv
          trigger inidata_sv eq complete
          edit MEM '890'
          edit NPES '42'
          edit THREADS '2'
          edit CPUTIME '20000'
        task sv2gg
          trigger sv eq complete
          edit EMOS_TYPE '0'
        task sv_te
          trigger sv2gg eq complete
        task sv2fdb
          trigger sv_te eq complete
      endfamily
      family case8
        trigger /dts/main/make eq complete and ./control eq complete
        edit FSFAMILY 'mc/case8'
        edit REGIONNR '8'
        edit INISTEP '48'
        edit FCLENGTH '24'
        inlimit /dts/limits:c1b
        task getini
        task inidata_sv
          trigger getini eq complete
          edit NPES '16'
          edit THREADS '8'
          edit MEM '1300'
          edit MRF '10'
          edit CPUTIME '1000'
          edit STACK '450'
        task sv
          trigger inidata_sv eq complete
          edit MEM '890'
          edit NPES '42'
          edit THREADS '2'
          edit CPUTIME '20000'
        task sv2gg
          trigger sv eq complete
          edit EMOS_TYPE '0'
        task sv_te
          trigger sv2gg eq complete
        task sv2fdb
          trigger sv_te eq complete
      endfamily
      family case9
        trigger /dts/main/make eq complete and ./control eq complete
        edit FSFAMILY 'mc/case9'
        edit REGIONNR '9'
        edit INISTEP '48'
        edit FCLENGTH '24'
        inlimit /dts/limits:c1b
        task getini
        task inidata_sv
          trigger getini eq complete
          edit NPES '16'
          edit THREADS '8'
          edit MEM '1300'
          edit MRF '10'
          edit CPUTIME '1000'
          edit STACK '450'
        task sv
          trigger inidata_sv eq complete
          edit MEM '890'
          edit NPES '42'
          edit THREADS '2'
          edit CPUTIME '20000'
        task sv2gg
          trigger sv eq complete
          edit EMOS_TYPE '0'
        task sv_te
          trigger sv2gg eq complete
        task sv2fdb
          trigger sv_te eq complete
      endfamily
      family case10
        trigger /dts/main/make eq complete and ./control eq complete
        edit FSFAMILY 'mc/case10'
        edit REGIONNR '10'
        edit INISTEP '48'
        edit FCLENGTH '24'
        inlimit /dts/limits:c1b
        task getini
        task inidata_sv
          trigger getini eq complete
          edit NPES '16'
          edit THREADS '8'
          edit MEM '1300'
          edit MRF '10'
          edit CPUTIME '1000'
          edit STACK '450'
        task sv
          trigger inidata_sv eq complete
          edit MEM '890'
          edit NPES '42'
          edit THREADS '2'
          edit CPUTIME '20000'
        task sv2gg
          trigger sv eq complete
          edit EMOS_TYPE '0'
        task sv_te
          trigger sv2gg eq complete
        task sv2fdb
          trigger sv_te eq complete
      endfamily
    endfamily
    task logfilestore
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit EMOS_BASE '00'
      edit KEEPLOGS 'no'
      edit LOGSTOECFS 'no'
      edit MOVE_TO_ECFS 'yes'
      edit LOGTASK 'main'
    task logfiles
      late -s +00:05 -c +00:15
      trigger ./00 eq complete
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit MOVE_TO_ECFS 'no'
      edit LOGTASK 'main'
      time 15:00
  endfamily
  family lag
    trigger ( main:YMD ge lag:YMD)
    repeat date YMD 20080901 20121212 1
    inlimit /dts/limits:mars
    family 00
      edit EMOS_BASE '00'
      edit SUBFSFAMILY ''
      task archivecontrol
        trigger /dts/main/00/control eq complete
        edit QUEUE 'emos'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        inlimit /dts/limits:swarm
      family case1
        trigger ./archivecontrol eq complete and /dts/main/00/case1 eq complete
        edit FSFAMILY 'mc/case1'
        edit REGIONNR '1'
        edit INISTEP '42'
        edit FCLENGTH '30'
        edit DELTA_HOUR '42'
        task svsave
        task archive_sv
      endfamily
      family case2
        trigger ./archivecontrol eq complete and /dts/main/00/case2 eq complete
        edit FSFAMILY 'mc/case2'
        edit REGIONNR '2'
        edit INISTEP '54'
        edit FCLENGTH '30'
        edit DELTA_HOUR '54'
        task svsave
        task archive_sv
      endfamily
      family case3
        trigger ./archivecontrol eq complete and /dts/main/00/case3 eq complete
        edit FSFAMILY 'mc/case3'
        edit REGIONNR '3'
        edit INISTEP '66'
        edit FCLENGTH '30'
        edit DELTA_HOUR '66'
        task svsave
        task archive_sv
      endfamily
      family case4
        trigger ./archivecontrol eq complete and /dts/main/00/case4 eq complete
        edit FSFAMILY 'mc/case4'
        edit REGIONNR '4'
        edit INISTEP '42'
        edit FCLENGTH '42'
        edit DELTA_HOUR '42'
        task svsave
        task archive_sv
      endfamily
      family case5
        trigger ./archivecontrol eq complete and /dts/main/00/case5 eq complete
        edit FSFAMILY 'mc/case5'
        edit REGIONNR '5'
        edit INISTEP '54'
        edit FCLENGTH '30'
        edit DELTA_HOUR '54'
        task svsave
        task archive_sv
      endfamily
      family case6
        trigger ./archivecontrol eq complete and /dts/main/00/case6 eq complete
        edit FSFAMILY 'mc/case6'
        edit REGIONNR '6'
        edit INISTEP '48'
        edit FCLENGTH '24'
        task svsave
        task archive_sv
      endfamily
      family case7
        trigger ./archivecontrol eq complete and /dts/main/00/case7 eq complete
        edit FSFAMILY 'mc/case7'
        edit REGIONNR '7'
        edit INISTEP '48'
        edit FCLENGTH '24'
        task svsave
        task archive_sv
      endfamily
      family case8
        trigger ./archivecontrol eq complete and /dts/main/00/case8 eq complete
        edit FSFAMILY 'mc/case8'
        edit REGIONNR '8'
        edit INISTEP '48'
        edit FCLENGTH '24'
        task svsave
        task archive_sv
      endfamily
      family case9
        trigger ./archivecontrol eq complete and /dts/main/00/case9 eq complete
        edit FSFAMILY 'mc/case9'
        edit REGIONNR '9'
        edit INISTEP '48'
        edit FCLENGTH '24'
        task svsave
        task archive_sv
      endfamily
      family case10
        trigger ./archivecontrol eq complete and /dts/main/00/case10 eq complete
        edit FSFAMILY 'mc/case10'
        edit REGIONNR '10'
        edit INISTEP '48'
        edit FCLENGTH '24'
        task svsave
        task archive_sv
      endfamily
    endfamily
    task logfilestore
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit LOGTASK 'lag'
      edit KEEPLOGS 'no'
      edit LOGSTOECFS 'no'
      edit MOVE_TO_ECFS 'yes'
    task logfiles
      late -s +00:05 -c +00:15
      trigger ./00 eq complete
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit MOVE_TO_ECFS 'no'
      edit LOGTASK 'lag'
      time 13:00
  endfamily
  family pop
    trigger ( main:YMD ge pop:YMD)
    repeat date YMD 20080901 20121212 1
    edit EMOS_BASE '00'
    inlimit /dts/limits:webplots
    family 00
      edit SUBFSFAMILY ''
      edit HOUR '0'
      task plotcontrol
        trigger /dts/main/00/control eq complete
        edit QUEUE 'emos'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        inlimit /dts/limits:swarm
      task getobslist
        trigger plotcontrol eq complete
        edit QUEUE 'emos'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        inlimit /dts/limits:swarm
      family case1
        trigger /dts/main/make eq complete and ./plotcontrol eq complete and /dts/main/00/case1 eq complete
        edit FSFAMILY 'mc/case1'
        edit INISTEP '42'
        edit FCLENGTH '30'
        edit DELTA_HOUR '42'
        task plotftg
          edit QUEUE 'emos'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /dts/limits:swarm
      endfamily
      family case2
        trigger /dts/main/make eq complete and ./plotcontrol eq complete and /dts/main/00/case2 eq complete
        edit FSFAMILY 'mc/case2'
        edit INISTEP '54'
        edit FCLENGTH '30'
        edit DELTA_HOUR '54'
        task plotftg
          edit QUEUE 'emos'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /dts/limits:swarm
      endfamily
      family case3
        trigger /dts/main/make eq complete and ./plotcontrol eq complete and /dts/main/00/case3 eq complete
        edit FSFAMILY 'mc/case3'
        edit INISTEP '66'
        edit FCLENGTH '30'
        edit DELTA_HOUR '66'
        task plotftg
          edit QUEUE 'emos'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /dts/limits:swarm
      endfamily
      family case4
        trigger /dts/main/make eq complete and ./plotcontrol eq complete and /dts/main/00/case4 eq complete
        edit FSFAMILY 'mc/case4'
        edit INISTEP '42'
        edit FCLENGTH '42'
        edit DELTA_HOUR '42'
        task plotftg
          edit QUEUE 'emos'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /dts/limits:swarm
      endfamily
      family case5
        trigger /dts/main/make eq complete and ./plotcontrol eq complete and /dts/main/00/case5 eq complete
        edit FSFAMILY 'mc/case5'
        edit INISTEP '54'
        edit FCLENGTH '30'
        edit DELTA_HOUR '54'
        task plotftg
          edit QUEUE 'emos'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /dts/limits:swarm
      endfamily
      family case6
        trigger /dts/main/make eq complete and ./plotcontrol eq complete and /dts/main/00/case6 eq complete
        edit FSFAMILY 'mc/case6'
        edit INISTEP '48'
        edit FCLENGTH '24'
        task plotftg
          edit QUEUE 'emos'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /dts/limits:swarm
      endfamily
      family case7
        trigger /dts/main/make eq complete and ./plotcontrol eq complete and /dts/main/00/case7 eq complete
        edit FSFAMILY 'mc/case7'
        edit INISTEP '48'
        edit FCLENGTH '24'
        task plotftg
          edit QUEUE 'emos'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /dts/limits:swarm
      endfamily
      family case8
        trigger /dts/main/make eq complete and ./plotcontrol eq complete and /dts/main/00/case8 eq complete
        edit FSFAMILY 'mc/case8'
        edit INISTEP '48'
        edit FCLENGTH '24'
        task plotftg
          edit QUEUE 'emos'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /dts/limits:swarm
      endfamily
      family case9
        trigger /dts/main/make eq complete and ./plotcontrol eq complete and /dts/main/00/case9 eq complete
        edit FSFAMILY 'mc/case9'
        edit INISTEP '48'
        edit FCLENGTH '24'
        task plotftg
          edit QUEUE 'emos'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /dts/limits:swarm
      endfamily
      family case10
        trigger /dts/main/make eq complete and ./plotcontrol eq complete and /dts/main/00/case10 eq complete
        edit FSFAMILY 'mc/case10'
        edit INISTEP '48'
        edit FCLENGTH '24'
        task plotftg
          edit QUEUE 'emos'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /dts/limits:swarm
      endfamily
    endfamily
    task cases2ftp
      trigger /dts/main/00/control eq complete
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      inlimit /dts/limits:swarm
    task cases2web
      trigger ./00 eq complete
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      inlimit /dts/limits:swarm
    task clean
      trigger ./00 eq complete and cases2web eq complete
    task clean_ws
      trigger ./00 eq complete and cases2web eq complete
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      inlimit /dts/limits:swarm
    task logfilestore
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit KEEPLOGS 'no'
      edit LOGSTOECFS 'no'
      edit MOVE_TO_ECFS 'yes'
      edit LOGTASK 'pop'
    task logfiles
      late -s +00:05 -c +00:15
      trigger ./00 eq complete and ./clean eq complete
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit MOVE_TO_ECFS 'no'
      edit LOGTASK 'pop'
      time 15:00
  endfamily
  family sap_other_centres
    trigger ( main:YMD ge pop:YMD)
    repeat date YMD 20080901 20121212 1
    edit EMOS_BASE '00'
    family 00
      family ukmo
        edit CENTRE 'metoffice'
        edit SUBFSFAMILY ''
        edit FSFAMILY 'mc/%CENTRE%'
        edit HOUR '0'
        edit INISTEP '48'
        edit FCLENGTH '24'
        edit CENTRE_ORIGIN 'egrr'
        edit SAP_METHOD '0'
        edit SAP_TYPE 'svar'
        edit QUEUE 'emos'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        inlimit /dts/limits:swarm
        task getdata
          time 11:00
        task plotftg_other
          trigger getdata eq complete
        task sap_other2web
          trigger plotftg_other eq complete and /dts/pop/cases2web eq complete
        task archive_other
          trigger sap_other2web eq complete
      endfamily
      family meteofrance
        edit CENTRE 'meteofrance'
        edit SUBFSFAMILY ''
        edit FSFAMILY 'mc/%CENTRE%'
        edit HOUR '0'
        edit INISTEP '48'
        edit FCLENGTH '24'
        edit CENTRE_ORIGIN 'lfpw'
        edit SAP_METHOD '0'
        edit SAP_TYPE 'svar'
        edit QUEUE 'emos'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        inlimit /dts/limits:swarm
        task getdata
          time 11:00
        task getfields
          trigger getdata eq complete
        task plotftg_other
          trigger getfields eq complete
        task sap_meteofrance2web
          trigger plotftg_other eq complete and /dts/pop/cases2web eq complete
        task archive_other
          trigger sap_meteofrance2web eq complete
      endfamily
      family uib
        edit CENTRE 'uib'
        edit SUBFSFAMILY ''
        edit FSFAMILY 'mc/%CENTRE%'
        edit HOUR '0'
        edit INISTEP '48'
        edit FCLENGTH '24'
        edit CENTRE_ORIGIN 'uib'
        edit SAP_METHOD '0'
        edit SAP_TYPE 'svar'
        edit QUEUE 'emos'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        inlimit /dts/limits:swarm
        task getdata
          time 11:00
        task getfields
          trigger getdata eq complete
        task plotftg_other
          trigger getfields eq complete
          edit SMSPASS 'stripped'
        task sap_uib2web
          trigger plotftg_other eq complete and /dts/pop/cases2web eq complete
        task archive_other
          trigger sap_uib2web eq complete
      endfamily
      family cnrm
        edit CENTRE 'cnrm'
        edit SUBFSFAMILY ''
        edit FSFAMILY 'mc/%CENTRE%'
        edit HOUR '0'
        edit INISTEP '48'
        edit FCLENGTH '24'
        edit CENTRE_ORIGIN 'cnrm'
        edit SAP_METHOD '0'
        edit SAP_TYPE 'svar'
        edit QUEUE 'emos'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        inlimit /dts/limits:swarm
        task getdata
          time 11:00
        task getfields
          trigger getdata eq complete
        task plotftg_other
          trigger getfields eq complete
        task sap_cnrm2web
          trigger plotftg_other eq complete and /dts/pop/cases2web eq complete
        task archive_other
          trigger sap_cnrm2web eq complete
      endfamily
    endfamily
    task logfilestore
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit KEEPLOGS 'no'
      edit LOGSTOECFS 'no'
      edit MOVE_TO_ECFS 'yes'
      edit LOGTASK 'sap_other_centres'
    task logfiles
      late -s +00:05 -c +00:15
      trigger ./00 eq complete
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit MOVE_TO_ECFS 'no'
      edit LOGTASK 'sap_other_centres'
      time 15:00
  endfamily
  family post_process
    trigger ( main:YMD ge pop:YMD)
    repeat date YMD 20080901 20121212 1
    edit EMOS_BASE '00'
    edit QUEUE 'emos'
    edit QUEUE_EPILOG 'emos-exp'
    edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
    edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
    edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
    edit LOGDIR '/vol/emos_nc/output'
    edit SMSOUT '/vol/emos_nc/output'
    inlimit /dts/limits:swarm
    task monitor
      time 07:00
    task extraobs
      time 14:05
    task sendrequest
      trigger extraobs eq complete
    task logfilestore
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit KEEPLOGS 'no'
      edit LOGSTOECFS 'no'
      edit MOVE_TO_ECFS 'yes'
      edit LOGTASK 'post_process'
    task logfiles
      late -s +00:05 -c +00:15
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit MOVE_TO_ECFS 'no'
      edit LOGTASK 'post_process'
      time 15:00
  endfamily
endsuite
suite seas_fr3_back
  edit VERSION '0001'
  edit EXPVER '0001'
  edit SCHOST 'hpce'
  edit WSHOST 'swarm'
  edit WSLOGDIR '/vol/emos/output'
  edit SCLOGDIR '/hpce/emos_dir/0001/log'
  edit STREAM 'MMSF'
  edit REINITIALIZE 'false'
  edit EPSTYPE 'fc'
  edit EMOS_BASE '00'
  edit DATEMASK '*.*.*'
  edit START_DAY '19781101'
  edit END_DAY '20071201'
  edit YMD '19781101'
  edit USE_YMD 'true'
  edit FCLENGTH '5160'
  edit FCTOTAL '11'
  edit OCTOTAL '1'
  edit OCGROUP 'an0'
  edit RUNCOUPLE '1'
  edit OCSYSTEM '0'
  edit OCMETHOD '1'
  edit ORIGIN 'LFPW'
  edit FDB_ROOT_FR '/fdb'
  edit SMSTRIES '2'
  edit DELTA_DAY '0'
  edit USER_PRIORITY '50'
  edit FIRST_DAY '0'
  edit EMOS_TYPE 'fc'
  edit EMOS_TIME_STEP_H '0000'
  edit EMOS_STREAM 'DA'
  edit MARS_DB 'marsod'
  edit FCNPES '1'
  edit ANNPES '1'
  edit SMSHOME '/vol/emos/output'
  edit SMSFILES '/home/ma/emos/def/seas_fr3_back/smsfiles'
  edit SMSINCLUDE '/home/ma/emos/def/seas_fr3_back/include'
  edit ACCOUNT 'oosu'
  edit USER 'emos'
  edit MEMBER '0'
  edit RUNBACK '1'
  edit QUEUE 'ns'
  edit QUEUE_EPILOG 'ns'
  edit SMSCMD 'smssubmit %USER% %SCHOST% %SMSJOB%'
  edit SMSKILL 'smskill %USER% %SCHOST% %SMSRID% %SMSJOB%'
  edit LOGDIR '/hpce/emos_dir/0001/log'
  edit SMSOUT '/hpce/emos_dir/0001/log'
  edit SMSLOGHOST 'hpce-batch'
  edit SMSLOGPORT '9316'
  edit HOUR '00'
  edit FSFAMILY 'seas'
  edit SLEEPTIME '1'
  inlimit /limits:hpce
  clock hybrid 
  family limits
    limit retrieve 2
    limit mars 2
    limit ws 6
  endfamily
  family make
    task setup
    task mmsf_tools
      trigger setup eq complete
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID%'
      edit LOGDIR '/vol/emos/output'
      edit SMSOUT '/vol/emos/output'
      inlimit /limits:swarm
      inlimit /seas_fr3_back/limits:ws
    task sc_tools
      trigger setup eq complete
    task logfiles
      trigger setup eq complete and sc_tools eq complete
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID%'
      edit LOGDIR '/vol/emos/output'
      edit SMSOUT '/vol/emos/output'
      edit LOGTASK 'make'
  endfamily
  family main
    trigger make eq complete
    repeat enumerated YMD "19790101" "19800101" "19810101" "19820101" "19830101" "19840101" "19850101" "19860101" "19870101" "19880101" "19890101" "19900101" "19910101" "19920101" "19930101" "19940101" "19950101" "19960101" "19970101" "19980101" "19990101" "20000101" "20010101" "20020101" "20030101" "20040101" "20050101" "20060101" "20070101" "19790201" "19800201" "19810201" "19820201" "19830201" "19840201" "19850201" "19860201" "19870201" "19880201" "19890201" "19900201" "19910201" "19920201" "19930201" "19940201" "19950201" "19960201" "19970201" "19980201" "19990201" "20000201" "20010201" "20020201" "20030201" "20040201" "20050201" "20060201" "20070201" "19790301" "19800301" "19810301" "19820301" "19830301" "19840301" "19850301" "19860301" "19870301" "19880301" "19890301" "19900301" "19910301" "19920301" "19930301" "19940301" "19950301" "19960301" "19970301" "19980301" "19990301" "20000301" "20010301" "20020301" "20030301" "20040301" "20050301" "20060301" "20070301" "19790401" "19800401" "19810401" "19820401" "19830401" "19840401" "19850401" "19860401" "19870401" "19880401" "19890401" "19900401" "19910401" "19920401" "19930401" "19940401" "19950401" "19960401" "19970401" "19980401" "19990401" "20000401" "20010401" "20020401" "20030401" "20040401" "20050401" "20060401" "20070401" "19790501" "19800501" "19810501" "19820501" "19830501" "19840501" "19850501" "19860501" "19870501" "19880501" "19890501" "19900501" "19910501" "19920501" "19930501" "19940501" "19950501" "19960501" "19970501" "19980501" "19990501" "20000501" "20010501" "20020501" "20030501" "20040501" "20050501" "20060501" "20070501" "19790601" "19800601" "19810601" "19820601" "19830601" "19840601" "19850601" "19860601" "19870601" "19880601" "19890601" "19900601" "19910601" "19920601" "19930601" "19940601" "19950601" "19960601" "19970601" "19980601" "19990601" "20000601" "20010601" "20020601" "20030601" "20040601" "20050601" "20060601" "20070601" "19790701" "19800701" "19810701" "19820701" "19830701" "19840701" "19850701" "19860701" "19870701" "19880701" "19890701" "19900701" "19910701" "19920701" "19930701" "19940701" "19950701" "19960701" "19970701" "19980701" "19990701" "20000701" "20010701" "20020701" "20030701" "20040701" "20050701" "20060701" "20070701" "19790801" "19800801" "19810801" "19820801" "19830801" "19840801" "19850801" "19860801" "19870801" "19880801" "19890801" "19900801" "19910801" "19920801" "19930801" "19940801" "19950801" "19960801" "19970801" "19980801" "19990801" "20000801" "20010801" "20020801" "20030801" "20040801" "20050801" "20060801" "20070801" "19790901" "19800901" "19810901" "19820901" "19830901" "19840901" "19850901" "19860901" "19870901" "19880901" "19890901" "19900901" "19910901" "19920901" "19930901" "19940901" "19950901" "19960901" "19970901" "19980901" "19990901" "20000901" "20010901" "20020901" "20030901" "20040901" "20050901" "20060901" "20070901" "19791001" "19801001" "19811001" "19821001" "19831001" "19841001" "19851001" "19861001" "19871001" "19881001" "19891001" "19901001" "19911001" "19921001" "19931001" "19941001" "19951001" "19961001" "19971001" "19981001" "19991001" "20001001" "20011001" "20021001" "20031001" "20041001" "20051001" "20061001" "20071001" "19781101" "19791101" "19801101" "19811101" "19821101" "19831101" "19841101" "19851101" "19861101" "19871101" "19881101" "19891101" "19901101" "19911101" "19921101" "19931101" "19941101" "19951101" "19961101" "19971101" "19981101" "19991101" "20001101" "20011101" "20021101" "20031101" "20041101" "20051101" "20061101" "20071101" "19781201" "19791201" "19801201" "19811201" "19821201" "19831201" "19841201" "19851201" "19861201" "19871201" "19881201" "19891201" "19901201" "19911201" "19921201" "19931201" "19941201" "19951201" "19961201" "19971201" "19981201" "19991201" "20001201" "20011201" "20021201" "20031201" "20041201" "20051201" "20061201" "20071201"
    edit LOGTASK 'main'
    family fc
      family fc0
        edit FCNUM 'fcnum'
        edit FCGROUP 'fc0'
        task fr2fdb
          inlimit /seas_fr3_back/limits:retrieve
        task extrafields_fr
          trigger fr2fdb eq complete
        task mm_sfc_fr
          trigger fr2fdb eq complete and extrafields_fr eq complete
        task mm_ua_fr
          trigger fr2fdb eq complete and mm_sfc_fr eq complete
      endfamily
      family fc1
        edit FCNUM 'fcnum'
        edit FCGROUP 'fc1'
        task fr2fdb
          inlimit /seas_fr3_back/limits:retrieve
        task extrafields_fr
          trigger fr2fdb eq complete
        task mm_sfc_fr
          trigger fr2fdb eq complete and extrafields_fr eq complete
        task mm_ua_fr
          trigger fr2fdb eq complete and mm_sfc_fr eq complete
      endfamily
      family fc2
        edit FCNUM 'fcnum'
        edit FCGROUP 'fc2'
        task fr2fdb
          inlimit /seas_fr3_back/limits:retrieve
        task extrafields_fr
          trigger fr2fdb eq complete
        task mm_sfc_fr
          trigger fr2fdb eq complete and extrafields_fr eq complete
        task mm_ua_fr
          trigger fr2fdb eq complete and mm_sfc_fr eq complete
      endfamily
      family fc3
        edit FCNUM 'fcnum'
        edit FCGROUP 'fc3'
        task fr2fdb
          inlimit /seas_fr3_back/limits:retrieve
        task extrafields_fr
          trigger fr2fdb eq complete
        task mm_sfc_fr
          trigger fr2fdb eq complete and extrafields_fr eq complete
        task mm_ua_fr
          trigger fr2fdb eq complete and mm_sfc_fr eq complete
      endfamily
      family fc4
        edit FCNUM 'fcnum'
        edit FCGROUP 'fc4'
        task fr2fdb
          inlimit /seas_fr3_back/limits:retrieve
        task extrafields_fr
          trigger fr2fdb eq complete
        task mm_sfc_fr
          trigger fr2fdb eq complete and extrafields_fr eq complete
        task mm_ua_fr
          trigger fr2fdb eq complete and mm_sfc_fr eq complete
      endfamily
      family fc5
        edit FCNUM 'fcnum'
        edit FCGROUP 'fc5'
        task fr2fdb
          inlimit /seas_fr3_back/limits:retrieve
        task extrafields_fr
          trigger fr2fdb eq complete
        task mm_sfc_fr
          trigger fr2fdb eq complete and extrafields_fr eq complete
        task mm_ua_fr
          trigger fr2fdb eq complete and mm_sfc_fr eq complete
      endfamily
      family fc6
        edit FCNUM 'fcnum'
        edit FCGROUP 'fc6'
        task fr2fdb
          inlimit /seas_fr3_back/limits:retrieve
        task extrafields_fr
          trigger fr2fdb eq complete
        task mm_sfc_fr
          trigger fr2fdb eq complete and extrafields_fr eq complete
        task mm_ua_fr
          trigger fr2fdb eq complete and mm_sfc_fr eq complete
      endfamily
      family fc7
        edit FCNUM 'fcnum'
        edit FCGROUP 'fc7'
        task fr2fdb
          inlimit /seas_fr3_back/limits:retrieve
        task extrafields_fr
          trigger fr2fdb eq complete
        task mm_sfc_fr
          trigger fr2fdb eq complete and extrafields_fr eq complete
        task mm_ua_fr
          trigger fr2fdb eq complete and mm_sfc_fr eq complete
      endfamily
      family fc8
        edit FCNUM 'fcnum'
        edit FCGROUP 'fc8'
        task fr2fdb
          inlimit /seas_fr3_back/limits:retrieve
        task extrafields_fr
          trigger fr2fdb eq complete
        task mm_sfc_fr
          trigger fr2fdb eq complete and extrafields_fr eq complete
        task mm_ua_fr
          trigger fr2fdb eq complete and mm_sfc_fr eq complete
      endfamily
      family fc9
        edit FCNUM 'fcnum'
        edit FCGROUP 'fc9'
        task fr2fdb
          inlimit /seas_fr3_back/limits:retrieve
        task extrafields_fr
          trigger fr2fdb eq complete
        task mm_sfc_fr
          trigger fr2fdb eq complete and extrafields_fr eq complete
        task mm_ua_fr
          trigger fr2fdb eq complete and mm_sfc_fr eq complete
      endfamily
      family fc10
        edit FCNUM 'fcnum'
        edit FCGROUP 'fc10'
        task fr2fdb
          inlimit /seas_fr3_back/limits:retrieve
        task extrafields_fr
          trigger fr2fdb eq complete
        task mm_sfc_fr
          trigger fr2fdb eq complete and extrafields_fr eq complete
        task mm_ua_fr
          trigger fr2fdb eq complete and mm_sfc_fr eq complete
      endfamily
    endfamily
    task logfilestore
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID%'
      edit LOGDIR '/vol/emos/output'
      edit SMSOUT '/vol/emos/output'
      edit KEEPLOGS 'no'
      edit LOGSTOECFS 'no'
      edit MOVE_TO_ECFS 'yes'
    task logfiles
      trigger fc eq complete and logfilestore eq complete
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID%'
      edit LOGDIR '/vol/emos/output'
      edit SMSOUT '/vol/emos/output'
      edit MOVE_TO_ECFS 'no'
  endfamily
  family lag
    trigger make eq complete
    repeat enumerated YMD "19790101" "19800101" "19810101" "19820101" "19830101" "19840101" "19850101" "19860101" "19870101" "19880101" "19890101" "19900101" "19910101" "19920101" "19930101" "19940101" "19950101" "19960101" "19970101" "19980101" "19990101" "20000101" "20010101" "20020101" "20030101" "20040101" "20050101" "20060101" "20070101" "19790201" "19800201" "19810201" "19820201" "19830201" "19840201" "19850201" "19860201" "19870201" "19880201" "19890201" "19900201" "19910201" "19920201" "19930201" "19940201" "19950201" "19960201" "19970201" "19980201" "19990201" "20000201" "20010201" "20020201" "20030201" "20040201" "20050201" "20060201" "20070201" "19790301" "19800301" "19810301" "19820301" "19830301" "19840301" "19850301" "19860301" "19870301" "19880301" "19890301" "19900301" "19910301" "19920301" "19930301" "19940301" "19950301" "19960301" "19970301" "19980301" "19990301" "20000301" "20010301" "20020301" "20030301" "20040301" "20050301" "20060301" "20070301" "19790401" "19800401" "19810401" "19820401" "19830401" "19840401" "19850401" "19860401" "19870401" "19880401" "19890401" "19900401" "19910401" "19920401" "19930401" "19940401" "19950401" "19960401" "19970401" "19980401" "19990401" "20000401" "20010401" "20020401" "20030401" "20040401" "20050401" "20060401" "20070401" "19790501" "19800501" "19810501" "19820501" "19830501" "19840501" "19850501" "19860501" "19870501" "19880501" "19890501" "19900501" "19910501" "19920501" "19930501" "19940501" "19950501" "19960501" "19970501" "19980501" "19990501" "20000501" "20010501" "20020501" "20030501" "20040501" "20050501" "20060501" "20070501" "19790601" "19800601" "19810601" "19820601" "19830601" "19840601" "19850601" "19860601" "19870601" "19880601" "19890601" "19900601" "19910601" "19920601" "19930601" "19940601" "19950601" "19960601" "19970601" "19980601" "19990601" "20000601" "20010601" "20020601" "20030601" "20040601" "20050601" "20060601" "20070601" "19790701" "19800701" "19810701" "19820701" "19830701" "19840701" "19850701" "19860701" "19870701" "19880701" "19890701" "19900701" "19910701" "19920701" "19930701" "19940701" "19950701" "19960701" "19970701" "19980701" "19990701" "20000701" "20010701" "20020701" "20030701" "20040701" "20050701" "20060701" "20070701" "19790801" "19800801" "19810801" "19820801" "19830801" "19840801" "19850801" "19860801" "19870801" "19880801" "19890801" "19900801" "19910801" "19920801" "19930801" "19940801" "19950801" "19960801" "19970801" "19980801" "19990801" "20000801" "20010801" "20020801" "20030801" "20040801" "20050801" "20060801" "20070801" "19790901" "19800901" "19810901" "19820901" "19830901" "19840901" "19850901" "19860901" "19870901" "19880901" "19890901" "19900901" "19910901" "19920901" "19930901" "19940901" "19950901" "19960901" "19970901" "19980901" "19990901" "20000901" "20010901" "20020901" "20030901" "20040901" "20050901" "20060901" "20070901" "19791001" "19801001" "19811001" "19821001" "19831001" "19841001" "19851001" "19861001" "19871001" "19881001" "19891001" "19901001" "19911001" "19921001" "19931001" "19941001" "19951001" "19961001" "19971001" "19981001" "19991001" "20001001" "20011001" "20021001" "20031001" "20041001" "20051001" "20061001" "20071001" "19781101" "19791101" "19801101" "19811101" "19821101" "19831101" "19841101" "19851101" "19861101" "19871101" "19881101" "19891101" "19901101" "19911101" "19921101" "19931101" "19941101" "19951101" "19961101" "19971101" "19981101" "19991101" "20001101" "20011101" "20021101" "20031101" "20041101" "20051101" "20061101" "20071101" "19781201" "19791201" "19801201" "19811201" "19821201" "19831201" "19841201" "19851201" "19861201" "19871201" "19881201" "19891201" "19901201" "19911201" "19921201" "19931201" "19941201" "19951201" "19961201" "19971201" "19981201" "19991201" "20001201" "20011201" "20021201" "20031201" "20041201" "20051201" "20061201" "20071201"
    edit LOGTASK 'lag'
    family archive
      trigger ( /seas_fr3_back/main:YMD gt /seas_fr3_back/lag:YMD) or ( /seas_fr3_back/main:YMD eq /seas_fr3_back/lag:YMD and /seas_fr3_back/main/fc eq complete)
      edit SMSFILES '/home/ma/emos/def/seas_fr3_back/smsfiles/archive'
      inlimit /limits:mars
      inlimit /seas_fr3_back/limits:mars
      task daily_sfc_fr
        label files ""
        event fdb
      task daily_ua_fr
        trigger daily_sfc_fr eq complete
        label files ""
        event fdb
      task mm_sfc_fr
        label files ""
        event fdb
      task mm_ua_fr
        label files ""
        event fdb
    endfamily
    family process
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID%'
      edit LOGDIR '/vol/emos/output'
      edit SMSOUT '/vol/emos/output'
      inlimit /limits:swarm
      inlimit /seas_fr3_back/limits:ws
      family fc0
        trigger ( /seas_fr3_back/main:YMD gt /seas_fr3_back/lag:YMD) or ( /seas_fr3_back/main:YMD eq /seas_fr3_back/lag:YMD and /seas_fr3_back/main/fc/fc0 eq complete)
        edit FCGROUP 'fc0'
        edit FCNUM 'fcnum'
        task nino_sst_fr
        task nino_atmos_fr
          trigger nino_sst_fr eq complete
        task tcyc_fr
          trigger nino_atmos_fr eq complete
      endfamily
      family fc1
        trigger ( /seas_fr3_back/main:YMD gt /seas_fr3_back/lag:YMD) or ( /seas_fr3_back/main:YMD eq /seas_fr3_back/lag:YMD and /seas_fr3_back/main/fc/fc1 eq complete)
        edit FCGROUP 'fc1'
        edit FCNUM 'fcnum'
        task nino_sst_fr
        task nino_atmos_fr
          trigger nino_sst_fr eq complete
        task tcyc_fr
          trigger nino_atmos_fr eq complete
      endfamily
      family fc2
        trigger ( /seas_fr3_back/main:YMD gt /seas_fr3_back/lag:YMD) or ( /seas_fr3_back/main:YMD eq /seas_fr3_back/lag:YMD and /seas_fr3_back/main/fc/fc2 eq complete)
        edit FCGROUP 'fc2'
        edit FCNUM 'fcnum'
        task nino_sst_fr
        task nino_atmos_fr
          trigger nino_sst_fr eq complete
        task tcyc_fr
          trigger nino_atmos_fr eq complete
      endfamily
      family fc3
        trigger ( /seas_fr3_back/main:YMD gt /seas_fr3_back/lag:YMD) or ( /seas_fr3_back/main:YMD eq /seas_fr3_back/lag:YMD and /seas_fr3_back/main/fc/fc3 eq complete)
        edit FCGROUP 'fc3'
        edit FCNUM 'fcnum'
        task nino_sst_fr
        task nino_atmos_fr
          trigger nino_sst_fr eq complete
        task tcyc_fr
          trigger nino_atmos_fr eq complete
      endfamily
      family fc4
        trigger ( /seas_fr3_back/main:YMD gt /seas_fr3_back/lag:YMD) or ( /seas_fr3_back/main:YMD eq /seas_fr3_back/lag:YMD and /seas_fr3_back/main/fc/fc4 eq complete)
        edit FCGROUP 'fc4'
        edit FCNUM 'fcnum'
        task nino_sst_fr
        task nino_atmos_fr
          trigger nino_sst_fr eq complete
        task tcyc_fr
          trigger nino_atmos_fr eq complete
      endfamily
      family fc5
        trigger ( /seas_fr3_back/main:YMD gt /seas_fr3_back/lag:YMD) or ( /seas_fr3_back/main:YMD eq /seas_fr3_back/lag:YMD and /seas_fr3_back/main/fc/fc5 eq complete)
        edit FCGROUP 'fc5'
        edit FCNUM 'fcnum'
        task nino_sst_fr
        task nino_atmos_fr
          trigger nino_sst_fr eq complete
        task tcyc_fr
          trigger nino_atmos_fr eq complete
      endfamily
      family fc6
        trigger ( /seas_fr3_back/main:YMD gt /seas_fr3_back/lag:YMD) or ( /seas_fr3_back/main:YMD eq /seas_fr3_back/lag:YMD and /seas_fr3_back/main/fc/fc6 eq complete)
        edit FCGROUP 'fc6'
        edit FCNUM 'fcnum'
        task nino_sst_fr
        task nino_atmos_fr
          trigger nino_sst_fr eq complete
        task tcyc_fr
          trigger nino_atmos_fr eq complete
      endfamily
      family fc7
        trigger ( /seas_fr3_back/main:YMD gt /seas_fr3_back/lag:YMD) or ( /seas_fr3_back/main:YMD eq /seas_fr3_back/lag:YMD and /seas_fr3_back/main/fc/fc7 eq complete)
        edit FCGROUP 'fc7'
        edit FCNUM 'fcnum'
        task nino_sst_fr
        task nino_atmos_fr
          trigger nino_sst_fr eq complete
        task tcyc_fr
          trigger nino_atmos_fr eq complete
      endfamily
      family fc8
        trigger ( /seas_fr3_back/main:YMD gt /seas_fr3_back/lag:YMD) or ( /seas_fr3_back/main:YMD eq /seas_fr3_back/lag:YMD and /seas_fr3_back/main/fc/fc8 eq complete)
        edit FCGROUP 'fc8'
        edit FCNUM 'fcnum'
        task nino_sst_fr
        task nino_atmos_fr
          trigger nino_sst_fr eq complete
        task tcyc_fr
          trigger nino_atmos_fr eq complete
      endfamily
      family fc9
        trigger ( /seas_fr3_back/main:YMD gt /seas_fr3_back/lag:YMD) or ( /seas_fr3_back/main:YMD eq /seas_fr3_back/lag:YMD and /seas_fr3_back/main/fc/fc9 eq complete)
        edit FCGROUP 'fc9'
        edit FCNUM 'fcnum'
        task nino_sst_fr
        task nino_atmos_fr
          trigger nino_sst_fr eq complete
        task tcyc_fr
          trigger nino_atmos_fr eq complete
      endfamily
      family fc10
        trigger ( /seas_fr3_back/main:YMD gt /seas_fr3_back/lag:YMD) or ( /seas_fr3_back/main:YMD eq /seas_fr3_back/lag:YMD and /seas_fr3_back/main/fc/fc10 eq complete)
        edit FCGROUP 'fc10'
        edit FCNUM 'fcnum'
        task nino_sst_fr
        task nino_atmos_fr
          trigger nino_sst_fr eq complete
        task tcyc_fr
          trigger nino_atmos_fr eq complete
      endfamily
    endfamily
    family clean
      trigger archive eq complete and process eq complete
      task postclean_fdb
      task clean_sc
      task clean_ws
        edit QUEUE 'emos'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID%'
        edit LOGDIR '/vol/emos/output'
        edit SMSOUT '/vol/emos/output'
        inlimit /limits:swarm
        inlimit /seas_fr3_back/limits:ws
    endfamily
    task logfilestore
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID%'
      edit LOGDIR '/vol/emos/output'
      edit SMSOUT '/vol/emos/output'
      edit KEEPLOGS 'no'
      edit LOGSTOECFS 'no'
      edit MOVE_TO_ECFS 'yes'
    task logfiles
      trigger archive eq complete and process eq complete and clean eq complete and logfilestore eq complete
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID%'
      edit LOGDIR '/vol/emos/output'
      edit SMSOUT '/vol/emos/output'
      edit MOVE_TO_ECFS 'no'
  endfamily
endsuite
suite cweb
  edit USER 'emos'
  edit SCHOST 'c1a'
  edit SCHOST_BKUP 'c1b'
  edit STHOST '/s1a'
  edit STHOST_BKUP '/s1b'
  edit WSHOST 'linux_cluster'
  edit WSLOGDIR '/vol/emos_nc/output'
  edit SCLOGDIR '/s1a/emos_esuite/emos_data/9001/log'
  edit DATEMASK '*.*.*'
  edit QUEUE_EPILOG 'ns'
  edit SMSLOGHOST 'c1a'
  edit SMSLOGPORT '9316'
  edit ACCOUNT 'oocweb'
  edit VERSION '9001'
  edit EPSVERSION '0001'
  edit RFVERSION '0001'
  edit SMSTRIES '1'
  edit DELTA_DAY '0'
  edit EMOS_BASE '00'
  edit FIRST_DAY '0'
  edit USE_YMD 'true'
  edit SMSHOME '/vol/emos_nc/output'
  edit SMSFILES '/home/ma/emos/def/cweb'
  edit SMSINCLUDE '/home/ma/emos/def/cweb/include'
  edit FSFAMILY '/'
  edit XBINS '/home/ma/emos/bin'
  edit STREAM 'undef'
  edit EPSTYPE 'undef'
  edit MEMBER '0'
  edit EMOS_TIME_STEP_H '00'
  edit SUITE_START '20091001'
  edit LOW_RESOLUTION '1.5'
  edit SLEEPTIME '1'
  clock hybrid 
  family limits
    limit retrieve 10
    limit purge 1
  endfamily
  family o
    repeat date YMD 20091001 20121212 1
    edit STREAM 'oper'
    family 00
      trigger ( /cweb/o:YMD lt /o/main:YMD) or ( /cweb/o:YMD eq /o/main:YMD and /o/main/00/fc eq complete)
      edit QUEUE 'ns'
      edit QUEUE_EPILOG 'ns'
      edit SMSCMD 'smssubmit %USER% %SCHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %SCHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %SCHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/s1a/emos_esuite/emos_data/9001/log'
      edit SMSOUT '/s1a/emos_esuite/emos_data/9001/log'
      edit DELTA_DAY '1'
      edit EMOS_BASE '00'
      edit SMSFILES '/home/ma/emos/def/cweb/cweb'
      inlimit /limits:c1a
      task purge
        inlimit /cweb/limits:purge
        label info ""
      family feed
        trigger purge eq complete
        inlimit /cweb/limits:retrieve
        task an_sfc
          label info ""
          label grid ""
        task an_pl
          label info ""
          label grid ""
        task an_ml
          label info ""
          label grid ""
        task fc_sfc
          label info ""
          label grid ""
        task fc_pl
          label info ""
          label grid ""
        task fc_ml
          label info ""
          label grid ""
        task fc_wave
          label info ""
          label grid ""
        task sim
          trigger ( /o/lag:YMD gt /cweb/o:YMD) or ( /o/lag:YMD eq /cweb/o:YMD and /o/lag/00/satim eq complete)
          label info ""
          label grid ""
      endfamily
      task check
        trigger feed eq complete
        label info ""
      task ready
        trigger check eq complete
        label info ""
    endfamily
    family 12
      trigger ( /cweb/o:YMD lt /o/main:YMD) or ( /cweb/o:YMD eq /o/main:YMD and /o/main/12/fc eq complete)
      edit QUEUE 'ns'
      edit QUEUE_EPILOG 'ns'
      edit SMSCMD 'smssubmit %USER% %SCHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %SCHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %SCHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/s1a/emos_esuite/emos_data/9001/log'
      edit SMSOUT '/s1a/emos_esuite/emos_data/9001/log'
      edit DELTA_DAY '0'
      edit EMOS_BASE '12'
      edit SMSFILES '/home/ma/emos/def/cweb/cweb'
      inlimit /limits:c1a
      task purge
        inlimit /cweb/limits:purge
        label info ""
      family feed
        trigger purge eq complete
        inlimit /cweb/limits:retrieve
        task an_sfc
          label info ""
          label grid ""
        task an_pl
          label info ""
          label grid ""
        task an_ml
          label info ""
          label grid ""
        task fc_sfc
          label info ""
          label grid ""
        task fc_pl
          label info ""
          label grid ""
        task fc_ml
          label info ""
          label grid ""
        task fc_wave
          label info ""
          label grid ""
        task sim
          trigger ( /o/lag:YMD gt /cweb/o:YMD) or ( /o/lag:YMD eq /cweb/o:YMD and /o/lag/12/satim eq complete)
          label info ""
          label grid ""
      endfamily
      task check
        trigger feed eq complete
        label info ""
      task ready
        trigger check eq complete
        label info ""
    endfamily
    task logfilestore
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit SMSFILES '/home/ma/emos/def/cweb'
      edit LOGTASK 'o'
      edit KEEPLOGS 'no'
      edit LOGSTOECFS 'no'
      edit MOVE_TO_ECFS 'yes'
    task logfiles
      late -s +00:05 -c +00:15
      trigger ./00 eq complete and ./12 eq complete and logfilestore eq complete
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit SMSFILES '/home/ma/emos/def/cweb'
      edit LOGTASK 'o'
      edit MOVE_TO_ECFS 'no'
  endfamily
  family mc
    repeat date YMD 20091001 20121212 1
    edit STREAM 'enfo'
    family 00
      trigger ( /cweb/mc:YMD lt /mc/main:YMD) or ( /cweb/mc:YMD eq /mc/main:YMD and /mc/main/00/legB/postproc eq complete)
      edit QUEUE 'ns'
      edit QUEUE_EPILOG 'ns'
      edit SMSCMD 'smssubmit %USER% %SCHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %SCHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %SCHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/s1a/emos_esuite/emos_data/9001/log'
      edit SMSOUT '/s1a/emos_esuite/emos_data/9001/log'
      edit DELTA_DAY '1'
      edit EMOS_BASE '00'
      edit SMSFILES '/home/ma/emos/def/cweb/cweb'
      inlimit /limits:c1a
      task purge
        inlimit /cweb/limits:purge
        label info ""
      family feed
        trigger purge eq complete
        inlimit /cweb/limits:retrieve
        family leg_A
          edit LEG 'A'
          task cf_sfc
            label info ""
            label grid ""
          task cf_pl
            label info ""
            label grid ""
          task cf_wave_sfc
            label info ""
            label grid ""
          task pf_sfc
            label info ""
            label grid ""
          task pf_pl
            label info ""
            label grid ""
          task pf_wave_sfc
            label info ""
            label grid ""
          task efi_sfc
            label info ""
            label grid ""
        endfamily
        family leg_B
          edit LEG 'B'
          task cf_sfc
            label info ""
            label grid ""
          task cf_pl
            label info ""
            label grid ""
          task cf_wave_sfc
            label info ""
            label grid ""
          task pf_sfc
            label info ""
            label grid ""
          task pf_pl
            label info ""
            label grid ""
          task pf_wave_sfc
            label info ""
            label grid ""
        endfamily
        family leg_C
          edit LEG 'C'
          task cf_sfc
            label info ""
            label grid ""
          task cf_pl
            label info ""
            label grid ""
          task cf_wave_sfc
            label info ""
            label grid ""
          task pf_sfc
            label info ""
            label grid ""
          task pf_pl
            label info ""
            label grid ""
          task pf_wave_sfc
            label info ""
            label grid ""
        endfamily
      endfamily
      task check
        trigger feed eq complete
        label info ""
      task ready
        trigger check eq complete
        label info ""
    endfamily
    family 12
      trigger ( /cweb/mc:YMD lt /mc/main:YMD) or ( /cweb/mc:YMD eq /mc/main:YMD and /mc/main/12/legB/postproc eq complete)
      edit QUEUE 'ns'
      edit QUEUE_EPILOG 'ns'
      edit SMSCMD 'smssubmit %USER% %SCHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %SCHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %SCHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/s1a/emos_esuite/emos_data/9001/log'
      edit SMSOUT '/s1a/emos_esuite/emos_data/9001/log'
      edit DELTA_DAY '0'
      edit EMOS_BASE '12'
      edit SMSFILES '/home/ma/emos/def/cweb/cweb'
      inlimit /limits:c1a
      task purge
        inlimit /cweb/limits:purge
        label info ""
      family feed
        trigger purge eq complete
        inlimit /cweb/limits:retrieve
        family leg_A
          edit LEG 'A'
          task cf_sfc
            label info ""
            label grid ""
          task cf_pl
            label info ""
            label grid ""
          task cf_wave_sfc
            label info ""
            label grid ""
          task pf_sfc
            label info ""
            label grid ""
          task pf_pl
            label info ""
            label grid ""
          task pf_wave_sfc
            label info ""
            label grid ""
          task efi_sfc
            label info ""
            label grid ""
        endfamily
        family leg_B
          edit LEG 'B'
          task cf_sfc
            label info ""
            label grid ""
          task cf_pl
            label info ""
            label grid ""
          task cf_wave_sfc
            label info ""
            label grid ""
          task pf_sfc
            label info ""
            label grid ""
          task pf_pl
            label info ""
            label grid ""
          task pf_wave_sfc
            label info ""
            label grid ""
        endfamily
        family leg_C
          edit LEG 'C'
          task cf_sfc
            label info ""
            label grid ""
          task cf_pl
            label info ""
            label grid ""
          task cf_wave_sfc
            label info ""
            label grid ""
          task pf_sfc
            label info ""
            label grid ""
          task pf_pl
            label info ""
            label grid ""
          task pf_wave_sfc
            label info ""
            label grid ""
        endfamily
      endfamily
      task check
        trigger feed eq complete
        label info ""
      task ready
        trigger check eq complete
        label info ""
    endfamily
    task logfilestore
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit SMSFILES '/home/ma/emos/def/cweb'
      edit LOGTASK 'mc'
      edit KEEPLOGS 'no'
      edit LOGSTOECFS 'no'
      edit MOVE_TO_ECFS 'yes'
    task logfiles
      late -s +00:05 -c +00:15
      trigger ./00 eq complete and ./12 eq complete and logfilestore eq complete
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit SMSFILES '/home/ma/emos/def/cweb'
      edit LOGTASK 'mc'
      edit MOVE_TO_ECFS 'no'
  endfamily
endsuite
suite emc_33r1
  edit USER 'emos'
  edit SCHOST 'c1a'
  edit SCHOST_BKUP 'hpcf'
  edit WSHOST 'linux_cluster'
  edit WSLOGDIR '/vol/emos_nc/output'
  edit SCLOGDIR '/emos_esuite/emos_data/0039/log'
  edit USE_YMD 'true'
  edit YMD '20080203'
  edit DATEMASK '*.*.*'
  edit VERSION '0039'
  edit INIEXPVER '0039'
  edit TCINIEXPVER '0039'
  edit FIRST_DAY '1'
  edit SMSTRIES '2'
  edit DELTA_DAY '0'
  edit EMOS_BASE '12'
  edit EMOS_TIME_STEP_H '0000'
  edit FSFAMILY 'mc'
  edit IMPL_DATE '2012123112'
  edit EVERSION '0001'
  edit USER_PRIORITY '70'
  edit ININPES '16'
  edit FCNPES '16'
  edit PGNPES '32'
  edit ANNPES '4'
  edit REINITIALIZE 'false'
  edit SMSHOME '/vol/emos_nc/output'
  edit SMSFILES '/home/ma/emos/def/emc_33r1'
  edit SMSINCLUDE '/home/ma/emos/def/emc_33r1/include'
  edit ACCOUNT 'oeenfo'
  edit STREAM 'EF'
  edit EMOS_STREAM 'EF'
  edit ENSEMBLES '50'
  edit ENS_SIZE '50'
  edit MARS_DB 'marsod'
  edit EPSTYPE 'fc'
  edit MEMBER '0'
  edit INILEVELS '91'
  edit INIPERIOD_4D '6'
  edit IS_REAL_TIME 'is_real_time'
  edit RUN_3HOURS_PP '1'
  edit RUNHINDCAST '0'
  edit ERA40END '2002073100'
  edit NH_YEARS '18'
  edit NH_WEEKS '2'
  edit DELTA_DAY_M '-15'
  edit DELTA_DAY_P '25'
  edit FCGTYPE 'l_2'
  edit EPSFCGTYPE 'l_2'
  edit LUSE_RSTRHBIAS 'true'
  edit FCGROUP 'fc0'
  edit SMSLOGHOST 'c1a'
  edit SMSLOGPORT '9316'
  edit OCNPES_OC '4'
  edit OCNPES_IFS '28'
  edit OD_PROJ 'od'
  edit SMSURLBASE 'http://datasvc.ecmwf.int/twiki/bin/view/Main'
  edit SMSURL 'CurrentAndOldESuites'
  edit SUITE_START '20080204'
  edit PRODUCE_EPS_ML '0'
  edit QUEUE 'ns'
  edit QUEUE_EPILOG 'ns'
  edit SMSCMD 'smssubmit.35r3 %USER% %SCHOST% %SMSJOB% %SMSJOBOUT%'
  edit SMSKILL 'smskill %USER% %SCHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
  edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %SCHOST% %SMSRID% %SMSJOB%'
  edit LOGDIR '/s1a/emos_esuite/emos_data/0039/log'
  edit SMSOUT '/s1a/emos_esuite/emos_data/0039/log'
  edit STHOST '/s1a'
  edit COURTESY '0'
  edit SLEEPTIME '1'
  inlimit /limits:hpce
  clock hybrid 
  family efi_curop
    edit SMSFILES '/home/ma/emos/def/o/mc/efi_hdc'
    edit SMSURLBASE 'http://diss.ecmwf.int/do/monitoring/summary'
    edit SMSURL 'GENFO/00'
    limit ext 30
    inlimit ext
    family hind
      repeat date YMD 20080403 20121212 1
      edit EMOS_BASE '00'
      family op
        family stage
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit.legc %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit ACC '24'
          edit EFI_STEP1 '00'
          edit EFI_STEP2 '24'
          edit RETRIEVE 'stage'
          inlimit /limits:linux_cluster
          family all
            edit RF_NAME 'all'
            task prep
          endfamily
        endfamily
        family d0105_24
          edit EFI_STEP1 '000 012 024 036 048 060 072 084 096'
          edit EFI_STEP2 '024 036 048 060 072 084 096 108 120'
          edit ACC '24'
          family 2t
            edit RF_NAME '2t'
            task prep
          endfamily
          family 10ff
            edit RF_NAME '10ff'
            task prep
          endfamily
          family 10fg
            edit RF_NAME '10fg'
            task prep
          endfamily
          family tp
            edit RF_NAME 'tp'
            task prep
          endfamily
        endfamily
        family d0610_24
          edit EFI_STEP1 '108 120 132 144 168 192 216'
          edit EFI_STEP2 '132 144 156 168 192 216 240'
          edit ACC '24'
          family 2t
            edit RF_NAME '2t'
            task prep
          endfamily
          family 10ff
            edit RF_NAME '10ff'
            task prep
          endfamily
          family 10fg
            edit RF_NAME '10fg'
            task prep
          endfamily
          family tp
            edit RF_NAME 'tp'
            task prep
          endfamily
        endfamily
      endfamily
      family nop
        family stage
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit.legc %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit ACC '24'
          edit EFI_STEP1 '00'
          edit EFI_STEP2 '24'
          edit RETRIEVE 'stage'
          inlimit /limits:linux_cluster
          family all
            edit RF_NAME 'all'
            task prep
          endfamily
        endfamily
        family d0105_24
          edit EFI_STEP1 '000 012 024 036 048 060 072 084 096'
          edit EFI_STEP2 '024 036 048 060 072 084 096 108 120'
          edit ACC '24'
          family 2tmin
            edit RF_NAME '2tmin'
            task prep
          endfamily
          family 2tmax
            edit RF_NAME '2tmax'
            task prep
          endfamily
          family 10dd
            edit RF_NAME '10dd'
            task prep
          endfamily
          family ts
            edit RF_NAME 'ts'
            task prep
          endfamily
          family tcc
            edit RF_NAME 'tcc'
            task prep
          endfamily
          family hsttmax
            edit RF_NAME 'hsttmax'
            task prep
          endfamily
        endfamily
        family d0610_24
          edit EFI_STEP1 '108 120 132 144 168 192 216'
          edit EFI_STEP2 '132 144 156 168 192 216 240'
          edit ACC '24'
          family 2tmin
            edit RF_NAME '2tmin'
            task prep
          endfamily
          family 2tmax
            edit RF_NAME '2tmax'
            task prep
          endfamily
          family 10dd
            edit RF_NAME '10dd'
            task prep
          endfamily
          family ts
            edit RF_NAME 'ts'
            task prep
          endfamily
          family tcc
            edit RF_NAME 'tcc'
            task prep
          endfamily
          family hsttmax
            edit RF_NAME 'hsttmax'
            task prep
          endfamily
        endfamily
        family d1115_24
          edit EFI_STEP1 '240 264 288 312 336'
          edit EFI_STEP2 '264 288 312 336 360'
          edit ACC '24'
          family 2t
            edit RF_NAME '2t'
            task prep
          endfamily
          family 10ff
            edit RF_NAME '10ff'
            task prep
          endfamily
          family 10fg
            edit RF_NAME '10fg'
            task prep
          endfamily
          family tp
            edit RF_NAME 'tp'
            task prep
          endfamily
          family 2tmin
            edit RF_NAME '2tmin'
            task prep
          endfamily
          family 2tmax
            edit RF_NAME '2tmax'
            task prep
          endfamily
          family 10dd
            edit RF_NAME '10dd'
            task prep
          endfamily
          family ts
            edit RF_NAME 'ts'
            task prep
          endfamily
          family tcc
            edit RF_NAME 'tcc'
            task prep
          endfamily
          family hsttmax
            edit RF_NAME 'hsttmax'
            task prep
          endfamily
        endfamily
        family d0102_06
          edit EFI_STEP1 '006 012 018 024 030 036 042 048'
          edit EFI_STEP2 '006 012 018 024 030 036 042 048'
          edit ACC '0'
          family 2t
            edit RF_NAME '2t'
            task prep
          endfamily
          family 10ff
            edit RF_NAME '10ff'
            task prep
          endfamily
          family 10fg
            edit RF_NAME '10fg'
            task prep
          endfamily
          family tp
            edit RF_NAME 'tp'
            task prep
          endfamily
          family 2tmin
            edit RF_NAME '2tmin'
            task prep
          endfamily
          family 2tmax
            edit RF_NAME '2tmax'
            task prep
          endfamily
          family 10dd
            edit RF_NAME '10dd'
            task prep
          endfamily
          family ts
            edit RF_NAME 'ts'
            task prep
          endfamily
          family tcc
            edit RF_NAME 'tcc'
            task prep
          endfamily
          family hsttmax
            edit RF_NAME 'hsttmax'
            task prep
          endfamily
        endfamily
        family d0304_06
          edit EFI_STEP1 '054 060 066 072 078 084 090 096'
          edit EFI_STEP2 '054 060 066 072 078 084 090 096'
          edit ACC '0'
          family 2t
            edit RF_NAME '2t'
            task prep
          endfamily
          family 10ff
            edit RF_NAME '10ff'
            task prep
          endfamily
          family 10fg
            edit RF_NAME '10fg'
            task prep
          endfamily
          family tp
            edit RF_NAME 'tp'
            task prep
          endfamily
          family 2tmin
            edit RF_NAME '2tmin'
            task prep
          endfamily
          family 2tmax
            edit RF_NAME '2tmax'
            task prep
          endfamily
          family 10dd
            edit RF_NAME '10dd'
            task prep
          endfamily
          family ts
            edit RF_NAME 'ts'
            task prep
          endfamily
          family tcc
            edit RF_NAME 'tcc'
            task prep
          endfamily
          family hsttmax
            edit RF_NAME 'hsttmax'
            task prep
          endfamily
        endfamily
        family d0506_06
          edit EFI_STEP1 '102 108 114 120 126 132 138 144'
          edit EFI_STEP2 '102 108 114 120 126 132 138 144'
          edit ACC '0'
          family 2t
            edit RF_NAME '2t'
            task prep
          endfamily
          family 10ff
            edit RF_NAME '10ff'
            task prep
          endfamily
          family 10fg
            edit RF_NAME '10fg'
            task prep
          endfamily
          family tp
            edit RF_NAME 'tp'
            task prep
          endfamily
          family 2tmin
            edit RF_NAME '2tmin'
            task prep
          endfamily
          family 2tmax
            edit RF_NAME '2tmax'
            task prep
          endfamily
          family 10dd
            edit RF_NAME '10dd'
            task prep
          endfamily
          family ts
            edit RF_NAME 'ts'
            task prep
          endfamily
          family tcc
            edit RF_NAME 'tcc'
            task prep
          endfamily
          family hsttmax
            edit RF_NAME 'hsttmax'
            task prep
          endfamily
        endfamily
        family d0708_06
          edit EFI_STEP1 '150 156 162 168 174 180 186 192'
          edit EFI_STEP2 '150 156 162 168 174 180 186 192'
          edit ACC '0'
          family 2t
            edit RF_NAME '2t'
            task prep
          endfamily
          family 10ff
            edit RF_NAME '10ff'
            task prep
          endfamily
          family 10fg
            edit RF_NAME '10fg'
            task prep
          endfamily
          family tp
            edit RF_NAME 'tp'
            task prep
          endfamily
          family 2tmin
            edit RF_NAME '2tmin'
            task prep
          endfamily
          family 2tmax
            edit RF_NAME '2tmax'
            task prep
          endfamily
          family 10dd
            edit RF_NAME '10dd'
            task prep
          endfamily
          family ts
            edit RF_NAME 'ts'
            task prep
          endfamily
          family tcc
            edit RF_NAME 'tcc'
            task prep
          endfamily
          family hsttmax
            edit RF_NAME 'hsttmax'
            task prep
          endfamily
        endfamily
        family d0910_06
          edit EFI_STEP1 '198 204 210 216 222 228 234 240'
          edit EFI_STEP2 '198 204 210 216 222 228 234 240'
          edit ACC '0'
          family 2t
            edit RF_NAME '2t'
            task prep
          endfamily
          family 10ff
            edit RF_NAME '10ff'
            task prep
          endfamily
          family 10fg
            edit RF_NAME '10fg'
            task prep
          endfamily
          family tp
            edit RF_NAME 'tp'
            task prep
          endfamily
          family 2tmin
            edit RF_NAME '2tmin'
            task prep
          endfamily
          family 2tmax
            edit RF_NAME '2tmax'
            task prep
          endfamily
          family 10dd
            edit RF_NAME '10dd'
            task prep
          endfamily
          family ts
            edit RF_NAME 'ts'
            task prep
          endfamily
          family tcc
            edit RF_NAME 'tcc'
            task prep
          endfamily
          family hsttmax
            edit RF_NAME 'hsttmax'
            task prep
          endfamily
        endfamily
      endfamily
    endfamily
    family clim
      trigger ./clim:YMD lt ./hind:YMD
      repeat date YMD 20080403 20121212 1
      edit EMOS_BASE '00'
      family op
        family steps24
          family 000_024
            edit ACC '24'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '024'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 012_036
            edit ACC '24'
            edit EFI_STEP1 '012'
            edit EFI_STEP2 '036'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 024_048
            edit ACC '24'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '048'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 036_060
            edit ACC '24'
            edit EFI_STEP1 '036'
            edit EFI_STEP2 '060'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 048_072
            edit ACC '24'
            edit EFI_STEP1 '048'
            edit EFI_STEP2 '072'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 060_084
            edit ACC '24'
            edit EFI_STEP1 '060'
            edit EFI_STEP2 '084'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 072_096
            edit ACC '24'
            edit EFI_STEP1 '072'
            edit EFI_STEP2 '096'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 084_108
            edit ACC '24'
            edit EFI_STEP1 '084'
            edit EFI_STEP2 '108'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 096_120
            edit ACC '24'
            edit EFI_STEP1 '096'
            edit EFI_STEP2 '120'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 108_132
            edit ACC '24'
            edit EFI_STEP1 '108'
            edit EFI_STEP2 '132'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 120_144
            edit ACC '24'
            edit EFI_STEP1 '120'
            edit EFI_STEP2 '144'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
        endfamily
        family 000_120
          edit ACC '120'
          edit EFI_STEP1 '000'
          edit EFI_STEP2 '120'
          family tp
            edit RF_NAME 'tp'
            task prep_sort
          endfamily
        endfamily
        family 024_144
          edit ACC '120'
          edit EFI_STEP1 '024'
          edit EFI_STEP2 '144'
          family tp
            edit RF_NAME 'tp'
            task prep_sort
          endfamily
        endfamily
        family 000_240
          edit ACC '240'
          edit EFI_STEP1 '000'
          edit EFI_STEP2 '240'
          family tp
            edit RF_NAME 'tp'
            task prep_sort
          endfamily
        endfamily
      endfamily
      family nop
        family steps24
          family 000_024
            edit ACC '24'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '024'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 012_036
            edit ACC '24'
            edit EFI_STEP1 '012'
            edit EFI_STEP2 '036'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 024_048
            edit ACC '24'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '048'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 036_060
            edit ACC '24'
            edit EFI_STEP1 '036'
            edit EFI_STEP2 '060'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 048_072
            edit ACC '24'
            edit EFI_STEP1 '048'
            edit EFI_STEP2 '072'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 060_084
            edit ACC '24'
            edit EFI_STEP1 '060'
            edit EFI_STEP2 '084'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 072_096
            edit ACC '24'
            edit EFI_STEP1 '072'
            edit EFI_STEP2 '096'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 084_108
            edit ACC '24'
            edit EFI_STEP1 '084'
            edit EFI_STEP2 '108'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 096_120
            edit ACC '24'
            edit EFI_STEP1 '096'
            edit EFI_STEP2 '120'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 108_132
            edit ACC '24'
            edit EFI_STEP1 '108'
            edit EFI_STEP2 '132'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 120_144
            edit ACC '24'
            edit EFI_STEP1 '120'
            edit EFI_STEP2 '144'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 144_168
            edit ACC '24'
            edit EFI_STEP1 '144'
            edit EFI_STEP2 '168'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 168_192
            edit ACC '24'
            edit EFI_STEP1 '168'
            edit EFI_STEP2 '192'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 192_216
            edit ACC '24'
            edit EFI_STEP1 '192'
            edit EFI_STEP2 '216'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 216_240
            edit ACC '24'
            edit EFI_STEP1 '216'
            edit EFI_STEP2 '240'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 240_264
            edit ACC '24'
            edit EFI_STEP1 '240'
            edit EFI_STEP2 '264'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 264_288
            edit ACC '24'
            edit EFI_STEP1 '264'
            edit EFI_STEP2 '288'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 288_312
            edit ACC '24'
            edit EFI_STEP1 '288'
            edit EFI_STEP2 '312'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 312_336
            edit ACC '24'
            edit EFI_STEP1 '312'
            edit EFI_STEP2 '336'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 336_360
            edit ACC '24'
            edit EFI_STEP1 '336'
            edit EFI_STEP2 '360'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
        endfamily
        family steps06
          family 006_006
            edit ACC '0'
            edit EFI_STEP1 '006'
            edit EFI_STEP2 '006'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 012_012
            edit ACC '0'
            edit EFI_STEP1 '012'
            edit EFI_STEP2 '012'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 018_018
            edit ACC '0'
            edit EFI_STEP1 '018'
            edit EFI_STEP2 '018'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 024_024
            edit ACC '0'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '024'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 030_030
            edit ACC '0'
            edit EFI_STEP1 '030'
            edit EFI_STEP2 '030'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 036_036
            edit ACC '0'
            edit EFI_STEP1 '036'
            edit EFI_STEP2 '036'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 042_042
            edit ACC '0'
            edit EFI_STEP1 '042'
            edit EFI_STEP2 '042'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 048_048
            edit ACC '0'
            edit EFI_STEP1 '048'
            edit EFI_STEP2 '048'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 054_054
            edit ACC '0'
            edit EFI_STEP1 '054'
            edit EFI_STEP2 '054'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 060_060
            edit ACC '0'
            edit EFI_STEP1 '060'
            edit EFI_STEP2 '060'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 066_066
            edit ACC '0'
            edit EFI_STEP1 '066'
            edit EFI_STEP2 '066'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 072_072
            edit ACC '0'
            edit EFI_STEP1 '072'
            edit EFI_STEP2 '072'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 078_078
            edit ACC '0'
            edit EFI_STEP1 '078'
            edit EFI_STEP2 '078'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 084_084
            edit ACC '0'
            edit EFI_STEP1 '084'
            edit EFI_STEP2 '084'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 090_090
            edit ACC '0'
            edit EFI_STEP1 '090'
            edit EFI_STEP2 '090'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 096_096
            edit ACC '0'
            edit EFI_STEP1 '096'
            edit EFI_STEP2 '096'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 102_102
            edit ACC '0'
            edit EFI_STEP1 '102'
            edit EFI_STEP2 '102'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 108_108
            edit ACC '0'
            edit EFI_STEP1 '108'
            edit EFI_STEP2 '108'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 114_114
            edit ACC '0'
            edit EFI_STEP1 '114'
            edit EFI_STEP2 '114'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 120_120
            edit ACC '0'
            edit EFI_STEP1 '120'
            edit EFI_STEP2 '120'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 126_126
            edit ACC '0'
            edit EFI_STEP1 '126'
            edit EFI_STEP2 '126'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 132_132
            edit ACC '0'
            edit EFI_STEP1 '132'
            edit EFI_STEP2 '132'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 138_138
            edit ACC '0'
            edit EFI_STEP1 '138'
            edit EFI_STEP2 '138'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 144_144
            edit ACC '0'
            edit EFI_STEP1 '144'
            edit EFI_STEP2 '144'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 150_150
            edit ACC '0'
            edit EFI_STEP1 '150'
            edit EFI_STEP2 '150'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 156_156
            edit ACC '0'
            edit EFI_STEP1 '156'
            edit EFI_STEP2 '156'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 162_162
            edit ACC '0'
            edit EFI_STEP1 '162'
            edit EFI_STEP2 '162'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 168_168
            edit ACC '0'
            edit EFI_STEP1 '168'
            edit EFI_STEP2 '168'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 174_174
            edit ACC '0'
            edit EFI_STEP1 '174'
            edit EFI_STEP2 '174'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 180_180
            edit ACC '0'
            edit EFI_STEP1 '180'
            edit EFI_STEP2 '180'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 186_186
            edit ACC '0'
            edit EFI_STEP1 '186'
            edit EFI_STEP2 '186'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 192_192
            edit ACC '0'
            edit EFI_STEP1 '192'
            edit EFI_STEP2 '192'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 198_198
            edit ACC '0'
            edit EFI_STEP1 '198'
            edit EFI_STEP2 '198'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 204_204
            edit ACC '0'
            edit EFI_STEP1 '204'
            edit EFI_STEP2 '204'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 210_210
            edit ACC '0'
            edit EFI_STEP1 '210'
            edit EFI_STEP2 '210'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 216_216
            edit ACC '0'
            edit EFI_STEP1 '216'
            edit EFI_STEP2 '216'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 222_222
            edit ACC '0'
            edit EFI_STEP1 '222'
            edit EFI_STEP2 '222'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 228_228
            edit ACC '0'
            edit EFI_STEP1 '228'
            edit EFI_STEP2 '228'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 234_234
            edit ACC '0'
            edit EFI_STEP1 '234'
            edit EFI_STEP2 '234'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 240_240
            edit ACC '0'
            edit EFI_STEP1 '240'
            edit EFI_STEP2 '240'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
        endfamily
        family 120_168
          trigger ./steps24 eq complete
          edit ACC '48'
          edit EFI_STEP1 '120'
          edit EFI_STEP2 '168'
          family 2t
            edit RF_NAME '2t'
            task prep_sort
          endfamily
          family 10ff
            edit RF_NAME '10ff'
            task prep_sort
          endfamily
          family 10fg
            edit RF_NAME '10fg'
            task prep_sort
          endfamily
          family tp
            edit RF_NAME 'tp'
            task prep_sort
          endfamily
          family 2tmin
            edit RF_NAME '2tmin'
            task prep_sort
          endfamily
          family 2tmax
            edit RF_NAME '2tmax'
            task prep_sort
          endfamily
          family 10dd
            edit RF_NAME '10dd'
            task prep_sort
          endfamily
          family ts
            edit RF_NAME 'ts'
            task prep_sort
          endfamily
          family tcc
            edit RF_NAME 'tcc'
            task prep_sort
          endfamily
          family hsttmax
            edit RF_NAME 'hsttmax'
            task prep_sort
          endfamily
        endfamily
        family 168_240
          trigger ./steps24 eq complete
          edit ACC '72'
          edit EFI_STEP1 '168'
          edit EFI_STEP2 '240'
          family 2t
            edit RF_NAME '2t'
            task prep_sort
          endfamily
          family 10ff
            edit RF_NAME '10ff'
            task prep_sort
          endfamily
          family 10fg
            edit RF_NAME '10fg'
            task prep_sort
          endfamily
          family tp
            edit RF_NAME 'tp'
            task prep_sort
          endfamily
          family 2tmin
            edit RF_NAME '2tmin'
            task prep_sort
          endfamily
          family 2tmax
            edit RF_NAME '2tmax'
            task prep_sort
          endfamily
          family 10dd
            edit RF_NAME '10dd'
            task prep_sort
          endfamily
          family ts
            edit RF_NAME 'ts'
            task prep_sort
          endfamily
          family tcc
            edit RF_NAME 'tcc'
            task prep_sort
          endfamily
          family hsttmax
            edit RF_NAME 'hsttmax'
            task prep_sort
          endfamily
        endfamily
        family 000_168
          trigger ./steps24 eq complete
          edit ACC '168'
          edit EFI_STEP1 '000'
          edit EFI_STEP2 '168'
          family 2t
            edit RF_NAME '2t'
            task prep_sort
          endfamily
          family 10ff
            edit RF_NAME '10ff'
            task prep_sort
          endfamily
          family 10fg
            edit RF_NAME '10fg'
            task prep_sort
          endfamily
          family tp
            edit RF_NAME 'tp'
            task prep_sort
          endfamily
          family 2tmin
            edit RF_NAME '2tmin'
            task prep_sort
          endfamily
          family 2tmax
            edit RF_NAME '2tmax'
            task prep_sort
          endfamily
          family 10dd
            edit RF_NAME '10dd'
            task prep_sort
          endfamily
          family ts
            edit RF_NAME 'ts'
            task prep_sort
          endfamily
          family tcc
            edit RF_NAME 'tcc'
            task prep_sort
          endfamily
          family hsttmax
            edit RF_NAME 'hsttmax'
            task prep_sort
          endfamily
        endfamily
        family 240_336
          trigger ./steps24 eq complete
          edit ACC '96'
          edit EFI_STEP1 '240'
          edit EFI_STEP2 '336'
          family 2t
            edit RF_NAME '2t'
            task prep_sort
          endfamily
          family 10ff
            edit RF_NAME '10ff'
            task prep_sort
          endfamily
          family 10fg
            edit RF_NAME '10fg'
            task prep_sort
          endfamily
          family tp
            edit RF_NAME 'tp'
            task prep_sort
          endfamily
          family 2tmin
            edit RF_NAME '2tmin'
            task prep_sort
          endfamily
          family 2tmax
            edit RF_NAME '2tmax'
            task prep_sort
          endfamily
          family 10dd
            edit RF_NAME '10dd'
            task prep_sort
          endfamily
          family ts
            edit RF_NAME 'ts'
            task prep_sort
          endfamily
          family tcc
            edit RF_NAME 'tcc'
            task prep_sort
          endfamily
          family hsttmax
            edit RF_NAME 'hsttmax'
            task prep_sort
          endfamily
        endfamily
        family 168_336
          trigger ./steps24 eq complete
          edit ACC '168'
          edit EFI_STEP1 '168'
          edit EFI_STEP2 '336'
          family 2t
            edit RF_NAME '2t'
            task prep_sort
          endfamily
          family 10ff
            edit RF_NAME '10ff'
            task prep_sort
          endfamily
          family 10fg
            edit RF_NAME '10fg'
            task prep_sort
          endfamily
          family tp
            edit RF_NAME 'tp'
            task prep_sort
          endfamily
          family 2tmin
            edit RF_NAME '2tmin'
            task prep_sort
          endfamily
          family 2tmax
            edit RF_NAME '2tmax'
            task prep_sort
          endfamily
          family 10dd
            edit RF_NAME '10dd'
            task prep_sort
          endfamily
          family ts
            edit RF_NAME 'ts'
            task prep_sort
          endfamily
          family tcc
            edit RF_NAME 'tcc'
            task prep_sort
          endfamily
          family hsttmax
            edit RF_NAME 'hsttmax'
            task prep_sort
          endfamily
        endfamily
      endfamily
    endfamily
    family mn
      trigger ./mn:YMD lt ./clim:YMD
      repeat date YMD 20080221 20121212 1
      family 12
        edit EMOS_BASE '12'
        family op
          task stage
            edit NC_DAYS '31'
          family 000_120
            trigger ./stage eq complete
            edit ACC '120'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '120'
            family tp
              edit RF_NAME 'tp'
              task extindex
            endfamily
          endfamily
          family 024_144
            trigger ./stage eq complete
            edit ACC '120'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '144'
            family tp
              edit RF_NAME 'tp'
              task extindex
            endfamily
          endfamily
          family 000_240
            trigger ./stage eq complete
            edit ACC '240'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '240'
            family tp
              edit RF_NAME 'tp'
              task extindex
            endfamily
          endfamily
          family 012_036
            trigger ./stage eq complete
            edit ACC '24'
            edit EFI_STEP1 '012'
            edit EFI_STEP2 '036'
            family 2t
              edit RF_NAME '2t'
              task extindex
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task extindex
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task extindex
            endfamily
            family tp
              edit RF_NAME 'tp'
              task extindex
            endfamily
          endfamily
          family 036_060
            trigger ./stage eq complete
            edit ACC '24'
            edit EFI_STEP1 '036'
            edit EFI_STEP2 '060'
            family 2t
              edit RF_NAME '2t'
              task extindex
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task extindex
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task extindex
            endfamily
            family tp
              edit RF_NAME 'tp'
              task extindex
            endfamily
          endfamily
          family 060_084
            trigger ./stage eq complete
            edit ACC '24'
            edit EFI_STEP1 '060'
            edit EFI_STEP2 '084'
            family 2t
              edit RF_NAME '2t'
              task extindex
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task extindex
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task extindex
            endfamily
            family tp
              edit RF_NAME 'tp'
              task extindex
            endfamily
          endfamily
          family 084_108
            trigger ./stage eq complete
            edit ACC '24'
            edit EFI_STEP1 '084'
            edit EFI_STEP2 '108'
            family 2t
              edit RF_NAME '2t'
              task extindex
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task extindex
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task extindex
            endfamily
            family tp
              edit RF_NAME 'tp'
              task extindex
            endfamily
          endfamily
          family 108_132
            trigger ./stage eq complete
            edit ACC '24'
            edit EFI_STEP1 '108'
            edit EFI_STEP2 '132'
            family 2t
              edit RF_NAME '2t'
              task extindex
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task extindex
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task extindex
            endfamily
            family tp
              edit RF_NAME 'tp'
              task extindex
            endfamily
          endfamily
        endfamily
        family nop
          task stage
            edit NC_DAYS '31'
          family 228_324
            trigger ./stage eq complete
            edit ACC '96'
            edit EFI_STEP1 '228'
            edit EFI_STEP2 '324'
            family 2t
              edit RF_NAME '2t'
              task extindex
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task extindex
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task extindex
            endfamily
            family tp
              edit RF_NAME 'tp'
              task extindex
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task extindex
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task extindex
            endfamily
            family ts
              edit RF_NAME 'ts'
              task extindex
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task extindex
            endfamily
          endfamily
          family 168_336
            trigger ./stage eq complete
            edit ACC '168'
            edit EFI_STEP1 '168'
            edit EFI_STEP2 '336'
            family 2t
              edit RF_NAME '2t'
              task extindex
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task extindex
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task extindex
            endfamily
            family tp
              edit RF_NAME 'tp'
              task extindex
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task extindex
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task extindex
            endfamily
            family ts
              edit RF_NAME 'ts'
              task extindex
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task extindex
            endfamily
          endfamily
          family 000_024
            trigger ./stage eq complete
            edit ACC '24'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '024'
            family 2t
              edit RF_NAME '2t'
              task extindex
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task extindex
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task extindex
            endfamily
            family tp
              edit RF_NAME 'tp'
              task extindex
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task extindex
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task extindex
            endfamily
            family ts
              edit RF_NAME 'ts'
              task extindex
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task extindex
            endfamily
          endfamily
          family 024_048
            trigger ./stage eq complete
            edit ACC '24'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '048'
            family 2t
              edit RF_NAME '2t'
              task extindex
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task extindex
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task extindex
            endfamily
            family tp
              edit RF_NAME 'tp'
              task extindex
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task extindex
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task extindex
            endfamily
            family ts
              edit RF_NAME 'ts'
              task extindex
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task extindex
            endfamily
          endfamily
          family 048_072
            trigger ./stage eq complete
            edit ACC '24'
            edit EFI_STEP1 '048'
            edit EFI_STEP2 '072'
            family 2t
              edit RF_NAME '2t'
              task extindex
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task extindex
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task extindex
            endfamily
            family tp
              edit RF_NAME 'tp'
              task extindex
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task extindex
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task extindex
            endfamily
            family ts
              edit RF_NAME 'ts'
              task extindex
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task extindex
            endfamily
          endfamily
          family 072_096
            trigger ./stage eq complete
            edit ACC '24'
            edit EFI_STEP1 '072'
            edit EFI_STEP2 '096'
            family 2t
              edit RF_NAME '2t'
              task extindex
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task extindex
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task extindex
            endfamily
            family tp
              edit RF_NAME 'tp'
              task extindex
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task extindex
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task extindex
            endfamily
            family ts
              edit RF_NAME 'ts'
              task extindex
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task extindex
            endfamily
          endfamily
          family 096_120
            trigger ./stage eq complete
            edit ACC '24'
            edit EFI_STEP1 '096'
            edit EFI_STEP2 '120'
            family 2t
              edit RF_NAME '2t'
              task extindex
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task extindex
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task extindex
            endfamily
            family tp
              edit RF_NAME 'tp'
              task extindex
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task extindex
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task extindex
            endfamily
            family ts
              edit RF_NAME 'ts'
              task extindex
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task extindex
            endfamily
          endfamily
          family 120_144
            trigger ./stage eq complete
            edit ACC '24'
            edit EFI_STEP1 '120'
            edit EFI_STEP2 '144'
            family 2t
              edit RF_NAME '2t'
              task extindex
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task extindex
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task extindex
            endfamily
            family tp
              edit RF_NAME 'tp'
              task extindex
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task extindex
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task extindex
            endfamily
            family ts
              edit RF_NAME 'ts'
              task extindex
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task extindex
            endfamily
          endfamily
          family 108_156
            trigger ./stage eq complete
            edit ACC '48'
            edit EFI_STEP1 '108'
            edit EFI_STEP2 '156'
            family 2t
              edit RF_NAME '2t'
              task extindex
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task extindex
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task extindex
            endfamily
            family tp
              edit RF_NAME 'tp'
              task extindex
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task extindex
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task extindex
            endfamily
            family ts
              edit RF_NAME 'ts'
              task extindex
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task extindex
            endfamily
          endfamily
          family 156_228
            trigger ./stage eq complete
            edit ACC '72'
            edit EFI_STEP1 '156'
            edit EFI_STEP2 '228'
            family 2t
              edit RF_NAME '2t'
              task extindex
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task extindex
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task extindex
            endfamily
            family tp
              edit RF_NAME 'tp'
              task extindex
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task extindex
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task extindex
            endfamily
            family ts
              edit RF_NAME 'ts'
              task extindex
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task extindex
            endfamily
          endfamily
          family 000_168
            trigger ./stage eq complete
            edit ACC '168'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '168'
            family 2t
              edit RF_NAME '2t'
              task extindex
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task extindex
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task extindex
            endfamily
            family tp
              edit RF_NAME 'tp'
              task extindex
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task extindex
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task extindex
            endfamily
            family ts
              edit RF_NAME 'ts'
              task extindex
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task extindex
            endfamily
          endfamily
          family 012_036
            trigger ./stage eq complete
            edit ACC '24'
            edit EFI_STEP1 '012'
            edit EFI_STEP2 '036'
            family 2tmin
              edit RF_NAME '2tmin'
              task extindex
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task extindex
            endfamily
            family ts
              edit RF_NAME 'ts'
              task extindex
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task extindex
            endfamily
          endfamily
          family 036_060
            trigger ./stage eq complete
            edit ACC '24'
            edit EFI_STEP1 '036'
            edit EFI_STEP2 '060'
            family 2tmin
              edit RF_NAME '2tmin'
              task extindex
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task extindex
            endfamily
            family ts
              edit RF_NAME 'ts'
              task extindex
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task extindex
            endfamily
          endfamily
          family 060_084
            trigger ./stage eq complete
            edit ACC '24'
            edit EFI_STEP1 '060'
            edit EFI_STEP2 '084'
            family 2tmin
              edit RF_NAME '2tmin'
              task extindex
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task extindex
            endfamily
            family ts
              edit RF_NAME 'ts'
              task extindex
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task extindex
            endfamily
          endfamily
          family 084_108
            trigger ./stage eq complete
            edit ACC '24'
            edit EFI_STEP1 '084'
            edit EFI_STEP2 '108'
            family 2tmin
              edit RF_NAME '2tmin'
              task extindex
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task extindex
            endfamily
            family ts
              edit RF_NAME 'ts'
              task extindex
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task extindex
            endfamily
          endfamily
          family 108_132
            trigger ./stage eq complete
            edit ACC '24'
            edit EFI_STEP1 '108'
            edit EFI_STEP2 '132'
            family 2tmin
              edit RF_NAME '2tmin'
              task extindex
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task extindex
            endfamily
            family ts
              edit RF_NAME 'ts'
              task extindex
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task extindex
            endfamily
          endfamily
        endfamily
        family arch_op
          trigger op eq complete
          edit SMSFILES '/home/ma/emos/def/emc_33r1/archive'
          task efi2fdb
            edit SMSFILES '/home/ma/emos/def/emc_33r1/mc/efi_hdc'
            event none
          family pf
            trigger ./efi2fdb eq complete
            edit EMOS_TYPE 'pf'
            edit EFI_TYPE 'efi'
            family 2t
              edit RF_NAME '2t'
              task efi
                label info ""
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task efi
                label info ""
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task efi
                label info ""
            endfamily
            family tp
              edit RF_NAME 'tp'
              task efi
                label info ""
            endfamily
          endfamily
          family cf
            trigger ./efi2fdb eq complete
            edit EMOS_TYPE 'cf'
            edit EFI_TYPE 'efic'
            family 2t
              edit RF_NAME '2t'
              task efi
                label info ""
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task efi
                label info ""
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task efi
                label info ""
            endfamily
            family tp
              edit RF_NAME 'tp'
              task efi
                label info ""
            endfamily
          endfamily
        endfamily
        family arch_nop
          trigger nop eq complete
          edit SMSFILES '/home/ma/emos/def/emc_33r1/archive'
          task efi2fdb
            edit SMSFILES '/home/ma/emos/def/emc_33r1/mc/efi_hdc'
            event none
          family pf
            trigger ./efi2fdb eq complete
            edit EMOS_TYPE 'pf'
            edit EFI_TYPE 'efi'
            family 2t
              edit RF_NAME '2t'
              task efi
                label info ""
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task efi
                label info ""
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task efi
                label info ""
            endfamily
            family tp
              edit RF_NAME 'tp'
              task efi
                label info ""
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task efi
                label info ""
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task efi
                label info ""
            endfamily
            family ts
              edit RF_NAME 'ts'
              task efi
                label info ""
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task efi
                label info ""
            endfamily
          endfamily
          family cf
            trigger ./efi2fdb eq complete
            edit EMOS_TYPE 'cf'
            edit EFI_TYPE 'efic'
            family 2t
              edit RF_NAME '2t'
              task efi
                label info ""
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task efi
                label info ""
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task efi
                label info ""
            endfamily
            family tp
              edit RF_NAME 'tp'
              task efi
                label info ""
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task efi
                label info ""
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task efi
                label info ""
            endfamily
            family ts
              edit RF_NAME 'ts'
              task efi
                label info ""
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task efi
                label info ""
            endfamily
          endfamily
        endfamily
      endfamily
      family 00
        edit EMOS_BASE '00'
        edit DELTA_DAY '1'
        family op
          task stage
            edit NC_DAYS '31'
          family 000_120
            trigger ./stage eq complete
            edit ACC '120'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '120'
            family tp
              edit RF_NAME 'tp'
              task extindex
            endfamily
          endfamily
          family 024_144
            trigger ./stage eq complete
            edit ACC '120'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '144'
            family tp
              edit RF_NAME 'tp'
              task extindex
            endfamily
          endfamily
          family 000_240
            trigger ./stage eq complete
            edit ACC '240'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '240'
            family tp
              edit RF_NAME 'tp'
              task extindex
            endfamily
          endfamily
          family 000_024
            trigger ./stage eq complete
            edit ACC '24'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '024'
            family 2t
              edit RF_NAME '2t'
              task extindex
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task extindex
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task extindex
            endfamily
            family tp
              edit RF_NAME 'tp'
              task extindex
            endfamily
          endfamily
          family 024_048
            trigger ./stage eq complete
            edit ACC '24'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '048'
            family 2t
              edit RF_NAME '2t'
              task extindex
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task extindex
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task extindex
            endfamily
            family tp
              edit RF_NAME 'tp'
              task extindex
            endfamily
          endfamily
          family 048_072
            trigger ./stage eq complete
            edit ACC '24'
            edit EFI_STEP1 '048'
            edit EFI_STEP2 '072'
            family 2t
              edit RF_NAME '2t'
              task extindex
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task extindex
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task extindex
            endfamily
            family tp
              edit RF_NAME 'tp'
              task extindex
            endfamily
          endfamily
          family 072_096
            trigger ./stage eq complete
            edit ACC '24'
            edit EFI_STEP1 '072'
            edit EFI_STEP2 '096'
            family 2t
              edit RF_NAME '2t'
              task extindex
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task extindex
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task extindex
            endfamily
            family tp
              edit RF_NAME 'tp'
              task extindex
            endfamily
          endfamily
          family 096_120
            trigger ./stage eq complete
            edit ACC '24'
            edit EFI_STEP1 '096'
            edit EFI_STEP2 '120'
            family 2t
              edit RF_NAME '2t'
              task extindex
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task extindex
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task extindex
            endfamily
            family tp
              edit RF_NAME 'tp'
              task extindex
            endfamily
          endfamily
        endfamily
        family nop
          task stage
            edit NC_DAYS '31'
          family 240_336
            trigger ./stage eq complete
            edit ACC '96'
            edit EFI_STEP1 '240'
            edit EFI_STEP2 '336'
            family 2t
              edit RF_NAME '2t'
              task extindex
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task extindex
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task extindex
            endfamily
            family tp
              edit RF_NAME 'tp'
              task extindex
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task extindex
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task extindex
            endfamily
            family ts
              edit RF_NAME 'ts'
              task extindex
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task extindex
            endfamily
          endfamily
          family 168_336
            trigger ./stage eq complete
            edit ACC '168'
            edit EFI_STEP1 '168'
            edit EFI_STEP2 '336'
            family 2t
              edit RF_NAME '2t'
              task extindex
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task extindex
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task extindex
            endfamily
            family tp
              edit RF_NAME 'tp'
              task extindex
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task extindex
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task extindex
            endfamily
            family ts
              edit RF_NAME 'ts'
              task extindex
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task extindex
            endfamily
          endfamily
          family 012_036
            trigger ./stage eq complete
            edit ACC '24'
            edit EFI_STEP1 '012'
            edit EFI_STEP2 '036'
            family 2t
              edit RF_NAME '2t'
              task extindex
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task extindex
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task extindex
            endfamily
            family tp
              edit RF_NAME 'tp'
              task extindex
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task extindex
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task extindex
            endfamily
            family ts
              edit RF_NAME 'ts'
              task extindex
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task extindex
            endfamily
          endfamily
          family 036_060
            trigger ./stage eq complete
            edit ACC '24'
            edit EFI_STEP1 '036'
            edit EFI_STEP2 '060'
            family 2t
              edit RF_NAME '2t'
              task extindex
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task extindex
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task extindex
            endfamily
            family tp
              edit RF_NAME 'tp'
              task extindex
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task extindex
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task extindex
            endfamily
            family ts
              edit RF_NAME 'ts'
              task extindex
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task extindex
            endfamily
          endfamily
          family 060_084
            trigger ./stage eq complete
            edit ACC '24'
            edit EFI_STEP1 '060'
            edit EFI_STEP2 '084'
            family 2t
              edit RF_NAME '2t'
              task extindex
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task extindex
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task extindex
            endfamily
            family tp
              edit RF_NAME 'tp'
              task extindex
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task extindex
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task extindex
            endfamily
            family ts
              edit RF_NAME 'ts'
              task extindex
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task extindex
            endfamily
          endfamily
          family 084_108
            trigger ./stage eq complete
            edit ACC '24'
            edit EFI_STEP1 '084'
            edit EFI_STEP2 '108'
            family 2t
              edit RF_NAME '2t'
              task extindex
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task extindex
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task extindex
            endfamily
            family tp
              edit RF_NAME 'tp'
              task extindex
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task extindex
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task extindex
            endfamily
            family ts
              edit RF_NAME 'ts'
              task extindex
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task extindex
            endfamily
          endfamily
          family 108_132
            trigger ./stage eq complete
            edit ACC '24'
            edit EFI_STEP1 '108'
            edit EFI_STEP2 '132'
            family 2t
              edit RF_NAME '2t'
              task extindex
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task extindex
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task extindex
            endfamily
            family tp
              edit RF_NAME 'tp'
              task extindex
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task extindex
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task extindex
            endfamily
            family ts
              edit RF_NAME 'ts'
              task extindex
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task extindex
            endfamily
          endfamily
          family 120_144
            trigger ./stage eq complete
            edit ACC '24'
            edit EFI_STEP1 '120'
            edit EFI_STEP2 '144'
            family 2t
              edit RF_NAME '2t'
              task extindex
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task extindex
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task extindex
            endfamily
            family tp
              edit RF_NAME 'tp'
              task extindex
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task extindex
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task extindex
            endfamily
            family ts
              edit RF_NAME 'ts'
              task extindex
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task extindex
            endfamily
          endfamily
          family 120_168
            trigger ./stage eq complete
            edit ACC '48'
            edit EFI_STEP1 '120'
            edit EFI_STEP2 '168'
            family 2t
              edit RF_NAME '2t'
              task extindex
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task extindex
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task extindex
            endfamily
            family tp
              edit RF_NAME 'tp'
              task extindex
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task extindex
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task extindex
            endfamily
            family ts
              edit RF_NAME 'ts'
              task extindex
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task extindex
            endfamily
          endfamily
          family 168_240
            trigger ./stage eq complete
            edit ACC '72'
            edit EFI_STEP1 '168'
            edit EFI_STEP2 '240'
            family 2t
              edit RF_NAME '2t'
              task extindex
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task extindex
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task extindex
            endfamily
            family tp
              edit RF_NAME 'tp'
              task extindex
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task extindex
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task extindex
            endfamily
            family ts
              edit RF_NAME 'ts'
              task extindex
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task extindex
            endfamily
          endfamily
          family 000_168
            trigger ./stage eq complete
            edit ACC '168'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '168'
            family 2t
              edit RF_NAME '2t'
              task extindex
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task extindex
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task extindex
            endfamily
            family tp
              edit RF_NAME 'tp'
              task extindex
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task extindex
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task extindex
            endfamily
            family ts
              edit RF_NAME 'ts'
              task extindex
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task extindex
            endfamily
          endfamily
          family 000_024
            trigger ./stage eq complete
            edit ACC '24'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '024'
            family 2tmin
              edit RF_NAME '2tmin'
              task extindex
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task extindex
            endfamily
            family ts
              edit RF_NAME 'ts'
              task extindex
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task extindex
            endfamily
          endfamily
          family 024_048
            trigger ./stage eq complete
            edit ACC '24'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '048'
            family 2tmin
              edit RF_NAME '2tmin'
              task extindex
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task extindex
            endfamily
            family ts
              edit RF_NAME 'ts'
              task extindex
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task extindex
            endfamily
          endfamily
          family 048_072
            trigger ./stage eq complete
            edit ACC '24'
            edit EFI_STEP1 '048'
            edit EFI_STEP2 '072'
            family 2tmin
              edit RF_NAME '2tmin'
              task extindex
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task extindex
            endfamily
            family ts
              edit RF_NAME 'ts'
              task extindex
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task extindex
            endfamily
          endfamily
          family 072_096
            trigger ./stage eq complete
            edit ACC '24'
            edit EFI_STEP1 '072'
            edit EFI_STEP2 '096'
            family 2tmin
              edit RF_NAME '2tmin'
              task extindex
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task extindex
            endfamily
            family ts
              edit RF_NAME 'ts'
              task extindex
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task extindex
            endfamily
          endfamily
          family 096_120
            trigger ./stage eq complete
            edit ACC '24'
            edit EFI_STEP1 '096'
            edit EFI_STEP2 '120'
            family 2tmin
              edit RF_NAME '2tmin'
              task extindex
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task extindex
            endfamily
            family ts
              edit RF_NAME 'ts'
              task extindex
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task extindex
            endfamily
          endfamily
        endfamily
        family arch_op
          trigger op eq complete
          edit SMSFILES '/home/ma/emos/def/emc_33r1/archive'
          task efi2fdb
            edit SMSFILES '/home/ma/emos/def/emc_33r1/mc/efi_hdc'
            event none
          family pf
            trigger ./efi2fdb eq complete
            edit EMOS_TYPE 'pf'
            edit EFI_TYPE 'efi'
            family 2t
              edit RF_NAME '2t'
              task efi
                label info ""
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task efi
                label info ""
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task efi
                label info ""
            endfamily
            family tp
              edit RF_NAME 'tp'
              task efi
                label info ""
            endfamily
          endfamily
          family cf
            trigger ./efi2fdb eq complete
            edit EMOS_TYPE 'cf'
            edit EFI_TYPE 'efic'
            family 2t
              edit RF_NAME '2t'
              task efi
                label info ""
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task efi
                label info ""
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task efi
                label info ""
            endfamily
            family tp
              edit RF_NAME 'tp'
              task efi
                label info ""
            endfamily
          endfamily
        endfamily
        family arch_nop
          trigger nop eq complete
          edit SMSFILES '/home/ma/emos/def/emc_33r1/archive'
          task efi2fdb
            edit SMSFILES '/home/ma/emos/def/emc_33r1/mc/efi_hdc'
            event none
          family pf
            trigger ./efi2fdb eq complete
            edit EMOS_TYPE 'pf'
            edit EFI_TYPE 'efi'
            family 2t
              edit RF_NAME '2t'
              task efi
                label info ""
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task efi
                label info ""
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task efi
                label info ""
            endfamily
            family tp
              edit RF_NAME 'tp'
              task efi
                label info ""
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task efi
                label info ""
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task efi
                label info ""
            endfamily
            family ts
              edit RF_NAME 'ts'
              task efi
                label info ""
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task efi
                label info ""
            endfamily
          endfamily
          family cf
            trigger ./efi2fdb eq complete
            edit EMOS_TYPE 'cf'
            edit EFI_TYPE 'efic'
            family 2t
              edit RF_NAME '2t'
              task efi
                label info ""
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task efi
                label info ""
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task efi
                label info ""
            endfamily
            family tp
              edit RF_NAME 'tp'
              task efi
                label info ""
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task efi
                label info ""
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task efi
                label info ""
            endfamily
            family ts
              edit RF_NAME 'ts'
              task efi
                label info ""
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task efi
                label info ""
            endfamily
          endfamily
        endfamily
      endfamily
    endfamily
  endfamily
  family efi_next
    edit VERSION '0039'
    edit SMSFILES '/home/ma/emos/def/emc_33r1/mc/efi'
    edit EVERSION '0001'
    edit IMPL_DATE '2012121212'
    edit EMOS_BASE '00'
    edit MARS_DB 'marsod'
    edit EMOS_TYPE 'undef'
    edit EPS_VERSION '0001'
    edit SMSURLBASE 'http://diss.ecmwf.int/do/monitoring/summary'
    edit SMSURL 'GENFO/00'
    edit MARS_NEW '0'
    edit UPDATE_WEB_NOW 'yes'
    edit SCHOST 'c1a'
    edit STHOST '/s1a'
    edit SCHOST_BKUP 'c1b'
    edit SCLOGDIR '/s1a/emos_esuite/emos_data/0039/log'
    edit LOGDIR '/s1a/emos_esuite/emos_data/0039/log'
    edit SMSOUT '/s1a/emos_esuite/emos_data/0039/log'
    edit SMSLOGHOST 'c1a'
    limit ext 35
    inlimit ext
    family mn
      trigger ./mn:YMD lt ./clim:YMD
      repeat date YMD 20080221 20121212 1
      edit EFINENS '50'
      family deps
        edit SMSFILES '/home/ma/emos/def/emc_33r1/ref'
        family 00
          time 09:30
          task dummy
            edit SMSPASS 'stripped'
        endfamily
        family 12
          time 21:30
          task dummy
        endfamily
      endfamily
      family 12
        trigger ./deps/12 eq complete
        edit EMOS_BASE '12'
        family op
          edit EPSLEG '1'
          task stage
          family 024
            family 012_036
              edit ACC '024'
              edit EFI_STEP1 '012'
              edit EFI_STEP2 '036'
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 036_060
              edit ACC '024'
              edit EFI_STEP1 '036'
              edit EFI_STEP2 '060'
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 060_084
              edit ACC '024'
              edit EFI_STEP1 '060'
              edit EFI_STEP2 '084'
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 084_108
              edit ACC '024'
              edit EFI_STEP1 '084'
              edit EFI_STEP2 '108'
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 108_132
              edit ACC '024'
              edit EFI_STEP1 '108'
              edit EFI_STEP2 '132'
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 000_024
              edit ACC '024'
              edit EFI_STEP1 '000'
              edit EFI_STEP2 '024'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
            endfamily
            family 024_048
              edit ACC '024'
              edit EFI_STEP1 '024'
              edit EFI_STEP2 '048'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
            endfamily
            family 048_072
              edit ACC '024'
              edit EFI_STEP1 '048'
              edit EFI_STEP2 '072'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
            endfamily
            family 072_096
              edit ACC '024'
              edit EFI_STEP1 '072'
              edit EFI_STEP2 '096'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
            endfamily
            family 096_120
              edit ACC '024'
              edit EFI_STEP1 '096'
              edit EFI_STEP2 '120'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
            endfamily
            family 120_144
              edit ACC '024'
              edit EFI_STEP1 '120'
              edit EFI_STEP2 '144'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
            endfamily
            family 144_168
              edit ACC '024'
              edit EFI_STEP1 '144'
              edit EFI_STEP2 '168'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
            endfamily
            family 168_192
              edit ACC '024'
              edit EFI_STEP1 '168'
              edit EFI_STEP2 '192'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
            endfamily
            family 192_216
              edit ACC '024'
              edit EFI_STEP1 '192'
              edit EFI_STEP2 '216'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
            endfamily
            family 216_240
              edit ACC '024'
              edit EFI_STEP1 '216'
              edit EFI_STEP2 '240'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
            endfamily
          endfamily
          family 120
            trigger ./024 eq complete
            family 000_120
              edit ACC '120'
              edit EFI_STEP1 '000'
              edit EFI_STEP2 '120'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 024_144
              edit ACC '120'
              edit EFI_STEP1 '024'
              edit EFI_STEP2 '144'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
          endfamily
          family 240
            trigger ./024 eq complete
            family 000_240
              edit ACC '240'
              edit EFI_STEP1 '000'
              edit EFI_STEP2 '240'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
          endfamily
        endfamily
        family nop
          edit EPSLEG '2'
          task stage
          family 024
            family 012_036
              edit ACC '024'
              edit EFI_STEP1 '012'
              edit EFI_STEP2 '036'
              family 2tmin
                edit RF_NAME '2tmin'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 036_060
              edit ACC '024'
              edit EFI_STEP1 '036'
              edit EFI_STEP2 '060'
              family 2tmin
                edit RF_NAME '2tmin'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 060_084
              edit ACC '024'
              edit EFI_STEP1 '060'
              edit EFI_STEP2 '084'
              family 2tmin
                edit RF_NAME '2tmin'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 084_108
              edit ACC '024'
              edit EFI_STEP1 '084'
              edit EFI_STEP2 '108'
              family 2tmin
                edit RF_NAME '2tmin'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 108_132
              edit ACC '024'
              edit EFI_STEP1 '108'
              edit EFI_STEP2 '132'
              family 2tmin
                edit RF_NAME '2tmin'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 132_156
              edit ACC '024'
              edit EFI_STEP1 '132'
              edit EFI_STEP2 '156'
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 156_180
              edit ACC '024'
              edit EFI_STEP1 '156'
              edit EFI_STEP2 '180'
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 000_024
              edit ACC '024'
              edit EFI_STEP1 '000'
              edit EFI_STEP2 '024'
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 024_048
              edit ACC '024'
              edit EFI_STEP1 '024'
              edit EFI_STEP2 '048'
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 048_072
              edit ACC '024'
              edit EFI_STEP1 '048'
              edit EFI_STEP2 '072'
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 072_096
              edit ACC '024'
              edit EFI_STEP1 '072'
              edit EFI_STEP2 '096'
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 096_120
              edit ACC '024'
              edit EFI_STEP1 '096'
              edit EFI_STEP2 '120'
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 120_144
              edit ACC '024'
              edit EFI_STEP1 '120'
              edit EFI_STEP2 '144'
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 144_168
              edit ACC '024'
              edit EFI_STEP1 '144'
              edit EFI_STEP2 '168'
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 180_204
              edit ACC '024'
              edit EFI_STEP1 '180'
              edit EFI_STEP2 '204'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 204_228
              edit ACC '024'
              edit EFI_STEP1 '204'
              edit EFI_STEP2 '228'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 168_192
              edit ACC '24'
              edit EFI_STEP1 '168'
              edit EFI_STEP2 '192'
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 192_216
              edit ACC '24'
              edit EFI_STEP1 '192'
              edit EFI_STEP2 '216'
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 216_240
              edit ACC '24'
              edit EFI_STEP1 '216'
              edit EFI_STEP2 '240'
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 240_264
              edit ACC '24'
              edit EFI_STEP1 '240'
              edit EFI_STEP2 '264'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 264_288
              edit ACC '24'
              edit EFI_STEP1 '264'
              edit EFI_STEP2 '288'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 288_312
              edit ACC '24'
              edit EFI_STEP1 '288'
              edit EFI_STEP2 '312'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 312_336
              edit ACC '24'
              edit EFI_STEP1 '312'
              edit EFI_STEP2 '336'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 336_360
              edit ACC '24'
              edit EFI_STEP1 '336'
              edit EFI_STEP2 '360'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
          endfamily
          family 072
            family 012_084
              edit ACC '072'
              edit EFI_STEP1 '012'
              edit EFI_STEP2 '084'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 036_108
              edit ACC '072'
              edit EFI_STEP1 '036'
              edit EFI_STEP2 '108'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 060_132
              edit ACC '072'
              edit EFI_STEP1 '060'
              edit EFI_STEP2 '132'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 084_156
              edit ACC '072'
              edit EFI_STEP1 '084'
              edit EFI_STEP2 '156'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 108_180
              edit ACC '072'
              edit EFI_STEP1 '108'
              edit EFI_STEP2 '180'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 132_204
              edit ACC '072'
              edit EFI_STEP1 '132'
              edit EFI_STEP2 '204'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 156_228
              edit ACC '072'
              edit EFI_STEP1 '156'
              edit EFI_STEP2 '228'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
          endfamily
          family 120
            family 012_132
              edit ACC '120'
              edit EFI_STEP1 '012'
              edit EFI_STEP2 '132'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 036_156
              edit ACC '120'
              edit EFI_STEP1 '036'
              edit EFI_STEP2 '156'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 060_180
              edit ACC '120'
              edit EFI_STEP1 '060'
              edit EFI_STEP2 '180'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 084_204
              edit ACC '120'
              edit EFI_STEP1 '084'
              edit EFI_STEP2 '204'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 108_228
              edit ACC '120'
              edit EFI_STEP1 '108'
              edit EFI_STEP2 '228'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 240_360
              edit ACC '120'
              edit EFI_STEP1 '240'
              edit EFI_STEP2 '360'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
          endfamily
          family 240
            family 000_240
              edit ACC '240'
              edit EFI_STEP1 '000'
              edit EFI_STEP2 '240'
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
          endfamily
          family 360
            family 000_360
              edit ACC '360'
              edit EFI_STEP1 '000'
              edit EFI_STEP2 '360'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
          endfamily
        endfamily
        family arch_op
          trigger op eq complete
          edit EPSLEG '1'
          edit SMSFILES '/home/ma/emos/def/emc_33r1/archive'
          task efi2fdb
            edit SMSFILES '/home/ma/emos/def/emc_33r1/mc/efi_hdc'
            event none
          family arch
            trigger ./efi2fdb eq complete
            family 2t
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME '2t'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
            family 10ff
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME '10ff'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
            family 10fg
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME '10fg'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
            family tp
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME 'tp'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
          endfamily
        endfamily
        family arch_nop
          trigger nop eq complete
          edit EPSLEG '2'
          edit SMSFILES '/home/ma/emos/def/emc_33r1/archive'
          task efi2fdb
            edit SMSFILES '/home/ma/emos/def/emc_33r1/mc/efi_hdc'
            event none
          family arch
            trigger ./efi2fdb eq complete
            family 2t
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME '2t'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
            family 10ff
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME '10ff'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
            family 10fg
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME '10fg'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
            family tp
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME 'tp'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
            family 2tmin
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME '2tmin'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
            family 2tmax
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME '2tmax'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
            family ts
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME 'ts'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
            family hsttmax
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME 'hsttmax'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
          endfamily
        endfamily
        family data
          trigger ./op eq complete and ./nop eq complete and ./arch_nop eq complete
          edit WSHOST 'linux_cluster'
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit.legc %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit SMSFILES '/home/ma/emos/def/emc_33r1/smsfiles/web'
          inlimit /limits:linux_cluster
          family 2t
            edit PARAM '2t'
            task efidata
          endfamily
          family 2tmin
            edit PARAM '2tmin'
            task efidata
          endfamily
          family 2tmax
            edit PARAM '2tmax'
            task efidata
          endfamily
          family tp
            edit PARAM 'tp'
            task efidata
          endfamily
          family 10ff
            edit PARAM '10ff'
            task efidata
          endfamily
          family 10fg
            edit PARAM '10fg'
            task efidata
          endfamily
          family sf
            edit PARAM 'sf'
            task efidata
          endfamily
          family hsstmax
            edit PARAM 'hsttmax'
            task efidata
          endfamily
        endfamily
        family plot
          trigger ./op eq complete and ./nop eq complete and data eq complete
          edit WSHOST 'linux_cluster'
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit.legc %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit SMSFILES '/home/ma/emos/def/emc_33r1/smsfiles/web'
          edit METVIEW 'metview_new'
          inlimit /limits:linux_cluster
          family 2t
            edit PARAM '2t'
            task efiplot
              edit SMSPASS 'stripped'
          endfamily
          family 2tmin
            edit PARAM '2tmin'
            task efiplot
          endfamily
          family 2tmax
            edit PARAM '2tmax'
            task efiplot
          endfamily
          family tp
            edit PARAM 'tp'
            task efiplot
          endfamily
          family 10ff
            edit PARAM '10ff'
            task efiplot
          endfamily
          family 10fg
            edit PARAM '10fg'
            task efiplot
          endfamily
          family sf
            edit PARAM 'sf'
            task efiplot
          endfamily
          family hsttmax
            edit PARAM 'hsttmax'
            task efiplot
          endfamily
        endfamily
      endfamily
      family 00
        trigger ./deps/00 eq complete and ./12 eq complete
        edit EMOS_BASE '00'
        edit DELTA_DAY '1'
        family op
          edit EPSLEG '1'
          task stage
          family 024
            family 000_024
              edit ACC '024'
              edit EFI_STEP1 '000'
              edit EFI_STEP2 '024'
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 024_048
              edit ACC '024'
              edit EFI_STEP1 '024'
              edit EFI_STEP2 '048'
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 048_072
              edit ACC '024'
              edit EFI_STEP1 '048'
              edit EFI_STEP2 '072'
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 072_096
              edit ACC '024'
              edit EFI_STEP1 '072'
              edit EFI_STEP2 '096'
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 096_120
              edit ACC '024'
              edit EFI_STEP1 '096'
              edit EFI_STEP2 '120'
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 120_144
              edit ACC '024'
              edit EFI_STEP1 '120'
              edit EFI_STEP2 '144'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
            endfamily
            family 144_168
              edit ACC '024'
              edit EFI_STEP1 '144'
              edit EFI_STEP2 '168'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
            endfamily
            family 168_192
              edit ACC '024'
              edit EFI_STEP1 '168'
              edit EFI_STEP2 '192'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
            endfamily
            family 192_216
              edit ACC '024'
              edit EFI_STEP1 '192'
              edit EFI_STEP2 '216'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
            endfamily
            family 216_240
              edit ACC '024'
              edit EFI_STEP1 '216'
              edit EFI_STEP2 '240'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
            endfamily
          endfamily
          family 120
            trigger ./024 eq complete
            family 000_120
              edit ACC '120'
              edit EFI_STEP1 '000'
              edit EFI_STEP2 '120'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 024_144
              edit ACC '120'
              edit EFI_STEP1 '024'
              edit EFI_STEP2 '144'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
          endfamily
          family 240
            trigger ./024 eq complete
            family 000_240
              edit ACC '240'
              edit EFI_STEP1 '000'
              edit EFI_STEP2 '240'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
          endfamily
        endfamily
        family nop
          edit EPSLEG '2'
          task stage
          family 024
            family 000_024
              edit ACC '024'
              edit EFI_STEP1 '000'
              edit EFI_STEP2 '024'
              family 2tmin
                edit RF_NAME '2tmin'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 024_048
              edit ACC '024'
              edit EFI_STEP1 '024'
              edit EFI_STEP2 '048'
              family 2tmin
                edit RF_NAME '2tmin'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 048_072
              edit ACC '024'
              edit EFI_STEP1 '048'
              edit EFI_STEP2 '072'
              family 2tmin
                edit RF_NAME '2tmin'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 072_096
              edit ACC '024'
              edit EFI_STEP1 '072'
              edit EFI_STEP2 '096'
              family 2tmin
                edit RF_NAME '2tmin'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 096_120
              edit ACC '024'
              edit EFI_STEP1 '096'
              edit EFI_STEP2 '120'
              family 2tmin
                edit RF_NAME '2tmin'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family reproc
              family 120_144
                edit ACC '024'
                edit EFI_STEP1 '120'
                edit EFI_STEP2 '144'
                family tp
                  edit RF_NAME 'tp'
                  task extindex
                endfamily
              endfamily
              family 144_168
                edit ACC '024'
                edit EFI_STEP1 '144'
                edit EFI_STEP2 '168'
                family tp
                  edit RF_NAME 'tp'
                  task extindex
                endfamily
              endfamily
            endfamily
            family 120_144
              edit ACC '024'
              edit EFI_STEP1 '120'
              edit EFI_STEP2 '144'
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 144_168
              edit ACC '024'
              edit EFI_STEP1 '144'
              edit EFI_STEP2 '168'
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 168_192
              edit ACC '24'
              edit EFI_STEP1 '168'
              edit EFI_STEP2 '192'
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 192_216
              edit ACC '24'
              edit EFI_STEP1 '192'
              edit EFI_STEP2 '216'
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 216_240
              edit ACC '24'
              edit EFI_STEP1 '216'
              edit EFI_STEP2 '240'
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 240_264
              edit ACC '24'
              edit EFI_STEP1 '240'
              edit EFI_STEP2 '264'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 264_288
              edit ACC '24'
              edit EFI_STEP1 '264'
              edit EFI_STEP2 '288'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 288_312
              edit ACC '24'
              edit EFI_STEP1 '288'
              edit EFI_STEP2 '312'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 312_336
              edit ACC '24'
              edit EFI_STEP1 '312'
              edit EFI_STEP2 '336'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 336_360
              edit ACC '24'
              edit EFI_STEP1 '336'
              edit EFI_STEP2 '360'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
          endfamily
          family 072
            family 000_072
              edit ACC '072'
              edit EFI_STEP1 '000'
              edit EFI_STEP2 '072'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 024_096
              edit ACC '072'
              edit EFI_STEP1 '024'
              edit EFI_STEP2 '096'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 048_120
              edit ACC '072'
              edit EFI_STEP1 '048'
              edit EFI_STEP2 '120'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 072_144
              edit ACC '072'
              edit EFI_STEP1 '072'
              edit EFI_STEP2 '144'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 096_168
              edit ACC '072'
              edit EFI_STEP1 '096'
              edit EFI_STEP2 '168'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 120_192
              edit ACC '072'
              edit EFI_STEP1 '120'
              edit EFI_STEP2 '192'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 144_216
              edit ACC '072'
              edit EFI_STEP1 '144'
              edit EFI_STEP2 '216'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
          endfamily
          family 120
            family 000_120
              edit ACC '120'
              edit EFI_STEP1 '000'
              edit EFI_STEP2 '120'
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 024_144
              edit ACC '120'
              edit EFI_STEP1 '024'
              edit EFI_STEP2 '144'
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 048_168
              edit ACC '120'
              edit EFI_STEP1 '048'
              edit EFI_STEP2 '168'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 072_192
              edit ACC '120'
              edit EFI_STEP1 '072'
              edit EFI_STEP2 '192'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 096_216
              edit ACC '120'
              edit EFI_STEP1 '096'
              edit EFI_STEP2 '216'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 240_360
              edit ACC '120'
              edit EFI_STEP1 '240'
              edit EFI_STEP2 '360'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
          endfamily
          family 240
            family 000_240
              edit ACC '240'
              edit EFI_STEP1 '000'
              edit EFI_STEP2 '240'
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
          endfamily
          family 360
            family 000_360
              edit ACC '360'
              edit EFI_STEP1 '000'
              edit EFI_STEP2 '360'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
          endfamily
        endfamily
        family arch_op
          trigger op eq complete
          edit EPSLEG '1'
          edit SMSFILES '/home/ma/emos/def/emc_33r1/archive'
          task efi2fdb
            edit SMSFILES '/home/ma/emos/def/emc_33r1/mc/efi_hdc'
            event none
          family arch
            trigger ./efi2fdb eq complete
            family 2t
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME '2t'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
            family 10ff
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME '10ff'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
            family 10fg
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME '10fg'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
            family tp
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME 'tp'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
          endfamily
        endfamily
        family arch_nop
          trigger nop eq complete
          edit EPSLEG '2'
          edit SMSFILES '/home/ma/emos/def/emc_33r1/archive'
          task efi2fdb
            edit SMSFILES '/home/ma/emos/def/emc_33r1/mc/efi_hdc'
            event none
          family arch
            trigger ./efi2fdb eq complete
            family 2t
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME '2t'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
            family 10ff
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME '10ff'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
            family 10fg
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME '10fg'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
            family tp
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME 'tp'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
            family 2tmin
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME '2tmin'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
            family 2tmax
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME '2tmax'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
            family ts
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME 'ts'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
            family hsttmax
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME 'hsttmax'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
          endfamily
        endfamily
        family data
          trigger ./op eq complete and ./nop eq complete and ./arch_nop eq complete
          edit WSHOST 'linux_cluster'
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit.legc %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit SMSFILES '/home/ma/emos/def/emc_33r1/smsfiles/web'
          inlimit /limits:linux_cluster
          family 2t
            edit PARAM '2t'
            task efidata
          endfamily
          family 2tmin
            edit PARAM '2tmin'
            task efidata
          endfamily
          family 2tmax
            edit PARAM '2tmax'
            task efidata
          endfamily
          family tp
            edit PARAM 'tp'
            task efidata
          endfamily
          family 10ff
            edit PARAM '10ff'
            task efidata
          endfamily
          family 10fg
            edit PARAM '10fg'
            task efidata
          endfamily
          family sf
            edit PARAM 'sf'
            task efidata
          endfamily
          family hsstmax
            edit PARAM 'hsttmax'
            task efidata
          endfamily
        endfamily
        family plot
          trigger ./op eq complete and ./nop eq complete and data eq complete
          edit WSHOST 'linux_cluster'
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit.legc %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit SMSFILES '/home/ma/emos/def/emc_33r1/smsfiles/web'
          edit METVIEW 'metview_new'
          inlimit /limits:linux_cluster
          family 2t
            edit PARAM '2t'
            task efiplot
          endfamily
          family 2tmin
            edit PARAM '2tmin'
            task efiplot
          endfamily
          family 2tmax
            edit PARAM '2tmax'
            task efiplot
          endfamily
          family tp
            edit PARAM 'tp'
            task efiplot
          endfamily
          family 10ff
            edit PARAM '10ff'
            task efiplot
          endfamily
          family 10fg
            edit PARAM '10fg'
            task efiplot
          endfamily
          family sf
            edit PARAM 'sf'
            task efiplot
          endfamily
          family hsttmax
            edit PARAM 'hsttmax'
            task efiplot
          endfamily
        endfamily
      endfamily
      task efi_cleaner
        trigger ./12 eq complete and ./00 eq complete
    endfamily
    family hind
      repeat date YMD 20080717 20121212 7
      edit DELTA_DAY '0'
      time 09:30
      family stage
        edit RF_NAME 'all'
        edit RETRIEVE 'stage'
        task prep
      endfamily
      family op
        trigger ./stage eq complete
        edit EPSLEG '1'
        family 024
          family 000_024
            edit ACC '024'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '024'
            family 2t
              edit RF_NAME '2t'
              task prep
                edit SMSPASS 'stripped'
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
                edit SMSPASS 'stripped'
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
                edit SMSPASS 'stripped'
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 012_036
            edit ACC '024'
            edit EFI_STEP1 '012'
            edit EFI_STEP2 '036'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 024_048
            edit ACC '024'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '048'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 036_060
            edit ACC '024'
            edit EFI_STEP1 '036'
            edit EFI_STEP2 '060'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 048_072
            edit ACC '024'
            edit EFI_STEP1 '048'
            edit EFI_STEP2 '072'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 060_084
            edit ACC '024'
            edit EFI_STEP1 '060'
            edit EFI_STEP2 '084'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 072_096
            edit ACC '024'
            edit EFI_STEP1 '072'
            edit EFI_STEP2 '096'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 084_108
            edit ACC '024'
            edit EFI_STEP1 '084'
            edit EFI_STEP2 '108'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 096_120
            edit ACC '024'
            edit EFI_STEP1 '096'
            edit EFI_STEP2 '120'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 108_132
            edit ACC '024'
            edit EFI_STEP1 '108'
            edit EFI_STEP2 '132'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 120_144
            edit ACC '024'
            edit EFI_STEP1 '120'
            edit EFI_STEP2 '144'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 144_168
            edit ACC '024'
            edit EFI_STEP1 '144'
            edit EFI_STEP2 '168'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 168_192
            edit ACC '024'
            edit EFI_STEP1 '168'
            edit EFI_STEP2 '192'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 192_216
            edit ACC '024'
            edit EFI_STEP1 '192'
            edit EFI_STEP2 '216'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 216_240
            edit ACC '024'
            edit EFI_STEP1 '216'
            edit EFI_STEP2 '240'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
        endfamily
        family 120
          trigger ./024 eq complete
          family 000_120
            edit ACC '120'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '120'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 024_144
            edit ACC '120'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '144'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
        endfamily
        family 240
          trigger ./024 eq complete
          family 000_240
            edit ACC '240'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '240'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
        endfamily
      endfamily
      family nop
        trigger ./stage eq complete
        edit EPSLEG '2'
        family 024
          family 000_024
            edit ACC '024'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '024'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
          endfamily
          family 012_036
            edit ACC '024'
            edit EFI_STEP1 '012'
            edit EFI_STEP2 '036'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
          endfamily
          family 024_048
            edit ACC '024'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '048'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
          endfamily
          family 036_060
            edit ACC '024'
            edit EFI_STEP1 '036'
            edit EFI_STEP2 '060'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
          endfamily
          family 048_072
            edit ACC '024'
            edit EFI_STEP1 '048'
            edit EFI_STEP2 '072'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
          endfamily
          family 060_084
            edit ACC '024'
            edit EFI_STEP1 '060'
            edit EFI_STEP2 '084'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
          endfamily
          family 072_096
            edit ACC '024'
            edit EFI_STEP1 '072'
            edit EFI_STEP2 '096'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
          endfamily
          family 084_108
            edit ACC '024'
            edit EFI_STEP1 '084'
            edit EFI_STEP2 '108'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
          endfamily
          family 096_120
            edit ACC '024'
            edit EFI_STEP1 '096'
            edit EFI_STEP2 '120'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
          endfamily
          family 108_132
            edit ACC '024'
            edit EFI_STEP1 '108'
            edit EFI_STEP2 '132'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
          endfamily
          family 120_144
            edit ACC '024'
            edit EFI_STEP1 '120'
            edit EFI_STEP2 '144'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
          endfamily
          family 144_168
            edit ACC '024'
            edit EFI_STEP1 '144'
            edit EFI_STEP2 '168'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
          endfamily
          family 168_192
            edit ACC '024'
            edit EFI_STEP1 '168'
            edit EFI_STEP2 '192'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
          endfamily
          family 192_216
            edit ACC '024'
            edit EFI_STEP1 '192'
            edit EFI_STEP2 '216'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
          endfamily
          family 216_240
            edit ACC '024'
            edit EFI_STEP1 '216'
            edit EFI_STEP2 '240'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
          endfamily
          family 240_264
            edit ACC '024'
            edit EFI_STEP1 '240'
            edit EFI_STEP2 '264'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
          endfamily
          family 264_288
            edit ACC '024'
            edit EFI_STEP1 '264'
            edit EFI_STEP2 '288'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
          endfamily
          family 288_312
            edit ACC '024'
            edit EFI_STEP1 '288'
            edit EFI_STEP2 '312'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
          endfamily
          family 312_336
            edit ACC '024'
            edit EFI_STEP1 '312'
            edit EFI_STEP2 '336'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
          endfamily
          family 336_360
            edit ACC '024'
            edit EFI_STEP1 '336'
            edit EFI_STEP2 '360'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
          endfamily
        endfamily
        family 000
          family 006_006
            edit ACC '000'
            edit EFI_STEP1 '006'
            edit EFI_STEP2 '006'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 012_012
            edit ACC '000'
            edit EFI_STEP1 '012'
            edit EFI_STEP2 '012'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 018_018
            edit ACC '000'
            edit EFI_STEP1 '018'
            edit EFI_STEP2 '018'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 024_024
            edit ACC '000'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '024'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 030_030
            edit ACC '000'
            edit EFI_STEP1 '030'
            edit EFI_STEP2 '030'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 036_036
            edit ACC '000'
            edit EFI_STEP1 '036'
            edit EFI_STEP2 '036'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 042_042
            edit ACC '000'
            edit EFI_STEP1 '042'
            edit EFI_STEP2 '042'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 048_048
            edit ACC '000'
            edit EFI_STEP1 '048'
            edit EFI_STEP2 '048'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 054_054
            edit ACC '000'
            edit EFI_STEP1 '054'
            edit EFI_STEP2 '054'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 060_060
            edit ACC '000'
            edit EFI_STEP1 '060'
            edit EFI_STEP2 '060'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 066_066
            edit ACC '000'
            edit EFI_STEP1 '066'
            edit EFI_STEP2 '066'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 072_072
            edit ACC '000'
            edit EFI_STEP1 '072'
            edit EFI_STEP2 '072'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 078_078
            edit ACC '000'
            edit EFI_STEP1 '078'
            edit EFI_STEP2 '078'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 084_084
            edit ACC '000'
            edit EFI_STEP1 '084'
            edit EFI_STEP2 '084'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 090_090
            edit ACC '000'
            edit EFI_STEP1 '090'
            edit EFI_STEP2 '090'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 096_096
            edit ACC '000'
            edit EFI_STEP1 '096'
            edit EFI_STEP2 '096'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 102_102
            edit ACC '000'
            edit EFI_STEP1 '102'
            edit EFI_STEP2 '102'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 108_108
            edit ACC '000'
            edit EFI_STEP1 '108'
            edit EFI_STEP2 '108'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 114_114
            edit ACC '000'
            edit EFI_STEP1 '114'
            edit EFI_STEP2 '114'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 120_120
            edit ACC '000'
            edit EFI_STEP1 '120'
            edit EFI_STEP2 '120'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 126_126
            edit ACC '000'
            edit EFI_STEP1 '126'
            edit EFI_STEP2 '126'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 132_132
            edit ACC '000'
            edit EFI_STEP1 '132'
            edit EFI_STEP2 '132'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 138_138
            edit ACC '000'
            edit EFI_STEP1 '138'
            edit EFI_STEP2 '138'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 144_144
            edit ACC '000'
            edit EFI_STEP1 '144'
            edit EFI_STEP2 '144'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 150_150
            edit ACC '000'
            edit EFI_STEP1 '150'
            edit EFI_STEP2 '150'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 156_156
            edit ACC '000'
            edit EFI_STEP1 '156'
            edit EFI_STEP2 '156'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 162_162
            edit ACC '000'
            edit EFI_STEP1 '162'
            edit EFI_STEP2 '162'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 168_168
            edit ACC '000'
            edit EFI_STEP1 '168'
            edit EFI_STEP2 '168'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 174_174
            edit ACC '000'
            edit EFI_STEP1 '174'
            edit EFI_STEP2 '174'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 180_180
            edit ACC '000'
            edit EFI_STEP1 '180'
            edit EFI_STEP2 '180'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 186_186
            edit ACC '000'
            edit EFI_STEP1 '186'
            edit EFI_STEP2 '186'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 192_192
            edit ACC '000'
            edit EFI_STEP1 '192'
            edit EFI_STEP2 '192'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 198_198
            edit ACC '000'
            edit EFI_STEP1 '198'
            edit EFI_STEP2 '198'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 204_204
            edit ACC '000'
            edit EFI_STEP1 '204'
            edit EFI_STEP2 '204'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 210_210
            edit ACC '000'
            edit EFI_STEP1 '210'
            edit EFI_STEP2 '210'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 216_216
            edit ACC '000'
            edit EFI_STEP1 '216'
            edit EFI_STEP2 '216'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 222_222
            edit ACC '000'
            edit EFI_STEP1 '222'
            edit EFI_STEP2 '222'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 228_228
            edit ACC '000'
            edit EFI_STEP1 '228'
            edit EFI_STEP2 '228'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 234_234
            edit ACC '000'
            edit EFI_STEP1 '234'
            edit EFI_STEP2 '234'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 240_240
            edit ACC '000'
            edit EFI_STEP1 '240'
            edit EFI_STEP2 '240'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 252_252
            edit ACC '000'
            edit EFI_STEP1 '252'
            edit EFI_STEP2 '252'
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 264_264
            edit ACC '000'
            edit EFI_STEP1 '264'
            edit EFI_STEP2 '264'
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 276_276
            edit ACC '000'
            edit EFI_STEP1 '276'
            edit EFI_STEP2 '276'
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 288_288
            edit ACC '000'
            edit EFI_STEP1 '288'
            edit EFI_STEP2 '288'
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 300_300
            edit ACC '000'
            edit EFI_STEP1 '300'
            edit EFI_STEP2 '300'
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 312_312
            edit ACC '000'
            edit EFI_STEP1 '312'
            edit EFI_STEP2 '312'
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 324_324
            edit ACC '000'
            edit EFI_STEP1 '324'
            edit EFI_STEP2 '324'
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 336_336
            edit ACC '000'
            edit EFI_STEP1 '336'
            edit EFI_STEP2 '336'
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 348_348
            edit ACC '000'
            edit EFI_STEP1 '348'
            edit EFI_STEP2 '348'
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 360_360
            edit ACC '000'
            edit EFI_STEP1 '360'
            edit EFI_STEP2 '360'
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
        endfamily
        family 006
          family 000_006
            edit ACC '006'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '006'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 006_012
            edit ACC '006'
            edit EFI_STEP1 '006'
            edit EFI_STEP2 '012'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 012_018
            edit ACC '006'
            edit EFI_STEP1 '012'
            edit EFI_STEP2 '018'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 018_024
            edit ACC '006'
            edit EFI_STEP1 '018'
            edit EFI_STEP2 '024'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 024_030
            edit ACC '006'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '030'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 030_036
            edit ACC '006'
            edit EFI_STEP1 '030'
            edit EFI_STEP2 '036'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 036_042
            edit ACC '006'
            edit EFI_STEP1 '036'
            edit EFI_STEP2 '042'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 042_048
            edit ACC '006'
            edit EFI_STEP1 '042'
            edit EFI_STEP2 '048'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 048_054
            edit ACC '006'
            edit EFI_STEP1 '048'
            edit EFI_STEP2 '054'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 054_060
            edit ACC '006'
            edit EFI_STEP1 '054'
            edit EFI_STEP2 '060'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 060_066
            edit ACC '006'
            edit EFI_STEP1 '060'
            edit EFI_STEP2 '066'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 066_072
            edit ACC '006'
            edit EFI_STEP1 '066'
            edit EFI_STEP2 '072'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 072_078
            edit ACC '006'
            edit EFI_STEP1 '072'
            edit EFI_STEP2 '078'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 078_084
            edit ACC '006'
            edit EFI_STEP1 '078'
            edit EFI_STEP2 '084'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 084_090
            edit ACC '006'
            edit EFI_STEP1 '084'
            edit EFI_STEP2 '090'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 090_096
            edit ACC '006'
            edit EFI_STEP1 '090'
            edit EFI_STEP2 '096'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 096_102
            edit ACC '006'
            edit EFI_STEP1 '096'
            edit EFI_STEP2 '102'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 102_108
            edit ACC '006'
            edit EFI_STEP1 '102'
            edit EFI_STEP2 '108'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 108_114
            edit ACC '006'
            edit EFI_STEP1 '108'
            edit EFI_STEP2 '114'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 114_120
            edit ACC '006'
            edit EFI_STEP1 '114'
            edit EFI_STEP2 '120'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 120_126
            edit ACC '006'
            edit EFI_STEP1 '120'
            edit EFI_STEP2 '126'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 126_132
            edit ACC '006'
            edit EFI_STEP1 '126'
            edit EFI_STEP2 '132'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 132_138
            edit ACC '006'
            edit EFI_STEP1 '132'
            edit EFI_STEP2 '138'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 138_144
            edit ACC '006'
            edit EFI_STEP1 '138'
            edit EFI_STEP2 '144'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 144_150
            edit ACC '006'
            edit EFI_STEP1 '144'
            edit EFI_STEP2 '150'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 150_156
            edit ACC '006'
            edit EFI_STEP1 '150'
            edit EFI_STEP2 '156'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 156_162
            edit ACC '006'
            edit EFI_STEP1 '156'
            edit EFI_STEP2 '162'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 162_168
            edit ACC '006'
            edit EFI_STEP1 '162'
            edit EFI_STEP2 '168'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 168_174
            edit ACC '006'
            edit EFI_STEP1 '168'
            edit EFI_STEP2 '174'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 174_180
            edit ACC '006'
            edit EFI_STEP1 '174'
            edit EFI_STEP2 '180'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 180_186
            edit ACC '006'
            edit EFI_STEP1 '180'
            edit EFI_STEP2 '186'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 186_192
            edit ACC '006'
            edit EFI_STEP1 '186'
            edit EFI_STEP2 '192'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 192_198
            edit ACC '006'
            edit EFI_STEP1 '192'
            edit EFI_STEP2 '198'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 198_204
            edit ACC '006'
            edit EFI_STEP1 '198'
            edit EFI_STEP2 '204'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 204_210
            edit ACC '006'
            edit EFI_STEP1 '204'
            edit EFI_STEP2 '210'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 210_216
            edit ACC '006'
            edit EFI_STEP1 '210'
            edit EFI_STEP2 '216'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 216_222
            edit ACC '006'
            edit EFI_STEP1 '216'
            edit EFI_STEP2 '222'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 222_228
            edit ACC '006'
            edit EFI_STEP1 '222'
            edit EFI_STEP2 '228'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 228_234
            edit ACC '006'
            edit EFI_STEP1 '228'
            edit EFI_STEP2 '234'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 234_240
            edit ACC '006'
            edit EFI_STEP1 '234'
            edit EFI_STEP2 '240'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
        endfamily
        family 012
          family 000_012
            edit ACC '012'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '012'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 012_024
            edit ACC '012'
            edit EFI_STEP1 '012'
            edit EFI_STEP2 '024'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 024_036
            edit ACC '012'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '036'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 036_048
            edit ACC '012'
            edit EFI_STEP1 '036'
            edit EFI_STEP2 '048'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 048_060
            edit ACC '012'
            edit EFI_STEP1 '048'
            edit EFI_STEP2 '060'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 060_072
            edit ACC '012'
            edit EFI_STEP1 '060'
            edit EFI_STEP2 '072'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 072_084
            edit ACC '012'
            edit EFI_STEP1 '072'
            edit EFI_STEP2 '084'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 084_096
            edit ACC '012'
            edit EFI_STEP1 '084'
            edit EFI_STEP2 '096'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 096_108
            edit ACC '012'
            edit EFI_STEP1 '096'
            edit EFI_STEP2 '108'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 108_120
            edit ACC '012'
            edit EFI_STEP1 '108'
            edit EFI_STEP2 '120'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 120_132
            edit ACC '012'
            edit EFI_STEP1 '120'
            edit EFI_STEP2 '132'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 132_144
            edit ACC '012'
            edit EFI_STEP1 '132'
            edit EFI_STEP2 '144'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 144_156
            edit ACC '012'
            edit EFI_STEP1 '144'
            edit EFI_STEP2 '156'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 156_168
            edit ACC '012'
            edit EFI_STEP1 '156'
            edit EFI_STEP2 '168'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 168_180
            edit ACC '012'
            edit EFI_STEP1 '168'
            edit EFI_STEP2 '180'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 180_192
            edit ACC '012'
            edit EFI_STEP1 '180'
            edit EFI_STEP2 '192'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 192_204
            edit ACC '012'
            edit EFI_STEP1 '192'
            edit EFI_STEP2 '204'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 204_216
            edit ACC '012'
            edit EFI_STEP1 '204'
            edit EFI_STEP2 '216'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 216_228
            edit ACC '012'
            edit EFI_STEP1 '216'
            edit EFI_STEP2 '228'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 228_240
            edit ACC '012'
            edit EFI_STEP1 '228'
            edit EFI_STEP2 '240'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 240_252
            edit ACC '012'
            edit EFI_STEP1 '240'
            edit EFI_STEP2 '252'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 252_264
            edit ACC '012'
            edit EFI_STEP1 '252'
            edit EFI_STEP2 '264'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 264_276
            edit ACC '012'
            edit EFI_STEP1 '264'
            edit EFI_STEP2 '276'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 276_288
            edit ACC '012'
            edit EFI_STEP1 '276'
            edit EFI_STEP2 '288'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 288_300
            edit ACC '012'
            edit EFI_STEP1 '288'
            edit EFI_STEP2 '300'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 300_312
            edit ACC '012'
            edit EFI_STEP1 '300'
            edit EFI_STEP2 '312'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 312_324
            edit ACC '012'
            edit EFI_STEP1 '312'
            edit EFI_STEP2 '324'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 324_336
            edit ACC '012'
            edit EFI_STEP1 '324'
            edit EFI_STEP2 '336'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 336_348
            edit ACC '012'
            edit EFI_STEP1 '336'
            edit EFI_STEP2 '348'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 348_360
            edit ACC '012'
            edit EFI_STEP1 '348'
            edit EFI_STEP2 '360'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
        endfamily
        family 072
          trigger ./024 eq complete
          family 000_072
            edit ACC '072'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '072'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 024_096
            edit ACC '072'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '096'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 048_120
            edit ACC '072'
            edit EFI_STEP1 '048'
            edit EFI_STEP2 '120'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 072_144
            edit ACC '072'
            edit EFI_STEP1 '072'
            edit EFI_STEP2 '144'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 096_168
            edit ACC '072'
            edit EFI_STEP1 '096'
            edit EFI_STEP2 '168'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 120_192
            edit ACC '072'
            edit EFI_STEP1 '120'
            edit EFI_STEP2 '192'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 144_216
            edit ACC '072'
            edit EFI_STEP1 '144'
            edit EFI_STEP2 '216'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 168_240
            edit ACC '072'
            edit EFI_STEP1 '168'
            edit EFI_STEP2 '240'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
        endfamily
        family 120
          trigger ./024 eq complete
          family 000_120
            edit ACC '120'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '120'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 024_144
            edit ACC '120'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '144'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 048_168
            edit ACC '120'
            edit EFI_STEP1 '048'
            edit EFI_STEP2 '168'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 072_192
            edit ACC '120'
            edit EFI_STEP1 '072'
            edit EFI_STEP2 '192'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 096_216
            edit ACC '120'
            edit EFI_STEP1 '096'
            edit EFI_STEP2 '216'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 120_240
            edit ACC '120'
            edit EFI_STEP1 '120'
            edit EFI_STEP2 '240'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 240_360
            edit ACC '120'
            edit EFI_STEP1 '240'
            edit EFI_STEP2 '360'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
        endfamily
        family 240
          trigger ./024 eq complete
          family 000_240
            edit ACC '240'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '240'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 120_360
            edit ACC '240'
            edit EFI_STEP1 '120'
            edit EFI_STEP2 '360'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
        endfamily
        family 360
          trigger ./024 eq complete
          family 000_360
            edit ACC '360'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '360'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
        endfamily
      endfamily
      family check_nrt
        complete ./check_nrt/go eq complete
        edit REFPATH '/mofc/hc2'
        family catchup
          task check_nrt
            event ok
        endfamily
        family rt
          trigger catchup eq complete
          task check_nrt
            event ok
            cron -w 1 08:00
        endfamily
        task go
          complete ./rt/check_nrt:ok or ./catchup/check_nrt:ok
          trigger 0 eq 1
      endfamily
    endfamily
    family clim
      trigger ./clim:YMD lt ./hind:YMD
      repeat date YMD 20080717 20121212 7
      edit EFI_REMOVE_GRIB 'echo'
      edit DELTA_DAY '-7'
      time 09:30
      family op
        edit EPSLEG '1'
        family 024
          family 000_024
            edit ACC '024'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '024'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 012_036
            edit ACC '024'
            edit EFI_STEP1 '012'
            edit EFI_STEP2 '036'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 024_048
            edit ACC '024'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '048'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 036_060
            edit ACC '024'
            edit EFI_STEP1 '036'
            edit EFI_STEP2 '060'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 048_072
            edit ACC '024'
            edit EFI_STEP1 '048'
            edit EFI_STEP2 '072'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 060_084
            edit ACC '024'
            edit EFI_STEP1 '060'
            edit EFI_STEP2 '084'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 072_096
            edit ACC '024'
            edit EFI_STEP1 '072'
            edit EFI_STEP2 '096'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 084_108
            edit ACC '024'
            edit EFI_STEP1 '084'
            edit EFI_STEP2 '108'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 096_120
            edit ACC '024'
            edit EFI_STEP1 '096'
            edit EFI_STEP2 '120'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 108_132
            edit ACC '024'
            edit EFI_STEP1 '108'
            edit EFI_STEP2 '132'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 120_144
            edit ACC '024'
            edit EFI_STEP1 '120'
            edit EFI_STEP2 '144'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 144_168
            edit ACC '024'
            edit EFI_STEP1 '144'
            edit EFI_STEP2 '168'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 168_192
            edit ACC '024'
            edit EFI_STEP1 '168'
            edit EFI_STEP2 '192'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 192_216
            edit ACC '024'
            edit EFI_STEP1 '192'
            edit EFI_STEP2 '216'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 216_240
            edit ACC '024'
            edit EFI_STEP1 '216'
            edit EFI_STEP2 '240'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
        endfamily
        family 120
          family 000_120
            edit ACC '120'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '120'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 024_144
            edit ACC '120'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '144'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
        endfamily
        family 240
          family 000_240
            edit ACC '240'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '240'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
        endfamily
      endfamily
      family op2fdb
        trigger op eq complete
        edit EPSLEG '1'
        task efi2fdb
          edit KIND 'clim1'
          edit SMSFILES '/home/ma/emos/def/emc_33r1/mc/efi_hdc'
      endfamily
      family nop
        edit EPSLEG '2'
        family 024
          family 000_024
            edit ACC '024'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '024'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
          endfamily
          family 012_036
            edit ACC '024'
            edit EFI_STEP1 '012'
            edit EFI_STEP2 '036'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
          endfamily
          family 024_048
            edit ACC '024'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '048'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
          endfamily
          family 036_060
            edit ACC '024'
            edit EFI_STEP1 '036'
            edit EFI_STEP2 '060'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
          endfamily
          family 048_072
            edit ACC '024'
            edit EFI_STEP1 '048'
            edit EFI_STEP2 '072'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
          endfamily
          family 060_084
            edit ACC '024'
            edit EFI_STEP1 '060'
            edit EFI_STEP2 '084'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
          endfamily
          family 072_096
            edit ACC '024'
            edit EFI_STEP1 '072'
            edit EFI_STEP2 '096'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
          endfamily
          family 084_108
            edit ACC '024'
            edit EFI_STEP1 '084'
            edit EFI_STEP2 '108'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
          endfamily
          family 096_120
            edit ACC '024'
            edit EFI_STEP1 '096'
            edit EFI_STEP2 '120'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
          endfamily
          family 108_132
            edit ACC '024'
            edit EFI_STEP1 '108'
            edit EFI_STEP2 '132'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
          endfamily
          family 120_144
            edit ACC '024'
            edit EFI_STEP1 '120'
            edit EFI_STEP2 '144'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
          endfamily
          family 144_168
            edit ACC '024'
            edit EFI_STEP1 '144'
            edit EFI_STEP2 '168'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
          endfamily
          family 168_192
            edit ACC '024'
            edit EFI_STEP1 '168'
            edit EFI_STEP2 '192'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
          endfamily
          family 192_216
            edit ACC '024'
            edit EFI_STEP1 '192'
            edit EFI_STEP2 '216'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
          endfamily
          family 216_240
            edit ACC '024'
            edit EFI_STEP1 '216'
            edit EFI_STEP2 '240'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
          endfamily
          family 240_264
            edit ACC '024'
            edit EFI_STEP1 '240'
            edit EFI_STEP2 '264'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
          endfamily
          family 264_288
            edit ACC '024'
            edit EFI_STEP1 '264'
            edit EFI_STEP2 '288'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
          endfamily
          family 288_312
            edit ACC '024'
            edit EFI_STEP1 '288'
            edit EFI_STEP2 '312'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
          endfamily
          family 312_336
            edit ACC '024'
            edit EFI_STEP1 '312'
            edit EFI_STEP2 '336'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
          endfamily
          family 336_360
            edit ACC '024'
            edit EFI_STEP1 '336'
            edit EFI_STEP2 '360'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
          endfamily
        endfamily
        family 000
          family 006_006
            edit ACC '000'
            edit EFI_STEP1 '006'
            edit EFI_STEP2 '006'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 012_012
            edit ACC '000'
            edit EFI_STEP1 '012'
            edit EFI_STEP2 '012'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 018_018
            edit ACC '000'
            edit EFI_STEP1 '018'
            edit EFI_STEP2 '018'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 024_024
            edit ACC '000'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '024'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 030_030
            edit ACC '000'
            edit EFI_STEP1 '030'
            edit EFI_STEP2 '030'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 036_036
            edit ACC '000'
            edit EFI_STEP1 '036'
            edit EFI_STEP2 '036'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 042_042
            edit ACC '000'
            edit EFI_STEP1 '042'
            edit EFI_STEP2 '042'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 048_048
            edit ACC '000'
            edit EFI_STEP1 '048'
            edit EFI_STEP2 '048'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 054_054
            edit ACC '000'
            edit EFI_STEP1 '054'
            edit EFI_STEP2 '054'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 060_060
            edit ACC '000'
            edit EFI_STEP1 '060'
            edit EFI_STEP2 '060'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 066_066
            edit ACC '000'
            edit EFI_STEP1 '066'
            edit EFI_STEP2 '066'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 072_072
            edit ACC '000'
            edit EFI_STEP1 '072'
            edit EFI_STEP2 '072'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 078_078
            edit ACC '000'
            edit EFI_STEP1 '078'
            edit EFI_STEP2 '078'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 084_084
            edit ACC '000'
            edit EFI_STEP1 '084'
            edit EFI_STEP2 '084'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 090_090
            edit ACC '000'
            edit EFI_STEP1 '090'
            edit EFI_STEP2 '090'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 096_096
            edit ACC '000'
            edit EFI_STEP1 '096'
            edit EFI_STEP2 '096'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 102_102
            edit ACC '000'
            edit EFI_STEP1 '102'
            edit EFI_STEP2 '102'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 108_108
            edit ACC '000'
            edit EFI_STEP1 '108'
            edit EFI_STEP2 '108'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 114_114
            edit ACC '000'
            edit EFI_STEP1 '114'
            edit EFI_STEP2 '114'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 120_120
            edit ACC '000'
            edit EFI_STEP1 '120'
            edit EFI_STEP2 '120'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 126_126
            edit ACC '000'
            edit EFI_STEP1 '126'
            edit EFI_STEP2 '126'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 132_132
            edit ACC '000'
            edit EFI_STEP1 '132'
            edit EFI_STEP2 '132'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 138_138
            edit ACC '000'
            edit EFI_STEP1 '138'
            edit EFI_STEP2 '138'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 144_144
            edit ACC '000'
            edit EFI_STEP1 '144'
            edit EFI_STEP2 '144'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 150_150
            edit ACC '000'
            edit EFI_STEP1 '150'
            edit EFI_STEP2 '150'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 156_156
            edit ACC '000'
            edit EFI_STEP1 '156'
            edit EFI_STEP2 '156'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 162_162
            edit ACC '000'
            edit EFI_STEP1 '162'
            edit EFI_STEP2 '162'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 168_168
            edit ACC '000'
            edit EFI_STEP1 '168'
            edit EFI_STEP2 '168'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 174_174
            edit ACC '000'
            edit EFI_STEP1 '174'
            edit EFI_STEP2 '174'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 180_180
            edit ACC '000'
            edit EFI_STEP1 '180'
            edit EFI_STEP2 '180'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 186_186
            edit ACC '000'
            edit EFI_STEP1 '186'
            edit EFI_STEP2 '186'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 192_192
            edit ACC '000'
            edit EFI_STEP1 '192'
            edit EFI_STEP2 '192'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 198_198
            edit ACC '000'
            edit EFI_STEP1 '198'
            edit EFI_STEP2 '198'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 204_204
            edit ACC '000'
            edit EFI_STEP1 '204'
            edit EFI_STEP2 '204'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 210_210
            edit ACC '000'
            edit EFI_STEP1 '210'
            edit EFI_STEP2 '210'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 216_216
            edit ACC '000'
            edit EFI_STEP1 '216'
            edit EFI_STEP2 '216'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 222_222
            edit ACC '000'
            edit EFI_STEP1 '222'
            edit EFI_STEP2 '222'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 228_228
            edit ACC '000'
            edit EFI_STEP1 '228'
            edit EFI_STEP2 '228'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 234_234
            edit ACC '000'
            edit EFI_STEP1 '234'
            edit EFI_STEP2 '234'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 240_240
            edit ACC '000'
            edit EFI_STEP1 '240'
            edit EFI_STEP2 '240'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 252_252
            edit ACC '000'
            edit EFI_STEP1 '252'
            edit EFI_STEP2 '252'
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 264_264
            edit ACC '000'
            edit EFI_STEP1 '264'
            edit EFI_STEP2 '264'
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 276_276
            edit ACC '000'
            edit EFI_STEP1 '276'
            edit EFI_STEP2 '276'
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 288_288
            edit ACC '000'
            edit EFI_STEP1 '288'
            edit EFI_STEP2 '288'
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 300_300
            edit ACC '000'
            edit EFI_STEP1 '300'
            edit EFI_STEP2 '300'
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 312_312
            edit ACC '000'
            edit EFI_STEP1 '312'
            edit EFI_STEP2 '312'
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 324_324
            edit ACC '000'
            edit EFI_STEP1 '324'
            edit EFI_STEP2 '324'
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 336_336
            edit ACC '000'
            edit EFI_STEP1 '336'
            edit EFI_STEP2 '336'
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 348_348
            edit ACC '000'
            edit EFI_STEP1 '348'
            edit EFI_STEP2 '348'
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 360_360
            edit ACC '000'
            edit EFI_STEP1 '360'
            edit EFI_STEP2 '360'
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
        endfamily
        family 006
          family 000_006
            edit ACC '006'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '006'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 006_012
            edit ACC '006'
            edit EFI_STEP1 '006'
            edit EFI_STEP2 '012'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 012_018
            edit ACC '006'
            edit EFI_STEP1 '012'
            edit EFI_STEP2 '018'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 018_024
            edit ACC '006'
            edit EFI_STEP1 '018'
            edit EFI_STEP2 '024'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 024_030
            edit ACC '006'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '030'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 030_036
            edit ACC '006'
            edit EFI_STEP1 '030'
            edit EFI_STEP2 '036'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 036_042
            edit ACC '006'
            edit EFI_STEP1 '036'
            edit EFI_STEP2 '042'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 042_048
            edit ACC '006'
            edit EFI_STEP1 '042'
            edit EFI_STEP2 '048'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 048_054
            edit ACC '006'
            edit EFI_STEP1 '048'
            edit EFI_STEP2 '054'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 054_060
            edit ACC '006'
            edit EFI_STEP1 '054'
            edit EFI_STEP2 '060'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 060_066
            edit ACC '006'
            edit EFI_STEP1 '060'
            edit EFI_STEP2 '066'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 066_072
            edit ACC '006'
            edit EFI_STEP1 '066'
            edit EFI_STEP2 '072'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 072_078
            edit ACC '006'
            edit EFI_STEP1 '072'
            edit EFI_STEP2 '078'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 078_084
            edit ACC '006'
            edit EFI_STEP1 '078'
            edit EFI_STEP2 '084'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 084_090
            edit ACC '006'
            edit EFI_STEP1 '084'
            edit EFI_STEP2 '090'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 090_096
            edit ACC '006'
            edit EFI_STEP1 '090'
            edit EFI_STEP2 '096'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 096_102
            edit ACC '006'
            edit EFI_STEP1 '096'
            edit EFI_STEP2 '102'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 102_108
            edit ACC '006'
            edit EFI_STEP1 '102'
            edit EFI_STEP2 '108'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 108_114
            edit ACC '006'
            edit EFI_STEP1 '108'
            edit EFI_STEP2 '114'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 114_120
            edit ACC '006'
            edit EFI_STEP1 '114'
            edit EFI_STEP2 '120'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 120_126
            edit ACC '006'
            edit EFI_STEP1 '120'
            edit EFI_STEP2 '126'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 126_132
            edit ACC '006'
            edit EFI_STEP1 '126'
            edit EFI_STEP2 '132'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 132_138
            edit ACC '006'
            edit EFI_STEP1 '132'
            edit EFI_STEP2 '138'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 138_144
            edit ACC '006'
            edit EFI_STEP1 '138'
            edit EFI_STEP2 '144'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 144_150
            edit ACC '006'
            edit EFI_STEP1 '144'
            edit EFI_STEP2 '150'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 150_156
            edit ACC '006'
            edit EFI_STEP1 '150'
            edit EFI_STEP2 '156'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 156_162
            edit ACC '006'
            edit EFI_STEP1 '156'
            edit EFI_STEP2 '162'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 162_168
            edit ACC '006'
            edit EFI_STEP1 '162'
            edit EFI_STEP2 '168'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 168_174
            edit ACC '006'
            edit EFI_STEP1 '168'
            edit EFI_STEP2 '174'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 174_180
            edit ACC '006'
            edit EFI_STEP1 '174'
            edit EFI_STEP2 '180'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 180_186
            edit ACC '006'
            edit EFI_STEP1 '180'
            edit EFI_STEP2 '186'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 186_192
            edit ACC '006'
            edit EFI_STEP1 '186'
            edit EFI_STEP2 '192'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 192_198
            edit ACC '006'
            edit EFI_STEP1 '192'
            edit EFI_STEP2 '198'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 198_204
            edit ACC '006'
            edit EFI_STEP1 '198'
            edit EFI_STEP2 '204'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 204_210
            edit ACC '006'
            edit EFI_STEP1 '204'
            edit EFI_STEP2 '210'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 210_216
            edit ACC '006'
            edit EFI_STEP1 '210'
            edit EFI_STEP2 '216'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 216_222
            edit ACC '006'
            edit EFI_STEP1 '216'
            edit EFI_STEP2 '222'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 222_228
            edit ACC '006'
            edit EFI_STEP1 '222'
            edit EFI_STEP2 '228'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 228_234
            edit ACC '006'
            edit EFI_STEP1 '228'
            edit EFI_STEP2 '234'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 234_240
            edit ACC '006'
            edit EFI_STEP1 '234'
            edit EFI_STEP2 '240'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
        endfamily
        family 012
          family 000_012
            edit ACC '012'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '012'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 012_024
            edit ACC '012'
            edit EFI_STEP1 '012'
            edit EFI_STEP2 '024'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 024_036
            edit ACC '012'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '036'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 036_048
            edit ACC '012'
            edit EFI_STEP1 '036'
            edit EFI_STEP2 '048'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 048_060
            edit ACC '012'
            edit EFI_STEP1 '048'
            edit EFI_STEP2 '060'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 060_072
            edit ACC '012'
            edit EFI_STEP1 '060'
            edit EFI_STEP2 '072'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 072_084
            edit ACC '012'
            edit EFI_STEP1 '072'
            edit EFI_STEP2 '084'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 084_096
            edit ACC '012'
            edit EFI_STEP1 '084'
            edit EFI_STEP2 '096'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 096_108
            edit ACC '012'
            edit EFI_STEP1 '096'
            edit EFI_STEP2 '108'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 108_120
            edit ACC '012'
            edit EFI_STEP1 '108'
            edit EFI_STEP2 '120'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 120_132
            edit ACC '012'
            edit EFI_STEP1 '120'
            edit EFI_STEP2 '132'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 132_144
            edit ACC '012'
            edit EFI_STEP1 '132'
            edit EFI_STEP2 '144'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 144_156
            edit ACC '012'
            edit EFI_STEP1 '144'
            edit EFI_STEP2 '156'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 156_168
            edit ACC '012'
            edit EFI_STEP1 '156'
            edit EFI_STEP2 '168'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 168_180
            edit ACC '012'
            edit EFI_STEP1 '168'
            edit EFI_STEP2 '180'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 180_192
            edit ACC '012'
            edit EFI_STEP1 '180'
            edit EFI_STEP2 '192'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 192_204
            edit ACC '012'
            edit EFI_STEP1 '192'
            edit EFI_STEP2 '204'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 204_216
            edit ACC '012'
            edit EFI_STEP1 '204'
            edit EFI_STEP2 '216'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 216_228
            edit ACC '012'
            edit EFI_STEP1 '216'
            edit EFI_STEP2 '228'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 228_240
            edit ACC '012'
            edit EFI_STEP1 '228'
            edit EFI_STEP2 '240'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 240_252
            edit ACC '012'
            edit EFI_STEP1 '240'
            edit EFI_STEP2 '252'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 252_264
            edit ACC '012'
            edit EFI_STEP1 '252'
            edit EFI_STEP2 '264'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 264_276
            edit ACC '012'
            edit EFI_STEP1 '264'
            edit EFI_STEP2 '276'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 276_288
            edit ACC '012'
            edit EFI_STEP1 '276'
            edit EFI_STEP2 '288'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 288_300
            edit ACC '012'
            edit EFI_STEP1 '288'
            edit EFI_STEP2 '300'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 300_312
            edit ACC '012'
            edit EFI_STEP1 '300'
            edit EFI_STEP2 '312'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 312_324
            edit ACC '012'
            edit EFI_STEP1 '312'
            edit EFI_STEP2 '324'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 324_336
            edit ACC '012'
            edit EFI_STEP1 '324'
            edit EFI_STEP2 '336'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 336_348
            edit ACC '012'
            edit EFI_STEP1 '336'
            edit EFI_STEP2 '348'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 348_360
            edit ACC '012'
            edit EFI_STEP1 '348'
            edit EFI_STEP2 '360'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
        endfamily
        family 072
          family 000_072
            edit ACC '072'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '072'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 024_096
            edit ACC '072'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '096'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 048_120
            edit ACC '072'
            edit EFI_STEP1 '048'
            edit EFI_STEP2 '120'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 072_144
            edit ACC '072'
            edit EFI_STEP1 '072'
            edit EFI_STEP2 '144'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 096_168
            edit ACC '072'
            edit EFI_STEP1 '096'
            edit EFI_STEP2 '168'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 120_192
            edit ACC '072'
            edit EFI_STEP1 '120'
            edit EFI_STEP2 '192'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 144_216
            edit ACC '072'
            edit EFI_STEP1 '144'
            edit EFI_STEP2 '216'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 168_240
            edit ACC '072'
            edit EFI_STEP1 '168'
            edit EFI_STEP2 '240'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
        endfamily
        family 120
          family 000_120
            edit ACC '120'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '120'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 024_144
            edit ACC '120'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '144'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 048_168
            edit ACC '120'
            edit EFI_STEP1 '048'
            edit EFI_STEP2 '168'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 072_192
            edit ACC '120'
            edit EFI_STEP1 '072'
            edit EFI_STEP2 '192'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 096_216
            edit ACC '120'
            edit EFI_STEP1 '096'
            edit EFI_STEP2 '216'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 120_240
            edit ACC '120'
            edit EFI_STEP1 '120'
            edit EFI_STEP2 '240'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 240_360
            edit ACC '120'
            edit EFI_STEP1 '240'
            edit EFI_STEP2 '360'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
        endfamily
        family 240
          family 000_240
            edit ACC '240'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '240'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 120_360
            edit ACC '240'
            edit EFI_STEP1 '120'
            edit EFI_STEP2 '360'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
        endfamily
        family 360
          family 000_360
            edit ACC '360'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '360'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
        endfamily
      endfamily
      family nop2fdb
        trigger nop eq complete and ./op2fdb eq complete
        edit EPSLEG '2'
        task efi2fdb
          edit KIND 'clim2'
          edit SMSFILES '/home/ma/emos/def/emc_33r1/mc/efi_hdc'
      endfamily
      family arch_nop
        trigger nop2fdb eq complete
        edit EPSLEG '2'
        edit SMSFILES '/home/ma/emos/def/emc_33r1/archive'
        family 2t
          edit step1 '24'
          edit step2 '24'
          edit RF_NAME '2t'
          task eficlim
            label info ""
        endfamily
        family 10ff
          edit step1 '24'
          edit step2 '24'
          edit RF_NAME '10ff'
          task eficlim
            label info ""
        endfamily
        family 10fg
          edit step1 '24'
          edit step2 '24'
          edit RF_NAME '10fg'
          task eficlim
            label info ""
        endfamily
        family tp
          edit step1 '24'
          edit step2 '24'
          edit RF_NAME 'tp'
          task eficlim
            label info ""
        endfamily
        family 2tmin
          edit step1 '24'
          edit step2 '24'
          edit RF_NAME '2tmin'
          task eficlim
            label info ""
        endfamily
        family 2tmax
          edit step1 '24'
          edit step2 '24'
          edit RF_NAME '2tmax'
          task eficlim
            label info ""
        endfamily
        family ts
          edit step1 '24'
          edit step2 '24'
          edit RF_NAME 'ts'
          task eficlim
            label info ""
        endfamily
        family hsttmax
          edit step1 '24'
          edit step2 '24'
          edit RF_NAME 'hsttmax'
          task eficlim
            label info ""
        endfamily
        family 10dd
          edit step1 '24'
          edit step2 '24'
          edit RF_NAME '10dd'
          task eficlim
            label info ""
        endfamily
        family 850t
          edit step1 '24'
          edit step2 '24'
          edit RF_NAME '850t'
          task eficlim
            label info ""
        endfamily
        family 500z
          edit step1 '24'
          edit step2 '24'
          edit RF_NAME '500z'
          task eficlim
            label info ""
        endfamily
        family 2tmean
          edit step2 '24'
          edit RF_NAME '2t'
          edit step1 '0'
          task eficlim
            label info ""
        endfamily
        family 10ffmean
          edit step2 '24'
          edit RF_NAME '10ff'
          edit step1 '0'
          task eficlim
            label info ""
        endfamily
        family tcc
          edit step1 '24'
          edit step2 '24'
          edit RF_NAME 'tcc'
          task eficlim
            label info ""
        endfamily
        family tccmean
          edit step2 '24'
          edit RF_NAME 'tcc'
          edit step1 '0'
          task eficlim
            label info ""
        endfamily
      endfamily
      task efi_cleaner
        trigger arch_nop eq complete and op eq complete and nop eq complete and op2fdb eq complete and nop2fdb eq complete
        edit DELTA_DAY '-43'
      family gdb
        trigger ( op eq complete and nop eq complete) and ( arch_nop eq complete)
        edit RHOST 'epsgrams'
        edit RDIR '/webapps/epsgrams/data/'
        edit KIND 'clim_15_days'
        edit WSHOST 'linux_cluster'
        edit QUEUE 'emos-e'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        inlimit /limits:linux_cluster
        task sync_climate
        task gclean
          trigger sync_climate eq complete
          edit SMSFILES '/home/ma/emos/def/emc_33r1/pop/data/epsgrams/grams'
          edit KEEP_DAYS '14'
          edit REMOTE_KEEP_DAYS '14'
        task gdb
          trigger gclean eq complete and ( /mc/main:YMD ge /emc_33r1/efi_next/mn:YMD and /mc/main/00 eq complete)
          edit SMSFILES '/home/ma/emos/def/emc_33r1/pop/data/epsgrams/grams'
          label files "="
          label info "="
          label local "="
          label remote "="
          event fdb
        task glink
          trigger gdb eq complete
          edit SMSFILES '/home/ma/emos/def/emc_33r1/pop/data/epsgrams/grams'
      endfamily
      family check_nrt
        complete ./check_nrt/go eq complete
        edit REFPATH '/mofc/hc2'
        family catchup
          task check_nrt
            event ok
        endfamily
        family rt
          trigger catchup eq complete
          task check_nrt
            event ok
            cron -w 3 06:30
        endfamily
        task go
          complete ./rt/check_nrt:ok or ./catchup/check_nrt:ok
          trigger 0 eq 1
      endfamily
    endfamily
  endfamily
  family limits
    limit mars 4
    limit pftotal 20
    limit pertic 13
    limit modelepsA 13
    limit modelepsB 15
    limit pl 3
    limit sfc 2
    limit prodeps 5
    limit prodwave 4
    limit pdb 5
    limit q2diss 5
    limit efi 13
    limit efih 13
    limit refc 10
    limit tc 5
    limit webplots 10
    limit dp 16
    limit hind 14
    limit pesh 20
  endfamily
endsuite
suite metops_test
  repeat date YMD 20090401 20191210 1
  edit USER 'emos'
  edit SCHOST 'c1a'
  edit WSHOST 'linux_cluster'
  edit WSLOGDIR '/vol/emos_nc/output'
  edit SCLOGDIR '/c1a/emos_dir/0001/log'
  edit USE_YMD 'true'
  edit VERSION '0001'
  edit FIRST_DAY '0'
  edit SMSTRIES '2'
  edit DELTA_DAY '0'
  edit EMOS_BASE '00'
  edit EMOS_TIME_STEP_H '0000'
  edit FSFAMILY ''
  edit EPSTYPE 'fc'
  edit MEMBER '0'
  edit PGNPES '3'
  edit SVNPES '8'
  edit REINITIALIZE 'false'
  edit DATEMASK '*.*.*'
  edit REMOVE_ADIR 'false'
  edit SMSHOME '/vol/emos_nc/output'
  edit SMSFILES '/home/ma/emos/def/metops_test/smsfiles'
  edit SMSINCLUDE '/home/ma/emos/def/metops_test/include'
  edit SMSURLBASE 'http://www.ecmwf.int'
  edit SMSURL 'publications/manuals/sms'
  edit ACCOUNT 'oosu'
  edit STREAM 'DA'
  edit EMOS_STREAM 'DA'
  edit FAMILY 'metops_test'
  edit MARS_DB 'marsod'
  edit OBSHOST 'dummy'
  edit SUITE_START '2000022100'
  edit QUEUE 'plots-medium'
  edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
  edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOBOUT%'
  edit LOGDIR '/vol/emos_nc/output'
  edit SMSOUT '/vol/emos_nc/output'
  edit SLEEPTIME '1'
  inlimit /limits:linux_cluster
  clock hybrid 
  family 00
    edit EMOS_BASE '00'
    family web
      family quartiles
        edit PLOT_AREAS 'Europe,North_America,Asia,South_Atlantic,South_Pacific,Australia'
        edit STEP_LIST '18/TO/126/by/12'
        edit STEP_PLOT '12/TO/120/by/12'
        edit ENSEMBLES '50'
        time 10:00
        task w_tp_quartiles
        task w_wind_quartiles
        task w_wgust_quartiles
        task w_wind925_quartiles
        task w_cape_quartiles
      endfamily
      family genesis
        task w_tc_genesis
          edit UPDATE_WEB_NOW 'yes'
          time 10:00
      endfamily
      family swfdp_pacific
        edit SMSFILES '/home/ma/emos/def/metops_test/smsfiles/swfdp_pacific'
        edit PLOTDOM '-40/150/10/-120'
        edit STEPS '06/12/18/24/30/36/42/48/54/60/66/72/84/96/108/120/132/144'
        edit UPDATE_WEB_NOW 'yes'
        edit SMSINCLUDE '/home/ma/emos/def/o/include'
        edit PLOT_STEPS '12/to/144/by/12'
        time 10:00
        family proba
          task swfdp_pacific_proba
          task swfdp_pacific_tp
            trigger swfdp_pacific_proba eq complete
          task swfdp_pacific_ff
            trigger swfdp_pacific_proba eq complete
          task swfdp_pacific_tp24
            trigger swfdp_pacific_proba eq complete
            edit STEPS '24/48/72/96/120'
          task swfdp_pacific_proba_clean
            trigger swfdp_pacific_tp eq complete and swfdp_pacific_ff eq complete and swfdp_pacific_tp24 eq complete
        endfamily
        family efi
          edit QUEUE 'plots-medium'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOBOUT%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit SMSINCLUDE '/home/ma/emos/def/o/include'
          inlimit /limits:linux_cluster
          family swfdp_pacific_tpi
            task w_efi
              edit PARAMETER 'tpi'
          endfamily
          family swfdp_pacific_10wsi
            task w_efi
              edit PARAMETER '10wsi'
          endfamily
          family swfdp_pacific_2ti
            task w_efi
              edit PARAMETER '2ti'
          endfamily
          family swfdp_pacific_10fgi
            task w_efi
              edit PARAMETER '10fgi'
          endfamily
        endfamily
        task pb_swh
        task pb_mwp
        task swfdp_pacific_wavegrams
          edit SMSINCLUDE '/home/ma/emos/def/metops_test/include'
        family maps
          edit PLOT_LOCATION '/webplots_e/data/swfdp_pacific/maps'
          edit STEPS '06/12/18/24/30/36/42/48/54/60/66/72/84/96/108/120/132/144'
          task swfdp_pacific_maps
          task swfdp_pacific_toweb1
            trigger swfdp_pacific_maps eq complete
          task swfdp_pacific_toweb2
            trigger swfdp_pacific_maps eq complete
          task swfdp_pacific_toweb3
            trigger swfdp_pacific_maps eq complete
          task swfdp_pacific_clean
            trigger swfdp_pacific_toweb1 eq complete and swfdp_pacific_toweb2 eq complete and swfdp_pacific_toweb3 eq complete
        endfamily
        family waves
          edit PLOT_LEGS '1'
          edit STEPS '6/12/18/24/30/36/42/48/54/60/66/72/84/96/108/120/132/144'
          task wave_data
          family plots
            task swfdp_pacific_swh
              trigger ../wave_data eq complete
            task swfdp_pacific_mwp
              trigger ../wave_data eq complete
            task swfdp_pacific_wwsh
              trigger ../wave_data eq complete
            task swfdp_pacific_wwmp
              trigger ../wave_data eq complete
            task swfdp_pacific_tssh
              trigger ../wave_data eq complete
            task swfdp_pacific_tsmp
              trigger ../wave_data eq complete
          endfamily
          task swfdp_pacific_wave_clean
            trigger ./plots eq complete
        endfamily
      endfamily
      family swdfp
        edit SMSFILES '/home/ma/emos/def/metops_test/smsfiles/swdfp'
        edit PLOTDOM '-45/-10/6/62'
        edit UPDATE_WEB_NOW 'yes'
        edit SMSINCLUDE '/home/ma/emos/def/o/include'
        edit PLOT_STEPS '12/to/144/by/12'
        time 10:00
        family waves
          edit SMSINCLUDE '/home/ma/emos/def/o/include'
          edit PLOT_LEGS '1'
          edit STEPS '6/12/18/24/30/36/42/48/54/60/66/72/84/96/108/120/132/144'
          task wave_data
          family plots
            task swdfp_swh
              trigger ../wave_data eq complete
            task swdfp_mwp
              trigger ../wave_data eq complete
            task swdfp_wwsh
              trigger ../wave_data eq complete
            task swdfp_wwmp
              trigger ../wave_data eq complete
            task swdfp_tssh
              trigger ../wave_data eq complete
            task swdfp_tsmp
              trigger ../wave_data eq complete
          endfamily
          task swdfp_wave_clean
            trigger ./plots eq complete
        endfamily
        task pb_swdfp_swh
        task pb_swdfp_mwp
      endfamily
      family wave_verif
        edit UPDATE_WEB_NOW 'yes'
        time 10:00
        task w_wave_timeseries
        task w_wave_annualscores
        task w_wave_intercomparison
      endfamily
      family newcluster
        task cluster_plot
          edit UPDATE_WEB_NOW 'yes'
          time 10:00
      endfamily
    endfamily
  endfamily
  family 12
    edit EMOS_BASE '12'
    family web
      family quartiles
        edit PLOT_AREAS 'Europe,North_America,Asia,South_Atlantic,South_Pacific,Australia'
        edit STEP_LIST '18/TO/126/by/12'
        edit STEP_PLOT '12/TO/120/by/12'
        edit ENSEMBLES '50'
        time 20:45
        task w_tp_quartiles
        task w_wind_quartiles
        task w_wgust_quartiles
        task w_wind925_quartiles
        task w_cape_quartiles
      endfamily
      family genesis
        task w_tc_genesis
          edit UPDATE_WEB_NOW 'yes'
          time 20:45
      endfamily
      family swfdp_pacific
        edit SMSFILES '/home/ma/emos/def/metops_test/smsfiles/swfdp_pacific'
        edit PLOTDOM '-40/150/10/-120'
        edit STEPS '06/12/18/24/30/36/42/48/54/60/66/72/84/96/108/120/132/144'
        edit UPDATE_WEB_NOW 'yes'
        edit SMSINCLUDE '/home/ma/emos/def/o/include'
        edit PLOT_STEPS '12/to/144/by/12'
        time 20:45
        family proba
          task swfdp_pacific_proba
          task swfdp_pacific_tp
            trigger swfdp_pacific_proba eq complete
          task swfdp_pacific_ff
            trigger swfdp_pacific_proba eq complete
          task swfdp_pacific_tp24
            trigger swfdp_pacific_proba eq complete
            edit STEPS '24/48/72/96/120'
          task swfdp_pacific_proba_clean
            trigger swfdp_pacific_tp eq complete and swfdp_pacific_ff eq complete and swfdp_pacific_tp24 eq complete
        endfamily
        family efi
          edit QUEUE 'plots-medium'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOBOUT%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit SMSINCLUDE '/home/ma/emos/def/o/include'
          inlimit /limits:linux_cluster
          family swfdp_pacific_tpi
            task w_efi
              edit PARAMETER 'tpi'
          endfamily
          family swfdp_pacific_10wsi
            task w_efi
              edit PARAMETER '10wsi'
          endfamily
          family swfdp_pacific_2ti
            task w_efi
              edit PARAMETER '2ti'
          endfamily
          family swfdp_pacific_10fgi
            task w_efi
              edit PARAMETER '10fgi'
          endfamily
        endfamily
        task pb_swh
        task pb_mwp
        task swfdp_pacific_wavegrams
          edit SMSINCLUDE '/home/ma/emos/def/metops_test/include'
        family maps
          edit PLOT_LOCATION '/webplots_e/data/swfdp_pacific/maps'
          edit STEPS '06/12/18/24/30/36/42/48/54/60/66/72/84/96/108/120/132/144'
          task swfdp_pacific_maps
          task swfdp_pacific_toweb1
            trigger swfdp_pacific_maps eq complete
          task swfdp_pacific_toweb2
            trigger swfdp_pacific_maps eq complete
          task swfdp_pacific_toweb3
            trigger swfdp_pacific_maps eq complete
          task swfdp_pacific_clean
            trigger swfdp_pacific_toweb1 eq complete and swfdp_pacific_toweb2 eq complete and swfdp_pacific_toweb3 eq complete
        endfamily
        family waves
          edit PLOT_LEGS '1'
          edit STEPS '6/12/18/24/30/36/42/48/54/60/66/72/84/96/108/120/132/144'
          task wave_data
          family plots
            task swfdp_pacific_swh
              trigger ../wave_data eq complete
            task swfdp_pacific_mwp
              trigger ../wave_data eq complete
            task swfdp_pacific_wwsh
              trigger ../wave_data eq complete
            task swfdp_pacific_wwmp
              trigger ../wave_data eq complete
            task swfdp_pacific_tssh
              trigger ../wave_data eq complete
            task swfdp_pacific_tsmp
              trigger ../wave_data eq complete
          endfamily
          task swfdp_pacific_wave_clean
            trigger ./plots eq complete
        endfamily
      endfamily
      family swdfp
        edit SMSFILES '/home/ma/emos/def/metops_test/smsfiles/swdfp'
        edit PLOTDOM '-45/-10/6/62'
        edit UPDATE_WEB_NOW 'yes'
        time 20:45
        family waves
          edit PLOT_LEGS '1'
          edit STEPS '6/12/18/24/30/36/42/48/54/60/66/72/84/96/108/120/132/144'
          task wave_data
          family plots
            task swdfp_wwsh
              trigger ../wave_data eq complete
          endfamily
        endfamily
      endfamily
    endfamily
  endfamily
endsuite
suite verify
  edit METPY_VERSION '3.4.2.emos'
  edit METPY 'metpy_001'
  edit VERIFY 'metpy_001'
  edit FIRST_DATE '20091008'
  edit LAST_DATE '20111031'
  edit DATE_INCREMENT '1'
  edit RUNS_PER_DAY '1'
  edit OPERATIONAL 'yes'
  edit USER 'emos'
  edit SCHOST 'c1a'
  edit WSLOGDIR '/vol/emos_nc/output'
  edit SCLOGDIR '/c1a/emos_dir/0001/log'
  edit USE_YMD 'true'
  edit VERSION '0001'
  edit FIRST_DAY '0'
  edit SMSTRIES '1'
  edit DELTA_DAY '0'
  edit EMOS_BASE '12'
  edit EMOS_TIME_STEP_H '0000'
  edit FSFAMILY ''
  edit STHOST_BKUP 'toto'
  edit EPSTYPE 'fc'
  edit MEMBER '0'
  edit PGNPES '3'
  edit SVNPES '8'
  edit REINITIALIZE 'false'
  edit DATEMASK '*.*.*'
  edit YMD 'dummy'
  edit REMOVE_ADIR 'false'
  edit ANNPES '18'
  edit ANMEM '3584'
  edit ANSTACK '700'
  edit ANMRFS '160'
  edit FCNPES '24'
  edit FCMEM '3584'
  edit FCSTACK '690'
  edit FCMRFS '32'
  edit PERIOD_4D '12'
  edit WINDOW_LENGTH_4D '12'
  edit WINDOW_OFFSET_4D '12'
  edit MKCMANPES '6'
  edit MKCMAMEM '3284'
  edit MKCMASTACK '600'
  edit MKCMAMRFS '600'
  edit SMSHOME '/home/ma/emos/def/verify'
  edit SMSFILES '/home/ma/emos/def/verify'
  edit SMSINCLUDE '/home/ma/emos/def/verify/include'
  edit ACCOUNT 'oosu'
  edit STREAM 'DA'
  edit EMOS_STREAM 'DA'
  edit MARS_DB 'marsod'
  edit OBSHOST 'ha_pp'
  edit SUITE_START '19971117'
  edit SCRIPTFILES '/home/ma/emos/def/verify/scripts'
  edit VERIFY_CACHE '/var/tmp/verify'
  edit WSHOST 'swarm'
  edit QUEUE 'verify'
  edit SMSCMD 'smssubmit.claude1 %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
  edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOBOUT%'
  edit LOGDIR '/vol/emos_nc/output'
  edit SMSOUT '/vol/emos_nc/output'
  edit SLEEPTIME '1'
  clock hybrid 
  family 00
    repeat date YMD 20091009 20111031 1
    edit SMSFILES '/home/ma/emos/def/verify/verify_oper'
    edit EMOS_BASE '00'
    family hr
      trigger /o/lag/00/archive/fc240 eq complete
      edit WSHOST 'swarm'
      edit QUEUE 'verify@scores2'
      edit SMSCMD 'smssubmit.claude1 %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOBOUT%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit STREAM 'oper'
      limit count 3
      inlimit count
      task hr_retrieve
        edit SMSTRIES '2'
      family z
        trigger hr_retrieve eq complete
        edit PARAMETER 'z'
        task hr_upperair
      endfamily
      family t
        trigger hr_retrieve eq complete
        edit PARAMETER 't'
        task hr_upperair
      endfamily
      family u_v_ff
        trigger hr_retrieve eq complete
        edit PARAMETER 'u,v,ff'
        task hr_upperair
          edit SMSPASS 'stripped'
      endfamily
      family r
        trigger hr_retrieve eq complete
        edit PARAMETER 'r'
        task hr_upperair
      endfamily
      family q
        trigger hr_retrieve eq complete
        edit PARAMETER 'q'
        task hr_upperair
      endfamily
      family o3
        trigger hr_retrieve eq complete
        edit PARAMETER 'o3'
        task hr_upperair
      endfamily
      task hr_sfc
        trigger hr_retrieve eq complete
      task hr_to_ct
        trigger hr_retrieve eq complete
      task hr_from_ct
        trigger hr_to_ct eq complete
      task hr_to_ct_wind_850
        trigger hr_retrieve eq complete
      task hr_from_ct_wind_850
        trigger hr_to_ct_wind_850 eq complete
      task hr_to_ct_wind_250
        trigger hr_retrieve eq complete
      task hr_from_ct_wind_250
        trigger hr_to_ct_wind_250 eq complete
    endfamily
    family hrobs
      trigger ./hr eq complete
      edit WSHOST 'swarm'
      edit QUEUE 'verify@scores1'
      edit SMSCMD 'smssubmit.claude1 %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOBOUT%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit FDB_HOST 'scores1'
      edit STREAM 'oper'
      edit DELTA_DAY '-2'
      limit count 2
      inlimit count
      task hrobs_retrieve1
        edit SMSTRIES '2'
      task hrobs_retrieve2
        edit SMSTRIES '2'
      family z
        trigger hrobs_retrieve1 eq complete and hrobs_retrieve2 eq complete
        edit PARAMETER 'z'
        task hrobs
      endfamily
      family t
        trigger hrobs_retrieve1 eq complete and hrobs_retrieve2 eq complete
        edit PARAMETER 't'
        task hrobs
      endfamily
      family u_v_ff
        trigger hrobs_retrieve1 eq complete and hrobs_retrieve2 eq complete
        edit PARAMETER 'u,v,ff'
        task hrobs
      endfamily
      family r
        trigger hrobs_retrieve1 eq complete and hrobs_retrieve2 eq complete
        edit PARAMETER 'r'
        task hrobs
      endfamily
      family q
        trigger hrobs_retrieve1 eq complete and hrobs_retrieve2 eq complete
        edit PARAMETER 'q'
        task hrobs
      endfamily
    endfamily
    family ctrl
      trigger /mc/main/00/legB/postproc eq complete
      edit WSHOST 'swarm'
      edit QUEUE 'verify@scores2'
      edit SMSCMD 'smssubmit.claude1 %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOBOUT%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit STREAM 'enfo'
      limit count 2
      inlimit count
      task ctrl_retrieve
      family z
        trigger ctrl_retrieve eq complete
        edit PARAMETER 'z'
        task ctrl_upperair
      endfamily
      family t
        trigger ctrl_retrieve eq complete
        edit PARAMETER 't'
        task ctrl_upperair
      endfamily
      family u_v_ff
        trigger ctrl_retrieve eq complete
        edit PARAMETER 'u,v,ff'
        task ctrl_upperair
      endfamily
      family r
        trigger ctrl_retrieve eq complete
        edit PARAMETER 'r'
        task ctrl_upperair
      endfamily
      family q
        trigger ctrl_retrieve eq complete
        edit PARAMETER 'q'
        task ctrl_upperair
      endfamily
      task ctrl_sfc
        trigger ctrl_retrieve eq complete
      task ctrl_to_ct
        trigger ctrl_retrieve eq complete
      task ctrl_from_ct
        trigger ctrl_to_ct eq complete
      task ctrl_to_ct_wind_850
        trigger ctrl_retrieve eq complete
      task ctrl_from_ct_wind_850
        trigger ctrl_to_ct_wind_850 eq complete
      task ctrl_to_ct_wind_250
        trigger ctrl_retrieve eq complete
      task ctrl_from_ct_wind_250
        trigger ctrl_to_ct_wind_250 eq complete
    endfamily
    family ctrlobs
      trigger ./ctrl eq complete
      edit WSHOST 'swarm'
      edit QUEUE 'verify@scores1'
      edit SMSCMD 'smssubmit.claude1 %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOBOUT%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit FDB_HOST 'scores1'
      edit STREAM 'enfo'
      edit DELTA_DAY '-2'
      limit count 2
      inlimit count
      task ctrlobs_retrieve1
        edit SMSTRIES '2'
      task ctrlobs_retrieve2
        edit SMSTRIES '2'
      family z
        trigger ctrlobs_retrieve1 eq complete and ctrlobs_retrieve2 eq complete
        edit PARAMETER 'z'
        task ctrlobs
      endfamily
      family t
        trigger ctrlobs_retrieve1 eq complete and ctrlobs_retrieve2 eq complete
        edit PARAMETER 't'
        task ctrlobs
      endfamily
      family r
        trigger ctrlobs_retrieve1 eq complete and ctrlobs_retrieve2 eq complete
        edit PARAMETER 'r'
        task ctrlobs
      endfamily
      family q
        trigger ctrlobs_retrieve1 eq complete and ctrlobs_retrieve2 eq complete
        edit PARAMETER 'q'
        task ctrlobs
      endfamily
      family u_v_ff
        trigger ctrlobs_retrieve1 eq complete and ctrlobs_retrieve2 eq complete
        edit PARAMETER 'u,v,ff'
        task ctrlobs
      endfamily
    endfamily
    family eps
      trigger /mc/main/00/legB/postproc eq complete
      edit WSHOST 'swarm'
      edit QUEUE 'verify@scores2'
      edit SMSCMD 'smssubmit.claude1 %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOBOUT%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit STREAM 'enfo'
      limit count 4
      inlimit count
      task eps_retrieve
        edit SMSTRIES '2'
      task eps_to_ct_z
        trigger eps_retrieve eq complete
      task eps_to_ct_msl
        trigger eps_retrieve eq complete
      task eps_to_ct_t
        trigger eps_retrieve eq complete
      task eps_from_ct
        trigger eps_to_ct_z eq complete and eps_to_ct_t eq complete and eps_to_ct_msl eq complete
      task eps_to_ct_wind_850
        trigger eps_retrieve eq complete
      task eps_from_ct_wind_850
        trigger eps_to_ct_wind_850 eq complete
      task eps_to_ct_wind_250
        trigger eps_retrieve eq complete
      task eps_from_ct_wind_250
        trigger eps_to_ct_wind_250 eq complete
      task eps_spread_score1
        trigger eps_to_ct_z eq complete and eps_to_ct_t eq complete and eps_to_ct_msl eq complete
      task eps_spread_score2
        trigger eps_to_ct_wind_850 eq complete and eps_to_ct_wind_250 eq complete
      task eps_cal_val
      task eps_mean_spread_z
        trigger eps_to_ct_z eq complete
      task eps_mean_spread_t
        trigger eps_to_ct_t eq complete
      task eps_mean_spread_msl
        trigger eps_to_ct_msl eq complete
      task eps_mean_spread_wind_850
        trigger eps_to_ct_wind_850 eq complete
      task eps_mean_spread_wind_250
        trigger eps_to_ct_wind_250 eq complete
    endfamily
    family clim
      trigger /o/main/00/an eq complete
      edit STREAM 'oper'
      limit count 1
      inlimit count
      task clim_z
      task clim_msl
      task clim_t
      task clim_wind_850
      task clim_wind_250
      task clim_waves_mwp
      task clim_waves_swh
    endfamily
    family waves
      trigger /mc/main/00/legB/postproc eq complete
      edit WSHOST 'swarm'
      edit QUEUE 'verify@scores2'
      edit SMSCMD 'smssubmit.claude1 %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOBOUT%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      limit count 4
      inlimit count
      task waves_retrieve
        edit SMSTRIES '2'
      task waves_retrieve_ff
        edit SMSTRIES '2'
      task waves_eps_retrieve
        edit SMSTRIES '2'
      task waves_ctrl_retrieve
        edit SMSTRIES '2'
      task waves
        trigger waves_retrieve eq complete
      task waves_ff
        trigger waves_retrieve_ff eq complete
      task waves_ff_sea
        trigger waves_ff eq complete
      task waves_to_ct
        trigger waves eq complete
      task waves_from_ct
        trigger waves_to_ct eq complete
      task waves_ctrl
        trigger waves_ctrl_retrieve eq complete
      task waves_ctrl_to_ct
        trigger waves_ctrl eq complete
      task waves_ctrl_from_ct
        trigger waves_ctrl_to_ct eq complete
      task waves_eps_to_ct
        trigger waves_eps_retrieve eq complete
      task waves_eps_from_ct
        trigger waves_eps_to_ct eq complete
      task waves_eps_mean_spread
        trigger waves_eps_to_ct eq complete
      task waves_eps_spread_score
        trigger waves_eps_to_ct eq complete
    endfamily
    family other
      edit WSHOST 'swarm'
      edit QUEUE 'verify@scores2'
      edit SMSCMD 'smssubmit.claude1 %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOBOUT%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit STREAM 'oper'
      limit count 3
      inlimit count
      time 08:00
      task other_ukmo_fc
      task other_ukmo_an
      task other_ncep_fc
      task other_ncep_an
      task other_dwd_fc
      task other_dwd_an
      task other_metfr_fc
      task other_metfr_an
      task other_jma_fc
      task other_jma_an
      task other_cmc_fc
      task other_cmc_an
      task other_bom_fc
      task other_bom_an
    endfamily
    task restart
      time 17:00
  endfamily
  family 12
    repeat date YMD 20091008 20111031 1
    edit SMSFILES '/home/ma/emos/def/verify/verify_oper'
    edit EMOS_BASE '12'
    edit OVERWRITE 'yes'
    family hr
      trigger /o/lag/12/archive/fc240 eq complete
      edit WSHOST 'swarm'
      edit QUEUE 'verify@scores2'
      edit SMSCMD 'smssubmit.claude1 %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOBOUT%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit STREAM 'oper'
      limit count 3
      inlimit count
      task hr_retrieve
        edit SMSTRIES '2'
      family z
        trigger hr_retrieve eq complete
        edit PARAMETER 'z'
        task hr_upperair
      endfamily
      family t
        trigger hr_retrieve eq complete
        edit PARAMETER 't'
        task hr_upperair
      endfamily
      family u_v_ff
        trigger hr_retrieve eq complete
        edit PARAMETER 'u,v,ff'
        task hr_upperair
      endfamily
      family r
        trigger hr_retrieve eq complete
        edit PARAMETER 'r'
        task hr_upperair
      endfamily
      family q
        trigger hr_retrieve eq complete
        edit PARAMETER 'q'
        task hr_upperair
      endfamily
      family o3
        trigger hr_retrieve eq complete
        edit PARAMETER 'o3'
        task hr_upperair
      endfamily
      task hr_sfc
        trigger hr_retrieve eq complete
      task hr_to_ct
        trigger hr_retrieve eq complete
      task hr_from_ct
        trigger hr_to_ct eq complete
      task hr_to_ct_wind_850
        trigger hr_retrieve eq complete
      task hr_from_ct_wind_850
        trigger hr_to_ct_wind_850 eq complete
      task hr_to_ct_wind_250
        trigger hr_retrieve eq complete
      task hr_from_ct_wind_250
        trigger hr_to_ct_wind_250 eq complete
    endfamily
    family hrobs
      trigger ./hr eq complete
      edit WSHOST 'swarm'
      edit QUEUE 'verify@scores1'
      edit SMSCMD 'smssubmit.claude1 %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOBOUT%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit FDB_HOST 'scores1'
      edit STREAM 'oper'
      edit DELTA_DAY '-2'
      edit OVERWRITE 'yes'
      limit count 2
      inlimit count
      task hrobs_retrieve1
        edit SMSTRIES '2'
      task hrobs_retrieve2
        edit SMSTRIES '2'
      family z
        trigger hrobs_retrieve1 eq complete and hrobs_retrieve2 eq complete
        edit PARAMETER 'z'
        task hrobs
      endfamily
      family t
        trigger hrobs_retrieve1 eq complete and hrobs_retrieve2 eq complete
        edit PARAMETER 't'
        task hrobs
      endfamily
      family u_v_ff
        trigger hrobs_retrieve1 eq complete and hrobs_retrieve2 eq complete
        edit PARAMETER 'u,v,ff'
        task hrobs
      endfamily
      family r
        trigger hrobs_retrieve1 eq complete and hrobs_retrieve2 eq complete
        edit PARAMETER 'r'
        task hrobs
      endfamily
      family q
        trigger hrobs_retrieve1 eq complete and hrobs_retrieve2 eq complete
        edit PARAMETER 'q'
        task hrobs
      endfamily
    endfamily
    family ctrl
      trigger /mc/main/12/legB/postproc eq complete
      edit WSHOST 'swarm'
      edit QUEUE 'verify@scores2'
      edit SMSCMD 'smssubmit.claude1 %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOBOUT%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit STREAM 'enfo'
      limit count 2
      inlimit count
      task ctrl_retrieve
        edit SMSTRIES '2'
      family z
        trigger ctrl_retrieve eq complete
        edit PARAMETER 'z'
        task ctrl_upperair
      endfamily
      family t
        trigger ctrl_retrieve eq complete
        edit PARAMETER 't'
        task ctrl_upperair
      endfamily
      family u_v_ff
        trigger ctrl_retrieve eq complete
        edit PARAMETER 'u,v,ff'
        task ctrl_upperair
      endfamily
      family r
        trigger ctrl_retrieve eq complete
        edit PARAMETER 'r'
        task ctrl_upperair
      endfamily
      family q
        trigger ctrl_retrieve eq complete
        edit PARAMETER 'q'
        task ctrl_upperair
      endfamily
      task ctrl_sfc
        trigger ctrl_retrieve eq complete
      task ctrl_to_ct
        trigger ctrl_retrieve eq complete
      task ctrl_from_ct
        trigger ctrl_to_ct eq complete
      task ctrl_to_ct_wind_850
        trigger ctrl_retrieve eq complete
      task ctrl_from_ct_wind_850
        trigger ctrl_to_ct_wind_850 eq complete
      task ctrl_to_ct_wind_250
        trigger ctrl_retrieve eq complete
      task ctrl_from_ct_wind_250
        trigger ctrl_to_ct_wind_250 eq complete
    endfamily
    family ctrlobs
      trigger ./ctrl eq complete
      edit WSHOST 'swarm'
      edit QUEUE 'verify@scores1'
      edit SMSCMD 'smssubmit.claude1 %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOBOUT%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit FDB_HOST 'scores1'
      edit STREAM 'enfo'
      edit DELTA_DAY '-2'
      edit OVERWRITE 'yes'
      limit count 2
      inlimit count
      task ctrlobs_retrieve1
        edit SMSTRIES '2'
      task ctrlobs_retrieve2
        edit SMSTRIES '2'
      family z
        trigger ctrlobs_retrieve1 eq complete and ctrlobs_retrieve2 eq complete
        edit PARAMETER 'z'
        task ctrlobs
      endfamily
      family t
        trigger ctrlobs_retrieve1 eq complete and ctrlobs_retrieve2 eq complete
        edit PARAMETER 't'
        task ctrlobs
      endfamily
      family r
        trigger ctrlobs_retrieve1 eq complete and ctrlobs_retrieve2 eq complete
        edit PARAMETER 'r'
        task ctrlobs
      endfamily
      family q
        trigger ctrlobs_retrieve1 eq complete and ctrlobs_retrieve2 eq complete
        edit PARAMETER 'q'
        task ctrlobs
      endfamily
      family u_v_ff
        trigger ctrlobs_retrieve1 eq complete and ctrlobs_retrieve2 eq complete
        edit PARAMETER 'u,v,ff'
        task ctrlobs
      endfamily
    endfamily
    family eps
      trigger /mc/main/12/legB/postproc eq complete
      edit WSHOST 'swarm'
      edit QUEUE 'verify@scores2'
      edit SMSCMD 'smssubmit.claude1 %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOBOUT%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit STREAM 'enfo'
      limit count 2
      inlimit count
      task eps_retrieve
        edit SMSTRIES '2'
      task eps_to_ct_z
        trigger eps_retrieve eq complete
      task eps_to_ct_msl
        trigger eps_retrieve eq complete
      task eps_to_ct_t
        trigger eps_retrieve eq complete
      task eps_from_ct
        trigger eps_to_ct_z eq complete and eps_to_ct_t eq complete and eps_to_ct_msl eq complete
      task eps_to_ct_wind_850
        trigger eps_retrieve eq complete
      task eps_from_ct_wind_850
        trigger eps_to_ct_wind_850 eq complete
      task eps_to_ct_wind_250
        trigger eps_retrieve eq complete
      task eps_from_ct_wind_250
        trigger eps_to_ct_wind_250 eq complete
      task eps_spread_score1
        trigger eps_to_ct_z eq complete and eps_to_ct_t eq complete and eps_to_ct_msl eq complete
      task eps_spread_score2
        trigger eps_to_ct_wind_850 eq complete and eps_to_ct_wind_250 eq complete
      task eps_cal_val
      task eps_mean_spread_z
        trigger eps_to_ct_z eq complete
      task eps_mean_spread_t
        trigger eps_to_ct_t eq complete
      task eps_mean_spread_msl
        trigger eps_to_ct_msl eq complete
      task eps_mean_spread_wind_850
        trigger eps_to_ct_wind_850 eq complete
      task eps_mean_spread_wind_250
        trigger eps_to_ct_wind_250 eq complete
    endfamily
    family clim
      trigger /o/main/12/an eq complete
      edit STREAM 'oper'
      limit count 1
      inlimit count
      task clim_z
      task clim_msl
      task clim_t
      task clim_wind_850
      task clim_wind_250
      task clim_waves_mwp
      task clim_waves_swh
    endfamily
    family waves
      trigger /mc/main/12/legB/postproc eq complete
      edit WSHOST 'swarm'
      edit QUEUE 'verify@scores2'
      edit SMSCMD 'smssubmit.claude1 %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOBOUT%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit OVERWRITE 'yes'
      limit count 4
      inlimit count
      task waves_retrieve
        edit SMSTRIES '2'
      task waves_retrieve_ff
        edit SMSTRIES '2'
      task waves_eps_retrieve
        edit SMSTRIES '2'
      task waves_ctrl_retrieve
        edit SMSTRIES '2'
      task waves
        trigger waves_retrieve eq complete
        edit SMSSTATUS '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      task waves_ff
        trigger waves_retrieve_ff eq complete
      task waves_ff_sea
        trigger waves_ff eq complete
      task waves_to_ct
        trigger waves eq complete
      task waves_from_ct
        trigger waves_to_ct eq complete
      task waves_ctrl
        trigger waves_ctrl_retrieve eq complete
      task waves_ctrl_to_ct
        trigger waves_ctrl eq complete
      task waves_ctrl_from_ct
        trigger waves_ctrl_to_ct eq complete
      task waves_eps_to_ct
        trigger waves_eps_retrieve eq complete
      task waves_eps_from_ct
        trigger waves_eps_to_ct eq complete
      task waves_eps_mean_spread
        trigger waves_eps_to_ct eq complete
      task waves_eps_spread_score
        trigger waves_eps_to_ct eq complete
    endfamily
    family other
      edit WSHOST 'swarm'
      edit QUEUE 'verify@scores2'
      edit SMSCMD 'smssubmit.claude1 %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOBOUT%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit STREAM 'oper'
      limit count 3
      inlimit count
      time 22:00
      task other_ukmo_fc
      task other_ukmo_an
      task other_ncep_fc
      task other_ncep_an
      task other_dwd_fc
      task other_dwd_an
      task other_metfr_fc
      task other_metfr_an
      task other_jma_fc
      task other_jma_an
      task other_cmc_fc
      task other_cmc_an
      task other_bom_fc
      task other_bom_an
    endfamily
    task restart
      time 17:30
  endfamily
  family 06
    repeat date YMD 20091008 20111031 1
    edit SMSFILES '/home/ma/emos/def/verify/verify_oper'
    edit EMOS_BASE '06'
    family bc
      edit WSHOST 'swarm'
      edit QUEUE 'verify@scores2'
      edit SMSCMD 'smssubmit.claude1 %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOBOUT%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit STREAM 'scda'
      limit count 1
      inlimit count
      task bc_upperair
      task bc_sfc
    endfamily
    task restart
      time 17:00
  endfamily
  family 18
    repeat date YMD 20091008 20111031 1
    edit SMSFILES '/home/ma/emos/def/verify/verify_oper'
    edit EMOS_BASE '18'
    family bc
      edit WSHOST 'swarm'
      edit QUEUE 'verify@scores2'
      edit SMSCMD 'smssubmit.claude1 %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOBOUT%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit STREAM 'scda'
      limit count 1
      inlimit count
      task bc_upperair
      task bc_sfc
    endfamily
    task restart
      time 09:00
  endfamily
endsuite
suite macc
  edit USER 'emos'
  edit SCHOST 'bee01'
  edit SCHOST_BKUP 'drn01'
  edit WSHOST 'swarm'
  edit WSLOGDIR '/vol/emos_nc/output'
  edit SCLOGDIR '/emos_dir/f93i/log'
  edit USE_YMD 'true'
  edit YMD '20090601'
  edit DATEMASK '*.*.*'
  edit VERSION 'f93i'
  edit INIEXPVER '0001'
  edit TCINIEXPVER '0001'
  edit FIRST_DAY '0'
  edit SMSTRIES '2'
  edit DELTA_DAY '0'
  edit EMOS_BASE '00'
  edit EMOS_TIME_STEP_H '0000'
  edit FSFAMILY 'tc'
  edit USER_PRIORITY '70'
  edit REINITIALIZE 'false'
  edit SMSHOME '/vol/emos_nc/output'
  edit SMSFILES '/home/ma/emos/def/macc/smsfiles'
  edit SMSINCLUDE '/home/ma/emos/def/macc/include'
  edit ACCOUNT 'oosu'
  edit STREAM 'EF'
  edit EMOS_STREAM 'EF'
  edit ENSEMBLES '50'
  edit ENS_SIZE '50'
  edit MARS_DB 'marsod'
  edit EPSTYPE 'fc'
  edit MEMBER '0'
  edit IS_REAL_TIME 'is_real_time'
  edit RUN_3HOURS_PP '1'
  edit SMSLOGHOST 'c1a'
  edit SMSLOGPORT '9316'
  edit OD_PROJ 'od'
  edit SUITE_START '20090602'
  edit QUEUE 'emos-e'
  edit QUEUE_EPILOG 'emos-exp'
  edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB%'
  edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB%'
  edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
  edit LOGDIR '/vol/emos_nc/output'
  edit SMSOUT '/vol/emos_nc/output'
  edit SLEEPTIME '1'
  inlimit /limits:swarm
  clock hybrid 
  family limits
    limit linux_cluster 8
    limit raq_ensemble 2
    limit calchas 4
  endfamily
  family gac
    inlimit /macc/limits:linux_cluster
    family plot
      edit METPY 'metpy'
      edit UPDATE_WEB_NOW 'yes'
      edit DO_PUSH_TO_WEB 'yes'
      family nrt
        repeat date YMD 20090928 20111215 1
        edit METPY 'metpy'
        edit VERSION 'f93i'
        edit PARAMETERS 'co,go3'
        edit LEVELS '850,500,300,50'
        task nrt_fields
          trigger ( /ecgems/f93i/an/lag:YMD gt /macc/gac/plot/nrt:YMD) or ( /ecgems/f93i/an/lag:YMD eq /macc/gac/plot/nrt:YMD and /ecgems/f93i/an/lag/00/archive eq complete)
        task nrt_totalcolumns
          trigger nrt_fields eq complete
          edit PARAMETERS 'tcco,gtco3'
        task nrt_zonalmean
          trigger nrt_totalcolumns eq complete
          edit METPY 'metpy_new'
        task nrt_opticaldepth
          trigger nrt_zonalmean eq complete
          edit AERTYPES 'total,natural,anthropogenic'
          edit METPY 'metpy_new'
        task nrt_uvindex
          trigger nrt_opticaldepth eq complete
      endfamily
      family fcampaign
        repeat date YMD 20090928 20111215 1
        edit VERSION 'f7kn'
        edit OWNER 'nal'
        family deps
          task dummy
            time 09:00
        endfamily
        task fcampaign_retrieve_and_process
          trigger deps eq complete
        task fcampaign_make_horiz_plots
          trigger fcampaign_retrieve_and_process eq complete
        task fcampaign_make_xs_plots
          trigger fcampaign_make_horiz_plots eq complete
        task fcampaign_cleanup
          trigger fcampaign_make_xs_plots eq complete
      endfamily
      family gems_reanalysis
        repeat enumerated month "200301" "200302" "200303" "200304" "200305" "200306" "200307" "200308" "200309" "200310" "200311" "200312" "200401" "200402" "200403" "200404" "200405" "200406" "200407" "200408" "200409" "200410" "200411" "200412" "200501" "200502" "200503" "200504" "200505" "200506" "200507" "200508" "200509" "200510" "200511" "200512" "200601" "200602" "200603" "200604" "200605" "200606" "200607" "200608" "200609" "200610" "200611" "200612" "200701" "200702" "200703" "200704" "200705" "200706" "200707" "200708" "200709" "200710" "200711" "200712" "200801" "200802" "200803" "200804" "200805" "200806" "200807" "200808" "200809" "200810" "200811" "200812" "200901" "200902" "200903" "200904"
        edit VERSION 'f026'
        edit LEVELS '850,500,300,50'
        edit title 'GEMS_Reanalysis'
        edit PARAMETERS 'co,go3,hcho,ch4,nox'
        task gems_monthly_fields
        task gems_monthly_zonalmean
          trigger gems_monthly_fields eq complete
          edit METPY 'metpy_new'
        task gems_monthly_totalcolumns
          trigger gems_monthly_zonalmean eq complete
          edit PARAMETERS 'tcco,gtco3,tchcho,tcch4'
      endfamily
      family fire_products
        repeat date YMD 20090928 20111215 1
        edit METPY 'metpy_new'
        family deps
          task dummy
            time 07:00
        endfamily
        task fire_radiative_power
          trigger deps eq complete
      endfamily
    endfamily
    family verif
      family aer
        family nrt
          repeat date YMD 20090928 20111215 1
          edit OWNER 'nal'
          task aerover_nrt_make_plots
            trigger ( /ecgems/f93i/an/lag:YMD gt /macc/gac/verif/aer/nrt:YMD) or ( /ecgems/f93i/an/lag:YMD eq /macc/gac/verif/aer/nrt:YMD and /ecgems/f93i/an/lag/00/archive eq complete)
          family publish
            repeat integer imonth 1 0 -1
            task aerover_nrt_push_statistic_plots
              trigger ../aerover_nrt_make_plots eq complete
            task aerover_nrt_push_station_plots
              trigger aerover_nrt_push_statistic_plots eq complete
          endfamily
        endfamily
      endfamily
    endfamily
  endfamily
  family diss
    task fcampaign_diss
      trigger ( /macc/gac/plot/fcampaign:YMD gt /macc/diss/fcampaign_diss:YMD) or ( /macc/gac/plot/fcampaign:YMD eq /macc/diss/fcampaign_diss:YMD and /macc/gac/plot/fcampaign eq complete)
      repeat date YMD 20090928 20111215 1
    task nrt_diss
      trigger ( /macc/gac/plot/nrt:YMD gt /macc/diss/nrt_diss:YMD) or ( /macc/gac/plot/nrt:YMD eq /macc/diss/nrt_diss:YMD and /macc/gac/plot/nrt eq complete)
      repeat date YMD 20090928 20111215 1
  endfamily
endsuite
suite ecgems
  edit SMSFILES '/home/ma/emos/def/sync'
  edit SLEEPTIME '1'
  clock hybrid 
  family f93i
    family an
      family lag
        edit YMD '20100114'
        family 00
          task archive
        endfamily
        family 12
          task archive
        endfamily
      endfamily
    endfamily
  endfamily
  family f8na
    family an
      family lag
        edit YMD '20010101'
        family 00
          task archive
        endfamily
        family 12
          task archive
        endfamily
      endfamily
    endfamily
  endfamily
endsuite
suite e_36r1
  edit CALC_PSBIAS 'true'
  edit CALC_RSTRHBIAS 'false'
  edit OBREALTIME 'true'
  edit LFULL_IASI 'true'
  edit USER 'emos'
  edit SCHOST 'c1b'
  edit STHOST '/s1b'
  edit SCHOST_BKUP 'c1a'
  edit STHOST_BKUP '/s1a'
  edit WSHOST 'linux_cluster'
  edit WSLOGDIR '/vol/emos_nc/output'
  edit SCLOGDIR '/s1b/emos_esuite/emos_data/0046/log'
  edit SMSLOGHOST 'c1b'
  edit SMSLOGPORT '9316'
  edit RESERVATION_ON '0'
  edit VERSION '0046'
  edit INIEXPVER '0045'
  edit INISTREAM 'DCDA'
  edit INIOFFSET_4D '3'
  edit INICLASS 'od'
  edit INILEVELS '91'
  edit INIRESOL '1279'
  edit FIRST_DAY '0'
  edit SMSTRIES '2'
  edit DELTA_DAY '0'
  edit EMOS_BASE '12'
  edit EMOS_TIME_STEP_H '0000'
  edit FSFAMILY ''
  edit INITIME '2009111800'
  edit EPSTYPE 'fc'
  edit MEMBER '0'
  edit PGNPES '64'
  edit SVNPES '8'
  edit ININPES '64'
  edit REINITIALIZE 'false'
  edit REINIEXPVER '0046'
  edit REINIPERIOD_4D '12'
  edit REINISTREAM 'DCDA'
  edit REINIOFFSET '6'
  edit REINI_NOT_ED 'false'
  edit DATEMASK '*.*.*'
  edit YMD '20091118'
  edit USE_YMD 'true'
  edit PARALLEL '4'
  edit USER_PRIORITY '70'
  edit FC_USER_PRIORITY '90'
  edit ANNPES '192'
  edit NPES_CCMA '%ANNPES%'
  edit ANMEM '3584'
  edit ANSTACK '700'
  edit ANMRFS '160'
  edit FCNPES '192'
  edit FCMEM '3584'
  edit FCSTACK '690'
  edit FCMRFS '32'
  edit PERIOD_4D '12'
  edit WINDOW_LENGTH_4D '12'
  edit WINDOW_OFFSET_4D '3'
  edit SMSHOME '/vol/emos_nc/output'
  edit SMSFILES '/home/ma/emos/def/e_36r1'
  edit SMSINCLUDE '/home/ma/emos/def/e_36r1/include'
  edit ACCOUNT 'oesu'
  edit STREAM 'DA'
  edit EMOS_STREAM 'DA'
  edit PGEN_BINDIR '/home/ma/emos/bin/diss'
  edit PGEN_OBJDIR '/home/ma/emos/bin/diss'
  edit DISS_REQDB_ROOT '/diss/requirements'
  edit DISS_BINDIR '/diss/bin/DISS'
  edit MARS_DB 'marsod'
  edit OBSHOST 'pp1'
  edit OBS_USER 'emos'
  edit OBS_BASE '/pp1'
  edit LANONLY 'true'
  edit LAR_4V_EF 'true'
  edit LFC_DIR_FREQ 'true'
  edit SMSLOGTIMEOUT '10'
  edit INIVARBC 'true'
  edit RUN_3HOURS_PP '1'
  edit VERIFY 'verify'
  edit SUITE_START '2009100212'
  edit QUEUE 'ns'
  edit QUEUE_EPILOG 'ns'
  edit SMSCMD 'smssubmit.36r1 %USER% %SCHOST% %SMSJOB% %SMSJOBOUT%'
  edit SMSKILL 'smskill %USER% %SCHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
  edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %SCHOST% %SMSRID% %SMSJOB%'
  edit LOGDIR '/s1b/emos_esuite/emos_data/0046/log'
  edit SMSOUT '/s1b/emos_esuite/emos_data/0046/log'
  edit SLEEPTIME '1'
  inlimit /limits:c1b
  clock hybrid 
  family limits
    limit mars 7
    limit obs 1
    limit ml 3
    limit prod 20
    limit q2diss 8
    limit w2diss 2
    limit pdb 8
    limit prod_bc 6
    limit prod_bc2 6
    limit q2diss_bc 5
    limit pdb_bc 2
    limit prodcheck 1
    limit backupprod 4
    limit dissbackup 4
    limit webplots 4
    limit satmon 7
    limit smon 3
    limit obs_get 5
    limit sync_fdb 1
    limit sync_data 1
    limit reservation 1
    limit satimsim 20
    limit satimsim_ret 5
  endfamily
  family main
    trigger ( main:YMD le ( lag:YMD + 3))
    repeat date YMD 20091118 20121212 1
    family make
      edit DELTA_DAY '-1'
      edit EMOS_BASE '18'
      edit STREAM 'DCDA'
      task setup
      task links
        trigger setup eq complete
        edit FSFAMILY 'an'
        edit WSHOST 'bee01'
      task datalinks
        trigger setup eq complete and links eq complete
        edit FSFAMILY 'an'
      family bins
        trigger ./links eq complete and ./datalinks eq complete
        edit SMSFILES '/home/ma/emos/def/e_36r1/make'
        task odbsql
        task odbtools
          trigger odbsql eq complete
        task b2otools
          trigger odbtools eq complete
        task black
          edit FSFAMILY 'an'
        task ifs
          trigger black eq complete and b2otools eq complete
          edit FSFAMILY 'an'
        task ssaabs
          trigger ifs eq complete
          edit FSFAMILY 'surf_anal'
        task wamabs
        task satrad
          trigger ifs eq complete
        task scat
        task obstat
          trigger odbsql eq complete
        task satmon
          trigger odbtools eq complete
        task prepdata
          trigger b2otools eq complete
        task satimbin
      endfamily
      family const
        edit SMSFILES '/home/ma/emos/def/e_36r1/make'
        task wconst
          trigger ../bins/wamabs eq complete
        task wavini
          trigger wconst eq complete
          edit FSFAMILY 'fc'
      endfamily
      task obstat_ws
        trigger links eq complete and datalinks eq complete
        edit QUEUE 'emos-e'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        edit SMSTRIES '1'
        inlimit /limits:linux_cluster
      task sync_basedir
      task getini
        trigger links eq complete and datalinks eq complete and ./bins/prepdata eq complete
        edit FSFAMILY 'fc'
      task inidata
        trigger ./bins/ifs eq complete and getini eq complete
        edit THREADS '8'
        edit FSFAMILY 'fc'
      family fc
        trigger inidata eq complete and const eq complete
        edit FSFAMILY 'fc'
        task getfcdata
        task model
          trigger getfcdata eq complete
          edit FCLENGTH '15'
          edit THREADS '8'
          edit USER_PRIORITY '94'
          meter step -1 15 15
        task clean
          trigger model eq complete
          edit DELTA_DAY '-2'
      endfamily
      family prod
        edit SMSFILES '/home/ma/emos/def/e_36r1/prod/make'
        edit EMOS_DOMAIN 'g'
        edit EMOS_TYPE 'fc'
        edit EMOS_STREAM 'OPER'
        task wparam
          edit EMOS_TYPE 'wp'
        task tf
          edit EMOS_TYPE 'tf'
        task sortreq
        task bitmap
        task clima
        task disssch
        task tools
        task oper
        task wave
          edit EMOS_STREAM 'WAVE'
      endfamily
      task plotdb
        trigger links eq complete and datalinks eq complete
        edit QUEUE 'emos-e'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        inlimit /limits:linux_cluster
    endfamily
    family 00dc
      trigger make eq complete
      edit EMOS_BASE '00'
      edit MXUP_TRAJ '3'
      edit STREAM 'DCDA'
      edit WINDOW_LENGTH_4D '12'
      edit WINDOW_OFFSET_4D '3'
      edit EDFAMILY 'false'
      edit PERIOD_4D '12'
      edit FCLENGTH '15'
      edit DELTA_DAY '0'
      family obs
        edit FSFAMILY 'an'
        edit SMSFILES '/home/ma/emos/def/e_36r1/obs'
        family get
          inlimit /e_36r1/limits:obs_get
          inlimit /limits:obs_get
          task get_nonbin
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc00/nonbin eq complete
            edit OBSFILE_OBSHOST 'BUFR'
            edit OBSFILE_HPCF 'bufr'
          task get_ers1_wa
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc00/ers1_wa eq complete
            edit OBSFILE_OBSHOST 'UWAX'
            edit OBSFILE_HPCF 'wavescat'
          task get_eas1
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc00/eas1 eq complete
            edit OBSFILE_OBSHOST 'EAS1'
            edit OBSFILE_HPCF 'wavesar'
          task get_ers1_wi
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc00/ers1_wi eq complete
            edit OBSFILE_OBSHOST 'UWIX'
            edit OBSFILE_HPCF 'scat'
          task get_ers1_ra
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc00/ers1_ra eq complete
            edit OBSFILE_OBSHOST 'URAX'
            edit OBSFILE_HPCF 'ralt'
          task get_rtovs
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc00/rtovs eq complete
            edit OBSFILE_OBSHOST 'TOVH'
            edit OBSFILE_HPCF 'tovh'
          task get_ssbt
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc00/ssbt eq complete
            edit OBSFILE_OBSHOST 'SSBT'
            edit OBSFILE_HPCF 'ssbt'
          task get_sst
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc00/sst/sst_fdb_c1b eq complete
            edit OBSFILE_OBSHOST 'nil'
            edit OBSFILE_HPCF 'nil'
          task get_atov
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc00/atov eq complete
            edit OBSFILE_OBSHOST 'ATOV'
            edit OBSFILE_HPCF 'atov'
          task get_grad
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc00/grad eq complete
            edit OBSFILE_OBSHOST 'GRAD'
            edit OBSFILE_HPCF 'grad'
          task get_stor
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc00/stor eq complete
            edit OBSFILE_OBSHOST 'STOR'
            edit OBSFILE_HPCF 'storm'
          task get_trmm
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc00/trmm eq complete
            edit OBSFILE_OBSHOST 'TRMM'
            edit OBSFILE_HPCF 'tmi'
          task get_qsca
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc00/qsca eq complete
            edit OBSFILE_OBSHOST 'QSCA'
            edit OBSFILE_HPCF 'qsca'
          task get_sbuv
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc00/sbuv eq complete
            edit OBSFILE_OBSHOST 'SBUV'
            edit OBSFILE_HPCF 'reo3'
          task get_airs
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc00/airs eq complete
            edit OBSFILE_OBSHOST 'AIRS'
            edit OBSFILE_HPCF 'airs'
          task get_ice
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc00/ice/ice_fdb_c1b eq complete
            edit OBSFILE_OBSHOST 'nil'
            edit OBSFILE_HPCF 'nil'
          task get_snow
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc00/snow eq complete
            edit OBSFILE_OBSHOST 'SNOW'
            edit OBSFILE_HPCF 'imssnow'
          task get_ateu
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc00/ateu eq complete
            edit OBSFILE_OBSHOST 'ATEU'
            edit OBSFILE_HPCF 'amsu_ears'
          task get_bice
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc00/bice eq complete
            edit OBSFILE_OBSHOST 'BICE'
            edit OBSFILE_HPCF 'baltic_ci'
          task get_amsa
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc00/amsa eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'AMSA'
            edit OBSFILE_HPCF 'amsa'
          task get_mhsa
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc00/mhsa eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'MHSA'
            edit OBSFILE_HPCF 'mhsa'
          task get_hirs
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc00/hirs eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'HIRS'
            edit OBSFILE_HPCF 'hirs'
          task get_amsr
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc00/amsr eq complete
            edit OBSFILE_OBSHOST 'AMSR'
            edit OBSFILE_HPCF 'amsre_aq'
          task get_rado
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc00/rado eq complete
            edit OBSFILE_OBSHOST 'RADO'
            edit OBSFILE_HPCF 'gpsro'
          task get_smin
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc00/smin eq complete
            edit OBSFILE_OBSHOST 'SMIN'
            edit OBSFILE_HPCF 'ssmisbt'
          task get_iasi
            late -s +00:02 -c +00:07
            trigger /pp2/main/extractions/dc00/iasi eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'IASI'
            edit OBSFILE_HPCF 'iasi'
          task get_gras
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc00/gras eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'GRAS'
            edit OBSFILE_HPCF 'gras'
          task get_modw
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc00/modw eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'MODW'
            edit OBSFILE_HPCF 'dbmodiswinds'
          task get_avhn
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc00/avhn eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'AVHN'
            edit OBSFILE_HPCF 'avhrrwinds'
          task get_atap
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc00/atap eq complete
            edit OBSFILE_OBSHOST 'ATAP'
            edit OBSFILE_HPCF 'amsu_pacrars'
          task get_g2to
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc00/g2to eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'G2TO'
            edit OBSFILE_HPCF 'reo3_gome2'
          task get_grst
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc00/grst eq complete
            edit OBSFILE_OBSHOST 'GRST'
            edit OBSFILE_HPCF 'ostia'
          task get_emer
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc00/emer eq complete
            edit OBSFILE_OBSHOST 'EMER'
            edit OBSFILE_HPCF 'envisat_meris'
          task get_j2nb
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc00/j2nb eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'J2NB'
            edit OBSFILE_HPCF 'j2nb'
          task get_j2eo
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc00/j2eo eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'J2EO'
            edit OBSFILE_HPCF 'j2eo'
          task get_asc2
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc00/asc2 eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'ASC2'
            edit OBSFILE_HPCF 'ascat_l2'
          task get_asel
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc00/asel eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'ASEL'
            edit OBSFILE_HPCF 'ascat_ears'
          task get_ascs
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc00/ascs eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'ASCS'
            edit OBSFILE_HPCF 'ascat_sm'
          task atovairs
            trigger get_atov eq complete and get_airs eq complete and get_ateu eq complete and get_amsa eq complete and get_mhsa eq complete and get_hirs eq complete and get_atap eq complete
          task snowlinks
            trigger get_nonbin eq complete and get_ers1_wi eq complete and get_rtovs eq complete and get_snow eq complete
          task ro_combine
            trigger get_rado eq complete and get_gras eq complete
          task reo3_combine
            trigger get_sbuv eq complete and get_g2to eq complete
          task bufr_combine
            trigger get_nonbin eq complete and get_modw eq complete and get_avhn eq complete
          task jason2_combine
            trigger get_ers1_ra eq complete and get_j2nb eq complete and get_j2eo eq complete
          task ascat_combine
            trigger get_asc2 eq complete and get_asel eq complete and get_ascs eq complete
        endfamily
        family getplus6
          edit EMOS_BASE '06'
          inlimit /e_36r1/limits:obs_get
          inlimit /limits:obs_get
          task get_nonbin
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc06/nonbin eq complete
            edit OBSFILE_OBSHOST 'BUFR'
            edit OBSFILE_HPCF 'bufr'
          task get_ers1_wa
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc06/ers1_wa eq complete
            edit OBSFILE_OBSHOST 'UWAX'
            edit OBSFILE_HPCF 'wavescat'
          task get_eas1
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc06/eas1 eq complete
            edit OBSFILE_OBSHOST 'EAS1'
            edit OBSFILE_HPCF 'wavesar'
          task get_ers1_wi
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc06/ers1_wi eq complete
            edit OBSFILE_OBSHOST 'UWIX'
            edit OBSFILE_HPCF 'scat'
          task get_ers1_ra
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc06/ers1_ra eq complete
            edit OBSFILE_OBSHOST 'URAX'
            edit OBSFILE_HPCF 'ralt'
          task get_rtovs
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc06/rtovs eq complete
            edit OBSFILE_OBSHOST 'TOVH'
            edit OBSFILE_HPCF 'tovh'
          task get_ssbt
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc06/ssbt eq complete
            edit OBSFILE_OBSHOST 'SSBT'
            edit OBSFILE_HPCF 'ssbt'
          task get_sst
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc06/sst/sst_fdb_c1b eq complete
            edit OBSFILE_OBSHOST 'nil'
            edit OBSFILE_HPCF 'nil'
          task get_atov
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc06/atov eq complete
            edit OBSFILE_OBSHOST 'ATOV'
            edit OBSFILE_HPCF 'atov'
          task get_grad
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc06/grad eq complete
            edit OBSFILE_OBSHOST 'GRAD'
            edit OBSFILE_HPCF 'grad'
          task get_stor
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc06/stor eq complete
            edit OBSFILE_OBSHOST 'STOR'
            edit OBSFILE_HPCF 'storm'
          task get_trmm
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc06/trmm eq complete
            edit OBSFILE_OBSHOST 'TRMM'
            edit OBSFILE_HPCF 'tmi'
          task get_qsca
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc06/qsca eq complete
            edit OBSFILE_OBSHOST 'QSCA'
            edit OBSFILE_HPCF 'qsca'
          task get_sbuv
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc06/sbuv eq complete
            edit OBSFILE_OBSHOST 'SBUV'
            edit OBSFILE_HPCF 'reo3'
          task get_airs
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc06/airs eq complete
            edit OBSFILE_OBSHOST 'AIRS'
            edit OBSFILE_HPCF 'airs'
          task get_ice
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc06/ice/ice_fdb_c1b eq complete
            edit OBSFILE_OBSHOST 'nil'
            edit OBSFILE_HPCF 'nil'
          task get_snow
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc06/snow eq complete
            edit OBSFILE_OBSHOST 'SNOW'
            edit OBSFILE_HPCF 'imssnow'
          task get_ateu
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc06/ateu eq complete
            edit OBSFILE_OBSHOST 'ATEU'
            edit OBSFILE_HPCF 'amsu_ears'
          task get_bice
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc06/bice eq complete
            edit OBSFILE_OBSHOST 'BICE'
            edit OBSFILE_HPCF 'baltic_ci'
          task get_amsa
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc06/amsa eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'AMSA'
            edit OBSFILE_HPCF 'amsa'
          task get_mhsa
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc06/mhsa eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'MHSA'
            edit OBSFILE_HPCF 'mhsa'
          task get_hirs
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc06/hirs eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'HIRS'
            edit OBSFILE_HPCF 'hirs'
          task get_amsr
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc06/amsr eq complete
            edit OBSFILE_OBSHOST 'AMSR'
            edit OBSFILE_HPCF 'amsre_aq'
          task get_rado
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc06/rado eq complete
            edit OBSFILE_OBSHOST 'RADO'
            edit OBSFILE_HPCF 'gpsro'
          task get_smin
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc06/smin eq complete
            edit OBSFILE_OBSHOST 'SMIN'
            edit OBSFILE_HPCF 'ssmisbt'
          task get_iasi
            late -s +00:02 -c +00:07
            trigger /pp2/main/extractions/dc06/iasi eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'IASI'
            edit OBSFILE_HPCF 'iasi'
          task get_gras
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc06/gras eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'GRAS'
            edit OBSFILE_HPCF 'gras'
          task get_modw
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc06/modw eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'MODW'
            edit OBSFILE_HPCF 'dbmodiswinds'
          task get_avhn
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc06/avhn eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'AVHN'
            edit OBSFILE_HPCF 'avhrrwinds'
          task get_atap
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc06/atap eq complete
            edit OBSFILE_OBSHOST 'ATAP'
            edit OBSFILE_HPCF 'amsu_pacrars'
          task get_g2to
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc06/g2to eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'G2TO'
            edit OBSFILE_HPCF 'reo3_gome2'
          task get_grst
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc06/grst eq complete
            edit OBSFILE_OBSHOST 'GRST'
            edit OBSFILE_HPCF 'ostia'
          task get_emer
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc06/emer eq complete
            edit OBSFILE_OBSHOST 'EMER'
            edit OBSFILE_HPCF 'envisat_meris'
          task get_j2nb
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc06/j2nb eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'J2NB'
            edit OBSFILE_HPCF 'j2nb'
          task get_j2eo
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc06/j2eo eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'J2EO'
            edit OBSFILE_HPCF 'j2eo'
          task get_asc2
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc06/asc2 eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'ASC2'
            edit OBSFILE_HPCF 'ascat_l2'
          task get_asel
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc06/asel eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'ASEL'
            edit OBSFILE_HPCF 'ascat_ears'
          task get_ascs
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc06/ascs eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'ASCS'
            edit OBSFILE_HPCF 'ascat_sm'
          task atovairs
            trigger get_atov eq complete and get_airs eq complete and get_ateu eq complete and get_amsa eq complete and get_mhsa eq complete and get_hirs eq complete and get_atap eq complete
          task snowlinks
            trigger get_nonbin eq complete and get_ers1_wi eq complete and get_rtovs eq complete and get_snow eq complete
          task ro_combine
            trigger get_rado eq complete and get_gras eq complete
          task reo3_combine
            trigger get_sbuv eq complete and get_g2to eq complete
          task bufr_combine
            trigger get_nonbin eq complete and get_modw eq complete and get_avhn eq complete
          task jason2_combine
            trigger get_ers1_ra eq complete and get_j2nb eq complete and get_j2eo eq complete
          task ascat_combine
            trigger get_asc2 eq complete and get_asel eq complete and get_ascs eq complete
        endfamily
        task fetchmars
          trigger ./get eq complete and ./getplus6 eq complete
        family prepare_obs
          trigger ./get eq complete and ./getplus6 eq complete
          task preobs
            edit mode 'not1crad'
          task premwimg
          task preobs_wave
            edit mode 'wave'
          task pre1crad_prepare
            edit mode '1crad'
          task pre1crad_airs
            trigger pre1crad_prepare eq complete
          family pre1crad_iasi
            trigger pre1crad_prepare eq complete
            edit CPUTIME '1500'
            edit MAXTASK '16'
            task pre1crad_iasi_split_1
              edit TASKNO '1'
            task pre1crad_iasi_split_2
              edit TASKNO '2'
            task pre1crad_iasi_1
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '1'
            task pre1crad_iasi_2
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '2'
            task pre1crad_iasi_3
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '3'
            task pre1crad_iasi_4
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '4'
            task pre1crad_iasi_5
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '5'
            task pre1crad_iasi_6
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '6'
            task pre1crad_iasi_7
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '7'
            task pre1crad_iasi_8
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '8'
            task pre1crad_iasi_9
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '9'
            task pre1crad_iasi_10
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '10'
            task pre1crad_iasi_11
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '11'
            task pre1crad_iasi_12
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '12'
            task pre1crad_iasi_13
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '13'
            task pre1crad_iasi_14
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '14'
            task pre1crad_iasi_15
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '15'
            task pre1crad_iasi_16
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '16'
          endfamily
          task pre1crad_hirs
            trigger pre1crad_prepare eq complete
          task pre1crad_msu
            trigger pre1crad_prepare eq complete
          task pre1crad_ssu
            trigger pre1crad_prepare eq complete
          task pre1crad_amsua
            trigger pre1crad_prepare eq complete
          task pre1crad_amsub
            trigger pre1crad_prepare eq complete
          task pre1crad_mhs
            trigger pre1crad_prepare eq complete
          task pregeos
          task prereo3
          task prescat
        endfamily
        family reservation
          late -c +00:03
          trigger ./get/get_stor eq complete
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
          inlimit /e_36r1/limits:reservation
          task adv_res
            late -s +00:03 -c +00:02
            edit RESERVE_FOR '30'
            edit RESERVE_IN '30'
            edit RESERVED_NODES '48'
            edit RESERVATION_LABEL 'emos00dcA'
            label info ""
        endfamily
      endfamily
      family an
        trigger ./obs/fetchmars eq complete and ./obs/prepare_obs eq complete
        edit FSFAMILY 'an'
        edit SMSFILES '/home/ma/emos/def/e_36r1/an'
        edit PERIOD_4D '12'
        edit WINDOW_LENGTH_4D '12'
        edit WINDOW_OFFSET_4D '3'
        edit ACCOUNT 'oedcdaan'
        family makeodb
          edit NPES '16'
          task cleanodb
          family bufr2odb
            trigger cleanodb eq complete
            edit THREADS '1'
            edit NPOOLS '%ANNPES%'
            edit THINFAC '1'
            edit SENSOR '-1'
            edit REBALANCE '1'
            edit SUBTYPES_INCLUDE ''
            edit SUBTYPES_EXCLUDE ''
            task b2o_conv
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_EXCLUDE '85 82 83 87'
              edit USER_PRIORITY '90'
            task b2o_surf_conv
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_EXCLUDE '142 143 144 145 85 82 83 87'
            task b2o_airs
              late -s +00:05 -c +00:15
              edit THREADS '2'
              edit SENSOR '11'
              edit SUBTYPES_INCLUDE '57'
              edit NPOOLS '$((2* %ANNPES%))'
            task b2o_iasi
              late -s +00:05 -c +00:15
              edit SENSOR '993'
              edit SUBTYPES_INCLUDE '240'
              edit THREADS '4'
              edit NPOOLS '$((2* %ANNPES%))'
            task b2o_hirs
              late -s +00:05 -c +00:15
              edit SENSOR '0'
              edit SUBTYPES_INCLUDE '54 55'
              edit NPOOLS '$((2* %ANNPES%))'
            task b2o_msu
              late -s +00:05 -c +00:15
              edit SENSOR '1'
              edit SUBTYPES_INCLUDE '54 55'
            task b2o_ssu
              late -s +00:05 -c +00:15
              edit SENSOR '2'
              edit SUBTYPES_INCLUDE '54 55'
            task b2o_amsua
              late -s +00:05 -c +00:15
              edit SENSOR '3'
              edit SUBTYPES_INCLUDE '54 55'
            task b2o_amsub
              late -s +00:05 -c +00:15
              edit SENSOR '4'
              edit SUBTYPES_INCLUDE '54 55'
            task b2o_mhs
              late -s +00:05 -c +00:15
              edit SENSOR '11'
              edit SUBTYPES_INCLUDE '54 55'
            task b2o_satob
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '82 83 87'
            task b2o_reo3
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '205 206 207 208 215'
            task b2o_ssmi
              late -s +00:05 -c +00:15
              edit SENSOR '6'
              edit SUBTYPES_INCLUDE '55 127'
            task b2o_ssmis
              late -s +00:05 -c +00:15
              edit SENSOR '10'
              edit SUBTYPES_INCLUDE '49'
            task b2o_amsre
              late -s +00:05 -c +00:15
              edit SENSOR '17'
              edit SUBTYPES_INCLUDE '59'
            task b2o_tmi
              late -s +00:05 -c +00:15
              edit SENSOR '9'
              edit SUBTYPES_INCLUDE '129'
            task b2o_scatt
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '122 137 139'
            task b2o_geos
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '189 89 76 77 78'
            task b2o_gpsro
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '250'
            task b2o_meris
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '212'
          endfamily
          task mergeodb
            trigger bufr2odb eq complete
          task mergeodb_surf
            trigger ./bufr2odb/b2o_surf_conv eq complete
            edit NPES '16'
            edit THREADS '1'
        endfamily
        task restartodb
          trigger makeodb eq complete
        task preCleanFDB
        task vardata
          event hires
        task fetcherr
        task lowres
          trigger vardata:hires
          edit THREADS '4'
          edit ININPES '32'
        family surf_anal
          trigger makeodb eq complete and restartodb eq complete and preCleanFDB eq complete and vardata:hires
          edit SMSFILES '/home/ma/emos/def/e_36r1/an'
          edit FSFAMILY 'surf_anal'
          task t2ana
            edit mode 't2m'
          task rh2ana
            trigger t2ana eq complete
            edit mode 'rh2m'
          task snow
            trigger rh2ana eq complete
            edit mode 'snow'
          task sst
            edit THREADS '4'
          task slwet
            trigger sst eq complete and snow eq complete
        endfamily
        family 4dvar
          trigger makeodb eq complete and restartodb eq complete and vardata eq complete and fetcherr eq complete and lowres eq complete and preCleanFDB eq complete
          edit THREADS '8'
          edit VAR_DELAY '00'
          edit EMOS_SLEEP '00'
          family uptraj_0
            edit uptraj '0'
            task ifstraj
              late -s +00:02
              edit mode 'traj'
              edit USER_PRIORITY '91'
              edit SMSTRIES '1'
            task ifsmin
              late -s +00:02
              trigger ifstraj eq complete
              edit mode 'min'
              edit USER_PRIORITY '92'
            task ifstsave
              trigger ifstraj eq complete
              edit THREADS '1'
              edit USER_PRIORITY '93'
          endfamily
          family uptraj_1
            trigger uptraj_0 eq complete
            edit uptraj '1'
            task ifstraj
              late -s +00:02
              edit mode 'traj'
              edit USER_PRIORITY '93'
            task ifsmin
              late -s +00:02
              trigger ifstraj eq complete
              edit mode 'min'
              edit USER_PRIORITY '94'
            task ifstsave
              trigger ifstraj eq complete
              edit THREADS '1'
              edit USER_PRIORITY '95'
          endfamily
          family uptraj_2
            trigger uptraj_1 eq complete
            edit uptraj '2'
            task ifstraj
              late -s +00:02
              edit mode 'traj'
              edit USER_PRIORITY '95'
            task ifsmin
              late -s +00:02
              trigger ifstraj eq complete
              edit mode 'min'
              edit USER_PRIORITY '96'
            task ifstsave
              trigger ifstraj eq complete
              edit THREADS '1'
              edit USER_PRIORITY '97'
          endfamily
          task ifstraj
            late -s +00:02
            trigger uptraj_2 eq complete
            edit uptraj '3'
            edit mode 'traj'
            edit USER_PRIORITY '97'
            event finaltraj
            event finalwave
        endfamily
        task forceinv
          trigger ./4dvar/ifstraj:finaltraj and ./surf_anal eq complete
        task emiskf
          trigger ./4dvar/uptraj_0/ifstraj eq complete
          edit THREADS '1'
        task update_psbias
          trigger ./4dvar/ifstraj:finalwave
        task update_rstrhbias
          trigger ./4dvar/ifstraj:finalwave
        task dcda_to_da
          trigger ./4dvar eq complete and surf_anal eq complete
          edit EMOS_TYPE 'an'
          edit OPHOST 'c1b'
        family reservation1
          late -c +00:03
          trigger ./4dvar/uptraj_0/ifstraj eq active or ./4dvar/uptraj_0/ifstraj eq complete
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
          inlimit /e_36r1/limits:reservation
          task reserve_cpu
            late -s +00:02 -c +00:02
            edit RESERVED_CPUS '3328'
            edit RESERVE_FOR '65'
            label info ""
          task rm_res
            late -s +00:02 -c +00:02
            edit RESERVATION_LABEL 'emos00'
            label info ""
          task set_class
            late -s +00:02 -c +00:02
            edit RESERVATION_DEFINITION 'default_nodiag'
            label info ""
        endfamily
      endfamily
      family fc
        trigger an/forceinv eq complete and ( an/4dvar/ifstraj:finalwave or an/4dvar eq complete)
        edit EMOS_BASE '06'
        edit ACCOUNT 'oedcdafc'
        edit FSFAMILY 'fc'
        task getfcdata
        task model
          late -s +00:02
          trigger getfcdata eq complete
          edit FCLENGTH '15'
          edit THREADS '8'
          edit USER_PRIORITY '94'
          meter step -1 15 15
        task ddh
          trigger model eq complete
        task clean
          trigger ../an eq complete and getfcdata eq complete and model eq complete and ddh eq complete
        family reservation1
          late -c +00:03
          trigger ./getfcdata eq active or ./getfcdata eq complete
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
          inlimit /e_36r1/limits:reservation
          task reserve_cpu
            late -s +00:03 -c +00:02
            edit RESERVED_CPUS '3072'
            edit RESERVE_FOR '4'
            label info ""
        endfamily
        family reservation2
          late -c +00:03
          trigger ./model eq complete
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
          inlimit /e_36r1/limits:reservation
          task reset_cpu
            late -s +00:03 -c +00:02
            complete /e_36r1/main/12/an/vardata eq active or /e_36r1/main/12/an/vardata eq complete
            edit RESERVED_CPUS '0'
            label info ""
        endfamily
      endfamily
    endfamily
    family 06
      trigger make eq complete
      edit EMOS_BASE '06'
      edit MXUP_TRAJ '3'
      edit REINISTREAM 'DA'
      edit LAR_4V_EF 'false'
      edit LFC_DIR_FREQ 'false'
      edit WINDOW_LENGTH_4D '6'
      edit WINDOW_OFFSET_4D '3'
      edit EDFAMILY 'true'
      edit PERIOD_4D '6'
      edit FCLENGTH '9'
      edit DELTA_DAY '0'
      family obs
        edit FSFAMILY 'an'
        edit SMSFILES '/home/ma/emos/def/e_36r1/obs'
        family get
          inlimit /e_36r1/limits:obs_get
          inlimit /limits:obs_get
          task get_nonbin
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/nonbin eq complete
            edit OBSFILE_OBSHOST 'BUFR'
            edit OBSFILE_HPCF 'bufr'
          task get_ers1_wa
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/ers1_wa eq complete
            edit OBSFILE_OBSHOST 'UWAX'
            edit OBSFILE_HPCF 'wavescat'
          task get_eas1
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/eas1 eq complete
            edit OBSFILE_OBSHOST 'EAS1'
            edit OBSFILE_HPCF 'wavesar'
          task get_ers1_wi
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/ers1_wi eq complete
            edit OBSFILE_OBSHOST 'UWIX'
            edit OBSFILE_HPCF 'scat'
          task get_ers1_ra
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/ers1_ra eq complete
            edit OBSFILE_OBSHOST 'URAX'
            edit OBSFILE_HPCF 'ralt'
          task get_rtovs
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/rtovs eq complete
            edit OBSFILE_OBSHOST 'TOVH'
            edit OBSFILE_HPCF 'tovh'
          task get_ssbt
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/ssbt eq complete
            edit OBSFILE_OBSHOST 'SSBT'
            edit OBSFILE_HPCF 'ssbt'
          task get_sst
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/sst/sst_fdb_c1b eq complete
            edit OBSFILE_OBSHOST 'nil'
            edit OBSFILE_HPCF 'nil'
          task get_atov
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/atov eq complete
            edit OBSFILE_OBSHOST 'ATOV'
            edit OBSFILE_HPCF 'atov'
          task get_grad
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/grad eq complete
            edit OBSFILE_OBSHOST 'GRAD'
            edit OBSFILE_HPCF 'grad'
          task get_stor
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/stor eq complete
            edit OBSFILE_OBSHOST 'STOR'
            edit OBSFILE_HPCF 'storm'
          task get_trmm
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/trmm eq complete
            edit OBSFILE_OBSHOST 'TRMM'
            edit OBSFILE_HPCF 'tmi'
          task get_qsca
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/qsca eq complete
            edit OBSFILE_OBSHOST 'QSCA'
            edit OBSFILE_HPCF 'qsca'
          task get_sbuv
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/sbuv eq complete
            edit OBSFILE_OBSHOST 'SBUV'
            edit OBSFILE_HPCF 'reo3'
          task get_airs
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/airs eq complete
            edit OBSFILE_OBSHOST 'AIRS'
            edit OBSFILE_HPCF 'airs'
          task get_ice
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/ice/ice_fdb_c1b eq complete
            edit OBSFILE_OBSHOST 'nil'
            edit OBSFILE_HPCF 'nil'
          task get_snow
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/snow eq complete
            edit OBSFILE_OBSHOST 'SNOW'
            edit OBSFILE_HPCF 'imssnow'
          task get_ateu
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/ateu eq complete
            edit OBSFILE_OBSHOST 'ATEU'
            edit OBSFILE_HPCF 'amsu_ears'
          task get_bice
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/bice eq complete
            edit OBSFILE_OBSHOST 'BICE'
            edit OBSFILE_HPCF 'baltic_ci'
          task get_amsa
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex06/amsa eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'AMSA'
            edit OBSFILE_HPCF 'amsa'
          task get_mhsa
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex06/mhsa eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'MHSA'
            edit OBSFILE_HPCF 'mhsa'
          task get_hirs
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex06/hirs eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'HIRS'
            edit OBSFILE_HPCF 'hirs'
          task get_amsr
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/amsr eq complete
            edit OBSFILE_OBSHOST 'AMSR'
            edit OBSFILE_HPCF 'amsre_aq'
          task get_rado
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/rado eq complete
            edit OBSFILE_OBSHOST 'RADO'
            edit OBSFILE_HPCF 'gpsro'
          task get_smin
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/smin eq complete
            edit OBSFILE_OBSHOST 'SMIN'
            edit OBSFILE_HPCF 'ssmisbt'
          task get_iasi
            late -s +00:02 -c +00:07
            trigger /pp2/main/extractions/ex06/iasi eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'IASI'
            edit OBSFILE_HPCF 'iasi'
          task get_gras
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex06/gras eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'GRAS'
            edit OBSFILE_HPCF 'gras'
          task get_modw
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex06/modw eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'MODW'
            edit OBSFILE_HPCF 'dbmodiswinds'
          task get_avhn
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex06/avhn eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'AVHN'
            edit OBSFILE_HPCF 'avhrrwinds'
          task get_atap
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/atap eq complete
            edit OBSFILE_OBSHOST 'ATAP'
            edit OBSFILE_HPCF 'amsu_pacrars'
          task get_g2to
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex06/g2to eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'G2TO'
            edit OBSFILE_HPCF 'reo3_gome2'
          task get_grst
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/grst eq complete
            edit OBSFILE_OBSHOST 'GRST'
            edit OBSFILE_HPCF 'ostia'
          task get_emer
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/emer eq complete
            edit OBSFILE_OBSHOST 'EMER'
            edit OBSFILE_HPCF 'envisat_meris'
          task get_j2nb
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex06/j2nb eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'J2NB'
            edit OBSFILE_HPCF 'j2nb'
          task get_j2eo
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex06/j2eo eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'J2EO'
            edit OBSFILE_HPCF 'j2eo'
          task get_asc2
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex06/asc2 eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'ASC2'
            edit OBSFILE_HPCF 'ascat_l2'
          task get_asel
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex06/asel eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'ASEL'
            edit OBSFILE_HPCF 'ascat_ears'
          task get_ascs
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex06/ascs eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'ASCS'
            edit OBSFILE_HPCF 'ascat_sm'
          task atovairs
            trigger get_atov eq complete and get_airs eq complete and get_ateu eq complete and get_amsa eq complete and get_mhsa eq complete and get_hirs eq complete and get_atap eq complete
          task snowlinks
            trigger get_nonbin eq complete and get_ers1_wi eq complete and get_rtovs eq complete and get_snow eq complete
          task ro_combine
            trigger get_rado eq complete and get_gras eq complete
          task reo3_combine
            trigger get_sbuv eq complete and get_g2to eq complete
          task bufr_combine
            trigger get_nonbin eq complete and get_modw eq complete and get_avhn eq complete
          task jason2_combine
            trigger get_ers1_ra eq complete and get_j2nb eq complete and get_j2eo eq complete
          task ascat_combine
            trigger get_asc2 eq complete and get_asel eq complete and get_ascs eq complete
        endfamily
        task fetchmars
          trigger ./get eq complete
        family prepare_obs
          trigger ./get eq complete
          task preobs
            edit mode 'not1crad'
          task premwimg
          task preobs_wave
            edit mode 'wave'
          task pre1crad_prepare
            edit mode '1crad'
          task pre1crad_airs
            trigger pre1crad_prepare eq complete
          family pre1crad_iasi
            trigger pre1crad_prepare eq complete
            edit CPUTIME '1500'
            edit MAXTASK '16'
            task pre1crad_iasi_split_1
              edit TASKNO '1'
            task pre1crad_iasi_split_2
              edit TASKNO '2'
            task pre1crad_iasi_1
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '1'
            task pre1crad_iasi_2
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '2'
            task pre1crad_iasi_3
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '3'
            task pre1crad_iasi_4
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '4'
            task pre1crad_iasi_5
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '5'
            task pre1crad_iasi_6
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '6'
            task pre1crad_iasi_7
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '7'
            task pre1crad_iasi_8
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '8'
            task pre1crad_iasi_9
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '9'
            task pre1crad_iasi_10
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '10'
            task pre1crad_iasi_11
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '11'
            task pre1crad_iasi_12
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '12'
            task pre1crad_iasi_13
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '13'
            task pre1crad_iasi_14
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '14'
            task pre1crad_iasi_15
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '15'
            task pre1crad_iasi_16
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '16'
          endfamily
          task pre1crad_hirs
            trigger pre1crad_prepare eq complete
          task pre1crad_msu
            trigger pre1crad_prepare eq complete
          task pre1crad_ssu
            trigger pre1crad_prepare eq complete
          task pre1crad_amsua
            trigger pre1crad_prepare eq complete
          task pre1crad_amsub
            trigger pre1crad_prepare eq complete
          task pre1crad_mhs
            trigger pre1crad_prepare eq complete
          task pregeos
          task prereo3
          task prescat
        endfamily
      endfamily
      family an
        trigger ./obs/fetchmars eq complete and ./obs/prepare_obs eq complete
        edit FSFAMILY 'an'
        edit SMSFILES '/home/ma/emos/def/e_36r1/an'
        edit ACCOUNT 'oeoperan'
        family makeodb
          edit NPES '16'
          task cleanodb
          family bufr2odb
            trigger cleanodb eq complete
            edit THREADS '1'
            edit NPOOLS '%ANNPES%'
            edit THINFAC '1'
            edit SENSOR '-1'
            edit REBALANCE '1'
            edit SUBTYPES_INCLUDE ''
            edit SUBTYPES_EXCLUDE ''
            task b2o_conv
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_EXCLUDE '85 82 83 87'
              edit USER_PRIORITY '90'
            task b2o_surf_conv
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_EXCLUDE '142 143 144 145 85 82 83 87'
            task b2o_airs
              late -s +00:05 -c +00:15
              edit THREADS '2'
              edit SENSOR '11'
              edit SUBTYPES_INCLUDE '57'
              edit NPOOLS '$((2* %ANNPES%))'
            task b2o_iasi
              late -s +00:05 -c +00:15
              edit SENSOR '993'
              edit SUBTYPES_INCLUDE '240'
              edit THREADS '4'
              edit NPOOLS '$((2* %ANNPES%))'
            task b2o_hirs
              late -s +00:05 -c +00:15
              edit SENSOR '0'
              edit SUBTYPES_INCLUDE '54 55'
              edit NPOOLS '$((2* %ANNPES%))'
            task b2o_msu
              late -s +00:05 -c +00:15
              edit SENSOR '1'
              edit SUBTYPES_INCLUDE '54 55'
            task b2o_ssu
              late -s +00:05 -c +00:15
              edit SENSOR '2'
              edit SUBTYPES_INCLUDE '54 55'
            task b2o_amsua
              late -s +00:05 -c +00:15
              edit SENSOR '3'
              edit SUBTYPES_INCLUDE '54 55'
            task b2o_amsub
              late -s +00:05 -c +00:15
              edit SENSOR '4'
              edit SUBTYPES_INCLUDE '54 55'
            task b2o_mhs
              late -s +00:05 -c +00:15
              edit SENSOR '11'
              edit SUBTYPES_INCLUDE '54 55'
            task b2o_satob
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '82 83 87'
            task b2o_reo3
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '205 206 207 208 215'
            task b2o_ssmi
              late -s +00:05 -c +00:15
              edit SENSOR '6'
              edit SUBTYPES_INCLUDE '55 127'
            task b2o_ssmis
              late -s +00:05 -c +00:15
              edit SENSOR '10'
              edit SUBTYPES_INCLUDE '49'
            task b2o_amsre
              late -s +00:05 -c +00:15
              edit SENSOR '17'
              edit SUBTYPES_INCLUDE '59'
            task b2o_tmi
              late -s +00:05 -c +00:15
              edit SENSOR '9'
              edit SUBTYPES_INCLUDE '129'
            task b2o_scatt
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '122 137 139'
            task b2o_geos
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '189 89 76 77 78'
            task b2o_gpsro
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '250'
            task b2o_meris
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '212'
          endfamily
          task mergeodb
            trigger bufr2odb eq complete
          task mergeodb_surf
            trigger ./bufr2odb/b2o_surf_conv eq complete
            edit NPES '16'
            edit THREADS '1'
        endfamily
        task restartodb
          trigger makeodb eq complete
        task preCleanFDB
        task vardata
          trigger ../../00dc/fc/model:step ge 6 or ../../00dc/fc/model eq complete
          event hires
        task fetcherr
          trigger ../../00dc/an eq complete
        task lowres
          trigger vardata:hires
          edit THREADS '4'
          edit ININPES '32'
        family surf_anal
          trigger makeodb eq complete and restartodb eq complete and preCleanFDB eq complete and vardata:hires
          edit SMSFILES '/home/ma/emos/def/e_36r1/an'
          edit FSFAMILY 'surf_anal'
          task t2ana
            edit mode 't2m'
          task rh2ana
            trigger t2ana eq complete
            edit mode 'rh2m'
          task snow
            trigger rh2ana eq complete
            edit mode 'snow'
          task sst
            edit THREADS '4'
          task slwet
            trigger sst eq complete and snow eq complete
        endfamily
        family 4dvar
          trigger makeodb eq complete and restartodb eq complete and vardata eq complete and fetcherr eq complete and lowres eq complete and preCleanFDB eq complete
          edit THREADS '8'
          edit VAR_DELAY '00'
          edit EMOS_SLEEP '00'
          family uptraj_0
            edit uptraj '0'
            task ifstraj
              late -s +00:02
              edit mode 'traj'
              edit USER_PRIORITY '91'
              edit SMSTRIES '1'
            task ifsmin
              late -s +00:02
              trigger ifstraj eq complete
              edit mode 'min'
              edit USER_PRIORITY '92'
            task ifstsave
              trigger ifstraj eq complete
              edit THREADS '1'
              edit USER_PRIORITY '93'
          endfamily
          family uptraj_1
            trigger uptraj_0 eq complete
            edit uptraj '1'
            task ifstraj
              late -s +00:02
              edit mode 'traj'
              edit USER_PRIORITY '93'
            task ifsmin
              late -s +00:02
              trigger ifstraj eq complete
              edit mode 'min'
              edit USER_PRIORITY '94'
            task ifstsave
              trigger ifstraj eq complete
              edit THREADS '1'
              edit USER_PRIORITY '95'
          endfamily
          family uptraj_2
            trigger uptraj_1 eq complete
            edit uptraj '2'
            task ifstraj
              late -s +00:02
              edit mode 'traj'
              edit USER_PRIORITY '95'
            task ifsmin
              late -s +00:02
              trigger ifstraj eq complete
              edit mode 'min'
              edit USER_PRIORITY '96'
            task ifstsave
              trigger ifstraj eq complete
              edit THREADS '1'
              edit USER_PRIORITY '97'
          endfamily
          task ifstraj
            late -s +00:02
            trigger uptraj_2 eq complete
            edit uptraj '3'
            edit mode 'traj'
            edit USER_PRIORITY '97'
            event finaltraj
            event finalwave
        endfamily
        task forceinv
          trigger ./4dvar/ifstraj:finaltraj and ./surf_anal eq complete
      endfamily
      family fc
        trigger an/forceinv eq complete and ( an/4dvar/ifstraj:finalwave or an/4dvar eq complete)
        edit ACCOUNT 'oeoperfc'
        edit FSFAMILY 'fc'
        task getfcdata
        task getpersSST
          trigger getfcdata eq complete
          edit SSTLEN '9'
          edit SSTINT '1'
          edit THREADS '2'
          edit NPES '16'
          edit MEM '1000'
          edit CPUTIME '960'
        task model
          late -s +00:02
          trigger ( getfcdata eq complete) and ( getpersSST eq complete)
          edit PERSST 'true'
          edit FCLENGTH '9'
          edit THREADS '8'
          edit USER_PRIORITY '94'
          meter step -1 9 9
        task ddh
          trigger model eq complete
        task clean
          trigger ../an eq complete and getfcdata eq complete and model eq complete and ddh eq complete
      endfamily
    endfamily
    family 12bc
      edit EMOS_BASE '12'
      edit EDFAMILY 'false'
      edit INISTREAM 'DA'
      edit REINITIALIZE 'false'
      edit REINIEXPVER '0046'
      edit INIEXPVER '0045'
      edit REINIPERIOD_4D '6'
      edit REINISTREAM 'DA'
      edit REINIOFFSET '6'
      edit RUN_HOURLY_PP '0'
      edit PERIOD_4D '6'
      edit WINDOW_LENGTH_4D '6'
      edit WINDOW_OFFSET_4D '3'
      edit MXUP_TRAJ '3'
      edit PARALLEL '4'
      edit LANONLY 'false'
      edit STREAM 'SCDA'
      edit EMOS_STREAM 'SCDA'
      edit DELTA_DAY '0'
      edit ACCOUNT 'opebcpg'
      edit SMSURL 'GSCDA/12'
      edit USER_PRIORITY '91'
      family getreq
        trigger ( /e_36r1/main/12/an/4dvar eq active or /e_36r1/main/12/an/4dvar eq complete) and ( ./prod eq queued or ./prod eq complete)
        edit SMSFILES '/home/ma/emos/def/e_36r1/prod'
        edit EMOS_TYPE 'fc'
        edit PROD_TASK '1'
        edit EMOS_DOMAIN 'G'
        edit FSFAMILY 'ppdi'
        edit EMOS_PROJECT '12'
        edit SMSTRIES '1'
        edit EMOS_STREAM 'SCDA'
        task collectreq
          edit QUEUE 'diss'
          edit USER 'emos'
          edit WSHOST 'diss'
          edit QUEUE_EPILOG 'emos-exp'
          edit LOGDIR '/diss/log'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit SMSOUT '/vol/emos_nc/output'
          edit EMOS_BASE '12'
        task req
          late -s +00:03 -c +00:10
          trigger collectreq eq complete
          label info ""
          event 1 
        task collectbitmap
          trigger collectreq eq complete
          edit QUEUE 'diss'
          edit USER 'emos'
          edit WSHOST 'diss'
          edit QUEUE_EPILOG 'emos-exp'
          edit LOGDIR '/diss/log'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit SMSOUT '/vol/emos_nc/output'
        task bitmap
          late -s +00:03 -c +00:10
          trigger collectbitmap eq complete and ( req:001 or req eq complete)
          label info ""
          event 1 
      endfamily
      family prod
        trigger ( getreq/req:001 or getreq/req eq complete)
        edit SMSFILES '/home/ma/emos/def/e_36r1/prod'
        edit EMOS_DOMAIN 'G'
        edit EMOS_TYPE 'fc'
        edit FSFAMILY 'ppdi'
        edit PROD_TASK '1'
        edit EMOS_PROJECT '12'
        edit SMSURLBASE 'http://diss.ecmwf.int/do/monitoring/summary'
        family an
          trigger /e_36r1/main/12/an/surf_anal eq complete and ( /e_36r1/main/12/an/4dvar/ifstraj:finalwave or /e_36r1/main/12/an/4dvar/ifstraj eq complete) and 0 eq 0
          edit EMOS_TYPE 'an'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ../090/prodgen eq complete and ../../prod eq queued and /e_36r1/main/12/prod eq complete
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 000
          trigger ( /e_36r1/main/12/fc/model:step ge 000 or /e_36r1/main/12/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0000'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/12/fc/model:step ge 12 or /e_36r1/main/12/fc/model eq complete) and ( ../an/backupprod eq complete or ../an/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 003
          trigger ( /e_36r1/main/12/fc/model:step ge 003 or /e_36r1/main/12/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0003'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/12/fc/model:step ge 15 or /e_36r1/main/12/fc/model eq complete) and ( ../000/backupprod eq complete or ../000/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 006
          trigger ( /e_36r1/main/12/fc/model:step ge 006 or /e_36r1/main/12/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0006'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/12/fc/model:step ge 18 or /e_36r1/main/12/fc/model eq complete) and ( ../003/backupprod eq complete or ../003/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 009
          trigger ( /e_36r1/main/12/fc/model:step ge 009 or /e_36r1/main/12/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0009'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/12/fc/model:step ge 21 or /e_36r1/main/12/fc/model eq complete) and ( ../006/backupprod eq complete or ../006/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 012
          trigger ( /e_36r1/main/12/fc/model:step ge 012 or /e_36r1/main/12/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0012'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/12/fc/model:step ge 24 or /e_36r1/main/12/fc/model eq complete) and ( ../009/backupprod eq complete or ../009/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 015
          trigger ( /e_36r1/main/12/fc/model:step ge 015 or /e_36r1/main/12/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0015'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/12/fc/model:step ge 27 or /e_36r1/main/12/fc/model eq complete) and ( ../012/backupprod eq complete or ../012/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 018
          trigger ( /e_36r1/main/12/fc/model:step ge 018 or /e_36r1/main/12/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0018'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/12/fc/model:step ge 30 or /e_36r1/main/12/fc/model eq complete) and ( ../015/backupprod eq complete or ../015/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 021
          trigger ( /e_36r1/main/12/fc/model:step ge 021 or /e_36r1/main/12/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0021'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/12/fc/model:step ge 33 or /e_36r1/main/12/fc/model eq complete) and ( ../018/backupprod eq complete or ../018/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 024
          trigger ( /e_36r1/main/12/fc/model:step ge 024 or /e_36r1/main/12/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0024'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/12/fc/model:step ge 36 or /e_36r1/main/12/fc/model eq complete) and ( ../021/backupprod eq complete or ../021/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 027
          trigger ( /e_36r1/main/12/fc/model:step ge 027 or /e_36r1/main/12/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0027'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/12/fc/model:step ge 39 or /e_36r1/main/12/fc/model eq complete) and ( ../024/backupprod eq complete or ../024/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 030
          trigger ( /e_36r1/main/12/fc/model:step ge 030 or /e_36r1/main/12/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0030'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/12/fc/model:step ge 42 or /e_36r1/main/12/fc/model eq complete) and ( ../027/backupprod eq complete or ../027/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 033
          trigger ( /e_36r1/main/12/fc/model:step ge 033 or /e_36r1/main/12/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0033'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/12/fc/model:step ge 45 or /e_36r1/main/12/fc/model eq complete) and ( ../030/backupprod eq complete or ../030/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 036
          trigger ( /e_36r1/main/12/fc/model:step ge 036 or /e_36r1/main/12/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0036'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/12/fc/model:step ge 48 or /e_36r1/main/12/fc/model eq complete) and ( ../033/backupprod eq complete or ../033/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 039
          trigger ( /e_36r1/main/12/fc/model:step ge 039 or /e_36r1/main/12/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0039'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/12/fc/model:step ge 51 or /e_36r1/main/12/fc/model eq complete) and ( ../036/backupprod eq complete or ../036/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 042
          trigger ( /e_36r1/main/12/fc/model:step ge 042 or /e_36r1/main/12/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0042'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/12/fc/model:step ge 54 or /e_36r1/main/12/fc/model eq complete) and ( ../039/backupprod eq complete or ../039/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 045
          trigger ( /e_36r1/main/12/fc/model:step ge 045 or /e_36r1/main/12/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0045'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/12/fc/model:step ge 57 or /e_36r1/main/12/fc/model eq complete) and ( ../042/backupprod eq complete or ../042/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 048
          trigger ( /e_36r1/main/12/fc/model:step ge 048 or /e_36r1/main/12/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0048'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/12/fc/model:step ge 60 or /e_36r1/main/12/fc/model eq complete) and ( ../045/backupprod eq complete or ../045/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 051
          trigger ( /e_36r1/main/12/fc/model:step ge 051 or /e_36r1/main/12/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0051'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/12/fc/model:step ge 63 or /e_36r1/main/12/fc/model eq complete) and ( ../048/backupprod eq complete or ../048/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 054
          trigger ( /e_36r1/main/12/fc/model:step ge 054 or /e_36r1/main/12/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0054'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/12/fc/model:step ge 66 or /e_36r1/main/12/fc/model eq complete) and ( ../051/backupprod eq complete or ../051/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 057
          trigger ( /e_36r1/main/12/fc/model:step ge 057 or /e_36r1/main/12/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0057'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/12/fc/model:step ge 69 or /e_36r1/main/12/fc/model eq complete) and ( ../054/backupprod eq complete or ../054/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 060
          trigger ( /e_36r1/main/12/fc/model:step ge 060 or /e_36r1/main/12/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0060'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/12/fc/model:step ge 72 or /e_36r1/main/12/fc/model eq complete) and ( ../057/backupprod eq complete or ../057/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 063
          trigger ( /e_36r1/main/12/fc/model:step ge 063 or /e_36r1/main/12/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0063'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/12/fc/model:step ge 75 or /e_36r1/main/12/fc/model eq complete) and ( ../060/backupprod eq complete or ../060/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 066
          trigger ( /e_36r1/main/12/fc/model:step ge 066 or /e_36r1/main/12/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0066'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/12/fc/model:step ge 78 or /e_36r1/main/12/fc/model eq complete) and ( ../063/backupprod eq complete or ../063/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 069
          trigger ( /e_36r1/main/12/fc/model:step ge 069 or /e_36r1/main/12/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0069'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/12/fc/model:step ge 81 or /e_36r1/main/12/fc/model eq complete) and ( ../066/backupprod eq complete or ../066/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 072
          trigger ( /e_36r1/main/12/fc/model:step ge 072 or /e_36r1/main/12/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0072'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/12/fc/model:step ge 84 or /e_36r1/main/12/fc/model eq complete) and ( ../069/backupprod eq complete or ../069/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 075
          trigger ( /e_36r1/main/12/fc/model:step ge 075 or /e_36r1/main/12/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0075'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/12/fc/model:step ge 87 or /e_36r1/main/12/fc/model eq complete) and ( ../072/backupprod eq complete or ../072/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 078
          trigger ( /e_36r1/main/12/fc/model:step ge 078 or /e_36r1/main/12/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0078'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/12/fc/model:step ge 90 or /e_36r1/main/12/fc/model eq complete) and ( ../075/backupprod eq complete or ../075/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 081
          trigger ( /e_36r1/main/12/fc/model:step ge 081 or /e_36r1/main/12/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0081'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
        endfamily
        family 084
          trigger ( /e_36r1/main/12/fc/model:step ge 084 or /e_36r1/main/12/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0084'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
        endfamily
        family 087
          trigger ( /e_36r1/main/12/fc/model:step ge 087 or /e_36r1/main/12/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0087'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
        endfamily
        family 090
          trigger ( /e_36r1/main/12/fc/model:step ge 090 or /e_36r1/main/12/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0090'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
        endfamily
      endfamily
      family prod2diss
        edit USER_PRIORITY '90'
        edit SMSURLBASE 'http://diss.ecmwf.int/do/monitoring/summary'
        edit SMSFILES '/home/ma/emos/def/e_36r1/prod'
        task operation_is_late
          event yes
        family oper
          edit EMOS_DOMAIN 'G'
          edit EMOS_TYPE 'fc'
          edit FSFAMILY 'ppdi'
          edit PROD_TASK '1'
          edit USER_PRIORITY '92'
          edit SMSURL 'GSCDA/12'
          edit EMOS_PROJECT '12'
          family an
            task pdb
              trigger ../../../prod/an/prodgen eq complete
              edit EMOS_TYPE 'an'
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete
              edit EMOS_TYPE 'an'
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../090/q2diss eq complete and ../../../prod/an/backupprod eq complete
              edit EMOS_TYPE 'an'
              label info ""
              event 1 
          endfamily
          family 000
            edit EMOS_TIME_STEP_H '0000'
            task pdb
              trigger ../../../prod/000/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ../an/q2diss eq complete
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
          endfamily
          family 003
            edit EMOS_TIME_STEP_H '0003'
            task pdb
              trigger ../../../prod/003/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../000/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/003/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 006
            edit EMOS_TIME_STEP_H '0006'
            task pdb
              trigger ../../../prod/006/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../003/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/006/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 009
            edit EMOS_TIME_STEP_H '0009'
            task pdb
              trigger ../../../prod/009/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../006/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/009/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 012
            edit EMOS_TIME_STEP_H '0012'
            task pdb
              trigger ../../../prod/012/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../009/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/012/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 015
            edit EMOS_TIME_STEP_H '0015'
            task pdb
              trigger ../../../prod/015/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../012/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/015/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 018
            edit EMOS_TIME_STEP_H '0018'
            task pdb
              trigger ../../../prod/018/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../015/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/018/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 021
            edit EMOS_TIME_STEP_H '0021'
            task pdb
              trigger ../../../prod/021/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../018/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/021/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 024
            edit EMOS_TIME_STEP_H '0024'
            task pdb
              trigger ../../../prod/024/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../021/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/024/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 027
            edit EMOS_TIME_STEP_H '0027'
            task pdb
              trigger ../../../prod/027/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../024/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/027/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 030
            edit EMOS_TIME_STEP_H '0030'
            task pdb
              trigger ../../../prod/030/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../027/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/030/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 033
            edit EMOS_TIME_STEP_H '0033'
            task pdb
              trigger ../../../prod/033/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../030/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/033/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 036
            edit EMOS_TIME_STEP_H '0036'
            task pdb
              trigger ../../../prod/036/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../033/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/036/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 039
            edit EMOS_TIME_STEP_H '0039'
            task pdb
              trigger ../../../prod/039/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../036/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/039/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 042
            edit EMOS_TIME_STEP_H '0042'
            task pdb
              trigger ../../../prod/042/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../039/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/042/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 045
            edit EMOS_TIME_STEP_H '0045'
            task pdb
              trigger ../../../prod/045/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../042/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/045/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 048
            edit EMOS_TIME_STEP_H '0048'
            task pdb
              trigger ../../../prod/048/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../045/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/048/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 051
            edit EMOS_TIME_STEP_H '0051'
            task pdb
              trigger ../../../prod/051/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../048/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/051/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 054
            edit EMOS_TIME_STEP_H '0054'
            task pdb
              trigger ../../../prod/054/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../051/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/054/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 057
            edit EMOS_TIME_STEP_H '0057'
            task pdb
              trigger ../../../prod/057/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../054/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/057/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 060
            edit EMOS_TIME_STEP_H '0060'
            task pdb
              trigger ../../../prod/060/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../057/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/060/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 063
            edit EMOS_TIME_STEP_H '0063'
            task pdb
              trigger ../../../prod/063/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../060/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/063/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 066
            edit EMOS_TIME_STEP_H '0066'
            task pdb
              trigger ../../../prod/066/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../063/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/066/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 069
            edit EMOS_TIME_STEP_H '0069'
            task pdb
              trigger ../../../prod/069/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../066/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/069/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 072
            edit EMOS_TIME_STEP_H '0072'
            task pdb
              trigger ../../../prod/072/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../069/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/072/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 075
            edit EMOS_TIME_STEP_H '0075'
            task pdb
              trigger ../../../prod/075/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../072/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/075/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 078
            edit EMOS_TIME_STEP_H '0078'
            task pdb
              trigger ../../../prod/078/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../075/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/078/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 081
            edit EMOS_TIME_STEP_H '0081'
            task pdb
              trigger ../../../prod/081/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../078/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
          endfamily
          family 084
            edit EMOS_TIME_STEP_H '0084'
            task pdb
              trigger ../../../prod/084/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../081/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
          endfamily
          family 087
            edit EMOS_TIME_STEP_H '0087'
            task pdb
              trigger ../../../prod/087/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../084/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
          endfamily
          family 090
            edit EMOS_TIME_STEP_H '0090'
            task pdb
              trigger ../../../prod/090/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../087/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
          endfamily
          task check
            trigger ./078 eq complete
        endfamily
      endfamily
    endfamily
    family 12
      trigger make eq complete
      edit EMOS_BASE '12'
      edit MXUP_TRAJ '3'
      edit REINISTREAM 'DCDA'
      edit WINDOW_LENGTH_4D '6'
      edit WINDOW_OFFSET_4D '3'
      edit REINITIALIZE 'true'
      edit EDFAMILY 'true'
      edit PERIOD_4D '6'
      edit FCLENGTH '9'
      edit DELTA_DAY '0'
      edit ACCOUNT 'oeoperpg'
      edit SMSURL 'GOPER/12'
      edit USER_PRIORITY '90'
      family obs
        edit FSFAMILY 'an'
        edit SMSFILES '/home/ma/emos/def/e_36r1/obs'
        family get
          inlimit /e_36r1/limits:obs_get
          inlimit /limits:obs_get
          task get_nonbin
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex12/nonbin eq complete
            edit OBSFILE_OBSHOST 'BUFR'
            edit OBSFILE_HPCF 'bufr'
          task get_ers1_wa
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex12/ers1_wa eq complete
            edit OBSFILE_OBSHOST 'UWAX'
            edit OBSFILE_HPCF 'wavescat'
          task get_eas1
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex12/eas1 eq complete
            edit OBSFILE_OBSHOST 'EAS1'
            edit OBSFILE_HPCF 'wavesar'
          task get_ers1_wi
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex12/ers1_wi eq complete
            edit OBSFILE_OBSHOST 'UWIX'
            edit OBSFILE_HPCF 'scat'
          task get_ers1_ra
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex12/ers1_ra eq complete
            edit OBSFILE_OBSHOST 'URAX'
            edit OBSFILE_HPCF 'ralt'
          task get_rtovs
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex12/rtovs eq complete
            edit OBSFILE_OBSHOST 'TOVH'
            edit OBSFILE_HPCF 'tovh'
          task get_ssbt
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex12/ssbt eq complete
            edit OBSFILE_OBSHOST 'SSBT'
            edit OBSFILE_HPCF 'ssbt'
          task get_sst
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex12/sst/sst_fdb_c1b eq complete
            edit OBSFILE_OBSHOST 'nil'
            edit OBSFILE_HPCF 'nil'
          task get_atov
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex12/atov eq complete
            edit OBSFILE_OBSHOST 'ATOV'
            edit OBSFILE_HPCF 'atov'
          task get_grad
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex12/grad eq complete
            edit OBSFILE_OBSHOST 'GRAD'
            edit OBSFILE_HPCF 'grad'
          task get_stor
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex12/stor eq complete
            edit OBSFILE_OBSHOST 'STOR'
            edit OBSFILE_HPCF 'storm'
          task get_trmm
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex12/trmm eq complete
            edit OBSFILE_OBSHOST 'TRMM'
            edit OBSFILE_HPCF 'tmi'
          task get_qsca
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex12/qsca eq complete
            edit OBSFILE_OBSHOST 'QSCA'
            edit OBSFILE_HPCF 'qsca'
          task get_sbuv
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex12/sbuv eq complete
            edit OBSFILE_OBSHOST 'SBUV'
            edit OBSFILE_HPCF 'reo3'
          task get_airs
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex12/airs eq complete
            edit OBSFILE_OBSHOST 'AIRS'
            edit OBSFILE_HPCF 'airs'
          task get_ice
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex12/ice/ice_fdb_c1b eq complete
            edit OBSFILE_OBSHOST 'nil'
            edit OBSFILE_HPCF 'nil'
          task get_snow
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex12/snow eq complete
            edit OBSFILE_OBSHOST 'SNOW'
            edit OBSFILE_HPCF 'imssnow'
          task get_ateu
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex12/ateu eq complete
            edit OBSFILE_OBSHOST 'ATEU'
            edit OBSFILE_HPCF 'amsu_ears'
          task get_bice
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex12/bice eq complete
            edit OBSFILE_OBSHOST 'BICE'
            edit OBSFILE_HPCF 'baltic_ci'
          task get_amsa
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex12/amsa eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'AMSA'
            edit OBSFILE_HPCF 'amsa'
          task get_mhsa
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex12/mhsa eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'MHSA'
            edit OBSFILE_HPCF 'mhsa'
          task get_hirs
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex12/hirs eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'HIRS'
            edit OBSFILE_HPCF 'hirs'
          task get_amsr
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex12/amsr eq complete
            edit OBSFILE_OBSHOST 'AMSR'
            edit OBSFILE_HPCF 'amsre_aq'
          task get_rado
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex12/rado eq complete
            edit OBSFILE_OBSHOST 'RADO'
            edit OBSFILE_HPCF 'gpsro'
          task get_smin
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex12/smin eq complete
            edit OBSFILE_OBSHOST 'SMIN'
            edit OBSFILE_HPCF 'ssmisbt'
          task get_iasi
            late -s +00:02 -c +00:07
            trigger /pp2/main/extractions/ex12/iasi eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'IASI'
            edit OBSFILE_HPCF 'iasi'
          task get_gras
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex12/gras eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'GRAS'
            edit OBSFILE_HPCF 'gras'
          task get_modw
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex12/modw eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'MODW'
            edit OBSFILE_HPCF 'dbmodiswinds'
          task get_avhn
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex12/avhn eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'AVHN'
            edit OBSFILE_HPCF 'avhrrwinds'
          task get_atap
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex12/atap eq complete
            edit OBSFILE_OBSHOST 'ATAP'
            edit OBSFILE_HPCF 'amsu_pacrars'
          task get_g2to
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex12/g2to eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'G2TO'
            edit OBSFILE_HPCF 'reo3_gome2'
          task get_grst
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex12/grst eq complete
            edit OBSFILE_OBSHOST 'GRST'
            edit OBSFILE_HPCF 'ostia'
          task get_emer
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex12/emer eq complete
            edit OBSFILE_OBSHOST 'EMER'
            edit OBSFILE_HPCF 'envisat_meris'
          task get_j2nb
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex12/j2nb eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'J2NB'
            edit OBSFILE_HPCF 'j2nb'
          task get_j2eo
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex12/j2eo eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'J2EO'
            edit OBSFILE_HPCF 'j2eo'
          task get_asc2
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex12/asc2 eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'ASC2'
            edit OBSFILE_HPCF 'ascat_l2'
          task get_asel
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex12/asel eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'ASEL'
            edit OBSFILE_HPCF 'ascat_ears'
          task get_ascs
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex12/ascs eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'ASCS'
            edit OBSFILE_HPCF 'ascat_sm'
          task atovairs
            trigger get_atov eq complete and get_airs eq complete and get_ateu eq complete and get_amsa eq complete and get_mhsa eq complete and get_hirs eq complete and get_atap eq complete
          task snowlinks
            trigger get_nonbin eq complete and get_ers1_wi eq complete and get_rtovs eq complete and get_snow eq complete
          task ro_combine
            trigger get_rado eq complete and get_gras eq complete
          task reo3_combine
            trigger get_sbuv eq complete and get_g2to eq complete
          task bufr_combine
            trigger get_nonbin eq complete and get_modw eq complete and get_avhn eq complete
          task jason2_combine
            trigger get_ers1_ra eq complete and get_j2nb eq complete and get_j2eo eq complete
          task ascat_combine
            trigger get_asc2 eq complete and get_asel eq complete and get_ascs eq complete
        endfamily
        task fetchmars
          trigger ./get eq complete
        family prepare_obs
          trigger ./get eq complete
          task preobs
            trigger /e_36r1/main/00dc/an/update_psbias eq complete and /e_36r1/main/00dc/an/update_rstrhbias eq complete
            edit mode 'not1crad'
          task premwimg
          task preobs_wave
            edit mode 'wave'
          task pre1crad_prepare
            edit mode '1crad'
          task pre1crad_airs
            trigger pre1crad_prepare eq complete
          family pre1crad_iasi
            trigger pre1crad_prepare eq complete
            edit CPUTIME '1500'
            edit MAXTASK '16'
            task pre1crad_iasi_split_1
              edit TASKNO '1'
            task pre1crad_iasi_split_2
              edit TASKNO '2'
            task pre1crad_iasi_1
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '1'
            task pre1crad_iasi_2
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '2'
            task pre1crad_iasi_3
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '3'
            task pre1crad_iasi_4
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '4'
            task pre1crad_iasi_5
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '5'
            task pre1crad_iasi_6
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '6'
            task pre1crad_iasi_7
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '7'
            task pre1crad_iasi_8
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '8'
            task pre1crad_iasi_9
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '9'
            task pre1crad_iasi_10
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '10'
            task pre1crad_iasi_11
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '11'
            task pre1crad_iasi_12
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '12'
            task pre1crad_iasi_13
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '13'
            task pre1crad_iasi_14
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '14'
            task pre1crad_iasi_15
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '15'
            task pre1crad_iasi_16
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '16'
          endfamily
          task pre1crad_hirs
            trigger pre1crad_prepare eq complete
          task pre1crad_msu
            trigger pre1crad_prepare eq complete
          task pre1crad_ssu
            trigger pre1crad_prepare eq complete
          task pre1crad_amsua
            trigger pre1crad_prepare eq complete
          task pre1crad_amsub
            trigger pre1crad_prepare eq complete
          task pre1crad_mhs
            trigger pre1crad_prepare eq complete
          task pregeos
          task prereo3
          task prescat
        endfamily
        family reservation
          late -c +00:03
          trigger ./get/get_stor eq complete
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
          inlimit /e_36r1/limits:reservation
          task adv_res
            late -s +00:03 -c +00:02
            edit RESERVE_FOR '30'
            edit RESERVE_IN '60'
            edit RESERVED_NODES '45'
            edit RESERVATION_LABEL 'emos12XP'
            label info ""
          task adv_resA
            late -s +00:03 -c +00:02
            edit RESERVE_FOR '80'
            edit RESERVE_IN '11'
            edit RESERVED_NODES '48'
            edit RESERVATION_LABEL 'emos12A'
            label info ""
          task adv_resB
            late -s +00:03 -c +00:02
            edit RESERVE_FOR '30'
            edit RESERVE_IN '48'
            edit RESERVED_NODES '48'
            edit RESERVATION_LABEL 'emos12B'
            label info ""
          task adv_resC
            late -s +00:03 -c +00:02
            edit RESERVE_FOR '50'
            edit RESERVE_IN '55'
            edit RESERVED_NODES '132'
            edit RESERVATION_LABEL 'emos12C'
            label info ""
          task loadl_cycle_limit
            late -s +00:03 -c +00:02
            edit LOADL_CYCLE '65'
            label info ""
        endfamily
      endfamily
      family an
        trigger ./obs/fetchmars eq complete and ./obs/prepare_obs eq complete
        edit FSFAMILY 'an'
        edit SMSFILES '/home/ma/emos/def/e_36r1/an'
        edit PERIOD_4D '6'
        edit WINDOW_LENGTH_4D '6'
        edit WINDOW_OFFSET_4D '3'
        edit ACCOUNT 'oeoperan'
        family makeodb
          edit NPES '16'
          task cleanodb
          family bufr2odb
            trigger cleanodb eq complete
            edit THREADS '1'
            edit NPOOLS '%ANNPES%'
            edit THINFAC '1'
            edit SENSOR '-1'
            edit REBALANCE '1'
            edit SUBTYPES_INCLUDE ''
            edit SUBTYPES_EXCLUDE ''
            task b2o_conv
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_EXCLUDE '85 82 83 87'
              edit USER_PRIORITY '90'
            task b2o_surf_conv
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_EXCLUDE '142 143 144 145 85 82 83 87'
            task b2o_airs
              late -s +00:05 -c +00:15
              edit THREADS '2'
              edit SENSOR '11'
              edit SUBTYPES_INCLUDE '57'
              edit NPOOLS '$((2* %ANNPES%))'
            task b2o_iasi
              late -s +00:05 -c +00:15
              edit SENSOR '993'
              edit SUBTYPES_INCLUDE '240'
              edit THREADS '4'
              edit NPOOLS '$((2* %ANNPES%))'
            task b2o_hirs
              late -s +00:05 -c +00:15
              edit SENSOR '0'
              edit SUBTYPES_INCLUDE '54 55'
              edit NPOOLS '$((2* %ANNPES%))'
            task b2o_msu
              late -s +00:05 -c +00:15
              edit SENSOR '1'
              edit SUBTYPES_INCLUDE '54 55'
            task b2o_ssu
              late -s +00:05 -c +00:15
              edit SENSOR '2'
              edit SUBTYPES_INCLUDE '54 55'
            task b2o_amsua
              late -s +00:05 -c +00:15
              edit SENSOR '3'
              edit SUBTYPES_INCLUDE '54 55'
            task b2o_amsub
              late -s +00:05 -c +00:15
              edit SENSOR '4'
              edit SUBTYPES_INCLUDE '54 55'
            task b2o_mhs
              late -s +00:05 -c +00:15
              edit SENSOR '11'
              edit SUBTYPES_INCLUDE '54 55'
            task b2o_satob
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '82 83 87'
            task b2o_reo3
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '205 206 207 208 215'
            task b2o_ssmi
              late -s +00:05 -c +00:15
              edit SENSOR '6'
              edit SUBTYPES_INCLUDE '55 127'
            task b2o_ssmis
              late -s +00:05 -c +00:15
              edit SENSOR '10'
              edit SUBTYPES_INCLUDE '49'
            task b2o_amsre
              late -s +00:05 -c +00:15
              edit SENSOR '17'
              edit SUBTYPES_INCLUDE '59'
            task b2o_tmi
              late -s +00:05 -c +00:15
              edit SENSOR '9'
              edit SUBTYPES_INCLUDE '129'
            task b2o_scatt
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '122 137 139'
            task b2o_geos
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '189 89 76 77 78'
            task b2o_gpsro
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '250'
            task b2o_meris
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '212'
          endfamily
          task mergeodb
            trigger bufr2odb eq complete
          task mergeodb_surf
            trigger ./bufr2odb/b2o_surf_conv eq complete
            edit NPES '16'
            edit THREADS '1'
        endfamily
        task restartodb
          trigger makeodb eq complete
        task preCleanFDB
        task vardata
          trigger ( ../../00dc/fc/model:step ge 6 or ../../00dc/fc/model eq complete) and ../../06/an eq complete
          event hires
        task fetcherr
          trigger ../../06/an eq complete and ../../00dc/an eq complete
        task lowres
          trigger vardata:hires
          edit THREADS '4'
          edit ININPES '32'
        family surf_anal
          trigger makeodb eq complete and restartodb eq complete and preCleanFDB eq complete and vardata:hires
          edit SMSFILES '/home/ma/emos/def/e_36r1/an'
          edit FSFAMILY 'surf_anal'
          task t2ana
            edit mode 't2m'
          task rh2ana
            trigger t2ana eq complete
            edit mode 'rh2m'
          task snow
            trigger rh2ana eq complete
            edit mode 'snow'
          task sst
            edit THREADS '4'
          task slwet
            trigger sst eq complete and snow eq complete
        endfamily
        family 4dvar
          trigger makeodb eq complete and restartodb eq complete and vardata eq complete and fetcherr eq complete and lowres eq complete and preCleanFDB eq complete
          edit THREADS '8'
          edit VAR_DELAY '00'
          edit EMOS_SLEEP '00'
          family uptraj_0
            edit uptraj '0'
            task ifstraj
              late -s +00:02
              edit mode 'traj'
              edit USER_PRIORITY '91'
              edit SMSTRIES '1'
            task ifsmin
              late -s +00:02
              trigger ifstraj eq complete
              edit mode 'min'
              edit USER_PRIORITY '92'
            task ifstsave
              trigger ifstraj eq complete
              edit THREADS '1'
              edit USER_PRIORITY '93'
          endfamily
          family uptraj_1
            trigger uptraj_0 eq complete
            edit uptraj '1'
            task ifstraj
              late -s +00:02
              edit mode 'traj'
              edit USER_PRIORITY '93'
            task ifsmin
              late -s +00:02
              trigger ifstraj eq complete
              edit mode 'min'
              edit USER_PRIORITY '94'
            task ifstsave
              trigger ifstraj eq complete
              edit THREADS '1'
              edit USER_PRIORITY '95'
          endfamily
          family uptraj_2
            trigger uptraj_1 eq complete
            edit uptraj '2'
            task ifstraj
              late -s +00:02
              edit mode 'traj'
              edit USER_PRIORITY '95'
            task ifsmin
              late -s +00:02
              trigger ifstraj eq complete
              edit mode 'min'
              edit USER_PRIORITY '96'
            task ifstsave
              trigger ifstraj eq complete
              edit THREADS '1'
              edit USER_PRIORITY '97'
          endfamily
          task ifstraj
            late -s +00:02
            trigger uptraj_2 eq complete
            edit uptraj '3'
            edit mode 'traj'
            edit USER_PRIORITY '97'
            event finaltraj
            event finalwave
        endfamily
        task forceinv
          trigger ./4dvar/ifstraj:finaltraj and ./surf_anal eq complete
        family reservation1
          late -c +00:03
          trigger ./4dvar/uptraj_0/ifstraj eq active or ./4dvar/uptraj_0/ifstraj eq complete
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
          inlimit /e_36r1/limits:reservation
          task reserve_cpu
            late -s +00:03 -c +00:02
            edit RESERVED_CPUS '3328'
            edit RESERVE_FOR '80'
            label info ""
          task rm_res
            late -s +00:03 -c +00:02
            edit RESERVATION_LABEL 'emos12A'
            label info ""
          task ms_cpu
            late -s +00:03 -c +00:02
            edit MS_CPUS '$((64*30))'
            label info ""
        endfamily
        family reservation3
          late -c +00:03
          trigger ./4dvar/uptraj_1 eq complete
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
          inlimit /e_36r1/limits:reservation
          task set_class
            late -s +00:03 -c +00:02
            edit RESERVATION_DEFINITION 'default'
            label info ""
        endfamily
      endfamily
      family fc
        trigger an/forceinv eq complete and ( an/4dvar/ifstraj:finalwave or an/4dvar eq complete)
        edit ACCOUNT 'oeoperfc'
        edit FSFAMILY 'fc'
        task getfcdata
        task getpersSST
          trigger getfcdata eq complete
          edit SSTLEN '240'
          edit SSTINT '1'
          edit THREADS '2'
          edit NPES '16'
          edit MEM '1000'
          edit CPUTIME '960'
        task model
          late -s +00:02
          trigger ( getfcdata eq complete) and ( getpersSST eq complete)
          edit PERSST 'true'
          edit FCLENGTH '240'
          edit THREADS '8'
          edit USER_PRIORITY '94'
          meter step -1 240 240
        task ddh
          trigger model eq complete
        task imet
          trigger ddh eq complete
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /limits:linux_cluster
        task ddh_arm
          trigger ddh eq complete
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /limits:linux_cluster
        task clean
          trigger ../an eq complete and getfcdata eq complete and model eq complete and ddh eq complete and ../../00dc/fc eq complete
        family reservation1
          late -c +00:03
          trigger ./getfcdata eq active or ./getfcdata eq complete
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
          inlimit /e_36r1/limits:reservation
          family xp
            task set_xp
              late -s +00:03 -c +00:02
              edit RESERVATION_LABEL 'emos12XP'
              label info ""
            task rm_res
              late -s +00:03 -c +00:02
              trigger ./set_xp eq complete
              edit RESERVATION_LABEL 'emos12XP'
              label info ""
          endfamily
          task reserve_cpu
            late -s +00:03 -c +00:02
            edit RESERVED_CPUS '12160'
            edit RESERVE_FOR '2'
            label info ""
          task rm_res
            late -s +00:03 -c +00:02
            edit RESERVATION_LABEL 'emos12B'
            label info ""
        endfamily
      endfamily
      family getreq
        trigger ( /e_36r1/main/12/an/4dvar eq active or /e_36r1/main/12/an/4dvar eq complete) and ( ./prod eq queued or ./prod eq complete)
        edit SMSFILES '/home/ma/emos/def/e_36r1/prod'
        edit EMOS_TYPE 'fc'
        edit PROD_TASK '1'
        edit EMOS_DOMAIN 'G'
        edit FSFAMILY 'ppdi'
        edit EMOS_PROJECT '12'
        edit SMSTRIES '1'
        edit EMOS_STREAM 'OPER'
        task collectreq
          edit QUEUE 'diss'
          edit USER 'emos'
          edit WSHOST 'diss'
          edit QUEUE_EPILOG 'emos-exp'
          edit LOGDIR '/diss/log'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit SMSOUT '/vol/emos_nc/output'
          edit EMOS_BASE '12'
        task req
          late -s +00:03 -c +00:10
          trigger collectreq eq complete
          label info ""
          event 1 
        task collectbitmap
          trigger collectreq eq complete
          edit QUEUE 'diss'
          edit USER 'emos'
          edit WSHOST 'diss'
          edit QUEUE_EPILOG 'emos-exp'
          edit LOGDIR '/diss/log'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit SMSOUT '/vol/emos_nc/output'
        task bitmap
          late -s +00:03 -c +00:10
          trigger collectbitmap eq complete and ( req:001 or req eq complete)
          label info ""
          event 1 
        task collectwave
          trigger ./collectbitmap eq complete
          edit QUEUE 'diss'
          edit USER 'emos'
          edit WSHOST 'diss'
          edit QUEUE_EPILOG 'emos-exp'
          edit LOGDIR '/diss/log'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit SMSOUT '/vol/emos_nc/output'
          edit EMOS_BASE '12'
          edit EMOS_STREAM 'WAVE'
        task wavereq
          late -s +00:03 -c +00:10
          trigger ./collectwave eq complete
          edit EMOS_STREAM 'WAVE'
          label info ""
          event 1 
        task pgclima
          late -s +00:03 -c +00:10
          event 1 
        task cleanpdb
      endfamily
      family prod
        trigger ( getreq/req:001 or getreq/req eq complete) and ( getreq/bitmap:001 or getreq/bitmap eq complete) and ( getreq/wavereq:001 or getreq/wavereq eq complete)
        edit SMSFILES '/home/ma/emos/def/e_36r1/prod'
        edit EMOS_DOMAIN 'G'
        edit EMOS_TYPE 'fc'
        edit FSFAMILY 'ppdi'
        edit PROD_TASK '1'
        edit EMOS_PROJECT '12'
        edit SMSURLBASE 'http://diss.ecmwf.int/do/monitoring/summary'
        family 06an
          trigger /e_36r1/main/00dc/an/dcda_to_da eq complete
          edit EMOS_TYPE 'an'
          edit EMOS_BASE '06'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            edit MEMBER 'off'
            edit STEPS '0'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 12an
          trigger /e_36r1/main/12/an/surf_anal eq complete and ( /e_36r1/main/12/an/4dvar/ifstraj:finalwave or /e_36r1/main/12/an/4dvar/ifstraj eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TYPE 'an'
          edit PGNPES '90'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            edit SMSPASS 'stripped'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ( ./derived_param:SMOS or ./derived_param eq complete) and ( ( /e_36r1/main/12/an/4dvar/ifstraj:finalwave or /e_36r1/main/12/an/4dvar/ifstraj eq complete))
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            edit MEMBER 'off'
            edit STEPS '0'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 000
          trigger ( /e_36r1/main/12/fc/model:step ge 000 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0000'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            edit MEMBER 'off'
            edit STEPS '000'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 003
          trigger ( /e_36r1/main/12/fc/model:step ge 003 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0003'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../000/derived_param:SMOS or ../000/derived_param eq complete) or ( ../000/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '003'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 006
          trigger ( /e_36r1/main/12/fc/model:step ge 006 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0006'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../003/derived_param:SMOS or ../003/derived_param eq complete) or ( ../003/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '006'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 009
          trigger ( /e_36r1/main/12/fc/model:step ge 009 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0009'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../006/derived_param:SMOS or ../006/derived_param eq complete) or ( ../006/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '009'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 012
          trigger ( /e_36r1/main/12/fc/model:step ge 012 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0012'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../009/derived_param:SMOS or ../009/derived_param eq complete) or ( ../009/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '012'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 015
          trigger ( /e_36r1/main/12/fc/model:step ge 015 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0015'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../012/derived_param:SMOS or ../012/derived_param eq complete) or ( ../012/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '015'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 018
          trigger ( /e_36r1/main/12/fc/model:step ge 018 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0018'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../015/derived_param:SMOS or ../015/derived_param eq complete) or ( ../015/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '018'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 021
          trigger ( /e_36r1/main/12/fc/model:step ge 021 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0021'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../018/derived_param:SMOS or ../018/derived_param eq complete) or ( ../018/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '021'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 024
          trigger ( /e_36r1/main/12/fc/model:step ge 024 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0024'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../021/derived_param:SMOS or ../021/derived_param eq complete) or ( ../021/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '024'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 027
          trigger ( /e_36r1/main/12/fc/model:step ge 027 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0027'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../024/derived_param:SMOS or ../024/derived_param eq complete) or ( ../024/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '027'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 030
          trigger ( /e_36r1/main/12/fc/model:step ge 030 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0030'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../027/derived_param:SMOS or ../027/derived_param eq complete) or ( ../027/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '030'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 033
          trigger ( /e_36r1/main/12/fc/model:step ge 033 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0033'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../030/derived_param:SMOS or ../030/derived_param eq complete) or ( ../030/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '033'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 036
          trigger ( /e_36r1/main/12/fc/model:step ge 036 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0036'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../033/derived_param:SMOS or ../033/derived_param eq complete) or ( ../033/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '036'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 039
          trigger ( /e_36r1/main/12/fc/model:step ge 039 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0039'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../036/derived_param:SMOS or ../036/derived_param eq complete) or ( ../036/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '039'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 042
          trigger ( /e_36r1/main/12/fc/model:step ge 042 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0042'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../039/derived_param:SMOS or ../039/derived_param eq complete) or ( ../039/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '042'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 045
          trigger ( /e_36r1/main/12/fc/model:step ge 045 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0045'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../042/derived_param:SMOS or ../042/derived_param eq complete) or ( ../042/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '045'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 048
          trigger ( /e_36r1/main/12/fc/model:step ge 048 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0048'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../045/derived_param:SMOS or ../045/derived_param eq complete) or ( ../045/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '048'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 051
          trigger ( /e_36r1/main/12/fc/model:step ge 051 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0051'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../048/derived_param:SMOS or ../048/derived_param eq complete) or ( ../048/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '051'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 054
          trigger ( /e_36r1/main/12/fc/model:step ge 054 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0054'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../051/derived_param:SMOS or ../051/derived_param eq complete) or ( ../051/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '054'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 057
          trigger ( /e_36r1/main/12/fc/model:step ge 057 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0057'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../054/derived_param:SMOS or ../054/derived_param eq complete) or ( ../054/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '057'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 060
          trigger ( /e_36r1/main/12/fc/model:step ge 060 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0060'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../057/derived_param:SMOS or ../057/derived_param eq complete) or ( ../057/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '060'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 063
          trigger ( /e_36r1/main/12/fc/model:step ge 063 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0063'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../060/derived_param:SMOS or ../060/derived_param eq complete) or ( ../060/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '063'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 066
          trigger ( /e_36r1/main/12/fc/model:step ge 066 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0066'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../063/derived_param:SMOS or ../063/derived_param eq complete) or ( ../063/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '066'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 069
          trigger ( /e_36r1/main/12/fc/model:step ge 069 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0069'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../066/derived_param:SMOS or ../066/derived_param eq complete) or ( ../066/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '069'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 072
          trigger ( /e_36r1/main/12/fc/model:step ge 072 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0072'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../069/derived_param:SMOS or ../069/derived_param eq complete) or ( ../069/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '072'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 075
          trigger ( /e_36r1/main/12/fc/model:step ge 075 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0075'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../072/derived_param:SMOS or ../072/derived_param eq complete) or ( ../072/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '075'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 078
          trigger ( /e_36r1/main/12/fc/model:step ge 078 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0078'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../075/derived_param:SMOS or ../075/derived_param eq complete) or ( ../075/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '078'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 081
          trigger ( /e_36r1/main/12/fc/model:step ge 081 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0081'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../078/derived_param:SMOS or ../078/derived_param eq complete) or ( ../078/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '081'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 084
          trigger ( /e_36r1/main/12/fc/model:step ge 084 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0084'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../081/derived_param:SMOS or ../081/derived_param eq complete) or ( ../081/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '084'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 087
          trigger ( /e_36r1/main/12/fc/model:step ge 087 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0087'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../084/derived_param:SMOS or ../084/derived_param eq complete) or ( ../084/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '087'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 090
          trigger ( /e_36r1/main/12/fc/model:step ge 090 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0090'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../087/derived_param:SMOS or ../087/derived_param eq complete) or ( ../087/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '090'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 093
          trigger ( /e_36r1/main/12/fc/model:step ge 093 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0093'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../090/derived_param:SMOS or ../090/derived_param eq complete) or ( ../090/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '093'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 096
          trigger ( /e_36r1/main/12/fc/model:step ge 096 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0096'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../093/derived_param:SMOS or ../093/derived_param eq complete) or ( ../093/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '096'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 099
          trigger ( /e_36r1/main/12/fc/model:step ge 099 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0099'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../096/derived_param:SMOS or ../096/derived_param eq complete) or ( ../096/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '099'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 102
          trigger ( /e_36r1/main/12/fc/model:step ge 102 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0102'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../099/derived_param:SMOS or ../099/derived_param eq complete) or ( ../099/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '102'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 105
          trigger ( /e_36r1/main/12/fc/model:step ge 105 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0105'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../102/derived_param:SMOS or ../102/derived_param eq complete) or ( ../102/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '105'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 108
          trigger ( /e_36r1/main/12/fc/model:step ge 108 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0108'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../105/derived_param:SMOS or ../105/derived_param eq complete) or ( ../105/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '108'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 111
          trigger ( /e_36r1/main/12/fc/model:step ge 111 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0111'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../108/derived_param:SMOS or ../108/derived_param eq complete) or ( ../108/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '111'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 114
          trigger ( /e_36r1/main/12/fc/model:step ge 114 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0114'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../111/derived_param:SMOS or ../111/derived_param eq complete) or ( ../111/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '114'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 117
          trigger ( /e_36r1/main/12/fc/model:step ge 117 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0117'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../114/derived_param:SMOS or ../114/derived_param eq complete) or ( ../114/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '117'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 120
          trigger ( /e_36r1/main/12/fc/model:step ge 120 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0120'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../117/derived_param:SMOS or ../117/derived_param eq complete) or ( ../117/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '120'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 123
          trigger ( /e_36r1/main/12/fc/model:step ge 123 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0123'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../120/derived_param:SMOS or ../120/derived_param eq complete) or ( ../120/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '123'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 126
          trigger ( /e_36r1/main/12/fc/model:step ge 126 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0126'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../123/derived_param:SMOS or ../123/derived_param eq complete) or ( ../123/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '126'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 129
          trigger ( /e_36r1/main/12/fc/model:step ge 129 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0129'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../126/derived_param:SMOS or ../126/derived_param eq complete) or ( ../126/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '129'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 132
          trigger ( /e_36r1/main/12/fc/model:step ge 132 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0132'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../129/derived_param:SMOS or ../129/derived_param eq complete) or ( ../129/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '132'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 135
          trigger ( /e_36r1/main/12/fc/model:step ge 135 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0135'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../132/derived_param:SMOS or ../132/derived_param eq complete) or ( ../132/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '135'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 138
          trigger ( /e_36r1/main/12/fc/model:step ge 138 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0138'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../135/derived_param:SMOS or ../135/derived_param eq complete) or ( ../135/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '138'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 141
          trigger ( /e_36r1/main/12/fc/model:step ge 141 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0141'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../138/derived_param:SMOS or ../138/derived_param eq complete) or ( ../138/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '141'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 144
          trigger ( /e_36r1/main/12/fc/model:step ge 144 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0144'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../141/derived_param:SMOS or ../141/derived_param eq complete) or ( ../141/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '144'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 150
          trigger ( /e_36r1/main/12/fc/model:step ge 150 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0150'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../144/derived_param:SMOS or ../144/derived_param eq complete) or ( ../144/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '150'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 156
          trigger ( /e_36r1/main/12/fc/model:step ge 156 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0156'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../150/derived_param:SMOS or ../150/derived_param eq complete) or ( ../150/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '156'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 162
          trigger ( /e_36r1/main/12/fc/model:step ge 162 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0162'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../156/derived_param:SMOS or ../156/derived_param eq complete) or ( ../156/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '162'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 168
          trigger ( /e_36r1/main/12/fc/model:step ge 168 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0168'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../162/derived_param:SMOS or ../162/derived_param eq complete) or ( ../162/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '168'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 174
          trigger ( /e_36r1/main/12/fc/model:step ge 174 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0174'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../168/derived_param:SMOS or ../168/derived_param eq complete) or ( ../168/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '174'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 180
          trigger ( /e_36r1/main/12/fc/model:step ge 180 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0180'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../174/derived_param:SMOS or ../174/derived_param eq complete) or ( ../174/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '180'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 186
          trigger ( /e_36r1/main/12/fc/model:step ge 186 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0186'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../180/derived_param:SMOS or ../180/derived_param eq complete) or ( ../180/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '186'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 192
          trigger ( /e_36r1/main/12/fc/model:step ge 192 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0192'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../186/derived_param:SMOS or ../186/derived_param eq complete) or ( ../186/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '192'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 198
          trigger ( /e_36r1/main/12/fc/model:step ge 198 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0198'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../192/derived_param:SMOS or ../192/derived_param eq complete) or ( ../192/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '198'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 204
          trigger ( /e_36r1/main/12/fc/model:step ge 204 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0204'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../198/derived_param:SMOS or ../198/derived_param eq complete) or ( ../198/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '204'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 210
          trigger ( /e_36r1/main/12/fc/model:step ge 210 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0210'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../204/derived_param:SMOS or ../204/derived_param eq complete) or ( ../204/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '210'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 216
          trigger ( /e_36r1/main/12/fc/model:step ge 216 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0216'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../210/derived_param:SMOS or ../210/derived_param eq complete) or ( ../210/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '216'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 222
          trigger ( /e_36r1/main/12/fc/model:step ge 222 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0222'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../216/derived_param:SMOS or ../216/derived_param eq complete) or ( ../216/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '222'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 228
          trigger ( /e_36r1/main/12/fc/model:step ge 228 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0228'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../222/derived_param:SMOS or ../222/derived_param eq complete) or ( ../222/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '228'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 234
          trigger ( /e_36r1/main/12/fc/model:step ge 234 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0234'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../228/derived_param:SMOS or ../228/derived_param eq complete) or ( ../228/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '234'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 240
          trigger ( /e_36r1/main/12/fc/model:step ge 240 or /e_36r1/main/12/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete) and ../getreq/cleanpdb eq complete
          edit EMOS_TIME_STEP_H '0240'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodmet
            late -s +00:05 -c +00:15
            edit PROD_TASK '5'
            edit PGNPES '1'
            edit EMOS_TYPE 'mt'
            inlimit /e_36r1/limits:prod
          task prod_wparam
            late -s +00:05 -c +00:15
            trigger ./derived_param eq complete
            edit PROD_TASK 'w'
            edit EMOS_TYPE 'wp'
            edit PGNPES '64'
            edit INIMEM '800'
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../234/derived_param:SMOS or ../234/derived_param eq complete) or ( ../234/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '240'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family reservation1
          late -c +00:03
          trigger /e_36r1/main/12bc/prod/078/prodgen eq active or /e_36r1/main/12bc/prod/078/prodgen eq complete
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
          inlimit /e_36r1/limits:reservation
          task set_xp
            late -s +00:03 -c +00:02
            edit RESERVATION_NODES '45'
            label info ""
        endfamily
        family reservation2
          late -c +00:03
          trigger /e_36r1/main/12/prod/240/prodgen eq active or /e_36r1/main/12/prod/240/prodgen eq complete
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
          inlimit /e_36r1/limits:reservation
          task set_xp
            late -s +00:03 -c +00:02
            edit RESERVATION_NODES '20'
            label info ""
        endfamily
        family reservation3
          late -c +00:03
          trigger /e_36r1/main/12/prod/240/prodgen eq active or /e_36r1/main/12/prod/240/prodgen eq complete
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
          inlimit /e_36r1/limits:reservation
          task reset_xp
            late -s +00:03 -c +00:02
            trigger /emc_36r1/main/12/legB/prod eq complete
            edit RESERVATION_NODES '0'
            label info ""
        endfamily
      endfamily
      family prod2diss
        edit USER_PRIORITY '90'
        edit SMSURLBASE 'http://diss.ecmwf.int/do/monitoring/summary'
        edit SMSFILES '/home/ma/emos/def/e_36r1/prod'
        task operation_is_late
          event yes
        family oper
          edit EMOS_DOMAIN 'G'
          edit EMOS_TYPE 'fc'
          edit FSFAMILY 'ppdi'
          edit PROD_TASK '1'
          edit USER_PRIORITY '92'
          edit EMOS_PROJECT '12'
          edit SMSURL 'GSCDA/12'
          family 06an
            edit EMOS_BASE '06'
            edit EMOS_TYPE 'an'
            edit SMSURL 'GOPER/06'
            task pdb
              trigger ../../../prod/06an/prodgen eq complete
              edit EMOS_TYPE 'an'
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete
              edit EMOS_TYPE 'an'
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 12an
            edit EMOS_TYPE 'an'
            edit SMSURL 'GOPER/12'
            task pdb
              trigger ../../../prod/12an/prodgen eq complete
              edit EMOS_TYPE 'an'
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete
              edit EMOS_TYPE 'an'
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 000
            edit EMOS_TIME_STEP_H '0000'
            task pdb
              trigger ../../../prod/000/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../12an/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 003
            edit EMOS_TIME_STEP_H '0003'
            task pdb
              trigger ../../../prod/003/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../000/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 006
            edit EMOS_TIME_STEP_H '0006'
            task pdb
              trigger ../../../prod/006/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../003/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 009
            edit EMOS_TIME_STEP_H '0009'
            task pdb
              trigger ../../../prod/009/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../006/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 012
            edit EMOS_TIME_STEP_H '0012'
            task pdb
              trigger ../../../prod/012/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../009/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 015
            edit EMOS_TIME_STEP_H '0015'
            task pdb
              trigger ../../../prod/015/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../012/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 018
            edit EMOS_TIME_STEP_H '0018'
            task pdb
              trigger ../../../prod/018/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../015/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 021
            edit EMOS_TIME_STEP_H '0021'
            task pdb
              trigger ../../../prod/021/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../018/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 024
            edit EMOS_TIME_STEP_H '0024'
            task pdb
              trigger ../../../prod/024/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../021/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 027
            edit EMOS_TIME_STEP_H '0027'
            task pdb
              trigger ../../../prod/027/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../024/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 030
            edit EMOS_TIME_STEP_H '0030'
            task pdb
              trigger ../../../prod/030/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../027/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 033
            edit EMOS_TIME_STEP_H '0033'
            task pdb
              trigger ../../../prod/033/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../030/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 036
            edit EMOS_TIME_STEP_H '0036'
            task pdb
              trigger ../../../prod/036/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../033/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 039
            edit EMOS_TIME_STEP_H '0039'
            task pdb
              trigger ../../../prod/039/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../036/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 042
            edit EMOS_TIME_STEP_H '0042'
            task pdb
              trigger ../../../prod/042/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../039/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 045
            edit EMOS_TIME_STEP_H '0045'
            task pdb
              trigger ../../../prod/045/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../042/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 048
            edit EMOS_TIME_STEP_H '0048'
            task pdb
              trigger ../../../prod/048/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../045/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 051
            edit EMOS_TIME_STEP_H '0051'
            task pdb
              trigger ../../../prod/051/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../048/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 054
            edit EMOS_TIME_STEP_H '0054'
            task pdb
              trigger ../../../prod/054/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../051/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 057
            edit EMOS_TIME_STEP_H '0057'
            task pdb
              trigger ../../../prod/057/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../054/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 060
            edit EMOS_TIME_STEP_H '0060'
            task pdb
              trigger ../../../prod/060/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../057/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 063
            edit EMOS_TIME_STEP_H '0063'
            task pdb
              trigger ../../../prod/063/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../060/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 066
            edit EMOS_TIME_STEP_H '0066'
            task pdb
              trigger ../../../prod/066/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../063/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 069
            edit EMOS_TIME_STEP_H '0069'
            task pdb
              trigger ../../../prod/069/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../066/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 072
            edit EMOS_TIME_STEP_H '0072'
            task pdb
              trigger ../../../prod/072/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../069/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 075
            edit EMOS_TIME_STEP_H '0075'
            task pdb
              trigger ../../../prod/075/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../072/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 078
            edit EMOS_TIME_STEP_H '0078'
            task pdb
              trigger ../../../prod/078/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../075/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 081
            edit EMOS_TIME_STEP_H '0081'
            task pdb
              trigger ../../../prod/081/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../078/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 084
            edit EMOS_TIME_STEP_H '0084'
            task pdb
              trigger ../../../prod/084/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../081/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 087
            edit EMOS_TIME_STEP_H '0087'
            task pdb
              trigger ../../../prod/087/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../084/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 090
            edit EMOS_TIME_STEP_H '0090'
            task pdb
              trigger ../../../prod/090/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../087/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 093
            edit EMOS_TIME_STEP_H '0093'
            task pdb
              trigger ../../../prod/093/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../090/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 096
            edit EMOS_TIME_STEP_H '0096'
            task pdb
              trigger ../../../prod/096/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../093/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 099
            edit EMOS_TIME_STEP_H '0099'
            task pdb
              trigger ../../../prod/099/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../096/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 102
            edit EMOS_TIME_STEP_H '0102'
            task pdb
              trigger ../../../prod/102/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../099/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 105
            edit EMOS_TIME_STEP_H '0105'
            task pdb
              trigger ../../../prod/105/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../102/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 108
            edit EMOS_TIME_STEP_H '0108'
            task pdb
              trigger ../../../prod/108/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../105/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 111
            edit EMOS_TIME_STEP_H '0111'
            task pdb
              trigger ../../../prod/111/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../108/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 114
            edit EMOS_TIME_STEP_H '0114'
            task pdb
              trigger ../../../prod/114/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../111/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 117
            edit EMOS_TIME_STEP_H '0117'
            task pdb
              trigger ../../../prod/117/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../114/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 120
            edit EMOS_TIME_STEP_H '0120'
            task pdb
              trigger ../../../prod/120/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../117/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 123
            edit EMOS_TIME_STEP_H '0123'
            task pdb
              trigger ../../../prod/123/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../120/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 126
            edit EMOS_TIME_STEP_H '0126'
            task pdb
              trigger ../../../prod/126/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../123/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 129
            edit EMOS_TIME_STEP_H '0129'
            task pdb
              trigger ../../../prod/129/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../126/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 132
            edit EMOS_TIME_STEP_H '0132'
            task pdb
              trigger ../../../prod/132/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../129/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 135
            edit EMOS_TIME_STEP_H '0135'
            task pdb
              trigger ../../../prod/135/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../132/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 138
            edit EMOS_TIME_STEP_H '0138'
            task pdb
              trigger ../../../prod/138/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../135/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 141
            edit EMOS_TIME_STEP_H '0141'
            task pdb
              trigger ../../../prod/141/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../138/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 144
            edit EMOS_TIME_STEP_H '0144'
            task pdb
              trigger ../../../prod/144/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../141/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 150
            edit EMOS_TIME_STEP_H '0150'
            task pdb
              trigger ../../../prod/150/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../144/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 156
            edit EMOS_TIME_STEP_H '0156'
            task pdb
              trigger ../../../prod/156/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../150/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 162
            edit EMOS_TIME_STEP_H '0162'
            task pdb
              trigger ../../../prod/162/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../156/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 168
            edit EMOS_TIME_STEP_H '0168'
            task pdb
              trigger ../../../prod/168/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../162/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 174
            edit EMOS_TIME_STEP_H '0174'
            task pdb
              trigger ../../../prod/174/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../168/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 180
            edit EMOS_TIME_STEP_H '0180'
            task pdb
              trigger ../../../prod/180/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../174/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 186
            edit EMOS_TIME_STEP_H '0186'
            task pdb
              trigger ../../../prod/186/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../180/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 192
            edit EMOS_TIME_STEP_H '0192'
            task pdb
              trigger ../../../prod/192/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../186/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 198
            edit EMOS_TIME_STEP_H '0198'
            task pdb
              trigger ../../../prod/198/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../192/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 204
            edit EMOS_TIME_STEP_H '0204'
            task pdb
              trigger ../../../prod/204/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../198/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 210
            edit EMOS_TIME_STEP_H '0210'
            task pdb
              trigger ../../../prod/210/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../204/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 216
            edit EMOS_TIME_STEP_H '0216'
            task pdb
              trigger ../../../prod/216/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../210/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 222
            edit EMOS_TIME_STEP_H '0222'
            task pdb
              trigger ../../../prod/222/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../216/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 228
            edit EMOS_TIME_STEP_H '0228'
            task pdb
              trigger ../../../prod/228/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../222/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 234
            edit EMOS_TIME_STEP_H '0234'
            task pdb
              trigger ../../../prod/234/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../228/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 240
            edit EMOS_TIME_STEP_H '0240'
            task pdb
              trigger ../../../prod/240/prodgen eq complete and ../../../prod/240/prodmet eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../234/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
            task diss_wparam
              late -s +00:05 -c +00:15
              trigger ../../../prod/240/prod_wparam eq complete
              edit PROD_TASK 'w'
              edit EMOS_TYPE 'wp'
              label info ""
              event pdb
          endfamily
          task check
            trigger ./240 eq complete
        endfamily
        family wave
          edit EMOS_DOMAIN 'G'
          edit EMOS_TYPE 'fc'
          edit FSFAMILY 'ppdi'
          edit PROD_TASK '1'
          edit EMOS_PROJECT '12'
          edit EMOS_STREAM 'WV'
          edit SMSURL 'GWAVE/12'
          inlimit /e_36r1/limits:w2diss
          inlimit /limits:q2diss
          family 06an
            edit EMOS_TYPE 'an'
            edit EMOS_BASE '06'
            task diss
              late -s +00:05 -c +00:15
              trigger ../../../prod/06an/prodwave eq complete
              edit EMOS_TYPE 'an'
              label info ""
              event pdb
          endfamily
          family 12an
            edit EMOS_TYPE 'an'
            task diss
              late -s +00:05 -c +00:15
              trigger ../../../prod/12an/prodwave eq complete
              edit EMOS_TYPE 'an'
              label info ""
              event pdb
          endfamily
          family 000
            trigger ../../prod/000/prodwave eq complete and ./06an eq complete and ./12an eq complete
            edit EMOS_TIME_STEP_H '0000'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 003
            trigger ../../prod/003/prodwave eq complete and ( ./000/diss:pdb or ./000 eq complete)
            edit EMOS_TIME_STEP_H '0003'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 006
            trigger ../../prod/006/prodwave eq complete and ( ./003/diss:pdb or ./003 eq complete)
            edit EMOS_TIME_STEP_H '0006'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 009
            trigger ../../prod/009/prodwave eq complete and ( ./006/diss:pdb or ./006 eq complete)
            edit EMOS_TIME_STEP_H '0009'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 012
            trigger ../../prod/012/prodwave eq complete and ( ./009/diss:pdb or ./009 eq complete)
            edit EMOS_TIME_STEP_H '0012'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 015
            trigger ../../prod/015/prodwave eq complete and ( ./012/diss:pdb or ./012 eq complete)
            edit EMOS_TIME_STEP_H '0015'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 018
            trigger ../../prod/018/prodwave eq complete and ( ./015/diss:pdb or ./015 eq complete)
            edit EMOS_TIME_STEP_H '0018'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 021
            trigger ../../prod/021/prodwave eq complete and ( ./018/diss:pdb or ./018 eq complete)
            edit EMOS_TIME_STEP_H '0021'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 024
            trigger ../../prod/024/prodwave eq complete and ( ./021/diss:pdb or ./021 eq complete)
            edit EMOS_TIME_STEP_H '0024'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 027
            trigger ../../prod/027/prodwave eq complete and ( ./024/diss:pdb or ./024 eq complete)
            edit EMOS_TIME_STEP_H '0027'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 030
            trigger ../../prod/030/prodwave eq complete and ( ./027/diss:pdb or ./027 eq complete)
            edit EMOS_TIME_STEP_H '0030'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 033
            trigger ../../prod/033/prodwave eq complete and ( ./030/diss:pdb or ./030 eq complete)
            edit EMOS_TIME_STEP_H '0033'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 036
            trigger ../../prod/036/prodwave eq complete and ( ./033/diss:pdb or ./033 eq complete)
            edit EMOS_TIME_STEP_H '0036'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 039
            trigger ../../prod/039/prodwave eq complete and ( ./036/diss:pdb or ./036 eq complete)
            edit EMOS_TIME_STEP_H '0039'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 042
            trigger ../../prod/042/prodwave eq complete and ( ./039/diss:pdb or ./039 eq complete)
            edit EMOS_TIME_STEP_H '0042'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 045
            trigger ../../prod/045/prodwave eq complete and ( ./042/diss:pdb or ./042 eq complete)
            edit EMOS_TIME_STEP_H '0045'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 048
            trigger ../../prod/048/prodwave eq complete and ( ./045/diss:pdb or ./045 eq complete)
            edit EMOS_TIME_STEP_H '0048'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 051
            trigger ../../prod/051/prodwave eq complete and ( ./048/diss:pdb or ./048 eq complete)
            edit EMOS_TIME_STEP_H '0051'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 054
            trigger ../../prod/054/prodwave eq complete and ( ./051/diss:pdb or ./051 eq complete)
            edit EMOS_TIME_STEP_H '0054'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 057
            trigger ../../prod/057/prodwave eq complete and ( ./054/diss:pdb or ./054 eq complete)
            edit EMOS_TIME_STEP_H '0057'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 060
            trigger ../../prod/060/prodwave eq complete and ( ./057/diss:pdb or ./057 eq complete)
            edit EMOS_TIME_STEP_H '0060'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 063
            trigger ../../prod/063/prodwave eq complete and ( ./060/diss:pdb or ./060 eq complete)
            edit EMOS_TIME_STEP_H '0063'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 066
            trigger ../../prod/066/prodwave eq complete and ( ./063/diss:pdb or ./063 eq complete)
            edit EMOS_TIME_STEP_H '0066'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 069
            trigger ../../prod/069/prodwave eq complete and ( ./066/diss:pdb or ./066 eq complete)
            edit EMOS_TIME_STEP_H '0069'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 072
            trigger ../../prod/072/prodwave eq complete and ( ./069/diss:pdb or ./069 eq complete)
            edit EMOS_TIME_STEP_H '0072'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 075
            trigger ../../prod/075/prodwave eq complete and ( ./072/diss:pdb or ./072 eq complete)
            edit EMOS_TIME_STEP_H '0075'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 078
            trigger ../../prod/078/prodwave eq complete and ( ./075/diss:pdb or ./075 eq complete)
            edit EMOS_TIME_STEP_H '0078'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 081
            trigger ../../prod/081/prodwave eq complete and ( ./078/diss:pdb or ./078 eq complete)
            edit EMOS_TIME_STEP_H '0081'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 084
            trigger ../../prod/084/prodwave eq complete and ( ./081/diss:pdb or ./081 eq complete)
            edit EMOS_TIME_STEP_H '0084'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 087
            trigger ../../prod/087/prodwave eq complete and ( ./084/diss:pdb or ./084 eq complete)
            edit EMOS_TIME_STEP_H '0087'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 090
            trigger ../../prod/090/prodwave eq complete and ( ./087/diss:pdb or ./087 eq complete)
            edit EMOS_TIME_STEP_H '0090'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 093
            trigger ../../prod/093/prodwave eq complete and ( ./090/diss:pdb or ./090 eq complete)
            edit EMOS_TIME_STEP_H '0093'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 096
            trigger ../../prod/096/prodwave eq complete and ( ./093/diss:pdb or ./093 eq complete)
            edit EMOS_TIME_STEP_H '0096'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 099
            trigger ../../prod/099/prodwave eq complete and ( ./096/diss:pdb or ./096 eq complete)
            edit EMOS_TIME_STEP_H '0099'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 102
            trigger ../../prod/102/prodwave eq complete and ( ./099/diss:pdb or ./099 eq complete)
            edit EMOS_TIME_STEP_H '0102'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 105
            trigger ../../prod/105/prodwave eq complete and ( ./102/diss:pdb or ./102 eq complete)
            edit EMOS_TIME_STEP_H '0105'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 108
            trigger ../../prod/108/prodwave eq complete and ( ./105/diss:pdb or ./105 eq complete)
            edit EMOS_TIME_STEP_H '0108'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 111
            trigger ../../prod/111/prodwave eq complete and ( ./108/diss:pdb or ./108 eq complete)
            edit EMOS_TIME_STEP_H '0111'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 114
            trigger ../../prod/114/prodwave eq complete and ( ./111/diss:pdb or ./111 eq complete)
            edit EMOS_TIME_STEP_H '0114'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 117
            trigger ../../prod/117/prodwave eq complete and ( ./114/diss:pdb or ./114 eq complete)
            edit EMOS_TIME_STEP_H '0117'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 120
            trigger ../../prod/120/prodwave eq complete and ( ./117/diss:pdb or ./117 eq complete)
            edit EMOS_TIME_STEP_H '0120'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 123
            trigger ../../prod/123/prodwave eq complete and ( ./120/diss:pdb or ./120 eq complete)
            edit EMOS_TIME_STEP_H '0123'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 126
            trigger ../../prod/126/prodwave eq complete and ( ./123/diss:pdb or ./123 eq complete)
            edit EMOS_TIME_STEP_H '0126'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 129
            trigger ../../prod/129/prodwave eq complete and ( ./126/diss:pdb or ./126 eq complete)
            edit EMOS_TIME_STEP_H '0129'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 132
            trigger ../../prod/132/prodwave eq complete and ( ./129/diss:pdb or ./129 eq complete)
            edit EMOS_TIME_STEP_H '0132'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 135
            trigger ../../prod/135/prodwave eq complete and ( ./132/diss:pdb or ./132 eq complete)
            edit EMOS_TIME_STEP_H '0135'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 138
            trigger ../../prod/138/prodwave eq complete and ( ./135/diss:pdb or ./135 eq complete)
            edit EMOS_TIME_STEP_H '0138'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 141
            trigger ../../prod/141/prodwave eq complete and ( ./138/diss:pdb or ./138 eq complete)
            edit EMOS_TIME_STEP_H '0141'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 144
            trigger ../../prod/144/prodwave eq complete and ( ./141/diss:pdb or ./141 eq complete)
            edit EMOS_TIME_STEP_H '0144'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 150
            trigger ../../prod/150/prodwave eq complete and ( ./144/diss:pdb or ./144 eq complete)
            edit EMOS_TIME_STEP_H '0150'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 156
            trigger ../../prod/156/prodwave eq complete and ( ./150/diss:pdb or ./150 eq complete)
            edit EMOS_TIME_STEP_H '0156'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 162
            trigger ../../prod/162/prodwave eq complete and ( ./156/diss:pdb or ./156 eq complete)
            edit EMOS_TIME_STEP_H '0162'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 168
            trigger ../../prod/168/prodwave eq complete and ( ./162/diss:pdb or ./162 eq complete)
            edit EMOS_TIME_STEP_H '0168'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 174
            trigger ../../prod/174/prodwave eq complete and ( ./168/diss:pdb or ./168 eq complete)
            edit EMOS_TIME_STEP_H '0174'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 180
            trigger ../../prod/180/prodwave eq complete and ( ./174/diss:pdb or ./174 eq complete)
            edit EMOS_TIME_STEP_H '0180'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 186
            trigger ../../prod/186/prodwave eq complete and ( ./180/diss:pdb or ./180 eq complete)
            edit EMOS_TIME_STEP_H '0186'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 192
            trigger ../../prod/192/prodwave eq complete and ( ./186/diss:pdb or ./186 eq complete)
            edit EMOS_TIME_STEP_H '0192'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 198
            trigger ../../prod/198/prodwave eq complete and ( ./192/diss:pdb or ./192 eq complete)
            edit EMOS_TIME_STEP_H '0198'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 204
            trigger ../../prod/204/prodwave eq complete and ( ./198/diss:pdb or ./198 eq complete)
            edit EMOS_TIME_STEP_H '0204'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 210
            trigger ../../prod/210/prodwave eq complete and ( ./204/diss:pdb or ./204 eq complete)
            edit EMOS_TIME_STEP_H '0210'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 216
            trigger ../../prod/216/prodwave eq complete and ( ./210/diss:pdb or ./210 eq complete)
            edit EMOS_TIME_STEP_H '0216'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 222
            trigger ../../prod/222/prodwave eq complete and ( ./216/diss:pdb or ./216 eq complete)
            edit EMOS_TIME_STEP_H '0222'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 228
            trigger ../../prod/228/prodwave eq complete and ( ./222/diss:pdb or ./222 eq complete)
            edit EMOS_TIME_STEP_H '0228'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 234
            trigger ../../prod/234/prodwave eq complete and ( ./228/diss:pdb or ./228 eq complete)
            edit EMOS_TIME_STEP_H '0234'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 240
            trigger ../../prod/240/prodwave eq complete and ( ./234/diss:pdb or ./234 eq complete)
            edit EMOS_TIME_STEP_H '0240'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          task check
            trigger ./240 eq complete
        endfamily
      endfamily
    endfamily
    family 18bc
      trigger ( ./make eq complete) and ( /e_36r1/main/12/fc eq complete)
      edit EMOS_BASE '18'
      edit EDFAMILY 'false'
      edit INISTREAM 'DA'
      edit REINITIALIZE 'false'
      edit REINIEXPVER '0046'
      edit INIEXPVER '0045'
      edit REINIPERIOD_4D '6'
      edit REINISTREAM 'DA'
      edit REINIOFFSET '6'
      edit RUN_HOURLY_PP '0'
      edit PERIOD_4D '6'
      edit WINDOW_LENGTH_4D '6'
      edit WINDOW_OFFSET_4D '3'
      edit MXUP_TRAJ '3'
      edit PARALLEL '4'
      edit LANONLY 'false'
      edit STREAM 'SCDA'
      edit EMOS_STREAM 'SCDA'
      edit DELTA_DAY '0'
      edit ACCOUNT 'opebcpg'
      edit SMSURL 'GSCDA/18'
      edit USER_PRIORITY '91'
      family obs
        edit FSFAMILY 'an'
        edit SMSFILES '/home/ma/emos/def/e_36r1/obs'
        family get
          inlimit /e_36r1/limits:obs_get
          inlimit /limits:obs_get
          task get_nonbin
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/nonbin eq complete
            edit OBSFILE_OBSHOST 'BUFR'
            edit OBSFILE_HPCF 'bufr'
          task get_ers1_wa
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/ers1_wa eq complete
            edit OBSFILE_OBSHOST 'UWAX'
            edit OBSFILE_HPCF 'wavescat'
          task get_eas1
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/eas1 eq complete
            edit OBSFILE_OBSHOST 'EAS1'
            edit OBSFILE_HPCF 'wavesar'
          task get_ers1_wi
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/ers1_wi eq complete
            edit OBSFILE_OBSHOST 'UWIX'
            edit OBSFILE_HPCF 'scat'
          task get_ers1_ra
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/ers1_ra eq complete
            edit OBSFILE_OBSHOST 'URAX'
            edit OBSFILE_HPCF 'ralt'
          task get_rtovs
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/rtovs eq complete
            edit OBSFILE_OBSHOST 'TOVH'
            edit OBSFILE_HPCF 'tovh'
          task get_ssbt
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/ssbt eq complete
            edit OBSFILE_OBSHOST 'SSBT'
            edit OBSFILE_HPCF 'ssbt'
          task get_sst
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/sst/sst_fdb_c1b eq complete
            edit OBSFILE_OBSHOST 'nil'
            edit OBSFILE_HPCF 'nil'
          task get_atov
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/atov eq complete
            edit OBSFILE_OBSHOST 'ATOV'
            edit OBSFILE_HPCF 'atov'
          task get_grad
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/grad eq complete
            edit OBSFILE_OBSHOST 'GRAD'
            edit OBSFILE_HPCF 'grad'
          task get_stor
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/stor eq complete
            edit OBSFILE_OBSHOST 'STOR'
            edit OBSFILE_HPCF 'storm'
          task get_trmm
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/trmm eq complete
            edit OBSFILE_OBSHOST 'TRMM'
            edit OBSFILE_HPCF 'tmi'
          task get_qsca
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/qsca eq complete
            edit OBSFILE_OBSHOST 'QSCA'
            edit OBSFILE_HPCF 'qsca'
          task get_sbuv
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/sbuv eq complete
            edit OBSFILE_OBSHOST 'SBUV'
            edit OBSFILE_HPCF 'reo3'
          task get_airs
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/airs eq complete
            edit OBSFILE_OBSHOST 'AIRS'
            edit OBSFILE_HPCF 'airs'
          task get_ice
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/ice/ice_fdb_c1b eq complete
            edit OBSFILE_OBSHOST 'nil'
            edit OBSFILE_HPCF 'nil'
          task get_snow
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/snow eq complete
            edit OBSFILE_OBSHOST 'SNOW'
            edit OBSFILE_HPCF 'imssnow'
          task get_ateu
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/ateu eq complete
            edit OBSFILE_OBSHOST 'ATEU'
            edit OBSFILE_HPCF 'amsu_ears'
          task get_bice
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/bice eq complete
            edit OBSFILE_OBSHOST 'BICE'
            edit OBSFILE_HPCF 'baltic_ci'
          task get_amsa
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex18/amsa eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'AMSA'
            edit OBSFILE_HPCF 'amsa'
          task get_mhsa
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex18/mhsa eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'MHSA'
            edit OBSFILE_HPCF 'mhsa'
          task get_hirs
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex18/hirs eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'HIRS'
            edit OBSFILE_HPCF 'hirs'
          task get_amsr
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/amsr eq complete
            edit OBSFILE_OBSHOST 'AMSR'
            edit OBSFILE_HPCF 'amsre_aq'
          task get_rado
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/rado eq complete
            edit OBSFILE_OBSHOST 'RADO'
            edit OBSFILE_HPCF 'gpsro'
          task get_smin
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/smin eq complete
            edit OBSFILE_OBSHOST 'SMIN'
            edit OBSFILE_HPCF 'ssmisbt'
          task get_iasi
            late -s +00:02 -c +00:07
            trigger /pp2/main/extractions/ex18/iasi eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'IASI'
            edit OBSFILE_HPCF 'iasi'
          task get_gras
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex18/gras eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'GRAS'
            edit OBSFILE_HPCF 'gras'
          task get_modw
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex18/modw eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'MODW'
            edit OBSFILE_HPCF 'dbmodiswinds'
          task get_avhn
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex18/avhn eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'AVHN'
            edit OBSFILE_HPCF 'avhrrwinds'
          task get_atap
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/atap eq complete
            edit OBSFILE_OBSHOST 'ATAP'
            edit OBSFILE_HPCF 'amsu_pacrars'
          task get_g2to
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex18/g2to eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'G2TO'
            edit OBSFILE_HPCF 'reo3_gome2'
          task get_grst
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/grst eq complete
            edit OBSFILE_OBSHOST 'GRST'
            edit OBSFILE_HPCF 'ostia'
          task get_emer
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/emer eq complete
            edit OBSFILE_OBSHOST 'EMER'
            edit OBSFILE_HPCF 'envisat_meris'
          task get_j2nb
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex18/j2nb eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'J2NB'
            edit OBSFILE_HPCF 'j2nb'
          task get_j2eo
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex18/j2eo eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'J2EO'
            edit OBSFILE_HPCF 'j2eo'
          task get_asc2
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex18/asc2 eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'ASC2'
            edit OBSFILE_HPCF 'ascat_l2'
          task get_asel
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex18/asel eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'ASEL'
            edit OBSFILE_HPCF 'ascat_ears'
          task get_ascs
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex18/ascs eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'ASCS'
            edit OBSFILE_HPCF 'ascat_sm'
          task atovairs
            trigger get_atov eq complete and get_airs eq complete and get_ateu eq complete and get_amsa eq complete and get_mhsa eq complete and get_hirs eq complete and get_atap eq complete
          task snowlinks
            trigger get_nonbin eq complete and get_ers1_wi eq complete and get_rtovs eq complete and get_snow eq complete
          task ro_combine
            trigger get_rado eq complete and get_gras eq complete
          task reo3_combine
            trigger get_sbuv eq complete and get_g2to eq complete
          task bufr_combine
            trigger get_nonbin eq complete and get_modw eq complete and get_avhn eq complete
          task jason2_combine
            trigger get_ers1_ra eq complete and get_j2nb eq complete and get_j2eo eq complete
          task ascat_combine
            trigger get_asc2 eq complete and get_asel eq complete and get_ascs eq complete
        endfamily
        task fetchmars
          trigger ./get eq complete
        family prepare_obs
          trigger ./get eq complete
          task preobs
            trigger /e_36r1/main/00dc/an/update_psbias eq complete and /e_36r1/main/00dc/an/update_rstrhbias eq complete
            edit mode 'not1crad'
            edit REINITIALIZE 'true'
            edit REINIPERIOD_4D '6'
          task premwimg
          task preobs_wave
            edit mode 'wave'
          task pre1crad_prepare
            edit mode '1crad'
          task pre1crad_airs
            trigger pre1crad_prepare eq complete
          family pre1crad_iasi
            trigger pre1crad_prepare eq complete
            edit CPUTIME '1500'
            edit MAXTASK '16'
            task pre1crad_iasi_split_1
              edit TASKNO '1'
            task pre1crad_iasi_split_2
              edit TASKNO '2'
            task pre1crad_iasi_1
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '1'
            task pre1crad_iasi_2
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '2'
            task pre1crad_iasi_3
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '3'
            task pre1crad_iasi_4
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '4'
            task pre1crad_iasi_5
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '5'
            task pre1crad_iasi_6
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '6'
            task pre1crad_iasi_7
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '7'
            task pre1crad_iasi_8
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '8'
            task pre1crad_iasi_9
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '9'
            task pre1crad_iasi_10
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '10'
            task pre1crad_iasi_11
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '11'
            task pre1crad_iasi_12
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '12'
            task pre1crad_iasi_13
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '13'
            task pre1crad_iasi_14
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '14'
            task pre1crad_iasi_15
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '15'
            task pre1crad_iasi_16
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '16'
          endfamily
          task pre1crad_hirs
            trigger pre1crad_prepare eq complete
          task pre1crad_msu
            trigger pre1crad_prepare eq complete
          task pre1crad_ssu
            trigger pre1crad_prepare eq complete
          task pre1crad_amsua
            trigger pre1crad_prepare eq complete
          task pre1crad_amsub
            trigger pre1crad_prepare eq complete
          task pre1crad_mhs
            trigger pre1crad_prepare eq complete
          task pregeos
          task prereo3
          task prescat
        endfamily
        family reservation
          late -c +00:03
          trigger ./get/get_stor eq complete
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
          inlimit /e_36r1/limits:reservation
          task adv_res
            late -s +00:03 -c +00:02
            edit RESERVE_FOR '30'
            edit RESERVE_IN '11'
            edit RESERVED_NODES '32'
            edit RESERVATION_LABEL 'emos18A'
            label info ""
          task loadl_cycle_limit
            late -s +00:03 -c +00:02
            edit LOADL_CYCLE '65'
            label info ""
        endfamily
      endfamily
      family an
        trigger ./obs/fetchmars eq complete and ./obs/prepare_obs eq complete
        edit FSFAMILY 'an'
        edit SMSFILES '/home/ma/emos/def/e_36r1/an'
        edit PERIOD_4D '6'
        edit WINDOW_LENGTH_4D '6'
        edit WINDOW_OFFSET_4D '3'
        edit ACCOUNT 'opebcan'
        edit REINITIALIZE 'true'
        edit ANNPES '144'
        family makeodb
          edit NPES '16'
          task cleanodb
          family bufr2odb
            trigger cleanodb eq complete
            edit THREADS '1'
            edit NPOOLS '%ANNPES%'
            edit THINFAC '1'
            edit SENSOR '-1'
            edit REBALANCE '1'
            edit SUBTYPES_INCLUDE ''
            edit SUBTYPES_EXCLUDE ''
            task b2o_conv
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_EXCLUDE '85 82 83 87'
              edit USER_PRIORITY '90'
            task b2o_surf_conv
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_EXCLUDE '142 143 144 145 85 82 83 87'
            task b2o_airs
              late -s +00:05 -c +00:15
              edit THREADS '2'
              edit SENSOR '11'
              edit SUBTYPES_INCLUDE '57'
              edit NPOOLS '$((2* %ANNPES%))'
            task b2o_iasi
              late -s +00:05 -c +00:15
              edit SENSOR '993'
              edit SUBTYPES_INCLUDE '240'
              edit THREADS '4'
              edit NPOOLS '$((2* %ANNPES%))'
            task b2o_hirs
              late -s +00:05 -c +00:15
              edit SENSOR '0'
              edit SUBTYPES_INCLUDE '54 55'
              edit NPOOLS '$((2* %ANNPES%))'
            task b2o_msu
              late -s +00:05 -c +00:15
              edit SENSOR '1'
              edit SUBTYPES_INCLUDE '54 55'
            task b2o_ssu
              late -s +00:05 -c +00:15
              edit SENSOR '2'
              edit SUBTYPES_INCLUDE '54 55'
            task b2o_amsua
              late -s +00:05 -c +00:15
              edit SENSOR '3'
              edit SUBTYPES_INCLUDE '54 55'
            task b2o_amsub
              late -s +00:05 -c +00:15
              edit SENSOR '4'
              edit SUBTYPES_INCLUDE '54 55'
            task b2o_mhs
              late -s +00:05 -c +00:15
              edit SENSOR '11'
              edit SUBTYPES_INCLUDE '54 55'
            task b2o_satob
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '82 83 87'
            task b2o_reo3
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '205 206 207 208 215'
            task b2o_ssmi
              late -s +00:05 -c +00:15
              edit SENSOR '6'
              edit SUBTYPES_INCLUDE '55 127'
            task b2o_ssmis
              late -s +00:05 -c +00:15
              edit SENSOR '10'
              edit SUBTYPES_INCLUDE '49'
            task b2o_amsre
              late -s +00:05 -c +00:15
              edit SENSOR '17'
              edit SUBTYPES_INCLUDE '59'
            task b2o_tmi
              late -s +00:05 -c +00:15
              edit SENSOR '9'
              edit SUBTYPES_INCLUDE '129'
            task b2o_scatt
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '122 137 139'
            task b2o_geos
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '189 89 76 77 78'
            task b2o_gpsro
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '250'
            task b2o_meris
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '212'
          endfamily
          task mergeodb
            trigger bufr2odb eq complete
          task mergeodb_surf
            trigger ./bufr2odb/b2o_surf_conv eq complete
            edit NPES '16'
            edit THREADS '1'
        endfamily
        task restartodb
          trigger makeodb eq complete
        task preCleanFDB
        task vardata
          trigger ( ../../12/fc/model:step ge 9 or ../../12/fc/model eq complete)
          event hires
        task fetcherr
        task lowres
          trigger vardata:hires
          edit THREADS '4'
          edit ININPES '32'
        family surf_anal
          trigger makeodb eq complete and restartodb eq complete and preCleanFDB eq complete and vardata:hires
          edit SMSFILES '/home/ma/emos/def/e_36r1/an'
          edit FSFAMILY 'surf_anal'
          task t2ana
            edit mode 't2m'
          task rh2ana
            trigger t2ana eq complete
            edit mode 'rh2m'
          task snow
            trigger rh2ana eq complete
            edit mode 'snow'
          task sst
            edit THREADS '4'
          task slwet
            trigger sst eq complete and snow eq complete
        endfamily
        family 4dvar
          trigger makeodb eq complete and restartodb eq complete and vardata eq complete and fetcherr eq complete and lowres eq complete and preCleanFDB eq complete
          edit THREADS '8'
          edit VAR_DELAY '00'
          edit EMOS_SLEEP '00'
          family uptraj_0
            edit uptraj '0'
            task ifstraj
              late -s +00:02
              edit mode 'traj'
              edit USER_PRIORITY '91'
              edit SMSTRIES '1'
            task ifsmin
              late -s +00:02
              trigger ifstraj eq complete
              edit mode 'min'
              edit USER_PRIORITY '92'
            task ifstsave
              trigger ifstraj eq complete
              edit THREADS '1'
              edit USER_PRIORITY '93'
          endfamily
          family uptraj_1
            trigger uptraj_0 eq complete
            edit uptraj '1'
            task ifstraj
              late -s +00:02
              edit mode 'traj'
              edit USER_PRIORITY '93'
            task ifsmin
              late -s +00:02
              trigger ifstraj eq complete
              edit mode 'min'
              edit USER_PRIORITY '94'
            task ifstsave
              trigger ifstraj eq complete
              edit THREADS '1'
              edit USER_PRIORITY '95'
          endfamily
          family uptraj_2
            trigger uptraj_1 eq complete
            edit uptraj '2'
            task ifstraj
              late -s +00:02
              edit mode 'traj'
              edit USER_PRIORITY '95'
            task ifsmin
              late -s +00:02
              trigger ifstraj eq complete
              edit mode 'min'
              edit USER_PRIORITY '96'
            task ifstsave
              trigger ifstraj eq complete
              edit THREADS '1'
              edit USER_PRIORITY '97'
          endfamily
          task ifstraj
            late -s +00:02
            trigger uptraj_2 eq complete
            edit uptraj '3'
            edit mode 'traj'
            edit USER_PRIORITY '97'
            event finaltraj
            event finalwave
        endfamily
        task forceinv
          trigger ./4dvar/ifstraj:finaltraj and ./surf_anal eq complete
        family reservation1
          late -c +00:03
          trigger ./4dvar/uptraj_0/ifstraj eq active or ./4dvar/uptraj_0/ifstraj eq complete
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
          inlimit /e_36r1/limits:reservation
          task reserve_cpu
            late -s +00:03 -c +00:02
            edit RESERVED_CPUS '2432'
            edit RESERVE_FOR '35'
            label info ""
          task rm_res
            late -s +00:03 -c +00:02
            edit RESERVATION_LABEL 'emos18A'
            label info ""
        endfamily
      endfamily
      family fc
        trigger an/forceinv eq complete and ( an/4dvar/ifstraj:finalwave or an/4dvar eq complete)
        edit ACCOUNT 'opebcfc'
        edit REINITIALIZE 'true'
        edit FSFAMILY 'fc'
        task getfcdata
        task getpersSST
          trigger getfcdata eq complete
          edit SSTLEN '90'
          edit SSTINT '1'
          edit THREADS '2'
          edit NPES '16'
          edit MEM '1000'
          edit CPUTIME '960'
        task model
          late -s +00:02
          trigger ( getfcdata eq complete) and ( getpersSST eq complete)
          edit PERSST 'true'
          edit FCLENGTH '90'
          edit THREADS '8'
          edit USER_PRIORITY '94'
          edit FCNPES '144'
          meter step -1 90 90
        task clean
          trigger ../an eq complete and getfcdata eq complete and model eq complete
        family reservation1
          late -c +00:03
          trigger ./getfcdata eq complete
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
          inlimit /e_36r1/limits:reservation
          task reserve_cpu
            late -s +00:03 -c +00:02
            edit RESERVED_CPUS '3072'
            edit RESERVE_FOR '20'
            label info ""
        endfamily
        family reservation2
          late -c +00:03
          trigger ./model eq complete
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
          inlimit /e_36r1/limits:reservation
          task reset_cpu
            late -s +00:03 -c +00:02
            edit RESERVED_CPUS '0'
            label info ""
        endfamily
      endfamily
      family getreq
        trigger ( /e_36r1/main/18bc/an/4dvar eq active or /e_36r1/main/18bc/an/4dvar eq complete) and ( ./prod eq queued or ./prod eq complete)
        edit SMSFILES '/home/ma/emos/def/e_36r1/prod'
        edit EMOS_TYPE 'fc'
        edit PROD_TASK '1'
        edit EMOS_DOMAIN 'G'
        edit FSFAMILY 'ppdi'
        edit EMOS_PROJECT '12'
        edit SMSTRIES '1'
        edit EMOS_STREAM 'SCDA'
        task collectreq
          edit QUEUE 'diss'
          edit USER 'emos'
          edit WSHOST 'diss'
          edit QUEUE_EPILOG 'emos-exp'
          edit LOGDIR '/diss/log'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit SMSOUT '/vol/emos_nc/output'
          edit EMOS_BASE '18'
        task req
          late -s +00:03 -c +00:10
          trigger collectreq eq complete
          label info ""
          event 1 
        task collectbitmap
          trigger collectreq eq complete
          edit QUEUE 'diss'
          edit USER 'emos'
          edit WSHOST 'diss'
          edit QUEUE_EPILOG 'emos-exp'
          edit LOGDIR '/diss/log'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit SMSOUT '/vol/emos_nc/output'
        task bitmap
          late -s +00:03 -c +00:10
          trigger collectbitmap eq complete and ( req:001 or req eq complete)
          label info ""
          event 1 
      endfamily
      family prod
        trigger ( getreq/req:001 or getreq/req eq complete)
        edit SMSFILES '/home/ma/emos/def/e_36r1/prod'
        edit EMOS_DOMAIN 'G'
        edit EMOS_TYPE 'fc'
        edit FSFAMILY 'ppdi'
        edit PROD_TASK '1'
        edit EMOS_PROJECT '12'
        edit SMSURLBASE 'http://diss.ecmwf.int/do/monitoring/summary'
        family an
          trigger /e_36r1/main/18bc/an/surf_anal eq complete and ( /e_36r1/main/18bc/an/4dvar/ifstraj:finalwave or /e_36r1/main/18bc/an/4dvar/ifstraj eq complete) and 0 eq 0
          edit EMOS_TYPE 'an'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ../090/prodgen eq complete and ( /e_36r1/main/18bc/fc/model:step ge 12 or /e_36r1/main/18bc/fc/model eq complete)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            edit MEMBER 'off'
            edit STEPS '0'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 000
          trigger ( /e_36r1/main/18bc/fc/model:step ge 000 or /e_36r1/main/18bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0000'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/18bc/fc/model:step ge 12 or /e_36r1/main/18bc/fc/model eq complete) and ( ../an/backupprod eq complete or ../an/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            edit MEMBER 'off'
            edit STEPS '000'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 003
          trigger ( /e_36r1/main/18bc/fc/model:step ge 003 or /e_36r1/main/18bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0003'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/18bc/fc/model:step ge 15 or /e_36r1/main/18bc/fc/model eq complete) and ( ../000/backupprod eq complete or ../000/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../000/derived_param:SMOS or ../000/derived_param eq complete) or ( ../000/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '003'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 006
          trigger ( /e_36r1/main/18bc/fc/model:step ge 006 or /e_36r1/main/18bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0006'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/18bc/fc/model:step ge 18 or /e_36r1/main/18bc/fc/model eq complete) and ( ../003/backupprod eq complete or ../003/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../003/derived_param:SMOS or ../003/derived_param eq complete) or ( ../003/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '006'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 009
          trigger ( /e_36r1/main/18bc/fc/model:step ge 009 or /e_36r1/main/18bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0009'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/18bc/fc/model:step ge 21 or /e_36r1/main/18bc/fc/model eq complete) and ( ../006/backupprod eq complete or ../006/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../006/derived_param:SMOS or ../006/derived_param eq complete) or ( ../006/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '009'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 012
          trigger ( /e_36r1/main/18bc/fc/model:step ge 012 or /e_36r1/main/18bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0012'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/18bc/fc/model:step ge 24 or /e_36r1/main/18bc/fc/model eq complete) and ( ../009/backupprod eq complete or ../009/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../009/derived_param:SMOS or ../009/derived_param eq complete) or ( ../009/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '012'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 015
          trigger ( /e_36r1/main/18bc/fc/model:step ge 015 or /e_36r1/main/18bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0015'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/18bc/fc/model:step ge 27 or /e_36r1/main/18bc/fc/model eq complete) and ( ../012/backupprod eq complete or ../012/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../012/derived_param:SMOS or ../012/derived_param eq complete) or ( ../012/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '015'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 018
          trigger ( /e_36r1/main/18bc/fc/model:step ge 018 or /e_36r1/main/18bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0018'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/18bc/fc/model:step ge 30 or /e_36r1/main/18bc/fc/model eq complete) and ( ../015/backupprod eq complete or ../015/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../015/derived_param:SMOS or ../015/derived_param eq complete) or ( ../015/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '018'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 021
          trigger ( /e_36r1/main/18bc/fc/model:step ge 021 or /e_36r1/main/18bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0021'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/18bc/fc/model:step ge 33 or /e_36r1/main/18bc/fc/model eq complete) and ( ../018/backupprod eq complete or ../018/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../018/derived_param:SMOS or ../018/derived_param eq complete) or ( ../018/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '021'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 024
          trigger ( /e_36r1/main/18bc/fc/model:step ge 024 or /e_36r1/main/18bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0024'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/18bc/fc/model:step ge 36 or /e_36r1/main/18bc/fc/model eq complete) and ( ../021/backupprod eq complete or ../021/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../021/derived_param:SMOS or ../021/derived_param eq complete) or ( ../021/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '024'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 027
          trigger ( /e_36r1/main/18bc/fc/model:step ge 027 or /e_36r1/main/18bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0027'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/18bc/fc/model:step ge 39 or /e_36r1/main/18bc/fc/model eq complete) and ( ../024/backupprod eq complete or ../024/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../024/derived_param:SMOS or ../024/derived_param eq complete) or ( ../024/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '027'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 030
          trigger ( /e_36r1/main/18bc/fc/model:step ge 030 or /e_36r1/main/18bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0030'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/18bc/fc/model:step ge 42 or /e_36r1/main/18bc/fc/model eq complete) and ( ../027/backupprod eq complete or ../027/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../027/derived_param:SMOS or ../027/derived_param eq complete) or ( ../027/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '030'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 033
          trigger ( /e_36r1/main/18bc/fc/model:step ge 033 or /e_36r1/main/18bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0033'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/18bc/fc/model:step ge 45 or /e_36r1/main/18bc/fc/model eq complete) and ( ../030/backupprod eq complete or ../030/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../030/derived_param:SMOS or ../030/derived_param eq complete) or ( ../030/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '033'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 036
          trigger ( /e_36r1/main/18bc/fc/model:step ge 036 or /e_36r1/main/18bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0036'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/18bc/fc/model:step ge 48 or /e_36r1/main/18bc/fc/model eq complete) and ( ../033/backupprod eq complete or ../033/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../033/derived_param:SMOS or ../033/derived_param eq complete) or ( ../033/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '036'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 039
          trigger ( /e_36r1/main/18bc/fc/model:step ge 039 or /e_36r1/main/18bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0039'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/18bc/fc/model:step ge 51 or /e_36r1/main/18bc/fc/model eq complete) and ( ../036/backupprod eq complete or ../036/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../036/derived_param:SMOS or ../036/derived_param eq complete) or ( ../036/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '039'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 042
          trigger ( /e_36r1/main/18bc/fc/model:step ge 042 or /e_36r1/main/18bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0042'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/18bc/fc/model:step ge 54 or /e_36r1/main/18bc/fc/model eq complete) and ( ../039/backupprod eq complete or ../039/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../039/derived_param:SMOS or ../039/derived_param eq complete) or ( ../039/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '042'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 045
          trigger ( /e_36r1/main/18bc/fc/model:step ge 045 or /e_36r1/main/18bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0045'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/18bc/fc/model:step ge 57 or /e_36r1/main/18bc/fc/model eq complete) and ( ../042/backupprod eq complete or ../042/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../042/derived_param:SMOS or ../042/derived_param eq complete) or ( ../042/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '045'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 048
          trigger ( /e_36r1/main/18bc/fc/model:step ge 048 or /e_36r1/main/18bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0048'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/18bc/fc/model:step ge 60 or /e_36r1/main/18bc/fc/model eq complete) and ( ../045/backupprod eq complete or ../045/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../045/derived_param:SMOS or ../045/derived_param eq complete) or ( ../045/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '048'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 051
          trigger ( /e_36r1/main/18bc/fc/model:step ge 051 or /e_36r1/main/18bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0051'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/18bc/fc/model:step ge 63 or /e_36r1/main/18bc/fc/model eq complete) and ( ../048/backupprod eq complete or ../048/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../048/derived_param:SMOS or ../048/derived_param eq complete) or ( ../048/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '051'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 054
          trigger ( /e_36r1/main/18bc/fc/model:step ge 054 or /e_36r1/main/18bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0054'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/18bc/fc/model:step ge 66 or /e_36r1/main/18bc/fc/model eq complete) and ( ../051/backupprod eq complete or ../051/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../051/derived_param:SMOS or ../051/derived_param eq complete) or ( ../051/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '054'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 057
          trigger ( /e_36r1/main/18bc/fc/model:step ge 057 or /e_36r1/main/18bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0057'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/18bc/fc/model:step ge 69 or /e_36r1/main/18bc/fc/model eq complete) and ( ../054/backupprod eq complete or ../054/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../054/derived_param:SMOS or ../054/derived_param eq complete) or ( ../054/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '057'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 060
          trigger ( /e_36r1/main/18bc/fc/model:step ge 060 or /e_36r1/main/18bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0060'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/18bc/fc/model:step ge 72 or /e_36r1/main/18bc/fc/model eq complete) and ( ../057/backupprod eq complete or ../057/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../057/derived_param:SMOS or ../057/derived_param eq complete) or ( ../057/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '060'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 063
          trigger ( /e_36r1/main/18bc/fc/model:step ge 063 or /e_36r1/main/18bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0063'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/18bc/fc/model:step ge 75 or /e_36r1/main/18bc/fc/model eq complete) and ( ../060/backupprod eq complete or ../060/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../060/derived_param:SMOS or ../060/derived_param eq complete) or ( ../060/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '063'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 066
          trigger ( /e_36r1/main/18bc/fc/model:step ge 066 or /e_36r1/main/18bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0066'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/18bc/fc/model:step ge 78 or /e_36r1/main/18bc/fc/model eq complete) and ( ../063/backupprod eq complete or ../063/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../063/derived_param:SMOS or ../063/derived_param eq complete) or ( ../063/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '066'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 069
          trigger ( /e_36r1/main/18bc/fc/model:step ge 069 or /e_36r1/main/18bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0069'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/18bc/fc/model:step ge 81 or /e_36r1/main/18bc/fc/model eq complete) and ( ../066/backupprod eq complete or ../066/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../066/derived_param:SMOS or ../066/derived_param eq complete) or ( ../066/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '069'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 072
          trigger ( /e_36r1/main/18bc/fc/model:step ge 072 or /e_36r1/main/18bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0072'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/18bc/fc/model:step ge 84 or /e_36r1/main/18bc/fc/model eq complete) and ( ../069/backupprod eq complete or ../069/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../069/derived_param:SMOS or ../069/derived_param eq complete) or ( ../069/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '072'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 075
          trigger ( /e_36r1/main/18bc/fc/model:step ge 075 or /e_36r1/main/18bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0075'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/18bc/fc/model:step ge 87 or /e_36r1/main/18bc/fc/model eq complete) and ( ../072/backupprod eq complete or ../072/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../072/derived_param:SMOS or ../072/derived_param eq complete) or ( ../072/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '075'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 078
          trigger ( /e_36r1/main/18bc/fc/model:step ge 078 or /e_36r1/main/18bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0078'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/18bc/fc/model:step ge 90 or /e_36r1/main/18bc/fc/model eq complete) and ( ../075/backupprod eq complete or ../075/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../075/derived_param:SMOS or ../075/derived_param eq complete) or ( ../075/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '078'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 081
          trigger ( /e_36r1/main/18bc/fc/model:step ge 081 or /e_36r1/main/18bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0081'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task derived_param
            trigger ( ../078/derived_param:SMOS or ../078/derived_param eq complete) or ( ../078/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '081'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 084
          trigger ( /e_36r1/main/18bc/fc/model:step ge 084 or /e_36r1/main/18bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0084'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task derived_param
            trigger ( ../081/derived_param:SMOS or ../081/derived_param eq complete) or ( ../081/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '084'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 087
          trigger ( /e_36r1/main/18bc/fc/model:step ge 087 or /e_36r1/main/18bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0087'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task derived_param
            trigger ( ../084/derived_param:SMOS or ../084/derived_param eq complete) or ( ../084/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '087'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 090
          trigger ( /e_36r1/main/18bc/fc/model:step ge 090 or /e_36r1/main/18bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0090'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task derived_param
            trigger ( ../087/derived_param:SMOS or ../087/derived_param eq complete) or ( ../087/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '090'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
      endfamily
      family prod2diss
        edit USER_PRIORITY '90'
        edit SMSURLBASE 'http://diss.ecmwf.int/do/monitoring/summary'
        edit SMSFILES '/home/ma/emos/def/e_36r1/prod'
        task operation_is_late
          event yes
        family oper
          edit EMOS_DOMAIN 'G'
          edit EMOS_TYPE 'fc'
          edit FSFAMILY 'ppdi'
          edit PROD_TASK '1'
          edit USER_PRIORITY '92'
          edit SMSURL 'GSCDA/18'
          edit EMOS_PROJECT '12'
          family an
            task pdb
              trigger ../../../prod/an/prodgen eq complete
              edit EMOS_TYPE 'an'
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete
              edit EMOS_TYPE 'an'
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../090/q2diss eq complete and ../../../prod/an/backupprod eq complete
              edit EMOS_TYPE 'an'
              label info ""
              event 1 
          endfamily
          family 000
            edit EMOS_TIME_STEP_H '0000'
            task pdb
              trigger ../../../prod/000/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ../an/q2diss eq complete
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
          endfamily
          family 003
            edit EMOS_TIME_STEP_H '0003'
            task pdb
              trigger ../../../prod/003/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../000/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/003/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 006
            edit EMOS_TIME_STEP_H '0006'
            task pdb
              trigger ../../../prod/006/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../003/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/006/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 009
            edit EMOS_TIME_STEP_H '0009'
            task pdb
              trigger ../../../prod/009/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../006/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/009/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 012
            edit EMOS_TIME_STEP_H '0012'
            task pdb
              trigger ../../../prod/012/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../009/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/012/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 015
            edit EMOS_TIME_STEP_H '0015'
            task pdb
              trigger ../../../prod/015/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../012/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/015/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 018
            edit EMOS_TIME_STEP_H '0018'
            task pdb
              trigger ../../../prod/018/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../015/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/018/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 021
            edit EMOS_TIME_STEP_H '0021'
            task pdb
              trigger ../../../prod/021/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../018/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/021/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 024
            edit EMOS_TIME_STEP_H '0024'
            task pdb
              trigger ../../../prod/024/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../021/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/024/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 027
            edit EMOS_TIME_STEP_H '0027'
            task pdb
              trigger ../../../prod/027/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../024/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/027/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 030
            edit EMOS_TIME_STEP_H '0030'
            task pdb
              trigger ../../../prod/030/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../027/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/030/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 033
            edit EMOS_TIME_STEP_H '0033'
            task pdb
              trigger ../../../prod/033/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../030/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/033/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 036
            edit EMOS_TIME_STEP_H '0036'
            task pdb
              trigger ../../../prod/036/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../033/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/036/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 039
            edit EMOS_TIME_STEP_H '0039'
            task pdb
              trigger ../../../prod/039/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../036/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/039/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 042
            edit EMOS_TIME_STEP_H '0042'
            task pdb
              trigger ../../../prod/042/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../039/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/042/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 045
            edit EMOS_TIME_STEP_H '0045'
            task pdb
              trigger ../../../prod/045/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../042/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/045/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 048
            edit EMOS_TIME_STEP_H '0048'
            task pdb
              trigger ../../../prod/048/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../045/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/048/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 051
            edit EMOS_TIME_STEP_H '0051'
            task pdb
              trigger ../../../prod/051/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../048/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/051/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 054
            edit EMOS_TIME_STEP_H '0054'
            task pdb
              trigger ../../../prod/054/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../051/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/054/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 057
            edit EMOS_TIME_STEP_H '0057'
            task pdb
              trigger ../../../prod/057/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../054/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/057/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 060
            edit EMOS_TIME_STEP_H '0060'
            task pdb
              trigger ../../../prod/060/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../057/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/060/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 063
            edit EMOS_TIME_STEP_H '0063'
            task pdb
              trigger ../../../prod/063/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../060/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/063/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 066
            edit EMOS_TIME_STEP_H '0066'
            task pdb
              trigger ../../../prod/066/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../063/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/066/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 069
            edit EMOS_TIME_STEP_H '0069'
            task pdb
              trigger ../../../prod/069/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../066/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/069/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 072
            edit EMOS_TIME_STEP_H '0072'
            task pdb
              trigger ../../../prod/072/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../069/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/072/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 075
            edit EMOS_TIME_STEP_H '0075'
            task pdb
              trigger ../../../prod/075/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../072/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/075/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 078
            edit EMOS_TIME_STEP_H '0078'
            task pdb
              trigger ../../../prod/078/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../075/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/078/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 081
            edit EMOS_TIME_STEP_H '0081'
            task pdb
              trigger ../../../prod/081/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../078/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
          endfamily
          family 084
            edit EMOS_TIME_STEP_H '0084'
            task pdb
              trigger ../../../prod/084/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../081/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
          endfamily
          family 087
            edit EMOS_TIME_STEP_H '0087'
            task pdb
              trigger ../../../prod/087/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../084/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
          endfamily
          family 090
            edit EMOS_TIME_STEP_H '0090'
            task pdb
              trigger ../../../prod/090/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../087/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
          endfamily
          task check
            trigger ./078 eq complete
        endfamily
      endfamily
    endfamily
    family 12dc
      trigger make eq complete
      edit EMOS_BASE '12'
      edit MXUP_TRAJ '3'
      edit STREAM 'DCDA'
      edit WINDOW_LENGTH_4D '12'
      edit WINDOW_OFFSET_4D '3'
      edit EDFAMILY 'false'
      edit PERIOD_4D '12'
      edit FCLENGTH '15'
      edit DELTA_DAY '0'
      family obs
        edit FSFAMILY 'an'
        edit SMSFILES '/home/ma/emos/def/e_36r1/obs'
        family get
          inlimit /e_36r1/limits:obs_get
          inlimit /limits:obs_get
          task get_nonbin
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc12/nonbin eq complete
            edit OBSFILE_OBSHOST 'BUFR'
            edit OBSFILE_HPCF 'bufr'
          task get_ers1_wa
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc12/ers1_wa eq complete
            edit OBSFILE_OBSHOST 'UWAX'
            edit OBSFILE_HPCF 'wavescat'
          task get_eas1
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc12/eas1 eq complete
            edit OBSFILE_OBSHOST 'EAS1'
            edit OBSFILE_HPCF 'wavesar'
          task get_ers1_wi
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc12/ers1_wi eq complete
            edit OBSFILE_OBSHOST 'UWIX'
            edit OBSFILE_HPCF 'scat'
          task get_ers1_ra
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc12/ers1_ra eq complete
            edit OBSFILE_OBSHOST 'URAX'
            edit OBSFILE_HPCF 'ralt'
          task get_rtovs
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc12/rtovs eq complete
            edit OBSFILE_OBSHOST 'TOVH'
            edit OBSFILE_HPCF 'tovh'
          task get_ssbt
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc12/ssbt eq complete
            edit OBSFILE_OBSHOST 'SSBT'
            edit OBSFILE_HPCF 'ssbt'
          task get_sst
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc12/sst/sst_fdb_c1b eq complete
            edit OBSFILE_OBSHOST 'nil'
            edit OBSFILE_HPCF 'nil'
          task get_atov
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc12/atov eq complete
            edit OBSFILE_OBSHOST 'ATOV'
            edit OBSFILE_HPCF 'atov'
          task get_grad
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc12/grad eq complete
            edit OBSFILE_OBSHOST 'GRAD'
            edit OBSFILE_HPCF 'grad'
          task get_stor
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc12/stor eq complete
            edit OBSFILE_OBSHOST 'STOR'
            edit OBSFILE_HPCF 'storm'
          task get_trmm
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc12/trmm eq complete
            edit OBSFILE_OBSHOST 'TRMM'
            edit OBSFILE_HPCF 'tmi'
          task get_qsca
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc12/qsca eq complete
            edit OBSFILE_OBSHOST 'QSCA'
            edit OBSFILE_HPCF 'qsca'
          task get_sbuv
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc12/sbuv eq complete
            edit OBSFILE_OBSHOST 'SBUV'
            edit OBSFILE_HPCF 'reo3'
          task get_airs
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc12/airs eq complete
            edit OBSFILE_OBSHOST 'AIRS'
            edit OBSFILE_HPCF 'airs'
          task get_ice
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc12/ice/ice_fdb_c1b eq complete
            edit OBSFILE_OBSHOST 'nil'
            edit OBSFILE_HPCF 'nil'
          task get_snow
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc12/snow eq complete
            edit OBSFILE_OBSHOST 'SNOW'
            edit OBSFILE_HPCF 'imssnow'
          task get_ateu
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc12/ateu eq complete
            edit OBSFILE_OBSHOST 'ATEU'
            edit OBSFILE_HPCF 'amsu_ears'
          task get_bice
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc12/bice eq complete
            edit OBSFILE_OBSHOST 'BICE'
            edit OBSFILE_HPCF 'baltic_ci'
          task get_amsa
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc12/amsa eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'AMSA'
            edit OBSFILE_HPCF 'amsa'
          task get_mhsa
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc12/mhsa eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'MHSA'
            edit OBSFILE_HPCF 'mhsa'
          task get_hirs
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc12/hirs eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'HIRS'
            edit OBSFILE_HPCF 'hirs'
          task get_amsr
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc12/amsr eq complete
            edit OBSFILE_OBSHOST 'AMSR'
            edit OBSFILE_HPCF 'amsre_aq'
          task get_rado
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc12/rado eq complete
            edit OBSFILE_OBSHOST 'RADO'
            edit OBSFILE_HPCF 'gpsro'
          task get_smin
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc12/smin eq complete
            edit OBSFILE_OBSHOST 'SMIN'
            edit OBSFILE_HPCF 'ssmisbt'
          task get_iasi
            late -s +00:02 -c +00:07
            trigger /pp2/main/extractions/dc12/iasi eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'IASI'
            edit OBSFILE_HPCF 'iasi'
          task get_gras
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc12/gras eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'GRAS'
            edit OBSFILE_HPCF 'gras'
          task get_modw
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc12/modw eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'MODW'
            edit OBSFILE_HPCF 'dbmodiswinds'
          task get_avhn
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc12/avhn eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'AVHN'
            edit OBSFILE_HPCF 'avhrrwinds'
          task get_atap
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc12/atap eq complete
            edit OBSFILE_OBSHOST 'ATAP'
            edit OBSFILE_HPCF 'amsu_pacrars'
          task get_g2to
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc12/g2to eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'G2TO'
            edit OBSFILE_HPCF 'reo3_gome2'
          task get_grst
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc12/grst eq complete
            edit OBSFILE_OBSHOST 'GRST'
            edit OBSFILE_HPCF 'ostia'
          task get_emer
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc12/emer eq complete
            edit OBSFILE_OBSHOST 'EMER'
            edit OBSFILE_HPCF 'envisat_meris'
          task get_j2nb
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc12/j2nb eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'J2NB'
            edit OBSFILE_HPCF 'j2nb'
          task get_j2eo
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc12/j2eo eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'J2EO'
            edit OBSFILE_HPCF 'j2eo'
          task get_asc2
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc12/asc2 eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'ASC2'
            edit OBSFILE_HPCF 'ascat_l2'
          task get_asel
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc12/asel eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'ASEL'
            edit OBSFILE_HPCF 'ascat_ears'
          task get_ascs
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc12/ascs eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'ASCS'
            edit OBSFILE_HPCF 'ascat_sm'
          task atovairs
            trigger get_atov eq complete and get_airs eq complete and get_ateu eq complete and get_amsa eq complete and get_mhsa eq complete and get_hirs eq complete and get_atap eq complete
          task snowlinks
            trigger get_nonbin eq complete and get_ers1_wi eq complete and get_rtovs eq complete and get_snow eq complete
          task ro_combine
            trigger get_rado eq complete and get_gras eq complete
          task reo3_combine
            trigger get_sbuv eq complete and get_g2to eq complete
          task bufr_combine
            trigger get_nonbin eq complete and get_modw eq complete and get_avhn eq complete
          task jason2_combine
            trigger get_ers1_ra eq complete and get_j2nb eq complete and get_j2eo eq complete
          task ascat_combine
            trigger get_asc2 eq complete and get_asel eq complete and get_ascs eq complete
        endfamily
        family getplus6
          edit EMOS_BASE '18'
          inlimit /e_36r1/limits:obs_get
          inlimit /limits:obs_get
          task get_nonbin
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc18/nonbin eq complete
            edit OBSFILE_OBSHOST 'BUFR'
            edit OBSFILE_HPCF 'bufr'
          task get_ers1_wa
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc18/ers1_wa eq complete
            edit OBSFILE_OBSHOST 'UWAX'
            edit OBSFILE_HPCF 'wavescat'
          task get_eas1
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc18/eas1 eq complete
            edit OBSFILE_OBSHOST 'EAS1'
            edit OBSFILE_HPCF 'wavesar'
          task get_ers1_wi
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc18/ers1_wi eq complete
            edit OBSFILE_OBSHOST 'UWIX'
            edit OBSFILE_HPCF 'scat'
          task get_ers1_ra
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc18/ers1_ra eq complete
            edit OBSFILE_OBSHOST 'URAX'
            edit OBSFILE_HPCF 'ralt'
          task get_rtovs
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc18/rtovs eq complete
            edit OBSFILE_OBSHOST 'TOVH'
            edit OBSFILE_HPCF 'tovh'
          task get_ssbt
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc18/ssbt eq complete
            edit OBSFILE_OBSHOST 'SSBT'
            edit OBSFILE_HPCF 'ssbt'
          task get_sst
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc18/sst/sst_fdb_c1b eq complete
            edit OBSFILE_OBSHOST 'nil'
            edit OBSFILE_HPCF 'nil'
          task get_atov
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc18/atov eq complete
            edit OBSFILE_OBSHOST 'ATOV'
            edit OBSFILE_HPCF 'atov'
          task get_grad
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc18/grad eq complete
            edit OBSFILE_OBSHOST 'GRAD'
            edit OBSFILE_HPCF 'grad'
          task get_stor
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc18/stor eq complete
            edit OBSFILE_OBSHOST 'STOR'
            edit OBSFILE_HPCF 'storm'
          task get_trmm
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc18/trmm eq complete
            edit OBSFILE_OBSHOST 'TRMM'
            edit OBSFILE_HPCF 'tmi'
          task get_qsca
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc18/qsca eq complete
            edit OBSFILE_OBSHOST 'QSCA'
            edit OBSFILE_HPCF 'qsca'
          task get_sbuv
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc18/sbuv eq complete
            edit OBSFILE_OBSHOST 'SBUV'
            edit OBSFILE_HPCF 'reo3'
          task get_airs
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc18/airs eq complete
            edit OBSFILE_OBSHOST 'AIRS'
            edit OBSFILE_HPCF 'airs'
          task get_ice
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc18/ice/ice_fdb_c1b eq complete
            edit OBSFILE_OBSHOST 'nil'
            edit OBSFILE_HPCF 'nil'
          task get_snow
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc18/snow eq complete
            edit OBSFILE_OBSHOST 'SNOW'
            edit OBSFILE_HPCF 'imssnow'
          task get_ateu
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc18/ateu eq complete
            edit OBSFILE_OBSHOST 'ATEU'
            edit OBSFILE_HPCF 'amsu_ears'
          task get_bice
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc18/bice eq complete
            edit OBSFILE_OBSHOST 'BICE'
            edit OBSFILE_HPCF 'baltic_ci'
          task get_amsa
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc18/amsa eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'AMSA'
            edit OBSFILE_HPCF 'amsa'
          task get_mhsa
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc18/mhsa eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'MHSA'
            edit OBSFILE_HPCF 'mhsa'
          task get_hirs
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc18/hirs eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'HIRS'
            edit OBSFILE_HPCF 'hirs'
          task get_amsr
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc18/amsr eq complete
            edit OBSFILE_OBSHOST 'AMSR'
            edit OBSFILE_HPCF 'amsre_aq'
          task get_rado
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc18/rado eq complete
            edit OBSFILE_OBSHOST 'RADO'
            edit OBSFILE_HPCF 'gpsro'
          task get_smin
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc18/smin eq complete
            edit OBSFILE_OBSHOST 'SMIN'
            edit OBSFILE_HPCF 'ssmisbt'
          task get_iasi
            late -s +00:02 -c +00:07
            trigger /pp2/main/extractions/dc18/iasi eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'IASI'
            edit OBSFILE_HPCF 'iasi'
          task get_gras
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc18/gras eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'GRAS'
            edit OBSFILE_HPCF 'gras'
          task get_modw
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc18/modw eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'MODW'
            edit OBSFILE_HPCF 'dbmodiswinds'
          task get_avhn
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc18/avhn eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'AVHN'
            edit OBSFILE_HPCF 'avhrrwinds'
          task get_atap
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc18/atap eq complete
            edit OBSFILE_OBSHOST 'ATAP'
            edit OBSFILE_HPCF 'amsu_pacrars'
          task get_g2to
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc18/g2to eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'G2TO'
            edit OBSFILE_HPCF 'reo3_gome2'
          task get_grst
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc18/grst eq complete
            edit OBSFILE_OBSHOST 'GRST'
            edit OBSFILE_HPCF 'ostia'
          task get_emer
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/dc18/emer eq complete
            edit OBSFILE_OBSHOST 'EMER'
            edit OBSFILE_HPCF 'envisat_meris'
          task get_j2nb
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc18/j2nb eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'J2NB'
            edit OBSFILE_HPCF 'j2nb'
          task get_j2eo
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc18/j2eo eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'J2EO'
            edit OBSFILE_HPCF 'j2eo'
          task get_asc2
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc18/asc2 eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'ASC2'
            edit OBSFILE_HPCF 'ascat_l2'
          task get_asel
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc18/asel eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'ASEL'
            edit OBSFILE_HPCF 'ascat_ears'
          task get_ascs
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/dc18/ascs eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'ASCS'
            edit OBSFILE_HPCF 'ascat_sm'
          task atovairs
            trigger get_atov eq complete and get_airs eq complete and get_ateu eq complete and get_amsa eq complete and get_mhsa eq complete and get_hirs eq complete and get_atap eq complete
          task snowlinks
            trigger get_nonbin eq complete and get_ers1_wi eq complete and get_rtovs eq complete and get_snow eq complete
          task ro_combine
            trigger get_rado eq complete and get_gras eq complete
          task reo3_combine
            trigger get_sbuv eq complete and get_g2to eq complete
          task bufr_combine
            trigger get_nonbin eq complete and get_modw eq complete and get_avhn eq complete
          task jason2_combine
            trigger get_ers1_ra eq complete and get_j2nb eq complete and get_j2eo eq complete
          task ascat_combine
            trigger get_asc2 eq complete and get_asel eq complete and get_ascs eq complete
        endfamily
        task fetchmars
          trigger ./get eq complete and ./getplus6 eq complete
        family prepare_obs
          trigger ./get eq complete and ./getplus6 eq complete
          task preobs
            trigger /e_36r1/main/00dc/an/update_psbias eq complete and /e_36r1/main/00dc/an/update_rstrhbias eq complete
            edit mode 'not1crad'
          task premwimg
          task preobs_wave
            edit mode 'wave'
          task pre1crad_prepare
            edit mode '1crad'
          task pre1crad_airs
            trigger pre1crad_prepare eq complete
          family pre1crad_iasi
            trigger pre1crad_prepare eq complete
            edit CPUTIME '1500'
            edit MAXTASK '16'
            task pre1crad_iasi_split_1
              edit TASKNO '1'
            task pre1crad_iasi_split_2
              edit TASKNO '2'
            task pre1crad_iasi_1
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '1'
            task pre1crad_iasi_2
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '2'
            task pre1crad_iasi_3
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '3'
            task pre1crad_iasi_4
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '4'
            task pre1crad_iasi_5
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '5'
            task pre1crad_iasi_6
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '6'
            task pre1crad_iasi_7
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '7'
            task pre1crad_iasi_8
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '8'
            task pre1crad_iasi_9
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '9'
            task pre1crad_iasi_10
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '10'
            task pre1crad_iasi_11
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '11'
            task pre1crad_iasi_12
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '12'
            task pre1crad_iasi_13
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '13'
            task pre1crad_iasi_14
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '14'
            task pre1crad_iasi_15
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '15'
            task pre1crad_iasi_16
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '16'
          endfamily
          task pre1crad_hirs
            trigger pre1crad_prepare eq complete
          task pre1crad_msu
            trigger pre1crad_prepare eq complete
          task pre1crad_ssu
            trigger pre1crad_prepare eq complete
          task pre1crad_amsua
            trigger pre1crad_prepare eq complete
          task pre1crad_amsub
            trigger pre1crad_prepare eq complete
          task pre1crad_mhs
            trigger pre1crad_prepare eq complete
          task pregeos
          task prereo3
          task prescat
        endfamily
        family reservation
          late -c +00:03
          trigger ./get/get_stor eq complete
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
          inlimit /e_36r1/limits:reservation
          task adv_res
            late -s +00:03 -c +00:02
            edit RESERVE_FOR '30'
            edit RESERVE_IN '30'
            edit RESERVED_NODES '48'
            edit RESERVATION_LABEL 'emos12dcA'
            label info ""
        endfamily
      endfamily
      family an
        trigger ./obs/fetchmars eq complete and ./obs/prepare_obs eq complete
        edit FSFAMILY 'an'
        edit SMSFILES '/home/ma/emos/def/e_36r1/an'
        edit PERIOD_4D '12'
        edit WINDOW_LENGTH_4D '12'
        edit WINDOW_OFFSET_4D '3'
        edit ACCOUNT 'oedcdaan'
        family makeodb
          edit NPES '16'
          task cleanodb
          family bufr2odb
            trigger cleanodb eq complete
            edit THREADS '1'
            edit NPOOLS '%ANNPES%'
            edit THINFAC '1'
            edit SENSOR '-1'
            edit REBALANCE '1'
            edit SUBTYPES_INCLUDE ''
            edit SUBTYPES_EXCLUDE ''
            task b2o_conv
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_EXCLUDE '85 82 83 87'
              edit USER_PRIORITY '90'
            task b2o_surf_conv
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_EXCLUDE '142 143 144 145 85 82 83 87'
            task b2o_airs
              late -s +00:05 -c +00:15
              edit THREADS '2'
              edit SENSOR '11'
              edit SUBTYPES_INCLUDE '57'
              edit NPOOLS '$((2* %ANNPES%))'
            task b2o_iasi
              late -s +00:05 -c +00:15
              edit SENSOR '993'
              edit SUBTYPES_INCLUDE '240'
              edit THREADS '4'
              edit NPOOLS '$((2* %ANNPES%))'
            task b2o_hirs
              late -s +00:05 -c +00:15
              edit SENSOR '0'
              edit SUBTYPES_INCLUDE '54 55'
              edit NPOOLS '$((2* %ANNPES%))'
            task b2o_msu
              late -s +00:05 -c +00:15
              edit SENSOR '1'
              edit SUBTYPES_INCLUDE '54 55'
            task b2o_ssu
              late -s +00:05 -c +00:15
              edit SENSOR '2'
              edit SUBTYPES_INCLUDE '54 55'
            task b2o_amsua
              late -s +00:05 -c +00:15
              edit SENSOR '3'
              edit SUBTYPES_INCLUDE '54 55'
            task b2o_amsub
              late -s +00:05 -c +00:15
              edit SENSOR '4'
              edit SUBTYPES_INCLUDE '54 55'
            task b2o_mhs
              late -s +00:05 -c +00:15
              edit SENSOR '11'
              edit SUBTYPES_INCLUDE '54 55'
            task b2o_satob
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '82 83 87'
            task b2o_reo3
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '205 206 207 208 215'
            task b2o_ssmi
              late -s +00:05 -c +00:15
              edit SENSOR '6'
              edit SUBTYPES_INCLUDE '55 127'
            task b2o_ssmis
              late -s +00:05 -c +00:15
              edit SENSOR '10'
              edit SUBTYPES_INCLUDE '49'
            task b2o_amsre
              late -s +00:05 -c +00:15
              edit SENSOR '17'
              edit SUBTYPES_INCLUDE '59'
            task b2o_tmi
              late -s +00:05 -c +00:15
              edit SENSOR '9'
              edit SUBTYPES_INCLUDE '129'
            task b2o_scatt
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '122 137 139'
            task b2o_geos
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '189 89 76 77 78'
            task b2o_gpsro
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '250'
            task b2o_meris
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '212'
          endfamily
          task mergeodb
            trigger bufr2odb eq complete
          task mergeodb_surf
            trigger ./bufr2odb/b2o_surf_conv eq complete
            edit NPES '16'
            edit THREADS '1'
        endfamily
        task restartodb
          trigger makeodb eq complete
        task preCleanFDB
        task vardata
          trigger ( ../../00dc/fc/model:step ge 9 or ../../00dc/fc/model eq complete)
          event hires
        task fetcherr
          trigger ../../00dc/an eq complete
        task lowres
          trigger vardata:hires
          edit THREADS '4'
          edit ININPES '32'
        family surf_anal
          trigger makeodb eq complete and restartodb eq complete and preCleanFDB eq complete and vardata:hires
          edit SMSFILES '/home/ma/emos/def/e_36r1/an'
          edit FSFAMILY 'surf_anal'
          task t2ana
            edit mode 't2m'
          task rh2ana
            trigger t2ana eq complete
            edit mode 'rh2m'
          task snow
            trigger rh2ana eq complete
            edit mode 'snow'
          task sst
            edit THREADS '4'
          task slwet
            trigger sst eq complete and snow eq complete
        endfamily
        family 4dvar
          trigger makeodb eq complete and restartodb eq complete and vardata eq complete and fetcherr eq complete and lowres eq complete and preCleanFDB eq complete
          edit THREADS '8'
          edit VAR_DELAY '00'
          edit EMOS_SLEEP '00'
          family uptraj_0
            edit uptraj '0'
            task ifstraj
              late -s +00:02
              edit mode 'traj'
              edit USER_PRIORITY '91'
              edit SMSTRIES '1'
            task ifsmin
              late -s +00:02
              trigger ifstraj eq complete
              edit mode 'min'
              edit USER_PRIORITY '92'
            task ifstsave
              trigger ifstraj eq complete
              edit THREADS '1'
              edit USER_PRIORITY '93'
          endfamily
          family uptraj_1
            trigger uptraj_0 eq complete
            edit uptraj '1'
            task ifstraj
              late -s +00:02
              edit mode 'traj'
              edit USER_PRIORITY '93'
            task ifsmin
              late -s +00:02
              trigger ifstraj eq complete
              edit mode 'min'
              edit USER_PRIORITY '94'
            task ifstsave
              trigger ifstraj eq complete
              edit THREADS '1'
              edit USER_PRIORITY '95'
          endfamily
          family uptraj_2
            trigger uptraj_1 eq complete
            edit uptraj '2'
            task ifstraj
              late -s +00:02
              edit mode 'traj'
              edit USER_PRIORITY '95'
            task ifsmin
              late -s +00:02
              trigger ifstraj eq complete
              edit mode 'min'
              edit USER_PRIORITY '96'
            task ifstsave
              trigger ifstraj eq complete
              edit THREADS '1'
              edit USER_PRIORITY '97'
          endfamily
          task ifstraj
            late -s +00:02
            trigger uptraj_2 eq complete
            edit uptraj '3'
            edit mode 'traj'
            edit USER_PRIORITY '97'
            event finaltraj
            event finalwave
        endfamily
        task forceinv
          trigger ./4dvar/ifstraj:finaltraj and ./surf_anal eq complete
        task emiskf
          trigger ./4dvar/uptraj_0/ifstraj eq complete
          edit THREADS '1'
        task update_psbias
          trigger ./4dvar/ifstraj:finalwave
        task update_rstrhbias
          trigger ./4dvar/ifstraj:finalwave
        task dcda_to_da
          trigger ./4dvar eq complete and surf_anal eq complete
          edit EMOS_TYPE 'an'
          edit OPHOST 'c1b'
        family reservation1
          late -c +00:03
          trigger ./4dvar/uptraj_0/ifstraj eq active or ./4dvar/uptraj_0/ifstraj eq complete
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
          inlimit /e_36r1/limits:reservation
          task reserve_cpu
            late -s +00:02 -c +00:02
            edit RESERVED_CPUS '3328'
            edit RESERVE_FOR '65'
            label info ""
          task rm_res
            late -s +00:02 -c +00:02
            edit RESERVATION_LABEL 'emos12'
            label info ""
          task set_class
            late -s +00:02 -c +00:02
            edit RESERVATION_DEFINITION 'default_nodiag'
            label info ""
        endfamily
      endfamily
      family fc
        trigger an/forceinv eq complete and ( an/4dvar/ifstraj:finalwave or an/4dvar eq complete)
        edit EMOS_BASE '18'
        edit ACCOUNT 'oedcdafc'
        edit FSFAMILY 'fc'
        task getfcdata
        task model
          late -s +00:02
          trigger getfcdata eq complete
          edit FCLENGTH '15'
          edit THREADS '8'
          edit USER_PRIORITY '94'
          meter step -1 15 15
        task ddh
          trigger model eq complete
        task clean
          trigger ../an eq complete and getfcdata eq complete and model eq complete and ddh eq complete and ../../12/fc eq complete
        family reservation1
          late -c +00:03
          trigger ./getfcdata eq active or ./getfcdata eq complete
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
          inlimit /e_36r1/limits:reservation
          task reserve_cpu
            late -s +00:03 -c +00:02
            edit RESERVED_CPUS '3072'
            edit RESERVE_FOR '4'
            label info ""
        endfamily
        family reservation2
          late -c +00:03
          trigger ./model eq complete
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
          inlimit /e_36r1/limits:reservation
          task reset_cpu
            late -s +00:03 -c +00:02
            complete /e_36r1/main/00/an/vardata eq active or /e_36r1/main/00/an/vardata eq complete
            edit RESERVED_CPUS '0'
            label info ""
        endfamily
      endfamily
    endfamily
    family 18
      trigger make eq complete
      edit EMOS_BASE '18'
      edit MXUP_TRAJ '3'
      edit REINISTREAM 'DA'
      edit LAR_4V_EF 'false'
      edit LFC_DIR_FREQ 'false'
      edit WINDOW_LENGTH_4D '6'
      edit WINDOW_OFFSET_4D '3'
      edit EDFAMILY 'true'
      edit PERIOD_4D '6'
      edit FCLENGTH '9'
      edit DELTA_DAY '0'
      family obs
        edit FSFAMILY 'an'
        edit SMSFILES '/home/ma/emos/def/e_36r1/obs'
        family get
          inlimit /e_36r1/limits:obs_get
          inlimit /limits:obs_get
          task get_nonbin
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/nonbin eq complete
            edit OBSFILE_OBSHOST 'BUFR'
            edit OBSFILE_HPCF 'bufr'
          task get_ers1_wa
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/ers1_wa eq complete
            edit OBSFILE_OBSHOST 'UWAX'
            edit OBSFILE_HPCF 'wavescat'
          task get_eas1
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/eas1 eq complete
            edit OBSFILE_OBSHOST 'EAS1'
            edit OBSFILE_HPCF 'wavesar'
          task get_ers1_wi
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/ers1_wi eq complete
            edit OBSFILE_OBSHOST 'UWIX'
            edit OBSFILE_HPCF 'scat'
          task get_ers1_ra
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/ers1_ra eq complete
            edit OBSFILE_OBSHOST 'URAX'
            edit OBSFILE_HPCF 'ralt'
          task get_rtovs
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/rtovs eq complete
            edit OBSFILE_OBSHOST 'TOVH'
            edit OBSFILE_HPCF 'tovh'
          task get_ssbt
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/ssbt eq complete
            edit OBSFILE_OBSHOST 'SSBT'
            edit OBSFILE_HPCF 'ssbt'
          task get_sst
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/sst/sst_fdb_c1b eq complete
            edit OBSFILE_OBSHOST 'nil'
            edit OBSFILE_HPCF 'nil'
          task get_atov
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/atov eq complete
            edit OBSFILE_OBSHOST 'ATOV'
            edit OBSFILE_HPCF 'atov'
          task get_grad
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/grad eq complete
            edit OBSFILE_OBSHOST 'GRAD'
            edit OBSFILE_HPCF 'grad'
          task get_stor
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/stor eq complete
            edit OBSFILE_OBSHOST 'STOR'
            edit OBSFILE_HPCF 'storm'
          task get_trmm
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/trmm eq complete
            edit OBSFILE_OBSHOST 'TRMM'
            edit OBSFILE_HPCF 'tmi'
          task get_qsca
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/qsca eq complete
            edit OBSFILE_OBSHOST 'QSCA'
            edit OBSFILE_HPCF 'qsca'
          task get_sbuv
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/sbuv eq complete
            edit OBSFILE_OBSHOST 'SBUV'
            edit OBSFILE_HPCF 'reo3'
          task get_airs
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/airs eq complete
            edit OBSFILE_OBSHOST 'AIRS'
            edit OBSFILE_HPCF 'airs'
          task get_ice
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/ice/ice_fdb_c1b eq complete
            edit OBSFILE_OBSHOST 'nil'
            edit OBSFILE_HPCF 'nil'
          task get_snow
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/snow eq complete
            edit OBSFILE_OBSHOST 'SNOW'
            edit OBSFILE_HPCF 'imssnow'
          task get_ateu
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/ateu eq complete
            edit OBSFILE_OBSHOST 'ATEU'
            edit OBSFILE_HPCF 'amsu_ears'
          task get_bice
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/bice eq complete
            edit OBSFILE_OBSHOST 'BICE'
            edit OBSFILE_HPCF 'baltic_ci'
          task get_amsa
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex18/amsa eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'AMSA'
            edit OBSFILE_HPCF 'amsa'
          task get_mhsa
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex18/mhsa eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'MHSA'
            edit OBSFILE_HPCF 'mhsa'
          task get_hirs
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex18/hirs eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'HIRS'
            edit OBSFILE_HPCF 'hirs'
          task get_amsr
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/amsr eq complete
            edit OBSFILE_OBSHOST 'AMSR'
            edit OBSFILE_HPCF 'amsre_aq'
          task get_rado
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/rado eq complete
            edit OBSFILE_OBSHOST 'RADO'
            edit OBSFILE_HPCF 'gpsro'
          task get_smin
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/smin eq complete
            edit OBSFILE_OBSHOST 'SMIN'
            edit OBSFILE_HPCF 'ssmisbt'
          task get_iasi
            late -s +00:02 -c +00:07
            trigger /pp2/main/extractions/ex18/iasi eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'IASI'
            edit OBSFILE_HPCF 'iasi'
          task get_gras
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex18/gras eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'GRAS'
            edit OBSFILE_HPCF 'gras'
          task get_modw
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex18/modw eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'MODW'
            edit OBSFILE_HPCF 'dbmodiswinds'
          task get_avhn
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex18/avhn eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'AVHN'
            edit OBSFILE_HPCF 'avhrrwinds'
          task get_atap
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/atap eq complete
            edit OBSFILE_OBSHOST 'ATAP'
            edit OBSFILE_HPCF 'amsu_pacrars'
          task get_g2to
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex18/g2to eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'G2TO'
            edit OBSFILE_HPCF 'reo3_gome2'
          task get_grst
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/grst eq complete
            edit OBSFILE_OBSHOST 'GRST'
            edit OBSFILE_HPCF 'ostia'
          task get_emer
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex18/emer eq complete
            edit OBSFILE_OBSHOST 'EMER'
            edit OBSFILE_HPCF 'envisat_meris'
          task get_j2nb
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex18/j2nb eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'J2NB'
            edit OBSFILE_HPCF 'j2nb'
          task get_j2eo
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex18/j2eo eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'J2EO'
            edit OBSFILE_HPCF 'j2eo'
          task get_asc2
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex18/asc2 eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'ASC2'
            edit OBSFILE_HPCF 'ascat_l2'
          task get_asel
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex18/asel eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'ASEL'
            edit OBSFILE_HPCF 'ascat_ears'
          task get_ascs
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex18/ascs eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'ASCS'
            edit OBSFILE_HPCF 'ascat_sm'
          task atovairs
            trigger get_atov eq complete and get_airs eq complete and get_ateu eq complete and get_amsa eq complete and get_mhsa eq complete and get_hirs eq complete and get_atap eq complete
          task snowlinks
            trigger get_nonbin eq complete and get_ers1_wi eq complete and get_rtovs eq complete and get_snow eq complete
          task ro_combine
            trigger get_rado eq complete and get_gras eq complete
          task reo3_combine
            trigger get_sbuv eq complete and get_g2to eq complete
          task bufr_combine
            trigger get_nonbin eq complete and get_modw eq complete and get_avhn eq complete
          task jason2_combine
            trigger get_ers1_ra eq complete and get_j2nb eq complete and get_j2eo eq complete
          task ascat_combine
            trigger get_asc2 eq complete and get_asel eq complete and get_ascs eq complete
        endfamily
        task fetchmars
          trigger ./get eq complete
        family prepare_obs
          trigger ./get eq complete
          task preobs
            edit mode 'not1crad'
          task premwimg
          task preobs_wave
            edit mode 'wave'
          task pre1crad_prepare
            edit mode '1crad'
          task pre1crad_airs
            trigger pre1crad_prepare eq complete
          family pre1crad_iasi
            trigger pre1crad_prepare eq complete
            edit CPUTIME '1500'
            edit MAXTASK '16'
            task pre1crad_iasi_split_1
              edit TASKNO '1'
            task pre1crad_iasi_split_2
              edit TASKNO '2'
            task pre1crad_iasi_1
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '1'
            task pre1crad_iasi_2
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '2'
            task pre1crad_iasi_3
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '3'
            task pre1crad_iasi_4
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '4'
            task pre1crad_iasi_5
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '5'
            task pre1crad_iasi_6
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '6'
            task pre1crad_iasi_7
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '7'
            task pre1crad_iasi_8
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '8'
            task pre1crad_iasi_9
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '9'
            task pre1crad_iasi_10
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '10'
            task pre1crad_iasi_11
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '11'
            task pre1crad_iasi_12
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '12'
            task pre1crad_iasi_13
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '13'
            task pre1crad_iasi_14
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '14'
            task pre1crad_iasi_15
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '15'
            task pre1crad_iasi_16
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '16'
          endfamily
          task pre1crad_hirs
            trigger pre1crad_prepare eq complete
          task pre1crad_msu
            trigger pre1crad_prepare eq complete
          task pre1crad_ssu
            trigger pre1crad_prepare eq complete
          task pre1crad_amsua
            trigger pre1crad_prepare eq complete
          task pre1crad_amsub
            trigger pre1crad_prepare eq complete
          task pre1crad_mhs
            trigger pre1crad_prepare eq complete
          task pregeos
          task prereo3
          task prescat
        endfamily
        family reservation
          late -c +00:03
          trigger ./get/get_stor eq complete
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
          inlimit /e_36r1/limits:reservation
        endfamily
      endfamily
      family an
        trigger ./obs/fetchmars eq complete and ./obs/prepare_obs eq complete
        edit FSFAMILY 'an'
        edit SMSFILES '/home/ma/emos/def/e_36r1/an'
        edit ACCOUNT 'oeoperan'
        family makeodb
          edit NPES '16'
          task cleanodb
          family bufr2odb
            trigger cleanodb eq complete
            edit THREADS '1'
            edit NPOOLS '%ANNPES%'
            edit THINFAC '1'
            edit SENSOR '-1'
            edit REBALANCE '1'
            edit SUBTYPES_INCLUDE ''
            edit SUBTYPES_EXCLUDE ''
            task b2o_conv
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_EXCLUDE '85 82 83 87'
              edit USER_PRIORITY '90'
            task b2o_surf_conv
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_EXCLUDE '142 143 144 145 85 82 83 87'
            task b2o_airs
              late -s +00:05 -c +00:15
              edit THREADS '2'
              edit SENSOR '11'
              edit SUBTYPES_INCLUDE '57'
              edit NPOOLS '$((2* %ANNPES%))'
            task b2o_iasi
              late -s +00:05 -c +00:15
              edit SENSOR '993'
              edit SUBTYPES_INCLUDE '240'
              edit THREADS '4'
              edit NPOOLS '$((2* %ANNPES%))'
            task b2o_hirs
              late -s +00:05 -c +00:15
              edit SENSOR '0'
              edit SUBTYPES_INCLUDE '54 55'
              edit NPOOLS '$((2* %ANNPES%))'
            task b2o_msu
              late -s +00:05 -c +00:15
              edit SENSOR '1'
              edit SUBTYPES_INCLUDE '54 55'
            task b2o_ssu
              late -s +00:05 -c +00:15
              edit SENSOR '2'
              edit SUBTYPES_INCLUDE '54 55'
            task b2o_amsua
              late -s +00:05 -c +00:15
              edit SENSOR '3'
              edit SUBTYPES_INCLUDE '54 55'
            task b2o_amsub
              late -s +00:05 -c +00:15
              edit SENSOR '4'
              edit SUBTYPES_INCLUDE '54 55'
            task b2o_mhs
              late -s +00:05 -c +00:15
              edit SENSOR '11'
              edit SUBTYPES_INCLUDE '54 55'
            task b2o_satob
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '82 83 87'
            task b2o_reo3
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '205 206 207 208 215'
            task b2o_ssmi
              late -s +00:05 -c +00:15
              edit SENSOR '6'
              edit SUBTYPES_INCLUDE '55 127'
            task b2o_ssmis
              late -s +00:05 -c +00:15
              edit SENSOR '10'
              edit SUBTYPES_INCLUDE '49'
            task b2o_amsre
              late -s +00:05 -c +00:15
              edit SENSOR '17'
              edit SUBTYPES_INCLUDE '59'
            task b2o_tmi
              late -s +00:05 -c +00:15
              edit SENSOR '9'
              edit SUBTYPES_INCLUDE '129'
            task b2o_scatt
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '122 137 139'
            task b2o_geos
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '189 89 76 77 78'
            task b2o_gpsro
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '250'
            task b2o_meris
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '212'
          endfamily
          task mergeodb
            trigger bufr2odb eq complete
          task mergeodb_surf
            trigger ./bufr2odb/b2o_surf_conv eq complete
            edit NPES '16'
            edit THREADS '1'
        endfamily
        task restartodb
          trigger makeodb eq complete
        task preCleanFDB
        task vardata
          trigger ../../12dc/fc/model:step ge 6 or ../../12dc/fc/model eq complete
          event hires
        task fetcherr
          trigger ../../12dc/an eq complete
        task lowres
          trigger vardata:hires
          edit THREADS '4'
          edit ININPES '32'
        family surf_anal
          trigger makeodb eq complete and restartodb eq complete and preCleanFDB eq complete and vardata:hires
          edit SMSFILES '/home/ma/emos/def/e_36r1/an'
          edit FSFAMILY 'surf_anal'
          task t2ana
            edit mode 't2m'
          task rh2ana
            trigger t2ana eq complete
            edit mode 'rh2m'
          task snow
            trigger rh2ana eq complete
            edit mode 'snow'
          task sst
            edit THREADS '4'
          task slwet
            trigger sst eq complete and snow eq complete
        endfamily
        family 4dvar
          trigger makeodb eq complete and restartodb eq complete and vardata eq complete and fetcherr eq complete and lowres eq complete and preCleanFDB eq complete
          edit THREADS '8'
          edit VAR_DELAY '00'
          edit EMOS_SLEEP '00'
          family uptraj_0
            edit uptraj '0'
            task ifstraj
              late -s +00:02
              edit mode 'traj'
              edit USER_PRIORITY '91'
              edit SMSTRIES '1'
            task ifsmin
              late -s +00:02
              trigger ifstraj eq complete
              edit mode 'min'
              edit USER_PRIORITY '92'
            task ifstsave
              trigger ifstraj eq complete
              edit THREADS '1'
              edit USER_PRIORITY '93'
          endfamily
          family uptraj_1
            trigger uptraj_0 eq complete
            edit uptraj '1'
            task ifstraj
              late -s +00:02
              edit mode 'traj'
              edit USER_PRIORITY '93'
            task ifsmin
              late -s +00:02
              trigger ifstraj eq complete
              edit mode 'min'
              edit USER_PRIORITY '94'
            task ifstsave
              trigger ifstraj eq complete
              edit THREADS '1'
              edit USER_PRIORITY '95'
          endfamily
          family uptraj_2
            trigger uptraj_1 eq complete
            edit uptraj '2'
            task ifstraj
              late -s +00:02
              edit mode 'traj'
              edit USER_PRIORITY '95'
            task ifsmin
              late -s +00:02
              trigger ifstraj eq complete
              edit mode 'min'
              edit USER_PRIORITY '96'
            task ifstsave
              trigger ifstraj eq complete
              edit THREADS '1'
              edit USER_PRIORITY '97'
          endfamily
          task ifstraj
            late -s +00:02
            trigger uptraj_2 eq complete
            edit uptraj '3'
            edit mode 'traj'
            edit USER_PRIORITY '97'
            event finaltraj
            event finalwave
        endfamily
        task forceinv
          trigger ./4dvar/ifstraj:finaltraj and ./surf_anal eq complete
        family reservation1
          late -c +00:03
          trigger ./4dvar/uptraj_0/ifstraj eq active or ./4dvar/uptraj_0/ifstraj eq complete
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
          inlimit /e_36r1/limits:reservation
        endfamily
      endfamily
      family fc
        trigger an/forceinv eq complete and ( an/4dvar/ifstraj:finalwave or an/4dvar eq complete)
        edit ACCOUNT 'oeoperfc'
        edit FSFAMILY 'fc'
        task getfcdata
        task getpersSST
          trigger getfcdata eq complete
          edit SSTLEN '9'
          edit SSTINT '1'
          edit THREADS '2'
          edit NPES '16'
          edit MEM '1000'
          edit CPUTIME '960'
        task model
          late -s +00:02
          trigger ( getfcdata eq complete) and ( getpersSST eq complete)
          edit PERSST 'true'
          edit FCLENGTH '9'
          edit THREADS '8'
          edit USER_PRIORITY '94'
          meter step -1 9 9
        task ddh
          trigger model eq complete
        task clean
          trigger ../an eq complete and getfcdata eq complete and model eq complete and ddh eq complete
        family reservation1
          late -c +00:03
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
          inlimit /e_36r1/limits:reservation
        endfamily
      endfamily
    endfamily
    family 00bc
      edit EMOS_BASE '00'
      edit EDFAMILY 'false'
      edit INISTREAM 'DA'
      edit REINITIALIZE 'false'
      edit REINIEXPVER '0046'
      edit INIEXPVER '0045'
      edit REINIPERIOD_4D '6'
      edit REINISTREAM 'DA'
      edit REINIOFFSET '6'
      edit RUN_HOURLY_PP '0'
      edit PERIOD_4D '6'
      edit WINDOW_LENGTH_4D '6'
      edit WINDOW_OFFSET_4D '3'
      edit MXUP_TRAJ '3'
      edit PARALLEL '4'
      edit LANONLY 'false'
      edit STREAM 'SCDA'
      edit EMOS_STREAM 'SCDA'
      edit DELTA_DAY '1'
      edit ACCOUNT 'opebcpg'
      edit SMSURL 'GSCDA/00'
      edit USER_PRIORITY '91'
      family getreq
        trigger ( /e_36r1/main/00/an/4dvar eq active or /e_36r1/main/00/an/4dvar eq complete) and ( ./prod eq queued or ./prod eq complete)
        edit SMSFILES '/home/ma/emos/def/e_36r1/prod'
        edit EMOS_TYPE 'fc'
        edit PROD_TASK '1'
        edit EMOS_DOMAIN 'G'
        edit FSFAMILY 'ppdi'
        edit EMOS_PROJECT '12'
        edit SMSTRIES '1'
        edit EMOS_STREAM 'SCDA'
        task collectreq
          edit QUEUE 'diss'
          edit USER 'emos'
          edit WSHOST 'diss'
          edit QUEUE_EPILOG 'emos-exp'
          edit LOGDIR '/diss/log'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit SMSOUT '/vol/emos_nc/output'
          edit EMOS_BASE '00'
        task req
          late -s +00:03 -c +00:10
          trigger collectreq eq complete
          label info ""
          event 1 
        task collectbitmap
          trigger collectreq eq complete
          edit QUEUE 'diss'
          edit USER 'emos'
          edit WSHOST 'diss'
          edit QUEUE_EPILOG 'emos-exp'
          edit LOGDIR '/diss/log'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit SMSOUT '/vol/emos_nc/output'
        task bitmap
          late -s +00:03 -c +00:10
          trigger collectbitmap eq complete and ( req:001 or req eq complete)
          label info ""
          event 1 
      endfamily
      family prod
        trigger ( getreq/req:001 or getreq/req eq complete)
        edit SMSFILES '/home/ma/emos/def/e_36r1/prod'
        edit EMOS_DOMAIN 'G'
        edit EMOS_TYPE 'fc'
        edit FSFAMILY 'ppdi'
        edit PROD_TASK '1'
        edit EMOS_PROJECT '12'
        edit SMSURLBASE 'http://diss.ecmwf.int/do/monitoring/summary'
        edit EMOS_BASE '00'
        family an
          trigger /e_36r1/main/00/an/surf_anal eq complete and ( /e_36r1/main/00/an/4dvar/ifstraj:finalwave or /e_36r1/main/00/an/4dvar/ifstraj eq complete) and 0 eq 0
          edit EMOS_TYPE 'an'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ../090/prodgen eq complete and ../../prod eq queued and /e_36r1/main/00/prod eq complete
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 000
          trigger ( /e_36r1/main/00/fc/model:step ge 000 or /e_36r1/main/00/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0000'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/00/fc/model:step ge 12 or /e_36r1/main/00/fc/model eq complete) and ( ../an/backupprod eq complete or ../an/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 003
          trigger ( /e_36r1/main/00/fc/model:step ge 003 or /e_36r1/main/00/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0003'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/00/fc/model:step ge 15 or /e_36r1/main/00/fc/model eq complete) and ( ../000/backupprod eq complete or ../000/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 006
          trigger ( /e_36r1/main/00/fc/model:step ge 006 or /e_36r1/main/00/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0006'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/00/fc/model:step ge 18 or /e_36r1/main/00/fc/model eq complete) and ( ../003/backupprod eq complete or ../003/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 009
          trigger ( /e_36r1/main/00/fc/model:step ge 009 or /e_36r1/main/00/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0009'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/00/fc/model:step ge 21 or /e_36r1/main/00/fc/model eq complete) and ( ../006/backupprod eq complete or ../006/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 012
          trigger ( /e_36r1/main/00/fc/model:step ge 012 or /e_36r1/main/00/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0012'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/00/fc/model:step ge 24 or /e_36r1/main/00/fc/model eq complete) and ( ../009/backupprod eq complete or ../009/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 015
          trigger ( /e_36r1/main/00/fc/model:step ge 015 or /e_36r1/main/00/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0015'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/00/fc/model:step ge 27 or /e_36r1/main/00/fc/model eq complete) and ( ../012/backupprod eq complete or ../012/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 018
          trigger ( /e_36r1/main/00/fc/model:step ge 018 or /e_36r1/main/00/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0018'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/00/fc/model:step ge 30 or /e_36r1/main/00/fc/model eq complete) and ( ../015/backupprod eq complete or ../015/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 021
          trigger ( /e_36r1/main/00/fc/model:step ge 021 or /e_36r1/main/00/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0021'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/00/fc/model:step ge 33 or /e_36r1/main/00/fc/model eq complete) and ( ../018/backupprod eq complete or ../018/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 024
          trigger ( /e_36r1/main/00/fc/model:step ge 024 or /e_36r1/main/00/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0024'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/00/fc/model:step ge 36 or /e_36r1/main/00/fc/model eq complete) and ( ../021/backupprod eq complete or ../021/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 027
          trigger ( /e_36r1/main/00/fc/model:step ge 027 or /e_36r1/main/00/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0027'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/00/fc/model:step ge 39 or /e_36r1/main/00/fc/model eq complete) and ( ../024/backupprod eq complete or ../024/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 030
          trigger ( /e_36r1/main/00/fc/model:step ge 030 or /e_36r1/main/00/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0030'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/00/fc/model:step ge 42 or /e_36r1/main/00/fc/model eq complete) and ( ../027/backupprod eq complete or ../027/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 033
          trigger ( /e_36r1/main/00/fc/model:step ge 033 or /e_36r1/main/00/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0033'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/00/fc/model:step ge 45 or /e_36r1/main/00/fc/model eq complete) and ( ../030/backupprod eq complete or ../030/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 036
          trigger ( /e_36r1/main/00/fc/model:step ge 036 or /e_36r1/main/00/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0036'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/00/fc/model:step ge 48 or /e_36r1/main/00/fc/model eq complete) and ( ../033/backupprod eq complete or ../033/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 039
          trigger ( /e_36r1/main/00/fc/model:step ge 039 or /e_36r1/main/00/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0039'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/00/fc/model:step ge 51 or /e_36r1/main/00/fc/model eq complete) and ( ../036/backupprod eq complete or ../036/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 042
          trigger ( /e_36r1/main/00/fc/model:step ge 042 or /e_36r1/main/00/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0042'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/00/fc/model:step ge 54 or /e_36r1/main/00/fc/model eq complete) and ( ../039/backupprod eq complete or ../039/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 045
          trigger ( /e_36r1/main/00/fc/model:step ge 045 or /e_36r1/main/00/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0045'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/00/fc/model:step ge 57 or /e_36r1/main/00/fc/model eq complete) and ( ../042/backupprod eq complete or ../042/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 048
          trigger ( /e_36r1/main/00/fc/model:step ge 048 or /e_36r1/main/00/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0048'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/00/fc/model:step ge 60 or /e_36r1/main/00/fc/model eq complete) and ( ../045/backupprod eq complete or ../045/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 051
          trigger ( /e_36r1/main/00/fc/model:step ge 051 or /e_36r1/main/00/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0051'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/00/fc/model:step ge 63 or /e_36r1/main/00/fc/model eq complete) and ( ../048/backupprod eq complete or ../048/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 054
          trigger ( /e_36r1/main/00/fc/model:step ge 054 or /e_36r1/main/00/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0054'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/00/fc/model:step ge 66 or /e_36r1/main/00/fc/model eq complete) and ( ../051/backupprod eq complete or ../051/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 057
          trigger ( /e_36r1/main/00/fc/model:step ge 057 or /e_36r1/main/00/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0057'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/00/fc/model:step ge 69 or /e_36r1/main/00/fc/model eq complete) and ( ../054/backupprod eq complete or ../054/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 060
          trigger ( /e_36r1/main/00/fc/model:step ge 060 or /e_36r1/main/00/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0060'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/00/fc/model:step ge 72 or /e_36r1/main/00/fc/model eq complete) and ( ../057/backupprod eq complete or ../057/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 063
          trigger ( /e_36r1/main/00/fc/model:step ge 063 or /e_36r1/main/00/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0063'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/00/fc/model:step ge 75 or /e_36r1/main/00/fc/model eq complete) and ( ../060/backupprod eq complete or ../060/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 066
          trigger ( /e_36r1/main/00/fc/model:step ge 066 or /e_36r1/main/00/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0066'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/00/fc/model:step ge 78 or /e_36r1/main/00/fc/model eq complete) and ( ../063/backupprod eq complete or ../063/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 069
          trigger ( /e_36r1/main/00/fc/model:step ge 069 or /e_36r1/main/00/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0069'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/00/fc/model:step ge 81 or /e_36r1/main/00/fc/model eq complete) and ( ../066/backupprod eq complete or ../066/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 072
          trigger ( /e_36r1/main/00/fc/model:step ge 072 or /e_36r1/main/00/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0072'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/00/fc/model:step ge 84 or /e_36r1/main/00/fc/model eq complete) and ( ../069/backupprod eq complete or ../069/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 075
          trigger ( /e_36r1/main/00/fc/model:step ge 075 or /e_36r1/main/00/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0075'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/00/fc/model:step ge 87 or /e_36r1/main/00/fc/model eq complete) and ( ../072/backupprod eq complete or ../072/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 078
          trigger ( /e_36r1/main/00/fc/model:step ge 078 or /e_36r1/main/00/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0078'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/00/fc/model:step ge 90 or /e_36r1/main/00/fc/model eq complete) and ( ../075/backupprod eq complete or ../075/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
        endfamily
        family 081
          trigger ( /e_36r1/main/00/fc/model:step ge 081 or /e_36r1/main/00/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0081'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
        endfamily
        family 084
          trigger ( /e_36r1/main/00/fc/model:step ge 084 or /e_36r1/main/00/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0084'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
        endfamily
        family 087
          trigger ( /e_36r1/main/00/fc/model:step ge 087 or /e_36r1/main/00/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0087'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
        endfamily
        family 090
          trigger ( /e_36r1/main/00/fc/model:step ge 090 or /e_36r1/main/00/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0090'
          task prodgen
            late -s +00:05 -c +00:15
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc
        endfamily
      endfamily
      family prod2diss
        edit USER_PRIORITY '90'
        edit SMSURLBASE 'http://diss.ecmwf.int/do/monitoring/summary'
        edit SMSFILES '/home/ma/emos/def/e_36r1/prod'
        task operation_is_late
          event yes
        family oper
          edit EMOS_DOMAIN 'G'
          edit EMOS_TYPE 'fc'
          edit FSFAMILY 'ppdi'
          edit PROD_TASK '1'
          edit USER_PRIORITY '92'
          edit SMSURL 'GSCDA/00'
          edit EMOS_BASE '00'
          edit EMOS_PROJECT '12'
          family an
            task pdb
              trigger ../../../prod/an/prodgen eq complete
              edit EMOS_TYPE 'an'
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete
              edit EMOS_TYPE 'an'
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../090/q2diss eq complete and ../../../prod/an/backupprod eq complete
              edit EMOS_TYPE 'an'
              label info ""
              event 1 
          endfamily
          family 000
            edit EMOS_TIME_STEP_H '0000'
            task pdb
              trigger ../../../prod/000/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ../an/q2diss eq complete
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
          endfamily
          family 003
            edit EMOS_TIME_STEP_H '0003'
            task pdb
              trigger ../../../prod/003/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../000/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/003/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 006
            edit EMOS_TIME_STEP_H '0006'
            task pdb
              trigger ../../../prod/006/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../003/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/006/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 009
            edit EMOS_TIME_STEP_H '0009'
            task pdb
              trigger ../../../prod/009/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../006/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/009/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 012
            edit EMOS_TIME_STEP_H '0012'
            task pdb
              trigger ../../../prod/012/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../009/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/012/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 015
            edit EMOS_TIME_STEP_H '0015'
            task pdb
              trigger ../../../prod/015/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../012/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/015/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 018
            edit EMOS_TIME_STEP_H '0018'
            task pdb
              trigger ../../../prod/018/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../015/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/018/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 021
            edit EMOS_TIME_STEP_H '0021'
            task pdb
              trigger ../../../prod/021/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../018/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/021/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 024
            edit EMOS_TIME_STEP_H '0024'
            task pdb
              trigger ../../../prod/024/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../021/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/024/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 027
            edit EMOS_TIME_STEP_H '0027'
            task pdb
              trigger ../../../prod/027/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../024/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/027/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 030
            edit EMOS_TIME_STEP_H '0030'
            task pdb
              trigger ../../../prod/030/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../027/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/030/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 033
            edit EMOS_TIME_STEP_H '0033'
            task pdb
              trigger ../../../prod/033/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../030/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/033/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 036
            edit EMOS_TIME_STEP_H '0036'
            task pdb
              trigger ../../../prod/036/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../033/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/036/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 039
            edit EMOS_TIME_STEP_H '0039'
            task pdb
              trigger ../../../prod/039/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../036/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/039/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 042
            edit EMOS_TIME_STEP_H '0042'
            task pdb
              trigger ../../../prod/042/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../039/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/042/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 045
            edit EMOS_TIME_STEP_H '0045'
            task pdb
              trigger ../../../prod/045/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../042/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/045/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 048
            edit EMOS_TIME_STEP_H '0048'
            task pdb
              trigger ../../../prod/048/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../045/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/048/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 051
            edit EMOS_TIME_STEP_H '0051'
            task pdb
              trigger ../../../prod/051/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../048/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/051/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 054
            edit EMOS_TIME_STEP_H '0054'
            task pdb
              trigger ../../../prod/054/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../051/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/054/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 057
            edit EMOS_TIME_STEP_H '0057'
            task pdb
              trigger ../../../prod/057/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../054/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/057/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 060
            edit EMOS_TIME_STEP_H '0060'
            task pdb
              trigger ../../../prod/060/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../057/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/060/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 063
            edit EMOS_TIME_STEP_H '0063'
            task pdb
              trigger ../../../prod/063/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../060/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/063/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 066
            edit EMOS_TIME_STEP_H '0066'
            task pdb
              trigger ../../../prod/066/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../063/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/066/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 069
            edit EMOS_TIME_STEP_H '0069'
            task pdb
              trigger ../../../prod/069/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../066/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/069/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 072
            edit EMOS_TIME_STEP_H '0072'
            task pdb
              trigger ../../../prod/072/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../069/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/072/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 075
            edit EMOS_TIME_STEP_H '0075'
            task pdb
              trigger ../../../prod/075/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../072/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/075/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 078
            edit EMOS_TIME_STEP_H '0078'
            task pdb
              trigger ../../../prod/078/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../075/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/078/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 081
            edit EMOS_TIME_STEP_H '0081'
            task pdb
              trigger ../../../prod/081/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../078/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
          endfamily
          family 084
            edit EMOS_TIME_STEP_H '0084'
            task pdb
              trigger ../../../prod/084/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../081/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
          endfamily
          family 087
            edit EMOS_TIME_STEP_H '0087'
            task pdb
              trigger ../../../prod/087/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../084/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
          endfamily
          family 090
            edit EMOS_TIME_STEP_H '0090'
            task pdb
              trigger ../../../prod/090/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../087/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
          endfamily
          task check
            trigger ./078 eq complete
        endfamily
      endfamily
    endfamily
    family 00
      trigger make eq complete
      edit EMOS_BASE '00'
      edit MXUP_TRAJ '3'
      edit REINISTREAM 'DCDA'
      edit WINDOW_LENGTH_4D '6'
      edit WINDOW_OFFSET_4D '3'
      edit REINITIALIZE 'true'
      edit EDFAMILY 'true'
      edit PERIOD_4D '6'
      edit FCLENGTH '9'
      edit DELTA_DAY '1'
      edit ACCOUNT 'oeoperpg'
      edit SMSURL 'GOPER/00'
      edit USER_PRIORITY '90'
      family obs
        edit FSFAMILY 'an'
        edit SMSFILES '/home/ma/emos/def/e_36r1/obs'
        family get
          inlimit /e_36r1/limits:obs_get
          inlimit /limits:obs_get
          task get_nonbin
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex00/nonbin eq complete
            edit OBSFILE_OBSHOST 'BUFR'
            edit OBSFILE_HPCF 'bufr'
          task get_ers1_wa
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex00/ers1_wa eq complete
            edit OBSFILE_OBSHOST 'UWAX'
            edit OBSFILE_HPCF 'wavescat'
          task get_eas1
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex00/eas1 eq complete
            edit OBSFILE_OBSHOST 'EAS1'
            edit OBSFILE_HPCF 'wavesar'
          task get_ers1_wi
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex00/ers1_wi eq complete
            edit OBSFILE_OBSHOST 'UWIX'
            edit OBSFILE_HPCF 'scat'
          task get_ers1_ra
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex00/ers1_ra eq complete
            edit OBSFILE_OBSHOST 'URAX'
            edit OBSFILE_HPCF 'ralt'
          task get_rtovs
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex00/rtovs eq complete
            edit OBSFILE_OBSHOST 'TOVH'
            edit OBSFILE_HPCF 'tovh'
          task get_ssbt
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex00/ssbt eq complete
            edit OBSFILE_OBSHOST 'SSBT'
            edit OBSFILE_HPCF 'ssbt'
          task get_sst
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex00/sst/sst_fdb_c1b eq complete
            edit OBSFILE_OBSHOST 'nil'
            edit OBSFILE_HPCF 'nil'
          task get_atov
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex00/atov eq complete
            edit OBSFILE_OBSHOST 'ATOV'
            edit OBSFILE_HPCF 'atov'
          task get_grad
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex00/grad eq complete
            edit OBSFILE_OBSHOST 'GRAD'
            edit OBSFILE_HPCF 'grad'
          task get_stor
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex00/stor eq complete
            edit OBSFILE_OBSHOST 'STOR'
            edit OBSFILE_HPCF 'storm'
          task get_trmm
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex00/trmm eq complete
            edit OBSFILE_OBSHOST 'TRMM'
            edit OBSFILE_HPCF 'tmi'
          task get_qsca
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex00/qsca eq complete
            edit OBSFILE_OBSHOST 'QSCA'
            edit OBSFILE_HPCF 'qsca'
          task get_sbuv
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex00/sbuv eq complete
            edit OBSFILE_OBSHOST 'SBUV'
            edit OBSFILE_HPCF 'reo3'
          task get_airs
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex00/airs eq complete
            edit OBSFILE_OBSHOST 'AIRS'
            edit OBSFILE_HPCF 'airs'
          task get_ice
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex00/ice/ice_fdb_c1b eq complete
            edit OBSFILE_OBSHOST 'nil'
            edit OBSFILE_HPCF 'nil'
          task get_snow
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex00/snow eq complete
            edit OBSFILE_OBSHOST 'SNOW'
            edit OBSFILE_HPCF 'imssnow'
          task get_ateu
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex00/ateu eq complete
            edit OBSFILE_OBSHOST 'ATEU'
            edit OBSFILE_HPCF 'amsu_ears'
          task get_bice
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex00/bice eq complete
            edit OBSFILE_OBSHOST 'BICE'
            edit OBSFILE_HPCF 'baltic_ci'
          task get_amsa
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex00/amsa eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'AMSA'
            edit OBSFILE_HPCF 'amsa'
          task get_mhsa
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex00/mhsa eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'MHSA'
            edit OBSFILE_HPCF 'mhsa'
          task get_hirs
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex00/hirs eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'HIRS'
            edit OBSFILE_HPCF 'hirs'
          task get_amsr
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex00/amsr eq complete
            edit OBSFILE_OBSHOST 'AMSR'
            edit OBSFILE_HPCF 'amsre_aq'
          task get_rado
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex00/rado eq complete
            edit OBSFILE_OBSHOST 'RADO'
            edit OBSFILE_HPCF 'gpsro'
          task get_smin
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex00/smin eq complete
            edit OBSFILE_OBSHOST 'SMIN'
            edit OBSFILE_HPCF 'ssmisbt'
          task get_iasi
            late -s +00:02 -c +00:07
            trigger /pp2/main/extractions/ex00/iasi eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'IASI'
            edit OBSFILE_HPCF 'iasi'
          task get_gras
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex00/gras eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'GRAS'
            edit OBSFILE_HPCF 'gras'
          task get_modw
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex00/modw eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'MODW'
            edit OBSFILE_HPCF 'dbmodiswinds'
          task get_avhn
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex00/avhn eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'AVHN'
            edit OBSFILE_HPCF 'avhrrwinds'
          task get_atap
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex00/atap eq complete
            edit OBSFILE_OBSHOST 'ATAP'
            edit OBSFILE_HPCF 'amsu_pacrars'
          task get_g2to
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex00/g2to eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'G2TO'
            edit OBSFILE_HPCF 'reo3_gome2'
          task get_grst
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex00/grst eq complete
            edit OBSFILE_OBSHOST 'GRST'
            edit OBSFILE_HPCF 'ostia'
          task get_emer
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex00/emer eq complete
            edit OBSFILE_OBSHOST 'EMER'
            edit OBSFILE_HPCF 'envisat_meris'
          task get_j2nb
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex00/j2nb eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'J2NB'
            edit OBSFILE_HPCF 'j2nb'
          task get_j2eo
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex00/j2eo eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'J2EO'
            edit OBSFILE_HPCF 'j2eo'
          task get_asc2
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex00/asc2 eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'ASC2'
            edit OBSFILE_HPCF 'ascat_l2'
          task get_asel
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex00/asel eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'ASEL'
            edit OBSFILE_HPCF 'ascat_ears'
          task get_ascs
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex00/ascs eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'ASCS'
            edit OBSFILE_HPCF 'ascat_sm'
          task atovairs
            trigger get_atov eq complete and get_airs eq complete and get_ateu eq complete and get_amsa eq complete and get_mhsa eq complete and get_hirs eq complete and get_atap eq complete
          task snowlinks
            trigger get_nonbin eq complete and get_ers1_wi eq complete and get_rtovs eq complete and get_snow eq complete
          task ro_combine
            trigger get_rado eq complete and get_gras eq complete
          task reo3_combine
            trigger get_sbuv eq complete and get_g2to eq complete
          task bufr_combine
            trigger get_nonbin eq complete and get_modw eq complete and get_avhn eq complete
          task jason2_combine
            trigger get_ers1_ra eq complete and get_j2nb eq complete and get_j2eo eq complete
          task ascat_combine
            trigger get_asc2 eq complete and get_asel eq complete and get_ascs eq complete
        endfamily
        task fetchmars
          trigger ./get eq complete
        family prepare_obs
          trigger ./get eq complete
          task preobs
            trigger /e_36r1/main/12dc/an/update_psbias eq complete and /e_36r1/main/12dc/an/update_rstrhbias eq complete
            edit mode 'not1crad'
          task premwimg
          task preobs_wave
            edit mode 'wave'
          task pre1crad_prepare
            edit mode '1crad'
          task pre1crad_airs
            trigger pre1crad_prepare eq complete
          family pre1crad_iasi
            trigger pre1crad_prepare eq complete
            edit CPUTIME '1500'
            edit MAXTASK '16'
            task pre1crad_iasi_split_1
              edit TASKNO '1'
            task pre1crad_iasi_split_2
              edit TASKNO '2'
            task pre1crad_iasi_1
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '1'
            task pre1crad_iasi_2
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '2'
            task pre1crad_iasi_3
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '3'
            task pre1crad_iasi_4
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '4'
            task pre1crad_iasi_5
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '5'
            task pre1crad_iasi_6
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '6'
            task pre1crad_iasi_7
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '7'
            task pre1crad_iasi_8
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '8'
            task pre1crad_iasi_9
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '9'
            task pre1crad_iasi_10
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '10'
            task pre1crad_iasi_11
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '11'
            task pre1crad_iasi_12
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '12'
            task pre1crad_iasi_13
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '13'
            task pre1crad_iasi_14
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '14'
            task pre1crad_iasi_15
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '15'
            task pre1crad_iasi_16
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '16'
          endfamily
          task pre1crad_hirs
            trigger pre1crad_prepare eq complete
          task pre1crad_msu
            trigger pre1crad_prepare eq complete
          task pre1crad_ssu
            trigger pre1crad_prepare eq complete
          task pre1crad_amsua
            trigger pre1crad_prepare eq complete
          task pre1crad_amsub
            trigger pre1crad_prepare eq complete
          task pre1crad_mhs
            trigger pre1crad_prepare eq complete
          task pregeos
          task prereo3
          task prescat
        endfamily
        family reservation
          late -c +00:03
          trigger ./get/get_stor eq complete
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
          inlimit /e_36r1/limits:reservation
          task adv_res
            late -s +00:03 -c +00:02
            edit RESERVE_FOR '40'
            edit RESERVE_IN '60'
            edit RESERVED_NODES '24'
            edit RESERVATION_LABEL 'emos00XP'
            label info ""
          task adv_resA
            late -s +00:03 -c +00:02
            edit RESERVE_FOR '100'
            edit RESERVE_IN '11'
            edit RESERVED_NODES '48'
            edit RESERVATION_LABEL 'emos00A'
            label info ""
          task adv_resB
            late -s +00:03 -c +00:02
            edit RESERVE_FOR '30'
            edit RESERVE_IN '48'
            edit RESERVED_NODES '48'
            edit RESERVATION_LABEL 'emos00B'
            label info ""
          task adv_resC
            late -s +00:03 -c +00:02
            edit RESERVE_FOR '80'
            edit RESERVE_IN '60'
            edit RESERVED_NODES '120'
            edit RESERVATION_LABEL 'emos00C'
            label info ""
          task loadl_cycle_limit
            late -s +00:03 -c +00:02
            edit LOADL_CYCLE '65'
            label info ""
        endfamily
      endfamily
      family an
        trigger ./obs/fetchmars eq complete and ./obs/prepare_obs eq complete
        edit FSFAMILY 'an'
        edit SMSFILES '/home/ma/emos/def/e_36r1/an'
        edit PERIOD_4D '6'
        edit WINDOW_LENGTH_4D '6'
        edit WINDOW_OFFSET_4D '3'
        edit ACCOUNT 'oeoperan'
        family makeodb
          edit NPES '16'
          task cleanodb
          family bufr2odb
            trigger cleanodb eq complete
            edit THREADS '1'
            edit NPOOLS '%ANNPES%'
            edit THINFAC '1'
            edit SENSOR '-1'
            edit REBALANCE '1'
            edit SUBTYPES_INCLUDE ''
            edit SUBTYPES_EXCLUDE ''
            task b2o_conv
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_EXCLUDE '85 82 83 87'
              edit USER_PRIORITY '90'
            task b2o_surf_conv
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_EXCLUDE '142 143 144 145 85 82 83 87'
            task b2o_airs
              late -s +00:05 -c +00:15
              edit THREADS '2'
              edit SENSOR '11'
              edit SUBTYPES_INCLUDE '57'
              edit NPOOLS '$((2* %ANNPES%))'
            task b2o_iasi
              late -s +00:05 -c +00:15
              edit SENSOR '993'
              edit SUBTYPES_INCLUDE '240'
              edit THREADS '4'
              edit NPOOLS '$((2* %ANNPES%))'
            task b2o_hirs
              late -s +00:05 -c +00:15
              edit SENSOR '0'
              edit SUBTYPES_INCLUDE '54 55'
              edit NPOOLS '$((2* %ANNPES%))'
            task b2o_msu
              late -s +00:05 -c +00:15
              edit SENSOR '1'
              edit SUBTYPES_INCLUDE '54 55'
            task b2o_ssu
              late -s +00:05 -c +00:15
              edit SENSOR '2'
              edit SUBTYPES_INCLUDE '54 55'
            task b2o_amsua
              late -s +00:05 -c +00:15
              edit SENSOR '3'
              edit SUBTYPES_INCLUDE '54 55'
            task b2o_amsub
              late -s +00:05 -c +00:15
              edit SENSOR '4'
              edit SUBTYPES_INCLUDE '54 55'
            task b2o_mhs
              late -s +00:05 -c +00:15
              edit SENSOR '11'
              edit SUBTYPES_INCLUDE '54 55'
            task b2o_satob
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '82 83 87'
            task b2o_reo3
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '205 206 207 208 215'
            task b2o_ssmi
              late -s +00:05 -c +00:15
              edit SENSOR '6'
              edit SUBTYPES_INCLUDE '55 127'
            task b2o_ssmis
              late -s +00:05 -c +00:15
              edit SENSOR '10'
              edit SUBTYPES_INCLUDE '49'
            task b2o_amsre
              late -s +00:05 -c +00:15
              edit SENSOR '17'
              edit SUBTYPES_INCLUDE '59'
            task b2o_tmi
              late -s +00:05 -c +00:15
              edit SENSOR '9'
              edit SUBTYPES_INCLUDE '129'
            task b2o_scatt
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '122 137 139'
            task b2o_geos
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '189 89 76 77 78'
            task b2o_gpsro
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '250'
            task b2o_meris
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '212'
          endfamily
          task mergeodb
            trigger bufr2odb eq complete
          task mergeodb_surf
            trigger ./bufr2odb/b2o_surf_conv eq complete
            edit NPES '16'
            edit THREADS '1'
        endfamily
        task restartodb
          trigger makeodb eq complete
        task preCleanFDB
        task vardata
          trigger ( ../../12dc/fc/model:step ge 6 or ../../12dc/fc/model eq complete) and ../../18/an eq complete
          event hires
        task fetcherr
          trigger ../../18/an eq complete and ../../12dc/an eq complete
        task lowres
          trigger vardata:hires
          edit THREADS '4'
          edit ININPES '32'
        family surf_anal
          trigger makeodb eq complete and restartodb eq complete and preCleanFDB eq complete and vardata:hires
          edit SMSFILES '/home/ma/emos/def/e_36r1/an'
          edit FSFAMILY 'surf_anal'
          task t2ana
            edit mode 't2m'
          task rh2ana
            trigger t2ana eq complete
            edit mode 'rh2m'
          task snow
            trigger rh2ana eq complete
            edit mode 'snow'
          task sst
            edit THREADS '4'
          task slwet
            trigger sst eq complete and snow eq complete
        endfamily
        family 4dvar
          trigger makeodb eq complete and restartodb eq complete and vardata eq complete and fetcherr eq complete and lowres eq complete and preCleanFDB eq complete
          edit THREADS '8'
          edit VAR_DELAY '00'
          edit EMOS_SLEEP '00'
          family uptraj_0
            edit uptraj '0'
            task ifstraj
              late -s +00:02
              edit mode 'traj'
              edit USER_PRIORITY '91'
              edit SMSTRIES '1'
            task ifsmin
              late -s +00:02
              trigger ifstraj eq complete
              edit mode 'min'
              edit USER_PRIORITY '92'
            task ifstsave
              trigger ifstraj eq complete
              edit THREADS '1'
              edit USER_PRIORITY '93'
          endfamily
          family uptraj_1
            trigger uptraj_0 eq complete
            edit uptraj '1'
            task ifstraj
              late -s +00:02
              edit mode 'traj'
              edit USER_PRIORITY '93'
            task ifsmin
              late -s +00:02
              trigger ifstraj eq complete
              edit mode 'min'
              edit USER_PRIORITY '94'
            task ifstsave
              trigger ifstraj eq complete
              edit THREADS '1'
              edit USER_PRIORITY '95'
          endfamily
          family uptraj_2
            trigger uptraj_1 eq complete
            edit uptraj '2'
            task ifstraj
              late -s +00:02
              edit mode 'traj'
              edit USER_PRIORITY '95'
            task ifsmin
              late -s +00:02
              trigger ifstraj eq complete
              edit mode 'min'
              edit USER_PRIORITY '96'
            task ifstsave
              trigger ifstraj eq complete
              edit THREADS '1'
              edit USER_PRIORITY '97'
          endfamily
          task ifstraj
            late -s +00:02
            trigger uptraj_2 eq complete
            edit uptraj '3'
            edit mode 'traj'
            edit USER_PRIORITY '97'
            event finaltraj
            event finalwave
        endfamily
        task forceinv
          trigger ./4dvar/ifstraj:finaltraj and ./surf_anal eq complete
        family reservation1
          late -c +00:03
          trigger ./4dvar/uptraj_0/ifstraj eq active or ./4dvar/uptraj_0/ifstraj eq complete
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
          inlimit /e_36r1/limits:reservation
          task reserve_cpu
            late -s +00:03 -c +00:02
            edit RESERVED_CPUS '3072'
            edit RESERVE_FOR '100'
            label info ""
          task rm_res
            late -s +00:03 -c +00:02
            edit RESERVATION_LABEL 'emos00A'
            label info ""
          task ms_cpu
            late -s +00:03 -c +00:02
            edit MS_CPUS '$((64*30))'
            label info ""
        endfamily
        family reservation3
          late -c +00:03
          trigger ./4dvar/uptraj_1 eq complete
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
          inlimit /e_36r1/limits:reservation
          task set_class
            late -s +00:03 -c +00:02
            edit RESERVATION_DEFINITION 'default'
            label info ""
        endfamily
      endfamily
      family fc
        trigger an/forceinv eq complete and ( an/4dvar/ifstraj:finalwave or an/4dvar eq complete)
        edit ACCOUNT 'oeoperfc'
        edit FSFAMILY 'fc'
        task getfcdata
        task getpersSST
          trigger getfcdata eq complete
          edit SSTLEN '240'
          edit SSTINT '1'
          edit THREADS '2'
          edit NPES '16'
          edit MEM '1000'
          edit CPUTIME '960'
        task model
          late -s +00:02
          trigger ( getfcdata eq complete) and ( getpersSST eq complete)
          edit PERSST 'true'
          edit FCLENGTH '240'
          edit THREADS '8'
          edit USER_PRIORITY '94'
          meter step -1 240 240
        task ddh
          trigger model eq complete
        task clean
          trigger ../an eq complete and getfcdata eq complete and model eq complete and ddh eq complete and ../../12dc/fc eq complete
        family reservation1
          late -c +00:03
          trigger ./getfcdata eq active or ./getfcdata eq complete
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
          inlimit /e_36r1/limits:reservation
          family xp
            task set_xp
              late -s +00:03 -c +00:02
              edit RESERVATION_LABEL 'emos00XP'
              label info ""
            task rm_res
              late -s +00:03 -c +00:02
              trigger ./set_xp eq complete
              edit RESERVATION_LABEL 'emos00XP'
              label info ""
          endfamily
          task reserve_cpu
            late -s +00:03 -c +00:02
            edit RESERVED_CPUS '12160'
            edit RESERVE_FOR '2'
            label info ""
          task rm_res
            late -s +00:03 -c +00:02
            edit RESERVATION_LABEL 'emos00B'
            label info ""
        endfamily
      endfamily
      family getreq
        trigger ( /e_36r1/main/00/an/4dvar eq active or /e_36r1/main/00/an/4dvar eq complete) and ( ./prod eq queued or ./prod eq complete)
        edit SMSFILES '/home/ma/emos/def/e_36r1/prod'
        edit EMOS_TYPE 'fc'
        edit PROD_TASK '1'
        edit EMOS_DOMAIN 'G'
        edit FSFAMILY 'ppdi'
        edit EMOS_PROJECT '12'
        edit SMSTRIES '1'
        edit EMOS_STREAM 'OPER'
        task collectreq
          edit QUEUE 'diss'
          edit USER 'emos'
          edit WSHOST 'diss'
          edit QUEUE_EPILOG 'emos-exp'
          edit LOGDIR '/diss/log'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit SMSOUT '/vol/emos_nc/output'
          edit EMOS_BASE '00'
        task req
          late -s +00:03 -c +00:10
          trigger collectreq eq complete
          label info ""
          event 1 
        task collectbitmap
          trigger collectreq eq complete
          edit QUEUE 'diss'
          edit USER 'emos'
          edit WSHOST 'diss'
          edit QUEUE_EPILOG 'emos-exp'
          edit LOGDIR '/diss/log'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit SMSOUT '/vol/emos_nc/output'
        task bitmap
          late -s +00:03 -c +00:10
          trigger collectbitmap eq complete and ( req:001 or req eq complete)
          label info ""
          event 1 
        task collectwave
          trigger ./collectbitmap eq complete
          edit QUEUE 'diss'
          edit USER 'emos'
          edit WSHOST 'diss'
          edit QUEUE_EPILOG 'emos-exp'
          edit LOGDIR '/diss/log'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit SMSOUT '/vol/emos_nc/output'
          edit EMOS_BASE '00'
          edit EMOS_STREAM 'WAVE'
        task wavereq
          late -s +00:03 -c +00:10
          trigger ./collectwave eq complete
          edit EMOS_STREAM 'WAVE'
          label info ""
          event 1 
        task pgclima
          late -s +00:03 -c +00:10
          event 1 
      endfamily
      family prod
        trigger ( getreq/req:001 or getreq/req eq complete) and ( getreq/bitmap:001 or getreq/bitmap eq complete) and ( getreq/wavereq:001 or getreq/wavereq eq complete)
        edit SMSFILES '/home/ma/emos/def/e_36r1/prod'
        edit EMOS_DOMAIN 'G'
        edit EMOS_TYPE 'fc'
        edit FSFAMILY 'ppdi'
        edit PROD_TASK '1'
        edit EMOS_PROJECT '12'
        edit SMSURLBASE 'http://diss.ecmwf.int/do/monitoring/summary'
        edit EMOS_BASE '00'
        family 18an
          trigger /e_36r1/main/12dc/an/dcda_to_da eq complete
          edit EMOS_TYPE 'an'
          edit EMOS_BASE '18'
          edit DELTA_DAY '0'
          edit SMSURL 'GOPER/18'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            edit MEMBER 'off'
            edit STEPS '0'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 00an
          trigger /e_36r1/main/00/an/surf_anal eq complete and ( /e_36r1/main/00/an/4dvar/ifstraj:finalwave or /e_36r1/main/00/an/4dvar/ifstraj eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TYPE 'an'
          edit PGNPES '90'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ( ./derived_param:SMOS or ./derived_param eq complete) and ( ( /e_36r1/main/00/an/4dvar/ifstraj:finalwave or /e_36r1/main/00/an/4dvar/ifstraj eq complete))
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            edit MEMBER 'off'
            edit STEPS '0'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 000
          trigger ( /e_36r1/main/00/fc/model:step ge 000 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0000'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            edit MEMBER 'off'
            edit STEPS '000'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 003
          trigger ( /e_36r1/main/00/fc/model:step ge 003 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0003'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../000/derived_param:SMOS or ../000/derived_param eq complete) or ( ../000/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '003'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 006
          trigger ( /e_36r1/main/00/fc/model:step ge 006 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0006'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../003/derived_param:SMOS or ../003/derived_param eq complete) or ( ../003/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '006'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 009
          trigger ( /e_36r1/main/00/fc/model:step ge 009 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0009'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../006/derived_param:SMOS or ../006/derived_param eq complete) or ( ../006/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '009'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 012
          trigger ( /e_36r1/main/00/fc/model:step ge 012 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0012'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../009/derived_param:SMOS or ../009/derived_param eq complete) or ( ../009/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '012'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 015
          trigger ( /e_36r1/main/00/fc/model:step ge 015 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0015'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../012/derived_param:SMOS or ../012/derived_param eq complete) or ( ../012/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '015'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 018
          trigger ( /e_36r1/main/00/fc/model:step ge 018 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0018'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../015/derived_param:SMOS or ../015/derived_param eq complete) or ( ../015/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '018'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 021
          trigger ( /e_36r1/main/00/fc/model:step ge 021 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0021'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../018/derived_param:SMOS or ../018/derived_param eq complete) or ( ../018/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '021'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 024
          trigger ( /e_36r1/main/00/fc/model:step ge 024 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0024'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../021/derived_param:SMOS or ../021/derived_param eq complete) or ( ../021/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '024'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 027
          trigger ( /e_36r1/main/00/fc/model:step ge 027 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0027'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../024/derived_param:SMOS or ../024/derived_param eq complete) or ( ../024/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '027'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 030
          trigger ( /e_36r1/main/00/fc/model:step ge 030 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0030'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../027/derived_param:SMOS or ../027/derived_param eq complete) or ( ../027/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '030'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 033
          trigger ( /e_36r1/main/00/fc/model:step ge 033 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0033'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../030/derived_param:SMOS or ../030/derived_param eq complete) or ( ../030/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '033'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 036
          trigger ( /e_36r1/main/00/fc/model:step ge 036 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0036'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../033/derived_param:SMOS or ../033/derived_param eq complete) or ( ../033/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '036'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 039
          trigger ( /e_36r1/main/00/fc/model:step ge 039 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0039'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../036/derived_param:SMOS or ../036/derived_param eq complete) or ( ../036/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '039'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 042
          trigger ( /e_36r1/main/00/fc/model:step ge 042 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0042'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../039/derived_param:SMOS or ../039/derived_param eq complete) or ( ../039/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '042'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 045
          trigger ( /e_36r1/main/00/fc/model:step ge 045 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0045'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../042/derived_param:SMOS or ../042/derived_param eq complete) or ( ../042/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '045'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 048
          trigger ( /e_36r1/main/00/fc/model:step ge 048 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0048'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../045/derived_param:SMOS or ../045/derived_param eq complete) or ( ../045/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '048'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 051
          trigger ( /e_36r1/main/00/fc/model:step ge 051 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0051'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../048/derived_param:SMOS or ../048/derived_param eq complete) or ( ../048/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '051'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 054
          trigger ( /e_36r1/main/00/fc/model:step ge 054 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0054'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../051/derived_param:SMOS or ../051/derived_param eq complete) or ( ../051/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '054'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 057
          trigger ( /e_36r1/main/00/fc/model:step ge 057 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0057'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../054/derived_param:SMOS or ../054/derived_param eq complete) or ( ../054/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '057'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 060
          trigger ( /e_36r1/main/00/fc/model:step ge 060 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0060'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../057/derived_param:SMOS or ../057/derived_param eq complete) or ( ../057/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '060'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 063
          trigger ( /e_36r1/main/00/fc/model:step ge 063 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0063'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../060/derived_param:SMOS or ../060/derived_param eq complete) or ( ../060/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '063'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 066
          trigger ( /e_36r1/main/00/fc/model:step ge 066 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0066'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../063/derived_param:SMOS or ../063/derived_param eq complete) or ( ../063/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '066'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 069
          trigger ( /e_36r1/main/00/fc/model:step ge 069 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0069'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../066/derived_param:SMOS or ../066/derived_param eq complete) or ( ../066/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '069'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 072
          trigger ( /e_36r1/main/00/fc/model:step ge 072 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0072'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../069/derived_param:SMOS or ../069/derived_param eq complete) or ( ../069/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '072'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 075
          trigger ( /e_36r1/main/00/fc/model:step ge 075 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0075'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../072/derived_param:SMOS or ../072/derived_param eq complete) or ( ../072/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '075'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 078
          trigger ( /e_36r1/main/00/fc/model:step ge 078 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0078'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../075/derived_param:SMOS or ../075/derived_param eq complete) or ( ../075/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '078'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 081
          trigger ( /e_36r1/main/00/fc/model:step ge 081 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0081'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../078/derived_param:SMOS or ../078/derived_param eq complete) or ( ../078/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '081'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 084
          trigger ( /e_36r1/main/00/fc/model:step ge 084 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0084'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../081/derived_param:SMOS or ../081/derived_param eq complete) or ( ../081/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '084'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 087
          trigger ( /e_36r1/main/00/fc/model:step ge 087 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0087'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../084/derived_param:SMOS or ../084/derived_param eq complete) or ( ../084/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '087'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 090
          trigger ( /e_36r1/main/00/fc/model:step ge 090 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0090'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../087/derived_param:SMOS or ../087/derived_param eq complete) or ( ../087/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '090'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 093
          trigger ( /e_36r1/main/00/fc/model:step ge 093 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0093'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../090/derived_param:SMOS or ../090/derived_param eq complete) or ( ../090/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '093'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 096
          trigger ( /e_36r1/main/00/fc/model:step ge 096 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0096'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../093/derived_param:SMOS or ../093/derived_param eq complete) or ( ../093/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '096'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 099
          trigger ( /e_36r1/main/00/fc/model:step ge 099 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0099'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../096/derived_param:SMOS or ../096/derived_param eq complete) or ( ../096/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '099'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 102
          trigger ( /e_36r1/main/00/fc/model:step ge 102 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0102'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../099/derived_param:SMOS or ../099/derived_param eq complete) or ( ../099/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '102'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 105
          trigger ( /e_36r1/main/00/fc/model:step ge 105 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0105'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../102/derived_param:SMOS or ../102/derived_param eq complete) or ( ../102/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '105'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 108
          trigger ( /e_36r1/main/00/fc/model:step ge 108 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0108'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../105/derived_param:SMOS or ../105/derived_param eq complete) or ( ../105/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '108'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 111
          trigger ( /e_36r1/main/00/fc/model:step ge 111 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0111'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../108/derived_param:SMOS or ../108/derived_param eq complete) or ( ../108/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '111'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 114
          trigger ( /e_36r1/main/00/fc/model:step ge 114 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0114'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../111/derived_param:SMOS or ../111/derived_param eq complete) or ( ../111/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '114'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 117
          trigger ( /e_36r1/main/00/fc/model:step ge 117 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0117'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../114/derived_param:SMOS or ../114/derived_param eq complete) or ( ../114/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '117'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 120
          trigger ( /e_36r1/main/00/fc/model:step ge 120 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0120'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../117/derived_param:SMOS or ../117/derived_param eq complete) or ( ../117/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '120'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 123
          trigger ( /e_36r1/main/00/fc/model:step ge 123 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0123'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../120/derived_param:SMOS or ../120/derived_param eq complete) or ( ../120/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '123'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 126
          trigger ( /e_36r1/main/00/fc/model:step ge 126 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0126'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../123/derived_param:SMOS or ../123/derived_param eq complete) or ( ../123/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '126'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 129
          trigger ( /e_36r1/main/00/fc/model:step ge 129 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0129'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../126/derived_param:SMOS or ../126/derived_param eq complete) or ( ../126/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '129'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 132
          trigger ( /e_36r1/main/00/fc/model:step ge 132 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0132'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../129/derived_param:SMOS or ../129/derived_param eq complete) or ( ../129/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '132'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 135
          trigger ( /e_36r1/main/00/fc/model:step ge 135 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0135'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../132/derived_param:SMOS or ../132/derived_param eq complete) or ( ../132/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '135'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 138
          trigger ( /e_36r1/main/00/fc/model:step ge 138 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0138'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../135/derived_param:SMOS or ../135/derived_param eq complete) or ( ../135/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '138'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 141
          trigger ( /e_36r1/main/00/fc/model:step ge 141 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0141'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../138/derived_param:SMOS or ../138/derived_param eq complete) or ( ../138/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '141'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 144
          trigger ( /e_36r1/main/00/fc/model:step ge 144 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0144'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../141/derived_param:SMOS or ../141/derived_param eq complete) or ( ../141/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '144'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 150
          trigger ( /e_36r1/main/00/fc/model:step ge 150 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0150'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../144/derived_param:SMOS or ../144/derived_param eq complete) or ( ../144/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '150'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 156
          trigger ( /e_36r1/main/00/fc/model:step ge 156 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0156'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../150/derived_param:SMOS or ../150/derived_param eq complete) or ( ../150/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '156'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 162
          trigger ( /e_36r1/main/00/fc/model:step ge 162 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0162'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../156/derived_param:SMOS or ../156/derived_param eq complete) or ( ../156/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '162'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 168
          trigger ( /e_36r1/main/00/fc/model:step ge 168 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0168'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../162/derived_param:SMOS or ../162/derived_param eq complete) or ( ../162/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '168'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 174
          trigger ( /e_36r1/main/00/fc/model:step ge 174 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0174'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../168/derived_param:SMOS or ../168/derived_param eq complete) or ( ../168/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '174'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 180
          trigger ( /e_36r1/main/00/fc/model:step ge 180 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0180'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../174/derived_param:SMOS or ../174/derived_param eq complete) or ( ../174/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '180'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 186
          trigger ( /e_36r1/main/00/fc/model:step ge 186 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0186'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../180/derived_param:SMOS or ../180/derived_param eq complete) or ( ../180/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '186'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 192
          trigger ( /e_36r1/main/00/fc/model:step ge 192 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0192'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../186/derived_param:SMOS or ../186/derived_param eq complete) or ( ../186/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '192'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 198
          trigger ( /e_36r1/main/00/fc/model:step ge 198 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0198'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../192/derived_param:SMOS or ../192/derived_param eq complete) or ( ../192/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '198'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 204
          trigger ( /e_36r1/main/00/fc/model:step ge 204 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0204'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../198/derived_param:SMOS or ../198/derived_param eq complete) or ( ../198/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '204'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 210
          trigger ( /e_36r1/main/00/fc/model:step ge 210 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0210'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../204/derived_param:SMOS or ../204/derived_param eq complete) or ( ../204/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '210'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 216
          trigger ( /e_36r1/main/00/fc/model:step ge 216 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0216'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../210/derived_param:SMOS or ../210/derived_param eq complete) or ( ../210/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '216'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 222
          trigger ( /e_36r1/main/00/fc/model:step ge 222 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0222'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../216/derived_param:SMOS or ../216/derived_param eq complete) or ( ../216/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '222'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 228
          trigger ( /e_36r1/main/00/fc/model:step ge 228 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0228'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../222/derived_param:SMOS or ../222/derived_param eq complete) or ( ../222/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '228'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 234
          trigger ( /e_36r1/main/00/fc/model:step ge 234 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0234'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../228/derived_param:SMOS or ../228/derived_param eq complete) or ( ../228/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '234'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 240
          trigger ( /e_36r1/main/00/fc/model:step ge 240 or /e_36r1/main/00/fc/model eq complete) and ( ../getreq/pgclima:001 or ../getreq/pgclima eq complete)
          edit EMOS_TIME_STEP_H '0240'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '91'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task prod_wparam
            late -s +00:05 -c +00:15
            trigger ./derived_param eq complete
            edit PROD_TASK 'w'
            edit EMOS_TYPE 'wp'
            edit PGNPES '64'
            edit INIMEM '800'
            inlimit /e_36r1/limits:prod
          task prodwave
            late -s +00:05 -c +00:15
            edit EMOS_STREAM 'WV'
            edit PGNPES '2'
            inlimit /limits:prod
            inlimit /e_36r1/limits:prod
          task derived_param
            trigger ( ../234/derived_param:SMOS or ../234/derived_param eq complete) or ( ../234/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '240'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family reservation1
          late -c +00:03
          trigger /e_36r1/main/00bc/prod/078/prodgen eq active or /e_36r1/main/00bc/prod/078/prodgen eq complete
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
          inlimit /e_36r1/limits:reservation
          task set_xp
            late -s +00:03 -c +00:02
            edit RESERVATION_NODES '18'
            label info ""
        endfamily
        family reservation2
          late -c +00:03
          trigger /e_36r1/main/00/prod/240/prodgen eq active or /e_36r1/main/00/prod/240/prodgen eq complete
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
          inlimit /e_36r1/limits:reservation
          task set_xp
            late -s +00:03 -c +00:02
            edit RESERVATION_NODES '18'
            label info ""
        endfamily
        family reservation3
          late -c +00:03
          trigger /e_36r1/main/00/prod/240/prodgen eq active or /e_36r1/main/00/prod/240/prodgen eq complete
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
          inlimit /e_36r1/limits:reservation
          task reset_xp
            late -s +00:03 -c +00:02
            trigger /emc_36r1/main/00/legB/prod eq complete
            edit RESERVATION_NODES '0'
            label info ""
        endfamily
      endfamily
      family prod2diss
        edit USER_PRIORITY '90'
        edit SMSURLBASE 'http://diss.ecmwf.int/do/monitoring/summary'
        edit SMSFILES '/home/ma/emos/def/e_36r1/prod'
        task operation_is_late
          event yes
        family oper
          edit EMOS_DOMAIN 'G'
          edit EMOS_TYPE 'fc'
          edit FSFAMILY 'ppdi'
          edit PROD_TASK '1'
          edit USER_PRIORITY '92'
          edit EMOS_PROJECT '12'
          edit SMSURL 'GOPER/00'
          edit EMOS_BASE '00'
          family 18an
            edit EMOS_TYPE 'an'
            edit EMOS_BASE '18'
            edit SMSURL 'GOPER/18'
            edit DELTA_DAY '0'
            task pdb
              trigger ../../../prod/18an/prodgen eq complete
              edit EMOS_TYPE 'an'
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete
              edit EMOS_TYPE 'an'
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
          endfamily
          family 00an
            edit EMOS_TYPE 'an'
            task pdb
              trigger ../../../prod/00an/prodgen eq complete
              edit EMOS_TYPE 'an'
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete
              edit EMOS_TYPE 'an'
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
          endfamily
          family 000
            edit EMOS_TIME_STEP_H '0000'
            task pdb
              trigger ../../../prod/000/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../00an/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 003
            edit EMOS_TIME_STEP_H '0003'
            task pdb
              trigger ../../../prod/003/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../000/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 006
            edit EMOS_TIME_STEP_H '0006'
            task pdb
              trigger ../../../prod/006/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../003/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 009
            edit EMOS_TIME_STEP_H '0009'
            task pdb
              trigger ../../../prod/009/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../006/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 012
            edit EMOS_TIME_STEP_H '0012'
            task pdb
              trigger ../../../prod/012/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../009/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 015
            edit EMOS_TIME_STEP_H '0015'
            task pdb
              trigger ../../../prod/015/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../012/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 018
            edit EMOS_TIME_STEP_H '0018'
            task pdb
              trigger ../../../prod/018/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../015/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 021
            edit EMOS_TIME_STEP_H '0021'
            task pdb
              trigger ../../../prod/021/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../018/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 024
            edit EMOS_TIME_STEP_H '0024'
            task pdb
              trigger ../../../prod/024/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../021/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 027
            edit EMOS_TIME_STEP_H '0027'
            task pdb
              trigger ../../../prod/027/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../024/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 030
            edit EMOS_TIME_STEP_H '0030'
            task pdb
              trigger ../../../prod/030/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../027/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 033
            edit EMOS_TIME_STEP_H '0033'
            task pdb
              trigger ../../../prod/033/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../030/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 036
            edit EMOS_TIME_STEP_H '0036'
            task pdb
              trigger ../../../prod/036/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../033/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 039
            edit EMOS_TIME_STEP_H '0039'
            task pdb
              trigger ../../../prod/039/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../036/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 042
            edit EMOS_TIME_STEP_H '0042'
            task pdb
              trigger ../../../prod/042/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../039/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 045
            edit EMOS_TIME_STEP_H '0045'
            task pdb
              trigger ../../../prod/045/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../042/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 048
            edit EMOS_TIME_STEP_H '0048'
            task pdb
              trigger ../../../prod/048/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../045/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 051
            edit EMOS_TIME_STEP_H '0051'
            task pdb
              trigger ../../../prod/051/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../048/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 054
            edit EMOS_TIME_STEP_H '0054'
            task pdb
              trigger ../../../prod/054/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../051/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 057
            edit EMOS_TIME_STEP_H '0057'
            task pdb
              trigger ../../../prod/057/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../054/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 060
            edit EMOS_TIME_STEP_H '0060'
            task pdb
              trigger ../../../prod/060/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../057/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 063
            edit EMOS_TIME_STEP_H '0063'
            task pdb
              trigger ../../../prod/063/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../060/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 066
            edit EMOS_TIME_STEP_H '0066'
            task pdb
              trigger ../../../prod/066/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../063/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 069
            edit EMOS_TIME_STEP_H '0069'
            task pdb
              trigger ../../../prod/069/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../066/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 072
            edit EMOS_TIME_STEP_H '0072'
            task pdb
              trigger ../../../prod/072/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../069/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 075
            edit EMOS_TIME_STEP_H '0075'
            task pdb
              trigger ../../../prod/075/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../072/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 078
            edit EMOS_TIME_STEP_H '0078'
            task pdb
              trigger ../../../prod/078/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../075/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 081
            edit EMOS_TIME_STEP_H '0081'
            task pdb
              trigger ../../../prod/081/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../078/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 084
            edit EMOS_TIME_STEP_H '0084'
            task pdb
              trigger ../../../prod/084/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../081/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 087
            edit EMOS_TIME_STEP_H '0087'
            task pdb
              trigger ../../../prod/087/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../084/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 090
            edit EMOS_TIME_STEP_H '0090'
            task pdb
              trigger ../../../prod/090/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../087/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 093
            edit EMOS_TIME_STEP_H '0093'
            task pdb
              trigger ../../../prod/093/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../090/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 096
            edit EMOS_TIME_STEP_H '0096'
            task pdb
              trigger ../../../prod/096/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../093/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 099
            edit EMOS_TIME_STEP_H '0099'
            task pdb
              trigger ../../../prod/099/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../096/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 102
            edit EMOS_TIME_STEP_H '0102'
            task pdb
              trigger ../../../prod/102/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../099/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 105
            edit EMOS_TIME_STEP_H '0105'
            task pdb
              trigger ../../../prod/105/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../102/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 108
            edit EMOS_TIME_STEP_H '0108'
            task pdb
              trigger ../../../prod/108/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../105/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 111
            edit EMOS_TIME_STEP_H '0111'
            task pdb
              trigger ../../../prod/111/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../108/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 114
            edit EMOS_TIME_STEP_H '0114'
            task pdb
              trigger ../../../prod/114/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../111/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 117
            edit EMOS_TIME_STEP_H '0117'
            task pdb
              trigger ../../../prod/117/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../114/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 120
            edit EMOS_TIME_STEP_H '0120'
            task pdb
              trigger ../../../prod/120/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../117/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 123
            edit EMOS_TIME_STEP_H '0123'
            task pdb
              trigger ../../../prod/123/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../120/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 126
            edit EMOS_TIME_STEP_H '0126'
            task pdb
              trigger ../../../prod/126/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../123/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 129
            edit EMOS_TIME_STEP_H '0129'
            task pdb
              trigger ../../../prod/129/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../126/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 132
            edit EMOS_TIME_STEP_H '0132'
            task pdb
              trigger ../../../prod/132/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../129/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 135
            edit EMOS_TIME_STEP_H '0135'
            task pdb
              trigger ../../../prod/135/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../132/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 138
            edit EMOS_TIME_STEP_H '0138'
            task pdb
              trigger ../../../prod/138/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../135/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 141
            edit EMOS_TIME_STEP_H '0141'
            task pdb
              trigger ../../../prod/141/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../138/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 144
            edit EMOS_TIME_STEP_H '0144'
            task pdb
              trigger ../../../prod/144/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../141/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 150
            edit EMOS_TIME_STEP_H '0150'
            task pdb
              trigger ../../../prod/150/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../144/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 156
            edit EMOS_TIME_STEP_H '0156'
            task pdb
              trigger ../../../prod/156/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../150/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 162
            edit EMOS_TIME_STEP_H '0162'
            task pdb
              trigger ../../../prod/162/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../156/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 168
            edit EMOS_TIME_STEP_H '0168'
            task pdb
              trigger ../../../prod/168/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../162/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 174
            edit EMOS_TIME_STEP_H '0174'
            task pdb
              trigger ../../../prod/174/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../168/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 180
            edit EMOS_TIME_STEP_H '0180'
            task pdb
              trigger ../../../prod/180/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../174/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 186
            edit EMOS_TIME_STEP_H '0186'
            task pdb
              trigger ../../../prod/186/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../180/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 192
            edit EMOS_TIME_STEP_H '0192'
            task pdb
              trigger ../../../prod/192/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../186/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 198
            edit EMOS_TIME_STEP_H '0198'
            task pdb
              trigger ../../../prod/198/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../192/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 204
            edit EMOS_TIME_STEP_H '0204'
            task pdb
              trigger ../../../prod/204/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../198/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 210
            edit EMOS_TIME_STEP_H '0210'
            task pdb
              trigger ../../../prod/210/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../204/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 216
            edit EMOS_TIME_STEP_H '0216'
            task pdb
              trigger ../../../prod/216/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../210/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 222
            edit EMOS_TIME_STEP_H '0222'
            task pdb
              trigger ../../../prod/222/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../216/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 228
            edit EMOS_TIME_STEP_H '0228'
            task pdb
              trigger ../../../prod/228/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../222/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 234
            edit EMOS_TIME_STEP_H '0234'
            task pdb
              trigger ../../../prod/234/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../228/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family 240
            edit EMOS_TIME_STEP_H '0240'
            task pdb
              trigger ../../../prod/240/prodgen eq complete
              inlimit /e_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../234/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
            task diss_wparam
              late -s +00:05 -c +00:15
              trigger ../../../prod/240/prod_wparam eq complete
              edit PROD_TASK 'w'
              edit EMOS_TYPE 'wp'
              label info ""
              event pdb
          endfamily
          task check
            trigger ./240 eq complete
        endfamily
        family wave
          edit EMOS_DOMAIN 'G'
          edit EMOS_TYPE 'fc'
          edit FSFAMILY 'ppdi'
          edit PROD_TASK '1'
          edit EMOS_PROJECT '12'
          edit EMOS_STREAM 'WV'
          edit SMSURL 'GWAVE/00'
          inlimit /e_36r1/limits:w2diss
          inlimit /limits:q2diss
          family 18an
            edit EMOS_TYPE 'an'
            edit EMOS_BASE '18'
            edit SMSURL 'GWAVE/18'
            edit DELTA_DAY '0'
            task diss
              late -s +00:05 -c +00:15
              trigger ../../../prod/18an/prodwave eq complete
              edit EMOS_TYPE 'an'
              label info ""
              event pdb
          endfamily
          family 00an
            edit EMOS_TYPE 'an'
            task diss
              late -s +00:05 -c +00:15
              trigger ../../../prod/00an/prodwave eq complete
              edit EMOS_TYPE 'an'
              label info ""
              event pdb
          endfamily
          family 000
            trigger ../../prod/000/prodwave eq complete and ./18an eq complete and ./00an eq complete
            edit EMOS_TIME_STEP_H '0000'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 003
            trigger ../../prod/003/prodwave eq complete and ( ./000/diss:pdb or ./000 eq complete)
            edit EMOS_TIME_STEP_H '0003'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 006
            trigger ../../prod/006/prodwave eq complete and ( ./003/diss:pdb or ./003 eq complete)
            edit EMOS_TIME_STEP_H '0006'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 009
            trigger ../../prod/009/prodwave eq complete and ( ./006/diss:pdb or ./006 eq complete)
            edit EMOS_TIME_STEP_H '0009'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 012
            trigger ../../prod/012/prodwave eq complete and ( ./009/diss:pdb or ./009 eq complete)
            edit EMOS_TIME_STEP_H '0012'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 015
            trigger ../../prod/015/prodwave eq complete and ( ./012/diss:pdb or ./012 eq complete)
            edit EMOS_TIME_STEP_H '0015'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 018
            trigger ../../prod/018/prodwave eq complete and ( ./015/diss:pdb or ./015 eq complete)
            edit EMOS_TIME_STEP_H '0018'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 021
            trigger ../../prod/021/prodwave eq complete and ( ./018/diss:pdb or ./018 eq complete)
            edit EMOS_TIME_STEP_H '0021'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 024
            trigger ../../prod/024/prodwave eq complete and ( ./021/diss:pdb or ./021 eq complete)
            edit EMOS_TIME_STEP_H '0024'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 027
            trigger ../../prod/027/prodwave eq complete and ( ./024/diss:pdb or ./024 eq complete)
            edit EMOS_TIME_STEP_H '0027'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 030
            trigger ../../prod/030/prodwave eq complete and ( ./027/diss:pdb or ./027 eq complete)
            edit EMOS_TIME_STEP_H '0030'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 033
            trigger ../../prod/033/prodwave eq complete and ( ./030/diss:pdb or ./030 eq complete)
            edit EMOS_TIME_STEP_H '0033'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 036
            trigger ../../prod/036/prodwave eq complete and ( ./033/diss:pdb or ./033 eq complete)
            edit EMOS_TIME_STEP_H '0036'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 039
            trigger ../../prod/039/prodwave eq complete and ( ./036/diss:pdb or ./036 eq complete)
            edit EMOS_TIME_STEP_H '0039'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 042
            trigger ../../prod/042/prodwave eq complete and ( ./039/diss:pdb or ./039 eq complete)
            edit EMOS_TIME_STEP_H '0042'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 045
            trigger ../../prod/045/prodwave eq complete and ( ./042/diss:pdb or ./042 eq complete)
            edit EMOS_TIME_STEP_H '0045'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 048
            trigger ../../prod/048/prodwave eq complete and ( ./045/diss:pdb or ./045 eq complete)
            edit EMOS_TIME_STEP_H '0048'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 051
            trigger ../../prod/051/prodwave eq complete and ( ./048/diss:pdb or ./048 eq complete)
            edit EMOS_TIME_STEP_H '0051'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 054
            trigger ../../prod/054/prodwave eq complete and ( ./051/diss:pdb or ./051 eq complete)
            edit EMOS_TIME_STEP_H '0054'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 057
            trigger ../../prod/057/prodwave eq complete and ( ./054/diss:pdb or ./054 eq complete)
            edit EMOS_TIME_STEP_H '0057'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 060
            trigger ../../prod/060/prodwave eq complete and ( ./057/diss:pdb or ./057 eq complete)
            edit EMOS_TIME_STEP_H '0060'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 063
            trigger ../../prod/063/prodwave eq complete and ( ./060/diss:pdb or ./060 eq complete)
            edit EMOS_TIME_STEP_H '0063'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 066
            trigger ../../prod/066/prodwave eq complete and ( ./063/diss:pdb or ./063 eq complete)
            edit EMOS_TIME_STEP_H '0066'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 069
            trigger ../../prod/069/prodwave eq complete and ( ./066/diss:pdb or ./066 eq complete)
            edit EMOS_TIME_STEP_H '0069'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 072
            trigger ../../prod/072/prodwave eq complete and ( ./069/diss:pdb or ./069 eq complete)
            edit EMOS_TIME_STEP_H '0072'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 075
            trigger ../../prod/075/prodwave eq complete and ( ./072/diss:pdb or ./072 eq complete)
            edit EMOS_TIME_STEP_H '0075'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 078
            trigger ../../prod/078/prodwave eq complete and ( ./075/diss:pdb or ./075 eq complete)
            edit EMOS_TIME_STEP_H '0078'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 081
            trigger ../../prod/081/prodwave eq complete and ( ./078/diss:pdb or ./078 eq complete)
            edit EMOS_TIME_STEP_H '0081'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 084
            trigger ../../prod/084/prodwave eq complete and ( ./081/diss:pdb or ./081 eq complete)
            edit EMOS_TIME_STEP_H '0084'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 087
            trigger ../../prod/087/prodwave eq complete and ( ./084/diss:pdb or ./084 eq complete)
            edit EMOS_TIME_STEP_H '0087'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 090
            trigger ../../prod/090/prodwave eq complete and ( ./087/diss:pdb or ./087 eq complete)
            edit EMOS_TIME_STEP_H '0090'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 093
            trigger ../../prod/093/prodwave eq complete and ( ./090/diss:pdb or ./090 eq complete)
            edit EMOS_TIME_STEP_H '0093'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 096
            trigger ../../prod/096/prodwave eq complete and ( ./093/diss:pdb or ./093 eq complete)
            edit EMOS_TIME_STEP_H '0096'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 099
            trigger ../../prod/099/prodwave eq complete and ( ./096/diss:pdb or ./096 eq complete)
            edit EMOS_TIME_STEP_H '0099'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 102
            trigger ../../prod/102/prodwave eq complete and ( ./099/diss:pdb or ./099 eq complete)
            edit EMOS_TIME_STEP_H '0102'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 105
            trigger ../../prod/105/prodwave eq complete and ( ./102/diss:pdb or ./102 eq complete)
            edit EMOS_TIME_STEP_H '0105'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 108
            trigger ../../prod/108/prodwave eq complete and ( ./105/diss:pdb or ./105 eq complete)
            edit EMOS_TIME_STEP_H '0108'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 111
            trigger ../../prod/111/prodwave eq complete and ( ./108/diss:pdb or ./108 eq complete)
            edit EMOS_TIME_STEP_H '0111'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 114
            trigger ../../prod/114/prodwave eq complete and ( ./111/diss:pdb or ./111 eq complete)
            edit EMOS_TIME_STEP_H '0114'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 117
            trigger ../../prod/117/prodwave eq complete and ( ./114/diss:pdb or ./114 eq complete)
            edit EMOS_TIME_STEP_H '0117'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 120
            trigger ../../prod/120/prodwave eq complete and ( ./117/diss:pdb or ./117 eq complete)
            edit EMOS_TIME_STEP_H '0120'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 123
            trigger ../../prod/123/prodwave eq complete and ( ./120/diss:pdb or ./120 eq complete)
            edit EMOS_TIME_STEP_H '0123'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 126
            trigger ../../prod/126/prodwave eq complete and ( ./123/diss:pdb or ./123 eq complete)
            edit EMOS_TIME_STEP_H '0126'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 129
            trigger ../../prod/129/prodwave eq complete and ( ./126/diss:pdb or ./126 eq complete)
            edit EMOS_TIME_STEP_H '0129'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 132
            trigger ../../prod/132/prodwave eq complete and ( ./129/diss:pdb or ./129 eq complete)
            edit EMOS_TIME_STEP_H '0132'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 135
            trigger ../../prod/135/prodwave eq complete and ( ./132/diss:pdb or ./132 eq complete)
            edit EMOS_TIME_STEP_H '0135'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 138
            trigger ../../prod/138/prodwave eq complete and ( ./135/diss:pdb or ./135 eq complete)
            edit EMOS_TIME_STEP_H '0138'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 141
            trigger ../../prod/141/prodwave eq complete and ( ./138/diss:pdb or ./138 eq complete)
            edit EMOS_TIME_STEP_H '0141'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 144
            trigger ../../prod/144/prodwave eq complete and ( ./141/diss:pdb or ./141 eq complete)
            edit EMOS_TIME_STEP_H '0144'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 150
            trigger ../../prod/150/prodwave eq complete and ( ./144/diss:pdb or ./144 eq complete)
            edit EMOS_TIME_STEP_H '0150'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 156
            trigger ../../prod/156/prodwave eq complete and ( ./150/diss:pdb or ./150 eq complete)
            edit EMOS_TIME_STEP_H '0156'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 162
            trigger ../../prod/162/prodwave eq complete and ( ./156/diss:pdb or ./156 eq complete)
            edit EMOS_TIME_STEP_H '0162'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 168
            trigger ../../prod/168/prodwave eq complete and ( ./162/diss:pdb or ./162 eq complete)
            edit EMOS_TIME_STEP_H '0168'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 174
            trigger ../../prod/174/prodwave eq complete and ( ./168/diss:pdb or ./168 eq complete)
            edit EMOS_TIME_STEP_H '0174'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 180
            trigger ../../prod/180/prodwave eq complete and ( ./174/diss:pdb or ./174 eq complete)
            edit EMOS_TIME_STEP_H '0180'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 186
            trigger ../../prod/186/prodwave eq complete and ( ./180/diss:pdb or ./180 eq complete)
            edit EMOS_TIME_STEP_H '0186'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 192
            trigger ../../prod/192/prodwave eq complete and ( ./186/diss:pdb or ./186 eq complete)
            edit EMOS_TIME_STEP_H '0192'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 198
            trigger ../../prod/198/prodwave eq complete and ( ./192/diss:pdb or ./192 eq complete)
            edit EMOS_TIME_STEP_H '0198'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 204
            trigger ../../prod/204/prodwave eq complete and ( ./198/diss:pdb or ./198 eq complete)
            edit EMOS_TIME_STEP_H '0204'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 210
            trigger ../../prod/210/prodwave eq complete and ( ./204/diss:pdb or ./204 eq complete)
            edit EMOS_TIME_STEP_H '0210'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 216
            trigger ../../prod/216/prodwave eq complete and ( ./210/diss:pdb or ./210 eq complete)
            edit EMOS_TIME_STEP_H '0216'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 222
            trigger ../../prod/222/prodwave eq complete and ( ./216/diss:pdb or ./216 eq complete)
            edit EMOS_TIME_STEP_H '0222'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 228
            trigger ../../prod/228/prodwave eq complete and ( ./222/diss:pdb or ./222 eq complete)
            edit EMOS_TIME_STEP_H '0228'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 234
            trigger ../../prod/234/prodwave eq complete and ( ./228/diss:pdb or ./228 eq complete)
            edit EMOS_TIME_STEP_H '0234'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 240
            trigger ../../prod/240/prodwave eq complete and ( ./234/diss:pdb or ./234 eq complete)
            edit EMOS_TIME_STEP_H '0240'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          task check
            trigger ./240 eq complete
        endfamily
      endfamily
    endfamily
    family 06bc
      trigger ( ( ./make eq complete) and ( /e_36r1/main/00/fc eq complete)) and ( /e_36r1/main/18bc/fc eq complete)
      edit EMOS_BASE '06'
      edit EDFAMILY 'false'
      edit INISTREAM 'DA'
      edit REINITIALIZE 'false'
      edit REINIEXPVER '0046'
      edit INIEXPVER '0045'
      edit REINIPERIOD_4D '6'
      edit REINISTREAM 'DA'
      edit REINIOFFSET '6'
      edit RUN_HOURLY_PP '0'
      edit PERIOD_4D '6'
      edit WINDOW_LENGTH_4D '6'
      edit WINDOW_OFFSET_4D '3'
      edit MXUP_TRAJ '3'
      edit PARALLEL '4'
      edit LANONLY 'false'
      edit STREAM 'SCDA'
      edit EMOS_STREAM 'SCDA'
      edit DELTA_DAY '1'
      edit ACCOUNT 'opebcpg'
      edit SMSURL 'GSCDA/06'
      edit USER_PRIORITY '91'
      family obs
        edit FSFAMILY 'an'
        edit SMSFILES '/home/ma/emos/def/e_36r1/obs'
        family get
          inlimit /e_36r1/limits:obs_get
          inlimit /limits:obs_get
          task get_nonbin
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/nonbin eq complete
            edit OBSFILE_OBSHOST 'BUFR'
            edit OBSFILE_HPCF 'bufr'
          task get_ers1_wa
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/ers1_wa eq complete
            edit OBSFILE_OBSHOST 'UWAX'
            edit OBSFILE_HPCF 'wavescat'
          task get_eas1
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/eas1 eq complete
            edit OBSFILE_OBSHOST 'EAS1'
            edit OBSFILE_HPCF 'wavesar'
          task get_ers1_wi
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/ers1_wi eq complete
            edit OBSFILE_OBSHOST 'UWIX'
            edit OBSFILE_HPCF 'scat'
          task get_ers1_ra
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/ers1_ra eq complete
            edit OBSFILE_OBSHOST 'URAX'
            edit OBSFILE_HPCF 'ralt'
          task get_rtovs
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/rtovs eq complete
            edit OBSFILE_OBSHOST 'TOVH'
            edit OBSFILE_HPCF 'tovh'
          task get_ssbt
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/ssbt eq complete
            edit OBSFILE_OBSHOST 'SSBT'
            edit OBSFILE_HPCF 'ssbt'
          task get_sst
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/sst/sst_fdb_c1b eq complete
            edit OBSFILE_OBSHOST 'nil'
            edit OBSFILE_HPCF 'nil'
          task get_atov
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/atov eq complete
            edit OBSFILE_OBSHOST 'ATOV'
            edit OBSFILE_HPCF 'atov'
          task get_grad
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/grad eq complete
            edit OBSFILE_OBSHOST 'GRAD'
            edit OBSFILE_HPCF 'grad'
          task get_stor
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/stor eq complete
            edit OBSFILE_OBSHOST 'STOR'
            edit OBSFILE_HPCF 'storm'
          task get_trmm
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/trmm eq complete
            edit OBSFILE_OBSHOST 'TRMM'
            edit OBSFILE_HPCF 'tmi'
          task get_qsca
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/qsca eq complete
            edit OBSFILE_OBSHOST 'QSCA'
            edit OBSFILE_HPCF 'qsca'
          task get_sbuv
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/sbuv eq complete
            edit OBSFILE_OBSHOST 'SBUV'
            edit OBSFILE_HPCF 'reo3'
          task get_airs
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/airs eq complete
            edit OBSFILE_OBSHOST 'AIRS'
            edit OBSFILE_HPCF 'airs'
          task get_ice
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/ice/ice_fdb_c1b eq complete
            edit OBSFILE_OBSHOST 'nil'
            edit OBSFILE_HPCF 'nil'
          task get_snow
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/snow eq complete
            edit OBSFILE_OBSHOST 'SNOW'
            edit OBSFILE_HPCF 'imssnow'
          task get_ateu
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/ateu eq complete
            edit OBSFILE_OBSHOST 'ATEU'
            edit OBSFILE_HPCF 'amsu_ears'
          task get_bice
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/bice eq complete
            edit OBSFILE_OBSHOST 'BICE'
            edit OBSFILE_HPCF 'baltic_ci'
          task get_amsa
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex06/amsa eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'AMSA'
            edit OBSFILE_HPCF 'amsa'
          task get_mhsa
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex06/mhsa eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'MHSA'
            edit OBSFILE_HPCF 'mhsa'
          task get_hirs
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex06/hirs eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'HIRS'
            edit OBSFILE_HPCF 'hirs'
          task get_amsr
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/amsr eq complete
            edit OBSFILE_OBSHOST 'AMSR'
            edit OBSFILE_HPCF 'amsre_aq'
          task get_rado
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/rado eq complete
            edit OBSFILE_OBSHOST 'RADO'
            edit OBSFILE_HPCF 'gpsro'
          task get_smin
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/smin eq complete
            edit OBSFILE_OBSHOST 'SMIN'
            edit OBSFILE_HPCF 'ssmisbt'
          task get_iasi
            late -s +00:02 -c +00:07
            trigger /pp2/main/extractions/ex06/iasi eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'IASI'
            edit OBSFILE_HPCF 'iasi'
          task get_gras
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex06/gras eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'GRAS'
            edit OBSFILE_HPCF 'gras'
          task get_modw
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex06/modw eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'MODW'
            edit OBSFILE_HPCF 'dbmodiswinds'
          task get_avhn
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex06/avhn eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'AVHN'
            edit OBSFILE_HPCF 'avhrrwinds'
          task get_atap
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/atap eq complete
            edit OBSFILE_OBSHOST 'ATAP'
            edit OBSFILE_HPCF 'amsu_pacrars'
          task get_g2to
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex06/g2to eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'G2TO'
            edit OBSFILE_HPCF 'reo3_gome2'
          task get_grst
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/grst eq complete
            edit OBSFILE_OBSHOST 'GRST'
            edit OBSFILE_HPCF 'ostia'
          task get_emer
            late -s +00:02 -c +00:04
            trigger /pp1/main/extractions/ex06/emer eq complete
            edit OBSFILE_OBSHOST 'EMER'
            edit OBSFILE_HPCF 'envisat_meris'
          task get_j2nb
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex06/j2nb eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'J2NB'
            edit OBSFILE_HPCF 'j2nb'
          task get_j2eo
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex06/j2eo eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'J2EO'
            edit OBSFILE_HPCF 'j2eo'
          task get_asc2
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex06/asc2 eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'ASC2'
            edit OBSFILE_HPCF 'ascat_l2'
          task get_asel
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex06/asel eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'ASEL'
            edit OBSFILE_HPCF 'ascat_ears'
          task get_ascs
            late -s +00:02 -c +00:04
            trigger /pp2/main/extractions/ex06/ascs eq complete
            edit OBSHOST 'pp2'
            edit OBS_BASE '/pp2'
            edit OBSFILE_OBSHOST 'ASCS'
            edit OBSFILE_HPCF 'ascat_sm'
          task atovairs
            trigger get_atov eq complete and get_airs eq complete and get_ateu eq complete and get_amsa eq complete and get_mhsa eq complete and get_hirs eq complete and get_atap eq complete
          task snowlinks
            trigger get_nonbin eq complete and get_ers1_wi eq complete and get_rtovs eq complete and get_snow eq complete
          task ro_combine
            trigger get_rado eq complete and get_gras eq complete
          task reo3_combine
            trigger get_sbuv eq complete and get_g2to eq complete
          task bufr_combine
            trigger get_nonbin eq complete and get_modw eq complete and get_avhn eq complete
          task jason2_combine
            trigger get_ers1_ra eq complete and get_j2nb eq complete and get_j2eo eq complete
          task ascat_combine
            trigger get_asc2 eq complete and get_asel eq complete and get_ascs eq complete
        endfamily
        task fetchmars
          trigger ./get eq complete
        family prepare_obs
          trigger ./get eq complete
          task preobs
            trigger /e_36r1/main/12dc/an/update_psbias eq complete and /e_36r1/main/12dc/an/update_rstrhbias eq complete
            edit mode 'not1crad'
            edit REINITIALIZE 'true'
            edit REINIPERIOD_4D '6'
          task premwimg
          task preobs_wave
            edit mode 'wave'
          task pre1crad_prepare
            edit mode '1crad'
          task pre1crad_airs
            trigger pre1crad_prepare eq complete
          family pre1crad_iasi
            trigger pre1crad_prepare eq complete
            edit CPUTIME '1500'
            edit MAXTASK '16'
            task pre1crad_iasi_split_1
              edit TASKNO '1'
            task pre1crad_iasi_split_2
              edit TASKNO '2'
            task pre1crad_iasi_1
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '1'
            task pre1crad_iasi_2
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '2'
            task pre1crad_iasi_3
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '3'
            task pre1crad_iasi_4
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '4'
            task pre1crad_iasi_5
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '5'
            task pre1crad_iasi_6
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '6'
            task pre1crad_iasi_7
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '7'
            task pre1crad_iasi_8
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '8'
            task pre1crad_iasi_9
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '9'
            task pre1crad_iasi_10
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '10'
            task pre1crad_iasi_11
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '11'
            task pre1crad_iasi_12
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '12'
            task pre1crad_iasi_13
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '13'
            task pre1crad_iasi_14
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '14'
            task pre1crad_iasi_15
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '15'
            task pre1crad_iasi_16
              trigger pre1crad_iasi_split_1 eq complete and pre1crad_iasi_split_2 eq complete
              edit TASKNO '16'
          endfamily
          task pre1crad_hirs
            trigger pre1crad_prepare eq complete
          task pre1crad_msu
            trigger pre1crad_prepare eq complete
          task pre1crad_ssu
            trigger pre1crad_prepare eq complete
          task pre1crad_amsua
            trigger pre1crad_prepare eq complete
          task pre1crad_amsub
            trigger pre1crad_prepare eq complete
          task pre1crad_mhs
            trigger pre1crad_prepare eq complete
          task pregeos
          task prereo3
          task prescat
        endfamily
        family reservation
          late -c +00:03
          trigger ./get/get_stor eq complete
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
          inlimit /e_36r1/limits:reservation
          task adv_res
            late -s +00:03 -c +00:02
            edit RESERVE_FOR '30'
            edit RESERVE_IN '11'
            edit RESERVED_NODES '32'
            edit RESERVATION_LABEL 'emos06A'
            label info ""
          task loadl_cycle_limit
            late -s +00:03 -c +00:02
            edit LOADL_CYCLE '65'
            label info ""
        endfamily
      endfamily
      family an
        trigger ./obs/fetchmars eq complete and ./obs/prepare_obs eq complete
        edit FSFAMILY 'an'
        edit SMSFILES '/home/ma/emos/def/e_36r1/an'
        edit PERIOD_4D '6'
        edit WINDOW_LENGTH_4D '6'
        edit WINDOW_OFFSET_4D '3'
        edit ACCOUNT 'opebcan'
        edit REINITIALIZE 'true'
        edit ANNPES '144'
        family makeodb
          edit NPES '16'
          task cleanodb
          family bufr2odb
            trigger cleanodb eq complete
            edit THREADS '1'
            edit NPOOLS '%ANNPES%'
            edit THINFAC '1'
            edit SENSOR '-1'
            edit REBALANCE '1'
            edit SUBTYPES_INCLUDE ''
            edit SUBTYPES_EXCLUDE ''
            task b2o_conv
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_EXCLUDE '85 82 83 87'
              edit USER_PRIORITY '90'
            task b2o_surf_conv
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_EXCLUDE '142 143 144 145 85 82 83 87'
            task b2o_airs
              late -s +00:05 -c +00:15
              edit THREADS '2'
              edit SENSOR '11'
              edit SUBTYPES_INCLUDE '57'
              edit NPOOLS '$((2* %ANNPES%))'
            task b2o_iasi
              late -s +00:05 -c +00:15
              edit SENSOR '993'
              edit SUBTYPES_INCLUDE '240'
              edit THREADS '4'
              edit NPOOLS '$((2* %ANNPES%))'
            task b2o_hirs
              late -s +00:05 -c +00:15
              edit SENSOR '0'
              edit SUBTYPES_INCLUDE '54 55'
              edit NPOOLS '$((2* %ANNPES%))'
            task b2o_msu
              late -s +00:05 -c +00:15
              edit SENSOR '1'
              edit SUBTYPES_INCLUDE '54 55'
            task b2o_ssu
              late -s +00:05 -c +00:15
              edit SENSOR '2'
              edit SUBTYPES_INCLUDE '54 55'
            task b2o_amsua
              late -s +00:05 -c +00:15
              edit SENSOR '3'
              edit SUBTYPES_INCLUDE '54 55'
            task b2o_amsub
              late -s +00:05 -c +00:15
              edit SENSOR '4'
              edit SUBTYPES_INCLUDE '54 55'
            task b2o_mhs
              late -s +00:05 -c +00:15
              edit SENSOR '11'
              edit SUBTYPES_INCLUDE '54 55'
            task b2o_satob
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '82 83 87'
            task b2o_reo3
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '205 206 207 208 215'
            task b2o_ssmi
              late -s +00:05 -c +00:15
              edit SENSOR '6'
              edit SUBTYPES_INCLUDE '55 127'
            task b2o_ssmis
              late -s +00:05 -c +00:15
              edit SENSOR '10'
              edit SUBTYPES_INCLUDE '49'
            task b2o_amsre
              late -s +00:05 -c +00:15
              edit SENSOR '17'
              edit SUBTYPES_INCLUDE '59'
            task b2o_tmi
              late -s +00:05 -c +00:15
              edit SENSOR '9'
              edit SUBTYPES_INCLUDE '129'
            task b2o_scatt
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '122 137 139'
            task b2o_geos
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '189 89 76 77 78'
            task b2o_gpsro
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '250'
            task b2o_meris
              late -s +00:05 -c +00:15
              edit SENSOR '-1'
              edit SUBTYPES_INCLUDE '212'
          endfamily
          task mergeodb
            trigger bufr2odb eq complete
          task mergeodb_surf
            trigger ./bufr2odb/b2o_surf_conv eq complete
            edit NPES '16'
            edit THREADS '1'
        endfamily
        task restartodb
          trigger makeodb eq complete
        task preCleanFDB
        task vardata
          trigger ( ../../00/fc/model:step ge 9 or ../../00/fc/model eq complete)
          event hires
        task fetcherr
        task lowres
          trigger vardata:hires
          edit THREADS '4'
          edit ININPES '32'
        family surf_anal
          trigger makeodb eq complete and restartodb eq complete and preCleanFDB eq complete and vardata:hires
          edit SMSFILES '/home/ma/emos/def/e_36r1/an'
          edit FSFAMILY 'surf_anal'
          task t2ana
            edit mode 't2m'
          task rh2ana
            trigger t2ana eq complete
            edit mode 'rh2m'
          task snow
            trigger rh2ana eq complete
            edit mode 'snow'
          task sst
            edit THREADS '4'
          task slwet
            trigger sst eq complete and snow eq complete
        endfamily
        family 4dvar
          trigger makeodb eq complete and restartodb eq complete and vardata eq complete and fetcherr eq complete and lowres eq complete and preCleanFDB eq complete
          edit THREADS '8'
          edit VAR_DELAY '00'
          edit EMOS_SLEEP '00'
          family uptraj_0
            edit uptraj '0'
            task ifstraj
              late -s +00:02
              edit mode 'traj'
              edit USER_PRIORITY '91'
              edit SMSTRIES '1'
            task ifsmin
              late -s +00:02
              trigger ifstraj eq complete
              edit mode 'min'
              edit USER_PRIORITY '92'
            task ifstsave
              trigger ifstraj eq complete
              edit THREADS '1'
              edit USER_PRIORITY '93'
          endfamily
          family uptraj_1
            trigger uptraj_0 eq complete
            edit uptraj '1'
            task ifstraj
              late -s +00:02
              edit mode 'traj'
              edit USER_PRIORITY '93'
            task ifsmin
              late -s +00:02
              trigger ifstraj eq complete
              edit mode 'min'
              edit USER_PRIORITY '94'
            task ifstsave
              trigger ifstraj eq complete
              edit THREADS '1'
              edit USER_PRIORITY '95'
          endfamily
          family uptraj_2
            trigger uptraj_1 eq complete
            edit uptraj '2'
            task ifstraj
              late -s +00:02
              edit mode 'traj'
              edit USER_PRIORITY '95'
            task ifsmin
              late -s +00:02
              trigger ifstraj eq complete
              edit mode 'min'
              edit USER_PRIORITY '96'
            task ifstsave
              trigger ifstraj eq complete
              edit THREADS '1'
              edit USER_PRIORITY '97'
          endfamily
          task ifstraj
            late -s +00:02
            trigger uptraj_2 eq complete
            edit uptraj '3'
            edit mode 'traj'
            edit USER_PRIORITY '97'
            event finaltraj
            event finalwave
        endfamily
        task forceinv
          trigger ./4dvar/ifstraj:finaltraj and ./surf_anal eq complete
        family reservation1
          late -c +00:03
          trigger ./4dvar/uptraj_0/ifstraj eq active or ./4dvar/uptraj_0/ifstraj eq complete
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
          inlimit /e_36r1/limits:reservation
          task reserve_cpu
            late -s +00:03 -c +00:02
            edit RESERVED_CPUS '2432'
            edit RESERVE_FOR '35'
            label info ""
          task rm_res
            late -s +00:03 -c +00:02
            edit RESERVATION_LABEL 'emos06A'
            label info ""
        endfamily
      endfamily
      family fc
        trigger an/forceinv eq complete and ( an/4dvar/ifstraj:finalwave or an/4dvar eq complete)
        edit ACCOUNT 'opebcfc'
        edit REINITIALIZE 'true'
        edit FSFAMILY 'fc'
        task getfcdata
        task getpersSST
          trigger getfcdata eq complete
          edit SSTLEN '90'
          edit SSTINT '1'
          edit THREADS '2'
          edit NPES '16'
          edit MEM '1000'
          edit CPUTIME '960'
        task model
          late -s +00:02
          trigger ( getfcdata eq complete) and ( getpersSST eq complete)
          edit PERSST 'true'
          edit FCLENGTH '90'
          edit THREADS '8'
          edit USER_PRIORITY '94'
          edit FCNPES '144'
          meter step -1 90 90
        task clean
          trigger ../an eq complete and getfcdata eq complete and model eq complete and ../../18bc/fc eq complete
        family reservation1
          late -c +00:03
          trigger ./getfcdata eq complete
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
          inlimit /e_36r1/limits:reservation
          task reserve_cpu
            late -s +00:03 -c +00:02
            edit RESERVED_CPUS '3072'
            edit RESERVE_FOR '20'
            label info ""
        endfamily
        family reservation2
          late -c +00:03
          trigger ./model eq complete
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
          inlimit /e_36r1/limits:reservation
          task reset_cpu
            late -s +00:03 -c +00:02
            edit RESERVED_CPUS '0'
            label info ""
        endfamily
      endfamily
      family getreq
        trigger ( /e_36r1/main/06bc/an/4dvar eq active or /e_36r1/main/06bc/an/4dvar eq complete) and ( ./prod eq queued or ./prod eq complete)
        edit SMSFILES '/home/ma/emos/def/e_36r1/prod'
        edit EMOS_TYPE 'fc'
        edit PROD_TASK '1'
        edit EMOS_DOMAIN 'G'
        edit FSFAMILY 'ppdi'
        edit EMOS_PROJECT '12'
        edit SMSTRIES '1'
        edit EMOS_STREAM 'SCDA'
        task collectreq
          edit QUEUE 'diss'
          edit USER 'emos'
          edit WSHOST 'diss'
          edit QUEUE_EPILOG 'emos-exp'
          edit LOGDIR '/diss/log'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit SMSOUT '/vol/emos_nc/output'
          edit EMOS_BASE '06'
        task req
          late -s +00:03 -c +00:10
          trigger collectreq eq complete
          label info ""
          event 1 
        task collectbitmap
          trigger collectreq eq complete
          edit QUEUE 'diss'
          edit USER 'emos'
          edit WSHOST 'diss'
          edit QUEUE_EPILOG 'emos-exp'
          edit LOGDIR '/diss/log'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit SMSOUT '/vol/emos_nc/output'
        task bitmap
          late -s +00:03 -c +00:10
          trigger collectbitmap eq complete and ( req:001 or req eq complete)
          label info ""
          event 1 
      endfamily
      family prod
        trigger ( getreq/req:001 or getreq/req eq complete)
        edit SMSFILES '/home/ma/emos/def/e_36r1/prod'
        edit EMOS_DOMAIN 'G'
        edit EMOS_TYPE 'fc'
        edit FSFAMILY 'ppdi'
        edit PROD_TASK '1'
        edit EMOS_PROJECT '12'
        edit SMSURLBASE 'http://diss.ecmwf.int/do/monitoring/summary'
        family an
          trigger /e_36r1/main/06bc/an/surf_anal eq complete and ( /e_36r1/main/06bc/an/4dvar/ifstraj:finalwave or /e_36r1/main/06bc/an/4dvar/ifstraj eq complete) and 0 eq 0
          edit EMOS_TYPE 'an'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ../090/prodgen eq complete and ( /e_36r1/main/06bc/fc/model:step ge 12 or /e_36r1/main/06bc/fc/model eq complete)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            edit MEMBER 'off'
            edit STEPS '0'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 000
          trigger ( /e_36r1/main/06bc/fc/model:step ge 000 or /e_36r1/main/06bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0000'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/06bc/fc/model:step ge 12 or /e_36r1/main/06bc/fc/model eq complete) and ( ../an/backupprod eq complete or ../an/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            edit MEMBER 'off'
            edit STEPS '000'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 003
          trigger ( /e_36r1/main/06bc/fc/model:step ge 003 or /e_36r1/main/06bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0003'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/06bc/fc/model:step ge 15 or /e_36r1/main/06bc/fc/model eq complete) and ( ../000/backupprod eq complete or ../000/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../000/derived_param:SMOS or ../000/derived_param eq complete) or ( ../000/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '003'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 006
          trigger ( /e_36r1/main/06bc/fc/model:step ge 006 or /e_36r1/main/06bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0006'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/06bc/fc/model:step ge 18 or /e_36r1/main/06bc/fc/model eq complete) and ( ../003/backupprod eq complete or ../003/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../003/derived_param:SMOS or ../003/derived_param eq complete) or ( ../003/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '006'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 009
          trigger ( /e_36r1/main/06bc/fc/model:step ge 009 or /e_36r1/main/06bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0009'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/06bc/fc/model:step ge 21 or /e_36r1/main/06bc/fc/model eq complete) and ( ../006/backupprod eq complete or ../006/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../006/derived_param:SMOS or ../006/derived_param eq complete) or ( ../006/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '009'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 012
          trigger ( /e_36r1/main/06bc/fc/model:step ge 012 or /e_36r1/main/06bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0012'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/06bc/fc/model:step ge 24 or /e_36r1/main/06bc/fc/model eq complete) and ( ../009/backupprod eq complete or ../009/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../009/derived_param:SMOS or ../009/derived_param eq complete) or ( ../009/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '012'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 015
          trigger ( /e_36r1/main/06bc/fc/model:step ge 015 or /e_36r1/main/06bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0015'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/06bc/fc/model:step ge 27 or /e_36r1/main/06bc/fc/model eq complete) and ( ../012/backupprod eq complete or ../012/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../012/derived_param:SMOS or ../012/derived_param eq complete) or ( ../012/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '015'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 018
          trigger ( /e_36r1/main/06bc/fc/model:step ge 018 or /e_36r1/main/06bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0018'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/06bc/fc/model:step ge 30 or /e_36r1/main/06bc/fc/model eq complete) and ( ../015/backupprod eq complete or ../015/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../015/derived_param:SMOS or ../015/derived_param eq complete) or ( ../015/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '018'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 021
          trigger ( /e_36r1/main/06bc/fc/model:step ge 021 or /e_36r1/main/06bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0021'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/06bc/fc/model:step ge 33 or /e_36r1/main/06bc/fc/model eq complete) and ( ../018/backupprod eq complete or ../018/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../018/derived_param:SMOS or ../018/derived_param eq complete) or ( ../018/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '021'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 024
          trigger ( /e_36r1/main/06bc/fc/model:step ge 024 or /e_36r1/main/06bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0024'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/06bc/fc/model:step ge 36 or /e_36r1/main/06bc/fc/model eq complete) and ( ../021/backupprod eq complete or ../021/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../021/derived_param:SMOS or ../021/derived_param eq complete) or ( ../021/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '024'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 027
          trigger ( /e_36r1/main/06bc/fc/model:step ge 027 or /e_36r1/main/06bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0027'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/06bc/fc/model:step ge 39 or /e_36r1/main/06bc/fc/model eq complete) and ( ../024/backupprod eq complete or ../024/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../024/derived_param:SMOS or ../024/derived_param eq complete) or ( ../024/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '027'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 030
          trigger ( /e_36r1/main/06bc/fc/model:step ge 030 or /e_36r1/main/06bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0030'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/06bc/fc/model:step ge 42 or /e_36r1/main/06bc/fc/model eq complete) and ( ../027/backupprod eq complete or ../027/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../027/derived_param:SMOS or ../027/derived_param eq complete) or ( ../027/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '030'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 033
          trigger ( /e_36r1/main/06bc/fc/model:step ge 033 or /e_36r1/main/06bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0033'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/06bc/fc/model:step ge 45 or /e_36r1/main/06bc/fc/model eq complete) and ( ../030/backupprod eq complete or ../030/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../030/derived_param:SMOS or ../030/derived_param eq complete) or ( ../030/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '033'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 036
          trigger ( /e_36r1/main/06bc/fc/model:step ge 036 or /e_36r1/main/06bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0036'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/06bc/fc/model:step ge 48 or /e_36r1/main/06bc/fc/model eq complete) and ( ../033/backupprod eq complete or ../033/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../033/derived_param:SMOS or ../033/derived_param eq complete) or ( ../033/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '036'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 039
          trigger ( /e_36r1/main/06bc/fc/model:step ge 039 or /e_36r1/main/06bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0039'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/06bc/fc/model:step ge 51 or /e_36r1/main/06bc/fc/model eq complete) and ( ../036/backupprod eq complete or ../036/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../036/derived_param:SMOS or ../036/derived_param eq complete) or ( ../036/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '039'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 042
          trigger ( /e_36r1/main/06bc/fc/model:step ge 042 or /e_36r1/main/06bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0042'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/06bc/fc/model:step ge 54 or /e_36r1/main/06bc/fc/model eq complete) and ( ../039/backupprod eq complete or ../039/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../039/derived_param:SMOS or ../039/derived_param eq complete) or ( ../039/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '042'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 045
          trigger ( /e_36r1/main/06bc/fc/model:step ge 045 or /e_36r1/main/06bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0045'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/06bc/fc/model:step ge 57 or /e_36r1/main/06bc/fc/model eq complete) and ( ../042/backupprod eq complete or ../042/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../042/derived_param:SMOS or ../042/derived_param eq complete) or ( ../042/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '045'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 048
          trigger ( /e_36r1/main/06bc/fc/model:step ge 048 or /e_36r1/main/06bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0048'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/06bc/fc/model:step ge 60 or /e_36r1/main/06bc/fc/model eq complete) and ( ../045/backupprod eq complete or ../045/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../045/derived_param:SMOS or ../045/derived_param eq complete) or ( ../045/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '048'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 051
          trigger ( /e_36r1/main/06bc/fc/model:step ge 051 or /e_36r1/main/06bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0051'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/06bc/fc/model:step ge 63 or /e_36r1/main/06bc/fc/model eq complete) and ( ../048/backupprod eq complete or ../048/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../048/derived_param:SMOS or ../048/derived_param eq complete) or ( ../048/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '051'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 054
          trigger ( /e_36r1/main/06bc/fc/model:step ge 054 or /e_36r1/main/06bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0054'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/06bc/fc/model:step ge 66 or /e_36r1/main/06bc/fc/model eq complete) and ( ../051/backupprod eq complete or ../051/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../051/derived_param:SMOS or ../051/derived_param eq complete) or ( ../051/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '054'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 057
          trigger ( /e_36r1/main/06bc/fc/model:step ge 057 or /e_36r1/main/06bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0057'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/06bc/fc/model:step ge 69 or /e_36r1/main/06bc/fc/model eq complete) and ( ../054/backupprod eq complete or ../054/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../054/derived_param:SMOS or ../054/derived_param eq complete) or ( ../054/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '057'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 060
          trigger ( /e_36r1/main/06bc/fc/model:step ge 060 or /e_36r1/main/06bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0060'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/06bc/fc/model:step ge 72 or /e_36r1/main/06bc/fc/model eq complete) and ( ../057/backupprod eq complete or ../057/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../057/derived_param:SMOS or ../057/derived_param eq complete) or ( ../057/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '060'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 063
          trigger ( /e_36r1/main/06bc/fc/model:step ge 063 or /e_36r1/main/06bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0063'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/06bc/fc/model:step ge 75 or /e_36r1/main/06bc/fc/model eq complete) and ( ../060/backupprod eq complete or ../060/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../060/derived_param:SMOS or ../060/derived_param eq complete) or ( ../060/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '063'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 066
          trigger ( /e_36r1/main/06bc/fc/model:step ge 066 or /e_36r1/main/06bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0066'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/06bc/fc/model:step ge 78 or /e_36r1/main/06bc/fc/model eq complete) and ( ../063/backupprod eq complete or ../063/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../063/derived_param:SMOS or ../063/derived_param eq complete) or ( ../063/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '066'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 069
          trigger ( /e_36r1/main/06bc/fc/model:step ge 069 or /e_36r1/main/06bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0069'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/06bc/fc/model:step ge 81 or /e_36r1/main/06bc/fc/model eq complete) and ( ../066/backupprod eq complete or ../066/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../066/derived_param:SMOS or ../066/derived_param eq complete) or ( ../066/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '069'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 072
          trigger ( /e_36r1/main/06bc/fc/model:step ge 072 or /e_36r1/main/06bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0072'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/06bc/fc/model:step ge 84 or /e_36r1/main/06bc/fc/model eq complete) and ( ../069/backupprod eq complete or ../069/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../069/derived_param:SMOS or ../069/derived_param eq complete) or ( ../069/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '072'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 075
          trigger ( /e_36r1/main/06bc/fc/model:step ge 075 or /e_36r1/main/06bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0075'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/06bc/fc/model:step ge 87 or /e_36r1/main/06bc/fc/model eq complete) and ( ../072/backupprod eq complete or ../072/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../072/derived_param:SMOS or ../072/derived_param eq complete) or ( ../072/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '075'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 078
          trigger ( /e_36r1/main/06bc/fc/model:step ge 078 or /e_36r1/main/06bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0078'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task backupprod
            late -s +00:05 -c +00:20
            trigger ( /e_36r1/main/06bc/fc/model:step ge 90 or /e_36r1/main/06bc/fc/model eq complete) and ( ../075/backupprod eq complete or ../075/backupprod eq active)
            edit PGNPES '16'
            inlimit /e_36r1/limits:backupprod
          task derived_param
            trigger ( ../075/derived_param:SMOS or ../075/derived_param eq complete) or ( ../075/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '078'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 081
          trigger ( /e_36r1/main/06bc/fc/model:step ge 081 or /e_36r1/main/06bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0081'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task derived_param
            trigger ( ../078/derived_param:SMOS or ../078/derived_param eq complete) or ( ../078/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '081'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 084
          trigger ( /e_36r1/main/06bc/fc/model:step ge 084 or /e_36r1/main/06bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0084'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task derived_param
            trigger ( ../081/derived_param:SMOS or ../081/derived_param eq complete) or ( ../081/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '084'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 087
          trigger ( /e_36r1/main/06bc/fc/model:step ge 087 or /e_36r1/main/06bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0087'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task derived_param
            trigger ( ../084/derived_param:SMOS or ../084/derived_param eq complete) or ( ../084/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '087'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
        family 090
          trigger ( /e_36r1/main/06bc/fc/model:step ge 090 or /e_36r1/main/06bc/fc/model eq complete) and 0 eq 0
          edit EMOS_TIME_STEP_H '0090'
          task prodgen
            late -s +00:05 -c +00:15
            trigger ./derived_param:SMOS or ./derived_param eq complete
            edit USER_PRIORITY '92'
            inlimit /e_36r1/limits:prod
            inlimit /e_36r1/limits:prod_bc2
          task derived_param
            trigger ( ../087/derived_param:SMOS or ../087/derived_param eq complete) or ( ../087/derived_param eq aborted)
            edit MEMBER 'off'
            edit STEPS '090'
            edit SMSFILES '/home/ma/emos/def/e_36r1/mc/fc'
            event SMOS
        endfamily
      endfamily
      family prod2diss
        edit USER_PRIORITY '90'
        edit SMSURLBASE 'http://diss.ecmwf.int/do/monitoring/summary'
        edit SMSFILES '/home/ma/emos/def/e_36r1/prod'
        task operation_is_late
          event yes
        family oper
          edit EMOS_DOMAIN 'G'
          edit EMOS_TYPE 'fc'
          edit FSFAMILY 'ppdi'
          edit PROD_TASK '1'
          edit USER_PRIORITY '92'
          edit SMSURL 'GSCDA/06'
          edit EMOS_PROJECT '12'
          family an
            task pdb
              trigger ../../../prod/an/prodgen eq complete
              edit EMOS_TYPE 'an'
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete
              edit EMOS_TYPE 'an'
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../090/q2diss eq complete and ../../../prod/an/backupprod eq complete
              edit EMOS_TYPE 'an'
              label info ""
              event 1 
          endfamily
          family 000
            edit EMOS_TIME_STEP_H '0000'
            task pdb
              trigger ../../../prod/000/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ../an/q2diss eq complete
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
          endfamily
          family 003
            edit EMOS_TIME_STEP_H '0003'
            task pdb
              trigger ../../../prod/003/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../000/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/003/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 006
            edit EMOS_TIME_STEP_H '0006'
            task pdb
              trigger ../../../prod/006/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../003/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/006/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 009
            edit EMOS_TIME_STEP_H '0009'
            task pdb
              trigger ../../../prod/009/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../006/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/009/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 012
            edit EMOS_TIME_STEP_H '0012'
            task pdb
              trigger ../../../prod/012/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../009/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/012/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 015
            edit EMOS_TIME_STEP_H '0015'
            task pdb
              trigger ../../../prod/015/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../012/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/015/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 018
            edit EMOS_TIME_STEP_H '0018'
            task pdb
              trigger ../../../prod/018/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../015/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/018/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 021
            edit EMOS_TIME_STEP_H '0021'
            task pdb
              trigger ../../../prod/021/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../018/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/021/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 024
            edit EMOS_TIME_STEP_H '0024'
            task pdb
              trigger ../../../prod/024/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../021/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/024/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 027
            edit EMOS_TIME_STEP_H '0027'
            task pdb
              trigger ../../../prod/027/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../024/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/027/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 030
            edit EMOS_TIME_STEP_H '0030'
            task pdb
              trigger ../../../prod/030/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../027/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/030/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 033
            edit EMOS_TIME_STEP_H '0033'
            task pdb
              trigger ../../../prod/033/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../030/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/033/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 036
            edit EMOS_TIME_STEP_H '0036'
            task pdb
              trigger ../../../prod/036/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../033/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/036/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 039
            edit EMOS_TIME_STEP_H '0039'
            task pdb
              trigger ../../../prod/039/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../036/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/039/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 042
            edit EMOS_TIME_STEP_H '0042'
            task pdb
              trigger ../../../prod/042/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../039/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/042/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 045
            edit EMOS_TIME_STEP_H '0045'
            task pdb
              trigger ../../../prod/045/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../042/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/045/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 048
            edit EMOS_TIME_STEP_H '0048'
            task pdb
              trigger ../../../prod/048/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../045/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/048/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 051
            edit EMOS_TIME_STEP_H '0051'
            task pdb
              trigger ../../../prod/051/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../048/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/051/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 054
            edit EMOS_TIME_STEP_H '0054'
            task pdb
              trigger ../../../prod/054/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../051/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/054/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 057
            edit EMOS_TIME_STEP_H '0057'
            task pdb
              trigger ../../../prod/057/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../054/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/057/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 060
            edit EMOS_TIME_STEP_H '0060'
            task pdb
              trigger ../../../prod/060/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../057/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/060/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 063
            edit EMOS_TIME_STEP_H '0063'
            task pdb
              trigger ../../../prod/063/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../060/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/063/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 066
            edit EMOS_TIME_STEP_H '0066'
            task pdb
              trigger ../../../prod/066/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../063/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/066/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 069
            edit EMOS_TIME_STEP_H '0069'
            task pdb
              trigger ../../../prod/069/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../066/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/069/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 072
            edit EMOS_TIME_STEP_H '0072'
            task pdb
              trigger ../../../prod/072/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../069/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/072/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 075
            edit EMOS_TIME_STEP_H '0075'
            task pdb
              trigger ../../../prod/075/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../072/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/075/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 078
            edit EMOS_TIME_STEP_H '0078'
            task pdb
              trigger ../../../prod/078/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../075/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
            task dissbackup
              late -s +00:05 -c +00:15
              trigger ../../../prod/078/backupprod eq complete
              inlimit /e_36r1/limits:dissbackup
              label info ""
              event 1 
          endfamily
          family 081
            edit EMOS_TIME_STEP_H '0081'
            task pdb
              trigger ../../../prod/081/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../078/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
          endfamily
          family 084
            edit EMOS_TIME_STEP_H '0084'
            task pdb
              trigger ../../../prod/084/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../081/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
          endfamily
          family 087
            edit EMOS_TIME_STEP_H '0087'
            task pdb
              trigger ../../../prod/087/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../084/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
          endfamily
          family 090
            edit EMOS_TIME_STEP_H '0090'
            task pdb
              trigger ../../../prod/090/prodgen eq complete
              inlimit /e_36r1/limits:pdb_bc
              inlimit /limits:pdb_bc
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete and ( not ../../../prod2diss/operation_is_late:yes or ../087/q2diss eq complete)
              edit USER_PRIORITY '93'
              inlimit /e_36r1/limits:q2diss_bc
              inlimit /limits:q2diss_bc
              label info ""
              event 1 
          endfamily
          task check
            trigger ./078 eq complete
        endfamily
      endfamily
    endfamily
    task logfilestore
      edit QUEUE 'emos-e'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit SMSFILES '/home/ma/emos/def/e_36r1'
      edit LOGTASK 'main'
      edit KEEPLOGS 'no'
      edit LOGSTOECFS 'no'
      edit MOVE_TO_ECFS 'yes'
    task logfiles
      late -s +00:05 -c +00:15
      trigger make eq complete and ./00dc eq complete and ./06 eq complete and ./12 eq complete and ./12dc eq complete and ./18 eq complete and ./00 eq complete and ./12bc eq complete and ./18bc eq complete and ./00bc eq complete and ./06bc eq complete and logfilestore eq complete
      edit QUEUE 'emos-e'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit SMSFILES '/home/ma/emos/def/e_36r1'
      edit LOGTASK 'main'
      edit MOVE_TO_ECFS 'no'
      time 12:00
  endfamily
  family lag
    repeat date YMD 20091118 20121212 1
    edit MARS_DB 'marsod'
    family 00dc
      edit EMOS_BASE '00'
      edit MXUP_TRAJ '3'
      edit STREAM 'DCDA'
      edit WINDOW_LENGTH_4D '12'
      edit WINDOW_OFFSET_4D '3'
      edit EDFAMILY 'false'
      edit PERIOD_4D '12'
      edit FCLENGTH '15'
      edit DELTA_DAY '0'
      family ddh
        trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/00dc/fc eq complete)
        edit EMOS_BASE '06'
        edit SMSFILES '/home/ma/emos/def/e_36r1/fc'
        task ddh
      endfamily
      family sync
        trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/00dc/an eq complete)
        edit EMOS_TYPE 'an'
        edit FSFAMILY 'an'
        edit SMSFILES '/home/ma/emos/def/e_36r1/sync'
        task dcda_to_da_bk
          edit QUEUE 'ns'
          edit QUEUE_EPILOG 'ns'
          edit SCHOST 'c1a'
          edit STHOST '/s1a'
          edit SCHOST_BKUP 'c1b'
          edit STHOST_BKUP '/s1b'
          edit SMSLOGHOST 'c1a'
          edit LOGDIR '/s1a/emos_esuite/emos_data/0046/log'
          edit SMSOUT '/s1a/emos_esuite/emos_data/0046/log'
          edit OPHOST 'c1b'
          edit SMSFILES '/home/ma/emos/def/e_36r1/an'
          inlimit /limits:c1a
          label files ""
          label info ""
          event fdb
        task sync_satmon
          trigger ../satmon eq complete
        task sync_bias
        task sync_fc
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/00dc/fc/model:step gt 6 or /e_36r1/main/00dc/fc/model eq complete))
          edit QUEUE 'ns'
          edit QUEUE_EPILOG 'ns'
          edit SCHOST 'c1a'
          edit STHOST '/s1a'
          edit SCHOST_BKUP 'c1b'
          edit STHOST_BKUP '/s1b'
          edit SMSLOGHOST 'c1a'
          edit LOGDIR '/s1a/emos_esuite/emos_data/0046/log'
          edit SMSOUT '/s1a/emos_esuite/emos_data/0046/log'
          edit EMOS_BASE '06'
          inlimit /limits:c1a
          label files ""
          label info ""
          event fdb
        task sync_an
          edit QUEUE 'ns'
          edit QUEUE_EPILOG 'ns'
          edit SCHOST 'c1a'
          edit STHOST '/s1a'
          edit SCHOST_BKUP 'c1b'
          edit STHOST_BKUP '/s1b'
          edit SMSLOGHOST 'c1a'
          edit LOGDIR '/s1a/emos_esuite/emos_data/0046/log'
          edit SMSOUT '/s1a/emos_esuite/emos_data/0046/log'
          edit EMOS_BASE '06'
          inlimit /limits:c1a
          label files ""
          label info ""
          event fdb
        task sync_wv
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/00dc/fc/model:step gt 3 or /e_36r1/main/00dc/fc/model eq complete))
          edit QUEUE 'ns'
          edit QUEUE_EPILOG 'ns'
          edit SCHOST 'c1a'
          edit STHOST '/s1a'
          edit SCHOST_BKUP 'c1b'
          edit STHOST_BKUP '/s1b'
          edit SMSLOGHOST 'c1a'
          edit LOGDIR '/s1a/emos_esuite/emos_data/0046/log'
          edit SMSOUT '/s1a/emos_esuite/emos_data/0046/log'
          edit EMOS_BASE '06'
          inlimit /limits:c1a
          label files ""
          label info ""
          event fdb
        task sync_clean
          edit QUEUE 'ns'
          edit QUEUE_EPILOG 'ns'
          edit SCHOST 'c1a'
          edit STHOST '/s1a'
          edit SCHOST_BKUP 'c1b'
          edit STHOST_BKUP '/s1b'
          edit SMSLOGHOST 'c1a'
          edit LOGDIR '/s1a/emos_esuite/emos_data/0046/log'
          edit SMSOUT '/s1a/emos_esuite/emos_data/0046/log'
          inlimit /limits:c1a
      endfamily
      family archive
        trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ../../main/00dc/an eq complete and ( ../../main/00dc/fc/model:step ge 15 or ../../main/00dc/fc/model eq complete))
        edit FSFAMILY 'mars'
        edit SMSFILES '/home/ma/emos/def/e_36r1/archive'
        edit EMOS_TYPE 'an'
        inlimit /limits:mars
        inlimit /e_36r1/limits:mars
        task sp2fdb
          edit STEPS '3/6/9/12/18'
          label files ""
          label info ""
          event fdb
        task ansfc
          trigger sp2fdb eq complete
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task anml
          trigger ansfc:fdb or ansfc eq complete
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          inlimit /e_36r1/limits:ml
          label files ""
          label info ""
          event fdb
        task anpl
          trigger anml:fdb or anml eq complete
          edit FCLENGTH '15'
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task anwave
          trigger ansfc:fdb or ansfc eq complete
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task anil
          trigger anpl:fdb or anpl eq complete
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task derivedp
          trigger ./ansfc eq complete
          edit STEPS '0'
          label files ""
          label info ""
          event fdb
        family obs
          edit FSFAMILY 'an'
          inlimit /e_36r1/limits:obs
          task ai
            trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/00dc/an/4dvar/ifstraj:finaltraj or /e_36r1/main/00dc/an/4dvar eq complete))
            edit TYPE 'ai'
            label files ""
            label info ""
            event fdb
          task af
            trigger ../../fb/feedback eq complete
            edit TYPE 'af'
            label files ""
            label info ""
            event fdb
          task fb
            trigger ../../fb/feedback eq complete
            edit TYPE 'fb'
            label files ""
            label info ""
            event fdb
        endfamily
        task ice_nmc
          trigger ansfc:fdb or ansfc eq complete
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task sst
          trigger ansfc:fdb or ansfc eq complete
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        family fc015
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/00dc/fc/model:step ge 21 or /e_36r1/main/00dc/fc/model eq complete)
          edit STEPS '3/6/9/12/15'
          edit EMOS_BASE '06'
          edit EMOS_TYPE 'fc'
          task sp2fdb
            label files ""
            label info ""
            event fdb
          task fcsfc
            trigger sp2fdb eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcml
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            inlimit /e_36r1/limits:ml
            label files ""
            label info ""
            event fdb
          task fcpl
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcil
            trigger fcpl:fdb or fcpl eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
        endfamily
        task fc12g
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/00dc/fc/model eq complete)
          edit STEPS_SPECTRAL '6/12'
          edit EMOS_BASE '06'
          edit STEPS '0/3/6/9/12/15'
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      family fb
        trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/00dc/an eq complete)
        edit FSFAMILY 'an'
        edit SMSFILES '/home/ma/emos/def/e_36r1/fb'
        task sync_rstrhbias
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
        task biassave
          trigger sync_rstrhbias eq complete
        family feedback
          edit ACCOUNT 'oedcdaan'
          edit NPES '8'
          edit FBNEW2OLD '1'
          edit LCONV '1'
          edit LSATOB '1'
          edit LREO3 '1'
          edit LAEOLUS '0'
          edit LSSMI '1'
          edit LSSMIS '1'
          edit LAMSRE '1'
          edit LTMI '1'
          edit LHIRS '1'
          edit LMSU '1'
          edit LSSU '1'
          edit LAMSUA '1'
          edit LAMSUB '1'
          edit LMHS '1'
          edit LGEOS '1'
          edit LSCAT '1'
          edit LAIRS '1'
          edit LVTPR1 '0'
          edit LVTPR2 '0'
          edit LGPSRO '1'
          edit LIASI '1'
          edit LMERIS '1'
          task revmatchup
            edit NPES '128'
            edit THREADS '2'
          task matchup
            trigger revmatchup eq complete
            edit NPES '128'
            edit THREADS '2'
          task odbcmp_prepare
            edit THREADS '1'
            edit NPES '1'
            edit MEM '1000'
            edit CPUTIME '1000'
            edit PRIORITY '90'
          family odb_compress
            trigger odbcmp_prepare eq complete and matchup eq complete
            edit THREADS '2'
            edit NPES '1'
            task odbcmp_conv
            task odbcmp_surf_conv
            task odbcmp_airs
            task odbcmp_iasi
            task odbcmp_hirs
            task odbcmp_msu
            task odbcmp_ssu
            task odbcmp_amsua
            task odbcmp_amsub
            task odbcmp_mhs
            task odbcmp_satob
            task odbcmp_reo3
            task odbcmp_ssmi
            task odbcmp_ssmis
            task odbcmp_amsre
            task odbcmp_tmi
            task odbcmp_scatt
            task odbcmp_geos
            task odbcmp_gpsro
            task odbcmp_meris
          endfamily
          family odb2bufr
            trigger matchup eq complete
            edit NPES '1'
            edit SUBTYPES_INCLUDE ''
            edit SUBTYPES_EXCLUDE ''
            task o2b_conv
              edit SUBTYPES_EXCLUDE '85 82 83 87'
              edit NPES '2'
            task o2b_surf_conv
              edit SUBTYPES_EXCLUDE '142 143 144 145 85 82 83 87'
              edit NPES '2'
            task o2b_airs
              edit SUBTYPES_INCLUDE '57'
              edit NPES '8'
              edit THREADS '2'
            task o2b_iasi
              edit SUBTYPES_INCLUDE '240'
              edit NPES '8'
              edit THREADS '2'
            task o2b_hirs
              edit SUBTYPES_INCLUDE '54 55'
              edit NPES '2'
            task o2b_msu
              edit SUBTYPES_INCLUDE '54 55'
            task o2b_ssu
              edit SUBTYPES_INCLUDE '54 55'
            task o2b_amsua
              edit SUBTYPES_INCLUDE '54 55'
              edit NPES '2'
            task o2b_amsub
              edit SUBTYPES_INCLUDE '54 55'
              edit NPES '2'
            task o2b_mhs
              edit SUBTYPES_INCLUDE '54 55'
              edit NPES '2'
            task o2b_satob
              edit SUBTYPES_INCLUDE '82 83 87'
              edit NPES '2'
            task o2b_reo3
              edit SUBTYPES_INCLUDE '205 206 207 208 215'
              edit NPES '4'
              edit THREADS '2'
            task o2b_ssmi
              edit SUBTYPES_INCLUDE '55 127'
            task o2b_ssmis
              edit SUBTYPES_INCLUDE '49'
            task o2b_amsre
              edit SUBTYPES_INCLUDE '59'
            task o2b_tmi
              edit SUBTYPES_INCLUDE '129'
            task o2b_scatt
              edit SUBTYPES_INCLUDE '122 137 139'
            task o2b_geos
              edit SUBTYPES_INCLUDE '189 89 76 77 78'
              edit NPES '2'
            task o2b_gpsro
              edit SUBTYPES_INCLUDE '250'
            task o2b_meris
              edit SUBTYPES_INCLUDE '212'
          endfamily
          task mergebufr
            trigger odb2bufr eq complete
        endfamily
        family obstat
          trigger feedback/odb2bufr eq complete
          edit NPES '1'
          task obstat_conv
          task obstat_surf_conv
          task obstat_airs
            edit NPES '8'
            edit THREADS '2'
          task obstat_iasi
            edit NPES '8'
            edit THREADS '2'
          task obstat_hirs
            edit NPES '4'
          task obstat_msu
          task obstat_ssu
          task obstat_amsua
            edit NPES '2'
          task obstat_amsub
          task obstat_mhs
          task obstat_satob
          task obstat_reo3
          task obstat_ssmi
            edit NPES '2'
          task obstat_ssmis
            edit NPES '2'
          task obstat_amsre
            edit NPES '2'
          task obstat_tmi
            edit NPES '2'
          task obstat_scatt
          task obstat_geos
            edit NPES '2'
          task obstat_gpsro
            edit NPES '2'
          task obstat_meris
            edit NPES '2'
        endfamily
        task fdbksave
          trigger feedback eq complete and obstat eq complete
        task clean_odbcmp
          trigger feedback eq complete and obstat eq complete and fdbksave eq complete
        task wavesave
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ../../../main/00dc/an/4dvar eq complete)
        task obstat_merge
          trigger obstat eq complete
          edit WSHOST 'bee01'
          edit PRIORITY '51'
      endfamily
      family traj
        task feed
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ../../../main/00dc/fc eq complete)
        task cp_to_other
          trigger feed eq complete
      endfamily
      family satmon
        edit FSFAMILY 'an'
        inlimit /e_36r1/limits:smon
        task getsmon
          trigger ../fb/fdbksave eq complete and ../archive/obs/af eq complete
        task smon_meris
          trigger getsmon eq complete
          edit smon_smsmode 'MERIS'
        task smon_amsre
          trigger getsmon eq complete
          edit smon_smsmode 'AMSRE'
        task smon_ssmis
          trigger getsmon eq complete
          edit smon_smsmode 'SSMIS'
        task smon_tmi
          trigger getsmon eq complete
          edit smon_smsmode 'TMI'
        task smon_gpsro
          trigger getsmon eq complete
          edit smon_smsmode 'GPSRO'
        task smon_pacrars
          trigger getsmon eq complete
          edit smon_smsmode 'PACRARS'
        task smon_ears
          trigger getsmon eq complete
          edit smon_smsmode 'EARS'
        task smon_o3msg
          trigger getsmon eq complete
          edit smon_smsmode 'O3MSG'
        task smon_o3knmi
          trigger getsmon eq complete
          edit smon_smsmode 'O3KNMI'
        task smon_pwc
          trigger getsmon eq complete
          edit smon_smsmode 'PWC'
        task smon_temp
          trigger getsmon eq complete
          edit smon_smsmode 'TEMP'
        task smon_iasi
          trigger getsmon eq complete
          edit smon_smsmode 'IASI'
        task smon_airs
          trigger getsmon eq complete
          edit smon_smsmode 'AIRS'
        task smon_dbamv
          trigger getsmon eq complete
          edit smon_smsmode 'DBAMV'
        task smon_amv
          trigger getsmon eq complete
          edit smon_smsmode 'AMV'
        task smon_o3
          trigger getsmon eq complete
          edit smon_smsmode 'O3'
        task smon_dmsp
          trigger getsmon eq complete
          edit smon_smsmode 'DMSP'
        task smon_noaa
          trigger getsmon eq complete
          edit smon_smsmode 'NOAA'
        task smon_geos
          trigger getsmon eq complete
          edit smon_smsmode 'GEOS'
        task smon_scatt
          trigger getsmon eq complete
          edit smon_smsmode 'SCATT'
        task smon_slmoist
          trigger getsmon eq complete
          edit smon_smsmode 'SLMOIST'
        task smon_clean
          trigger getsmon eq complete and smon_meris eq complete and smon_amsre eq complete and smon_ssmis eq complete and smon_tmi eq complete and smon_gpsro eq complete and smon_pacrars eq complete and smon_ears eq complete and smon_o3msg eq complete and smon_o3knmi eq complete and smon_pwc eq complete and smon_temp eq complete and smon_iasi eq complete and smon_airs eq complete and smon_dbamv eq complete and smon_amv eq complete and smon_o3 eq complete and smon_dmsp eq complete and smon_noaa eq complete and smon_geos eq complete and smon_scatt eq complete and smon_slmoist eq complete
      endfamily
      task clean
        trigger fb eq complete and satmon eq complete
        edit FSFAMILY 'an'
      task cleanplus6
        trigger clean eq complete
        edit EMOS_BASE '06'
        edit FSFAMILY 'an'
      family verif
        edit QUEUE 'emos-e'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        edit SMSFILES '/home/ma/emos/def/e_36r1/verif'
        edit DELTA_DAY '1'
        inlimit /limits:linux_cluster
        family fc
          task v12li1
          task v12li2
          task v12up1
            trigger v12li1 eq complete
          task v12up2
            trigger v12up1 eq complete and v12li2 eq complete
          task v12pl
            trigger v12up2 eq complete
          task v12memv
            trigger v12up2 eq complete
            edit DATEMASK '02.*.*'
          task v12memb
            trigger v12up2 eq complete
            edit DATEMASK '11.*.*'
          task v12upmv
            trigger v12memv eq complete
            edit DATEMASK '02.*.*'
          task v12upmb
            trigger v12memb eq complete
            edit DATEMASK '11.*.*'
          task v12crmv
            edit DATEMASK '28.*.*'
          task v12crmb
            edit DATEMASK '28.*.*'
          task v12crd
            edit DATEMASK '28.*.*'
        endfamily
        family oc
          trigger ./fc eq complete or ./fc eq aborted
          task mk_dummy_fields
            edit QUEUE 'emos'
            edit USER 'emos'
            edit PP_BASE '/pp1'
            edit WSHOST 'pp1'
            edit QUEUE_EPILOG 'emos-exp'
            edit LOGDIR '/pp1/log'
            edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
            edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
            edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
            edit SMSOUT '/vol/emos_nc/output'
          task insert_fields
            trigger mk_dummy_fields eq complete
            edit QUEUE 'emos-e'
            edit QUEUE_EPILOG 'emos-exp'
            edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
            edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
            edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
            edit LOGDIR '/vol/emos_nc/output'
            edit SMSOUT '/vol/emos_nc/output'
            edit PP_BASE '/pp1'
            inlimit /limits:linux_cluster
          task vbrli1
            trigger insert_fields eq complete
          task vbrli2
            trigger vbrli1 eq complete
          task vwali1
            trigger insert_fields eq complete
          task vwali2
            trigger vwali1 eq complete
          task vwali3
            trigger vwali2 eq complete
          task vofli
            trigger insert_fields eq complete
          task vtlli
            trigger insert_fields eq complete
          task vtoli
            trigger insert_fields eq complete
          task vmoli
            trigger insert_fields eq complete
          task vbrup1
            trigger vbrli1 eq complete
          task vbrup2
            trigger vbrup1 eq complete and vbrli2 eq complete
          task vwaup1
            trigger vwali1 eq complete
          task vwaup2
            trigger vwaup1 eq complete and vwali2 eq complete
          task vwaup3
            trigger vwaup2 eq complete and vwali3 eq complete
          task vofup
            trigger vofli eq complete
          task vtlup
            trigger vtlli eq complete
          task vtoup
            trigger vtoli eq complete
          task vmoup
            trigger vmoli eq complete
          task vmeli
            trigger insert_fields eq complete
          task vmeup
            trigger vmeli eq complete
          task voccrd
            edit DATEMASK '28.*.*'
          task voccrm
            edit DATEMASK '28.*.*'
          task vocmem
            edit DATEMASK '11.*.*'
          task vocupm
            trigger vocmem eq complete
            edit DATEMASK '11.*.*'
          task cleanup
            trigger vmeup eq complete and vbrup2 eq complete and vwaup3 eq complete and vofup eq complete and vtlup eq complete and vtoup eq complete and vmoup eq complete and voccrd eq complete and voccrm eq complete and vocupm eq complete
            edit QUEUE 'emos-e'
            edit QUEUE_EPILOG 'emos-exp'
            edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
            edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
            edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
            edit LOGDIR '/vol/emos_nc/output'
            edit SMSOUT '/vol/emos_nc/output'
            inlimit /limits:linux_cluster
        endfamily
        family cf
          trigger oc eq complete or oc eq aborted
          task vcnli1
            late -c +02:00
          task vcnli2
          task vcnup1
            trigger vcnli1 eq complete
          task vcnup2
            trigger vcnup1 eq complete and vcnli2 eq complete
          task vcnpl
            trigger vcnup2 eq complete
          task vcncrd
          task vcncrm
            edit DATEMASK '28.*.*'
          task vcnmem
            trigger vcnup2 eq complete
            edit DATEMASK '11.*.*'
          task vcnupm
            trigger vcnmem eq complete
            edit DATEMASK '11.*.*'
          task vcnli3
          task vcnli4
          task vcnup3
            trigger vcnup2 eq complete and vcnli3 eq complete
          task vcnup4
            trigger vcnup3 eq complete and vcnli4 eq complete
          task vcnmem2
            trigger vcnup4 eq complete
            edit DATEMASK '22.*.*'
          task vcnupm2
            trigger vcnmem2 eq complete
            edit DATEMASK '22.*.*'
        endfamily
        family ob
          trigger oc eq complete or oc eq aborted
          time 05:00
          family stage1
            task vobli1
            task vobli2
            task vobup1
              trigger vobli1 eq complete
            task vobup2
              trigger vobup1 eq complete and vobli2 eq complete
            task vobnobs
              trigger vobup2 eq complete
              edit DATEMASK '03.*.*'
            task vobmem
              trigger ./vobup2 eq complete
              edit DATEMASK '03.*.*'
            task vobupm
              trigger vobmem eq complete
              edit DATEMASK '03.*.*'
          endfamily
          family stage2
            trigger ./stage1 eq complete
            edit EMOS_BASE '00'
            task vobli1
            task vobli2
            task vobup1
              trigger vobli1 eq complete
            task vobup2
              trigger vobup1 eq complete and vobli2 eq complete
            task vobnobs
              trigger vobup2 eq complete
              edit DATEMASK '03.*.*'
            task vobmem
              trigger ./vobup2 eq complete
              edit DATEMASK '03.*.*'
            task vobupm
              trigger vobmem eq complete
              edit DATEMASK '03.*.*'
          endfamily
          task vobcrd
            edit DATEMASK '28.*.*'
          task vobcrm
            edit DATEMASK '28.*.*'
        endfamily
        family em
          task vemli
            late -c +02:00
          task vemup
            trigger vemli eq complete
          task vemcrd
            edit DATEMASK '28.*.*'
          task vemcrm
            edit DATEMASK '28.*.*'
          task vemmem
            trigger vemup eq complete
            edit DATEMASK '11.*.*'
          task vemupm
            trigger vemmem eq complete
            edit DATEMASK '11.*.*'
        endfamily
        family es
          trigger em eq complete
          task vesli
            late -c +02:00
          task vesup
            trigger vesli eq complete
          task vescrd
            edit DATEMASK '28.*.*'
          task vescrm
            edit DATEMASK '28.*.*'
          task vesmem
            trigger vesup eq complete
            edit DATEMASK '11.*.*'
          task vesupm
            trigger vesmem eq complete
            edit DATEMASK '11.*.*'
        endfamily
        family cv
          trigger es eq complete
          task vcvli
            late -c +02:00
          task vcvup
            trigger vcvli eq complete
          task vcvcrd
            edit DATEMASK '28.*.*'
          task vcvcrm
            edit DATEMASK '28.*.*'
          task vcvmem
            trigger vcvup eq complete
            edit DATEMASK '11.*.*'
          task vcvupm
            trigger vcvmem eq complete
            edit DATEMASK '11.*.*'
        endfamily
        family wv
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ../../../main/00dc/fc/model eq complete)
          task glver
          task glupr
            trigger glver eq complete
          task glcrr
            edit DATEMASK '28.*.*'
          task ff10ver
            trigger glver eq complete
          task ff10upr
            trigger ff10ver eq complete and glupr eq complete
          task glmem
            trigger ff10upr eq complete
            edit DATEMASK '11.*.*'
          task glupm
            trigger glmem eq complete
            edit DATEMASK '11.*.*'
          task glcrm
            edit DATEMASK '28.*.*'
        endfamily
        family sf
          trigger fc eq complete or fc eq aborted
          task vseli1
          task vseli2
          task vseup1
            trigger vseli1 eq complete
          task vseup2
            trigger vseup1 eq complete and vseli2 eq complete
          task vsecrd
            edit DATEMASK '28.*.*'
          task vsecrm
            edit DATEMASK '28.*.*'
          task vsemem
            trigger vseup2 eq complete
            edit DATEMASK '12.*.*'
          task vseupm
            trigger vsemem eq complete
            edit DATEMASK '12.*.*'
        endfamily
      endfamily
    endfamily
    family 06
      edit EMOS_BASE '06'
      edit MXUP_TRAJ '3'
      edit REINISTREAM 'DA'
      edit LAR_4V_EF 'false'
      edit LFC_DIR_FREQ 'false'
      edit WINDOW_LENGTH_4D '6'
      edit WINDOW_OFFSET_4D '3'
      edit EDFAMILY 'true'
      edit PERIOD_4D '6'
      edit FCLENGTH '9'
      edit DELTA_DAY '0'
      family archive
        trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ../../main/12/prod/06an eq complete)
        edit FSFAMILY 'mars'
        edit SMSFILES '/home/ma/emos/def/e_36r1/archive'
        edit EMOS_TYPE 'an'
        inlimit /limits:mars
        inlimit /e_36r1/limits:mars
        task ansfc
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/12/prod/06an/derived_param eq complete))
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task anml
          trigger ansfc:fdb or ansfc eq complete
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          inlimit /e_36r1/limits:ml
          label files ""
          label info ""
          event fdb
        task anpl
          trigger anml:fdb or anml eq complete
          edit FCLENGTH '6'
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task anwave
          trigger ansfc:fdb or ansfc eq complete
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task anil
          trigger anpl:fdb or anpl eq complete
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task derivedp
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/12/prod/06an/derived_param eq complete))
          edit STEPS '0'
          label files ""
          label info ""
          event fdb
        family obs
          edit FSFAMILY 'an'
          inlimit /e_36r1/limits:obs
          task highres
            edit QUEUE 'emos-e'
            edit QUEUE_EPILOG 'emos-exp'
            edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
            edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
            edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
            edit LOGDIR '/vol/emos_nc/output'
            edit SMSOUT '/vol/emos_nc/output'
            inlimit /limits:linux_cluster
          task ai
            trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/06/an/4dvar/ifstraj:finaltraj or /e_36r1/main/06/an/4dvar eq complete))
            edit TYPE 'ai'
            label files ""
            label info ""
            event fdb
          task af
            trigger ../../fb/feedback eq complete
            edit TYPE 'af'
            label files ""
            label info ""
            event fdb
          task fb
            trigger ../../fb/feedback eq complete
            edit TYPE 'fb'
            label files ""
            label info ""
            event fdb
        endfamily
        task ice_nmc
          trigger ansfc:fdb or ansfc eq complete
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task sst
          trigger ansfc:fdb or ansfc eq complete
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        family fc015
          edit STEPS '3/6/9/12/15'
          edit EMOS_TYPE 'fc'
          task fcsfc
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcml
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            inlimit /e_36r1/limits:ml
            label files ""
            label info ""
            event fdb
          task fcpl
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcil
            trigger fcpl:fdb or fcpl eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task derivedp
            trigger fcil:fdb or fcil eq complete
            label files ""
            label info ""
            event fdb
        endfamily
        task fc12g
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/06/fc/model eq complete)
          edit STEPS '0/3/6/9/12/15'
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      family fb
        trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/06/an eq complete)
        edit FSFAMILY 'an'
        edit SMSFILES '/home/ma/emos/def/e_36r1/fb'
        task sync_rstrhbias
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
        task biassave
          trigger sync_rstrhbias eq complete
        family feedback
          edit ACCOUNT 'oeoperan'
          edit NPES '8'
          edit FBNEW2OLD '1'
          edit LCONV '1'
          edit LSATOB '1'
          edit LREO3 '1'
          edit LAEOLUS '0'
          edit LSSMI '1'
          edit LSSMIS '1'
          edit LAMSRE '1'
          edit LTMI '1'
          edit LHIRS '1'
          edit LMSU '1'
          edit LSSU '1'
          edit LAMSUA '1'
          edit LAMSUB '1'
          edit LMHS '1'
          edit LGEOS '1'
          edit LSCAT '1'
          edit LAIRS '1'
          edit LVTPR1 '0'
          edit LVTPR2 '0'
          edit LGPSRO '1'
          edit LIASI '1'
          edit LMERIS '1'
          task revmatchup
            edit NPES '128'
            edit THREADS '2'
          task matchup
            trigger revmatchup eq complete
            edit NPES '128'
            edit THREADS '2'
          task odbcmp_prepare
            edit THREADS '1'
            edit NPES '1'
            edit MEM '1000'
            edit CPUTIME '1000'
            edit PRIORITY '90'
          family odb_compress
            trigger odbcmp_prepare eq complete and matchup eq complete
            edit THREADS '2'
            edit NPES '1'
            task odbcmp_conv
            task odbcmp_surf_conv
            task odbcmp_airs
            task odbcmp_iasi
            task odbcmp_hirs
            task odbcmp_msu
            task odbcmp_ssu
            task odbcmp_amsua
            task odbcmp_amsub
            task odbcmp_mhs
            task odbcmp_satob
            task odbcmp_reo3
            task odbcmp_ssmi
            task odbcmp_ssmis
            task odbcmp_amsre
            task odbcmp_tmi
            task odbcmp_scatt
            task odbcmp_geos
            task odbcmp_gpsro
            task odbcmp_meris
          endfamily
          family odb2bufr
            trigger matchup eq complete
            edit NPES '1'
            edit SUBTYPES_INCLUDE ''
            edit SUBTYPES_EXCLUDE ''
            task o2b_conv
              edit SUBTYPES_EXCLUDE '85 82 83 87'
              edit NPES '2'
            task o2b_surf_conv
              edit SUBTYPES_EXCLUDE '142 143 144 145 85 82 83 87'
              edit NPES '2'
            task o2b_airs
              edit SUBTYPES_INCLUDE '57'
              edit NPES '8'
              edit THREADS '2'
            task o2b_iasi
              edit SUBTYPES_INCLUDE '240'
              edit NPES '8'
              edit THREADS '2'
            task o2b_hirs
              edit SUBTYPES_INCLUDE '54 55'
              edit NPES '2'
            task o2b_msu
              edit SUBTYPES_INCLUDE '54 55'
            task o2b_ssu
              edit SUBTYPES_INCLUDE '54 55'
            task o2b_amsua
              edit SUBTYPES_INCLUDE '54 55'
              edit NPES '2'
            task o2b_amsub
              edit SUBTYPES_INCLUDE '54 55'
              edit NPES '2'
            task o2b_mhs
              edit SUBTYPES_INCLUDE '54 55'
              edit NPES '2'
            task o2b_satob
              edit SUBTYPES_INCLUDE '82 83 87'
              edit NPES '2'
            task o2b_reo3
              edit SUBTYPES_INCLUDE '205 206 207 208 215'
              edit NPES '4'
              edit THREADS '2'
            task o2b_ssmi
              edit SUBTYPES_INCLUDE '55 127'
            task o2b_ssmis
              edit SUBTYPES_INCLUDE '49'
            task o2b_amsre
              edit SUBTYPES_INCLUDE '59'
            task o2b_tmi
              edit SUBTYPES_INCLUDE '129'
            task o2b_scatt
              edit SUBTYPES_INCLUDE '122 137 139'
            task o2b_geos
              edit SUBTYPES_INCLUDE '189 89 76 77 78'
              edit NPES '2'
            task o2b_gpsro
              edit SUBTYPES_INCLUDE '250'
            task o2b_meris
              edit SUBTYPES_INCLUDE '212'
          endfamily
          task mergebufr
            trigger odb2bufr eq complete
        endfamily
        family obstat
          trigger feedback/odb2bufr eq complete
          edit NPES '1'
          task obstat_conv
          task obstat_surf_conv
          task obstat_airs
            edit NPES '8'
            edit THREADS '2'
          task obstat_iasi
            edit NPES '8'
            edit THREADS '2'
          task obstat_hirs
            edit NPES '4'
          task obstat_msu
          task obstat_ssu
          task obstat_amsua
            edit NPES '2'
          task obstat_amsub
          task obstat_mhs
          task obstat_satob
          task obstat_reo3
          task obstat_ssmi
            edit NPES '2'
          task obstat_ssmis
            edit NPES '2'
          task obstat_amsre
            edit NPES '2'
          task obstat_tmi
            edit NPES '2'
          task obstat_scatt
          task obstat_geos
            edit NPES '2'
          task obstat_gpsro
            edit NPES '2'
          task obstat_meris
            edit NPES '2'
        endfamily
        task fdbksave
          trigger feedback eq complete and obstat eq complete
        task clean_odbcmp
          trigger feedback eq complete and obstat eq complete and fdbksave eq complete
        task wavesave
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ../../../main/06/an/4dvar eq complete)
        task obstat_merge
          trigger obstat eq complete
          edit WSHOST 'bee01'
          edit PRIORITY '51'
      endfamily
      family traj
        task feed
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ../../../main/06/fc eq complete)
        task cp_to_other
          trigger feed eq complete
      endfamily
      family satmon
        trigger ../00dc/satmon eq complete
        edit FSFAMILY 'an'
        inlimit /e_36r1/limits:smon
        task getsmon
          trigger ../fb/fdbksave eq complete and ../archive/obs/af eq complete
        task smon_meris
          trigger getsmon eq complete
          edit smon_smsmode 'MERIS'
        task smon_amsre
          trigger getsmon eq complete
          edit smon_smsmode 'AMSRE'
        task smon_ssmis
          trigger getsmon eq complete
          edit smon_smsmode 'SSMIS'
        task smon_tmi
          trigger getsmon eq complete
          edit smon_smsmode 'TMI'
        task smon_gpsro
          trigger getsmon eq complete
          edit smon_smsmode 'GPSRO'
        task smon_pacrars
          trigger getsmon eq complete
          edit smon_smsmode 'PACRARS'
        task smon_ears
          trigger getsmon eq complete
          edit smon_smsmode 'EARS'
        task smon_o3msg
          trigger getsmon eq complete
          edit smon_smsmode 'O3MSG'
        task smon_o3knmi
          trigger getsmon eq complete
          edit smon_smsmode 'O3KNMI'
        task smon_pwc
          trigger getsmon eq complete
          edit smon_smsmode 'PWC'
        task smon_temp
          trigger getsmon eq complete
          edit smon_smsmode 'TEMP'
        task smon_iasi
          trigger getsmon eq complete
          edit smon_smsmode 'IASI'
        task smon_airs
          trigger getsmon eq complete
          edit smon_smsmode 'AIRS'
        task smon_dbamv
          trigger getsmon eq complete
          edit smon_smsmode 'DBAMV'
        task smon_amv
          trigger getsmon eq complete
          edit smon_smsmode 'AMV'
        task smon_o3
          trigger getsmon eq complete
          edit smon_smsmode 'O3'
        task smon_dmsp
          trigger getsmon eq complete
          edit smon_smsmode 'DMSP'
        task smon_noaa
          trigger getsmon eq complete
          edit smon_smsmode 'NOAA'
        task smon_geos
          trigger getsmon eq complete
          edit smon_smsmode 'GEOS'
        task smon_scatt
          trigger getsmon eq complete
          edit smon_smsmode 'SCATT'
        task smon_slmoist
          trigger getsmon eq complete
          edit smon_smsmode 'SLMOIST'
        task smon_clean
          trigger getsmon eq complete and smon_meris eq complete and smon_amsre eq complete and smon_ssmis eq complete and smon_tmi eq complete and smon_gpsro eq complete and smon_pacrars eq complete and smon_ears eq complete and smon_o3msg eq complete and smon_o3knmi eq complete and smon_pwc eq complete and smon_temp eq complete and smon_iasi eq complete and smon_airs eq complete and smon_dbamv eq complete and smon_amv eq complete and smon_o3 eq complete and smon_dmsp eq complete and smon_noaa eq complete and smon_geos eq complete and smon_scatt eq complete and smon_slmoist eq complete
      endfamily
      task clean
        trigger fb eq complete and ../00dc/clean eq complete
        edit FSFAMILY 'an'
      family verif
        edit QUEUE 'emos-e'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        edit SMSFILES '/home/ma/emos/def/e_36r1/verif'
        edit DELTA_DAY '1'
        inlimit /limits:linux_cluster
        family fc
          task v12li1
          task v12li2
          task v12up1
            trigger v12li1 eq complete
          task v12up2
            trigger v12up1 eq complete and v12li2 eq complete
          task v12pl
            trigger v12up2 eq complete
          task v12memv
            trigger v12up2 eq complete
            edit DATEMASK '02.*.*'
          task v12memb
            trigger v12up2 eq complete
            edit DATEMASK '11.*.*'
          task v12upmv
            trigger v12memv eq complete
            edit DATEMASK '02.*.*'
          task v12upmb
            trigger v12memb eq complete
            edit DATEMASK '11.*.*'
          task v12crmv
            edit DATEMASK '28.*.*'
          task v12crmb
            edit DATEMASK '28.*.*'
          task v12crd
            edit DATEMASK '28.*.*'
        endfamily
        family oc
          trigger ./fc eq complete or ./fc eq aborted
          task mk_dummy_fields
            edit QUEUE 'emos'
            edit USER 'emos'
            edit PP_BASE '/pp1'
            edit WSHOST 'pp1'
            edit QUEUE_EPILOG 'emos-exp'
            edit LOGDIR '/pp1/log'
            edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
            edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
            edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
            edit SMSOUT '/vol/emos_nc/output'
          task insert_fields
            trigger mk_dummy_fields eq complete
            edit QUEUE 'emos-e'
            edit QUEUE_EPILOG 'emos-exp'
            edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
            edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
            edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
            edit LOGDIR '/vol/emos_nc/output'
            edit SMSOUT '/vol/emos_nc/output'
            edit PP_BASE '/pp1'
            inlimit /limits:linux_cluster
          task vbrli1
            trigger insert_fields eq complete
          task vbrli2
            trigger vbrli1 eq complete
          task vwali1
            trigger insert_fields eq complete
          task vwali2
            trigger vwali1 eq complete
          task vwali3
            trigger vwali2 eq complete
          task vofli
            trigger insert_fields eq complete
          task vtlli
            trigger insert_fields eq complete
          task vtoli
            trigger insert_fields eq complete
          task vmoli
            trigger insert_fields eq complete
          task vbrup1
            trigger vbrli1 eq complete
          task vbrup2
            trigger vbrup1 eq complete and vbrli2 eq complete
          task vwaup1
            trigger vwali1 eq complete
          task vwaup2
            trigger vwaup1 eq complete and vwali2 eq complete
          task vwaup3
            trigger vwaup2 eq complete and vwali3 eq complete
          task vofup
            trigger vofli eq complete
          task vtlup
            trigger vtlli eq complete
          task vtoup
            trigger vtoli eq complete
          task vmoup
            trigger vmoli eq complete
          task vmeli
            trigger insert_fields eq complete
          task vmeup
            trigger vmeli eq complete
          task voccrd
            edit DATEMASK '28.*.*'
          task voccrm
            edit DATEMASK '28.*.*'
          task vocmem
            edit DATEMASK '11.*.*'
          task vocupm
            trigger vocmem eq complete
            edit DATEMASK '11.*.*'
          task cleanup
            trigger vmeup eq complete and vbrup2 eq complete and vwaup3 eq complete and vofup eq complete and vtlup eq complete and vtoup eq complete and vmoup eq complete and voccrd eq complete and voccrm eq complete and vocupm eq complete
            edit QUEUE 'emos-e'
            edit QUEUE_EPILOG 'emos-exp'
            edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
            edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
            edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
            edit LOGDIR '/vol/emos_nc/output'
            edit SMSOUT '/vol/emos_nc/output'
            inlimit /limits:linux_cluster
        endfamily
        family cf
          trigger oc eq complete or oc eq aborted
          task vcnli1
            late -c +02:00
          task vcnli2
          task vcnup1
            trigger vcnli1 eq complete
          task vcnup2
            trigger vcnup1 eq complete and vcnli2 eq complete
          task vcnpl
            trigger vcnup2 eq complete
          task vcncrd
          task vcncrm
            edit DATEMASK '28.*.*'
          task vcnmem
            trigger vcnup2 eq complete
            edit DATEMASK '11.*.*'
          task vcnupm
            trigger vcnmem eq complete
            edit DATEMASK '11.*.*'
          task vcnli3
          task vcnli4
          task vcnup3
            trigger vcnup2 eq complete and vcnli3 eq complete
          task vcnup4
            trigger vcnup3 eq complete and vcnli4 eq complete
          task vcnmem2
            trigger vcnup4 eq complete
            edit DATEMASK '22.*.*'
          task vcnupm2
            trigger vcnmem2 eq complete
            edit DATEMASK '22.*.*'
        endfamily
        family ob
          trigger oc eq complete or oc eq aborted
          time 05:00
          family stage1
            task vobli1
            task vobli2
            task vobup1
              trigger vobli1 eq complete
            task vobup2
              trigger vobup1 eq complete and vobli2 eq complete
            task vobnobs
              trigger vobup2 eq complete
              edit DATEMASK '03.*.*'
            task vobmem
              trigger ./vobup2 eq complete
              edit DATEMASK '03.*.*'
            task vobupm
              trigger vobmem eq complete
              edit DATEMASK '03.*.*'
          endfamily
          family stage2
            trigger ./stage1 eq complete
            edit EMOS_BASE '00'
            task vobli1
            task vobli2
            task vobup1
              trigger vobli1 eq complete
            task vobup2
              trigger vobup1 eq complete and vobli2 eq complete
            task vobnobs
              trigger vobup2 eq complete
              edit DATEMASK '03.*.*'
            task vobmem
              trigger ./vobup2 eq complete
              edit DATEMASK '03.*.*'
            task vobupm
              trigger vobmem eq complete
              edit DATEMASK '03.*.*'
          endfamily
          task vobcrd
            edit DATEMASK '28.*.*'
          task vobcrm
            edit DATEMASK '28.*.*'
        endfamily
        family em
          task vemli
            late -c +02:00
          task vemup
            trigger vemli eq complete
          task vemcrd
            edit DATEMASK '28.*.*'
          task vemcrm
            edit DATEMASK '28.*.*'
          task vemmem
            trigger vemup eq complete
            edit DATEMASK '11.*.*'
          task vemupm
            trigger vemmem eq complete
            edit DATEMASK '11.*.*'
        endfamily
        family es
          trigger em eq complete
          task vesli
            late -c +02:00
          task vesup
            trigger vesli eq complete
          task vescrd
            edit DATEMASK '28.*.*'
          task vescrm
            edit DATEMASK '28.*.*'
          task vesmem
            trigger vesup eq complete
            edit DATEMASK '11.*.*'
          task vesupm
            trigger vesmem eq complete
            edit DATEMASK '11.*.*'
        endfamily
        family cv
          trigger es eq complete
          task vcvli
            late -c +02:00
          task vcvup
            trigger vcvli eq complete
          task vcvcrd
            edit DATEMASK '28.*.*'
          task vcvcrm
            edit DATEMASK '28.*.*'
          task vcvmem
            trigger vcvup eq complete
            edit DATEMASK '11.*.*'
          task vcvupm
            trigger vcvmem eq complete
            edit DATEMASK '11.*.*'
        endfamily
        family wv
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ../../../main/06/fc/model eq complete)
          task glver
          task glupr
            trigger glver eq complete
          task glcrr
            edit DATEMASK '28.*.*'
          task ff10ver
            trigger glver eq complete
          task ff10upr
            trigger ff10ver eq complete and glupr eq complete
          task glmem
            trigger ff10upr eq complete
            edit DATEMASK '11.*.*'
          task glupm
            trigger glmem eq complete
            edit DATEMASK '11.*.*'
          task glcrm
            edit DATEMASK '28.*.*'
        endfamily
        family sf
          trigger fc eq complete or fc eq aborted
          task vseli1
          task vseli2
          task vseup1
            trigger vseli1 eq complete
          task vseup2
            trigger vseup1 eq complete and vseli2 eq complete
          task vsecrd
            edit DATEMASK '28.*.*'
          task vsecrm
            edit DATEMASK '28.*.*'
          task vsemem
            trigger vseup2 eq complete
            edit DATEMASK '12.*.*'
          task vseupm
            trigger vsemem eq complete
            edit DATEMASK '12.*.*'
        endfamily
      endfamily
    endfamily
    family 12
      edit EMOS_BASE '12'
      edit MXUP_TRAJ '3'
      edit REINISTREAM 'DCDA'
      edit WINDOW_LENGTH_4D '6'
      edit WINDOW_OFFSET_4D '3'
      edit REINITIALIZE 'true'
      edit EDFAMILY 'true'
      edit PERIOD_4D '6'
      edit FCLENGTH '9'
      edit DELTA_DAY '0'
      family ddh
        trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/12/fc eq complete)
        edit SMSFILES '/home/ma/emos/def/e_36r1/fc'
        task ddh
      endfamily
      family sync
        trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/12/an eq complete)
        edit EMOS_TYPE 'an'
        edit FSFAMILY 'an'
        edit SMSFILES '/home/ma/emos/def/e_36r1/sync'
        task sync_bias
          edit CALC_PSBIAS 'false'
        task sync_fc
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/12/fc/model:step gt 6 or /e_36r1/main/12/fc/model eq complete))
          edit QUEUE 'ns'
          edit QUEUE_EPILOG 'ns'
          edit SCHOST 'c1a'
          edit STHOST '/s1a'
          edit SCHOST_BKUP 'c1b'
          edit STHOST_BKUP '/s1b'
          edit SMSLOGHOST 'c1a'
          edit LOGDIR '/s1a/emos_esuite/emos_data/0046/log'
          edit SMSOUT '/s1a/emos_esuite/emos_data/0046/log'
          inlimit /limits:c1a
          label files ""
          label info ""
          event fdb
        task sync_an
          edit QUEUE 'ns'
          edit QUEUE_EPILOG 'ns'
          edit SCHOST 'c1a'
          edit STHOST '/s1a'
          edit SCHOST_BKUP 'c1b'
          edit STHOST_BKUP '/s1b'
          edit SMSLOGHOST 'c1a'
          edit LOGDIR '/s1a/emos_esuite/emos_data/0046/log'
          edit SMSOUT '/s1a/emos_esuite/emos_data/0046/log'
          inlimit /limits:c1a
          label files ""
          label info ""
          event fdb
        task sync_wv
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/12/fc/model:step gt 3 or /e_36r1/main/12/fc/model eq complete))
          edit QUEUE 'ns'
          edit QUEUE_EPILOG 'ns'
          edit SCHOST 'c1a'
          edit STHOST '/s1a'
          edit SCHOST_BKUP 'c1b'
          edit STHOST_BKUP '/s1b'
          edit SMSLOGHOST 'c1a'
          edit LOGDIR '/s1a/emos_esuite/emos_data/0046/log'
          edit SMSOUT '/s1a/emos_esuite/emos_data/0046/log'
          inlimit /limits:c1a
          label files ""
          label info ""
          event fdb
        task sync_clean
          edit QUEUE 'ns'
          edit QUEUE_EPILOG 'ns'
          edit SCHOST 'c1a'
          edit STHOST '/s1a'
          edit SCHOST_BKUP 'c1b'
          edit STHOST_BKUP '/s1b'
          edit SMSLOGHOST 'c1a'
          edit LOGDIR '/s1a/emos_esuite/emos_data/0046/log'
          edit SMSOUT '/s1a/emos_esuite/emos_data/0046/log'
          inlimit /limits:c1a
      endfamily
      family archive
        trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ../../main/12/an eq complete and ( ../../main/12/fc/model:step ge 9 or ../../main/12/fc/model eq complete))
        edit FSFAMILY 'mars'
        edit SMSFILES '/home/ma/emos/def/e_36r1/archive'
        edit EMOS_TYPE 'an'
        inlimit /limits:mars
        inlimit /e_36r1/limits:mars
        task ansfc
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/12/prod/000 eq complete and /e_36r1/main/12/prod/12an eq complete)
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task anml
          trigger ansfc:fdb or ansfc eq complete
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          inlimit /e_36r1/limits:ml
          label files ""
          label info ""
          event fdb
        task anpl
          trigger anml:fdb or anml eq complete
          edit FCLENGTH '240'
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task anwave
          trigger ansfc:fdb or ansfc eq complete
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task anil
          trigger anpl:fdb or anpl eq complete
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task derivedp
          trigger ./ansfc eq complete
          edit STEPS '0'
          label files ""
          label info ""
          event fdb
        family obs
          edit FSFAMILY 'an'
          inlimit /e_36r1/limits:obs
          task ai
            trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/12/an/4dvar/ifstraj:finaltraj or /e_36r1/main/12/an/4dvar eq complete))
            edit TYPE 'ai'
            label files ""
            label info ""
            event fdb
          task af
            trigger ../../fb/feedback eq complete
            edit TYPE 'af'
            label files ""
            label info ""
            event fdb
          task fb
            trigger ../../fb/feedback eq complete
            edit TYPE 'fb'
            label files ""
            label info ""
            event fdb
        endfamily
        task ice_nmc
          trigger ansfc:fdb or ansfc eq complete
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task sst
          trigger ansfc:fdb or ansfc eq complete
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        family fc024
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/12/prod/024 eq complete)
          edit STEPS '0/3/006/009/012/015/018/021/024'
          edit EMOS_TYPE 'fc'
          task fcsfc
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcml
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            inlimit /e_36r1/limits:ml
            label files ""
            label info ""
            event fdb
          task fcpl
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcil
            trigger fcpl:fdb or fcpl eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task derivedp
            trigger fcil:fdb or fcil eq complete
            label files ""
            label info ""
            event fdb
        endfamily
        family fc048
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/12/prod/048 eq complete)
          edit STEPS '027/030/033/036/039/042/045/048'
          edit EMOS_TYPE 'fc'
          task fcsfc
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcml
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            inlimit /e_36r1/limits:ml
            label files ""
            label info ""
            event fdb
          task fcpl
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcil
            trigger fcpl:fdb or fcpl eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task derivedp
            trigger fcil:fdb or fcil eq complete
            label files ""
            label info ""
            event fdb
        endfamily
        family fc072
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/12/prod/072 eq complete)
          edit STEPS '051/054/057/060/063/066/069/072'
          edit EMOS_TYPE 'fc'
          task fcsfc
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcml
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            inlimit /e_36r1/limits:ml
            label files ""
            label info ""
            event fdb
          task fcpl
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcil
            trigger fcpl:fdb or fcpl eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task derivedp
            trigger fcil:fdb or fcil eq complete
            label files ""
            label info ""
            event fdb
        endfamily
        family fc096
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/12/prod/096 eq complete)
          edit STEPS '075/078/081/084/087/090/093/096'
          edit EMOS_TYPE 'fc'
          task fcsfc
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcml
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            inlimit /e_36r1/limits:ml
            label files ""
            label info ""
            event fdb
          task fcpl
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcil
            trigger fcpl:fdb or fcpl eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task derivedp
            trigger fcil:fdb or fcil eq complete
            label files ""
            label info ""
            event fdb
        endfamily
        family fc120
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/12/prod/120 eq complete)
          edit STEPS '099/102/105/108/111/114/117/120'
          edit EMOS_TYPE 'fc'
          task fcsfc
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcml
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            inlimit /e_36r1/limits:ml
            label files ""
            label info ""
            event fdb
          task fcpl
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcil
            trigger fcpl:fdb or fcpl eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task derivedp
            trigger fcil:fdb or fcil eq complete
            label files ""
            label info ""
            event fdb
        endfamily
        family fc144
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/12/prod/144 eq complete)
          edit STEPS '123/126/129/132/135/138/141/144'
          edit EMOS_TYPE 'fc'
          task fcsfc
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcml
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            inlimit /e_36r1/limits:ml
            label files ""
            label info ""
            event fdb
          task fcpl
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcil
            trigger fcpl:fdb or fcpl eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task derivedp
            trigger fcil:fdb or fcil eq complete
            label files ""
            label info ""
            event fdb
        endfamily
        family fc192
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/12/prod/192 eq complete)
          edit STEPS '150/156/162/168/174/180/186/192'
          edit EMOS_TYPE 'fc'
          task fcsfc
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcml
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            inlimit /e_36r1/limits:ml
            label files ""
            label info ""
            event fdb
          task fcpl
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcil
            trigger fcpl:fdb or fcpl eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task derivedp
            trigger fcil:fdb or fcil eq complete
            label files ""
            label info ""
            event fdb
        endfamily
        family fc240
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/12/prod/240 eq complete)
          edit STEPS '198/204/210/216/222/228/234/240'
          edit EMOS_TYPE 'fc'
          task fcsfc
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcml
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            inlimit /e_36r1/limits:ml
            label files ""
            label info ""
            event fdb
          task fcpl
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcil
            trigger fcpl:fdb or fcpl eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task derivedp
            trigger fcil:fdb or fcil eq complete
            label files ""
            label info ""
            event fdb
        endfamily
        family satim1
          task satim
            trigger ../../satim eq complete
            edit STEPS '0/to/48/by/3'
            label files ""
            label info ""
            event fdb
        endfamily
        family satim2
          task satim
            trigger ../../satim eq complete
            edit STEPS '54/to/120/by/6'
            label files ""
            label info ""
            event fdb
        endfamily
        task fc12g
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/12/fc/model eq complete)
          edit STEPS_SPECTRAL '6/12/18/24/30/36/42/48/54/60/66/72/78/84/90/96/102/108/114/120'
          edit STEPS '0/3/6/9/12/15/18/21/24/27/30/33/36/39/42/45/48/51/54/57/60/63/66/69/72/75/78/81/84/87/90/93/96/99/102/105/108/111/114/117/120/123/126/129/132/135/138/141/144/150/156/162/168/174/180/186/192/198/204/210/216/222/228/234/240'
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task rmfdb
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/12/obs eq complete)
        task flush
          edit STREAM 'oper'
          label files ""
          label info ""
          event fdb
      endfamily
      family fb
        trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/12/an eq complete)
        edit FSFAMILY 'an'
        edit SMSFILES '/home/ma/emos/def/e_36r1/fb'
        task sync_rstrhbias
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
        task biassave
          trigger sync_rstrhbias eq complete
        family feedback
          edit ACCOUNT 'oeoperan'
          edit NPES '8'
          edit FBNEW2OLD '1'
          edit LCONV '1'
          edit LSATOB '1'
          edit LREO3 '1'
          edit LAEOLUS '0'
          edit LSSMI '1'
          edit LSSMIS '1'
          edit LAMSRE '1'
          edit LTMI '1'
          edit LHIRS '1'
          edit LMSU '1'
          edit LSSU '1'
          edit LAMSUA '1'
          edit LAMSUB '1'
          edit LMHS '1'
          edit LGEOS '1'
          edit LSCAT '1'
          edit LAIRS '1'
          edit LVTPR1 '0'
          edit LVTPR2 '0'
          edit LGPSRO '1'
          edit LIASI '1'
          edit LMERIS '1'
          task revmatchup
            edit NPES '128'
            edit THREADS '2'
          task matchup
            trigger revmatchup eq complete
            edit NPES '128'
            edit THREADS '2'
          task odbcmp_prepare
            edit THREADS '1'
            edit NPES '1'
            edit MEM '1000'
            edit CPUTIME '1000'
            edit PRIORITY '90'
          family odb_compress
            trigger odbcmp_prepare eq complete and matchup eq complete
            edit THREADS '2'
            edit NPES '1'
            task odbcmp_conv
            task odbcmp_surf_conv
            task odbcmp_airs
            task odbcmp_iasi
            task odbcmp_hirs
            task odbcmp_msu
            task odbcmp_ssu
            task odbcmp_amsua
            task odbcmp_amsub
            task odbcmp_mhs
            task odbcmp_satob
            task odbcmp_reo3
            task odbcmp_ssmi
            task odbcmp_ssmis
            task odbcmp_amsre
            task odbcmp_tmi
            task odbcmp_scatt
            task odbcmp_geos
            task odbcmp_gpsro
            task odbcmp_meris
          endfamily
          family odb2bufr
            trigger matchup eq complete
            edit NPES '1'
            edit SUBTYPES_INCLUDE ''
            edit SUBTYPES_EXCLUDE ''
            task o2b_conv
              edit SUBTYPES_EXCLUDE '85 82 83 87'
              edit NPES '2'
            task o2b_surf_conv
              edit SUBTYPES_EXCLUDE '142 143 144 145 85 82 83 87'
              edit NPES '2'
            task o2b_airs
              edit SUBTYPES_INCLUDE '57'
              edit NPES '8'
              edit THREADS '2'
            task o2b_iasi
              edit SUBTYPES_INCLUDE '240'
              edit NPES '8'
              edit THREADS '2'
            task o2b_hirs
              edit SUBTYPES_INCLUDE '54 55'
              edit NPES '2'
            task o2b_msu
              edit SUBTYPES_INCLUDE '54 55'
            task o2b_ssu
              edit SUBTYPES_INCLUDE '54 55'
            task o2b_amsua
              edit SUBTYPES_INCLUDE '54 55'
              edit NPES '2'
            task o2b_amsub
              edit SUBTYPES_INCLUDE '54 55'
              edit NPES '2'
            task o2b_mhs
              edit SUBTYPES_INCLUDE '54 55'
              edit NPES '2'
            task o2b_satob
              edit SUBTYPES_INCLUDE '82 83 87'
              edit NPES '2'
            task o2b_reo3
              edit SUBTYPES_INCLUDE '205 206 207 208 215'
              edit NPES '4'
              edit THREADS '2'
            task o2b_ssmi
              edit SUBTYPES_INCLUDE '55 127'
            task o2b_ssmis
              edit SUBTYPES_INCLUDE '49'
            task o2b_amsre
              edit SUBTYPES_INCLUDE '59'
            task o2b_tmi
              edit SUBTYPES_INCLUDE '129'
            task o2b_scatt
              edit SUBTYPES_INCLUDE '122 137 139'
            task o2b_geos
              edit SUBTYPES_INCLUDE '189 89 76 77 78'
              edit NPES '2'
            task o2b_gpsro
              edit SUBTYPES_INCLUDE '250'
            task o2b_meris
              edit SUBTYPES_INCLUDE '212'
          endfamily
          task mergebufr
            trigger odb2bufr eq complete
        endfamily
        family obstat
          trigger feedback/odb2bufr eq complete
          edit NPES '1'
          task obstat_conv
          task obstat_surf_conv
          task obstat_airs
            edit NPES '8'
            edit THREADS '2'
          task obstat_iasi
            edit NPES '8'
            edit THREADS '2'
          task obstat_hirs
            edit NPES '4'
          task obstat_msu
          task obstat_ssu
          task obstat_amsua
            edit NPES '2'
          task obstat_amsub
          task obstat_mhs
          task obstat_satob
          task obstat_reo3
          task obstat_ssmi
            edit NPES '2'
          task obstat_ssmis
            edit NPES '2'
          task obstat_amsre
            edit NPES '2'
          task obstat_tmi
            edit NPES '2'
          task obstat_scatt
          task obstat_geos
            edit NPES '2'
          task obstat_gpsro
            edit NPES '2'
          task obstat_meris
            edit NPES '2'
        endfamily
        task fdbksave
          trigger feedback eq complete and obstat eq complete
        task clean_odbcmp
          trigger feedback eq complete and obstat eq complete and fdbksave eq complete
        task wavesave
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ../../../main/12/an/4dvar eq complete)
        task obstat_merge
          trigger obstat eq complete
          edit WSHOST 'bee01'
          edit PRIORITY '51'
      endfamily
      family traj
        task feed
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ../../../main/12/fc eq complete)
        task cp_to_other
          trigger feed eq complete
      endfamily
      family satim
        edit QUEUE 'ns'
        edit QUEUE_EPILOG 'ns'
        edit SMSCMD 'smssubmit %USER% %SCHOST% %SMSJOB% %SMSJOBOUT%'
        edit SMSKILL 'smskill %USER% %SCHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %SCHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/s1b/emos_esuite/emos_data/0046/log'
        edit SMSOUT '/s1b/emos_esuite/emos_data/0046/log'
        edit MEM '11000'
        edit SMSFILES '/home/ma/emos/def/e_36r1/plot'
        edit SMSURLBASE 'http://intra.ecmwf.int/publications/cms/get/metapps'
        inlimit /limits:c1b
        task display_seaice
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/12/an eq complete)
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit SMSFILES '/home/ma/emos/def/e_36r1/admin'
          edit FSFAMILY 'undef'
          edit METVIEW 'metview_emos'
          edit SMSURL 'seaice'
          inlimit /limits:linux_cluster
        task display_sst
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit FSFAMILY 'undef'
          edit SMSURL 'sst'
          inlimit /limits:linux_cluster
        family satim_0
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/12/prod eq complete))
          edit IMAGETIME '0'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_3
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/12/prod eq complete))
          edit IMAGETIME '3'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_6
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/12/prod eq complete))
          edit IMAGETIME '6'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_9
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/12/prod eq complete))
          edit IMAGETIME '9'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_12
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/12/prod eq complete))
          edit IMAGETIME '12'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_15
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/12/prod eq complete))
          edit IMAGETIME '15'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_18
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/12/prod eq complete))
          edit IMAGETIME '18'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_21
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/12/prod eq complete))
          edit IMAGETIME '21'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_24
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/12/prod eq complete))
          edit IMAGETIME '24'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_27
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/12/prod eq complete))
          edit IMAGETIME '27'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_30
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/12/prod eq complete))
          edit IMAGETIME '30'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_33
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/12/prod eq complete))
          edit IMAGETIME '33'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_36
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/12/prod eq complete))
          edit IMAGETIME '36'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_39
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/12/prod eq complete))
          edit IMAGETIME '39'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_42
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/12/prod eq complete))
          edit IMAGETIME '42'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_45
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/12/prod eq complete))
          edit IMAGETIME '45'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_48
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/12/prod eq complete))
          edit IMAGETIME '48'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_54
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/12/prod eq complete))
          edit IMAGETIME '54'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_60
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/12/prod eq complete))
          edit IMAGETIME '60'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_66
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/12/prod eq complete))
          edit IMAGETIME '66'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_72
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/12/prod eq complete))
          edit IMAGETIME '72'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_78
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/12/prod eq complete))
          edit IMAGETIME '78'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_84
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/12/prod eq complete))
          edit IMAGETIME '84'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_90
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/12/prod eq complete))
          edit IMAGETIME '90'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_96
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/12/prod eq complete))
          edit IMAGETIME '96'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_102
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/12/prod eq complete))
          edit IMAGETIME '102'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_108
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/12/prod eq complete))
          edit IMAGETIME '108'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_114
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/12/prod eq complete))
          edit IMAGETIME '114'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_120
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/12/prod eq complete))
          edit IMAGETIME '120'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
      endfamily
      family satmon
        trigger ../00dc/satmon eq complete
        edit FSFAMILY 'an'
        inlimit /e_36r1/limits:smon
        task getsmon
          trigger ../fb/fdbksave eq complete and ../archive/obs/af eq complete
        task smon_meris
          trigger getsmon eq complete
          edit smon_smsmode 'MERIS'
        task smon_amsre
          trigger getsmon eq complete
          edit smon_smsmode 'AMSRE'
        task smon_ssmis
          trigger getsmon eq complete
          edit smon_smsmode 'SSMIS'
        task smon_tmi
          trigger getsmon eq complete
          edit smon_smsmode 'TMI'
        task smon_gpsro
          trigger getsmon eq complete
          edit smon_smsmode 'GPSRO'
        task smon_pacrars
          trigger getsmon eq complete
          edit smon_smsmode 'PACRARS'
        task smon_ears
          trigger getsmon eq complete
          edit smon_smsmode 'EARS'
        task smon_o3msg
          trigger getsmon eq complete
          edit smon_smsmode 'O3MSG'
        task smon_o3knmi
          trigger getsmon eq complete
          edit smon_smsmode 'O3KNMI'
        task smon_pwc
          trigger getsmon eq complete
          edit smon_smsmode 'PWC'
        task smon_temp
          trigger getsmon eq complete
          edit smon_smsmode 'TEMP'
        task smon_iasi
          trigger getsmon eq complete
          edit smon_smsmode 'IASI'
        task smon_airs
          trigger getsmon eq complete
          edit smon_smsmode 'AIRS'
        task smon_dbamv
          trigger getsmon eq complete
          edit smon_smsmode 'DBAMV'
        task smon_amv
          trigger getsmon eq complete
          edit smon_smsmode 'AMV'
        task smon_o3
          trigger getsmon eq complete
          edit smon_smsmode 'O3'
        task smon_dmsp
          trigger getsmon eq complete
          edit smon_smsmode 'DMSP'
        task smon_noaa
          trigger getsmon eq complete
          edit smon_smsmode 'NOAA'
        task smon_geos
          trigger getsmon eq complete
          edit smon_smsmode 'GEOS'
        task smon_scatt
          trigger getsmon eq complete
          edit smon_smsmode 'SCATT'
        task smon_slmoist
          trigger getsmon eq complete
          edit smon_smsmode 'SLMOIST'
        task smon_clean
          trigger getsmon eq complete and smon_meris eq complete and smon_amsre eq complete and smon_ssmis eq complete and smon_tmi eq complete and smon_gpsro eq complete and smon_pacrars eq complete and smon_ears eq complete and smon_o3msg eq complete and smon_o3knmi eq complete and smon_pwc eq complete and smon_temp eq complete and smon_iasi eq complete and smon_airs eq complete and smon_dbamv eq complete and smon_amv eq complete and smon_o3 eq complete and smon_dmsp eq complete and smon_noaa eq complete and smon_geos eq complete and smon_scatt eq complete and smon_slmoist eq complete
      endfamily
      task clean
        trigger fb eq complete and ../06/clean eq complete
        edit FSFAMILY 'an'
      family verif
        edit QUEUE 'emos-e'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        edit SMSFILES '/home/ma/emos/def/e_36r1/verif'
        edit DELTA_DAY '1'
        inlimit /limits:linux_cluster
        time 02:00
        family fc
          task v12li1
          task v12li2
          task v12up1
            trigger v12li1 eq complete
          task v12up2
            trigger v12up1 eq complete and v12li2 eq complete
          task v12pl
            trigger v12up2 eq complete
          task v12memv
            trigger v12up2 eq complete
            edit DATEMASK '02.*.*'
          task v12memb
            trigger v12up2 eq complete
            edit DATEMASK '11.*.*'
          task v12upmv
            trigger v12memv eq complete
            edit DATEMASK '02.*.*'
          task v12upmb
            trigger v12memb eq complete
            edit DATEMASK '11.*.*'
          task v12crmv
            edit DATEMASK '28.*.*'
          task v12crmb
            edit DATEMASK '28.*.*'
          task v12crd
            edit DATEMASK '28.*.*'
        endfamily
        family oc
          trigger ./fc eq complete or ./fc eq aborted
          task mk_dummy_fields
            edit QUEUE 'emos'
            edit USER 'emos'
            edit PP_BASE '/pp1'
            edit WSHOST 'pp1'
            edit QUEUE_EPILOG 'emos-exp'
            edit LOGDIR '/pp1/log'
            edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
            edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
            edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
            edit SMSOUT '/vol/emos_nc/output'
          task insert_fields
            trigger mk_dummy_fields eq complete
            edit QUEUE 'emos-e'
            edit QUEUE_EPILOG 'emos-exp'
            edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
            edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
            edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
            edit LOGDIR '/vol/emos_nc/output'
            edit SMSOUT '/vol/emos_nc/output'
            edit PP_BASE '/pp1'
            inlimit /limits:linux_cluster
          task vbrli1
            trigger insert_fields eq complete
          task vbrli2
            trigger vbrli1 eq complete
          task vwali1
            trigger insert_fields eq complete
          task vwali2
            trigger vwali1 eq complete
          task vwali3
            trigger vwali2 eq complete
          task vofli
            trigger insert_fields eq complete
          task vtlli
            trigger insert_fields eq complete
          task vtoli
            trigger insert_fields eq complete
          task vmoli
            trigger insert_fields eq complete
          task vbrup1
            trigger vbrli1 eq complete
          task vbrup2
            trigger vbrup1 eq complete and vbrli2 eq complete
          task vwaup1
            trigger vwali1 eq complete
          task vwaup2
            trigger vwaup1 eq complete and vwali2 eq complete
          task vwaup3
            trigger vwaup2 eq complete and vwali3 eq complete
          task vofup
            trigger vofli eq complete
          task vtlup
            trigger vtlli eq complete
          task vtoup
            trigger vtoli eq complete
          task vmoup
            trigger vmoli eq complete
          task vmeli
            trigger insert_fields eq complete
          task vmeup
            trigger vmeli eq complete
          task voccrd
            edit DATEMASK '28.*.*'
          task voccrm
            edit DATEMASK '28.*.*'
          task vocmem
            edit DATEMASK '11.*.*'
          task vocupm
            trigger vocmem eq complete
            edit DATEMASK '11.*.*'
          task cleanup
            trigger vmeup eq complete and vbrup2 eq complete and vwaup3 eq complete and vofup eq complete and vtlup eq complete and vtoup eq complete and vmoup eq complete and voccrd eq complete and voccrm eq complete and vocupm eq complete
            edit QUEUE 'emos-e'
            edit QUEUE_EPILOG 'emos-exp'
            edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
            edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
            edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
            edit LOGDIR '/vol/emos_nc/output'
            edit SMSOUT '/vol/emos_nc/output'
            inlimit /limits:linux_cluster
        endfamily
        family cf
          trigger oc eq complete or oc eq aborted
          task vcnli1
            late -c +02:00
          task vcnli2
          task vcnup1
            trigger vcnli1 eq complete
          task vcnup2
            trigger vcnup1 eq complete and vcnli2 eq complete
          task vcnpl
            trigger vcnup2 eq complete
          task vcncrd
          task vcncrm
            edit DATEMASK '28.*.*'
          task vcnmem
            trigger vcnup2 eq complete
            edit DATEMASK '11.*.*'
          task vcnupm
            trigger vcnmem eq complete
            edit DATEMASK '11.*.*'
          task vcnli3
          task vcnli4
          task vcnup3
            trigger vcnup2 eq complete and vcnli3 eq complete
          task vcnup4
            trigger vcnup3 eq complete and vcnli4 eq complete
          task vcnmem2
            trigger vcnup4 eq complete
            edit DATEMASK '22.*.*'
          task vcnupm2
            trigger vcnmem2 eq complete
            edit DATEMASK '22.*.*'
        endfamily
        family ob
          trigger oc eq complete or oc eq aborted
          time 05:00
          family stage1
            task vobli1
            task vobli2
            task vobup1
              trigger vobli1 eq complete
            task vobup2
              trigger vobup1 eq complete and vobli2 eq complete
            task vobnobs
              trigger vobup2 eq complete
              edit DATEMASK '03.*.*'
            task vobmem
              trigger ./vobup2 eq complete
              edit DATEMASK '03.*.*'
            task vobupm
              trigger vobmem eq complete
              edit DATEMASK '03.*.*'
          endfamily
          family stage2
            trigger ./stage1 eq complete
            edit EMOS_BASE '00'
            task vobli1
            task vobli2
            task vobup1
              trigger vobli1 eq complete
            task vobup2
              trigger vobup1 eq complete and vobli2 eq complete
            task vobnobs
              trigger vobup2 eq complete
              edit DATEMASK '03.*.*'
            task vobmem
              trigger ./vobup2 eq complete
              edit DATEMASK '03.*.*'
            task vobupm
              trigger vobmem eq complete
              edit DATEMASK '03.*.*'
          endfamily
          task vobcrd
            edit DATEMASK '28.*.*'
          task vobcrm
            edit DATEMASK '28.*.*'
        endfamily
        family em
          task vemli
            late -c +02:00
          task vemup
            trigger vemli eq complete
          task vemcrd
            edit DATEMASK '28.*.*'
          task vemcrm
            edit DATEMASK '28.*.*'
          task vemmem
            trigger vemup eq complete
            edit DATEMASK '11.*.*'
          task vemupm
            trigger vemmem eq complete
            edit DATEMASK '11.*.*'
        endfamily
        family es
          trigger em eq complete
          task vesli
            late -c +02:00
          task vesup
            trigger vesli eq complete
          task vescrd
            edit DATEMASK '28.*.*'
          task vescrm
            edit DATEMASK '28.*.*'
          task vesmem
            trigger vesup eq complete
            edit DATEMASK '11.*.*'
          task vesupm
            trigger vesmem eq complete
            edit DATEMASK '11.*.*'
        endfamily
        family cv
          trigger es eq complete
          task vcvli
            late -c +02:00
          task vcvup
            trigger vcvli eq complete
          task vcvcrd
            edit DATEMASK '28.*.*'
          task vcvcrm
            edit DATEMASK '28.*.*'
          task vcvmem
            trigger vcvup eq complete
            edit DATEMASK '11.*.*'
          task vcvupm
            trigger vcvmem eq complete
            edit DATEMASK '11.*.*'
        endfamily
        family wv
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ../../../main/12/fc/model eq complete)
          task glver
          task glupr
            trigger glver eq complete
          task glcrr
            edit DATEMASK '28.*.*'
          task ff10ver
            trigger glver eq complete
          task ff10upr
            trigger ff10ver eq complete and glupr eq complete
          task glmem
            trigger ff10upr eq complete
            edit DATEMASK '11.*.*'
          task glupm
            trigger glmem eq complete
            edit DATEMASK '11.*.*'
          task glcrm
            edit DATEMASK '28.*.*'
        endfamily
        family sf
          trigger fc eq complete or fc eq aborted
          task vseli1
          task vseli2
          task vseup1
            trigger vseli1 eq complete
          task vseup2
            trigger vseup1 eq complete and vseli2 eq complete
          task vsecrd
            edit DATEMASK '28.*.*'
          task vsecrm
            edit DATEMASK '28.*.*'
          task vsemem
            trigger vseup2 eq complete
            edit DATEMASK '12.*.*'
          task vseupm
            trigger vsemem eq complete
            edit DATEMASK '12.*.*'
        endfamily
      endfamily
    endfamily
    family 12dc
      edit EMOS_BASE '12'
      edit MXUP_TRAJ '3'
      edit STREAM 'DCDA'
      edit WINDOW_LENGTH_4D '12'
      edit WINDOW_OFFSET_4D '3'
      edit EDFAMILY 'false'
      edit PERIOD_4D '12'
      edit FCLENGTH '15'
      edit DELTA_DAY '0'
      family ddh
        trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/12dc/fc eq complete)
        edit EMOS_BASE '18'
        edit SMSFILES '/home/ma/emos/def/e_36r1/fc'
        task ddh
      endfamily
      family sync
        trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/12dc/an eq complete)
        edit EMOS_TYPE 'an'
        edit FSFAMILY 'an'
        edit SMSFILES '/home/ma/emos/def/e_36r1/sync'
        task dcda_to_da_bk
          edit QUEUE 'ns'
          edit QUEUE_EPILOG 'ns'
          edit SCHOST 'c1a'
          edit STHOST '/s1a'
          edit SCHOST_BKUP 'c1b'
          edit STHOST_BKUP '/s1b'
          edit SMSLOGHOST 'c1a'
          edit LOGDIR '/s1a/emos_esuite/emos_data/0046/log'
          edit SMSOUT '/s1a/emos_esuite/emos_data/0046/log'
          edit OPHOST 'c1b'
          edit SMSFILES '/home/ma/emos/def/e_36r1/an'
          inlimit /limits:c1a
          label files ""
          label info ""
          event fdb
        task sync_satmon
          trigger ../satmon eq complete
        task sync_bias
        task sync_fc
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/12dc/fc/model:step gt 6 or /e_36r1/main/12dc/fc/model eq complete))
          edit QUEUE 'ns'
          edit QUEUE_EPILOG 'ns'
          edit SCHOST 'c1a'
          edit STHOST '/s1a'
          edit SCHOST_BKUP 'c1b'
          edit STHOST_BKUP '/s1b'
          edit SMSLOGHOST 'c1a'
          edit LOGDIR '/s1a/emos_esuite/emos_data/0046/log'
          edit SMSOUT '/s1a/emos_esuite/emos_data/0046/log'
          edit EMOS_BASE '18'
          inlimit /limits:c1a
          label files ""
          label info ""
          event fdb
        task sync_an
          edit QUEUE 'ns'
          edit QUEUE_EPILOG 'ns'
          edit SCHOST 'c1a'
          edit STHOST '/s1a'
          edit SCHOST_BKUP 'c1b'
          edit STHOST_BKUP '/s1b'
          edit SMSLOGHOST 'c1a'
          edit LOGDIR '/s1a/emos_esuite/emos_data/0046/log'
          edit SMSOUT '/s1a/emos_esuite/emos_data/0046/log'
          edit EMOS_BASE '18'
          inlimit /limits:c1a
          label files ""
          label info ""
          event fdb
        task sync_wv
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/12dc/fc/model:step gt 3 or /e_36r1/main/12dc/fc/model eq complete))
          edit QUEUE 'ns'
          edit QUEUE_EPILOG 'ns'
          edit SCHOST 'c1a'
          edit STHOST '/s1a'
          edit SCHOST_BKUP 'c1b'
          edit STHOST_BKUP '/s1b'
          edit SMSLOGHOST 'c1a'
          edit LOGDIR '/s1a/emos_esuite/emos_data/0046/log'
          edit SMSOUT '/s1a/emos_esuite/emos_data/0046/log'
          edit EMOS_BASE '18'
          inlimit /limits:c1a
          label files ""
          label info ""
          event fdb
        task sync_clean
          edit QUEUE 'ns'
          edit QUEUE_EPILOG 'ns'
          edit SCHOST 'c1a'
          edit STHOST '/s1a'
          edit SCHOST_BKUP 'c1b'
          edit STHOST_BKUP '/s1b'
          edit SMSLOGHOST 'c1a'
          edit LOGDIR '/s1a/emos_esuite/emos_data/0046/log'
          edit SMSOUT '/s1a/emos_esuite/emos_data/0046/log'
          inlimit /limits:c1a
      endfamily
      family archive
        trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ../../main/12dc/an eq complete and ( ../../main/12dc/fc/model:step ge 15 or ../../main/12dc/fc/model eq complete))
        edit FSFAMILY 'mars'
        edit SMSFILES '/home/ma/emos/def/e_36r1/archive'
        edit EMOS_TYPE 'an'
        inlimit /limits:mars
        inlimit /e_36r1/limits:mars
        task sp2fdb
          edit STEPS '3/6/9/12/18'
          label files ""
          label info ""
          event fdb
        task ansfc
          trigger sp2fdb eq complete
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task anml
          trigger ansfc:fdb or ansfc eq complete
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          inlimit /e_36r1/limits:ml
          label files ""
          label info ""
          event fdb
        task anpl
          trigger anml:fdb or anml eq complete
          edit FCLENGTH '15'
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task anwave
          trigger ansfc:fdb or ansfc eq complete
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task anil
          trigger anpl:fdb or anpl eq complete
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task derivedp
          trigger ./ansfc eq complete
          edit STEPS '0'
          label files ""
          label info ""
          event fdb
        family obs
          edit FSFAMILY 'an'
          inlimit /e_36r1/limits:obs
          task ai
            trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/12dc/an/4dvar/ifstraj:finaltraj or /e_36r1/main/12dc/an/4dvar eq complete))
            edit TYPE 'ai'
            label files ""
            label info ""
            event fdb
          task af
            trigger ../../fb/feedback eq complete
            edit TYPE 'af'
            label files ""
            label info ""
            event fdb
          task fb
            trigger ../../fb/feedback eq complete
            edit TYPE 'fb'
            label files ""
            label info ""
            event fdb
        endfamily
        task ice_nmc
          trigger ansfc:fdb or ansfc eq complete
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task sst
          trigger ansfc:fdb or ansfc eq complete
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        family fc015
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/12dc/fc/model:step ge 21 or /e_36r1/main/12dc/fc/model eq complete)
          edit STEPS '3/6/9/12/15'
          edit EMOS_BASE '18'
          edit EMOS_TYPE 'fc'
          task sp2fdb
            label files ""
            label info ""
            event fdb
          task fcsfc
            trigger sp2fdb eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcml
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            inlimit /e_36r1/limits:ml
            label files ""
            label info ""
            event fdb
          task fcpl
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcil
            trigger fcpl:fdb or fcpl eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
        endfamily
        task fc12g
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/12dc/fc/model eq complete)
          edit STEPS_SPECTRAL '6/12'
          edit EMOS_BASE '18'
          edit STEPS '0/3/6/9/12/15'
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task rmfdb
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/12dc/obs eq complete)
        task flush
          edit STREAM 'oper'
          label files ""
          label info ""
          event fdb
      endfamily
      family fb
        trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/12dc/an eq complete)
        edit FSFAMILY 'an'
        edit SMSFILES '/home/ma/emos/def/e_36r1/fb'
        task sync_rstrhbias
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
        task biassave
          trigger sync_rstrhbias eq complete
        family feedback
          edit ACCOUNT 'oedcdaan'
          edit NPES '8'
          edit FBNEW2OLD '1'
          edit LCONV '1'
          edit LSATOB '1'
          edit LREO3 '1'
          edit LAEOLUS '0'
          edit LSSMI '1'
          edit LSSMIS '1'
          edit LAMSRE '1'
          edit LTMI '1'
          edit LHIRS '1'
          edit LMSU '1'
          edit LSSU '1'
          edit LAMSUA '1'
          edit LAMSUB '1'
          edit LMHS '1'
          edit LGEOS '1'
          edit LSCAT '1'
          edit LAIRS '1'
          edit LVTPR1 '0'
          edit LVTPR2 '0'
          edit LGPSRO '1'
          edit LIASI '1'
          edit LMERIS '1'
          task revmatchup
            edit NPES '128'
            edit THREADS '2'
          task matchup
            trigger revmatchup eq complete
            edit NPES '128'
            edit THREADS '2'
          task odbcmp_prepare
            edit THREADS '1'
            edit NPES '1'
            edit MEM '1000'
            edit CPUTIME '1000'
            edit PRIORITY '90'
          family odb_compress
            trigger odbcmp_prepare eq complete and matchup eq complete
            edit THREADS '2'
            edit NPES '1'
            task odbcmp_conv
            task odbcmp_surf_conv
            task odbcmp_airs
            task odbcmp_iasi
            task odbcmp_hirs
            task odbcmp_msu
            task odbcmp_ssu
            task odbcmp_amsua
            task odbcmp_amsub
            task odbcmp_mhs
            task odbcmp_satob
            task odbcmp_reo3
            task odbcmp_ssmi
            task odbcmp_ssmis
            task odbcmp_amsre
            task odbcmp_tmi
            task odbcmp_scatt
            task odbcmp_geos
            task odbcmp_gpsro
            task odbcmp_meris
          endfamily
          family odb2bufr
            trigger matchup eq complete
            edit NPES '1'
            edit SUBTYPES_INCLUDE ''
            edit SUBTYPES_EXCLUDE ''
            task o2b_conv
              edit SUBTYPES_EXCLUDE '85 82 83 87'
              edit NPES '2'
            task o2b_surf_conv
              edit SUBTYPES_EXCLUDE '142 143 144 145 85 82 83 87'
              edit NPES '2'
            task o2b_airs
              edit SUBTYPES_INCLUDE '57'
              edit NPES '8'
              edit THREADS '2'
            task o2b_iasi
              edit SUBTYPES_INCLUDE '240'
              edit NPES '8'
              edit THREADS '2'
            task o2b_hirs
              edit SUBTYPES_INCLUDE '54 55'
              edit NPES '2'
            task o2b_msu
              edit SUBTYPES_INCLUDE '54 55'
            task o2b_ssu
              edit SUBTYPES_INCLUDE '54 55'
            task o2b_amsua
              edit SUBTYPES_INCLUDE '54 55'
              edit NPES '2'
            task o2b_amsub
              edit SUBTYPES_INCLUDE '54 55'
              edit NPES '2'
            task o2b_mhs
              edit SUBTYPES_INCLUDE '54 55'
              edit NPES '2'
            task o2b_satob
              edit SUBTYPES_INCLUDE '82 83 87'
              edit NPES '2'
            task o2b_reo3
              edit SUBTYPES_INCLUDE '205 206 207 208 215'
              edit NPES '4'
              edit THREADS '2'
            task o2b_ssmi
              edit SUBTYPES_INCLUDE '55 127'
            task o2b_ssmis
              edit SUBTYPES_INCLUDE '49'
            task o2b_amsre
              edit SUBTYPES_INCLUDE '59'
            task o2b_tmi
              edit SUBTYPES_INCLUDE '129'
            task o2b_scatt
              edit SUBTYPES_INCLUDE '122 137 139'
            task o2b_geos
              edit SUBTYPES_INCLUDE '189 89 76 77 78'
              edit NPES '2'
            task o2b_gpsro
              edit SUBTYPES_INCLUDE '250'
            task o2b_meris
              edit SUBTYPES_INCLUDE '212'
          endfamily
          task mergebufr
            trigger odb2bufr eq complete
        endfamily
        family obstat
          trigger feedback/odb2bufr eq complete
          edit NPES '1'
          task obstat_conv
          task obstat_surf_conv
          task obstat_airs
            edit NPES '8'
            edit THREADS '2'
          task obstat_iasi
            edit NPES '8'
            edit THREADS '2'
          task obstat_hirs
            edit NPES '4'
          task obstat_msu
          task obstat_ssu
          task obstat_amsua
            edit NPES '2'
          task obstat_amsub
          task obstat_mhs
          task obstat_satob
          task obstat_reo3
          task obstat_ssmi
            edit NPES '2'
          task obstat_ssmis
            edit NPES '2'
          task obstat_amsre
            edit NPES '2'
          task obstat_tmi
            edit NPES '2'
          task obstat_scatt
          task obstat_geos
            edit NPES '2'
          task obstat_gpsro
            edit NPES '2'
          task obstat_meris
            edit NPES '2'
        endfamily
        task fdbksave
          trigger feedback eq complete and obstat eq complete
        task clean_odbcmp
          trigger feedback eq complete and obstat eq complete and fdbksave eq complete
        task wavesave
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ../../../main/12dc/an/4dvar eq complete)
        task obstat_merge
          trigger obstat eq complete
          edit WSHOST 'bee01'
          edit PRIORITY '51'
      endfamily
      family traj
        task feed
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ../../../main/12dc/fc eq complete)
        task cp_to_other
          trigger feed eq complete
      endfamily
      family satmon
        trigger ../00dc/satmon eq complete
        edit FSFAMILY 'an'
        inlimit /e_36r1/limits:smon
        task getsmon
          trigger ../fb/fdbksave eq complete and ../archive/obs/af eq complete
        task smon_meris
          trigger getsmon eq complete
          edit smon_smsmode 'MERIS'
        task smon_amsre
          trigger getsmon eq complete
          edit smon_smsmode 'AMSRE'
        task smon_ssmis
          trigger getsmon eq complete
          edit smon_smsmode 'SSMIS'
        task smon_tmi
          trigger getsmon eq complete
          edit smon_smsmode 'TMI'
        task smon_gpsro
          trigger getsmon eq complete
          edit smon_smsmode 'GPSRO'
        task smon_pacrars
          trigger getsmon eq complete
          edit smon_smsmode 'PACRARS'
        task smon_ears
          trigger getsmon eq complete
          edit smon_smsmode 'EARS'
        task smon_o3msg
          trigger getsmon eq complete
          edit smon_smsmode 'O3MSG'
        task smon_o3knmi
          trigger getsmon eq complete
          edit smon_smsmode 'O3KNMI'
        task smon_pwc
          trigger getsmon eq complete
          edit smon_smsmode 'PWC'
        task smon_temp
          trigger getsmon eq complete
          edit smon_smsmode 'TEMP'
        task smon_iasi
          trigger getsmon eq complete
          edit smon_smsmode 'IASI'
        task smon_airs
          trigger getsmon eq complete
          edit smon_smsmode 'AIRS'
        task smon_dbamv
          trigger getsmon eq complete
          edit smon_smsmode 'DBAMV'
        task smon_amv
          trigger getsmon eq complete
          edit smon_smsmode 'AMV'
        task smon_o3
          trigger getsmon eq complete
          edit smon_smsmode 'O3'
        task smon_dmsp
          trigger getsmon eq complete
          edit smon_smsmode 'DMSP'
        task smon_noaa
          trigger getsmon eq complete
          edit smon_smsmode 'NOAA'
        task smon_geos
          trigger getsmon eq complete
          edit smon_smsmode 'GEOS'
        task smon_scatt
          trigger getsmon eq complete
          edit smon_smsmode 'SCATT'
        task smon_slmoist
          trigger getsmon eq complete
          edit smon_smsmode 'SLMOIST'
        task smon_clean
          trigger getsmon eq complete and smon_meris eq complete and smon_amsre eq complete and smon_ssmis eq complete and smon_tmi eq complete and smon_gpsro eq complete and smon_pacrars eq complete and smon_ears eq complete and smon_o3msg eq complete and smon_o3knmi eq complete and smon_pwc eq complete and smon_temp eq complete and smon_iasi eq complete and smon_airs eq complete and smon_dbamv eq complete and smon_amv eq complete and smon_o3 eq complete and smon_dmsp eq complete and smon_noaa eq complete and smon_geos eq complete and smon_scatt eq complete and smon_slmoist eq complete
      endfamily
      task clean
        trigger fb eq complete and ../12/clean eq complete
        edit FSFAMILY 'an'
      task cleanplus6
        trigger clean eq complete
        edit EMOS_BASE '18'
        edit FSFAMILY 'an'
      family verif
        edit QUEUE 'emos-e'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        edit SMSFILES '/home/ma/emos/def/e_36r1/verif'
        edit DELTA_DAY '1'
        inlimit /limits:linux_cluster
        family fc
          task v12li1
          task v12li2
          task v12up1
            trigger v12li1 eq complete
          task v12up2
            trigger v12up1 eq complete and v12li2 eq complete
          task v12pl
            trigger v12up2 eq complete
          task v12memv
            trigger v12up2 eq complete
            edit DATEMASK '02.*.*'
          task v12memb
            trigger v12up2 eq complete
            edit DATEMASK '11.*.*'
          task v12upmv
            trigger v12memv eq complete
            edit DATEMASK '02.*.*'
          task v12upmb
            trigger v12memb eq complete
            edit DATEMASK '11.*.*'
          task v12crmv
            edit DATEMASK '28.*.*'
          task v12crmb
            edit DATEMASK '28.*.*'
          task v12crd
            edit DATEMASK '28.*.*'
        endfamily
        family oc
          trigger ./fc eq complete or ./fc eq aborted
          task mk_dummy_fields
            edit QUEUE 'emos'
            edit USER 'emos'
            edit PP_BASE '/pp1'
            edit WSHOST 'pp1'
            edit QUEUE_EPILOG 'emos-exp'
            edit LOGDIR '/pp1/log'
            edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
            edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
            edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
            edit SMSOUT '/vol/emos_nc/output'
          task insert_fields
            trigger mk_dummy_fields eq complete
            edit QUEUE 'emos-e'
            edit QUEUE_EPILOG 'emos-exp'
            edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
            edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
            edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
            edit LOGDIR '/vol/emos_nc/output'
            edit SMSOUT '/vol/emos_nc/output'
            edit PP_BASE '/pp1'
            inlimit /limits:linux_cluster
          task vbrli1
            trigger insert_fields eq complete
          task vbrli2
            trigger vbrli1 eq complete
          task vwali1
            trigger insert_fields eq complete
          task vwali2
            trigger vwali1 eq complete
          task vwali3
            trigger vwali2 eq complete
          task vofli
            trigger insert_fields eq complete
          task vtlli
            trigger insert_fields eq complete
          task vtoli
            trigger insert_fields eq complete
          task vmoli
            trigger insert_fields eq complete
          task vbrup1
            trigger vbrli1 eq complete
          task vbrup2
            trigger vbrup1 eq complete and vbrli2 eq complete
          task vwaup1
            trigger vwali1 eq complete
          task vwaup2
            trigger vwaup1 eq complete and vwali2 eq complete
          task vwaup3
            trigger vwaup2 eq complete and vwali3 eq complete
          task vofup
            trigger vofli eq complete
          task vtlup
            trigger vtlli eq complete
          task vtoup
            trigger vtoli eq complete
          task vmoup
            trigger vmoli eq complete
          task vmeli
            trigger insert_fields eq complete
          task vmeup
            trigger vmeli eq complete
          task voccrd
            edit DATEMASK '28.*.*'
          task voccrm
            edit DATEMASK '28.*.*'
          task vocmem
            edit DATEMASK '11.*.*'
          task vocupm
            trigger vocmem eq complete
            edit DATEMASK '11.*.*'
          task cleanup
            trigger vmeup eq complete and vbrup2 eq complete and vwaup3 eq complete and vofup eq complete and vtlup eq complete and vtoup eq complete and vmoup eq complete and voccrd eq complete and voccrm eq complete and vocupm eq complete
            edit QUEUE 'emos-e'
            edit QUEUE_EPILOG 'emos-exp'
            edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
            edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
            edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
            edit LOGDIR '/vol/emos_nc/output'
            edit SMSOUT '/vol/emos_nc/output'
            inlimit /limits:linux_cluster
        endfamily
        family cf
          trigger oc eq complete or oc eq aborted
          task vcnli1
            late -c +02:00
          task vcnli2
          task vcnup1
            trigger vcnli1 eq complete
          task vcnup2
            trigger vcnup1 eq complete and vcnli2 eq complete
          task vcnpl
            trigger vcnup2 eq complete
          task vcncrd
          task vcncrm
            edit DATEMASK '28.*.*'
          task vcnmem
            trigger vcnup2 eq complete
            edit DATEMASK '11.*.*'
          task vcnupm
            trigger vcnmem eq complete
            edit DATEMASK '11.*.*'
          task vcnli3
          task vcnli4
          task vcnup3
            trigger vcnup2 eq complete and vcnli3 eq complete
          task vcnup4
            trigger vcnup3 eq complete and vcnli4 eq complete
          task vcnmem2
            trigger vcnup4 eq complete
            edit DATEMASK '22.*.*'
          task vcnupm2
            trigger vcnmem2 eq complete
            edit DATEMASK '22.*.*'
        endfamily
        family ob
          trigger oc eq complete or oc eq aborted
          time 05:00
          family stage1
            task vobli1
            task vobli2
            task vobup1
              trigger vobli1 eq complete
            task vobup2
              trigger vobup1 eq complete and vobli2 eq complete
            task vobnobs
              trigger vobup2 eq complete
              edit DATEMASK '03.*.*'
            task vobmem
              trigger ./vobup2 eq complete
              edit DATEMASK '03.*.*'
            task vobupm
              trigger vobmem eq complete
              edit DATEMASK '03.*.*'
          endfamily
          family stage2
            trigger ./stage1 eq complete
            edit EMOS_BASE '00'
            task vobli1
            task vobli2
            task vobup1
              trigger vobli1 eq complete
            task vobup2
              trigger vobup1 eq complete and vobli2 eq complete
            task vobnobs
              trigger vobup2 eq complete
              edit DATEMASK '03.*.*'
            task vobmem
              trigger ./vobup2 eq complete
              edit DATEMASK '03.*.*'
            task vobupm
              trigger vobmem eq complete
              edit DATEMASK '03.*.*'
          endfamily
          task vobcrd
            edit DATEMASK '28.*.*'
          task vobcrm
            edit DATEMASK '28.*.*'
        endfamily
        family em
          task vemli
            late -c +02:00
          task vemup
            trigger vemli eq complete
          task vemcrd
            edit DATEMASK '28.*.*'
          task vemcrm
            edit DATEMASK '28.*.*'
          task vemmem
            trigger vemup eq complete
            edit DATEMASK '11.*.*'
          task vemupm
            trigger vemmem eq complete
            edit DATEMASK '11.*.*'
        endfamily
        family es
          trigger em eq complete
          task vesli
            late -c +02:00
          task vesup
            trigger vesli eq complete
          task vescrd
            edit DATEMASK '28.*.*'
          task vescrm
            edit DATEMASK '28.*.*'
          task vesmem
            trigger vesup eq complete
            edit DATEMASK '11.*.*'
          task vesupm
            trigger vesmem eq complete
            edit DATEMASK '11.*.*'
        endfamily
        family cv
          trigger es eq complete
          task vcvli
            late -c +02:00
          task vcvup
            trigger vcvli eq complete
          task vcvcrd
            edit DATEMASK '28.*.*'
          task vcvcrm
            edit DATEMASK '28.*.*'
          task vcvmem
            trigger vcvup eq complete
            edit DATEMASK '11.*.*'
          task vcvupm
            trigger vcvmem eq complete
            edit DATEMASK '11.*.*'
        endfamily
        family wv
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ../../../main/12dc/fc/model eq complete)
          task glver
          task glupr
            trigger glver eq complete
          task glcrr
            edit DATEMASK '28.*.*'
          task ff10ver
            trigger glver eq complete
          task ff10upr
            trigger ff10ver eq complete and glupr eq complete
          task glmem
            trigger ff10upr eq complete
            edit DATEMASK '11.*.*'
          task glupm
            trigger glmem eq complete
            edit DATEMASK '11.*.*'
          task glcrm
            edit DATEMASK '28.*.*'
        endfamily
        family sf
          trigger fc eq complete or fc eq aborted
          task vseli1
          task vseli2
          task vseup1
            trigger vseli1 eq complete
          task vseup2
            trigger vseup1 eq complete and vseli2 eq complete
          task vsecrd
            edit DATEMASK '28.*.*'
          task vsecrm
            edit DATEMASK '28.*.*'
          task vsemem
            trigger vseup2 eq complete
            edit DATEMASK '12.*.*'
          task vseupm
            trigger vsemem eq complete
            edit DATEMASK '12.*.*'
        endfamily
      endfamily
    endfamily
    family 18
      edit EMOS_BASE '18'
      edit MXUP_TRAJ '3'
      edit REINISTREAM 'DA'
      edit LAR_4V_EF 'false'
      edit LFC_DIR_FREQ 'false'
      edit WINDOW_LENGTH_4D '6'
      edit WINDOW_OFFSET_4D '3'
      edit EDFAMILY 'true'
      edit PERIOD_4D '6'
      edit FCLENGTH '9'
      edit DELTA_DAY '0'
      family archive
        trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ../../main/00/prod/18an eq complete)
        edit FSFAMILY 'mars'
        edit SMSFILES '/home/ma/emos/def/e_36r1/archive'
        edit EMOS_TYPE 'an'
        inlimit /limits:mars
        inlimit /e_36r1/limits:mars
        task ansfc
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/00/prod/18an/derived_param eq complete))
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task anml
          trigger ansfc:fdb or ansfc eq complete
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          inlimit /e_36r1/limits:ml
          label files ""
          label info ""
          event fdb
        task anpl
          trigger anml:fdb or anml eq complete
          edit FCLENGTH '6'
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task anwave
          trigger ansfc:fdb or ansfc eq complete
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task anil
          trigger anpl:fdb or anpl eq complete
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task derivedp
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/00/prod/18an/derived_param eq complete))
          edit STEPS '0'
          label files ""
          label info ""
          event fdb
        family obs
          edit FSFAMILY 'an'
          inlimit /e_36r1/limits:obs
          task ai
            trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/18/an/4dvar/ifstraj:finaltraj or /e_36r1/main/18/an/4dvar eq complete))
            edit TYPE 'ai'
            label files ""
            label info ""
            event fdb
          task af
            trigger ../../fb/feedback eq complete
            edit TYPE 'af'
            label files ""
            label info ""
            event fdb
          task fb
            trigger ../../fb/feedback eq complete
            edit TYPE 'fb'
            label files ""
            label info ""
            event fdb
        endfamily
        task ice_nmc
          trigger ansfc:fdb or ansfc eq complete
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task sst
          trigger ansfc:fdb or ansfc eq complete
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        family fc015
          edit STEPS '3/6/9/12/15'
          edit EMOS_TYPE 'fc'
          task fcsfc
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcml
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            inlimit /e_36r1/limits:ml
            label files ""
            label info ""
            event fdb
          task fcpl
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcil
            trigger fcpl:fdb or fcpl eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task derivedp
            trigger fcil:fdb or fcil eq complete
            label files ""
            label info ""
            event fdb
        endfamily
        task fc12g
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/18/fc/model eq complete)
          edit STEPS '0/3/6/9/12/15'
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      family fb
        trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/18/an eq complete)
        edit FSFAMILY 'an'
        edit SMSFILES '/home/ma/emos/def/e_36r1/fb'
        task sync_rstrhbias
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
        task biassave
          trigger sync_rstrhbias eq complete
        family feedback
          edit ACCOUNT 'oeoperan'
          edit NPES '8'
          edit FBNEW2OLD '1'
          edit LCONV '1'
          edit LSATOB '1'
          edit LREO3 '1'
          edit LAEOLUS '0'
          edit LSSMI '1'
          edit LSSMIS '1'
          edit LAMSRE '1'
          edit LTMI '1'
          edit LHIRS '1'
          edit LMSU '1'
          edit LSSU '1'
          edit LAMSUA '1'
          edit LAMSUB '1'
          edit LMHS '1'
          edit LGEOS '1'
          edit LSCAT '1'
          edit LAIRS '1'
          edit LVTPR1 '0'
          edit LVTPR2 '0'
          edit LGPSRO '1'
          edit LIASI '1'
          edit LMERIS '1'
          task revmatchup
            edit NPES '128'
            edit THREADS '2'
          task matchup
            trigger revmatchup eq complete
            edit NPES '128'
            edit THREADS '2'
          task odbcmp_prepare
            edit THREADS '1'
            edit NPES '1'
            edit MEM '1000'
            edit CPUTIME '1000'
            edit PRIORITY '90'
          family odb_compress
            trigger odbcmp_prepare eq complete and matchup eq complete
            edit THREADS '2'
            edit NPES '1'
            task odbcmp_conv
            task odbcmp_surf_conv
            task odbcmp_airs
            task odbcmp_iasi
            task odbcmp_hirs
            task odbcmp_msu
            task odbcmp_ssu
            task odbcmp_amsua
            task odbcmp_amsub
            task odbcmp_mhs
            task odbcmp_satob
            task odbcmp_reo3
            task odbcmp_ssmi
            task odbcmp_ssmis
            task odbcmp_amsre
            task odbcmp_tmi
            task odbcmp_scatt
            task odbcmp_geos
            task odbcmp_gpsro
            task odbcmp_meris
          endfamily
          family odb2bufr
            trigger matchup eq complete
            edit NPES '1'
            edit SUBTYPES_INCLUDE ''
            edit SUBTYPES_EXCLUDE ''
            task o2b_conv
              edit SUBTYPES_EXCLUDE '85 82 83 87'
              edit NPES '2'
            task o2b_surf_conv
              edit SUBTYPES_EXCLUDE '142 143 144 145 85 82 83 87'
              edit NPES '2'
            task o2b_airs
              edit SUBTYPES_INCLUDE '57'
              edit NPES '8'
              edit THREADS '2'
            task o2b_iasi
              edit SUBTYPES_INCLUDE '240'
              edit NPES '8'
              edit THREADS '2'
            task o2b_hirs
              edit SUBTYPES_INCLUDE '54 55'
              edit NPES '2'
            task o2b_msu
              edit SUBTYPES_INCLUDE '54 55'
            task o2b_ssu
              edit SUBTYPES_INCLUDE '54 55'
            task o2b_amsua
              edit SUBTYPES_INCLUDE '54 55'
              edit NPES '2'
            task o2b_amsub
              edit SUBTYPES_INCLUDE '54 55'
              edit NPES '2'
            task o2b_mhs
              edit SUBTYPES_INCLUDE '54 55'
              edit NPES '2'
            task o2b_satob
              edit SUBTYPES_INCLUDE '82 83 87'
              edit NPES '2'
            task o2b_reo3
              edit SUBTYPES_INCLUDE '205 206 207 208 215'
              edit NPES '4'
              edit THREADS '2'
            task o2b_ssmi
              edit SUBTYPES_INCLUDE '55 127'
            task o2b_ssmis
              edit SUBTYPES_INCLUDE '49'
            task o2b_amsre
              edit SUBTYPES_INCLUDE '59'
            task o2b_tmi
              edit SUBTYPES_INCLUDE '129'
            task o2b_scatt
              edit SUBTYPES_INCLUDE '122 137 139'
            task o2b_geos
              edit SUBTYPES_INCLUDE '189 89 76 77 78'
              edit NPES '2'
            task o2b_gpsro
              edit SUBTYPES_INCLUDE '250'
            task o2b_meris
              edit SUBTYPES_INCLUDE '212'
          endfamily
          task mergebufr
            trigger odb2bufr eq complete
        endfamily
        family obstat
          trigger feedback/odb2bufr eq complete
          edit NPES '1'
          task obstat_conv
          task obstat_surf_conv
          task obstat_airs
            edit NPES '8'
            edit THREADS '2'
          task obstat_iasi
            edit NPES '8'
            edit THREADS '2'
          task obstat_hirs
            edit NPES '4'
          task obstat_msu
          task obstat_ssu
          task obstat_amsua
            edit NPES '2'
          task obstat_amsub
          task obstat_mhs
          task obstat_satob
          task obstat_reo3
          task obstat_ssmi
            edit NPES '2'
          task obstat_ssmis
            edit NPES '2'
          task obstat_amsre
            edit NPES '2'
          task obstat_tmi
            edit NPES '2'
          task obstat_scatt
          task obstat_geos
            edit NPES '2'
          task obstat_gpsro
            edit NPES '2'
          task obstat_meris
            edit NPES '2'
        endfamily
        task fdbksave
          trigger feedback eq complete and obstat eq complete
        task clean_odbcmp
          trigger feedback eq complete and obstat eq complete and fdbksave eq complete
        task wavesave
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ../../../main/18/an/4dvar eq complete)
        task obstat_merge
          trigger obstat eq complete
          edit WSHOST 'bee01'
          edit PRIORITY '51'
      endfamily
      family traj
        task feed
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ../../../main/18/fc eq complete)
        task cp_to_other
          trigger feed eq complete
      endfamily
      family satmon
        trigger ../00dc/satmon eq complete
        edit FSFAMILY 'an'
        inlimit /e_36r1/limits:smon
        task getsmon
          trigger ../fb/fdbksave eq complete and ../archive/obs/af eq complete
        task smon_meris
          trigger getsmon eq complete
          edit smon_smsmode 'MERIS'
        task smon_amsre
          trigger getsmon eq complete
          edit smon_smsmode 'AMSRE'
        task smon_ssmis
          trigger getsmon eq complete
          edit smon_smsmode 'SSMIS'
        task smon_tmi
          trigger getsmon eq complete
          edit smon_smsmode 'TMI'
        task smon_gpsro
          trigger getsmon eq complete
          edit smon_smsmode 'GPSRO'
        task smon_pacrars
          trigger getsmon eq complete
          edit smon_smsmode 'PACRARS'
        task smon_ears
          trigger getsmon eq complete
          edit smon_smsmode 'EARS'
        task smon_o3msg
          trigger getsmon eq complete
          edit smon_smsmode 'O3MSG'
        task smon_o3knmi
          trigger getsmon eq complete
          edit smon_smsmode 'O3KNMI'
        task smon_pwc
          trigger getsmon eq complete
          edit smon_smsmode 'PWC'
        task smon_temp
          trigger getsmon eq complete
          edit smon_smsmode 'TEMP'
        task smon_iasi
          trigger getsmon eq complete
          edit smon_smsmode 'IASI'
        task smon_airs
          trigger getsmon eq complete
          edit smon_smsmode 'AIRS'
        task smon_dbamv
          trigger getsmon eq complete
          edit smon_smsmode 'DBAMV'
        task smon_amv
          trigger getsmon eq complete
          edit smon_smsmode 'AMV'
        task smon_o3
          trigger getsmon eq complete
          edit smon_smsmode 'O3'
        task smon_dmsp
          trigger getsmon eq complete
          edit smon_smsmode 'DMSP'
        task smon_noaa
          trigger getsmon eq complete
          edit smon_smsmode 'NOAA'
        task smon_geos
          trigger getsmon eq complete
          edit smon_smsmode 'GEOS'
        task smon_scatt
          trigger getsmon eq complete
          edit smon_smsmode 'SCATT'
        task smon_slmoist
          trigger getsmon eq complete
          edit smon_smsmode 'SLMOIST'
        task smon_clean
          trigger getsmon eq complete and smon_meris eq complete and smon_amsre eq complete and smon_ssmis eq complete and smon_tmi eq complete and smon_gpsro eq complete and smon_pacrars eq complete and smon_ears eq complete and smon_o3msg eq complete and smon_o3knmi eq complete and smon_pwc eq complete and smon_temp eq complete and smon_iasi eq complete and smon_airs eq complete and smon_dbamv eq complete and smon_amv eq complete and smon_o3 eq complete and smon_dmsp eq complete and smon_noaa eq complete and smon_geos eq complete and smon_scatt eq complete and smon_slmoist eq complete
      endfamily
      task clean
        trigger fb eq complete and ../12dc/clean eq complete
        edit FSFAMILY 'an'
      family verif
        edit QUEUE 'emos-e'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        edit SMSFILES '/home/ma/emos/def/e_36r1/verif'
        edit DELTA_DAY '1'
        inlimit /limits:linux_cluster
        family fc
          task v12li1
          task v12li2
          task v12up1
            trigger v12li1 eq complete
          task v12up2
            trigger v12up1 eq complete and v12li2 eq complete
          task v12pl
            trigger v12up2 eq complete
          task v12memv
            trigger v12up2 eq complete
            edit DATEMASK '02.*.*'
          task v12memb
            trigger v12up2 eq complete
            edit DATEMASK '11.*.*'
          task v12upmv
            trigger v12memv eq complete
            edit DATEMASK '02.*.*'
          task v12upmb
            trigger v12memb eq complete
            edit DATEMASK '11.*.*'
          task v12crmv
            edit DATEMASK '28.*.*'
          task v12crmb
            edit DATEMASK '28.*.*'
          task v12crd
            edit DATEMASK '28.*.*'
        endfamily
        family oc
          trigger ./fc eq complete or ./fc eq aborted
          task mk_dummy_fields
            edit QUEUE 'emos'
            edit USER 'emos'
            edit PP_BASE '/pp1'
            edit WSHOST 'pp1'
            edit QUEUE_EPILOG 'emos-exp'
            edit LOGDIR '/pp1/log'
            edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
            edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
            edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
            edit SMSOUT '/vol/emos_nc/output'
          task insert_fields
            trigger mk_dummy_fields eq complete
            edit QUEUE 'emos-e'
            edit QUEUE_EPILOG 'emos-exp'
            edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
            edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
            edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
            edit LOGDIR '/vol/emos_nc/output'
            edit SMSOUT '/vol/emos_nc/output'
            edit PP_BASE '/pp1'
            inlimit /limits:linux_cluster
          task vbrli1
            trigger insert_fields eq complete
          task vbrli2
            trigger vbrli1 eq complete
          task vwali1
            trigger insert_fields eq complete
          task vwali2
            trigger vwali1 eq complete
          task vwali3
            trigger vwali2 eq complete
          task vofli
            trigger insert_fields eq complete
          task vtlli
            trigger insert_fields eq complete
          task vtoli
            trigger insert_fields eq complete
          task vmoli
            trigger insert_fields eq complete
          task vbrup1
            trigger vbrli1 eq complete
          task vbrup2
            trigger vbrup1 eq complete and vbrli2 eq complete
          task vwaup1
            trigger vwali1 eq complete
          task vwaup2
            trigger vwaup1 eq complete and vwali2 eq complete
          task vwaup3
            trigger vwaup2 eq complete and vwali3 eq complete
          task vofup
            trigger vofli eq complete
          task vtlup
            trigger vtlli eq complete
          task vtoup
            trigger vtoli eq complete
          task vmoup
            trigger vmoli eq complete
          task vmeli
            trigger insert_fields eq complete
          task vmeup
            trigger vmeli eq complete
          task voccrd
            edit DATEMASK '28.*.*'
          task voccrm
            edit DATEMASK '28.*.*'
          task vocmem
            edit DATEMASK '11.*.*'
          task vocupm
            trigger vocmem eq complete
            edit DATEMASK '11.*.*'
          task cleanup
            trigger vmeup eq complete and vbrup2 eq complete and vwaup3 eq complete and vofup eq complete and vtlup eq complete and vtoup eq complete and vmoup eq complete and voccrd eq complete and voccrm eq complete and vocupm eq complete
            edit QUEUE 'emos-e'
            edit QUEUE_EPILOG 'emos-exp'
            edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
            edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
            edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
            edit LOGDIR '/vol/emos_nc/output'
            edit SMSOUT '/vol/emos_nc/output'
            inlimit /limits:linux_cluster
        endfamily
        family cf
          trigger oc eq complete or oc eq aborted
          task vcnli1
            late -c +02:00
          task vcnli2
          task vcnup1
            trigger vcnli1 eq complete
          task vcnup2
            trigger vcnup1 eq complete and vcnli2 eq complete
          task vcnpl
            trigger vcnup2 eq complete
          task vcncrd
          task vcncrm
            edit DATEMASK '28.*.*'
          task vcnmem
            trigger vcnup2 eq complete
            edit DATEMASK '11.*.*'
          task vcnupm
            trigger vcnmem eq complete
            edit DATEMASK '11.*.*'
          task vcnli3
          task vcnli4
          task vcnup3
            trigger vcnup2 eq complete and vcnli3 eq complete
          task vcnup4
            trigger vcnup3 eq complete and vcnli4 eq complete
          task vcnmem2
            trigger vcnup4 eq complete
            edit DATEMASK '22.*.*'
          task vcnupm2
            trigger vcnmem2 eq complete
            edit DATEMASK '22.*.*'
        endfamily
        family ob
          trigger oc eq complete or oc eq aborted
          time 05:00
          family stage1
            task vobli1
            task vobli2
            task vobup1
              trigger vobli1 eq complete
            task vobup2
              trigger vobup1 eq complete and vobli2 eq complete
            task vobnobs
              trigger vobup2 eq complete
              edit DATEMASK '03.*.*'
            task vobmem
              trigger ./vobup2 eq complete
              edit DATEMASK '03.*.*'
            task vobupm
              trigger vobmem eq complete
              edit DATEMASK '03.*.*'
          endfamily
          family stage2
            trigger ./stage1 eq complete
            edit EMOS_BASE '00'
            task vobli1
            task vobli2
            task vobup1
              trigger vobli1 eq complete
            task vobup2
              trigger vobup1 eq complete and vobli2 eq complete
            task vobnobs
              trigger vobup2 eq complete
              edit DATEMASK '03.*.*'
            task vobmem
              trigger ./vobup2 eq complete
              edit DATEMASK '03.*.*'
            task vobupm
              trigger vobmem eq complete
              edit DATEMASK '03.*.*'
          endfamily
          task vobcrd
            edit DATEMASK '28.*.*'
          task vobcrm
            edit DATEMASK '28.*.*'
        endfamily
        family em
          task vemli
            late -c +02:00
          task vemup
            trigger vemli eq complete
          task vemcrd
            edit DATEMASK '28.*.*'
          task vemcrm
            edit DATEMASK '28.*.*'
          task vemmem
            trigger vemup eq complete
            edit DATEMASK '11.*.*'
          task vemupm
            trigger vemmem eq complete
            edit DATEMASK '11.*.*'
        endfamily
        family es
          trigger em eq complete
          task vesli
            late -c +02:00
          task vesup
            trigger vesli eq complete
          task vescrd
            edit DATEMASK '28.*.*'
          task vescrm
            edit DATEMASK '28.*.*'
          task vesmem
            trigger vesup eq complete
            edit DATEMASK '11.*.*'
          task vesupm
            trigger vesmem eq complete
            edit DATEMASK '11.*.*'
        endfamily
        family cv
          trigger es eq complete
          task vcvli
            late -c +02:00
          task vcvup
            trigger vcvli eq complete
          task vcvcrd
            edit DATEMASK '28.*.*'
          task vcvcrm
            edit DATEMASK '28.*.*'
          task vcvmem
            trigger vcvup eq complete
            edit DATEMASK '11.*.*'
          task vcvupm
            trigger vcvmem eq complete
            edit DATEMASK '11.*.*'
        endfamily
        family wv
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ../../../main/18/fc/model eq complete)
          task glver
          task glupr
            trigger glver eq complete
          task glcrr
            edit DATEMASK '28.*.*'
          task ff10ver
            trigger glver eq complete
          task ff10upr
            trigger ff10ver eq complete and glupr eq complete
          task glmem
            trigger ff10upr eq complete
            edit DATEMASK '11.*.*'
          task glupm
            trigger glmem eq complete
            edit DATEMASK '11.*.*'
          task glcrm
            edit DATEMASK '28.*.*'
        endfamily
        family sf
          trigger fc eq complete or fc eq aborted
          task vseli1
          task vseli2
          task vseup1
            trigger vseli1 eq complete
          task vseup2
            trigger vseup1 eq complete and vseli2 eq complete
          task vsecrd
            edit DATEMASK '28.*.*'
          task vsecrm
            edit DATEMASK '28.*.*'
          task vsemem
            trigger vseup2 eq complete
            edit DATEMASK '12.*.*'
          task vseupm
            trigger vsemem eq complete
            edit DATEMASK '12.*.*'
        endfamily
      endfamily
    endfamily
    family 00
      edit EMOS_BASE '00'
      edit MXUP_TRAJ '3'
      edit REINISTREAM 'DCDA'
      edit WINDOW_LENGTH_4D '6'
      edit WINDOW_OFFSET_4D '3'
      edit REINITIALIZE 'true'
      edit EDFAMILY 'true'
      edit PERIOD_4D '6'
      edit FCLENGTH '9'
      edit DELTA_DAY '1'
      family ddh
        trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/00/fc eq complete)
        edit SMSFILES '/home/ma/emos/def/e_36r1/fc'
        task ddh
      endfamily
      family sync
        trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/00/an eq complete)
        edit EMOS_TYPE 'an'
        edit FSFAMILY 'an'
        edit SMSFILES '/home/ma/emos/def/e_36r1/sync'
        task sync_bias
          edit CALC_PSBIAS 'false'
        task sync_fc
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/00/fc/model:step gt 6 or /e_36r1/main/00/fc/model eq complete))
          edit QUEUE 'ns'
          edit QUEUE_EPILOG 'ns'
          edit SCHOST 'c1a'
          edit STHOST '/s1a'
          edit SCHOST_BKUP 'c1b'
          edit STHOST_BKUP '/s1b'
          edit SMSLOGHOST 'c1a'
          edit LOGDIR '/s1a/emos_esuite/emos_data/0046/log'
          edit SMSOUT '/s1a/emos_esuite/emos_data/0046/log'
          inlimit /limits:c1a
          label files ""
          label info ""
          event fdb
        task sync_an
          edit QUEUE 'ns'
          edit QUEUE_EPILOG 'ns'
          edit SCHOST 'c1a'
          edit STHOST '/s1a'
          edit SCHOST_BKUP 'c1b'
          edit STHOST_BKUP '/s1b'
          edit SMSLOGHOST 'c1a'
          edit LOGDIR '/s1a/emos_esuite/emos_data/0046/log'
          edit SMSOUT '/s1a/emos_esuite/emos_data/0046/log'
          inlimit /limits:c1a
          label files ""
          label info ""
          event fdb
        task sync_wv
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/00/fc/model:step gt 3 or /e_36r1/main/00/fc/model eq complete))
          edit QUEUE 'ns'
          edit QUEUE_EPILOG 'ns'
          edit SCHOST 'c1a'
          edit STHOST '/s1a'
          edit SCHOST_BKUP 'c1b'
          edit STHOST_BKUP '/s1b'
          edit SMSLOGHOST 'c1a'
          edit LOGDIR '/s1a/emos_esuite/emos_data/0046/log'
          edit SMSOUT '/s1a/emos_esuite/emos_data/0046/log'
          inlimit /limits:c1a
          label files ""
          label info ""
          event fdb
        task sync_clean
          edit QUEUE 'ns'
          edit QUEUE_EPILOG 'ns'
          edit SCHOST 'c1a'
          edit STHOST '/s1a'
          edit SCHOST_BKUP 'c1b'
          edit STHOST_BKUP '/s1b'
          edit SMSLOGHOST 'c1a'
          edit LOGDIR '/s1a/emos_esuite/emos_data/0046/log'
          edit SMSOUT '/s1a/emos_esuite/emos_data/0046/log'
          inlimit /limits:c1a
      endfamily
      family archive
        trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ../../main/00/an eq complete and ( ../../main/00/fc/model:step ge 9 or ../../main/00/fc/model eq complete))
        edit FSFAMILY 'mars'
        edit SMSFILES '/home/ma/emos/def/e_36r1/archive'
        edit EMOS_TYPE 'an'
        inlimit /limits:mars
        inlimit /e_36r1/limits:mars
        task ansfc
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/00/prod/000 eq complete and /e_36r1/main/00/prod/00an eq complete)
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task anml
          trigger ansfc:fdb or ansfc eq complete
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          inlimit /e_36r1/limits:ml
          label files ""
          label info ""
          event fdb
        task anpl
          trigger anml:fdb or anml eq complete
          edit FCLENGTH '240'
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task anwave
          trigger ansfc:fdb or ansfc eq complete
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task anil
          trigger anpl:fdb or anpl eq complete
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task derivedp
          trigger ./ansfc eq complete
          edit STEPS '0'
          label files ""
          label info ""
          event fdb
        family obs
          edit FSFAMILY 'an'
          inlimit /e_36r1/limits:obs
          task ai
            trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/00/an/4dvar/ifstraj:finaltraj or /e_36r1/main/00/an/4dvar eq complete))
            edit TYPE 'ai'
            label files ""
            label info ""
            event fdb
          task af
            trigger ../../fb/feedback eq complete
            edit TYPE 'af'
            label files ""
            label info ""
            event fdb
          task fb
            trigger ../../fb/feedback eq complete
            edit TYPE 'fb'
            label files ""
            label info ""
            event fdb
        endfamily
        task ice_nmc
          trigger ansfc:fdb or ansfc eq complete
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task sst
          trigger ansfc:fdb or ansfc eq complete
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        family fc024
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/00/prod/024 eq complete)
          edit STEPS '0/3/006/009/012/015/018/021/024'
          edit EMOS_TYPE 'fc'
          task fcsfc
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcml
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            inlimit /e_36r1/limits:ml
            label files ""
            label info ""
            event fdb
          task fcpl
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcil
            trigger fcpl:fdb or fcpl eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task derivedp
            trigger fcil:fdb or fcil eq complete
            label files ""
            label info ""
            event fdb
        endfamily
        family fc048
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/00/prod/048 eq complete)
          edit STEPS '027/030/033/036/039/042/045/048'
          edit EMOS_TYPE 'fc'
          task fcsfc
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcml
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            inlimit /e_36r1/limits:ml
            label files ""
            label info ""
            event fdb
          task fcpl
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcil
            trigger fcpl:fdb or fcpl eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task derivedp
            trigger fcil:fdb or fcil eq complete
            label files ""
            label info ""
            event fdb
        endfamily
        family fc072
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/00/prod/072 eq complete)
          edit STEPS '051/054/057/060/063/066/069/072'
          edit EMOS_TYPE 'fc'
          task fcsfc
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcml
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            inlimit /e_36r1/limits:ml
            label files ""
            label info ""
            event fdb
          task fcpl
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcil
            trigger fcpl:fdb or fcpl eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task derivedp
            trigger fcil:fdb or fcil eq complete
            label files ""
            label info ""
            event fdb
        endfamily
        family fc096
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/00/prod/096 eq complete)
          edit STEPS '075/078/081/084/087/090/093/096'
          edit EMOS_TYPE 'fc'
          task fcsfc
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcml
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            inlimit /e_36r1/limits:ml
            label files ""
            label info ""
            event fdb
          task fcpl
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcil
            trigger fcpl:fdb or fcpl eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task derivedp
            trigger fcil:fdb or fcil eq complete
            label files ""
            label info ""
            event fdb
        endfamily
        family fc120
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/00/prod/120 eq complete)
          edit STEPS '099/102/105/108/111/114/117/120'
          edit EMOS_TYPE 'fc'
          task fcsfc
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcml
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            inlimit /e_36r1/limits:ml
            label files ""
            label info ""
            event fdb
          task fcpl
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcil
            trigger fcpl:fdb or fcpl eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task derivedp
            trigger fcil:fdb or fcil eq complete
            label files ""
            label info ""
            event fdb
        endfamily
        family fc144
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/00/prod/144 eq complete)
          edit STEPS '123/126/129/132/135/138/141/144'
          edit EMOS_TYPE 'fc'
          task fcsfc
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcml
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            inlimit /e_36r1/limits:ml
            label files ""
            label info ""
            event fdb
          task fcpl
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcil
            trigger fcpl:fdb or fcpl eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task derivedp
            trigger fcil:fdb or fcil eq complete
            label files ""
            label info ""
            event fdb
        endfamily
        family fc192
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/00/prod/192 eq complete)
          edit STEPS '150/156/162/168/174/180/186/192'
          edit EMOS_TYPE 'fc'
          task fcsfc
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcml
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            inlimit /e_36r1/limits:ml
            label files ""
            label info ""
            event fdb
          task fcpl
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcil
            trigger fcpl:fdb or fcpl eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task derivedp
            trigger fcil:fdb or fcil eq complete
            label files ""
            label info ""
            event fdb
        endfamily
        family fc240
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/00/prod/240 eq complete)
          edit STEPS '198/204/210/216/222/228/234/240'
          edit EMOS_TYPE 'fc'
          task fcsfc
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcml
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            inlimit /e_36r1/limits:ml
            label files ""
            label info ""
            event fdb
          task fcpl
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcil
            trigger fcpl:fdb or fcpl eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task derivedp
            trigger fcil:fdb or fcil eq complete
            label files ""
            label info ""
            event fdb
        endfamily
        family satim1
          task satim
            trigger ../../satim eq complete
            edit STEPS '0/to/48/by/3'
            label files ""
            label info ""
            event fdb
        endfamily
        family satim2
          task satim
            trigger ../../satim eq complete
            edit STEPS '54/to/120/by/6'
            label files ""
            label info ""
            event fdb
        endfamily
        task fc12g
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/00/fc/model eq complete)
          edit STEPS_SPECTRAL '6/12/18/24/30/36/42/48/54/60/66/72/78/84/90/96/102/108/114/120'
          edit STEPS '0/3/6/9/12/15/18/21/24/27/30/33/36/39/42/45/48/51/54/57/60/63/66/69/72/75/78/81/84/87/90/93/96/99/102/105/108/111/114/117/120/123/126/129/132/135/138/141/144/150/156/162/168/174/180/186/192/198/204/210/216/222/228/234/240'
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      family fb
        trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/00/an eq complete)
        edit FSFAMILY 'an'
        edit SMSFILES '/home/ma/emos/def/e_36r1/fb'
        task sync_rstrhbias
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
        task biassave
          trigger sync_rstrhbias eq complete
        family feedback
          edit ACCOUNT 'oeoperan'
          edit NPES '8'
          edit FBNEW2OLD '1'
          edit LCONV '1'
          edit LSATOB '1'
          edit LREO3 '1'
          edit LAEOLUS '0'
          edit LSSMI '1'
          edit LSSMIS '1'
          edit LAMSRE '1'
          edit LTMI '1'
          edit LHIRS '1'
          edit LMSU '1'
          edit LSSU '1'
          edit LAMSUA '1'
          edit LAMSUB '1'
          edit LMHS '1'
          edit LGEOS '1'
          edit LSCAT '1'
          edit LAIRS '1'
          edit LVTPR1 '0'
          edit LVTPR2 '0'
          edit LGPSRO '1'
          edit LIASI '1'
          edit LMERIS '1'
          task revmatchup
            edit NPES '128'
            edit THREADS '2'
          task matchup
            trigger revmatchup eq complete
            edit NPES '128'
            edit THREADS '2'
          task odbcmp_prepare
            edit THREADS '1'
            edit NPES '1'
            edit MEM '1000'
            edit CPUTIME '1000'
            edit PRIORITY '90'
          family odb_compress
            trigger odbcmp_prepare eq complete and matchup eq complete
            edit THREADS '2'
            edit NPES '1'
            task odbcmp_conv
            task odbcmp_surf_conv
            task odbcmp_airs
            task odbcmp_iasi
            task odbcmp_hirs
            task odbcmp_msu
            task odbcmp_ssu
            task odbcmp_amsua
            task odbcmp_amsub
            task odbcmp_mhs
            task odbcmp_satob
            task odbcmp_reo3
            task odbcmp_ssmi
            task odbcmp_ssmis
            task odbcmp_amsre
            task odbcmp_tmi
            task odbcmp_scatt
            task odbcmp_geos
            task odbcmp_gpsro
            task odbcmp_meris
          endfamily
          family odb2bufr
            trigger matchup eq complete
            edit NPES '1'
            edit SUBTYPES_INCLUDE ''
            edit SUBTYPES_EXCLUDE ''
            task o2b_conv
              edit SUBTYPES_EXCLUDE '85 82 83 87'
              edit NPES '2'
            task o2b_surf_conv
              edit SUBTYPES_EXCLUDE '142 143 144 145 85 82 83 87'
              edit NPES '2'
            task o2b_airs
              edit SUBTYPES_INCLUDE '57'
              edit NPES '8'
              edit THREADS '2'
            task o2b_iasi
              edit SUBTYPES_INCLUDE '240'
              edit NPES '8'
              edit THREADS '2'
            task o2b_hirs
              edit SUBTYPES_INCLUDE '54 55'
              edit NPES '2'
            task o2b_msu
              edit SUBTYPES_INCLUDE '54 55'
            task o2b_ssu
              edit SUBTYPES_INCLUDE '54 55'
            task o2b_amsua
              edit SUBTYPES_INCLUDE '54 55'
              edit NPES '2'
            task o2b_amsub
              edit SUBTYPES_INCLUDE '54 55'
              edit NPES '2'
            task o2b_mhs
              edit SUBTYPES_INCLUDE '54 55'
              edit NPES '2'
            task o2b_satob
              edit SUBTYPES_INCLUDE '82 83 87'
              edit NPES '2'
            task o2b_reo3
              edit SUBTYPES_INCLUDE '205 206 207 208 215'
              edit NPES '4'
              edit THREADS '2'
            task o2b_ssmi
              edit SUBTYPES_INCLUDE '55 127'
            task o2b_ssmis
              edit SUBTYPES_INCLUDE '49'
            task o2b_amsre
              edit SUBTYPES_INCLUDE '59'
            task o2b_tmi
              edit SUBTYPES_INCLUDE '129'
            task o2b_scatt
              edit SUBTYPES_INCLUDE '122 137 139'
            task o2b_geos
              edit SUBTYPES_INCLUDE '189 89 76 77 78'
              edit NPES '2'
            task o2b_gpsro
              edit SUBTYPES_INCLUDE '250'
            task o2b_meris
              edit SUBTYPES_INCLUDE '212'
          endfamily
          task mergebufr
            trigger odb2bufr eq complete
        endfamily
        family obstat
          trigger feedback/odb2bufr eq complete
          edit NPES '1'
          task obstat_conv
          task obstat_surf_conv
          task obstat_airs
            edit NPES '8'
            edit THREADS '2'
          task obstat_iasi
            edit NPES '8'
            edit THREADS '2'
          task obstat_hirs
            edit NPES '4'
          task obstat_msu
          task obstat_ssu
          task obstat_amsua
            edit NPES '2'
          task obstat_amsub
          task obstat_mhs
          task obstat_satob
          task obstat_reo3
          task obstat_ssmi
            edit NPES '2'
          task obstat_ssmis
            edit NPES '2'
          task obstat_amsre
            edit NPES '2'
          task obstat_tmi
            edit NPES '2'
          task obstat_scatt
          task obstat_geos
            edit NPES '2'
          task obstat_gpsro
            edit NPES '2'
          task obstat_meris
            edit NPES '2'
        endfamily
        task fdbksave
          trigger feedback eq complete and obstat eq complete
        task clean_odbcmp
          trigger feedback eq complete and obstat eq complete and fdbksave eq complete
        task wavesave
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ../../../main/00/an/4dvar eq complete)
        task obstat_merge
          trigger obstat eq complete
          edit WSHOST 'bee01'
          edit PRIORITY '51'
      endfamily
      family traj
        task feed
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ../../../main/00/fc eq complete)
        task cp_to_other
          trigger feed eq complete
      endfamily
      family satim
        edit QUEUE 'ns'
        edit QUEUE_EPILOG 'ns'
        edit SMSCMD 'smssubmit %USER% %SCHOST% %SMSJOB% %SMSJOBOUT%'
        edit SMSKILL 'smskill %USER% %SCHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %SCHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/s1b/emos_esuite/emos_data/0046/log'
        edit SMSOUT '/s1b/emos_esuite/emos_data/0046/log'
        edit MEM '11000'
        edit SMSFILES '/home/ma/emos/def/e_36r1/plot'
        edit SMSURLBASE 'http://intra.ecmwf.int/publications/cms/get/metapps'
        inlimit /limits:c1b
        task display_seaice
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/00/an eq complete)
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit SMSFILES '/home/ma/emos/def/e_36r1/admin'
          edit FSFAMILY 'undef'
          edit METVIEW 'metview_emos'
          edit SMSURL 'seaice'
          inlimit /limits:linux_cluster
        task display_sst
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit FSFAMILY 'undef'
          edit SMSURL 'sst'
          inlimit /limits:linux_cluster
        family satim_0
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/00/prod eq complete))
          edit IMAGETIME '0'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_3
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/00/prod eq complete))
          edit IMAGETIME '3'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_6
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/00/prod eq complete))
          edit IMAGETIME '6'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_9
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/00/prod eq complete))
          edit IMAGETIME '9'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_12
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/00/prod eq complete))
          edit IMAGETIME '12'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_15
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/00/prod eq complete))
          edit IMAGETIME '15'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_18
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/00/prod eq complete))
          edit IMAGETIME '18'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_21
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/00/prod eq complete))
          edit IMAGETIME '21'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_24
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/00/prod eq complete))
          edit IMAGETIME '24'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_27
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/00/prod eq complete))
          edit IMAGETIME '27'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_30
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/00/prod eq complete))
          edit IMAGETIME '30'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_33
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/00/prod eq complete))
          edit IMAGETIME '33'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_36
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/00/prod eq complete))
          edit IMAGETIME '36'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_39
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/00/prod eq complete))
          edit IMAGETIME '39'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_42
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/00/prod eq complete))
          edit IMAGETIME '42'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_45
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/00/prod eq complete))
          edit IMAGETIME '45'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_48
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/00/prod eq complete))
          edit IMAGETIME '48'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_54
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/00/prod eq complete))
          edit IMAGETIME '54'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_60
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/00/prod eq complete))
          edit IMAGETIME '60'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_66
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/00/prod eq complete))
          edit IMAGETIME '66'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_72
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/00/prod eq complete))
          edit IMAGETIME '72'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_78
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/00/prod eq complete))
          edit IMAGETIME '78'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_84
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/00/prod eq complete))
          edit IMAGETIME '84'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_90
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/00/prod eq complete))
          edit IMAGETIME '90'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_96
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/00/prod eq complete))
          edit IMAGETIME '96'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_102
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/00/prod eq complete))
          edit IMAGETIME '102'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_108
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/00/prod eq complete))
          edit IMAGETIME '108'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_114
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/00/prod eq complete))
          edit IMAGETIME '114'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
        family satim_120
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/00/prod eq complete))
          edit IMAGETIME '120'
          edit WMOID 'undef'
          family retr
            edit SATIM_MODE '1'
            inlimit /e_36r1/limits:satimsim_ret
            task satimsim
              edit MEM '781'
          endfamily
          family sat
            trigger ./retr eq complete
            edit SATIM_MODE '2'
            inlimit /e_36r1/limits:satimsim
            family sat_54
              edit WMOID '54'
              task satimsim
            endfamily
            family sat_56
              edit WMOID '56'
              task satimsim
            endfamily
            family sat_171
              edit WMOID '171'
              task satimsim
            endfamily
            family sat_255
              edit WMOID '255'
              task satimsim
            endfamily
            family sat_256
              edit WMOID '256'
              task satimsim
            endfamily
          endfamily
          family satim_flush
            trigger ./sat eq complete
            edit SATIM_MODE '3'
            task satimsim
              edit MEM '781'
          endfamily
        endfamily
      endfamily
      family satmon
        trigger ../00dc/satmon eq complete
        edit FSFAMILY 'an'
        inlimit /e_36r1/limits:smon
        task getsmon
          trigger ../fb/fdbksave eq complete and ../archive/obs/af eq complete
        task smon_meris
          trigger getsmon eq complete
          edit smon_smsmode 'MERIS'
        task smon_amsre
          trigger getsmon eq complete
          edit smon_smsmode 'AMSRE'
        task smon_ssmis
          trigger getsmon eq complete
          edit smon_smsmode 'SSMIS'
        task smon_tmi
          trigger getsmon eq complete
          edit smon_smsmode 'TMI'
        task smon_gpsro
          trigger getsmon eq complete
          edit smon_smsmode 'GPSRO'
        task smon_pacrars
          trigger getsmon eq complete
          edit smon_smsmode 'PACRARS'
        task smon_ears
          trigger getsmon eq complete
          edit smon_smsmode 'EARS'
        task smon_o3msg
          trigger getsmon eq complete
          edit smon_smsmode 'O3MSG'
        task smon_o3knmi
          trigger getsmon eq complete
          edit smon_smsmode 'O3KNMI'
        task smon_pwc
          trigger getsmon eq complete
          edit smon_smsmode 'PWC'
        task smon_temp
          trigger getsmon eq complete
          edit smon_smsmode 'TEMP'
        task smon_iasi
          trigger getsmon eq complete
          edit smon_smsmode 'IASI'
        task smon_airs
          trigger getsmon eq complete
          edit smon_smsmode 'AIRS'
        task smon_dbamv
          trigger getsmon eq complete
          edit smon_smsmode 'DBAMV'
        task smon_amv
          trigger getsmon eq complete
          edit smon_smsmode 'AMV'
        task smon_o3
          trigger getsmon eq complete
          edit smon_smsmode 'O3'
        task smon_dmsp
          trigger getsmon eq complete
          edit smon_smsmode 'DMSP'
        task smon_noaa
          trigger getsmon eq complete
          edit smon_smsmode 'NOAA'
        task smon_geos
          trigger getsmon eq complete
          edit smon_smsmode 'GEOS'
        task smon_scatt
          trigger getsmon eq complete
          edit smon_smsmode 'SCATT'
        task smon_slmoist
          trigger getsmon eq complete
          edit smon_smsmode 'SLMOIST'
        task smon_clean
          trigger getsmon eq complete and smon_meris eq complete and smon_amsre eq complete and smon_ssmis eq complete and smon_tmi eq complete and smon_gpsro eq complete and smon_pacrars eq complete and smon_ears eq complete and smon_o3msg eq complete and smon_o3knmi eq complete and smon_pwc eq complete and smon_temp eq complete and smon_iasi eq complete and smon_airs eq complete and smon_dbamv eq complete and smon_amv eq complete and smon_o3 eq complete and smon_dmsp eq complete and smon_noaa eq complete and smon_geos eq complete and smon_scatt eq complete and smon_slmoist eq complete
      endfamily
      task clean
        trigger fb eq complete and ../18/clean eq complete
        edit FSFAMILY 'an'
      family verif
        edit QUEUE 'emos-e'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        edit SMSFILES '/home/ma/emos/def/e_36r1/verif'
        edit DELTA_DAY '1'
        inlimit /limits:linux_cluster
        time 09:00
        family fc
          task v12li1
          task v12li2
          task v12up1
            trigger v12li1 eq complete
          task v12up2
            trigger v12up1 eq complete and v12li2 eq complete
          task v12pl
            trigger v12up2 eq complete
          task v12memv
            trigger v12up2 eq complete
            edit DATEMASK '02.*.*'
          task v12memb
            trigger v12up2 eq complete
            edit DATEMASK '11.*.*'
          task v12upmv
            trigger v12memv eq complete
            edit DATEMASK '02.*.*'
          task v12upmb
            trigger v12memb eq complete
            edit DATEMASK '11.*.*'
          task v12crmv
            edit DATEMASK '28.*.*'
          task v12crmb
            edit DATEMASK '28.*.*'
          task v12crd
            edit DATEMASK '28.*.*'
        endfamily
        family oc
          trigger ./fc eq complete or ./fc eq aborted
          task mk_dummy_fields
            edit QUEUE 'emos'
            edit USER 'emos'
            edit PP_BASE '/pp1'
            edit WSHOST 'pp1'
            edit QUEUE_EPILOG 'emos-exp'
            edit LOGDIR '/pp1/log'
            edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
            edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
            edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
            edit SMSOUT '/vol/emos_nc/output'
          task insert_fields
            trigger mk_dummy_fields eq complete
            edit QUEUE 'emos-e'
            edit QUEUE_EPILOG 'emos-exp'
            edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
            edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
            edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
            edit LOGDIR '/vol/emos_nc/output'
            edit SMSOUT '/vol/emos_nc/output'
            edit PP_BASE '/pp1'
            inlimit /limits:linux_cluster
          task vbrli1
            trigger insert_fields eq complete
          task vbrli2
            trigger vbrli1 eq complete
          task vwali1
            trigger insert_fields eq complete
          task vwali2
            trigger vwali1 eq complete
          task vwali3
            trigger vwali2 eq complete
          task vofli
            trigger insert_fields eq complete
          task vtlli
            trigger insert_fields eq complete
          task vtoli
            trigger insert_fields eq complete
          task vmoli
            trigger insert_fields eq complete
          task vbrup1
            trigger vbrli1 eq complete
          task vbrup2
            trigger vbrup1 eq complete and vbrli2 eq complete
          task vwaup1
            trigger vwali1 eq complete
          task vwaup2
            trigger vwaup1 eq complete and vwali2 eq complete
          task vwaup3
            trigger vwaup2 eq complete and vwali3 eq complete
          task vofup
            trigger vofli eq complete
          task vtlup
            trigger vtlli eq complete
          task vtoup
            trigger vtoli eq complete
          task vmoup
            trigger vmoli eq complete
          task vmeli
            trigger insert_fields eq complete
          task vmeup
            trigger vmeli eq complete
          task voccrd
            edit DATEMASK '28.*.*'
          task voccrm
            edit DATEMASK '28.*.*'
          task vocmem
            edit DATEMASK '11.*.*'
          task vocupm
            trigger vocmem eq complete
            edit DATEMASK '11.*.*'
          task cleanup
            trigger vmeup eq complete and vbrup2 eq complete and vwaup3 eq complete and vofup eq complete and vtlup eq complete and vtoup eq complete and vmoup eq complete and voccrd eq complete and voccrm eq complete and vocupm eq complete
            edit QUEUE 'emos-e'
            edit QUEUE_EPILOG 'emos-exp'
            edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
            edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
            edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
            edit LOGDIR '/vol/emos_nc/output'
            edit SMSOUT '/vol/emos_nc/output'
            inlimit /limits:linux_cluster
        endfamily
        family cf
          trigger oc eq complete or oc eq aborted
          task vcnli1
            late -c +02:00
          task vcnli2
          task vcnup1
            trigger vcnli1 eq complete
          task vcnup2
            trigger vcnup1 eq complete and vcnli2 eq complete
          task vcnpl
            trigger vcnup2 eq complete
          task vcncrd
          task vcncrm
            edit DATEMASK '28.*.*'
          task vcnmem
            trigger vcnup2 eq complete
            edit DATEMASK '11.*.*'
          task vcnupm
            trigger vcnmem eq complete
            edit DATEMASK '11.*.*'
          task vcnli3
          task vcnli4
          task vcnup3
            trigger vcnup2 eq complete and vcnli3 eq complete
          task vcnup4
            trigger vcnup3 eq complete and vcnli4 eq complete
          task vcnmem2
            trigger vcnup4 eq complete
            edit DATEMASK '22.*.*'
          task vcnupm2
            trigger vcnmem2 eq complete
            edit DATEMASK '22.*.*'
        endfamily
        family ob
          trigger oc eq complete or oc eq aborted
          time 05:00
          family stage1
            task vobli1
            task vobli2
            task vobup1
              trigger vobli1 eq complete
            task vobup2
              trigger vobup1 eq complete and vobli2 eq complete
            task vobnobs
              trigger vobup2 eq complete
              edit DATEMASK '03.*.*'
            task vobmem
              trigger ./vobup2 eq complete
              edit DATEMASK '03.*.*'
            task vobupm
              trigger vobmem eq complete
              edit DATEMASK '03.*.*'
          endfamily
          family stage2
            trigger ./stage1 eq complete
            edit EMOS_BASE '00'
            task vobli1
            task vobli2
            task vobup1
              trigger vobli1 eq complete
            task vobup2
              trigger vobup1 eq complete and vobli2 eq complete
            task vobnobs
              trigger vobup2 eq complete
              edit DATEMASK '03.*.*'
            task vobmem
              trigger ./vobup2 eq complete
              edit DATEMASK '03.*.*'
            task vobupm
              trigger vobmem eq complete
              edit DATEMASK '03.*.*'
          endfamily
          task vobcrd
            edit DATEMASK '28.*.*'
          task vobcrm
            edit DATEMASK '28.*.*'
        endfamily
        family em
          task vemli
            late -c +02:00
          task vemup
            trigger vemli eq complete
          task vemcrd
            edit DATEMASK '28.*.*'
          task vemcrm
            edit DATEMASK '28.*.*'
          task vemmem
            trigger vemup eq complete
            edit DATEMASK '11.*.*'
          task vemupm
            trigger vemmem eq complete
            edit DATEMASK '11.*.*'
        endfamily
        family es
          trigger em eq complete
          task vesli
            late -c +02:00
          task vesup
            trigger vesli eq complete
          task vescrd
            edit DATEMASK '28.*.*'
          task vescrm
            edit DATEMASK '28.*.*'
          task vesmem
            trigger vesup eq complete
            edit DATEMASK '11.*.*'
          task vesupm
            trigger vesmem eq complete
            edit DATEMASK '11.*.*'
        endfamily
        family cv
          trigger es eq complete
          task vcvli
            late -c +02:00
          task vcvup
            trigger vcvli eq complete
          task vcvcrd
            edit DATEMASK '28.*.*'
          task vcvcrm
            edit DATEMASK '28.*.*'
          task vcvmem
            trigger vcvup eq complete
            edit DATEMASK '11.*.*'
          task vcvupm
            trigger vcvmem eq complete
            edit DATEMASK '11.*.*'
        endfamily
        family wv
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ../../../main/00/fc/model eq complete)
          task glver
          task glupr
            trigger glver eq complete
          task glcrr
            edit DATEMASK '28.*.*'
          task ff10ver
            trigger glver eq complete
          task ff10upr
            trigger ff10ver eq complete and glupr eq complete
          task glmem
            trigger ff10upr eq complete
            edit DATEMASK '11.*.*'
          task glupm
            trigger glmem eq complete
            edit DATEMASK '11.*.*'
          task glcrm
            edit DATEMASK '28.*.*'
        endfamily
        family sf
          trigger fc eq complete or fc eq aborted
          task vseli1
          task vseli2
          task vseup1
            trigger vseli1 eq complete
          task vseup2
            trigger vseup1 eq complete and vseli2 eq complete
          task vsecrd
            edit DATEMASK '28.*.*'
          task vsecrm
            edit DATEMASK '28.*.*'
          task vsemem
            trigger vseup2 eq complete
            edit DATEMASK '12.*.*'
          task vseupm
            trigger vsemem eq complete
            edit DATEMASK '12.*.*'
        endfamily
      endfamily
    endfamily
    family 12bc
      edit EMOS_BASE '12'
      edit EDFAMILY 'false'
      edit INISTREAM 'DA'
      edit REINITIALIZE 'false'
      edit REINIEXPVER '0046'
      edit INIEXPVER '0045'
      edit REINIPERIOD_4D '6'
      edit REINISTREAM 'DA'
      edit REINIOFFSET '6'
      edit RUN_HOURLY_PP '0'
      edit PERIOD_4D '6'
      edit WINDOW_LENGTH_4D '6'
      edit WINDOW_OFFSET_4D '3'
      edit MXUP_TRAJ '3'
      edit PARALLEL '4'
      edit LANONLY 'false'
      edit ACCOUNT 'opebcmis'
      edit STREAM 'SCDA'
      edit EMOS_STREAM 'SCDA'
      edit DELTA_DAY '0'
      family verif
        trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/12/fc eq complete)
        edit QUEUE 'emos-e'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        edit SMSFILES '/home/ma/emos/def/e_36r1/bcverif'
        edit SCORE_STREAM 'BC_0046'
        inlimit /limits:linux_cluster
        task create_daily
        task create_monthly
          edit DATEMASK '28.*.*'
        task clim
          trigger create_daily eq complete
        task non_clim
          trigger create_daily eq complete
        task clim_update
          trigger clim eq complete
        task non_clim_update
          trigger non_clim eq complete and clim eq complete
        task plot
          trigger non_clim_update eq complete
        task mean
          trigger non_clim_update eq complete
          edit DATEMASK '10.*.*'
        task mean_update
          trigger mean eq complete
          edit DATEMASK '10.*.*'
      endfamily
    endfamily
    family 18bc
      edit EMOS_BASE '18'
      edit EDFAMILY 'false'
      edit INISTREAM 'DA'
      edit REINITIALIZE 'false'
      edit REINIEXPVER '0046'
      edit INIEXPVER '0045'
      edit REINIPERIOD_4D '6'
      edit REINISTREAM 'DA'
      edit REINIOFFSET '6'
      edit RUN_HOURLY_PP '0'
      edit PERIOD_4D '6'
      edit WINDOW_LENGTH_4D '6'
      edit WINDOW_OFFSET_4D '3'
      edit MXUP_TRAJ '3'
      edit PARALLEL '4'
      edit LANONLY 'false'
      edit ACCOUNT 'opebcmis'
      edit STREAM 'SCDA'
      edit EMOS_STREAM 'SCDA'
      edit DELTA_DAY '0'
      family archive
        trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ../../main/18bc/an eq complete and ( ../../main/18bc/fc/model:step ge 90 or ../../main/18bc/fc/model eq complete))
        edit FSFAMILY 'mars'
        edit SMSFILES '/home/ma/emos/def/e_36r1/archive'
        edit EMOS_TYPE 'an'
        inlimit /limits:mars
        inlimit /e_36r1/limits:mars
        task ansfc
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task anml
          trigger ansfc:fdb or ansfc eq complete
          edit EDFAMILY 'true'
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          inlimit /e_36r1/limits:ml
          label files ""
          label info ""
          event fdb
        task anpl
          trigger anml:fdb or anml eq complete
          edit FCLENGTH '9'
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task anwave
          trigger ansfc:fdb or ansfc eq complete
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task anil
          trigger anpl:fdb or anpl eq complete
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task derivedp
          trigger ./ansfc eq complete
          edit STEPS '0'
          label files ""
          label info ""
          event fdb
        family obs
          edit FSFAMILY 'an'
          inlimit /e_36r1/limits:obs
          task ai
            trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/18bc/an/4dvar/ifstraj:finaltraj or /e_36r1/main/18bc/an/4dvar eq complete))
            edit TYPE 'ai'
            label files ""
            label info ""
            event fdb
          task af
            trigger ../../fb/feedback eq complete
            edit TYPE 'af'
            label files ""
            label info ""
            event fdb
          task fb
            trigger ../../fb/feedback eq complete
            edit TYPE 'fb'
            label files ""
            label info ""
            event fdb
        endfamily
        task ice_nmc
          trigger ansfc:fdb or ansfc eq complete
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task sst
          trigger ansfc:fdb or ansfc eq complete
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        family fc024
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/18bc/prod/024 eq complete)
          edit STEPS '0/3/06/09/12/15/18/21/24'
          edit EMOS_TYPE 'fc'
          task fcsfc
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcml
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            inlimit /e_36r1/limits:ml
            label files ""
            label info ""
            event fdb
          task fcpl
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcil
            trigger fcpl:fdb or fcpl eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task derivedp
            trigger fcil:fdb or fcil eq complete
            label files ""
            label info ""
            event fdb
        endfamily
        family fc048
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/18bc/prod/048 eq complete)
          edit STEPS '27/30/33/36/39/42/45/48'
          edit EMOS_TYPE 'fc'
          task fcsfc
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcml
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            inlimit /e_36r1/limits:ml
            label files ""
            label info ""
            event fdb
          task fcpl
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcil
            trigger fcpl:fdb or fcpl eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task derivedp
            trigger fcil:fdb or fcil eq complete
            label files ""
            label info ""
            event fdb
        endfamily
        family fc072
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/18bc/prod/072 eq complete)
          edit STEPS '51/54/57/60/63/66/69/72'
          edit EMOS_TYPE 'fc'
          task fcsfc
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcml
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            inlimit /e_36r1/limits:ml
            label files ""
            label info ""
            event fdb
          task fcpl
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcil
            trigger fcpl:fdb or fcpl eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task derivedp
            trigger fcil:fdb or fcil eq complete
            label files ""
            label info ""
            event fdb
        endfamily
        family fc090
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/18bc/prod/090 eq complete)
          edit STEPS '75/78/81/84/87/90'
          edit EMOS_TYPE 'fc'
          task fcsfc
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcml
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            inlimit /e_36r1/limits:ml
            label files ""
            label info ""
            event fdb
          task fcpl
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcil
            trigger fcpl:fdb or fcpl eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task derivedp
            trigger fcil:fdb or fcil eq complete
            label files ""
            label info ""
            event fdb
        endfamily
        task fc12g
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/18bc/fc/model eq complete)
          edit STEPS_SPECTRAL '6/12/18/24/30/36/42/48/54/60/66/72/78/84/90'
          edit STEPS '0/6/12/18/24/30/36/42/48/54/60/66/72/78/84/90'
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      family fb
        trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/18bc/an eq complete)
        edit FSFAMILY 'an'
        edit SMSFILES '/home/ma/emos/def/e_36r1/fb'
        task sync_rstrhbias
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
        task biassave
          trigger sync_rstrhbias eq complete
        family feedback
          edit ACCOUNT 'opebcan'
          edit NPES '8'
          edit FBNEW2OLD '1'
          edit LCONV '1'
          edit LSATOB '1'
          edit LREO3 '1'
          edit LAEOLUS '0'
          edit LSSMI '1'
          edit LSSMIS '1'
          edit LAMSRE '1'
          edit LTMI '1'
          edit LHIRS '1'
          edit LMSU '1'
          edit LSSU '1'
          edit LAMSUA '1'
          edit LAMSUB '1'
          edit LMHS '1'
          edit LGEOS '1'
          edit LSCAT '1'
          edit LAIRS '1'
          edit LVTPR1 '0'
          edit LVTPR2 '0'
          edit LGPSRO '1'
          edit LIASI '1'
          edit LMERIS '1'
          task revmatchup
            edit NPES '128'
            edit THREADS '2'
          task matchup
            trigger revmatchup eq complete
            edit NPES '128'
            edit THREADS '2'
          task odbcmp_prepare
            edit THREADS '1'
            edit NPES '1'
            edit MEM '1000'
            edit CPUTIME '1000'
            edit PRIORITY '90'
          family odb_compress
            trigger odbcmp_prepare eq complete and matchup eq complete
            edit THREADS '2'
            edit NPES '1'
            task odbcmp_conv
            task odbcmp_surf_conv
            task odbcmp_airs
            task odbcmp_iasi
            task odbcmp_hirs
            task odbcmp_msu
            task odbcmp_ssu
            task odbcmp_amsua
            task odbcmp_amsub
            task odbcmp_mhs
            task odbcmp_satob
            task odbcmp_reo3
            task odbcmp_ssmi
            task odbcmp_ssmis
            task odbcmp_amsre
            task odbcmp_tmi
            task odbcmp_scatt
            task odbcmp_geos
            task odbcmp_gpsro
            task odbcmp_meris
          endfamily
          family odb2bufr
            trigger matchup eq complete
            edit NPES '1'
            edit SUBTYPES_INCLUDE ''
            edit SUBTYPES_EXCLUDE ''
            task o2b_conv
              edit SUBTYPES_EXCLUDE '85 82 83 87'
              edit NPES '2'
            task o2b_surf_conv
              edit SUBTYPES_EXCLUDE '142 143 144 145 85 82 83 87'
              edit NPES '2'
            task o2b_airs
              edit SUBTYPES_INCLUDE '57'
              edit NPES '8'
              edit THREADS '2'
            task o2b_iasi
              edit SUBTYPES_INCLUDE '240'
              edit NPES '8'
              edit THREADS '2'
            task o2b_hirs
              edit SUBTYPES_INCLUDE '54 55'
              edit NPES '2'
            task o2b_msu
              edit SUBTYPES_INCLUDE '54 55'
            task o2b_ssu
              edit SUBTYPES_INCLUDE '54 55'
            task o2b_amsua
              edit SUBTYPES_INCLUDE '54 55'
              edit NPES '2'
            task o2b_amsub
              edit SUBTYPES_INCLUDE '54 55'
              edit NPES '2'
            task o2b_mhs
              edit SUBTYPES_INCLUDE '54 55'
              edit NPES '2'
            task o2b_satob
              edit SUBTYPES_INCLUDE '82 83 87'
              edit NPES '2'
            task o2b_reo3
              edit SUBTYPES_INCLUDE '205 206 207 208 215'
              edit NPES '4'
              edit THREADS '2'
            task o2b_ssmi
              edit SUBTYPES_INCLUDE '55 127'
            task o2b_ssmis
              edit SUBTYPES_INCLUDE '49'
            task o2b_amsre
              edit SUBTYPES_INCLUDE '59'
            task o2b_tmi
              edit SUBTYPES_INCLUDE '129'
            task o2b_scatt
              edit SUBTYPES_INCLUDE '122 137 139'
            task o2b_geos
              edit SUBTYPES_INCLUDE '189 89 76 77 78'
              edit NPES '2'
            task o2b_gpsro
              edit SUBTYPES_INCLUDE '250'
            task o2b_meris
              edit SUBTYPES_INCLUDE '212'
          endfamily
          task mergebufr
            trigger odb2bufr eq complete
        endfamily
        family obstat
          trigger feedback/odb2bufr eq complete
          edit NPES '1'
          task obstat_conv
          task obstat_surf_conv
          task obstat_airs
            edit NPES '8'
            edit THREADS '2'
          task obstat_iasi
            edit NPES '8'
            edit THREADS '2'
          task obstat_hirs
            edit NPES '4'
          task obstat_msu
          task obstat_ssu
          task obstat_amsua
            edit NPES '2'
          task obstat_amsub
          task obstat_mhs
          task obstat_satob
          task obstat_reo3
          task obstat_ssmi
            edit NPES '2'
          task obstat_ssmis
            edit NPES '2'
          task obstat_amsre
            edit NPES '2'
          task obstat_tmi
            edit NPES '2'
          task obstat_scatt
          task obstat_geos
            edit NPES '2'
          task obstat_gpsro
            edit NPES '2'
          task obstat_meris
            edit NPES '2'
        endfamily
        task fdbksave
          trigger feedback eq complete and obstat eq complete
        task clean_odbcmp
          trigger feedback eq complete and obstat eq complete and fdbksave eq complete
        task wavesave
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ../../../main/18bc/an/4dvar eq complete)
        task obstat_merge
          trigger obstat eq complete
          edit WSHOST 'bee01'
          edit PRIORITY '51'
      endfamily
      task clean
        trigger fb eq complete
        edit FSFAMILY 'an'
      family verif
        trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/18bc/fc eq complete)
        edit QUEUE 'emos-e'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        edit SMSFILES '/home/ma/emos/def/e_36r1/bcverif'
        edit SCORE_STREAM 'BC_0046'
        inlimit /limits:linux_cluster
        task create_daily
        task create_monthly
          edit DATEMASK '28.*.*'
        task clim
          trigger create_daily eq complete
        task non_clim
          trigger create_daily eq complete
        task clim_update
          trigger clim eq complete
        task non_clim_update
          trigger non_clim eq complete and clim eq complete
        task plot
          trigger non_clim_update eq complete
        task mean
          trigger non_clim_update eq complete
          edit DATEMASK '10.*.*'
        task mean_update
          trigger mean eq complete
          edit DATEMASK '10.*.*'
      endfamily
    endfamily
    family 00bc
      edit EMOS_BASE '00'
      edit EDFAMILY 'false'
      edit INISTREAM 'DA'
      edit REINITIALIZE 'false'
      edit REINIEXPVER '0046'
      edit INIEXPVER '0045'
      edit REINIPERIOD_4D '6'
      edit REINISTREAM 'DA'
      edit REINIOFFSET '6'
      edit RUN_HOURLY_PP '0'
      edit PERIOD_4D '6'
      edit WINDOW_LENGTH_4D '6'
      edit WINDOW_OFFSET_4D '3'
      edit MXUP_TRAJ '3'
      edit PARALLEL '4'
      edit LANONLY 'false'
      edit ACCOUNT 'opebcmis'
      edit STREAM 'SCDA'
      edit EMOS_STREAM 'SCDA'
      edit DELTA_DAY '1'
      family verif
        trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/00/fc eq complete)
        edit QUEUE 'emos-e'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        edit SMSFILES '/home/ma/emos/def/e_36r1/bcverif'
        edit SCORE_STREAM 'BC_0046'
        inlimit /limits:linux_cluster
        task create_daily
        task create_monthly
          edit DATEMASK '28.*.*'
        task clim
          trigger create_daily eq complete
        task non_clim
          trigger create_daily eq complete
        task clim_update
          trigger clim eq complete
        task non_clim_update
          trigger non_clim eq complete and clim eq complete
        task plot
          trigger non_clim_update eq complete
        task mean
          trigger non_clim_update eq complete
          edit DATEMASK '10.*.*'
        task mean_update
          trigger mean eq complete
          edit DATEMASK '10.*.*'
      endfamily
    endfamily
    family 06bc
      edit EMOS_BASE '06'
      edit EDFAMILY 'false'
      edit INISTREAM 'DA'
      edit REINITIALIZE 'false'
      edit REINIEXPVER '0046'
      edit INIEXPVER '0045'
      edit REINIPERIOD_4D '6'
      edit REINISTREAM 'DA'
      edit REINIOFFSET '6'
      edit RUN_HOURLY_PP '0'
      edit PERIOD_4D '6'
      edit WINDOW_LENGTH_4D '6'
      edit WINDOW_OFFSET_4D '3'
      edit MXUP_TRAJ '3'
      edit PARALLEL '4'
      edit LANONLY 'false'
      edit ACCOUNT 'opebcmis'
      edit STREAM 'SCDA'
      edit EMOS_STREAM 'SCDA'
      edit DELTA_DAY '1'
      family archive
        trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ../../main/06bc/an eq complete and ( ../../main/06bc/fc/model:step ge 90 or ../../main/06bc/fc/model eq complete))
        edit FSFAMILY 'mars'
        edit SMSFILES '/home/ma/emos/def/e_36r1/archive'
        edit EMOS_TYPE 'an'
        inlimit /limits:mars
        inlimit /e_36r1/limits:mars
        task ansfc
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task anml
          trigger ansfc:fdb or ansfc eq complete
          edit EDFAMILY 'true'
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          inlimit /e_36r1/limits:ml
          label files ""
          label info ""
          event fdb
        task anpl
          trigger anml:fdb or anml eq complete
          edit FCLENGTH '9'
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task anwave
          trigger ansfc:fdb or ansfc eq complete
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task anil
          trigger anpl:fdb or anpl eq complete
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task derivedp
          trigger ./ansfc eq complete
          edit STEPS '0'
          label files ""
          label info ""
          event fdb
        family obs
          edit FSFAMILY 'an'
          inlimit /e_36r1/limits:obs
          task highres
            edit QUEUE 'emos-e'
            edit QUEUE_EPILOG 'emos-exp'
            edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
            edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
            edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
            edit LOGDIR '/vol/emos_nc/output'
            edit SMSOUT '/vol/emos_nc/output'
            inlimit /limits:linux_cluster
          task ai
            trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ( /e_36r1/main/06bc/an/4dvar/ifstraj:finaltraj or /e_36r1/main/06bc/an/4dvar eq complete))
            edit TYPE 'ai'
            label files ""
            label info ""
            event fdb
          task af
            trigger ../../fb/feedback eq complete
            edit TYPE 'af'
            label files ""
            label info ""
            event fdb
          task fb
            trigger ../../fb/feedback eq complete
            edit TYPE 'fb'
            label files ""
            label info ""
            event fdb
        endfamily
        task ice_nmc
          trigger ansfc:fdb or ansfc eq complete
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task sst
          trigger ansfc:fdb or ansfc eq complete
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        family fc024
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/06bc/prod/024 eq complete)
          edit STEPS '0/3/06/09/12/15/18/21/24'
          edit EMOS_TYPE 'fc'
          task fcsfc
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcml
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            inlimit /e_36r1/limits:ml
            label files ""
            label info ""
            event fdb
          task fcpl
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcil
            trigger fcpl:fdb or fcpl eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task derivedp
            trigger fcil:fdb or fcil eq complete
            label files ""
            label info ""
            event fdb
        endfamily
        family fc048
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/06bc/prod/048 eq complete)
          edit STEPS '27/30/33/36/39/42/45/48'
          edit EMOS_TYPE 'fc'
          task fcsfc
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcml
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            inlimit /e_36r1/limits:ml
            label files ""
            label info ""
            event fdb
          task fcpl
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcil
            trigger fcpl:fdb or fcpl eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task derivedp
            trigger fcil:fdb or fcil eq complete
            label files ""
            label info ""
            event fdb
        endfamily
        family fc072
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/06bc/prod/072 eq complete)
          edit STEPS '51/54/57/60/63/66/69/72'
          edit EMOS_TYPE 'fc'
          task fcsfc
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcml
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            inlimit /e_36r1/limits:ml
            label files ""
            label info ""
            event fdb
          task fcpl
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcil
            trigger fcpl:fdb or fcpl eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task derivedp
            trigger fcil:fdb or fcil eq complete
            label files ""
            label info ""
            event fdb
        endfamily
        family fc090
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/06bc/prod/090 eq complete)
          edit STEPS '75/78/81/84/87/90'
          edit EMOS_TYPE 'fc'
          task fcsfc
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcml
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            inlimit /e_36r1/limits:ml
            label files ""
            label info ""
            event fdb
          task fcpl
            trigger fcsfc:fdb or fcsfc eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task fcil
            trigger fcpl:fdb or fcpl eq complete
            inlimit /limits:mars
            inlimit /e_36r1/limits:mars
            label files ""
            label info ""
            event fdb
          task derivedp
            trigger fcil:fdb or fcil eq complete
            label files ""
            label info ""
            event fdb
        endfamily
        task fc12g
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/06bc/fc/model eq complete)
          edit STEPS_SPECTRAL '6/12/18/24/30/36/42/48/54/60/66/72/78/84/90'
          edit STEPS '0/6/12/18/24/30/36/42/48/54/60/66/72/78/84/90'
          inlimit /limits:mars
          inlimit /e_36r1/limits:mars
          label files ""
          label info ""
          event fdb
        task rmfdb
          trigger ansfc eq complete and anml eq complete and anpl eq complete and anwave eq complete and obs eq complete
      endfamily
      family fb
        trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/06bc/an eq complete)
        edit FSFAMILY 'an'
        edit SMSFILES '/home/ma/emos/def/e_36r1/fb'
        task sync_rstrhbias
          edit SMSFILES '/home/ma/emos/def/e_36r1/smsfiles'
        task biassave
          trigger sync_rstrhbias eq complete
        family feedback
          edit ACCOUNT 'opebcan'
          edit NPES '8'
          edit FBNEW2OLD '1'
          edit LCONV '1'
          edit LSATOB '1'
          edit LREO3 '1'
          edit LAEOLUS '0'
          edit LSSMI '1'
          edit LSSMIS '1'
          edit LAMSRE '1'
          edit LTMI '1'
          edit LHIRS '1'
          edit LMSU '1'
          edit LSSU '1'
          edit LAMSUA '1'
          edit LAMSUB '1'
          edit LMHS '1'
          edit LGEOS '1'
          edit LSCAT '1'
          edit LAIRS '1'
          edit LVTPR1 '0'
          edit LVTPR2 '0'
          edit LGPSRO '1'
          edit LIASI '1'
          edit LMERIS '1'
          task revmatchup
            edit NPES '128'
            edit THREADS '2'
          task matchup
            trigger revmatchup eq complete
            edit NPES '128'
            edit THREADS '2'
          task odbcmp_prepare
            edit THREADS '1'
            edit NPES '1'
            edit MEM '1000'
            edit CPUTIME '1000'
            edit PRIORITY '90'
          family odb_compress
            trigger odbcmp_prepare eq complete and matchup eq complete
            edit THREADS '2'
            edit NPES '1'
            task odbcmp_conv
            task odbcmp_surf_conv
            task odbcmp_airs
            task odbcmp_iasi
            task odbcmp_hirs
            task odbcmp_msu
            task odbcmp_ssu
            task odbcmp_amsua
            task odbcmp_amsub
            task odbcmp_mhs
            task odbcmp_satob
            task odbcmp_reo3
            task odbcmp_ssmi
            task odbcmp_ssmis
            task odbcmp_amsre
            task odbcmp_tmi
            task odbcmp_scatt
            task odbcmp_geos
            task odbcmp_gpsro
            task odbcmp_meris
          endfamily
          family odb2bufr
            trigger matchup eq complete
            edit NPES '1'
            edit SUBTYPES_INCLUDE ''
            edit SUBTYPES_EXCLUDE ''
            task o2b_conv
              edit SUBTYPES_EXCLUDE '85 82 83 87'
              edit NPES '2'
            task o2b_surf_conv
              edit SUBTYPES_EXCLUDE '142 143 144 145 85 82 83 87'
              edit NPES '2'
            task o2b_airs
              edit SUBTYPES_INCLUDE '57'
              edit NPES '8'
              edit THREADS '2'
            task o2b_iasi
              edit SUBTYPES_INCLUDE '240'
              edit NPES '8'
              edit THREADS '2'
            task o2b_hirs
              edit SUBTYPES_INCLUDE '54 55'
              edit NPES '2'
            task o2b_msu
              edit SUBTYPES_INCLUDE '54 55'
            task o2b_ssu
              edit SUBTYPES_INCLUDE '54 55'
            task o2b_amsua
              edit SUBTYPES_INCLUDE '54 55'
              edit NPES '2'
            task o2b_amsub
              edit SUBTYPES_INCLUDE '54 55'
              edit NPES '2'
            task o2b_mhs
              edit SUBTYPES_INCLUDE '54 55'
              edit NPES '2'
            task o2b_satob
              edit SUBTYPES_INCLUDE '82 83 87'
              edit NPES '2'
            task o2b_reo3
              edit SUBTYPES_INCLUDE '205 206 207 208 215'
              edit NPES '4'
              edit THREADS '2'
            task o2b_ssmi
              edit SUBTYPES_INCLUDE '55 127'
            task o2b_ssmis
              edit SUBTYPES_INCLUDE '49'
            task o2b_amsre
              edit SUBTYPES_INCLUDE '59'
            task o2b_tmi
              edit SUBTYPES_INCLUDE '129'
            task o2b_scatt
              edit SUBTYPES_INCLUDE '122 137 139'
            task o2b_geos
              edit SUBTYPES_INCLUDE '189 89 76 77 78'
              edit NPES '2'
            task o2b_gpsro
              edit SUBTYPES_INCLUDE '250'
            task o2b_meris
              edit SUBTYPES_INCLUDE '212'
          endfamily
          task mergebufr
            trigger odb2bufr eq complete
        endfamily
        family obstat
          trigger feedback/odb2bufr eq complete
          edit NPES '1'
          task obstat_conv
          task obstat_surf_conv
          task obstat_airs
            edit NPES '8'
            edit THREADS '2'
          task obstat_iasi
            edit NPES '8'
            edit THREADS '2'
          task obstat_hirs
            edit NPES '4'
          task obstat_msu
          task obstat_ssu
          task obstat_amsua
            edit NPES '2'
          task obstat_amsub
          task obstat_mhs
          task obstat_satob
          task obstat_reo3
          task obstat_ssmi
            edit NPES '2'
          task obstat_ssmis
            edit NPES '2'
          task obstat_amsre
            edit NPES '2'
          task obstat_tmi
            edit NPES '2'
          task obstat_scatt
          task obstat_geos
            edit NPES '2'
          task obstat_gpsro
            edit NPES '2'
          task obstat_meris
            edit NPES '2'
        endfamily
        task fdbksave
          trigger feedback eq complete and obstat eq complete
        task clean_odbcmp
          trigger feedback eq complete and obstat eq complete and fdbksave eq complete
        task wavesave
          trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and ../../../main/06bc/an/4dvar eq complete)
        task obstat_merge
          trigger obstat eq complete
          edit WSHOST 'bee01'
          edit PRIORITY '51'
      endfamily
      task clean
        trigger fb eq complete
        edit FSFAMILY 'an'
      family verif
        trigger ( /e_36r1/main:YMD gt /e_36r1/lag:YMD) or ( /e_36r1/main:YMD eq /e_36r1/lag:YMD and /e_36r1/main/06bc/fc eq complete)
        edit QUEUE 'emos-e'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        edit SMSFILES '/home/ma/emos/def/e_36r1/bcverif'
        edit SCORE_STREAM 'BC_0046'
        inlimit /limits:linux_cluster
        task create_daily
        task create_monthly
          edit DATEMASK '28.*.*'
        task clim
          trigger create_daily eq complete
        task non_clim
          trigger create_daily eq complete
        task clim_update
          trigger clim eq complete
        task non_clim_update
          trigger non_clim eq complete and clim eq complete
        task plot
          trigger non_clim_update eq complete
        task mean
          trigger non_clim_update eq complete
          edit DATEMASK '10.*.*'
        task mean_update
          trigger mean eq complete
          edit DATEMASK '10.*.*'
      endfamily
    endfamily
    task checkarchive
      trigger ./06 eq complete and ./12 eq complete and ./12dc eq complete and ./18 eq complete and ./00 eq complete and ./00dc eq complete
      edit QUEUE 'emos-e'
      edit QUEUE_EPILOG 'emos-exp'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit FSFAMILY 'mars'
      edit EMOS_TYPE 'an'
      edit OSTREAN 'DCDA DCWV'
      edit WSHOST 'linux_cluster'
      inlimit /limits:linux_cluster
    task checkarchivebc
      trigger ./18bc eq complete and ./06bc eq complete
      edit QUEUE 'emos-e'
      edit QUEUE_EPILOG 'emos-exp'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit STREAM 'SCDA'
      edit FSFAMILY 'mars'
      edit EMOS_TYPE 'an'
      edit WSTREAN 'DCDA DCWV'
      edit WSHOST 'linux_cluster'
      inlimit /limits:linux_cluster
    task logfilestore
      edit QUEUE 'emos-e'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit SMSFILES '/home/ma/emos/def/e_36r1'
      edit LOGTASK 'lag'
      edit KEEPLOGS 'no'
      edit LOGSTOECFS 'no'
      edit MOVE_TO_ECFS 'yes'
    task logfiles
      late -s +00:05 -c +00:15
      trigger /e_36r1/main/make eq complete and ./00dc eq complete and ./06 eq complete and ./12 eq complete and ./12dc eq complete and ./18 eq complete and ./00 eq complete and ./12bc eq complete and ./18bc eq complete and ./00bc eq complete and ./06bc eq complete and checkarchive eq complete and checkarchivebc eq complete and logfilestore eq complete
      edit QUEUE 'emos-e'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit SMSFILES '/home/ma/emos/def/e_36r1'
      edit LOGTASK 'lag'
      edit MOVE_TO_ECFS 'no'
      time 11:45
  endfamily
  family verify
    repeat date YMD 20091118 20121212 1
    edit VERIFY 'verify_test'
    edit UPDATE_WEB_NOW 'yes'
    edit SMSFILES '/home/ma/emos/def/e_36r1/verify'
    edit FORECAST_DELTA '1'
    edit RUN_PARALLEL '0'
    edit EXPVERV '0046'
    edit EXPVERP '0046'
    edit EXPVERF '0046'
    edit CLASSV 'od'
    edit CLASSP 'od'
    edit CLASSF 'od'
    edit QUEUE 'emos-e'
    edit QUEUE_EPILOG 'emos-exp'
    edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
    edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
    edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
    edit LOGDIR '/vol/emos_nc/output'
    edit SMSOUT '/vol/emos_nc/output'
    inlimit /limits:mars
    inlimit /limits:linux_cluster
    family 12
      edit EMOS_BASE '12'
      family forecast
        trigger ( /e_36r1/lag:YMD gt /e_36r1/verify:YMD) or ( /e_36r1/lag:YMD eq /e_36r1/verify:YMD and ../../lag/12/archive eq complete)
        edit QUEUE 'emos-e'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        edit SCORE_STREAM '36r1_0046'
        edit OPER_STREAM 'OPER'
        edit INDEX_SIZE '102400'
        inlimit /limits:linux_cluster
        task create_score
        family clim_levels
          trigger create_score eq complete
          task fc_clim_levels
          task update_score
            trigger fc_clim_levels eq complete
        endfamily
        family non_clim_levels
          trigger create_score eq complete
          task fc_non_clim_levels
          task update_score
            trigger fc_non_clim_levels eq complete and ../clim_levels eq complete
        endfamily
        family others
          trigger create_score eq complete
          task fc_others
          task update_score
            trigger fc_others eq complete and ../non_clim_levels eq complete
        endfamily
      endfamily
      family observations
        trigger forecast eq complete
        edit QUEUE 'emos-e'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        edit FORECAST_DELTA '3'
        edit SCORE_STREAM '36r1_0046_OB'
        edit OPER_STREAM 'O_ECMWF'
        edit INDEX_SIZE '102400'
        inlimit /limits:linux_cluster
        task create_score
        family clim_levels
          trigger create_score eq complete
          task ob_clim_levels
          task update_score
            trigger ob_clim_levels eq complete
        endfamily
        family non_clim_levels
          trigger create_score eq complete
          task ob_non_clim_levels
          task update_score
            trigger ob_non_clim_levels eq complete and ../clim_levels eq complete
        endfamily
      endfamily
      family waves
        trigger forecast eq complete
        edit QUEUE 'emos-e'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        edit INDEX_SIZE '102400'
        edit SCORE_STREAM '36r1_0046_WV'
        edit OPER_STREAM 'WAVE_G'
        inlimit /limits:linux_cluster
        family height
          edit STREAM 'WV'
          task create_score
          task wv_height
            trigger create_score eq complete
          task update_score
            trigger wv_height eq complete
        endfamily
        family wind
          trigger height eq complete
          task wv_wind_all
          task update_score
            trigger wv_wind_all eq complete
        endfamily
        family wind_over_sea
          trigger wind eq complete
          edit INDEX_SIZE '102400'
          edit SCORE_STREAM '36r1_0046_WV_SEA'
          task create_score
          task wv_wind_sea
            trigger create_score eq complete
          task update_score
            trigger wv_wind_sea eq complete
        endfamily
      endfamily
    endfamily
    family 00
      trigger ./12 eq complete
      edit EMOS_BASE '00'
      family forecast
        trigger ( /e_36r1/lag:YMD gt /e_36r1/verify:YMD) or ( /e_36r1/lag:YMD eq /e_36r1/verify:YMD and ../../lag/00/archive eq complete)
        edit QUEUE 'emos-e'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        edit SCORE_STREAM '36r1_0046'
        edit OPER_STREAM 'OPER'
        edit INDEX_SIZE '102400'
        inlimit /limits:linux_cluster
        task create_score
        family clim_levels
          trigger create_score eq complete
          task fc_clim_levels
          task update_score
            trigger fc_clim_levels eq complete
        endfamily
        family non_clim_levels
          trigger create_score eq complete
          task fc_non_clim_levels
          task update_score
            trigger fc_non_clim_levels eq complete and ../clim_levels eq complete
        endfamily
        family others
          trigger create_score eq complete
          task fc_others
          task update_score
            trigger fc_others eq complete and ../non_clim_levels eq complete
        endfamily
      endfamily
      family observations
        trigger forecast eq complete
        edit QUEUE 'emos-e'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        edit FORECAST_DELTA '3'
        edit SCORE_STREAM '36r1_0046_OB'
        edit OPER_STREAM 'O_ECMWF'
        edit INDEX_SIZE '102400'
        inlimit /limits:linux_cluster
        task create_score
        family clim_levels
          trigger create_score eq complete
          task ob_clim_levels
          task update_score
            trigger ob_clim_levels eq complete
        endfamily
        family non_clim_levels
          trigger create_score eq complete
          task ob_non_clim_levels
          task update_score
            trigger ob_non_clim_levels eq complete and ../clim_levels eq complete
        endfamily
      endfamily
      family waves
        trigger forecast eq complete
        edit QUEUE 'emos-e'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        edit INDEX_SIZE '102400'
        edit SCORE_STREAM '36r1_0046_WV'
        edit OPER_STREAM 'WAVE_G'
        inlimit /limits:linux_cluster
        family height
          edit STREAM 'WV'
          task create_score
          task wv_height
            trigger create_score eq complete
          task update_score
            trigger wv_height eq complete
        endfamily
        family wind
          trigger height eq complete
          task wv_wind_all
          task update_score
            trigger wv_wind_all eq complete
        endfamily
        family wind_over_sea
          trigger wind eq complete
          edit INDEX_SIZE '102400'
          edit SCORE_STREAM '36r1_0046_WV_SEA'
          task create_score
          task wv_wind_sea
            trigger create_score eq complete
          task update_score
            trigger wv_wind_sea eq complete
        endfamily
      endfamily
    endfamily
    family plot
      trigger ./00 eq complete and ./12 eq complete
      edit WSHOST 'linux_cluster'
      edit QUEUE 'serial'
      limit plot 3
      inlimit plot
      family forecast
        edit SCORE_STREAM '36r1_0046'
        edit OPER_STREAM 'OPER'
        edit FORECAST_DELTA '1'
        task fc_scocom
        task fc_scocat
        task fc_pltime
        task fc_error
      endfamily
      family observations
        edit SCORE_STREAM '36r1_0046_OB'
        edit OPER_STREAM 'O_ECMWF'
        edit FORECAST_DELTA '3'
        task ob_scocom
        task ob_scocat
        task ob_pltime
        task ob_error
      endfamily
      family waves
        edit SCORE_STREAM '36r1_0046_WV'
        edit OPER_STREAM 'WAVE_G'
        edit FORECAST_DELTA '1'
        task wv_scocom
        task wv_scocat
        task wv_pltime
        task wv_error
      endfamily
    endfamily
    task logfilestore
      edit QUEUE 'emos-e'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit SMSFILES '/home/ma/emos/def/e_36r1'
      edit LOGTASK 'verify'
      edit KEEPLOGS 'no'
      edit LOGSTOECFS 'no'
      edit MOVE_TO_ECFS 'yes'
    task logfiles
      late -s +00:05 -c +00:15
      trigger ./00 eq complete and ./12 eq complete and plot eq complete and logfilestore eq complete
      edit QUEUE 'emos-e'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit SMSFILES '/home/ma/emos/def/e_36r1'
      edit LOGTASK 'verify'
      edit MOVE_TO_ECFS 'no'
      edit LOGSTOECFS 'no'
      time 11:45
  endfamily
  family pop
    repeat date YMD 20091118 20121212 1
    edit VERIFY 'verify_test'
    edit QUEUE 'emos-e'
    edit QUEUE_EPILOG 'emos-exp'
    edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
    edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
    edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
    edit LOGDIR '/vol/emos_nc/output'
    edit SMSOUT '/vol/emos_nc/output'
    edit METVIEW '/usr/local/share/metview'
    inlimit /limits:linux_cluster
    family 00
      edit EMOS_BASE '00'
      edit MXUP_TRAJ '3'
      edit REINISTREAM 'DCDA'
      edit WINDOW_LENGTH_4D '12'
      edit WINDOW_OFFSET_4D '9'
      edit REINITIALIZE 'true'
      edit EDFAMILY 'true'
      edit PERIOD_4D '6'
      edit FCLENGTH '9'
      edit DELTA_DAY '1'
      edit SMSFILES '/home/ma/emos/def/e_36r1/pop'
      family data
        edit SMSFILES '/home/ma/emos/def/e_36r1/pop/'
        family fields
          trigger ( /e_36r1/main:YMD gt /e_36r1/pop:YMD) or ( /e_36r1/main:YMD eq /e_36r1/pop:YMD and /e_36r1/main/00/fc/model eq complete)
          edit WSHOST 'linux_cluster'
          edit QUEUE 'plots-medium'
          edit STEP_LIST '12/to/240/by/12'
          inlimit /e_36r1/limits:webplots
          task z500
          task mslp
          task mslp_gg
          task t850
          task rain
          task t2m
          task w30m
          task cloud
          task waves
          task rainp
          task w10m
          task wind
            edit STEP_LIST '24/to/240/by/24'
          task rh
            edit STEP_LIST '24/to/240/by/24'
          task vorticity
            edit STEP_LIST '24/to/240/by/24'
          task divergence
            edit STEP_LIST '24/to/240/by/24'
        endfamily
        family metgram
          trigger ( /e_36r1/main:YMD gt /e_36r1/pop:YMD) or ( /e_36r1/main:YMD eq /e_36r1/pop:YMD and /e_36r1/main/00/fc/model eq complete and /mc/main/00/legA/fc eq complete)
          edit WSHOST 'linux_cluster'
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit RHOST 'epsgrams'
          edit SMSFILES '/home/ma/emos/def/e_36r1/plot/metgram'
          inlimit /limits:linux_cluster
          task metret
            trigger ( /e_36r1/main:YMD gt /e_36r1/pop:YMD) or ( /e_36r1/main:YMD eq /e_36r1/pop:YMD and /e_36r1/main/00/fc/model eq complete )
          task metclean
            trigger metret eq complete
            edit DELTA_DAY '-2'
          family fc
            edit KIND 'metgram_fc'
            edit ENSEMBLES '0'
            edit SMSFILES '/home/ma/emos/def/e_36r1/pop/data/epsgrams/grams'
            task gclean
              edit REMOTE_KEEP_DAYS '3'
              edit KEEP_DAYS '3'
            task gdb
              trigger gclean eq complete
              label files ""
              label info ""
              event 1 
              event fdb
            task glink
              trigger gdb eq complete
          endfamily
        endfamily
        family bufr
          edit SMSFILES '/home/ma/emos/def/e_36r1/pop/data/bufr'
          edit WSHOST 'linux_cluster'
          edit QUEUE 'plots-medium'
          task feedback
            trigger ( /e_36r1/lag:YMD gt /e_36r1/pop:YMD) or ( /e_36r1/lag:YMD eq /e_36r1/pop:YMD and /e_36r1/lag/00/archive/obs eq complete)
        endfamily
      endfamily
      family web
        edit SMSFILES '/home/ma/emos/def/e_36r1/pop/web'
        family fc
          trigger ../data/fields eq complete
          edit WSHOST 'linux_cluster'
          edit QUEUE 'plots-medium'
          inlimit /e_36r1/limits:webplots
          family members
            edit PLOT_AREAS 'Europe,North_America,South_America,Asia,Australia,Africa'
            edit PLOT_STEPS '0,12,24,36,48,60,72,84,96,108,120,132,144,156,168,180,192,204,216,228,240'
            edit PLOT_LEGS '1'
            task w_z500t850
            task w_mslrain
            task w_t2m30mw
            task w_clouds
            task w_waves
              edit PLOT_AREAS 'Global,Europe,Pacific,Indian,North_Atlantic,South_Atlantic'
            task w_mwp
              edit PLOT_AREAS 'Global,Europe,Pacific,Indian,North_Atlantic,South_Atlantic'
          endfamily
          family additional
            edit PLOT_AREAS 'Europe,North_America,South_America,Asia,Australia,Africa,North_Hemisphere,South_Hemisphere'
            edit PLOT_STEPS '0,24,48,72,96,120,144,168'
            task w_uv_rh
            task w_msl_uv200
            task w_z500t850_gts
              edit PLOT_STEPS '0,24,48,72,96,120,144,168,192,216,240'
            task w_vordir
              edit PLOT_STEPS '0,24,48,72,96,120,144'
            family wmo_swh_dir
              edit PLOT_STEPS '0,24,48,72,96,120,144,168'
              edit PLOT_AREAS 'Global'
              edit PLOT_TYPE '_wmo'
              task w_waves
            endfamily
            family wmo_mwp_dir
              edit PLOT_STEPS '0,24,48,72,96,120,144,168'
              edit PLOT_AREAS 'Global'
              edit PLOT_TYPE '_wmo'
              task w_mwp
            endfamily
          endfamily
          family essential
            edit PLOT_AREAS 'Europe,North_America,South_America,Asia,Australia,Africa,North_Hemisphere,South_Hemisphere'
            edit PLOT_STEPS '72,96,120,144,168,192,216,240'
            task public_plots
          endfamily
          task web_update
            trigger members eq complete and additional eq complete and essential eq complete
            edit WSHOST 'linux_cluster'
            edit QUEUE 'plots-medium'
            edit PACKAGES 'esuite'
            inlimit /e_36r1/limits:webplots
        endfamily
        family satim
          edit SMSFILES '/home/ma/emos/def/e_36r1/pop/web'
          edit UPDATE_WEB_NOW 'no'
          edit WSHOST 'linux_cluster'
          edit QUEUE 'plots-medium'
          edit PLOT_AREAS 'Europe,North_America,South_America,Africa,Indian_Ocean,Australia,West_Pacific,East_Pacific'
          task w_simulated
            trigger ( /e_36r1/lag:YMD gt /e_36r1/pop:YMD) or ( /e_36r1/lag:YMD eq /e_36r1/pop:YMD and /e_36r1/lag/00/satim eq complete)
          task web_update
            trigger ./w_simulated eq complete and ../fc eq complete
            edit PACKAGES 'esuite'
        endfamily
        family an
          trigger ( /e_36r1/main:YMD gt /e_36r1/pop:YMD) or ( /e_36r1/main:YMD eq /e_36r1/pop:YMD and /e_36r1/main/00/fc/model eq complete)
          edit SMSFILES '/home/ma/emos/def/e_36r1/plot/plot240'
          edit WSHOST 'linux_cluster'
          edit QUEUE 'plots-medium'
          inlimit /e_36r1/limits:webplots
          task w_albedo
            trigger ( /e_36r1/main:YMD gt /e_36r1/pop:YMD) or ( /e_36r1/main:YMD eq /e_36r1/pop:YMD and /e_36r1/main/00/fc/model eq complete and ../../data/fields eq complete)
            edit WEEKDAY 'tuesday'
            edit SMSFILES '/home/ma/emos/def/e_36r1/pop/web/'
          task psst
            edit WSHOST 'linux_cluster'
            edit QUEUE 'plots-medium'
            inlimit /e_36r1/limits:webplots
          task psw123
          task web_update
            trigger w_albedo eq complete and psst eq complete and psw123 eq complete
            edit WSHOST 'linux_cluster'
            edit QUEUE 'plots-medium'
            edit PACKAGES 'esuite'
            edit SMSFILES '/home/ma/emos/def/e_36r1/pop/web/'
            inlimit /e_36r1/limits:webplots
        endfamily
        family monitoring
          edit UPDATE_WEB_NOW 'yes'
          edit WSHOST 'linux_cluster'
          edit QUEUE 'plots-medium'
          task w_coverage
            trigger /e_36r1/pop/00/data/bufr eq complete
          task w_tscover
            time 07:10
          task w_guan
            edit DATEMASK '03.*.*'
        endfamily
        family cloudsat
          trigger fc eq complete
          edit UPDATE_WEB_NOW 'yes'
          edit WSHOST 'linux_cluster'
          edit QUEUE 'plots-medium'
          edit PLOT_AREAS 'South_Atlantic,Pacific,Indian'
          edit PLOT_STEPS '0,12,24,36,48,60,72,84,96,108,120,132,144,156,168'
          task csat_clouds
          task csat_waves
          task csat_wind
        endfamily
        family amma
          edit SMSFILES '/home/ma/emos/def/e_36r1/pop/web/amma'
          family monitoring
            edit UPDATE_WEB_NOW 'yes'
            edit QUEUE 'emos-e'
            edit QUEUE_EPILOG 'emos-exp'
            edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
            edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
            edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
            edit LOGDIR '/vol/emos_nc/output'
            edit SMSOUT '/vol/emos_nc/output'
            edit WSHOST 'linux_cluster'
            inlimit /limits:linux_cluster
            time 10:00
            task amma_cover
              edit DATEMASK '03.*.*'
            task amma_pilot
              edit DATEMASK '03.*.*'
            task amma_rs
              edit DATEMASK '03.*.*'
            task amma_tephigram
              edit METVIEW '/usr/local/share/metview_old'
          endfamily
          family fc
            trigger ( /e_36r1/main:YMD gt /e_36r1/pop:YMD) or ( /e_36r1/main:YMD eq /e_36r1/pop:YMD and /e_36r1/main/00/fc/model eq complete)
            edit UPDATE_WEB_NOW 'yes'
            edit WSHOST 'linux_cluster'
            edit QUEUE 'serial'
            edit PLOTDOM '-30/-35/45/40'
            time 09:30
            task satplot
              edit STEPS '00/06/12/15/18/24/30/36/39/42/48'
            task maps
              edit STEPS '06/12/18/24/30/36/42/48/54/60/66/72/78/84/90/96/102/108/114/120'
            task tephi
              edit METVIEW '/usr/local/share/metview_old'
              edit STEPS '00/06/12/18/24/30/36/42/48/54/60/66/72'
            task trajectories
              edit STEPS '24/48/72'
            task cross
              edit STEPS '00/06/12/18/24/30/36/42/48/54/60/66/72/78/84/90/96/102/108/114/120'
            task hovm
              edit STEPS '06/12/18/24/30/36/42/48/54/60/66/72/78/84/90/96/102/108/114/120/126/132/138/144/150/156/162/168/174/180/186/192/198/204/210/216/222/228/234/240'
            task obs_mslp
            task metgram
            task prob
              edit STEPS '60/72/84/96/108/120/132/144/156/168/180/192/204/216/228/240'
          endfamily
        endfamily
        family swdfp
          trigger ( /e_36r1/main:YMD gt /e_36r1/pop:YMD) or ( /e_36r1/main:YMD eq /e_36r1/pop:YMD and /e_36r1/main/00/fc/model eq complete)
          edit SMSFILES '/home/ma/emos/def/e_36r1/pop/web/swdfp'
          edit UPDATE_WEB_NOW 'yes'
          edit WSHOST 'linux_cluster'
          edit QUEUE 'serial'
          edit PLOTDOM '-45/-10/6/62'
          time 10:30
          family maps
            edit PLOT_LOCATION '/webplots_e/data/swdfp/maps'
            edit STEPS '06/12/18/24/30/36/42/48/54/60/66/72/84/96/108/120/132/144'
            task swdfp_maps
            task smaps_toweb1
              trigger swdfp_maps eq complete
            task smaps_toweb2
              trigger swdfp_maps eq complete
            task smaps_toweb3
              trigger swdfp_maps eq complete
            task swdfp_clean
              trigger smaps_toweb1 eq complete and smaps_toweb2 eq complete and smaps_toweb3 eq complete
          endfamily
          task swdfp_satplot
            edit STEPS '00/06/12/15/18/24/30/36/39/42/48'
          task swdfp_tephi
            edit METVIEW '/usr/local/share/metview_old'
            edit STEPS '00/06/12/18/24/30/36/42/48/54/60/66/72'
          task swdfp_obs_mslp
            edit PLOTDOM '-37/3/-3/55'
          task swdfp_metgram
        endfamily
      endfamily
      family paper
        task pobstat
          trigger ( /e_36r1/lag:YMD gt /e_36r1/pop:YMD) or ( /e_36r1/lag:YMD eq /e_36r1/pop:YMD and /e_36r1/lag/00/fb/obstat_merge eq complete)
          edit FSFAMILY 'an'
          edit WSHOST 'linux_cluster'
          edit SMSFILES '/home/ma/emos/def/e_36r1/plot'
      endfamily
      task cleanup
        trigger web eq complete
        edit WSHOST 'linux_cluster'
        edit QUEUE 'plots-medium'
        inlimit /e_36r1/limits:webplots
    endfamily
    family 12
      edit EMOS_BASE '12'
      edit MXUP_TRAJ '3'
      edit REINISTREAM 'DCDA'
      edit WINDOW_LENGTH_4D '12'
      edit WINDOW_OFFSET_4D '9'
      edit REINITIALIZE 'true'
      edit EDFAMILY 'true'
      edit PERIOD_4D '6'
      edit FCLENGTH '9'
      edit DELTA_DAY '0'
      edit SMSFILES '/home/ma/emos/def/e_36r1/pop'
      family proc
        edit SMSFILES '/home/ma/emos/def/e_36r1/plot'
        task buoycol
          trigger ( /e_36r1/main:YMD gt /e_36r1/pop:YMD) or ( /e_36r1/main:YMD eq /e_36r1/pop:YMD and /e_36r1/main/12/obs eq complete)
          edit FSFAMILY 'fc'
      endfamily
      family data
        edit SMSFILES '/home/ma/emos/def/e_36r1/pop/'
        family fields
          trigger ( /e_36r1/main:YMD gt /e_36r1/pop:YMD) or ( /e_36r1/main:YMD eq /e_36r1/pop:YMD and /e_36r1/main/12/fc/model eq complete)
          edit WSHOST 'linux_cluster'
          edit QUEUE 'plots-medium'
          edit STEP_LIST '12/to/240/by/12'
          inlimit /e_36r1/limits:webplots
          task z500
          task mslp
          task mslp_gg
          task t850
          task rain
          task t2m
          task w30m
          task cloud
          task waves
          task rainp
          task w10m
          task wind
            edit STEP_LIST '24/to/240/by/24'
          task rh
            edit STEP_LIST '24/to/240/by/24'
          task vorticity
            edit STEP_LIST '24/to/240/by/24'
          task divergence
            edit STEP_LIST '24/to/240/by/24'
        endfamily
        family metgram
          trigger ( /e_36r1/main:YMD gt /e_36r1/pop:YMD) or ( /e_36r1/main:YMD eq /e_36r1/pop:YMD and /e_36r1/main/12/fc/model eq complete and /mc/main/12/legA/fc eq complete)
          edit WSHOST 'linux_cluster'
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit RHOST 'epsgrams'
          edit SMSFILES '/home/ma/emos/def/e_36r1/plot/metgram'
          inlimit /limits:linux_cluster
          task metret
            trigger ( /e_36r1/main:YMD gt /e_36r1/pop:YMD) or ( /e_36r1/main:YMD eq /e_36r1/pop:YMD and /e_36r1/main/12/fc/model eq complete )
          task metclean
            trigger metret eq complete
            edit DELTA_DAY '-3'
          family fc
            edit KIND 'metgram_fc'
            edit ENSEMBLES '0'
            edit SMSFILES '/home/ma/emos/def/e_36r1/pop/data/epsgrams/grams'
            task gclean
              edit REMOTE_KEEP_DAYS '3'
              edit KEEP_DAYS '3'
            task gdb
              trigger gclean eq complete
              label files ""
              label info ""
              event 1 
              event fdb
            task glink
              trigger gdb eq complete
          endfamily
        endfamily
        family bufr
          edit SMSFILES '/home/ma/emos/def/e_36r1/pop/data/bufr'
          edit WSHOST 'linux_cluster'
          edit QUEUE 'plots-medium'
          task feedback
            trigger ( /e_36r1/lag:YMD gt /e_36r1/pop:YMD) or ( /e_36r1/lag:YMD eq /e_36r1/pop:YMD and /e_36r1/lag/12/archive/obs eq complete)
        endfamily
      endfamily
      family web
        edit SMSFILES '/home/ma/emos/def/e_36r1/pop/web'
        family fc
          trigger ../data/fields eq complete
          edit WSHOST 'linux_cluster'
          edit QUEUE 'plots-medium'
          inlimit /e_36r1/limits:webplots
          family members
            edit PLOT_AREAS 'Europe,North_America,South_America,Asia,Australia,Africa'
            edit PLOT_STEPS '0,12,24,36,48,60,72,84,96,108,120,132,144,156,168,180,192,204,216,228,240'
            edit PLOT_LEGS '1'
            task w_z500t850
            task w_mslrain
            task w_t2m30mw
            task w_clouds
            task w_waves
              edit PLOT_AREAS 'Global,Europe,Pacific,Indian,North_Atlantic,South_Atlantic'
            task w_mwp
              edit PLOT_AREAS 'Global,Europe,Pacific,Indian,North_Atlantic,South_Atlantic'
          endfamily
          family additional
            edit PLOT_AREAS 'Europe,North_America,South_America,Asia,Australia,Africa,North_Hemisphere,South_Hemisphere'
            edit PLOT_STEPS '0,24,48,72,96,120,144,168'
            task w_uv_rh
            task w_msl_uv200
            task w_z500t850_gts
              edit PLOT_STEPS '0,24,48,72,96,120,144,168,192,216,240'
            task w_vordir
              edit PLOT_STEPS '0,24,48,72,96,120,144'
            family wmo_swh_dir
              edit PLOT_STEPS '0,24,48,72,96,120,144,168'
              edit PLOT_AREAS 'Global'
              edit PLOT_TYPE '_wmo'
              task w_waves
            endfamily
            family wmo_mwp_dir
              edit PLOT_STEPS '0,24,48,72,96,120,144,168'
              edit PLOT_AREAS 'Global'
              edit PLOT_TYPE '_wmo'
              task w_mwp
            endfamily
          endfamily
          family essential
            edit PLOT_AREAS 'Europe,North_America,South_America,Asia,Australia,Africa,North_Hemisphere,South_Hemisphere'
            edit PLOT_STEPS '72,96,120,144,168,192,216,240'
            task public_plots
          endfamily
          task web_update
            trigger members eq complete and additional eq complete and essential eq complete
            edit WSHOST 'linux_cluster'
            edit QUEUE 'plots-medium'
            edit PACKAGES 'esuite'
            inlimit /e_36r1/limits:webplots
        endfamily
        family satim
          edit SMSFILES '/home/ma/emos/def/e_36r1/pop/web'
          edit UPDATE_WEB_NOW 'no'
          edit WSHOST 'linux_cluster'
          edit QUEUE 'plots-medium'
          edit PLOT_AREAS 'Europe,North_America,South_America,Africa,Indian_Ocean,Australia,West_Pacific,East_Pacific'
          task w_simulated
            trigger ( /e_36r1/lag:YMD gt /e_36r1/pop:YMD) or ( /e_36r1/lag:YMD eq /e_36r1/pop:YMD and /e_36r1/lag/12/satim eq complete)
          task web_update
            trigger ./w_simulated eq complete and ../fc eq complete
            edit PACKAGES 'esuite'
        endfamily
        family verification
          edit SMSFILES '/home/ma/emos/def/e_36r1/pop/web/verification'
          edit UPDATE_WEB_NOW 'yes'
          family monthly
            edit DATEMASK '11.*.*'
            task w_2t
            task w_cloud_humid
            task w_preci
            task w_scores
            task w_scores_mean
            task w_wmomeans
            task w_pltime_waves
          endfamily
          family daily
            trigger /e_36r1/lag/12/verif/oc eq complete
            task w_allcentres
            task w_weekly_centres
              trigger w_allcentres eq complete
              edit WEEKDAY 'monday'
          endfamily
        endfamily
        family monitoring
          edit UPDATE_WEB_NOW 'yes'
          edit WSHOST 'linux_cluster'
          edit QUEUE 'plots-medium'
          task w_coverage
            trigger /e_36r1/pop/12/data/bufr eq complete
        endfamily
        family cloudsat
          trigger fc eq complete
          edit UPDATE_WEB_NOW 'yes'
          edit WSHOST 'linux_cluster'
          edit QUEUE 'plots-medium'
          edit PLOT_AREAS 'South_Atlantic,Pacific,Indian'
          edit PLOT_STEPS '0,12,24,36,48,60,72,84,96,108,120,132,144,156,168'
          task csat_clouds
          task csat_waves
          task csat_wind
        endfamily
        family amma
          edit SMSFILES '/home/ma/emos/def/e_36r1/pop/web/amma'
          family fc
            trigger ( /e_36r1/main:YMD gt /e_36r1/pop:YMD) or ( /e_36r1/main:YMD eq /e_36r1/pop:YMD and /e_36r1/main/12/fc/model eq complete)
            edit UPDATE_WEB_NOW 'yes'
            edit WSHOST 'linux_cluster'
            edit QUEUE 'serial'
            edit PLOTDOM '-30/-35/45/40'
            time 00:30
            task satplot
              edit STEPS '00/06/12/15/18/24/30/36/39/42/48'
            task maps
              edit STEPS '06/12/18/24/30/36/42/48/54/60/66/72/78/84/90/96/102/108/114/120'
            task tephi
              edit METVIEW '/usr/local/share/metview_old'
              edit STEPS '00/06/12/18/24/30/36/42/48/54/60/66/72'
            task trajectories
              edit STEPS '24/48/72'
            task cross
              edit STEPS '00/06/12/18/24/30/36/42/48/54/60/66/72/78/84/90/96/102/108/114/120'
            task hovm
              edit STEPS '06/12/18/24/30/36/42/48/54/60/66/72/78/84/90/96/102/108/114/120/126/132/138/144/150/156/162/168/174/180/186/192/198/204/210/216/222/228/234/240'
            task obs_mslp
            task metgram
            task prob
              edit STEPS '60/72/84/96/108/120/132/144/156/168/180/192/204/216/228/240'
          endfamily
        endfamily
        family swdfp
          trigger ( /e_36r1/main:YMD gt /e_36r1/pop:YMD) or ( /e_36r1/main:YMD eq /e_36r1/pop:YMD and /e_36r1/main/12/fc/model eq complete)
          edit SMSFILES '/home/ma/emos/def/e_36r1/pop/web/swdfp'
          edit UPDATE_WEB_NOW 'yes'
          edit WSHOST 'linux_cluster'
          edit QUEUE 'serial'
          edit PLOTDOM '-45/-10/6/62'
          time 01:30
          family maps
            edit PLOT_LOCATION '/webplots_e/data/swdfp/maps'
            edit STEPS '06/12/18/24/30/36/42/48/54/60/66/72/84/96/108/120/132/144'
            task swdfp_maps
            task smaps_toweb1
              trigger swdfp_maps eq complete
            task smaps_toweb2
              trigger swdfp_maps eq complete
            task smaps_toweb3
              trigger swdfp_maps eq complete
            task swdfp_clean
              trigger smaps_toweb1 eq complete and smaps_toweb2 eq complete and smaps_toweb3 eq complete
          endfamily
          task swdfp_satplot
            edit STEPS '00/06/12/15/18/24/30/36/39/42/48'
          task swdfp_tephi
            edit METVIEW '/usr/local/share/metview_old'
            edit STEPS '00/06/12/18/24/30/36/42/48/54/60/66/72'
          task swdfp_obs_mslp
            edit PLOTDOM '-37/3/-3/55'
          task swdfp_metgram
        endfamily
      endfamily
      family paper
        task pobstat
          trigger ( /e_36r1/lag:YMD gt /e_36r1/pop:YMD) or ( /e_36r1/lag:YMD eq /e_36r1/pop:YMD and /e_36r1/lag/12/fb/obstat_merge eq complete)
          edit FSFAMILY 'an'
          edit WSHOST 'linux_cluster'
          edit SMSFILES '/home/ma/emos/def/e_36r1/plot'
      endfamily
      family verify
        edit SMSFILES '/home/ma/emos/def/e_36r1/pop/verify'
        family consistency
          edit ECFS_OPTIONS '-o'
          edit ECMWF 'ECMWF24'
          edit BRACKNELL 'BRAKL24'
          edit OFFENBACH 'OFFNB24'
          edit WASHINGTON 'WASHN24'
          edit DATEMASK '02.*.*'
          task create_daily_files
            edit DATEMASK '02.*.*'
          task create_monthly_files
            trigger create_daily_files eq complete
            edit DATEMASK '02.01.*'
          task ecmwf
            trigger create_monthly_files eq complete
            edit DATEMASK '02.*.*'
          task bracknell
            trigger create_monthly_files eq complete
            edit DATEMASK '02.*.*'
          task offenbach
            trigger create_monthly_files eq complete
            edit DATEMASK '02.*.*'
          task washington
            trigger create_monthly_files eq complete
            edit DATEMASK '02.*.*'
        endfamily
      endfamily
      task cleanup
        trigger web eq complete
        edit WSHOST 'linux_cluster'
        edit QUEUE 'plots-medium'
        inlimit /e_36r1/limits:webplots
    endfamily
    family 00dc
      edit EMOS_BASE '00'
      edit MXUP_TRAJ '3'
      edit STREAM 'DCDA'
      edit WINDOW_LENGTH_4D '12'
      edit WINDOW_OFFSET_4D '3'
      edit EDFAMILY 'false'
      edit PERIOD_4D '12'
      edit FCLENGTH '15'
      edit DELTA_DAY '0'
      edit SMSFILES '/home/ma/emos/def/e_36r1/pop'
      family web
        edit SMSFILES '/home/ma/emos/def/e_36r1/pop/web'
        family an
          task pincr
            trigger ( /e_36r1/main:YMD gt /e_36r1/pop:YMD) or ( /e_36r1/main:YMD eq /e_36r1/pop:YMD and /e_36r1/main/00dc/an eq complete)
            edit SMSFILES '/home/ma/emos/def/e_36r1/plot'
            edit UPDATE_WEB_NOW 'yes'
            edit WSHOST 'linux_cluster'
            edit QUEUE 'plots-medium'
            inlimit /e_36r1/limits:webplots
        endfamily
        family monitoring
          edit UPDATE_WEB_NOW 'yes'
          family satmon
            edit QUEUE_EPILOG 'emos-exp'
            edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
            edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
            edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
            edit LOGDIR '/vol/emos_nc/output'
            edit SMSOUT '/vol/emos_nc/output'
            edit WSHOST 'linux_cluster'
            edit QUEUE 'plots-medium'
            inlimit /limits:linux_cluster
            task satmon
              trigger ( /e_36r1/lag:YMD gt /e_36r1/pop:YMD) or ( /e_36r1/lag:YMD eq /e_36r1/pop:YMD and /e_36r1/lag/00dc/satmon eq complete)
            family airs
              trigger satmon eq complete
              edit SATELLITE 'airs'
              inlimit /e_36r1/limits:satmon
              task w_satmon
                edit PLOTALL 'true'
                edit PLOTGEO 'false'
              task w_satmon_geo
                edit WEEKDAY 'saturday'
                edit PLOTALL 'false'
                edit PLOTGEO 'true'
            endfamily
            family noaa
              trigger satmon eq complete
              edit SATELLITE 'noaa'
              inlimit /e_36r1/limits:satmon
              task w_satmon
                edit PLOTALL 'true'
                edit PLOTGEO 'false'
              task w_satmon_geo
                edit WEEKDAY 'saturday'
                edit PLOTALL 'false'
                edit PLOTGEO 'true'
            endfamily
            family ears
              trigger satmon eq complete
              edit SATELLITE 'ears'
              inlimit /e_36r1/limits:satmon
              task w_satmon
                edit PLOTALL 'true'
                edit PLOTGEO 'false'
              task w_satmon_geo
                edit WEEKDAY 'saturday'
                edit PLOTALL 'false'
                edit PLOTGEO 'true'
            endfamily
            family geos
              trigger satmon eq complete
              edit SATELLITE 'geos'
              inlimit /e_36r1/limits:satmon
              task w_satmon
                edit PLOTALL 'true'
                edit PLOTGEO 'false'
              task w_satmon_geo
                edit WEEKDAY 'saturday'
                edit PLOTALL 'false'
                edit PLOTGEO 'true'
            endfamily
            family o3
              trigger satmon eq complete
              edit SATELLITE 'o3'
              inlimit /e_36r1/limits:satmon
              task w_satmon
                edit PLOTALL 'true'
                edit PLOTGEO 'false'
              task w_satmon_geo
                edit WEEKDAY 'saturday'
                edit PLOTALL 'false'
                edit PLOTGEO 'true'
            endfamily
            family o3knmi
              trigger satmon eq complete
              edit SATELLITE 'o3knmi'
              inlimit /e_36r1/limits:satmon
              task w_satmon
                edit PLOTALL 'true'
                edit PLOTGEO 'false'
              task w_satmon_geo
                edit WEEKDAY 'saturday'
                edit PLOTALL 'false'
                edit PLOTGEO 'true'
            endfamily
            family temp
              trigger satmon eq complete
              edit SATELLITE 'temp'
              inlimit /e_36r1/limits:satmon
              task w_satmon
                edit PLOTALL 'true'
                edit PLOTGEO 'false'
              task w_satmon_geo
                edit WEEKDAY 'saturday'
                edit PLOTALL 'false'
                edit PLOTGEO 'true'
            endfamily
            family pwc
              trigger satmon eq complete
              edit SATELLITE 'pwc'
              inlimit /e_36r1/limits:satmon
              task w_satmon
                edit PLOTALL 'true'
                edit PLOTGEO 'false'
              task w_satmon_geo
                edit WEEKDAY 'saturday'
                edit PLOTALL 'false'
                edit PLOTGEO 'true'
            endfamily
            family amv_speed
              trigger satmon eq complete
              edit SATELLITE 'amv_speed'
              inlimit /e_36r1/limits:satmon
              task w_satmon
                edit PLOTALL 'true'
                edit PLOTGEO 'false'
              task w_satmon_geo
                edit WEEKDAY 'saturday'
                edit PLOTALL 'false'
                edit PLOTGEO 'true'
            endfamily
            family dmsp
              trigger satmon eq complete
              edit SATELLITE 'dmsp'
              inlimit /e_36r1/limits:satmon
              task w_satmon
                edit PLOTALL 'true'
                edit PLOTGEO 'false'
              task w_satmon_geo
                edit WEEKDAY 'saturday'
                edit PLOTALL 'false'
                edit PLOTGEO 'true'
            endfamily
            family gpsro
              trigger satmon eq complete
              edit SATELLITE 'gpsro'
              inlimit /e_36r1/limits:satmon
              task w_satmon
                edit PLOTALL 'true'
                edit PLOTGEO 'false'
              task w_satmon_geo
                edit WEEKDAY 'saturday'
                edit PLOTALL 'false'
                edit PLOTGEO 'true'
            endfamily
            family tmi
              trigger satmon eq complete
              edit SATELLITE 'tmi'
              inlimit /e_36r1/limits:satmon
              task w_satmon
                edit PLOTALL 'true'
                edit PLOTGEO 'false'
              task w_satmon_geo
                edit WEEKDAY 'saturday'
                edit PLOTALL 'false'
                edit PLOTGEO 'true'
            endfamily
            family ssmis
              trigger satmon eq complete
              edit SATELLITE 'ssmis'
              inlimit /e_36r1/limits:satmon
              task w_satmon
                edit PLOTALL 'true'
                edit PLOTGEO 'false'
              task w_satmon_geo
                edit WEEKDAY 'saturday'
                edit PLOTALL 'false'
                edit PLOTGEO 'true'
            endfamily
            family amsre
              trigger satmon eq complete
              edit SATELLITE 'amsre'
              inlimit /e_36r1/limits:satmon
              task w_satmon
                edit PLOTALL 'true'
                edit PLOTGEO 'false'
              task w_satmon_geo
                edit WEEKDAY 'saturday'
                edit PLOTALL 'false'
                edit PLOTGEO 'true'
            endfamily
            family iasi
              trigger satmon eq complete
              edit SATELLITE 'iasi'
              inlimit /e_36r1/limits:satmon
              task w_satmon
                edit PLOTALL 'true'
                edit PLOTGEO 'false'
              task w_satmon_geo
                edit WEEKDAY 'saturday'
                edit PLOTALL 'false'
                edit PLOTGEO 'true'
            endfamily
            family pacrars
              trigger satmon eq complete
              edit SATELLITE 'pacrars'
              inlimit /e_36r1/limits:satmon
              task w_satmon
                edit PLOTALL 'true'
                edit PLOTGEO 'false'
              task w_satmon_geo
                edit WEEKDAY 'saturday'
                edit PLOTALL 'false'
                edit PLOTGEO 'true'
            endfamily
            family dbamv_speed
              trigger satmon eq complete
              edit SATELLITE 'dbamv_speed'
              inlimit /e_36r1/limits:satmon
              task w_satmon
                edit PLOTALL 'true'
                edit PLOTGEO 'false'
              task w_satmon_geo
                edit WEEKDAY 'saturday'
                edit PLOTALL 'false'
                edit PLOTGEO 'true'
            endfamily
            family meris
              trigger satmon eq complete
              edit SATELLITE 'meris'
              inlimit /e_36r1/limits:satmon
              task w_satmon
                edit PLOTALL 'true'
                edit PLOTGEO 'false'
              task w_satmon_geo
                edit WEEKDAY 'saturday'
                edit PLOTALL 'false'
                edit PLOTGEO 'true'
            endfamily
            family scatt_speed
              trigger satmon eq complete
              edit SATELLITE 'scatt_speed'
              inlimit /e_36r1/limits:satmon
              task w_satmon
                edit PLOTALL 'true'
                edit PLOTGEO 'false'
              task w_satmon_geo
                edit WEEKDAY 'saturday'
                edit PLOTALL 'false'
                edit PLOTGEO 'true'
            endfamily
            family slmoist
              trigger satmon eq complete
              edit SATELLITE 'slmoist'
              inlimit /e_36r1/limits:satmon
              task w_satmon
                edit PLOTALL 'true'
                edit PLOTGEO 'false'
              task w_satmon_geo
                edit WEEKDAY 'saturday'
                edit PLOTALL 'false'
                edit PLOTGEO 'true'
            endfamily
            task web_daily
              trigger airs eq complete and noaa eq complete and ears eq complete and geos eq complete and o3 eq complete and o3knmi eq complete and temp eq complete and pwc eq complete and amv_speed eq complete and dmsp eq complete and gpsro eq complete and tmi eq complete and ssmis eq complete and amsre eq complete and iasi eq complete and pacrars eq complete and dbamv_speed eq complete and meris eq complete and scatt_speed eq complete and slmoist eq complete
            task web_weekly
              trigger airs eq complete and noaa eq complete and ears eq complete and geos eq complete and o3 eq complete and o3knmi eq complete and temp eq complete and pwc eq complete and amv_speed eq complete and dmsp eq complete and gpsro eq complete and tmi eq complete and ssmis eq complete and amsre eq complete and iasi eq complete and pacrars eq complete and dbamv_speed eq complete and meris eq complete and scatt_speed eq complete and slmoist eq complete
              edit WEEKDAY 'saturday'
          endfamily
        endfamily
      endfamily
    endfamily
    family 06dc
      edit EMOS_BASE '06'
      edit MXUP_TRAJ '3'
      edit STREAM 'DCDA'
      edit WINDOW_LENGTH_4D '12'
      edit WINDOW_OFFSET_4D '3'
      edit EDFAMILY 'false'
      edit PERIOD_4D '12'
      edit FCLENGTH '15'
      edit DELTA_DAY '0'
      edit SMSFILES '/home/ma/emos/def/e_36r1/pop'
      family data
        edit SMSFILES '/home/ma/emos/def/e_36r1/pop/'
        family bufr
          edit SMSFILES '/home/ma/emos/def/e_36r1/pop/data/bufr'
          edit WSHOST 'linux_cluster'
          edit QUEUE 'plots-medium'
          task feedback
            trigger ( /e_36r1/lag:YMD gt /e_36r1/pop:YMD) or ( /e_36r1/lag:YMD eq /e_36r1/pop:YMD and /e_36r1/lag/00dc/archive/obs eq complete)
        endfamily
      endfamily
      family web
        edit SMSFILES '/home/ma/emos/def/e_36r1/pop/web'
        family an
          task pincr
            trigger ( /e_36r1/main:YMD gt /e_36r1/pop:YMD) or ( /e_36r1/main:YMD eq /e_36r1/pop:YMD and /e_36r1/main/00dc/an eq complete and /e_36r1/pop/00dc/web/an/pincr eq complete)
            edit SMSFILES '/home/ma/emos/def/e_36r1/plot'
            edit UPDATE_WEB_NOW 'yes'
            edit WSHOST 'linux_cluster'
            edit QUEUE 'plots-medium'
            inlimit /e_36r1/limits:webplots
        endfamily
        family monitoring
          edit UPDATE_WEB_NOW 'yes'
          edit WSHOST 'linux_cluster'
          edit QUEUE 'plots-medium'
          task w_coverage
            trigger /e_36r1/pop/06dc/data/bufr eq complete
        endfamily
      endfamily
      task cleanup
        trigger web eq complete
        edit WSHOST 'linux_cluster'
        edit QUEUE 'plots-medium'
        inlimit /e_36r1/limits:webplots
    endfamily
    family 12dc
      edit EMOS_BASE '12'
      edit MXUP_TRAJ '3'
      edit STREAM 'DCDA'
      edit WINDOW_LENGTH_4D '12'
      edit WINDOW_OFFSET_4D '3'
      edit EDFAMILY 'false'
      edit PERIOD_4D '12'
      edit FCLENGTH '15'
      edit DELTA_DAY '0'
      edit SMSFILES '/home/ma/emos/def/e_36r1/pop'
      family web
        edit SMSFILES '/home/ma/emos/def/e_36r1/pop/web'
        family an
          task pincr
            trigger ( /e_36r1/main:YMD gt /e_36r1/pop:YMD) or ( /e_36r1/main:YMD eq /e_36r1/pop:YMD and /e_36r1/main/12dc/an eq complete)
            edit SMSFILES '/home/ma/emos/def/e_36r1/plot'
            edit UPDATE_WEB_NOW 'yes'
            edit WSHOST 'linux_cluster'
            edit QUEUE 'plots-medium'
            inlimit /e_36r1/limits:webplots
        endfamily
      endfamily
    endfamily
    family 18dc
      edit EMOS_BASE '18'
      edit MXUP_TRAJ '3'
      edit STREAM 'DCDA'
      edit WINDOW_LENGTH_4D '12'
      edit WINDOW_OFFSET_4D '3'
      edit EDFAMILY 'false'
      edit PERIOD_4D '12'
      edit FCLENGTH '15'
      edit DELTA_DAY '0'
      edit SMSFILES '/home/ma/emos/def/e_36r1/pop'
      family data
        edit SMSFILES '/home/ma/emos/def/e_36r1/pop/'
        family bufr
          edit SMSFILES '/home/ma/emos/def/e_36r1/pop/data/bufr'
          edit WSHOST 'linux_cluster'
          edit QUEUE 'plots-medium'
          task feedback
            trigger ( /e_36r1/lag:YMD gt /e_36r1/pop:YMD) or ( /e_36r1/lag:YMD eq /e_36r1/pop:YMD and /e_36r1/lag/12dc/archive/obs eq complete)
        endfamily
      endfamily
      family web
        edit SMSFILES '/home/ma/emos/def/e_36r1/pop/web'
        family an
          task pincr
            trigger ( /e_36r1/main:YMD gt /e_36r1/pop:YMD) or ( /e_36r1/main:YMD eq /e_36r1/pop:YMD and /e_36r1/main/12dc/an eq complete and /e_36r1/pop/12dc/web/an/pincr eq complete)
            edit SMSFILES '/home/ma/emos/def/e_36r1/plot'
            edit UPDATE_WEB_NOW 'yes'
            edit WSHOST 'linux_cluster'
            edit QUEUE 'plots-medium'
            inlimit /e_36r1/limits:webplots
        endfamily
        family monitoring
          edit UPDATE_WEB_NOW 'yes'
          edit WSHOST 'linux_cluster'
          edit QUEUE 'plots-medium'
          task w_coverage
            trigger /e_36r1/pop/18dc/data/bufr eq complete
        endfamily
      endfamily
      task cleanup
        trigger web eq complete
        edit WSHOST 'linux_cluster'
        edit QUEUE 'plots-medium'
        inlimit /e_36r1/limits:webplots
    endfamily
    family 18bc
      edit EDFAMILY 'false'
      edit INISTREAM 'DA'
      edit REINITIALIZE 'false'
      edit REINIEXPVER '0046'
      edit INIEXPVER '0045'
      edit REINIPERIOD_4D '6'
      edit REINISTREAM 'DA'
      edit REINIOFFSET '6'
      edit RUN_HOURLY_PP '0'
      edit PERIOD_4D '12'
      edit WINDOW_LENGTH_4D '12'
      edit WINDOW_OFFSET_4D '9'
      edit MXUP_TRAJ '3'
      edit PARALLEL '4'
      edit LANONLY 'false'
      edit ACCOUNT 'opebcmis'
      edit STREAM 'SCDA'
      edit EMOS_STREAM 'SCDA'
      edit DELTA_DAY '0'
      edit SMSFILES '/home/ma/emos/def/e_36r1/plot'
      edit EMOS_BASE '18'
      family paper
        family an
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit FSFAMILY 'an'
          inlimit /limits:linux_cluster
          task pobstat
            trigger ( /e_36r1/lag:YMD gt /e_36r1/pop:YMD) or ( /e_36r1/lag:YMD eq /e_36r1/pop:YMD and /e_36r1/lag/18bc/fb/obstat_merge eq complete)
            edit PRIORITY '51'
        endfamily
      endfamily
      task prod_2t
        edit WSHOST 'linux_cluster'
        edit QUEUE 'serial'
        edit SMSFILES '/home/mo/moz/sms/other_smsfiles/global_efi/daily'
        edit REFDATE '0'
        edit EFI_PARAM '2t'
    endfamily
    family 06bc
      edit EDFAMILY 'false'
      edit INISTREAM 'DA'
      edit REINITIALIZE 'false'
      edit REINIEXPVER '0046'
      edit INIEXPVER '0045'
      edit REINIPERIOD_4D '6'
      edit REINISTREAM 'DA'
      edit REINIOFFSET '6'
      edit RUN_HOURLY_PP '0'
      edit PERIOD_4D '12'
      edit WINDOW_LENGTH_4D '12'
      edit WINDOW_OFFSET_4D '9'
      edit MXUP_TRAJ '3'
      edit PARALLEL '4'
      edit LANONLY 'false'
      edit ACCOUNT 'opebcmis'
      edit STREAM 'SCDA'
      edit EMOS_STREAM 'SCDA'
      edit DELTA_DAY '1'
      edit SMSFILES '/home/ma/emos/def/e_36r1/plot'
      edit EMOS_BASE '06'
      family paper
        family an
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit FSFAMILY 'an'
          inlimit /limits:linux_cluster
          task pobstat
            trigger ( /e_36r1/lag:YMD gt /e_36r1/pop:YMD) or ( /e_36r1/lag:YMD eq /e_36r1/pop:YMD and /e_36r1/lag/06bc/fb/obstat_merge eq complete)
            edit PRIORITY '51'
        endfamily
      endfamily
      task prod_2t
        edit WSHOST 'linux_cluster'
        edit QUEUE 'serial'
        edit SMSFILES '/home/mo/moz/sms/other_smsfiles/global_efi/daily'
        edit REFDATE '0'
        edit EFI_PARAM '2t'
    endfamily
    task logfilestore
      edit QUEUE 'emos-e'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit SMSFILES '/home/ma/emos/def/e_36r1'
      edit LOGTASK 'pop'
      edit KEEPLOGS 'no'
      edit LOGSTOECFS 'no'
      edit MOVE_TO_ECFS 'yes'
    task logfiles
      late -s +00:05 -c +00:15
      trigger ../main/make eq complete and ./00 eq complete and ./12 eq complete and ./00dc eq complete and ./06dc eq complete and ./12dc eq complete and ./18dc eq complete and ./18bc eq complete and ./06bc eq complete and logfilestore eq complete
      edit QUEUE 'emos-e'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit SMSFILES '/home/ma/emos/def/e_36r1'
      edit LOGTASK 'pop'
      edit MOVE_TO_ECFS 'no'
      time 14:15
  endfamily
  family msjobs
    complete ( /e_36r1/msjobs:YMD le /e_36r1/main:YMD) and /e_36r1/main/logfiles eq active
    trigger ( /e_36r1/main:YMD gt /e_36r1/msjobs:YMD) or ( /e_36r1/main:YMD eq /e_36r1/msjobs:YMD or ( /e_36r1/lag:YMD gt /e_36r1/msjobs:YMD)) or ( /e_36r1/lag:YMD eq /e_36r1/msjobs:YMD or ( /e_36r1/pop:YMD gt /e_36r1/msjobs:YMD)) or ( /e_36r1/pop:YMD eq /e_36r1/msjobs:YMD)
    repeat date YMD 20091118 20121212 1
    edit SMSURLBASE 'http://ecgate.ecmwf.int:9080/'
    edit SMSURL 'do/events'
    family 12
      edit EMOS_BASE '12'
      edit MXUP_TRAJ '3'
      edit REINISTREAM 'DCDA'
      edit WINDOW_LENGTH_4D '12'
      edit WINDOW_OFFSET_4D '9'
      edit REINITIALIZE 'true'
      edit EDFAMILY 'true'
      edit PERIOD_4D '6'
      edit FCLENGTH '9'
      edit DELTA_DAY '-1'
      edit SMSFILES '/home/ma/emos/def/e_36r1/msjobs'
      edit SMSTRYNO '1'
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'ecgate'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit SMSOUT '/vol/emos_nc/output'
      edit LOGDIR '/vol/emos_nc/output'
      family mstraj
        trigger ( /e_36r1/lag:YMD gt /e_36r1/msjobs:YMD) or ( /e_36r1/lag:YMD eq /e_36r1/msjobs:YMD and /e_36r1/lag/12/traj/cp_to_other eq complete)
        edit MSJ_DIR 'tr12h000'
        time 18:55
        task ms_ec
          edit SMSTRIES '1'
      endfamily
      family ms_an06
        trigger ( /e_36r1/main:YMD gt /e_36r1/msjobs:YMD) or ( /e_36r1/main:YMD eq /e_36r1/msjobs:YMD and /e_36r1/main/12/an eq complete and ( ../../main/12/fc/model:step gt 000 or ../../main/12/fc/model eq complete))
        edit MSJ_DIR 'an06h000'
        edit EMOS_BASE '06'
        time 17:40
        task ms_ec
          edit SMSTRIES '1'
      endfamily
      family ms000
        trigger ( /e_36r1/main:YMD gt /e_36r1/msjobs:YMD) or ( /e_36r1/main:YMD eq /e_36r1/msjobs:YMD and /e_36r1/main/12/an eq complete and ( ../../main/12/fc/model:step gt 009 or ../../main/12/fc/model eq complete))
        edit MSJ_DIR 'an12h000'
        time 17:40
        task ms_ec
          edit SMSTRIES '1'
      endfamily
      family ms036
        trigger ( /e_36r1/main:YMD gt /e_36r1/msjobs:YMD) or ( /e_36r1/main:YMD eq /e_36r1/msjobs:YMD and ( /e_36r1/main/12/fc/model:step gt 036 or /e_36r1/main/12/fc/model eq complete) and /e_36r1/main/12/prod/036/derived_param eq complete)
        edit EMOS_TIME_STEP_H '0036'
        edit MSJ_DIR 'fc12h036'
        time 17:56
        task ms_ec
          edit SMSTRIES '1'
      endfamily
      family ms072
        trigger ( /e_36r1/main:YMD gt /e_36r1/msjobs:YMD) or ( /e_36r1/main:YMD eq /e_36r1/msjobs:YMD and ( /e_36r1/main/12/fc/model:step gt 072 or /e_36r1/main/12/fc/model eq complete) and /e_36r1/main/12/prod/072/derived_param eq complete)
        edit EMOS_TIME_STEP_H '0072'
        edit MSJ_DIR 'fc12h072'
        time 18:06
        task ms_ec
          edit SMSTRIES '1'
      endfamily
      family ms144
        trigger ( /e_36r1/main:YMD gt /e_36r1/msjobs:YMD) or ( /e_36r1/main:YMD eq /e_36r1/msjobs:YMD and ( /e_36r1/main/12/fc/model:step gt 144 or /e_36r1/main/12/fc/model eq complete) and /e_36r1/main/12/prod/144/derived_param eq complete)
        edit EMOS_TIME_STEP_H '0144'
        edit MSJ_DIR 'fc12h144'
        time 18:27
        task ms_ec
          edit SMSTRIES '1'
      endfamily
      family ms240
        trigger ( /e_36r1/main:YMD gt /e_36r1/msjobs:YMD) or ( /e_36r1/main:YMD eq /e_36r1/msjobs:YMD and ( /e_36r1/main/12/fc/model:step gt 240 or /e_36r1/main/12/fc/model eq complete) and /e_36r1/main/12/prod/240/derived_param eq complete)
        edit EMOS_TIME_STEP_H '0240'
        edit MSJ_DIR 'fc12h240'
        time 18:55
        task ms_ec
          edit SMSTRIES '1'
      endfamily
      family mswave
        trigger ( /e_36r1/main:YMD gt /e_36r1/msjobs:YMD) or ( /e_36r1/main:YMD eq /e_36r1/msjobs:YMD and /e_36r1/main/12/fc/model eq complete)
        edit EMOS_TIME_STEP_H '240'
        edit MSJ_DIR 'wg12h240'
        time 18:55
        task ms_ec
          edit SMSTRIES '1'
      endfamily
      family msmetgram
        trigger ( /e_36r1/pop:YMD gt /e_36r1/msjobs:YMD) or ( /e_36r1/pop:YMD eq /e_36r1/msjobs:YMD and /e_36r1/pop/12/data/metgram eq complete)
        edit MSJ_DIR 'fc12hmetgram'
        time 18:55
        task ms_ec
          edit SMSTRIES '1'
      endfamily
      task test_ll
        edit MSJ_DIR 'test_ll'
        edit USE_NEW_MSJ 'true'
    endfamily
    family 00
      edit EMOS_BASE '00'
      edit MXUP_TRAJ '3'
      edit REINISTREAM 'DCDA'
      edit WINDOW_LENGTH_4D '12'
      edit WINDOW_OFFSET_4D '9'
      edit REINITIALIZE 'true'
      edit EDFAMILY 'true'
      edit PERIOD_4D '6'
      edit FCLENGTH '9'
      edit DELTA_DAY '0'
      edit SMSFILES '/home/ma/emos/def/e_36r1/msjobs'
      edit SMSTRYNO '1'
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'ecgate'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit SMSOUT '/vol/emos_nc/output'
      edit LOGDIR '/vol/emos_nc/output'
      family mstraj
        trigger ( /e_36r1/lag:YMD gt /e_36r1/msjobs:YMD) or ( /e_36r1/lag:YMD eq /e_36r1/msjobs:YMD and /e_36r1/lag/00/traj/cp_to_other eq complete)
        edit MSJ_DIR 'tr12h000'
        time 06:55
        task ms_ec
          edit SMSTRIES '1'
      endfamily
      family ms_an18
        trigger ( /e_36r1/main:YMD gt /e_36r1/msjobs:YMD) or ( /e_36r1/main:YMD eq /e_36r1/msjobs:YMD and /e_36r1/main/00/an eq complete and ( ../../main/00/fc/model:step gt 000 or ../../main/00/fc/model eq complete))
        edit DELTA_DAY '0'
        edit EMOS_BASE '18'
        edit MSJ_DIR 'an18h000'
        time 05:40
        task ms_ec
          edit SMSTRIES '1'
      endfamily
      family ms000
        trigger ( /e_36r1/main:YMD gt /e_36r1/msjobs:YMD) or ( /e_36r1/main:YMD eq /e_36r1/msjobs:YMD and /e_36r1/main/00/an eq complete and ( ../../main/00/fc/model:step gt 009 or ../../main/00/fc/model eq complete))
        edit MSJ_DIR 'an00h000'
        time 05:40
        task ms_ec
          edit SMSTRIES '1'
      endfamily
      family ms036
        trigger ( /e_36r1/main:YMD gt /e_36r1/msjobs:YMD) or ( /e_36r1/main:YMD eq /e_36r1/msjobs:YMD and ( /e_36r1/main/00/fc/model:step gt 036 or /e_36r1/main/00/fc/model eq complete) and /e_36r1/main/00/prod/036/derived_param eq complete)
        edit EMOS_TIME_STEP_H '0036'
        edit MSJ_DIR 'fc00h036'
        time 05:56
        task ms_ec
          edit SMSTRIES '1'
      endfamily
      family ms072
        trigger ( /e_36r1/main:YMD gt /e_36r1/msjobs:YMD) or ( /e_36r1/main:YMD eq /e_36r1/msjobs:YMD and ( /e_36r1/main/00/fc/model:step gt 072 or /e_36r1/main/00/fc/model eq complete) and /e_36r1/main/00/prod/072/derived_param eq complete)
        edit EMOS_TIME_STEP_H '0072'
        edit MSJ_DIR 'fc00h072'
        time 06:06
        task ms_ec
          edit SMSTRIES '1'
      endfamily
      family ms144
        trigger ( /e_36r1/main:YMD gt /e_36r1/msjobs:YMD) or ( /e_36r1/main:YMD eq /e_36r1/msjobs:YMD and ( /e_36r1/main/00/fc/model:step gt 144 or /e_36r1/main/00/fc/model eq complete) and /e_36r1/main/00/prod/144/derived_param eq complete)
        edit EMOS_TIME_STEP_H '0144'
        edit MSJ_DIR 'fc00h144'
        time 06:27
        task ms_ec
          edit SMSTRIES '1'
      endfamily
      family ms240
        trigger ( /e_36r1/main:YMD gt /e_36r1/msjobs:YMD) or ( /e_36r1/main:YMD eq /e_36r1/msjobs:YMD and ( /e_36r1/main/00/fc/model:step gt 240 or /e_36r1/main/00/fc/model eq complete) and /e_36r1/main/00/prod/240/derived_param eq complete)
        edit EMOS_TIME_STEP_H '0240'
        edit MSJ_DIR 'fc00h240'
        time 06:55
        task ms_ec
          edit SMSTRIES '1'
      endfamily
      family mswave
        trigger ( /e_36r1/main:YMD gt /e_36r1/msjobs:YMD) or ( /e_36r1/main:YMD eq /e_36r1/msjobs:YMD and /e_36r1/main/00/fc/model eq complete)
        edit EMOS_TIME_STEP_H '240'
        edit MSJ_DIR 'wg00h240'
        time 06:55
        task ms_ec
          edit SMSTRIES '1'
      endfamily
      family msmetgram
        trigger ( /e_36r1/pop:YMD gt /e_36r1/msjobs:YMD) or ( /e_36r1/pop:YMD eq /e_36r1/msjobs:YMD and /e_36r1/pop/00/data/metgram eq complete)
        edit MSJ_DIR 'fc00hmetgram'
        time 06:55
        task ms_ec
          edit SMSTRIES '1'
      endfamily
      task test_ll
        edit MSJ_DIR 'test_ll'
        edit USE_NEW_MSJ 'true'
    endfamily
    family 12bc
      edit EMOS_BASE '12'
      edit EDFAMILY 'false'
      edit INISTREAM 'DA'
      edit REINITIALIZE 'false'
      edit REINIEXPVER '0046'
      edit INIEXPVER '0045'
      edit REINIPERIOD_4D '6'
      edit REINISTREAM 'DA'
      edit REINIOFFSET '6'
      edit RUN_HOURLY_PP '0'
      edit PERIOD_4D '6'
      edit WINDOW_LENGTH_4D '12'
      edit WINDOW_OFFSET_4D '9'
      edit MXUP_TRAJ '3'
      edit PARALLEL '4'
      edit LANONLY 'false'
      edit ACCOUNT 'opebcmis'
      edit STREAM 'SCDA'
      edit EMOS_STREAM 'SCDA'
      edit DELTA_DAY '-1'
      edit SMSFILES '/home/ma/emos/def/e_36r1/msjobs'
      edit SMSTRYNO '1'
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'ecgate'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit SMSOUT '/vol/emos_nc/output'
      edit LOGDIR '/vol/emos_nc/output'
      family ms012
        trigger ( /e_36r1/main:YMD gt /e_36r1/msjobs:YMD) or ( /e_36r1/main:YMD eq /e_36r1/msjobs:YMD and ( /e_36r1/main/12/fc eq complete or /e_36r1/main/12/fc/model:step gt 012))
        edit EMOS_TIME_STEP_H '0012'
        edit MSJ_DIR 'bc12h012'
        time 17:49
        task ms_ec
          edit SMSTRIES '1'
      endfamily
      family ms072
        trigger ( /e_36r1/main:YMD gt /e_36r1/msjobs:YMD) or ( /e_36r1/main:YMD eq /e_36r1/msjobs:YMD and ( /e_36r1/main/12/fc eq complete or /e_36r1/main/12/fc/model:step gt 072))
        edit EMOS_TIME_STEP_H '0072'
        edit MSJ_DIR 'bc12h072'
        time 18:06
        task ms_ec
          edit SMSTRIES '1'
      endfamily
    endfamily
    family 18bc
      edit EMOS_BASE '18'
      edit EDFAMILY 'false'
      edit INISTREAM 'DA'
      edit REINITIALIZE 'false'
      edit REINIEXPVER '0046'
      edit INIEXPVER '0045'
      edit REINIPERIOD_4D '6'
      edit REINISTREAM 'DA'
      edit REINIOFFSET '6'
      edit RUN_HOURLY_PP '0'
      edit PERIOD_4D '6'
      edit WINDOW_LENGTH_4D '12'
      edit WINDOW_OFFSET_4D '9'
      edit MXUP_TRAJ '3'
      edit PARALLEL '4'
      edit LANONLY 'false'
      edit ACCOUNT 'opebcmis'
      edit STREAM 'SCDA'
      edit EMOS_STREAM 'SCDA'
      edit DELTA_DAY '-1'
      edit SMSFILES '/home/ma/emos/def/e_36r1/msjobs'
      edit SMSTRYNO '1'
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'ecgate'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit SMSOUT '/vol/emos_nc/output'
      edit LOGDIR '/vol/emos_nc/output'
      family ms012
        trigger ( /e_36r1/main:YMD gt /e_36r1/msjobs:YMD) or ( /e_36r1/main:YMD eq /e_36r1/msjobs:YMD and ( /e_36r1/main/18bc/fc eq complete or /e_36r1/main/18bc/fc/model:step gt 012))
        edit EMOS_TIME_STEP_H '0012'
        edit MSJ_DIR 'bc18h012'
        time 23:49
        task ms_ec
          edit SMSTRIES '1'
      endfamily
      family ms072
        trigger ( /e_36r1/main:YMD gt /e_36r1/msjobs:YMD) or ( /e_36r1/main:YMD eq /e_36r1/msjobs:YMD and ( /e_36r1/main/18bc/fc eq complete or /e_36r1/main/18bc/fc/model:step gt 072))
        edit EMOS_TIME_STEP_H '0072'
        edit MSJ_DIR 'bc18h072'
        time 00:06
        task ms_ec
          edit SMSTRIES '1'
      endfamily
    endfamily
    family 00bc
      edit EMOS_BASE '00'
      edit EDFAMILY 'false'
      edit INISTREAM 'DA'
      edit REINITIALIZE 'false'
      edit REINIEXPVER '0046'
      edit INIEXPVER '0045'
      edit REINIPERIOD_4D '6'
      edit REINISTREAM 'DA'
      edit REINIOFFSET '6'
      edit RUN_HOURLY_PP '0'
      edit PERIOD_4D '6'
      edit WINDOW_LENGTH_4D '12'
      edit WINDOW_OFFSET_4D '9'
      edit MXUP_TRAJ '3'
      edit PARALLEL '4'
      edit LANONLY 'false'
      edit ACCOUNT 'opebcmis'
      edit STREAM 'SCDA'
      edit EMOS_STREAM 'SCDA'
      edit DELTA_DAY '0'
      edit SMSFILES '/home/ma/emos/def/e_36r1/msjobs'
      edit SMSTRYNO '1'
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'ecgate'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit SMSOUT '/vol/emos_nc/output'
      edit LOGDIR '/vol/emos_nc/output'
      family ms012
        trigger ( /e_36r1/main:YMD gt /e_36r1/msjobs:YMD) or ( /e_36r1/main:YMD eq /e_36r1/msjobs:YMD and ( /e_36r1/main/00/fc eq complete or /e_36r1/main/00/fc/model:step gt 012))
        edit EMOS_TIME_STEP_H '0012'
        edit MSJ_DIR 'bc00h012'
        time 05:49
        task ms_ec
          edit SMSTRIES '1'
      endfamily
      family ms072
        trigger ( /e_36r1/main:YMD gt /e_36r1/msjobs:YMD) or ( /e_36r1/main:YMD eq /e_36r1/msjobs:YMD and ( /e_36r1/main/00/fc eq complete or /e_36r1/main/00/fc/model:step gt 072))
        edit EMOS_TIME_STEP_H '0072'
        edit MSJ_DIR 'bc00h072'
        time 06:06
        task ms_ec
          edit SMSTRIES '1'
      endfamily
    endfamily
    family 06bc
      edit EMOS_BASE '06'
      edit EDFAMILY 'false'
      edit INISTREAM 'DA'
      edit REINITIALIZE 'false'
      edit REINIEXPVER '0046'
      edit INIEXPVER '0045'
      edit REINIPERIOD_4D '6'
      edit REINISTREAM 'DA'
      edit REINIOFFSET '6'
      edit RUN_HOURLY_PP '0'
      edit PERIOD_4D '6'
      edit WINDOW_LENGTH_4D '12'
      edit WINDOW_OFFSET_4D '9'
      edit MXUP_TRAJ '3'
      edit PARALLEL '4'
      edit LANONLY 'false'
      edit ACCOUNT 'opebcmis'
      edit STREAM 'SCDA'
      edit EMOS_STREAM 'SCDA'
      edit DELTA_DAY '0'
      edit SMSFILES '/home/ma/emos/def/e_36r1/msjobs'
      edit SMSTRYNO '1'
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'ecgate'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit SMSOUT '/vol/emos_nc/output'
      edit LOGDIR '/vol/emos_nc/output'
      family ms012
        trigger ( /e_36r1/main:YMD gt /e_36r1/msjobs:YMD) or ( /e_36r1/main:YMD eq /e_36r1/msjobs:YMD and ( /e_36r1/main/06bc/fc eq complete or /e_36r1/main/06bc/fc/model:step gt 012))
        edit EMOS_TIME_STEP_H '0012'
        edit MSJ_DIR 'bc06h012'
        time 11:49
        task ms_ec
          edit SMSTRIES '1'
      endfamily
      family ms072
        trigger ( /e_36r1/main:YMD gt /e_36r1/msjobs:YMD) or ( /e_36r1/main:YMD eq /e_36r1/msjobs:YMD and ( /e_36r1/main/06bc/fc eq complete or /e_36r1/main/06bc/fc/model:step gt 072))
        edit EMOS_TIME_STEP_H '0072'
        edit MSJ_DIR 'bc06h072'
        time 12:06
        task ms_ec
          edit SMSTRIES '1'
      endfamily
    endfamily
    task logfilestore
      edit QUEUE 'emos-e'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit SMSFILES '/home/ma/emos/def/e_36r1'
      edit LOGTASK 'msjobs'
      edit KEEPLOGS 'no'
      edit LOGSTOECFS 'no'
      edit MOVE_TO_ECFS 'yes'
    task logfiles
      late -s +00:05 -c +00:15
      trigger ./12 eq complete and ./00 eq complete and ./12bc eq complete and ./18bc eq complete and ./00bc eq complete and ./06bc eq complete and logfilestore eq complete
      edit QUEUE 'emos-e'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit SMSFILES '/home/ma/emos/def/e_36r1'
      edit LOGTASK 'msjobs'
      edit MOVE_TO_ECFS 'no'
      time 14:20
  endfamily
endsuite
suite emc_36r1
  edit USER 'emos'
  edit SCHOST 'c1b'
  edit STHOST '/s1b'
  edit SCHOST_BKUP 'c1a'
  edit STHOST_BKUP '/s1a'
  edit WSHOST 'linux_cluster'
  edit WSLOGDIR '/vol/emos_nc/output'
  edit SCLOGDIR '/s1b/emos_esuite/emos_data/0046/log'
  edit USE_YMD 'true'
  edit YMD '20091003'
  edit DATEMASK '*.*.*'
  edit VERSION '0046'
  edit INIEXPVER '0046'
  edit TCINIEXPVER '0046'
  edit FIRST_DAY '1'
  edit SMSTRIES '2'
  edit DELTA_DAY '0'
  edit EMOS_BASE '12'
  edit EMOS_TIME_STEP_H '0000'
  edit FSFAMILY 'mc'
  edit IMPL_DATE '2012123112'
  edit EVERSION '0001'
  edit USER_PRIORITY '70'
  edit ININPES '24'
  edit FCNPES '32'
  edit PGNPES '32'
  edit ANNPES '4'
  edit REINITIALIZE 'false'
  edit RESERVATION_ON '0'
  edit SMSHOME '/vol/emos_nc/output'
  edit SMSFILES '/home/ma/emos/def/emc_36r1'
  edit SMSINCLUDE '/home/ma/emos/def/emc_36r1/include'
  edit ACCOUNT 'oeenfo'
  edit STREAM 'EF'
  edit EMOS_STREAM 'EF'
  edit ENSEMBLES '50'
  edit ENS_SIZE '50'
  edit MARS_DB 'marsod'
  edit EPSTYPE 'fc'
  edit MEMBER '0'
  edit INILEVELS '91'
  edit INIPERIOD_4D '6'
  edit IS_REAL_TIME 'is_real_time'
  edit RUN_3HOURS_PP '1'
  edit RUNHINDCAST '0'
  edit ERA40END '2002073100'
  edit NH_YEARS '18'
  edit NH_WEEKS '2'
  edit PGEN_BINDIR '/home/ma/emos/bin/diss'
  edit PGEN_OBJDIR '/home/ma/emos/bin/diss'
  edit DISS_REQDB_ROOT '/diss/requirements'
  edit DISS_BINDIR '/diss/bin/DISS'
  edit DELTA_DAY_M '-15'
  edit DELTA_DAY_P '25'
  edit FCGTYPE 'l_2'
  edit EPSFCGTYPE 'l_2'
  edit LUSE_RSTRHBIAS 'true'
  edit FCGROUP 'fc0'
  edit SMSLOGHOST 'c1b'
  edit SMSLOGPORT '9316'
  edit COURTESY '0'
  edit OCNPES_OC '8'
  edit OCNPES_IFS '56'
  edit OD_PROJ 'od'
  edit SMSURLBASE 'http://datasvc.ecmwf.int/twiki/bin/view/Main'
  edit SMSURL 'CurrentAndOldESuites'
  edit SUITE_START '20091004'
  edit PRODUCE_EPS_ML '1'
  edit QUEUE 'ns'
  edit QUEUE_EPILOG 'ns'
  edit SMSCMD 'smssubmit.map %USER% %SCHOST% %SMSJOB% %SMSJOBOUT%'
  edit SMSKILL 'smskill %USER% %SCHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
  edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %SCHOST% %SMSRID% %SMSJOB%'
  edit LOGDIR '/s1b/emos_esuite/emos_data/0046/log'
  edit SMSOUT '/s1b/emos_esuite/emos_data/0046/log'
  edit NOEFI '0'
  edit SLEEPTIME '1'
  inlimit /limits:c1b
  clock hybrid 
  family hc0
    repeat date YMD 20091008 20121212 7
    edit ACCOUNT 'oeenfo'
    edit LOGTASK 'hc0'
    edit QUEUE 'ns'
    edit QUEUE_EPILOG 'ns'
    edit ENSEMBLES '4'
    edit ENS_SIZE '4'
    edit RUNHINDCAST '1'
    edit STREAM 'ENFH'
    edit EMOS_STREAM 'ENFH'
    edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
    edit EMOS_BASE '00'
    edit INICLASS 'EI'
    edit INIEXPVER '0001'
    edit INILEVELS '60'
    edit DELETE_FAMILY 'true'
    family back
      trigger /emc_36r1/main/make eq complete
      edit RUNVARFC '1'
      edit YMD '20091217'
      inlimit /emc_36r1/limits:hind
      task stage_hc
        edit RUNVARFC '1'
        edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
      task rmfdb
        trigger stage_hc eq complete
        edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
        edit NH_WEEKS_KEPT '1'
      task lock_fam
        trigger 0 eq 1
      task lock_1701
        complete ./18 eq complete
        trigger 0 eq 1
      family 18
        trigger ./rmfdb eq complete
        edit DELTA_YEAR '-18'
        family m2
          edit DELTA_DAY '-2'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '3000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit FCNPES '16'
          edit EPSSVTSTEP '900'
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit CPUTIME '12000'
          edit STACK '350'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
          task svsh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '73'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '74'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnhevo
            trigger svnh eq complete
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit CPUTIME '120000'
          task svshevo
            trigger svsh eq complete
            edit EVOLVE 'yes'
          task svsave
            trigger svshevo eq complete and svnhevo eq complete
          task clean
            trigger svnhevo eq complete and svshevo eq complete and svsave eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
        endfamily
        family sv
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '2000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit NPES '16'
          edit EPSSVTSTEP '900'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit THREADS '8'
          task targets
            late -s +00:03 -c +00:10
            edit EPSTCBB '1'
            label da_info ""
            label dcda_info ""
            event sv
          task svsh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '73'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '74'
            inlimit /emc_36r1/limits:pesh 2
          task subspace
            trigger svnh eq complete and svsh eq complete and ( targets eq complete or targets:sv)
          task sv1
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '1'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '72'
            inlimit /emc_36r1/limits:pesh 2
          task sv2
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '2'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '71'
            inlimit /emc_36r1/limits:pesh 2
          task sv3
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '3'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv4
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '4'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv5
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '5'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv6
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '6'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task getsvevo
            event ecfs
          task getsvs
            trigger getsvevo eq complete and svnh eq complete and svsh eq complete and sv1 eq complete and sv2 eq complete and sv3 eq complete and sv4 eq complete and sv5 eq complete and sv6 eq complete and ../m2 eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTROP '1'
          task svnhevo
            trigger svnh eq complete and ../legA/fc eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit THREADS '4'
            edit STACK '350'
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
          task svshevo
            trigger svsh eq complete and ../legA/fc eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit STACK '450'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
        endfamily
        family legA
          family fc
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '240'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit TSTEP '1200'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '1'
            task getini
              edit LINIENDA '0'
              edit INIDATEENDA '0'
              edit EPSCOUPLE '0'
              edit N_PERT_AN '10'
              edit EPSCOUPL_A '0'
              edit EPSCOUPL_B '0'
              edit EPSCOUPL_C '0'
            task inidata
              trigger getini eq complete
              edit THREADS '8'
            task getae
              event 1 
            task rot
              trigger ../../sv/getsvs eq complete and ( getae:1 or getae eq complete)
              edit EPSTROP '1'
            family pert_ic
              trigger rot eq complete and inidata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTROP '1'
              inlimit /emc_36r1/limits:pertic
              family 00
                task pertinic
                  edit MEMBER '0'
              endfamily
              family 01
                task pertinic
                  edit MEMBER '1'
              endfamily
              family 02
                task pertinic
                  edit MEMBER '2'
              endfamily
              family 03
                task pertinic
                  edit MEMBER '3'
              endfamily
              family 04
                task pertinic
                  edit MEMBER '4'
              endfamily
            endfamily
            task cp_pert
              trigger pert_ic eq complete
              event 1 
            task wavini
              edit FSFAMILY 'mc'
              edit PERSST 'true'
              edit SSTINT '1'
            task wavfcdata
              trigger ./getpersSST eq complete and ../../legA/fc/wavini eq complete
              edit EPSCOUPLE '1'
            task getpersSST
              trigger ./inidata eq complete
              edit EPSCOUPLE '1'
              edit EPS_INI_STREAM 'DA'
              edit EPSFCRES '639'
            family pf
              trigger ./getpersSST eq complete and ./wavfcdata eq complete and wavini eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              meter step -1 240 240
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task getiniLeg
                  trigger ../../pert_ic/01/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task getiniLeg
                  trigger ../../pert_ic/02/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task getiniLeg
                  trigger ../../pert_ic/03/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task getiniLeg
                  trigger ../../pert_ic/04/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
            endfamily
            family cf
              trigger ./pert_ic/00/pertinic eq complete and ./getpersSST eq complete and ./wavfcdata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              task getiniLeg
                trigger ../inidata eq complete and ../wavini eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit FCNPES '32'
                edit THREADS '4'
                inlimit /emc_36r1/limits:pesh 6
                meter step -1 240 240
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../fc eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:dp
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        family legBcpl
          trigger ./legA eq complete
          family fc
            edit EPSCOUPLE '1'
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '144'
            edit EPSINISTEP_PREVIOUS '216'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit TSTEP '2700'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '2'
            family pf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit THREADS '2'
              edit RUNVARFC '1'
              meter step -1 552 552
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/01/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/01 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/02/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/02 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/03/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/03 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/04/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/04 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
            endfamily
            family cf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc0'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger ../../../../legA/fc/cf/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:pesh
              endfamily
              task getvarepsdata
                trigger ../../../legA/fc/cf eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '552'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit SMSTRIES '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:pesh
                meter step -1 552 552
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../../legA/postproc/dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:dp
              family dp
                trigger ./cf eq complete and ./pf eq complete
                edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
                edit OVERLAP '1'
                edit STEPS '216:240:6'
                inlimit /emc_36r1/limits:dp
                family cf
                  edit EMOS_TYPE 'cf'
                  task derived_param
                    edit MEMBER 'off'
                endfamily
                family pf
                  edit EMOS_TYPE 'pf'
                  family 01
                    edit EPSMEMBER '1'
                    edit MEMBER '1'
                    edit SUBFSFAMILY '/pf001'
                    task derived_param
                  endfamily
                  family 02
                    edit EPSMEMBER '2'
                    edit MEMBER '2'
                    edit SUBFSFAMILY '/pf002'
                    task derived_param
                  endfamily
                  family 03
                    edit EPSMEMBER '3'
                    edit MEMBER '3'
                    edit SUBFSFAMILY '/pf003'
                    task derived_param
                  endfamily
                  family 04
                    edit EPSMEMBER '4'
                    edit MEMBER '4'
                    edit SUBFSFAMILY '/pf004'
                    task derived_param
                  endfamily
                endfamily
              endfamily
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                edit EPSFCRES '319'
                task extrafields
                  edit FCLENGTH '768'
                  edit FCLENGTH0 '240'
                task wm_sfc
                  trigger extrafields eq complete
                  edit FCLENGTH '768'
                task wm_ua
                  trigger extrafields eq complete and wm_sfc eq complete
                  edit FCLENGTH '768'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        task save_logs
          trigger m2 eq complete and sv eq complete and legA eq complete and legBcpl eq complete
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
      endfamily
      family 17
        trigger ( ./rmfdb eq complete) and ( ./lock_1701 eq complete)
        edit DELTA_YEAR '-17'
        family m2
          edit DELTA_DAY '-2'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '3000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit FCNPES '16'
          edit EPSSVTSTEP '900'
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit CPUTIME '12000'
          edit STACK '350'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
          task svsh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '73'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '74'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnhevo
            trigger svnh eq complete
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit CPUTIME '120000'
          task svshevo
            trigger svsh eq complete
            edit EVOLVE 'yes'
          task svsave
            trigger svshevo eq complete and svnhevo eq complete
          task clean
            trigger svnhevo eq complete and svshevo eq complete and svsave eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
        endfamily
        family sv
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '2000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit NPES '16'
          edit EPSSVTSTEP '900'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit THREADS '8'
          task targets
            late -s +00:03 -c +00:10
            edit EPSTCBB '1'
            label da_info ""
            label dcda_info ""
            event sv
          task svsh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '73'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '74'
            inlimit /emc_36r1/limits:pesh 2
          task subspace
            trigger svnh eq complete and svsh eq complete and ( targets eq complete or targets:sv)
          task sv1
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '1'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '72'
            inlimit /emc_36r1/limits:pesh 2
          task sv2
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '2'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '71'
            inlimit /emc_36r1/limits:pesh 2
          task sv3
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '3'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv4
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '4'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv5
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '5'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv6
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '6'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task getsvevo
            event ecfs
          task getsvs
            trigger getsvevo eq complete and svnh eq complete and svsh eq complete and sv1 eq complete and sv2 eq complete and sv3 eq complete and sv4 eq complete and sv5 eq complete and sv6 eq complete and ../m2 eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTROP '1'
          task svnhevo
            trigger svnh eq complete and ../legA/fc eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit THREADS '4'
            edit STACK '350'
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
          task svshevo
            trigger svsh eq complete and ../legA/fc eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit STACK '450'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
        endfamily
        family legA
          family fc
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '240'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit TSTEP '1200'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '1'
            task getini
              edit LINIENDA '0'
              edit INIDATEENDA '0'
              edit EPSCOUPLE '0'
              edit N_PERT_AN '10'
              edit EPSCOUPL_A '0'
              edit EPSCOUPL_B '0'
              edit EPSCOUPL_C '0'
            task inidata
              trigger getini eq complete
              edit THREADS '8'
            task getae
              event 1 
            task rot
              trigger ../../sv/getsvs eq complete and ( getae:1 or getae eq complete)
              edit EPSTROP '1'
            family pert_ic
              trigger rot eq complete and inidata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTROP '1'
              inlimit /emc_36r1/limits:pertic
              family 00
                task pertinic
                  edit MEMBER '0'
              endfamily
              family 01
                task pertinic
                  edit MEMBER '1'
              endfamily
              family 02
                task pertinic
                  edit MEMBER '2'
              endfamily
              family 03
                task pertinic
                  edit MEMBER '3'
              endfamily
              family 04
                task pertinic
                  edit MEMBER '4'
              endfamily
            endfamily
            task cp_pert
              trigger pert_ic eq complete
              event 1 
            task wavini
              edit FSFAMILY 'mc'
              edit PERSST 'true'
              edit SSTINT '1'
            task wavfcdata
              trigger ./getpersSST eq complete and ../../legA/fc/wavini eq complete
              edit EPSCOUPLE '1'
            task getpersSST
              trigger ./inidata eq complete
              edit EPSCOUPLE '1'
              edit EPS_INI_STREAM 'DA'
              edit EPSFCRES '639'
            family pf
              trigger ./getpersSST eq complete and ./wavfcdata eq complete and wavini eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              meter step -1 240 240
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task getiniLeg
                  trigger ../../pert_ic/01/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task getiniLeg
                  trigger ../../pert_ic/02/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task getiniLeg
                  trigger ../../pert_ic/03/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task getiniLeg
                  trigger ../../pert_ic/04/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
            endfamily
            family cf
              trigger ./pert_ic/00/pertinic eq complete and ./getpersSST eq complete and ./wavfcdata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              task getiniLeg
                trigger ../inidata eq complete and ../wavini eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit FCNPES '32'
                edit THREADS '4'
                inlimit /emc_36r1/limits:pesh 6
                meter step -1 240 240
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../fc eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:dp
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        family legBcpl
          trigger ./legA eq complete
          family fc
            edit EPSCOUPLE '1'
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '144'
            edit EPSINISTEP_PREVIOUS '216'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit TSTEP '2700'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '2'
            family pf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit THREADS '2'
              edit RUNVARFC '1'
              meter step -1 552 552
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/01/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/01 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/02/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/02 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/03/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/03 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/04/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/04 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
            endfamily
            family cf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc0'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger ../../../../legA/fc/cf/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:pesh
              endfamily
              task getvarepsdata
                trigger ../../../legA/fc/cf eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '552'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit SMSTRIES '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:pesh
                meter step -1 552 552
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../../legA/postproc/dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:dp
              family dp
                trigger ./cf eq complete and ./pf eq complete
                edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
                edit OVERLAP '1'
                edit STEPS '216:240:6'
                inlimit /emc_36r1/limits:dp
                family cf
                  edit EMOS_TYPE 'cf'
                  task derived_param
                    edit MEMBER 'off'
                endfamily
                family pf
                  edit EMOS_TYPE 'pf'
                  family 01
                    edit EPSMEMBER '1'
                    edit MEMBER '1'
                    edit SUBFSFAMILY '/pf001'
                    task derived_param
                  endfamily
                  family 02
                    edit EPSMEMBER '2'
                    edit MEMBER '2'
                    edit SUBFSFAMILY '/pf002'
                    task derived_param
                  endfamily
                  family 03
                    edit EPSMEMBER '3'
                    edit MEMBER '3'
                    edit SUBFSFAMILY '/pf003'
                    task derived_param
                  endfamily
                  family 04
                    edit EPSMEMBER '4'
                    edit MEMBER '4'
                    edit SUBFSFAMILY '/pf004'
                    task derived_param
                  endfamily
                endfamily
              endfamily
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                edit EPSFCRES '319'
                task extrafields
                  edit FCLENGTH '768'
                  edit FCLENGTH0 '240'
                task wm_sfc
                  trigger extrafields eq complete
                  edit FCLENGTH '768'
                task wm_ua
                  trigger extrafields eq complete and wm_sfc eq complete
                  edit FCLENGTH '768'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        task save_logs
          trigger m2 eq complete and sv eq complete and legA eq complete and legBcpl eq complete
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
      endfamily
      family 16
        trigger ( ./rmfdb eq complete) and ( ./lock_1701 eq complete)
        edit DELTA_YEAR '-16'
        family m2
          edit DELTA_DAY '-2'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '3000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit FCNPES '16'
          edit EPSSVTSTEP '900'
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit CPUTIME '12000'
          edit STACK '350'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
          task svsh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '73'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '74'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnhevo
            trigger svnh eq complete
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit CPUTIME '120000'
          task svshevo
            trigger svsh eq complete
            edit EVOLVE 'yes'
          task svsave
            trigger svshevo eq complete and svnhevo eq complete
          task clean
            trigger svnhevo eq complete and svshevo eq complete and svsave eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
        endfamily
        family sv
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '2000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit NPES '16'
          edit EPSSVTSTEP '900'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit THREADS '8'
          task targets
            late -s +00:03 -c +00:10
            edit EPSTCBB '1'
            label da_info ""
            label dcda_info ""
            event sv
          task svsh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '73'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '74'
            inlimit /emc_36r1/limits:pesh 2
          task subspace
            trigger svnh eq complete and svsh eq complete and ( targets eq complete or targets:sv)
          task sv1
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '1'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '72'
            inlimit /emc_36r1/limits:pesh 2
          task sv2
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '2'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '71'
            inlimit /emc_36r1/limits:pesh 2
          task sv3
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '3'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv4
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '4'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv5
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '5'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv6
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '6'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task getsvevo
            event ecfs
          task getsvs
            trigger getsvevo eq complete and svnh eq complete and svsh eq complete and sv1 eq complete and sv2 eq complete and sv3 eq complete and sv4 eq complete and sv5 eq complete and sv6 eq complete and ../m2 eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTROP '1'
          task svnhevo
            trigger svnh eq complete and ../legA/fc eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit THREADS '4'
            edit STACK '350'
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
          task svshevo
            trigger svsh eq complete and ../legA/fc eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit STACK '450'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
        endfamily
        family legA
          family fc
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '240'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit TSTEP '1200'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '1'
            task getini
              edit LINIENDA '0'
              edit INIDATEENDA '0'
              edit EPSCOUPLE '0'
              edit N_PERT_AN '10'
              edit EPSCOUPL_A '0'
              edit EPSCOUPL_B '0'
              edit EPSCOUPL_C '0'
            task inidata
              trigger getini eq complete
              edit THREADS '8'
            task getae
              event 1 
            task rot
              trigger ../../sv/getsvs eq complete and ( getae:1 or getae eq complete)
              edit EPSTROP '1'
            family pert_ic
              trigger rot eq complete and inidata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTROP '1'
              inlimit /emc_36r1/limits:pertic
              family 00
                task pertinic
                  edit MEMBER '0'
              endfamily
              family 01
                task pertinic
                  edit MEMBER '1'
              endfamily
              family 02
                task pertinic
                  edit MEMBER '2'
              endfamily
              family 03
                task pertinic
                  edit MEMBER '3'
              endfamily
              family 04
                task pertinic
                  edit MEMBER '4'
              endfamily
            endfamily
            task cp_pert
              trigger pert_ic eq complete
              event 1 
            task wavini
              edit FSFAMILY 'mc'
              edit PERSST 'true'
              edit SSTINT '1'
            task wavfcdata
              trigger ./getpersSST eq complete and ../../legA/fc/wavini eq complete
              edit EPSCOUPLE '1'
            task getpersSST
              trigger ./inidata eq complete
              edit EPSCOUPLE '1'
              edit EPS_INI_STREAM 'DA'
              edit EPSFCRES '639'
            family pf
              trigger ./getpersSST eq complete and ./wavfcdata eq complete and wavini eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              meter step -1 240 240
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task getiniLeg
                  trigger ../../pert_ic/01/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task getiniLeg
                  trigger ../../pert_ic/02/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task getiniLeg
                  trigger ../../pert_ic/03/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task getiniLeg
                  trigger ../../pert_ic/04/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
            endfamily
            family cf
              trigger ./pert_ic/00/pertinic eq complete and ./getpersSST eq complete and ./wavfcdata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              task getiniLeg
                trigger ../inidata eq complete and ../wavini eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit FCNPES '32'
                edit THREADS '4'
                inlimit /emc_36r1/limits:pesh 6
                meter step -1 240 240
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../fc eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:dp
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        family legBcpl
          trigger ./legA eq complete
          family fc
            edit EPSCOUPLE '1'
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '144'
            edit EPSINISTEP_PREVIOUS '216'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit TSTEP '2700'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '2'
            family pf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit THREADS '2'
              edit RUNVARFC '1'
              meter step -1 552 552
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/01/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/01 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/02/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/02 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/03/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/03 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/04/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/04 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
            endfamily
            family cf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc0'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger ../../../../legA/fc/cf/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:pesh
              endfamily
              task getvarepsdata
                trigger ../../../legA/fc/cf eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '552'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit SMSTRIES '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:pesh
                meter step -1 552 552
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../../legA/postproc/dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:dp
              family dp
                trigger ./cf eq complete and ./pf eq complete
                edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
                edit OVERLAP '1'
                edit STEPS '216:240:6'
                inlimit /emc_36r1/limits:dp
                family cf
                  edit EMOS_TYPE 'cf'
                  task derived_param
                    edit MEMBER 'off'
                endfamily
                family pf
                  edit EMOS_TYPE 'pf'
                  family 01
                    edit EPSMEMBER '1'
                    edit MEMBER '1'
                    edit SUBFSFAMILY '/pf001'
                    task derived_param
                  endfamily
                  family 02
                    edit EPSMEMBER '2'
                    edit MEMBER '2'
                    edit SUBFSFAMILY '/pf002'
                    task derived_param
                  endfamily
                  family 03
                    edit EPSMEMBER '3'
                    edit MEMBER '3'
                    edit SUBFSFAMILY '/pf003'
                    task derived_param
                  endfamily
                  family 04
                    edit EPSMEMBER '4'
                    edit MEMBER '4'
                    edit SUBFSFAMILY '/pf004'
                    task derived_param
                  endfamily
                endfamily
              endfamily
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                edit EPSFCRES '319'
                task extrafields
                  edit FCLENGTH '768'
                  edit FCLENGTH0 '240'
                task wm_sfc
                  trigger extrafields eq complete
                  edit FCLENGTH '768'
                task wm_ua
                  trigger extrafields eq complete and wm_sfc eq complete
                  edit FCLENGTH '768'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        task save_logs
          trigger m2 eq complete and sv eq complete and legA eq complete and legBcpl eq complete
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
      endfamily
      family 15
        trigger ( ./rmfdb eq complete) and ( ./lock_1701 eq complete)
        edit DELTA_YEAR '-15'
        family m2
          edit DELTA_DAY '-2'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '3000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit FCNPES '16'
          edit EPSSVTSTEP '900'
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit CPUTIME '12000'
          edit STACK '350'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
          task svsh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '73'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '74'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnhevo
            trigger svnh eq complete
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit CPUTIME '120000'
          task svshevo
            trigger svsh eq complete
            edit EVOLVE 'yes'
          task svsave
            trigger svshevo eq complete and svnhevo eq complete
          task clean
            trigger svnhevo eq complete and svshevo eq complete and svsave eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
        endfamily
        family sv
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '2000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit NPES '16'
          edit EPSSVTSTEP '900'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit THREADS '8'
          task targets
            late -s +00:03 -c +00:10
            edit EPSTCBB '1'
            label da_info ""
            label dcda_info ""
            event sv
          task svsh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '73'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '74'
            inlimit /emc_36r1/limits:pesh 2
          task subspace
            trigger svnh eq complete and svsh eq complete and ( targets eq complete or targets:sv)
          task sv1
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '1'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '72'
            inlimit /emc_36r1/limits:pesh 2
          task sv2
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '2'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '71'
            inlimit /emc_36r1/limits:pesh 2
          task sv3
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '3'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv4
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '4'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv5
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '5'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv6
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '6'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task getsvevo
            event ecfs
          task getsvs
            trigger getsvevo eq complete and svnh eq complete and svsh eq complete and sv1 eq complete and sv2 eq complete and sv3 eq complete and sv4 eq complete and sv5 eq complete and sv6 eq complete and ../m2 eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTROP '1'
          task svnhevo
            trigger svnh eq complete and ../legA/fc eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit THREADS '4'
            edit STACK '350'
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
          task svshevo
            trigger svsh eq complete and ../legA/fc eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit STACK '450'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
        endfamily
        family legA
          family fc
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '240'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit TSTEP '1200'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '1'
            task getini
              edit LINIENDA '0'
              edit INIDATEENDA '0'
              edit EPSCOUPLE '0'
              edit N_PERT_AN '10'
              edit EPSCOUPL_A '0'
              edit EPSCOUPL_B '0'
              edit EPSCOUPL_C '0'
            task inidata
              trigger getini eq complete
              edit THREADS '8'
            task getae
              event 1 
            task rot
              trigger ../../sv/getsvs eq complete and ( getae:1 or getae eq complete)
              edit EPSTROP '1'
            family pert_ic
              trigger rot eq complete and inidata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTROP '1'
              inlimit /emc_36r1/limits:pertic
              family 00
                task pertinic
                  edit MEMBER '0'
              endfamily
              family 01
                task pertinic
                  edit MEMBER '1'
              endfamily
              family 02
                task pertinic
                  edit MEMBER '2'
              endfamily
              family 03
                task pertinic
                  edit MEMBER '3'
              endfamily
              family 04
                task pertinic
                  edit MEMBER '4'
              endfamily
            endfamily
            task cp_pert
              trigger pert_ic eq complete
              event 1 
            task wavini
              edit FSFAMILY 'mc'
              edit PERSST 'true'
              edit SSTINT '1'
            task wavfcdata
              trigger ./getpersSST eq complete and ../../legA/fc/wavini eq complete
              edit EPSCOUPLE '1'
            task getpersSST
              trigger ./inidata eq complete
              edit EPSCOUPLE '1'
              edit EPS_INI_STREAM 'DA'
              edit EPSFCRES '639'
            family pf
              trigger ./getpersSST eq complete and ./wavfcdata eq complete and wavini eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              meter step -1 240 240
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task getiniLeg
                  trigger ../../pert_ic/01/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task getiniLeg
                  trigger ../../pert_ic/02/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task getiniLeg
                  trigger ../../pert_ic/03/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task getiniLeg
                  trigger ../../pert_ic/04/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
            endfamily
            family cf
              trigger ./pert_ic/00/pertinic eq complete and ./getpersSST eq complete and ./wavfcdata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              task getiniLeg
                trigger ../inidata eq complete and ../wavini eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit FCNPES '32'
                edit THREADS '4'
                inlimit /emc_36r1/limits:pesh 6
                meter step -1 240 240
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../fc eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:dp
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        family legBcpl
          trigger ./legA eq complete
          family fc
            edit EPSCOUPLE '1'
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '144'
            edit EPSINISTEP_PREVIOUS '216'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit TSTEP '2700'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '2'
            family pf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit THREADS '2'
              edit RUNVARFC '1'
              meter step -1 552 552
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/01/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/01 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/02/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/02 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/03/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/03 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/04/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/04 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
            endfamily
            family cf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc0'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger ../../../../legA/fc/cf/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:pesh
              endfamily
              task getvarepsdata
                trigger ../../../legA/fc/cf eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '552'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit SMSTRIES '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:pesh
                meter step -1 552 552
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../../legA/postproc/dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:dp
              family dp
                trigger ./cf eq complete and ./pf eq complete
                edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
                edit OVERLAP '1'
                edit STEPS '216:240:6'
                inlimit /emc_36r1/limits:dp
                family cf
                  edit EMOS_TYPE 'cf'
                  task derived_param
                    edit MEMBER 'off'
                endfamily
                family pf
                  edit EMOS_TYPE 'pf'
                  family 01
                    edit EPSMEMBER '1'
                    edit MEMBER '1'
                    edit SUBFSFAMILY '/pf001'
                    task derived_param
                  endfamily
                  family 02
                    edit EPSMEMBER '2'
                    edit MEMBER '2'
                    edit SUBFSFAMILY '/pf002'
                    task derived_param
                  endfamily
                  family 03
                    edit EPSMEMBER '3'
                    edit MEMBER '3'
                    edit SUBFSFAMILY '/pf003'
                    task derived_param
                  endfamily
                  family 04
                    edit EPSMEMBER '4'
                    edit MEMBER '4'
                    edit SUBFSFAMILY '/pf004'
                    task derived_param
                  endfamily
                endfamily
              endfamily
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                edit EPSFCRES '319'
                task extrafields
                  edit FCLENGTH '768'
                  edit FCLENGTH0 '240'
                task wm_sfc
                  trigger extrafields eq complete
                  edit FCLENGTH '768'
                task wm_ua
                  trigger extrafields eq complete and wm_sfc eq complete
                  edit FCLENGTH '768'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        task save_logs
          trigger m2 eq complete and sv eq complete and legA eq complete and legBcpl eq complete
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
      endfamily
      family 14
        trigger ( ./rmfdb eq complete) and ( ./lock_1701 eq complete)
        edit DELTA_YEAR '-14'
        family m2
          edit DELTA_DAY '-2'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '3000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit FCNPES '16'
          edit EPSSVTSTEP '900'
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit CPUTIME '12000'
          edit STACK '350'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
          task svsh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '73'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '74'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnhevo
            trigger svnh eq complete
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit CPUTIME '120000'
          task svshevo
            trigger svsh eq complete
            edit EVOLVE 'yes'
          task svsave
            trigger svshevo eq complete and svnhevo eq complete
          task clean
            trigger svnhevo eq complete and svshevo eq complete and svsave eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
        endfamily
        family sv
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '2000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit NPES '16'
          edit EPSSVTSTEP '900'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit THREADS '8'
          task targets
            late -s +00:03 -c +00:10
            edit EPSTCBB '1'
            label da_info ""
            label dcda_info ""
            event sv
          task svsh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '73'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '74'
            inlimit /emc_36r1/limits:pesh 2
          task subspace
            trigger svnh eq complete and svsh eq complete and ( targets eq complete or targets:sv)
          task sv1
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '1'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '72'
            inlimit /emc_36r1/limits:pesh 2
          task sv2
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '2'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '71'
            inlimit /emc_36r1/limits:pesh 2
          task sv3
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '3'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv4
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '4'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv5
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '5'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv6
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '6'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task getsvevo
            event ecfs
          task getsvs
            trigger getsvevo eq complete and svnh eq complete and svsh eq complete and sv1 eq complete and sv2 eq complete and sv3 eq complete and sv4 eq complete and sv5 eq complete and sv6 eq complete and ../m2 eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTROP '1'
          task svnhevo
            trigger svnh eq complete and ../legA/fc eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit THREADS '4'
            edit STACK '350'
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
          task svshevo
            trigger svsh eq complete and ../legA/fc eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit STACK '450'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
        endfamily
        family legA
          family fc
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '240'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit TSTEP '1200'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '1'
            task getini
              edit LINIENDA '0'
              edit INIDATEENDA '0'
              edit EPSCOUPLE '0'
              edit N_PERT_AN '10'
              edit EPSCOUPL_A '0'
              edit EPSCOUPL_B '0'
              edit EPSCOUPL_C '0'
            task inidata
              trigger getini eq complete
              edit THREADS '8'
            task getae
              event 1 
            task rot
              trigger ../../sv/getsvs eq complete and ( getae:1 or getae eq complete)
              edit EPSTROP '1'
            family pert_ic
              trigger rot eq complete and inidata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTROP '1'
              inlimit /emc_36r1/limits:pertic
              family 00
                task pertinic
                  edit MEMBER '0'
              endfamily
              family 01
                task pertinic
                  edit MEMBER '1'
              endfamily
              family 02
                task pertinic
                  edit MEMBER '2'
              endfamily
              family 03
                task pertinic
                  edit MEMBER '3'
              endfamily
              family 04
                task pertinic
                  edit MEMBER '4'
              endfamily
            endfamily
            task cp_pert
              trigger pert_ic eq complete
              event 1 
            task wavini
              edit FSFAMILY 'mc'
              edit PERSST 'true'
              edit SSTINT '1'
            task wavfcdata
              trigger ./getpersSST eq complete and ../../legA/fc/wavini eq complete
              edit EPSCOUPLE '1'
            task getpersSST
              trigger ./inidata eq complete
              edit EPSCOUPLE '1'
              edit EPS_INI_STREAM 'DA'
              edit EPSFCRES '639'
            family pf
              trigger ./getpersSST eq complete and ./wavfcdata eq complete and wavini eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              meter step -1 240 240
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task getiniLeg
                  trigger ../../pert_ic/01/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task getiniLeg
                  trigger ../../pert_ic/02/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task getiniLeg
                  trigger ../../pert_ic/03/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task getiniLeg
                  trigger ../../pert_ic/04/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
            endfamily
            family cf
              trigger ./pert_ic/00/pertinic eq complete and ./getpersSST eq complete and ./wavfcdata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              task getiniLeg
                trigger ../inidata eq complete and ../wavini eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit FCNPES '32'
                edit THREADS '4'
                inlimit /emc_36r1/limits:pesh 6
                meter step -1 240 240
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../fc eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:dp
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        family legBcpl
          trigger ./legA eq complete
          family fc
            edit EPSCOUPLE '1'
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '144'
            edit EPSINISTEP_PREVIOUS '216'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit TSTEP '2700'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '2'
            family pf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit THREADS '2'
              edit RUNVARFC '1'
              meter step -1 552 552
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/01/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/01 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/02/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/02 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/03/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/03 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/04/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/04 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
            endfamily
            family cf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc0'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger ../../../../legA/fc/cf/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:pesh
              endfamily
              task getvarepsdata
                trigger ../../../legA/fc/cf eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '552'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit SMSTRIES '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:pesh
                meter step -1 552 552
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../../legA/postproc/dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:dp
              family dp
                trigger ./cf eq complete and ./pf eq complete
                edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
                edit OVERLAP '1'
                edit STEPS '216:240:6'
                inlimit /emc_36r1/limits:dp
                family cf
                  edit EMOS_TYPE 'cf'
                  task derived_param
                    edit MEMBER 'off'
                endfamily
                family pf
                  edit EMOS_TYPE 'pf'
                  family 01
                    edit EPSMEMBER '1'
                    edit MEMBER '1'
                    edit SUBFSFAMILY '/pf001'
                    task derived_param
                  endfamily
                  family 02
                    edit EPSMEMBER '2'
                    edit MEMBER '2'
                    edit SUBFSFAMILY '/pf002'
                    task derived_param
                  endfamily
                  family 03
                    edit EPSMEMBER '3'
                    edit MEMBER '3'
                    edit SUBFSFAMILY '/pf003'
                    task derived_param
                  endfamily
                  family 04
                    edit EPSMEMBER '4'
                    edit MEMBER '4'
                    edit SUBFSFAMILY '/pf004'
                    task derived_param
                  endfamily
                endfamily
              endfamily
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                edit EPSFCRES '319'
                task extrafields
                  edit FCLENGTH '768'
                  edit FCLENGTH0 '240'
                task wm_sfc
                  trigger extrafields eq complete
                  edit FCLENGTH '768'
                task wm_ua
                  trigger extrafields eq complete and wm_sfc eq complete
                  edit FCLENGTH '768'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        task save_logs
          trigger m2 eq complete and sv eq complete and legA eq complete and legBcpl eq complete
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
      endfamily
      family 13
        trigger ( ./rmfdb eq complete) and ( ./lock_1701 eq complete)
        edit DELTA_YEAR '-13'
        family m2
          edit DELTA_DAY '-2'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '3000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit FCNPES '16'
          edit EPSSVTSTEP '900'
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit CPUTIME '12000'
          edit STACK '350'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
          task svsh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '73'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '74'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnhevo
            trigger svnh eq complete
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit CPUTIME '120000'
          task svshevo
            trigger svsh eq complete
            edit EVOLVE 'yes'
          task svsave
            trigger svshevo eq complete and svnhevo eq complete
          task clean
            trigger svnhevo eq complete and svshevo eq complete and svsave eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
        endfamily
        family sv
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '2000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit NPES '16'
          edit EPSSVTSTEP '900'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit THREADS '8'
          task targets
            late -s +00:03 -c +00:10
            edit EPSTCBB '1'
            label da_info ""
            label dcda_info ""
            event sv
          task svsh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '73'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '74'
            inlimit /emc_36r1/limits:pesh 2
          task subspace
            trigger svnh eq complete and svsh eq complete and ( targets eq complete or targets:sv)
          task sv1
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '1'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '72'
            inlimit /emc_36r1/limits:pesh 2
          task sv2
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '2'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '71'
            inlimit /emc_36r1/limits:pesh 2
          task sv3
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '3'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv4
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '4'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv5
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '5'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv6
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '6'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task getsvevo
            event ecfs
          task getsvs
            trigger getsvevo eq complete and svnh eq complete and svsh eq complete and sv1 eq complete and sv2 eq complete and sv3 eq complete and sv4 eq complete and sv5 eq complete and sv6 eq complete and ../m2 eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTROP '1'
          task svnhevo
            trigger svnh eq complete and ../legA/fc eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit THREADS '4'
            edit STACK '350'
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
          task svshevo
            trigger svsh eq complete and ../legA/fc eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit STACK '450'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
        endfamily
        family legA
          family fc
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '240'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit TSTEP '1200'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '1'
            task getini
              edit LINIENDA '0'
              edit INIDATEENDA '0'
              edit EPSCOUPLE '0'
              edit N_PERT_AN '10'
              edit EPSCOUPL_A '0'
              edit EPSCOUPL_B '0'
              edit EPSCOUPL_C '0'
            task inidata
              trigger getini eq complete
              edit THREADS '8'
            task getae
              event 1 
            task rot
              trigger ../../sv/getsvs eq complete and ( getae:1 or getae eq complete)
              edit EPSTROP '1'
            family pert_ic
              trigger rot eq complete and inidata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTROP '1'
              inlimit /emc_36r1/limits:pertic
              family 00
                task pertinic
                  edit MEMBER '0'
              endfamily
              family 01
                task pertinic
                  edit MEMBER '1'
              endfamily
              family 02
                task pertinic
                  edit MEMBER '2'
              endfamily
              family 03
                task pertinic
                  edit MEMBER '3'
              endfamily
              family 04
                task pertinic
                  edit MEMBER '4'
              endfamily
            endfamily
            task cp_pert
              trigger pert_ic eq complete
              event 1 
            task wavini
              edit FSFAMILY 'mc'
              edit PERSST 'true'
              edit SSTINT '1'
            task wavfcdata
              trigger ./getpersSST eq complete and ../../legA/fc/wavini eq complete
              edit EPSCOUPLE '1'
            task getpersSST
              trigger ./inidata eq complete
              edit EPSCOUPLE '1'
              edit EPS_INI_STREAM 'DA'
              edit EPSFCRES '639'
            family pf
              trigger ./getpersSST eq complete and ./wavfcdata eq complete and wavini eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              meter step -1 240 240
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task getiniLeg
                  trigger ../../pert_ic/01/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task getiniLeg
                  trigger ../../pert_ic/02/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task getiniLeg
                  trigger ../../pert_ic/03/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task getiniLeg
                  trigger ../../pert_ic/04/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
            endfamily
            family cf
              trigger ./pert_ic/00/pertinic eq complete and ./getpersSST eq complete and ./wavfcdata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              task getiniLeg
                trigger ../inidata eq complete and ../wavini eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit FCNPES '32'
                edit THREADS '4'
                inlimit /emc_36r1/limits:pesh 6
                meter step -1 240 240
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../fc eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:dp
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        family legBcpl
          trigger ./legA eq complete
          family fc
            edit EPSCOUPLE '1'
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '144'
            edit EPSINISTEP_PREVIOUS '216'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit TSTEP '2700'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '2'
            family pf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit THREADS '2'
              edit RUNVARFC '1'
              meter step -1 552 552
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/01/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/01 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/02/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/02 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/03/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/03 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/04/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/04 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
            endfamily
            family cf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc0'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger ../../../../legA/fc/cf/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:pesh
              endfamily
              task getvarepsdata
                trigger ../../../legA/fc/cf eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '552'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit SMSTRIES '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:pesh
                meter step -1 552 552
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../../legA/postproc/dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:dp
              family dp
                trigger ./cf eq complete and ./pf eq complete
                edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
                edit OVERLAP '1'
                edit STEPS '216:240:6'
                inlimit /emc_36r1/limits:dp
                family cf
                  edit EMOS_TYPE 'cf'
                  task derived_param
                    edit MEMBER 'off'
                endfamily
                family pf
                  edit EMOS_TYPE 'pf'
                  family 01
                    edit EPSMEMBER '1'
                    edit MEMBER '1'
                    edit SUBFSFAMILY '/pf001'
                    task derived_param
                  endfamily
                  family 02
                    edit EPSMEMBER '2'
                    edit MEMBER '2'
                    edit SUBFSFAMILY '/pf002'
                    task derived_param
                  endfamily
                  family 03
                    edit EPSMEMBER '3'
                    edit MEMBER '3'
                    edit SUBFSFAMILY '/pf003'
                    task derived_param
                  endfamily
                  family 04
                    edit EPSMEMBER '4'
                    edit MEMBER '4'
                    edit SUBFSFAMILY '/pf004'
                    task derived_param
                  endfamily
                endfamily
              endfamily
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                edit EPSFCRES '319'
                task extrafields
                  edit FCLENGTH '768'
                  edit FCLENGTH0 '240'
                task wm_sfc
                  trigger extrafields eq complete
                  edit FCLENGTH '768'
                task wm_ua
                  trigger extrafields eq complete and wm_sfc eq complete
                  edit FCLENGTH '768'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        task save_logs
          trigger m2 eq complete and sv eq complete and legA eq complete and legBcpl eq complete
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
      endfamily
      family 12
        trigger ( ./rmfdb eq complete) and ( ./lock_1701 eq complete)
        edit DELTA_YEAR '-12'
        family m2
          edit DELTA_DAY '-2'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '3000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit FCNPES '16'
          edit EPSSVTSTEP '900'
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit CPUTIME '12000'
          edit STACK '350'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
          task svsh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '73'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '74'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnhevo
            trigger svnh eq complete
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit CPUTIME '120000'
          task svshevo
            trigger svsh eq complete
            edit EVOLVE 'yes'
          task svsave
            trigger svshevo eq complete and svnhevo eq complete
          task clean
            trigger svnhevo eq complete and svshevo eq complete and svsave eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
        endfamily
        family sv
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '2000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit NPES '16'
          edit EPSSVTSTEP '900'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit THREADS '8'
          task targets
            late -s +00:03 -c +00:10
            edit EPSTCBB '1'
            label da_info ""
            label dcda_info ""
            event sv
          task svsh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '73'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '74'
            inlimit /emc_36r1/limits:pesh 2
          task subspace
            trigger svnh eq complete and svsh eq complete and ( targets eq complete or targets:sv)
          task sv1
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '1'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '72'
            inlimit /emc_36r1/limits:pesh 2
          task sv2
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '2'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '71'
            inlimit /emc_36r1/limits:pesh 2
          task sv3
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '3'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv4
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '4'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv5
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '5'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv6
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '6'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task getsvevo
            event ecfs
          task getsvs
            trigger getsvevo eq complete and svnh eq complete and svsh eq complete and sv1 eq complete and sv2 eq complete and sv3 eq complete and sv4 eq complete and sv5 eq complete and sv6 eq complete and ../m2 eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTROP '1'
          task svnhevo
            trigger svnh eq complete and ../legA/fc eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit THREADS '4'
            edit STACK '350'
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
          task svshevo
            trigger svsh eq complete and ../legA/fc eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit STACK '450'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
        endfamily
        family legA
          family fc
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '240'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit TSTEP '1200'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '1'
            task getini
              edit LINIENDA '0'
              edit INIDATEENDA '0'
              edit EPSCOUPLE '0'
              edit N_PERT_AN '10'
              edit EPSCOUPL_A '0'
              edit EPSCOUPL_B '0'
              edit EPSCOUPL_C '0'
            task inidata
              trigger getini eq complete
              edit THREADS '8'
            task getae
              event 1 
            task rot
              trigger ../../sv/getsvs eq complete and ( getae:1 or getae eq complete)
              edit EPSTROP '1'
            family pert_ic
              trigger rot eq complete and inidata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTROP '1'
              inlimit /emc_36r1/limits:pertic
              family 00
                task pertinic
                  edit MEMBER '0'
              endfamily
              family 01
                task pertinic
                  edit MEMBER '1'
              endfamily
              family 02
                task pertinic
                  edit MEMBER '2'
              endfamily
              family 03
                task pertinic
                  edit MEMBER '3'
              endfamily
              family 04
                task pertinic
                  edit MEMBER '4'
              endfamily
            endfamily
            task cp_pert
              trigger pert_ic eq complete
              event 1 
            task wavini
              edit FSFAMILY 'mc'
              edit PERSST 'true'
              edit SSTINT '1'
            task wavfcdata
              trigger ./getpersSST eq complete and ../../legA/fc/wavini eq complete
              edit EPSCOUPLE '1'
            task getpersSST
              trigger ./inidata eq complete
              edit EPSCOUPLE '1'
              edit EPS_INI_STREAM 'DA'
              edit EPSFCRES '639'
            family pf
              trigger ./getpersSST eq complete and ./wavfcdata eq complete and wavini eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              meter step -1 240 240
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task getiniLeg
                  trigger ../../pert_ic/01/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task getiniLeg
                  trigger ../../pert_ic/02/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task getiniLeg
                  trigger ../../pert_ic/03/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task getiniLeg
                  trigger ../../pert_ic/04/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
            endfamily
            family cf
              trigger ./pert_ic/00/pertinic eq complete and ./getpersSST eq complete and ./wavfcdata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              task getiniLeg
                trigger ../inidata eq complete and ../wavini eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit FCNPES '32'
                edit THREADS '4'
                inlimit /emc_36r1/limits:pesh 6
                meter step -1 240 240
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../fc eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:dp
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        family legBcpl
          trigger ./legA eq complete
          family fc
            edit EPSCOUPLE '1'
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '144'
            edit EPSINISTEP_PREVIOUS '216'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit TSTEP '2700'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '2'
            family pf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit THREADS '2'
              edit RUNVARFC '1'
              meter step -1 552 552
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/01/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/01 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/02/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/02 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/03/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/03 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/04/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/04 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
            endfamily
            family cf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc0'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger ../../../../legA/fc/cf/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:pesh
              endfamily
              task getvarepsdata
                trigger ../../../legA/fc/cf eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '552'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit SMSTRIES '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:pesh
                meter step -1 552 552
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../../legA/postproc/dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:dp
              family dp
                trigger ./cf eq complete and ./pf eq complete
                edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
                edit OVERLAP '1'
                edit STEPS '216:240:6'
                inlimit /emc_36r1/limits:dp
                family cf
                  edit EMOS_TYPE 'cf'
                  task derived_param
                    edit MEMBER 'off'
                endfamily
                family pf
                  edit EMOS_TYPE 'pf'
                  family 01
                    edit EPSMEMBER '1'
                    edit MEMBER '1'
                    edit SUBFSFAMILY '/pf001'
                    task derived_param
                  endfamily
                  family 02
                    edit EPSMEMBER '2'
                    edit MEMBER '2'
                    edit SUBFSFAMILY '/pf002'
                    task derived_param
                  endfamily
                  family 03
                    edit EPSMEMBER '3'
                    edit MEMBER '3'
                    edit SUBFSFAMILY '/pf003'
                    task derived_param
                  endfamily
                  family 04
                    edit EPSMEMBER '4'
                    edit MEMBER '4'
                    edit SUBFSFAMILY '/pf004'
                    task derived_param
                  endfamily
                endfamily
              endfamily
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                edit EPSFCRES '319'
                task extrafields
                  edit FCLENGTH '768'
                  edit FCLENGTH0 '240'
                task wm_sfc
                  trigger extrafields eq complete
                  edit FCLENGTH '768'
                task wm_ua
                  trigger extrafields eq complete and wm_sfc eq complete
                  edit FCLENGTH '768'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        task save_logs
          trigger m2 eq complete and sv eq complete and legA eq complete and legBcpl eq complete
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
      endfamily
      family 11
        trigger ( ./rmfdb eq complete) and ( ./lock_1701 eq complete)
        edit DELTA_YEAR '-11'
        family m2
          edit DELTA_DAY '-2'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '3000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit FCNPES '16'
          edit EPSSVTSTEP '900'
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit CPUTIME '12000'
          edit STACK '350'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
          task svsh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '73'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '74'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnhevo
            trigger svnh eq complete
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit CPUTIME '120000'
          task svshevo
            trigger svsh eq complete
            edit EVOLVE 'yes'
          task svsave
            trigger svshevo eq complete and svnhevo eq complete
          task clean
            trigger svnhevo eq complete and svshevo eq complete and svsave eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
        endfamily
        family sv
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '2000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit NPES '16'
          edit EPSSVTSTEP '900'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit THREADS '8'
          task targets
            late -s +00:03 -c +00:10
            edit EPSTCBB '1'
            label da_info ""
            label dcda_info ""
            event sv
          task svsh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '73'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '74'
            inlimit /emc_36r1/limits:pesh 2
          task subspace
            trigger svnh eq complete and svsh eq complete and ( targets eq complete or targets:sv)
          task sv1
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '1'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '72'
            inlimit /emc_36r1/limits:pesh 2
          task sv2
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '2'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '71'
            inlimit /emc_36r1/limits:pesh 2
          task sv3
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '3'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv4
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '4'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv5
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '5'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv6
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '6'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task getsvevo
            event ecfs
          task getsvs
            trigger getsvevo eq complete and svnh eq complete and svsh eq complete and sv1 eq complete and sv2 eq complete and sv3 eq complete and sv4 eq complete and sv5 eq complete and sv6 eq complete and ../m2 eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTROP '1'
          task svnhevo
            trigger svnh eq complete and ../legA/fc eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit THREADS '4'
            edit STACK '350'
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
          task svshevo
            trigger svsh eq complete and ../legA/fc eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit STACK '450'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
        endfamily
        family legA
          family fc
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '240'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit TSTEP '1200'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '1'
            task getini
              edit LINIENDA '0'
              edit INIDATEENDA '0'
              edit EPSCOUPLE '0'
              edit N_PERT_AN '10'
              edit EPSCOUPL_A '0'
              edit EPSCOUPL_B '0'
              edit EPSCOUPL_C '0'
            task inidata
              trigger getini eq complete
              edit THREADS '8'
            task getae
              event 1 
            task rot
              trigger ../../sv/getsvs eq complete and ( getae:1 or getae eq complete)
              edit EPSTROP '1'
            family pert_ic
              trigger rot eq complete and inidata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTROP '1'
              inlimit /emc_36r1/limits:pertic
              family 00
                task pertinic
                  edit MEMBER '0'
              endfamily
              family 01
                task pertinic
                  edit MEMBER '1'
              endfamily
              family 02
                task pertinic
                  edit MEMBER '2'
              endfamily
              family 03
                task pertinic
                  edit MEMBER '3'
              endfamily
              family 04
                task pertinic
                  edit MEMBER '4'
              endfamily
            endfamily
            task cp_pert
              trigger pert_ic eq complete
              event 1 
            task wavini
              edit FSFAMILY 'mc'
              edit PERSST 'true'
              edit SSTINT '1'
            task wavfcdata
              trigger ./getpersSST eq complete and ../../legA/fc/wavini eq complete
              edit EPSCOUPLE '1'
            task getpersSST
              trigger ./inidata eq complete
              edit EPSCOUPLE '1'
              edit EPS_INI_STREAM 'DA'
              edit EPSFCRES '639'
            family pf
              trigger ./getpersSST eq complete and ./wavfcdata eq complete and wavini eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              meter step -1 240 240
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task getiniLeg
                  trigger ../../pert_ic/01/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task getiniLeg
                  trigger ../../pert_ic/02/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task getiniLeg
                  trigger ../../pert_ic/03/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task getiniLeg
                  trigger ../../pert_ic/04/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
            endfamily
            family cf
              trigger ./pert_ic/00/pertinic eq complete and ./getpersSST eq complete and ./wavfcdata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              task getiniLeg
                trigger ../inidata eq complete and ../wavini eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit FCNPES '32'
                edit THREADS '4'
                inlimit /emc_36r1/limits:pesh 6
                meter step -1 240 240
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../fc eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:dp
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        family legBcpl
          trigger ./legA eq complete
          family fc
            edit EPSCOUPLE '1'
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '144'
            edit EPSINISTEP_PREVIOUS '216'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit TSTEP '2700'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '2'
            family pf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit THREADS '2'
              edit RUNVARFC '1'
              meter step -1 552 552
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/01/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/01 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/02/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/02 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/03/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/03 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/04/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/04 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
            endfamily
            family cf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc0'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger ../../../../legA/fc/cf/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:pesh
              endfamily
              task getvarepsdata
                trigger ../../../legA/fc/cf eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '552'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit SMSTRIES '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:pesh
                meter step -1 552 552
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../../legA/postproc/dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:dp
              family dp
                trigger ./cf eq complete and ./pf eq complete
                edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
                edit OVERLAP '1'
                edit STEPS '216:240:6'
                inlimit /emc_36r1/limits:dp
                family cf
                  edit EMOS_TYPE 'cf'
                  task derived_param
                    edit MEMBER 'off'
                endfamily
                family pf
                  edit EMOS_TYPE 'pf'
                  family 01
                    edit EPSMEMBER '1'
                    edit MEMBER '1'
                    edit SUBFSFAMILY '/pf001'
                    task derived_param
                  endfamily
                  family 02
                    edit EPSMEMBER '2'
                    edit MEMBER '2'
                    edit SUBFSFAMILY '/pf002'
                    task derived_param
                  endfamily
                  family 03
                    edit EPSMEMBER '3'
                    edit MEMBER '3'
                    edit SUBFSFAMILY '/pf003'
                    task derived_param
                  endfamily
                  family 04
                    edit EPSMEMBER '4'
                    edit MEMBER '4'
                    edit SUBFSFAMILY '/pf004'
                    task derived_param
                  endfamily
                endfamily
              endfamily
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                edit EPSFCRES '319'
                task extrafields
                  edit FCLENGTH '768'
                  edit FCLENGTH0 '240'
                task wm_sfc
                  trigger extrafields eq complete
                  edit FCLENGTH '768'
                task wm_ua
                  trigger extrafields eq complete and wm_sfc eq complete
                  edit FCLENGTH '768'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        task save_logs
          trigger m2 eq complete and sv eq complete and legA eq complete and legBcpl eq complete
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
      endfamily
      family 10
        trigger ( ./rmfdb eq complete) and ( ./lock_1701 eq complete)
        edit DELTA_YEAR '-10'
        family m2
          edit DELTA_DAY '-2'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '3000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit FCNPES '16'
          edit EPSSVTSTEP '900'
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit CPUTIME '12000'
          edit STACK '350'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
          task svsh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '73'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '74'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnhevo
            trigger svnh eq complete
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit CPUTIME '120000'
          task svshevo
            trigger svsh eq complete
            edit EVOLVE 'yes'
          task svsave
            trigger svshevo eq complete and svnhevo eq complete
          task clean
            trigger svnhevo eq complete and svshevo eq complete and svsave eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
        endfamily
        family sv
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '2000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit NPES '16'
          edit EPSSVTSTEP '900'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit THREADS '8'
          task targets
            late -s +00:03 -c +00:10
            edit EPSTCBB '1'
            label da_info ""
            label dcda_info ""
            event sv
          task svsh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '73'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '74'
            inlimit /emc_36r1/limits:pesh 2
          task subspace
            trigger svnh eq complete and svsh eq complete and ( targets eq complete or targets:sv)
          task sv1
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '1'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '72'
            inlimit /emc_36r1/limits:pesh 2
          task sv2
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '2'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '71'
            inlimit /emc_36r1/limits:pesh 2
          task sv3
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '3'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv4
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '4'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv5
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '5'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv6
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '6'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task getsvevo
            event ecfs
          task getsvs
            trigger getsvevo eq complete and svnh eq complete and svsh eq complete and sv1 eq complete and sv2 eq complete and sv3 eq complete and sv4 eq complete and sv5 eq complete and sv6 eq complete and ../m2 eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTROP '1'
          task svnhevo
            trigger svnh eq complete and ../legA/fc eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit THREADS '4'
            edit STACK '350'
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
          task svshevo
            trigger svsh eq complete and ../legA/fc eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit STACK '450'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
        endfamily
        family legA
          family fc
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '240'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit TSTEP '1200'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '1'
            task getini
              edit LINIENDA '0'
              edit INIDATEENDA '0'
              edit EPSCOUPLE '0'
              edit N_PERT_AN '10'
              edit EPSCOUPL_A '0'
              edit EPSCOUPL_B '0'
              edit EPSCOUPL_C '0'
            task inidata
              trigger getini eq complete
              edit THREADS '8'
            task getae
              event 1 
            task rot
              trigger ../../sv/getsvs eq complete and ( getae:1 or getae eq complete)
              edit EPSTROP '1'
            family pert_ic
              trigger rot eq complete and inidata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTROP '1'
              inlimit /emc_36r1/limits:pertic
              family 00
                task pertinic
                  edit MEMBER '0'
              endfamily
              family 01
                task pertinic
                  edit MEMBER '1'
              endfamily
              family 02
                task pertinic
                  edit MEMBER '2'
              endfamily
              family 03
                task pertinic
                  edit MEMBER '3'
              endfamily
              family 04
                task pertinic
                  edit MEMBER '4'
              endfamily
            endfamily
            task cp_pert
              trigger pert_ic eq complete
              event 1 
            task wavini
              edit FSFAMILY 'mc'
              edit PERSST 'true'
              edit SSTINT '1'
            task wavfcdata
              trigger ./getpersSST eq complete and ../../legA/fc/wavini eq complete
              edit EPSCOUPLE '1'
            task getpersSST
              trigger ./inidata eq complete
              edit EPSCOUPLE '1'
              edit EPS_INI_STREAM 'DA'
              edit EPSFCRES '639'
            family pf
              trigger ./getpersSST eq complete and ./wavfcdata eq complete and wavini eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              meter step -1 240 240
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task getiniLeg
                  trigger ../../pert_ic/01/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task getiniLeg
                  trigger ../../pert_ic/02/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task getiniLeg
                  trigger ../../pert_ic/03/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task getiniLeg
                  trigger ../../pert_ic/04/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
            endfamily
            family cf
              trigger ./pert_ic/00/pertinic eq complete and ./getpersSST eq complete and ./wavfcdata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              task getiniLeg
                trigger ../inidata eq complete and ../wavini eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit FCNPES '32'
                edit THREADS '4'
                inlimit /emc_36r1/limits:pesh 6
                meter step -1 240 240
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../fc eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:dp
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        family legBcpl
          trigger ./legA eq complete
          family fc
            edit EPSCOUPLE '1'
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '144'
            edit EPSINISTEP_PREVIOUS '216'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit TSTEP '2700'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '2'
            family pf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit THREADS '2'
              edit RUNVARFC '1'
              meter step -1 552 552
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/01/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/01 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/02/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/02 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/03/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/03 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/04/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/04 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
            endfamily
            family cf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc0'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger ../../../../legA/fc/cf/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:pesh
              endfamily
              task getvarepsdata
                trigger ../../../legA/fc/cf eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '552'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit SMSTRIES '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:pesh
                meter step -1 552 552
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../../legA/postproc/dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:dp
              family dp
                trigger ./cf eq complete and ./pf eq complete
                edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
                edit OVERLAP '1'
                edit STEPS '216:240:6'
                inlimit /emc_36r1/limits:dp
                family cf
                  edit EMOS_TYPE 'cf'
                  task derived_param
                    edit MEMBER 'off'
                endfamily
                family pf
                  edit EMOS_TYPE 'pf'
                  family 01
                    edit EPSMEMBER '1'
                    edit MEMBER '1'
                    edit SUBFSFAMILY '/pf001'
                    task derived_param
                  endfamily
                  family 02
                    edit EPSMEMBER '2'
                    edit MEMBER '2'
                    edit SUBFSFAMILY '/pf002'
                    task derived_param
                  endfamily
                  family 03
                    edit EPSMEMBER '3'
                    edit MEMBER '3'
                    edit SUBFSFAMILY '/pf003'
                    task derived_param
                  endfamily
                  family 04
                    edit EPSMEMBER '4'
                    edit MEMBER '4'
                    edit SUBFSFAMILY '/pf004'
                    task derived_param
                  endfamily
                endfamily
              endfamily
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                edit EPSFCRES '319'
                task extrafields
                  edit FCLENGTH '768'
                  edit FCLENGTH0 '240'
                task wm_sfc
                  trigger extrafields eq complete
                  edit FCLENGTH '768'
                task wm_ua
                  trigger extrafields eq complete and wm_sfc eq complete
                  edit FCLENGTH '768'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        task save_logs
          trigger m2 eq complete and sv eq complete and legA eq complete and legBcpl eq complete
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
      endfamily
      family 09
        trigger ( ./rmfdb eq complete) and ( ./lock_1701 eq complete)
        edit DELTA_YEAR '-09'
        family m2
          edit DELTA_DAY '-2'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '3000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit FCNPES '16'
          edit EPSSVTSTEP '900'
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit CPUTIME '12000'
          edit STACK '350'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
          task svsh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '73'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '74'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnhevo
            trigger svnh eq complete
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit CPUTIME '120000'
          task svshevo
            trigger svsh eq complete
            edit EVOLVE 'yes'
          task svsave
            trigger svshevo eq complete and svnhevo eq complete
          task clean
            trigger svnhevo eq complete and svshevo eq complete and svsave eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
        endfamily
        family sv
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '2000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit NPES '16'
          edit EPSSVTSTEP '900'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit THREADS '8'
          task targets
            late -s +00:03 -c +00:10
            edit EPSTCBB '1'
            label da_info ""
            label dcda_info ""
            event sv
          task svsh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '73'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '74'
            inlimit /emc_36r1/limits:pesh 2
          task subspace
            trigger svnh eq complete and svsh eq complete and ( targets eq complete or targets:sv)
          task sv1
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '1'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '72'
            inlimit /emc_36r1/limits:pesh 2
          task sv2
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '2'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '71'
            inlimit /emc_36r1/limits:pesh 2
          task sv3
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '3'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv4
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '4'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv5
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '5'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv6
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '6'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task getsvevo
            event ecfs
          task getsvs
            trigger getsvevo eq complete and svnh eq complete and svsh eq complete and sv1 eq complete and sv2 eq complete and sv3 eq complete and sv4 eq complete and sv5 eq complete and sv6 eq complete and ../m2 eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTROP '1'
          task svnhevo
            trigger svnh eq complete and ../legA/fc eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit THREADS '4'
            edit STACK '350'
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
          task svshevo
            trigger svsh eq complete and ../legA/fc eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit STACK '450'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
        endfamily
        family legA
          family fc
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '240'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit TSTEP '1200'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '1'
            task getini
              edit LINIENDA '0'
              edit INIDATEENDA '0'
              edit EPSCOUPLE '0'
              edit N_PERT_AN '10'
              edit EPSCOUPL_A '0'
              edit EPSCOUPL_B '0'
              edit EPSCOUPL_C '0'
            task inidata
              trigger getini eq complete
              edit THREADS '8'
            task getae
              event 1 
            task rot
              trigger ../../sv/getsvs eq complete and ( getae:1 or getae eq complete)
              edit EPSTROP '1'
            family pert_ic
              trigger rot eq complete and inidata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTROP '1'
              inlimit /emc_36r1/limits:pertic
              family 00
                task pertinic
                  edit MEMBER '0'
              endfamily
              family 01
                task pertinic
                  edit MEMBER '1'
              endfamily
              family 02
                task pertinic
                  edit MEMBER '2'
              endfamily
              family 03
                task pertinic
                  edit MEMBER '3'
              endfamily
              family 04
                task pertinic
                  edit MEMBER '4'
              endfamily
            endfamily
            task cp_pert
              trigger pert_ic eq complete
              event 1 
            task wavini
              edit FSFAMILY 'mc'
              edit PERSST 'true'
              edit SSTINT '1'
            task wavfcdata
              trigger ./getpersSST eq complete and ../../legA/fc/wavini eq complete
              edit EPSCOUPLE '1'
            task getpersSST
              trigger ./inidata eq complete
              edit EPSCOUPLE '1'
              edit EPS_INI_STREAM 'DA'
              edit EPSFCRES '639'
            family pf
              trigger ./getpersSST eq complete and ./wavfcdata eq complete and wavini eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              meter step -1 240 240
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task getiniLeg
                  trigger ../../pert_ic/01/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task getiniLeg
                  trigger ../../pert_ic/02/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task getiniLeg
                  trigger ../../pert_ic/03/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task getiniLeg
                  trigger ../../pert_ic/04/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
            endfamily
            family cf
              trigger ./pert_ic/00/pertinic eq complete and ./getpersSST eq complete and ./wavfcdata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              task getiniLeg
                trigger ../inidata eq complete and ../wavini eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit FCNPES '32'
                edit THREADS '4'
                inlimit /emc_36r1/limits:pesh 6
                meter step -1 240 240
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../fc eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:dp
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        family legBcpl
          trigger ./legA eq complete
          family fc
            edit EPSCOUPLE '1'
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '144'
            edit EPSINISTEP_PREVIOUS '216'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit TSTEP '2700'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '2'
            family pf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit THREADS '2'
              edit RUNVARFC '1'
              meter step -1 552 552
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/01/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/01 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/02/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/02 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/03/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/03 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/04/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/04 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
            endfamily
            family cf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc0'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger ../../../../legA/fc/cf/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:pesh
              endfamily
              task getvarepsdata
                trigger ../../../legA/fc/cf eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '552'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit SMSTRIES '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:pesh
                meter step -1 552 552
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../../legA/postproc/dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:dp
              family dp
                trigger ./cf eq complete and ./pf eq complete
                edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
                edit OVERLAP '1'
                edit STEPS '216:240:6'
                inlimit /emc_36r1/limits:dp
                family cf
                  edit EMOS_TYPE 'cf'
                  task derived_param
                    edit MEMBER 'off'
                endfamily
                family pf
                  edit EMOS_TYPE 'pf'
                  family 01
                    edit EPSMEMBER '1'
                    edit MEMBER '1'
                    edit SUBFSFAMILY '/pf001'
                    task derived_param
                  endfamily
                  family 02
                    edit EPSMEMBER '2'
                    edit MEMBER '2'
                    edit SUBFSFAMILY '/pf002'
                    task derived_param
                  endfamily
                  family 03
                    edit EPSMEMBER '3'
                    edit MEMBER '3'
                    edit SUBFSFAMILY '/pf003'
                    task derived_param
                  endfamily
                  family 04
                    edit EPSMEMBER '4'
                    edit MEMBER '4'
                    edit SUBFSFAMILY '/pf004'
                    task derived_param
                  endfamily
                endfamily
              endfamily
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                edit EPSFCRES '319'
                task extrafields
                  edit FCLENGTH '768'
                  edit FCLENGTH0 '240'
                task wm_sfc
                  trigger extrafields eq complete
                  edit FCLENGTH '768'
                task wm_ua
                  trigger extrafields eq complete and wm_sfc eq complete
                  edit FCLENGTH '768'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        task save_logs
          trigger m2 eq complete and sv eq complete and legA eq complete and legBcpl eq complete
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
      endfamily
      family 08
        trigger ( ./rmfdb eq complete) and ( ./lock_1701 eq complete)
        edit DELTA_YEAR '-08'
        family m2
          edit DELTA_DAY '-2'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '3000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit FCNPES '16'
          edit EPSSVTSTEP '900'
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit CPUTIME '12000'
          edit STACK '350'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
          task svsh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '73'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '74'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnhevo
            trigger svnh eq complete
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit CPUTIME '120000'
          task svshevo
            trigger svsh eq complete
            edit EVOLVE 'yes'
          task svsave
            trigger svshevo eq complete and svnhevo eq complete
          task clean
            trigger svnhevo eq complete and svshevo eq complete and svsave eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
        endfamily
        family sv
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '2000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit NPES '16'
          edit EPSSVTSTEP '900'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit THREADS '8'
          task targets
            late -s +00:03 -c +00:10
            edit EPSTCBB '1'
            label da_info ""
            label dcda_info ""
            event sv
          task svsh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '73'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '74'
            inlimit /emc_36r1/limits:pesh 2
          task subspace
            trigger svnh eq complete and svsh eq complete and ( targets eq complete or targets:sv)
          task sv1
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '1'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '72'
            inlimit /emc_36r1/limits:pesh 2
          task sv2
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '2'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '71'
            inlimit /emc_36r1/limits:pesh 2
          task sv3
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '3'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv4
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '4'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv5
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '5'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv6
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '6'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task getsvevo
            event ecfs
          task getsvs
            trigger getsvevo eq complete and svnh eq complete and svsh eq complete and sv1 eq complete and sv2 eq complete and sv3 eq complete and sv4 eq complete and sv5 eq complete and sv6 eq complete and ../m2 eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTROP '1'
          task svnhevo
            trigger svnh eq complete and ../legA/fc eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit THREADS '4'
            edit STACK '350'
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
          task svshevo
            trigger svsh eq complete and ../legA/fc eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit STACK '450'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
        endfamily
        family legA
          family fc
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '240'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit TSTEP '1200'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '1'
            task getini
              edit LINIENDA '0'
              edit INIDATEENDA '0'
              edit EPSCOUPLE '0'
              edit N_PERT_AN '10'
              edit EPSCOUPL_A '0'
              edit EPSCOUPL_B '0'
              edit EPSCOUPL_C '0'
            task inidata
              trigger getini eq complete
              edit THREADS '8'
            task getae
              event 1 
            task rot
              trigger ../../sv/getsvs eq complete and ( getae:1 or getae eq complete)
              edit EPSTROP '1'
            family pert_ic
              trigger rot eq complete and inidata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTROP '1'
              inlimit /emc_36r1/limits:pertic
              family 00
                task pertinic
                  edit MEMBER '0'
              endfamily
              family 01
                task pertinic
                  edit MEMBER '1'
              endfamily
              family 02
                task pertinic
                  edit MEMBER '2'
              endfamily
              family 03
                task pertinic
                  edit MEMBER '3'
              endfamily
              family 04
                task pertinic
                  edit MEMBER '4'
              endfamily
            endfamily
            task cp_pert
              trigger pert_ic eq complete
              event 1 
            task wavini
              edit FSFAMILY 'mc'
              edit PERSST 'true'
              edit SSTINT '1'
            task wavfcdata
              trigger ./getpersSST eq complete and ../../legA/fc/wavini eq complete
              edit EPSCOUPLE '1'
            task getpersSST
              trigger ./inidata eq complete
              edit EPSCOUPLE '1'
              edit EPS_INI_STREAM 'DA'
              edit EPSFCRES '639'
            family pf
              trigger ./getpersSST eq complete and ./wavfcdata eq complete and wavini eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              meter step -1 240 240
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task getiniLeg
                  trigger ../../pert_ic/01/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task getiniLeg
                  trigger ../../pert_ic/02/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task getiniLeg
                  trigger ../../pert_ic/03/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task getiniLeg
                  trigger ../../pert_ic/04/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
            endfamily
            family cf
              trigger ./pert_ic/00/pertinic eq complete and ./getpersSST eq complete and ./wavfcdata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              task getiniLeg
                trigger ../inidata eq complete and ../wavini eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit FCNPES '32'
                edit THREADS '4'
                inlimit /emc_36r1/limits:pesh 6
                meter step -1 240 240
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../fc eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:dp
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        family legBcpl
          trigger ./legA eq complete
          family fc
            edit EPSCOUPLE '1'
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '144'
            edit EPSINISTEP_PREVIOUS '216'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit TSTEP '2700'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '2'
            family pf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit THREADS '2'
              edit RUNVARFC '1'
              meter step -1 552 552
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/01/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/01 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/02/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/02 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/03/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/03 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/04/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/04 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
            endfamily
            family cf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc0'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger ../../../../legA/fc/cf/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:pesh
              endfamily
              task getvarepsdata
                trigger ../../../legA/fc/cf eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '552'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit SMSTRIES '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:pesh
                meter step -1 552 552
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../../legA/postproc/dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:dp
              family dp
                trigger ./cf eq complete and ./pf eq complete
                edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
                edit OVERLAP '1'
                edit STEPS '216:240:6'
                inlimit /emc_36r1/limits:dp
                family cf
                  edit EMOS_TYPE 'cf'
                  task derived_param
                    edit MEMBER 'off'
                endfamily
                family pf
                  edit EMOS_TYPE 'pf'
                  family 01
                    edit EPSMEMBER '1'
                    edit MEMBER '1'
                    edit SUBFSFAMILY '/pf001'
                    task derived_param
                  endfamily
                  family 02
                    edit EPSMEMBER '2'
                    edit MEMBER '2'
                    edit SUBFSFAMILY '/pf002'
                    task derived_param
                  endfamily
                  family 03
                    edit EPSMEMBER '3'
                    edit MEMBER '3'
                    edit SUBFSFAMILY '/pf003'
                    task derived_param
                  endfamily
                  family 04
                    edit EPSMEMBER '4'
                    edit MEMBER '4'
                    edit SUBFSFAMILY '/pf004'
                    task derived_param
                  endfamily
                endfamily
              endfamily
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                edit EPSFCRES '319'
                task extrafields
                  edit FCLENGTH '768'
                  edit FCLENGTH0 '240'
                task wm_sfc
                  trigger extrafields eq complete
                  edit FCLENGTH '768'
                task wm_ua
                  trigger extrafields eq complete and wm_sfc eq complete
                  edit FCLENGTH '768'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        task save_logs
          trigger m2 eq complete and sv eq complete and legA eq complete and legBcpl eq complete
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
      endfamily
      family 07
        trigger ( ./rmfdb eq complete) and ( ./lock_1701 eq complete)
        edit DELTA_YEAR '-07'
        family m2
          edit DELTA_DAY '-2'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '3000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit FCNPES '16'
          edit EPSSVTSTEP '900'
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit CPUTIME '12000'
          edit STACK '350'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
          task svsh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '73'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '74'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnhevo
            trigger svnh eq complete
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit CPUTIME '120000'
          task svshevo
            trigger svsh eq complete
            edit EVOLVE 'yes'
          task svsave
            trigger svshevo eq complete and svnhevo eq complete
          task clean
            trigger svnhevo eq complete and svshevo eq complete and svsave eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
        endfamily
        family sv
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '2000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit NPES '16'
          edit EPSSVTSTEP '900'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit THREADS '8'
          task targets
            late -s +00:03 -c +00:10
            edit EPSTCBB '1'
            label da_info ""
            label dcda_info ""
            event sv
          task svsh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '73'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '74'
            inlimit /emc_36r1/limits:pesh 2
          task subspace
            trigger svnh eq complete and svsh eq complete and ( targets eq complete or targets:sv)
          task sv1
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '1'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '72'
            inlimit /emc_36r1/limits:pesh 2
          task sv2
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '2'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '71'
            inlimit /emc_36r1/limits:pesh 2
          task sv3
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '3'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv4
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '4'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv5
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '5'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv6
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '6'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task getsvevo
            event ecfs
          task getsvs
            trigger getsvevo eq complete and svnh eq complete and svsh eq complete and sv1 eq complete and sv2 eq complete and sv3 eq complete and sv4 eq complete and sv5 eq complete and sv6 eq complete and ../m2 eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTROP '1'
          task svnhevo
            trigger svnh eq complete and ../legA/fc eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit THREADS '4'
            edit STACK '350'
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
          task svshevo
            trigger svsh eq complete and ../legA/fc eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit STACK '450'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
        endfamily
        family legA
          family fc
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '240'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit TSTEP '1200'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '1'
            task getini
              edit LINIENDA '0'
              edit INIDATEENDA '0'
              edit EPSCOUPLE '0'
              edit N_PERT_AN '10'
              edit EPSCOUPL_A '0'
              edit EPSCOUPL_B '0'
              edit EPSCOUPL_C '0'
            task inidata
              trigger getini eq complete
              edit THREADS '8'
            task getae
              event 1 
            task rot
              trigger ../../sv/getsvs eq complete and ( getae:1 or getae eq complete)
              edit EPSTROP '1'
            family pert_ic
              trigger rot eq complete and inidata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTROP '1'
              inlimit /emc_36r1/limits:pertic
              family 00
                task pertinic
                  edit MEMBER '0'
              endfamily
              family 01
                task pertinic
                  edit MEMBER '1'
              endfamily
              family 02
                task pertinic
                  edit MEMBER '2'
              endfamily
              family 03
                task pertinic
                  edit MEMBER '3'
              endfamily
              family 04
                task pertinic
                  edit MEMBER '4'
              endfamily
            endfamily
            task cp_pert
              trigger pert_ic eq complete
              event 1 
            task wavini
              edit FSFAMILY 'mc'
              edit PERSST 'true'
              edit SSTINT '1'
            task wavfcdata
              trigger ./getpersSST eq complete and ../../legA/fc/wavini eq complete
              edit EPSCOUPLE '1'
            task getpersSST
              trigger ./inidata eq complete
              edit EPSCOUPLE '1'
              edit EPS_INI_STREAM 'DA'
              edit EPSFCRES '639'
            family pf
              trigger ./getpersSST eq complete and ./wavfcdata eq complete and wavini eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              meter step -1 240 240
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task getiniLeg
                  trigger ../../pert_ic/01/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task getiniLeg
                  trigger ../../pert_ic/02/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task getiniLeg
                  trigger ../../pert_ic/03/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task getiniLeg
                  trigger ../../pert_ic/04/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
            endfamily
            family cf
              trigger ./pert_ic/00/pertinic eq complete and ./getpersSST eq complete and ./wavfcdata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              task getiniLeg
                trigger ../inidata eq complete and ../wavini eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit FCNPES '32'
                edit THREADS '4'
                inlimit /emc_36r1/limits:pesh 6
                meter step -1 240 240
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../fc eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:dp
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        family legBcpl
          trigger ./legA eq complete
          family fc
            edit EPSCOUPLE '1'
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '144'
            edit EPSINISTEP_PREVIOUS '216'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit TSTEP '2700'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '2'
            family pf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit THREADS '2'
              edit RUNVARFC '1'
              meter step -1 552 552
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/01/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/01 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/02/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/02 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/03/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/03 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/04/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/04 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
            endfamily
            family cf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc0'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger ../../../../legA/fc/cf/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:pesh
              endfamily
              task getvarepsdata
                trigger ../../../legA/fc/cf eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '552'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit SMSTRIES '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:pesh
                meter step -1 552 552
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../../legA/postproc/dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:dp
              family dp
                trigger ./cf eq complete and ./pf eq complete
                edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
                edit OVERLAP '1'
                edit STEPS '216:240:6'
                inlimit /emc_36r1/limits:dp
                family cf
                  edit EMOS_TYPE 'cf'
                  task derived_param
                    edit MEMBER 'off'
                endfamily
                family pf
                  edit EMOS_TYPE 'pf'
                  family 01
                    edit EPSMEMBER '1'
                    edit MEMBER '1'
                    edit SUBFSFAMILY '/pf001'
                    task derived_param
                  endfamily
                  family 02
                    edit EPSMEMBER '2'
                    edit MEMBER '2'
                    edit SUBFSFAMILY '/pf002'
                    task derived_param
                  endfamily
                  family 03
                    edit EPSMEMBER '3'
                    edit MEMBER '3'
                    edit SUBFSFAMILY '/pf003'
                    task derived_param
                  endfamily
                  family 04
                    edit EPSMEMBER '4'
                    edit MEMBER '4'
                    edit SUBFSFAMILY '/pf004'
                    task derived_param
                  endfamily
                endfamily
              endfamily
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                edit EPSFCRES '319'
                task extrafields
                  edit FCLENGTH '768'
                  edit FCLENGTH0 '240'
                task wm_sfc
                  trigger extrafields eq complete
                  edit FCLENGTH '768'
                task wm_ua
                  trigger extrafields eq complete and wm_sfc eq complete
                  edit FCLENGTH '768'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        task save_logs
          trigger m2 eq complete and sv eq complete and legA eq complete and legBcpl eq complete
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
      endfamily
      family 06
        trigger ( ./rmfdb eq complete) and ( ./lock_1701 eq complete)
        edit DELTA_YEAR '-06'
        family m2
          edit DELTA_DAY '-2'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '3000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit FCNPES '16'
          edit EPSSVTSTEP '900'
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit CPUTIME '12000'
          edit STACK '350'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
          task svsh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '73'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '74'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnhevo
            trigger svnh eq complete
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit CPUTIME '120000'
          task svshevo
            trigger svsh eq complete
            edit EVOLVE 'yes'
          task svsave
            trigger svshevo eq complete and svnhevo eq complete
          task clean
            trigger svnhevo eq complete and svshevo eq complete and svsave eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
        endfamily
        family sv
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '2000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit NPES '16'
          edit EPSSVTSTEP '900'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit THREADS '8'
          task targets
            late -s +00:03 -c +00:10
            edit EPSTCBB '1'
            label da_info ""
            label dcda_info ""
            event sv
          task svsh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '73'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '74'
            inlimit /emc_36r1/limits:pesh 2
          task subspace
            trigger svnh eq complete and svsh eq complete and ( targets eq complete or targets:sv)
          task sv1
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '1'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '72'
            inlimit /emc_36r1/limits:pesh 2
          task sv2
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '2'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '71'
            inlimit /emc_36r1/limits:pesh 2
          task sv3
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '3'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv4
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '4'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv5
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '5'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv6
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '6'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task getsvevo
            event ecfs
          task getsvs
            trigger getsvevo eq complete and svnh eq complete and svsh eq complete and sv1 eq complete and sv2 eq complete and sv3 eq complete and sv4 eq complete and sv5 eq complete and sv6 eq complete and ../m2 eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTROP '1'
          task svnhevo
            trigger svnh eq complete and ../legA/fc eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit THREADS '4'
            edit STACK '350'
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
          task svshevo
            trigger svsh eq complete and ../legA/fc eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit STACK '450'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
        endfamily
        family legA
          family fc
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '240'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit TSTEP '1200'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '1'
            task getini
              edit LINIENDA '0'
              edit INIDATEENDA '0'
              edit EPSCOUPLE '0'
              edit N_PERT_AN '10'
              edit EPSCOUPL_A '0'
              edit EPSCOUPL_B '0'
              edit EPSCOUPL_C '0'
            task inidata
              trigger getini eq complete
              edit THREADS '8'
            task getae
              event 1 
            task rot
              trigger ../../sv/getsvs eq complete and ( getae:1 or getae eq complete)
              edit EPSTROP '1'
            family pert_ic
              trigger rot eq complete and inidata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTROP '1'
              inlimit /emc_36r1/limits:pertic
              family 00
                task pertinic
                  edit MEMBER '0'
              endfamily
              family 01
                task pertinic
                  edit MEMBER '1'
              endfamily
              family 02
                task pertinic
                  edit MEMBER '2'
              endfamily
              family 03
                task pertinic
                  edit MEMBER '3'
              endfamily
              family 04
                task pertinic
                  edit MEMBER '4'
              endfamily
            endfamily
            task cp_pert
              trigger pert_ic eq complete
              event 1 
            task wavini
              edit FSFAMILY 'mc'
              edit PERSST 'true'
              edit SSTINT '1'
            task wavfcdata
              trigger ./getpersSST eq complete and ../../legA/fc/wavini eq complete
              edit EPSCOUPLE '1'
            task getpersSST
              trigger ./inidata eq complete
              edit EPSCOUPLE '1'
              edit EPS_INI_STREAM 'DA'
              edit EPSFCRES '639'
            family pf
              trigger ./getpersSST eq complete and ./wavfcdata eq complete and wavini eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              meter step -1 240 240
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task getiniLeg
                  trigger ../../pert_ic/01/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task getiniLeg
                  trigger ../../pert_ic/02/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task getiniLeg
                  trigger ../../pert_ic/03/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task getiniLeg
                  trigger ../../pert_ic/04/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
            endfamily
            family cf
              trigger ./pert_ic/00/pertinic eq complete and ./getpersSST eq complete and ./wavfcdata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              task getiniLeg
                trigger ../inidata eq complete and ../wavini eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit FCNPES '32'
                edit THREADS '4'
                inlimit /emc_36r1/limits:pesh 6
                meter step -1 240 240
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../fc eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:dp
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        family legBcpl
          trigger ./legA eq complete
          family fc
            edit EPSCOUPLE '1'
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '144'
            edit EPSINISTEP_PREVIOUS '216'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit TSTEP '2700'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '2'
            family pf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit THREADS '2'
              edit RUNVARFC '1'
              meter step -1 552 552
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/01/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/01 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/02/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/02 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/03/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/03 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/04/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/04 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
            endfamily
            family cf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc0'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger ../../../../legA/fc/cf/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:pesh
              endfamily
              task getvarepsdata
                trigger ../../../legA/fc/cf eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '552'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit SMSTRIES '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:pesh
                meter step -1 552 552
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../../legA/postproc/dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:dp
              family dp
                trigger ./cf eq complete and ./pf eq complete
                edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
                edit OVERLAP '1'
                edit STEPS '216:240:6'
                inlimit /emc_36r1/limits:dp
                family cf
                  edit EMOS_TYPE 'cf'
                  task derived_param
                    edit MEMBER 'off'
                endfamily
                family pf
                  edit EMOS_TYPE 'pf'
                  family 01
                    edit EPSMEMBER '1'
                    edit MEMBER '1'
                    edit SUBFSFAMILY '/pf001'
                    task derived_param
                  endfamily
                  family 02
                    edit EPSMEMBER '2'
                    edit MEMBER '2'
                    edit SUBFSFAMILY '/pf002'
                    task derived_param
                  endfamily
                  family 03
                    edit EPSMEMBER '3'
                    edit MEMBER '3'
                    edit SUBFSFAMILY '/pf003'
                    task derived_param
                  endfamily
                  family 04
                    edit EPSMEMBER '4'
                    edit MEMBER '4'
                    edit SUBFSFAMILY '/pf004'
                    task derived_param
                  endfamily
                endfamily
              endfamily
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                edit EPSFCRES '319'
                task extrafields
                  edit FCLENGTH '768'
                  edit FCLENGTH0 '240'
                task wm_sfc
                  trigger extrafields eq complete
                  edit FCLENGTH '768'
                task wm_ua
                  trigger extrafields eq complete and wm_sfc eq complete
                  edit FCLENGTH '768'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        task save_logs
          trigger m2 eq complete and sv eq complete and legA eq complete and legBcpl eq complete
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
      endfamily
      family 05
        trigger ( ./rmfdb eq complete) and ( ./lock_1701 eq complete)
        edit DELTA_YEAR '-05'
        family m2
          edit DELTA_DAY '-2'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '3000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit FCNPES '16'
          edit EPSSVTSTEP '900'
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit CPUTIME '12000'
          edit STACK '350'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
          task svsh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '73'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '74'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnhevo
            trigger svnh eq complete
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit CPUTIME '120000'
          task svshevo
            trigger svsh eq complete
            edit EVOLVE 'yes'
          task svsave
            trigger svshevo eq complete and svnhevo eq complete
          task clean
            trigger svnhevo eq complete and svshevo eq complete and svsave eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
        endfamily
        family sv
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '2000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit NPES '16'
          edit EPSSVTSTEP '900'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit THREADS '8'
          task targets
            late -s +00:03 -c +00:10
            edit EPSTCBB '1'
            label da_info ""
            label dcda_info ""
            event sv
          task svsh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '73'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '74'
            inlimit /emc_36r1/limits:pesh 2
          task subspace
            trigger svnh eq complete and svsh eq complete and ( targets eq complete or targets:sv)
          task sv1
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '1'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '72'
            inlimit /emc_36r1/limits:pesh 2
          task sv2
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '2'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '71'
            inlimit /emc_36r1/limits:pesh 2
          task sv3
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '3'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv4
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '4'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv5
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '5'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv6
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '6'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task getsvevo
            event ecfs
          task getsvs
            trigger getsvevo eq complete and svnh eq complete and svsh eq complete and sv1 eq complete and sv2 eq complete and sv3 eq complete and sv4 eq complete and sv5 eq complete and sv6 eq complete and ../m2 eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTROP '1'
          task svnhevo
            trigger svnh eq complete and ../legA/fc eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit THREADS '4'
            edit STACK '350'
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
          task svshevo
            trigger svsh eq complete and ../legA/fc eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit STACK '450'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
        endfamily
        family legA
          family fc
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '240'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit TSTEP '1200'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '1'
            task getini
              edit LINIENDA '0'
              edit INIDATEENDA '0'
              edit EPSCOUPLE '0'
              edit N_PERT_AN '10'
              edit EPSCOUPL_A '0'
              edit EPSCOUPL_B '0'
              edit EPSCOUPL_C '0'
            task inidata
              trigger getini eq complete
              edit THREADS '8'
            task getae
              event 1 
            task rot
              trigger ../../sv/getsvs eq complete and ( getae:1 or getae eq complete)
              edit EPSTROP '1'
            family pert_ic
              trigger rot eq complete and inidata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTROP '1'
              inlimit /emc_36r1/limits:pertic
              family 00
                task pertinic
                  edit MEMBER '0'
              endfamily
              family 01
                task pertinic
                  edit MEMBER '1'
              endfamily
              family 02
                task pertinic
                  edit MEMBER '2'
              endfamily
              family 03
                task pertinic
                  edit MEMBER '3'
              endfamily
              family 04
                task pertinic
                  edit MEMBER '4'
              endfamily
            endfamily
            task cp_pert
              trigger pert_ic eq complete
              event 1 
            task wavini
              edit FSFAMILY 'mc'
              edit PERSST 'true'
              edit SSTINT '1'
            task wavfcdata
              trigger ./getpersSST eq complete and ../../legA/fc/wavini eq complete
              edit EPSCOUPLE '1'
            task getpersSST
              trigger ./inidata eq complete
              edit EPSCOUPLE '1'
              edit EPS_INI_STREAM 'DA'
              edit EPSFCRES '639'
            family pf
              trigger ./getpersSST eq complete and ./wavfcdata eq complete and wavini eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              meter step -1 240 240
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task getiniLeg
                  trigger ../../pert_ic/01/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task getiniLeg
                  trigger ../../pert_ic/02/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task getiniLeg
                  trigger ../../pert_ic/03/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task getiniLeg
                  trigger ../../pert_ic/04/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
            endfamily
            family cf
              trigger ./pert_ic/00/pertinic eq complete and ./getpersSST eq complete and ./wavfcdata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              task getiniLeg
                trigger ../inidata eq complete and ../wavini eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit FCNPES '32'
                edit THREADS '4'
                inlimit /emc_36r1/limits:pesh 6
                meter step -1 240 240
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../fc eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:dp
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        family legBcpl
          trigger ./legA eq complete
          family fc
            edit EPSCOUPLE '1'
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '144'
            edit EPSINISTEP_PREVIOUS '216'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit TSTEP '2700'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '2'
            family pf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit THREADS '2'
              edit RUNVARFC '1'
              meter step -1 552 552
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/01/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/01 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/02/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/02 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/03/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/03 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/04/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/04 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
            endfamily
            family cf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc0'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger ../../../../legA/fc/cf/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:pesh
              endfamily
              task getvarepsdata
                trigger ../../../legA/fc/cf eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '552'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit SMSTRIES '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:pesh
                meter step -1 552 552
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../../legA/postproc/dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:dp
              family dp
                trigger ./cf eq complete and ./pf eq complete
                edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
                edit OVERLAP '1'
                edit STEPS '216:240:6'
                inlimit /emc_36r1/limits:dp
                family cf
                  edit EMOS_TYPE 'cf'
                  task derived_param
                    edit MEMBER 'off'
                endfamily
                family pf
                  edit EMOS_TYPE 'pf'
                  family 01
                    edit EPSMEMBER '1'
                    edit MEMBER '1'
                    edit SUBFSFAMILY '/pf001'
                    task derived_param
                  endfamily
                  family 02
                    edit EPSMEMBER '2'
                    edit MEMBER '2'
                    edit SUBFSFAMILY '/pf002'
                    task derived_param
                  endfamily
                  family 03
                    edit EPSMEMBER '3'
                    edit MEMBER '3'
                    edit SUBFSFAMILY '/pf003'
                    task derived_param
                  endfamily
                  family 04
                    edit EPSMEMBER '4'
                    edit MEMBER '4'
                    edit SUBFSFAMILY '/pf004'
                    task derived_param
                  endfamily
                endfamily
              endfamily
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                edit EPSFCRES '319'
                task extrafields
                  edit FCLENGTH '768'
                  edit FCLENGTH0 '240'
                task wm_sfc
                  trigger extrafields eq complete
                  edit FCLENGTH '768'
                task wm_ua
                  trigger extrafields eq complete and wm_sfc eq complete
                  edit FCLENGTH '768'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        task save_logs
          trigger m2 eq complete and sv eq complete and legA eq complete and legBcpl eq complete
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
      endfamily
      family 04
        trigger ( ./rmfdb eq complete) and ( ./lock_1701 eq complete)
        edit DELTA_YEAR '-04'
        family m2
          edit DELTA_DAY '-2'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '3000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit FCNPES '16'
          edit EPSSVTSTEP '900'
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit CPUTIME '12000'
          edit STACK '350'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
          task svsh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '73'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '74'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnhevo
            trigger svnh eq complete
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit CPUTIME '120000'
          task svshevo
            trigger svsh eq complete
            edit EVOLVE 'yes'
          task svsave
            trigger svshevo eq complete and svnhevo eq complete
          task clean
            trigger svnhevo eq complete and svshevo eq complete and svsave eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
        endfamily
        family sv
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '2000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit NPES '16'
          edit EPSSVTSTEP '900'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit THREADS '8'
          task targets
            late -s +00:03 -c +00:10
            edit EPSTCBB '1'
            label da_info ""
            label dcda_info ""
            event sv
          task svsh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '73'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '74'
            inlimit /emc_36r1/limits:pesh 2
          task subspace
            trigger svnh eq complete and svsh eq complete and ( targets eq complete or targets:sv)
          task sv1
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '1'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '72'
            inlimit /emc_36r1/limits:pesh 2
          task sv2
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '2'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '71'
            inlimit /emc_36r1/limits:pesh 2
          task sv3
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '3'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv4
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '4'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv5
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '5'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv6
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '6'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task getsvevo
            event ecfs
          task getsvs
            trigger getsvevo eq complete and svnh eq complete and svsh eq complete and sv1 eq complete and sv2 eq complete and sv3 eq complete and sv4 eq complete and sv5 eq complete and sv6 eq complete and ../m2 eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTROP '1'
          task svnhevo
            trigger svnh eq complete and ../legA/fc eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit THREADS '4'
            edit STACK '350'
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
          task svshevo
            trigger svsh eq complete and ../legA/fc eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit STACK '450'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
        endfamily
        family legA
          family fc
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '240'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit TSTEP '1200'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '1'
            task getini
              edit LINIENDA '0'
              edit INIDATEENDA '0'
              edit EPSCOUPLE '0'
              edit N_PERT_AN '10'
              edit EPSCOUPL_A '0'
              edit EPSCOUPL_B '0'
              edit EPSCOUPL_C '0'
            task inidata
              trigger getini eq complete
              edit THREADS '8'
            task getae
              event 1 
            task rot
              trigger ../../sv/getsvs eq complete and ( getae:1 or getae eq complete)
              edit EPSTROP '1'
            family pert_ic
              trigger rot eq complete and inidata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTROP '1'
              inlimit /emc_36r1/limits:pertic
              family 00
                task pertinic
                  edit MEMBER '0'
              endfamily
              family 01
                task pertinic
                  edit MEMBER '1'
              endfamily
              family 02
                task pertinic
                  edit MEMBER '2'
              endfamily
              family 03
                task pertinic
                  edit MEMBER '3'
              endfamily
              family 04
                task pertinic
                  edit MEMBER '4'
              endfamily
            endfamily
            task cp_pert
              trigger pert_ic eq complete
              event 1 
            task wavini
              edit FSFAMILY 'mc'
              edit PERSST 'true'
              edit SSTINT '1'
            task wavfcdata
              trigger ./getpersSST eq complete and ../../legA/fc/wavini eq complete
              edit EPSCOUPLE '1'
            task getpersSST
              trigger ./inidata eq complete
              edit EPSCOUPLE '1'
              edit EPS_INI_STREAM 'DA'
              edit EPSFCRES '639'
            family pf
              trigger ./getpersSST eq complete and ./wavfcdata eq complete and wavini eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              meter step -1 240 240
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task getiniLeg
                  trigger ../../pert_ic/01/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task getiniLeg
                  trigger ../../pert_ic/02/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task getiniLeg
                  trigger ../../pert_ic/03/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task getiniLeg
                  trigger ../../pert_ic/04/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
            endfamily
            family cf
              trigger ./pert_ic/00/pertinic eq complete and ./getpersSST eq complete and ./wavfcdata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              task getiniLeg
                trigger ../inidata eq complete and ../wavini eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit FCNPES '32'
                edit THREADS '4'
                inlimit /emc_36r1/limits:pesh 6
                meter step -1 240 240
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../fc eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:dp
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        family legBcpl
          trigger ./legA eq complete
          family fc
            edit EPSCOUPLE '1'
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '144'
            edit EPSINISTEP_PREVIOUS '216'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit TSTEP '2700'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '2'
            family pf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit THREADS '2'
              edit RUNVARFC '1'
              meter step -1 552 552
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/01/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/01 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/02/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/02 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/03/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/03 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/04/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/04 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
            endfamily
            family cf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc0'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger ../../../../legA/fc/cf/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:pesh
              endfamily
              task getvarepsdata
                trigger ../../../legA/fc/cf eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '552'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit SMSTRIES '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:pesh
                meter step -1 552 552
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../../legA/postproc/dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:dp
              family dp
                trigger ./cf eq complete and ./pf eq complete
                edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
                edit OVERLAP '1'
                edit STEPS '216:240:6'
                inlimit /emc_36r1/limits:dp
                family cf
                  edit EMOS_TYPE 'cf'
                  task derived_param
                    edit MEMBER 'off'
                endfamily
                family pf
                  edit EMOS_TYPE 'pf'
                  family 01
                    edit EPSMEMBER '1'
                    edit MEMBER '1'
                    edit SUBFSFAMILY '/pf001'
                    task derived_param
                  endfamily
                  family 02
                    edit EPSMEMBER '2'
                    edit MEMBER '2'
                    edit SUBFSFAMILY '/pf002'
                    task derived_param
                  endfamily
                  family 03
                    edit EPSMEMBER '3'
                    edit MEMBER '3'
                    edit SUBFSFAMILY '/pf003'
                    task derived_param
                  endfamily
                  family 04
                    edit EPSMEMBER '4'
                    edit MEMBER '4'
                    edit SUBFSFAMILY '/pf004'
                    task derived_param
                  endfamily
                endfamily
              endfamily
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                edit EPSFCRES '319'
                task extrafields
                  edit FCLENGTH '768'
                  edit FCLENGTH0 '240'
                task wm_sfc
                  trigger extrafields eq complete
                  edit FCLENGTH '768'
                task wm_ua
                  trigger extrafields eq complete and wm_sfc eq complete
                  edit FCLENGTH '768'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        task save_logs
          trigger m2 eq complete and sv eq complete and legA eq complete and legBcpl eq complete
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
      endfamily
      family 03
        trigger ( ./rmfdb eq complete) and ( ./lock_1701 eq complete)
        edit DELTA_YEAR '-03'
        family m2
          edit DELTA_DAY '-2'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '3000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit FCNPES '16'
          edit EPSSVTSTEP '900'
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit CPUTIME '12000'
          edit STACK '350'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
          task svsh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '73'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '74'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnhevo
            trigger svnh eq complete
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit CPUTIME '120000'
          task svshevo
            trigger svsh eq complete
            edit EVOLVE 'yes'
          task svsave
            trigger svshevo eq complete and svnhevo eq complete
          task clean
            trigger svnhevo eq complete and svshevo eq complete and svsave eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
        endfamily
        family sv
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '2000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit NPES '16'
          edit EPSSVTSTEP '900'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit THREADS '8'
          task targets
            late -s +00:03 -c +00:10
            edit EPSTCBB '1'
            label da_info ""
            label dcda_info ""
            event sv
          task svsh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '73'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '74'
            inlimit /emc_36r1/limits:pesh 2
          task subspace
            trigger svnh eq complete and svsh eq complete and ( targets eq complete or targets:sv)
          task sv1
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '1'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '72'
            inlimit /emc_36r1/limits:pesh 2
          task sv2
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '2'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '71'
            inlimit /emc_36r1/limits:pesh 2
          task sv3
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '3'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv4
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '4'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv5
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '5'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv6
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '6'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task getsvevo
            event ecfs
          task getsvs
            trigger getsvevo eq complete and svnh eq complete and svsh eq complete and sv1 eq complete and sv2 eq complete and sv3 eq complete and sv4 eq complete and sv5 eq complete and sv6 eq complete and ../m2 eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTROP '1'
          task svnhevo
            trigger svnh eq complete and ../legA/fc eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit THREADS '4'
            edit STACK '350'
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
          task svshevo
            trigger svsh eq complete and ../legA/fc eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit STACK '450'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
        endfamily
        family legA
          family fc
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '240'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit TSTEP '1200'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '1'
            task getini
              edit LINIENDA '0'
              edit INIDATEENDA '0'
              edit EPSCOUPLE '0'
              edit N_PERT_AN '10'
              edit EPSCOUPL_A '0'
              edit EPSCOUPL_B '0'
              edit EPSCOUPL_C '0'
            task inidata
              trigger getini eq complete
              edit THREADS '8'
            task getae
              event 1 
            task rot
              trigger ../../sv/getsvs eq complete and ( getae:1 or getae eq complete)
              edit EPSTROP '1'
            family pert_ic
              trigger rot eq complete and inidata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTROP '1'
              inlimit /emc_36r1/limits:pertic
              family 00
                task pertinic
                  edit MEMBER '0'
              endfamily
              family 01
                task pertinic
                  edit MEMBER '1'
              endfamily
              family 02
                task pertinic
                  edit MEMBER '2'
              endfamily
              family 03
                task pertinic
                  edit MEMBER '3'
              endfamily
              family 04
                task pertinic
                  edit MEMBER '4'
              endfamily
            endfamily
            task cp_pert
              trigger pert_ic eq complete
              event 1 
            task wavini
              edit FSFAMILY 'mc'
              edit PERSST 'true'
              edit SSTINT '1'
            task wavfcdata
              trigger ./getpersSST eq complete and ../../legA/fc/wavini eq complete
              edit EPSCOUPLE '1'
            task getpersSST
              trigger ./inidata eq complete
              edit EPSCOUPLE '1'
              edit EPS_INI_STREAM 'DA'
              edit EPSFCRES '639'
            family pf
              trigger ./getpersSST eq complete and ./wavfcdata eq complete and wavini eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              meter step -1 240 240
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task getiniLeg
                  trigger ../../pert_ic/01/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task getiniLeg
                  trigger ../../pert_ic/02/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task getiniLeg
                  trigger ../../pert_ic/03/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task getiniLeg
                  trigger ../../pert_ic/04/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
            endfamily
            family cf
              trigger ./pert_ic/00/pertinic eq complete and ./getpersSST eq complete and ./wavfcdata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              task getiniLeg
                trigger ../inidata eq complete and ../wavini eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit FCNPES '32'
                edit THREADS '4'
                inlimit /emc_36r1/limits:pesh 6
                meter step -1 240 240
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../fc eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:dp
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        family legBcpl
          trigger ./legA eq complete
          family fc
            edit EPSCOUPLE '1'
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '144'
            edit EPSINISTEP_PREVIOUS '216'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit TSTEP '2700'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '2'
            family pf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit THREADS '2'
              edit RUNVARFC '1'
              meter step -1 552 552
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/01/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/01 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/02/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/02 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/03/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/03 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/04/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/04 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
            endfamily
            family cf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc0'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger ../../../../legA/fc/cf/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:pesh
              endfamily
              task getvarepsdata
                trigger ../../../legA/fc/cf eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '552'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit SMSTRIES '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:pesh
                meter step -1 552 552
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../../legA/postproc/dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:dp
              family dp
                trigger ./cf eq complete and ./pf eq complete
                edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
                edit OVERLAP '1'
                edit STEPS '216:240:6'
                inlimit /emc_36r1/limits:dp
                family cf
                  edit EMOS_TYPE 'cf'
                  task derived_param
                    edit MEMBER 'off'
                endfamily
                family pf
                  edit EMOS_TYPE 'pf'
                  family 01
                    edit EPSMEMBER '1'
                    edit MEMBER '1'
                    edit SUBFSFAMILY '/pf001'
                    task derived_param
                  endfamily
                  family 02
                    edit EPSMEMBER '2'
                    edit MEMBER '2'
                    edit SUBFSFAMILY '/pf002'
                    task derived_param
                  endfamily
                  family 03
                    edit EPSMEMBER '3'
                    edit MEMBER '3'
                    edit SUBFSFAMILY '/pf003'
                    task derived_param
                  endfamily
                  family 04
                    edit EPSMEMBER '4'
                    edit MEMBER '4'
                    edit SUBFSFAMILY '/pf004'
                    task derived_param
                  endfamily
                endfamily
              endfamily
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                edit EPSFCRES '319'
                task extrafields
                  edit FCLENGTH '768'
                  edit FCLENGTH0 '240'
                task wm_sfc
                  trigger extrafields eq complete
                  edit FCLENGTH '768'
                task wm_ua
                  trigger extrafields eq complete and wm_sfc eq complete
                  edit FCLENGTH '768'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        task save_logs
          trigger m2 eq complete and sv eq complete and legA eq complete and legBcpl eq complete
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
      endfamily
      family 02
        trigger ( ./rmfdb eq complete) and ( ./lock_1701 eq complete)
        edit DELTA_YEAR '-02'
        family m2
          edit DELTA_DAY '-2'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '3000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit FCNPES '16'
          edit EPSSVTSTEP '900'
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit CPUTIME '12000'
          edit STACK '350'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
          task svsh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '73'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '74'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnhevo
            trigger svnh eq complete
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit CPUTIME '120000'
          task svshevo
            trigger svsh eq complete
            edit EVOLVE 'yes'
          task svsave
            trigger svshevo eq complete and svnhevo eq complete
          task clean
            trigger svnhevo eq complete and svshevo eq complete and svsave eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
        endfamily
        family sv
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '2000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit NPES '16'
          edit EPSSVTSTEP '900'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit THREADS '8'
          task targets
            late -s +00:03 -c +00:10
            edit EPSTCBB '1'
            label da_info ""
            label dcda_info ""
            event sv
          task svsh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '73'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '74'
            inlimit /emc_36r1/limits:pesh 2
          task subspace
            trigger svnh eq complete and svsh eq complete and ( targets eq complete or targets:sv)
          task sv1
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '1'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '72'
            inlimit /emc_36r1/limits:pesh 2
          task sv2
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '2'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '71'
            inlimit /emc_36r1/limits:pesh 2
          task sv3
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '3'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv4
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '4'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv5
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '5'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv6
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '6'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task getsvevo
            event ecfs
          task getsvs
            trigger getsvevo eq complete and svnh eq complete and svsh eq complete and sv1 eq complete and sv2 eq complete and sv3 eq complete and sv4 eq complete and sv5 eq complete and sv6 eq complete and ../m2 eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTROP '1'
          task svnhevo
            trigger svnh eq complete and ../legA/fc eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit THREADS '4'
            edit STACK '350'
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
          task svshevo
            trigger svsh eq complete and ../legA/fc eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit STACK '450'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
        endfamily
        family legA
          family fc
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '240'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit TSTEP '1200'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '1'
            task getini
              edit LINIENDA '0'
              edit INIDATEENDA '0'
              edit EPSCOUPLE '0'
              edit N_PERT_AN '10'
              edit EPSCOUPL_A '0'
              edit EPSCOUPL_B '0'
              edit EPSCOUPL_C '0'
            task inidata
              trigger getini eq complete
              edit THREADS '8'
            task getae
              event 1 
            task rot
              trigger ../../sv/getsvs eq complete and ( getae:1 or getae eq complete)
              edit EPSTROP '1'
            family pert_ic
              trigger rot eq complete and inidata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTROP '1'
              inlimit /emc_36r1/limits:pertic
              family 00
                task pertinic
                  edit MEMBER '0'
              endfamily
              family 01
                task pertinic
                  edit MEMBER '1'
              endfamily
              family 02
                task pertinic
                  edit MEMBER '2'
              endfamily
              family 03
                task pertinic
                  edit MEMBER '3'
              endfamily
              family 04
                task pertinic
                  edit MEMBER '4'
              endfamily
            endfamily
            task cp_pert
              trigger pert_ic eq complete
              event 1 
            task wavini
              edit FSFAMILY 'mc'
              edit PERSST 'true'
              edit SSTINT '1'
            task wavfcdata
              trigger ./getpersSST eq complete and ../../legA/fc/wavini eq complete
              edit EPSCOUPLE '1'
            task getpersSST
              trigger ./inidata eq complete
              edit EPSCOUPLE '1'
              edit EPS_INI_STREAM 'DA'
              edit EPSFCRES '639'
            family pf
              trigger ./getpersSST eq complete and ./wavfcdata eq complete and wavini eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              meter step -1 240 240
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task getiniLeg
                  trigger ../../pert_ic/01/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task getiniLeg
                  trigger ../../pert_ic/02/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task getiniLeg
                  trigger ../../pert_ic/03/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task getiniLeg
                  trigger ../../pert_ic/04/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
            endfamily
            family cf
              trigger ./pert_ic/00/pertinic eq complete and ./getpersSST eq complete and ./wavfcdata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              task getiniLeg
                trigger ../inidata eq complete and ../wavini eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit FCNPES '32'
                edit THREADS '4'
                inlimit /emc_36r1/limits:pesh 6
                meter step -1 240 240
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../fc eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:dp
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        family legBcpl
          trigger ./legA eq complete
          family fc
            edit EPSCOUPLE '1'
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '144'
            edit EPSINISTEP_PREVIOUS '216'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit TSTEP '2700'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '2'
            family pf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit THREADS '2'
              edit RUNVARFC '1'
              meter step -1 552 552
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/01/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/01 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/02/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/02 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/03/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/03 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/04/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/04 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
            endfamily
            family cf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc0'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger ../../../../legA/fc/cf/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:pesh
              endfamily
              task getvarepsdata
                trigger ../../../legA/fc/cf eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '552'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit SMSTRIES '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:pesh
                meter step -1 552 552
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../../legA/postproc/dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:dp
              family dp
                trigger ./cf eq complete and ./pf eq complete
                edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
                edit OVERLAP '1'
                edit STEPS '216:240:6'
                inlimit /emc_36r1/limits:dp
                family cf
                  edit EMOS_TYPE 'cf'
                  task derived_param
                    edit MEMBER 'off'
                endfamily
                family pf
                  edit EMOS_TYPE 'pf'
                  family 01
                    edit EPSMEMBER '1'
                    edit MEMBER '1'
                    edit SUBFSFAMILY '/pf001'
                    task derived_param
                  endfamily
                  family 02
                    edit EPSMEMBER '2'
                    edit MEMBER '2'
                    edit SUBFSFAMILY '/pf002'
                    task derived_param
                  endfamily
                  family 03
                    edit EPSMEMBER '3'
                    edit MEMBER '3'
                    edit SUBFSFAMILY '/pf003'
                    task derived_param
                  endfamily
                  family 04
                    edit EPSMEMBER '4'
                    edit MEMBER '4'
                    edit SUBFSFAMILY '/pf004'
                    task derived_param
                  endfamily
                endfamily
              endfamily
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                edit EPSFCRES '319'
                task extrafields
                  edit FCLENGTH '768'
                  edit FCLENGTH0 '240'
                task wm_sfc
                  trigger extrafields eq complete
                  edit FCLENGTH '768'
                task wm_ua
                  trigger extrafields eq complete and wm_sfc eq complete
                  edit FCLENGTH '768'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        task save_logs
          trigger m2 eq complete and sv eq complete and legA eq complete and legBcpl eq complete
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
      endfamily
      family 01
        trigger ( ./rmfdb eq complete) and ( ./lock_1701 eq complete)
        edit DELTA_YEAR '-01'
        family m2
          edit DELTA_DAY '-2'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '3000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit FCNPES '16'
          edit EPSSVTSTEP '900'
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit CPUTIME '12000'
          edit STACK '350'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
          task svsh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '73'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '74'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnhevo
            trigger svnh eq complete
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit CPUTIME '120000'
          task svshevo
            trigger svsh eq complete
            edit EVOLVE 'yes'
          task svsave
            trigger svshevo eq complete and svnhevo eq complete
          task clean
            trigger svnhevo eq complete and svshevo eq complete and svsave eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
        endfamily
        family sv
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '2000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit NPES '16'
          edit EPSSVTSTEP '900'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit THREADS '8'
          task targets
            late -s +00:03 -c +00:10
            edit EPSTCBB '1'
            label da_info ""
            label dcda_info ""
            event sv
          task svsh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '73'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '74'
            inlimit /emc_36r1/limits:pesh 2
          task subspace
            trigger svnh eq complete and svsh eq complete and ( targets eq complete or targets:sv)
          task sv1
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '1'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '72'
            inlimit /emc_36r1/limits:pesh 2
          task sv2
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '2'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '71'
            inlimit /emc_36r1/limits:pesh 2
          task sv3
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '3'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv4
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '4'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv5
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '5'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv6
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '6'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task getsvevo
            event ecfs
          task getsvs
            trigger getsvevo eq complete and svnh eq complete and svsh eq complete and sv1 eq complete and sv2 eq complete and sv3 eq complete and sv4 eq complete and sv5 eq complete and sv6 eq complete and ../m2 eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTROP '1'
          task svnhevo
            trigger svnh eq complete and ../legA/fc eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit THREADS '4'
            edit STACK '350'
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
          task svshevo
            trigger svsh eq complete and ../legA/fc eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit STACK '450'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
        endfamily
        family legA
          family fc
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '240'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit TSTEP '1200'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '1'
            task getini
              edit LINIENDA '0'
              edit INIDATEENDA '0'
              edit EPSCOUPLE '0'
              edit N_PERT_AN '10'
              edit EPSCOUPL_A '0'
              edit EPSCOUPL_B '0'
              edit EPSCOUPL_C '0'
            task inidata
              trigger getini eq complete
              edit THREADS '8'
            task getae
              event 1 
            task rot
              trigger ../../sv/getsvs eq complete and ( getae:1 or getae eq complete)
              edit EPSTROP '1'
            family pert_ic
              trigger rot eq complete and inidata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTROP '1'
              inlimit /emc_36r1/limits:pertic
              family 00
                task pertinic
                  edit MEMBER '0'
              endfamily
              family 01
                task pertinic
                  edit MEMBER '1'
              endfamily
              family 02
                task pertinic
                  edit MEMBER '2'
              endfamily
              family 03
                task pertinic
                  edit MEMBER '3'
              endfamily
              family 04
                task pertinic
                  edit MEMBER '4'
              endfamily
            endfamily
            task cp_pert
              trigger pert_ic eq complete
              event 1 
            task wavini
              edit FSFAMILY 'mc'
              edit PERSST 'true'
              edit SSTINT '1'
            task wavfcdata
              trigger ./getpersSST eq complete and ../../legA/fc/wavini eq complete
              edit EPSCOUPLE '1'
            task getpersSST
              trigger ./inidata eq complete
              edit EPSCOUPLE '1'
              edit EPS_INI_STREAM 'DA'
              edit EPSFCRES '639'
            family pf
              trigger ./getpersSST eq complete and ./wavfcdata eq complete and wavini eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              meter step -1 240 240
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task getiniLeg
                  trigger ../../pert_ic/01/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task getiniLeg
                  trigger ../../pert_ic/02/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task getiniLeg
                  trigger ../../pert_ic/03/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task getiniLeg
                  trigger ../../pert_ic/04/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
            endfamily
            family cf
              trigger ./pert_ic/00/pertinic eq complete and ./getpersSST eq complete and ./wavfcdata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              task getiniLeg
                trigger ../inidata eq complete and ../wavini eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit FCNPES '32'
                edit THREADS '4'
                inlimit /emc_36r1/limits:pesh 6
                meter step -1 240 240
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../fc eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:dp
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        family legBcpl
          trigger ./legA eq complete
          family fc
            edit EPSCOUPLE '1'
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '144'
            edit EPSINISTEP_PREVIOUS '216'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit TSTEP '2700'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '2'
            family pf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit THREADS '2'
              edit RUNVARFC '1'
              meter step -1 552 552
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/01/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/01 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/02/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/02 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/03/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/03 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/04/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/04 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
            endfamily
            family cf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc0'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger ../../../../legA/fc/cf/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:pesh
              endfamily
              task getvarepsdata
                trigger ../../../legA/fc/cf eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '552'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit SMSTRIES '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:pesh
                meter step -1 552 552
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../../legA/postproc/dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:dp
              family dp
                trigger ./cf eq complete and ./pf eq complete
                edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
                edit OVERLAP '1'
                edit STEPS '216:240:6'
                inlimit /emc_36r1/limits:dp
                family cf
                  edit EMOS_TYPE 'cf'
                  task derived_param
                    edit MEMBER 'off'
                endfamily
                family pf
                  edit EMOS_TYPE 'pf'
                  family 01
                    edit EPSMEMBER '1'
                    edit MEMBER '1'
                    edit SUBFSFAMILY '/pf001'
                    task derived_param
                  endfamily
                  family 02
                    edit EPSMEMBER '2'
                    edit MEMBER '2'
                    edit SUBFSFAMILY '/pf002'
                    task derived_param
                  endfamily
                  family 03
                    edit EPSMEMBER '3'
                    edit MEMBER '3'
                    edit SUBFSFAMILY '/pf003'
                    task derived_param
                  endfamily
                  family 04
                    edit EPSMEMBER '4'
                    edit MEMBER '4'
                    edit SUBFSFAMILY '/pf004'
                    task derived_param
                  endfamily
                endfamily
              endfamily
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                edit EPSFCRES '319'
                task extrafields
                  edit FCLENGTH '768'
                  edit FCLENGTH0 '240'
                task wm_sfc
                  trigger extrafields eq complete
                  edit FCLENGTH '768'
                task wm_ua
                  trigger extrafields eq complete and wm_sfc eq complete
                  edit FCLENGTH '768'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        task save_logs
          trigger m2 eq complete and sv eq complete and legA eq complete and legBcpl eq complete
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
      endfamily
    endfamily
    family arch
      edit EMOS_TYPE 'an'
      edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
      edit FSFAMILY 'mars'
      edit OVERLAP '0'
      edit YMD '20100121'
      limit mutex 1
      inlimit /emc_36r1/limits:mars
      inlimit /limits:mars
      family 18
        trigger ( ../back/18 eq complete)
        edit DELTA_YEAR '-18'
        family legA
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf144
              edit STEPS '0/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/144/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf240
              edit STEPS '150/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf072
              edit STEPS '0/to/072/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/072/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf144
              edit STEPS '078/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf192
              edit STEPS '150/to/192/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf240
              edit STEPS '198/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
              task pfml
                edit STEPS '216'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit RUNVARFC '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family coupled
            trigger ./cf eq complete and ./pf eq complete
            edit EPSLEG '1'
            edit OCLEG '1'
            edit OCEPS '1'
            edit FCCHUNK '1'
            edit FCLENGTH '240'
            edit FCLENGTH0 '0'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '0'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legB
          edit EPSLEG '2'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            task cfml
              edit STEPS '216'
            family cf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task pfml
              edit STEPS '360'
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family coupled
            edit EPSLEG '2'
            edit OCLEG '2'
            edit OCEPS '2'
            edit FCCHUNK '1'
            edit FCLENGTH '360'
            edit FCLENGTH0 '216'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '1'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legC
          edit EPSLEG '3'
          edit FSFAMILY 'mars'
          edit OVERLAP '0'
          edit EMOS_TYPE 'an'
          edit EPSFCRES '255'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit FCCHUNK '1'
            edit FCLENGTH '768'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger /emc_36r1/00/legC/postproc/dp eq complete and /emc_36r1/00/legC/postproc/extra eq complete
              edit RUNVARFC '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              edit EMOS_TYPE 'pf'
              edit STEPS '366/to/768/by/6'
              family cf
                edit EMOS_TYPE 'cf'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family extra
              trigger /emc_36r1/hc2/back/18/legBcpl/postproc/extra eq complete
              edit RUNVARFC '1'
              task wm_sfc_arc
                label files ""
                label info ""
                event fdb
              task wm_ua_arc
                label files ""
                label info ""
                event fdb
            endfamily
            family nino
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit DAILY '0'
              edit RUNVARFC '1'
              edit FCLENGTH1 '360'
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit FCLENGTH '768'
              edit EPSCOUPLE '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              limit semaphore 10
              inlimit /limits:linux_cluster
              inlimit semaphore
              family 00
                edit OCGROUP 'an00'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '0'
                edit MEMBER '0'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 01
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '1'
                edit MEMBER '1'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 02
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '2'
                edit MEMBER '2'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 03
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '3'
                edit MEMBER '3'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 04
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '4'
                edit MEMBER '4'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
            endfamily
            task tcyctoecfs
              trigger nino eq complete
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit EPSCOUPLE '1'
              edit FCLENGTH '768'
              inlimit /limits:linux_cluster
          endfamily
        endfamily
        task save_logs
          trigger legA eq complete and legB eq complete and legC eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit SCHOST 'localhost'
      endfamily
      family 17
        trigger ( ../back/17 eq complete)
        edit DELTA_YEAR '-17'
        family legA
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf144
              edit STEPS '0/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/144/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf240
              edit STEPS '150/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf072
              edit STEPS '0/to/072/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/072/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf144
              edit STEPS '078/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf192
              edit STEPS '150/to/192/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf240
              edit STEPS '198/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
              task pfml
                edit STEPS '216'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit RUNVARFC '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family coupled
            trigger ./cf eq complete and ./pf eq complete
            edit EPSLEG '1'
            edit OCLEG '1'
            edit OCEPS '1'
            edit FCCHUNK '1'
            edit FCLENGTH '240'
            edit FCLENGTH0 '0'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '0'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legB
          edit EPSLEG '2'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            task cfml
              edit STEPS '216'
            family cf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task pfml
              edit STEPS '360'
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family coupled
            edit EPSLEG '2'
            edit OCLEG '2'
            edit OCEPS '2'
            edit FCCHUNK '1'
            edit FCLENGTH '360'
            edit FCLENGTH0 '216'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '1'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legC
          edit EPSLEG '3'
          edit FSFAMILY 'mars'
          edit OVERLAP '0'
          edit EMOS_TYPE 'an'
          edit EPSFCRES '255'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit FCCHUNK '1'
            edit FCLENGTH '768'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger /emc_36r1/00/legC/postproc/dp eq complete and /emc_36r1/00/legC/postproc/extra eq complete
              edit RUNVARFC '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              edit EMOS_TYPE 'pf'
              edit STEPS '366/to/768/by/6'
              family cf
                edit EMOS_TYPE 'cf'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family extra
              trigger /emc_36r1/hc2/back/17/legBcpl/postproc/extra eq complete
              edit RUNVARFC '1'
              task wm_sfc_arc
                label files ""
                label info ""
                event fdb
              task wm_ua_arc
                label files ""
                label info ""
                event fdb
            endfamily
            family nino
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit DAILY '0'
              edit RUNVARFC '1'
              edit FCLENGTH1 '360'
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit FCLENGTH '768'
              edit EPSCOUPLE '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              limit semaphore 10
              inlimit /limits:linux_cluster
              inlimit semaphore
              family 00
                edit OCGROUP 'an00'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '0'
                edit MEMBER '0'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 01
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '1'
                edit MEMBER '1'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 02
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '2'
                edit MEMBER '2'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 03
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '3'
                edit MEMBER '3'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 04
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '4'
                edit MEMBER '4'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
            endfamily
            task tcyctoecfs
              trigger nino eq complete
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit EPSCOUPLE '1'
              edit FCLENGTH '768'
              inlimit /limits:linux_cluster
          endfamily
        endfamily
        task save_logs
          trigger legA eq complete and legB eq complete and legC eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit SCHOST 'localhost'
      endfamily
      family 16
        trigger ( ../back/16 eq complete)
        edit DELTA_YEAR '-16'
        family legA
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf144
              edit STEPS '0/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/144/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf240
              edit STEPS '150/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf072
              edit STEPS '0/to/072/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/072/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf144
              edit STEPS '078/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf192
              edit STEPS '150/to/192/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf240
              edit STEPS '198/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
              task pfml
                edit STEPS '216'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit RUNVARFC '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family coupled
            trigger ./cf eq complete and ./pf eq complete
            edit EPSLEG '1'
            edit OCLEG '1'
            edit OCEPS '1'
            edit FCCHUNK '1'
            edit FCLENGTH '240'
            edit FCLENGTH0 '0'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '0'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legB
          edit EPSLEG '2'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            task cfml
              edit STEPS '216'
            family cf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task pfml
              edit STEPS '360'
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family coupled
            edit EPSLEG '2'
            edit OCLEG '2'
            edit OCEPS '2'
            edit FCCHUNK '1'
            edit FCLENGTH '360'
            edit FCLENGTH0 '216'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '1'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legC
          edit EPSLEG '3'
          edit FSFAMILY 'mars'
          edit OVERLAP '0'
          edit EMOS_TYPE 'an'
          edit EPSFCRES '255'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit FCCHUNK '1'
            edit FCLENGTH '768'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger /emc_36r1/00/legC/postproc/dp eq complete and /emc_36r1/00/legC/postproc/extra eq complete
              edit RUNVARFC '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              edit EMOS_TYPE 'pf'
              edit STEPS '366/to/768/by/6'
              family cf
                edit EMOS_TYPE 'cf'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family extra
              trigger /emc_36r1/hc2/back/16/legBcpl/postproc/extra eq complete
              edit RUNVARFC '1'
              task wm_sfc_arc
                label files ""
                label info ""
                event fdb
              task wm_ua_arc
                label files ""
                label info ""
                event fdb
            endfamily
            family nino
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit DAILY '0'
              edit RUNVARFC '1'
              edit FCLENGTH1 '360'
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit FCLENGTH '768'
              edit EPSCOUPLE '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              limit semaphore 10
              inlimit /limits:linux_cluster
              inlimit semaphore
              family 00
                edit OCGROUP 'an00'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '0'
                edit MEMBER '0'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 01
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '1'
                edit MEMBER '1'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 02
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '2'
                edit MEMBER '2'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 03
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '3'
                edit MEMBER '3'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 04
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '4'
                edit MEMBER '4'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
            endfamily
            task tcyctoecfs
              trigger nino eq complete
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit EPSCOUPLE '1'
              edit FCLENGTH '768'
              inlimit /limits:linux_cluster
          endfamily
        endfamily
        task save_logs
          trigger legA eq complete and legB eq complete and legC eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit SCHOST 'localhost'
      endfamily
      family 15
        trigger ( ../back/15 eq complete)
        edit DELTA_YEAR '-15'
        family legA
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf144
              edit STEPS '0/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/144/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf240
              edit STEPS '150/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf072
              edit STEPS '0/to/072/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/072/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf144
              edit STEPS '078/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf192
              edit STEPS '150/to/192/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf240
              edit STEPS '198/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
              task pfml
                edit STEPS '216'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit RUNVARFC '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family coupled
            trigger ./cf eq complete and ./pf eq complete
            edit EPSLEG '1'
            edit OCLEG '1'
            edit OCEPS '1'
            edit FCCHUNK '1'
            edit FCLENGTH '240'
            edit FCLENGTH0 '0'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '0'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legB
          edit EPSLEG '2'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            task cfml
              edit STEPS '216'
            family cf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task pfml
              edit STEPS '360'
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family coupled
            edit EPSLEG '2'
            edit OCLEG '2'
            edit OCEPS '2'
            edit FCCHUNK '1'
            edit FCLENGTH '360'
            edit FCLENGTH0 '216'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '1'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legC
          edit EPSLEG '3'
          edit FSFAMILY 'mars'
          edit OVERLAP '0'
          edit EMOS_TYPE 'an'
          edit EPSFCRES '255'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit FCCHUNK '1'
            edit FCLENGTH '768'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger /emc_36r1/00/legC/postproc/dp eq complete and /emc_36r1/00/legC/postproc/extra eq complete
              edit RUNVARFC '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              edit EMOS_TYPE 'pf'
              edit STEPS '366/to/768/by/6'
              family cf
                edit EMOS_TYPE 'cf'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family extra
              trigger /emc_36r1/hc2/back/15/legBcpl/postproc/extra eq complete
              edit RUNVARFC '1'
              task wm_sfc_arc
                label files ""
                label info ""
                event fdb
              task wm_ua_arc
                label files ""
                label info ""
                event fdb
            endfamily
            family nino
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit DAILY '0'
              edit RUNVARFC '1'
              edit FCLENGTH1 '360'
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit FCLENGTH '768'
              edit EPSCOUPLE '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              limit semaphore 10
              inlimit /limits:linux_cluster
              inlimit semaphore
              family 00
                edit OCGROUP 'an00'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '0'
                edit MEMBER '0'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 01
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '1'
                edit MEMBER '1'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 02
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '2'
                edit MEMBER '2'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 03
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '3'
                edit MEMBER '3'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 04
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '4'
                edit MEMBER '4'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
            endfamily
            task tcyctoecfs
              trigger nino eq complete
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit EPSCOUPLE '1'
              edit FCLENGTH '768'
              inlimit /limits:linux_cluster
          endfamily
        endfamily
        task save_logs
          trigger legA eq complete and legB eq complete and legC eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit SCHOST 'localhost'
      endfamily
      family 14
        trigger ( ../back/14 eq complete)
        edit DELTA_YEAR '-14'
        family legA
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf144
              edit STEPS '0/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/144/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf240
              edit STEPS '150/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf072
              edit STEPS '0/to/072/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/072/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf144
              edit STEPS '078/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf192
              edit STEPS '150/to/192/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf240
              edit STEPS '198/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
              task pfml
                edit STEPS '216'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit RUNVARFC '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family coupled
            trigger ./cf eq complete and ./pf eq complete
            edit EPSLEG '1'
            edit OCLEG '1'
            edit OCEPS '1'
            edit FCCHUNK '1'
            edit FCLENGTH '240'
            edit FCLENGTH0 '0'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '0'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legB
          edit EPSLEG '2'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            task cfml
              edit STEPS '216'
            family cf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task pfml
              edit STEPS '360'
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family coupled
            edit EPSLEG '2'
            edit OCLEG '2'
            edit OCEPS '2'
            edit FCCHUNK '1'
            edit FCLENGTH '360'
            edit FCLENGTH0 '216'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '1'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legC
          edit EPSLEG '3'
          edit FSFAMILY 'mars'
          edit OVERLAP '0'
          edit EMOS_TYPE 'an'
          edit EPSFCRES '255'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit FCCHUNK '1'
            edit FCLENGTH '768'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger /emc_36r1/00/legC/postproc/dp eq complete and /emc_36r1/00/legC/postproc/extra eq complete
              edit RUNVARFC '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              edit EMOS_TYPE 'pf'
              edit STEPS '366/to/768/by/6'
              family cf
                edit EMOS_TYPE 'cf'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family extra
              trigger /emc_36r1/hc2/back/14/legBcpl/postproc/extra eq complete
              edit RUNVARFC '1'
              task wm_sfc_arc
                label files ""
                label info ""
                event fdb
              task wm_ua_arc
                label files ""
                label info ""
                event fdb
            endfamily
            family nino
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit DAILY '0'
              edit RUNVARFC '1'
              edit FCLENGTH1 '360'
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit FCLENGTH '768'
              edit EPSCOUPLE '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              limit semaphore 10
              inlimit /limits:linux_cluster
              inlimit semaphore
              family 00
                edit OCGROUP 'an00'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '0'
                edit MEMBER '0'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 01
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '1'
                edit MEMBER '1'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 02
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '2'
                edit MEMBER '2'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 03
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '3'
                edit MEMBER '3'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 04
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '4'
                edit MEMBER '4'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
            endfamily
            task tcyctoecfs
              trigger nino eq complete
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit EPSCOUPLE '1'
              edit FCLENGTH '768'
              inlimit /limits:linux_cluster
          endfamily
        endfamily
        task save_logs
          trigger legA eq complete and legB eq complete and legC eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit SCHOST 'localhost'
      endfamily
      family 13
        trigger ( ../back/13 eq complete)
        edit DELTA_YEAR '-13'
        family legA
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf144
              edit STEPS '0/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/144/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf240
              edit STEPS '150/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf072
              edit STEPS '0/to/072/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/072/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf144
              edit STEPS '078/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf192
              edit STEPS '150/to/192/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf240
              edit STEPS '198/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
              task pfml
                edit STEPS '216'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit RUNVARFC '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family coupled
            trigger ./cf eq complete and ./pf eq complete
            edit EPSLEG '1'
            edit OCLEG '1'
            edit OCEPS '1'
            edit FCCHUNK '1'
            edit FCLENGTH '240'
            edit FCLENGTH0 '0'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '0'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legB
          edit EPSLEG '2'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            task cfml
              edit STEPS '216'
            family cf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task pfml
              edit STEPS '360'
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family coupled
            edit EPSLEG '2'
            edit OCLEG '2'
            edit OCEPS '2'
            edit FCCHUNK '1'
            edit FCLENGTH '360'
            edit FCLENGTH0 '216'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '1'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legC
          edit EPSLEG '3'
          edit FSFAMILY 'mars'
          edit OVERLAP '0'
          edit EMOS_TYPE 'an'
          edit EPSFCRES '255'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit FCCHUNK '1'
            edit FCLENGTH '768'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger /emc_36r1/00/legC/postproc/dp eq complete and /emc_36r1/00/legC/postproc/extra eq complete
              edit RUNVARFC '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              edit EMOS_TYPE 'pf'
              edit STEPS '366/to/768/by/6'
              family cf
                edit EMOS_TYPE 'cf'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family extra
              trigger /emc_36r1/hc2/back/13/legBcpl/postproc/extra eq complete
              edit RUNVARFC '1'
              task wm_sfc_arc
                label files ""
                label info ""
                event fdb
              task wm_ua_arc
                label files ""
                label info ""
                event fdb
            endfamily
            family nino
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit DAILY '0'
              edit RUNVARFC '1'
              edit FCLENGTH1 '360'
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit FCLENGTH '768'
              edit EPSCOUPLE '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              limit semaphore 10
              inlimit /limits:linux_cluster
              inlimit semaphore
              family 00
                edit OCGROUP 'an00'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '0'
                edit MEMBER '0'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 01
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '1'
                edit MEMBER '1'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 02
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '2'
                edit MEMBER '2'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 03
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '3'
                edit MEMBER '3'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 04
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '4'
                edit MEMBER '4'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
            endfamily
            task tcyctoecfs
              trigger nino eq complete
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit EPSCOUPLE '1'
              edit FCLENGTH '768'
              inlimit /limits:linux_cluster
          endfamily
        endfamily
        task save_logs
          trigger legA eq complete and legB eq complete and legC eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit SCHOST 'localhost'
      endfamily
      family 12
        trigger ( ../back/12 eq complete)
        edit DELTA_YEAR '-12'
        family legA
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf144
              edit STEPS '0/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/144/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf240
              edit STEPS '150/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf072
              edit STEPS '0/to/072/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/072/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf144
              edit STEPS '078/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf192
              edit STEPS '150/to/192/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf240
              edit STEPS '198/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
              task pfml
                edit STEPS '216'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit RUNVARFC '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family coupled
            trigger ./cf eq complete and ./pf eq complete
            edit EPSLEG '1'
            edit OCLEG '1'
            edit OCEPS '1'
            edit FCCHUNK '1'
            edit FCLENGTH '240'
            edit FCLENGTH0 '0'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '0'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legB
          edit EPSLEG '2'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            task cfml
              edit STEPS '216'
            family cf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task pfml
              edit STEPS '360'
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family coupled
            edit EPSLEG '2'
            edit OCLEG '2'
            edit OCEPS '2'
            edit FCCHUNK '1'
            edit FCLENGTH '360'
            edit FCLENGTH0 '216'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '1'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legC
          edit EPSLEG '3'
          edit FSFAMILY 'mars'
          edit OVERLAP '0'
          edit EMOS_TYPE 'an'
          edit EPSFCRES '255'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit FCCHUNK '1'
            edit FCLENGTH '768'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger /emc_36r1/00/legC/postproc/dp eq complete and /emc_36r1/00/legC/postproc/extra eq complete
              edit RUNVARFC '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              edit EMOS_TYPE 'pf'
              edit STEPS '366/to/768/by/6'
              family cf
                edit EMOS_TYPE 'cf'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family extra
              trigger /emc_36r1/hc2/back/12/legBcpl/postproc/extra eq complete
              edit RUNVARFC '1'
              task wm_sfc_arc
                label files ""
                label info ""
                event fdb
              task wm_ua_arc
                label files ""
                label info ""
                event fdb
            endfamily
            family nino
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit DAILY '0'
              edit RUNVARFC '1'
              edit FCLENGTH1 '360'
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit FCLENGTH '768'
              edit EPSCOUPLE '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              limit semaphore 10
              inlimit /limits:linux_cluster
              inlimit semaphore
              family 00
                edit OCGROUP 'an00'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '0'
                edit MEMBER '0'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 01
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '1'
                edit MEMBER '1'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 02
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '2'
                edit MEMBER '2'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 03
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '3'
                edit MEMBER '3'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 04
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '4'
                edit MEMBER '4'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
            endfamily
            task tcyctoecfs
              trigger nino eq complete
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit EPSCOUPLE '1'
              edit FCLENGTH '768'
              inlimit /limits:linux_cluster
          endfamily
        endfamily
        task save_logs
          trigger legA eq complete and legB eq complete and legC eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit SCHOST 'localhost'
      endfamily
      family 11
        trigger ( ../back/11 eq complete)
        edit DELTA_YEAR '-11'
        family legA
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf144
              edit STEPS '0/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/144/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf240
              edit STEPS '150/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf072
              edit STEPS '0/to/072/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/072/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf144
              edit STEPS '078/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf192
              edit STEPS '150/to/192/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf240
              edit STEPS '198/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
              task pfml
                edit STEPS '216'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit RUNVARFC '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family coupled
            trigger ./cf eq complete and ./pf eq complete
            edit EPSLEG '1'
            edit OCLEG '1'
            edit OCEPS '1'
            edit FCCHUNK '1'
            edit FCLENGTH '240'
            edit FCLENGTH0 '0'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '0'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legB
          edit EPSLEG '2'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            task cfml
              edit STEPS '216'
            family cf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task pfml
              edit STEPS '360'
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family coupled
            edit EPSLEG '2'
            edit OCLEG '2'
            edit OCEPS '2'
            edit FCCHUNK '1'
            edit FCLENGTH '360'
            edit FCLENGTH0 '216'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '1'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legC
          edit EPSLEG '3'
          edit FSFAMILY 'mars'
          edit OVERLAP '0'
          edit EMOS_TYPE 'an'
          edit EPSFCRES '255'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit FCCHUNK '1'
            edit FCLENGTH '768'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger /emc_36r1/00/legC/postproc/dp eq complete and /emc_36r1/00/legC/postproc/extra eq complete
              edit RUNVARFC '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              edit EMOS_TYPE 'pf'
              edit STEPS '366/to/768/by/6'
              family cf
                edit EMOS_TYPE 'cf'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family extra
              trigger /emc_36r1/hc2/back/11/legBcpl/postproc/extra eq complete
              edit RUNVARFC '1'
              task wm_sfc_arc
                label files ""
                label info ""
                event fdb
              task wm_ua_arc
                label files ""
                label info ""
                event fdb
            endfamily
            family nino
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit DAILY '0'
              edit RUNVARFC '1'
              edit FCLENGTH1 '360'
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit FCLENGTH '768'
              edit EPSCOUPLE '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              limit semaphore 10
              inlimit /limits:linux_cluster
              inlimit semaphore
              family 00
                edit OCGROUP 'an00'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '0'
                edit MEMBER '0'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 01
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '1'
                edit MEMBER '1'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 02
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '2'
                edit MEMBER '2'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 03
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '3'
                edit MEMBER '3'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 04
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '4'
                edit MEMBER '4'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
            endfamily
            task tcyctoecfs
              trigger nino eq complete
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit EPSCOUPLE '1'
              edit FCLENGTH '768'
              inlimit /limits:linux_cluster
          endfamily
        endfamily
        task save_logs
          trigger legA eq complete and legB eq complete and legC eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit SCHOST 'localhost'
      endfamily
      family 10
        trigger ( ../back/10 eq complete)
        edit DELTA_YEAR '-10'
        family legA
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf144
              edit STEPS '0/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/144/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf240
              edit STEPS '150/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf072
              edit STEPS '0/to/072/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/072/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf144
              edit STEPS '078/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf192
              edit STEPS '150/to/192/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf240
              edit STEPS '198/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
              task pfml
                edit STEPS '216'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit RUNVARFC '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family coupled
            trigger ./cf eq complete and ./pf eq complete
            edit EPSLEG '1'
            edit OCLEG '1'
            edit OCEPS '1'
            edit FCCHUNK '1'
            edit FCLENGTH '240'
            edit FCLENGTH0 '0'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '0'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legB
          edit EPSLEG '2'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            task cfml
              edit STEPS '216'
            family cf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task pfml
              edit STEPS '360'
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family coupled
            edit EPSLEG '2'
            edit OCLEG '2'
            edit OCEPS '2'
            edit FCCHUNK '1'
            edit FCLENGTH '360'
            edit FCLENGTH0 '216'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '1'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legC
          edit EPSLEG '3'
          edit FSFAMILY 'mars'
          edit OVERLAP '0'
          edit EMOS_TYPE 'an'
          edit EPSFCRES '255'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit FCCHUNK '1'
            edit FCLENGTH '768'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger /emc_36r1/00/legC/postproc/dp eq complete and /emc_36r1/00/legC/postproc/extra eq complete
              edit RUNVARFC '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              edit EMOS_TYPE 'pf'
              edit STEPS '366/to/768/by/6'
              family cf
                edit EMOS_TYPE 'cf'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family extra
              trigger /emc_36r1/hc2/back/10/legBcpl/postproc/extra eq complete
              edit RUNVARFC '1'
              task wm_sfc_arc
                label files ""
                label info ""
                event fdb
              task wm_ua_arc
                label files ""
                label info ""
                event fdb
            endfamily
            family nino
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit DAILY '0'
              edit RUNVARFC '1'
              edit FCLENGTH1 '360'
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit FCLENGTH '768'
              edit EPSCOUPLE '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              limit semaphore 10
              inlimit /limits:linux_cluster
              inlimit semaphore
              family 00
                edit OCGROUP 'an00'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '0'
                edit MEMBER '0'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 01
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '1'
                edit MEMBER '1'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 02
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '2'
                edit MEMBER '2'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 03
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '3'
                edit MEMBER '3'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 04
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '4'
                edit MEMBER '4'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
            endfamily
            task tcyctoecfs
              trigger nino eq complete
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit EPSCOUPLE '1'
              edit FCLENGTH '768'
              inlimit /limits:linux_cluster
          endfamily
        endfamily
        task save_logs
          trigger legA eq complete and legB eq complete and legC eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit SCHOST 'localhost'
      endfamily
      family 09
        trigger ( ../back/09 eq complete)
        edit DELTA_YEAR '-09'
        family legA
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf144
              edit STEPS '0/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/144/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf240
              edit STEPS '150/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf072
              edit STEPS '0/to/072/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/072/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf144
              edit STEPS '078/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf192
              edit STEPS '150/to/192/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf240
              edit STEPS '198/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
              task pfml
                edit STEPS '216'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit RUNVARFC '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family coupled
            trigger ./cf eq complete and ./pf eq complete
            edit EPSLEG '1'
            edit OCLEG '1'
            edit OCEPS '1'
            edit FCCHUNK '1'
            edit FCLENGTH '240'
            edit FCLENGTH0 '0'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '0'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legB
          edit EPSLEG '2'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            task cfml
              edit STEPS '216'
            family cf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task pfml
              edit STEPS '360'
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family coupled
            edit EPSLEG '2'
            edit OCLEG '2'
            edit OCEPS '2'
            edit FCCHUNK '1'
            edit FCLENGTH '360'
            edit FCLENGTH0 '216'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '1'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legC
          edit EPSLEG '3'
          edit FSFAMILY 'mars'
          edit OVERLAP '0'
          edit EMOS_TYPE 'an'
          edit EPSFCRES '255'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit FCCHUNK '1'
            edit FCLENGTH '768'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger /emc_36r1/00/legC/postproc/dp eq complete and /emc_36r1/00/legC/postproc/extra eq complete
              edit RUNVARFC '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              edit EMOS_TYPE 'pf'
              edit STEPS '366/to/768/by/6'
              family cf
                edit EMOS_TYPE 'cf'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family extra
              trigger /emc_36r1/hc2/back/09/legBcpl/postproc/extra eq complete
              edit RUNVARFC '1'
              task wm_sfc_arc
                label files ""
                label info ""
                event fdb
              task wm_ua_arc
                label files ""
                label info ""
                event fdb
            endfamily
            family nino
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit DAILY '0'
              edit RUNVARFC '1'
              edit FCLENGTH1 '360'
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit FCLENGTH '768'
              edit EPSCOUPLE '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              limit semaphore 10
              inlimit /limits:linux_cluster
              inlimit semaphore
              family 00
                edit OCGROUP 'an00'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '0'
                edit MEMBER '0'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 01
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '1'
                edit MEMBER '1'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 02
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '2'
                edit MEMBER '2'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 03
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '3'
                edit MEMBER '3'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 04
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '4'
                edit MEMBER '4'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
            endfamily
            task tcyctoecfs
              trigger nino eq complete
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit EPSCOUPLE '1'
              edit FCLENGTH '768'
              inlimit /limits:linux_cluster
          endfamily
        endfamily
        task save_logs
          trigger legA eq complete and legB eq complete and legC eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit SCHOST 'localhost'
      endfamily
      family 08
        trigger ( ../back/08 eq complete)
        edit DELTA_YEAR '-08'
        family legA
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf144
              edit STEPS '0/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/144/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf240
              edit STEPS '150/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf072
              edit STEPS '0/to/072/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/072/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf144
              edit STEPS '078/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf192
              edit STEPS '150/to/192/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf240
              edit STEPS '198/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
              task pfml
                edit STEPS '216'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit RUNVARFC '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family coupled
            trigger ./cf eq complete and ./pf eq complete
            edit EPSLEG '1'
            edit OCLEG '1'
            edit OCEPS '1'
            edit FCCHUNK '1'
            edit FCLENGTH '240'
            edit FCLENGTH0 '0'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '0'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legB
          edit EPSLEG '2'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            task cfml
              edit STEPS '216'
            family cf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task pfml
              edit STEPS '360'
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family coupled
            edit EPSLEG '2'
            edit OCLEG '2'
            edit OCEPS '2'
            edit FCCHUNK '1'
            edit FCLENGTH '360'
            edit FCLENGTH0 '216'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '1'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legC
          edit EPSLEG '3'
          edit FSFAMILY 'mars'
          edit OVERLAP '0'
          edit EMOS_TYPE 'an'
          edit EPSFCRES '255'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit FCCHUNK '1'
            edit FCLENGTH '768'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger /emc_36r1/00/legC/postproc/dp eq complete and /emc_36r1/00/legC/postproc/extra eq complete
              edit RUNVARFC '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              edit EMOS_TYPE 'pf'
              edit STEPS '366/to/768/by/6'
              family cf
                edit EMOS_TYPE 'cf'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family extra
              trigger /emc_36r1/hc2/back/08/legBcpl/postproc/extra eq complete
              edit RUNVARFC '1'
              task wm_sfc_arc
                label files ""
                label info ""
                event fdb
              task wm_ua_arc
                label files ""
                label info ""
                event fdb
            endfamily
            family nino
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit DAILY '0'
              edit RUNVARFC '1'
              edit FCLENGTH1 '360'
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit FCLENGTH '768'
              edit EPSCOUPLE '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              limit semaphore 10
              inlimit /limits:linux_cluster
              inlimit semaphore
              family 00
                edit OCGROUP 'an00'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '0'
                edit MEMBER '0'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 01
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '1'
                edit MEMBER '1'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 02
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '2'
                edit MEMBER '2'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 03
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '3'
                edit MEMBER '3'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 04
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '4'
                edit MEMBER '4'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
            endfamily
            task tcyctoecfs
              trigger nino eq complete
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit EPSCOUPLE '1'
              edit FCLENGTH '768'
              inlimit /limits:linux_cluster
          endfamily
        endfamily
        task save_logs
          trigger legA eq complete and legB eq complete and legC eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit SCHOST 'localhost'
      endfamily
      family 07
        trigger ( ../back/07 eq complete)
        edit DELTA_YEAR '-07'
        family legA
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf144
              edit STEPS '0/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/144/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf240
              edit STEPS '150/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf072
              edit STEPS '0/to/072/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/072/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf144
              edit STEPS '078/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf192
              edit STEPS '150/to/192/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf240
              edit STEPS '198/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
              task pfml
                edit STEPS '216'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit RUNVARFC '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family coupled
            trigger ./cf eq complete and ./pf eq complete
            edit EPSLEG '1'
            edit OCLEG '1'
            edit OCEPS '1'
            edit FCCHUNK '1'
            edit FCLENGTH '240'
            edit FCLENGTH0 '0'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '0'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legB
          edit EPSLEG '2'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            task cfml
              edit STEPS '216'
            family cf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task pfml
              edit STEPS '360'
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family coupled
            edit EPSLEG '2'
            edit OCLEG '2'
            edit OCEPS '2'
            edit FCCHUNK '1'
            edit FCLENGTH '360'
            edit FCLENGTH0 '216'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '1'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legC
          edit EPSLEG '3'
          edit FSFAMILY 'mars'
          edit OVERLAP '0'
          edit EMOS_TYPE 'an'
          edit EPSFCRES '255'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit FCCHUNK '1'
            edit FCLENGTH '768'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger /emc_36r1/00/legC/postproc/dp eq complete and /emc_36r1/00/legC/postproc/extra eq complete
              edit RUNVARFC '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              edit EMOS_TYPE 'pf'
              edit STEPS '366/to/768/by/6'
              family cf
                edit EMOS_TYPE 'cf'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family extra
              trigger /emc_36r1/hc2/back/07/legBcpl/postproc/extra eq complete
              edit RUNVARFC '1'
              task wm_sfc_arc
                label files ""
                label info ""
                event fdb
              task wm_ua_arc
                label files ""
                label info ""
                event fdb
            endfamily
            family nino
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit DAILY '0'
              edit RUNVARFC '1'
              edit FCLENGTH1 '360'
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit FCLENGTH '768'
              edit EPSCOUPLE '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              limit semaphore 10
              inlimit /limits:linux_cluster
              inlimit semaphore
              family 00
                edit OCGROUP 'an00'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '0'
                edit MEMBER '0'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 01
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '1'
                edit MEMBER '1'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 02
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '2'
                edit MEMBER '2'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 03
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '3'
                edit MEMBER '3'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 04
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '4'
                edit MEMBER '4'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
            endfamily
            task tcyctoecfs
              trigger nino eq complete
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit EPSCOUPLE '1'
              edit FCLENGTH '768'
              inlimit /limits:linux_cluster
          endfamily
        endfamily
        task save_logs
          trigger legA eq complete and legB eq complete and legC eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit SCHOST 'localhost'
      endfamily
      family 06
        trigger ( ../back/06 eq complete)
        edit DELTA_YEAR '-06'
        family legA
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf144
              edit STEPS '0/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/144/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf240
              edit STEPS '150/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf072
              edit STEPS '0/to/072/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/072/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf144
              edit STEPS '078/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf192
              edit STEPS '150/to/192/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf240
              edit STEPS '198/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
              task pfml
                edit STEPS '216'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit RUNVARFC '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family coupled
            trigger ./cf eq complete and ./pf eq complete
            edit EPSLEG '1'
            edit OCLEG '1'
            edit OCEPS '1'
            edit FCCHUNK '1'
            edit FCLENGTH '240'
            edit FCLENGTH0 '0'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '0'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legB
          edit EPSLEG '2'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            task cfml
              edit STEPS '216'
            family cf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task pfml
              edit STEPS '360'
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family coupled
            edit EPSLEG '2'
            edit OCLEG '2'
            edit OCEPS '2'
            edit FCCHUNK '1'
            edit FCLENGTH '360'
            edit FCLENGTH0 '216'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '1'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legC
          edit EPSLEG '3'
          edit FSFAMILY 'mars'
          edit OVERLAP '0'
          edit EMOS_TYPE 'an'
          edit EPSFCRES '255'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit FCCHUNK '1'
            edit FCLENGTH '768'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger /emc_36r1/00/legC/postproc/dp eq complete and /emc_36r1/00/legC/postproc/extra eq complete
              edit RUNVARFC '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              edit EMOS_TYPE 'pf'
              edit STEPS '366/to/768/by/6'
              family cf
                edit EMOS_TYPE 'cf'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family extra
              trigger /emc_36r1/hc2/back/06/legBcpl/postproc/extra eq complete
              edit RUNVARFC '1'
              task wm_sfc_arc
                label files ""
                label info ""
                event fdb
              task wm_ua_arc
                label files ""
                label info ""
                event fdb
            endfamily
            family nino
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit DAILY '0'
              edit RUNVARFC '1'
              edit FCLENGTH1 '360'
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit FCLENGTH '768'
              edit EPSCOUPLE '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              limit semaphore 10
              inlimit /limits:linux_cluster
              inlimit semaphore
              family 00
                edit OCGROUP 'an00'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '0'
                edit MEMBER '0'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 01
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '1'
                edit MEMBER '1'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 02
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '2'
                edit MEMBER '2'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 03
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '3'
                edit MEMBER '3'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 04
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '4'
                edit MEMBER '4'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
            endfamily
            task tcyctoecfs
              trigger nino eq complete
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit EPSCOUPLE '1'
              edit FCLENGTH '768'
              inlimit /limits:linux_cluster
          endfamily
        endfamily
        task save_logs
          trigger legA eq complete and legB eq complete and legC eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit SCHOST 'localhost'
      endfamily
      family 05
        trigger ( ../back/05 eq complete)
        edit DELTA_YEAR '-05'
        family legA
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf144
              edit STEPS '0/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/144/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf240
              edit STEPS '150/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf072
              edit STEPS '0/to/072/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/072/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf144
              edit STEPS '078/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf192
              edit STEPS '150/to/192/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf240
              edit STEPS '198/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
              task pfml
                edit STEPS '216'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit RUNVARFC '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family coupled
            trigger ./cf eq complete and ./pf eq complete
            edit EPSLEG '1'
            edit OCLEG '1'
            edit OCEPS '1'
            edit FCCHUNK '1'
            edit FCLENGTH '240'
            edit FCLENGTH0 '0'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '0'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legB
          edit EPSLEG '2'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            task cfml
              edit STEPS '216'
            family cf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task pfml
              edit STEPS '360'
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family coupled
            edit EPSLEG '2'
            edit OCLEG '2'
            edit OCEPS '2'
            edit FCCHUNK '1'
            edit FCLENGTH '360'
            edit FCLENGTH0 '216'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '1'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legC
          edit EPSLEG '3'
          edit FSFAMILY 'mars'
          edit OVERLAP '0'
          edit EMOS_TYPE 'an'
          edit EPSFCRES '255'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit FCCHUNK '1'
            edit FCLENGTH '768'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger /emc_36r1/00/legC/postproc/dp eq complete and /emc_36r1/00/legC/postproc/extra eq complete
              edit RUNVARFC '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              edit EMOS_TYPE 'pf'
              edit STEPS '366/to/768/by/6'
              family cf
                edit EMOS_TYPE 'cf'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family extra
              trigger /emc_36r1/hc2/back/05/legBcpl/postproc/extra eq complete
              edit RUNVARFC '1'
              task wm_sfc_arc
                label files ""
                label info ""
                event fdb
              task wm_ua_arc
                label files ""
                label info ""
                event fdb
            endfamily
            family nino
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit DAILY '0'
              edit RUNVARFC '1'
              edit FCLENGTH1 '360'
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit FCLENGTH '768'
              edit EPSCOUPLE '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              limit semaphore 10
              inlimit /limits:linux_cluster
              inlimit semaphore
              family 00
                edit OCGROUP 'an00'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '0'
                edit MEMBER '0'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 01
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '1'
                edit MEMBER '1'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 02
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '2'
                edit MEMBER '2'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 03
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '3'
                edit MEMBER '3'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 04
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '4'
                edit MEMBER '4'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
            endfamily
            task tcyctoecfs
              trigger nino eq complete
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit EPSCOUPLE '1'
              edit FCLENGTH '768'
              inlimit /limits:linux_cluster
          endfamily
        endfamily
        task save_logs
          trigger legA eq complete and legB eq complete and legC eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit SCHOST 'localhost'
      endfamily
      family 04
        trigger ( ../back/04 eq complete)
        edit DELTA_YEAR '-04'
        family legA
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf144
              edit STEPS '0/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/144/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf240
              edit STEPS '150/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf072
              edit STEPS '0/to/072/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/072/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf144
              edit STEPS '078/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf192
              edit STEPS '150/to/192/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf240
              edit STEPS '198/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
              task pfml
                edit STEPS '216'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit RUNVARFC '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family coupled
            trigger ./cf eq complete and ./pf eq complete
            edit EPSLEG '1'
            edit OCLEG '1'
            edit OCEPS '1'
            edit FCCHUNK '1'
            edit FCLENGTH '240'
            edit FCLENGTH0 '0'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '0'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legB
          edit EPSLEG '2'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            task cfml
              edit STEPS '216'
            family cf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task pfml
              edit STEPS '360'
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family coupled
            edit EPSLEG '2'
            edit OCLEG '2'
            edit OCEPS '2'
            edit FCCHUNK '1'
            edit FCLENGTH '360'
            edit FCLENGTH0 '216'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '1'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legC
          edit EPSLEG '3'
          edit FSFAMILY 'mars'
          edit OVERLAP '0'
          edit EMOS_TYPE 'an'
          edit EPSFCRES '255'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit FCCHUNK '1'
            edit FCLENGTH '768'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger /emc_36r1/00/legC/postproc/dp eq complete and /emc_36r1/00/legC/postproc/extra eq complete
              edit RUNVARFC '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              edit EMOS_TYPE 'pf'
              edit STEPS '366/to/768/by/6'
              family cf
                edit EMOS_TYPE 'cf'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family extra
              trigger /emc_36r1/hc2/back/04/legBcpl/postproc/extra eq complete
              edit RUNVARFC '1'
              task wm_sfc_arc
                label files ""
                label info ""
                event fdb
              task wm_ua_arc
                label files ""
                label info ""
                event fdb
            endfamily
            family nino
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit DAILY '0'
              edit RUNVARFC '1'
              edit FCLENGTH1 '360'
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit FCLENGTH '768'
              edit EPSCOUPLE '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              limit semaphore 10
              inlimit /limits:linux_cluster
              inlimit semaphore
              family 00
                edit OCGROUP 'an00'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '0'
                edit MEMBER '0'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 01
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '1'
                edit MEMBER '1'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 02
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '2'
                edit MEMBER '2'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 03
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '3'
                edit MEMBER '3'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 04
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '4'
                edit MEMBER '4'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
            endfamily
            task tcyctoecfs
              trigger nino eq complete
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit EPSCOUPLE '1'
              edit FCLENGTH '768'
              inlimit /limits:linux_cluster
          endfamily
        endfamily
        task save_logs
          trigger legA eq complete and legB eq complete and legC eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit SCHOST 'localhost'
      endfamily
      family 03
        trigger ( ../back/03 eq complete)
        edit DELTA_YEAR '-03'
        family legA
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf144
              edit STEPS '0/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/144/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf240
              edit STEPS '150/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf072
              edit STEPS '0/to/072/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/072/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf144
              edit STEPS '078/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf192
              edit STEPS '150/to/192/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf240
              edit STEPS '198/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
              task pfml
                edit STEPS '216'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit RUNVARFC '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family coupled
            trigger ./cf eq complete and ./pf eq complete
            edit EPSLEG '1'
            edit OCLEG '1'
            edit OCEPS '1'
            edit FCCHUNK '1'
            edit FCLENGTH '240'
            edit FCLENGTH0 '0'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '0'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legB
          edit EPSLEG '2'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            task cfml
              edit STEPS '216'
            family cf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task pfml
              edit STEPS '360'
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family coupled
            edit EPSLEG '2'
            edit OCLEG '2'
            edit OCEPS '2'
            edit FCCHUNK '1'
            edit FCLENGTH '360'
            edit FCLENGTH0 '216'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '1'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legC
          edit EPSLEG '3'
          edit FSFAMILY 'mars'
          edit OVERLAP '0'
          edit EMOS_TYPE 'an'
          edit EPSFCRES '255'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit FCCHUNK '1'
            edit FCLENGTH '768'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger /emc_36r1/00/legC/postproc/dp eq complete and /emc_36r1/00/legC/postproc/extra eq complete
              edit RUNVARFC '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              edit EMOS_TYPE 'pf'
              edit STEPS '366/to/768/by/6'
              family cf
                edit EMOS_TYPE 'cf'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family extra
              trigger /emc_36r1/hc2/back/03/legBcpl/postproc/extra eq complete
              edit RUNVARFC '1'
              task wm_sfc_arc
                label files ""
                label info ""
                event fdb
              task wm_ua_arc
                label files ""
                label info ""
                event fdb
            endfamily
            family nino
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit DAILY '0'
              edit RUNVARFC '1'
              edit FCLENGTH1 '360'
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit FCLENGTH '768'
              edit EPSCOUPLE '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              limit semaphore 10
              inlimit /limits:linux_cluster
              inlimit semaphore
              family 00
                edit OCGROUP 'an00'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '0'
                edit MEMBER '0'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 01
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '1'
                edit MEMBER '1'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 02
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '2'
                edit MEMBER '2'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 03
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '3'
                edit MEMBER '3'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 04
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '4'
                edit MEMBER '4'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
            endfamily
            task tcyctoecfs
              trigger nino eq complete
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit EPSCOUPLE '1'
              edit FCLENGTH '768'
              inlimit /limits:linux_cluster
          endfamily
        endfamily
        task save_logs
          trigger legA eq complete and legB eq complete and legC eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit SCHOST 'localhost'
      endfamily
      family 02
        trigger ( ../back/02 eq complete)
        edit DELTA_YEAR '-02'
        family legA
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf144
              edit STEPS '0/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/144/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf240
              edit STEPS '150/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf072
              edit STEPS '0/to/072/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/072/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf144
              edit STEPS '078/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf192
              edit STEPS '150/to/192/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf240
              edit STEPS '198/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
              task pfml
                edit STEPS '216'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit RUNVARFC '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family coupled
            trigger ./cf eq complete and ./pf eq complete
            edit EPSLEG '1'
            edit OCLEG '1'
            edit OCEPS '1'
            edit FCCHUNK '1'
            edit FCLENGTH '240'
            edit FCLENGTH0 '0'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '0'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legB
          edit EPSLEG '2'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            task cfml
              edit STEPS '216'
            family cf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task pfml
              edit STEPS '360'
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family coupled
            edit EPSLEG '2'
            edit OCLEG '2'
            edit OCEPS '2'
            edit FCCHUNK '1'
            edit FCLENGTH '360'
            edit FCLENGTH0 '216'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '1'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legC
          edit EPSLEG '3'
          edit FSFAMILY 'mars'
          edit OVERLAP '0'
          edit EMOS_TYPE 'an'
          edit EPSFCRES '255'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit FCCHUNK '1'
            edit FCLENGTH '768'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger /emc_36r1/00/legC/postproc/dp eq complete and /emc_36r1/00/legC/postproc/extra eq complete
              edit RUNVARFC '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              edit EMOS_TYPE 'pf'
              edit STEPS '366/to/768/by/6'
              family cf
                edit EMOS_TYPE 'cf'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family extra
              trigger /emc_36r1/hc2/back/02/legBcpl/postproc/extra eq complete
              edit RUNVARFC '1'
              task wm_sfc_arc
                label files ""
                label info ""
                event fdb
              task wm_ua_arc
                label files ""
                label info ""
                event fdb
            endfamily
            family nino
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit DAILY '0'
              edit RUNVARFC '1'
              edit FCLENGTH1 '360'
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit FCLENGTH '768'
              edit EPSCOUPLE '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              limit semaphore 10
              inlimit /limits:linux_cluster
              inlimit semaphore
              family 00
                edit OCGROUP 'an00'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '0'
                edit MEMBER '0'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 01
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '1'
                edit MEMBER '1'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 02
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '2'
                edit MEMBER '2'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 03
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '3'
                edit MEMBER '3'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 04
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '4'
                edit MEMBER '4'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
            endfamily
            task tcyctoecfs
              trigger nino eq complete
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit EPSCOUPLE '1'
              edit FCLENGTH '768'
              inlimit /limits:linux_cluster
          endfamily
        endfamily
        task save_logs
          trigger legA eq complete and legB eq complete and legC eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit SCHOST 'localhost'
      endfamily
      family 01
        trigger ( ../back/01 eq complete)
        edit DELTA_YEAR '-01'
        family legA
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf144
              edit STEPS '0/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/144/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf240
              edit STEPS '150/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf072
              edit STEPS '0/to/072/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/072/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf144
              edit STEPS '078/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf192
              edit STEPS '150/to/192/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf240
              edit STEPS '198/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
              task pfml
                edit STEPS '216'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit RUNVARFC '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family coupled
            trigger ./cf eq complete and ./pf eq complete
            edit EPSLEG '1'
            edit OCLEG '1'
            edit OCEPS '1'
            edit FCCHUNK '1'
            edit FCLENGTH '240'
            edit FCLENGTH0 '0'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '0'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legB
          edit EPSLEG '2'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            task cfml
              edit STEPS '216'
            family cf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task pfml
              edit STEPS '360'
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family coupled
            edit EPSLEG '2'
            edit OCLEG '2'
            edit OCEPS '2'
            edit FCCHUNK '1'
            edit FCLENGTH '360'
            edit FCLENGTH0 '216'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '1'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legC
          edit EPSLEG '3'
          edit FSFAMILY 'mars'
          edit OVERLAP '0'
          edit EMOS_TYPE 'an'
          edit EPSFCRES '255'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit FCCHUNK '1'
            edit FCLENGTH '768'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger /emc_36r1/00/legC/postproc/dp eq complete and /emc_36r1/00/legC/postproc/extra eq complete
              edit RUNVARFC '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              edit EMOS_TYPE 'pf'
              edit STEPS '366/to/768/by/6'
              family cf
                edit EMOS_TYPE 'cf'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family extra
              trigger /emc_36r1/hc2/back/01/legBcpl/postproc/extra eq complete
              edit RUNVARFC '1'
              task wm_sfc_arc
                label files ""
                label info ""
                event fdb
              task wm_ua_arc
                label files ""
                label info ""
                event fdb
            endfamily
            family nino
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit DAILY '0'
              edit RUNVARFC '1'
              edit FCLENGTH1 '360'
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit FCLENGTH '768'
              edit EPSCOUPLE '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              limit semaphore 10
              inlimit /limits:linux_cluster
              inlimit semaphore
              family 00
                edit OCGROUP 'an00'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '0'
                edit MEMBER '0'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 01
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '1'
                edit MEMBER '1'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 02
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '2'
                edit MEMBER '2'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 03
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '3'
                edit MEMBER '3'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 04
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '4'
                edit MEMBER '4'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
            endfamily
            task tcyctoecfs
              trigger nino eq complete
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit EPSCOUPLE '1'
              edit FCLENGTH '768'
              inlimit /limits:linux_cluster
          endfamily
        endfamily
        task save_logs
          trigger legA eq complete and legB eq complete and legC eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit SCHOST 'localhost'
      endfamily
    endfamily
    family postproc
      trigger ./back eq complete and ( ../hc2:YMD gt ../hc0:YMD or ( ../hc2:YMD eq ../hc0:YMD and ../hc2/back eq complete))
      edit EMOS_TYPE '00'
      edit FCCHUNK '1'
      edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/leg'
      inlimit /emc_36r1/limits:postproc
      task stage_hc
        edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
      family legA
        trigger ./stage_hc eq complete
        edit FCLENGTH0 '0'
        edit FCLENGTH '240'
        edit EPSFCRES '399'
        edit FCCHUNK '1'
        edit STEP1 '0'
        edit STEP2 '240'
        task ensm
        family reord
          edit EPSCOUPLE '1'
          edit EMOS_TYPE 'pf'
          edit FCLENGTH0 '0'
          edit FCLENGTH '240'
          edit EPSFCRES '399'
          edit FCCHUNK '1'
          edit STEP1 '0'
          edit STEP2 '240'
          edit STEPS '-1'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/leg'
          edit RUNVARFC '1'
          limit reord 1
          inlimit reord
          label info ""
          family 130
            edit GENERATE '1'
            edit ARCHIVE '0'
            edit PARAMS '130.128'
            task reord
              label info ""
          endfamily
          family 228
            edit GENERATE '1'
            edit ARCHIVE '0'
            edit PARAMS '228.128'
            task reord
              label info ""
          endfamily
          family 207
            edit GENERATE '1'
            edit ARCHIVE '0'
            edit PARAMS '207.128'
            task reord
              label info ""
          endfamily
          family 167
            edit GENERATE '1'
            edit ARCHIVE '0'
            edit PARAMS '167.128'
            task reord
              label info ""
          endfamily
          family 164
            edit GENERATE '1'
            edit ARCHIVE '0'
            edit PARAMS '164.128'
            task reord
              label info ""
          endfamily
          family arch
            trigger ./130 eq complete and ./228 eq complete and ./207 eq complete and ./167 eq complete and ./164 eq complete
            edit GENERATE '0'
            edit ARCHIVE '1'
            inlimit /emc_36r1/lag/ref/00:mutex
            task reord
              label info ""
          endfamily
        endfamily
      endfamily
      family legB
        trigger legA eq complete
        edit FCLENGTH0 '240'
        edit FCLENGTH '768'
        edit EPSFCRES '255'
        edit FCCHUNK '1'
        edit STEP1 '240'
        edit STEP2 '768'
        task ensm
        family reord
          edit EPSCOUPLE '1'
          edit EMOS_TYPE 'pf'
          edit FCLENGTH0 '240'
          edit FCLENGTH '768'
          edit EPSFCRES '255'
          edit FCCHUNK '1'
          edit STEP1 '240'
          edit STEP2 '768'
          edit STEPS '-1'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/leg'
          edit RUNVARFC '1'
          limit reord 1
          inlimit reord
          label info ""
          family 130
            edit GENERATE '1'
            edit ARCHIVE '0'
            edit PARAMS '130.128'
            task reord
              label info ""
          endfamily
          family 228
            edit GENERATE '1'
            edit ARCHIVE '0'
            edit PARAMS '228.128'
            task reord
              label info ""
          endfamily
          family 207
            edit GENERATE '1'
            edit ARCHIVE '0'
            edit PARAMS '207.128'
            task reord
              label info ""
          endfamily
          family 167
            edit GENERATE '1'
            edit ARCHIVE '0'
            edit PARAMS '167.128'
            task reord
              label info ""
          endfamily
          family 164
            edit GENERATE '1'
            edit ARCHIVE '0'
            edit PARAMS '164.128'
            task reord
              label info ""
          endfamily
          family arch
            trigger ./130 eq complete and ./228 eq complete and ./207 eq complete and ./167 eq complete and ./164 eq complete
            edit GENERATE '0'
            edit ARCHIVE '1'
            inlimit /emc_36r1/lag/ref/00:mutex
            task reord
              label info ""
          endfamily
        endfamily
      endfamily
      task wmem_veps
        trigger legA eq complete and legB eq complete
        edit EMOS_TYPE 'UNDEF'
        edit FCCHUNK '1'
        edit FCLENGTH '768'
      task wmem_archive
        trigger wmem_veps eq complete and ./legA eq complete and ./legB eq complete
        edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
        edit FCLENGTH '768'
        label info ""
      family clim
        trigger ../arch eq complete and ./stage_hc eq complete
        edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/leg/clim'
        edit CLIM_FC_YEARS '0'
        edit OCSUITE 'vareps'
        edit ERA40 'YES'
        edit START_DAY '20091008'
        edit END_DAY '20121213'
        edit RUNVARFC '1'
        edit QUEUE 'emos-e'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        inlimit /limits:linux_cluster
        family init
          task listclim
          task getdata
            trigger listclim eq complete
          task getobsclim
            trigger getdata eq complete
        endfamily
        family bias
          trigger init eq complete
          task cal_bias
          task plot_bias
            trigger cal_bias eq complete
          task plot_clim
            trigger plot_bias eq complete
          task cal_zondia
            trigger plot_clim eq complete
          task plot_zondia
            trigger cal_zondia eq complete
          task migrate_bias
            trigger plot_zondia eq complete
          task clean_bias
            trigger migrate_bias eq complete
          task clean
            trigger clean_bias eq complete
        endfamily
      endfamily
    endfamily
    family prod
      trigger ./back eq complete and ./postproc eq complete
      edit SMSFILES '/home/ma/emos/def/mc/prod'
      edit EMOS_DOMAIN 'G'
      edit EMOS_TYPE 'FC'
      edit PROD_TASK '1'
      edit EMOS_PROJECT '00'
      edit EMOS_BASE '00'
      edit PGNPES '32'
      edit EMOS_STREAM 'ENFH'
      family getreq
        trigger ../back eq complete
        edit EMOS_TYPE 'cm'
        edit EMOS_DOMAIN 'G'
        edit EMOS_PROJECT '12'
        edit PROD_TASK '1'
        edit SMSTRIES '1'
        edit EMOS_STREAM 'ENFH'
        task collectreq
          edit QUEUE 'diss'
          edit USER 'emos'
          edit WSHOST 'diss'
          edit QUEUE_EPILOG 'emos-exp'
          edit LOGDIR '/diss/log'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit SMSOUT '/vol/emos_nc/output'
          edit EMOS_BASE '00'
        task eps
          trigger collectreq eq complete
          label info ""
          event 1 
        task collectwave
          edit QUEUE 'diss'
          edit USER 'emos'
          edit WSHOST 'diss'
          edit QUEUE_EPILOG 'emos-exp'
          edit LOGDIR '/diss/log'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit SMSOUT '/vol/emos_nc/output'
          edit EMOS_STREAM 'ENWH'
        task wave_eps
          trigger collectwave eq complete
          edit EMOS_STREAM 'ENWH'
          label info ""
          event 1 
      endfamily
      family eps
        trigger ../back eq complete and ( ./getreq/eps eq complete or ./getreq/eps:001)
        edit EMOS_STREAM 'ENFH'
        family day01
          edit EMOS_TIME_STEP_H '0001'
          task prodgen
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ./pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day02
          edit EMOS_TIME_STEP_H '0002'
          task prodgen
            trigger ../day01/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day01/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day03
          edit EMOS_TIME_STEP_H '0003'
          task prodgen
            trigger ../day02/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day02/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day04
          edit EMOS_TIME_STEP_H '0004'
          task prodgen
            trigger ../day03/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day03/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day05
          edit EMOS_TIME_STEP_H '0005'
          task prodgen
            trigger ../day04/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day04/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day06
          edit EMOS_TIME_STEP_H '0006'
          task prodgen
            trigger ../day05/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day05/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day07
          edit EMOS_TIME_STEP_H '0007'
          task prodgen
            trigger ../day06/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day06/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day08
          edit EMOS_TIME_STEP_H '0008'
          task prodgen
            trigger ../day07/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day07/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day09
          edit EMOS_TIME_STEP_H '0009'
          task prodgen
            trigger ../day08/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day08/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day10
          edit EMOS_TIME_STEP_H '0010'
          task prodgen
            trigger ../day09/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day09/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day11
          edit EMOS_TIME_STEP_H '0011'
          task prodgen
            trigger ../day10/prodgen eq complete
            edit MEM '1900'
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day10/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day12
          edit EMOS_TIME_STEP_H '0012'
          task prodgen
            trigger ../day11/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day11/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day13
          edit EMOS_TIME_STEP_H '0013'
          task prodgen
            trigger ../day12/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day12/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day14
          edit EMOS_TIME_STEP_H '0014'
          task prodgen
            trigger ../day13/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day13/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day15
          edit EMOS_TIME_STEP_H '0015'
          task prodgen
            trigger ../day14/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day14/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day16
          edit EMOS_TIME_STEP_H '0016'
          task prodgen
            trigger ../day15/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day15/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day17
          edit EMOS_TIME_STEP_H '0017'
          task prodgen
            trigger ../day16/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day16/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day18
          edit EMOS_TIME_STEP_H '0018'
          task prodgen
            trigger ../day17/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day17/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day19
          edit EMOS_TIME_STEP_H '0019'
          task prodgen
            trigger ../day18/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day18/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day20
          edit EMOS_TIME_STEP_H '0020'
          task prodgen
            trigger ../day19/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day19/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day21
          edit EMOS_TIME_STEP_H '0021'
          task prodgen
            trigger ../day20/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day20/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day22
          edit EMOS_TIME_STEP_H '0022'
          task prodgen
            trigger ../day21/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day21/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day23
          edit EMOS_TIME_STEP_H '0023'
          task prodgen
            trigger ../day22/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day22/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day24
          edit EMOS_TIME_STEP_H '0024'
          task prodgen
            trigger ../day23/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day23/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day25
          edit EMOS_TIME_STEP_H '0025'
          task prodgen
            trigger ../day24/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day24/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day26
          edit EMOS_TIME_STEP_H '0026'
          task prodgen
            trigger ../day25/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day25/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day27
          edit EMOS_TIME_STEP_H '0027'
          task prodgen
            trigger ../day26/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day26/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day28
          edit EMOS_TIME_STEP_H '0028'
          task prodgen
            trigger ../day27/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day27/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day29
          edit EMOS_TIME_STEP_H '0029'
          task prodgen
            trigger ../day28/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day28/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day30
          edit EMOS_TIME_STEP_H '0030'
          task prodgen
            trigger ../day29/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day29/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day31
          edit EMOS_TIME_STEP_H '0031'
          task prodgen
            trigger ../day30/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day30/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day32
          edit EMOS_TIME_STEP_H '0032'
          task prodgen
            trigger ../day31/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day31/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day33
          edit EMOS_TIME_STEP_H '0033'
          task prodgen
            trigger ../day32/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day32/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
      endfamily
      family wave
        trigger ../back eq complete and ( ./getreq/wave_eps eq complete or ./getreq/eps:001)
        edit PGNPES '4'
        edit EMOS_STREAM 'ENWH'
        family day01
          edit EMOS_TIME_STEP_H '0001'
          task prodgen
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ./pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day02
          edit EMOS_TIME_STEP_H '0002'
          task prodgen
            trigger ../day01/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day01/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day03
          edit EMOS_TIME_STEP_H '0003'
          task prodgen
            trigger ../day02/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day02/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day04
          edit EMOS_TIME_STEP_H '0004'
          task prodgen
            trigger ../day03/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day03/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day05
          edit EMOS_TIME_STEP_H '0005'
          task prodgen
            trigger ../day04/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day04/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day06
          edit EMOS_TIME_STEP_H '0006'
          task prodgen
            trigger ../day05/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day05/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day07
          edit EMOS_TIME_STEP_H '0007'
          task prodgen
            trigger ../day06/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day06/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day08
          edit EMOS_TIME_STEP_H '0008'
          task prodgen
            trigger ../day07/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day07/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day09
          edit EMOS_TIME_STEP_H '0009'
          task prodgen
            trigger ../day08/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day08/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day10
          edit EMOS_TIME_STEP_H '0010'
          task prodgen
            trigger ../day09/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day09/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day11
          edit EMOS_TIME_STEP_H '0011'
          task prodgen
            trigger ../day10/prodgen eq complete
            edit MEM '1900'
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day10/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day12
          edit EMOS_TIME_STEP_H '0012'
          task prodgen
            trigger ../day11/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day11/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day13
          edit EMOS_TIME_STEP_H '0013'
          task prodgen
            trigger ../day12/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day12/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day14
          edit EMOS_TIME_STEP_H '0014'
          task prodgen
            trigger ../day13/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day13/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day15
          edit EMOS_TIME_STEP_H '0015'
          task prodgen
            trigger ../day14/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day14/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day16
          edit EMOS_TIME_STEP_H '0016'
          task prodgen
            trigger ../day15/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day15/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day17
          edit EMOS_TIME_STEP_H '0017'
          task prodgen
            trigger ../day16/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day16/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day18
          edit EMOS_TIME_STEP_H '0018'
          task prodgen
            trigger ../day17/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day17/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day19
          edit EMOS_TIME_STEP_H '0019'
          task prodgen
            trigger ../day18/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day18/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day20
          edit EMOS_TIME_STEP_H '0020'
          task prodgen
            trigger ../day19/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day19/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day21
          edit EMOS_TIME_STEP_H '0021'
          task prodgen
            trigger ../day20/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day20/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day22
          edit EMOS_TIME_STEP_H '0022'
          task prodgen
            trigger ../day21/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day21/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day23
          edit EMOS_TIME_STEP_H '0023'
          task prodgen
            trigger ../day22/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day22/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day24
          edit EMOS_TIME_STEP_H '0024'
          task prodgen
            trigger ../day23/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day23/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day25
          edit EMOS_TIME_STEP_H '0025'
          task prodgen
            trigger ../day24/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day24/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day26
          edit EMOS_TIME_STEP_H '0026'
          task prodgen
            trigger ../day25/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day25/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day27
          edit EMOS_TIME_STEP_H '0027'
          task prodgen
            trigger ../day26/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day26/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day28
          edit EMOS_TIME_STEP_H '0028'
          task prodgen
            trigger ../day27/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day27/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day29
          edit EMOS_TIME_STEP_H '0029'
          task prodgen
            trigger ../day28/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day28/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day30
          edit EMOS_TIME_STEP_H '0030'
          task prodgen
            trigger ../day29/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day29/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day31
          edit EMOS_TIME_STEP_H '0031'
          task prodgen
            trigger ../day30/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day30/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day32
          edit EMOS_TIME_STEP_H '0032'
          task prodgen
            trigger ../day31/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day31/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
        family day33
          edit EMOS_TIME_STEP_H '0033'
          task prodgen
            trigger ../day32/prodgen eq complete
          task pdb
            trigger prodgen eq complete
            inlimit /emc_36r1/limits:pdb
            inlimit /limits:pdb
            label info ""
            event 1 
          task q2diss
            late -s +00:05 -c +00:15
            trigger ../day32/q2diss eq complete and pdb eq complete
            edit USER_PRIORITY '93'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event 1 
        endfamily
      endfamily
    endfamily
    family adm
      task hindsync
        trigger ../back eq complete and ../arch eq complete and ../postproc eq complete
        edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/efi_hdc'
        edit NH_WEEKS_KEPT '3'
        edit BACKUP_HCFDB '0'
      family fri
        complete ./fri/compl eq complete and ./fri/dummy eq queued
        edit SMSFILES '/home/ma/emos/def/emc_36r1/ref'
        task dummy
          cron -w 5 09:00
        task compl
          complete dummy eq active or dummy eq submitted or dummy eq complete
          trigger 1 eq 0
      endfamily
    endfamily
    task logfilestore
      edit QUEUE 'emos-e'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit KEEPLOGS 'no'
      edit LOGSTOECFS 'no'
      edit MOVE_TO_ECFS 'yes'
    task logfiles
      late -s +00:05 -c +00:25
      trigger logfilestore eq complete and back eq complete and arch eq complete and postproc eq complete and prod eq complete and adm eq complete
      edit QUEUE 'emos-e'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit MOVE_TO_ECFS 'no'
  endfamily
  family 00
    repeat date YMD 20091007 20121212 7
    edit DELTA_DAY '1'
    edit ACCOUNT 'oeenfo'
    edit LOGTASK '00'
    edit EMOS_BASE '00'
    edit QUEUE 'os'
    family ref
      edit SMSFILES '/home/ma/emos/def/emc_36r1/ref'
      family thu1000
        complete ./thu1000/compl eq complete and ./thu1000/dummy eq queued
        task dummy
        task compl
          complete ./dummy eq active or ./dummy eq submitted
          trigger 1 eq 0
      endfamily
      family thu2200
        complete ./thu2200/compl eq complete and ./thu2200/dummy eq queued
        task dummy
        task compl
          complete ./dummy eq active or ./dummy eq submitted
          trigger 1 eq 0
      endfamily
      family wed
        complete ./wed/compl eq complete and ./wed/dummy eq queued
        task dummy
          cron -w 3 18:30
        task compl
          complete ./dummy eq active or ./dummy eq submitted
          trigger 1 eq 0
      endfamily
      family tue
        complete ./tue/compl eq complete and ./tue/dummy eq queued
        task dummy
          cron -w 2 08:00
        task compl
          complete ./dummy eq active or ./dummy eq submitted
          trigger 1 eq 0
      endfamily
    endfamily
    family legC
      trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legA/fc/pf eq complete and /emc_36r1/main/00/legA/fc/cf eq complete and /emc_36r1/main/make eq complete)
      edit EPSCOUPLE '1'
      edit RUNCOUPLE '1'
      edit EPSLEG '3'
      edit EPSFCLENGTH_C '408'
      family fc
        edit FCMEM '2240'
        edit FCSTACK '700'
        edit FCMRFS '0'
        edit PRIORITY '69'
        edit USER_PRIORITY '69'
        edit N_PERT_AN '10'
        edit EPSWAMNFRE '25'
        edit EPSWAMNANG '12'
        edit TSTEP '2700'
        edit EPSFCRES '639'
        edit RESOL '639'
        edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
        edit CPUTIME '1000'
        edit MEM '960'
        edit THREADS '8'
        edit NPES '8'
        edit EPSLEG '3'
        family pf
          complete /emc_36r1/main/ref/00/events:cf
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
          edit EPSTYPE 'pf'
          edit EMOS_TYPE 'pf'
          edit USER_PRIORITY '69'
          edit PRIORITY '69'
          edit FCNPES '32'
          edit EPSNLEGS '3'
          edit EPSFCLEV '62'
          edit EPSWAMNSTPW '1'
          edit EPSINISTEP_PREVIOUS '0'
          edit EPSWAMRESOL 'global50'
          edit EPSFCRES '639'
          edit EPSWAMNFRE '30'
          edit EPSWAMNANG '24'
          edit FCLENGTH '240'
          edit TSTEP '1200'
          edit EPSLEG '3'
          edit CLASS 'od'
          edit THREADS '2'
          edit RUNVARFC '1'
          inlimit /emc_36r1/limits:pftotal
          meter step -1 408 408
          family 01
            edit EPSMEMBER '1'
            edit MEMBER '1'
            edit SUBFSFAMILY '/pf001'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/01 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 02
            edit EPSMEMBER '2'
            edit MEMBER '2'
            edit SUBFSFAMILY '/pf002'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/02 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 03
            edit EPSMEMBER '3'
            edit MEMBER '3'
            edit SUBFSFAMILY '/pf003'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/03 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 04
            edit EPSMEMBER '4'
            edit MEMBER '4'
            edit SUBFSFAMILY '/pf004'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/04 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 05
            edit EPSMEMBER '5'
            edit MEMBER '5'
            edit SUBFSFAMILY '/pf005'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/05 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 06
            edit EPSMEMBER '6'
            edit MEMBER '6'
            edit SUBFSFAMILY '/pf006'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/06 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 07
            edit EPSMEMBER '7'
            edit MEMBER '7'
            edit SUBFSFAMILY '/pf007'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/07 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 08
            edit EPSMEMBER '8'
            edit MEMBER '8'
            edit SUBFSFAMILY '/pf008'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/08 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 09
            edit EPSMEMBER '9'
            edit MEMBER '9'
            edit SUBFSFAMILY '/pf009'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/09 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 10
            edit EPSMEMBER '10'
            edit MEMBER '10'
            edit SUBFSFAMILY '/pf010'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/10 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 11
            edit EPSMEMBER '11'
            edit MEMBER '11'
            edit SUBFSFAMILY '/pf011'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/11 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 12
            edit EPSMEMBER '12'
            edit MEMBER '12'
            edit SUBFSFAMILY '/pf012'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/12 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 13
            edit EPSMEMBER '13'
            edit MEMBER '13'
            edit SUBFSFAMILY '/pf013'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/13 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 14
            edit EPSMEMBER '14'
            edit MEMBER '14'
            edit SUBFSFAMILY '/pf014'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/14 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 15
            edit EPSMEMBER '15'
            edit MEMBER '15'
            edit SUBFSFAMILY '/pf015'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/15 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 16
            edit EPSMEMBER '16'
            edit MEMBER '16'
            edit SUBFSFAMILY '/pf016'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/16 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 17
            edit EPSMEMBER '17'
            edit MEMBER '17'
            edit SUBFSFAMILY '/pf017'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/17 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 18
            edit EPSMEMBER '18'
            edit MEMBER '18'
            edit SUBFSFAMILY '/pf018'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/18 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 19
            edit EPSMEMBER '19'
            edit MEMBER '19'
            edit SUBFSFAMILY '/pf019'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/19 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 20
            edit EPSMEMBER '20'
            edit MEMBER '20'
            edit SUBFSFAMILY '/pf020'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/20 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 21
            edit EPSMEMBER '21'
            edit MEMBER '21'
            edit SUBFSFAMILY '/pf021'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/21 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 22
            edit EPSMEMBER '22'
            edit MEMBER '22'
            edit SUBFSFAMILY '/pf022'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/22 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 23
            edit EPSMEMBER '23'
            edit MEMBER '23'
            edit SUBFSFAMILY '/pf023'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/23 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 24
            edit EPSMEMBER '24'
            edit MEMBER '24'
            edit SUBFSFAMILY '/pf024'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/24 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 25
            edit EPSMEMBER '25'
            edit MEMBER '25'
            edit SUBFSFAMILY '/pf025'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/25 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 26
            edit EPSMEMBER '26'
            edit MEMBER '26'
            edit SUBFSFAMILY '/pf026'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/26 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 27
            edit EPSMEMBER '27'
            edit MEMBER '27'
            edit SUBFSFAMILY '/pf027'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/27 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 28
            edit EPSMEMBER '28'
            edit MEMBER '28'
            edit SUBFSFAMILY '/pf028'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/28 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 29
            edit EPSMEMBER '29'
            edit MEMBER '29'
            edit SUBFSFAMILY '/pf029'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/29 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 30
            edit EPSMEMBER '30'
            edit MEMBER '30'
            edit SUBFSFAMILY '/pf030'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/30 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 31
            edit EPSMEMBER '31'
            edit MEMBER '31'
            edit SUBFSFAMILY '/pf031'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/31 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 32
            edit EPSMEMBER '32'
            edit MEMBER '32'
            edit SUBFSFAMILY '/pf032'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/32 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 33
            edit EPSMEMBER '33'
            edit MEMBER '33'
            edit SUBFSFAMILY '/pf033'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/33 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 34
            edit EPSMEMBER '34'
            edit MEMBER '34'
            edit SUBFSFAMILY '/pf034'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/34 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 35
            edit EPSMEMBER '35'
            edit MEMBER '35'
            edit SUBFSFAMILY '/pf035'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/35 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 36
            edit EPSMEMBER '36'
            edit MEMBER '36'
            edit SUBFSFAMILY '/pf036'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/36 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 37
            edit EPSMEMBER '37'
            edit MEMBER '37'
            edit SUBFSFAMILY '/pf037'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/37 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 38
            edit EPSMEMBER '38'
            edit MEMBER '38'
            edit SUBFSFAMILY '/pf038'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/38 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 39
            edit EPSMEMBER '39'
            edit MEMBER '39'
            edit SUBFSFAMILY '/pf039'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/39 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 40
            edit EPSMEMBER '40'
            edit MEMBER '40'
            edit SUBFSFAMILY '/pf040'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/40 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 41
            edit EPSMEMBER '41'
            edit MEMBER '41'
            edit SUBFSFAMILY '/pf041'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/41 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 42
            edit EPSMEMBER '42'
            edit MEMBER '42'
            edit SUBFSFAMILY '/pf042'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/42 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 43
            edit EPSMEMBER '43'
            edit MEMBER '43'
            edit SUBFSFAMILY '/pf043'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/43 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 44
            edit EPSMEMBER '44'
            edit MEMBER '44'
            edit SUBFSFAMILY '/pf044'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/44 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 45
            edit EPSMEMBER '45'
            edit MEMBER '45'
            edit SUBFSFAMILY '/pf045'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/45 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 46
            edit EPSMEMBER '46'
            edit MEMBER '46'
            edit SUBFSFAMILY '/pf046'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/46 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 47
            edit EPSMEMBER '47'
            edit MEMBER '47'
            edit SUBFSFAMILY '/pf047'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/47 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 48
            edit EPSMEMBER '48'
            edit MEMBER '48'
            edit SUBFSFAMILY '/pf048'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/48 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 49
            edit EPSMEMBER '49'
            edit MEMBER '49'
            edit SUBFSFAMILY '/pf049'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/49 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
          family 50
            edit EPSMEMBER '50'
            edit MEMBER '50'
            edit SUBFSFAMILY '/pf050'
            task getvarepsdata
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/pf/50 eq complete)
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSFCRESN '319'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit THREADS '8'
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '1200'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '360'
              edit EPSFCRES '319'
              edit EPSLEG '3'
              edit NPES '1'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '3'
              edit EPSFCRES '319'
              edit EPSINISTEP '408'
              edit EPSINISTEP_PREVIOUS '360'
              edit FCLENGTH '408'
              edit MEM '1600'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit EPSFCRESN '255'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNSTPW '1'
              edit EPSFCLENGTH_B '360'
              edit TSTEP '2700'
              edit EPSFCLENGTH_A '240'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 408 408
          endfamily
        endfamily
        family cf
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
          edit MEMBER '0'
          edit EPSMEMBER '0'
          edit SUBFSFAMILY '/cf000'
          edit EPSTYPE 'cf'
          edit EMOS_TYPE 'cf'
          edit USER_PRIORITY '70'
          edit EPSNLEGS '3'
          edit EPSFCLEV '62'
          edit EPSWAMNSTPW '1'
          edit EPSINISTEP_PREVIOUS '0'
          edit EPSWAMRESOL 'global50'
          edit EPSFCRES '639'
          edit EPSWAMNFRE '30'
          edit EPSWAMNANG '24'
          edit TSTEP '1200'
          edit EPSLEG '3'
          edit CLASS 'od'
          edit OCGROUP 'an0'
          edit FCGROUP 'fc0'
          edit FCCHUNK '1'
          edit OCTOTAL '5'
          edit FCTOTAL '51'
          edit DAILY '0'
          edit RUNVARFC '1'
          edit FCLENGTH '408'
          edit FCLENGTH1 '360'
          task getvarepsdata
            trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legBcpl/fc/cf eq complete)
            edit EPSNLEGS '3'
            edit EPSFCLEV '62'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMRESOL 'global50'
            edit TSTEP '1200'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit EPSLEG '3'
            edit EPSINISTEP '360'
            edit EPSFCRES '319'
            edit EPSFCRESN '319'
            edit EPSWAMNSTPW '1'
            edit EPSFCLENGTH_B '360'
            edit CLASS 'od'
          task intHtoL
            trigger getvarepsdata eq complete
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
            edit EPSNLEGS '3'
            edit EPSFCLEV '62'
            edit EPSWAMNSTPW '1'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMRESOL 'global50'
            edit TSTEP '1200'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit EPSINISTEP '360'
            edit EPSFCRES '319'
            edit EPSLEG '3'
            edit NPES '1'
            edit CLASS 'od'
          task modeleps
            trigger ./intHtoL eq complete and ./getvarepsdata eq complete
            edit EPSNLEGS '3'
            edit EPSFCLEV '62'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit EPSLEG '3'
            edit EPSFCRES '319'
            edit EPSINISTEP '408'
            edit EPSINISTEP_PREVIOUS '360'
            edit FCLENGTH '408'
            edit MEM '1600'
            edit MRF '10'
            edit STACK '350'
            edit FCCHUNK '1'
            edit EPSFCRESN '255'
            edit EPSWAMRESOL 'global50'
            edit EPSWAMNSTPW '1'
            edit EPSFCLENGTH_B '360'
            edit TSTEP '2700'
            edit EPSFCLENGTH_A '240'
            edit SMSTRIES '1'
            edit NPES '1'
            edit THREADS '64'
            edit FCNPES '1'
            edit CLASS 'od'
            inlimit /emc_36r1/limits:modelepsB
            meter step -1 408 408
        endfamily
      endfamily
      family postproc
        trigger ./fc eq active or ./fc eq complete
        edit STEPS '372/768/12/24 768/768/12/120'
        edit FCCHUNK '1'
        edit OCTOTAL '5'
        edit FCTOTAL '51'
        edit DAILY '0'
        edit RUNVARFC '1'
        edit FCLENGTH1 '360'
        edit FCLENGTH '768'
        edit FCGROUP 'fc0'
        inlimit /emc_36r1/limits:postproc
        family dp
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
          edit STEPS '366:768:6'
          inlimit /emc_36r1/limits:dp
          family cf
            trigger ../../fc/cf eq complete
            edit EMOS_TYPE 'cf'
            task derived_param
              edit MEMBER 'off'
          endfamily
          family pf
            complete /emc_36r1/main/ref/00/events:cf
            edit EMOS_TYPE 'pf'
            family 01
              trigger ../../../fc/pf/01 eq complete
              edit EPSMEMBER '1'
              edit MEMBER '1'
              edit SUBFSFAMILY '/pf001'
              task derived_param
            endfamily
            family 02
              trigger ../../../fc/pf/02 eq complete
              edit EPSMEMBER '2'
              edit MEMBER '2'
              edit SUBFSFAMILY '/pf002'
              task derived_param
            endfamily
            family 03
              trigger ../../../fc/pf/03 eq complete
              edit EPSMEMBER '3'
              edit MEMBER '3'
              edit SUBFSFAMILY '/pf003'
              task derived_param
            endfamily
            family 04
              trigger ../../../fc/pf/04 eq complete
              edit EPSMEMBER '4'
              edit MEMBER '4'
              edit SUBFSFAMILY '/pf004'
              task derived_param
            endfamily
            family 05
              trigger ../../../fc/pf/05 eq complete
              edit EPSMEMBER '5'
              edit MEMBER '5'
              edit SUBFSFAMILY '/pf005'
              task derived_param
            endfamily
            family 06
              trigger ../../../fc/pf/06 eq complete
              edit EPSMEMBER '6'
              edit MEMBER '6'
              edit SUBFSFAMILY '/pf006'
              task derived_param
            endfamily
            family 07
              trigger ../../../fc/pf/07 eq complete
              edit EPSMEMBER '7'
              edit MEMBER '7'
              edit SUBFSFAMILY '/pf007'
              task derived_param
            endfamily
            family 08
              trigger ../../../fc/pf/08 eq complete
              edit EPSMEMBER '8'
              edit MEMBER '8'
              edit SUBFSFAMILY '/pf008'
              task derived_param
            endfamily
            family 09
              trigger ../../../fc/pf/09 eq complete
              edit EPSMEMBER '9'
              edit MEMBER '9'
              edit SUBFSFAMILY '/pf009'
              task derived_param
            endfamily
            family 10
              trigger ../../../fc/pf/10 eq complete
              edit EPSMEMBER '10'
              edit MEMBER '10'
              edit SUBFSFAMILY '/pf010'
              task derived_param
            endfamily
            family 11
              trigger ../../../fc/pf/11 eq complete
              edit EPSMEMBER '11'
              edit MEMBER '11'
              edit SUBFSFAMILY '/pf011'
              task derived_param
            endfamily
            family 12
              trigger ../../../fc/pf/12 eq complete
              edit EPSMEMBER '12'
              edit MEMBER '12'
              edit SUBFSFAMILY '/pf012'
              task derived_param
            endfamily
            family 13
              trigger ../../../fc/pf/13 eq complete
              edit EPSMEMBER '13'
              edit MEMBER '13'
              edit SUBFSFAMILY '/pf013'
              task derived_param
            endfamily
            family 14
              trigger ../../../fc/pf/14 eq complete
              edit EPSMEMBER '14'
              edit MEMBER '14'
              edit SUBFSFAMILY '/pf014'
              task derived_param
            endfamily
            family 15
              trigger ../../../fc/pf/15 eq complete
              edit EPSMEMBER '15'
              edit MEMBER '15'
              edit SUBFSFAMILY '/pf015'
              task derived_param
            endfamily
            family 16
              trigger ../../../fc/pf/16 eq complete
              edit EPSMEMBER '16'
              edit MEMBER '16'
              edit SUBFSFAMILY '/pf016'
              task derived_param
            endfamily
            family 17
              trigger ../../../fc/pf/17 eq complete
              edit EPSMEMBER '17'
              edit MEMBER '17'
              edit SUBFSFAMILY '/pf017'
              task derived_param
            endfamily
            family 18
              trigger ../../../fc/pf/18 eq complete
              edit EPSMEMBER '18'
              edit MEMBER '18'
              edit SUBFSFAMILY '/pf018'
              task derived_param
            endfamily
            family 19
              trigger ../../../fc/pf/19 eq complete
              edit EPSMEMBER '19'
              edit MEMBER '19'
              edit SUBFSFAMILY '/pf019'
              task derived_param
            endfamily
            family 20
              trigger ../../../fc/pf/20 eq complete
              edit EPSMEMBER '20'
              edit MEMBER '20'
              edit SUBFSFAMILY '/pf020'
              task derived_param
            endfamily
            family 21
              trigger ../../../fc/pf/21 eq complete
              edit EPSMEMBER '21'
              edit MEMBER '21'
              edit SUBFSFAMILY '/pf021'
              task derived_param
            endfamily
            family 22
              trigger ../../../fc/pf/22 eq complete
              edit EPSMEMBER '22'
              edit MEMBER '22'
              edit SUBFSFAMILY '/pf022'
              task derived_param
            endfamily
            family 23
              trigger ../../../fc/pf/23 eq complete
              edit EPSMEMBER '23'
              edit MEMBER '23'
              edit SUBFSFAMILY '/pf023'
              task derived_param
            endfamily
            family 24
              trigger ../../../fc/pf/24 eq complete
              edit EPSMEMBER '24'
              edit MEMBER '24'
              edit SUBFSFAMILY '/pf024'
              task derived_param
            endfamily
            family 25
              trigger ../../../fc/pf/25 eq complete
              edit EPSMEMBER '25'
              edit MEMBER '25'
              edit SUBFSFAMILY '/pf025'
              task derived_param
            endfamily
            family 26
              trigger ../../../fc/pf/26 eq complete
              edit EPSMEMBER '26'
              edit MEMBER '26'
              edit SUBFSFAMILY '/pf026'
              task derived_param
            endfamily
            family 27
              trigger ../../../fc/pf/27 eq complete
              edit EPSMEMBER '27'
              edit MEMBER '27'
              edit SUBFSFAMILY '/pf027'
              task derived_param
            endfamily
            family 28
              trigger ../../../fc/pf/28 eq complete
              edit EPSMEMBER '28'
              edit MEMBER '28'
              edit SUBFSFAMILY '/pf028'
              task derived_param
            endfamily
            family 29
              trigger ../../../fc/pf/29 eq complete
              edit EPSMEMBER '29'
              edit MEMBER '29'
              edit SUBFSFAMILY '/pf029'
              task derived_param
            endfamily
            family 30
              trigger ../../../fc/pf/30 eq complete
              edit EPSMEMBER '30'
              edit MEMBER '30'
              edit SUBFSFAMILY '/pf030'
              task derived_param
            endfamily
            family 31
              trigger ../../../fc/pf/31 eq complete
              edit EPSMEMBER '31'
              edit MEMBER '31'
              edit SUBFSFAMILY '/pf031'
              task derived_param
            endfamily
            family 32
              trigger ../../../fc/pf/32 eq complete
              edit EPSMEMBER '32'
              edit MEMBER '32'
              edit SUBFSFAMILY '/pf032'
              task derived_param
            endfamily
            family 33
              trigger ../../../fc/pf/33 eq complete
              edit EPSMEMBER '33'
              edit MEMBER '33'
              edit SUBFSFAMILY '/pf033'
              task derived_param
            endfamily
            family 34
              trigger ../../../fc/pf/34 eq complete
              edit EPSMEMBER '34'
              edit MEMBER '34'
              edit SUBFSFAMILY '/pf034'
              task derived_param
            endfamily
            family 35
              trigger ../../../fc/pf/35 eq complete
              edit EPSMEMBER '35'
              edit MEMBER '35'
              edit SUBFSFAMILY '/pf035'
              task derived_param
            endfamily
            family 36
              trigger ../../../fc/pf/36 eq complete
              edit EPSMEMBER '36'
              edit MEMBER '36'
              edit SUBFSFAMILY '/pf036'
              task derived_param
            endfamily
            family 37
              trigger ../../../fc/pf/37 eq complete
              edit EPSMEMBER '37'
              edit MEMBER '37'
              edit SUBFSFAMILY '/pf037'
              task derived_param
            endfamily
            family 38
              trigger ../../../fc/pf/38 eq complete
              edit EPSMEMBER '38'
              edit MEMBER '38'
              edit SUBFSFAMILY '/pf038'
              task derived_param
            endfamily
            family 39
              trigger ../../../fc/pf/39 eq complete
              edit EPSMEMBER '39'
              edit MEMBER '39'
              edit SUBFSFAMILY '/pf039'
              task derived_param
            endfamily
            family 40
              trigger ../../../fc/pf/40 eq complete
              edit EPSMEMBER '40'
              edit MEMBER '40'
              edit SUBFSFAMILY '/pf040'
              task derived_param
            endfamily
            family 41
              trigger ../../../fc/pf/41 eq complete
              edit EPSMEMBER '41'
              edit MEMBER '41'
              edit SUBFSFAMILY '/pf041'
              task derived_param
            endfamily
            family 42
              trigger ../../../fc/pf/42 eq complete
              edit EPSMEMBER '42'
              edit MEMBER '42'
              edit SUBFSFAMILY '/pf042'
              task derived_param
            endfamily
            family 43
              trigger ../../../fc/pf/43 eq complete
              edit EPSMEMBER '43'
              edit MEMBER '43'
              edit SUBFSFAMILY '/pf043'
              task derived_param
            endfamily
            family 44
              trigger ../../../fc/pf/44 eq complete
              edit EPSMEMBER '44'
              edit MEMBER '44'
              edit SUBFSFAMILY '/pf044'
              task derived_param
            endfamily
            family 45
              trigger ../../../fc/pf/45 eq complete
              edit EPSMEMBER '45'
              edit MEMBER '45'
              edit SUBFSFAMILY '/pf045'
              task derived_param
            endfamily
            family 46
              trigger ../../../fc/pf/46 eq complete
              edit EPSMEMBER '46'
              edit MEMBER '46'
              edit SUBFSFAMILY '/pf046'
              task derived_param
            endfamily
            family 47
              trigger ../../../fc/pf/47 eq complete
              edit EPSMEMBER '47'
              edit MEMBER '47'
              edit SUBFSFAMILY '/pf047'
              task derived_param
            endfamily
            family 48
              trigger ../../../fc/pf/48 eq complete
              edit EPSMEMBER '48'
              edit MEMBER '48'
              edit SUBFSFAMILY '/pf048'
              task derived_param
            endfamily
            family 49
              trigger ../../../fc/pf/49 eq complete
              edit EPSMEMBER '49'
              edit MEMBER '49'
              edit SUBFSFAMILY '/pf049'
              task derived_param
            endfamily
            family 50
              trigger ../../../fc/pf/50 eq complete
              edit EPSMEMBER '50'
              edit MEMBER '50'
              edit SUBFSFAMILY '/pf050'
              task derived_param
            endfamily
          endfamily
        endfamily
        family extra
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
          edit STEPS '366:768:6'
          inlimit /emc_36r1/limits:extra
          family cf
            trigger ../dp/cf eq complete
            edit EMOS_TYPE 'cf'
            task extrafields
              edit EPSFCRES '319'
              edit FCLENGTH '768'
              edit FCLENGTH0 '360'
            task wm_sfc
              trigger extrafields eq complete
              edit FCLENGTH '768'
            task wm_ua
              trigger wm_sfc eq complete
              edit FCLENGTH '768'
          endfamily
          family pf
            complete /emc_36r1/main/ref/00/events:cf
            edit EMOS_TYPE 'pf'
            family 01
              trigger ../../dp/pf/01 eq complete
              edit EPSMEMBER '1'
              edit MEMBER '1'
              edit SUBFSFAMILY '/pf001'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 02
              trigger ../../dp/pf/02 eq complete
              edit EPSMEMBER '2'
              edit MEMBER '2'
              edit SUBFSFAMILY '/pf002'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 03
              trigger ../../dp/pf/03 eq complete
              edit EPSMEMBER '3'
              edit MEMBER '3'
              edit SUBFSFAMILY '/pf003'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 04
              trigger ../../dp/pf/04 eq complete
              edit EPSMEMBER '4'
              edit MEMBER '4'
              edit SUBFSFAMILY '/pf004'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 05
              trigger ../../dp/pf/05 eq complete
              edit EPSMEMBER '5'
              edit MEMBER '5'
              edit SUBFSFAMILY '/pf005'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 06
              trigger ../../dp/pf/06 eq complete
              edit EPSMEMBER '6'
              edit MEMBER '6'
              edit SUBFSFAMILY '/pf006'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 07
              trigger ../../dp/pf/07 eq complete
              edit EPSMEMBER '7'
              edit MEMBER '7'
              edit SUBFSFAMILY '/pf007'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 08
              trigger ../../dp/pf/08 eq complete
              edit EPSMEMBER '8'
              edit MEMBER '8'
              edit SUBFSFAMILY '/pf008'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 09
              trigger ../../dp/pf/09 eq complete
              edit EPSMEMBER '9'
              edit MEMBER '9'
              edit SUBFSFAMILY '/pf009'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 10
              trigger ../../dp/pf/10 eq complete
              edit EPSMEMBER '10'
              edit MEMBER '10'
              edit SUBFSFAMILY '/pf010'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 11
              trigger ../../dp/pf/11 eq complete
              edit EPSMEMBER '11'
              edit MEMBER '11'
              edit SUBFSFAMILY '/pf011'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 12
              trigger ../../dp/pf/12 eq complete
              edit EPSMEMBER '12'
              edit MEMBER '12'
              edit SUBFSFAMILY '/pf012'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 13
              trigger ../../dp/pf/13 eq complete
              edit EPSMEMBER '13'
              edit MEMBER '13'
              edit SUBFSFAMILY '/pf013'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 14
              trigger ../../dp/pf/14 eq complete
              edit EPSMEMBER '14'
              edit MEMBER '14'
              edit SUBFSFAMILY '/pf014'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 15
              trigger ../../dp/pf/15 eq complete
              edit EPSMEMBER '15'
              edit MEMBER '15'
              edit SUBFSFAMILY '/pf015'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 16
              trigger ../../dp/pf/16 eq complete
              edit EPSMEMBER '16'
              edit MEMBER '16'
              edit SUBFSFAMILY '/pf016'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 17
              trigger ../../dp/pf/17 eq complete
              edit EPSMEMBER '17'
              edit MEMBER '17'
              edit SUBFSFAMILY '/pf017'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 18
              trigger ../../dp/pf/18 eq complete
              edit EPSMEMBER '18'
              edit MEMBER '18'
              edit SUBFSFAMILY '/pf018'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 19
              trigger ../../dp/pf/19 eq complete
              edit EPSMEMBER '19'
              edit MEMBER '19'
              edit SUBFSFAMILY '/pf019'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 20
              trigger ../../dp/pf/20 eq complete
              edit EPSMEMBER '20'
              edit MEMBER '20'
              edit SUBFSFAMILY '/pf020'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 21
              trigger ../../dp/pf/21 eq complete
              edit EPSMEMBER '21'
              edit MEMBER '21'
              edit SUBFSFAMILY '/pf021'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 22
              trigger ../../dp/pf/22 eq complete
              edit EPSMEMBER '22'
              edit MEMBER '22'
              edit SUBFSFAMILY '/pf022'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 23
              trigger ../../dp/pf/23 eq complete
              edit EPSMEMBER '23'
              edit MEMBER '23'
              edit SUBFSFAMILY '/pf023'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 24
              trigger ../../dp/pf/24 eq complete
              edit EPSMEMBER '24'
              edit MEMBER '24'
              edit SUBFSFAMILY '/pf024'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 25
              trigger ../../dp/pf/25 eq complete
              edit EPSMEMBER '25'
              edit MEMBER '25'
              edit SUBFSFAMILY '/pf025'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 26
              trigger ../../dp/pf/26 eq complete
              edit EPSMEMBER '26'
              edit MEMBER '26'
              edit SUBFSFAMILY '/pf026'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 27
              trigger ../../dp/pf/27 eq complete
              edit EPSMEMBER '27'
              edit MEMBER '27'
              edit SUBFSFAMILY '/pf027'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 28
              trigger ../../dp/pf/28 eq complete
              edit EPSMEMBER '28'
              edit MEMBER '28'
              edit SUBFSFAMILY '/pf028'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 29
              trigger ../../dp/pf/29 eq complete
              edit EPSMEMBER '29'
              edit MEMBER '29'
              edit SUBFSFAMILY '/pf029'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 30
              trigger ../../dp/pf/30 eq complete
              edit EPSMEMBER '30'
              edit MEMBER '30'
              edit SUBFSFAMILY '/pf030'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 31
              trigger ../../dp/pf/31 eq complete
              edit EPSMEMBER '31'
              edit MEMBER '31'
              edit SUBFSFAMILY '/pf031'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 32
              trigger ../../dp/pf/32 eq complete
              edit EPSMEMBER '32'
              edit MEMBER '32'
              edit SUBFSFAMILY '/pf032'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 33
              trigger ../../dp/pf/33 eq complete
              edit EPSMEMBER '33'
              edit MEMBER '33'
              edit SUBFSFAMILY '/pf033'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 34
              trigger ../../dp/pf/34 eq complete
              edit EPSMEMBER '34'
              edit MEMBER '34'
              edit SUBFSFAMILY '/pf034'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 35
              trigger ../../dp/pf/35 eq complete
              edit EPSMEMBER '35'
              edit MEMBER '35'
              edit SUBFSFAMILY '/pf035'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 36
              trigger ../../dp/pf/36 eq complete
              edit EPSMEMBER '36'
              edit MEMBER '36'
              edit SUBFSFAMILY '/pf036'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 37
              trigger ../../dp/pf/37 eq complete
              edit EPSMEMBER '37'
              edit MEMBER '37'
              edit SUBFSFAMILY '/pf037'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 38
              trigger ../../dp/pf/38 eq complete
              edit EPSMEMBER '38'
              edit MEMBER '38'
              edit SUBFSFAMILY '/pf038'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 39
              trigger ../../dp/pf/39 eq complete
              edit EPSMEMBER '39'
              edit MEMBER '39'
              edit SUBFSFAMILY '/pf039'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 40
              trigger ../../dp/pf/40 eq complete
              edit EPSMEMBER '40'
              edit MEMBER '40'
              edit SUBFSFAMILY '/pf040'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 41
              trigger ../../dp/pf/41 eq complete
              edit EPSMEMBER '41'
              edit MEMBER '41'
              edit SUBFSFAMILY '/pf041'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 42
              trigger ../../dp/pf/42 eq complete
              edit EPSMEMBER '42'
              edit MEMBER '42'
              edit SUBFSFAMILY '/pf042'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 43
              trigger ../../dp/pf/43 eq complete
              edit EPSMEMBER '43'
              edit MEMBER '43'
              edit SUBFSFAMILY '/pf043'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 44
              trigger ../../dp/pf/44 eq complete
              edit EPSMEMBER '44'
              edit MEMBER '44'
              edit SUBFSFAMILY '/pf044'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 45
              trigger ../../dp/pf/45 eq complete
              edit EPSMEMBER '45'
              edit MEMBER '45'
              edit SUBFSFAMILY '/pf045'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 46
              trigger ../../dp/pf/46 eq complete
              edit EPSMEMBER '46'
              edit MEMBER '46'
              edit SUBFSFAMILY '/pf046'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 47
              trigger ../../dp/pf/47 eq complete
              edit EPSMEMBER '47'
              edit MEMBER '47'
              edit SUBFSFAMILY '/pf047'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 48
              trigger ../../dp/pf/48 eq complete
              edit EPSMEMBER '48'
              edit MEMBER '48'
              edit SUBFSFAMILY '/pf048'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 49
              trigger ../../dp/pf/49 eq complete
              edit EPSMEMBER '49'
              edit MEMBER '49'
              edit SUBFSFAMILY '/pf049'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
            family 50
              trigger ../../dp/pf/50 eq complete
              edit EPSMEMBER '50'
              edit MEMBER '50'
              edit SUBFSFAMILY '/pf050'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '768'
                edit FCLENGTH0 '360'
              task wm_sfc
                trigger extrafields eq complete
                edit FCLENGTH '768'
              task wm_ua
                trigger wm_sfc eq complete
                edit FCLENGTH '768'
            endfamily
          endfamily
        endfamily
        family mofc
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/leg'
          edit EMOS_TYPE '00'
          edit EXPVER_ENFH '0046'
          edit EXPVER_EFHS '0046'
          edit NH_ENSEMBLES '4'
          task wmem_veps
            trigger ../dp eq complete and ../extra eq complete and ../ensms eq complete
          family wmanom
            trigger ( /emc_36r1/hc0:YMD gt /emc_36r1/00:YMD or ( /emc_36r1/hc0:YMD eq /emc_36r1/00:YMD and /emc_36r1/hc0/postproc/wmem_veps eq complete)) and wmem_veps eq complete and ../ensms eq complete
            edit WRITE_FDB '0'
            family fcmean
              edit TYPES 'fcmean'
              family 129
                edit PARAMS '129'
                task wmanom
              endfamily
              family all
                edit PARAMS '130 133 001 078 228'
                task wmanom
              endfamily
            endfamily
            family em
              edit TYPES 'em'
              family all
                edit PARAMS 'all'
                task wmanom
              endfamily
            endfamily
            task prob_fdb
              trigger fcmean eq complete and em eq complete
              edit PARAM '129 130 133 001 078 228 144 all'
              edit MARS_FDB_SET_TABLE '1'
          endfamily
          task signi
            trigger wmanom eq complete
          task prob_thr
            trigger wmanom eq complete
          task prob_perc
            trigger wmanom eq complete
        endfamily
        family ensms
          complete /emc_36r1/main/ref/00/events:cf
          trigger ../fc eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/leg'
          family limits
            limit fdb 1
            limit gen 8
          endfamily
          task ensms
            edit ENS_M 'false'
            edit ENS_S 'false'
            edit ENS_MARS 'false'
            event stage
          family 498
            trigger ./ensms:stage or ./ensms eq complete
            edit STEP1 '366'
            edit STEP2 '498'
            edit BY '6'
            edit STEPS '0'
            family T
              edit ENS_M 'false'
              edit ENS_S 'false'
              edit LIST_PARAM 'T'
              inlimit ../limits:gen
              family 850
                task ensms
                  edit LIST_LEV '850'
              endfamily
              family 500
                task ensms
                  edit LIST_LEV '500'
              endfamily
            endfamily
            family Z
              edit ENS_M 'false'
              edit ENS_S 'false'
              edit LIST_PARAM 'Z'
              inlimit ../limits:gen
              family 1000
                task ensms
                  edit LIST_LEV '1000'
              endfamily
              family 500
                task ensms
                  edit LIST_LEV '500'
              endfamily
            endfamily
            task ensms
              trigger T eq complete and Z eq complete
              edit ENS_MARS 'false'
              inlimit ../limits:fdb
          endfamily
          family 636
            trigger ./ensms:stage or ./ensms eq complete
            edit STEP1 '504'
            edit STEP2 '636'
            edit BY '6'
            edit STEPS '0'
            family T
              edit ENS_M 'false'
              edit ENS_S 'false'
              edit LIST_PARAM 'T'
              inlimit ../limits:gen
              family 850
                task ensms
                  edit LIST_LEV '850'
              endfamily
              family 500
                task ensms
                  edit LIST_LEV '500'
              endfamily
            endfamily
            family Z
              edit ENS_M 'false'
              edit ENS_S 'false'
              edit LIST_PARAM 'Z'
              inlimit ../limits:gen
              family 1000
                task ensms
                  edit LIST_LEV '1000'
              endfamily
              family 500
                task ensms
                  edit LIST_LEV '500'
              endfamily
            endfamily
            task ensms
              trigger T eq complete and Z eq complete
              edit ENS_MARS 'false'
              inlimit ../limits:fdb
          endfamily
          family 768
            trigger ./ensms:stage or ./ensms eq complete
            edit STEP1 '642'
            edit STEP2 '768'
            edit BY '6'
            edit STEPS '0'
            family T
              edit ENS_M 'false'
              edit ENS_S 'false'
              edit LIST_PARAM 'T'
              inlimit ../limits:gen
              family 850
                task ensms
                  edit LIST_LEV '850'
              endfamily
              family 500
                task ensms
                  edit LIST_LEV '500'
              endfamily
            endfamily
            family Z
              edit ENS_M 'false'
              edit ENS_S 'false'
              edit LIST_PARAM 'Z'
              inlimit ../limits:gen
              family 1000
                task ensms
                  edit LIST_LEV '1000'
              endfamily
              family 500
                task ensms
                  edit LIST_LEV '500'
              endfamily
            endfamily
            task ensms
              trigger T eq complete and Z eq complete
              edit ENS_MARS 'false'
              inlimit ../limits:fdb
          endfamily
        endfamily
      endfamily
      family prod
        trigger fc eq complete and postproc eq complete
        edit SMSFILES '/home/ma/emos/def/emc_36r1/prod'
        edit EMOS_STREAM 'ENFO'
        edit EMOS_DOMAIN 'G'
        edit EMOS_TYPE 'FC'
        edit PROD_TASK '1'
        edit EMOS_PROJECT '00'
        edit EMOS_BASE '00'
        edit PGNPES '32'
        family eps
          trigger ( ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and ( /emc_36r1/main/00/legA/prod/getreq/eps eq complete or /emc_36r1/main/00/legA/prod/getreq/eps:001) and ( /emc_36r1/main/00/legA/prod/getreq/bitmap eq complete or /emc_36r1/main/00/legA/prod/getreq/bitmap:001)))
          edit EMOS_STREAM 'ENFO'
          family day01
            edit EMOS_TIME_STEP_H '0001'
            task prodgen
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day02
            edit EMOS_TIME_STEP_H '0002'
            task prodgen
              trigger ../day01/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day01/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day03
            edit EMOS_TIME_STEP_H '0003'
            task prodgen
              trigger ../day02/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day02/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day04
            edit EMOS_TIME_STEP_H '0004'
            task prodgen
              trigger ../day03/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day03/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day05
            edit EMOS_TIME_STEP_H '0005'
            task prodgen
              trigger ../day04/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day04/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day06
            edit EMOS_TIME_STEP_H '0006'
            task prodgen
              trigger ../day05/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day05/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day07
            edit EMOS_TIME_STEP_H '0007'
            task prodgen
              trigger ../day06/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day06/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day08
            edit EMOS_TIME_STEP_H '0008'
            task prodgen
              trigger ../day07/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day07/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day09
            edit EMOS_TIME_STEP_H '0009'
            task prodgen
              trigger ../day08/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day08/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day10
            edit EMOS_TIME_STEP_H '0010'
            task prodgen
              trigger ../day09/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day09/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day11
            edit EMOS_TIME_STEP_H '0011'
            task prodgen
              trigger ../day10/prodgen eq complete
              edit MEM '1900'
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day10/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day12
            edit EMOS_TIME_STEP_H '0012'
            task prodgen
              trigger ../day11/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day11/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day13
            edit EMOS_TIME_STEP_H '0013'
            task prodgen
              trigger ../day12/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day12/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day14
            edit EMOS_TIME_STEP_H '0014'
            task prodgen
              trigger ../day13/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day13/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day15
            edit EMOS_TIME_STEP_H '0015'
            task prodgen
              trigger ../day14/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day14/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day16
            edit EMOS_TIME_STEP_H '0016'
            task prodgen
              trigger ../day15/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day15/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day17
            edit EMOS_TIME_STEP_H '0017'
            task prodgen
              trigger ../day16/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day16/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day18
            edit EMOS_TIME_STEP_H '0018'
            task prodgen
              trigger ../day17/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day17/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day19
            edit EMOS_TIME_STEP_H '0019'
            task prodgen
              trigger ../day18/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day18/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day20
            edit EMOS_TIME_STEP_H '0020'
            task prodgen
              trigger ../day19/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day19/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day21
            edit EMOS_TIME_STEP_H '0021'
            task prodgen
              trigger ../day20/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day20/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day22
            edit EMOS_TIME_STEP_H '0022'
            task prodgen
              trigger ../day21/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day21/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day23
            edit EMOS_TIME_STEP_H '0023'
            task prodgen
              trigger ../day22/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day22/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day24
            edit EMOS_TIME_STEP_H '0024'
            task prodgen
              trigger ../day23/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day23/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day25
            edit EMOS_TIME_STEP_H '0025'
            task prodgen
              trigger ../day24/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day24/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day26
            edit EMOS_TIME_STEP_H '0026'
            task prodgen
              trigger ../day25/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day25/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day27
            edit EMOS_TIME_STEP_H '0027'
            task prodgen
              trigger ../day26/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day26/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day28
            edit EMOS_TIME_STEP_H '0028'
            task prodgen
              trigger ../day27/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day27/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day29
            edit EMOS_TIME_STEP_H '0029'
            task prodgen
              trigger ../day28/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day28/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day30
            edit EMOS_TIME_STEP_H '0030'
            task prodgen
              trigger ../day29/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day29/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day31
            edit EMOS_TIME_STEP_H '0031'
            task prodgen
              trigger ../day30/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day30/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day32
            edit EMOS_TIME_STEP_H '0032'
            task prodgen
              trigger ../day31/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day31/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day33
            edit EMOS_TIME_STEP_H '0033'
            task prodgen
              trigger ../day32/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day32/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
        endfamily
        family wave
          trigger ./eps/day33/prodgen eq complete and ( ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and ( /emc_36r1/main/00/legA/prod/getreq/wave_eps eq complete or /emc_36r1/main/00/legA/prod/getreq/wave_eps:001)))
          edit PGNPES '4'
          edit EMOS_STREAM 'WAEF'
          family day01
            edit EMOS_TIME_STEP_H '0001'
            task prodgen
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ./pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day02
            edit EMOS_TIME_STEP_H '0002'
            task prodgen
              trigger ../day01/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day01/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day03
            edit EMOS_TIME_STEP_H '0003'
            task prodgen
              trigger ../day02/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day02/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day04
            edit EMOS_TIME_STEP_H '0004'
            task prodgen
              trigger ../day03/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day03/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day05
            edit EMOS_TIME_STEP_H '0005'
            task prodgen
              trigger ../day04/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day04/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day06
            edit EMOS_TIME_STEP_H '0006'
            task prodgen
              trigger ../day05/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day05/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day07
            edit EMOS_TIME_STEP_H '0007'
            task prodgen
              trigger ../day06/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day06/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day08
            edit EMOS_TIME_STEP_H '0008'
            task prodgen
              trigger ../day07/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day07/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day09
            edit EMOS_TIME_STEP_H '0009'
            task prodgen
              trigger ../day08/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day08/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day10
            edit EMOS_TIME_STEP_H '0010'
            task prodgen
              trigger ../day09/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day09/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day11
            edit EMOS_TIME_STEP_H '0011'
            task prodgen
              trigger ../day10/prodgen eq complete
              edit MEM '1900'
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day10/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day12
            edit EMOS_TIME_STEP_H '0012'
            task prodgen
              trigger ../day11/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day11/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day13
            edit EMOS_TIME_STEP_H '0013'
            task prodgen
              trigger ../day12/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day12/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day14
            edit EMOS_TIME_STEP_H '0014'
            task prodgen
              trigger ../day13/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day13/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day15
            edit EMOS_TIME_STEP_H '0015'
            task prodgen
              trigger ../day14/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day14/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day16
            edit EMOS_TIME_STEP_H '0016'
            task prodgen
              trigger ../day15/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day15/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day17
            edit EMOS_TIME_STEP_H '0017'
            task prodgen
              trigger ../day16/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day16/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day18
            edit EMOS_TIME_STEP_H '0018'
            task prodgen
              trigger ../day17/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day17/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day19
            edit EMOS_TIME_STEP_H '0019'
            task prodgen
              trigger ../day18/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day18/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day20
            edit EMOS_TIME_STEP_H '0020'
            task prodgen
              trigger ../day19/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day19/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day21
            edit EMOS_TIME_STEP_H '0021'
            task prodgen
              trigger ../day20/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day20/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day22
            edit EMOS_TIME_STEP_H '0022'
            task prodgen
              trigger ../day21/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day21/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day23
            edit EMOS_TIME_STEP_H '0023'
            task prodgen
              trigger ../day22/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day22/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day24
            edit EMOS_TIME_STEP_H '0024'
            task prodgen
              trigger ../day23/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day23/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day25
            edit EMOS_TIME_STEP_H '0025'
            task prodgen
              trigger ../day24/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day24/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day26
            edit EMOS_TIME_STEP_H '0026'
            task prodgen
              trigger ../day25/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day25/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day27
            edit EMOS_TIME_STEP_H '0027'
            task prodgen
              trigger ../day26/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day26/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day28
            edit EMOS_TIME_STEP_H '0028'
            task prodgen
              trigger ../day27/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day27/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day29
            edit EMOS_TIME_STEP_H '0029'
            task prodgen
              trigger ../day28/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day28/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day30
            edit EMOS_TIME_STEP_H '0030'
            task prodgen
              trigger ../day29/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day29/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day31
            edit EMOS_TIME_STEP_H '0031'
            task prodgen
              trigger ../day30/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day30/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day32
            edit EMOS_TIME_STEP_H '0032'
            task prodgen
              trigger ../day31/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day31/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
          family day33
            edit EMOS_TIME_STEP_H '0033'
            task prodgen
              trigger ../day32/prodgen eq complete
            task pdb
              trigger prodgen eq complete
              inlimit /emc_36r1/limits:pdb
              inlimit /limits:pdb
              label info ""
              event 1 
            task q2diss
              late -s +00:05 -c +00:15
              trigger ../day32/q2diss eq complete and pdb eq complete
              edit USER_PRIORITY '93'
              inlimit /emc_36r1/limits:q2diss
              inlimit /limits:q2diss
              label info ""
              event 1 
          endfamily
        endfamily
      endfamily
    endfamily
    family arch
      trigger ./legC/fc eq complete
      edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
      edit EPSLEG '3'
      edit FSFAMILY 'mars'
      edit OVERLAP '0'
      edit EMOS_TYPE 'an'
      edit EPSFCRES '255'
      inlimit /emc_36r1/limits:mars
      inlimit /limits:mars
      family cf
        edit EMOS_TYPE 'cf'
        edit NUMBER '0'
        family cf408
          edit STEPS '366/to/408/by/6'
          task pfpl
            inlimit /emc_36r1/limits:pl
            label files ""
            label info ""
            event fdb
          task pfsfc
            trigger ../../postproc/dp eq complete
            inlimit /emc_36r1/limits:sfc
            label files ""
            label info ""
            event fdb
          task pfil
            label files ""
            label info ""
            event fdb
          task pfwv
            edit STREAM 'WAEF'
            edit EMOS_STREAM 'WAEF'
            label files ""
            label info ""
            event fdb
        endfamily
        family cf480
          edit STEPS '414/to/480/by/6'
          task pfpl
            inlimit /emc_36r1/limits:pl
            label files ""
            label info ""
            event fdb
          task pfsfc
            trigger ../../postproc/dp eq complete
            inlimit /emc_36r1/limits:sfc
            label files ""
            label info ""
            event fdb
          task pfil
            label files ""
            label info ""
            event fdb
          task pfwv
            edit STREAM 'WAEF'
            edit EMOS_STREAM 'WAEF'
            label files ""
            label info ""
            event fdb
        endfamily
        family cf600
          edit STEPS '486/to/600/by/6'
          task pfpl
            inlimit /emc_36r1/limits:pl
            label files ""
            label info ""
            event fdb
          task pfsfc
            trigger ../../postproc/dp eq complete
            inlimit /emc_36r1/limits:sfc
            label files ""
            label info ""
            event fdb
          task pfil
            label files ""
            label info ""
            event fdb
          task pfwv
            edit STREAM 'WAEF'
            edit EMOS_STREAM 'WAEF'
            label files ""
            label info ""
            event fdb
        endfamily
        family cf768
          edit STEPS '606/to/768/by/6'
          task pfpl
            inlimit /emc_36r1/limits:pl
            label files ""
            label info ""
            event fdb
          task pfsfc
            trigger ../../postproc/dp eq complete
            inlimit /emc_36r1/limits:sfc
            label files ""
            label info ""
            event fdb
          task pfil
            label files ""
            label info ""
            event fdb
          task pfwv
            edit STREAM 'WAEF'
            edit EMOS_STREAM 'WAEF'
            label files ""
            label info ""
            event fdb
        endfamily
      endfamily
      family pf
        complete /emc_36r1/main/ref/00/events:cf
        edit EMOS_TYPE 'pf'
        family pf408
          edit STEPS '366/to/408/by/6'
          task pfpl
            inlimit /emc_36r1/limits:pl
            label files ""
            label info ""
            event fdb
          task pfsfc
            trigger ../../postproc/dp eq complete
            inlimit /emc_36r1/limits:sfc
            label files ""
            label info ""
            event fdb
          task pfil
            label files ""
            label info ""
            event fdb
          task pfwv
            edit STREAM 'WAEF'
            edit EMOS_STREAM 'WAEF'
            label files ""
            label info ""
            event fdb
        endfamily
        family pf480
          edit STEPS '414/to/480/by/6'
          task pfpl
            inlimit /emc_36r1/limits:pl
            label files ""
            label info ""
            event fdb
          task pfsfc
            trigger ../../postproc/dp eq complete
            inlimit /emc_36r1/limits:sfc
            label files ""
            label info ""
            event fdb
          task pfil
            label files ""
            label info ""
            event fdb
          task pfwv
            edit STREAM 'WAEF'
            edit EMOS_STREAM 'WAEF'
            label files ""
            label info ""
            event fdb
        endfamily
        family pf600
          edit STEPS '486/to/600/by/6'
          task pfpl
            inlimit /emc_36r1/limits:pl
            label files ""
            label info ""
            event fdb
          task pfsfc
            trigger ../../postproc/dp eq complete
            inlimit /emc_36r1/limits:sfc
            label files ""
            label info ""
            event fdb
          task pfil
            label files ""
            label info ""
            event fdb
          task pfwv
            edit STREAM 'WAEF'
            edit EMOS_STREAM 'WAEF'
            label files ""
            label info ""
            event fdb
        endfamily
        family pf768
          edit STEPS '606/to/768/by/6'
          task pfpl
            inlimit /emc_36r1/limits:pl
            label files ""
            label info ""
            event fdb
          task pfsfc
            trigger ../../postproc/dp eq complete
            inlimit /emc_36r1/limits:sfc
            label files ""
            label info ""
            event fdb
          task pfil
            label files ""
            label info ""
            event fdb
          task pfwv
            edit STREAM 'WAEF'
            edit EMOS_STREAM 'WAEF'
            label files ""
            label info ""
            event fdb
        endfamily
      endfamily
      family postproc
        complete /emc_36r1/main/ref/00/events:cf
        edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
        edit FCCHUNK '1'
        edit FCLENGTH '768'
        inlimit /emc_36r1/limits:postproc
        family mofc
          complete /emc_36r1/main/ref/00/events:cf
          trigger /emc_36r1/00/legC/postproc/mofc eq complete
          task prob_archive
            label info ""
          task wmem_archive
            label info ""
          task wmanom_archive
            label info ""
          task signi_archive
            trigger wmanom_archive eq complete
            label info ""
        endfamily
        family ensms
          trigger /emc_36r1/00/legC/postproc/ensms eq complete
          edit EMOS_TYPE 'an'
          family 768
            edit STEPS '366/to/768/by/6'
            task mean
              edit TYPE 'em'
              label files ""
              label info ""
              event fdb
            task stdev
              edit TYPE 'es'
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family dp
          trigger /emc_36r1/00/legC/postproc/dp eq complete and /emc_36r1/00/legC/postproc/extra eq complete
          edit RUNVARFC '1'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
          edit EMOS_TYPE 'pf'
          edit STEPS '366/to/768/by/6'
          family cf
            edit EMOS_TYPE 'cf'
            task derivedp
              label files ""
              label info ""
              event fdb
            task extra_arc
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit STEPS '360/to/768'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf
            task derivedp
              label files ""
              label info ""
              event fdb
            task extra_arc
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit STEPS '360/to/768'
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family extra
          trigger /emc_36r1/00/legC/postproc/extra eq complete
          edit RUNVARFC '1'
          task wm_sfc_arc
            label files ""
            label info ""
            event fdb
          task wm_ua_arc
            label files ""
            label info ""
            event fdb
        endfamily
        family reord
          complete /emc_36r1/lag/ref/00/events:cf
          edit EPSCOUPLE '1'
          edit EMOS_TYPE 'pf'
          edit FCLENGTH0 '360'
          edit FCLENGTH '768'
          edit EPSFCRES '255'
          edit FCCHUNK '1'
          edit STEP1 '360'
          edit STEP2 '768'
          edit STEPS '-1'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/leg'
          limit reord 1
          inlimit reord
          label info ""
          family 130
            edit GENERATE '1'
            edit ARCHIVE '0'
            edit PARAMS '130.128'
            task reord
              label info ""
          endfamily
          family 228
            edit GENERATE '1'
            edit ARCHIVE '0'
            edit PARAMS '228.128'
            task reord
              label info ""
          endfamily
          family 207
            edit GENERATE '1'
            edit ARCHIVE '0'
            edit PARAMS '207.128'
            task reord
              label info ""
          endfamily
          family 167
            edit GENERATE '1'
            edit ARCHIVE '0'
            edit PARAMS '167.128'
            task reord
              label info ""
          endfamily
          family 164
            edit GENERATE '1'
            edit ARCHIVE '0'
            edit PARAMS '164.128'
            task reord
              label info ""
          endfamily
          family arch
            trigger ./130 eq complete and ./228 eq complete and ./207 eq complete and ./167 eq complete and ./164 eq complete
            edit GENERATE '0'
            edit ARCHIVE '1'
            inlimit /emc_36r1/lag/ref/00:mutex
            task reord
              label info ""
          endfamily
        endfamily
        family nino
          edit FCCHUNK '1'
          edit OCTOTAL '5'
          edit FCTOTAL '51'
          edit DAILY '0'
          edit RUNVARFC '1'
          edit FCLENGTH1 '360'
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit FCLENGTH '768'
          edit EPSCOUPLE '1'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
          limit semaphore 10
          inlimit /limits:linux_cluster
          inlimit semaphore
          family 00
            edit OCGROUP 'an00'
            edit FCGROUP 'fc0'
            edit EPSMEMBER '0'
            edit MEMBER '0'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 01
            edit OCGROUP 'an1'
            edit FCGROUP 'fc0'
            edit EPSMEMBER '1'
            edit MEMBER '1'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 02
            edit OCGROUP 'an2'
            edit FCGROUP 'fc0'
            edit EPSMEMBER '2'
            edit MEMBER '2'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 03
            edit OCGROUP 'an3'
            edit FCGROUP 'fc0'
            edit EPSMEMBER '3'
            edit MEMBER '3'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 04
            edit OCGROUP 'an4'
            edit FCGROUP 'fc0'
            edit EPSMEMBER '4'
            edit MEMBER '4'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 05
            edit OCGROUP 'an0'
            edit FCGROUP 'fc1'
            edit EPSMEMBER '5'
            edit MEMBER '5'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 06
            edit OCGROUP 'an1'
            edit FCGROUP 'fc1'
            edit EPSMEMBER '6'
            edit MEMBER '6'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 07
            edit OCGROUP 'an2'
            edit FCGROUP 'fc1'
            edit EPSMEMBER '7'
            edit MEMBER '7'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 08
            edit OCGROUP 'an3'
            edit FCGROUP 'fc1'
            edit EPSMEMBER '8'
            edit MEMBER '8'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 09
            edit OCGROUP 'an4'
            edit FCGROUP 'fc1'
            edit EPSMEMBER '9'
            edit MEMBER '9'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 10
            edit OCGROUP 'an0'
            edit FCGROUP 'fc2'
            edit EPSMEMBER '10'
            edit MEMBER '10'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 11
            edit OCGROUP 'an1'
            edit FCGROUP 'fc2'
            edit EPSMEMBER '11'
            edit MEMBER '11'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 12
            edit OCGROUP 'an2'
            edit FCGROUP 'fc2'
            edit EPSMEMBER '12'
            edit MEMBER '12'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 13
            edit OCGROUP 'an3'
            edit FCGROUP 'fc2'
            edit EPSMEMBER '13'
            edit MEMBER '13'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 14
            edit OCGROUP 'an4'
            edit FCGROUP 'fc2'
            edit EPSMEMBER '14'
            edit MEMBER '14'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 15
            edit OCGROUP 'an0'
            edit FCGROUP 'fc3'
            edit EPSMEMBER '15'
            edit MEMBER '15'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 16
            edit OCGROUP 'an1'
            edit FCGROUP 'fc3'
            edit EPSMEMBER '16'
            edit MEMBER '16'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 17
            edit OCGROUP 'an2'
            edit FCGROUP 'fc3'
            edit EPSMEMBER '17'
            edit MEMBER '17'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 18
            edit OCGROUP 'an3'
            edit FCGROUP 'fc3'
            edit EPSMEMBER '18'
            edit MEMBER '18'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 19
            edit OCGROUP 'an4'
            edit FCGROUP 'fc3'
            edit EPSMEMBER '19'
            edit MEMBER '19'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 20
            edit OCGROUP 'an0'
            edit FCGROUP 'fc4'
            edit EPSMEMBER '20'
            edit MEMBER '20'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 21
            edit OCGROUP 'an1'
            edit FCGROUP 'fc4'
            edit EPSMEMBER '21'
            edit MEMBER '21'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 22
            edit OCGROUP 'an2'
            edit FCGROUP 'fc4'
            edit EPSMEMBER '22'
            edit MEMBER '22'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 23
            edit OCGROUP 'an3'
            edit FCGROUP 'fc4'
            edit EPSMEMBER '23'
            edit MEMBER '23'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 24
            edit OCGROUP 'an4'
            edit FCGROUP 'fc4'
            edit EPSMEMBER '24'
            edit MEMBER '24'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 25
            edit OCGROUP 'an0'
            edit FCGROUP 'fc5'
            edit EPSMEMBER '25'
            edit MEMBER '25'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 26
            edit OCGROUP 'an1'
            edit FCGROUP 'fc5'
            edit EPSMEMBER '26'
            edit MEMBER '26'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 27
            edit OCGROUP 'an2'
            edit FCGROUP 'fc5'
            edit EPSMEMBER '27'
            edit MEMBER '27'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 28
            edit OCGROUP 'an3'
            edit FCGROUP 'fc5'
            edit EPSMEMBER '28'
            edit MEMBER '28'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 29
            edit OCGROUP 'an4'
            edit FCGROUP 'fc5'
            edit EPSMEMBER '29'
            edit MEMBER '29'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 30
            edit OCGROUP 'an0'
            edit FCGROUP 'fc6'
            edit EPSMEMBER '30'
            edit MEMBER '30'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 31
            edit OCGROUP 'an1'
            edit FCGROUP 'fc6'
            edit EPSMEMBER '31'
            edit MEMBER '31'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 32
            edit OCGROUP 'an2'
            edit FCGROUP 'fc6'
            edit EPSMEMBER '32'
            edit MEMBER '32'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 33
            edit OCGROUP 'an3'
            edit FCGROUP 'fc6'
            edit EPSMEMBER '33'
            edit MEMBER '33'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 34
            edit OCGROUP 'an4'
            edit FCGROUP 'fc6'
            edit EPSMEMBER '34'
            edit MEMBER '34'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 35
            edit OCGROUP 'an0'
            edit FCGROUP 'fc7'
            edit EPSMEMBER '35'
            edit MEMBER '35'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 36
            edit OCGROUP 'an1'
            edit FCGROUP 'fc7'
            edit EPSMEMBER '36'
            edit MEMBER '36'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 37
            edit OCGROUP 'an2'
            edit FCGROUP 'fc7'
            edit EPSMEMBER '37'
            edit MEMBER '37'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 38
            edit OCGROUP 'an3'
            edit FCGROUP 'fc7'
            edit EPSMEMBER '38'
            edit MEMBER '38'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 39
            edit OCGROUP 'an4'
            edit FCGROUP 'fc7'
            edit EPSMEMBER '39'
            edit MEMBER '39'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 40
            edit OCGROUP 'an0'
            edit FCGROUP 'fc8'
            edit EPSMEMBER '40'
            edit MEMBER '40'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 41
            edit OCGROUP 'an1'
            edit FCGROUP 'fc8'
            edit EPSMEMBER '41'
            edit MEMBER '41'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 42
            edit OCGROUP 'an2'
            edit FCGROUP 'fc8'
            edit EPSMEMBER '42'
            edit MEMBER '42'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 43
            edit OCGROUP 'an3'
            edit FCGROUP 'fc8'
            edit EPSMEMBER '43'
            edit MEMBER '43'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 44
            edit OCGROUP 'an4'
            edit FCGROUP 'fc8'
            edit EPSMEMBER '44'
            edit MEMBER '44'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 45
            edit OCGROUP 'an0'
            edit FCGROUP 'fc9'
            edit EPSMEMBER '45'
            edit MEMBER '45'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 46
            edit OCGROUP 'an1'
            edit FCGROUP 'fc9'
            edit EPSMEMBER '46'
            edit MEMBER '46'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 47
            edit OCGROUP 'an2'
            edit FCGROUP 'fc9'
            edit EPSMEMBER '47'
            edit MEMBER '47'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 48
            edit OCGROUP 'an3'
            edit FCGROUP 'fc9'
            edit EPSMEMBER '48'
            edit MEMBER '48'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 49
            edit OCGROUP 'an4'
            edit FCGROUP 'fc9'
            edit EPSMEMBER '49'
            edit MEMBER '49'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
          family 50
            edit OCGROUP 'an0'
            edit FCGROUP 'fc10'
            edit EPSMEMBER '50'
            edit MEMBER '50'
            task nino_daily
            task tcyc
              trigger nino_daily eq complete or nino_daily eq aborted
          endfamily
        endfamily
        task tcyctoecfs
          trigger nino eq complete
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
          edit EPSCOUPLE '1'
          edit FCLENGTH '768'
          inlimit /limits:linux_cluster
      endfamily
      family coupled
        complete /emc_36r1/main/ref/00/events:cf
        edit EPSLEG '3'
        edit OCLEG '3'
        edit OCEPS '3'
        edit FCCHUNK '1'
        edit FCLENGTH '768'
        edit FCLENGTH0 '360'
        edit OCEPSTOT '3'
        edit FCTOTAL '51'
        edit RUNVARFC '1'
        edit EPSCOUPLE '1'
        edit RUNCOUPLE '1'
        edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
        inlimit /emc_36r1/lag/ref/00:mutex
        task preparcoce
        task arcfc
          trigger preparcoce eq complete
          label files ""
          label info ""
          event fdb
      endfamily
    endfamily
    family msjobs
      trigger ( /emc_36r1/main:YMD gt /emc_36r1/00:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/00:YMD and /emc_36r1/main/00/legA/fc/pf eq complete and /emc_36r1/main/00/legA/fc/cf eq complete and /emc_36r1/main/make eq complete)
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'ecgate'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit SMSOUT '/vol/emos_nc/output'
      edit LOGDIR '/vol/emos_nc/output'
      edit VERSION '0046'
      edit STREAM 'EF'
      edit EMOS_STREAM 'EF'
      edit SMSURLBASE 'http://ecgate.ecmwf.int:9080'
      edit SMSURL 'do/events'
      edit SMSFILES '/home/ma/emos/def/emc_36r1/msjobs'
      edit DELAY '120'
      edit EMOS_TIME_STEP_H '768'
      family hindcast
        trigger /emc_36r1/hc0/postproc eq complete and /emc_36r1/00/ref/thu1000 eq complete
        edit STREAM 'ENFH'
        edit EMOS_STREAM 'ENFH'
        edit MSJ_DIR 'mofcrefc'
        task ms_ec
      endfamily
      family legC
        trigger ../legC eq complete and /emc_36r1/00/ref/thu2200 eq complete
        edit MSJ_DIR 'mofc_all'
        task ms_ec
        family noaa
          edit WSHOST 'linux_cluster'
          edit QUEUE 'plots-medium'
          edit RUNVARFC '1'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/msjobs/noaa'
          edit DISS_HOUR '2200'
          edit SMSURLBASE 'http://diss.ecmwf.int/do/monitoring/arrival'
          edit SMSURL 'CDC/EF/00'
          edit CALL_FTP 'false'
          edit CALL_ECPDS 'true'
          task delete
            trigger ncep_weekly eq complete and /emc_36r1/00/web eq complete and /emc_36r1/00/ref/tue eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/smsfiles/web'
          task ncep_weekly
            trigger /emc_36r1/00/web/process/mjo eq complete
            edit SMSURL 'NO9/EF/00'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/smsfiles/web'
            edit FCLENGTH '768'
            edit ERA40END '2002033000'
        endfamily
      endfamily
    endfamily
    family web
      trigger ../00:YMD + 14 le ../hc2:YMD
      edit QUEUE_EPILOG 'emos-exp'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit WSHOST 'linux_cluster'
      edit QUEUE 'plots-medium'
      edit MARS_DB 'c1b fdb'
      edit EPSLEG '3'
      edit SMSFILES '/home/ma/emos/def/emc_36r1/smsfiles/web'
      edit EMOS_TYPE '00'
      edit RUNVARFC '1'
      edit NH_ENSEMBLES '4'
      edit EXPVER_ENFH '0046'
      edit EXPVER_EFHS '0046'
      inlimit /limits:linux_cluster
      inlimit /emc_36r1/limits:webplots
      family stage
        edit RETRIEVE 'stage'
        task chi200
        task olr
        task u200
        task u850
      endfamily
      family process
        trigger ../legC/fc eq complete and ../legC/postproc eq complete
        family anomaly
          family temp
            edit VAR 'TEMP'
            task plot_anom
          endfamily
          family rain
            edit VAR 'RAIN'
            task plot_anom
          endfamily
          family mslp
            edit VAR 'MSLP'
            task plot_anom
          endfamily
          family 2mtm
            edit VAR '2MTM'
            task plot_anom
          endfamily
          task toweb
            trigger temp eq complete and rain eq complete and mslp eq complete and ./2mtm eq complete
        endfamily
        family probanom
          family temp
            edit VAR 'TEMP'
            task plot_prob
          endfamily
          family rain
            edit VAR 'RAIN'
            task plot_prob
          endfamily
          family mslp
            edit VAR 'MSLP'
            task plot_prob
          endfamily
          family 2mtm
            edit VAR '2MTM'
            task plot_prob
          endfamily
          task toweb
            trigger temp eq complete and rain eq complete and mslp eq complete and ./2mtm eq complete
        endfamily
        family tercile
          family temp
            edit VAR 'TEMP'
            task plot_terc
          endfamily
          family rain
            edit VAR 'RAIN'
            task plot_terc
          endfamily
          family mslp
            edit VAR 'MSLP'
            task plot_terc
          endfamily
          family 2mtm
            edit VAR '2MTM'
            task plot_terc
          endfamily
          task toweb
            trigger temp eq complete and rain eq complete and mslp eq complete and ./2mtm eq complete
            edit TERCILES '.terc1. .terc2.'
        endfamily
        family eps
          trigger ./tercile eq complete or ./tercile/toweb eq aborted
          family members
            task mslp
              late -c +02:00
            task z500
            task mslp_wm
            task z500_wm
          endfamily
          family plumes
            task plume31
            task plumeter
              trigger plume31 eq complete
          endfamily
          family regime
            task kmeansclust
            task kmeansclust_week
            task plotkmeansclust
              trigger kmeansclust_week eq complete and kmeansclust eq complete
            task hovmoeller
            task hovmoeller_itr
            task flowz500
            task multivar
          endfamily
          task epstoweb
            trigger members eq complete and plumes eq complete and regime eq complete
        endfamily
        family mjo
          trigger ../stage eq complete
          edit FCLENGTH '768'
          edit ERA40END '2002033000'
          family process
            task chi200
            task olr
            task u850
            task u200
          endfamily
          task mjotoweb
            trigger ceof eq complete
          task ceof
            trigger process eq complete
          family ncep
            trigger process eq complete
            task olr_to_ncep
            task u200_to_ncep
            task u850_to_ncep
          endfamily
        endfamily
        family tcyc
          trigger /emc_36r1/00/arch/postproc/nino eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1/smsfiles/web/tcyc'
          family hr
            edit KIND 'hr'
            family clim
              edit PROC 'clim'
              task tcycweb
            endfamily
            family map
              edit PROC 'map'
              task tcycweb
            endfamily
            family diff
              trigger clim eq complete and map eq complete
              edit PROC 'diff'
              task tcycweb
            endfamily
          endfamily
          family ts
            edit KIND 'ts'
            family clim
              edit PROC 'clim'
              task tcycweb
            endfamily
            family map
              edit PROC 'map'
              task tcycweb
            endfamily
            family diff
              trigger clim eq complete and map eq complete
              edit PROC 'diff'
              task tcycweb
            endfamily
          endfamily
          family td
            edit KIND 'td'
            family clim
              edit PROC 'clim'
              task tcycweb
            endfamily
            family map
              edit PROC 'map'
              task tcycweb
            endfamily
            family diff
              trigger clim eq complete and map eq complete
              edit PROC 'diff'
              task tcycweb
            endfamily
          endfamily
          task freq
            trigger hr eq complete and ts eq complete and td eq complete
          task tcyctoweb
            trigger hr eq complete and ts eq complete and td eq complete and freq eq complete
        endfamily
        family tcyc_hc
          edit SMSFILES '/home/ma/emos/def/emc_36r1/smsfiles/web/tcyc'
          task tcyc_hc
            edit WSHOST 'localhost'
        endfamily
      endfamily
      task web_update
        trigger ./process eq complete and /emc_36r1/00/ref/thu2200 eq complete
      family clean
        trigger process eq complete and web_update eq complete
        task migrate
        task delete
          trigger migrate eq complete
      endfamily
    endfamily
    family amma
      trigger ./web eq complete and ./ref/thu2200 eq complete
      edit QUEUE_EPILOG 'emos-exp'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit UPDATE_WEB_NOW 'yes'
      edit WSHOST 'linux_cluster'
      edit QUEUE 'emos'
      edit FSFAMILY 'assim'
      edit FCTOTAL '51'
      edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/leg/amma'
      edit RUNVARFC '1'
      edit EMOS_TYPE '00'
      edit EMOS_BASE '00'
      inlimit /limits:linux_cluster
      family init
        task list
        task getfore
          trigger ./list eq complete
        task getclim
          trigger ./getfore eq complete
      endfamily
      family anomaly
        trigger ./init eq complete
        family sf200
          edit VAR 'SF200'
          task stats_a
          task plot_a
            trigger stats_a eq complete
        endfamily
        family vp200
          trigger ./sf200 eq complete
          edit VAR 'VP200'
          task stats_a
          task plot_a
            trigger stats_a eq complete
        endfamily
        task antoweb
          trigger ./vp200 eq complete
      endfamily
      family tercile
        trigger ./anomaly eq complete or ./anomaly/antoweb eq aborted
        family sf200
          edit VAR 'SF200'
          task stats_t
          task plot_t
            trigger stats_t eq complete
        endfamily
        family vp200
          trigger ./sf200 eq complete
          edit VAR 'VP200'
          task stats_t
          task plot_t
            trigger stats_t eq complete
        endfamily
        task trtoweb
          trigger ./vp200 eq complete
      endfamily
      family mjo
        trigger ./tercile eq complete or ./tercile/trtoweb eq aborted
        edit ERA40END '2002033000'
        task chi200
          edit FCLENGTH '768'
        task mjotoweb
          trigger ./chi200 eq complete
      endfamily
      task cleanup
        trigger ./mjo eq complete
    endfamily
    family efi
      trigger ../hc0:YMD + 14 le ../hc2:YMD
      edit MAXDAY '15'
      edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/efi_hdc'
      inlimit /emc_36r1/limits:efih
      family op
        family clim
          trigger /emc_36r1/hc2/efi/op eq complete
          family steps24
            family 000_024
              edit ACC '24'
              edit EFI_STEP1 '000'
              edit EFI_STEP2 '024'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
            endfamily
            family 012_036
              edit ACC '24'
              edit EFI_STEP1 '012'
              edit EFI_STEP2 '036'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
            endfamily
            family 024_048
              edit ACC '24'
              edit EFI_STEP1 '024'
              edit EFI_STEP2 '048'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
            endfamily
            family 036_060
              edit ACC '24'
              edit EFI_STEP1 '036'
              edit EFI_STEP2 '060'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
            endfamily
            family 048_072
              edit ACC '24'
              edit EFI_STEP1 '048'
              edit EFI_STEP2 '072'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
            endfamily
            family 060_084
              edit ACC '24'
              edit EFI_STEP1 '060'
              edit EFI_STEP2 '084'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
            endfamily
            family 072_096
              edit ACC '24'
              edit EFI_STEP1 '072'
              edit EFI_STEP2 '096'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
            endfamily
            family 084_108
              edit ACC '24'
              edit EFI_STEP1 '084'
              edit EFI_STEP2 '108'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
            endfamily
            family 096_120
              edit ACC '24'
              edit EFI_STEP1 '096'
              edit EFI_STEP2 '120'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
            endfamily
            family 108_132
              edit ACC '24'
              edit EFI_STEP1 '108'
              edit EFI_STEP2 '132'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
            endfamily
            family 120_144
              edit ACC '24'
              edit EFI_STEP1 '120'
              edit EFI_STEP2 '144'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
            endfamily
          endfamily
          family 000_120
            edit ACC '120'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '120'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 024_144
            edit ACC '120'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '144'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 000_240
            edit ACC '240'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '240'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
        endfamily
      endfamily
      family nop
        family clim
          trigger /emc_36r1/hc2/efi/nop eq complete
          family steps24
            family 000_024
              edit ACC '24'
              edit EFI_STEP1 '000'
              edit EFI_STEP2 '024'
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 012_036
              edit ACC '24'
              edit EFI_STEP1 '012'
              edit EFI_STEP2 '036'
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 024_048
              edit ACC '24'
              edit EFI_STEP1 '024'
              edit EFI_STEP2 '048'
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 036_060
              edit ACC '24'
              edit EFI_STEP1 '036'
              edit EFI_STEP2 '060'
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 048_072
              edit ACC '24'
              edit EFI_STEP1 '048'
              edit EFI_STEP2 '072'
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 060_084
              edit ACC '24'
              edit EFI_STEP1 '060'
              edit EFI_STEP2 '084'
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 072_096
              edit ACC '24'
              edit EFI_STEP1 '072'
              edit EFI_STEP2 '096'
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 084_108
              edit ACC '24'
              edit EFI_STEP1 '084'
              edit EFI_STEP2 '108'
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 096_120
              edit ACC '24'
              edit EFI_STEP1 '096'
              edit EFI_STEP2 '120'
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 108_132
              edit ACC '24'
              edit EFI_STEP1 '108'
              edit EFI_STEP2 '132'
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 120_144
              edit ACC '24'
              edit EFI_STEP1 '120'
              edit EFI_STEP2 '144'
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 144_168
              edit ACC '24'
              edit EFI_STEP1 '144'
              edit EFI_STEP2 '168'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 168_192
              edit ACC '24'
              edit EFI_STEP1 '168'
              edit EFI_STEP2 '192'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 192_216
              edit ACC '24'
              edit EFI_STEP1 '192'
              edit EFI_STEP2 '216'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 216_240
              edit ACC '24'
              edit EFI_STEP1 '216'
              edit EFI_STEP2 '240'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 240_264
              edit ACC '24'
              edit EFI_STEP1 '240'
              edit EFI_STEP2 '264'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 264_288
              edit ACC '24'
              edit EFI_STEP1 '264'
              edit EFI_STEP2 '288'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 288_312
              edit ACC '24'
              edit EFI_STEP1 '288'
              edit EFI_STEP2 '312'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 312_336
              edit ACC '24'
              edit EFI_STEP1 '312'
              edit EFI_STEP2 '336'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 336_360
              edit ACC '24'
              edit EFI_STEP1 '336'
              edit EFI_STEP2 '360'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
          endfamily
          family steps06
            family 006_006
              edit ACC '0'
              edit EFI_STEP1 '006'
              edit EFI_STEP2 '006'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 012_012
              edit ACC '0'
              edit EFI_STEP1 '012'
              edit EFI_STEP2 '012'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 018_018
              edit ACC '0'
              edit EFI_STEP1 '018'
              edit EFI_STEP2 '018'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 024_024
              edit ACC '0'
              edit EFI_STEP1 '024'
              edit EFI_STEP2 '024'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 030_030
              edit ACC '0'
              edit EFI_STEP1 '030'
              edit EFI_STEP2 '030'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 036_036
              edit ACC '0'
              edit EFI_STEP1 '036'
              edit EFI_STEP2 '036'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 042_042
              edit ACC '0'
              edit EFI_STEP1 '042'
              edit EFI_STEP2 '042'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 048_048
              edit ACC '0'
              edit EFI_STEP1 '048'
              edit EFI_STEP2 '048'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 054_054
              edit ACC '0'
              edit EFI_STEP1 '054'
              edit EFI_STEP2 '054'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 060_060
              edit ACC '0'
              edit EFI_STEP1 '060'
              edit EFI_STEP2 '060'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 066_066
              edit ACC '0'
              edit EFI_STEP1 '066'
              edit EFI_STEP2 '066'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 072_072
              edit ACC '0'
              edit EFI_STEP1 '072'
              edit EFI_STEP2 '072'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 078_078
              edit ACC '0'
              edit EFI_STEP1 '078'
              edit EFI_STEP2 '078'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 084_084
              edit ACC '0'
              edit EFI_STEP1 '084'
              edit EFI_STEP2 '084'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 090_090
              edit ACC '0'
              edit EFI_STEP1 '090'
              edit EFI_STEP2 '090'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 096_096
              edit ACC '0'
              edit EFI_STEP1 '096'
              edit EFI_STEP2 '096'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 102_102
              edit ACC '0'
              edit EFI_STEP1 '102'
              edit EFI_STEP2 '102'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 108_108
              edit ACC '0'
              edit EFI_STEP1 '108'
              edit EFI_STEP2 '108'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 114_114
              edit ACC '0'
              edit EFI_STEP1 '114'
              edit EFI_STEP2 '114'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 120_120
              edit ACC '0'
              edit EFI_STEP1 '120'
              edit EFI_STEP2 '120'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 126_126
              edit ACC '0'
              edit EFI_STEP1 '126'
              edit EFI_STEP2 '126'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 132_132
              edit ACC '0'
              edit EFI_STEP1 '132'
              edit EFI_STEP2 '132'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 138_138
              edit ACC '0'
              edit EFI_STEP1 '138'
              edit EFI_STEP2 '138'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 144_144
              edit ACC '0'
              edit EFI_STEP1 '144'
              edit EFI_STEP2 '144'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 150_150
              edit ACC '0'
              edit EFI_STEP1 '150'
              edit EFI_STEP2 '150'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 156_156
              edit ACC '0'
              edit EFI_STEP1 '156'
              edit EFI_STEP2 '156'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 162_162
              edit ACC '0'
              edit EFI_STEP1 '162'
              edit EFI_STEP2 '162'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 168_168
              edit ACC '0'
              edit EFI_STEP1 '168'
              edit EFI_STEP2 '168'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 174_174
              edit ACC '0'
              edit EFI_STEP1 '174'
              edit EFI_STEP2 '174'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 180_180
              edit ACC '0'
              edit EFI_STEP1 '180'
              edit EFI_STEP2 '180'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 186_186
              edit ACC '0'
              edit EFI_STEP1 '186'
              edit EFI_STEP2 '186'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 192_192
              edit ACC '0'
              edit EFI_STEP1 '192'
              edit EFI_STEP2 '192'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 198_198
              edit ACC '0'
              edit EFI_STEP1 '198'
              edit EFI_STEP2 '198'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 204_204
              edit ACC '0'
              edit EFI_STEP1 '204'
              edit EFI_STEP2 '204'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 210_210
              edit ACC '0'
              edit EFI_STEP1 '210'
              edit EFI_STEP2 '210'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 216_216
              edit ACC '0'
              edit EFI_STEP1 '216'
              edit EFI_STEP2 '216'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 222_222
              edit ACC '0'
              edit EFI_STEP1 '222'
              edit EFI_STEP2 '222'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 228_228
              edit ACC '0'
              edit EFI_STEP1 '228'
              edit EFI_STEP2 '228'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 234_234
              edit ACC '0'
              edit EFI_STEP1 '234'
              edit EFI_STEP2 '234'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
            family 240_240
              edit ACC '0'
              edit EFI_STEP1 '240'
              edit EFI_STEP2 '240'
              family 2t
                edit RF_NAME '2t'
                task prep_sort
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prep_sort
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prep_sort
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prep_sort
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prep_sort
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prep_sort
              endfamily
              family 10dd
                edit RF_NAME '10dd'
                task prep_sort
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prep_sort
              endfamily
              family tcc
                edit RF_NAME 'tcc'
                task prep_sort
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prep_sort
              endfamily
            endfamily
          endfamily
          family 120_168
            trigger ./steps24 eq complete
            edit ACC '48'
            edit EFI_STEP1 '120'
            edit EFI_STEP2 '168'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 168_240
            trigger ./steps24 eq complete
            edit ACC '72'
            edit EFI_STEP1 '168'
            edit EFI_STEP2 '240'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 000_168
            trigger ./steps24 eq complete
            edit ACC '168'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '168'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 240_336
            trigger ./steps24 eq complete
            edit ACC '96'
            edit EFI_STEP1 '240'
            edit EFI_STEP2 '336'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
          family 168_336
            trigger ./steps24 eq complete
            edit ACC '168'
            edit EFI_STEP1 '168'
            edit EFI_STEP2 '336'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
          endfamily
        endfamily
      endfamily
      task efi_clean
      task sync_climate
        trigger op eq complete and nop eq complete
        edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/efi'
        time 20:45
      family swarm
        trigger ./sync_climate eq complete
        edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/efi'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        edit WSHOST 'linux_cluster'
        edit QUEUE 'plots-medium'
        inlimit /limits:linux_cluster
        time 20:45
        task sync_climate
      endfamily
    endfamily
    task logfilestore
      edit QUEUE 'emos-e'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit KEEPLOGS 'no'
      edit LOGSTOECFS 'no'
      edit MOVE_TO_ECFS 'yes'
    task logfiles
      late -s +00:05 -c +00:35
      trigger ref eq complete and logfilestore eq complete and legC eq complete and arch eq complete and msjobs eq complete and web eq complete and amma eq complete and efi eq complete
      edit QUEUE 'emos-e'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit MOVE_TO_ECFS 'no'
  endfamily
  family limits
    limit mars 15
    limit marsret 10
    limit reservation 1
    limit pftotal 72
    limit pertic 26
    limit modelepsA 13
    limit modelepsB 52
    limit pl 3
    limit sfc 3
    limit prodeps 9
    limit prodwave 5
    limit pdb 5
    limit q2diss 5
    limit w2diss 2
    limit hind 60
    limit pesh 60
    limit efih 40
    limit refc 10
    limit webplots 9
    limit postproc 40
    limit dp 8
    limit tc 7
    limit efi 20
    limit extra 20
  endfamily
  family main
    repeat date YMD 20091003 20121212 1
    edit LOGTASK 'main'
    edit SUITE_START '20091124'
    family make
      task setup
      task links
        trigger setup eq complete
        edit FSFAMILY 'an'
        edit WSHOST 'bee01'
      task datalinks
        trigger setup eq complete and links eq complete
      family det_bins
        trigger links eq complete and datalinks eq complete
        edit SMSFILES '/home/ma/emos/def/emc_36r1/make'
        task odbsql
        task odbtools
          trigger odbsql eq complete
        task black
          edit FSFAMILY 'an'
        task b2otools
          trigger odbtools eq complete
          edit CALC_PSBIAS 'false'
          edit INIPSBIAS 'false'
          edit USE_PSBIAS 'false'
        task ifs
          trigger black eq complete and odbsql eq complete and odbtools eq complete and b2otools eq complete
          edit FSFAMILY 'an'
      endfamily
      family bins
        trigger links eq complete and datalinks eq complete and det_bins eq complete
        edit SMSFILES '/home/ma/emos/def/emc_36r1/make'
        task cluster
        task wamabs
        task prepdata
        task tc_ws
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /limits:linux_cluster
        task mc_tools
          edit SMSFILES '/home/ma/emos/def/emc_36r1/make'
          edit THREADS '1'
        task efi
      endfamily
      family prod
        edit SMSFILES '/home/ma/emos/def/emc_36r1/prod/make'
        edit EMOS_DOMAIN 'g'
        edit EMOS_TYPE 'fc'
        edit EMOS_STREAM 'OPER'
        task wparam
          edit EMOS_TYPE 'wp'
        task tf
          edit EMOS_TYPE 'tf'
        task sortreq
        task bitmap
        task clima
        task disssch
        task tools
        task ef
          edit EMOS_STREAM 'ENFO'
          edit EMOS_TYPE 'ef'
        task waef
          edit EMOS_STREAM 'WAEF'
          edit EMOS_TYPE 'ef'
      endfamily
      family mofc
        trigger links eq complete
        edit EPSCOUPLE '1'
        edit SMSFILES '/home/ma/emos/def/emc_36r1/make'
        edit RUNCOUPLE '1'
        edit RUNVARFC '1'
        task ocsetup
        task libsoc
          trigger ocsetup eq complete
        task oclinks
          trigger libsoc eq complete
        task oasis
          trigger ../det_bins/ifs eq complete
        task prepdata
          trigger oclinks eq complete
        task ma_tools
          trigger oclinks eq complete
        family ocean
          trigger ./ma_tools eq complete
          task getoce
          task getfr
            trigger getoce eq complete
          task ocean
            trigger ../libsoc eq complete
          task forcing
            trigger ../libsoc eq complete and ocean eq complete
          task getcouple
            trigger getoce eq complete
            edit RUNCOUPLE '1'
            edit FCGROUP 'fc0'
        endfamily
        task tools
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /limits:linux_cluster
        task mofc_tools
          edit EPSCOUPLE '1'
        task lib_pp
          edit EPSCOUPLE '1'
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /limits:linux_cluster
        task lib_verif
          edit EPSCOUPLE '1'
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /limits:linux_cluster
        task lib_clim
          edit EPSCOUPLE '1'
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit START_DAY '20091008'
          edit END_DAY '20121213'
          edit RUNHINDCAST '1'
          inlimit /limits:linux_cluster
      endfamily
      family wconst
        trigger bins/wamabs eq complete and links eq complete
        edit SMSFILES '/home/ma/emos/def/emc_36r1/make'
        edit EPSNLEGS '3'
        edit EPSFCLEV '62'
        edit EPSWAMNSTPW '1'
        edit EPSINISTEP_PREVIOUS '0'
        edit EPSWAMRESOL 'global50'
        edit EPSFCRES '639'
        edit TSTEP '1200'
        edit EPSWAMNFRE '30'
        edit EPSWAMNANG '24'
        edit CLASS 'od'
        task wconstA
          edit EPSWAMRESOL 'global50'
          edit EPSFCRES '639'
          edit EPSWAMNFRE '30'
          edit EPSWAMNANG '24'
        task wconstB
          trigger wconstA eq complete
          edit EPSWAMRESOL 'global50'
          edit EPSINISTEP '216'
          edit EPSFCRES '319'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
        task wconstC
          trigger wconstB eq complete
          edit EPSCOUPLE '1'
          edit EPSWAMRESOL 'global300'
          edit EPSFCRES '319'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
      endfamily
      task makesvevo
        trigger links eq complete
      task makesvevo_00
        trigger links eq complete
        edit DELTA_DAY '1'
        edit EMOS_BASE '00'
      task inistage
        edit SMSFILES '/home/ma/emos/def/emc_36r1/make'
      family make_sv
        trigger det_bins/ifs eq complete and links eq complete and inistage eq complete
        family m2_12
          edit EMOS_BASE '12'
          edit DELTA_DAY '-2'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '3000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit FCNPES '16'
          edit EPSSVTSTEP '900'
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit CPUTIME '12000'
          edit STACK '350'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
          task svsh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '73'
            edit MEMBER '0'
          task svnh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '74'
            edit MEMBER '0'
          task svnhevo
            trigger svnh eq complete
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit CPUTIME '120000'
          task svshevo
            trigger svsh eq complete
            edit EVOLVE 'yes'
          task svsave
            trigger svshevo eq complete and svnhevo eq complete
          task clean
            trigger svnhevo eq complete and svshevo eq complete and svsave eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
        endfamily
        family m2_00
          edit EMOS_BASE '00'
          edit DELTA_DAY '-1'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '3000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit FCNPES '16'
          edit EPSSVTSTEP '900'
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit CPUTIME '12000'
          edit STACK '350'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
          task svsh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '73'
            edit MEMBER '0'
          task svnh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '74'
            edit MEMBER '0'
          task svnhevo
            trigger svnh eq complete
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit CPUTIME '120000'
          task svshevo
            trigger svsh eq complete
            edit EVOLVE 'yes'
          task svsave
            trigger svshevo eq complete and svnhevo eq complete
          task clean
            trigger svnhevo eq complete and svshevo eq complete and svsave eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
        endfamily
        family m1_12
          edit EMOS_BASE '12'
          edit DELTA_DAY '-1'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '3000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit FCNPES '16'
          edit EPSSVTSTEP '900'
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit CPUTIME '12000'
          edit STACK '350'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
          task svsh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '73'
            edit MEMBER '0'
          task svnh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '74'
            edit MEMBER '0'
          task svnhevo
            trigger svnh eq complete
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit CPUTIME '120000'
          task svshevo
            trigger svsh eq complete
            edit EVOLVE 'yes'
          task svsave
            trigger svshevo eq complete and svnhevo eq complete
          task clean
            trigger svnhevo eq complete and svshevo eq complete and svsave eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
        endfamily
        family m1_00
          edit EMOS_BASE '00'
          edit DELTA_DAY '-0'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '3000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit FCNPES '16'
          edit EPSSVTSTEP '900'
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit CPUTIME '12000'
          edit STACK '350'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
          task svsh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '73'
            edit MEMBER '0'
          task svnh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '74'
            edit MEMBER '0'
          task svnhevo
            trigger svnh eq complete
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit CPUTIME '120000'
          task svshevo
            trigger svsh eq complete
            edit EVOLVE 'yes'
          task svsave
            trigger svshevo eq complete and svnhevo eq complete
          task clean
            trigger svnhevo eq complete and svshevo eq complete and svsave eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
        endfamily
      endfamily
      family calib
        trigger links eq complete and datalinks eq complete
        edit SMSFILES '/home/ma/emos/def/emc_36r1/make'
        family onws
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /limits:linux_cluster
          task tubes
            edit DELTA_DAY '0'
          task tubes_00
            edit EMOS_BASE '00'
            edit DELTA_DAY '1'
        endfamily
        family onsc
          edit QUEUE 'ns'
          edit QUEUE_EPILOG 'ns'
          edit SMSCMD 'smssubmit %USER% %SCHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %SCHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %SCHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/s1b/emos_esuite/emos_data/0046/log'
          edit SMSOUT '/s1b/emos_esuite/emos_data/0046/log'
          inlimit /limits:c1b
          task tubes
            edit DELTA_DAY '0'
          task tubes_00
            edit EMOS_BASE '00'
            edit DELTA_DAY '1'
        endfamily
      endfamily
    endfamily
    task inistage
      trigger ./make eq complete
      edit SMSFILES '/home/ma/emos/def/emc_36r1/make'
    task inistage_1
      trigger ./inistage eq complete
      edit DELTA_DAY '1'
      edit SMSFILES '/home/ma/emos/def/emc_36r1/make'
    family ref
      trigger /emc_36r1/main:YMD le /emc_36r1/main/ref:MC_STOP
      edit MC_STOP '20121213'
      edit SMSFILES '/home/ma/emos/def/emc_36r1/ref'
      limit mutex 1
      family 12
        edit EMOS_BASE '12'
        task events
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/ref'
          edit HINDCAST '1'
          edit ONLY_CF '0'
          event noefi
          event 240 
          event 360 
          event cpl
          event 768 
          event cf
      endfamily
      family 00
        edit EMOS_BASE '00'
        edit DELTA_DAY '1'
        task events
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/ref'
          edit HINDCAST '1'
          edit ONLY_CF '0'
          event noefi
          event 240 
          event 360 
          event cpl
          event 768 
          event cf
      endfamily
      task dummy
        time 13:00
      task cpl_nocpl
      task wait
        complete ( /emc_36r1/lag:YMD + 1) ge /emc_36r1/main:YMD
        trigger ( /emc_36r1/lag:YMD + 1) ge /emc_36r1/main:YMD
    endfamily
    family 12
      trigger ( ./make eq complete) and ( ./inistage eq complete)
      edit EMOS_BASE '12'
      edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
      family ref
        trigger ../make eq complete and ( /emc_36r1/main:YMD le /e_36r1/main:YMD)
        edit QUEUE_EPILOG 'emos-exp'
        edit WSHOST 'localhost'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        edit QUEUE 'emos-exp'
        limit mutex 1
        task an
          trigger /e_36r1/main/12/an eq complete
        task anfc
          complete ( ( /e_36r1/main/12/an/surf_anal/slwet eq complete and ( /e_36r1/main/12/an/4dvar/ifstraj:finalwave or /e_36r1/main/12/an/4dvar eq complete)) or /e_36r1/main/12/an eq complete)
          trigger 0 eq 1
        task model_active
          trigger /e_36r1/main/12/fc/model eq active or /e_36r1/main/12/fc/model eq complete
        task an_active
          trigger /e_36r1/main/12/an/4dvar/uptraj_0 eq active or /e_36r1/main/12/an/4dvar/uptraj_0 eq complete
        task model
          trigger /e_36r1/main/12/fc/model eq complete
        task dc
          trigger /e_36r1/main/00dc/fc/model:step ge 12 or /e_36r1/main/00dc/fc/model eq complete
        task obs
          trigger /e_36r1/main/12/obs/get/get_stor eq complete
      endfamily
      family sv
        trigger ./ref/dc eq complete and ./ref/an_active eq complete
        edit EVOLVE 'no'
        edit FASTEX 'no'
        edit EPSTYPE 'sv'
        edit FCMEM '2000'
        edit FCSTACK '400'
        edit FCMRFS '100'
        edit FCLENGTH '48'
        edit EPSSVTCSUB '1'
        edit EPSSVOP '1'
        edit NPES '16'
        edit EPSSVTSTEP '900'
        task getini
          edit LINIENDA '0'
          edit INIDATEENDA '0'
          edit EPSCOUPLE '0'
          edit N_PERT_AN '10'
          edit EPSCOUPL_A '0'
          edit EPSCOUPL_B '0'
          edit EPSCOUPL_C '0'
        task inidata_sv
          trigger getini eq complete
          edit THREADS '8'
        task targets
          late -s +00:03 -c +00:10
          trigger ../ref/obs eq complete
          edit EPSTCBB '1'
          label da_info ""
          label dcda_info ""
          event sv
        task svsh
          trigger inidata_sv eq complete
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit CPUTIME '12000'
          edit STACK '350'
          edit THREADS '8'
          edit USER_PRIORITY '73'
        task svnh
          trigger inidata_sv eq complete
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit CPUTIME '12000'
          edit STACK '350'
          edit THREADS '8'
          edit USER_PRIORITY '74'
        task subspace
          trigger svnh eq complete and svsh eq complete and ( targets eq complete or targets:sv)
        task sv1
          trigger ( targets eq complete or targets:sv) and subspace eq complete
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit CPUTIME '12000'
          edit STACK '350'
          edit MEMBER '1'
          edit EPSTROP '1'
          edit EPSSVDIAB 'true'
          edit THREADS '8'
          edit USER_PRIORITY '72'
        task sv2
          trigger ( targets eq complete or targets:sv) and subspace eq complete
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit CPUTIME '12000'
          edit STACK '350'
          edit MEMBER '2'
          edit EPSTROP '1'
          edit EPSSVDIAB 'true'
          edit THREADS '8'
          edit USER_PRIORITY '71'
        task sv3
          trigger ( targets eq complete or targets:sv) and subspace eq complete
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit CPUTIME '12000'
          edit STACK '350'
          edit MEMBER '3'
          edit EPSTROP '1'
          edit EPSSVDIAB 'true'
          edit THREADS '8'
        task sv4
          trigger ( targets eq complete or targets:sv) and subspace eq complete
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit CPUTIME '12000'
          edit STACK '350'
          edit MEMBER '4'
          edit EPSTROP '1'
          edit EPSSVDIAB 'true'
          edit THREADS '8'
        task sv5
          trigger ( targets eq complete or targets:sv) and subspace eq complete
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit CPUTIME '12000'
          edit STACK '350'
          edit MEMBER '5'
          edit EPSTROP '1'
          edit EPSSVDIAB 'true'
          edit THREADS '8'
        task sv6
          trigger ( targets eq complete or targets:sv) and subspace eq complete
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit CPUTIME '12000'
          edit STACK '350'
          edit MEMBER '6'
          edit EPSTROP '1'
          edit EPSSVDIAB 'true'
          edit THREADS '8'
        task getsvevo
          event ecfs
        task getsvs
          trigger getsvevo eq complete and svnh eq complete and svsh eq complete and sv1 eq complete and sv2 eq complete and sv3 eq complete and sv4 eq complete and sv5 eq complete and sv6 eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
          edit EPSTROP '1'
        task sv_value
          trigger getsvs eq complete
          inlimit /emc_36r1/main/ref:mutex
        task svnhevo
          trigger svnh eq complete and ../legA/fc eq complete
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit STACK '350'
          edit HEMISPHERE 'n'
          edit EVOLVE 'yes'
          edit ANNPES '8'
          edit CPUTIME '120000'
        task svshevo
          trigger svsh eq complete and ../legA/fc eq complete
          edit NPES '24'
          edit THREADS '8'
          edit MEM '1300'
          edit MRF '10'
          edit STACK '450'
          edit EVOLVE 'yes'
          edit ANNPES '8'
          edit CPUTIME '120000'
        task clean
          trigger ../legA/fc eq complete and svnhevo eq complete and svshevo eq complete
        task dcopy
          trigger clean eq complete
          edit SMSFILES '/home/ma/emos/def/o/system'
      endfamily
      family legA
        edit EPSLEG '1'
        family fc
          trigger /emc_36r1/main/12/ref/anfc eq complete
          edit FCMEM '2240'
          edit FCSTACK '700'
          edit FCMRFS '0'
          edit PRIORITY '72'
          edit USER_PRIORITY '72'
          edit N_PERT_AN '10'
          edit SSTLEN '240'
          edit EPSINISTEP_PREVIOUS '0'
          edit EPSWAMNFRE '30'
          edit EPSWAMNANG '24'
          edit TSTEP '1200'
          edit EPSFCRES '639'
          edit RESOL '639'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
          edit CPUTIME '1000'
          edit MEM '960'
          edit THREADS '8'
          edit NPES '8'
          edit EPSLEG '1'
          edit PERSST 'true'
          edit SSTINT '1'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata
            trigger getini eq complete
            edit THREADS '8'
          task getae
            event 1 
          task rot
            trigger ../../sv/getsvs eq complete and ( getae:1 or getae eq complete)
            edit EPSTROP '1'
          family pert_ic
            trigger rot eq complete and inidata eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTROP '1'
            inlimit /emc_36r1/limits:pertic
            family 00
              task pertinic
                edit MEMBER '0'
            endfamily
            family 01
              task pertinic
                edit MEMBER '1'
            endfamily
            family 02
              task pertinic
                edit MEMBER '2'
            endfamily
            family 03
              task pertinic
                edit MEMBER '3'
            endfamily
            family 04
              task pertinic
                edit MEMBER '4'
            endfamily
            family 05
              task pertinic
                edit MEMBER '5'
            endfamily
            family 06
              task pertinic
                edit MEMBER '6'
            endfamily
            family 07
              task pertinic
                edit MEMBER '7'
            endfamily
            family 08
              task pertinic
                edit MEMBER '8'
            endfamily
            family 09
              task pertinic
                edit MEMBER '9'
            endfamily
            family 10
              task pertinic
                edit MEMBER '10'
            endfamily
            family 11
              task pertinic
                edit MEMBER '11'
            endfamily
            family 12
              task pertinic
                edit MEMBER '12'
            endfamily
            family 13
              task pertinic
                edit MEMBER '13'
            endfamily
            family 14
              task pertinic
                edit MEMBER '14'
            endfamily
            family 15
              task pertinic
                edit MEMBER '15'
            endfamily
            family 16
              task pertinic
                edit MEMBER '16'
            endfamily
            family 17
              task pertinic
                edit MEMBER '17'
            endfamily
            family 18
              task pertinic
                edit MEMBER '18'
            endfamily
            family 19
              task pertinic
                edit MEMBER '19'
            endfamily
            family 20
              task pertinic
                edit MEMBER '20'
            endfamily
            family 21
              task pertinic
                edit MEMBER '21'
            endfamily
            family 22
              task pertinic
                edit MEMBER '22'
            endfamily
            family 23
              task pertinic
                edit MEMBER '23'
            endfamily
            family 24
              task pertinic
                edit MEMBER '24'
            endfamily
            family 25
              task pertinic
                edit MEMBER '25'
            endfamily
            family 26
              task pertinic
                edit MEMBER '26'
            endfamily
            family 27
              task pertinic
                edit MEMBER '27'
            endfamily
            family 28
              task pertinic
                edit MEMBER '28'
            endfamily
            family 29
              task pertinic
                edit MEMBER '29'
            endfamily
            family 30
              task pertinic
                edit MEMBER '30'
            endfamily
            family 31
              task pertinic
                edit MEMBER '31'
            endfamily
            family 32
              task pertinic
                edit MEMBER '32'
            endfamily
            family 33
              task pertinic
                edit MEMBER '33'
            endfamily
            family 34
              task pertinic
                edit MEMBER '34'
            endfamily
            family 35
              task pertinic
                edit MEMBER '35'
            endfamily
            family 36
              task pertinic
                edit MEMBER '36'
            endfamily
            family 37
              task pertinic
                edit MEMBER '37'
            endfamily
            family 38
              task pertinic
                edit MEMBER '38'
            endfamily
            family 39
              task pertinic
                edit MEMBER '39'
            endfamily
            family 40
              task pertinic
                edit MEMBER '40'
            endfamily
            family 41
              task pertinic
                edit MEMBER '41'
            endfamily
            family 42
              task pertinic
                edit MEMBER '42'
            endfamily
            family 43
              task pertinic
                edit MEMBER '43'
            endfamily
            family 44
              task pertinic
                edit MEMBER '44'
            endfamily
            family 45
              task pertinic
                edit MEMBER '45'
            endfamily
            family 46
              task pertinic
                edit MEMBER '46'
            endfamily
            family 47
              task pertinic
                edit MEMBER '47'
            endfamily
            family 48
              task pertinic
                edit MEMBER '48'
            endfamily
            family 49
              task pertinic
                edit MEMBER '49'
            endfamily
            family 50
              task pertinic
                edit MEMBER '50'
            endfamily
          endfamily
          task cp_pert
            trigger pert_ic eq complete
            event 1 
          task wavini
            trigger /emc_36r1/main/12/ref/an eq complete
            edit FSFAMILY 'mc'
          family cv
            trigger ./pert_ic/00 eq complete and ./inidata eq complete and ./wavini eq complete and /emc_36r1/main/12/ref/model_active eq complete
            edit EPSFCLEV '62'
            edit EPSWAMNSTPW '1'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMRESOL 'global50'
            edit EPSFCRES '639'
            edit TSTEP '1200'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit CLASS 'od'
            edit PERSST 'no'
            edit EPSTYPE 'cv'
            edit EMOS_TYPE 'cv'
            edit EPSNLEGS '1'
            edit EPSLEG '1'
            edit EPSINISTEP '360'
            edit FCLENGTH '360'
            edit USER_PRIORITY '73'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            inlimit /emc_36r1/limits:pftotal
            family control_1
              edit MEMBER '1'
              edit EPSMEMBER '1'
              edit SUBFSFAMILY '/cv001'
              task getiniLeg
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit FCNPES '32'
                edit THREADS '4'
                meter step -1 360 360
              task derived_param
                trigger modeleps eq complete
                edit STEPS '0:144:3 144:360:6'
            endfamily
            family control_2
              edit MEMBER '2'
              edit EPSMEMBER '2'
              edit SUBFSFAMILY '/cv002'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              task getiniLeg
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task intHtoL
                trigger getiniLeg eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '1200'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP_PREVIOUS '0'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit FCNPES '32'
                edit THREADS '4'
                meter step -1 360 360
              task derived_param
                trigger modeleps eq complete
                edit STEPS '0:144:3 144:360:6'
            endfamily
          endfamily
          task wavfcdata
            trigger ./getpersSST eq complete and ../../legA/fc/wavini eq complete
            edit EPSCOUPLE '1'
          task getpersSST
            trigger ./inidata eq complete
            edit EPSCOUPLE '1'
            edit EPS_INI_STREAM 'DA'
            edit EPSFCRES '639'
          family pf
            complete /emc_36r1/main/ref/12/events:cf
            trigger ./getpersSST eq complete and ./wavfcdata eq complete and wavini eq complete and ../../ref/model_active eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTYPE 'pf'
            edit EMOS_TYPE 'pf'
            edit USER_PRIORITY '72'
            edit PRIORITY '72'
            edit FCNPES '32'
            edit EPSNLEGS '3'
            edit EPSFCLEV '62'
            edit EPSWAMNSTPW '1'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMRESOL 'global50'
            edit EPSFCRES '639'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit EPSINISTEP '216'
            edit FCLENGTH '240'
            edit TSTEP '1200'
            edit EPSLEG '1'
            edit CLASS 'od'
            inlimit /emc_36r1/limits:pftotal
            meter step -1 240 240
            family 01
              edit EPSMEMBER '1'
              edit MEMBER '1'
              edit SUBFSFAMILY '/pf001'
              task getiniLeg
                trigger ../../pert_ic/01/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 02
              edit EPSMEMBER '2'
              edit MEMBER '2'
              edit SUBFSFAMILY '/pf002'
              task getiniLeg
                trigger ../../pert_ic/02/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 03
              edit EPSMEMBER '3'
              edit MEMBER '3'
              edit SUBFSFAMILY '/pf003'
              task getiniLeg
                trigger ../../pert_ic/03/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 04
              edit EPSMEMBER '4'
              edit MEMBER '4'
              edit SUBFSFAMILY '/pf004'
              task getiniLeg
                trigger ../../pert_ic/04/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 05
              edit EPSMEMBER '5'
              edit MEMBER '5'
              edit SUBFSFAMILY '/pf005'
              task getiniLeg
                trigger ../../pert_ic/05/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 06
              edit EPSMEMBER '6'
              edit MEMBER '6'
              edit SUBFSFAMILY '/pf006'
              task getiniLeg
                trigger ../../pert_ic/06/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 07
              edit EPSMEMBER '7'
              edit MEMBER '7'
              edit SUBFSFAMILY '/pf007'
              task getiniLeg
                trigger ../../pert_ic/07/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 08
              edit EPSMEMBER '8'
              edit MEMBER '8'
              edit SUBFSFAMILY '/pf008'
              task getiniLeg
                trigger ../../pert_ic/08/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 09
              edit EPSMEMBER '9'
              edit MEMBER '9'
              edit SUBFSFAMILY '/pf009'
              task getiniLeg
                trigger ../../pert_ic/09/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 10
              edit EPSMEMBER '10'
              edit MEMBER '10'
              edit SUBFSFAMILY '/pf010'
              task getiniLeg
                trigger ../../pert_ic/10/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 11
              edit EPSMEMBER '11'
              edit MEMBER '11'
              edit SUBFSFAMILY '/pf011'
              task getiniLeg
                trigger ../../pert_ic/11/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 12
              edit EPSMEMBER '12'
              edit MEMBER '12'
              edit SUBFSFAMILY '/pf012'
              task getiniLeg
                trigger ../../pert_ic/12/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 13
              edit EPSMEMBER '13'
              edit MEMBER '13'
              edit SUBFSFAMILY '/pf013'
              task getiniLeg
                trigger ../../pert_ic/13/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 14
              edit EPSMEMBER '14'
              edit MEMBER '14'
              edit SUBFSFAMILY '/pf014'
              task getiniLeg
                trigger ../../pert_ic/14/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 15
              edit EPSMEMBER '15'
              edit MEMBER '15'
              edit SUBFSFAMILY '/pf015'
              task getiniLeg
                trigger ../../pert_ic/15/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 16
              edit EPSMEMBER '16'
              edit MEMBER '16'
              edit SUBFSFAMILY '/pf016'
              task getiniLeg
                trigger ../../pert_ic/16/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 17
              edit EPSMEMBER '17'
              edit MEMBER '17'
              edit SUBFSFAMILY '/pf017'
              task getiniLeg
                trigger ../../pert_ic/17/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 18
              edit EPSMEMBER '18'
              edit MEMBER '18'
              edit SUBFSFAMILY '/pf018'
              task getiniLeg
                trigger ../../pert_ic/18/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 19
              edit EPSMEMBER '19'
              edit MEMBER '19'
              edit SUBFSFAMILY '/pf019'
              task getiniLeg
                trigger ../../pert_ic/19/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 20
              edit EPSMEMBER '20'
              edit MEMBER '20'
              edit SUBFSFAMILY '/pf020'
              task getiniLeg
                trigger ../../pert_ic/20/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 21
              edit EPSMEMBER '21'
              edit MEMBER '21'
              edit SUBFSFAMILY '/pf021'
              task getiniLeg
                trigger ../../pert_ic/21/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 22
              edit EPSMEMBER '22'
              edit MEMBER '22'
              edit SUBFSFAMILY '/pf022'
              task getiniLeg
                trigger ../../pert_ic/22/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 23
              edit EPSMEMBER '23'
              edit MEMBER '23'
              edit SUBFSFAMILY '/pf023'
              task getiniLeg
                trigger ../../pert_ic/23/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 24
              edit EPSMEMBER '24'
              edit MEMBER '24'
              edit SUBFSFAMILY '/pf024'
              task getiniLeg
                trigger ../../pert_ic/24/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 25
              edit EPSMEMBER '25'
              edit MEMBER '25'
              edit SUBFSFAMILY '/pf025'
              task getiniLeg
                trigger ../../pert_ic/25/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 26
              edit EPSMEMBER '26'
              edit MEMBER '26'
              edit SUBFSFAMILY '/pf026'
              task getiniLeg
                trigger ../../pert_ic/26/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 27
              edit EPSMEMBER '27'
              edit MEMBER '27'
              edit SUBFSFAMILY '/pf027'
              task getiniLeg
                trigger ../../pert_ic/27/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 28
              edit EPSMEMBER '28'
              edit MEMBER '28'
              edit SUBFSFAMILY '/pf028'
              task getiniLeg
                trigger ../../pert_ic/28/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 29
              edit EPSMEMBER '29'
              edit MEMBER '29'
              edit SUBFSFAMILY '/pf029'
              task getiniLeg
                trigger ../../pert_ic/29/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 30
              edit EPSMEMBER '30'
              edit MEMBER '30'
              edit SUBFSFAMILY '/pf030'
              task getiniLeg
                trigger ../../pert_ic/30/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 31
              edit EPSMEMBER '31'
              edit MEMBER '31'
              edit SUBFSFAMILY '/pf031'
              task getiniLeg
                trigger ../../pert_ic/31/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 32
              edit EPSMEMBER '32'
              edit MEMBER '32'
              edit SUBFSFAMILY '/pf032'
              task getiniLeg
                trigger ../../pert_ic/32/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 33
              edit EPSMEMBER '33'
              edit MEMBER '33'
              edit SUBFSFAMILY '/pf033'
              task getiniLeg
                trigger ../../pert_ic/33/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 34
              edit EPSMEMBER '34'
              edit MEMBER '34'
              edit SUBFSFAMILY '/pf034'
              task getiniLeg
                trigger ../../pert_ic/34/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 35
              edit EPSMEMBER '35'
              edit MEMBER '35'
              edit SUBFSFAMILY '/pf035'
              task getiniLeg
                trigger ../../pert_ic/35/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 36
              edit EPSMEMBER '36'
              edit MEMBER '36'
              edit SUBFSFAMILY '/pf036'
              task getiniLeg
                trigger ../../pert_ic/36/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 37
              edit EPSMEMBER '37'
              edit MEMBER '37'
              edit SUBFSFAMILY '/pf037'
              task getiniLeg
                trigger ../../pert_ic/37/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 38
              edit EPSMEMBER '38'
              edit MEMBER '38'
              edit SUBFSFAMILY '/pf038'
              task getiniLeg
                trigger ../../pert_ic/38/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 39
              edit EPSMEMBER '39'
              edit MEMBER '39'
              edit SUBFSFAMILY '/pf039'
              task getiniLeg
                trigger ../../pert_ic/39/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 40
              edit EPSMEMBER '40'
              edit MEMBER '40'
              edit SUBFSFAMILY '/pf040'
              task getiniLeg
                trigger ../../pert_ic/40/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 41
              edit EPSMEMBER '41'
              edit MEMBER '41'
              edit SUBFSFAMILY '/pf041'
              task getiniLeg
                trigger ../../pert_ic/41/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 42
              edit EPSMEMBER '42'
              edit MEMBER '42'
              edit SUBFSFAMILY '/pf042'
              task getiniLeg
                trigger ../../pert_ic/42/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 43
              edit EPSMEMBER '43'
              edit MEMBER '43'
              edit SUBFSFAMILY '/pf043'
              task getiniLeg
                trigger ../../pert_ic/43/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 44
              edit EPSMEMBER '44'
              edit MEMBER '44'
              edit SUBFSFAMILY '/pf044'
              task getiniLeg
                trigger ../../pert_ic/44/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 45
              edit EPSMEMBER '45'
              edit MEMBER '45'
              edit SUBFSFAMILY '/pf045'
              task getiniLeg
                trigger ../../pert_ic/45/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 46
              edit EPSMEMBER '46'
              edit MEMBER '46'
              edit SUBFSFAMILY '/pf046'
              task getiniLeg
                trigger ../../pert_ic/46/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 47
              edit EPSMEMBER '47'
              edit MEMBER '47'
              edit SUBFSFAMILY '/pf047'
              task getiniLeg
                trigger ../../pert_ic/47/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 48
              edit EPSMEMBER '48'
              edit MEMBER '48'
              edit SUBFSFAMILY '/pf048'
              task getiniLeg
                trigger ../../pert_ic/48/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 49
              edit EPSMEMBER '49'
              edit MEMBER '49'
              edit SUBFSFAMILY '/pf049'
              task getiniLeg
                trigger ../../pert_ic/49/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 50
              edit EPSMEMBER '50'
              edit MEMBER '50'
              edit SUBFSFAMILY '/pf050'
              task getiniLeg
                trigger ../../pert_ic/50/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
          endfamily
          family cf
            trigger ./pert_ic/00/pertinic eq complete and ./getpersSST eq complete and ./wavfcdata eq complete and /emc_36r1/main/12/ref/model_active eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit MEMBER '0'
            edit EPSMEMBER '0'
            edit SUBFSFAMILY '/cf000'
            edit EPSTYPE 'cf'
            edit EMOS_TYPE 'cf'
            edit USER_PRIORITY '73'
            edit EPSNLEGS '3'
            edit EPSFCLEV '62'
            edit EPSWAMNSTPW '1'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMRESOL 'global50'
            edit EPSFCRES '639'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit EPSINISTEP '216'
            edit FCLENGTH '240'
            edit TSTEP '1200'
            edit EPSLEG '1'
            edit CLASS 'od'
            task getiniLeg
              trigger ../inidata eq complete and ../wavini eq complete
              edit LINIENDA '0'
              edit INIDATEENDA '0'
              edit EPSCOUPLE '0'
              edit N_PERT_AN '10'
              edit EPSCOUPL_A '0'
              edit EPSCOUPL_B '0'
              edit EPSCOUPL_C '0'
            task modeleps
              trigger getiniLeg eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '1'
              edit EPSFCRES '639'
              edit EPSINISTEP '216'
              edit EPSINISTEP_PREVIOUS '0'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit CLASS 'od'
              edit FCNPES '32'
              edit THREADS '4'
              meter step -1 240 240
          endfamily
          family reservation1
            trigger ./pf/01/modeleps eq active or ./pf/01/modeleps eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/smsfiles'
            inlimit /emc_36r1/limits:reservation
            task reserve_cpu
              late -s +00:02 -c +00:02
              edit RESERVED_CPUS '10880'
              edit RESERVE_FOR '48'
              label info ""
            task rm_res
              late -s +00:02 -c +00:02
              trigger ./reserve_cpu eq complete
              edit RESERVATION_LABEL 'emos12'
              label info ""
          endfamily
          family reservation2
            trigger ( ./pf/46/modeleps eq active or ./pf/46/modeleps eq complete) and ( ./pf/47/modeleps eq active or ./pf/47/modeleps eq complete) and ( ./pf/48/modeleps eq active or ./pf/48/modeleps eq complete) and ( ./pf/49/modeleps eq active or ./pf/49/modeleps eq complete) and ( ./pf/50/modeleps eq active or ./pf/50/modeleps eq complete)
            edit SMSFILES '/home/ma/emos/def/emc_36r1/smsfiles'
            inlimit /emc_36r1/limits:reservation
            task reset_class
              late -s +00:02 -c +00:02
              edit RESERVATION_DEFINITION 'default'
              label info ""
            task ms_cpu
              late -s +00:02 -c +00:02
              edit MS_CPUS '12800'
              label info ""
            task reserve_cpu
              late -s +00:02 -c +00:02
              edit RESERVED_CPUS '6400'
              edit RESERVE_FOR '40'
              label info ""
          endfamily
          family reservation3
            trigger ( ./pf eq complete)
            edit SMSFILES '/home/ma/emos/def/emc_36r1/smsfiles'
            inlimit /emc_36r1/limits:reservation
            task reserve_cpu
              late -s +00:02 -c +00:02
              edit RESERVED_CPUS '1024'
              edit RESERVE_FOR '20'
              label info ""
          endfamily
        endfamily
        family fax
          trigger /emc_36r1/main/12/ref/an eq complete
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /limits:linux_cluster
          task ena4
            trigger ../postproc/clusters eq complete and ../postproc/ensms eq complete
            edit ANALYSIS 'NO'
          task plume
            trigger ../postproc/prob eq complete and /emc_36r1/main/12/ref/model eq complete
            edit ANALYSIS 'YES'
          task fax
            trigger plume eq complete and ena4 eq complete
            edit SMSTRIES '1'
          task manfax
            trigger plume eq complete and ena4 eq complete
            edit SMSTRIES '1'
            edit COUNTRY ''
          task faxcheck
            trigger fax eq complete
          task faxcheck2
            trigger fax eq complete
            time 08:00
        endfamily
        family postproc
          trigger ./fc ne queued
          edit STEPS '24/240/12/24 168/168/12/48 240/240/12/72 240/240/12/120'
          inlimit /emc_36r1/limits:postproc
          family prepare
            complete /emc_36r1/main/ref/12/events:cf
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/leg'
            family all
              complete ../../fc eq complete and ./query eq active
              trigger ../../fc eq complete and ( ./query eq queued or ./query eq complete)
              edit KIND 'all'
              edit STEP '-1'
              family t850
                edit PARAM 'T'
                edit LEVEL '850'
                task pp_prep
              endfamily
              family t500
                edit PARAM 'T'
                edit LEVEL '500'
                task pp_prep
              endfamily
              family z1000
                edit PARAM 'Z'
                edit LEVEL '1000'
                task pp_prep
              endfamily
              family z500
                edit PARAM 'Z'
                edit LEVEL '500'
                task pp_prep
              endfamily
            endfamily
            family query
              complete ../../fc eq complete and ../../fc/pf:step le 0
              trigger ( ( ../../fc/pf:step ge 0) and ( ../../fc/cf/modeleps:step ge 0)) or ( ../../fc/pf eq complete and ../../fc/cf eq complete)
              edit KIND 'query'
              family t850
                edit PARAM 'T'
                edit LEVEL '850'
                task pp_prep
                  meter step -1 240 240
              endfamily
              family t500
                edit PARAM 'T'
                edit LEVEL '500'
                task pp_prep
                  meter step -1 240 240
              endfamily
              family z1000
                edit PARAM 'Z'
                edit LEVEL '1000'
                task pp_prep
                  meter step -1 240 240
              endfamily
              family z500
                edit PARAM 'Z'
                edit LEVEL '500'
                task pp_prep
                  meter step -1 240 240
              endfamily
            endfamily
          endfamily
          family ensms
            complete /emc_36r1/main/ref/12/events:cf
            trigger ( ../fc eq complete) and ( ./prepare eq complete)
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/leg'
            family limits
              limit fdb 1
              limit gen 8
            endfamily
            task ensms
              edit ENS_M 'false'
              edit ENS_S 'false'
              edit ENS_MARS 'false'
              event stage
            family 048
              trigger ./ensms:stage or ./ensms eq complete
              edit STEP1 '003'
              edit STEP2 '048'
              edit BY '3'
              edit STEPS '0'
              family T
                edit ENS_M 'false'
                edit ENS_S 'false'
                edit LIST_PARAM 'T'
                inlimit ../limits:gen
                family 850
                  task ensms
                    edit LIST_LEV '850'
                endfamily
                family 500
                  task ensms
                    edit LIST_LEV '500'
                endfamily
              endfamily
              family Z
                edit ENS_M 'false'
                edit ENS_S 'false'
                edit LIST_PARAM 'Z'
                inlimit ../limits:gen
                family 1000
                  task ensms
                    edit LIST_LEV '1000'
                endfamily
                family 500
                  task ensms
                    edit LIST_LEV '500'
                endfamily
              endfamily
              task ensms
                trigger T eq complete and Z eq complete
                edit ENS_MARS 'false'
                inlimit ../limits:fdb
            endfamily
            family 096
              trigger ./ensms:stage or ./ensms eq complete
              edit STEP1 '051'
              edit STEP2 '096'
              edit BY '3'
              edit STEPS '0'
              family T
                edit ENS_M 'false'
                edit ENS_S 'false'
                edit LIST_PARAM 'T'
                inlimit ../limits:gen
                family 850
                  task ensms
                    edit LIST_LEV '850'
                endfamily
                family 500
                  task ensms
                    edit LIST_LEV '500'
                endfamily
              endfamily
              family Z
                edit ENS_M 'false'
                edit ENS_S 'false'
                edit LIST_PARAM 'Z'
                inlimit ../limits:gen
                family 1000
                  task ensms
                    edit LIST_LEV '1000'
                endfamily
                family 500
                  task ensms
                    edit LIST_LEV '500'
                endfamily
              endfamily
              task ensms
                trigger T eq complete and Z eq complete
                edit ENS_MARS 'false'
                inlimit ../limits:fdb
            endfamily
            family 144
              trigger ./ensms:stage or ./ensms eq complete
              edit STEP1 '099'
              edit STEP2 '144'
              edit BY '3'
              edit STEPS '0'
              family T
                edit ENS_M 'false'
                edit ENS_S 'false'
                edit LIST_PARAM 'T'
                inlimit ../limits:gen
                family 850
                  task ensms
                    edit LIST_LEV '850'
                endfamily
                family 500
                  task ensms
                    edit LIST_LEV '500'
                endfamily
              endfamily
              family Z
                edit ENS_M 'false'
                edit ENS_S 'false'
                edit LIST_PARAM 'Z'
                inlimit ../limits:gen
                family 1000
                  task ensms
                    edit LIST_LEV '1000'
                endfamily
                family 500
                  task ensms
                    edit LIST_LEV '500'
                endfamily
              endfamily
              task ensms
                trigger T eq complete and Z eq complete
                edit ENS_MARS 'false'
                inlimit ../limits:fdb
            endfamily
            family 240
              trigger ./ensms:stage or ./ensms eq complete
              edit STEP1 '150'
              edit STEP2 '240'
              edit BY '6'
              edit STEPS '0'
              family T
                edit ENS_M 'false'
                edit ENS_S 'false'
                edit LIST_PARAM 'T'
                inlimit ../limits:gen
                family 850
                  task ensms
                    edit LIST_LEV '850'
                endfamily
                family 500
                  task ensms
                    edit LIST_LEV '500'
                endfamily
              endfamily
              family Z
                edit ENS_M 'false'
                edit ENS_S 'false'
                edit LIST_PARAM 'Z'
                inlimit ../limits:gen
                family 1000
                  task ensms
                    edit LIST_LEV '1000'
                endfamily
                family 500
                  task ensms
                    edit LIST_LEV '500'
                endfamily
              endfamily
              task ensms
                trigger T eq complete and Z eq complete
                edit ENS_MARS 'false'
                inlimit ../limits:fdb
            endfamily
          endfamily
          family prob
            complete /emc_36r1/main/ref/12/events:cf
            trigger /emc_36r1/main/12/legA/fc eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/leg'
            edit STEP1 '12'
            edit STEP2 '240'
            edit BY '12'
            family 24
              edit STEP1 '24'
              edit STEP2 '240'
              edit BY '24'
              family t850
                trigger ../../prepare/all/t850 eq complete
                family 120_240
                  edit STEPS '120 240'
                  task t850
                    event 1 
                endfamily
                family 120_168
                  trigger ./120_240/t850:1 or ./120_240 eq complete
                  edit STEPS '120 168'
                  task t850
                    event 1 
                endfamily
                family 168_240
                  trigger ./120_240/t850:1 or ./120_240 eq complete
                  edit STEPS '168 240'
                  task t850
                    event 1 
                endfamily
              endfamily
              task rain
              task ff
            endfamily
            family 12
              edit STEP1 '12'
              edit STEP2 '228'
              edit BY '24'
              task t850
                trigger ../../prepare/all/t850 eq complete
              task rain
              task ff
            endfamily
            task 2t
            task swh
            task mwp
            task 10fg
              trigger ../dp eq complete
            task prob_fdb
              trigger ./2t eq complete and swh eq complete and mwp eq complete and ./10fg eq complete and ./12 eq complete and ./24 eq complete
          endfamily
          family tc
            complete /emc_36r1/main/ref/12/events:cf
            trigger ../fc eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/tc'
            edit TCINIEXPVER '0046'
            family obsdata
              trigger ( /emc_36r1/main/12/legA/fc/pf eq complete and /emc_36r1/main/12/legA/fc/cf eq complete)
              task retrieve
                late -c +00:10
                trigger ( /emc_36r1/main/12/sv/targets eq complete or /emc_36r1/main/12/sv/targets:sv)
                label info ""
                event notc
              task obs
                trigger retrieve eq complete
            endfamily
            family stage
              trigger ( /emc_36r1/main/12/legA/fc eq complete)
              task an
              task fc
              task pf
              task cf
            endfamily
            family tracking
              trigger obsdata/obs eq complete and stage eq complete
              inlimit /emc_36r1/limits:tc
              family 1
                edit MEMBER '1'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 2
                edit MEMBER '2'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 3
                edit MEMBER '3'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 4
                edit MEMBER '4'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 5
                edit MEMBER '5'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 6
                edit MEMBER '6'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 7
                edit MEMBER '7'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 8
                edit MEMBER '8'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 9
                edit MEMBER '9'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 10
                edit MEMBER '10'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 11
                edit MEMBER '11'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 12
                edit MEMBER '12'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 13
                edit MEMBER '13'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 14
                edit MEMBER '14'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 15
                edit MEMBER '15'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 16
                edit MEMBER '16'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 17
                edit MEMBER '17'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 18
                edit MEMBER '18'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 19
                edit MEMBER '19'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 20
                edit MEMBER '20'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 21
                edit MEMBER '21'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 22
                edit MEMBER '22'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 23
                edit MEMBER '23'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 24
                edit MEMBER '24'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 25
                edit MEMBER '25'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 26
                edit MEMBER '26'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 27
                edit MEMBER '27'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 28
                edit MEMBER '28'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 29
                edit MEMBER '29'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 30
                edit MEMBER '30'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 31
                edit MEMBER '31'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 32
                edit MEMBER '32'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 33
                edit MEMBER '33'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 34
                edit MEMBER '34'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 35
                edit MEMBER '35'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 36
                edit MEMBER '36'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 37
                edit MEMBER '37'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 38
                edit MEMBER '38'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 39
                edit MEMBER '39'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 40
                edit MEMBER '40'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 41
                edit MEMBER '41'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 42
                edit MEMBER '42'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 43
                edit MEMBER '43'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 44
                edit MEMBER '44'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 45
                edit MEMBER '45'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 46
                edit MEMBER '46'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 47
                edit MEMBER '47'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 48
                edit MEMBER '48'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 49
                edit MEMBER '49'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 50
                edit MEMBER '50'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family cf
                edit MEMBER '51'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family oper
                edit MEMBER '52'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
            endfamily
            family post_process
              task to_bufr
                trigger ../tracking eq complete
            endfamily
          endfamily
          family calibrate
            complete /emc_36r1/main/ref/12/events:cf
            trigger /emc_36r1/main/12/ref/an eq complete and ../fc eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/calibrate'
            family onsc
              edit QUEUE 'ns'
              edit QUEUE_EPILOG 'ns'
              edit SMSCMD 'smssubmit %USER% %SCHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %SCHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %SCHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/s1b/emos_esuite/emos_data/0046/log'
              edit SMSOUT '/s1b/emos_esuite/emos_data/0046/log'
              inlimit /limits:c1b
              task feedupcal
              task upcal
                trigger feedupcal eq complete
                edit Z500_EXPVER '0001'
                edit METVIEW 'metview'
              task revertupcal
            endfamily
          endfamily
          family clusters
            complete /emc_36r1/main/ref/12/events:cf
            trigger ./calibrate/onsc eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/clusters'
            edit USER_PRIORITY '81'
            edit QUEUE 'ns'
            edit QUEUE_EPILOG 'ns'
            edit SMSCMD 'smssubmit %USER% %SCHOST% %SMSJOB% %SMSJOBOUT%'
            edit SMSKILL 'smskill %USER% %SCHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
            edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %SCHOST% %SMSRID% %SMSJOB%'
            edit LOGDIR '/s1b/emos_esuite/emos_data/0046/log'
            edit SMSOUT '/s1b/emos_esuite/emos_data/0046/log'
            inlimit /limits:c1b
            task cthresh
            task cluster
              trigger cthresh eq complete and /emc_36r1/main/12/ref/model eq complete
              edit USER_PRIORITY '76'
            task tubcon
              trigger /emc_36r1/main/12/ref/model eq complete
            task tubgen
              late -c +00:20
              trigger tubcon eq complete
          endfamily
          family efi
            complete ( /emc_36r1/main/ref/12/events:cf) or ( /emc_36r1/main/ref/12/events:noefi)
            trigger /emc_36r1/main/12/legA/postproc/dp eq complete
            edit HC_VERSION '0046'
            edit HC_EXPVER '0046'
            edit REF_VERSION '0046'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/efi_hdc'
            edit NC_DAYS '31'
            edit EPSLEG '1'
            inlimit /emc_36r1/limits:efi
            task stage
              edit NC_DAYS '31'
            family 000_120
              trigger ./stage eq complete
              edit ACC '120'
              edit EFI_STEP1 '000'
              edit EFI_STEP2 '120'
              family tp
                edit RF_NAME 'tp'
                task extindex
              endfamily
            endfamily
            family 024_144
              trigger ./stage eq complete
              edit ACC '120'
              edit EFI_STEP1 '024'
              edit EFI_STEP2 '144'
              family tp
                edit RF_NAME 'tp'
                task extindex
              endfamily
            endfamily
            family 000_240
              trigger ./stage eq complete
              edit ACC '240'
              edit EFI_STEP1 '000'
              edit EFI_STEP2 '240'
              family tp
                edit RF_NAME 'tp'
                task extindex
              endfamily
            endfamily
            family 012_036
              trigger ./stage eq complete
              edit ACC '24'
              edit EFI_STEP1 '012'
              edit EFI_STEP2 '036'
              family 2t
                edit RF_NAME '2t'
                task extindex
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task extindex
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task extindex
              endfamily
              family tp
                edit RF_NAME 'tp'
                task extindex
              endfamily
            endfamily
            family 036_060
              trigger ./stage eq complete
              edit ACC '24'
              edit EFI_STEP1 '036'
              edit EFI_STEP2 '060'
              family 2t
                edit RF_NAME '2t'
                task extindex
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task extindex
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task extindex
              endfamily
              family tp
                edit RF_NAME 'tp'
                task extindex
              endfamily
            endfamily
            family 060_084
              trigger ./stage eq complete
              edit ACC '24'
              edit EFI_STEP1 '060'
              edit EFI_STEP2 '084'
              family 2t
                edit RF_NAME '2t'
                task extindex
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task extindex
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task extindex
              endfamily
              family tp
                edit RF_NAME 'tp'
                task extindex
              endfamily
            endfamily
            family 084_108
              trigger ./stage eq complete
              edit ACC '24'
              edit EFI_STEP1 '084'
              edit EFI_STEP2 '108'
              family 2t
                edit RF_NAME '2t'
                task extindex
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task extindex
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task extindex
              endfamily
              family tp
                edit RF_NAME 'tp'
                task extindex
              endfamily
            endfamily
            family 108_132
              trigger ./stage eq complete
              edit ACC '24'
              edit EFI_STEP1 '108'
              edit EFI_STEP2 '132'
              family 2t
                edit RF_NAME '2t'
                task extindex
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task extindex
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task extindex
              endfamily
              family tp
                edit RF_NAME 'tp'
                task extindex
              endfamily
            endfamily
          endfamily
          task efi2fdb
            complete ( /emc_36r1/main/ref/12/events:cf) or ( /emc_36r1/main/ref/12/events:noefi)
            trigger ./efi eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/efi_hdc'
            inlimit /emc_36r1/main/ref:mutex
            event none
          family dp
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit STEPS '0:144:3 144:240:6'
            inlimit /emc_36r1/limits:dp
            family cf
              trigger ../../fc/cf eq complete
              edit EMOS_TYPE 'cf'
              task derived_param
                edit MEMBER 'off'
            endfamily
            family pf
              complete /emc_36r1/main/ref/12/events:cf
              edit EMOS_TYPE 'pf'
              family 01
                trigger ../../../fc/pf/01 eq complete
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task derived_param
              endfamily
              family 02
                trigger ../../../fc/pf/02 eq complete
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task derived_param
              endfamily
              family 03
                trigger ../../../fc/pf/03 eq complete
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task derived_param
              endfamily
              family 04
                trigger ../../../fc/pf/04 eq complete
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task derived_param
              endfamily
              family 05
                trigger ../../../fc/pf/05 eq complete
                edit EPSMEMBER '5'
                edit MEMBER '5'
                edit SUBFSFAMILY '/pf005'
                task derived_param
              endfamily
              family 06
                trigger ../../../fc/pf/06 eq complete
                edit EPSMEMBER '6'
                edit MEMBER '6'
                edit SUBFSFAMILY '/pf006'
                task derived_param
              endfamily
              family 07
                trigger ../../../fc/pf/07 eq complete
                edit EPSMEMBER '7'
                edit MEMBER '7'
                edit SUBFSFAMILY '/pf007'
                task derived_param
              endfamily
              family 08
                trigger ../../../fc/pf/08 eq complete
                edit EPSMEMBER '8'
                edit MEMBER '8'
                edit SUBFSFAMILY '/pf008'
                task derived_param
              endfamily
              family 09
                trigger ../../../fc/pf/09 eq complete
                edit EPSMEMBER '9'
                edit MEMBER '9'
                edit SUBFSFAMILY '/pf009'
                task derived_param
              endfamily
              family 10
                trigger ../../../fc/pf/10 eq complete
                edit EPSMEMBER '10'
                edit MEMBER '10'
                edit SUBFSFAMILY '/pf010'
                task derived_param
              endfamily
              family 11
                trigger ../../../fc/pf/11 eq complete
                edit EPSMEMBER '11'
                edit MEMBER '11'
                edit SUBFSFAMILY '/pf011'
                task derived_param
              endfamily
              family 12
                trigger ../../../fc/pf/12 eq complete
                edit EPSMEMBER '12'
                edit MEMBER '12'
                edit SUBFSFAMILY '/pf012'
                task derived_param
              endfamily
              family 13
                trigger ../../../fc/pf/13 eq complete
                edit EPSMEMBER '13'
                edit MEMBER '13'
                edit SUBFSFAMILY '/pf013'
                task derived_param
              endfamily
              family 14
                trigger ../../../fc/pf/14 eq complete
                edit EPSMEMBER '14'
                edit MEMBER '14'
                edit SUBFSFAMILY '/pf014'
                task derived_param
              endfamily
              family 15
                trigger ../../../fc/pf/15 eq complete
                edit EPSMEMBER '15'
                edit MEMBER '15'
                edit SUBFSFAMILY '/pf015'
                task derived_param
              endfamily
              family 16
                trigger ../../../fc/pf/16 eq complete
                edit EPSMEMBER '16'
                edit MEMBER '16'
                edit SUBFSFAMILY '/pf016'
                task derived_param
              endfamily
              family 17
                trigger ../../../fc/pf/17 eq complete
                edit EPSMEMBER '17'
                edit MEMBER '17'
                edit SUBFSFAMILY '/pf017'
                task derived_param
              endfamily
              family 18
                trigger ../../../fc/pf/18 eq complete
                edit EPSMEMBER '18'
                edit MEMBER '18'
                edit SUBFSFAMILY '/pf018'
                task derived_param
              endfamily
              family 19
                trigger ../../../fc/pf/19 eq complete
                edit EPSMEMBER '19'
                edit MEMBER '19'
                edit SUBFSFAMILY '/pf019'
                task derived_param
              endfamily
              family 20
                trigger ../../../fc/pf/20 eq complete
                edit EPSMEMBER '20'
                edit MEMBER '20'
                edit SUBFSFAMILY '/pf020'
                task derived_param
              endfamily
              family 21
                trigger ../../../fc/pf/21 eq complete
                edit EPSMEMBER '21'
                edit MEMBER '21'
                edit SUBFSFAMILY '/pf021'
                task derived_param
              endfamily
              family 22
                trigger ../../../fc/pf/22 eq complete
                edit EPSMEMBER '22'
                edit MEMBER '22'
                edit SUBFSFAMILY '/pf022'
                task derived_param
              endfamily
              family 23
                trigger ../../../fc/pf/23 eq complete
                edit EPSMEMBER '23'
                edit MEMBER '23'
                edit SUBFSFAMILY '/pf023'
                task derived_param
              endfamily
              family 24
                trigger ../../../fc/pf/24 eq complete
                edit EPSMEMBER '24'
                edit MEMBER '24'
                edit SUBFSFAMILY '/pf024'
                task derived_param
              endfamily
              family 25
                trigger ../../../fc/pf/25 eq complete
                edit EPSMEMBER '25'
                edit MEMBER '25'
                edit SUBFSFAMILY '/pf025'
                task derived_param
              endfamily
              family 26
                trigger ../../../fc/pf/26 eq complete
                edit EPSMEMBER '26'
                edit MEMBER '26'
                edit SUBFSFAMILY '/pf026'
                task derived_param
              endfamily
              family 27
                trigger ../../../fc/pf/27 eq complete
                edit EPSMEMBER '27'
                edit MEMBER '27'
                edit SUBFSFAMILY '/pf027'
                task derived_param
              endfamily
              family 28
                trigger ../../../fc/pf/28 eq complete
                edit EPSMEMBER '28'
                edit MEMBER '28'
                edit SUBFSFAMILY '/pf028'
                task derived_param
              endfamily
              family 29
                trigger ../../../fc/pf/29 eq complete
                edit EPSMEMBER '29'
                edit MEMBER '29'
                edit SUBFSFAMILY '/pf029'
                task derived_param
              endfamily
              family 30
                trigger ../../../fc/pf/30 eq complete
                edit EPSMEMBER '30'
                edit MEMBER '30'
                edit SUBFSFAMILY '/pf030'
                task derived_param
              endfamily
              family 31
                trigger ../../../fc/pf/31 eq complete
                edit EPSMEMBER '31'
                edit MEMBER '31'
                edit SUBFSFAMILY '/pf031'
                task derived_param
              endfamily
              family 32
                trigger ../../../fc/pf/32 eq complete
                edit EPSMEMBER '32'
                edit MEMBER '32'
                edit SUBFSFAMILY '/pf032'
                task derived_param
              endfamily
              family 33
                trigger ../../../fc/pf/33 eq complete
                edit EPSMEMBER '33'
                edit MEMBER '33'
                edit SUBFSFAMILY '/pf033'
                task derived_param
              endfamily
              family 34
                trigger ../../../fc/pf/34 eq complete
                edit EPSMEMBER '34'
                edit MEMBER '34'
                edit SUBFSFAMILY '/pf034'
                task derived_param
              endfamily
              family 35
                trigger ../../../fc/pf/35 eq complete
                edit EPSMEMBER '35'
                edit MEMBER '35'
                edit SUBFSFAMILY '/pf035'
                task derived_param
              endfamily
              family 36
                trigger ../../../fc/pf/36 eq complete
                edit EPSMEMBER '36'
                edit MEMBER '36'
                edit SUBFSFAMILY '/pf036'
                task derived_param
              endfamily
              family 37
                trigger ../../../fc/pf/37 eq complete
                edit EPSMEMBER '37'
                edit MEMBER '37'
                edit SUBFSFAMILY '/pf037'
                task derived_param
              endfamily
              family 38
                trigger ../../../fc/pf/38 eq complete
                edit EPSMEMBER '38'
                edit MEMBER '38'
                edit SUBFSFAMILY '/pf038'
                task derived_param
              endfamily
              family 39
                trigger ../../../fc/pf/39 eq complete
                edit EPSMEMBER '39'
                edit MEMBER '39'
                edit SUBFSFAMILY '/pf039'
                task derived_param
              endfamily
              family 40
                trigger ../../../fc/pf/40 eq complete
                edit EPSMEMBER '40'
                edit MEMBER '40'
                edit SUBFSFAMILY '/pf040'
                task derived_param
              endfamily
              family 41
                trigger ../../../fc/pf/41 eq complete
                edit EPSMEMBER '41'
                edit MEMBER '41'
                edit SUBFSFAMILY '/pf041'
                task derived_param
              endfamily
              family 42
                trigger ../../../fc/pf/42 eq complete
                edit EPSMEMBER '42'
                edit MEMBER '42'
                edit SUBFSFAMILY '/pf042'
                task derived_param
              endfamily
              family 43
                trigger ../../../fc/pf/43 eq complete
                edit EPSMEMBER '43'
                edit MEMBER '43'
                edit SUBFSFAMILY '/pf043'
                task derived_param
              endfamily
              family 44
                trigger ../../../fc/pf/44 eq complete
                edit EPSMEMBER '44'
                edit MEMBER '44'
                edit SUBFSFAMILY '/pf044'
                task derived_param
              endfamily
              family 45
                trigger ../../../fc/pf/45 eq complete
                edit EPSMEMBER '45'
                edit MEMBER '45'
                edit SUBFSFAMILY '/pf045'
                task derived_param
              endfamily
              family 46
                trigger ../../../fc/pf/46 eq complete
                edit EPSMEMBER '46'
                edit MEMBER '46'
                edit SUBFSFAMILY '/pf046'
                task derived_param
              endfamily
              family 47
                trigger ../../../fc/pf/47 eq complete
                edit EPSMEMBER '47'
                edit MEMBER '47'
                edit SUBFSFAMILY '/pf047'
                task derived_param
              endfamily
              family 48
                trigger ../../../fc/pf/48 eq complete
                edit EPSMEMBER '48'
                edit MEMBER '48'
                edit SUBFSFAMILY '/pf048'
                task derived_param
              endfamily
              family 49
                trigger ../../../fc/pf/49 eq complete
                edit EPSMEMBER '49'
                edit MEMBER '49'
                edit SUBFSFAMILY '/pf049'
                task derived_param
              endfamily
              family 50
                trigger ../../../fc/pf/50 eq complete
                edit EPSMEMBER '50'
                edit MEMBER '50'
                edit SUBFSFAMILY '/pf050'
                task derived_param
              endfamily
            endfamily
          endfamily
          family extra
            trigger ./dp eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit STEPS '0:144:3 144:240:6'
            inlimit /emc_36r1/limits:extra
            family cf
              edit EMOS_TYPE 'cf'
              task extrafields
                edit EPSFCRES '639'
                edit FCLENGTH '240'
                edit FCLENGTH0 '0'
                edit RUNVARFC '1'
            endfamily
            family pf
              complete /emc_36r1/main/ref/12/events:cf
              edit EMOS_TYPE 'pf'
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 05
                edit EPSMEMBER '5'
                edit MEMBER '5'
                edit SUBFSFAMILY '/pf005'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 06
                edit EPSMEMBER '6'
                edit MEMBER '6'
                edit SUBFSFAMILY '/pf006'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 07
                edit EPSMEMBER '7'
                edit MEMBER '7'
                edit SUBFSFAMILY '/pf007'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 08
                edit EPSMEMBER '8'
                edit MEMBER '8'
                edit SUBFSFAMILY '/pf008'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 09
                edit EPSMEMBER '9'
                edit MEMBER '9'
                edit SUBFSFAMILY '/pf009'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 10
                edit EPSMEMBER '10'
                edit MEMBER '10'
                edit SUBFSFAMILY '/pf010'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 11
                edit EPSMEMBER '11'
                edit MEMBER '11'
                edit SUBFSFAMILY '/pf011'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 12
                edit EPSMEMBER '12'
                edit MEMBER '12'
                edit SUBFSFAMILY '/pf012'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 13
                edit EPSMEMBER '13'
                edit MEMBER '13'
                edit SUBFSFAMILY '/pf013'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 14
                edit EPSMEMBER '14'
                edit MEMBER '14'
                edit SUBFSFAMILY '/pf014'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 15
                edit EPSMEMBER '15'
                edit MEMBER '15'
                edit SUBFSFAMILY '/pf015'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 16
                edit EPSMEMBER '16'
                edit MEMBER '16'
                edit SUBFSFAMILY '/pf016'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 17
                edit EPSMEMBER '17'
                edit MEMBER '17'
                edit SUBFSFAMILY '/pf017'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 18
                edit EPSMEMBER '18'
                edit MEMBER '18'
                edit SUBFSFAMILY '/pf018'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 19
                edit EPSMEMBER '19'
                edit MEMBER '19'
                edit SUBFSFAMILY '/pf019'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 20
                edit EPSMEMBER '20'
                edit MEMBER '20'
                edit SUBFSFAMILY '/pf020'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 21
                edit EPSMEMBER '21'
                edit MEMBER '21'
                edit SUBFSFAMILY '/pf021'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 22
                edit EPSMEMBER '22'
                edit MEMBER '22'
                edit SUBFSFAMILY '/pf022'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 23
                edit EPSMEMBER '23'
                edit MEMBER '23'
                edit SUBFSFAMILY '/pf023'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 24
                edit EPSMEMBER '24'
                edit MEMBER '24'
                edit SUBFSFAMILY '/pf024'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 25
                edit EPSMEMBER '25'
                edit MEMBER '25'
                edit SUBFSFAMILY '/pf025'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 26
                edit EPSMEMBER '26'
                edit MEMBER '26'
                edit SUBFSFAMILY '/pf026'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 27
                edit EPSMEMBER '27'
                edit MEMBER '27'
                edit SUBFSFAMILY '/pf027'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 28
                edit EPSMEMBER '28'
                edit MEMBER '28'
                edit SUBFSFAMILY '/pf028'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 29
                edit EPSMEMBER '29'
                edit MEMBER '29'
                edit SUBFSFAMILY '/pf029'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 30
                edit EPSMEMBER '30'
                edit MEMBER '30'
                edit SUBFSFAMILY '/pf030'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 31
                edit EPSMEMBER '31'
                edit MEMBER '31'
                edit SUBFSFAMILY '/pf031'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 32
                edit EPSMEMBER '32'
                edit MEMBER '32'
                edit SUBFSFAMILY '/pf032'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 33
                edit EPSMEMBER '33'
                edit MEMBER '33'
                edit SUBFSFAMILY '/pf033'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 34
                edit EPSMEMBER '34'
                edit MEMBER '34'
                edit SUBFSFAMILY '/pf034'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 35
                edit EPSMEMBER '35'
                edit MEMBER '35'
                edit SUBFSFAMILY '/pf035'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 36
                edit EPSMEMBER '36'
                edit MEMBER '36'
                edit SUBFSFAMILY '/pf036'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 37
                edit EPSMEMBER '37'
                edit MEMBER '37'
                edit SUBFSFAMILY '/pf037'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 38
                edit EPSMEMBER '38'
                edit MEMBER '38'
                edit SUBFSFAMILY '/pf038'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 39
                edit EPSMEMBER '39'
                edit MEMBER '39'
                edit SUBFSFAMILY '/pf039'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 40
                edit EPSMEMBER '40'
                edit MEMBER '40'
                edit SUBFSFAMILY '/pf040'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 41
                edit EPSMEMBER '41'
                edit MEMBER '41'
                edit SUBFSFAMILY '/pf041'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 42
                edit EPSMEMBER '42'
                edit MEMBER '42'
                edit SUBFSFAMILY '/pf042'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 43
                edit EPSMEMBER '43'
                edit MEMBER '43'
                edit SUBFSFAMILY '/pf043'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 44
                edit EPSMEMBER '44'
                edit MEMBER '44'
                edit SUBFSFAMILY '/pf044'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 45
                edit EPSMEMBER '45'
                edit MEMBER '45'
                edit SUBFSFAMILY '/pf045'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 46
                edit EPSMEMBER '46'
                edit MEMBER '46'
                edit SUBFSFAMILY '/pf046'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 47
                edit EPSMEMBER '47'
                edit MEMBER '47'
                edit SUBFSFAMILY '/pf047'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 48
                edit EPSMEMBER '48'
                edit MEMBER '48'
                edit SUBFSFAMILY '/pf048'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 49
                edit EPSMEMBER '49'
                edit MEMBER '49'
                edit SUBFSFAMILY '/pf049'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 50
                edit EPSMEMBER '50'
                edit MEMBER '50'
                edit SUBFSFAMILY '/pf050'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
            endfamily
          endfamily
        endfamily
        family prod
          edit MEM '3000'
          edit ACCOUNT 'oeenfopg'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/prod'
          edit EMOS_DOMAIN 'G'
          edit EMOS_TYPE 'cp'
          edit EMOS_PROJECT '12'
          edit PROD_TASK '1'
          edit USER_PRIORITY '90'
          edit SMSURLBASE 'http://intra.ecmwf.int/plots/d/intra/'
          edit SMSURL 'sub=fixed/'
          inlimit /limits:prod
          family eps
            trigger ( ../../legA/prod/getreq/eps eq complete) and ( ../../legA/prod/getreq/bitmap:001 or ../../legA/prod/getreq/bitmap eq complete) and ../postproc/dp eq complete
            inlimit /emc_36r1/limits:prodeps
            family 0000
              edit EMOS_TIME_STEP_H '0000'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0003
              edit EMOS_TIME_STEP_H '0003'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0006
              edit EMOS_TIME_STEP_H '0006'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0009
              edit EMOS_TIME_STEP_H '0009'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0012
              edit EMOS_TIME_STEP_H '0012'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0015
              edit EMOS_TIME_STEP_H '0015'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0018
              edit EMOS_TIME_STEP_H '0018'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0021
              edit EMOS_TIME_STEP_H '0021'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0024
              edit EMOS_TIME_STEP_H '0024'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0027
              edit EMOS_TIME_STEP_H '0027'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0030
              edit EMOS_TIME_STEP_H '0030'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0033
              edit EMOS_TIME_STEP_H '0033'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0036
              edit EMOS_TIME_STEP_H '0036'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0039
              edit EMOS_TIME_STEP_H '0039'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0042
              edit EMOS_TIME_STEP_H '0042'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0045
              edit EMOS_TIME_STEP_H '0045'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0048
              edit EMOS_TIME_STEP_H '0048'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0051
              edit EMOS_TIME_STEP_H '0051'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0054
              edit EMOS_TIME_STEP_H '0054'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0057
              edit EMOS_TIME_STEP_H '0057'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0060
              edit EMOS_TIME_STEP_H '0060'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0063
              edit EMOS_TIME_STEP_H '0063'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0066
              edit EMOS_TIME_STEP_H '0066'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0069
              edit EMOS_TIME_STEP_H '0069'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0072
              edit EMOS_TIME_STEP_H '0072'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0075
              edit EMOS_TIME_STEP_H '0075'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0078
              edit EMOS_TIME_STEP_H '0078'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0081
              edit EMOS_TIME_STEP_H '0081'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0084
              edit EMOS_TIME_STEP_H '0084'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0087
              edit EMOS_TIME_STEP_H '0087'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0090
              edit EMOS_TIME_STEP_H '0090'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0093
              edit EMOS_TIME_STEP_H '0093'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0096
              edit EMOS_TIME_STEP_H '0096'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0099
              edit EMOS_TIME_STEP_H '0099'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0102
              edit EMOS_TIME_STEP_H '0102'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0105
              edit EMOS_TIME_STEP_H '0105'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0108
              edit EMOS_TIME_STEP_H '0108'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0111
              edit EMOS_TIME_STEP_H '0111'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0114
              edit EMOS_TIME_STEP_H '0114'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0117
              edit EMOS_TIME_STEP_H '0117'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0120
              edit EMOS_TIME_STEP_H '0120'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0123
              edit EMOS_TIME_STEP_H '0123'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0126
              edit EMOS_TIME_STEP_H '0126'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0129
              edit EMOS_TIME_STEP_H '0129'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0132
              edit EMOS_TIME_STEP_H '0132'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0135
              edit EMOS_TIME_STEP_H '0135'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0138
              edit EMOS_TIME_STEP_H '0138'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0141
              edit EMOS_TIME_STEP_H '0141'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0144
              edit EMOS_TIME_STEP_H '0144'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0150
              edit EMOS_TIME_STEP_H '0150'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0156
              edit EMOS_TIME_STEP_H '0156'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0162
              edit EMOS_TIME_STEP_H '0162'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0168
              edit EMOS_TIME_STEP_H '0168'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0174
              edit EMOS_TIME_STEP_H '0174'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0180
              edit EMOS_TIME_STEP_H '0180'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0186
              edit EMOS_TIME_STEP_H '0186'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0192
              edit EMOS_TIME_STEP_H '0192'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0198
              edit EMOS_TIME_STEP_H '0198'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0204
              edit EMOS_TIME_STEP_H '0204'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0210
              edit EMOS_TIME_STEP_H '0210'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0216
              edit EMOS_TIME_STEP_H '0216'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0222
              edit EMOS_TIME_STEP_H '0222'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0228
              edit EMOS_TIME_STEP_H '0228'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0234
              edit EMOS_TIME_STEP_H '0234'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0240
              edit EMOS_TIME_STEP_H '0240'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
          endfamily
          family wave
            trigger ( ../../legA/prod/getreq/wave_eps:001 or ../../legA/prod/getreq/wave_eps eq complete) and ( ../../legA/prod/getreq/bitmap:001 or ../../legA/prod/getreq/bitmap eq complete) and ../postproc/dp eq complete
            edit PGNPES '2'
            edit URL 'GWAEF/12'
            inlimit /emc_36r1/limits:prodwave
            family 0000
              edit EMOS_TIME_STEP_H '0000'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0003
              edit EMOS_TIME_STEP_H '0003'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0006
              edit EMOS_TIME_STEP_H '0006'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0009
              edit EMOS_TIME_STEP_H '0009'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0012
              edit EMOS_TIME_STEP_H '0012'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0015
              edit EMOS_TIME_STEP_H '0015'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0018
              edit EMOS_TIME_STEP_H '0018'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0021
              edit EMOS_TIME_STEP_H '0021'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0024
              edit EMOS_TIME_STEP_H '0024'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0027
              edit EMOS_TIME_STEP_H '0027'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0030
              edit EMOS_TIME_STEP_H '0030'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0033
              edit EMOS_TIME_STEP_H '0033'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0036
              edit EMOS_TIME_STEP_H '0036'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0039
              edit EMOS_TIME_STEP_H '0039'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0042
              edit EMOS_TIME_STEP_H '0042'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0045
              edit EMOS_TIME_STEP_H '0045'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0048
              edit EMOS_TIME_STEP_H '0048'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0051
              edit EMOS_TIME_STEP_H '0051'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0054
              edit EMOS_TIME_STEP_H '0054'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0057
              edit EMOS_TIME_STEP_H '0057'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0060
              edit EMOS_TIME_STEP_H '0060'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0063
              edit EMOS_TIME_STEP_H '0063'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0066
              edit EMOS_TIME_STEP_H '0066'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0069
              edit EMOS_TIME_STEP_H '0069'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0072
              edit EMOS_TIME_STEP_H '0072'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0075
              edit EMOS_TIME_STEP_H '0075'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0078
              edit EMOS_TIME_STEP_H '0078'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0081
              edit EMOS_TIME_STEP_H '0081'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0084
              edit EMOS_TIME_STEP_H '0084'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0087
              edit EMOS_TIME_STEP_H '0087'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0090
              edit EMOS_TIME_STEP_H '0090'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0093
              edit EMOS_TIME_STEP_H '0093'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0096
              edit EMOS_TIME_STEP_H '0096'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0099
              edit EMOS_TIME_STEP_H '0099'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0102
              edit EMOS_TIME_STEP_H '0102'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0105
              edit EMOS_TIME_STEP_H '0105'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0108
              edit EMOS_TIME_STEP_H '0108'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0111
              edit EMOS_TIME_STEP_H '0111'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0114
              edit EMOS_TIME_STEP_H '0114'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0117
              edit EMOS_TIME_STEP_H '0117'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0120
              edit EMOS_TIME_STEP_H '0120'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0123
              edit EMOS_TIME_STEP_H '0123'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0126
              edit EMOS_TIME_STEP_H '0126'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0129
              edit EMOS_TIME_STEP_H '0129'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0132
              edit EMOS_TIME_STEP_H '0132'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0135
              edit EMOS_TIME_STEP_H '0135'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0138
              edit EMOS_TIME_STEP_H '0138'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0141
              edit EMOS_TIME_STEP_H '0141'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0144
              edit EMOS_TIME_STEP_H '0144'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0150
              edit EMOS_TIME_STEP_H '0150'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0156
              edit EMOS_TIME_STEP_H '0156'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0162
              edit EMOS_TIME_STEP_H '0162'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0168
              edit EMOS_TIME_STEP_H '0168'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0174
              edit EMOS_TIME_STEP_H '0174'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0180
              edit EMOS_TIME_STEP_H '0180'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0186
              edit EMOS_TIME_STEP_H '0186'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0192
              edit EMOS_TIME_STEP_H '0192'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0198
              edit EMOS_TIME_STEP_H '0198'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0204
              edit EMOS_TIME_STEP_H '0204'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0210
              edit EMOS_TIME_STEP_H '0210'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0216
              edit EMOS_TIME_STEP_H '0216'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0222
              edit EMOS_TIME_STEP_H '0222'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0228
              edit EMOS_TIME_STEP_H '0228'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0234
              edit EMOS_TIME_STEP_H '0234'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0240
              edit EMOS_TIME_STEP_H '0240'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
          endfamily
          family getreq
            trigger ../fc/pf eq active or ../fc/pf/01 eq complete
            edit EMOS_TYPE 'cm'
            edit EMOS_DOMAIN 'G'
            edit EMOS_PROJECT '12'
            edit PROD_TASK '1'
            edit SMSTRIES '1'
            edit EMOS_STREAM 'ENFO'
            task collectreq
              edit QUEUE 'diss'
              edit USER 'emos'
              edit WSHOST 'diss'
              edit QUEUE_EPILOG 'emos-exp'
              edit LOGDIR '/diss/log'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit SMSOUT '/vol/emos_nc/output'
              edit EMOS_BASE '12'
            task eps
              late -s +00:03 -c +00:10
              trigger collectreq eq complete
              label info ""
              event 1 
            task collectwave
              trigger collectreq eq complete
              edit QUEUE 'diss'
              edit USER 'emos'
              edit WSHOST 'diss'
              edit QUEUE_EPILOG 'emos-exp'
              edit LOGDIR '/diss/log'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit SMSOUT '/vol/emos_nc/output'
              edit EMOS_STREAM 'WAEF'
              edit EMOS_BASE '12'
            task wave_eps
              late -s +00:03 -c +00:10
              trigger collectwave eq complete
              edit EMOS_STREAM 'WAEF'
              edit USER_PRIORITY '89'
              label info ""
              event 1 
            task collectbitmap
              trigger collectwave eq complete
              edit QUEUE 'diss'
              edit USER 'emos'
              edit WSHOST 'diss'
              edit QUEUE_EPILOG 'emos-exp'
              edit LOGDIR '/diss/log'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit SMSOUT '/vol/emos_nc/output'
            task bitmap
              late -s +00:03 -c +00:10
              trigger collectbitmap eq complete and ( eps:001 or eps eq complete)
              event 1 
          endfamily
          task enfo_tc
            late -s +00:05 -c +00:15
            trigger ( ./getreq/eps:001 or ./getreq/eps eq complete) and ../postproc/tc eq complete
            edit PROD_TASK 't'
            edit EMOS_TYPE 'tf'
            edit EMOS_TIME_STEP_H '0240'
            edit PGNPES '1'
          task oper_tc
            late -s +00:05 -c +00:15
            trigger ./enfo_tc eq complete
            edit PROD_TASK 't'
            edit EMOS_TYPE 'tf'
            edit EMOS_TIME_STEP_H '0240'
            edit PGNPES '1'
            edit EMOS_STREAM 'OPER'
          task prod_wparam
            late -s +00:05 -c +00:15
            trigger ( ../../legA/prod/getreq/eps eq complete) and ../postproc/dp eq complete and /e_36r1/main/12/prod/240/prod_wparam eq complete
            edit PROD_TASK 'w'
            edit EMOS_TYPE 'wp'
            edit EMOS_TIME_STEP_H '0240'
            edit PGNPES '52'
          task enfo_postproc
            late -s +00:05 -c +00:15
            trigger ../postproc eq complete and ( ../../legA/prod/getreq/eps eq complete) and ../../legA/postproc/efi eq complete and ../../legA/postproc/clusters eq complete
            edit PROD_TASK 'p'
            edit EMOS_TYPE 'ep'
            edit EMOS_TIME_STEP_H '0240'
            edit PGNPES '21'
          task waef_postproc
            late -s +00:05 -c +00:15
            trigger ( ../../legA/prod/getreq/wave_eps:001 or ../../legA/prod/getreq/wave_eps eq complete) and ../postproc/prob eq complete
            edit PROD_TASK 'p'
            edit EMOS_TYPE 'pw'
            edit EMOS_STREAM 'WE'
            edit EMOS_TIME_STEP_H '0240'
            edit PGNPES '1'
        endfamily
        family prod2diss
          edit USER_PRIORITY '90'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/prod'
          edit EMOS_DOMAIN 'G'
          edit EMOS_TYPE 'ef'
          edit EMOS_PROJECT '12'
          edit PROD_TASK '1'
          edit SMSURLBASE 'http://diss.ecmwf.int/do/monitoring/summary'
          edit SMSURL 'GENFO/12'
          task operation_is_late
            event yes
          family eps
            family 0000
              edit EMOS_TIME_STEP_H '0000'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0000/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0003
              edit EMOS_TIME_STEP_H '0003'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0003/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0000/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0006
              edit EMOS_TIME_STEP_H '0006'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0006/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0003/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0009
              edit EMOS_TIME_STEP_H '0009'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0009/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0006/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0012
              edit EMOS_TIME_STEP_H '0012'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0012/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0009/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0015
              edit EMOS_TIME_STEP_H '0015'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0015/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0012/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0018
              edit EMOS_TIME_STEP_H '0018'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0018/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0015/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0021
              edit EMOS_TIME_STEP_H '0021'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0021/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0018/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0024
              edit EMOS_TIME_STEP_H '0024'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0024/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0021/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0027
              edit EMOS_TIME_STEP_H '0027'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0027/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0024/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0030
              edit EMOS_TIME_STEP_H '0030'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0030/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0027/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0033
              edit EMOS_TIME_STEP_H '0033'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0033/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0030/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0036
              edit EMOS_TIME_STEP_H '0036'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0036/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0033/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0039
              edit EMOS_TIME_STEP_H '0039'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0039/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0036/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0042
              edit EMOS_TIME_STEP_H '0042'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0042/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0039/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0045
              edit EMOS_TIME_STEP_H '0045'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0045/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0042/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0048
              edit EMOS_TIME_STEP_H '0048'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0048/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0045/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0051
              edit EMOS_TIME_STEP_H '0051'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0051/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0048/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0054
              edit EMOS_TIME_STEP_H '0054'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0054/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0051/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0057
              edit EMOS_TIME_STEP_H '0057'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0057/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0054/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0060
              edit EMOS_TIME_STEP_H '0060'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0060/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0057/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0063
              edit EMOS_TIME_STEP_H '0063'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0063/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0060/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0066
              edit EMOS_TIME_STEP_H '0066'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0066/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0063/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0069
              edit EMOS_TIME_STEP_H '0069'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0069/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0066/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0072
              edit EMOS_TIME_STEP_H '0072'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0072/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0069/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0075
              edit EMOS_TIME_STEP_H '0075'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0075/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0072/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0078
              edit EMOS_TIME_STEP_H '0078'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0078/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0075/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0081
              edit EMOS_TIME_STEP_H '0081'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0081/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0078/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0084
              edit EMOS_TIME_STEP_H '0084'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0084/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0081/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0087
              edit EMOS_TIME_STEP_H '0087'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0087/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0084/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0090
              edit EMOS_TIME_STEP_H '0090'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0090/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0087/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0093
              edit EMOS_TIME_STEP_H '0093'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0093/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0090/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0096
              edit EMOS_TIME_STEP_H '0096'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0096/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0093/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0099
              edit EMOS_TIME_STEP_H '0099'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0099/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0096/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0102
              edit EMOS_TIME_STEP_H '0102'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0102/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0099/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0105
              edit EMOS_TIME_STEP_H '0105'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0105/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0102/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0108
              edit EMOS_TIME_STEP_H '0108'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0108/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0105/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0111
              edit EMOS_TIME_STEP_H '0111'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0111/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0108/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0114
              edit EMOS_TIME_STEP_H '0114'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0114/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0111/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0117
              edit EMOS_TIME_STEP_H '0117'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0117/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0114/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0120
              edit EMOS_TIME_STEP_H '0120'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0120/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0117/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0123
              edit EMOS_TIME_STEP_H '0123'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0123/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0120/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0126
              edit EMOS_TIME_STEP_H '0126'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0126/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0123/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0129
              edit EMOS_TIME_STEP_H '0129'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0129/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0126/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0132
              edit EMOS_TIME_STEP_H '0132'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0132/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0129/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0135
              edit EMOS_TIME_STEP_H '0135'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0135/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0132/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0138
              edit EMOS_TIME_STEP_H '0138'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0138/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0135/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0141
              edit EMOS_TIME_STEP_H '0141'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0141/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0138/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0144
              edit EMOS_TIME_STEP_H '0144'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0144/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0141/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0150
              edit EMOS_TIME_STEP_H '0150'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0150/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0144/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0156
              edit EMOS_TIME_STEP_H '0156'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0156/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0150/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0162
              edit EMOS_TIME_STEP_H '0162'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0162/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0156/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0168
              edit EMOS_TIME_STEP_H '0168'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0168/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0162/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0174
              edit EMOS_TIME_STEP_H '0174'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0174/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0168/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0180
              edit EMOS_TIME_STEP_H '0180'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0180/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0174/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0186
              edit EMOS_TIME_STEP_H '0186'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0186/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0180/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0192
              edit EMOS_TIME_STEP_H '0192'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0192/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0186/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0198
              edit EMOS_TIME_STEP_H '0198'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0198/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0192/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0204
              edit EMOS_TIME_STEP_H '0204'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0204/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0198/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0210
              edit EMOS_TIME_STEP_H '0210'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0210/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0204/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0216
              edit EMOS_TIME_STEP_H '0216'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0216/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0210/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0222
              edit EMOS_TIME_STEP_H '0222'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0222/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0216/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0228
              edit EMOS_TIME_STEP_H '0228'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0228/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0222/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0234
              edit EMOS_TIME_STEP_H '0234'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0234/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0228/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0240
              edit EMOS_TIME_STEP_H '0240'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0240/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0234/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            task check
              trigger ./0240 eq complete
          endfamily
          family wave
            edit EMOS_STREAM 'WE'
            edit EMOS_TYPE 'we'
            edit SMSURL 'GWAEF/12'
            family 0000
              trigger ../../prod/wave/0000/prodwave eq complete
              edit EMOS_TIME_STEP_H '0000'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0003
              trigger ../../prod/wave/0003/prodwave eq complete and ( ./0000 eq complete)
              edit EMOS_TIME_STEP_H '0003'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0006
              trigger ../../prod/wave/0006/prodwave eq complete and ( ./0003 eq complete)
              edit EMOS_TIME_STEP_H '0006'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0009
              trigger ../../prod/wave/0009/prodwave eq complete and ( ./0006 eq complete)
              edit EMOS_TIME_STEP_H '0009'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0012
              trigger ../../prod/wave/0012/prodwave eq complete and ( ./0009 eq complete)
              edit EMOS_TIME_STEP_H '0012'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0015
              trigger ../../prod/wave/0015/prodwave eq complete and ( ./0012 eq complete)
              edit EMOS_TIME_STEP_H '0015'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0018
              trigger ../../prod/wave/0018/prodwave eq complete and ( ./0015 eq complete)
              edit EMOS_TIME_STEP_H '0018'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0021
              trigger ../../prod/wave/0021/prodwave eq complete and ( ./0018 eq complete)
              edit EMOS_TIME_STEP_H '0021'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0024
              trigger ../../prod/wave/0024/prodwave eq complete and ( ./0021 eq complete)
              edit EMOS_TIME_STEP_H '0024'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0027
              trigger ../../prod/wave/0027/prodwave eq complete and ( ./0024 eq complete)
              edit EMOS_TIME_STEP_H '0027'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0030
              trigger ../../prod/wave/0030/prodwave eq complete and ( ./0027 eq complete)
              edit EMOS_TIME_STEP_H '0030'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0033
              trigger ../../prod/wave/0033/prodwave eq complete and ( ./0030 eq complete)
              edit EMOS_TIME_STEP_H '0033'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0036
              trigger ../../prod/wave/0036/prodwave eq complete and ( ./0033 eq complete)
              edit EMOS_TIME_STEP_H '0036'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0039
              trigger ../../prod/wave/0039/prodwave eq complete and ( ./0036 eq complete)
              edit EMOS_TIME_STEP_H '0039'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0042
              trigger ../../prod/wave/0042/prodwave eq complete and ( ./0039 eq complete)
              edit EMOS_TIME_STEP_H '0042'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0045
              trigger ../../prod/wave/0045/prodwave eq complete and ( ./0042 eq complete)
              edit EMOS_TIME_STEP_H '0045'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0048
              trigger ../../prod/wave/0048/prodwave eq complete and ( ./0045 eq complete)
              edit EMOS_TIME_STEP_H '0048'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0051
              trigger ../../prod/wave/0051/prodwave eq complete and ( ./0048 eq complete)
              edit EMOS_TIME_STEP_H '0051'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0054
              trigger ../../prod/wave/0054/prodwave eq complete and ( ./0051 eq complete)
              edit EMOS_TIME_STEP_H '0054'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0057
              trigger ../../prod/wave/0057/prodwave eq complete and ( ./0054 eq complete)
              edit EMOS_TIME_STEP_H '0057'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0060
              trigger ../../prod/wave/0060/prodwave eq complete and ( ./0057 eq complete)
              edit EMOS_TIME_STEP_H '0060'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0063
              trigger ../../prod/wave/0063/prodwave eq complete and ( ./0060 eq complete)
              edit EMOS_TIME_STEP_H '0063'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0066
              trigger ../../prod/wave/0066/prodwave eq complete and ( ./0063 eq complete)
              edit EMOS_TIME_STEP_H '0066'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0069
              trigger ../../prod/wave/0069/prodwave eq complete and ( ./0066 eq complete)
              edit EMOS_TIME_STEP_H '0069'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0072
              trigger ../../prod/wave/0072/prodwave eq complete and ( ./0069 eq complete)
              edit EMOS_TIME_STEP_H '0072'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0075
              trigger ../../prod/wave/0075/prodwave eq complete and ( ./0072 eq complete)
              edit EMOS_TIME_STEP_H '0075'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0078
              trigger ../../prod/wave/0078/prodwave eq complete and ( ./0075 eq complete)
              edit EMOS_TIME_STEP_H '0078'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0081
              trigger ../../prod/wave/0081/prodwave eq complete and ( ./0078 eq complete)
              edit EMOS_TIME_STEP_H '0081'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0084
              trigger ../../prod/wave/0084/prodwave eq complete and ( ./0081 eq complete)
              edit EMOS_TIME_STEP_H '0084'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0087
              trigger ../../prod/wave/0087/prodwave eq complete and ( ./0084 eq complete)
              edit EMOS_TIME_STEP_H '0087'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0090
              trigger ../../prod/wave/0090/prodwave eq complete and ( ./0087 eq complete)
              edit EMOS_TIME_STEP_H '0090'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0093
              trigger ../../prod/wave/0093/prodwave eq complete and ( ./0090 eq complete)
              edit EMOS_TIME_STEP_H '0093'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0096
              trigger ../../prod/wave/0096/prodwave eq complete and ( ./0093 eq complete)
              edit EMOS_TIME_STEP_H '0096'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0099
              trigger ../../prod/wave/0099/prodwave eq complete and ( ./0096 eq complete)
              edit EMOS_TIME_STEP_H '0099'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0102
              trigger ../../prod/wave/0102/prodwave eq complete and ( ./0099 eq complete)
              edit EMOS_TIME_STEP_H '0102'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0105
              trigger ../../prod/wave/0105/prodwave eq complete and ( ./0102 eq complete)
              edit EMOS_TIME_STEP_H '0105'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0108
              trigger ../../prod/wave/0108/prodwave eq complete and ( ./0105 eq complete)
              edit EMOS_TIME_STEP_H '0108'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0111
              trigger ../../prod/wave/0111/prodwave eq complete and ( ./0108 eq complete)
              edit EMOS_TIME_STEP_H '0111'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0114
              trigger ../../prod/wave/0114/prodwave eq complete and ( ./0111 eq complete)
              edit EMOS_TIME_STEP_H '0114'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0117
              trigger ../../prod/wave/0117/prodwave eq complete and ( ./0114 eq complete)
              edit EMOS_TIME_STEP_H '0117'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0120
              trigger ../../prod/wave/0120/prodwave eq complete and ( ./0117 eq complete)
              edit EMOS_TIME_STEP_H '0120'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0123
              trigger ../../prod/wave/0123/prodwave eq complete and ( ./0120 eq complete)
              edit EMOS_TIME_STEP_H '0123'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0126
              trigger ../../prod/wave/0126/prodwave eq complete and ( ./0123 eq complete)
              edit EMOS_TIME_STEP_H '0126'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0129
              trigger ../../prod/wave/0129/prodwave eq complete and ( ./0126 eq complete)
              edit EMOS_TIME_STEP_H '0129'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0132
              trigger ../../prod/wave/0132/prodwave eq complete and ( ./0129 eq complete)
              edit EMOS_TIME_STEP_H '0132'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0135
              trigger ../../prod/wave/0135/prodwave eq complete and ( ./0132 eq complete)
              edit EMOS_TIME_STEP_H '0135'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0138
              trigger ../../prod/wave/0138/prodwave eq complete and ( ./0135 eq complete)
              edit EMOS_TIME_STEP_H '0138'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0141
              trigger ../../prod/wave/0141/prodwave eq complete and ( ./0138 eq complete)
              edit EMOS_TIME_STEP_H '0141'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0144
              trigger ../../prod/wave/0144/prodwave eq complete and ( ./0141 eq complete)
              edit EMOS_TIME_STEP_H '0144'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0150
              trigger ../../prod/wave/0150/prodwave eq complete and ( ./0144 eq complete)
              edit EMOS_TIME_STEP_H '0150'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0156
              trigger ../../prod/wave/0156/prodwave eq complete and ( ./0150 eq complete)
              edit EMOS_TIME_STEP_H '0156'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0162
              trigger ../../prod/wave/0162/prodwave eq complete and ( ./0156 eq complete)
              edit EMOS_TIME_STEP_H '0162'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0168
              trigger ../../prod/wave/0168/prodwave eq complete and ( ./0162 eq complete)
              edit EMOS_TIME_STEP_H '0168'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0174
              trigger ../../prod/wave/0174/prodwave eq complete and ( ./0168 eq complete)
              edit EMOS_TIME_STEP_H '0174'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0180
              trigger ../../prod/wave/0180/prodwave eq complete and ( ./0174 eq complete)
              edit EMOS_TIME_STEP_H '0180'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0186
              trigger ../../prod/wave/0186/prodwave eq complete and ( ./0180 eq complete)
              edit EMOS_TIME_STEP_H '0186'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0192
              trigger ../../prod/wave/0192/prodwave eq complete and ( ./0186 eq complete)
              edit EMOS_TIME_STEP_H '0192'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0198
              trigger ../../prod/wave/0198/prodwave eq complete and ( ./0192 eq complete)
              edit EMOS_TIME_STEP_H '0198'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0204
              trigger ../../prod/wave/0204/prodwave eq complete and ( ./0198 eq complete)
              edit EMOS_TIME_STEP_H '0204'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0210
              trigger ../../prod/wave/0210/prodwave eq complete and ( ./0204 eq complete)
              edit EMOS_TIME_STEP_H '0210'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0216
              trigger ../../prod/wave/0216/prodwave eq complete and ( ./0210 eq complete)
              edit EMOS_TIME_STEP_H '0216'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0222
              trigger ../../prod/wave/0222/prodwave eq complete and ( ./0216 eq complete)
              edit EMOS_TIME_STEP_H '0222'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0228
              trigger ../../prod/wave/0228/prodwave eq complete and ( ./0222 eq complete)
              edit EMOS_TIME_STEP_H '0228'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0234
              trigger ../../prod/wave/0234/prodwave eq complete and ( ./0228 eq complete)
              edit EMOS_TIME_STEP_H '0234'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0240
              trigger ../../prod/wave/0240/prodwave eq complete and ( ./0234 eq complete)
              edit EMOS_TIME_STEP_H '0240'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            task check
              trigger ./0240 eq complete
          endfamily
          task diss_wparam
            late -s +00:05 -c +00:15
            trigger ../prod/prod_wparam eq complete
            edit PROD_TASK 'w'
            edit EMOS_TIME_STEP_H '0240'
            edit EMOS_TYPE 'wp'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event pdb
          task diss_waef_postproc
            late -s +00:05 -c +00:15
            trigger ../prod/waef_postproc eq complete
            edit PROD_TASK 'p'
            edit EMOS_TIME_STEP_H '0240'
            edit EMOS_TYPE 'pw'
            edit EMOS_STREAM 'WE'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event pdb
          task diss_postproc
            late -s +00:05 -c +00:15
            trigger ../prod/enfo_postproc eq complete
            edit PROD_TASK 'p'
            edit EMOS_TIME_STEP_H '0240'
            edit EMOS_TYPE 'ep'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event pdb
          task diss_enfotc
            late -s +00:05 -c +00:15
            trigger ../prod/enfo_tc eq complete
            edit PROD_TASK 't'
            edit EMOS_TIME_STEP_H '0240'
            edit EMOS_TYPE 'tf'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event pdb
          task diss_opertc
            late -s +00:05 -c +00:15
            trigger ../prod/oper_tc eq complete
            edit PROD_TASK 't'
            edit EMOS_TIME_STEP_H '0240'
            edit EMOS_TYPE 'tf'
            edit EMOS_STREAM 'DA'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event pdb
        endfamily
      endfamily
      family legB
        complete /emc_36r1/main/ref/12/events:240
        edit EPSLEG '2'
        family fc
          edit FCMEM '2240'
          edit FCSTACK '700'
          edit FCMRFS '0'
          edit PRIORITY '71'
          edit USER_PRIORITY '71'
          edit N_PERT_AN '10'
          edit SSTLEN '144'
          edit EPSINISTEP_PREVIOUS '216'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
          edit TSTEP '2700'
          edit EPSFCRES '639'
          edit RESOL '639'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
          edit CPUTIME '1000'
          edit MEM '960'
          edit THREADS '8'
          edit NPES '8'
          edit EPSLEG '2'
          edit PERSST 'true'
          edit SSTINT '1'
          task wavfcdata
            trigger ./getpersSST eq complete and ../../legA/fc/wavini eq complete
            edit EPSCOUPLE '1'
          task getpersSST
            trigger ../../legA/fc/inidata eq complete and ../../legA/fc/getpersSST eq complete
            edit EPSCOUPLE '1'
            edit EPS_INI_STREAM 'DA'
            edit EPSFCRES '319'
            edit RESOL '319'
          family pf
            complete /emc_36r1/main/ref/12/events:cf
            trigger ./getpersSST eq complete and ./wavfcdata eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTYPE 'pf'
            edit EMOS_TYPE 'pf'
            edit USER_PRIORITY '71'
            edit PRIORITY '71'
            edit FCNPES '32'
            edit FCCHUNK '1'
            edit OCTOTAL '5'
            edit FCTOTAL '51'
            edit FCLENGTH1 '216'
            edit DAILY '0'
            edit RUNVARFC '1'
            edit EPSNLEGS '3'
            edit EPSFCLEV '62'
            edit EPSWAMNSTPW '1'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSFCRES '319'
            edit EPSWAMRESOL 'global50'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit EPSINISTEP '360'
            edit FCLENGTH '240'
            edit TSTEP '1200'
            edit EPSLEG '2'
            edit CLASS 'od'
            edit THREADS '2'
            inlimit /emc_36r1/limits:pftotal
            meter step -1 144 144
            family 01
              trigger ../../../legA/fc/pf/01 eq complete
              edit EPSMEMBER '1'
              edit MEMBER '1'
              edit SUBFSFAMILY '/pf001'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 02
              trigger ../../../legA/fc/pf/02 eq complete
              edit EPSMEMBER '2'
              edit MEMBER '2'
              edit SUBFSFAMILY '/pf002'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 03
              trigger ../../../legA/fc/pf/03 eq complete
              edit EPSMEMBER '3'
              edit MEMBER '3'
              edit SUBFSFAMILY '/pf003'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 04
              trigger ../../../legA/fc/pf/04 eq complete
              edit EPSMEMBER '4'
              edit MEMBER '4'
              edit SUBFSFAMILY '/pf004'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 05
              trigger ../../../legA/fc/pf/05 eq complete
              edit EPSMEMBER '5'
              edit MEMBER '5'
              edit SUBFSFAMILY '/pf005'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 06
              trigger ../../../legA/fc/pf/06 eq complete
              edit EPSMEMBER '6'
              edit MEMBER '6'
              edit SUBFSFAMILY '/pf006'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 07
              trigger ../../../legA/fc/pf/07 eq complete
              edit EPSMEMBER '7'
              edit MEMBER '7'
              edit SUBFSFAMILY '/pf007'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 08
              trigger ../../../legA/fc/pf/08 eq complete
              edit EPSMEMBER '8'
              edit MEMBER '8'
              edit SUBFSFAMILY '/pf008'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 09
              trigger ../../../legA/fc/pf/09 eq complete
              edit EPSMEMBER '9'
              edit MEMBER '9'
              edit SUBFSFAMILY '/pf009'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 10
              trigger ../../../legA/fc/pf/10 eq complete
              edit EPSMEMBER '10'
              edit MEMBER '10'
              edit SUBFSFAMILY '/pf010'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 11
              trigger ../../../legA/fc/pf/11 eq complete
              edit EPSMEMBER '11'
              edit MEMBER '11'
              edit SUBFSFAMILY '/pf011'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 12
              trigger ../../../legA/fc/pf/12 eq complete
              edit EPSMEMBER '12'
              edit MEMBER '12'
              edit SUBFSFAMILY '/pf012'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 13
              trigger ../../../legA/fc/pf/13 eq complete
              edit EPSMEMBER '13'
              edit MEMBER '13'
              edit SUBFSFAMILY '/pf013'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 14
              trigger ../../../legA/fc/pf/14 eq complete
              edit EPSMEMBER '14'
              edit MEMBER '14'
              edit SUBFSFAMILY '/pf014'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 15
              trigger ../../../legA/fc/pf/15 eq complete
              edit EPSMEMBER '15'
              edit MEMBER '15'
              edit SUBFSFAMILY '/pf015'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 16
              trigger ../../../legA/fc/pf/16 eq complete
              edit EPSMEMBER '16'
              edit MEMBER '16'
              edit SUBFSFAMILY '/pf016'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 17
              trigger ../../../legA/fc/pf/17 eq complete
              edit EPSMEMBER '17'
              edit MEMBER '17'
              edit SUBFSFAMILY '/pf017'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 18
              trigger ../../../legA/fc/pf/18 eq complete
              edit EPSMEMBER '18'
              edit MEMBER '18'
              edit SUBFSFAMILY '/pf018'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 19
              trigger ../../../legA/fc/pf/19 eq complete
              edit EPSMEMBER '19'
              edit MEMBER '19'
              edit SUBFSFAMILY '/pf019'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 20
              trigger ../../../legA/fc/pf/20 eq complete
              edit EPSMEMBER '20'
              edit MEMBER '20'
              edit SUBFSFAMILY '/pf020'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 21
              trigger ../../../legA/fc/pf/21 eq complete
              edit EPSMEMBER '21'
              edit MEMBER '21'
              edit SUBFSFAMILY '/pf021'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 22
              trigger ../../../legA/fc/pf/22 eq complete
              edit EPSMEMBER '22'
              edit MEMBER '22'
              edit SUBFSFAMILY '/pf022'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 23
              trigger ../../../legA/fc/pf/23 eq complete
              edit EPSMEMBER '23'
              edit MEMBER '23'
              edit SUBFSFAMILY '/pf023'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 24
              trigger ../../../legA/fc/pf/24 eq complete
              edit EPSMEMBER '24'
              edit MEMBER '24'
              edit SUBFSFAMILY '/pf024'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 25
              trigger ../../../legA/fc/pf/25 eq complete
              edit EPSMEMBER '25'
              edit MEMBER '25'
              edit SUBFSFAMILY '/pf025'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 26
              trigger ../../../legA/fc/pf/26 eq complete
              edit EPSMEMBER '26'
              edit MEMBER '26'
              edit SUBFSFAMILY '/pf026'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 27
              trigger ../../../legA/fc/pf/27 eq complete
              edit EPSMEMBER '27'
              edit MEMBER '27'
              edit SUBFSFAMILY '/pf027'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 28
              trigger ../../../legA/fc/pf/28 eq complete
              edit EPSMEMBER '28'
              edit MEMBER '28'
              edit SUBFSFAMILY '/pf028'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 29
              trigger ../../../legA/fc/pf/29 eq complete
              edit EPSMEMBER '29'
              edit MEMBER '29'
              edit SUBFSFAMILY '/pf029'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 30
              trigger ../../../legA/fc/pf/30 eq complete
              edit EPSMEMBER '30'
              edit MEMBER '30'
              edit SUBFSFAMILY '/pf030'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 31
              trigger ../../../legA/fc/pf/31 eq complete
              edit EPSMEMBER '31'
              edit MEMBER '31'
              edit SUBFSFAMILY '/pf031'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 32
              trigger ../../../legA/fc/pf/32 eq complete
              edit EPSMEMBER '32'
              edit MEMBER '32'
              edit SUBFSFAMILY '/pf032'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 33
              trigger ../../../legA/fc/pf/33 eq complete
              edit EPSMEMBER '33'
              edit MEMBER '33'
              edit SUBFSFAMILY '/pf033'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 34
              trigger ../../../legA/fc/pf/34 eq complete
              edit EPSMEMBER '34'
              edit MEMBER '34'
              edit SUBFSFAMILY '/pf034'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 35
              trigger ../../../legA/fc/pf/35 eq complete
              edit EPSMEMBER '35'
              edit MEMBER '35'
              edit SUBFSFAMILY '/pf035'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 36
              trigger ../../../legA/fc/pf/36 eq complete
              edit EPSMEMBER '36'
              edit MEMBER '36'
              edit SUBFSFAMILY '/pf036'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 37
              trigger ../../../legA/fc/pf/37 eq complete
              edit EPSMEMBER '37'
              edit MEMBER '37'
              edit SUBFSFAMILY '/pf037'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 38
              trigger ../../../legA/fc/pf/38 eq complete
              edit EPSMEMBER '38'
              edit MEMBER '38'
              edit SUBFSFAMILY '/pf038'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 39
              trigger ../../../legA/fc/pf/39 eq complete
              edit EPSMEMBER '39'
              edit MEMBER '39'
              edit SUBFSFAMILY '/pf039'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 40
              trigger ../../../legA/fc/pf/40 eq complete
              edit EPSMEMBER '40'
              edit MEMBER '40'
              edit SUBFSFAMILY '/pf040'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 41
              trigger ../../../legA/fc/pf/41 eq complete
              edit EPSMEMBER '41'
              edit MEMBER '41'
              edit SUBFSFAMILY '/pf041'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 42
              trigger ../../../legA/fc/pf/42 eq complete
              edit EPSMEMBER '42'
              edit MEMBER '42'
              edit SUBFSFAMILY '/pf042'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 43
              trigger ../../../legA/fc/pf/43 eq complete
              edit EPSMEMBER '43'
              edit MEMBER '43'
              edit SUBFSFAMILY '/pf043'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 44
              trigger ../../../legA/fc/pf/44 eq complete
              edit EPSMEMBER '44'
              edit MEMBER '44'
              edit SUBFSFAMILY '/pf044'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 45
              trigger ../../../legA/fc/pf/45 eq complete
              edit EPSMEMBER '45'
              edit MEMBER '45'
              edit SUBFSFAMILY '/pf045'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 46
              trigger ../../../legA/fc/pf/46 eq complete
              edit EPSMEMBER '46'
              edit MEMBER '46'
              edit SUBFSFAMILY '/pf046'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 47
              trigger ../../../legA/fc/pf/47 eq complete
              edit EPSMEMBER '47'
              edit MEMBER '47'
              edit SUBFSFAMILY '/pf047'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 48
              trigger ../../../legA/fc/pf/48 eq complete
              edit EPSMEMBER '48'
              edit MEMBER '48'
              edit SUBFSFAMILY '/pf048'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 49
              trigger ../../../legA/fc/pf/49 eq complete
              edit EPSMEMBER '49'
              edit MEMBER '49'
              edit SUBFSFAMILY '/pf049'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 50
              trigger ../../../legA/fc/pf/50 eq complete
              edit EPSMEMBER '50'
              edit MEMBER '50'
              edit SUBFSFAMILY '/pf050'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
          endfamily
          family cf
            trigger ../../legA/fc/cf eq complete and not /emc_36r1/main/ref/12/events:cpl and ./getpersSST eq complete and ./wavfcdata eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit MEMBER '0'
            edit EPSMEMBER '0'
            edit SUBFSFAMILY '/cf000'
            edit EPSTYPE 'cf'
            edit EMOS_TYPE 'cf'
            edit USER_PRIORITY '72'
            edit EPSNLEGS '3'
            edit EPSFCLEV '62'
            edit EPSWAMNSTPW '1'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSFCRES '319'
            edit EPSWAMRESOL 'global50'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit EPSINISTEP '360'
            edit FCLENGTH '240'
            edit TSTEP '1200'
            edit EPSLEG '2'
            edit CLASS 'od'
            edit SMSTRIES '1'
            edit RUNCOUPLE '0'
            task getvarepsdata
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '2700'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSLEG '2'
              edit EPSINISTEP '216'
              edit EPSFCRES '639'
              edit EPSFCRESN '319'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '2700'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '216'
              edit EPSFCRES '319'
              edit NPES '24'
              edit FCNPES '24'
              edit THREADS '2'
              edit CLASS 'od'
            task modeleps
              trigger intHtoL eq complete and getvarepsdata eq complete
              edit FCNPES '32'
              edit THREADS '4'
              edit SMSTRIES '1'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSLEG '2'
              edit EPSFCRES '319'
              edit EPSINISTEP '144'
              edit EPSINISTEP_PREVIOUS '216'
              edit FCLENGTH '144'
              edit TSTEP '2700'
              edit CLASS 'od'
              meter step -1 144 144
          endfamily
          family reservation
            trigger ./pf/46/intHtoL eq complete and ./pf/47/intHtoL eq complete and ./pf/48/intHtoL eq complete and ./pf/49/intHtoL eq complete and ./pf/50/intHtoL eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/smsfiles'
            inlimit /emc_36r1/limits:reservation
            task reset_class
              late -s +00:02 -c +00:02
              edit RESERVATION_DEFINITION 'default'
              label info ""
            task reset_cpu
              late -s +00:02 -c +00:02
              edit RESERVED_CPUS '0'
              label info ""
          endfamily
        endfamily
        family postproc
          trigger ./fc/pf eq complete and ./fc/cf eq complete
          edit STEPS '252/360/12/24 360/360/12/120'
          inlimit /emc_36r1/limits:postproc
          family ensms
            complete /emc_36r1/main/ref/12/events:cf
            trigger ../fc eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/leg'
            family limits
              limit fdb 1
              limit gen 8
            endfamily
            task ensms
              edit ENS_M 'false'
              edit ENS_S 'false'
              edit ENS_MARS 'false'
              event stage
            family 300
              trigger ./ensms:stage or ./ensms eq complete
              edit STEP1 '246'
              edit STEP2 '300'
              edit BY '6'
              edit STEPS '0'
              family T
                edit ENS_M 'false'
                edit ENS_S 'false'
                edit LIST_PARAM 'T'
                inlimit ../limits:gen
                family 850
                  task ensms
                    edit LIST_LEV '850'
                endfamily
                family 500
                  task ensms
                    edit LIST_LEV '500'
                endfamily
              endfamily
              family Z
                edit ENS_M 'false'
                edit ENS_S 'false'
                edit LIST_PARAM 'Z'
                inlimit ../limits:gen
                family 1000
                  task ensms
                    edit LIST_LEV '1000'
                endfamily
                family 500
                  task ensms
                    edit LIST_LEV '500'
                endfamily
              endfamily
              task ensms
                trigger T eq complete and Z eq complete
                edit ENS_MARS 'false'
                inlimit ../limits:fdb
            endfamily
            family 360
              trigger ./ensms:stage or ./ensms eq complete
              edit STEP1 '306'
              edit STEP2 '360'
              edit BY '6'
              edit STEPS '0'
              family T
                edit ENS_M 'false'
                edit ENS_S 'false'
                edit LIST_PARAM 'T'
                inlimit ../limits:gen
                family 850
                  task ensms
                    edit LIST_LEV '850'
                endfamily
                family 500
                  task ensms
                    edit LIST_LEV '500'
                endfamily
              endfamily
              family Z
                edit ENS_M 'false'
                edit ENS_S 'false'
                edit LIST_PARAM 'Z'
                inlimit ../limits:gen
                family 1000
                  task ensms
                    edit LIST_LEV '1000'
                endfamily
                family 500
                  task ensms
                    edit LIST_LEV '500'
                endfamily
              endfamily
              task ensms
                trigger T eq complete and Z eq complete
                edit ENS_MARS 'false'
                inlimit ../limits:fdb
            endfamily
          endfamily
          family prob
            complete /emc_36r1/main/ref/12/events:cf
            trigger ( /emc_36r1/main/12/legB/fc eq complete) and ( /emc_36r1/main/12/legA/postproc/prob eq complete)
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/leg'
            edit STEP1 '252'
            edit STEP2 '360'
            edit BY '12'
            family 24
              edit STEP1 '240'
              edit STEP2 '360'
              edit BY '24'
              task t850
              task rain
              task ff
            endfamily
            family 12
              edit STEP1 '252'
              edit STEP2 '348'
              edit BY '24'
              task t850
              task rain
              task ff
            endfamily
            task 2t
            task swh
            task mwp
            task 10fg
              trigger ../dp eq complete
            task prob_fdb
              trigger ./2t eq complete and swh eq complete and mwp eq complete and ./10fg eq complete and ./12 eq complete and ./24 eq complete
          endfamily
          family dp
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit STEPS '246:360:6'
            inlimit /emc_36r1/limits:dp
            family dp
              trigger ./cf eq complete and ./pf eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit OVERLAP '1'
              edit STEPS '216:240:6'
              inlimit /emc_36r1/limits:dp
              family cf
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                complete /emc_36r1/main/ref/12/events:cf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
                family 05
                  edit EPSMEMBER '5'
                  edit MEMBER '5'
                  edit SUBFSFAMILY '/pf005'
                  task derived_param
                endfamily
                family 06
                  edit EPSMEMBER '6'
                  edit MEMBER '6'
                  edit SUBFSFAMILY '/pf006'
                  task derived_param
                endfamily
                family 07
                  edit EPSMEMBER '7'
                  edit MEMBER '7'
                  edit SUBFSFAMILY '/pf007'
                  task derived_param
                endfamily
                family 08
                  edit EPSMEMBER '8'
                  edit MEMBER '8'
                  edit SUBFSFAMILY '/pf008'
                  task derived_param
                endfamily
                family 09
                  edit EPSMEMBER '9'
                  edit MEMBER '9'
                  edit SUBFSFAMILY '/pf009'
                  task derived_param
                endfamily
                family 10
                  edit EPSMEMBER '10'
                  edit MEMBER '10'
                  edit SUBFSFAMILY '/pf010'
                  task derived_param
                endfamily
                family 11
                  edit EPSMEMBER '11'
                  edit MEMBER '11'
                  edit SUBFSFAMILY '/pf011'
                  task derived_param
                endfamily
                family 12
                  edit EPSMEMBER '12'
                  edit MEMBER '12'
                  edit SUBFSFAMILY '/pf012'
                  task derived_param
                endfamily
                family 13
                  edit EPSMEMBER '13'
                  edit MEMBER '13'
                  edit SUBFSFAMILY '/pf013'
                  task derived_param
                endfamily
                family 14
                  edit EPSMEMBER '14'
                  edit MEMBER '14'
                  edit SUBFSFAMILY '/pf014'
                  task derived_param
                endfamily
                family 15
                  edit EPSMEMBER '15'
                  edit MEMBER '15'
                  edit SUBFSFAMILY '/pf015'
                  task derived_param
                endfamily
                family 16
                  edit EPSMEMBER '16'
                  edit MEMBER '16'
                  edit SUBFSFAMILY '/pf016'
                  task derived_param
                endfamily
                family 17
                  edit EPSMEMBER '17'
                  edit MEMBER '17'
                  edit SUBFSFAMILY '/pf017'
                  task derived_param
                endfamily
                family 18
                  edit EPSMEMBER '18'
                  edit MEMBER '18'
                  edit SUBFSFAMILY '/pf018'
                  task derived_param
                endfamily
                family 19
                  edit EPSMEMBER '19'
                  edit MEMBER '19'
                  edit SUBFSFAMILY '/pf019'
                  task derived_param
                endfamily
                family 20
                  edit EPSMEMBER '20'
                  edit MEMBER '20'
                  edit SUBFSFAMILY '/pf020'
                  task derived_param
                endfamily
                family 21
                  edit EPSMEMBER '21'
                  edit MEMBER '21'
                  edit SUBFSFAMILY '/pf021'
                  task derived_param
                endfamily
                family 22
                  edit EPSMEMBER '22'
                  edit MEMBER '22'
                  edit SUBFSFAMILY '/pf022'
                  task derived_param
                endfamily
                family 23
                  edit EPSMEMBER '23'
                  edit MEMBER '23'
                  edit SUBFSFAMILY '/pf023'
                  task derived_param
                endfamily
                family 24
                  edit EPSMEMBER '24'
                  edit MEMBER '24'
                  edit SUBFSFAMILY '/pf024'
                  task derived_param
                endfamily
                family 25
                  edit EPSMEMBER '25'
                  edit MEMBER '25'
                  edit SUBFSFAMILY '/pf025'
                  task derived_param
                endfamily
                family 26
                  edit EPSMEMBER '26'
                  edit MEMBER '26'
                  edit SUBFSFAMILY '/pf026'
                  task derived_param
                endfamily
                family 27
                  edit EPSMEMBER '27'
                  edit MEMBER '27'
                  edit SUBFSFAMILY '/pf027'
                  task derived_param
                endfamily
                family 28
                  edit EPSMEMBER '28'
                  edit MEMBER '28'
                  edit SUBFSFAMILY '/pf028'
                  task derived_param
                endfamily
                family 29
                  edit EPSMEMBER '29'
                  edit MEMBER '29'
                  edit SUBFSFAMILY '/pf029'
                  task derived_param
                endfamily
                family 30
                  edit EPSMEMBER '30'
                  edit MEMBER '30'
                  edit SUBFSFAMILY '/pf030'
                  task derived_param
                endfamily
                family 31
                  edit EPSMEMBER '31'
                  edit MEMBER '31'
                  edit SUBFSFAMILY '/pf031'
                  task derived_param
                endfamily
                family 32
                  edit EPSMEMBER '32'
                  edit MEMBER '32'
                  edit SUBFSFAMILY '/pf032'
                  task derived_param
                endfamily
                family 33
                  edit EPSMEMBER '33'
                  edit MEMBER '33'
                  edit SUBFSFAMILY '/pf033'
                  task derived_param
                endfamily
                family 34
                  edit EPSMEMBER '34'
                  edit MEMBER '34'
                  edit SUBFSFAMILY '/pf034'
                  task derived_param
                endfamily
                family 35
                  edit EPSMEMBER '35'
                  edit MEMBER '35'
                  edit SUBFSFAMILY '/pf035'
                  task derived_param
                endfamily
                family 36
                  edit EPSMEMBER '36'
                  edit MEMBER '36'
                  edit SUBFSFAMILY '/pf036'
                  task derived_param
                endfamily
                family 37
                  edit EPSMEMBER '37'
                  edit MEMBER '37'
                  edit SUBFSFAMILY '/pf037'
                  task derived_param
                endfamily
                family 38
                  edit EPSMEMBER '38'
                  edit MEMBER '38'
                  edit SUBFSFAMILY '/pf038'
                  task derived_param
                endfamily
                family 39
                  edit EPSMEMBER '39'
                  edit MEMBER '39'
                  edit SUBFSFAMILY '/pf039'
                  task derived_param
                endfamily
                family 40
                  edit EPSMEMBER '40'
                  edit MEMBER '40'
                  edit SUBFSFAMILY '/pf040'
                  task derived_param
                endfamily
                family 41
                  edit EPSMEMBER '41'
                  edit MEMBER '41'
                  edit SUBFSFAMILY '/pf041'
                  task derived_param
                endfamily
                family 42
                  edit EPSMEMBER '42'
                  edit MEMBER '42'
                  edit SUBFSFAMILY '/pf042'
                  task derived_param
                endfamily
                family 43
                  edit EPSMEMBER '43'
                  edit MEMBER '43'
                  edit SUBFSFAMILY '/pf043'
                  task derived_param
                endfamily
                family 44
                  edit EPSMEMBER '44'
                  edit MEMBER '44'
                  edit SUBFSFAMILY '/pf044'
                  task derived_param
                endfamily
                family 45
                  edit EPSMEMBER '45'
                  edit MEMBER '45'
                  edit SUBFSFAMILY '/pf045'
                  task derived_param
                endfamily
                family 46
                  edit EPSMEMBER '46'
                  edit MEMBER '46'
                  edit SUBFSFAMILY '/pf046'
                  task derived_param
                endfamily
                family 47
                  edit EPSMEMBER '47'
                  edit MEMBER '47'
                  edit SUBFSFAMILY '/pf047'
                  task derived_param
                endfamily
                family 48
                  edit EPSMEMBER '48'
                  edit MEMBER '48'
                  edit SUBFSFAMILY '/pf048'
                  task derived_param
                endfamily
                family 49
                  edit EPSMEMBER '49'
                  edit MEMBER '49'
                  edit SUBFSFAMILY '/pf049'
                  task derived_param
                endfamily
                family 50
                  edit EPSMEMBER '50'
                  edit MEMBER '50'
                  edit SUBFSFAMILY '/pf050'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family cf
              trigger ../../fc/cf eq complete
              edit EMOS_TYPE 'cf'
              task derived_param
                edit MEMBER 'off'
            endfamily
            family pf
              complete /emc_36r1/main/ref/12/events:cf
              edit EMOS_TYPE 'pf'
              family 01
                trigger ../../../fc/pf/01 eq complete
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task derived_param
              endfamily
              family 02
                trigger ../../../fc/pf/02 eq complete
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task derived_param
              endfamily
              family 03
                trigger ../../../fc/pf/03 eq complete
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task derived_param
              endfamily
              family 04
                trigger ../../../fc/pf/04 eq complete
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task derived_param
              endfamily
              family 05
                trigger ../../../fc/pf/05 eq complete
                edit EPSMEMBER '5'
                edit MEMBER '5'
                edit SUBFSFAMILY '/pf005'
                task derived_param
              endfamily
              family 06
                trigger ../../../fc/pf/06 eq complete
                edit EPSMEMBER '6'
                edit MEMBER '6'
                edit SUBFSFAMILY '/pf006'
                task derived_param
              endfamily
              family 07
                trigger ../../../fc/pf/07 eq complete
                edit EPSMEMBER '7'
                edit MEMBER '7'
                edit SUBFSFAMILY '/pf007'
                task derived_param
              endfamily
              family 08
                trigger ../../../fc/pf/08 eq complete
                edit EPSMEMBER '8'
                edit MEMBER '8'
                edit SUBFSFAMILY '/pf008'
                task derived_param
              endfamily
              family 09
                trigger ../../../fc/pf/09 eq complete
                edit EPSMEMBER '9'
                edit MEMBER '9'
                edit SUBFSFAMILY '/pf009'
                task derived_param
              endfamily
              family 10
                trigger ../../../fc/pf/10 eq complete
                edit EPSMEMBER '10'
                edit MEMBER '10'
                edit SUBFSFAMILY '/pf010'
                task derived_param
              endfamily
              family 11
                trigger ../../../fc/pf/11 eq complete
                edit EPSMEMBER '11'
                edit MEMBER '11'
                edit SUBFSFAMILY '/pf011'
                task derived_param
              endfamily
              family 12
                trigger ../../../fc/pf/12 eq complete
                edit EPSMEMBER '12'
                edit MEMBER '12'
                edit SUBFSFAMILY '/pf012'
                task derived_param
              endfamily
              family 13
                trigger ../../../fc/pf/13 eq complete
                edit EPSMEMBER '13'
                edit MEMBER '13'
                edit SUBFSFAMILY '/pf013'
                task derived_param
              endfamily
              family 14
                trigger ../../../fc/pf/14 eq complete
                edit EPSMEMBER '14'
                edit MEMBER '14'
                edit SUBFSFAMILY '/pf014'
                task derived_param
              endfamily
              family 15
                trigger ../../../fc/pf/15 eq complete
                edit EPSMEMBER '15'
                edit MEMBER '15'
                edit SUBFSFAMILY '/pf015'
                task derived_param
              endfamily
              family 16
                trigger ../../../fc/pf/16 eq complete
                edit EPSMEMBER '16'
                edit MEMBER '16'
                edit SUBFSFAMILY '/pf016'
                task derived_param
              endfamily
              family 17
                trigger ../../../fc/pf/17 eq complete
                edit EPSMEMBER '17'
                edit MEMBER '17'
                edit SUBFSFAMILY '/pf017'
                task derived_param
              endfamily
              family 18
                trigger ../../../fc/pf/18 eq complete
                edit EPSMEMBER '18'
                edit MEMBER '18'
                edit SUBFSFAMILY '/pf018'
                task derived_param
              endfamily
              family 19
                trigger ../../../fc/pf/19 eq complete
                edit EPSMEMBER '19'
                edit MEMBER '19'
                edit SUBFSFAMILY '/pf019'
                task derived_param
              endfamily
              family 20
                trigger ../../../fc/pf/20 eq complete
                edit EPSMEMBER '20'
                edit MEMBER '20'
                edit SUBFSFAMILY '/pf020'
                task derived_param
              endfamily
              family 21
                trigger ../../../fc/pf/21 eq complete
                edit EPSMEMBER '21'
                edit MEMBER '21'
                edit SUBFSFAMILY '/pf021'
                task derived_param
              endfamily
              family 22
                trigger ../../../fc/pf/22 eq complete
                edit EPSMEMBER '22'
                edit MEMBER '22'
                edit SUBFSFAMILY '/pf022'
                task derived_param
              endfamily
              family 23
                trigger ../../../fc/pf/23 eq complete
                edit EPSMEMBER '23'
                edit MEMBER '23'
                edit SUBFSFAMILY '/pf023'
                task derived_param
              endfamily
              family 24
                trigger ../../../fc/pf/24 eq complete
                edit EPSMEMBER '24'
                edit MEMBER '24'
                edit SUBFSFAMILY '/pf024'
                task derived_param
              endfamily
              family 25
                trigger ../../../fc/pf/25 eq complete
                edit EPSMEMBER '25'
                edit MEMBER '25'
                edit SUBFSFAMILY '/pf025'
                task derived_param
              endfamily
              family 26
                trigger ../../../fc/pf/26 eq complete
                edit EPSMEMBER '26'
                edit MEMBER '26'
                edit SUBFSFAMILY '/pf026'
                task derived_param
              endfamily
              family 27
                trigger ../../../fc/pf/27 eq complete
                edit EPSMEMBER '27'
                edit MEMBER '27'
                edit SUBFSFAMILY '/pf027'
                task derived_param
              endfamily
              family 28
                trigger ../../../fc/pf/28 eq complete
                edit EPSMEMBER '28'
                edit MEMBER '28'
                edit SUBFSFAMILY '/pf028'
                task derived_param
              endfamily
              family 29
                trigger ../../../fc/pf/29 eq complete
                edit EPSMEMBER '29'
                edit MEMBER '29'
                edit SUBFSFAMILY '/pf029'
                task derived_param
              endfamily
              family 30
                trigger ../../../fc/pf/30 eq complete
                edit EPSMEMBER '30'
                edit MEMBER '30'
                edit SUBFSFAMILY '/pf030'
                task derived_param
              endfamily
              family 31
                trigger ../../../fc/pf/31 eq complete
                edit EPSMEMBER '31'
                edit MEMBER '31'
                edit SUBFSFAMILY '/pf031'
                task derived_param
              endfamily
              family 32
                trigger ../../../fc/pf/32 eq complete
                edit EPSMEMBER '32'
                edit MEMBER '32'
                edit SUBFSFAMILY '/pf032'
                task derived_param
              endfamily
              family 33
                trigger ../../../fc/pf/33 eq complete
                edit EPSMEMBER '33'
                edit MEMBER '33'
                edit SUBFSFAMILY '/pf033'
                task derived_param
              endfamily
              family 34
                trigger ../../../fc/pf/34 eq complete
                edit EPSMEMBER '34'
                edit MEMBER '34'
                edit SUBFSFAMILY '/pf034'
                task derived_param
              endfamily
              family 35
                trigger ../../../fc/pf/35 eq complete
                edit EPSMEMBER '35'
                edit MEMBER '35'
                edit SUBFSFAMILY '/pf035'
                task derived_param
              endfamily
              family 36
                trigger ../../../fc/pf/36 eq complete
                edit EPSMEMBER '36'
                edit MEMBER '36'
                edit SUBFSFAMILY '/pf036'
                task derived_param
              endfamily
              family 37
                trigger ../../../fc/pf/37 eq complete
                edit EPSMEMBER '37'
                edit MEMBER '37'
                edit SUBFSFAMILY '/pf037'
                task derived_param
              endfamily
              family 38
                trigger ../../../fc/pf/38 eq complete
                edit EPSMEMBER '38'
                edit MEMBER '38'
                edit SUBFSFAMILY '/pf038'
                task derived_param
              endfamily
              family 39
                trigger ../../../fc/pf/39 eq complete
                edit EPSMEMBER '39'
                edit MEMBER '39'
                edit SUBFSFAMILY '/pf039'
                task derived_param
              endfamily
              family 40
                trigger ../../../fc/pf/40 eq complete
                edit EPSMEMBER '40'
                edit MEMBER '40'
                edit SUBFSFAMILY '/pf040'
                task derived_param
              endfamily
              family 41
                trigger ../../../fc/pf/41 eq complete
                edit EPSMEMBER '41'
                edit MEMBER '41'
                edit SUBFSFAMILY '/pf041'
                task derived_param
              endfamily
              family 42
                trigger ../../../fc/pf/42 eq complete
                edit EPSMEMBER '42'
                edit MEMBER '42'
                edit SUBFSFAMILY '/pf042'
                task derived_param
              endfamily
              family 43
                trigger ../../../fc/pf/43 eq complete
                edit EPSMEMBER '43'
                edit MEMBER '43'
                edit SUBFSFAMILY '/pf043'
                task derived_param
              endfamily
              family 44
                trigger ../../../fc/pf/44 eq complete
                edit EPSMEMBER '44'
                edit MEMBER '44'
                edit SUBFSFAMILY '/pf044'
                task derived_param
              endfamily
              family 45
                trigger ../../../fc/pf/45 eq complete
                edit EPSMEMBER '45'
                edit MEMBER '45'
                edit SUBFSFAMILY '/pf045'
                task derived_param
              endfamily
              family 46
                trigger ../../../fc/pf/46 eq complete
                edit EPSMEMBER '46'
                edit MEMBER '46'
                edit SUBFSFAMILY '/pf046'
                task derived_param
              endfamily
              family 47
                trigger ../../../fc/pf/47 eq complete
                edit EPSMEMBER '47'
                edit MEMBER '47'
                edit SUBFSFAMILY '/pf047'
                task derived_param
              endfamily
              family 48
                trigger ../../../fc/pf/48 eq complete
                edit EPSMEMBER '48'
                edit MEMBER '48'
                edit SUBFSFAMILY '/pf048'
                task derived_param
              endfamily
              family 49
                trigger ../../../fc/pf/49 eq complete
                edit EPSMEMBER '49'
                edit MEMBER '49'
                edit SUBFSFAMILY '/pf049'
                task derived_param
              endfamily
              family 50
                trigger ../../../fc/pf/50 eq complete
                edit EPSMEMBER '50'
                edit MEMBER '50'
                edit SUBFSFAMILY '/pf050'
                task derived_param
              endfamily
            endfamily
          endfamily
          family extra
            trigger ./dp eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit STEPS '246:360:6'
            inlimit /emc_36r1/limits:extra
            family cf
              edit EMOS_TYPE 'cf'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '360'
                edit FCLENGTH0 '240'
                edit RUNVARFC '1'
            endfamily
            family pf
              complete /emc_36r1/main/ref/12/events:cf
              edit EMOS_TYPE 'pf'
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 05
                edit EPSMEMBER '5'
                edit MEMBER '5'
                edit SUBFSFAMILY '/pf005'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 06
                edit EPSMEMBER '6'
                edit MEMBER '6'
                edit SUBFSFAMILY '/pf006'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 07
                edit EPSMEMBER '7'
                edit MEMBER '7'
                edit SUBFSFAMILY '/pf007'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 08
                edit EPSMEMBER '8'
                edit MEMBER '8'
                edit SUBFSFAMILY '/pf008'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 09
                edit EPSMEMBER '9'
                edit MEMBER '9'
                edit SUBFSFAMILY '/pf009'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 10
                edit EPSMEMBER '10'
                edit MEMBER '10'
                edit SUBFSFAMILY '/pf010'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 11
                edit EPSMEMBER '11'
                edit MEMBER '11'
                edit SUBFSFAMILY '/pf011'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 12
                edit EPSMEMBER '12'
                edit MEMBER '12'
                edit SUBFSFAMILY '/pf012'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 13
                edit EPSMEMBER '13'
                edit MEMBER '13'
                edit SUBFSFAMILY '/pf013'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 14
                edit EPSMEMBER '14'
                edit MEMBER '14'
                edit SUBFSFAMILY '/pf014'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 15
                edit EPSMEMBER '15'
                edit MEMBER '15'
                edit SUBFSFAMILY '/pf015'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 16
                edit EPSMEMBER '16'
                edit MEMBER '16'
                edit SUBFSFAMILY '/pf016'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 17
                edit EPSMEMBER '17'
                edit MEMBER '17'
                edit SUBFSFAMILY '/pf017'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 18
                edit EPSMEMBER '18'
                edit MEMBER '18'
                edit SUBFSFAMILY '/pf018'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 19
                edit EPSMEMBER '19'
                edit MEMBER '19'
                edit SUBFSFAMILY '/pf019'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 20
                edit EPSMEMBER '20'
                edit MEMBER '20'
                edit SUBFSFAMILY '/pf020'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 21
                edit EPSMEMBER '21'
                edit MEMBER '21'
                edit SUBFSFAMILY '/pf021'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 22
                edit EPSMEMBER '22'
                edit MEMBER '22'
                edit SUBFSFAMILY '/pf022'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 23
                edit EPSMEMBER '23'
                edit MEMBER '23'
                edit SUBFSFAMILY '/pf023'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 24
                edit EPSMEMBER '24'
                edit MEMBER '24'
                edit SUBFSFAMILY '/pf024'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 25
                edit EPSMEMBER '25'
                edit MEMBER '25'
                edit SUBFSFAMILY '/pf025'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 26
                edit EPSMEMBER '26'
                edit MEMBER '26'
                edit SUBFSFAMILY '/pf026'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 27
                edit EPSMEMBER '27'
                edit MEMBER '27'
                edit SUBFSFAMILY '/pf027'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 28
                edit EPSMEMBER '28'
                edit MEMBER '28'
                edit SUBFSFAMILY '/pf028'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 29
                edit EPSMEMBER '29'
                edit MEMBER '29'
                edit SUBFSFAMILY '/pf029'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 30
                edit EPSMEMBER '30'
                edit MEMBER '30'
                edit SUBFSFAMILY '/pf030'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 31
                edit EPSMEMBER '31'
                edit MEMBER '31'
                edit SUBFSFAMILY '/pf031'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 32
                edit EPSMEMBER '32'
                edit MEMBER '32'
                edit SUBFSFAMILY '/pf032'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 33
                edit EPSMEMBER '33'
                edit MEMBER '33'
                edit SUBFSFAMILY '/pf033'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 34
                edit EPSMEMBER '34'
                edit MEMBER '34'
                edit SUBFSFAMILY '/pf034'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 35
                edit EPSMEMBER '35'
                edit MEMBER '35'
                edit SUBFSFAMILY '/pf035'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 36
                edit EPSMEMBER '36'
                edit MEMBER '36'
                edit SUBFSFAMILY '/pf036'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 37
                edit EPSMEMBER '37'
                edit MEMBER '37'
                edit SUBFSFAMILY '/pf037'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 38
                edit EPSMEMBER '38'
                edit MEMBER '38'
                edit SUBFSFAMILY '/pf038'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 39
                edit EPSMEMBER '39'
                edit MEMBER '39'
                edit SUBFSFAMILY '/pf039'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 40
                edit EPSMEMBER '40'
                edit MEMBER '40'
                edit SUBFSFAMILY '/pf040'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 41
                edit EPSMEMBER '41'
                edit MEMBER '41'
                edit SUBFSFAMILY '/pf041'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 42
                edit EPSMEMBER '42'
                edit MEMBER '42'
                edit SUBFSFAMILY '/pf042'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 43
                edit EPSMEMBER '43'
                edit MEMBER '43'
                edit SUBFSFAMILY '/pf043'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 44
                edit EPSMEMBER '44'
                edit MEMBER '44'
                edit SUBFSFAMILY '/pf044'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 45
                edit EPSMEMBER '45'
                edit MEMBER '45'
                edit SUBFSFAMILY '/pf045'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 46
                edit EPSMEMBER '46'
                edit MEMBER '46'
                edit SUBFSFAMILY '/pf046'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 47
                edit EPSMEMBER '47'
                edit MEMBER '47'
                edit SUBFSFAMILY '/pf047'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 48
                edit EPSMEMBER '48'
                edit MEMBER '48'
                edit SUBFSFAMILY '/pf048'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 49
                edit EPSMEMBER '49'
                edit MEMBER '49'
                edit SUBFSFAMILY '/pf049'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 50
                edit EPSMEMBER '50'
                edit MEMBER '50'
                edit SUBFSFAMILY '/pf050'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
            endfamily
          endfamily
          family calibrate
            complete /emc_36r1/main/ref/12/events:cf
            trigger /emc_36r1/main/12/ref/an eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/calibrate'
            family onws
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              inlimit /limits:linux_cluster
              task feedupcal
              task upcal
                trigger feedupcal eq complete
              task revertupcal
            endfamily
          endfamily
          family efi
            complete ( /emc_36r1/main/ref/12/events:cf) or ( /emc_36r1/main/ref/12/events:noefi)
            trigger /emc_36r1/main/12/legB/postproc/dp eq complete
            edit HC_VERSION '0046'
            edit HC_EXPVER '0046'
            edit REF_VERSION '0046'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/efi_hdc'
            edit NC_DAYS '31'
            edit EPSLEG '2'
            inlimit /emc_36r1/limits:efi
            task stage
              edit NC_DAYS '31'
            family 228_324
              trigger ./stage eq complete
              edit ACC '96'
              edit EFI_STEP1 '228'
              edit EFI_STEP2 '324'
              family 2t
                edit RF_NAME '2t'
                task extindex
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task extindex
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task extindex
              endfamily
              family tp
                edit RF_NAME 'tp'
                task extindex
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task extindex
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task extindex
              endfamily
              family ts
                edit RF_NAME 'ts'
                task extindex
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task extindex
              endfamily
            endfamily
            family 168_336
              trigger ./stage eq complete
              edit ACC '168'
              edit EFI_STEP1 '168'
              edit EFI_STEP2 '336'
              family 2t
                edit RF_NAME '2t'
                task extindex
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task extindex
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task extindex
              endfamily
              family tp
                edit RF_NAME 'tp'
                task extindex
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task extindex
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task extindex
              endfamily
              family ts
                edit RF_NAME 'ts'
                task extindex
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task extindex
              endfamily
            endfamily
            family 000_024
              trigger ./stage eq complete
              edit ACC '24'
              edit EFI_STEP1 '000'
              edit EFI_STEP2 '024'
              family 2t
                edit RF_NAME '2t'
                task extindex
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task extindex
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task extindex
              endfamily
              family tp
                edit RF_NAME 'tp'
                task extindex
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task extindex
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task extindex
              endfamily
              family ts
                edit RF_NAME 'ts'
                task extindex
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task extindex
              endfamily
            endfamily
            family 024_048
              trigger ./stage eq complete
              edit ACC '24'
              edit EFI_STEP1 '024'
              edit EFI_STEP2 '048'
              family 2t
                edit RF_NAME '2t'
                task extindex
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task extindex
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task extindex
              endfamily
              family tp
                edit RF_NAME 'tp'
                task extindex
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task extindex
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task extindex
              endfamily
              family ts
                edit RF_NAME 'ts'
                task extindex
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task extindex
              endfamily
            endfamily
            family 048_072
              trigger ./stage eq complete
              edit ACC '24'
              edit EFI_STEP1 '048'
              edit EFI_STEP2 '072'
              family 2t
                edit RF_NAME '2t'
                task extindex
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task extindex
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task extindex
              endfamily
              family tp
                edit RF_NAME 'tp'
                task extindex
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task extindex
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task extindex
              endfamily
              family ts
                edit RF_NAME 'ts'
                task extindex
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task extindex
              endfamily
            endfamily
            family 072_096
              trigger ./stage eq complete
              edit ACC '24'
              edit EFI_STEP1 '072'
              edit EFI_STEP2 '096'
              family 2t
                edit RF_NAME '2t'
                task extindex
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task extindex
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task extindex
              endfamily
              family tp
                edit RF_NAME 'tp'
                task extindex
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task extindex
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task extindex
              endfamily
              family ts
                edit RF_NAME 'ts'
                task extindex
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task extindex
              endfamily
            endfamily
            family 096_120
              trigger ./stage eq complete
              edit ACC '24'
              edit EFI_STEP1 '096'
              edit EFI_STEP2 '120'
              family 2t
                edit RF_NAME '2t'
                task extindex
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task extindex
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task extindex
              endfamily
              family tp
                edit RF_NAME 'tp'
                task extindex
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task extindex
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task extindex
              endfamily
              family ts
                edit RF_NAME 'ts'
                task extindex
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task extindex
              endfamily
            endfamily
            family 120_144
              trigger ./stage eq complete
              edit ACC '24'
              edit EFI_STEP1 '120'
              edit EFI_STEP2 '144'
              family 2t
                edit RF_NAME '2t'
                task extindex
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task extindex
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task extindex
              endfamily
              family tp
                edit RF_NAME 'tp'
                task extindex
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task extindex
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task extindex
              endfamily
              family ts
                edit RF_NAME 'ts'
                task extindex
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task extindex
              endfamily
            endfamily
            family 108_156
              trigger ./stage eq complete
              edit ACC '48'
              edit EFI_STEP1 '108'
              edit EFI_STEP2 '156'
              family 2t
                edit RF_NAME '2t'
                task extindex
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task extindex
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task extindex
              endfamily
              family tp
                edit RF_NAME 'tp'
                task extindex
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task extindex
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task extindex
              endfamily
              family ts
                edit RF_NAME 'ts'
                task extindex
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task extindex
              endfamily
            endfamily
            family 156_228
              trigger ./stage eq complete
              edit ACC '72'
              edit EFI_STEP1 '156'
              edit EFI_STEP2 '228'
              family 2t
                edit RF_NAME '2t'
                task extindex
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task extindex
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task extindex
              endfamily
              family tp
                edit RF_NAME 'tp'
                task extindex
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task extindex
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task extindex
              endfamily
              family ts
                edit RF_NAME 'ts'
                task extindex
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task extindex
              endfamily
            endfamily
            family 000_168
              trigger ./stage eq complete
              edit ACC '168'
              edit EFI_STEP1 '000'
              edit EFI_STEP2 '168'
              family 2t
                edit RF_NAME '2t'
                task extindex
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task extindex
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task extindex
              endfamily
              family tp
                edit RF_NAME 'tp'
                task extindex
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task extindex
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task extindex
              endfamily
              family ts
                edit RF_NAME 'ts'
                task extindex
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task extindex
              endfamily
            endfamily
            family 012_036
              trigger ./stage eq complete
              edit ACC '24'
              edit EFI_STEP1 '012'
              edit EFI_STEP2 '036'
              family 2tmin
                edit RF_NAME '2tmin'
                task extindex
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task extindex
              endfamily
              family ts
                edit RF_NAME 'ts'
                task extindex
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task extindex
              endfamily
            endfamily
            family 036_060
              trigger ./stage eq complete
              edit ACC '24'
              edit EFI_STEP1 '036'
              edit EFI_STEP2 '060'
              family 2tmin
                edit RF_NAME '2tmin'
                task extindex
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task extindex
              endfamily
              family ts
                edit RF_NAME 'ts'
                task extindex
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task extindex
              endfamily
            endfamily
            family 060_084
              trigger ./stage eq complete
              edit ACC '24'
              edit EFI_STEP1 '060'
              edit EFI_STEP2 '084'
              family 2tmin
                edit RF_NAME '2tmin'
                task extindex
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task extindex
              endfamily
              family ts
                edit RF_NAME 'ts'
                task extindex
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task extindex
              endfamily
            endfamily
            family 084_108
              trigger ./stage eq complete
              edit ACC '24'
              edit EFI_STEP1 '084'
              edit EFI_STEP2 '108'
              family 2tmin
                edit RF_NAME '2tmin'
                task extindex
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task extindex
              endfamily
              family ts
                edit RF_NAME 'ts'
                task extindex
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task extindex
              endfamily
            endfamily
            family 108_132
              trigger ./stage eq complete
              edit ACC '24'
              edit EFI_STEP1 '108'
              edit EFI_STEP2 '132'
              family 2tmin
                edit RF_NAME '2tmin'
                task extindex
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task extindex
              endfamily
              family ts
                edit RF_NAME 'ts'
                task extindex
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task extindex
              endfamily
            endfamily
          endfamily
          task efi2fdb
            complete ( /emc_36r1/main/ref/12/events:cf) or ( /emc_36r1/main/ref/12/events:noefi)
            trigger ./efi eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/efi_hdc'
            inlimit /emc_36r1/main/ref:mutex
            event none
        endfamily
        family prod
          edit MEM '3000'
          edit ACCOUNT 'oeenfopg'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/prod'
          edit EMOS_DOMAIN 'G'
          edit EMOS_TYPE 'cp'
          edit EMOS_PROJECT '12'
          edit PROD_TASK '1'
          edit USER_PRIORITY '90'
          edit SMSURLBASE 'http://intra.ecmwf.int/plots/d/intra/'
          edit SMSURL 'sub=fixed/'
          inlimit /limits:prod
          family eps
            trigger ( ../../legA/prod/getreq/eps eq complete) and ( ../../legA/prod/getreq/bitmap:001 or ../../legA/prod/getreq/bitmap eq complete) and ../postproc/dp eq complete
            inlimit /emc_36r1/limits:prodeps
            family 0246
              edit EMOS_TIME_STEP_H '0246'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0252
              edit EMOS_TIME_STEP_H '0252'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0258
              edit EMOS_TIME_STEP_H '0258'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0264
              edit EMOS_TIME_STEP_H '0264'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0270
              edit EMOS_TIME_STEP_H '0270'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0276
              edit EMOS_TIME_STEP_H '0276'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0282
              edit EMOS_TIME_STEP_H '0282'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0288
              edit EMOS_TIME_STEP_H '0288'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0294
              edit EMOS_TIME_STEP_H '0294'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0300
              edit EMOS_TIME_STEP_H '0300'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0306
              edit EMOS_TIME_STEP_H '0306'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0312
              edit EMOS_TIME_STEP_H '0312'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0318
              edit EMOS_TIME_STEP_H '0318'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0324
              edit EMOS_TIME_STEP_H '0324'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0330
              edit EMOS_TIME_STEP_H '0330'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0336
              edit EMOS_TIME_STEP_H '0336'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0342
              edit EMOS_TIME_STEP_H '0342'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0348
              edit EMOS_TIME_STEP_H '0348'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0354
              edit EMOS_TIME_STEP_H '0354'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0360
              edit EMOS_TIME_STEP_H '0360'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
          endfamily
          family wave
            trigger ( ../../legA/prod/getreq/wave_eps:001 or ../../legA/prod/getreq/wave_eps eq complete) and ( ../../legA/prod/getreq/bitmap:001 or ../../legA/prod/getreq/bitmap eq complete) and ../postproc/dp eq complete
            edit PGNPES '2'
            edit URL 'GWAEF/12'
            inlimit /emc_36r1/limits:prodwave
            family 0246
              edit EMOS_TIME_STEP_H '0246'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0252
              edit EMOS_TIME_STEP_H '0252'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0258
              edit EMOS_TIME_STEP_H '0258'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0264
              edit EMOS_TIME_STEP_H '0264'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0270
              edit EMOS_TIME_STEP_H '0270'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0276
              edit EMOS_TIME_STEP_H '0276'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0282
              edit EMOS_TIME_STEP_H '0282'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0288
              edit EMOS_TIME_STEP_H '0288'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0294
              edit EMOS_TIME_STEP_H '0294'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0300
              edit EMOS_TIME_STEP_H '0300'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0306
              edit EMOS_TIME_STEP_H '0306'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0312
              edit EMOS_TIME_STEP_H '0312'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0318
              edit EMOS_TIME_STEP_H '0318'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0324
              edit EMOS_TIME_STEP_H '0324'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0330
              edit EMOS_TIME_STEP_H '0330'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0336
              edit EMOS_TIME_STEP_H '0336'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0342
              edit EMOS_TIME_STEP_H '0342'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0348
              edit EMOS_TIME_STEP_H '0348'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0354
              edit EMOS_TIME_STEP_H '0354'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0360
              edit EMOS_TIME_STEP_H '0360'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
          endfamily
          task prod_wparam
            late -s +00:05 -c +00:15
            trigger ( ../../legA/prod/getreq/eps eq complete) and ../postproc/dp eq complete and ../../legA/prod/prod_wparam eq complete
            edit PROD_TASK 'w'
            edit EMOS_TYPE 'wp'
            edit EMOS_TIME_STEP_H '0360'
            edit PGNPES '52'
          task enfo_postproc
            late -s +00:05 -c +00:15
            trigger ../postproc/prob eq complete and ../postproc/ensms eq complete and ( ../../legA/prod/getreq/eps eq complete)
            edit PROD_TASK 'p'
            edit EMOS_TYPE 'ep'
            edit EMOS_TIME_STEP_H '0360'
            edit PGNPES '21'
          task waef_postproc
            late -s +00:05 -c +00:15
            trigger ( ../../legA/prod/getreq/wave_eps:001 or ../../legA/prod/getreq/wave_eps eq complete) and ../postproc/prob eq complete
            edit PROD_TASK 'p'
            edit EMOS_TYPE 'pw'
            edit EMOS_STREAM 'WE'
            edit EMOS_TIME_STEP_H '0360'
            edit PGNPES '1'
        endfamily
        family prod2diss
          edit USER_PRIORITY '90'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/prod'
          edit EMOS_DOMAIN 'G'
          edit EMOS_TYPE 'ef'
          edit EMOS_PROJECT '12'
          edit PROD_TASK '1'
          edit SMSURLBASE 'http://diss.ecmwf.int/do/monitoring/summary'
          edit SMSURL 'GENFO/12'
          family eps
            family 0246
              edit EMOS_TIME_STEP_H '0246'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0246/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../../../../legA/prod2diss/eps//0240/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0252
              edit EMOS_TIME_STEP_H '0252'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0252/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0246/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0258
              edit EMOS_TIME_STEP_H '0258'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0258/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0252/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0264
              edit EMOS_TIME_STEP_H '0264'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0264/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0258/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0270
              edit EMOS_TIME_STEP_H '0270'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0270/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0264/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0276
              edit EMOS_TIME_STEP_H '0276'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0276/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0270/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0282
              edit EMOS_TIME_STEP_H '0282'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0282/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0276/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0288
              edit EMOS_TIME_STEP_H '0288'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0288/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0282/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0294
              edit EMOS_TIME_STEP_H '0294'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0294/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0288/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0300
              edit EMOS_TIME_STEP_H '0300'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0300/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0294/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0306
              edit EMOS_TIME_STEP_H '0306'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0306/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0300/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0312
              edit EMOS_TIME_STEP_H '0312'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0312/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0306/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0318
              edit EMOS_TIME_STEP_H '0318'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0318/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0312/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0324
              edit EMOS_TIME_STEP_H '0324'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0324/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0318/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0330
              edit EMOS_TIME_STEP_H '0330'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0330/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0324/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0336
              edit EMOS_TIME_STEP_H '0336'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0336/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0330/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0342
              edit EMOS_TIME_STEP_H '0342'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0342/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0336/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0348
              edit EMOS_TIME_STEP_H '0348'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0348/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0342/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0354
              edit EMOS_TIME_STEP_H '0354'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0354/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0348/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0360
              edit EMOS_TIME_STEP_H '0360'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0360/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0354/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            task check
              trigger ./0360 eq complete
          endfamily
          family wave
            edit EMOS_STREAM 'WE'
            edit EMOS_TYPE 'we'
            edit SMSURL 'GWAEF/12'
            family 0246
              trigger ../../prod/wave/0246/prodwave eq complete and ( ../../../legA/prod2diss/wave/0240 eq complete)
              edit EMOS_TIME_STEP_H '0246'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0252
              trigger ../../prod/wave/0252/prodwave eq complete and ( ./0246 eq complete)
              edit EMOS_TIME_STEP_H '0252'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0258
              trigger ../../prod/wave/0258/prodwave eq complete and ( ./0252 eq complete)
              edit EMOS_TIME_STEP_H '0258'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0264
              trigger ../../prod/wave/0264/prodwave eq complete and ( ./0258 eq complete)
              edit EMOS_TIME_STEP_H '0264'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0270
              trigger ../../prod/wave/0270/prodwave eq complete and ( ./0264 eq complete)
              edit EMOS_TIME_STEP_H '0270'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0276
              trigger ../../prod/wave/0276/prodwave eq complete and ( ./0270 eq complete)
              edit EMOS_TIME_STEP_H '0276'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0282
              trigger ../../prod/wave/0282/prodwave eq complete and ( ./0276 eq complete)
              edit EMOS_TIME_STEP_H '0282'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0288
              trigger ../../prod/wave/0288/prodwave eq complete and ( ./0282 eq complete)
              edit EMOS_TIME_STEP_H '0288'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0294
              trigger ../../prod/wave/0294/prodwave eq complete and ( ./0288 eq complete)
              edit EMOS_TIME_STEP_H '0294'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0300
              trigger ../../prod/wave/0300/prodwave eq complete and ( ./0294 eq complete)
              edit EMOS_TIME_STEP_H '0300'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0306
              trigger ../../prod/wave/0306/prodwave eq complete and ( ./0300 eq complete)
              edit EMOS_TIME_STEP_H '0306'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0312
              trigger ../../prod/wave/0312/prodwave eq complete and ( ./0306 eq complete)
              edit EMOS_TIME_STEP_H '0312'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0318
              trigger ../../prod/wave/0318/prodwave eq complete and ( ./0312 eq complete)
              edit EMOS_TIME_STEP_H '0318'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0324
              trigger ../../prod/wave/0324/prodwave eq complete and ( ./0318 eq complete)
              edit EMOS_TIME_STEP_H '0324'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0330
              trigger ../../prod/wave/0330/prodwave eq complete and ( ./0324 eq complete)
              edit EMOS_TIME_STEP_H '0330'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0336
              trigger ../../prod/wave/0336/prodwave eq complete and ( ./0330 eq complete)
              edit EMOS_TIME_STEP_H '0336'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0342
              trigger ../../prod/wave/0342/prodwave eq complete and ( ./0336 eq complete)
              edit EMOS_TIME_STEP_H '0342'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0348
              trigger ../../prod/wave/0348/prodwave eq complete and ( ./0342 eq complete)
              edit EMOS_TIME_STEP_H '0348'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0354
              trigger ../../prod/wave/0354/prodwave eq complete and ( ./0348 eq complete)
              edit EMOS_TIME_STEP_H '0354'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0360
              trigger ../../prod/wave/0360/prodwave eq complete and ( ./0354 eq complete)
              edit EMOS_TIME_STEP_H '0360'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            task check
              trigger ./0360 eq complete
          endfamily
          task diss_wparam
            late -s +00:05 -c +00:15
            trigger ../prod/prod_wparam eq complete and ../../legA/prod2diss/diss_wparam eq complete
            edit PROD_TASK 'w'
            edit EMOS_TIME_STEP_H '0360'
            edit EMOS_TYPE 'wp'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event pdb
          task diss_waef_postproc
            late -s +00:05 -c +00:15
            trigger ../prod/waef_postproc eq complete and ../../legA/prod2diss/diss_waef_postproc eq complete
            edit PROD_TASK 'p'
            edit EMOS_TIME_STEP_H '0360'
            edit EMOS_TYPE 'pw'
            edit EMOS_STREAM 'WE'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event pdb
          task diss_postproc
            late -s +00:05 -c +00:15
            trigger ../prod/enfo_postproc eq complete and ../../legA/prod2diss/diss_postproc eq complete
            edit PROD_TASK 'p'
            edit EMOS_TIME_STEP_H '0360'
            edit EMOS_TYPE 'ep'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event pdb
        endfamily
      endfamily
    endfamily
    family 00
      trigger ( ./make eq complete) and ( ./inistage eq complete)
      edit DELTA_DAY '1'
      edit EMOS_BASE '00'
      edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
      family ref
        trigger ../make eq complete and ( /emc_36r1/main:YMD le /e_36r1/main:YMD)
        edit QUEUE_EPILOG 'emos-exp'
        edit WSHOST 'localhost'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        edit QUEUE 'emos-exp'
        limit mutex 1
        task an
          trigger /e_36r1/main/00/an eq complete
        task anfc
          complete ( ( /e_36r1/main/00/an/surf_anal/slwet eq complete and ( /e_36r1/main/00/an/4dvar/ifstraj:finalwave or /e_36r1/main/00/an/4dvar eq complete)) or /e_36r1/main/00/an eq complete)
          trigger 0 eq 1
        task model_active
          trigger /e_36r1/main/00/fc/model eq active or /e_36r1/main/00/fc/model eq complete
        task an_active
          trigger /e_36r1/main/00/an/4dvar/uptraj_0 eq active or /e_36r1/main/00/an/4dvar/uptraj_0 eq complete
        task model
          trigger /e_36r1/main/00/fc/model eq complete
        family oceini
          edit SMSFILES '/home/ma/emos/def/emc_36r1/ref'
          inlimit /emc_36r1/limits:marsret
          time 23:45
          task dummy
        endfamily
        task dc
          trigger /e_36r1/main/12dc/fc/model:step ge 12 or /e_36r1/main/12dc/fc/model eq complete
        task obs
          trigger /e_36r1/main/00/obs/get/get_stor eq complete
      endfamily
      family sv
        trigger ./ref/dc eq complete and ./ref/an_active eq complete
        edit EVOLVE 'no'
        edit FASTEX 'no'
        edit EPSTYPE 'sv'
        edit FCMEM '2000'
        edit FCSTACK '400'
        edit FCMRFS '100'
        edit FCLENGTH '48'
        edit EPSSVTCSUB '1'
        edit EPSSVOP '1'
        edit NPES '16'
        edit EPSSVTSTEP '900'
        task getini
          edit LINIENDA '0'
          edit INIDATEENDA '0'
          edit EPSCOUPLE '0'
          edit N_PERT_AN '10'
          edit EPSCOUPL_A '0'
          edit EPSCOUPL_B '0'
          edit EPSCOUPL_C '0'
        task inidata_sv
          trigger getini eq complete
          edit THREADS '8'
        task targets
          late -s +00:03 -c +00:10
          trigger ../ref/obs eq complete
          edit EPSTCBB '1'
          label da_info ""
          label dcda_info ""
          event sv
        task svsh
          trigger inidata_sv eq complete
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit CPUTIME '12000'
          edit STACK '350'
          edit THREADS '8'
          edit USER_PRIORITY '73'
        task svnh
          trigger inidata_sv eq complete
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit CPUTIME '12000'
          edit STACK '350'
          edit THREADS '8'
          edit USER_PRIORITY '74'
        task subspace
          trigger svnh eq complete and svsh eq complete and ( targets eq complete or targets:sv)
        task sv1
          trigger ( targets eq complete or targets:sv) and subspace eq complete
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit CPUTIME '12000'
          edit STACK '350'
          edit MEMBER '1'
          edit EPSTROP '1'
          edit EPSSVDIAB 'true'
          edit THREADS '8'
          edit USER_PRIORITY '72'
        task sv2
          trigger ( targets eq complete or targets:sv) and subspace eq complete
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit CPUTIME '12000'
          edit STACK '350'
          edit MEMBER '2'
          edit EPSTROP '1'
          edit EPSSVDIAB 'true'
          edit THREADS '8'
          edit USER_PRIORITY '71'
        task sv3
          trigger ( targets eq complete or targets:sv) and subspace eq complete
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit CPUTIME '12000'
          edit STACK '350'
          edit MEMBER '3'
          edit EPSTROP '1'
          edit EPSSVDIAB 'true'
          edit THREADS '8'
        task sv4
          trigger ( targets eq complete or targets:sv) and subspace eq complete
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit CPUTIME '12000'
          edit STACK '350'
          edit MEMBER '4'
          edit EPSTROP '1'
          edit EPSSVDIAB 'true'
          edit THREADS '8'
        task sv5
          trigger ( targets eq complete or targets:sv) and subspace eq complete
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit CPUTIME '12000'
          edit STACK '350'
          edit MEMBER '5'
          edit EPSTROP '1'
          edit EPSSVDIAB 'true'
          edit THREADS '8'
        task sv6
          trigger ( targets eq complete or targets:sv) and subspace eq complete
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit CPUTIME '12000'
          edit STACK '350'
          edit MEMBER '6'
          edit EPSTROP '1'
          edit EPSSVDIAB 'true'
          edit THREADS '8'
        task getsvevo
          event ecfs
        task getsvs
          trigger getsvevo eq complete and svnh eq complete and svsh eq complete and sv1 eq complete and sv2 eq complete and sv3 eq complete and sv4 eq complete and sv5 eq complete and sv6 eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
          edit EPSTROP '1'
        task sv_value
          trigger getsvs eq complete
          inlimit /emc_36r1/main/ref:mutex
        task svnhevo
          trigger svnh eq complete and ../legA/fc eq complete
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit STACK '350'
          edit HEMISPHERE 'n'
          edit EVOLVE 'yes'
          edit ANNPES '8'
          edit CPUTIME '120000'
        task svshevo
          trigger svsh eq complete and ../legA/fc eq complete
          edit NPES '24'
          edit THREADS '8'
          edit MEM '1300'
          edit MRF '10'
          edit STACK '450'
          edit EVOLVE 'yes'
          edit ANNPES '8'
          edit CPUTIME '120000'
        task clean
          trigger ../legA/fc eq complete and svnhevo eq complete and svshevo eq complete
        task dcopy
          trigger clean eq complete
          edit SMSFILES '/home/ma/emos/def/o/system'
      endfamily
      family legA
        edit EPSLEG '1'
        family fc
          trigger /emc_36r1/main/00/ref/anfc eq complete
          edit FCMEM '2240'
          edit FCSTACK '700'
          edit FCMRFS '0'
          edit PRIORITY '72'
          edit USER_PRIORITY '72'
          edit N_PERT_AN '10'
          edit SSTLEN '240'
          edit EPSINISTEP_PREVIOUS '0'
          edit EPSWAMNFRE '30'
          edit EPSWAMNANG '24'
          edit TSTEP '1200'
          edit EPSFCRES '639'
          edit RESOL '639'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
          edit CPUTIME '1000'
          edit MEM '960'
          edit THREADS '8'
          edit NPES '8'
          edit EPSLEG '1'
          edit PERSST 'true'
          edit SSTINT '1'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata
            trigger getini eq complete
            edit THREADS '8'
          task getae
            event 1 
          task rot
            trigger ../../sv/getsvs eq complete and ( getae:1 or getae eq complete)
            edit EPSTROP '1'
          family pert_ic
            trigger rot eq complete and inidata eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTROP '1'
            inlimit /emc_36r1/limits:pertic
            family 00
              task pertinic
                edit MEMBER '0'
            endfamily
            family 01
              task pertinic
                edit MEMBER '1'
            endfamily
            family 02
              task pertinic
                edit MEMBER '2'
            endfamily
            family 03
              task pertinic
                edit MEMBER '3'
            endfamily
            family 04
              task pertinic
                edit MEMBER '4'
            endfamily
            family 05
              task pertinic
                edit MEMBER '5'
            endfamily
            family 06
              task pertinic
                edit MEMBER '6'
            endfamily
            family 07
              task pertinic
                edit MEMBER '7'
            endfamily
            family 08
              task pertinic
                edit MEMBER '8'
            endfamily
            family 09
              task pertinic
                edit MEMBER '9'
            endfamily
            family 10
              task pertinic
                edit MEMBER '10'
            endfamily
            family 11
              task pertinic
                edit MEMBER '11'
            endfamily
            family 12
              task pertinic
                edit MEMBER '12'
            endfamily
            family 13
              task pertinic
                edit MEMBER '13'
            endfamily
            family 14
              task pertinic
                edit MEMBER '14'
            endfamily
            family 15
              task pertinic
                edit MEMBER '15'
            endfamily
            family 16
              task pertinic
                edit MEMBER '16'
            endfamily
            family 17
              task pertinic
                edit MEMBER '17'
            endfamily
            family 18
              task pertinic
                edit MEMBER '18'
            endfamily
            family 19
              task pertinic
                edit MEMBER '19'
            endfamily
            family 20
              task pertinic
                edit MEMBER '20'
            endfamily
            family 21
              task pertinic
                edit MEMBER '21'
            endfamily
            family 22
              task pertinic
                edit MEMBER '22'
            endfamily
            family 23
              task pertinic
                edit MEMBER '23'
            endfamily
            family 24
              task pertinic
                edit MEMBER '24'
            endfamily
            family 25
              task pertinic
                edit MEMBER '25'
            endfamily
            family 26
              task pertinic
                edit MEMBER '26'
            endfamily
            family 27
              task pertinic
                edit MEMBER '27'
            endfamily
            family 28
              task pertinic
                edit MEMBER '28'
            endfamily
            family 29
              task pertinic
                edit MEMBER '29'
            endfamily
            family 30
              task pertinic
                edit MEMBER '30'
            endfamily
            family 31
              task pertinic
                edit MEMBER '31'
            endfamily
            family 32
              task pertinic
                edit MEMBER '32'
            endfamily
            family 33
              task pertinic
                edit MEMBER '33'
            endfamily
            family 34
              task pertinic
                edit MEMBER '34'
            endfamily
            family 35
              task pertinic
                edit MEMBER '35'
            endfamily
            family 36
              task pertinic
                edit MEMBER '36'
            endfamily
            family 37
              task pertinic
                edit MEMBER '37'
            endfamily
            family 38
              task pertinic
                edit MEMBER '38'
            endfamily
            family 39
              task pertinic
                edit MEMBER '39'
            endfamily
            family 40
              task pertinic
                edit MEMBER '40'
            endfamily
            family 41
              task pertinic
                edit MEMBER '41'
            endfamily
            family 42
              task pertinic
                edit MEMBER '42'
            endfamily
            family 43
              task pertinic
                edit MEMBER '43'
            endfamily
            family 44
              task pertinic
                edit MEMBER '44'
            endfamily
            family 45
              task pertinic
                edit MEMBER '45'
            endfamily
            family 46
              task pertinic
                edit MEMBER '46'
            endfamily
            family 47
              task pertinic
                edit MEMBER '47'
            endfamily
            family 48
              task pertinic
                edit MEMBER '48'
            endfamily
            family 49
              task pertinic
                edit MEMBER '49'
            endfamily
            family 50
              task pertinic
                edit MEMBER '50'
            endfamily
          endfamily
          task cp_pert
            trigger pert_ic eq complete
            event 1 
          task wavini
            trigger /emc_36r1/main/00/ref/an eq complete
            edit FSFAMILY 'mc'
          family cv
            trigger ./pert_ic/00 eq complete and ./inidata eq complete and ./wavini eq complete and /emc_36r1/main/00/ref/model_active eq complete
            edit EPSFCLEV '62'
            edit EPSWAMNSTPW '1'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMRESOL 'global50'
            edit EPSFCRES '639'
            edit TSTEP '1200'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit CLASS 'od'
            edit PERSST 'no'
            edit EPSTYPE 'cv'
            edit EMOS_TYPE 'cv'
            edit EPSNLEGS '1'
            edit EPSLEG '1'
            edit EPSINISTEP '360'
            edit FCLENGTH '360'
            edit USER_PRIORITY '73'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            inlimit /emc_36r1/limits:pftotal
            family control_1
              edit MEMBER '1'
              edit EPSMEMBER '1'
              edit SUBFSFAMILY '/cv001'
              task getiniLeg
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit FCNPES '32'
                edit THREADS '4'
                meter step -1 360 360
              task derived_param
                trigger modeleps eq complete
                edit STEPS '0:144:3 144:360:6'
            endfamily
            family control_2
              edit MEMBER '2'
              edit EPSMEMBER '2'
              edit SUBFSFAMILY '/cv002'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              task getiniLeg
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task intHtoL
                trigger getiniLeg eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '1200'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP_PREVIOUS '0'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit FCNPES '32'
                edit THREADS '4'
                meter step -1 360 360
              task derived_param
                trigger modeleps eq complete
                edit STEPS '0:144:3 144:360:6'
            endfamily
          endfamily
          task wavfcdata
            trigger ./getpersSST eq complete and ../../legA/fc/wavini eq complete
            edit EPSCOUPLE '1'
          task getpersSST
            trigger ./inidata eq complete
            edit EPSCOUPLE '1'
            edit EPS_INI_STREAM 'DA'
            edit EPSFCRES '639'
          family pf
            complete /emc_36r1/main/ref/00/events:cf
            trigger ./getpersSST eq complete and ./wavfcdata eq complete and wavini eq complete and ../../ref/model_active eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTYPE 'pf'
            edit EMOS_TYPE 'pf'
            edit USER_PRIORITY '72'
            edit PRIORITY '72'
            edit FCNPES '64'
            edit EPSNLEGS '3'
            edit EPSFCLEV '62'
            edit EPSWAMNSTPW '1'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMRESOL 'global50'
            edit EPSFCRES '639'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit EPSINISTEP '216'
            edit FCLENGTH '240'
            edit TSTEP '1200'
            edit EPSLEG '1'
            edit CLASS 'od'
            inlimit /emc_36r1/limits:pftotal
            meter step -1 240 240
            family 01
              edit EPSMEMBER '1'
              edit MEMBER '1'
              edit SUBFSFAMILY '/pf001'
              task getiniLeg
                trigger ../../pert_ic/01/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 02
              edit EPSMEMBER '2'
              edit MEMBER '2'
              edit SUBFSFAMILY '/pf002'
              task getiniLeg
                trigger ../../pert_ic/02/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 03
              edit EPSMEMBER '3'
              edit MEMBER '3'
              edit SUBFSFAMILY '/pf003'
              task getiniLeg
                trigger ../../pert_ic/03/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 04
              edit EPSMEMBER '4'
              edit MEMBER '4'
              edit SUBFSFAMILY '/pf004'
              task getiniLeg
                trigger ../../pert_ic/04/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 05
              edit EPSMEMBER '5'
              edit MEMBER '5'
              edit SUBFSFAMILY '/pf005'
              task getiniLeg
                trigger ../../pert_ic/05/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 06
              edit EPSMEMBER '6'
              edit MEMBER '6'
              edit SUBFSFAMILY '/pf006'
              task getiniLeg
                trigger ../../pert_ic/06/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 07
              edit EPSMEMBER '7'
              edit MEMBER '7'
              edit SUBFSFAMILY '/pf007'
              task getiniLeg
                trigger ../../pert_ic/07/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 08
              edit EPSMEMBER '8'
              edit MEMBER '8'
              edit SUBFSFAMILY '/pf008'
              task getiniLeg
                trigger ../../pert_ic/08/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 09
              edit EPSMEMBER '9'
              edit MEMBER '9'
              edit SUBFSFAMILY '/pf009'
              task getiniLeg
                trigger ../../pert_ic/09/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 10
              edit EPSMEMBER '10'
              edit MEMBER '10'
              edit SUBFSFAMILY '/pf010'
              task getiniLeg
                trigger ../../pert_ic/10/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 11
              edit EPSMEMBER '11'
              edit MEMBER '11'
              edit SUBFSFAMILY '/pf011'
              task getiniLeg
                trigger ../../pert_ic/11/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 12
              edit EPSMEMBER '12'
              edit MEMBER '12'
              edit SUBFSFAMILY '/pf012'
              task getiniLeg
                trigger ../../pert_ic/12/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 13
              edit EPSMEMBER '13'
              edit MEMBER '13'
              edit SUBFSFAMILY '/pf013'
              task getiniLeg
                trigger ../../pert_ic/13/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 14
              edit EPSMEMBER '14'
              edit MEMBER '14'
              edit SUBFSFAMILY '/pf014'
              task getiniLeg
                trigger ../../pert_ic/14/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 15
              edit EPSMEMBER '15'
              edit MEMBER '15'
              edit SUBFSFAMILY '/pf015'
              task getiniLeg
                trigger ../../pert_ic/15/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 16
              edit EPSMEMBER '16'
              edit MEMBER '16'
              edit SUBFSFAMILY '/pf016'
              task getiniLeg
                trigger ../../pert_ic/16/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 17
              edit EPSMEMBER '17'
              edit MEMBER '17'
              edit SUBFSFAMILY '/pf017'
              task getiniLeg
                trigger ../../pert_ic/17/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 18
              edit EPSMEMBER '18'
              edit MEMBER '18'
              edit SUBFSFAMILY '/pf018'
              task getiniLeg
                trigger ../../pert_ic/18/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 19
              edit EPSMEMBER '19'
              edit MEMBER '19'
              edit SUBFSFAMILY '/pf019'
              task getiniLeg
                trigger ../../pert_ic/19/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 20
              edit EPSMEMBER '20'
              edit MEMBER '20'
              edit SUBFSFAMILY '/pf020'
              task getiniLeg
                trigger ../../pert_ic/20/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 21
              edit EPSMEMBER '21'
              edit MEMBER '21'
              edit SUBFSFAMILY '/pf021'
              task getiniLeg
                trigger ../../pert_ic/21/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 22
              edit EPSMEMBER '22'
              edit MEMBER '22'
              edit SUBFSFAMILY '/pf022'
              task getiniLeg
                trigger ../../pert_ic/22/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 23
              edit EPSMEMBER '23'
              edit MEMBER '23'
              edit SUBFSFAMILY '/pf023'
              task getiniLeg
                trigger ../../pert_ic/23/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 24
              edit EPSMEMBER '24'
              edit MEMBER '24'
              edit SUBFSFAMILY '/pf024'
              task getiniLeg
                trigger ../../pert_ic/24/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 25
              edit EPSMEMBER '25'
              edit MEMBER '25'
              edit SUBFSFAMILY '/pf025'
              task getiniLeg
                trigger ../../pert_ic/25/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 26
              edit EPSMEMBER '26'
              edit MEMBER '26'
              edit SUBFSFAMILY '/pf026'
              task getiniLeg
                trigger ../../pert_ic/26/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 27
              edit EPSMEMBER '27'
              edit MEMBER '27'
              edit SUBFSFAMILY '/pf027'
              task getiniLeg
                trigger ../../pert_ic/27/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 28
              edit EPSMEMBER '28'
              edit MEMBER '28'
              edit SUBFSFAMILY '/pf028'
              task getiniLeg
                trigger ../../pert_ic/28/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 29
              edit EPSMEMBER '29'
              edit MEMBER '29'
              edit SUBFSFAMILY '/pf029'
              task getiniLeg
                trigger ../../pert_ic/29/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 30
              edit EPSMEMBER '30'
              edit MEMBER '30'
              edit SUBFSFAMILY '/pf030'
              task getiniLeg
                trigger ../../pert_ic/30/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 31
              edit EPSMEMBER '31'
              edit MEMBER '31'
              edit SUBFSFAMILY '/pf031'
              task getiniLeg
                trigger ../../pert_ic/31/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 32
              edit EPSMEMBER '32'
              edit MEMBER '32'
              edit SUBFSFAMILY '/pf032'
              task getiniLeg
                trigger ../../pert_ic/32/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 33
              edit EPSMEMBER '33'
              edit MEMBER '33'
              edit SUBFSFAMILY '/pf033'
              task getiniLeg
                trigger ../../pert_ic/33/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 34
              edit EPSMEMBER '34'
              edit MEMBER '34'
              edit SUBFSFAMILY '/pf034'
              task getiniLeg
                trigger ../../pert_ic/34/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 35
              edit EPSMEMBER '35'
              edit MEMBER '35'
              edit SUBFSFAMILY '/pf035'
              task getiniLeg
                trigger ../../pert_ic/35/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 36
              edit EPSMEMBER '36'
              edit MEMBER '36'
              edit SUBFSFAMILY '/pf036'
              task getiniLeg
                trigger ../../pert_ic/36/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 37
              edit EPSMEMBER '37'
              edit MEMBER '37'
              edit SUBFSFAMILY '/pf037'
              task getiniLeg
                trigger ../../pert_ic/37/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 38
              edit EPSMEMBER '38'
              edit MEMBER '38'
              edit SUBFSFAMILY '/pf038'
              task getiniLeg
                trigger ../../pert_ic/38/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 39
              edit EPSMEMBER '39'
              edit MEMBER '39'
              edit SUBFSFAMILY '/pf039'
              task getiniLeg
                trigger ../../pert_ic/39/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 40
              edit EPSMEMBER '40'
              edit MEMBER '40'
              edit SUBFSFAMILY '/pf040'
              task getiniLeg
                trigger ../../pert_ic/40/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 41
              edit EPSMEMBER '41'
              edit MEMBER '41'
              edit SUBFSFAMILY '/pf041'
              task getiniLeg
                trigger ../../pert_ic/41/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 42
              edit EPSMEMBER '42'
              edit MEMBER '42'
              edit SUBFSFAMILY '/pf042'
              task getiniLeg
                trigger ../../pert_ic/42/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 43
              edit EPSMEMBER '43'
              edit MEMBER '43'
              edit SUBFSFAMILY '/pf043'
              task getiniLeg
                trigger ../../pert_ic/43/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 44
              edit EPSMEMBER '44'
              edit MEMBER '44'
              edit SUBFSFAMILY '/pf044'
              task getiniLeg
                trigger ../../pert_ic/44/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 45
              edit EPSMEMBER '45'
              edit MEMBER '45'
              edit SUBFSFAMILY '/pf045'
              task getiniLeg
                trigger ../../pert_ic/45/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 46
              edit EPSMEMBER '46'
              edit MEMBER '46'
              edit SUBFSFAMILY '/pf046'
              task getiniLeg
                trigger ../../pert_ic/46/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 47
              edit EPSMEMBER '47'
              edit MEMBER '47'
              edit SUBFSFAMILY '/pf047'
              task getiniLeg
                trigger ../../pert_ic/47/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 48
              edit EPSMEMBER '48'
              edit MEMBER '48'
              edit SUBFSFAMILY '/pf048'
              task getiniLeg
                trigger ../../pert_ic/48/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 49
              edit EPSMEMBER '49'
              edit MEMBER '49'
              edit SUBFSFAMILY '/pf049'
              task getiniLeg
                trigger ../../pert_ic/49/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
            family 50
              edit EPSMEMBER '50'
              edit MEMBER '50'
              edit SUBFSFAMILY '/pf050'
              task getiniLeg
                trigger ../../pert_ic/50/pertinic eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit THREADS '4'
                inlimit /emc_36r1/limits:modelepsA
                meter step -1 240 240
            endfamily
          endfamily
          family cf
            trigger ./pert_ic/00/pertinic eq complete and ./getpersSST eq complete and ./wavfcdata eq complete and /emc_36r1/main/00/ref/model_active eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit MEMBER '0'
            edit EPSMEMBER '0'
            edit SUBFSFAMILY '/cf000'
            edit EPSTYPE 'cf'
            edit EMOS_TYPE 'cf'
            edit USER_PRIORITY '73'
            edit EPSNLEGS '3'
            edit EPSFCLEV '62'
            edit EPSWAMNSTPW '1'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMRESOL 'global50'
            edit EPSFCRES '639'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit EPSINISTEP '216'
            edit FCLENGTH '240'
            edit TSTEP '1200'
            edit EPSLEG '1'
            edit CLASS 'od'
            task getiniLeg
              trigger ../inidata eq complete and ../wavini eq complete
              edit LINIENDA '0'
              edit INIDATEENDA '0'
              edit EPSCOUPLE '0'
              edit N_PERT_AN '10'
              edit EPSCOUPL_A '0'
              edit EPSCOUPL_B '0'
              edit EPSCOUPL_C '0'
            task modeleps
              trigger getiniLeg eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSLEG '1'
              edit EPSFCRES '639'
              edit EPSINISTEP '216'
              edit EPSINISTEP_PREVIOUS '0'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit CLASS 'od'
              edit FCNPES '32'
              edit THREADS '4'
              meter step -1 240 240
          endfamily
          family reservation1
            trigger ./pf/01/modeleps eq active or ./pf/01/modeleps eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/smsfiles'
            inlimit /emc_36r1/limits:reservation
            task reserve_cpu
              late -s +00:02 -c +00:02
              edit RESERVED_CPUS '10880'
              edit RESERVE_FOR '40'
              label info ""
            task rm_res
              late -s +00:02 -c +00:02
              trigger ./reserve_cpu eq complete
              edit RESERVATION_LABEL 'emos00'
              label info ""
          endfamily
          family reservation2
            trigger ( ./pf/46/modeleps eq active or ./pf/46/modeleps eq complete) and ( ./pf/47/modeleps eq active or ./pf/47/modeleps eq complete) and ( ./pf/48/modeleps eq active or ./pf/48/modeleps eq complete) and ( ./pf/49/modeleps eq active or ./pf/49/modeleps eq complete) and ( ./pf/50/modeleps eq active or ./pf/50/modeleps eq complete)
            edit SMSFILES '/home/ma/emos/def/emc_36r1/smsfiles'
            inlimit /emc_36r1/limits:reservation
            task reset_class
              late -s +00:02 -c +00:02
              edit RESERVATION_DEFINITION 'default'
              label info ""
            task ms_cpu
              late -s +00:02 -c +00:02
              edit MS_CPUS '$((64*80))'
              label info ""
            task reserve_cpu
              late -s +00:02 -c +00:02
              complete /emc_36r1/main/ref/00/events:768
              edit RESERVED_CPUS '6400'
              edit RESERVE_FOR '40'
              label info ""
            task reserve_cpu1
              late -s +00:02 -c +00:02
              complete ( /emc_36r1/main/ref/00/events eq complete) and ( not /emc_36r1/main/ref/00/events:768)
              edit RESERVED_CPUS '6400'
              edit RESERVE_FOR '100'
              label info ""
          endfamily
          family reservation3
            trigger ( ./pf eq complete)
            edit SMSFILES '/home/ma/emos/def/emc_36r1/smsfiles'
            inlimit /emc_36r1/limits:reservation
            task reserve_cpu
              late -s +00:02 -c +00:02
              complete /emc_36r1/main/ref/00/events:768
              edit RESERVED_CPUS '3200'
              edit RESERVE_FOR '30'
              label info ""
            task reserve_cpu1
              late -s +00:02 -c +00:02
              complete ( /emc_36r1/main/ref/00/events eq complete) and ( not /emc_36r1/main/ref/00/events:768)
              edit RESERVED_CPUS '3200'
              edit RESERVE_FOR '80'
              label info ""
          endfamily
        endfamily
        family fax
          trigger /emc_36r1/main/00/ref/an eq complete
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /limits:linux_cluster
          task ena4
            trigger ../postproc/clusters eq complete and ../postproc/ensms eq complete
            edit ANALYSIS 'NO'
          task plume
            trigger ../postproc/prob eq complete and /emc_36r1/main/00/ref/model eq complete
            edit ANALYSIS 'YES'
          task fax
            trigger plume eq complete and ena4 eq complete
            edit SMSTRIES '1'
          task manfax
            trigger plume eq complete and ena4 eq complete
            edit SMSTRIES '1'
            edit COUNTRY ''
          task faxcheck
            trigger fax eq complete
          task faxcheck2
            trigger fax eq complete
            time 08:00
        endfamily
        family postproc
          trigger ./fc ne queued
          edit STEPS '24/240/12/24 168/168/12/48 240/240/12/72 240/240/12/120'
          inlimit /emc_36r1/limits:postproc
          family prepare
            complete /emc_36r1/main/ref/00/events:cf
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/leg'
            family all
              complete ../../fc eq complete and ./query eq active
              trigger ../../fc eq complete and ( ./query eq queued or ./query eq complete)
              edit KIND 'all'
              edit STEP '-1'
              family t850
                edit PARAM 'T'
                edit LEVEL '850'
                task pp_prep
              endfamily
              family t500
                edit PARAM 'T'
                edit LEVEL '500'
                task pp_prep
              endfamily
              family z1000
                edit PARAM 'Z'
                edit LEVEL '1000'
                task pp_prep
              endfamily
              family z500
                edit PARAM 'Z'
                edit LEVEL '500'
                task pp_prep
              endfamily
            endfamily
            family query
              complete ../../fc eq complete and ../../fc/pf:step le 0
              trigger ( ( ../../fc/pf:step ge 0) and ( ../../fc/cf/modeleps:step ge 0)) or ( ../../fc/pf eq complete and ../../fc/cf eq complete)
              edit KIND 'query'
              family t850
                edit PARAM 'T'
                edit LEVEL '850'
                task pp_prep
                  meter step -1 240 240
              endfamily
              family t500
                edit PARAM 'T'
                edit LEVEL '500'
                task pp_prep
                  meter step -1 240 240
              endfamily
              family z1000
                edit PARAM 'Z'
                edit LEVEL '1000'
                task pp_prep
                  meter step -1 240 240
              endfamily
              family z500
                edit PARAM 'Z'
                edit LEVEL '500'
                task pp_prep
                  meter step -1 240 240
              endfamily
            endfamily
          endfamily
          family ensms
            complete /emc_36r1/main/ref/00/events:cf
            trigger ( ../fc eq complete) and ( ./prepare eq complete)
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/leg'
            family limits
              limit fdb 1
              limit gen 8
            endfamily
            task ensms
              edit ENS_M 'false'
              edit ENS_S 'false'
              edit ENS_MARS 'false'
              event stage
            family 048
              trigger ./ensms:stage or ./ensms eq complete
              edit STEP1 '003'
              edit STEP2 '048'
              edit BY '3'
              edit STEPS '0'
              family T
                edit ENS_M 'false'
                edit ENS_S 'false'
                edit LIST_PARAM 'T'
                inlimit ../limits:gen
                family 850
                  task ensms
                    edit LIST_LEV '850'
                endfamily
                family 500
                  task ensms
                    edit LIST_LEV '500'
                endfamily
              endfamily
              family Z
                edit ENS_M 'false'
                edit ENS_S 'false'
                edit LIST_PARAM 'Z'
                inlimit ../limits:gen
                family 1000
                  task ensms
                    edit LIST_LEV '1000'
                endfamily
                family 500
                  task ensms
                    edit LIST_LEV '500'
                endfamily
              endfamily
              task ensms
                trigger T eq complete and Z eq complete
                edit ENS_MARS 'false'
                inlimit ../limits:fdb
            endfamily
            family 096
              trigger ./ensms:stage or ./ensms eq complete
              edit STEP1 '051'
              edit STEP2 '096'
              edit BY '3'
              edit STEPS '0'
              family T
                edit ENS_M 'false'
                edit ENS_S 'false'
                edit LIST_PARAM 'T'
                inlimit ../limits:gen
                family 850
                  task ensms
                    edit LIST_LEV '850'
                endfamily
                family 500
                  task ensms
                    edit LIST_LEV '500'
                endfamily
              endfamily
              family Z
                edit ENS_M 'false'
                edit ENS_S 'false'
                edit LIST_PARAM 'Z'
                inlimit ../limits:gen
                family 1000
                  task ensms
                    edit LIST_LEV '1000'
                endfamily
                family 500
                  task ensms
                    edit LIST_LEV '500'
                endfamily
              endfamily
              task ensms
                trigger T eq complete and Z eq complete
                edit ENS_MARS 'false'
                inlimit ../limits:fdb
            endfamily
            family 144
              trigger ./ensms:stage or ./ensms eq complete
              edit STEP1 '099'
              edit STEP2 '144'
              edit BY '3'
              edit STEPS '0'
              family T
                edit ENS_M 'false'
                edit ENS_S 'false'
                edit LIST_PARAM 'T'
                inlimit ../limits:gen
                family 850
                  task ensms
                    edit LIST_LEV '850'
                endfamily
                family 500
                  task ensms
                    edit LIST_LEV '500'
                endfamily
              endfamily
              family Z
                edit ENS_M 'false'
                edit ENS_S 'false'
                edit LIST_PARAM 'Z'
                inlimit ../limits:gen
                family 1000
                  task ensms
                    edit LIST_LEV '1000'
                endfamily
                family 500
                  task ensms
                    edit LIST_LEV '500'
                endfamily
              endfamily
              task ensms
                trigger T eq complete and Z eq complete
                edit ENS_MARS 'false'
                inlimit ../limits:fdb
            endfamily
            family 240
              trigger ./ensms:stage or ./ensms eq complete
              edit STEP1 '150'
              edit STEP2 '240'
              edit BY '6'
              edit STEPS '0'
              family T
                edit ENS_M 'false'
                edit ENS_S 'false'
                edit LIST_PARAM 'T'
                inlimit ../limits:gen
                family 850
                  task ensms
                    edit LIST_LEV '850'
                endfamily
                family 500
                  task ensms
                    edit LIST_LEV '500'
                endfamily
              endfamily
              family Z
                edit ENS_M 'false'
                edit ENS_S 'false'
                edit LIST_PARAM 'Z'
                inlimit ../limits:gen
                family 1000
                  task ensms
                    edit LIST_LEV '1000'
                endfamily
                family 500
                  task ensms
                    edit LIST_LEV '500'
                endfamily
              endfamily
              task ensms
                trigger T eq complete and Z eq complete
                edit ENS_MARS 'false'
                inlimit ../limits:fdb
            endfamily
          endfamily
          family prob
            complete /emc_36r1/main/ref/00/events:cf
            trigger /emc_36r1/main/00/legA/fc eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/leg'
            edit STEP1 '12'
            edit STEP2 '240'
            edit BY '12'
            family 24
              edit STEP1 '24'
              edit STEP2 '240'
              edit BY '24'
              family t850
                trigger ../../prepare/all/t850 eq complete
                family 120_240
                  edit STEPS '120 240'
                  task t850
                    event 1 
                endfamily
                family 120_168
                  trigger ./120_240/t850:1 or ./120_240 eq complete
                  edit STEPS '120 168'
                  task t850
                    event 1 
                endfamily
                family 168_240
                  trigger ./120_240/t850:1 or ./120_240 eq complete
                  edit STEPS '168 240'
                  task t850
                    event 1 
                endfamily
              endfamily
              task rain
              task ff
            endfamily
            family 12
              edit STEP1 '12'
              edit STEP2 '228'
              edit BY '24'
              task t850
                trigger ../../prepare/all/t850 eq complete
              task rain
              task ff
            endfamily
            task 2t
            task swh
            task mwp
            task 10fg
              trigger ../dp eq complete
            task prob_fdb
              trigger ./2t eq complete and swh eq complete and mwp eq complete and ./10fg eq complete and ./12 eq complete and ./24 eq complete
          endfamily
          family tc
            complete /emc_36r1/main/ref/00/events:cf
            trigger ../fc eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/tc'
            edit TCINIEXPVER '0046'
            family obsdata
              trigger ( /emc_36r1/main/00/legA/fc/pf eq complete and /emc_36r1/main/00/legA/fc/cf eq complete)
              task retrieve
                late -c +00:10
                trigger ( /emc_36r1/main/00/sv/targets eq complete or /emc_36r1/main/00/sv/targets:sv)
                label info ""
                event notc
              task obs
                trigger retrieve eq complete
            endfamily
            family stage
              trigger ( /emc_36r1/main/00/legA/fc eq complete)
              task an
              task fc
              task pf
              task cf
            endfamily
            family tracking
              trigger obsdata/obs eq complete and stage eq complete
              inlimit /emc_36r1/limits:tc
              family 1
                edit MEMBER '1'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 2
                edit MEMBER '2'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 3
                edit MEMBER '3'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 4
                edit MEMBER '4'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 5
                edit MEMBER '5'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 6
                edit MEMBER '6'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 7
                edit MEMBER '7'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 8
                edit MEMBER '8'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 9
                edit MEMBER '9'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 10
                edit MEMBER '10'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 11
                edit MEMBER '11'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 12
                edit MEMBER '12'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 13
                edit MEMBER '13'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 14
                edit MEMBER '14'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 15
                edit MEMBER '15'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 16
                edit MEMBER '16'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 17
                edit MEMBER '17'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 18
                edit MEMBER '18'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 19
                edit MEMBER '19'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 20
                edit MEMBER '20'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 21
                edit MEMBER '21'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 22
                edit MEMBER '22'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 23
                edit MEMBER '23'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 24
                edit MEMBER '24'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 25
                edit MEMBER '25'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 26
                edit MEMBER '26'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 27
                edit MEMBER '27'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 28
                edit MEMBER '28'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 29
                edit MEMBER '29'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 30
                edit MEMBER '30'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 31
                edit MEMBER '31'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 32
                edit MEMBER '32'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 33
                edit MEMBER '33'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 34
                edit MEMBER '34'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 35
                edit MEMBER '35'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 36
                edit MEMBER '36'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 37
                edit MEMBER '37'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 38
                edit MEMBER '38'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 39
                edit MEMBER '39'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 40
                edit MEMBER '40'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 41
                edit MEMBER '41'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 42
                edit MEMBER '42'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 43
                edit MEMBER '43'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 44
                edit MEMBER '44'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 45
                edit MEMBER '45'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 46
                edit MEMBER '46'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 47
                edit MEMBER '47'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 48
                edit MEMBER '48'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 49
                edit MEMBER '49'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family 50
                edit MEMBER '50'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family cf
                edit MEMBER '51'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
              family oper
                edit MEMBER '52'
                edit SLEEP '1'
                task get_data
                task hilo
                  trigger get_data eq complete
                task tracking
                  trigger hilo eq complete
              endfamily
            endfamily
            family post_process
              task to_bufr
                trigger ../tracking eq complete
            endfamily
          endfamily
          family calibrate
            complete /emc_36r1/main/ref/00/events:cf
            trigger /emc_36r1/main/00/ref/an eq complete and ../fc eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/calibrate'
            family onsc
              edit QUEUE 'ns'
              edit QUEUE_EPILOG 'ns'
              edit SMSCMD 'smssubmit %USER% %SCHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %SCHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %SCHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/s1b/emos_esuite/emos_data/0046/log'
              edit SMSOUT '/s1b/emos_esuite/emos_data/0046/log'
              inlimit /limits:c1b
              task feedupcal
              task upcal
                trigger feedupcal eq complete
                edit Z500_EXPVER '0001'
                edit METVIEW 'metview'
              task revertupcal
            endfamily
          endfamily
          family clusters
            complete /emc_36r1/main/ref/00/events:cf
            trigger ./calibrate/onsc eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/clusters'
            edit USER_PRIORITY '81'
            edit QUEUE 'ns'
            edit QUEUE_EPILOG 'ns'
            edit SMSCMD 'smssubmit %USER% %SCHOST% %SMSJOB% %SMSJOBOUT%'
            edit SMSKILL 'smskill %USER% %SCHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
            edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %SCHOST% %SMSRID% %SMSJOB%'
            edit LOGDIR '/s1b/emos_esuite/emos_data/0046/log'
            edit SMSOUT '/s1b/emos_esuite/emos_data/0046/log'
            inlimit /limits:c1b
            task cthresh
            task cluster
              trigger cthresh eq complete and /emc_36r1/main/00/ref/model eq complete
              edit USER_PRIORITY '76'
            task tubcon
              trigger /emc_36r1/main/00/ref/model eq complete
            task tubgen
              late -c +00:20
              trigger tubcon eq complete
          endfamily
          family efi
            complete ( /emc_36r1/main/ref/00/events:cf) or ( /emc_36r1/main/ref/00/events:noefi)
            trigger /emc_36r1/main/00/legA/postproc/dp eq complete
            edit HC_VERSION '0046'
            edit HC_EXPVER '0046'
            edit REF_VERSION '0046'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/efi_hdc'
            edit NC_DAYS '31'
            edit EPSLEG '1'
            inlimit /emc_36r1/limits:efi
            task stage
              edit NC_DAYS '31'
            family 000_120
              trigger ./stage eq complete
              edit ACC '120'
              edit EFI_STEP1 '000'
              edit EFI_STEP2 '120'
              family tp
                edit RF_NAME 'tp'
                task extindex
              endfamily
            endfamily
            family 024_144
              trigger ./stage eq complete
              edit ACC '120'
              edit EFI_STEP1 '024'
              edit EFI_STEP2 '144'
              family tp
                edit RF_NAME 'tp'
                task extindex
              endfamily
            endfamily
            family 000_240
              trigger ./stage eq complete
              edit ACC '240'
              edit EFI_STEP1 '000'
              edit EFI_STEP2 '240'
              family tp
                edit RF_NAME 'tp'
                task extindex
              endfamily
            endfamily
            family 000_024
              trigger ./stage eq complete
              edit ACC '24'
              edit EFI_STEP1 '000'
              edit EFI_STEP2 '024'
              family 2t
                edit RF_NAME '2t'
                task extindex
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task extindex
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task extindex
              endfamily
              family tp
                edit RF_NAME 'tp'
                task extindex
              endfamily
            endfamily
            family 024_048
              trigger ./stage eq complete
              edit ACC '24'
              edit EFI_STEP1 '024'
              edit EFI_STEP2 '048'
              family 2t
                edit RF_NAME '2t'
                task extindex
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task extindex
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task extindex
              endfamily
              family tp
                edit RF_NAME 'tp'
                task extindex
              endfamily
            endfamily
            family 048_072
              trigger ./stage eq complete
              edit ACC '24'
              edit EFI_STEP1 '048'
              edit EFI_STEP2 '072'
              family 2t
                edit RF_NAME '2t'
                task extindex
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task extindex
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task extindex
              endfamily
              family tp
                edit RF_NAME 'tp'
                task extindex
              endfamily
            endfamily
            family 072_096
              trigger ./stage eq complete
              edit ACC '24'
              edit EFI_STEP1 '072'
              edit EFI_STEP2 '096'
              family 2t
                edit RF_NAME '2t'
                task extindex
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task extindex
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task extindex
              endfamily
              family tp
                edit RF_NAME 'tp'
                task extindex
              endfamily
            endfamily
            family 096_120
              trigger ./stage eq complete
              edit ACC '24'
              edit EFI_STEP1 '096'
              edit EFI_STEP2 '120'
              family 2t
                edit RF_NAME '2t'
                task extindex
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task extindex
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task extindex
              endfamily
              family tp
                edit RF_NAME 'tp'
                task extindex
              endfamily
            endfamily
          endfamily
          task efi2fdb
            complete ( /emc_36r1/main/ref/00/events:cf) or ( /emc_36r1/main/ref/00/events:noefi)
            trigger ./efi eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/efi_hdc'
            inlimit /emc_36r1/main/ref:mutex
            event none
          family dp
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit STEPS '0:144:3 144:240:6'
            inlimit /emc_36r1/limits:dp
            family cf
              trigger ../../fc/cf eq complete
              edit EMOS_TYPE 'cf'
              task derived_param
                edit MEMBER 'off'
            endfamily
            family pf
              complete /emc_36r1/main/ref/00/events:cf
              edit EMOS_TYPE 'pf'
              family 01
                trigger ../../../fc/pf/01 eq complete
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task derived_param
              endfamily
              family 02
                trigger ../../../fc/pf/02 eq complete
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task derived_param
              endfamily
              family 03
                trigger ../../../fc/pf/03 eq complete
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task derived_param
              endfamily
              family 04
                trigger ../../../fc/pf/04 eq complete
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task derived_param
              endfamily
              family 05
                trigger ../../../fc/pf/05 eq complete
                edit EPSMEMBER '5'
                edit MEMBER '5'
                edit SUBFSFAMILY '/pf005'
                task derived_param
              endfamily
              family 06
                trigger ../../../fc/pf/06 eq complete
                edit EPSMEMBER '6'
                edit MEMBER '6'
                edit SUBFSFAMILY '/pf006'
                task derived_param
              endfamily
              family 07
                trigger ../../../fc/pf/07 eq complete
                edit EPSMEMBER '7'
                edit MEMBER '7'
                edit SUBFSFAMILY '/pf007'
                task derived_param
              endfamily
              family 08
                trigger ../../../fc/pf/08 eq complete
                edit EPSMEMBER '8'
                edit MEMBER '8'
                edit SUBFSFAMILY '/pf008'
                task derived_param
              endfamily
              family 09
                trigger ../../../fc/pf/09 eq complete
                edit EPSMEMBER '9'
                edit MEMBER '9'
                edit SUBFSFAMILY '/pf009'
                task derived_param
              endfamily
              family 10
                trigger ../../../fc/pf/10 eq complete
                edit EPSMEMBER '10'
                edit MEMBER '10'
                edit SUBFSFAMILY '/pf010'
                task derived_param
              endfamily
              family 11
                trigger ../../../fc/pf/11 eq complete
                edit EPSMEMBER '11'
                edit MEMBER '11'
                edit SUBFSFAMILY '/pf011'
                task derived_param
              endfamily
              family 12
                trigger ../../../fc/pf/12 eq complete
                edit EPSMEMBER '12'
                edit MEMBER '12'
                edit SUBFSFAMILY '/pf012'
                task derived_param
              endfamily
              family 13
                trigger ../../../fc/pf/13 eq complete
                edit EPSMEMBER '13'
                edit MEMBER '13'
                edit SUBFSFAMILY '/pf013'
                task derived_param
              endfamily
              family 14
                trigger ../../../fc/pf/14 eq complete
                edit EPSMEMBER '14'
                edit MEMBER '14'
                edit SUBFSFAMILY '/pf014'
                task derived_param
              endfamily
              family 15
                trigger ../../../fc/pf/15 eq complete
                edit EPSMEMBER '15'
                edit MEMBER '15'
                edit SUBFSFAMILY '/pf015'
                task derived_param
              endfamily
              family 16
                trigger ../../../fc/pf/16 eq complete
                edit EPSMEMBER '16'
                edit MEMBER '16'
                edit SUBFSFAMILY '/pf016'
                task derived_param
              endfamily
              family 17
                trigger ../../../fc/pf/17 eq complete
                edit EPSMEMBER '17'
                edit MEMBER '17'
                edit SUBFSFAMILY '/pf017'
                task derived_param
              endfamily
              family 18
                trigger ../../../fc/pf/18 eq complete
                edit EPSMEMBER '18'
                edit MEMBER '18'
                edit SUBFSFAMILY '/pf018'
                task derived_param
              endfamily
              family 19
                trigger ../../../fc/pf/19 eq complete
                edit EPSMEMBER '19'
                edit MEMBER '19'
                edit SUBFSFAMILY '/pf019'
                task derived_param
              endfamily
              family 20
                trigger ../../../fc/pf/20 eq complete
                edit EPSMEMBER '20'
                edit MEMBER '20'
                edit SUBFSFAMILY '/pf020'
                task derived_param
              endfamily
              family 21
                trigger ../../../fc/pf/21 eq complete
                edit EPSMEMBER '21'
                edit MEMBER '21'
                edit SUBFSFAMILY '/pf021'
                task derived_param
              endfamily
              family 22
                trigger ../../../fc/pf/22 eq complete
                edit EPSMEMBER '22'
                edit MEMBER '22'
                edit SUBFSFAMILY '/pf022'
                task derived_param
              endfamily
              family 23
                trigger ../../../fc/pf/23 eq complete
                edit EPSMEMBER '23'
                edit MEMBER '23'
                edit SUBFSFAMILY '/pf023'
                task derived_param
              endfamily
              family 24
                trigger ../../../fc/pf/24 eq complete
                edit EPSMEMBER '24'
                edit MEMBER '24'
                edit SUBFSFAMILY '/pf024'
                task derived_param
              endfamily
              family 25
                trigger ../../../fc/pf/25 eq complete
                edit EPSMEMBER '25'
                edit MEMBER '25'
                edit SUBFSFAMILY '/pf025'
                task derived_param
              endfamily
              family 26
                trigger ../../../fc/pf/26 eq complete
                edit EPSMEMBER '26'
                edit MEMBER '26'
                edit SUBFSFAMILY '/pf026'
                task derived_param
              endfamily
              family 27
                trigger ../../../fc/pf/27 eq complete
                edit EPSMEMBER '27'
                edit MEMBER '27'
                edit SUBFSFAMILY '/pf027'
                task derived_param
              endfamily
              family 28
                trigger ../../../fc/pf/28 eq complete
                edit EPSMEMBER '28'
                edit MEMBER '28'
                edit SUBFSFAMILY '/pf028'
                task derived_param
              endfamily
              family 29
                trigger ../../../fc/pf/29 eq complete
                edit EPSMEMBER '29'
                edit MEMBER '29'
                edit SUBFSFAMILY '/pf029'
                task derived_param
              endfamily
              family 30
                trigger ../../../fc/pf/30 eq complete
                edit EPSMEMBER '30'
                edit MEMBER '30'
                edit SUBFSFAMILY '/pf030'
                task derived_param
              endfamily
              family 31
                trigger ../../../fc/pf/31 eq complete
                edit EPSMEMBER '31'
                edit MEMBER '31'
                edit SUBFSFAMILY '/pf031'
                task derived_param
              endfamily
              family 32
                trigger ../../../fc/pf/32 eq complete
                edit EPSMEMBER '32'
                edit MEMBER '32'
                edit SUBFSFAMILY '/pf032'
                task derived_param
              endfamily
              family 33
                trigger ../../../fc/pf/33 eq complete
                edit EPSMEMBER '33'
                edit MEMBER '33'
                edit SUBFSFAMILY '/pf033'
                task derived_param
              endfamily
              family 34
                trigger ../../../fc/pf/34 eq complete
                edit EPSMEMBER '34'
                edit MEMBER '34'
                edit SUBFSFAMILY '/pf034'
                task derived_param
              endfamily
              family 35
                trigger ../../../fc/pf/35 eq complete
                edit EPSMEMBER '35'
                edit MEMBER '35'
                edit SUBFSFAMILY '/pf035'
                task derived_param
              endfamily
              family 36
                trigger ../../../fc/pf/36 eq complete
                edit EPSMEMBER '36'
                edit MEMBER '36'
                edit SUBFSFAMILY '/pf036'
                task derived_param
              endfamily
              family 37
                trigger ../../../fc/pf/37 eq complete
                edit EPSMEMBER '37'
                edit MEMBER '37'
                edit SUBFSFAMILY '/pf037'
                task derived_param
              endfamily
              family 38
                trigger ../../../fc/pf/38 eq complete
                edit EPSMEMBER '38'
                edit MEMBER '38'
                edit SUBFSFAMILY '/pf038'
                task derived_param
              endfamily
              family 39
                trigger ../../../fc/pf/39 eq complete
                edit EPSMEMBER '39'
                edit MEMBER '39'
                edit SUBFSFAMILY '/pf039'
                task derived_param
              endfamily
              family 40
                trigger ../../../fc/pf/40 eq complete
                edit EPSMEMBER '40'
                edit MEMBER '40'
                edit SUBFSFAMILY '/pf040'
                task derived_param
              endfamily
              family 41
                trigger ../../../fc/pf/41 eq complete
                edit EPSMEMBER '41'
                edit MEMBER '41'
                edit SUBFSFAMILY '/pf041'
                task derived_param
              endfamily
              family 42
                trigger ../../../fc/pf/42 eq complete
                edit EPSMEMBER '42'
                edit MEMBER '42'
                edit SUBFSFAMILY '/pf042'
                task derived_param
              endfamily
              family 43
                trigger ../../../fc/pf/43 eq complete
                edit EPSMEMBER '43'
                edit MEMBER '43'
                edit SUBFSFAMILY '/pf043'
                task derived_param
              endfamily
              family 44
                trigger ../../../fc/pf/44 eq complete
                edit EPSMEMBER '44'
                edit MEMBER '44'
                edit SUBFSFAMILY '/pf044'
                task derived_param
              endfamily
              family 45
                trigger ../../../fc/pf/45 eq complete
                edit EPSMEMBER '45'
                edit MEMBER '45'
                edit SUBFSFAMILY '/pf045'
                task derived_param
              endfamily
              family 46
                trigger ../../../fc/pf/46 eq complete
                edit EPSMEMBER '46'
                edit MEMBER '46'
                edit SUBFSFAMILY '/pf046'
                task derived_param
              endfamily
              family 47
                trigger ../../../fc/pf/47 eq complete
                edit EPSMEMBER '47'
                edit MEMBER '47'
                edit SUBFSFAMILY '/pf047'
                task derived_param
              endfamily
              family 48
                trigger ../../../fc/pf/48 eq complete
                edit EPSMEMBER '48'
                edit MEMBER '48'
                edit SUBFSFAMILY '/pf048'
                task derived_param
              endfamily
              family 49
                trigger ../../../fc/pf/49 eq complete
                edit EPSMEMBER '49'
                edit MEMBER '49'
                edit SUBFSFAMILY '/pf049'
                task derived_param
              endfamily
              family 50
                trigger ../../../fc/pf/50 eq complete
                edit EPSMEMBER '50'
                edit MEMBER '50'
                edit SUBFSFAMILY '/pf050'
                task derived_param
              endfamily
            endfamily
          endfamily
          family extra
            trigger ./dp eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit STEPS '0:144:3 144:240:6'
            inlimit /emc_36r1/limits:extra
            family cf
              edit EMOS_TYPE 'cf'
              task extrafields
                edit EPSFCRES '639'
                edit FCLENGTH '240'
                edit FCLENGTH0 '0'
                edit RUNVARFC '1'
            endfamily
            family pf
              complete /emc_36r1/main/ref/00/events:cf
              edit EMOS_TYPE 'pf'
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 05
                edit EPSMEMBER '5'
                edit MEMBER '5'
                edit SUBFSFAMILY '/pf005'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 06
                edit EPSMEMBER '6'
                edit MEMBER '6'
                edit SUBFSFAMILY '/pf006'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 07
                edit EPSMEMBER '7'
                edit MEMBER '7'
                edit SUBFSFAMILY '/pf007'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 08
                edit EPSMEMBER '8'
                edit MEMBER '8'
                edit SUBFSFAMILY '/pf008'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 09
                edit EPSMEMBER '9'
                edit MEMBER '9'
                edit SUBFSFAMILY '/pf009'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 10
                edit EPSMEMBER '10'
                edit MEMBER '10'
                edit SUBFSFAMILY '/pf010'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 11
                edit EPSMEMBER '11'
                edit MEMBER '11'
                edit SUBFSFAMILY '/pf011'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 12
                edit EPSMEMBER '12'
                edit MEMBER '12'
                edit SUBFSFAMILY '/pf012'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 13
                edit EPSMEMBER '13'
                edit MEMBER '13'
                edit SUBFSFAMILY '/pf013'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 14
                edit EPSMEMBER '14'
                edit MEMBER '14'
                edit SUBFSFAMILY '/pf014'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 15
                edit EPSMEMBER '15'
                edit MEMBER '15'
                edit SUBFSFAMILY '/pf015'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 16
                edit EPSMEMBER '16'
                edit MEMBER '16'
                edit SUBFSFAMILY '/pf016'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 17
                edit EPSMEMBER '17'
                edit MEMBER '17'
                edit SUBFSFAMILY '/pf017'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 18
                edit EPSMEMBER '18'
                edit MEMBER '18'
                edit SUBFSFAMILY '/pf018'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 19
                edit EPSMEMBER '19'
                edit MEMBER '19'
                edit SUBFSFAMILY '/pf019'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 20
                edit EPSMEMBER '20'
                edit MEMBER '20'
                edit SUBFSFAMILY '/pf020'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 21
                edit EPSMEMBER '21'
                edit MEMBER '21'
                edit SUBFSFAMILY '/pf021'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 22
                edit EPSMEMBER '22'
                edit MEMBER '22'
                edit SUBFSFAMILY '/pf022'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 23
                edit EPSMEMBER '23'
                edit MEMBER '23'
                edit SUBFSFAMILY '/pf023'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 24
                edit EPSMEMBER '24'
                edit MEMBER '24'
                edit SUBFSFAMILY '/pf024'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 25
                edit EPSMEMBER '25'
                edit MEMBER '25'
                edit SUBFSFAMILY '/pf025'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 26
                edit EPSMEMBER '26'
                edit MEMBER '26'
                edit SUBFSFAMILY '/pf026'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 27
                edit EPSMEMBER '27'
                edit MEMBER '27'
                edit SUBFSFAMILY '/pf027'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 28
                edit EPSMEMBER '28'
                edit MEMBER '28'
                edit SUBFSFAMILY '/pf028'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 29
                edit EPSMEMBER '29'
                edit MEMBER '29'
                edit SUBFSFAMILY '/pf029'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 30
                edit EPSMEMBER '30'
                edit MEMBER '30'
                edit SUBFSFAMILY '/pf030'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 31
                edit EPSMEMBER '31'
                edit MEMBER '31'
                edit SUBFSFAMILY '/pf031'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 32
                edit EPSMEMBER '32'
                edit MEMBER '32'
                edit SUBFSFAMILY '/pf032'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 33
                edit EPSMEMBER '33'
                edit MEMBER '33'
                edit SUBFSFAMILY '/pf033'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 34
                edit EPSMEMBER '34'
                edit MEMBER '34'
                edit SUBFSFAMILY '/pf034'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 35
                edit EPSMEMBER '35'
                edit MEMBER '35'
                edit SUBFSFAMILY '/pf035'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 36
                edit EPSMEMBER '36'
                edit MEMBER '36'
                edit SUBFSFAMILY '/pf036'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 37
                edit EPSMEMBER '37'
                edit MEMBER '37'
                edit SUBFSFAMILY '/pf037'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 38
                edit EPSMEMBER '38'
                edit MEMBER '38'
                edit SUBFSFAMILY '/pf038'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 39
                edit EPSMEMBER '39'
                edit MEMBER '39'
                edit SUBFSFAMILY '/pf039'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 40
                edit EPSMEMBER '40'
                edit MEMBER '40'
                edit SUBFSFAMILY '/pf040'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 41
                edit EPSMEMBER '41'
                edit MEMBER '41'
                edit SUBFSFAMILY '/pf041'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 42
                edit EPSMEMBER '42'
                edit MEMBER '42'
                edit SUBFSFAMILY '/pf042'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 43
                edit EPSMEMBER '43'
                edit MEMBER '43'
                edit SUBFSFAMILY '/pf043'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 44
                edit EPSMEMBER '44'
                edit MEMBER '44'
                edit SUBFSFAMILY '/pf044'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 45
                edit EPSMEMBER '45'
                edit MEMBER '45'
                edit SUBFSFAMILY '/pf045'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 46
                edit EPSMEMBER '46'
                edit MEMBER '46'
                edit SUBFSFAMILY '/pf046'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 47
                edit EPSMEMBER '47'
                edit MEMBER '47'
                edit SUBFSFAMILY '/pf047'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 48
                edit EPSMEMBER '48'
                edit MEMBER '48'
                edit SUBFSFAMILY '/pf048'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 49
                edit EPSMEMBER '49'
                edit MEMBER '49'
                edit SUBFSFAMILY '/pf049'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family 50
                edit EPSMEMBER '50'
                edit MEMBER '50'
                edit SUBFSFAMILY '/pf050'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
            endfamily
          endfamily
        endfamily
        family prod
          edit MEM '3000'
          edit ACCOUNT 'oeenfopg'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/prod'
          edit EMOS_DOMAIN 'G'
          edit EMOS_TYPE 'cp'
          edit EMOS_PROJECT '12'
          edit PROD_TASK '1'
          edit USER_PRIORITY '90'
          edit SMSURLBASE 'http://intra.ecmwf.int/plots/d/intra/'
          edit SMSURL 'sub=fixed/'
          inlimit /limits:prod
          family eps
            trigger ( ../../legA/prod/getreq/eps eq complete) and ( ../../legA/prod/getreq/bitmap:001 or ../../legA/prod/getreq/bitmap eq complete) and ../postproc/dp eq complete
            inlimit /emc_36r1/limits:prodeps
            family 0000
              edit EMOS_TIME_STEP_H '0000'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0003
              edit EMOS_TIME_STEP_H '0003'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0006
              edit EMOS_TIME_STEP_H '0006'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0009
              edit EMOS_TIME_STEP_H '0009'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0012
              edit EMOS_TIME_STEP_H '0012'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0015
              edit EMOS_TIME_STEP_H '0015'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0018
              edit EMOS_TIME_STEP_H '0018'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0021
              edit EMOS_TIME_STEP_H '0021'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0024
              edit EMOS_TIME_STEP_H '0024'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0027
              edit EMOS_TIME_STEP_H '0027'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0030
              edit EMOS_TIME_STEP_H '0030'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0033
              edit EMOS_TIME_STEP_H '0033'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0036
              edit EMOS_TIME_STEP_H '0036'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0039
              edit EMOS_TIME_STEP_H '0039'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0042
              edit EMOS_TIME_STEP_H '0042'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0045
              edit EMOS_TIME_STEP_H '0045'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0048
              edit EMOS_TIME_STEP_H '0048'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0051
              edit EMOS_TIME_STEP_H '0051'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0054
              edit EMOS_TIME_STEP_H '0054'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0057
              edit EMOS_TIME_STEP_H '0057'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0060
              edit EMOS_TIME_STEP_H '0060'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0063
              edit EMOS_TIME_STEP_H '0063'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0066
              edit EMOS_TIME_STEP_H '0066'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0069
              edit EMOS_TIME_STEP_H '0069'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0072
              edit EMOS_TIME_STEP_H '0072'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0075
              edit EMOS_TIME_STEP_H '0075'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0078
              edit EMOS_TIME_STEP_H '0078'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0081
              edit EMOS_TIME_STEP_H '0081'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0084
              edit EMOS_TIME_STEP_H '0084'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0087
              edit EMOS_TIME_STEP_H '0087'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0090
              edit EMOS_TIME_STEP_H '0090'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0093
              edit EMOS_TIME_STEP_H '0093'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0096
              edit EMOS_TIME_STEP_H '0096'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0099
              edit EMOS_TIME_STEP_H '0099'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0102
              edit EMOS_TIME_STEP_H '0102'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0105
              edit EMOS_TIME_STEP_H '0105'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0108
              edit EMOS_TIME_STEP_H '0108'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0111
              edit EMOS_TIME_STEP_H '0111'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0114
              edit EMOS_TIME_STEP_H '0114'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0117
              edit EMOS_TIME_STEP_H '0117'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0120
              edit EMOS_TIME_STEP_H '0120'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0123
              edit EMOS_TIME_STEP_H '0123'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0126
              edit EMOS_TIME_STEP_H '0126'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0129
              edit EMOS_TIME_STEP_H '0129'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0132
              edit EMOS_TIME_STEP_H '0132'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0135
              edit EMOS_TIME_STEP_H '0135'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0138
              edit EMOS_TIME_STEP_H '0138'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0141
              edit EMOS_TIME_STEP_H '0141'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0144
              edit EMOS_TIME_STEP_H '0144'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0150
              edit EMOS_TIME_STEP_H '0150'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0156
              edit EMOS_TIME_STEP_H '0156'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0162
              edit EMOS_TIME_STEP_H '0162'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0168
              edit EMOS_TIME_STEP_H '0168'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0174
              edit EMOS_TIME_STEP_H '0174'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0180
              edit EMOS_TIME_STEP_H '0180'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0186
              edit EMOS_TIME_STEP_H '0186'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0192
              edit EMOS_TIME_STEP_H '0192'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0198
              edit EMOS_TIME_STEP_H '0198'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0204
              edit EMOS_TIME_STEP_H '0204'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0210
              edit EMOS_TIME_STEP_H '0210'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0216
              edit EMOS_TIME_STEP_H '0216'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0222
              edit EMOS_TIME_STEP_H '0222'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0228
              edit EMOS_TIME_STEP_H '0228'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0234
              edit EMOS_TIME_STEP_H '0234'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0240
              edit EMOS_TIME_STEP_H '0240'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
          endfamily
          family wave
            trigger ( ../../legA/prod/getreq/wave_eps:001 or ../../legA/prod/getreq/wave_eps eq complete) and ( ../../legA/prod/getreq/bitmap:001 or ../../legA/prod/getreq/bitmap eq complete) and ../postproc/dp eq complete
            edit PGNPES '2'
            edit URL 'GWAEF/00'
            inlimit /emc_36r1/limits:prodwave
            family 0000
              edit EMOS_TIME_STEP_H '0000'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0003
              edit EMOS_TIME_STEP_H '0003'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0006
              edit EMOS_TIME_STEP_H '0006'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0009
              edit EMOS_TIME_STEP_H '0009'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0012
              edit EMOS_TIME_STEP_H '0012'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0015
              edit EMOS_TIME_STEP_H '0015'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0018
              edit EMOS_TIME_STEP_H '0018'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0021
              edit EMOS_TIME_STEP_H '0021'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0024
              edit EMOS_TIME_STEP_H '0024'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0027
              edit EMOS_TIME_STEP_H '0027'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0030
              edit EMOS_TIME_STEP_H '0030'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0033
              edit EMOS_TIME_STEP_H '0033'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0036
              edit EMOS_TIME_STEP_H '0036'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0039
              edit EMOS_TIME_STEP_H '0039'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0042
              edit EMOS_TIME_STEP_H '0042'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0045
              edit EMOS_TIME_STEP_H '0045'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0048
              edit EMOS_TIME_STEP_H '0048'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0051
              edit EMOS_TIME_STEP_H '0051'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0054
              edit EMOS_TIME_STEP_H '0054'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0057
              edit EMOS_TIME_STEP_H '0057'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0060
              edit EMOS_TIME_STEP_H '0060'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0063
              edit EMOS_TIME_STEP_H '0063'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0066
              edit EMOS_TIME_STEP_H '0066'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0069
              edit EMOS_TIME_STEP_H '0069'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0072
              edit EMOS_TIME_STEP_H '0072'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0075
              edit EMOS_TIME_STEP_H '0075'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0078
              edit EMOS_TIME_STEP_H '0078'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0081
              edit EMOS_TIME_STEP_H '0081'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0084
              edit EMOS_TIME_STEP_H '0084'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0087
              edit EMOS_TIME_STEP_H '0087'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0090
              edit EMOS_TIME_STEP_H '0090'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0093
              edit EMOS_TIME_STEP_H '0093'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0096
              edit EMOS_TIME_STEP_H '0096'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0099
              edit EMOS_TIME_STEP_H '0099'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0102
              edit EMOS_TIME_STEP_H '0102'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0105
              edit EMOS_TIME_STEP_H '0105'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0108
              edit EMOS_TIME_STEP_H '0108'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0111
              edit EMOS_TIME_STEP_H '0111'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0114
              edit EMOS_TIME_STEP_H '0114'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0117
              edit EMOS_TIME_STEP_H '0117'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0120
              edit EMOS_TIME_STEP_H '0120'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0123
              edit EMOS_TIME_STEP_H '0123'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0126
              edit EMOS_TIME_STEP_H '0126'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0129
              edit EMOS_TIME_STEP_H '0129'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0132
              edit EMOS_TIME_STEP_H '0132'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0135
              edit EMOS_TIME_STEP_H '0135'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0138
              edit EMOS_TIME_STEP_H '0138'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0141
              edit EMOS_TIME_STEP_H '0141'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0144
              edit EMOS_TIME_STEP_H '0144'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0150
              edit EMOS_TIME_STEP_H '0150'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0156
              edit EMOS_TIME_STEP_H '0156'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0162
              edit EMOS_TIME_STEP_H '0162'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0168
              edit EMOS_TIME_STEP_H '0168'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0174
              edit EMOS_TIME_STEP_H '0174'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0180
              edit EMOS_TIME_STEP_H '0180'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0186
              edit EMOS_TIME_STEP_H '0186'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0192
              edit EMOS_TIME_STEP_H '0192'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0198
              edit EMOS_TIME_STEP_H '0198'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0204
              edit EMOS_TIME_STEP_H '0204'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0210
              edit EMOS_TIME_STEP_H '0210'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0216
              edit EMOS_TIME_STEP_H '0216'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0222
              edit EMOS_TIME_STEP_H '0222'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0228
              edit EMOS_TIME_STEP_H '0228'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0234
              edit EMOS_TIME_STEP_H '0234'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0240
              edit EMOS_TIME_STEP_H '0240'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
          endfamily
          family getreq
            trigger ../fc/pf eq active or ../fc/pf/01 eq complete
            edit EMOS_TYPE 'cm'
            edit EMOS_DOMAIN 'G'
            edit EMOS_PROJECT '12'
            edit PROD_TASK '1'
            edit SMSTRIES '1'
            edit EMOS_STREAM 'ENFO'
            task collectreq
              edit QUEUE 'diss'
              edit USER 'emos'
              edit WSHOST 'diss'
              edit QUEUE_EPILOG 'emos-exp'
              edit LOGDIR '/diss/log'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit SMSOUT '/vol/emos_nc/output'
              edit EMOS_BASE '00'
            task eps
              late -s +00:03 -c +00:10
              trigger collectreq eq complete
              label info ""
              event 1 
            task collectwave
              trigger collectreq eq complete
              edit QUEUE 'diss'
              edit USER 'emos'
              edit WSHOST 'diss'
              edit QUEUE_EPILOG 'emos-exp'
              edit LOGDIR '/diss/log'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit SMSOUT '/vol/emos_nc/output'
              edit EMOS_STREAM 'WAEF'
              edit EMOS_BASE '00'
            task wave_eps
              late -s +00:03 -c +00:10
              trigger collectwave eq complete
              edit EMOS_STREAM 'WAEF'
              edit USER_PRIORITY '89'
              label info ""
              event 1 
            task collectbitmap
              trigger collectwave eq complete
              edit QUEUE 'diss'
              edit USER 'emos'
              edit WSHOST 'diss'
              edit QUEUE_EPILOG 'emos-exp'
              edit LOGDIR '/diss/log'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit SMSOUT '/vol/emos_nc/output'
            task bitmap
              late -s +00:03 -c +00:10
              trigger collectbitmap eq complete and ( eps:001 or eps eq complete)
              event 1 
          endfamily
          task enfo_tc
            late -s +00:05 -c +00:15
            trigger ( ./getreq/eps:001 or ./getreq/eps eq complete) and ../postproc/tc eq complete
            edit PROD_TASK 't'
            edit EMOS_TYPE 'tf'
            edit EMOS_TIME_STEP_H '0240'
            edit PGNPES '1'
          task oper_tc
            late -s +00:05 -c +00:15
            trigger ./enfo_tc eq complete
            edit PROD_TASK 't'
            edit EMOS_TYPE 'tf'
            edit EMOS_TIME_STEP_H '0240'
            edit PGNPES '1'
            edit EMOS_STREAM 'OPER'
          task prod_wparam
            late -s +00:05 -c +00:15
            trigger ( ../../legA/prod/getreq/eps eq complete) and ../postproc/dp eq complete and /e_36r1/main/00/prod/240/prod_wparam eq complete
            edit PROD_TASK 'w'
            edit EMOS_TYPE 'wp'
            edit EMOS_TIME_STEP_H '0240'
            edit PGNPES '52'
          task enfo_postproc
            late -s +00:05 -c +00:15
            trigger ../postproc eq complete and ( ../../legA/prod/getreq/eps eq complete) and ../../legA/postproc/efi eq complete and ../../legA/postproc/clusters eq complete
            edit PROD_TASK 'p'
            edit EMOS_TYPE 'ep'
            edit EMOS_TIME_STEP_H '0240'
            edit PGNPES '21'
          task waef_postproc
            late -s +00:05 -c +00:15
            trigger ( ../../legA/prod/getreq/wave_eps:001 or ../../legA/prod/getreq/wave_eps eq complete) and ../postproc/prob eq complete
            edit PROD_TASK 'p'
            edit EMOS_TYPE 'pw'
            edit EMOS_STREAM 'WE'
            edit EMOS_TIME_STEP_H '0240'
            edit PGNPES '1'
        endfamily
        family prod2diss
          edit USER_PRIORITY '90'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/prod'
          edit EMOS_DOMAIN 'G'
          edit EMOS_TYPE 'ef'
          edit EMOS_PROJECT '12'
          edit PROD_TASK '1'
          edit SMSURLBASE 'http://diss.ecmwf.int/do/monitoring/summary'
          edit SMSURL 'GENFO/00'
          task operation_is_late
            event yes
          family eps
            family 0000
              edit EMOS_TIME_STEP_H '0000'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0000/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0003
              edit EMOS_TIME_STEP_H '0003'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0003/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0000/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0006
              edit EMOS_TIME_STEP_H '0006'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0006/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0003/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0009
              edit EMOS_TIME_STEP_H '0009'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0009/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0006/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0012
              edit EMOS_TIME_STEP_H '0012'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0012/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0009/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0015
              edit EMOS_TIME_STEP_H '0015'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0015/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0012/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0018
              edit EMOS_TIME_STEP_H '0018'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0018/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0015/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0021
              edit EMOS_TIME_STEP_H '0021'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0021/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0018/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0024
              edit EMOS_TIME_STEP_H '0024'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0024/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0021/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0027
              edit EMOS_TIME_STEP_H '0027'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0027/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0024/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0030
              edit EMOS_TIME_STEP_H '0030'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0030/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0027/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0033
              edit EMOS_TIME_STEP_H '0033'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0033/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0030/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0036
              edit EMOS_TIME_STEP_H '0036'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0036/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0033/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0039
              edit EMOS_TIME_STEP_H '0039'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0039/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0036/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0042
              edit EMOS_TIME_STEP_H '0042'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0042/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0039/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0045
              edit EMOS_TIME_STEP_H '0045'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0045/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0042/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0048
              edit EMOS_TIME_STEP_H '0048'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0048/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0045/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0051
              edit EMOS_TIME_STEP_H '0051'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0051/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0048/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0054
              edit EMOS_TIME_STEP_H '0054'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0054/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0051/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0057
              edit EMOS_TIME_STEP_H '0057'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0057/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0054/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0060
              edit EMOS_TIME_STEP_H '0060'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0060/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0057/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0063
              edit EMOS_TIME_STEP_H '0063'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0063/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0060/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0066
              edit EMOS_TIME_STEP_H '0066'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0066/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0063/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0069
              edit EMOS_TIME_STEP_H '0069'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0069/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0066/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0072
              edit EMOS_TIME_STEP_H '0072'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0072/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0069/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0075
              edit EMOS_TIME_STEP_H '0075'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0075/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0072/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0078
              edit EMOS_TIME_STEP_H '0078'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0078/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0075/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0081
              edit EMOS_TIME_STEP_H '0081'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0081/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0078/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0084
              edit EMOS_TIME_STEP_H '0084'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0084/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0081/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0087
              edit EMOS_TIME_STEP_H '0087'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0087/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0084/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0090
              edit EMOS_TIME_STEP_H '0090'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0090/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0087/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0093
              edit EMOS_TIME_STEP_H '0093'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0093/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0090/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0096
              edit EMOS_TIME_STEP_H '0096'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0096/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0093/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0099
              edit EMOS_TIME_STEP_H '0099'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0099/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0096/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0102
              edit EMOS_TIME_STEP_H '0102'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0102/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0099/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0105
              edit EMOS_TIME_STEP_H '0105'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0105/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0102/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0108
              edit EMOS_TIME_STEP_H '0108'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0108/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0105/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0111
              edit EMOS_TIME_STEP_H '0111'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0111/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0108/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0114
              edit EMOS_TIME_STEP_H '0114'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0114/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0111/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0117
              edit EMOS_TIME_STEP_H '0117'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0117/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0114/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0120
              edit EMOS_TIME_STEP_H '0120'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0120/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0117/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0123
              edit EMOS_TIME_STEP_H '0123'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0123/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0120/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0126
              edit EMOS_TIME_STEP_H '0126'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0126/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0123/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0129
              edit EMOS_TIME_STEP_H '0129'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0129/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0126/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0132
              edit EMOS_TIME_STEP_H '0132'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0132/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0129/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0135
              edit EMOS_TIME_STEP_H '0135'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0135/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0132/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0138
              edit EMOS_TIME_STEP_H '0138'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0138/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0135/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0141
              edit EMOS_TIME_STEP_H '0141'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0141/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0138/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0144
              edit EMOS_TIME_STEP_H '0144'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0144/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0141/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0150
              edit EMOS_TIME_STEP_H '0150'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0150/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0144/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0156
              edit EMOS_TIME_STEP_H '0156'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0156/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0150/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0162
              edit EMOS_TIME_STEP_H '0162'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0162/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0156/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0168
              edit EMOS_TIME_STEP_H '0168'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0168/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0162/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0174
              edit EMOS_TIME_STEP_H '0174'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0174/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0168/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0180
              edit EMOS_TIME_STEP_H '0180'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0180/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0174/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0186
              edit EMOS_TIME_STEP_H '0186'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0186/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0180/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0192
              edit EMOS_TIME_STEP_H '0192'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0192/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0186/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0198
              edit EMOS_TIME_STEP_H '0198'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0198/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0192/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0204
              edit EMOS_TIME_STEP_H '0204'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0204/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0198/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0210
              edit EMOS_TIME_STEP_H '0210'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0210/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0204/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0216
              edit EMOS_TIME_STEP_H '0216'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0216/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0210/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0222
              edit EMOS_TIME_STEP_H '0222'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0222/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0216/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0228
              edit EMOS_TIME_STEP_H '0228'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0228/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0222/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0234
              edit EMOS_TIME_STEP_H '0234'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0234/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0228/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0240
              edit EMOS_TIME_STEP_H '0240'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0240/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0234/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            task check
              trigger ./0240 eq complete
          endfamily
          family wave
            edit EMOS_STREAM 'WE'
            edit EMOS_TYPE 'we'
            edit SMSURL 'GWAEF/00'
            family 0000
              trigger ../../prod/wave/0000/prodwave eq complete
              edit EMOS_TIME_STEP_H '0000'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0003
              trigger ../../prod/wave/0003/prodwave eq complete and ( ./0000 eq complete)
              edit EMOS_TIME_STEP_H '0003'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0006
              trigger ../../prod/wave/0006/prodwave eq complete and ( ./0003 eq complete)
              edit EMOS_TIME_STEP_H '0006'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0009
              trigger ../../prod/wave/0009/prodwave eq complete and ( ./0006 eq complete)
              edit EMOS_TIME_STEP_H '0009'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0012
              trigger ../../prod/wave/0012/prodwave eq complete and ( ./0009 eq complete)
              edit EMOS_TIME_STEP_H '0012'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0015
              trigger ../../prod/wave/0015/prodwave eq complete and ( ./0012 eq complete)
              edit EMOS_TIME_STEP_H '0015'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0018
              trigger ../../prod/wave/0018/prodwave eq complete and ( ./0015 eq complete)
              edit EMOS_TIME_STEP_H '0018'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0021
              trigger ../../prod/wave/0021/prodwave eq complete and ( ./0018 eq complete)
              edit EMOS_TIME_STEP_H '0021'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0024
              trigger ../../prod/wave/0024/prodwave eq complete and ( ./0021 eq complete)
              edit EMOS_TIME_STEP_H '0024'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0027
              trigger ../../prod/wave/0027/prodwave eq complete and ( ./0024 eq complete)
              edit EMOS_TIME_STEP_H '0027'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0030
              trigger ../../prod/wave/0030/prodwave eq complete and ( ./0027 eq complete)
              edit EMOS_TIME_STEP_H '0030'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0033
              trigger ../../prod/wave/0033/prodwave eq complete and ( ./0030 eq complete)
              edit EMOS_TIME_STEP_H '0033'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0036
              trigger ../../prod/wave/0036/prodwave eq complete and ( ./0033 eq complete)
              edit EMOS_TIME_STEP_H '0036'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0039
              trigger ../../prod/wave/0039/prodwave eq complete and ( ./0036 eq complete)
              edit EMOS_TIME_STEP_H '0039'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0042
              trigger ../../prod/wave/0042/prodwave eq complete and ( ./0039 eq complete)
              edit EMOS_TIME_STEP_H '0042'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0045
              trigger ../../prod/wave/0045/prodwave eq complete and ( ./0042 eq complete)
              edit EMOS_TIME_STEP_H '0045'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0048
              trigger ../../prod/wave/0048/prodwave eq complete and ( ./0045 eq complete)
              edit EMOS_TIME_STEP_H '0048'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0051
              trigger ../../prod/wave/0051/prodwave eq complete and ( ./0048 eq complete)
              edit EMOS_TIME_STEP_H '0051'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0054
              trigger ../../prod/wave/0054/prodwave eq complete and ( ./0051 eq complete)
              edit EMOS_TIME_STEP_H '0054'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0057
              trigger ../../prod/wave/0057/prodwave eq complete and ( ./0054 eq complete)
              edit EMOS_TIME_STEP_H '0057'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0060
              trigger ../../prod/wave/0060/prodwave eq complete and ( ./0057 eq complete)
              edit EMOS_TIME_STEP_H '0060'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0063
              trigger ../../prod/wave/0063/prodwave eq complete and ( ./0060 eq complete)
              edit EMOS_TIME_STEP_H '0063'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0066
              trigger ../../prod/wave/0066/prodwave eq complete and ( ./0063 eq complete)
              edit EMOS_TIME_STEP_H '0066'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0069
              trigger ../../prod/wave/0069/prodwave eq complete and ( ./0066 eq complete)
              edit EMOS_TIME_STEP_H '0069'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0072
              trigger ../../prod/wave/0072/prodwave eq complete and ( ./0069 eq complete)
              edit EMOS_TIME_STEP_H '0072'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0075
              trigger ../../prod/wave/0075/prodwave eq complete and ( ./0072 eq complete)
              edit EMOS_TIME_STEP_H '0075'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0078
              trigger ../../prod/wave/0078/prodwave eq complete and ( ./0075 eq complete)
              edit EMOS_TIME_STEP_H '0078'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0081
              trigger ../../prod/wave/0081/prodwave eq complete and ( ./0078 eq complete)
              edit EMOS_TIME_STEP_H '0081'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0084
              trigger ../../prod/wave/0084/prodwave eq complete and ( ./0081 eq complete)
              edit EMOS_TIME_STEP_H '0084'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0087
              trigger ../../prod/wave/0087/prodwave eq complete and ( ./0084 eq complete)
              edit EMOS_TIME_STEP_H '0087'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0090
              trigger ../../prod/wave/0090/prodwave eq complete and ( ./0087 eq complete)
              edit EMOS_TIME_STEP_H '0090'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0093
              trigger ../../prod/wave/0093/prodwave eq complete and ( ./0090 eq complete)
              edit EMOS_TIME_STEP_H '0093'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0096
              trigger ../../prod/wave/0096/prodwave eq complete and ( ./0093 eq complete)
              edit EMOS_TIME_STEP_H '0096'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0099
              trigger ../../prod/wave/0099/prodwave eq complete and ( ./0096 eq complete)
              edit EMOS_TIME_STEP_H '0099'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0102
              trigger ../../prod/wave/0102/prodwave eq complete and ( ./0099 eq complete)
              edit EMOS_TIME_STEP_H '0102'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0105
              trigger ../../prod/wave/0105/prodwave eq complete and ( ./0102 eq complete)
              edit EMOS_TIME_STEP_H '0105'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0108
              trigger ../../prod/wave/0108/prodwave eq complete and ( ./0105 eq complete)
              edit EMOS_TIME_STEP_H '0108'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0111
              trigger ../../prod/wave/0111/prodwave eq complete and ( ./0108 eq complete)
              edit EMOS_TIME_STEP_H '0111'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0114
              trigger ../../prod/wave/0114/prodwave eq complete and ( ./0111 eq complete)
              edit EMOS_TIME_STEP_H '0114'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0117
              trigger ../../prod/wave/0117/prodwave eq complete and ( ./0114 eq complete)
              edit EMOS_TIME_STEP_H '0117'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0120
              trigger ../../prod/wave/0120/prodwave eq complete and ( ./0117 eq complete)
              edit EMOS_TIME_STEP_H '0120'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0123
              trigger ../../prod/wave/0123/prodwave eq complete and ( ./0120 eq complete)
              edit EMOS_TIME_STEP_H '0123'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0126
              trigger ../../prod/wave/0126/prodwave eq complete and ( ./0123 eq complete)
              edit EMOS_TIME_STEP_H '0126'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0129
              trigger ../../prod/wave/0129/prodwave eq complete and ( ./0126 eq complete)
              edit EMOS_TIME_STEP_H '0129'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0132
              trigger ../../prod/wave/0132/prodwave eq complete and ( ./0129 eq complete)
              edit EMOS_TIME_STEP_H '0132'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0135
              trigger ../../prod/wave/0135/prodwave eq complete and ( ./0132 eq complete)
              edit EMOS_TIME_STEP_H '0135'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0138
              trigger ../../prod/wave/0138/prodwave eq complete and ( ./0135 eq complete)
              edit EMOS_TIME_STEP_H '0138'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0141
              trigger ../../prod/wave/0141/prodwave eq complete and ( ./0138 eq complete)
              edit EMOS_TIME_STEP_H '0141'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0144
              trigger ../../prod/wave/0144/prodwave eq complete and ( ./0141 eq complete)
              edit EMOS_TIME_STEP_H '0144'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0150
              trigger ../../prod/wave/0150/prodwave eq complete and ( ./0144 eq complete)
              edit EMOS_TIME_STEP_H '0150'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0156
              trigger ../../prod/wave/0156/prodwave eq complete and ( ./0150 eq complete)
              edit EMOS_TIME_STEP_H '0156'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0162
              trigger ../../prod/wave/0162/prodwave eq complete and ( ./0156 eq complete)
              edit EMOS_TIME_STEP_H '0162'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0168
              trigger ../../prod/wave/0168/prodwave eq complete and ( ./0162 eq complete)
              edit EMOS_TIME_STEP_H '0168'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0174
              trigger ../../prod/wave/0174/prodwave eq complete and ( ./0168 eq complete)
              edit EMOS_TIME_STEP_H '0174'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0180
              trigger ../../prod/wave/0180/prodwave eq complete and ( ./0174 eq complete)
              edit EMOS_TIME_STEP_H '0180'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0186
              trigger ../../prod/wave/0186/prodwave eq complete and ( ./0180 eq complete)
              edit EMOS_TIME_STEP_H '0186'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0192
              trigger ../../prod/wave/0192/prodwave eq complete and ( ./0186 eq complete)
              edit EMOS_TIME_STEP_H '0192'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0198
              trigger ../../prod/wave/0198/prodwave eq complete and ( ./0192 eq complete)
              edit EMOS_TIME_STEP_H '0198'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0204
              trigger ../../prod/wave/0204/prodwave eq complete and ( ./0198 eq complete)
              edit EMOS_TIME_STEP_H '0204'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0210
              trigger ../../prod/wave/0210/prodwave eq complete and ( ./0204 eq complete)
              edit EMOS_TIME_STEP_H '0210'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0216
              trigger ../../prod/wave/0216/prodwave eq complete and ( ./0210 eq complete)
              edit EMOS_TIME_STEP_H '0216'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0222
              trigger ../../prod/wave/0222/prodwave eq complete and ( ./0216 eq complete)
              edit EMOS_TIME_STEP_H '0222'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0228
              trigger ../../prod/wave/0228/prodwave eq complete and ( ./0222 eq complete)
              edit EMOS_TIME_STEP_H '0228'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0234
              trigger ../../prod/wave/0234/prodwave eq complete and ( ./0228 eq complete)
              edit EMOS_TIME_STEP_H '0234'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0240
              trigger ../../prod/wave/0240/prodwave eq complete and ( ./0234 eq complete)
              edit EMOS_TIME_STEP_H '0240'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            task check
              trigger ./0240 eq complete
          endfamily
          task diss_wparam
            late -s +00:05 -c +00:15
            trigger ../prod/prod_wparam eq complete
            edit PROD_TASK 'w'
            edit EMOS_TIME_STEP_H '0240'
            edit EMOS_TYPE 'wp'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event pdb
          task diss_waef_postproc
            late -s +00:05 -c +00:15
            trigger ../prod/waef_postproc eq complete
            edit PROD_TASK 'p'
            edit EMOS_TIME_STEP_H '0240'
            edit EMOS_TYPE 'pw'
            edit EMOS_STREAM 'WE'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event pdb
          task diss_postproc
            late -s +00:05 -c +00:15
            trigger ../prod/enfo_postproc eq complete
            edit PROD_TASK 'p'
            edit EMOS_TIME_STEP_H '0240'
            edit EMOS_TYPE 'ep'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event pdb
          task diss_enfotc
            late -s +00:05 -c +00:15
            trigger ../prod/enfo_tc eq complete
            edit PROD_TASK 't'
            edit EMOS_TIME_STEP_H '0240'
            edit EMOS_TYPE 'tf'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event pdb
          task diss_opertc
            late -s +00:05 -c +00:15
            trigger ../prod/oper_tc eq complete
            edit PROD_TASK 't'
            edit EMOS_TIME_STEP_H '0240'
            edit EMOS_TYPE 'tf'
            edit EMOS_STREAM 'DA'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event pdb
        endfamily
      endfamily
      family legB
        complete /emc_36r1/main/ref/00/events:240
        edit EPSLEG '2'
        family fc
          complete /emc_36r1/main/ref/00/events:cpl and /emc_36r1/main/00/legBcpl/fc eq complete
          trigger /emc_36r1/main/ref/00/events eq complete and not /emc_36r1/main/ref/00/events:cpl
          edit FCMEM '2240'
          edit FCSTACK '700'
          edit FCMRFS '0'
          edit PRIORITY '71'
          edit USER_PRIORITY '71'
          edit N_PERT_AN '10'
          edit SSTLEN '144'
          edit EPSINISTEP_PREVIOUS '216'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
          edit TSTEP '2700'
          edit EPSFCRES '639'
          edit RESOL '639'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
          edit CPUTIME '1000'
          edit MEM '960'
          edit THREADS '8'
          edit NPES '8'
          edit EPSLEG '2'
          edit PERSST 'true'
          edit SSTINT '1'
          task wavfcdata
            trigger ./getpersSST eq complete and ../../legA/fc/wavini eq complete
            edit EPSCOUPLE '1'
          task getpersSST
            trigger ../../legA/fc/inidata eq complete and ../../legA/fc/getpersSST eq complete
            edit EPSCOUPLE '1'
            edit EPS_INI_STREAM 'DA'
            edit EPSFCRES '319'
            edit RESOL '319'
          family pf
            complete /emc_36r1/main/ref/00/events:cf
            trigger ./getpersSST eq complete and ./wavfcdata eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTYPE 'pf'
            edit EMOS_TYPE 'pf'
            edit USER_PRIORITY '71'
            edit PRIORITY '71'
            edit FCNPES '32'
            edit FCCHUNK '1'
            edit OCTOTAL '5'
            edit FCTOTAL '51'
            edit FCLENGTH1 '216'
            edit DAILY '0'
            edit RUNVARFC '1'
            edit EPSNLEGS '3'
            edit EPSFCLEV '62'
            edit EPSWAMNSTPW '1'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSFCRES '319'
            edit EPSWAMRESOL 'global50'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit EPSINISTEP '360'
            edit FCLENGTH '240'
            edit TSTEP '1200'
            edit EPSLEG '2'
            edit CLASS 'od'
            edit THREADS '2'
            inlimit /emc_36r1/limits:pftotal
            meter step -1 144 144
            family 01
              trigger ../../../legA/fc/pf/01 eq complete
              edit EPSMEMBER '1'
              edit MEMBER '1'
              edit SUBFSFAMILY '/pf001'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 02
              trigger ../../../legA/fc/pf/02 eq complete
              edit EPSMEMBER '2'
              edit MEMBER '2'
              edit SUBFSFAMILY '/pf002'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 03
              trigger ../../../legA/fc/pf/03 eq complete
              edit EPSMEMBER '3'
              edit MEMBER '3'
              edit SUBFSFAMILY '/pf003'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 04
              trigger ../../../legA/fc/pf/04 eq complete
              edit EPSMEMBER '4'
              edit MEMBER '4'
              edit SUBFSFAMILY '/pf004'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 05
              trigger ../../../legA/fc/pf/05 eq complete
              edit EPSMEMBER '5'
              edit MEMBER '5'
              edit SUBFSFAMILY '/pf005'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 06
              trigger ../../../legA/fc/pf/06 eq complete
              edit EPSMEMBER '6'
              edit MEMBER '6'
              edit SUBFSFAMILY '/pf006'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 07
              trigger ../../../legA/fc/pf/07 eq complete
              edit EPSMEMBER '7'
              edit MEMBER '7'
              edit SUBFSFAMILY '/pf007'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 08
              trigger ../../../legA/fc/pf/08 eq complete
              edit EPSMEMBER '8'
              edit MEMBER '8'
              edit SUBFSFAMILY '/pf008'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 09
              trigger ../../../legA/fc/pf/09 eq complete
              edit EPSMEMBER '9'
              edit MEMBER '9'
              edit SUBFSFAMILY '/pf009'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 10
              trigger ../../../legA/fc/pf/10 eq complete
              edit EPSMEMBER '10'
              edit MEMBER '10'
              edit SUBFSFAMILY '/pf010'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 11
              trigger ../../../legA/fc/pf/11 eq complete
              edit EPSMEMBER '11'
              edit MEMBER '11'
              edit SUBFSFAMILY '/pf011'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 12
              trigger ../../../legA/fc/pf/12 eq complete
              edit EPSMEMBER '12'
              edit MEMBER '12'
              edit SUBFSFAMILY '/pf012'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 13
              trigger ../../../legA/fc/pf/13 eq complete
              edit EPSMEMBER '13'
              edit MEMBER '13'
              edit SUBFSFAMILY '/pf013'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 14
              trigger ../../../legA/fc/pf/14 eq complete
              edit EPSMEMBER '14'
              edit MEMBER '14'
              edit SUBFSFAMILY '/pf014'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 15
              trigger ../../../legA/fc/pf/15 eq complete
              edit EPSMEMBER '15'
              edit MEMBER '15'
              edit SUBFSFAMILY '/pf015'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 16
              trigger ../../../legA/fc/pf/16 eq complete
              edit EPSMEMBER '16'
              edit MEMBER '16'
              edit SUBFSFAMILY '/pf016'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 17
              trigger ../../../legA/fc/pf/17 eq complete
              edit EPSMEMBER '17'
              edit MEMBER '17'
              edit SUBFSFAMILY '/pf017'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 18
              trigger ../../../legA/fc/pf/18 eq complete
              edit EPSMEMBER '18'
              edit MEMBER '18'
              edit SUBFSFAMILY '/pf018'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 19
              trigger ../../../legA/fc/pf/19 eq complete
              edit EPSMEMBER '19'
              edit MEMBER '19'
              edit SUBFSFAMILY '/pf019'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 20
              trigger ../../../legA/fc/pf/20 eq complete
              edit EPSMEMBER '20'
              edit MEMBER '20'
              edit SUBFSFAMILY '/pf020'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 21
              trigger ../../../legA/fc/pf/21 eq complete
              edit EPSMEMBER '21'
              edit MEMBER '21'
              edit SUBFSFAMILY '/pf021'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 22
              trigger ../../../legA/fc/pf/22 eq complete
              edit EPSMEMBER '22'
              edit MEMBER '22'
              edit SUBFSFAMILY '/pf022'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 23
              trigger ../../../legA/fc/pf/23 eq complete
              edit EPSMEMBER '23'
              edit MEMBER '23'
              edit SUBFSFAMILY '/pf023'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 24
              trigger ../../../legA/fc/pf/24 eq complete
              edit EPSMEMBER '24'
              edit MEMBER '24'
              edit SUBFSFAMILY '/pf024'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 25
              trigger ../../../legA/fc/pf/25 eq complete
              edit EPSMEMBER '25'
              edit MEMBER '25'
              edit SUBFSFAMILY '/pf025'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 26
              trigger ../../../legA/fc/pf/26 eq complete
              edit EPSMEMBER '26'
              edit MEMBER '26'
              edit SUBFSFAMILY '/pf026'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 27
              trigger ../../../legA/fc/pf/27 eq complete
              edit EPSMEMBER '27'
              edit MEMBER '27'
              edit SUBFSFAMILY '/pf027'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 28
              trigger ../../../legA/fc/pf/28 eq complete
              edit EPSMEMBER '28'
              edit MEMBER '28'
              edit SUBFSFAMILY '/pf028'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 29
              trigger ../../../legA/fc/pf/29 eq complete
              edit EPSMEMBER '29'
              edit MEMBER '29'
              edit SUBFSFAMILY '/pf029'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 30
              trigger ../../../legA/fc/pf/30 eq complete
              edit EPSMEMBER '30'
              edit MEMBER '30'
              edit SUBFSFAMILY '/pf030'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 31
              trigger ../../../legA/fc/pf/31 eq complete
              edit EPSMEMBER '31'
              edit MEMBER '31'
              edit SUBFSFAMILY '/pf031'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 32
              trigger ../../../legA/fc/pf/32 eq complete
              edit EPSMEMBER '32'
              edit MEMBER '32'
              edit SUBFSFAMILY '/pf032'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 33
              trigger ../../../legA/fc/pf/33 eq complete
              edit EPSMEMBER '33'
              edit MEMBER '33'
              edit SUBFSFAMILY '/pf033'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 34
              trigger ../../../legA/fc/pf/34 eq complete
              edit EPSMEMBER '34'
              edit MEMBER '34'
              edit SUBFSFAMILY '/pf034'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 35
              trigger ../../../legA/fc/pf/35 eq complete
              edit EPSMEMBER '35'
              edit MEMBER '35'
              edit SUBFSFAMILY '/pf035'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 36
              trigger ../../../legA/fc/pf/36 eq complete
              edit EPSMEMBER '36'
              edit MEMBER '36'
              edit SUBFSFAMILY '/pf036'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 37
              trigger ../../../legA/fc/pf/37 eq complete
              edit EPSMEMBER '37'
              edit MEMBER '37'
              edit SUBFSFAMILY '/pf037'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 38
              trigger ../../../legA/fc/pf/38 eq complete
              edit EPSMEMBER '38'
              edit MEMBER '38'
              edit SUBFSFAMILY '/pf038'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 39
              trigger ../../../legA/fc/pf/39 eq complete
              edit EPSMEMBER '39'
              edit MEMBER '39'
              edit SUBFSFAMILY '/pf039'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 40
              trigger ../../../legA/fc/pf/40 eq complete
              edit EPSMEMBER '40'
              edit MEMBER '40'
              edit SUBFSFAMILY '/pf040'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 41
              trigger ../../../legA/fc/pf/41 eq complete
              edit EPSMEMBER '41'
              edit MEMBER '41'
              edit SUBFSFAMILY '/pf041'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 42
              trigger ../../../legA/fc/pf/42 eq complete
              edit EPSMEMBER '42'
              edit MEMBER '42'
              edit SUBFSFAMILY '/pf042'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 43
              trigger ../../../legA/fc/pf/43 eq complete
              edit EPSMEMBER '43'
              edit MEMBER '43'
              edit SUBFSFAMILY '/pf043'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 44
              trigger ../../../legA/fc/pf/44 eq complete
              edit EPSMEMBER '44'
              edit MEMBER '44'
              edit SUBFSFAMILY '/pf044'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 45
              trigger ../../../legA/fc/pf/45 eq complete
              edit EPSMEMBER '45'
              edit MEMBER '45'
              edit SUBFSFAMILY '/pf045'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 46
              trigger ../../../legA/fc/pf/46 eq complete
              edit EPSMEMBER '46'
              edit MEMBER '46'
              edit SUBFSFAMILY '/pf046'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 47
              trigger ../../../legA/fc/pf/47 eq complete
              edit EPSMEMBER '47'
              edit MEMBER '47'
              edit SUBFSFAMILY '/pf047'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 48
              trigger ../../../legA/fc/pf/48 eq complete
              edit EPSMEMBER '48'
              edit MEMBER '48'
              edit SUBFSFAMILY '/pf048'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 49
              trigger ../../../legA/fc/pf/49 eq complete
              edit EPSMEMBER '49'
              edit MEMBER '49'
              edit SUBFSFAMILY '/pf049'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 50
              trigger ../../../legA/fc/pf/50 eq complete
              edit EPSMEMBER '50'
              edit MEMBER '50'
              edit SUBFSFAMILY '/pf050'
              edit SMSTRIES '1'
              task getvarepsdata
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger intHtoL eq complete and getvarepsdata eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '144'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
          endfamily
          family cf
            complete /emc_36r1/main/00/legBcpl/fc/cf/modeleps:step ge 144
            trigger ../../legA/fc/cf eq complete and not /emc_36r1/main/ref/00/events:cpl and ./getpersSST eq complete and ./wavfcdata eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit MEMBER '0'
            edit EPSMEMBER '0'
            edit SUBFSFAMILY '/cf000'
            edit EPSTYPE 'cf'
            edit EMOS_TYPE 'cf'
            edit USER_PRIORITY '72'
            edit EPSNLEGS '3'
            edit EPSFCLEV '62'
            edit EPSWAMNSTPW '1'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSFCRES '319'
            edit EPSWAMRESOL 'global50'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit EPSINISTEP '360'
            edit FCLENGTH '240'
            edit TSTEP '1200'
            edit EPSLEG '2'
            edit CLASS 'od'
            edit SMSTRIES '1'
            edit RUNCOUPLE '0'
            task getvarepsdata
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '2700'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSLEG '2'
              edit EPSINISTEP '216'
              edit EPSFCRES '639'
              edit EPSFCRESN '319'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '2700'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '216'
              edit EPSFCRES '319'
              edit NPES '24'
              edit FCNPES '24'
              edit THREADS '2'
              edit CLASS 'od'
            task modeleps
              trigger intHtoL eq complete and getvarepsdata eq complete
              edit FCNPES '32'
              edit THREADS '4'
              edit SMSTRIES '1'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSLEG '2'
              edit EPSFCRES '319'
              edit EPSINISTEP '144'
              edit EPSINISTEP_PREVIOUS '216'
              edit FCLENGTH '144'
              edit TSTEP '2700'
              edit CLASS 'od'
              meter step -1 144 144
          endfamily
        endfamily
        family postproc
          trigger ./fc/pf eq complete and ./fc/cf eq complete
          edit STEPS '252/360/12/24 360/360/12/120'
          inlimit /emc_36r1/limits:postproc
          family ensms
            complete /emc_36r1/main/ref/00/events:cf
            trigger ../fc eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/leg'
            family limits
              limit fdb 1
              limit gen 8
            endfamily
            task ensms
              edit ENS_M 'false'
              edit ENS_S 'false'
              edit ENS_MARS 'false'
              event stage
            family 300
              trigger ./ensms:stage or ./ensms eq complete
              edit STEP1 '246'
              edit STEP2 '300'
              edit BY '6'
              edit STEPS '0'
              family T
                edit ENS_M 'false'
                edit ENS_S 'false'
                edit LIST_PARAM 'T'
                inlimit ../limits:gen
                family 850
                  task ensms
                    edit LIST_LEV '850'
                endfamily
                family 500
                  task ensms
                    edit LIST_LEV '500'
                endfamily
              endfamily
              family Z
                edit ENS_M 'false'
                edit ENS_S 'false'
                edit LIST_PARAM 'Z'
                inlimit ../limits:gen
                family 1000
                  task ensms
                    edit LIST_LEV '1000'
                endfamily
                family 500
                  task ensms
                    edit LIST_LEV '500'
                endfamily
              endfamily
              task ensms
                trigger T eq complete and Z eq complete
                edit ENS_MARS 'false'
                inlimit ../limits:fdb
            endfamily
            family 360
              trigger ./ensms:stage or ./ensms eq complete
              edit STEP1 '306'
              edit STEP2 '360'
              edit BY '6'
              edit STEPS '0'
              family T
                edit ENS_M 'false'
                edit ENS_S 'false'
                edit LIST_PARAM 'T'
                inlimit ../limits:gen
                family 850
                  task ensms
                    edit LIST_LEV '850'
                endfamily
                family 500
                  task ensms
                    edit LIST_LEV '500'
                endfamily
              endfamily
              family Z
                edit ENS_M 'false'
                edit ENS_S 'false'
                edit LIST_PARAM 'Z'
                inlimit ../limits:gen
                family 1000
                  task ensms
                    edit LIST_LEV '1000'
                endfamily
                family 500
                  task ensms
                    edit LIST_LEV '500'
                endfamily
              endfamily
              task ensms
                trigger T eq complete and Z eq complete
                edit ENS_MARS 'false'
                inlimit ../limits:fdb
            endfamily
          endfamily
          family prob
            complete /emc_36r1/main/ref/00/events:cf
            trigger ( /emc_36r1/main/00/legB/fc eq complete) and ( /emc_36r1/main/00/legA/postproc/prob eq complete)
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/leg'
            edit STEP1 '252'
            edit STEP2 '360'
            edit BY '12'
            family 24
              edit STEP1 '240'
              edit STEP2 '360'
              edit BY '24'
              task t850
              task rain
              task ff
            endfamily
            family 12
              edit STEP1 '252'
              edit STEP2 '348'
              edit BY '24'
              task t850
              task rain
              task ff
            endfamily
            task 2t
            task swh
            task mwp
            task 10fg
              trigger ../dp eq complete
            task prob_fdb
              trigger ./2t eq complete and swh eq complete and mwp eq complete and ./10fg eq complete and ./12 eq complete and ./24 eq complete
          endfamily
          family dp
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit STEPS '246:360:6'
            inlimit /emc_36r1/limits:dp
            family dp
              trigger ./cf eq complete and ./pf eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit OVERLAP '1'
              edit STEPS '216:240:6'
              inlimit /emc_36r1/limits:dp
              family cf
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                complete /emc_36r1/main/ref/00/events:cf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
                family 05
                  edit EPSMEMBER '5'
                  edit MEMBER '5'
                  edit SUBFSFAMILY '/pf005'
                  task derived_param
                endfamily
                family 06
                  edit EPSMEMBER '6'
                  edit MEMBER '6'
                  edit SUBFSFAMILY '/pf006'
                  task derived_param
                endfamily
                family 07
                  edit EPSMEMBER '7'
                  edit MEMBER '7'
                  edit SUBFSFAMILY '/pf007'
                  task derived_param
                endfamily
                family 08
                  edit EPSMEMBER '8'
                  edit MEMBER '8'
                  edit SUBFSFAMILY '/pf008'
                  task derived_param
                endfamily
                family 09
                  edit EPSMEMBER '9'
                  edit MEMBER '9'
                  edit SUBFSFAMILY '/pf009'
                  task derived_param
                endfamily
                family 10
                  edit EPSMEMBER '10'
                  edit MEMBER '10'
                  edit SUBFSFAMILY '/pf010'
                  task derived_param
                endfamily
                family 11
                  edit EPSMEMBER '11'
                  edit MEMBER '11'
                  edit SUBFSFAMILY '/pf011'
                  task derived_param
                endfamily
                family 12
                  edit EPSMEMBER '12'
                  edit MEMBER '12'
                  edit SUBFSFAMILY '/pf012'
                  task derived_param
                endfamily
                family 13
                  edit EPSMEMBER '13'
                  edit MEMBER '13'
                  edit SUBFSFAMILY '/pf013'
                  task derived_param
                endfamily
                family 14
                  edit EPSMEMBER '14'
                  edit MEMBER '14'
                  edit SUBFSFAMILY '/pf014'
                  task derived_param
                endfamily
                family 15
                  edit EPSMEMBER '15'
                  edit MEMBER '15'
                  edit SUBFSFAMILY '/pf015'
                  task derived_param
                endfamily
                family 16
                  edit EPSMEMBER '16'
                  edit MEMBER '16'
                  edit SUBFSFAMILY '/pf016'
                  task derived_param
                endfamily
                family 17
                  edit EPSMEMBER '17'
                  edit MEMBER '17'
                  edit SUBFSFAMILY '/pf017'
                  task derived_param
                endfamily
                family 18
                  edit EPSMEMBER '18'
                  edit MEMBER '18'
                  edit SUBFSFAMILY '/pf018'
                  task derived_param
                endfamily
                family 19
                  edit EPSMEMBER '19'
                  edit MEMBER '19'
                  edit SUBFSFAMILY '/pf019'
                  task derived_param
                endfamily
                family 20
                  edit EPSMEMBER '20'
                  edit MEMBER '20'
                  edit SUBFSFAMILY '/pf020'
                  task derived_param
                endfamily
                family 21
                  edit EPSMEMBER '21'
                  edit MEMBER '21'
                  edit SUBFSFAMILY '/pf021'
                  task derived_param
                endfamily
                family 22
                  edit EPSMEMBER '22'
                  edit MEMBER '22'
                  edit SUBFSFAMILY '/pf022'
                  task derived_param
                endfamily
                family 23
                  edit EPSMEMBER '23'
                  edit MEMBER '23'
                  edit SUBFSFAMILY '/pf023'
                  task derived_param
                endfamily
                family 24
                  edit EPSMEMBER '24'
                  edit MEMBER '24'
                  edit SUBFSFAMILY '/pf024'
                  task derived_param
                endfamily
                family 25
                  edit EPSMEMBER '25'
                  edit MEMBER '25'
                  edit SUBFSFAMILY '/pf025'
                  task derived_param
                endfamily
                family 26
                  edit EPSMEMBER '26'
                  edit MEMBER '26'
                  edit SUBFSFAMILY '/pf026'
                  task derived_param
                endfamily
                family 27
                  edit EPSMEMBER '27'
                  edit MEMBER '27'
                  edit SUBFSFAMILY '/pf027'
                  task derived_param
                endfamily
                family 28
                  edit EPSMEMBER '28'
                  edit MEMBER '28'
                  edit SUBFSFAMILY '/pf028'
                  task derived_param
                endfamily
                family 29
                  edit EPSMEMBER '29'
                  edit MEMBER '29'
                  edit SUBFSFAMILY '/pf029'
                  task derived_param
                endfamily
                family 30
                  edit EPSMEMBER '30'
                  edit MEMBER '30'
                  edit SUBFSFAMILY '/pf030'
                  task derived_param
                endfamily
                family 31
                  edit EPSMEMBER '31'
                  edit MEMBER '31'
                  edit SUBFSFAMILY '/pf031'
                  task derived_param
                endfamily
                family 32
                  edit EPSMEMBER '32'
                  edit MEMBER '32'
                  edit SUBFSFAMILY '/pf032'
                  task derived_param
                endfamily
                family 33
                  edit EPSMEMBER '33'
                  edit MEMBER '33'
                  edit SUBFSFAMILY '/pf033'
                  task derived_param
                endfamily
                family 34
                  edit EPSMEMBER '34'
                  edit MEMBER '34'
                  edit SUBFSFAMILY '/pf034'
                  task derived_param
                endfamily
                family 35
                  edit EPSMEMBER '35'
                  edit MEMBER '35'
                  edit SUBFSFAMILY '/pf035'
                  task derived_param
                endfamily
                family 36
                  edit EPSMEMBER '36'
                  edit MEMBER '36'
                  edit SUBFSFAMILY '/pf036'
                  task derived_param
                endfamily
                family 37
                  edit EPSMEMBER '37'
                  edit MEMBER '37'
                  edit SUBFSFAMILY '/pf037'
                  task derived_param
                endfamily
                family 38
                  edit EPSMEMBER '38'
                  edit MEMBER '38'
                  edit SUBFSFAMILY '/pf038'
                  task derived_param
                endfamily
                family 39
                  edit EPSMEMBER '39'
                  edit MEMBER '39'
                  edit SUBFSFAMILY '/pf039'
                  task derived_param
                endfamily
                family 40
                  edit EPSMEMBER '40'
                  edit MEMBER '40'
                  edit SUBFSFAMILY '/pf040'
                  task derived_param
                endfamily
                family 41
                  edit EPSMEMBER '41'
                  edit MEMBER '41'
                  edit SUBFSFAMILY '/pf041'
                  task derived_param
                endfamily
                family 42
                  edit EPSMEMBER '42'
                  edit MEMBER '42'
                  edit SUBFSFAMILY '/pf042'
                  task derived_param
                endfamily
                family 43
                  edit EPSMEMBER '43'
                  edit MEMBER '43'
                  edit SUBFSFAMILY '/pf043'
                  task derived_param
                endfamily
                family 44
                  edit EPSMEMBER '44'
                  edit MEMBER '44'
                  edit SUBFSFAMILY '/pf044'
                  task derived_param
                endfamily
                family 45
                  edit EPSMEMBER '45'
                  edit MEMBER '45'
                  edit SUBFSFAMILY '/pf045'
                  task derived_param
                endfamily
                family 46
                  edit EPSMEMBER '46'
                  edit MEMBER '46'
                  edit SUBFSFAMILY '/pf046'
                  task derived_param
                endfamily
                family 47
                  edit EPSMEMBER '47'
                  edit MEMBER '47'
                  edit SUBFSFAMILY '/pf047'
                  task derived_param
                endfamily
                family 48
                  edit EPSMEMBER '48'
                  edit MEMBER '48'
                  edit SUBFSFAMILY '/pf048'
                  task derived_param
                endfamily
                family 49
                  edit EPSMEMBER '49'
                  edit MEMBER '49'
                  edit SUBFSFAMILY '/pf049'
                  task derived_param
                endfamily
                family 50
                  edit EPSMEMBER '50'
                  edit MEMBER '50'
                  edit SUBFSFAMILY '/pf050'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family cf
              trigger ../../fc/cf eq complete
              edit EMOS_TYPE 'cf'
              task derived_param
                edit MEMBER 'off'
            endfamily
            family pf
              complete /emc_36r1/main/ref/00/events:cf
              edit EMOS_TYPE 'pf'
              family 01
                trigger ../../../fc/pf/01 eq complete
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task derived_param
              endfamily
              family 02
                trigger ../../../fc/pf/02 eq complete
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task derived_param
              endfamily
              family 03
                trigger ../../../fc/pf/03 eq complete
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task derived_param
              endfamily
              family 04
                trigger ../../../fc/pf/04 eq complete
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task derived_param
              endfamily
              family 05
                trigger ../../../fc/pf/05 eq complete
                edit EPSMEMBER '5'
                edit MEMBER '5'
                edit SUBFSFAMILY '/pf005'
                task derived_param
              endfamily
              family 06
                trigger ../../../fc/pf/06 eq complete
                edit EPSMEMBER '6'
                edit MEMBER '6'
                edit SUBFSFAMILY '/pf006'
                task derived_param
              endfamily
              family 07
                trigger ../../../fc/pf/07 eq complete
                edit EPSMEMBER '7'
                edit MEMBER '7'
                edit SUBFSFAMILY '/pf007'
                task derived_param
              endfamily
              family 08
                trigger ../../../fc/pf/08 eq complete
                edit EPSMEMBER '8'
                edit MEMBER '8'
                edit SUBFSFAMILY '/pf008'
                task derived_param
              endfamily
              family 09
                trigger ../../../fc/pf/09 eq complete
                edit EPSMEMBER '9'
                edit MEMBER '9'
                edit SUBFSFAMILY '/pf009'
                task derived_param
              endfamily
              family 10
                trigger ../../../fc/pf/10 eq complete
                edit EPSMEMBER '10'
                edit MEMBER '10'
                edit SUBFSFAMILY '/pf010'
                task derived_param
              endfamily
              family 11
                trigger ../../../fc/pf/11 eq complete
                edit EPSMEMBER '11'
                edit MEMBER '11'
                edit SUBFSFAMILY '/pf011'
                task derived_param
              endfamily
              family 12
                trigger ../../../fc/pf/12 eq complete
                edit EPSMEMBER '12'
                edit MEMBER '12'
                edit SUBFSFAMILY '/pf012'
                task derived_param
              endfamily
              family 13
                trigger ../../../fc/pf/13 eq complete
                edit EPSMEMBER '13'
                edit MEMBER '13'
                edit SUBFSFAMILY '/pf013'
                task derived_param
              endfamily
              family 14
                trigger ../../../fc/pf/14 eq complete
                edit EPSMEMBER '14'
                edit MEMBER '14'
                edit SUBFSFAMILY '/pf014'
                task derived_param
              endfamily
              family 15
                trigger ../../../fc/pf/15 eq complete
                edit EPSMEMBER '15'
                edit MEMBER '15'
                edit SUBFSFAMILY '/pf015'
                task derived_param
              endfamily
              family 16
                trigger ../../../fc/pf/16 eq complete
                edit EPSMEMBER '16'
                edit MEMBER '16'
                edit SUBFSFAMILY '/pf016'
                task derived_param
              endfamily
              family 17
                trigger ../../../fc/pf/17 eq complete
                edit EPSMEMBER '17'
                edit MEMBER '17'
                edit SUBFSFAMILY '/pf017'
                task derived_param
              endfamily
              family 18
                trigger ../../../fc/pf/18 eq complete
                edit EPSMEMBER '18'
                edit MEMBER '18'
                edit SUBFSFAMILY '/pf018'
                task derived_param
              endfamily
              family 19
                trigger ../../../fc/pf/19 eq complete
                edit EPSMEMBER '19'
                edit MEMBER '19'
                edit SUBFSFAMILY '/pf019'
                task derived_param
              endfamily
              family 20
                trigger ../../../fc/pf/20 eq complete
                edit EPSMEMBER '20'
                edit MEMBER '20'
                edit SUBFSFAMILY '/pf020'
                task derived_param
              endfamily
              family 21
                trigger ../../../fc/pf/21 eq complete
                edit EPSMEMBER '21'
                edit MEMBER '21'
                edit SUBFSFAMILY '/pf021'
                task derived_param
              endfamily
              family 22
                trigger ../../../fc/pf/22 eq complete
                edit EPSMEMBER '22'
                edit MEMBER '22'
                edit SUBFSFAMILY '/pf022'
                task derived_param
              endfamily
              family 23
                trigger ../../../fc/pf/23 eq complete
                edit EPSMEMBER '23'
                edit MEMBER '23'
                edit SUBFSFAMILY '/pf023'
                task derived_param
              endfamily
              family 24
                trigger ../../../fc/pf/24 eq complete
                edit EPSMEMBER '24'
                edit MEMBER '24'
                edit SUBFSFAMILY '/pf024'
                task derived_param
              endfamily
              family 25
                trigger ../../../fc/pf/25 eq complete
                edit EPSMEMBER '25'
                edit MEMBER '25'
                edit SUBFSFAMILY '/pf025'
                task derived_param
              endfamily
              family 26
                trigger ../../../fc/pf/26 eq complete
                edit EPSMEMBER '26'
                edit MEMBER '26'
                edit SUBFSFAMILY '/pf026'
                task derived_param
              endfamily
              family 27
                trigger ../../../fc/pf/27 eq complete
                edit EPSMEMBER '27'
                edit MEMBER '27'
                edit SUBFSFAMILY '/pf027'
                task derived_param
              endfamily
              family 28
                trigger ../../../fc/pf/28 eq complete
                edit EPSMEMBER '28'
                edit MEMBER '28'
                edit SUBFSFAMILY '/pf028'
                task derived_param
              endfamily
              family 29
                trigger ../../../fc/pf/29 eq complete
                edit EPSMEMBER '29'
                edit MEMBER '29'
                edit SUBFSFAMILY '/pf029'
                task derived_param
              endfamily
              family 30
                trigger ../../../fc/pf/30 eq complete
                edit EPSMEMBER '30'
                edit MEMBER '30'
                edit SUBFSFAMILY '/pf030'
                task derived_param
              endfamily
              family 31
                trigger ../../../fc/pf/31 eq complete
                edit EPSMEMBER '31'
                edit MEMBER '31'
                edit SUBFSFAMILY '/pf031'
                task derived_param
              endfamily
              family 32
                trigger ../../../fc/pf/32 eq complete
                edit EPSMEMBER '32'
                edit MEMBER '32'
                edit SUBFSFAMILY '/pf032'
                task derived_param
              endfamily
              family 33
                trigger ../../../fc/pf/33 eq complete
                edit EPSMEMBER '33'
                edit MEMBER '33'
                edit SUBFSFAMILY '/pf033'
                task derived_param
              endfamily
              family 34
                trigger ../../../fc/pf/34 eq complete
                edit EPSMEMBER '34'
                edit MEMBER '34'
                edit SUBFSFAMILY '/pf034'
                task derived_param
              endfamily
              family 35
                trigger ../../../fc/pf/35 eq complete
                edit EPSMEMBER '35'
                edit MEMBER '35'
                edit SUBFSFAMILY '/pf035'
                task derived_param
              endfamily
              family 36
                trigger ../../../fc/pf/36 eq complete
                edit EPSMEMBER '36'
                edit MEMBER '36'
                edit SUBFSFAMILY '/pf036'
                task derived_param
              endfamily
              family 37
                trigger ../../../fc/pf/37 eq complete
                edit EPSMEMBER '37'
                edit MEMBER '37'
                edit SUBFSFAMILY '/pf037'
                task derived_param
              endfamily
              family 38
                trigger ../../../fc/pf/38 eq complete
                edit EPSMEMBER '38'
                edit MEMBER '38'
                edit SUBFSFAMILY '/pf038'
                task derived_param
              endfamily
              family 39
                trigger ../../../fc/pf/39 eq complete
                edit EPSMEMBER '39'
                edit MEMBER '39'
                edit SUBFSFAMILY '/pf039'
                task derived_param
              endfamily
              family 40
                trigger ../../../fc/pf/40 eq complete
                edit EPSMEMBER '40'
                edit MEMBER '40'
                edit SUBFSFAMILY '/pf040'
                task derived_param
              endfamily
              family 41
                trigger ../../../fc/pf/41 eq complete
                edit EPSMEMBER '41'
                edit MEMBER '41'
                edit SUBFSFAMILY '/pf041'
                task derived_param
              endfamily
              family 42
                trigger ../../../fc/pf/42 eq complete
                edit EPSMEMBER '42'
                edit MEMBER '42'
                edit SUBFSFAMILY '/pf042'
                task derived_param
              endfamily
              family 43
                trigger ../../../fc/pf/43 eq complete
                edit EPSMEMBER '43'
                edit MEMBER '43'
                edit SUBFSFAMILY '/pf043'
                task derived_param
              endfamily
              family 44
                trigger ../../../fc/pf/44 eq complete
                edit EPSMEMBER '44'
                edit MEMBER '44'
                edit SUBFSFAMILY '/pf044'
                task derived_param
              endfamily
              family 45
                trigger ../../../fc/pf/45 eq complete
                edit EPSMEMBER '45'
                edit MEMBER '45'
                edit SUBFSFAMILY '/pf045'
                task derived_param
              endfamily
              family 46
                trigger ../../../fc/pf/46 eq complete
                edit EPSMEMBER '46'
                edit MEMBER '46'
                edit SUBFSFAMILY '/pf046'
                task derived_param
              endfamily
              family 47
                trigger ../../../fc/pf/47 eq complete
                edit EPSMEMBER '47'
                edit MEMBER '47'
                edit SUBFSFAMILY '/pf047'
                task derived_param
              endfamily
              family 48
                trigger ../../../fc/pf/48 eq complete
                edit EPSMEMBER '48'
                edit MEMBER '48'
                edit SUBFSFAMILY '/pf048'
                task derived_param
              endfamily
              family 49
                trigger ../../../fc/pf/49 eq complete
                edit EPSMEMBER '49'
                edit MEMBER '49'
                edit SUBFSFAMILY '/pf049'
                task derived_param
              endfamily
              family 50
                trigger ../../../fc/pf/50 eq complete
                edit EPSMEMBER '50'
                edit MEMBER '50'
                edit SUBFSFAMILY '/pf050'
                task derived_param
              endfamily
            endfamily
          endfamily
          family extra
            trigger ./dp eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit STEPS '246:360:6'
            inlimit /emc_36r1/limits:extra
            family cf
              edit EMOS_TYPE 'cf'
              task extrafields
                edit EPSFCRES '319'
                edit FCLENGTH '360'
                edit FCLENGTH0 '240'
                edit RUNVARFC '1'
            endfamily
            family pf
              complete /emc_36r1/main/ref/00/events:cf
              edit EMOS_TYPE 'pf'
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 05
                edit EPSMEMBER '5'
                edit MEMBER '5'
                edit SUBFSFAMILY '/pf005'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 06
                edit EPSMEMBER '6'
                edit MEMBER '6'
                edit SUBFSFAMILY '/pf006'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 07
                edit EPSMEMBER '7'
                edit MEMBER '7'
                edit SUBFSFAMILY '/pf007'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 08
                edit EPSMEMBER '8'
                edit MEMBER '8'
                edit SUBFSFAMILY '/pf008'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 09
                edit EPSMEMBER '9'
                edit MEMBER '9'
                edit SUBFSFAMILY '/pf009'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 10
                edit EPSMEMBER '10'
                edit MEMBER '10'
                edit SUBFSFAMILY '/pf010'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 11
                edit EPSMEMBER '11'
                edit MEMBER '11'
                edit SUBFSFAMILY '/pf011'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 12
                edit EPSMEMBER '12'
                edit MEMBER '12'
                edit SUBFSFAMILY '/pf012'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 13
                edit EPSMEMBER '13'
                edit MEMBER '13'
                edit SUBFSFAMILY '/pf013'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 14
                edit EPSMEMBER '14'
                edit MEMBER '14'
                edit SUBFSFAMILY '/pf014'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 15
                edit EPSMEMBER '15'
                edit MEMBER '15'
                edit SUBFSFAMILY '/pf015'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 16
                edit EPSMEMBER '16'
                edit MEMBER '16'
                edit SUBFSFAMILY '/pf016'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 17
                edit EPSMEMBER '17'
                edit MEMBER '17'
                edit SUBFSFAMILY '/pf017'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 18
                edit EPSMEMBER '18'
                edit MEMBER '18'
                edit SUBFSFAMILY '/pf018'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 19
                edit EPSMEMBER '19'
                edit MEMBER '19'
                edit SUBFSFAMILY '/pf019'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 20
                edit EPSMEMBER '20'
                edit MEMBER '20'
                edit SUBFSFAMILY '/pf020'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 21
                edit EPSMEMBER '21'
                edit MEMBER '21'
                edit SUBFSFAMILY '/pf021'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 22
                edit EPSMEMBER '22'
                edit MEMBER '22'
                edit SUBFSFAMILY '/pf022'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 23
                edit EPSMEMBER '23'
                edit MEMBER '23'
                edit SUBFSFAMILY '/pf023'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 24
                edit EPSMEMBER '24'
                edit MEMBER '24'
                edit SUBFSFAMILY '/pf024'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 25
                edit EPSMEMBER '25'
                edit MEMBER '25'
                edit SUBFSFAMILY '/pf025'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 26
                edit EPSMEMBER '26'
                edit MEMBER '26'
                edit SUBFSFAMILY '/pf026'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 27
                edit EPSMEMBER '27'
                edit MEMBER '27'
                edit SUBFSFAMILY '/pf027'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 28
                edit EPSMEMBER '28'
                edit MEMBER '28'
                edit SUBFSFAMILY '/pf028'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 29
                edit EPSMEMBER '29'
                edit MEMBER '29'
                edit SUBFSFAMILY '/pf029'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 30
                edit EPSMEMBER '30'
                edit MEMBER '30'
                edit SUBFSFAMILY '/pf030'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 31
                edit EPSMEMBER '31'
                edit MEMBER '31'
                edit SUBFSFAMILY '/pf031'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 32
                edit EPSMEMBER '32'
                edit MEMBER '32'
                edit SUBFSFAMILY '/pf032'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 33
                edit EPSMEMBER '33'
                edit MEMBER '33'
                edit SUBFSFAMILY '/pf033'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 34
                edit EPSMEMBER '34'
                edit MEMBER '34'
                edit SUBFSFAMILY '/pf034'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 35
                edit EPSMEMBER '35'
                edit MEMBER '35'
                edit SUBFSFAMILY '/pf035'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 36
                edit EPSMEMBER '36'
                edit MEMBER '36'
                edit SUBFSFAMILY '/pf036'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 37
                edit EPSMEMBER '37'
                edit MEMBER '37'
                edit SUBFSFAMILY '/pf037'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 38
                edit EPSMEMBER '38'
                edit MEMBER '38'
                edit SUBFSFAMILY '/pf038'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 39
                edit EPSMEMBER '39'
                edit MEMBER '39'
                edit SUBFSFAMILY '/pf039'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 40
                edit EPSMEMBER '40'
                edit MEMBER '40'
                edit SUBFSFAMILY '/pf040'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 41
                edit EPSMEMBER '41'
                edit MEMBER '41'
                edit SUBFSFAMILY '/pf041'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 42
                edit EPSMEMBER '42'
                edit MEMBER '42'
                edit SUBFSFAMILY '/pf042'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 43
                edit EPSMEMBER '43'
                edit MEMBER '43'
                edit SUBFSFAMILY '/pf043'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 44
                edit EPSMEMBER '44'
                edit MEMBER '44'
                edit SUBFSFAMILY '/pf044'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 45
                edit EPSMEMBER '45'
                edit MEMBER '45'
                edit SUBFSFAMILY '/pf045'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 46
                edit EPSMEMBER '46'
                edit MEMBER '46'
                edit SUBFSFAMILY '/pf046'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 47
                edit EPSMEMBER '47'
                edit MEMBER '47'
                edit SUBFSFAMILY '/pf047'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 48
                edit EPSMEMBER '48'
                edit MEMBER '48'
                edit SUBFSFAMILY '/pf048'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 49
                edit EPSMEMBER '49'
                edit MEMBER '49'
                edit SUBFSFAMILY '/pf049'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
              family 50
                edit EPSMEMBER '50'
                edit MEMBER '50'
                edit SUBFSFAMILY '/pf050'
                task extrafields
                  edit EPSFCRES '319'
                  edit FCLENGTH '360'
                  edit FCLENGTH0 '240'
                  edit RUNVARFC '1'
              endfamily
            endfamily
          endfamily
          family calibrate
            complete /emc_36r1/main/ref/00/events:cf
            trigger /emc_36r1/main/00/ref/an eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/calibrate'
            family onws
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              inlimit /limits:linux_cluster
              task feedupcal
              task upcal
                trigger feedupcal eq complete
              task revertupcal
            endfamily
          endfamily
          family efi
            complete ( /emc_36r1/main/ref/00/events:cf) or ( /emc_36r1/main/ref/00/events:noefi)
            trigger /emc_36r1/main/00/legB/postproc/dp eq complete
            edit HC_VERSION '0046'
            edit HC_EXPVER '0046'
            edit REF_VERSION '0046'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/efi_hdc'
            edit NC_DAYS '31'
            edit EPSLEG '2'
            inlimit /emc_36r1/limits:efi
            task stage
              edit NC_DAYS '31'
            family 240_336
              trigger ./stage eq complete
              edit ACC '96'
              edit EFI_STEP1 '240'
              edit EFI_STEP2 '336'
              family 2t
                edit RF_NAME '2t'
                task extindex
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task extindex
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task extindex
              endfamily
              family tp
                edit RF_NAME 'tp'
                task extindex
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task extindex
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task extindex
              endfamily
              family ts
                edit RF_NAME 'ts'
                task extindex
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task extindex
              endfamily
            endfamily
            family 168_336
              trigger ./stage eq complete
              edit ACC '168'
              edit EFI_STEP1 '168'
              edit EFI_STEP2 '336'
              family 2t
                edit RF_NAME '2t'
                task extindex
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task extindex
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task extindex
              endfamily
              family tp
                edit RF_NAME 'tp'
                task extindex
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task extindex
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task extindex
              endfamily
              family ts
                edit RF_NAME 'ts'
                task extindex
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task extindex
              endfamily
            endfamily
            family 012_036
              trigger ./stage eq complete
              edit ACC '24'
              edit EFI_STEP1 '012'
              edit EFI_STEP2 '036'
              family 2t
                edit RF_NAME '2t'
                task extindex
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task extindex
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task extindex
              endfamily
              family tp
                edit RF_NAME 'tp'
                task extindex
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task extindex
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task extindex
              endfamily
              family ts
                edit RF_NAME 'ts'
                task extindex
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task extindex
              endfamily
            endfamily
            family 036_060
              trigger ./stage eq complete
              edit ACC '24'
              edit EFI_STEP1 '036'
              edit EFI_STEP2 '060'
              family 2t
                edit RF_NAME '2t'
                task extindex
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task extindex
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task extindex
              endfamily
              family tp
                edit RF_NAME 'tp'
                task extindex
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task extindex
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task extindex
              endfamily
              family ts
                edit RF_NAME 'ts'
                task extindex
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task extindex
              endfamily
            endfamily
            family 060_084
              trigger ./stage eq complete
              edit ACC '24'
              edit EFI_STEP1 '060'
              edit EFI_STEP2 '084'
              family 2t
                edit RF_NAME '2t'
                task extindex
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task extindex
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task extindex
              endfamily
              family tp
                edit RF_NAME 'tp'
                task extindex
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task extindex
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task extindex
              endfamily
              family ts
                edit RF_NAME 'ts'
                task extindex
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task extindex
              endfamily
            endfamily
            family 084_108
              trigger ./stage eq complete
              edit ACC '24'
              edit EFI_STEP1 '084'
              edit EFI_STEP2 '108'
              family 2t
                edit RF_NAME '2t'
                task extindex
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task extindex
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task extindex
              endfamily
              family tp
                edit RF_NAME 'tp'
                task extindex
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task extindex
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task extindex
              endfamily
              family ts
                edit RF_NAME 'ts'
                task extindex
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task extindex
              endfamily
            endfamily
            family 108_132
              trigger ./stage eq complete
              edit ACC '24'
              edit EFI_STEP1 '108'
              edit EFI_STEP2 '132'
              family 2t
                edit RF_NAME '2t'
                task extindex
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task extindex
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task extindex
              endfamily
              family tp
                edit RF_NAME 'tp'
                task extindex
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task extindex
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task extindex
              endfamily
              family ts
                edit RF_NAME 'ts'
                task extindex
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task extindex
              endfamily
            endfamily
            family 120_144
              trigger ./stage eq complete
              edit ACC '24'
              edit EFI_STEP1 '120'
              edit EFI_STEP2 '144'
              family 2t
                edit RF_NAME '2t'
                task extindex
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task extindex
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task extindex
              endfamily
              family tp
                edit RF_NAME 'tp'
                task extindex
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task extindex
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task extindex
              endfamily
              family ts
                edit RF_NAME 'ts'
                task extindex
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task extindex
              endfamily
            endfamily
            family 120_168
              trigger ./stage eq complete
              edit ACC '48'
              edit EFI_STEP1 '120'
              edit EFI_STEP2 '168'
              family 2t
                edit RF_NAME '2t'
                task extindex
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task extindex
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task extindex
              endfamily
              family tp
                edit RF_NAME 'tp'
                task extindex
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task extindex
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task extindex
              endfamily
              family ts
                edit RF_NAME 'ts'
                task extindex
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task extindex
              endfamily
            endfamily
            family 168_240
              trigger ./stage eq complete
              edit ACC '72'
              edit EFI_STEP1 '168'
              edit EFI_STEP2 '240'
              family 2t
                edit RF_NAME '2t'
                task extindex
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task extindex
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task extindex
              endfamily
              family tp
                edit RF_NAME 'tp'
                task extindex
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task extindex
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task extindex
              endfamily
              family ts
                edit RF_NAME 'ts'
                task extindex
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task extindex
              endfamily
            endfamily
            family 000_168
              trigger ./stage eq complete
              edit ACC '168'
              edit EFI_STEP1 '000'
              edit EFI_STEP2 '168'
              family 2t
                edit RF_NAME '2t'
                task extindex
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task extindex
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task extindex
              endfamily
              family tp
                edit RF_NAME 'tp'
                task extindex
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task extindex
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task extindex
              endfamily
              family ts
                edit RF_NAME 'ts'
                task extindex
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task extindex
              endfamily
            endfamily
            family 000_024
              trigger ./stage eq complete
              edit ACC '24'
              edit EFI_STEP1 '000'
              edit EFI_STEP2 '024'
              family 2tmin
                edit RF_NAME '2tmin'
                task extindex
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task extindex
              endfamily
              family ts
                edit RF_NAME 'ts'
                task extindex
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task extindex
              endfamily
            endfamily
            family 024_048
              trigger ./stage eq complete
              edit ACC '24'
              edit EFI_STEP1 '024'
              edit EFI_STEP2 '048'
              family 2tmin
                edit RF_NAME '2tmin'
                task extindex
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task extindex
              endfamily
              family ts
                edit RF_NAME 'ts'
                task extindex
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task extindex
              endfamily
            endfamily
            family 048_072
              trigger ./stage eq complete
              edit ACC '24'
              edit EFI_STEP1 '048'
              edit EFI_STEP2 '072'
              family 2tmin
                edit RF_NAME '2tmin'
                task extindex
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task extindex
              endfamily
              family ts
                edit RF_NAME 'ts'
                task extindex
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task extindex
              endfamily
            endfamily
            family 072_096
              trigger ./stage eq complete
              edit ACC '24'
              edit EFI_STEP1 '072'
              edit EFI_STEP2 '096'
              family 2tmin
                edit RF_NAME '2tmin'
                task extindex
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task extindex
              endfamily
              family ts
                edit RF_NAME 'ts'
                task extindex
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task extindex
              endfamily
            endfamily
            family 096_120
              trigger ./stage eq complete
              edit ACC '24'
              edit EFI_STEP1 '096'
              edit EFI_STEP2 '120'
              family 2tmin
                edit RF_NAME '2tmin'
                task extindex
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task extindex
              endfamily
              family ts
                edit RF_NAME 'ts'
                task extindex
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task extindex
              endfamily
            endfamily
          endfamily
          task efi2fdb
            complete ( /emc_36r1/main/ref/00/events:cf) or ( /emc_36r1/main/ref/00/events:noefi)
            trigger ./efi eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/efi_hdc'
            inlimit /emc_36r1/main/ref:mutex
            event none
        endfamily
        family prod
          edit MEM '3000'
          edit ACCOUNT 'oeenfopg'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/prod'
          edit EMOS_DOMAIN 'G'
          edit EMOS_TYPE 'cp'
          edit EMOS_PROJECT '12'
          edit PROD_TASK '1'
          edit USER_PRIORITY '90'
          edit SMSURLBASE 'http://intra.ecmwf.int/plots/d/intra/'
          edit SMSURL 'sub=fixed/'
          inlimit /limits:prod
          family eps
            trigger ( ../../legA/prod/getreq/eps eq complete) and ( ../../legA/prod/getreq/bitmap:001 or ../../legA/prod/getreq/bitmap eq complete) and ../postproc/dp eq complete
            inlimit /emc_36r1/limits:prodeps
            family 0246
              edit EMOS_TIME_STEP_H '0246'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0252
              edit EMOS_TIME_STEP_H '0252'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0258
              edit EMOS_TIME_STEP_H '0258'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0264
              edit EMOS_TIME_STEP_H '0264'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0270
              edit EMOS_TIME_STEP_H '0270'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0276
              edit EMOS_TIME_STEP_H '0276'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0282
              edit EMOS_TIME_STEP_H '0282'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0288
              edit EMOS_TIME_STEP_H '0288'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0294
              edit EMOS_TIME_STEP_H '0294'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0300
              edit EMOS_TIME_STEP_H '0300'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0306
              edit EMOS_TIME_STEP_H '0306'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0312
              edit EMOS_TIME_STEP_H '0312'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0318
              edit EMOS_TIME_STEP_H '0318'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0324
              edit EMOS_TIME_STEP_H '0324'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0330
              edit EMOS_TIME_STEP_H '0330'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0336
              edit EMOS_TIME_STEP_H '0336'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0342
              edit EMOS_TIME_STEP_H '0342'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0348
              edit EMOS_TIME_STEP_H '0348'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0354
              edit EMOS_TIME_STEP_H '0354'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
            family 0360
              edit EMOS_TIME_STEP_H '0360'
              task prodgen
                late -s +00:05 -c +00:15
            endfamily
          endfamily
          family wave
            trigger ( ../../legA/prod/getreq/wave_eps:001 or ../../legA/prod/getreq/wave_eps eq complete) and ( ../../legA/prod/getreq/bitmap:001 or ../../legA/prod/getreq/bitmap eq complete) and ../postproc/dp eq complete
            edit PGNPES '2'
            edit URL 'GWAEF/00'
            inlimit /emc_36r1/limits:prodwave
            family 0246
              edit EMOS_TIME_STEP_H '0246'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0252
              edit EMOS_TIME_STEP_H '0252'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0258
              edit EMOS_TIME_STEP_H '0258'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0264
              edit EMOS_TIME_STEP_H '0264'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0270
              edit EMOS_TIME_STEP_H '0270'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0276
              edit EMOS_TIME_STEP_H '0276'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0282
              edit EMOS_TIME_STEP_H '0282'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0288
              edit EMOS_TIME_STEP_H '0288'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0294
              edit EMOS_TIME_STEP_H '0294'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0300
              edit EMOS_TIME_STEP_H '0300'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0306
              edit EMOS_TIME_STEP_H '0306'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0312
              edit EMOS_TIME_STEP_H '0312'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0318
              edit EMOS_TIME_STEP_H '0318'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0324
              edit EMOS_TIME_STEP_H '0324'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0330
              edit EMOS_TIME_STEP_H '0330'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0336
              edit EMOS_TIME_STEP_H '0336'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0342
              edit EMOS_TIME_STEP_H '0342'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0348
              edit EMOS_TIME_STEP_H '0348'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0354
              edit EMOS_TIME_STEP_H '0354'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
            family 0360
              edit EMOS_TIME_STEP_H '0360'
              task prodwave
                late -s +00:05 -c +00:15
                edit EMOS_STREAM 'WE'
                edit EMOS_TYPE 'we'
            endfamily
          endfamily
          task prod_wparam
            late -s +00:05 -c +00:15
            trigger ( ../../legA/prod/getreq/eps eq complete) and ../postproc/dp eq complete and ../../legA/prod/prod_wparam eq complete
            edit PROD_TASK 'w'
            edit EMOS_TYPE 'wp'
            edit EMOS_TIME_STEP_H '0360'
            edit PGNPES '52'
          task enfo_postproc
            late -s +00:05 -c +00:15
            trigger ../postproc/prob eq complete and ../postproc/ensms eq complete and ( ../../legA/prod/getreq/eps eq complete)
            edit PROD_TASK 'p'
            edit EMOS_TYPE 'ep'
            edit EMOS_TIME_STEP_H '0360'
            edit PGNPES '21'
          task waef_postproc
            late -s +00:05 -c +00:15
            trigger ( ../../legA/prod/getreq/wave_eps:001 or ../../legA/prod/getreq/wave_eps eq complete) and ../postproc/prob eq complete
            edit PROD_TASK 'p'
            edit EMOS_TYPE 'pw'
            edit EMOS_STREAM 'WE'
            edit EMOS_TIME_STEP_H '0360'
            edit PGNPES '1'
        endfamily
        family prod2diss
          edit USER_PRIORITY '90'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/prod'
          edit EMOS_DOMAIN 'G'
          edit EMOS_TYPE 'ef'
          edit EMOS_PROJECT '12'
          edit PROD_TASK '1'
          edit SMSURLBASE 'http://diss.ecmwf.int/do/monitoring/summary'
          edit SMSURL 'GENFO/00'
          family eps
            family 0246
              edit EMOS_TIME_STEP_H '0246'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0246/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../../../../legA/prod2diss/eps//0240/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0252
              edit EMOS_TIME_STEP_H '0252'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0252/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0246/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0258
              edit EMOS_TIME_STEP_H '0258'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0258/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0252/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0264
              edit EMOS_TIME_STEP_H '0264'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0264/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0258/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0270
              edit EMOS_TIME_STEP_H '0270'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0270/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0264/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0276
              edit EMOS_TIME_STEP_H '0276'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0276/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0270/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0282
              edit EMOS_TIME_STEP_H '0282'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0282/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0276/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0288
              edit EMOS_TIME_STEP_H '0288'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0288/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0282/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0294
              edit EMOS_TIME_STEP_H '0294'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0294/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0288/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0300
              edit EMOS_TIME_STEP_H '0300'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0300/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0294/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0306
              edit EMOS_TIME_STEP_H '0306'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0306/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0300/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0312
              edit EMOS_TIME_STEP_H '0312'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0312/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0306/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0318
              edit EMOS_TIME_STEP_H '0318'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0318/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0312/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0324
              edit EMOS_TIME_STEP_H '0324'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0324/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0318/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0330
              edit EMOS_TIME_STEP_H '0330'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0330/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0324/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0336
              edit EMOS_TIME_STEP_H '0336'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0336/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0330/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0342
              edit EMOS_TIME_STEP_H '0342'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0342/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0336/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0348
              edit EMOS_TIME_STEP_H '0348'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0348/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0342/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0354
              edit EMOS_TIME_STEP_H '0354'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0354/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0348/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            family 0360
              edit EMOS_TIME_STEP_H '0360'
              task pdb
                late -s +00:05 -c +00:15
                trigger ../../../prod/eps/0360/prodgen eq complete
                inlimit /emc_36r1/limits:pdb
                inlimit /limits:pdb
                label info ""
                event 1 
              task q2diss
                late -s +00:05 -c +00:15
                trigger ./pdb eq complete and ( not ../../../../legA/prod2diss/operation_is_late:yes or ../0354/q2diss eq complete)
                inlimit /emc_36r1/limits:q2diss
                inlimit /limits:q2diss
                label info ""
                event 1 
            endfamily
            task check
              trigger ./0360 eq complete
          endfamily
          family wave
            edit EMOS_STREAM 'WE'
            edit EMOS_TYPE 'we'
            edit SMSURL 'GWAEF/00'
            family 0246
              trigger ../../prod/wave/0246/prodwave eq complete and ( ../../../legA/prod2diss/wave/0240 eq complete)
              edit EMOS_TIME_STEP_H '0246'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0252
              trigger ../../prod/wave/0252/prodwave eq complete and ( ./0246 eq complete)
              edit EMOS_TIME_STEP_H '0252'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0258
              trigger ../../prod/wave/0258/prodwave eq complete and ( ./0252 eq complete)
              edit EMOS_TIME_STEP_H '0258'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0264
              trigger ../../prod/wave/0264/prodwave eq complete and ( ./0258 eq complete)
              edit EMOS_TIME_STEP_H '0264'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0270
              trigger ../../prod/wave/0270/prodwave eq complete and ( ./0264 eq complete)
              edit EMOS_TIME_STEP_H '0270'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0276
              trigger ../../prod/wave/0276/prodwave eq complete and ( ./0270 eq complete)
              edit EMOS_TIME_STEP_H '0276'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0282
              trigger ../../prod/wave/0282/prodwave eq complete and ( ./0276 eq complete)
              edit EMOS_TIME_STEP_H '0282'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0288
              trigger ../../prod/wave/0288/prodwave eq complete and ( ./0282 eq complete)
              edit EMOS_TIME_STEP_H '0288'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0294
              trigger ../../prod/wave/0294/prodwave eq complete and ( ./0288 eq complete)
              edit EMOS_TIME_STEP_H '0294'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0300
              trigger ../../prod/wave/0300/prodwave eq complete and ( ./0294 eq complete)
              edit EMOS_TIME_STEP_H '0300'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0306
              trigger ../../prod/wave/0306/prodwave eq complete and ( ./0300 eq complete)
              edit EMOS_TIME_STEP_H '0306'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0312
              trigger ../../prod/wave/0312/prodwave eq complete and ( ./0306 eq complete)
              edit EMOS_TIME_STEP_H '0312'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0318
              trigger ../../prod/wave/0318/prodwave eq complete and ( ./0312 eq complete)
              edit EMOS_TIME_STEP_H '0318'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0324
              trigger ../../prod/wave/0324/prodwave eq complete and ( ./0318 eq complete)
              edit EMOS_TIME_STEP_H '0324'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0330
              trigger ../../prod/wave/0330/prodwave eq complete and ( ./0324 eq complete)
              edit EMOS_TIME_STEP_H '0330'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0336
              trigger ../../prod/wave/0336/prodwave eq complete and ( ./0330 eq complete)
              edit EMOS_TIME_STEP_H '0336'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0342
              trigger ../../prod/wave/0342/prodwave eq complete and ( ./0336 eq complete)
              edit EMOS_TIME_STEP_H '0342'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0348
              trigger ../../prod/wave/0348/prodwave eq complete and ( ./0342 eq complete)
              edit EMOS_TIME_STEP_H '0348'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0354
              trigger ../../prod/wave/0354/prodwave eq complete and ( ./0348 eq complete)
              edit EMOS_TIME_STEP_H '0354'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            family 0360
              trigger ../../prod/wave/0360/prodwave eq complete and ( ./0354 eq complete)
              edit EMOS_TIME_STEP_H '0360'
              task diss
                late -s +00:05 -c +00:15
                inlimit /emc_36r1/limits:w2diss
                inlimit /limits:q2diss
                label info ""
                event pdb
            endfamily
            task check
              trigger ./0360 eq complete
          endfamily
          task diss_wparam
            late -s +00:05 -c +00:15
            trigger ../prod/prod_wparam eq complete and ../../legA/prod2diss/diss_wparam eq complete
            edit PROD_TASK 'w'
            edit EMOS_TIME_STEP_H '0360'
            edit EMOS_TYPE 'wp'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event pdb
          task diss_waef_postproc
            late -s +00:05 -c +00:15
            trigger ../prod/waef_postproc eq complete and ../../legA/prod2diss/diss_waef_postproc eq complete
            edit PROD_TASK 'p'
            edit EMOS_TIME_STEP_H '0360'
            edit EMOS_TYPE 'pw'
            edit EMOS_STREAM 'WE'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event pdb
          task diss_postproc
            late -s +00:05 -c +00:15
            trigger ../prod/enfo_postproc eq complete and ../../legA/prod2diss/diss_postproc eq complete
            edit PROD_TASK 'p'
            edit EMOS_TIME_STEP_H '0360'
            edit EMOS_TYPE 'ep'
            inlimit /emc_36r1/limits:q2diss
            inlimit /limits:q2diss
            label info ""
            event pdb
        endfamily
      endfamily
      family legBcpl
        complete /emc_36r1/main/ref/00/events:240 or /emc_36r1/main/ref/00/events:360
        trigger /emc_36r1/main/ref/00/events eq complete and /emc_36r1/main/ref/00/events:cpl and /emc_36r1/main/00/ref/oceini eq complete
        edit EPSLEG '2'
        edit EPSCOUPLE '1'
        edit RUNCOUPLE '1'
        edit OC_DELAY_START '0'
        family fc
          edit FCMEM '2240'
          edit FCSTACK '700'
          edit FCMRFS '0'
          edit PRIORITY '71'
          edit USER_PRIORITY '71'
          edit N_PERT_AN '10'
          edit SSTLEN '144'
          edit EPSINISTEP_PREVIOUS '216'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
          edit TSTEP '2700'
          edit EPSFCRES '639'
          edit RESOL '639'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
          edit CPUTIME '1000'
          edit MEM '960'
          edit THREADS '8'
          edit NPES '8'
          edit EPSLEG '2'
          family pf
            complete /emc_36r1/main/ref/00/events:cf
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTYPE 'pf'
            edit EMOS_TYPE 'pf'
            edit USER_PRIORITY '71'
            edit PRIORITY '71'
            edit FCNPES '32'
            edit EPSNLEGS '3'
            edit EPSFCLEV '62'
            edit EPSWAMNSTPW '1'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSFCRES '319'
            edit EPSWAMRESOL 'global50'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit EPSINISTEP '360'
            edit FCLENGTH '240'
            edit TSTEP '1200'
            edit EPSLEG '2'
            edit CLASS 'od'
            edit THREADS '2'
            edit RUNVARFC '1'
            inlimit /emc_36r1/limits:pftotal
            meter step -1 144 144
            family 01
              edit EPSMEMBER '1'
              edit MEMBER '1'
              edit SUBFSFAMILY '/pf001'
              edit OCGROUP 'an1'
              edit FCGROUP 'fc0'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/01/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/01 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 02
              edit EPSMEMBER '2'
              edit MEMBER '2'
              edit SUBFSFAMILY '/pf002'
              edit OCGROUP 'an2'
              edit FCGROUP 'fc0'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/02/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/02 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 03
              edit EPSMEMBER '3'
              edit MEMBER '3'
              edit SUBFSFAMILY '/pf003'
              edit OCGROUP 'an3'
              edit FCGROUP 'fc0'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/03/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/03 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 04
              edit EPSMEMBER '4'
              edit MEMBER '4'
              edit SUBFSFAMILY '/pf004'
              edit OCGROUP 'an4'
              edit FCGROUP 'fc0'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/04/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/04 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 05
              edit EPSMEMBER '5'
              edit MEMBER '5'
              edit SUBFSFAMILY '/pf005'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc1'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/05/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/05 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 06
              edit EPSMEMBER '6'
              edit MEMBER '6'
              edit SUBFSFAMILY '/pf006'
              edit OCGROUP 'an1'
              edit FCGROUP 'fc1'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/06/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/06 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 07
              edit EPSMEMBER '7'
              edit MEMBER '7'
              edit SUBFSFAMILY '/pf007'
              edit OCGROUP 'an2'
              edit FCGROUP 'fc1'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/07/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/07 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 08
              edit EPSMEMBER '8'
              edit MEMBER '8'
              edit SUBFSFAMILY '/pf008'
              edit OCGROUP 'an3'
              edit FCGROUP 'fc1'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/08/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/08 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 09
              edit EPSMEMBER '9'
              edit MEMBER '9'
              edit SUBFSFAMILY '/pf009'
              edit OCGROUP 'an4'
              edit FCGROUP 'fc1'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/09/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/09 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 10
              edit EPSMEMBER '10'
              edit MEMBER '10'
              edit SUBFSFAMILY '/pf010'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc2'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/10/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/10 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 11
              edit EPSMEMBER '11'
              edit MEMBER '11'
              edit SUBFSFAMILY '/pf011'
              edit OCGROUP 'an1'
              edit FCGROUP 'fc2'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/11/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/11 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 12
              edit EPSMEMBER '12'
              edit MEMBER '12'
              edit SUBFSFAMILY '/pf012'
              edit OCGROUP 'an2'
              edit FCGROUP 'fc2'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/12/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/12 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 13
              edit EPSMEMBER '13'
              edit MEMBER '13'
              edit SUBFSFAMILY '/pf013'
              edit OCGROUP 'an3'
              edit FCGROUP 'fc2'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/13/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/13 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 14
              edit EPSMEMBER '14'
              edit MEMBER '14'
              edit SUBFSFAMILY '/pf014'
              edit OCGROUP 'an4'
              edit FCGROUP 'fc2'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/14/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/14 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 15
              edit EPSMEMBER '15'
              edit MEMBER '15'
              edit SUBFSFAMILY '/pf015'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc3'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/15/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/15 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 16
              edit EPSMEMBER '16'
              edit MEMBER '16'
              edit SUBFSFAMILY '/pf016'
              edit OCGROUP 'an1'
              edit FCGROUP 'fc3'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/16/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/16 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 17
              edit EPSMEMBER '17'
              edit MEMBER '17'
              edit SUBFSFAMILY '/pf017'
              edit OCGROUP 'an2'
              edit FCGROUP 'fc3'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/17/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/17 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 18
              edit EPSMEMBER '18'
              edit MEMBER '18'
              edit SUBFSFAMILY '/pf018'
              edit OCGROUP 'an3'
              edit FCGROUP 'fc3'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/18/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/18 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 19
              edit EPSMEMBER '19'
              edit MEMBER '19'
              edit SUBFSFAMILY '/pf019'
              edit OCGROUP 'an4'
              edit FCGROUP 'fc3'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/19/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/19 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 20
              edit EPSMEMBER '20'
              edit MEMBER '20'
              edit SUBFSFAMILY '/pf020'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc4'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/20/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/20 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 21
              edit EPSMEMBER '21'
              edit MEMBER '21'
              edit SUBFSFAMILY '/pf021'
              edit OCGROUP 'an1'
              edit FCGROUP 'fc4'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/21/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/21 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 22
              edit EPSMEMBER '22'
              edit MEMBER '22'
              edit SUBFSFAMILY '/pf022'
              edit OCGROUP 'an2'
              edit FCGROUP 'fc4'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/22/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/22 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 23
              edit EPSMEMBER '23'
              edit MEMBER '23'
              edit SUBFSFAMILY '/pf023'
              edit OCGROUP 'an3'
              edit FCGROUP 'fc4'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/23/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/23 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 24
              edit EPSMEMBER '24'
              edit MEMBER '24'
              edit SUBFSFAMILY '/pf024'
              edit OCGROUP 'an4'
              edit FCGROUP 'fc4'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/24/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/24 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 25
              edit EPSMEMBER '25'
              edit MEMBER '25'
              edit SUBFSFAMILY '/pf025'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc5'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/25/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/25 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 26
              edit EPSMEMBER '26'
              edit MEMBER '26'
              edit SUBFSFAMILY '/pf026'
              edit OCGROUP 'an1'
              edit FCGROUP 'fc5'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/26/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/26 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 27
              edit EPSMEMBER '27'
              edit MEMBER '27'
              edit SUBFSFAMILY '/pf027'
              edit OCGROUP 'an2'
              edit FCGROUP 'fc5'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/27/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/27 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 28
              edit EPSMEMBER '28'
              edit MEMBER '28'
              edit SUBFSFAMILY '/pf028'
              edit OCGROUP 'an3'
              edit FCGROUP 'fc5'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/28/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/28 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 29
              edit EPSMEMBER '29'
              edit MEMBER '29'
              edit SUBFSFAMILY '/pf029'
              edit OCGROUP 'an4'
              edit FCGROUP 'fc5'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/29/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/29 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 30
              edit EPSMEMBER '30'
              edit MEMBER '30'
              edit SUBFSFAMILY '/pf030'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc6'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/30/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/30 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 31
              edit EPSMEMBER '31'
              edit MEMBER '31'
              edit SUBFSFAMILY '/pf031'
              edit OCGROUP 'an1'
              edit FCGROUP 'fc6'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/31/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/31 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 32
              edit EPSMEMBER '32'
              edit MEMBER '32'
              edit SUBFSFAMILY '/pf032'
              edit OCGROUP 'an2'
              edit FCGROUP 'fc6'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/32/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/32 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 33
              edit EPSMEMBER '33'
              edit MEMBER '33'
              edit SUBFSFAMILY '/pf033'
              edit OCGROUP 'an3'
              edit FCGROUP 'fc6'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/33/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/33 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 34
              edit EPSMEMBER '34'
              edit MEMBER '34'
              edit SUBFSFAMILY '/pf034'
              edit OCGROUP 'an4'
              edit FCGROUP 'fc6'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/34/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/34 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 35
              edit EPSMEMBER '35'
              edit MEMBER '35'
              edit SUBFSFAMILY '/pf035'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc7'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/35/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/35 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 36
              edit EPSMEMBER '36'
              edit MEMBER '36'
              edit SUBFSFAMILY '/pf036'
              edit OCGROUP 'an1'
              edit FCGROUP 'fc7'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/36/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/36 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 37
              edit EPSMEMBER '37'
              edit MEMBER '37'
              edit SUBFSFAMILY '/pf037'
              edit OCGROUP 'an2'
              edit FCGROUP 'fc7'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/37/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/37 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 38
              edit EPSMEMBER '38'
              edit MEMBER '38'
              edit SUBFSFAMILY '/pf038'
              edit OCGROUP 'an3'
              edit FCGROUP 'fc7'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/38/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/38 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 39
              edit EPSMEMBER '39'
              edit MEMBER '39'
              edit SUBFSFAMILY '/pf039'
              edit OCGROUP 'an4'
              edit FCGROUP 'fc7'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/39/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/39 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 40
              edit EPSMEMBER '40'
              edit MEMBER '40'
              edit SUBFSFAMILY '/pf040'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc8'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/40/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/40 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 41
              edit EPSMEMBER '41'
              edit MEMBER '41'
              edit SUBFSFAMILY '/pf041'
              edit OCGROUP 'an1'
              edit FCGROUP 'fc8'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/41/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/41 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 42
              edit EPSMEMBER '42'
              edit MEMBER '42'
              edit SUBFSFAMILY '/pf042'
              edit OCGROUP 'an2'
              edit FCGROUP 'fc8'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/42/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/42 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 43
              edit EPSMEMBER '43'
              edit MEMBER '43'
              edit SUBFSFAMILY '/pf043'
              edit OCGROUP 'an3'
              edit FCGROUP 'fc8'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/43/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/43 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 44
              edit EPSMEMBER '44'
              edit MEMBER '44'
              edit SUBFSFAMILY '/pf044'
              edit OCGROUP 'an4'
              edit FCGROUP 'fc8'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/44/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/44 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 45
              edit EPSMEMBER '45'
              edit MEMBER '45'
              edit SUBFSFAMILY '/pf045'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc9'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/45/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/45 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 46
              edit EPSMEMBER '46'
              edit MEMBER '46'
              edit SUBFSFAMILY '/pf046'
              edit OCGROUP 'an1'
              edit FCGROUP 'fc9'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/46/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/46 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 47
              edit EPSMEMBER '47'
              edit MEMBER '47'
              edit SUBFSFAMILY '/pf047'
              edit OCGROUP 'an2'
              edit FCGROUP 'fc9'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/47/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/47 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 48
              edit EPSMEMBER '48'
              edit MEMBER '48'
              edit SUBFSFAMILY '/pf048'
              edit OCGROUP 'an3'
              edit FCGROUP 'fc9'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/48/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/48 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 49
              edit EPSMEMBER '49'
              edit MEMBER '49'
              edit SUBFSFAMILY '/pf049'
              edit OCGROUP 'an4'
              edit FCGROUP 'fc9'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/49/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/49 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
            family 50
              edit EPSMEMBER '50'
              edit MEMBER '50'
              edit SUBFSFAMILY '/pf050'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc10'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger /emc_36r1/main/00/legA/fc/pf/50/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:modelepsB
              endfamily
              task getvarepsdata
                trigger ../../../../legA/fc/pf/50 eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '144'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                edit SMSTRIES '1'
                inlimit /emc_36r1/limits:modelepsB
                meter step -1 144 144
            endfamily
          endfamily
          family cf
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit MEMBER '0'
            edit EPSMEMBER '0'
            edit SUBFSFAMILY '/cf000'
            edit EPSTYPE 'cf'
            edit EMOS_TYPE 'cf'
            edit USER_PRIORITY '72'
            edit EPSNLEGS '3'
            edit EPSFCLEV '62'
            edit EPSWAMNSTPW '1'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSFCRES '319'
            edit EPSWAMRESOL 'global50'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit EPSINISTEP '360'
            edit TSTEP '1200'
            edit EPSLEG '2'
            edit CLASS 'od'
            edit OCGROUP 'an0'
            edit FCGROUP 'fc0'
            edit FCCHUNK '1'
            edit OCTOTAL '5'
            edit FCTOTAL '51'
            edit FCLENGTH '144'
            edit FCLENGTH1 '216'
            edit DAILY '0'
            edit RUNVARFC '1'
            family ocean
              edit RUNACAN '0'
              edit RUNVARFC '1'
              edit RUNCOUPLE '0'
              edit ACAN_LAG '-12'
              edit EPSLEG '1'
              task oceini
                edit STREAM 'ocea'
              task linkoce
                trigger oceini eq complete
              task addforce
                trigger /emc_36r1/main/00/legA/fc/cf/modeleps eq complete
              task runhope
                trigger linkoce eq complete and addforce eq complete
                edit FCLENGTH '216'
                edit THREADS '16'
                inlimit /emc_36r1/limits:modelepsB
            endfamily
            task getvarepsdata
              trigger ../../../legA/fc/cf eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '2700'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSLEG '2'
              edit EPSINISTEP '216'
              edit EPSFCRES '639'
              edit EPSFCRESN '319'
              edit CLASS 'od'
            task intHtoL
              trigger getvarepsdata eq complete
              edit MEM '1300'
              edit MRF '10'
              edit CPUTIME '1000'
              edit STACK '450'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit TSTEP '2700'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '216'
              edit EPSFCRES '319'
              edit NPES '24'
              edit FCNPES '24'
              edit THREADS '2'
              edit CLASS 'od'
            task modeleps
              trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSLEG '2'
              edit EPSFCRES '319'
              edit EPSINISTEP '360'
              edit EPSINISTEP_PREVIOUS '216'
              edit FCLENGTH '144'
              edit EPSFCLENGTH_A '240'
              edit MEM '3200'
              edit MRF '10'
              edit STACK '350'
              edit FCCHUNK '1'
              edit SMSTRIES '1'
              edit NPES '1'
              edit THREADS '64'
              edit FCNPES '1'
              edit TSTEP '2700'
              edit CLASS 'od'
              inlimit /emc_36r1/limits:modelepsB
              meter step -1 144 144
          endfamily
          family reservation
            trigger ./pf/46/intHtoL eq complete and ./pf/47/intHtoL eq complete and ./pf/48/intHtoL eq complete and ./pf/49/intHtoL eq complete and ./pf/50/intHtoL eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/smsfiles'
            inlimit /emc_36r1/limits:reservation
            task reset_class
              late -s +00:02 -c +00:02
              edit RESERVATION_DEFINITION 'default'
              label info ""
            task reset_cpu
              late -s +00:02 -c +00:02
              complete /emc_36r1/main/ref/00/events:768
              edit RESERVED_CPUS '0'
              label info ""
          endfamily
        endfamily
      endfamily
      family legC
        trigger /emc_36r1/main/ref/00/events eq complete and /emc_36r1/main/ref/00/events:768
      endfamily
    endfamily
    task logfilestore
      edit QUEUE 'emos-e'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit KEEPLOGS 'no'
      edit LOGSTOECFS 'no'
      edit MOVE_TO_ECFS 'yes'
    task logfiles
      late -s +00:05 -c +00:35
      trigger logfilestore eq complete and /emc_36r1/main/make eq complete and ./12 eq complete and ./00 eq complete and /emc_36r1/main/ref/dummy eq complete
      edit QUEUE 'emos-e'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit MOVE_TO_ECFS 'no'
  endfamily
  family hc2
    repeat date YMD 20091008 20121212 7
    edit ACCOUNT 'oeenfo'
    edit LOGTASK 'hc2'
    edit QUEUE 'ns'
    edit QUEUE_EPILOG 'ns'
    edit ENSEMBLES '4'
    edit ENS_SIZE '4'
    edit RUNHINDCAST '1'
    edit STREAM 'ENFH'
    edit EMOS_STREAM 'ENFH'
    edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
    edit EMOS_BASE '00'
    edit INICLASS 'EI'
    edit INIEXPVER '0001'
    edit INILEVELS '60'
    edit DELETE_FAMILY 'true'
    family back
      trigger /emc_36r1/main/make eq complete
      inlimit /emc_36r1/limits:hind
      task stage_hc
        edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
      task rmfdb
        trigger ./stage_hc eq complete
        edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
      task lock_fam
        trigger 0 eq 1
      task lock_1701
        complete ./18 eq complete
        trigger 0 eq 1
      family 18
        trigger ./rmfdb eq complete
        edit DELTA_YEAR '-18'
        family m2
          edit DELTA_DAY '-2'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '3000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit FCNPES '16'
          edit EPSSVTSTEP '900'
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit CPUTIME '12000'
          edit STACK '350'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
          task svsh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '73'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '74'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnhevo
            trigger svnh eq complete
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit CPUTIME '120000'
          task svshevo
            trigger svsh eq complete
            edit EVOLVE 'yes'
          task svsave
            trigger svshevo eq complete and svnhevo eq complete
          task clean
            trigger svnhevo eq complete and svshevo eq complete and svsave eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
        endfamily
        family sv
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '2000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit NPES '16'
          edit EPSSVTSTEP '900'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit THREADS '8'
          task targets
            late -s +00:03 -c +00:10
            edit EPSTCBB '1'
            label da_info ""
            label dcda_info ""
            event sv
          task svsh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '73'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '74'
            inlimit /emc_36r1/limits:pesh 2
          task subspace
            trigger svnh eq complete and svsh eq complete and ( targets eq complete or targets:sv)
          task sv1
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '1'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '72'
            inlimit /emc_36r1/limits:pesh 2
          task sv2
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '2'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '71'
            inlimit /emc_36r1/limits:pesh 2
          task sv3
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '3'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv4
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '4'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv5
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '5'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv6
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '6'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task getsvevo
            event ecfs
          task getsvs
            trigger getsvevo eq complete and svnh eq complete and svsh eq complete and sv1 eq complete and sv2 eq complete and sv3 eq complete and sv4 eq complete and sv5 eq complete and sv6 eq complete and ../m2 eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTROP '1'
          task svnhevo
            trigger svnh eq complete and ../legA/fc eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit THREADS '4'
            edit STACK '350'
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
          task svshevo
            trigger svsh eq complete and ../legA/fc eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit STACK '450'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
        endfamily
        family legA
          family fc
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '240'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit TSTEP '1200'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '1'
            task getini
              edit LINIENDA '0'
              edit INIDATEENDA '0'
              edit EPSCOUPLE '0'
              edit N_PERT_AN '10'
              edit EPSCOUPL_A '0'
              edit EPSCOUPL_B '0'
              edit EPSCOUPL_C '0'
            task inidata
              trigger getini eq complete
              edit THREADS '8'
            task getae
              event 1 
            task rot
              trigger ../../sv/getsvs eq complete and ( getae:1 or getae eq complete)
              edit EPSTROP '1'
            family pert_ic
              trigger rot eq complete and inidata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTROP '1'
              inlimit /emc_36r1/limits:pertic
              family 00
                task pertinic
                  edit MEMBER '0'
              endfamily
              family 01
                task pertinic
                  edit MEMBER '1'
              endfamily
              family 02
                task pertinic
                  edit MEMBER '2'
              endfamily
              family 03
                task pertinic
                  edit MEMBER '3'
              endfamily
              family 04
                task pertinic
                  edit MEMBER '4'
              endfamily
            endfamily
            task cp_pert
              trigger pert_ic eq complete
              event 1 
            task wavini
              edit FSFAMILY 'mc'
              edit PERSST 'true'
              edit SSTINT '1'
            task wavfcdata
              trigger ./getpersSST eq complete and ../../legA/fc/wavini eq complete
              edit EPSCOUPLE '1'
            task getpersSST
              trigger ./inidata eq complete
              edit EPSCOUPLE '1'
              edit EPS_INI_STREAM 'DA'
              edit EPSFCRES '639'
            family pf
              trigger ./getpersSST eq complete and ./wavfcdata eq complete and wavini eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              meter step -1 240 240
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task getiniLeg
                  trigger ../../pert_ic/01/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task getiniLeg
                  trigger ../../pert_ic/02/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task getiniLeg
                  trigger ../../pert_ic/03/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task getiniLeg
                  trigger ../../pert_ic/04/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
            endfamily
            family cf
              trigger ./pert_ic/00/pertinic eq complete and ./getpersSST eq complete and ./wavfcdata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              task getiniLeg
                trigger ../inidata eq complete and ../wavini eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit FCNPES '32'
                edit THREADS '4'
                inlimit /emc_36r1/limits:pesh 6
                meter step -1 240 240
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../fc eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:dp
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                    edit SMSPASS 'stripped'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        family legBcpl
          trigger ./legA eq complete
          family fc
            edit EPSCOUPLE '1'
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '144'
            edit EPSINISTEP_PREVIOUS '216'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit TSTEP '2700'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '2'
            family pf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit THREADS '2'
              edit RUNVARFC '1'
              meter step -1 552 552
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/01/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/01 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/02/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/02 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/03/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/03 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/04/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/04 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
            endfamily
            family cf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc0'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger ../../../../legA/fc/cf/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:pesh
              endfamily
              task getvarepsdata
                trigger ../../../legA/fc/cf eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '552'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit SMSTRIES '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:pesh
                meter step -1 552 552
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../../legA/postproc/dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:dp
              family dp
                trigger ./cf eq complete and ./pf eq complete
                edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
                edit OVERLAP '1'
                edit STEPS '216:240:6'
                inlimit /emc_36r1/limits:dp
                family cf
                  edit EMOS_TYPE 'cf'
                  task derived_param
                    edit MEMBER 'off'
                endfamily
                family pf
                  edit EMOS_TYPE 'pf'
                  family 01
                    edit EPSMEMBER '1'
                    edit MEMBER '1'
                    edit SUBFSFAMILY '/pf001'
                    task derived_param
                  endfamily
                  family 02
                    edit EPSMEMBER '2'
                    edit MEMBER '2'
                    edit SUBFSFAMILY '/pf002'
                    task derived_param
                  endfamily
                  family 03
                    edit EPSMEMBER '3'
                    edit MEMBER '3'
                    edit SUBFSFAMILY '/pf003'
                    task derived_param
                  endfamily
                  family 04
                    edit EPSMEMBER '4'
                    edit MEMBER '4'
                    edit SUBFSFAMILY '/pf004'
                    task derived_param
                  endfamily
                endfamily
              endfamily
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                edit EPSFCRES '319'
                task extrafields
                  edit FCLENGTH '768'
                  edit FCLENGTH0 '240'
                task wm_sfc
                  trigger extrafields eq complete
                  edit FCLENGTH '768'
                task wm_ua
                  trigger extrafields eq complete and wm_sfc eq complete
                  edit FCLENGTH '768'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        task save_logs
          trigger m2 eq complete and sv eq complete and legA eq complete and legBcpl eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit SCHOST 'localhost'
      endfamily
      family 17
        trigger ./lock_1701 eq complete and ./rmfdb eq complete
        edit DELTA_YEAR '-17'
        family m2
          edit DELTA_DAY '-2'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '3000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit FCNPES '16'
          edit EPSSVTSTEP '900'
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit CPUTIME '12000'
          edit STACK '350'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
          task svsh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '73'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '74'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnhevo
            trigger svnh eq complete
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit CPUTIME '120000'
          task svshevo
            trigger svsh eq complete
            edit EVOLVE 'yes'
          task svsave
            trigger svshevo eq complete and svnhevo eq complete
          task clean
            trigger svnhevo eq complete and svshevo eq complete and svsave eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
        endfamily
        family sv
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '2000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit NPES '16'
          edit EPSSVTSTEP '900'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit THREADS '8'
          task targets
            late -s +00:03 -c +00:10
            edit EPSTCBB '1'
            label da_info ""
            label dcda_info ""
            event sv
          task svsh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '73'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '74'
            inlimit /emc_36r1/limits:pesh 2
          task subspace
            trigger svnh eq complete and svsh eq complete and ( targets eq complete or targets:sv)
          task sv1
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '1'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '72'
            inlimit /emc_36r1/limits:pesh 2
          task sv2
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '2'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '71'
            inlimit /emc_36r1/limits:pesh 2
          task sv3
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '3'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv4
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '4'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv5
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '5'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv6
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '6'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task getsvevo
            event ecfs
          task getsvs
            trigger getsvevo eq complete and svnh eq complete and svsh eq complete and sv1 eq complete and sv2 eq complete and sv3 eq complete and sv4 eq complete and sv5 eq complete and sv6 eq complete and ../m2 eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTROP '1'
          task svnhevo
            trigger svnh eq complete and ../legA/fc eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit THREADS '4'
            edit STACK '350'
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
          task svshevo
            trigger svsh eq complete and ../legA/fc eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit STACK '450'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
        endfamily
        family legA
          family fc
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '240'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit TSTEP '1200'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '1'
            task getini
              edit LINIENDA '0'
              edit INIDATEENDA '0'
              edit EPSCOUPLE '0'
              edit N_PERT_AN '10'
              edit EPSCOUPL_A '0'
              edit EPSCOUPL_B '0'
              edit EPSCOUPL_C '0'
            task inidata
              trigger getini eq complete
              edit THREADS '8'
            task getae
              event 1 
            task rot
              trigger ../../sv/getsvs eq complete and ( getae:1 or getae eq complete)
              edit EPSTROP '1'
            family pert_ic
              trigger rot eq complete and inidata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTROP '1'
              inlimit /emc_36r1/limits:pertic
              family 00
                task pertinic
                  edit MEMBER '0'
              endfamily
              family 01
                task pertinic
                  edit MEMBER '1'
              endfamily
              family 02
                task pertinic
                  edit MEMBER '2'
              endfamily
              family 03
                task pertinic
                  edit MEMBER '3'
              endfamily
              family 04
                task pertinic
                  edit MEMBER '4'
              endfamily
            endfamily
            task cp_pert
              trigger pert_ic eq complete
              event 1 
            task wavini
              edit FSFAMILY 'mc'
              edit PERSST 'true'
              edit SSTINT '1'
            task wavfcdata
              trigger ./getpersSST eq complete and ../../legA/fc/wavini eq complete
              edit EPSCOUPLE '1'
            task getpersSST
              trigger ./inidata eq complete
              edit EPSCOUPLE '1'
              edit EPS_INI_STREAM 'DA'
              edit EPSFCRES '639'
            family pf
              trigger ./getpersSST eq complete and ./wavfcdata eq complete and wavini eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              meter step -1 240 240
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task getiniLeg
                  trigger ../../pert_ic/01/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task getiniLeg
                  trigger ../../pert_ic/02/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task getiniLeg
                  trigger ../../pert_ic/03/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task getiniLeg
                  trigger ../../pert_ic/04/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
            endfamily
            family cf
              trigger ./pert_ic/00/pertinic eq complete and ./getpersSST eq complete and ./wavfcdata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              task getiniLeg
                trigger ../inidata eq complete and ../wavini eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit FCNPES '32'
                edit THREADS '4'
                inlimit /emc_36r1/limits:pesh 6
                meter step -1 240 240
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../fc eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:dp
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        family legBcpl
          trigger ./legA eq complete
          family fc
            edit EPSCOUPLE '1'
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '144'
            edit EPSINISTEP_PREVIOUS '216'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit TSTEP '2700'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '2'
            family pf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit THREADS '2'
              edit RUNVARFC '1'
              meter step -1 552 552
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/01/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/01 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/02/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/02 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/03/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/03 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/04/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/04 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
            endfamily
            family cf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc0'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger ../../../../legA/fc/cf/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:pesh
              endfamily
              task getvarepsdata
                trigger ../../../legA/fc/cf eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '552'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit SMSTRIES '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:pesh
                meter step -1 552 552
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../../legA/postproc/dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:dp
              family dp
                trigger ./cf eq complete and ./pf eq complete
                edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
                edit OVERLAP '1'
                edit STEPS '216:240:6'
                inlimit /emc_36r1/limits:dp
                family cf
                  edit EMOS_TYPE 'cf'
                  task derived_param
                    edit MEMBER 'off'
                endfamily
                family pf
                  edit EMOS_TYPE 'pf'
                  family 01
                    edit EPSMEMBER '1'
                    edit MEMBER '1'
                    edit SUBFSFAMILY '/pf001'
                    task derived_param
                  endfamily
                  family 02
                    edit EPSMEMBER '2'
                    edit MEMBER '2'
                    edit SUBFSFAMILY '/pf002'
                    task derived_param
                  endfamily
                  family 03
                    edit EPSMEMBER '3'
                    edit MEMBER '3'
                    edit SUBFSFAMILY '/pf003'
                    task derived_param
                  endfamily
                  family 04
                    edit EPSMEMBER '4'
                    edit MEMBER '4'
                    edit SUBFSFAMILY '/pf004'
                    task derived_param
                  endfamily
                endfamily
              endfamily
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                edit EPSFCRES '319'
                task extrafields
                  edit FCLENGTH '768'
                  edit FCLENGTH0 '240'
                task wm_sfc
                  trigger extrafields eq complete
                  edit FCLENGTH '768'
                task wm_ua
                  trigger extrafields eq complete and wm_sfc eq complete
                  edit FCLENGTH '768'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        task save_logs
          trigger m2 eq complete and sv eq complete and legA eq complete and legBcpl eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit SCHOST 'localhost'
      endfamily
      family 16
        trigger ./lock_1701 eq complete and ./rmfdb eq complete
        edit DELTA_YEAR '-16'
        family m2
          edit DELTA_DAY '-2'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '3000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit FCNPES '16'
          edit EPSSVTSTEP '900'
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit CPUTIME '12000'
          edit STACK '350'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
          task svsh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '73'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '74'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnhevo
            trigger svnh eq complete
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit CPUTIME '120000'
          task svshevo
            trigger svsh eq complete
            edit EVOLVE 'yes'
          task svsave
            trigger svshevo eq complete and svnhevo eq complete
          task clean
            trigger svnhevo eq complete and svshevo eq complete and svsave eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
        endfamily
        family sv
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '2000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit NPES '16'
          edit EPSSVTSTEP '900'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit THREADS '8'
          task targets
            late -s +00:03 -c +00:10
            edit EPSTCBB '1'
            label da_info ""
            label dcda_info ""
            event sv
          task svsh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '73'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '74'
            inlimit /emc_36r1/limits:pesh 2
          task subspace
            trigger svnh eq complete and svsh eq complete and ( targets eq complete or targets:sv)
          task sv1
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '1'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '72'
            inlimit /emc_36r1/limits:pesh 2
          task sv2
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '2'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '71'
            inlimit /emc_36r1/limits:pesh 2
          task sv3
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '3'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv4
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '4'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv5
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '5'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv6
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '6'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task getsvevo
            event ecfs
          task getsvs
            trigger getsvevo eq complete and svnh eq complete and svsh eq complete and sv1 eq complete and sv2 eq complete and sv3 eq complete and sv4 eq complete and sv5 eq complete and sv6 eq complete and ../m2 eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTROP '1'
          task svnhevo
            trigger svnh eq complete and ../legA/fc eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit THREADS '4'
            edit STACK '350'
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
          task svshevo
            trigger svsh eq complete and ../legA/fc eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit STACK '450'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
        endfamily
        family legA
          family fc
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '240'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit TSTEP '1200'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '1'
            task getini
              edit LINIENDA '0'
              edit INIDATEENDA '0'
              edit EPSCOUPLE '0'
              edit N_PERT_AN '10'
              edit EPSCOUPL_A '0'
              edit EPSCOUPL_B '0'
              edit EPSCOUPL_C '0'
            task inidata
              trigger getini eq complete
              edit THREADS '8'
            task getae
              event 1 
            task rot
              trigger ../../sv/getsvs eq complete and ( getae:1 or getae eq complete)
              edit EPSTROP '1'
            family pert_ic
              trigger rot eq complete and inidata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTROP '1'
              inlimit /emc_36r1/limits:pertic
              family 00
                task pertinic
                  edit MEMBER '0'
              endfamily
              family 01
                task pertinic
                  edit MEMBER '1'
              endfamily
              family 02
                task pertinic
                  edit MEMBER '2'
              endfamily
              family 03
                task pertinic
                  edit MEMBER '3'
              endfamily
              family 04
                task pertinic
                  edit MEMBER '4'
              endfamily
            endfamily
            task cp_pert
              trigger pert_ic eq complete
              event 1 
            task wavini
              edit FSFAMILY 'mc'
              edit PERSST 'true'
              edit SSTINT '1'
            task wavfcdata
              trigger ./getpersSST eq complete and ../../legA/fc/wavini eq complete
              edit EPSCOUPLE '1'
            task getpersSST
              trigger ./inidata eq complete
              edit EPSCOUPLE '1'
              edit EPS_INI_STREAM 'DA'
              edit EPSFCRES '639'
            family pf
              trigger ./getpersSST eq complete and ./wavfcdata eq complete and wavini eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              meter step -1 240 240
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task getiniLeg
                  trigger ../../pert_ic/01/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task getiniLeg
                  trigger ../../pert_ic/02/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task getiniLeg
                  trigger ../../pert_ic/03/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task getiniLeg
                  trigger ../../pert_ic/04/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
            endfamily
            family cf
              trigger ./pert_ic/00/pertinic eq complete and ./getpersSST eq complete and ./wavfcdata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              task getiniLeg
                trigger ../inidata eq complete and ../wavini eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit FCNPES '32'
                edit THREADS '4'
                inlimit /emc_36r1/limits:pesh 6
                meter step -1 240 240
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../fc eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:dp
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        family legBcpl
          trigger ./legA eq complete
          family fc
            edit EPSCOUPLE '1'
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '144'
            edit EPSINISTEP_PREVIOUS '216'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit TSTEP '2700'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '2'
            family pf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit THREADS '2'
              edit RUNVARFC '1'
              meter step -1 552 552
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/01/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/01 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/02/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/02 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/03/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/03 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/04/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/04 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
            endfamily
            family cf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc0'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger ../../../../legA/fc/cf/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:pesh
              endfamily
              task getvarepsdata
                trigger ../../../legA/fc/cf eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '552'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit SMSTRIES '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:pesh
                meter step -1 552 552
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../../legA/postproc/dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:dp
              family dp
                trigger ./cf eq complete and ./pf eq complete
                edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
                edit OVERLAP '1'
                edit STEPS '216:240:6'
                inlimit /emc_36r1/limits:dp
                family cf
                  edit EMOS_TYPE 'cf'
                  task derived_param
                    edit MEMBER 'off'
                endfamily
                family pf
                  edit EMOS_TYPE 'pf'
                  family 01
                    edit EPSMEMBER '1'
                    edit MEMBER '1'
                    edit SUBFSFAMILY '/pf001'
                    task derived_param
                  endfamily
                  family 02
                    edit EPSMEMBER '2'
                    edit MEMBER '2'
                    edit SUBFSFAMILY '/pf002'
                    task derived_param
                  endfamily
                  family 03
                    edit EPSMEMBER '3'
                    edit MEMBER '3'
                    edit SUBFSFAMILY '/pf003'
                    task derived_param
                  endfamily
                  family 04
                    edit EPSMEMBER '4'
                    edit MEMBER '4'
                    edit SUBFSFAMILY '/pf004'
                    task derived_param
                  endfamily
                endfamily
              endfamily
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                edit EPSFCRES '319'
                task extrafields
                  edit FCLENGTH '768'
                  edit FCLENGTH0 '240'
                task wm_sfc
                  trigger extrafields eq complete
                  edit FCLENGTH '768'
                task wm_ua
                  trigger extrafields eq complete and wm_sfc eq complete
                  edit FCLENGTH '768'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        task save_logs
          trigger m2 eq complete and sv eq complete and legA eq complete and legBcpl eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit SCHOST 'localhost'
      endfamily
      family 15
        trigger ./lock_1701 eq complete and ./rmfdb eq complete
        edit DELTA_YEAR '-15'
        family m2
          edit DELTA_DAY '-2'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '3000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit FCNPES '16'
          edit EPSSVTSTEP '900'
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit CPUTIME '12000'
          edit STACK '350'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
          task svsh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '73'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '74'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnhevo
            trigger svnh eq complete
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit CPUTIME '120000'
          task svshevo
            trigger svsh eq complete
            edit EVOLVE 'yes'
          task svsave
            trigger svshevo eq complete and svnhevo eq complete
          task clean
            trigger svnhevo eq complete and svshevo eq complete and svsave eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
        endfamily
        family sv
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '2000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit NPES '16'
          edit EPSSVTSTEP '900'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit THREADS '8'
          task targets
            late -s +00:03 -c +00:10
            edit EPSTCBB '1'
            label da_info ""
            label dcda_info ""
            event sv
          task svsh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '73'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '74'
            inlimit /emc_36r1/limits:pesh 2
          task subspace
            trigger svnh eq complete and svsh eq complete and ( targets eq complete or targets:sv)
          task sv1
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '1'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '72'
            inlimit /emc_36r1/limits:pesh 2
          task sv2
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '2'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '71'
            inlimit /emc_36r1/limits:pesh 2
          task sv3
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '3'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv4
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '4'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv5
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '5'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv6
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '6'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task getsvevo
            event ecfs
          task getsvs
            trigger getsvevo eq complete and svnh eq complete and svsh eq complete and sv1 eq complete and sv2 eq complete and sv3 eq complete and sv4 eq complete and sv5 eq complete and sv6 eq complete and ../m2 eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTROP '1'
          task svnhevo
            trigger svnh eq complete and ../legA/fc eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit THREADS '4'
            edit STACK '350'
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
          task svshevo
            trigger svsh eq complete and ../legA/fc eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit STACK '450'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
        endfamily
        family legA
          family fc
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '240'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit TSTEP '1200'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '1'
            task getini
              edit LINIENDA '0'
              edit INIDATEENDA '0'
              edit EPSCOUPLE '0'
              edit N_PERT_AN '10'
              edit EPSCOUPL_A '0'
              edit EPSCOUPL_B '0'
              edit EPSCOUPL_C '0'
            task inidata
              trigger getini eq complete
              edit THREADS '8'
            task getae
              event 1 
            task rot
              trigger ../../sv/getsvs eq complete and ( getae:1 or getae eq complete)
              edit EPSTROP '1'
            family pert_ic
              trigger rot eq complete and inidata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTROP '1'
              inlimit /emc_36r1/limits:pertic
              family 00
                task pertinic
                  edit MEMBER '0'
              endfamily
              family 01
                task pertinic
                  edit MEMBER '1'
              endfamily
              family 02
                task pertinic
                  edit MEMBER '2'
              endfamily
              family 03
                task pertinic
                  edit MEMBER '3'
              endfamily
              family 04
                task pertinic
                  edit MEMBER '4'
              endfamily
            endfamily
            task cp_pert
              trigger pert_ic eq complete
              event 1 
            task wavini
              edit FSFAMILY 'mc'
              edit PERSST 'true'
              edit SSTINT '1'
            task wavfcdata
              trigger ./getpersSST eq complete and ../../legA/fc/wavini eq complete
              edit EPSCOUPLE '1'
            task getpersSST
              trigger ./inidata eq complete
              edit EPSCOUPLE '1'
              edit EPS_INI_STREAM 'DA'
              edit EPSFCRES '639'
            family pf
              trigger ./getpersSST eq complete and ./wavfcdata eq complete and wavini eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              meter step -1 240 240
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task getiniLeg
                  trigger ../../pert_ic/01/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task getiniLeg
                  trigger ../../pert_ic/02/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task getiniLeg
                  trigger ../../pert_ic/03/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task getiniLeg
                  trigger ../../pert_ic/04/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
            endfamily
            family cf
              trigger ./pert_ic/00/pertinic eq complete and ./getpersSST eq complete and ./wavfcdata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              task getiniLeg
                trigger ../inidata eq complete and ../wavini eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit FCNPES '32'
                edit THREADS '4'
                inlimit /emc_36r1/limits:pesh 6
                meter step -1 240 240
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../fc eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:dp
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        family legBcpl
          trigger ./legA eq complete
          family fc
            edit EPSCOUPLE '1'
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '144'
            edit EPSINISTEP_PREVIOUS '216'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit TSTEP '2700'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '2'
            family pf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit THREADS '2'
              edit RUNVARFC '1'
              meter step -1 552 552
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/01/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/01 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/02/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/02 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/03/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/03 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/04/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/04 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
            endfamily
            family cf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc0'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger ../../../../legA/fc/cf/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:pesh
              endfamily
              task getvarepsdata
                trigger ../../../legA/fc/cf eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '552'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit SMSTRIES '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:pesh
                meter step -1 552 552
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../../legA/postproc/dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:dp
              family dp
                trigger ./cf eq complete and ./pf eq complete
                edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
                edit OVERLAP '1'
                edit STEPS '216:240:6'
                inlimit /emc_36r1/limits:dp
                family cf
                  edit EMOS_TYPE 'cf'
                  task derived_param
                    edit MEMBER 'off'
                endfamily
                family pf
                  edit EMOS_TYPE 'pf'
                  family 01
                    edit EPSMEMBER '1'
                    edit MEMBER '1'
                    edit SUBFSFAMILY '/pf001'
                    task derived_param
                  endfamily
                  family 02
                    edit EPSMEMBER '2'
                    edit MEMBER '2'
                    edit SUBFSFAMILY '/pf002'
                    task derived_param
                  endfamily
                  family 03
                    edit EPSMEMBER '3'
                    edit MEMBER '3'
                    edit SUBFSFAMILY '/pf003'
                    task derived_param
                  endfamily
                  family 04
                    edit EPSMEMBER '4'
                    edit MEMBER '4'
                    edit SUBFSFAMILY '/pf004'
                    task derived_param
                  endfamily
                endfamily
              endfamily
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                edit EPSFCRES '319'
                task extrafields
                  edit FCLENGTH '768'
                  edit FCLENGTH0 '240'
                task wm_sfc
                  trigger extrafields eq complete
                  edit FCLENGTH '768'
                task wm_ua
                  trigger extrafields eq complete and wm_sfc eq complete
                  edit FCLENGTH '768'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        task save_logs
          trigger m2 eq complete and sv eq complete and legA eq complete and legBcpl eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit SCHOST 'localhost'
      endfamily
      family 14
        trigger ./lock_1701 eq complete and ./rmfdb eq complete
        edit DELTA_YEAR '-14'
        family m2
          edit DELTA_DAY '-2'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '3000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit FCNPES '16'
          edit EPSSVTSTEP '900'
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit CPUTIME '12000'
          edit STACK '350'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
          task svsh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '73'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '74'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnhevo
            trigger svnh eq complete
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit CPUTIME '120000'
          task svshevo
            trigger svsh eq complete
            edit EVOLVE 'yes'
          task svsave
            trigger svshevo eq complete and svnhevo eq complete
          task clean
            trigger svnhevo eq complete and svshevo eq complete and svsave eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
        endfamily
        family sv
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '2000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit NPES '16'
          edit EPSSVTSTEP '900'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit THREADS '8'
          task targets
            late -s +00:03 -c +00:10
            edit EPSTCBB '1'
            label da_info ""
            label dcda_info ""
            event sv
          task svsh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '73'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '74'
            inlimit /emc_36r1/limits:pesh 2
          task subspace
            trigger svnh eq complete and svsh eq complete and ( targets eq complete or targets:sv)
          task sv1
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '1'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '72'
            inlimit /emc_36r1/limits:pesh 2
          task sv2
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '2'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '71'
            inlimit /emc_36r1/limits:pesh 2
          task sv3
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '3'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv4
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '4'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv5
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '5'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv6
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '6'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task getsvevo
            event ecfs
          task getsvs
            trigger getsvevo eq complete and svnh eq complete and svsh eq complete and sv1 eq complete and sv2 eq complete and sv3 eq complete and sv4 eq complete and sv5 eq complete and sv6 eq complete and ../m2 eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTROP '1'
          task svnhevo
            trigger svnh eq complete and ../legA/fc eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit THREADS '4'
            edit STACK '350'
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
          task svshevo
            trigger svsh eq complete and ../legA/fc eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit STACK '450'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
        endfamily
        family legA
          family fc
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '240'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit TSTEP '1200'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '1'
            task getini
              edit LINIENDA '0'
              edit INIDATEENDA '0'
              edit EPSCOUPLE '0'
              edit N_PERT_AN '10'
              edit EPSCOUPL_A '0'
              edit EPSCOUPL_B '0'
              edit EPSCOUPL_C '0'
            task inidata
              trigger getini eq complete
              edit THREADS '8'
            task getae
              event 1 
            task rot
              trigger ../../sv/getsvs eq complete and ( getae:1 or getae eq complete)
              edit EPSTROP '1'
            family pert_ic
              trigger rot eq complete and inidata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTROP '1'
              inlimit /emc_36r1/limits:pertic
              family 00
                task pertinic
                  edit MEMBER '0'
              endfamily
              family 01
                task pertinic
                  edit MEMBER '1'
              endfamily
              family 02
                task pertinic
                  edit MEMBER '2'
              endfamily
              family 03
                task pertinic
                  edit MEMBER '3'
              endfamily
              family 04
                task pertinic
                  edit MEMBER '4'
              endfamily
            endfamily
            task cp_pert
              trigger pert_ic eq complete
              event 1 
            task wavini
              edit FSFAMILY 'mc'
              edit PERSST 'true'
              edit SSTINT '1'
            task wavfcdata
              trigger ./getpersSST eq complete and ../../legA/fc/wavini eq complete
              edit EPSCOUPLE '1'
            task getpersSST
              trigger ./inidata eq complete
              edit EPSCOUPLE '1'
              edit EPS_INI_STREAM 'DA'
              edit EPSFCRES '639'
            family pf
              trigger ./getpersSST eq complete and ./wavfcdata eq complete and wavini eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              meter step -1 240 240
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task getiniLeg
                  trigger ../../pert_ic/01/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task getiniLeg
                  trigger ../../pert_ic/02/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task getiniLeg
                  trigger ../../pert_ic/03/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task getiniLeg
                  trigger ../../pert_ic/04/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
            endfamily
            family cf
              trigger ./pert_ic/00/pertinic eq complete and ./getpersSST eq complete and ./wavfcdata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              task getiniLeg
                trigger ../inidata eq complete and ../wavini eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit FCNPES '32'
                edit THREADS '4'
                inlimit /emc_36r1/limits:pesh 6
                meter step -1 240 240
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../fc eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:dp
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        family legBcpl
          trigger ./legA eq complete
          family fc
            edit EPSCOUPLE '1'
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '144'
            edit EPSINISTEP_PREVIOUS '216'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit TSTEP '2700'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '2'
            family pf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit THREADS '2'
              edit RUNVARFC '1'
              meter step -1 552 552
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/01/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/01 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/02/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/02 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/03/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/03 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/04/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/04 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
            endfamily
            family cf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc0'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger ../../../../legA/fc/cf/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:pesh
              endfamily
              task getvarepsdata
                trigger ../../../legA/fc/cf eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '552'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit SMSTRIES '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:pesh
                meter step -1 552 552
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../../legA/postproc/dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:dp
              family dp
                trigger ./cf eq complete and ./pf eq complete
                edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
                edit OVERLAP '1'
                edit STEPS '216:240:6'
                inlimit /emc_36r1/limits:dp
                family cf
                  edit EMOS_TYPE 'cf'
                  task derived_param
                    edit MEMBER 'off'
                endfamily
                family pf
                  edit EMOS_TYPE 'pf'
                  family 01
                    edit EPSMEMBER '1'
                    edit MEMBER '1'
                    edit SUBFSFAMILY '/pf001'
                    task derived_param
                  endfamily
                  family 02
                    edit EPSMEMBER '2'
                    edit MEMBER '2'
                    edit SUBFSFAMILY '/pf002'
                    task derived_param
                  endfamily
                  family 03
                    edit EPSMEMBER '3'
                    edit MEMBER '3'
                    edit SUBFSFAMILY '/pf003'
                    task derived_param
                  endfamily
                  family 04
                    edit EPSMEMBER '4'
                    edit MEMBER '4'
                    edit SUBFSFAMILY '/pf004'
                    task derived_param
                  endfamily
                endfamily
              endfamily
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                edit EPSFCRES '319'
                task extrafields
                  edit FCLENGTH '768'
                  edit FCLENGTH0 '240'
                task wm_sfc
                  trigger extrafields eq complete
                  edit FCLENGTH '768'
                task wm_ua
                  trigger extrafields eq complete and wm_sfc eq complete
                  edit FCLENGTH '768'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        task save_logs
          trigger m2 eq complete and sv eq complete and legA eq complete and legBcpl eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit SCHOST 'localhost'
      endfamily
      family 13
        trigger ./lock_1701 eq complete and ./rmfdb eq complete
        edit DELTA_YEAR '-13'
        family m2
          edit DELTA_DAY '-2'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '3000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit FCNPES '16'
          edit EPSSVTSTEP '900'
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit CPUTIME '12000'
          edit STACK '350'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
          task svsh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '73'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '74'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnhevo
            trigger svnh eq complete
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit CPUTIME '120000'
          task svshevo
            trigger svsh eq complete
            edit EVOLVE 'yes'
          task svsave
            trigger svshevo eq complete and svnhevo eq complete
          task clean
            trigger svnhevo eq complete and svshevo eq complete and svsave eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
        endfamily
        family sv
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '2000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit NPES '16'
          edit EPSSVTSTEP '900'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit THREADS '8'
          task targets
            late -s +00:03 -c +00:10
            edit EPSTCBB '1'
            label da_info ""
            label dcda_info ""
            event sv
          task svsh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '73'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '74'
            inlimit /emc_36r1/limits:pesh 2
          task subspace
            trigger svnh eq complete and svsh eq complete and ( targets eq complete or targets:sv)
          task sv1
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '1'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '72'
            inlimit /emc_36r1/limits:pesh 2
          task sv2
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '2'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '71'
            inlimit /emc_36r1/limits:pesh 2
          task sv3
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '3'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv4
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '4'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv5
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '5'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv6
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '6'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task getsvevo
            event ecfs
          task getsvs
            trigger getsvevo eq complete and svnh eq complete and svsh eq complete and sv1 eq complete and sv2 eq complete and sv3 eq complete and sv4 eq complete and sv5 eq complete and sv6 eq complete and ../m2 eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTROP '1'
          task svnhevo
            trigger svnh eq complete and ../legA/fc eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit THREADS '4'
            edit STACK '350'
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
          task svshevo
            trigger svsh eq complete and ../legA/fc eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit STACK '450'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
        endfamily
        family legA
          family fc
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '240'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit TSTEP '1200'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '1'
            task getini
              edit LINIENDA '0'
              edit INIDATEENDA '0'
              edit EPSCOUPLE '0'
              edit N_PERT_AN '10'
              edit EPSCOUPL_A '0'
              edit EPSCOUPL_B '0'
              edit EPSCOUPL_C '0'
            task inidata
              trigger getini eq complete
              edit THREADS '8'
            task getae
              event 1 
            task rot
              trigger ../../sv/getsvs eq complete and ( getae:1 or getae eq complete)
              edit EPSTROP '1'
            family pert_ic
              trigger rot eq complete and inidata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTROP '1'
              inlimit /emc_36r1/limits:pertic
              family 00
                task pertinic
                  edit MEMBER '0'
              endfamily
              family 01
                task pertinic
                  edit MEMBER '1'
              endfamily
              family 02
                task pertinic
                  edit MEMBER '2'
              endfamily
              family 03
                task pertinic
                  edit MEMBER '3'
              endfamily
              family 04
                task pertinic
                  edit MEMBER '4'
              endfamily
            endfamily
            task cp_pert
              trigger pert_ic eq complete
              event 1 
            task wavini
              edit FSFAMILY 'mc'
              edit PERSST 'true'
              edit SSTINT '1'
            task wavfcdata
              trigger ./getpersSST eq complete and ../../legA/fc/wavini eq complete
              edit EPSCOUPLE '1'
            task getpersSST
              trigger ./inidata eq complete
              edit EPSCOUPLE '1'
              edit EPS_INI_STREAM 'DA'
              edit EPSFCRES '639'
            family pf
              trigger ./getpersSST eq complete and ./wavfcdata eq complete and wavini eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              meter step -1 240 240
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task getiniLeg
                  trigger ../../pert_ic/01/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task getiniLeg
                  trigger ../../pert_ic/02/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task getiniLeg
                  trigger ../../pert_ic/03/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task getiniLeg
                  trigger ../../pert_ic/04/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
            endfamily
            family cf
              trigger ./pert_ic/00/pertinic eq complete and ./getpersSST eq complete and ./wavfcdata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              task getiniLeg
                trigger ../inidata eq complete and ../wavini eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit FCNPES '32'
                edit THREADS '4'
                inlimit /emc_36r1/limits:pesh 6
                meter step -1 240 240
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../fc eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:dp
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        family legBcpl
          trigger ./legA eq complete
          family fc
            edit EPSCOUPLE '1'
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '144'
            edit EPSINISTEP_PREVIOUS '216'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit TSTEP '2700'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '2'
            family pf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit THREADS '2'
              edit RUNVARFC '1'
              meter step -1 552 552
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/01/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/01 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/02/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/02 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/03/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/03 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/04/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/04 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
            endfamily
            family cf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc0'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger ../../../../legA/fc/cf/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:pesh
              endfamily
              task getvarepsdata
                trigger ../../../legA/fc/cf eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '552'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit SMSTRIES '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:pesh
                meter step -1 552 552
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../../legA/postproc/dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:dp
              family dp
                trigger ./cf eq complete and ./pf eq complete
                edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
                edit OVERLAP '1'
                edit STEPS '216:240:6'
                inlimit /emc_36r1/limits:dp
                family cf
                  edit EMOS_TYPE 'cf'
                  task derived_param
                    edit MEMBER 'off'
                endfamily
                family pf
                  edit EMOS_TYPE 'pf'
                  family 01
                    edit EPSMEMBER '1'
                    edit MEMBER '1'
                    edit SUBFSFAMILY '/pf001'
                    task derived_param
                  endfamily
                  family 02
                    edit EPSMEMBER '2'
                    edit MEMBER '2'
                    edit SUBFSFAMILY '/pf002'
                    task derived_param
                  endfamily
                  family 03
                    edit EPSMEMBER '3'
                    edit MEMBER '3'
                    edit SUBFSFAMILY '/pf003'
                    task derived_param
                  endfamily
                  family 04
                    edit EPSMEMBER '4'
                    edit MEMBER '4'
                    edit SUBFSFAMILY '/pf004'
                    task derived_param
                  endfamily
                endfamily
              endfamily
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                edit EPSFCRES '319'
                task extrafields
                  edit FCLENGTH '768'
                  edit FCLENGTH0 '240'
                task wm_sfc
                  trigger extrafields eq complete
                  edit FCLENGTH '768'
                task wm_ua
                  trigger extrafields eq complete and wm_sfc eq complete
                  edit FCLENGTH '768'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        task save_logs
          trigger m2 eq complete and sv eq complete and legA eq complete and legBcpl eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit SCHOST 'localhost'
      endfamily
      family 12
        trigger ./lock_1701 eq complete and ./rmfdb eq complete
        edit DELTA_YEAR '-12'
        family m2
          edit DELTA_DAY '-2'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '3000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit FCNPES '16'
          edit EPSSVTSTEP '900'
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit CPUTIME '12000'
          edit STACK '350'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
          task svsh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '73'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '74'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnhevo
            trigger svnh eq complete
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit CPUTIME '120000'
          task svshevo
            trigger svsh eq complete
            edit EVOLVE 'yes'
          task svsave
            trigger svshevo eq complete and svnhevo eq complete
          task clean
            trigger svnhevo eq complete and svshevo eq complete and svsave eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
        endfamily
        family sv
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '2000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit NPES '16'
          edit EPSSVTSTEP '900'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit THREADS '8'
          task targets
            late -s +00:03 -c +00:10
            edit EPSTCBB '1'
            label da_info ""
            label dcda_info ""
            event sv
          task svsh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '73'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '74'
            inlimit /emc_36r1/limits:pesh 2
          task subspace
            trigger svnh eq complete and svsh eq complete and ( targets eq complete or targets:sv)
          task sv1
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '1'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '72'
            inlimit /emc_36r1/limits:pesh 2
          task sv2
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '2'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '71'
            inlimit /emc_36r1/limits:pesh 2
          task sv3
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '3'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv4
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '4'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv5
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '5'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv6
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '6'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task getsvevo
            event ecfs
          task getsvs
            trigger getsvevo eq complete and svnh eq complete and svsh eq complete and sv1 eq complete and sv2 eq complete and sv3 eq complete and sv4 eq complete and sv5 eq complete and sv6 eq complete and ../m2 eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTROP '1'
          task svnhevo
            trigger svnh eq complete and ../legA/fc eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit THREADS '4'
            edit STACK '350'
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
          task svshevo
            trigger svsh eq complete and ../legA/fc eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit STACK '450'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
        endfamily
        family legA
          family fc
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '240'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit TSTEP '1200'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '1'
            task getini
              edit LINIENDA '0'
              edit INIDATEENDA '0'
              edit EPSCOUPLE '0'
              edit N_PERT_AN '10'
              edit EPSCOUPL_A '0'
              edit EPSCOUPL_B '0'
              edit EPSCOUPL_C '0'
            task inidata
              trigger getini eq complete
              edit THREADS '8'
            task getae
              event 1 
            task rot
              trigger ../../sv/getsvs eq complete and ( getae:1 or getae eq complete)
              edit EPSTROP '1'
            family pert_ic
              trigger rot eq complete and inidata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTROP '1'
              inlimit /emc_36r1/limits:pertic
              family 00
                task pertinic
                  edit MEMBER '0'
              endfamily
              family 01
                task pertinic
                  edit MEMBER '1'
              endfamily
              family 02
                task pertinic
                  edit MEMBER '2'
              endfamily
              family 03
                task pertinic
                  edit MEMBER '3'
              endfamily
              family 04
                task pertinic
                  edit MEMBER '4'
              endfamily
            endfamily
            task cp_pert
              trigger pert_ic eq complete
              event 1 
            task wavini
              edit FSFAMILY 'mc'
              edit PERSST 'true'
              edit SSTINT '1'
            task wavfcdata
              trigger ./getpersSST eq complete and ../../legA/fc/wavini eq complete
              edit EPSCOUPLE '1'
            task getpersSST
              trigger ./inidata eq complete
              edit EPSCOUPLE '1'
              edit EPS_INI_STREAM 'DA'
              edit EPSFCRES '639'
            family pf
              trigger ./getpersSST eq complete and ./wavfcdata eq complete and wavini eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              meter step -1 240 240
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task getiniLeg
                  trigger ../../pert_ic/01/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task getiniLeg
                  trigger ../../pert_ic/02/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task getiniLeg
                  trigger ../../pert_ic/03/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task getiniLeg
                  trigger ../../pert_ic/04/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
            endfamily
            family cf
              trigger ./pert_ic/00/pertinic eq complete and ./getpersSST eq complete and ./wavfcdata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              task getiniLeg
                trigger ../inidata eq complete and ../wavini eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit FCNPES '32'
                edit THREADS '4'
                inlimit /emc_36r1/limits:pesh 6
                meter step -1 240 240
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../fc eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:dp
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        family legBcpl
          trigger ./legA eq complete
          family fc
            edit EPSCOUPLE '1'
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '144'
            edit EPSINISTEP_PREVIOUS '216'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit TSTEP '2700'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '2'
            family pf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit THREADS '2'
              edit RUNVARFC '1'
              meter step -1 552 552
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/01/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/01 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/02/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/02 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/03/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/03 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/04/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/04 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
            endfamily
            family cf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc0'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger ../../../../legA/fc/cf/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:pesh
              endfamily
              task getvarepsdata
                trigger ../../../legA/fc/cf eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '552'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit SMSTRIES '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:pesh
                meter step -1 552 552
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../../legA/postproc/dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:dp
              family dp
                trigger ./cf eq complete and ./pf eq complete
                edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
                edit OVERLAP '1'
                edit STEPS '216:240:6'
                inlimit /emc_36r1/limits:dp
                family cf
                  edit EMOS_TYPE 'cf'
                  task derived_param
                    edit MEMBER 'off'
                endfamily
                family pf
                  edit EMOS_TYPE 'pf'
                  family 01
                    edit EPSMEMBER '1'
                    edit MEMBER '1'
                    edit SUBFSFAMILY '/pf001'
                    task derived_param
                  endfamily
                  family 02
                    edit EPSMEMBER '2'
                    edit MEMBER '2'
                    edit SUBFSFAMILY '/pf002'
                    task derived_param
                  endfamily
                  family 03
                    edit EPSMEMBER '3'
                    edit MEMBER '3'
                    edit SUBFSFAMILY '/pf003'
                    task derived_param
                  endfamily
                  family 04
                    edit EPSMEMBER '4'
                    edit MEMBER '4'
                    edit SUBFSFAMILY '/pf004'
                    task derived_param
                  endfamily
                endfamily
              endfamily
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                edit EPSFCRES '319'
                task extrafields
                  edit FCLENGTH '768'
                  edit FCLENGTH0 '240'
                task wm_sfc
                  trigger extrafields eq complete
                  edit FCLENGTH '768'
                task wm_ua
                  trigger extrafields eq complete and wm_sfc eq complete
                  edit FCLENGTH '768'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        task save_logs
          trigger m2 eq complete and sv eq complete and legA eq complete and legBcpl eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit SCHOST 'localhost'
      endfamily
      family 11
        trigger ./lock_1701 eq complete and ./rmfdb eq complete
        edit DELTA_YEAR '-11'
        family m2
          edit DELTA_DAY '-2'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '3000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit FCNPES '16'
          edit EPSSVTSTEP '900'
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit CPUTIME '12000'
          edit STACK '350'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
          task svsh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '73'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '74'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnhevo
            trigger svnh eq complete
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit CPUTIME '120000'
          task svshevo
            trigger svsh eq complete
            edit EVOLVE 'yes'
          task svsave
            trigger svshevo eq complete and svnhevo eq complete
          task clean
            trigger svnhevo eq complete and svshevo eq complete and svsave eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
        endfamily
        family sv
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '2000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit NPES '16'
          edit EPSSVTSTEP '900'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit THREADS '8'
          task targets
            late -s +00:03 -c +00:10
            edit EPSTCBB '1'
            label da_info ""
            label dcda_info ""
            event sv
          task svsh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '73'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '74'
            inlimit /emc_36r1/limits:pesh 2
          task subspace
            trigger svnh eq complete and svsh eq complete and ( targets eq complete or targets:sv)
          task sv1
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '1'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '72'
            inlimit /emc_36r1/limits:pesh 2
          task sv2
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '2'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '71'
            inlimit /emc_36r1/limits:pesh 2
          task sv3
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '3'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv4
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '4'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv5
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '5'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv6
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '6'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task getsvevo
            event ecfs
          task getsvs
            trigger getsvevo eq complete and svnh eq complete and svsh eq complete and sv1 eq complete and sv2 eq complete and sv3 eq complete and sv4 eq complete and sv5 eq complete and sv6 eq complete and ../m2 eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTROP '1'
          task svnhevo
            trigger svnh eq complete and ../legA/fc eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit THREADS '4'
            edit STACK '350'
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
          task svshevo
            trigger svsh eq complete and ../legA/fc eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit STACK '450'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
        endfamily
        family legA
          family fc
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '240'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit TSTEP '1200'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '1'
            task getini
              edit LINIENDA '0'
              edit INIDATEENDA '0'
              edit EPSCOUPLE '0'
              edit N_PERT_AN '10'
              edit EPSCOUPL_A '0'
              edit EPSCOUPL_B '0'
              edit EPSCOUPL_C '0'
            task inidata
              trigger getini eq complete
              edit THREADS '8'
            task getae
              event 1 
            task rot
              trigger ../../sv/getsvs eq complete and ( getae:1 or getae eq complete)
              edit EPSTROP '1'
            family pert_ic
              trigger rot eq complete and inidata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTROP '1'
              inlimit /emc_36r1/limits:pertic
              family 00
                task pertinic
                  edit MEMBER '0'
              endfamily
              family 01
                task pertinic
                  edit MEMBER '1'
              endfamily
              family 02
                task pertinic
                  edit MEMBER '2'
              endfamily
              family 03
                task pertinic
                  edit MEMBER '3'
              endfamily
              family 04
                task pertinic
                  edit MEMBER '4'
              endfamily
            endfamily
            task cp_pert
              trigger pert_ic eq complete
              event 1 
            task wavini
              edit FSFAMILY 'mc'
              edit PERSST 'true'
              edit SSTINT '1'
            task wavfcdata
              trigger ./getpersSST eq complete and ../../legA/fc/wavini eq complete
              edit EPSCOUPLE '1'
            task getpersSST
              trigger ./inidata eq complete
              edit EPSCOUPLE '1'
              edit EPS_INI_STREAM 'DA'
              edit EPSFCRES '639'
            family pf
              trigger ./getpersSST eq complete and ./wavfcdata eq complete and wavini eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              meter step -1 240 240
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task getiniLeg
                  trigger ../../pert_ic/01/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task getiniLeg
                  trigger ../../pert_ic/02/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task getiniLeg
                  trigger ../../pert_ic/03/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task getiniLeg
                  trigger ../../pert_ic/04/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
            endfamily
            family cf
              trigger ./pert_ic/00/pertinic eq complete and ./getpersSST eq complete and ./wavfcdata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              task getiniLeg
                trigger ../inidata eq complete and ../wavini eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit FCNPES '32'
                edit THREADS '4'
                inlimit /emc_36r1/limits:pesh 6
                meter step -1 240 240
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../fc eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:dp
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        family legBcpl
          trigger ./legA eq complete
          family fc
            edit EPSCOUPLE '1'
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '144'
            edit EPSINISTEP_PREVIOUS '216'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit TSTEP '2700'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '2'
            family pf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit THREADS '2'
              edit RUNVARFC '1'
              meter step -1 552 552
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/01/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/01 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/02/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/02 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/03/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/03 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/04/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/04 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
            endfamily
            family cf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc0'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger ../../../../legA/fc/cf/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:pesh
              endfamily
              task getvarepsdata
                trigger ../../../legA/fc/cf eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '552'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit SMSTRIES '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:pesh
                meter step -1 552 552
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../../legA/postproc/dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:dp
              family dp
                trigger ./cf eq complete and ./pf eq complete
                edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
                edit OVERLAP '1'
                edit STEPS '216:240:6'
                inlimit /emc_36r1/limits:dp
                family cf
                  edit EMOS_TYPE 'cf'
                  task derived_param
                    edit MEMBER 'off'
                endfamily
                family pf
                  edit EMOS_TYPE 'pf'
                  family 01
                    edit EPSMEMBER '1'
                    edit MEMBER '1'
                    edit SUBFSFAMILY '/pf001'
                    task derived_param
                  endfamily
                  family 02
                    edit EPSMEMBER '2'
                    edit MEMBER '2'
                    edit SUBFSFAMILY '/pf002'
                    task derived_param
                  endfamily
                  family 03
                    edit EPSMEMBER '3'
                    edit MEMBER '3'
                    edit SUBFSFAMILY '/pf003'
                    task derived_param
                  endfamily
                  family 04
                    edit EPSMEMBER '4'
                    edit MEMBER '4'
                    edit SUBFSFAMILY '/pf004'
                    task derived_param
                  endfamily
                endfamily
              endfamily
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                edit EPSFCRES '319'
                task extrafields
                  edit FCLENGTH '768'
                  edit FCLENGTH0 '240'
                task wm_sfc
                  trigger extrafields eq complete
                  edit FCLENGTH '768'
                task wm_ua
                  trigger extrafields eq complete and wm_sfc eq complete
                  edit FCLENGTH '768'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        task save_logs
          trigger m2 eq complete and sv eq complete and legA eq complete and legBcpl eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit SCHOST 'localhost'
      endfamily
      family 10
        trigger ./lock_1701 eq complete and ./rmfdb eq complete
        edit DELTA_YEAR '-10'
        family m2
          edit DELTA_DAY '-2'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '3000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit FCNPES '16'
          edit EPSSVTSTEP '900'
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit CPUTIME '12000'
          edit STACK '350'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
          task svsh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '73'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '74'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnhevo
            trigger svnh eq complete
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit CPUTIME '120000'
          task svshevo
            trigger svsh eq complete
            edit EVOLVE 'yes'
          task svsave
            trigger svshevo eq complete and svnhevo eq complete
          task clean
            trigger svnhevo eq complete and svshevo eq complete and svsave eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
        endfamily
        family sv
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '2000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit NPES '16'
          edit EPSSVTSTEP '900'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit THREADS '8'
          task targets
            late -s +00:03 -c +00:10
            edit EPSTCBB '1'
            label da_info ""
            label dcda_info ""
            event sv
          task svsh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '73'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '74'
            inlimit /emc_36r1/limits:pesh 2
          task subspace
            trigger svnh eq complete and svsh eq complete and ( targets eq complete or targets:sv)
          task sv1
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '1'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '72'
            inlimit /emc_36r1/limits:pesh 2
          task sv2
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '2'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '71'
            inlimit /emc_36r1/limits:pesh 2
          task sv3
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '3'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv4
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '4'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv5
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '5'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv6
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '6'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task getsvevo
            event ecfs
          task getsvs
            trigger getsvevo eq complete and svnh eq complete and svsh eq complete and sv1 eq complete and sv2 eq complete and sv3 eq complete and sv4 eq complete and sv5 eq complete and sv6 eq complete and ../m2 eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTROP '1'
          task svnhevo
            trigger svnh eq complete and ../legA/fc eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit THREADS '4'
            edit STACK '350'
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
          task svshevo
            trigger svsh eq complete and ../legA/fc eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit STACK '450'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
        endfamily
        family legA
          family fc
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '240'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit TSTEP '1200'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '1'
            task getini
              edit LINIENDA '0'
              edit INIDATEENDA '0'
              edit EPSCOUPLE '0'
              edit N_PERT_AN '10'
              edit EPSCOUPL_A '0'
              edit EPSCOUPL_B '0'
              edit EPSCOUPL_C '0'
            task inidata
              trigger getini eq complete
              edit THREADS '8'
            task getae
              event 1 
            task rot
              trigger ../../sv/getsvs eq complete and ( getae:1 or getae eq complete)
              edit EPSTROP '1'
            family pert_ic
              trigger rot eq complete and inidata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTROP '1'
              inlimit /emc_36r1/limits:pertic
              family 00
                task pertinic
                  edit MEMBER '0'
              endfamily
              family 01
                task pertinic
                  edit MEMBER '1'
              endfamily
              family 02
                task pertinic
                  edit MEMBER '2'
              endfamily
              family 03
                task pertinic
                  edit MEMBER '3'
              endfamily
              family 04
                task pertinic
                  edit MEMBER '4'
              endfamily
            endfamily
            task cp_pert
              trigger pert_ic eq complete
              event 1 
            task wavini
              edit FSFAMILY 'mc'
              edit PERSST 'true'
              edit SSTINT '1'
            task wavfcdata
              trigger ./getpersSST eq complete and ../../legA/fc/wavini eq complete
              edit EPSCOUPLE '1'
            task getpersSST
              trigger ./inidata eq complete
              edit EPSCOUPLE '1'
              edit EPS_INI_STREAM 'DA'
              edit EPSFCRES '639'
            family pf
              trigger ./getpersSST eq complete and ./wavfcdata eq complete and wavini eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              meter step -1 240 240
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task getiniLeg
                  trigger ../../pert_ic/01/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task getiniLeg
                  trigger ../../pert_ic/02/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task getiniLeg
                  trigger ../../pert_ic/03/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task getiniLeg
                  trigger ../../pert_ic/04/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
            endfamily
            family cf
              trigger ./pert_ic/00/pertinic eq complete and ./getpersSST eq complete and ./wavfcdata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              task getiniLeg
                trigger ../inidata eq complete and ../wavini eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit FCNPES '32'
                edit THREADS '4'
                inlimit /emc_36r1/limits:pesh 6
                meter step -1 240 240
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../fc eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:dp
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        family legBcpl
          trigger ./legA eq complete
          family fc
            edit EPSCOUPLE '1'
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '144'
            edit EPSINISTEP_PREVIOUS '216'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit TSTEP '2700'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '2'
            family pf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit THREADS '2'
              edit RUNVARFC '1'
              meter step -1 552 552
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/01/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/01 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/02/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/02 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/03/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/03 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/04/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/04 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
            endfamily
            family cf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc0'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger ../../../../legA/fc/cf/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:pesh
              endfamily
              task getvarepsdata
                trigger ../../../legA/fc/cf eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '552'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit SMSTRIES '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:pesh
                meter step -1 552 552
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../../legA/postproc/dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:dp
              family dp
                trigger ./cf eq complete and ./pf eq complete
                edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
                edit OVERLAP '1'
                edit STEPS '216:240:6'
                inlimit /emc_36r1/limits:dp
                family cf
                  edit EMOS_TYPE 'cf'
                  task derived_param
                    edit MEMBER 'off'
                endfamily
                family pf
                  edit EMOS_TYPE 'pf'
                  family 01
                    edit EPSMEMBER '1'
                    edit MEMBER '1'
                    edit SUBFSFAMILY '/pf001'
                    task derived_param
                  endfamily
                  family 02
                    edit EPSMEMBER '2'
                    edit MEMBER '2'
                    edit SUBFSFAMILY '/pf002'
                    task derived_param
                  endfamily
                  family 03
                    edit EPSMEMBER '3'
                    edit MEMBER '3'
                    edit SUBFSFAMILY '/pf003'
                    task derived_param
                  endfamily
                  family 04
                    edit EPSMEMBER '4'
                    edit MEMBER '4'
                    edit SUBFSFAMILY '/pf004'
                    task derived_param
                  endfamily
                endfamily
              endfamily
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                edit EPSFCRES '319'
                task extrafields
                  edit FCLENGTH '768'
                  edit FCLENGTH0 '240'
                task wm_sfc
                  trigger extrafields eq complete
                  edit FCLENGTH '768'
                task wm_ua
                  trigger extrafields eq complete and wm_sfc eq complete
                  edit FCLENGTH '768'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        task save_logs
          trigger m2 eq complete and sv eq complete and legA eq complete and legBcpl eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit SCHOST 'localhost'
      endfamily
      family 09
        trigger ./lock_1701 eq complete and ./rmfdb eq complete
        edit DELTA_YEAR '-09'
        family m2
          edit DELTA_DAY '-2'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '3000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit FCNPES '16'
          edit EPSSVTSTEP '900'
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit CPUTIME '12000'
          edit STACK '350'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
          task svsh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '73'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '74'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnhevo
            trigger svnh eq complete
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit CPUTIME '120000'
          task svshevo
            trigger svsh eq complete
            edit EVOLVE 'yes'
          task svsave
            trigger svshevo eq complete and svnhevo eq complete
          task clean
            trigger svnhevo eq complete and svshevo eq complete and svsave eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
        endfamily
        family sv
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '2000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit NPES '16'
          edit EPSSVTSTEP '900'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit THREADS '8'
          task targets
            late -s +00:03 -c +00:10
            edit EPSTCBB '1'
            label da_info ""
            label dcda_info ""
            event sv
          task svsh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '73'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '74'
            inlimit /emc_36r1/limits:pesh 2
          task subspace
            trigger svnh eq complete and svsh eq complete and ( targets eq complete or targets:sv)
          task sv1
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '1'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '72'
            inlimit /emc_36r1/limits:pesh 2
          task sv2
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '2'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '71'
            inlimit /emc_36r1/limits:pesh 2
          task sv3
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '3'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv4
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '4'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv5
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '5'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv6
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '6'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task getsvevo
            event ecfs
          task getsvs
            trigger getsvevo eq complete and svnh eq complete and svsh eq complete and sv1 eq complete and sv2 eq complete and sv3 eq complete and sv4 eq complete and sv5 eq complete and sv6 eq complete and ../m2 eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTROP '1'
          task svnhevo
            trigger svnh eq complete and ../legA/fc eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit THREADS '4'
            edit STACK '350'
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
          task svshevo
            trigger svsh eq complete and ../legA/fc eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit STACK '450'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
        endfamily
        family legA
          family fc
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '240'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit TSTEP '1200'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '1'
            task getini
              edit LINIENDA '0'
              edit INIDATEENDA '0'
              edit EPSCOUPLE '0'
              edit N_PERT_AN '10'
              edit EPSCOUPL_A '0'
              edit EPSCOUPL_B '0'
              edit EPSCOUPL_C '0'
            task inidata
              trigger getini eq complete
              edit THREADS '8'
            task getae
              event 1 
            task rot
              trigger ../../sv/getsvs eq complete and ( getae:1 or getae eq complete)
              edit EPSTROP '1'
            family pert_ic
              trigger rot eq complete and inidata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTROP '1'
              inlimit /emc_36r1/limits:pertic
              family 00
                task pertinic
                  edit MEMBER '0'
              endfamily
              family 01
                task pertinic
                  edit MEMBER '1'
              endfamily
              family 02
                task pertinic
                  edit MEMBER '2'
              endfamily
              family 03
                task pertinic
                  edit MEMBER '3'
              endfamily
              family 04
                task pertinic
                  edit MEMBER '4'
              endfamily
            endfamily
            task cp_pert
              trigger pert_ic eq complete
              event 1 
            task wavini
              edit FSFAMILY 'mc'
              edit PERSST 'true'
              edit SSTINT '1'
            task wavfcdata
              trigger ./getpersSST eq complete and ../../legA/fc/wavini eq complete
              edit EPSCOUPLE '1'
            task getpersSST
              trigger ./inidata eq complete
              edit EPSCOUPLE '1'
              edit EPS_INI_STREAM 'DA'
              edit EPSFCRES '639'
            family pf
              trigger ./getpersSST eq complete and ./wavfcdata eq complete and wavini eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              meter step -1 240 240
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task getiniLeg
                  trigger ../../pert_ic/01/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task getiniLeg
                  trigger ../../pert_ic/02/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task getiniLeg
                  trigger ../../pert_ic/03/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task getiniLeg
                  trigger ../../pert_ic/04/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
            endfamily
            family cf
              trigger ./pert_ic/00/pertinic eq complete and ./getpersSST eq complete and ./wavfcdata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              task getiniLeg
                trigger ../inidata eq complete and ../wavini eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit FCNPES '32'
                edit THREADS '4'
                inlimit /emc_36r1/limits:pesh 6
                meter step -1 240 240
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../fc eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:dp
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        family legBcpl
          trigger ./legA eq complete
          family fc
            edit EPSCOUPLE '1'
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '144'
            edit EPSINISTEP_PREVIOUS '216'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit TSTEP '2700'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '2'
            family pf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit THREADS '2'
              edit RUNVARFC '1'
              meter step -1 552 552
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/01/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/01 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/02/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/02 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/03/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/03 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/04/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/04 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
            endfamily
            family cf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc0'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger ../../../../legA/fc/cf/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:pesh
              endfamily
              task getvarepsdata
                trigger ../../../legA/fc/cf eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '552'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit SMSTRIES '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:pesh
                meter step -1 552 552
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../../legA/postproc/dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:dp
              family dp
                trigger ./cf eq complete and ./pf eq complete
                edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
                edit OVERLAP '1'
                edit STEPS '216:240:6'
                inlimit /emc_36r1/limits:dp
                family cf
                  edit EMOS_TYPE 'cf'
                  task derived_param
                    edit MEMBER 'off'
                endfamily
                family pf
                  edit EMOS_TYPE 'pf'
                  family 01
                    edit EPSMEMBER '1'
                    edit MEMBER '1'
                    edit SUBFSFAMILY '/pf001'
                    task derived_param
                  endfamily
                  family 02
                    edit EPSMEMBER '2'
                    edit MEMBER '2'
                    edit SUBFSFAMILY '/pf002'
                    task derived_param
                  endfamily
                  family 03
                    edit EPSMEMBER '3'
                    edit MEMBER '3'
                    edit SUBFSFAMILY '/pf003'
                    task derived_param
                  endfamily
                  family 04
                    edit EPSMEMBER '4'
                    edit MEMBER '4'
                    edit SUBFSFAMILY '/pf004'
                    task derived_param
                  endfamily
                endfamily
              endfamily
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                edit EPSFCRES '319'
                task extrafields
                  edit FCLENGTH '768'
                  edit FCLENGTH0 '240'
                task wm_sfc
                  trigger extrafields eq complete
                  edit FCLENGTH '768'
                task wm_ua
                  trigger extrafields eq complete and wm_sfc eq complete
                  edit FCLENGTH '768'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        task save_logs
          trigger m2 eq complete and sv eq complete and legA eq complete and legBcpl eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit SCHOST 'localhost'
      endfamily
      family 08
        trigger ./lock_1701 eq complete and ./rmfdb eq complete
        edit DELTA_YEAR '-08'
        family m2
          edit DELTA_DAY '-2'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '3000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit FCNPES '16'
          edit EPSSVTSTEP '900'
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit CPUTIME '12000'
          edit STACK '350'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
          task svsh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '73'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '74'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnhevo
            trigger svnh eq complete
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit CPUTIME '120000'
          task svshevo
            trigger svsh eq complete
            edit EVOLVE 'yes'
          task svsave
            trigger svshevo eq complete and svnhevo eq complete
          task clean
            trigger svnhevo eq complete and svshevo eq complete and svsave eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
        endfamily
        family sv
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '2000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit NPES '16'
          edit EPSSVTSTEP '900'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit THREADS '8'
          task targets
            late -s +00:03 -c +00:10
            edit EPSTCBB '1'
            label da_info ""
            label dcda_info ""
            event sv
          task svsh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '73'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '74'
            inlimit /emc_36r1/limits:pesh 2
          task subspace
            trigger svnh eq complete and svsh eq complete and ( targets eq complete or targets:sv)
          task sv1
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '1'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '72'
            inlimit /emc_36r1/limits:pesh 2
          task sv2
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '2'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '71'
            inlimit /emc_36r1/limits:pesh 2
          task sv3
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '3'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv4
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '4'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv5
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '5'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv6
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '6'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task getsvevo
            event ecfs
          task getsvs
            trigger getsvevo eq complete and svnh eq complete and svsh eq complete and sv1 eq complete and sv2 eq complete and sv3 eq complete and sv4 eq complete and sv5 eq complete and sv6 eq complete and ../m2 eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTROP '1'
          task svnhevo
            trigger svnh eq complete and ../legA/fc eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit THREADS '4'
            edit STACK '350'
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
          task svshevo
            trigger svsh eq complete and ../legA/fc eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit STACK '450'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
        endfamily
        family legA
          family fc
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '240'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit TSTEP '1200'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '1'
            task getini
              edit LINIENDA '0'
              edit INIDATEENDA '0'
              edit EPSCOUPLE '0'
              edit N_PERT_AN '10'
              edit EPSCOUPL_A '0'
              edit EPSCOUPL_B '0'
              edit EPSCOUPL_C '0'
            task inidata
              trigger getini eq complete
              edit THREADS '8'
            task getae
              event 1 
            task rot
              trigger ../../sv/getsvs eq complete and ( getae:1 or getae eq complete)
              edit EPSTROP '1'
            family pert_ic
              trigger rot eq complete and inidata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTROP '1'
              inlimit /emc_36r1/limits:pertic
              family 00
                task pertinic
                  edit MEMBER '0'
              endfamily
              family 01
                task pertinic
                  edit MEMBER '1'
              endfamily
              family 02
                task pertinic
                  edit MEMBER '2'
              endfamily
              family 03
                task pertinic
                  edit MEMBER '3'
              endfamily
              family 04
                task pertinic
                  edit MEMBER '4'
              endfamily
            endfamily
            task cp_pert
              trigger pert_ic eq complete
              event 1 
            task wavini
              edit FSFAMILY 'mc'
              edit PERSST 'true'
              edit SSTINT '1'
            task wavfcdata
              trigger ./getpersSST eq complete and ../../legA/fc/wavini eq complete
              edit EPSCOUPLE '1'
            task getpersSST
              trigger ./inidata eq complete
              edit EPSCOUPLE '1'
              edit EPS_INI_STREAM 'DA'
              edit EPSFCRES '639'
            family pf
              trigger ./getpersSST eq complete and ./wavfcdata eq complete and wavini eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              meter step -1 240 240
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task getiniLeg
                  trigger ../../pert_ic/01/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task getiniLeg
                  trigger ../../pert_ic/02/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task getiniLeg
                  trigger ../../pert_ic/03/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task getiniLeg
                  trigger ../../pert_ic/04/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
            endfamily
            family cf
              trigger ./pert_ic/00/pertinic eq complete and ./getpersSST eq complete and ./wavfcdata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              task getiniLeg
                trigger ../inidata eq complete and ../wavini eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit FCNPES '32'
                edit THREADS '4'
                inlimit /emc_36r1/limits:pesh 6
                meter step -1 240 240
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../fc eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:dp
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        family legBcpl
          trigger ./legA eq complete
          family fc
            edit EPSCOUPLE '1'
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '144'
            edit EPSINISTEP_PREVIOUS '216'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit TSTEP '2700'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '2'
            family pf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit THREADS '2'
              edit RUNVARFC '1'
              meter step -1 552 552
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/01/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/01 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/02/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/02 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/03/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/03 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/04/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/04 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
            endfamily
            family cf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc0'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger ../../../../legA/fc/cf/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:pesh
              endfamily
              task getvarepsdata
                trigger ../../../legA/fc/cf eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '552'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit SMSTRIES '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:pesh
                meter step -1 552 552
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../../legA/postproc/dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:dp
              family dp
                trigger ./cf eq complete and ./pf eq complete
                edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
                edit OVERLAP '1'
                edit STEPS '216:240:6'
                inlimit /emc_36r1/limits:dp
                family cf
                  edit EMOS_TYPE 'cf'
                  task derived_param
                    edit MEMBER 'off'
                endfamily
                family pf
                  edit EMOS_TYPE 'pf'
                  family 01
                    edit EPSMEMBER '1'
                    edit MEMBER '1'
                    edit SUBFSFAMILY '/pf001'
                    task derived_param
                  endfamily
                  family 02
                    edit EPSMEMBER '2'
                    edit MEMBER '2'
                    edit SUBFSFAMILY '/pf002'
                    task derived_param
                  endfamily
                  family 03
                    edit EPSMEMBER '3'
                    edit MEMBER '3'
                    edit SUBFSFAMILY '/pf003'
                    task derived_param
                  endfamily
                  family 04
                    edit EPSMEMBER '4'
                    edit MEMBER '4'
                    edit SUBFSFAMILY '/pf004'
                    task derived_param
                  endfamily
                endfamily
              endfamily
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                edit EPSFCRES '319'
                task extrafields
                  edit FCLENGTH '768'
                  edit FCLENGTH0 '240'
                task wm_sfc
                  trigger extrafields eq complete
                  edit FCLENGTH '768'
                task wm_ua
                  trigger extrafields eq complete and wm_sfc eq complete
                  edit FCLENGTH '768'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        task save_logs
          trigger m2 eq complete and sv eq complete and legA eq complete and legBcpl eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit SCHOST 'localhost'
      endfamily
      family 07
        trigger ./lock_1701 eq complete and ./rmfdb eq complete
        edit DELTA_YEAR '-07'
        family m2
          edit DELTA_DAY '-2'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '3000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit FCNPES '16'
          edit EPSSVTSTEP '900'
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit CPUTIME '12000'
          edit STACK '350'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
          task svsh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '73'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '74'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnhevo
            trigger svnh eq complete
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit CPUTIME '120000'
          task svshevo
            trigger svsh eq complete
            edit EVOLVE 'yes'
          task svsave
            trigger svshevo eq complete and svnhevo eq complete
          task clean
            trigger svnhevo eq complete and svshevo eq complete and svsave eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
        endfamily
        family sv
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '2000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit NPES '16'
          edit EPSSVTSTEP '900'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit THREADS '8'
          task targets
            late -s +00:03 -c +00:10
            edit EPSTCBB '1'
            label da_info ""
            label dcda_info ""
            event sv
          task svsh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '73'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '74'
            inlimit /emc_36r1/limits:pesh 2
          task subspace
            trigger svnh eq complete and svsh eq complete and ( targets eq complete or targets:sv)
          task sv1
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '1'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '72'
            inlimit /emc_36r1/limits:pesh 2
          task sv2
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '2'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '71'
            inlimit /emc_36r1/limits:pesh 2
          task sv3
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '3'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv4
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '4'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv5
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '5'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv6
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '6'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task getsvevo
            event ecfs
          task getsvs
            trigger getsvevo eq complete and svnh eq complete and svsh eq complete and sv1 eq complete and sv2 eq complete and sv3 eq complete and sv4 eq complete and sv5 eq complete and sv6 eq complete and ../m2 eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTROP '1'
          task svnhevo
            trigger svnh eq complete and ../legA/fc eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit THREADS '4'
            edit STACK '350'
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
          task svshevo
            trigger svsh eq complete and ../legA/fc eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit STACK '450'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
        endfamily
        family legA
          family fc
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '240'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit TSTEP '1200'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '1'
            task getini
              edit LINIENDA '0'
              edit INIDATEENDA '0'
              edit EPSCOUPLE '0'
              edit N_PERT_AN '10'
              edit EPSCOUPL_A '0'
              edit EPSCOUPL_B '0'
              edit EPSCOUPL_C '0'
            task inidata
              trigger getini eq complete
              edit THREADS '8'
            task getae
              event 1 
            task rot
              trigger ../../sv/getsvs eq complete and ( getae:1 or getae eq complete)
              edit EPSTROP '1'
            family pert_ic
              trigger rot eq complete and inidata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTROP '1'
              inlimit /emc_36r1/limits:pertic
              family 00
                task pertinic
                  edit MEMBER '0'
              endfamily
              family 01
                task pertinic
                  edit MEMBER '1'
              endfamily
              family 02
                task pertinic
                  edit MEMBER '2'
              endfamily
              family 03
                task pertinic
                  edit MEMBER '3'
              endfamily
              family 04
                task pertinic
                  edit MEMBER '4'
              endfamily
            endfamily
            task cp_pert
              trigger pert_ic eq complete
              event 1 
            task wavini
              edit FSFAMILY 'mc'
              edit PERSST 'true'
              edit SSTINT '1'
            task wavfcdata
              trigger ./getpersSST eq complete and ../../legA/fc/wavini eq complete
              edit EPSCOUPLE '1'
            task getpersSST
              trigger ./inidata eq complete
              edit EPSCOUPLE '1'
              edit EPS_INI_STREAM 'DA'
              edit EPSFCRES '639'
            family pf
              trigger ./getpersSST eq complete and ./wavfcdata eq complete and wavini eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              meter step -1 240 240
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task getiniLeg
                  trigger ../../pert_ic/01/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task getiniLeg
                  trigger ../../pert_ic/02/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task getiniLeg
                  trigger ../../pert_ic/03/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task getiniLeg
                  trigger ../../pert_ic/04/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
            endfamily
            family cf
              trigger ./pert_ic/00/pertinic eq complete and ./getpersSST eq complete and ./wavfcdata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              task getiniLeg
                trigger ../inidata eq complete and ../wavini eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit FCNPES '32'
                edit THREADS '4'
                inlimit /emc_36r1/limits:pesh 6
                meter step -1 240 240
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../fc eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:dp
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        family legBcpl
          trigger ./legA eq complete
          family fc
            edit EPSCOUPLE '1'
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '144'
            edit EPSINISTEP_PREVIOUS '216'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit TSTEP '2700'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '2'
            family pf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit THREADS '2'
              edit RUNVARFC '1'
              meter step -1 552 552
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/01/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/01 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/02/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/02 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/03/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/03 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/04/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/04 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
            endfamily
            family cf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc0'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger ../../../../legA/fc/cf/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:pesh
              endfamily
              task getvarepsdata
                trigger ../../../legA/fc/cf eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '552'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit SMSTRIES '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:pesh
                meter step -1 552 552
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../../legA/postproc/dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:dp
              family dp
                trigger ./cf eq complete and ./pf eq complete
                edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
                edit OVERLAP '1'
                edit STEPS '216:240:6'
                inlimit /emc_36r1/limits:dp
                family cf
                  edit EMOS_TYPE 'cf'
                  task derived_param
                    edit MEMBER 'off'
                endfamily
                family pf
                  edit EMOS_TYPE 'pf'
                  family 01
                    edit EPSMEMBER '1'
                    edit MEMBER '1'
                    edit SUBFSFAMILY '/pf001'
                    task derived_param
                  endfamily
                  family 02
                    edit EPSMEMBER '2'
                    edit MEMBER '2'
                    edit SUBFSFAMILY '/pf002'
                    task derived_param
                  endfamily
                  family 03
                    edit EPSMEMBER '3'
                    edit MEMBER '3'
                    edit SUBFSFAMILY '/pf003'
                    task derived_param
                  endfamily
                  family 04
                    edit EPSMEMBER '4'
                    edit MEMBER '4'
                    edit SUBFSFAMILY '/pf004'
                    task derived_param
                  endfamily
                endfamily
              endfamily
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                edit EPSFCRES '319'
                task extrafields
                  edit FCLENGTH '768'
                  edit FCLENGTH0 '240'
                task wm_sfc
                  trigger extrafields eq complete
                  edit FCLENGTH '768'
                task wm_ua
                  trigger extrafields eq complete and wm_sfc eq complete
                  edit FCLENGTH '768'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        task save_logs
          trigger m2 eq complete and sv eq complete and legA eq complete and legBcpl eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit SCHOST 'localhost'
      endfamily
      family 06
        trigger ./lock_1701 eq complete and ./rmfdb eq complete
        edit DELTA_YEAR '-06'
        family m2
          edit DELTA_DAY '-2'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '3000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit FCNPES '16'
          edit EPSSVTSTEP '900'
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit CPUTIME '12000'
          edit STACK '350'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
          task svsh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '73'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '74'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnhevo
            trigger svnh eq complete
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit CPUTIME '120000'
          task svshevo
            trigger svsh eq complete
            edit EVOLVE 'yes'
          task svsave
            trigger svshevo eq complete and svnhevo eq complete
          task clean
            trigger svnhevo eq complete and svshevo eq complete and svsave eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
        endfamily
        family sv
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '2000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit NPES '16'
          edit EPSSVTSTEP '900'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit THREADS '8'
          task targets
            late -s +00:03 -c +00:10
            edit EPSTCBB '1'
            label da_info ""
            label dcda_info ""
            event sv
          task svsh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '73'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '74'
            inlimit /emc_36r1/limits:pesh 2
          task subspace
            trigger svnh eq complete and svsh eq complete and ( targets eq complete or targets:sv)
          task sv1
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '1'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '72'
            inlimit /emc_36r1/limits:pesh 2
          task sv2
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '2'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '71'
            inlimit /emc_36r1/limits:pesh 2
          task sv3
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '3'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv4
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '4'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv5
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '5'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv6
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '6'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task getsvevo
            event ecfs
          task getsvs
            trigger getsvevo eq complete and svnh eq complete and svsh eq complete and sv1 eq complete and sv2 eq complete and sv3 eq complete and sv4 eq complete and sv5 eq complete and sv6 eq complete and ../m2 eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTROP '1'
          task svnhevo
            trigger svnh eq complete and ../legA/fc eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit THREADS '4'
            edit STACK '350'
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
          task svshevo
            trigger svsh eq complete and ../legA/fc eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit STACK '450'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
        endfamily
        family legA
          family fc
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '240'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit TSTEP '1200'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '1'
            task getini
              edit LINIENDA '0'
              edit INIDATEENDA '0'
              edit EPSCOUPLE '0'
              edit N_PERT_AN '10'
              edit EPSCOUPL_A '0'
              edit EPSCOUPL_B '0'
              edit EPSCOUPL_C '0'
            task inidata
              trigger getini eq complete
              edit THREADS '8'
            task getae
              event 1 
            task rot
              trigger ../../sv/getsvs eq complete and ( getae:1 or getae eq complete)
              edit EPSTROP '1'
            family pert_ic
              trigger rot eq complete and inidata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTROP '1'
              inlimit /emc_36r1/limits:pertic
              family 00
                task pertinic
                  edit MEMBER '0'
              endfamily
              family 01
                task pertinic
                  edit MEMBER '1'
              endfamily
              family 02
                task pertinic
                  edit MEMBER '2'
              endfamily
              family 03
                task pertinic
                  edit MEMBER '3'
              endfamily
              family 04
                task pertinic
                  edit MEMBER '4'
              endfamily
            endfamily
            task cp_pert
              trigger pert_ic eq complete
              event 1 
            task wavini
              edit FSFAMILY 'mc'
              edit PERSST 'true'
              edit SSTINT '1'
            task wavfcdata
              trigger ./getpersSST eq complete and ../../legA/fc/wavini eq complete
              edit EPSCOUPLE '1'
            task getpersSST
              trigger ./inidata eq complete
              edit EPSCOUPLE '1'
              edit EPS_INI_STREAM 'DA'
              edit EPSFCRES '639'
            family pf
              trigger ./getpersSST eq complete and ./wavfcdata eq complete and wavini eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              meter step -1 240 240
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task getiniLeg
                  trigger ../../pert_ic/01/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task getiniLeg
                  trigger ../../pert_ic/02/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task getiniLeg
                  trigger ../../pert_ic/03/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task getiniLeg
                  trigger ../../pert_ic/04/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
            endfamily
            family cf
              trigger ./pert_ic/00/pertinic eq complete and ./getpersSST eq complete and ./wavfcdata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              task getiniLeg
                trigger ../inidata eq complete and ../wavini eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit FCNPES '32'
                edit THREADS '4'
                inlimit /emc_36r1/limits:pesh 6
                meter step -1 240 240
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../fc eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:dp
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        family legBcpl
          trigger ./legA eq complete
          family fc
            edit EPSCOUPLE '1'
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '144'
            edit EPSINISTEP_PREVIOUS '216'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit TSTEP '2700'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '2'
            family pf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit THREADS '2'
              edit RUNVARFC '1'
              meter step -1 552 552
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/01/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/01 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/02/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/02 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/03/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/03 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/04/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/04 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
            endfamily
            family cf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc0'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger ../../../../legA/fc/cf/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:pesh
              endfamily
              task getvarepsdata
                trigger ../../../legA/fc/cf eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '552'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit SMSTRIES '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:pesh
                meter step -1 552 552
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../../legA/postproc/dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:dp
              family dp
                trigger ./cf eq complete and ./pf eq complete
                edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
                edit OVERLAP '1'
                edit STEPS '216:240:6'
                inlimit /emc_36r1/limits:dp
                family cf
                  edit EMOS_TYPE 'cf'
                  task derived_param
                    edit MEMBER 'off'
                endfamily
                family pf
                  edit EMOS_TYPE 'pf'
                  family 01
                    edit EPSMEMBER '1'
                    edit MEMBER '1'
                    edit SUBFSFAMILY '/pf001'
                    task derived_param
                  endfamily
                  family 02
                    edit EPSMEMBER '2'
                    edit MEMBER '2'
                    edit SUBFSFAMILY '/pf002'
                    task derived_param
                  endfamily
                  family 03
                    edit EPSMEMBER '3'
                    edit MEMBER '3'
                    edit SUBFSFAMILY '/pf003'
                    task derived_param
                  endfamily
                  family 04
                    edit EPSMEMBER '4'
                    edit MEMBER '4'
                    edit SUBFSFAMILY '/pf004'
                    task derived_param
                  endfamily
                endfamily
              endfamily
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                edit EPSFCRES '319'
                task extrafields
                  edit FCLENGTH '768'
                  edit FCLENGTH0 '240'
                task wm_sfc
                  trigger extrafields eq complete
                  edit FCLENGTH '768'
                task wm_ua
                  trigger extrafields eq complete and wm_sfc eq complete
                  edit FCLENGTH '768'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        task save_logs
          trigger m2 eq complete and sv eq complete and legA eq complete and legBcpl eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit SCHOST 'localhost'
      endfamily
      family 05
        trigger ./lock_1701 eq complete and ./rmfdb eq complete
        edit DELTA_YEAR '-05'
        family m2
          edit DELTA_DAY '-2'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '3000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit FCNPES '16'
          edit EPSSVTSTEP '900'
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit CPUTIME '12000'
          edit STACK '350'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
          task svsh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '73'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '74'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnhevo
            trigger svnh eq complete
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit CPUTIME '120000'
          task svshevo
            trigger svsh eq complete
            edit EVOLVE 'yes'
          task svsave
            trigger svshevo eq complete and svnhevo eq complete
          task clean
            trigger svnhevo eq complete and svshevo eq complete and svsave eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
        endfamily
        family sv
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '2000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit NPES '16'
          edit EPSSVTSTEP '900'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit THREADS '8'
          task targets
            late -s +00:03 -c +00:10
            edit EPSTCBB '1'
            label da_info ""
            label dcda_info ""
            event sv
          task svsh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '73'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '74'
            inlimit /emc_36r1/limits:pesh 2
          task subspace
            trigger svnh eq complete and svsh eq complete and ( targets eq complete or targets:sv)
          task sv1
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '1'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '72'
            inlimit /emc_36r1/limits:pesh 2
          task sv2
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '2'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '71'
            inlimit /emc_36r1/limits:pesh 2
          task sv3
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '3'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv4
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '4'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv5
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '5'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv6
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '6'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task getsvevo
            event ecfs
          task getsvs
            trigger getsvevo eq complete and svnh eq complete and svsh eq complete and sv1 eq complete and sv2 eq complete and sv3 eq complete and sv4 eq complete and sv5 eq complete and sv6 eq complete and ../m2 eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTROP '1'
          task svnhevo
            trigger svnh eq complete and ../legA/fc eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit THREADS '4'
            edit STACK '350'
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
          task svshevo
            trigger svsh eq complete and ../legA/fc eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit STACK '450'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
        endfamily
        family legA
          family fc
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '240'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit TSTEP '1200'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '1'
            task getini
              edit LINIENDA '0'
              edit INIDATEENDA '0'
              edit EPSCOUPLE '0'
              edit N_PERT_AN '10'
              edit EPSCOUPL_A '0'
              edit EPSCOUPL_B '0'
              edit EPSCOUPL_C '0'
            task inidata
              trigger getini eq complete
              edit THREADS '8'
            task getae
              event 1 
            task rot
              trigger ../../sv/getsvs eq complete and ( getae:1 or getae eq complete)
              edit EPSTROP '1'
            family pert_ic
              trigger rot eq complete and inidata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTROP '1'
              inlimit /emc_36r1/limits:pertic
              family 00
                task pertinic
                  edit MEMBER '0'
              endfamily
              family 01
                task pertinic
                  edit MEMBER '1'
              endfamily
              family 02
                task pertinic
                  edit MEMBER '2'
              endfamily
              family 03
                task pertinic
                  edit MEMBER '3'
              endfamily
              family 04
                task pertinic
                  edit MEMBER '4'
              endfamily
            endfamily
            task cp_pert
              trigger pert_ic eq complete
              event 1 
            task wavini
              edit FSFAMILY 'mc'
              edit PERSST 'true'
              edit SSTINT '1'
            task wavfcdata
              trigger ./getpersSST eq complete and ../../legA/fc/wavini eq complete
              edit EPSCOUPLE '1'
            task getpersSST
              trigger ./inidata eq complete
              edit EPSCOUPLE '1'
              edit EPS_INI_STREAM 'DA'
              edit EPSFCRES '639'
            family pf
              trigger ./getpersSST eq complete and ./wavfcdata eq complete and wavini eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              meter step -1 240 240
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task getiniLeg
                  trigger ../../pert_ic/01/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task getiniLeg
                  trigger ../../pert_ic/02/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task getiniLeg
                  trigger ../../pert_ic/03/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task getiniLeg
                  trigger ../../pert_ic/04/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
            endfamily
            family cf
              trigger ./pert_ic/00/pertinic eq complete and ./getpersSST eq complete and ./wavfcdata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              task getiniLeg
                trigger ../inidata eq complete and ../wavini eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit FCNPES '32'
                edit THREADS '4'
                inlimit /emc_36r1/limits:pesh 6
                meter step -1 240 240
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../fc eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:dp
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        family legBcpl
          trigger ./legA eq complete
          family fc
            edit EPSCOUPLE '1'
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '144'
            edit EPSINISTEP_PREVIOUS '216'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit TSTEP '2700'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '2'
            family pf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit THREADS '2'
              edit RUNVARFC '1'
              meter step -1 552 552
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/01/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/01 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/02/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/02 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/03/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/03 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/04/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/04 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
            endfamily
            family cf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc0'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger ../../../../legA/fc/cf/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:pesh
              endfamily
              task getvarepsdata
                trigger ../../../legA/fc/cf eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '552'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit SMSTRIES '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:pesh
                meter step -1 552 552
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../../legA/postproc/dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:dp
              family dp
                trigger ./cf eq complete and ./pf eq complete
                edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
                edit OVERLAP '1'
                edit STEPS '216:240:6'
                inlimit /emc_36r1/limits:dp
                family cf
                  edit EMOS_TYPE 'cf'
                  task derived_param
                    edit MEMBER 'off'
                endfamily
                family pf
                  edit EMOS_TYPE 'pf'
                  family 01
                    edit EPSMEMBER '1'
                    edit MEMBER '1'
                    edit SUBFSFAMILY '/pf001'
                    task derived_param
                  endfamily
                  family 02
                    edit EPSMEMBER '2'
                    edit MEMBER '2'
                    edit SUBFSFAMILY '/pf002'
                    task derived_param
                  endfamily
                  family 03
                    edit EPSMEMBER '3'
                    edit MEMBER '3'
                    edit SUBFSFAMILY '/pf003'
                    task derived_param
                  endfamily
                  family 04
                    edit EPSMEMBER '4'
                    edit MEMBER '4'
                    edit SUBFSFAMILY '/pf004'
                    task derived_param
                  endfamily
                endfamily
              endfamily
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                edit EPSFCRES '319'
                task extrafields
                  edit FCLENGTH '768'
                  edit FCLENGTH0 '240'
                task wm_sfc
                  trigger extrafields eq complete
                  edit FCLENGTH '768'
                task wm_ua
                  trigger extrafields eq complete and wm_sfc eq complete
                  edit FCLENGTH '768'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        task save_logs
          trigger m2 eq complete and sv eq complete and legA eq complete and legBcpl eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit SCHOST 'localhost'
      endfamily
      family 04
        trigger ./lock_1701 eq complete and ./rmfdb eq complete
        edit DELTA_YEAR '-04'
        family m2
          edit DELTA_DAY '-2'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '3000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit FCNPES '16'
          edit EPSSVTSTEP '900'
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit CPUTIME '12000'
          edit STACK '350'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
          task svsh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '73'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '74'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnhevo
            trigger svnh eq complete
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit CPUTIME '120000'
          task svshevo
            trigger svsh eq complete
            edit EVOLVE 'yes'
          task svsave
            trigger svshevo eq complete and svnhevo eq complete
          task clean
            trigger svnhevo eq complete and svshevo eq complete and svsave eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
        endfamily
        family sv
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '2000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit NPES '16'
          edit EPSSVTSTEP '900'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit THREADS '8'
          task targets
            late -s +00:03 -c +00:10
            edit EPSTCBB '1'
            label da_info ""
            label dcda_info ""
            event sv
          task svsh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '73'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '74'
            inlimit /emc_36r1/limits:pesh 2
          task subspace
            trigger svnh eq complete and svsh eq complete and ( targets eq complete or targets:sv)
          task sv1
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '1'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '72'
            inlimit /emc_36r1/limits:pesh 2
          task sv2
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '2'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '71'
            inlimit /emc_36r1/limits:pesh 2
          task sv3
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '3'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv4
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '4'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv5
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '5'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv6
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '6'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task getsvevo
            event ecfs
          task getsvs
            trigger getsvevo eq complete and svnh eq complete and svsh eq complete and sv1 eq complete and sv2 eq complete and sv3 eq complete and sv4 eq complete and sv5 eq complete and sv6 eq complete and ../m2 eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTROP '1'
          task svnhevo
            trigger svnh eq complete and ../legA/fc eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit THREADS '4'
            edit STACK '350'
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
          task svshevo
            trigger svsh eq complete and ../legA/fc eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit STACK '450'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
        endfamily
        family legA
          family fc
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '240'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit TSTEP '1200'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '1'
            task getini
              edit LINIENDA '0'
              edit INIDATEENDA '0'
              edit EPSCOUPLE '0'
              edit N_PERT_AN '10'
              edit EPSCOUPL_A '0'
              edit EPSCOUPL_B '0'
              edit EPSCOUPL_C '0'
            task inidata
              trigger getini eq complete
              edit THREADS '8'
            task getae
              event 1 
            task rot
              trigger ../../sv/getsvs eq complete and ( getae:1 or getae eq complete)
              edit EPSTROP '1'
            family pert_ic
              trigger rot eq complete and inidata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTROP '1'
              inlimit /emc_36r1/limits:pertic
              family 00
                task pertinic
                  edit MEMBER '0'
              endfamily
              family 01
                task pertinic
                  edit MEMBER '1'
              endfamily
              family 02
                task pertinic
                  edit MEMBER '2'
              endfamily
              family 03
                task pertinic
                  edit MEMBER '3'
              endfamily
              family 04
                task pertinic
                  edit MEMBER '4'
              endfamily
            endfamily
            task cp_pert
              trigger pert_ic eq complete
              event 1 
            task wavini
              edit FSFAMILY 'mc'
              edit PERSST 'true'
              edit SSTINT '1'
            task wavfcdata
              trigger ./getpersSST eq complete and ../../legA/fc/wavini eq complete
              edit EPSCOUPLE '1'
            task getpersSST
              trigger ./inidata eq complete
              edit EPSCOUPLE '1'
              edit EPS_INI_STREAM 'DA'
              edit EPSFCRES '639'
            family pf
              trigger ./getpersSST eq complete and ./wavfcdata eq complete and wavini eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              meter step -1 240 240
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task getiniLeg
                  trigger ../../pert_ic/01/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task getiniLeg
                  trigger ../../pert_ic/02/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task getiniLeg
                  trigger ../../pert_ic/03/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task getiniLeg
                  trigger ../../pert_ic/04/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
            endfamily
            family cf
              trigger ./pert_ic/00/pertinic eq complete and ./getpersSST eq complete and ./wavfcdata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              task getiniLeg
                trigger ../inidata eq complete and ../wavini eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit FCNPES '32'
                edit THREADS '4'
                inlimit /emc_36r1/limits:pesh 6
                meter step -1 240 240
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../fc eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:dp
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        family legBcpl
          trigger ./legA eq complete
          family fc
            edit EPSCOUPLE '1'
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '144'
            edit EPSINISTEP_PREVIOUS '216'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit TSTEP '2700'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '2'
            family pf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit THREADS '2'
              edit RUNVARFC '1'
              meter step -1 552 552
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/01/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/01 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/02/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/02 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/03/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/03 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/04/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/04 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
            endfamily
            family cf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc0'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger ../../../../legA/fc/cf/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:pesh
              endfamily
              task getvarepsdata
                trigger ../../../legA/fc/cf eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '552'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit SMSTRIES '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:pesh
                meter step -1 552 552
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../../legA/postproc/dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:dp
              family dp
                trigger ./cf eq complete and ./pf eq complete
                edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
                edit OVERLAP '1'
                edit STEPS '216:240:6'
                inlimit /emc_36r1/limits:dp
                family cf
                  edit EMOS_TYPE 'cf'
                  task derived_param
                    edit MEMBER 'off'
                endfamily
                family pf
                  edit EMOS_TYPE 'pf'
                  family 01
                    edit EPSMEMBER '1'
                    edit MEMBER '1'
                    edit SUBFSFAMILY '/pf001'
                    task derived_param
                  endfamily
                  family 02
                    edit EPSMEMBER '2'
                    edit MEMBER '2'
                    edit SUBFSFAMILY '/pf002'
                    task derived_param
                  endfamily
                  family 03
                    edit EPSMEMBER '3'
                    edit MEMBER '3'
                    edit SUBFSFAMILY '/pf003'
                    task derived_param
                  endfamily
                  family 04
                    edit EPSMEMBER '4'
                    edit MEMBER '4'
                    edit SUBFSFAMILY '/pf004'
                    task derived_param
                  endfamily
                endfamily
              endfamily
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                edit EPSFCRES '319'
                task extrafields
                  edit FCLENGTH '768'
                  edit FCLENGTH0 '240'
                task wm_sfc
                  trigger extrafields eq complete
                  edit FCLENGTH '768'
                task wm_ua
                  trigger extrafields eq complete and wm_sfc eq complete
                  edit FCLENGTH '768'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        task save_logs
          trigger m2 eq complete and sv eq complete and legA eq complete and legBcpl eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit SCHOST 'localhost'
      endfamily
      family 03
        trigger ./lock_1701 eq complete and ./rmfdb eq complete
        edit DELTA_YEAR '-03'
        family m2
          edit DELTA_DAY '-2'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '3000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit FCNPES '16'
          edit EPSSVTSTEP '900'
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit CPUTIME '12000'
          edit STACK '350'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
          task svsh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '73'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '74'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnhevo
            trigger svnh eq complete
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit CPUTIME '120000'
          task svshevo
            trigger svsh eq complete
            edit EVOLVE 'yes'
          task svsave
            trigger svshevo eq complete and svnhevo eq complete
          task clean
            trigger svnhevo eq complete and svshevo eq complete and svsave eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
        endfamily
        family sv
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '2000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit NPES '16'
          edit EPSSVTSTEP '900'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit THREADS '8'
          task targets
            late -s +00:03 -c +00:10
            edit EPSTCBB '1'
            label da_info ""
            label dcda_info ""
            event sv
          task svsh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '73'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '74'
            inlimit /emc_36r1/limits:pesh 2
          task subspace
            trigger svnh eq complete and svsh eq complete and ( targets eq complete or targets:sv)
          task sv1
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '1'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '72'
            inlimit /emc_36r1/limits:pesh 2
          task sv2
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '2'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '71'
            inlimit /emc_36r1/limits:pesh 2
          task sv3
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '3'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv4
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '4'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv5
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '5'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv6
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '6'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task getsvevo
            event ecfs
          task getsvs
            trigger getsvevo eq complete and svnh eq complete and svsh eq complete and sv1 eq complete and sv2 eq complete and sv3 eq complete and sv4 eq complete and sv5 eq complete and sv6 eq complete and ../m2 eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTROP '1'
          task svnhevo
            trigger svnh eq complete and ../legA/fc eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit THREADS '4'
            edit STACK '350'
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
          task svshevo
            trigger svsh eq complete and ../legA/fc eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit STACK '450'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
        endfamily
        family legA
          family fc
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '240'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit TSTEP '1200'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '1'
            task getini
              edit LINIENDA '0'
              edit INIDATEENDA '0'
              edit EPSCOUPLE '0'
              edit N_PERT_AN '10'
              edit EPSCOUPL_A '0'
              edit EPSCOUPL_B '0'
              edit EPSCOUPL_C '0'
            task inidata
              trigger getini eq complete
              edit THREADS '8'
            task getae
              event 1 
            task rot
              trigger ../../sv/getsvs eq complete and ( getae:1 or getae eq complete)
              edit EPSTROP '1'
            family pert_ic
              trigger rot eq complete and inidata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTROP '1'
              inlimit /emc_36r1/limits:pertic
              family 00
                task pertinic
                  edit MEMBER '0'
              endfamily
              family 01
                task pertinic
                  edit MEMBER '1'
              endfamily
              family 02
                task pertinic
                  edit MEMBER '2'
              endfamily
              family 03
                task pertinic
                  edit MEMBER '3'
              endfamily
              family 04
                task pertinic
                  edit MEMBER '4'
              endfamily
            endfamily
            task cp_pert
              trigger pert_ic eq complete
              event 1 
            task wavini
              edit FSFAMILY 'mc'
              edit PERSST 'true'
              edit SSTINT '1'
            task wavfcdata
              trigger ./getpersSST eq complete and ../../legA/fc/wavini eq complete
              edit EPSCOUPLE '1'
            task getpersSST
              trigger ./inidata eq complete
              edit EPSCOUPLE '1'
              edit EPS_INI_STREAM 'DA'
              edit EPSFCRES '639'
            family pf
              trigger ./getpersSST eq complete and ./wavfcdata eq complete and wavini eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              meter step -1 240 240
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task getiniLeg
                  trigger ../../pert_ic/01/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task getiniLeg
                  trigger ../../pert_ic/02/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task getiniLeg
                  trigger ../../pert_ic/03/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task getiniLeg
                  trigger ../../pert_ic/04/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
            endfamily
            family cf
              trigger ./pert_ic/00/pertinic eq complete and ./getpersSST eq complete and ./wavfcdata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              task getiniLeg
                trigger ../inidata eq complete and ../wavini eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit FCNPES '32'
                edit THREADS '4'
                inlimit /emc_36r1/limits:pesh 6
                meter step -1 240 240
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../fc eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:dp
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        family legBcpl
          trigger ./legA eq complete
          family fc
            edit EPSCOUPLE '1'
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '144'
            edit EPSINISTEP_PREVIOUS '216'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit TSTEP '2700'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '2'
            family pf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit THREADS '2'
              edit RUNVARFC '1'
              meter step -1 552 552
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/01/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/01 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/02/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/02 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/03/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/03 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/04/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/04 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
            endfamily
            family cf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc0'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger ../../../../legA/fc/cf/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:pesh
              endfamily
              task getvarepsdata
                trigger ../../../legA/fc/cf eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '552'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit SMSTRIES '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:pesh
                meter step -1 552 552
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../../legA/postproc/dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:dp
              family dp
                trigger ./cf eq complete and ./pf eq complete
                edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
                edit OVERLAP '1'
                edit STEPS '216:240:6'
                inlimit /emc_36r1/limits:dp
                family cf
                  edit EMOS_TYPE 'cf'
                  task derived_param
                    edit MEMBER 'off'
                endfamily
                family pf
                  edit EMOS_TYPE 'pf'
                  family 01
                    edit EPSMEMBER '1'
                    edit MEMBER '1'
                    edit SUBFSFAMILY '/pf001'
                    task derived_param
                  endfamily
                  family 02
                    edit EPSMEMBER '2'
                    edit MEMBER '2'
                    edit SUBFSFAMILY '/pf002'
                    task derived_param
                  endfamily
                  family 03
                    edit EPSMEMBER '3'
                    edit MEMBER '3'
                    edit SUBFSFAMILY '/pf003'
                    task derived_param
                  endfamily
                  family 04
                    edit EPSMEMBER '4'
                    edit MEMBER '4'
                    edit SUBFSFAMILY '/pf004'
                    task derived_param
                  endfamily
                endfamily
              endfamily
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                edit EPSFCRES '319'
                task extrafields
                  edit FCLENGTH '768'
                  edit FCLENGTH0 '240'
                task wm_sfc
                  trigger extrafields eq complete
                  edit FCLENGTH '768'
                task wm_ua
                  trigger extrafields eq complete and wm_sfc eq complete
                  edit FCLENGTH '768'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        task save_logs
          trigger m2 eq complete and sv eq complete and legA eq complete and legBcpl eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit SCHOST 'localhost'
      endfamily
      family 02
        trigger ./lock_1701 eq complete and ./rmfdb eq complete
        edit DELTA_YEAR '-02'
        family m2
          edit DELTA_DAY '-2'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '3000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit FCNPES '16'
          edit EPSSVTSTEP '900'
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit CPUTIME '12000'
          edit STACK '350'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
          task svsh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '73'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '74'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnhevo
            trigger svnh eq complete
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit CPUTIME '120000'
          task svshevo
            trigger svsh eq complete
            edit EVOLVE 'yes'
          task svsave
            trigger svshevo eq complete and svnhevo eq complete
          task clean
            trigger svnhevo eq complete and svshevo eq complete and svsave eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
        endfamily
        family sv
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '2000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit NPES '16'
          edit EPSSVTSTEP '900'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit THREADS '8'
          task targets
            late -s +00:03 -c +00:10
            edit EPSTCBB '1'
            label da_info ""
            label dcda_info ""
            event sv
          task svsh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '73'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '74'
            inlimit /emc_36r1/limits:pesh 2
          task subspace
            trigger svnh eq complete and svsh eq complete and ( targets eq complete or targets:sv)
          task sv1
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '1'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '72'
            inlimit /emc_36r1/limits:pesh 2
          task sv2
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '2'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '71'
            inlimit /emc_36r1/limits:pesh 2
          task sv3
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '3'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv4
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '4'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv5
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '5'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv6
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '6'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task getsvevo
            event ecfs
          task getsvs
            trigger getsvevo eq complete and svnh eq complete and svsh eq complete and sv1 eq complete and sv2 eq complete and sv3 eq complete and sv4 eq complete and sv5 eq complete and sv6 eq complete and ../m2 eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTROP '1'
          task svnhevo
            trigger svnh eq complete and ../legA/fc eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit THREADS '4'
            edit STACK '350'
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
          task svshevo
            trigger svsh eq complete and ../legA/fc eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit STACK '450'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
        endfamily
        family legA
          family fc
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '240'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit TSTEP '1200'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '1'
            task getini
              edit LINIENDA '0'
              edit INIDATEENDA '0'
              edit EPSCOUPLE '0'
              edit N_PERT_AN '10'
              edit EPSCOUPL_A '0'
              edit EPSCOUPL_B '0'
              edit EPSCOUPL_C '0'
            task inidata
              trigger getini eq complete
              edit THREADS '8'
            task getae
              event 1 
            task rot
              trigger ../../sv/getsvs eq complete and ( getae:1 or getae eq complete)
              edit EPSTROP '1'
            family pert_ic
              trigger rot eq complete and inidata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTROP '1'
              inlimit /emc_36r1/limits:pertic
              family 00
                task pertinic
                  edit MEMBER '0'
              endfamily
              family 01
                task pertinic
                  edit MEMBER '1'
              endfamily
              family 02
                task pertinic
                  edit MEMBER '2'
              endfamily
              family 03
                task pertinic
                  edit MEMBER '3'
              endfamily
              family 04
                task pertinic
                  edit MEMBER '4'
              endfamily
            endfamily
            task cp_pert
              trigger pert_ic eq complete
              event 1 
            task wavini
              edit FSFAMILY 'mc'
              edit PERSST 'true'
              edit SSTINT '1'
            task wavfcdata
              trigger ./getpersSST eq complete and ../../legA/fc/wavini eq complete
              edit EPSCOUPLE '1'
            task getpersSST
              trigger ./inidata eq complete
              edit EPSCOUPLE '1'
              edit EPS_INI_STREAM 'DA'
              edit EPSFCRES '639'
            family pf
              trigger ./getpersSST eq complete and ./wavfcdata eq complete and wavini eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              meter step -1 240 240
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task getiniLeg
                  trigger ../../pert_ic/01/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task getiniLeg
                  trigger ../../pert_ic/02/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task getiniLeg
                  trigger ../../pert_ic/03/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task getiniLeg
                  trigger ../../pert_ic/04/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
            endfamily
            family cf
              trigger ./pert_ic/00/pertinic eq complete and ./getpersSST eq complete and ./wavfcdata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              task getiniLeg
                trigger ../inidata eq complete and ../wavini eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit FCNPES '32'
                edit THREADS '4'
                inlimit /emc_36r1/limits:pesh 6
                meter step -1 240 240
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../fc eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:dp
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        family legBcpl
          trigger ./legA eq complete
          family fc
            edit EPSCOUPLE '1'
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '144'
            edit EPSINISTEP_PREVIOUS '216'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit TSTEP '2700'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '2'
            family pf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit THREADS '2'
              edit RUNVARFC '1'
              meter step -1 552 552
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/01/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/01 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/02/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/02 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/03/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/03 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/04/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/04 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
            endfamily
            family cf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc0'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger ../../../../legA/fc/cf/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:pesh
              endfamily
              task getvarepsdata
                trigger ../../../legA/fc/cf eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '552'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit SMSTRIES '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:pesh
                meter step -1 552 552
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../../legA/postproc/dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:dp
              family dp
                trigger ./cf eq complete and ./pf eq complete
                edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
                edit OVERLAP '1'
                edit STEPS '216:240:6'
                inlimit /emc_36r1/limits:dp
                family cf
                  edit EMOS_TYPE 'cf'
                  task derived_param
                    edit MEMBER 'off'
                endfamily
                family pf
                  edit EMOS_TYPE 'pf'
                  family 01
                    edit EPSMEMBER '1'
                    edit MEMBER '1'
                    edit SUBFSFAMILY '/pf001'
                    task derived_param
                  endfamily
                  family 02
                    edit EPSMEMBER '2'
                    edit MEMBER '2'
                    edit SUBFSFAMILY '/pf002'
                    task derived_param
                  endfamily
                  family 03
                    edit EPSMEMBER '3'
                    edit MEMBER '3'
                    edit SUBFSFAMILY '/pf003'
                    task derived_param
                  endfamily
                  family 04
                    edit EPSMEMBER '4'
                    edit MEMBER '4'
                    edit SUBFSFAMILY '/pf004'
                    task derived_param
                  endfamily
                endfamily
              endfamily
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                edit EPSFCRES '319'
                task extrafields
                  edit FCLENGTH '768'
                  edit FCLENGTH0 '240'
                task wm_sfc
                  trigger extrafields eq complete
                  edit FCLENGTH '768'
                task wm_ua
                  trigger extrafields eq complete and wm_sfc eq complete
                  edit FCLENGTH '768'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        task save_logs
          trigger m2 eq complete and sv eq complete and legA eq complete and legBcpl eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit SCHOST 'localhost'
      endfamily
      family 01
        trigger ./lock_1701 eq complete and ./rmfdb eq complete
        edit DELTA_YEAR '-01'
        family m2
          edit DELTA_DAY '-2'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '3000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit FCNPES '16'
          edit EPSSVTSTEP '900'
          edit MEM '1600'
          edit MRF '10'
          edit NPES '16'
          edit THREADS '4'
          edit CPUTIME '12000'
          edit STACK '350'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit CPUTIME '1000'
            edit STACK '450'
          task svsh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '73'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit THREADS '8'
            edit USER_PRIORITY '74'
            edit MEMBER '0'
            inlimit /emc_36r1/limits:pesh 2
          task svnhevo
            trigger svnh eq complete
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit CPUTIME '120000'
          task svshevo
            trigger svsh eq complete
            edit EVOLVE 'yes'
          task svsave
            trigger svshevo eq complete and svnhevo eq complete
          task clean
            trigger svnhevo eq complete and svshevo eq complete and svsave eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc'
        endfamily
        family sv
          edit EVOLVE 'no'
          edit FASTEX 'no'
          edit EPSTYPE 'sv'
          edit FCMEM '2000'
          edit FCSTACK '400'
          edit FCMRFS '100'
          edit FCLENGTH '48'
          edit EPSSVTCSUB '1'
          edit EPSSVOP '1'
          edit NPES '16'
          edit EPSSVTSTEP '900'
          task getini
            edit LINIENDA '0'
            edit INIDATEENDA '0'
            edit EPSCOUPLE '0'
            edit N_PERT_AN '10'
            edit EPSCOUPL_A '0'
            edit EPSCOUPL_B '0'
            edit EPSCOUPL_C '0'
          task inidata_sv
            trigger getini eq complete
            edit THREADS '8'
          task targets
            late -s +00:03 -c +00:10
            edit EPSTCBB '1'
            label da_info ""
            label dcda_info ""
            event sv
          task svsh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '73'
            inlimit /emc_36r1/limits:pesh 2
          task svnh
            trigger inidata_sv eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit THREADS '8'
            edit USER_PRIORITY '74'
            inlimit /emc_36r1/limits:pesh 2
          task subspace
            trigger svnh eq complete and svsh eq complete and ( targets eq complete or targets:sv)
          task sv1
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '1'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '72'
            inlimit /emc_36r1/limits:pesh 2
          task sv2
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '2'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            edit USER_PRIORITY '71'
            inlimit /emc_36r1/limits:pesh 2
          task sv3
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '3'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv4
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '4'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv5
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '5'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task sv6
            trigger ( targets eq complete or targets:sv) and subspace eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit CPUTIME '12000'
            edit STACK '350'
            edit MEMBER '6'
            edit EPSTROP '1'
            edit EPSSVDIAB 'true'
            edit THREADS '8'
            inlimit /emc_36r1/limits:pesh 2
          task getsvevo
            event ecfs
          task getsvs
            trigger getsvevo eq complete and svnh eq complete and svsh eq complete and sv1 eq complete and sv2 eq complete and sv3 eq complete and sv4 eq complete and sv5 eq complete and sv6 eq complete and ../m2 eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit EPSTROP '1'
          task svnhevo
            trigger svnh eq complete and ../legA/fc eq complete
            edit MEM '1600'
            edit MRF '10'
            edit NPES '16'
            edit THREADS '4'
            edit STACK '350'
            edit HEMISPHERE 'n'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
          task svshevo
            trigger svsh eq complete and ../legA/fc eq complete
            edit NPES '24'
            edit THREADS '8'
            edit MEM '1300'
            edit MRF '10'
            edit STACK '450'
            edit EVOLVE 'yes'
            edit ANNPES '8'
            edit CPUTIME '120000'
            inlimit /emc_36r1/limits:pesh 2
        endfamily
        family legA
          family fc
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '240'
            edit EPSINISTEP_PREVIOUS '0'
            edit EPSWAMNFRE '30'
            edit EPSWAMNANG '24'
            edit TSTEP '1200'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '1'
            task getini
              edit LINIENDA '0'
              edit INIDATEENDA '0'
              edit EPSCOUPLE '0'
              edit N_PERT_AN '10'
              edit EPSCOUPL_A '0'
              edit EPSCOUPL_B '0'
              edit EPSCOUPL_C '0'
            task inidata
              trigger getini eq complete
              edit THREADS '8'
            task getae
              event 1 
            task rot
              trigger ../../sv/getsvs eq complete and ( getae:1 or getae eq complete)
              edit EPSTROP '1'
            family pert_ic
              trigger rot eq complete and inidata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTROP '1'
              inlimit /emc_36r1/limits:pertic
              family 00
                task pertinic
                  edit MEMBER '0'
              endfamily
              family 01
                task pertinic
                  edit MEMBER '1'
              endfamily
              family 02
                task pertinic
                  edit MEMBER '2'
              endfamily
              family 03
                task pertinic
                  edit MEMBER '3'
              endfamily
              family 04
                task pertinic
                  edit MEMBER '4'
              endfamily
            endfamily
            task cp_pert
              trigger pert_ic eq complete
              event 1 
            task wavini
              edit FSFAMILY 'mc'
              edit PERSST 'true'
              edit SSTINT '1'
            task wavfcdata
              trigger ./getpersSST eq complete and ../../legA/fc/wavini eq complete
              edit EPSCOUPLE '1'
            task getpersSST
              trigger ./inidata eq complete
              edit EPSCOUPLE '1'
              edit EPS_INI_STREAM 'DA'
              edit EPSFCRES '639'
            family pf
              trigger ./getpersSST eq complete and ./wavfcdata eq complete and wavini eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              meter step -1 240 240
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                task getiniLeg
                  trigger ../../pert_ic/01/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                task getiniLeg
                  trigger ../../pert_ic/02/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                task getiniLeg
                  trigger ../../pert_ic/03/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                task getiniLeg
                  trigger ../../pert_ic/04/pertinic eq complete
                  edit LINIENDA '0'
                  edit INIDATEENDA '0'
                  edit EPSCOUPLE '0'
                  edit N_PERT_AN '10'
                  edit EPSCOUPL_A '0'
                  edit EPSCOUPL_B '0'
                  edit EPSCOUPL_C '0'
                task modeleps
                  trigger getiniLeg eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '30'
                  edit EPSWAMNANG '24'
                  edit EPSLEG '1'
                  edit EPSFCRES '639'
                  edit EPSINISTEP '216'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit FCLENGTH '240'
                  edit TSTEP '1200'
                  edit CLASS 'od'
                  edit THREADS '4'
                  inlimit /emc_36r1/limits:pesh 6
                  meter step -1 240 240
              endfamily
            endfamily
            family cf
              trigger ./pert_ic/00/pertinic eq complete and ./getpersSST eq complete and ./wavfcdata eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSWAMRESOL 'global50'
              edit EPSFCRES '639'
              edit EPSWAMNFRE '30'
              edit EPSWAMNANG '24'
              edit EPSINISTEP '216'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '1'
              edit CLASS 'od'
              task getiniLeg
                trigger ../inidata eq complete and ../wavini eq complete
                edit LINIENDA '0'
                edit INIDATEENDA '0'
                edit EPSCOUPLE '0'
                edit N_PERT_AN '10'
                edit EPSCOUPL_A '0'
                edit EPSCOUPL_B '0'
                edit EPSCOUPL_C '0'
              task modeleps
                trigger getiniLeg eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '30'
                edit EPSWAMNANG '24'
                edit EPSLEG '1'
                edit EPSFCRES '639'
                edit EPSINISTEP '216'
                edit EPSINISTEP_PREVIOUS '0'
                edit FCLENGTH '240'
                edit TSTEP '1200'
                edit CLASS 'od'
                edit FCNPES '32'
                edit THREADS '4'
                inlimit /emc_36r1/limits:pesh 6
                meter step -1 240 240
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../fc eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:dp
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '0:240:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                task extrafields
                  edit EPSFCRES '639'
                  edit FCLENGTH '240'
                  edit FCLENGTH0 '0'
                  edit RUNVARFC '1'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task extrafields
                    edit EPSFCRES '639'
                    edit FCLENGTH '240'
                    edit FCLENGTH0 '0'
                    edit RUNVARFC '1'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        family legBcpl
          trigger ./legA eq complete
          family fc
            edit EPSCOUPLE '1'
            edit FCMEM '2240'
            edit FCSTACK '700'
            edit FCMRFS '0'
            edit PRIORITY '67'
            edit USER_PRIORITY '67'
            edit N_PERT_AN '10'
            edit SSTLEN '144'
            edit EPSINISTEP_PREVIOUS '216'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            edit TSTEP '2700'
            edit EPSFCRES '639'
            edit RESOL '639'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
            edit CPUTIME '1000'
            edit MEM '960'
            edit THREADS '8'
            edit NPES '8'
            edit EPSLEG '2'
            family pf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit EPSTYPE 'pf'
              edit EMOS_TYPE 'pf'
              edit USER_PRIORITY '67'
              edit PRIORITY '67'
              edit FCNPES '32'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit FCLENGTH '240'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit THREADS '2'
              edit RUNVARFC '1'
              meter step -1 552 552
              family 01
                edit EPSMEMBER '1'
                edit MEMBER '1'
                edit SUBFSFAMILY '/pf001'
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/01/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/01 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 02
                edit EPSMEMBER '2'
                edit MEMBER '2'
                edit SUBFSFAMILY '/pf002'
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/02/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/02 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 03
                edit EPSMEMBER '3'
                edit MEMBER '3'
                edit SUBFSFAMILY '/pf003'
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/03/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/03 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
              family 04
                edit EPSMEMBER '4'
                edit MEMBER '4'
                edit SUBFSFAMILY '/pf004'
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit FCLENGTH '144'
                edit FCLENGTH1 '216'
                edit DAILY '0'
                edit RUNVARFC '1'
                family ocean
                  edit RUNACAN '0'
                  edit RUNVARFC '1'
                  edit RUNCOUPLE '0'
                  edit ACAN_LAG '-12'
                  edit EPSLEG '1'
                  task oceini
                    edit STREAM 'ocea'
                  task linkoce
                    trigger oceini eq complete
                  task addforce
                    trigger ../../../../../legA/fc/pf/04/modeleps eq complete
                  task runhope
                    trigger linkoce eq complete and addforce eq complete
                    edit FCLENGTH '216'
                    edit THREADS '16'
                    inlimit /emc_36r1/limits:pesh
                endfamily
                task getvarepsdata
                  trigger ../../../../legA/fc/pf/04 eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '639'
                  edit EPSFCRESN '319'
                  edit CLASS 'od'
                task intHtoL
                  trigger getvarepsdata eq complete
                  edit MEM '1300'
                  edit MRF '10'
                  edit CPUTIME '1000'
                  edit STACK '450'
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSINISTEP_PREVIOUS '0'
                  edit EPSWAMRESOL 'global50'
                  edit TSTEP '2700'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSINISTEP '216'
                  edit EPSFCRES '319'
                  edit NPES '24'
                  edit FCNPES '24'
                  edit THREADS '2'
                  edit CLASS 'od'
                task modeleps
                  trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                  edit EPSNLEGS '3'
                  edit EPSFCLEV '62'
                  edit EPSWAMNSTPW '1'
                  edit EPSWAMRESOL 'global50'
                  edit EPSWAMNFRE '25'
                  edit EPSWAMNANG '12'
                  edit EPSLEG '2'
                  edit EPSFCRES '319'
                  edit EPSINISTEP '360'
                  edit EPSINISTEP_PREVIOUS '216'
                  edit FCLENGTH '552'
                  edit EPSFCLENGTH_A '240'
                  edit MEM '3200'
                  edit MRF '10'
                  edit STACK '350'
                  edit FCCHUNK '1'
                  edit NPES '1'
                  edit THREADS '64'
                  edit FCNPES '1'
                  edit TSTEP '2700'
                  edit CLASS 'od'
                  edit SMSTRIES '1'
                  inlimit /emc_36r1/limits:pesh
                  meter step -1 552 552
              endfamily
            endfamily
            family cf
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit MEMBER '0'
              edit EPSMEMBER '0'
              edit SUBFSFAMILY '/cf000'
              edit EPSTYPE 'cf'
              edit EMOS_TYPE 'cf'
              edit USER_PRIORITY '78'
              edit EPSNLEGS '3'
              edit EPSFCLEV '62'
              edit EPSWAMNSTPW '1'
              edit EPSINISTEP_PREVIOUS '0'
              edit EPSFCRES '319'
              edit EPSWAMRESOL 'global50'
              edit EPSWAMNFRE '25'
              edit EPSWAMNANG '12'
              edit EPSINISTEP '552'
              edit TSTEP '1200'
              edit EPSLEG '2'
              edit CLASS 'od'
              edit OCGROUP 'an0'
              edit FCGROUP 'fc0'
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit FCLENGTH '144'
              edit FCLENGTH1 '216'
              edit DAILY '0'
              edit RUNVARFC '1'
              family ocean
                edit RUNACAN '0'
                edit RUNVARFC '1'
                edit RUNCOUPLE '0'
                edit ACAN_LAG '-12'
                edit EPSLEG '1'
                task oceini
                  edit STREAM 'ocea'
                task linkoce
                  trigger oceini eq complete
                task addforce
                  trigger ../../../../legA/fc/cf/modeleps eq complete
                task runhope
                  trigger linkoce eq complete and addforce eq complete
                  edit FCLENGTH '216'
                  edit THREADS '16'
                  inlimit /emc_36r1/limits:pesh
              endfamily
              task getvarepsdata
                trigger ../../../legA/fc/cf eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSINISTEP '216'
                edit EPSFCRES '639'
                edit EPSFCRESN '319'
                edit CLASS 'od'
              task intHtoL
                trigger getvarepsdata eq complete
                edit MEM '1300'
                edit MRF '10'
                edit CPUTIME '1000'
                edit STACK '450'
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSINISTEP_PREVIOUS '0'
                edit EPSWAMRESOL 'global50'
                edit TSTEP '2700'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSINISTEP '216'
                edit EPSFCRES '319'
                edit NPES '24'
                edit FCNPES '24'
                edit THREADS '2'
                edit CLASS 'od'
              task modeleps
                trigger ./intHtoL eq complete and ./getvarepsdata eq complete and ./ocean eq complete
                edit EPSNLEGS '3'
                edit EPSFCLEV '62'
                edit EPSWAMNSTPW '1'
                edit EPSWAMRESOL 'global50'
                edit EPSWAMNFRE '25'
                edit EPSWAMNANG '12'
                edit EPSLEG '2'
                edit EPSFCRES '319'
                edit EPSINISTEP '360'
                edit EPSINISTEP_PREVIOUS '216'
                edit FCLENGTH '552'
                edit EPSFCLENGTH_A '240'
                edit MEM '3200'
                edit MRF '10'
                edit STACK '350'
                edit FCCHUNK '1'
                edit SMSTRIES '1'
                edit NPES '1'
                edit THREADS '64'
                edit FCNPES '1'
                edit TSTEP '2700'
                edit CLASS 'od'
                inlimit /emc_36r1/limits:pesh
                meter step -1 552 552
            endfamily
          endfamily
          family postproc
            trigger ./fc eq complete
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              trigger ../../legA/postproc/dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:dp
              family dp
                trigger ./cf eq complete and ./pf eq complete
                edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
                edit OVERLAP '1'
                edit STEPS '216:240:6'
                inlimit /emc_36r1/limits:dp
                family cf
                  edit EMOS_TYPE 'cf'
                  task derived_param
                    edit MEMBER 'off'
                endfamily
                family pf
                  edit EMOS_TYPE 'pf'
                  family 01
                    edit EPSMEMBER '1'
                    edit MEMBER '1'
                    edit SUBFSFAMILY '/pf001'
                    task derived_param
                  endfamily
                  family 02
                    edit EPSMEMBER '2'
                    edit MEMBER '2'
                    edit SUBFSFAMILY '/pf002'
                    task derived_param
                  endfamily
                  family 03
                    edit EPSMEMBER '3'
                    edit MEMBER '3'
                    edit SUBFSFAMILY '/pf003'
                    task derived_param
                  endfamily
                  family 04
                    edit EPSMEMBER '4'
                    edit MEMBER '4'
                    edit SUBFSFAMILY '/pf004'
                    task derived_param
                  endfamily
                endfamily
              endfamily
              family cf
                trigger ../../fc/cf eq complete
                edit EMOS_TYPE 'cf'
                task derived_param
                  edit MEMBER 'off'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  trigger ../../../fc/pf/01 eq complete
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  task derived_param
                endfamily
                family 02
                  trigger ../../../fc/pf/02 eq complete
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  task derived_param
                endfamily
                family 03
                  trigger ../../../fc/pf/03 eq complete
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  task derived_param
                endfamily
                family 04
                  trigger ../../../fc/pf/04 eq complete
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  task derived_param
                endfamily
              endfamily
            endfamily
            family extra
              trigger ./dp eq complete
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/fc'
              edit STEPS '246:768:6'
              inlimit /emc_36r1/limits:extra
              family cf
                edit EMOS_TYPE 'cf'
                edit EPSFCRES '319'
                task extrafields
                  edit FCLENGTH '768'
                  edit FCLENGTH0 '240'
                task wm_sfc
                  trigger extrafields eq complete
                  edit FCLENGTH '768'
                task wm_ua
                  trigger extrafields eq complete and wm_sfc eq complete
                  edit FCLENGTH '768'
              endfamily
              family pf
                edit EMOS_TYPE 'pf'
                family 01
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  edit SUBFSFAMILY '/pf001'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 02
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  edit SUBFSFAMILY '/pf002'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 03
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  edit SUBFSFAMILY '/pf003'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
                family 04
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  edit SUBFSFAMILY '/pf004'
                  edit EPSFCRES '319'
                  task extrafields
                    edit FCLENGTH '768'
                    edit FCLENGTH0 '240'
                  task wm_sfc
                    trigger extrafields eq complete
                    edit FCLENGTH '768'
                  task wm_ua
                    trigger extrafields eq complete and wm_sfc eq complete
                    edit FCLENGTH '768'
                endfamily
              endfamily
            endfamily
          endfamily
        endfamily
        task save_logs
          trigger m2 eq complete and sv eq complete and legA eq complete and legBcpl eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit SCHOST 'localhost'
      endfamily
    endfamily
    family arch
      edit EMOS_TYPE 'an'
      edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
      edit FSFAMILY 'mars'
      edit OVERLAP '0'
      edit YMD '20100114'
      limit mutex 1
      inlimit /emc_36r1/limits:mars
      inlimit /limits:mars
      family 18
        trigger ( ../back/18 eq complete)
        edit DELTA_YEAR '-18'
        family legA
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf144
              edit STEPS '0/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/144/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf240
              edit STEPS '150/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf072
              edit STEPS '0/to/072/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/072/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf144
              edit STEPS '078/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf192
              edit STEPS '150/to/192/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf240
              edit STEPS '198/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
              task pfml
                edit STEPS '216'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit RUNVARFC '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family coupled
            trigger ./cf eq complete and ./pf eq complete
            edit EPSLEG '1'
            edit OCLEG '1'
            edit OCEPS '1'
            edit FCCHUNK '1'
            edit FCLENGTH '240'
            edit FCLENGTH0 '0'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '0'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legB
          edit EPSLEG '2'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            task cfml
              edit STEPS '216'
            family cf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task pfml
              edit STEPS '360'
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family coupled
            edit EPSLEG '2'
            edit OCLEG '2'
            edit OCEPS '2'
            edit FCCHUNK '1'
            edit FCLENGTH '360'
            edit FCLENGTH0 '216'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '1'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legC
          edit EPSLEG '3'
          edit FSFAMILY 'mars'
          edit OVERLAP '0'
          edit EMOS_TYPE 'an'
          edit EPSFCRES '255'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit FCCHUNK '1'
            edit FCLENGTH '768'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit RUNVARFC '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              edit EMOS_TYPE 'pf'
              edit STEPS '366/to/768/by/6'
              family cf
                edit EMOS_TYPE 'cf'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family extra
              trigger /emc_36r1/hc2/back/18/legBcpl/postproc/extra eq complete
              edit RUNVARFC '1'
              task wm_sfc_arc
                label files ""
                label info ""
                event fdb
              task wm_ua_arc
                label files ""
                label info ""
                event fdb
            endfamily
            family nino
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit DAILY '0'
              edit RUNVARFC '1'
              edit FCLENGTH1 '360'
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit FCLENGTH '768'
              edit EPSCOUPLE '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              limit semaphore 10
              inlimit /limits:linux_cluster
              inlimit semaphore
              family 00
                edit OCGROUP 'an00'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '0'
                edit MEMBER '0'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 01
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '1'
                edit MEMBER '1'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 02
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '2'
                edit MEMBER '2'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 03
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '3'
                edit MEMBER '3'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 04
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '4'
                edit MEMBER '4'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
            endfamily
            task tcyctoecfs
              trigger nino eq complete
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit EPSCOUPLE '1'
              edit FCLENGTH '768'
              inlimit /limits:linux_cluster
          endfamily
        endfamily
        task hc_clean
          trigger legA eq complete and legB eq complete and legC eq complete
        task save_logs
          trigger legA eq complete and legB eq complete and legC eq complete and hc_clean eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
      endfamily
      family 17
        trigger ( ../back/17 eq complete)
        edit DELTA_YEAR '-17'
        family legA
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf144
              edit STEPS '0/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/144/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf240
              edit STEPS '150/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf072
              edit STEPS '0/to/072/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/072/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf144
              edit STEPS '078/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf192
              edit STEPS '150/to/192/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf240
              edit STEPS '198/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
              task pfml
                edit STEPS '216'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit RUNVARFC '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family coupled
            trigger ./cf eq complete and ./pf eq complete
            edit EPSLEG '1'
            edit OCLEG '1'
            edit OCEPS '1'
            edit FCCHUNK '1'
            edit FCLENGTH '240'
            edit FCLENGTH0 '0'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '0'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legB
          edit EPSLEG '2'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            task cfml
              edit STEPS '216'
            family cf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task pfml
              edit STEPS '360'
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family coupled
            edit EPSLEG '2'
            edit OCLEG '2'
            edit OCEPS '2'
            edit FCCHUNK '1'
            edit FCLENGTH '360'
            edit FCLENGTH0 '216'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '1'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legC
          edit EPSLEG '3'
          edit FSFAMILY 'mars'
          edit OVERLAP '0'
          edit EMOS_TYPE 'an'
          edit EPSFCRES '255'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit FCCHUNK '1'
            edit FCLENGTH '768'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit RUNVARFC '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              edit EMOS_TYPE 'pf'
              edit STEPS '366/to/768/by/6'
              family cf
                edit EMOS_TYPE 'cf'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family extra
              trigger /emc_36r1/hc2/back/17/legBcpl/postproc/extra eq complete
              edit RUNVARFC '1'
              task wm_sfc_arc
                label files ""
                label info ""
                event fdb
              task wm_ua_arc
                label files ""
                label info ""
                event fdb
            endfamily
            family nino
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit DAILY '0'
              edit RUNVARFC '1'
              edit FCLENGTH1 '360'
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit FCLENGTH '768'
              edit EPSCOUPLE '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              limit semaphore 10
              inlimit /limits:linux_cluster
              inlimit semaphore
              family 00
                edit OCGROUP 'an00'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '0'
                edit MEMBER '0'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 01
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '1'
                edit MEMBER '1'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 02
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '2'
                edit MEMBER '2'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 03
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '3'
                edit MEMBER '3'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 04
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '4'
                edit MEMBER '4'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
            endfamily
            task tcyctoecfs
              trigger nino eq complete
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit EPSCOUPLE '1'
              edit FCLENGTH '768'
              inlimit /limits:linux_cluster
          endfamily
        endfamily
        task hc_clean
          trigger legA eq complete and legB eq complete and legC eq complete
        task save_logs
          trigger legA eq complete and legB eq complete and legC eq complete and hc_clean eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
      endfamily
      family 16
        trigger ( ../back/16 eq complete)
        edit DELTA_YEAR '-16'
        family legA
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf144
              edit STEPS '0/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/144/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf240
              edit STEPS '150/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf072
              edit STEPS '0/to/072/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/072/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf144
              edit STEPS '078/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf192
              edit STEPS '150/to/192/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf240
              edit STEPS '198/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
              task pfml
                edit STEPS '216'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit RUNVARFC '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family coupled
            trigger ./cf eq complete and ./pf eq complete
            edit EPSLEG '1'
            edit OCLEG '1'
            edit OCEPS '1'
            edit FCCHUNK '1'
            edit FCLENGTH '240'
            edit FCLENGTH0 '0'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '0'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legB
          edit EPSLEG '2'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            task cfml
              edit STEPS '216'
            family cf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task pfml
              edit STEPS '360'
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family coupled
            edit EPSLEG '2'
            edit OCLEG '2'
            edit OCEPS '2'
            edit FCCHUNK '1'
            edit FCLENGTH '360'
            edit FCLENGTH0 '216'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '1'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legC
          edit EPSLEG '3'
          edit FSFAMILY 'mars'
          edit OVERLAP '0'
          edit EMOS_TYPE 'an'
          edit EPSFCRES '255'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit FCCHUNK '1'
            edit FCLENGTH '768'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit RUNVARFC '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              edit EMOS_TYPE 'pf'
              edit STEPS '366/to/768/by/6'
              family cf
                edit EMOS_TYPE 'cf'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family extra
              trigger /emc_36r1/hc2/back/16/legBcpl/postproc/extra eq complete
              edit RUNVARFC '1'
              task wm_sfc_arc
                label files ""
                label info ""
                event fdb
              task wm_ua_arc
                label files ""
                label info ""
                event fdb
            endfamily
            family nino
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit DAILY '0'
              edit RUNVARFC '1'
              edit FCLENGTH1 '360'
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit FCLENGTH '768'
              edit EPSCOUPLE '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              limit semaphore 10
              inlimit /limits:linux_cluster
              inlimit semaphore
              family 00
                edit OCGROUP 'an00'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '0'
                edit MEMBER '0'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 01
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '1'
                edit MEMBER '1'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 02
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '2'
                edit MEMBER '2'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 03
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '3'
                edit MEMBER '3'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 04
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '4'
                edit MEMBER '4'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
            endfamily
            task tcyctoecfs
              trigger nino eq complete
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit EPSCOUPLE '1'
              edit FCLENGTH '768'
              inlimit /limits:linux_cluster
          endfamily
        endfamily
        task hc_clean
          trigger legA eq complete and legB eq complete and legC eq complete
        task save_logs
          trigger legA eq complete and legB eq complete and legC eq complete and hc_clean eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
      endfamily
      family 15
        trigger ( ../back/15 eq complete)
        edit DELTA_YEAR '-15'
        family legA
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf144
              edit STEPS '0/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/144/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf240
              edit STEPS '150/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf072
              edit STEPS '0/to/072/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/072/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf144
              edit STEPS '078/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf192
              edit STEPS '150/to/192/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf240
              edit STEPS '198/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
              task pfml
                edit STEPS '216'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit RUNVARFC '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family coupled
            trigger ./cf eq complete and ./pf eq complete
            edit EPSLEG '1'
            edit OCLEG '1'
            edit OCEPS '1'
            edit FCCHUNK '1'
            edit FCLENGTH '240'
            edit FCLENGTH0 '0'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '0'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legB
          edit EPSLEG '2'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            task cfml
              edit STEPS '216'
            family cf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task pfml
              edit STEPS '360'
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family coupled
            edit EPSLEG '2'
            edit OCLEG '2'
            edit OCEPS '2'
            edit FCCHUNK '1'
            edit FCLENGTH '360'
            edit FCLENGTH0 '216'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '1'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legC
          edit EPSLEG '3'
          edit FSFAMILY 'mars'
          edit OVERLAP '0'
          edit EMOS_TYPE 'an'
          edit EPSFCRES '255'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit FCCHUNK '1'
            edit FCLENGTH '768'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit RUNVARFC '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              edit EMOS_TYPE 'pf'
              edit STEPS '366/to/768/by/6'
              family cf
                edit EMOS_TYPE 'cf'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family extra
              trigger /emc_36r1/hc2/back/15/legBcpl/postproc/extra eq complete
              edit RUNVARFC '1'
              task wm_sfc_arc
                label files ""
                label info ""
                event fdb
              task wm_ua_arc
                label files ""
                label info ""
                event fdb
            endfamily
            family nino
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit DAILY '0'
              edit RUNVARFC '1'
              edit FCLENGTH1 '360'
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit FCLENGTH '768'
              edit EPSCOUPLE '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              limit semaphore 10
              inlimit /limits:linux_cluster
              inlimit semaphore
              family 00
                edit OCGROUP 'an00'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '0'
                edit MEMBER '0'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 01
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '1'
                edit MEMBER '1'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 02
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '2'
                edit MEMBER '2'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 03
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '3'
                edit MEMBER '3'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 04
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '4'
                edit MEMBER '4'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
            endfamily
            task tcyctoecfs
              trigger nino eq complete
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit EPSCOUPLE '1'
              edit FCLENGTH '768'
              inlimit /limits:linux_cluster
          endfamily
        endfamily
        task hc_clean
          trigger legA eq complete and legB eq complete and legC eq complete
        task save_logs
          trigger legA eq complete and legB eq complete and legC eq complete and hc_clean eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
      endfamily
      family 14
        trigger ( ../back/14 eq complete)
        edit DELTA_YEAR '-14'
        family legA
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf144
              edit STEPS '0/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/144/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf240
              edit STEPS '150/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf072
              edit STEPS '0/to/072/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/072/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf144
              edit STEPS '078/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf192
              edit STEPS '150/to/192/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf240
              edit STEPS '198/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
              task pfml
                edit STEPS '216'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit RUNVARFC '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family coupled
            trigger ./cf eq complete and ./pf eq complete
            edit EPSLEG '1'
            edit OCLEG '1'
            edit OCEPS '1'
            edit FCCHUNK '1'
            edit FCLENGTH '240'
            edit FCLENGTH0 '0'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '0'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legB
          edit EPSLEG '2'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            task cfml
              edit STEPS '216'
            family cf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task pfml
              edit STEPS '360'
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family coupled
            edit EPSLEG '2'
            edit OCLEG '2'
            edit OCEPS '2'
            edit FCCHUNK '1'
            edit FCLENGTH '360'
            edit FCLENGTH0 '216'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '1'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legC
          edit EPSLEG '3'
          edit FSFAMILY 'mars'
          edit OVERLAP '0'
          edit EMOS_TYPE 'an'
          edit EPSFCRES '255'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit FCCHUNK '1'
            edit FCLENGTH '768'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit RUNVARFC '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              edit EMOS_TYPE 'pf'
              edit STEPS '366/to/768/by/6'
              family cf
                edit EMOS_TYPE 'cf'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family extra
              trigger /emc_36r1/hc2/back/14/legBcpl/postproc/extra eq complete
              edit RUNVARFC '1'
              task wm_sfc_arc
                label files ""
                label info ""
                event fdb
              task wm_ua_arc
                label files ""
                label info ""
                event fdb
            endfamily
            family nino
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit DAILY '0'
              edit RUNVARFC '1'
              edit FCLENGTH1 '360'
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit FCLENGTH '768'
              edit EPSCOUPLE '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              limit semaphore 10
              inlimit /limits:linux_cluster
              inlimit semaphore
              family 00
                edit OCGROUP 'an00'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '0'
                edit MEMBER '0'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 01
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '1'
                edit MEMBER '1'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 02
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '2'
                edit MEMBER '2'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 03
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '3'
                edit MEMBER '3'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 04
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '4'
                edit MEMBER '4'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
            endfamily
            task tcyctoecfs
              trigger nino eq complete
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit EPSCOUPLE '1'
              edit FCLENGTH '768'
              inlimit /limits:linux_cluster
          endfamily
        endfamily
        task hc_clean
          trigger legA eq complete and legB eq complete and legC eq complete
        task save_logs
          trigger legA eq complete and legB eq complete and legC eq complete and hc_clean eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
      endfamily
      family 13
        trigger ( ../back/13 eq complete)
        edit DELTA_YEAR '-13'
        family legA
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf144
              edit STEPS '0/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/144/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf240
              edit STEPS '150/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf072
              edit STEPS '0/to/072/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/072/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf144
              edit STEPS '078/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf192
              edit STEPS '150/to/192/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf240
              edit STEPS '198/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
              task pfml
                edit STEPS '216'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit RUNVARFC '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family coupled
            trigger ./cf eq complete and ./pf eq complete
            edit EPSLEG '1'
            edit OCLEG '1'
            edit OCEPS '1'
            edit FCCHUNK '1'
            edit FCLENGTH '240'
            edit FCLENGTH0 '0'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '0'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legB
          edit EPSLEG '2'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            task cfml
              edit STEPS '216'
            family cf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task pfml
              edit STEPS '360'
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family coupled
            edit EPSLEG '2'
            edit OCLEG '2'
            edit OCEPS '2'
            edit FCCHUNK '1'
            edit FCLENGTH '360'
            edit FCLENGTH0 '216'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '1'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legC
          edit EPSLEG '3'
          edit FSFAMILY 'mars'
          edit OVERLAP '0'
          edit EMOS_TYPE 'an'
          edit EPSFCRES '255'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit FCCHUNK '1'
            edit FCLENGTH '768'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit RUNVARFC '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              edit EMOS_TYPE 'pf'
              edit STEPS '366/to/768/by/6'
              family cf
                edit EMOS_TYPE 'cf'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family extra
              trigger /emc_36r1/hc2/back/13/legBcpl/postproc/extra eq complete
              edit RUNVARFC '1'
              task wm_sfc_arc
                label files ""
                label info ""
                event fdb
              task wm_ua_arc
                label files ""
                label info ""
                event fdb
            endfamily
            family nino
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit DAILY '0'
              edit RUNVARFC '1'
              edit FCLENGTH1 '360'
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit FCLENGTH '768'
              edit EPSCOUPLE '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              limit semaphore 10
              inlimit /limits:linux_cluster
              inlimit semaphore
              family 00
                edit OCGROUP 'an00'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '0'
                edit MEMBER '0'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 01
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '1'
                edit MEMBER '1'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 02
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '2'
                edit MEMBER '2'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 03
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '3'
                edit MEMBER '3'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 04
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '4'
                edit MEMBER '4'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
            endfamily
            task tcyctoecfs
              trigger nino eq complete
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit EPSCOUPLE '1'
              edit FCLENGTH '768'
              inlimit /limits:linux_cluster
          endfamily
        endfamily
        task hc_clean
          trigger legA eq complete and legB eq complete and legC eq complete
        task save_logs
          trigger legA eq complete and legB eq complete and legC eq complete and hc_clean eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
      endfamily
      family 12
        trigger ( ../back/12 eq complete)
        edit DELTA_YEAR '-12'
        family legA
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf144
              edit STEPS '0/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/144/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf240
              edit STEPS '150/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf072
              edit STEPS '0/to/072/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/072/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf144
              edit STEPS '078/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf192
              edit STEPS '150/to/192/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf240
              edit STEPS '198/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
              task pfml
                edit STEPS '216'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit RUNVARFC '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family coupled
            trigger ./cf eq complete and ./pf eq complete
            edit EPSLEG '1'
            edit OCLEG '1'
            edit OCEPS '1'
            edit FCCHUNK '1'
            edit FCLENGTH '240'
            edit FCLENGTH0 '0'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '0'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legB
          edit EPSLEG '2'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            task cfml
              edit STEPS '216'
            family cf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task pfml
              edit STEPS '360'
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family coupled
            edit EPSLEG '2'
            edit OCLEG '2'
            edit OCEPS '2'
            edit FCCHUNK '1'
            edit FCLENGTH '360'
            edit FCLENGTH0 '216'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '1'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legC
          edit EPSLEG '3'
          edit FSFAMILY 'mars'
          edit OVERLAP '0'
          edit EMOS_TYPE 'an'
          edit EPSFCRES '255'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit FCCHUNK '1'
            edit FCLENGTH '768'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit RUNVARFC '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              edit EMOS_TYPE 'pf'
              edit STEPS '366/to/768/by/6'
              family cf
                edit EMOS_TYPE 'cf'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family extra
              trigger /emc_36r1/hc2/back/12/legBcpl/postproc/extra eq complete
              edit RUNVARFC '1'
              task wm_sfc_arc
                label files ""
                label info ""
                event fdb
              task wm_ua_arc
                label files ""
                label info ""
                event fdb
            endfamily
            family nino
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit DAILY '0'
              edit RUNVARFC '1'
              edit FCLENGTH1 '360'
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit FCLENGTH '768'
              edit EPSCOUPLE '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              limit semaphore 10
              inlimit /limits:linux_cluster
              inlimit semaphore
              family 00
                edit OCGROUP 'an00'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '0'
                edit MEMBER '0'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 01
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '1'
                edit MEMBER '1'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 02
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '2'
                edit MEMBER '2'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 03
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '3'
                edit MEMBER '3'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 04
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '4'
                edit MEMBER '4'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
            endfamily
            task tcyctoecfs
              trigger nino eq complete
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit EPSCOUPLE '1'
              edit FCLENGTH '768'
              inlimit /limits:linux_cluster
          endfamily
        endfamily
        task hc_clean
          trigger legA eq complete and legB eq complete and legC eq complete
        task save_logs
          trigger legA eq complete and legB eq complete and legC eq complete and hc_clean eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
      endfamily
      family 11
        trigger ( ../back/11 eq complete)
        edit DELTA_YEAR '-11'
        family legA
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf144
              edit STEPS '0/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/144/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf240
              edit STEPS '150/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf072
              edit STEPS '0/to/072/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/072/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf144
              edit STEPS '078/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf192
              edit STEPS '150/to/192/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf240
              edit STEPS '198/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
              task pfml
                edit STEPS '216'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit RUNVARFC '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family coupled
            trigger ./cf eq complete and ./pf eq complete
            edit EPSLEG '1'
            edit OCLEG '1'
            edit OCEPS '1'
            edit FCCHUNK '1'
            edit FCLENGTH '240'
            edit FCLENGTH0 '0'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '0'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legB
          edit EPSLEG '2'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            task cfml
              edit STEPS '216'
            family cf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task pfml
              edit STEPS '360'
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family coupled
            edit EPSLEG '2'
            edit OCLEG '2'
            edit OCEPS '2'
            edit FCCHUNK '1'
            edit FCLENGTH '360'
            edit FCLENGTH0 '216'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '1'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legC
          edit EPSLEG '3'
          edit FSFAMILY 'mars'
          edit OVERLAP '0'
          edit EMOS_TYPE 'an'
          edit EPSFCRES '255'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit FCCHUNK '1'
            edit FCLENGTH '768'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit RUNVARFC '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              edit EMOS_TYPE 'pf'
              edit STEPS '366/to/768/by/6'
              family cf
                edit EMOS_TYPE 'cf'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family extra
              trigger /emc_36r1/hc2/back/11/legBcpl/postproc/extra eq complete
              edit RUNVARFC '1'
              task wm_sfc_arc
                label files ""
                label info ""
                event fdb
              task wm_ua_arc
                label files ""
                label info ""
                event fdb
            endfamily
            family nino
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit DAILY '0'
              edit RUNVARFC '1'
              edit FCLENGTH1 '360'
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit FCLENGTH '768'
              edit EPSCOUPLE '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              limit semaphore 10
              inlimit /limits:linux_cluster
              inlimit semaphore
              family 00
                edit OCGROUP 'an00'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '0'
                edit MEMBER '0'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 01
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '1'
                edit MEMBER '1'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 02
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '2'
                edit MEMBER '2'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 03
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '3'
                edit MEMBER '3'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 04
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '4'
                edit MEMBER '4'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
            endfamily
            task tcyctoecfs
              trigger nino eq complete
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit EPSCOUPLE '1'
              edit FCLENGTH '768'
              inlimit /limits:linux_cluster
          endfamily
        endfamily
        task hc_clean
          trigger legA eq complete and legB eq complete and legC eq complete
        task save_logs
          trigger legA eq complete and legB eq complete and legC eq complete and hc_clean eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
      endfamily
      family 10
        trigger ( ../back/10 eq complete)
        edit DELTA_YEAR '-10'
        family legA
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf144
              edit STEPS '0/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/144/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf240
              edit STEPS '150/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf072
              edit STEPS '0/to/072/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/072/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf144
              edit STEPS '078/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf192
              edit STEPS '150/to/192/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf240
              edit STEPS '198/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
              task pfml
                edit STEPS '216'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit RUNVARFC '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family coupled
            trigger ./cf eq complete and ./pf eq complete
            edit EPSLEG '1'
            edit OCLEG '1'
            edit OCEPS '1'
            edit FCCHUNK '1'
            edit FCLENGTH '240'
            edit FCLENGTH0 '0'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '0'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legB
          edit EPSLEG '2'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            task cfml
              edit STEPS '216'
            family cf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task pfml
              edit STEPS '360'
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family coupled
            edit EPSLEG '2'
            edit OCLEG '2'
            edit OCEPS '2'
            edit FCCHUNK '1'
            edit FCLENGTH '360'
            edit FCLENGTH0 '216'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '1'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legC
          edit EPSLEG '3'
          edit FSFAMILY 'mars'
          edit OVERLAP '0'
          edit EMOS_TYPE 'an'
          edit EPSFCRES '255'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit FCCHUNK '1'
            edit FCLENGTH '768'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit RUNVARFC '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              edit EMOS_TYPE 'pf'
              edit STEPS '366/to/768/by/6'
              family cf
                edit EMOS_TYPE 'cf'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family extra
              trigger /emc_36r1/hc2/back/10/legBcpl/postproc/extra eq complete
              edit RUNVARFC '1'
              task wm_sfc_arc
                label files ""
                label info ""
                event fdb
              task wm_ua_arc
                label files ""
                label info ""
                event fdb
            endfamily
            family nino
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit DAILY '0'
              edit RUNVARFC '1'
              edit FCLENGTH1 '360'
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit FCLENGTH '768'
              edit EPSCOUPLE '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              limit semaphore 10
              inlimit /limits:linux_cluster
              inlimit semaphore
              family 00
                edit OCGROUP 'an00'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '0'
                edit MEMBER '0'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 01
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '1'
                edit MEMBER '1'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 02
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '2'
                edit MEMBER '2'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 03
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '3'
                edit MEMBER '3'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 04
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '4'
                edit MEMBER '4'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
            endfamily
            task tcyctoecfs
              trigger nino eq complete
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit EPSCOUPLE '1'
              edit FCLENGTH '768'
              inlimit /limits:linux_cluster
          endfamily
        endfamily
        task hc_clean
          trigger legA eq complete and legB eq complete and legC eq complete
        task save_logs
          trigger legA eq complete and legB eq complete and legC eq complete and hc_clean eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
      endfamily
      family 09
        trigger ( ../back/09 eq complete)
        edit DELTA_YEAR '-09'
        family legA
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf144
              edit STEPS '0/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/144/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf240
              edit STEPS '150/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf072
              edit STEPS '0/to/072/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/072/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf144
              edit STEPS '078/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf192
              edit STEPS '150/to/192/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf240
              edit STEPS '198/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
              task pfml
                edit STEPS '216'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit RUNVARFC '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family coupled
            trigger ./cf eq complete and ./pf eq complete
            edit EPSLEG '1'
            edit OCLEG '1'
            edit OCEPS '1'
            edit FCCHUNK '1'
            edit FCLENGTH '240'
            edit FCLENGTH0 '0'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '0'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legB
          edit EPSLEG '2'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            task cfml
              edit STEPS '216'
            family cf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task pfml
              edit STEPS '360'
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family coupled
            edit EPSLEG '2'
            edit OCLEG '2'
            edit OCEPS '2'
            edit FCCHUNK '1'
            edit FCLENGTH '360'
            edit FCLENGTH0 '216'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '1'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legC
          edit EPSLEG '3'
          edit FSFAMILY 'mars'
          edit OVERLAP '0'
          edit EMOS_TYPE 'an'
          edit EPSFCRES '255'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit FCCHUNK '1'
            edit FCLENGTH '768'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit RUNVARFC '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              edit EMOS_TYPE 'pf'
              edit STEPS '366/to/768/by/6'
              family cf
                edit EMOS_TYPE 'cf'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family extra
              trigger /emc_36r1/hc2/back/09/legBcpl/postproc/extra eq complete
              edit RUNVARFC '1'
              task wm_sfc_arc
                label files ""
                label info ""
                event fdb
              task wm_ua_arc
                label files ""
                label info ""
                event fdb
            endfamily
            family nino
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit DAILY '0'
              edit RUNVARFC '1'
              edit FCLENGTH1 '360'
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit FCLENGTH '768'
              edit EPSCOUPLE '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              limit semaphore 10
              inlimit /limits:linux_cluster
              inlimit semaphore
              family 00
                edit OCGROUP 'an00'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '0'
                edit MEMBER '0'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 01
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '1'
                edit MEMBER '1'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 02
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '2'
                edit MEMBER '2'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 03
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '3'
                edit MEMBER '3'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 04
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '4'
                edit MEMBER '4'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
            endfamily
            task tcyctoecfs
              trigger nino eq complete
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit EPSCOUPLE '1'
              edit FCLENGTH '768'
              inlimit /limits:linux_cluster
          endfamily
        endfamily
        task hc_clean
          trigger legA eq complete and legB eq complete and legC eq complete
        task save_logs
          trigger legA eq complete and legB eq complete and legC eq complete and hc_clean eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
      endfamily
      family 08
        trigger ( ../back/08 eq complete)
        edit DELTA_YEAR '-08'
        family legA
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf144
              edit STEPS '0/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/144/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf240
              edit STEPS '150/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf072
              edit STEPS '0/to/072/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/072/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf144
              edit STEPS '078/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf192
              edit STEPS '150/to/192/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf240
              edit STEPS '198/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
              task pfml
                edit STEPS '216'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit RUNVARFC '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family coupled
            trigger ./cf eq complete and ./pf eq complete
            edit EPSLEG '1'
            edit OCLEG '1'
            edit OCEPS '1'
            edit FCCHUNK '1'
            edit FCLENGTH '240'
            edit FCLENGTH0 '0'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '0'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legB
          edit EPSLEG '2'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            task cfml
              edit STEPS '216'
            family cf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task pfml
              edit STEPS '360'
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family coupled
            edit EPSLEG '2'
            edit OCLEG '2'
            edit OCEPS '2'
            edit FCCHUNK '1'
            edit FCLENGTH '360'
            edit FCLENGTH0 '216'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '1'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legC
          edit EPSLEG '3'
          edit FSFAMILY 'mars'
          edit OVERLAP '0'
          edit EMOS_TYPE 'an'
          edit EPSFCRES '255'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit FCCHUNK '1'
            edit FCLENGTH '768'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit RUNVARFC '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              edit EMOS_TYPE 'pf'
              edit STEPS '366/to/768/by/6'
              family cf
                edit EMOS_TYPE 'cf'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family extra
              trigger /emc_36r1/hc2/back/08/legBcpl/postproc/extra eq complete
              edit RUNVARFC '1'
              task wm_sfc_arc
                label files ""
                label info ""
                event fdb
              task wm_ua_arc
                label files ""
                label info ""
                event fdb
            endfamily
            family nino
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit DAILY '0'
              edit RUNVARFC '1'
              edit FCLENGTH1 '360'
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit FCLENGTH '768'
              edit EPSCOUPLE '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              limit semaphore 10
              inlimit /limits:linux_cluster
              inlimit semaphore
              family 00
                edit OCGROUP 'an00'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '0'
                edit MEMBER '0'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 01
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '1'
                edit MEMBER '1'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 02
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '2'
                edit MEMBER '2'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 03
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '3'
                edit MEMBER '3'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 04
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '4'
                edit MEMBER '4'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
            endfamily
            task tcyctoecfs
              trigger nino eq complete
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit EPSCOUPLE '1'
              edit FCLENGTH '768'
              inlimit /limits:linux_cluster
          endfamily
        endfamily
        task hc_clean
          trigger legA eq complete and legB eq complete and legC eq complete
        task save_logs
          trigger legA eq complete and legB eq complete and legC eq complete and hc_clean eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
      endfamily
      family 07
        trigger ( ../back/07 eq complete)
        edit DELTA_YEAR '-07'
        family legA
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf144
              edit STEPS '0/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/144/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf240
              edit STEPS '150/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf072
              edit STEPS '0/to/072/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/072/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf144
              edit STEPS '078/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf192
              edit STEPS '150/to/192/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf240
              edit STEPS '198/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
              task pfml
                edit STEPS '216'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit RUNVARFC '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family coupled
            trigger ./cf eq complete and ./pf eq complete
            edit EPSLEG '1'
            edit OCLEG '1'
            edit OCEPS '1'
            edit FCCHUNK '1'
            edit FCLENGTH '240'
            edit FCLENGTH0 '0'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '0'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legB
          edit EPSLEG '2'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            task cfml
              edit STEPS '216'
            family cf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task pfml
              edit STEPS '360'
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family coupled
            edit EPSLEG '2'
            edit OCLEG '2'
            edit OCEPS '2'
            edit FCCHUNK '1'
            edit FCLENGTH '360'
            edit FCLENGTH0 '216'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '1'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legC
          edit EPSLEG '3'
          edit FSFAMILY 'mars'
          edit OVERLAP '0'
          edit EMOS_TYPE 'an'
          edit EPSFCRES '255'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit FCCHUNK '1'
            edit FCLENGTH '768'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit RUNVARFC '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              edit EMOS_TYPE 'pf'
              edit STEPS '366/to/768/by/6'
              family cf
                edit EMOS_TYPE 'cf'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family extra
              trigger /emc_36r1/hc2/back/07/legBcpl/postproc/extra eq complete
              edit RUNVARFC '1'
              task wm_sfc_arc
                label files ""
                label info ""
                event fdb
              task wm_ua_arc
                label files ""
                label info ""
                event fdb
            endfamily
            family nino
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit DAILY '0'
              edit RUNVARFC '1'
              edit FCLENGTH1 '360'
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit FCLENGTH '768'
              edit EPSCOUPLE '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              limit semaphore 10
              inlimit /limits:linux_cluster
              inlimit semaphore
              family 00
                edit OCGROUP 'an00'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '0'
                edit MEMBER '0'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 01
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '1'
                edit MEMBER '1'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 02
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '2'
                edit MEMBER '2'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 03
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '3'
                edit MEMBER '3'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 04
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '4'
                edit MEMBER '4'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
            endfamily
            task tcyctoecfs
              trigger nino eq complete
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit EPSCOUPLE '1'
              edit FCLENGTH '768'
              inlimit /limits:linux_cluster
          endfamily
        endfamily
        task hc_clean
          trigger legA eq complete and legB eq complete and legC eq complete
        task save_logs
          trigger legA eq complete and legB eq complete and legC eq complete and hc_clean eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
      endfamily
      family 06
        trigger ( ../back/06 eq complete)
        edit DELTA_YEAR '-06'
        family legA
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf144
              edit STEPS '0/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/144/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf240
              edit STEPS '150/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf072
              edit STEPS '0/to/072/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/072/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf144
              edit STEPS '078/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf192
              edit STEPS '150/to/192/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf240
              edit STEPS '198/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
              task pfml
                edit STEPS '216'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit RUNVARFC '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family coupled
            trigger ./cf eq complete and ./pf eq complete
            edit EPSLEG '1'
            edit OCLEG '1'
            edit OCEPS '1'
            edit FCCHUNK '1'
            edit FCLENGTH '240'
            edit FCLENGTH0 '0'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '0'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legB
          edit EPSLEG '2'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            task cfml
              edit STEPS '216'
            family cf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task pfml
              edit STEPS '360'
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family coupled
            edit EPSLEG '2'
            edit OCLEG '2'
            edit OCEPS '2'
            edit FCCHUNK '1'
            edit FCLENGTH '360'
            edit FCLENGTH0 '216'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '1'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legC
          edit EPSLEG '3'
          edit FSFAMILY 'mars'
          edit OVERLAP '0'
          edit EMOS_TYPE 'an'
          edit EPSFCRES '255'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit FCCHUNK '1'
            edit FCLENGTH '768'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit RUNVARFC '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              edit EMOS_TYPE 'pf'
              edit STEPS '366/to/768/by/6'
              family cf
                edit EMOS_TYPE 'cf'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family extra
              trigger /emc_36r1/hc2/back/06/legBcpl/postproc/extra eq complete
              edit RUNVARFC '1'
              task wm_sfc_arc
                label files ""
                label info ""
                event fdb
              task wm_ua_arc
                label files ""
                label info ""
                event fdb
            endfamily
            family nino
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit DAILY '0'
              edit RUNVARFC '1'
              edit FCLENGTH1 '360'
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit FCLENGTH '768'
              edit EPSCOUPLE '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              limit semaphore 10
              inlimit /limits:linux_cluster
              inlimit semaphore
              family 00
                edit OCGROUP 'an00'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '0'
                edit MEMBER '0'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 01
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '1'
                edit MEMBER '1'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 02
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '2'
                edit MEMBER '2'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 03
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '3'
                edit MEMBER '3'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 04
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '4'
                edit MEMBER '4'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
            endfamily
            task tcyctoecfs
              trigger nino eq complete
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit EPSCOUPLE '1'
              edit FCLENGTH '768'
              inlimit /limits:linux_cluster
          endfamily
        endfamily
        task hc_clean
          trigger legA eq complete and legB eq complete and legC eq complete
        task save_logs
          trigger legA eq complete and legB eq complete and legC eq complete and hc_clean eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
      endfamily
      family 05
        trigger ( ../back/05 eq complete)
        edit DELTA_YEAR '-05'
        family legA
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf144
              edit STEPS '0/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/144/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf240
              edit STEPS '150/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf072
              edit STEPS '0/to/072/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/072/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf144
              edit STEPS '078/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf192
              edit STEPS '150/to/192/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf240
              edit STEPS '198/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
              task pfml
                edit STEPS '216'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit RUNVARFC '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family coupled
            trigger ./cf eq complete and ./pf eq complete
            edit EPSLEG '1'
            edit OCLEG '1'
            edit OCEPS '1'
            edit FCCHUNK '1'
            edit FCLENGTH '240'
            edit FCLENGTH0 '0'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '0'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legB
          edit EPSLEG '2'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            task cfml
              edit STEPS '216'
            family cf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task pfml
              edit STEPS '360'
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family coupled
            edit EPSLEG '2'
            edit OCLEG '2'
            edit OCEPS '2'
            edit FCCHUNK '1'
            edit FCLENGTH '360'
            edit FCLENGTH0 '216'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '1'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legC
          edit EPSLEG '3'
          edit FSFAMILY 'mars'
          edit OVERLAP '0'
          edit EMOS_TYPE 'an'
          edit EPSFCRES '255'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit FCCHUNK '1'
            edit FCLENGTH '768'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit RUNVARFC '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              edit EMOS_TYPE 'pf'
              edit STEPS '366/to/768/by/6'
              family cf
                edit EMOS_TYPE 'cf'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family extra
              trigger /emc_36r1/hc2/back/05/legBcpl/postproc/extra eq complete
              edit RUNVARFC '1'
              task wm_sfc_arc
                label files ""
                label info ""
                event fdb
              task wm_ua_arc
                label files ""
                label info ""
                event fdb
            endfamily
            family nino
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit DAILY '0'
              edit RUNVARFC '1'
              edit FCLENGTH1 '360'
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit FCLENGTH '768'
              edit EPSCOUPLE '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              limit semaphore 10
              inlimit /limits:linux_cluster
              inlimit semaphore
              family 00
                edit OCGROUP 'an00'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '0'
                edit MEMBER '0'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 01
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '1'
                edit MEMBER '1'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 02
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '2'
                edit MEMBER '2'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 03
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '3'
                edit MEMBER '3'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 04
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '4'
                edit MEMBER '4'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
            endfamily
            task tcyctoecfs
              trigger nino eq complete
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit EPSCOUPLE '1'
              edit FCLENGTH '768'
              inlimit /limits:linux_cluster
          endfamily
        endfamily
        task hc_clean
          trigger legA eq complete and legB eq complete and legC eq complete
        task save_logs
          trigger legA eq complete and legB eq complete and legC eq complete and hc_clean eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
      endfamily
      family 04
        trigger ( ../back/04 eq complete)
        edit DELTA_YEAR '-04'
        family legA
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf144
              edit STEPS '0/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/144/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf240
              edit STEPS '150/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf072
              edit STEPS '0/to/072/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/072/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf144
              edit STEPS '078/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf192
              edit STEPS '150/to/192/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf240
              edit STEPS '198/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
              task pfml
                edit STEPS '216'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit RUNVARFC '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family coupled
            trigger ./cf eq complete and ./pf eq complete
            edit EPSLEG '1'
            edit OCLEG '1'
            edit OCEPS '1'
            edit FCCHUNK '1'
            edit FCLENGTH '240'
            edit FCLENGTH0 '0'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '0'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legB
          edit EPSLEG '2'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            task cfml
              edit STEPS '216'
            family cf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task pfml
              edit STEPS '360'
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family coupled
            edit EPSLEG '2'
            edit OCLEG '2'
            edit OCEPS '2'
            edit FCCHUNK '1'
            edit FCLENGTH '360'
            edit FCLENGTH0 '216'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '1'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legC
          edit EPSLEG '3'
          edit FSFAMILY 'mars'
          edit OVERLAP '0'
          edit EMOS_TYPE 'an'
          edit EPSFCRES '255'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit FCCHUNK '1'
            edit FCLENGTH '768'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit RUNVARFC '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              edit EMOS_TYPE 'pf'
              edit STEPS '366/to/768/by/6'
              family cf
                edit EMOS_TYPE 'cf'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family extra
              trigger /emc_36r1/hc2/back/04/legBcpl/postproc/extra eq complete
              edit RUNVARFC '1'
              task wm_sfc_arc
                label files ""
                label info ""
                event fdb
              task wm_ua_arc
                label files ""
                label info ""
                event fdb
            endfamily
            family nino
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit DAILY '0'
              edit RUNVARFC '1'
              edit FCLENGTH1 '360'
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit FCLENGTH '768'
              edit EPSCOUPLE '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              limit semaphore 10
              inlimit /limits:linux_cluster
              inlimit semaphore
              family 00
                edit OCGROUP 'an00'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '0'
                edit MEMBER '0'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 01
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '1'
                edit MEMBER '1'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 02
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '2'
                edit MEMBER '2'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 03
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '3'
                edit MEMBER '3'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 04
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '4'
                edit MEMBER '4'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
            endfamily
            task tcyctoecfs
              trigger nino eq complete
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit EPSCOUPLE '1'
              edit FCLENGTH '768'
              inlimit /limits:linux_cluster
          endfamily
        endfamily
        task hc_clean
          trigger legA eq complete and legB eq complete and legC eq complete
        task save_logs
          trigger legA eq complete and legB eq complete and legC eq complete and hc_clean eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
      endfamily
      family 03
        trigger ( ../back/03 eq complete)
        edit DELTA_YEAR '-03'
        family legA
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf144
              edit STEPS '0/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/144/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf240
              edit STEPS '150/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf072
              edit STEPS '0/to/072/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/072/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf144
              edit STEPS '078/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf192
              edit STEPS '150/to/192/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf240
              edit STEPS '198/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
              task pfml
                edit STEPS '216'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit RUNVARFC '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family coupled
            trigger ./cf eq complete and ./pf eq complete
            edit EPSLEG '1'
            edit OCLEG '1'
            edit OCEPS '1'
            edit FCCHUNK '1'
            edit FCLENGTH '240'
            edit FCLENGTH0 '0'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '0'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legB
          edit EPSLEG '2'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            task cfml
              edit STEPS '216'
            family cf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task pfml
              edit STEPS '360'
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family coupled
            edit EPSLEG '2'
            edit OCLEG '2'
            edit OCEPS '2'
            edit FCCHUNK '1'
            edit FCLENGTH '360'
            edit FCLENGTH0 '216'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '1'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legC
          edit EPSLEG '3'
          edit FSFAMILY 'mars'
          edit OVERLAP '0'
          edit EMOS_TYPE 'an'
          edit EPSFCRES '255'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit FCCHUNK '1'
            edit FCLENGTH '768'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit RUNVARFC '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              edit EMOS_TYPE 'pf'
              edit STEPS '366/to/768/by/6'
              family cf
                edit EMOS_TYPE 'cf'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family extra
              trigger /emc_36r1/hc2/back/03/legBcpl/postproc/extra eq complete
              edit RUNVARFC '1'
              task wm_sfc_arc
                label files ""
                label info ""
                event fdb
              task wm_ua_arc
                label files ""
                label info ""
                event fdb
            endfamily
            family nino
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit DAILY '0'
              edit RUNVARFC '1'
              edit FCLENGTH1 '360'
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit FCLENGTH '768'
              edit EPSCOUPLE '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              limit semaphore 10
              inlimit /limits:linux_cluster
              inlimit semaphore
              family 00
                edit OCGROUP 'an00'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '0'
                edit MEMBER '0'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 01
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '1'
                edit MEMBER '1'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 02
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '2'
                edit MEMBER '2'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 03
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '3'
                edit MEMBER '3'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 04
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '4'
                edit MEMBER '4'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
            endfamily
            task tcyctoecfs
              trigger nino eq complete
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit EPSCOUPLE '1'
              edit FCLENGTH '768'
              inlimit /limits:linux_cluster
          endfamily
        endfamily
        task hc_clean
          trigger legA eq complete and legB eq complete and legC eq complete
        task save_logs
          trigger legA eq complete and legB eq complete and legC eq complete and hc_clean eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
      endfamily
      family 02
        trigger ( ../back/02 eq complete)
        edit DELTA_YEAR '-02'
        family legA
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf144
              edit STEPS '0/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/144/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf240
              edit STEPS '150/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf072
              edit STEPS '0/to/072/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/072/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf144
              edit STEPS '078/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf192
              edit STEPS '150/to/192/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf240
              edit STEPS '198/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
              task pfml
                edit STEPS '216'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit RUNVARFC '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family coupled
            trigger ./cf eq complete and ./pf eq complete
            edit EPSLEG '1'
            edit OCLEG '1'
            edit OCEPS '1'
            edit FCCHUNK '1'
            edit FCLENGTH '240'
            edit FCLENGTH0 '0'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '0'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legB
          edit EPSLEG '2'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            task cfml
              edit STEPS '216'
            family cf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task pfml
              edit STEPS '360'
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family coupled
            edit EPSLEG '2'
            edit OCLEG '2'
            edit OCEPS '2'
            edit FCCHUNK '1'
            edit FCLENGTH '360'
            edit FCLENGTH0 '216'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '1'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legC
          edit EPSLEG '3'
          edit FSFAMILY 'mars'
          edit OVERLAP '0'
          edit EMOS_TYPE 'an'
          edit EPSFCRES '255'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit FCCHUNK '1'
            edit FCLENGTH '768'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit RUNVARFC '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              edit EMOS_TYPE 'pf'
              edit STEPS '366/to/768/by/6'
              family cf
                edit EMOS_TYPE 'cf'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family extra
              trigger /emc_36r1/hc2/back/02/legBcpl/postproc/extra eq complete
              edit RUNVARFC '1'
              task wm_sfc_arc
                label files ""
                label info ""
                event fdb
              task wm_ua_arc
                label files ""
                label info ""
                event fdb
            endfamily
            family nino
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit DAILY '0'
              edit RUNVARFC '1'
              edit FCLENGTH1 '360'
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit FCLENGTH '768'
              edit EPSCOUPLE '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              limit semaphore 10
              inlimit /limits:linux_cluster
              inlimit semaphore
              family 00
                edit OCGROUP 'an00'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '0'
                edit MEMBER '0'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 01
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '1'
                edit MEMBER '1'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 02
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '2'
                edit MEMBER '2'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 03
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '3'
                edit MEMBER '3'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 04
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '4'
                edit MEMBER '4'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
            endfamily
            task tcyctoecfs
              trigger nino eq complete
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit EPSCOUPLE '1'
              edit FCLENGTH '768'
              inlimit /limits:linux_cluster
          endfamily
        endfamily
        task hc_clean
          trigger legA eq complete and legB eq complete and legC eq complete
        task save_logs
          trigger legA eq complete and legB eq complete and legC eq complete and hc_clean eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
      endfamily
      family 01
        trigger ( ../back/01 eq complete)
        edit DELTA_YEAR '-01'
        family legA
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf144
              edit STEPS '0/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/144/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf240
              edit STEPS '150/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf072
              edit STEPS '0/to/072/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                edit STEPS '006/to/072/by/6'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf144
              edit STEPS '078/to/144/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf192
              edit STEPS '150/to/192/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf240
              edit STEPS '198/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
              task pfml
                edit STEPS '216'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit RUNVARFC '1'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 144
                  edit STEPS '006/to/144/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family coupled
            trigger ./cf eq complete and ./pf eq complete
            edit EPSLEG '1'
            edit OCLEG '1'
            edit OCEPS '1'
            edit FCCHUNK '1'
            edit FCLENGTH '240'
            edit FCLENGTH0 '0'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '0'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legB
          edit EPSLEG '2'
          edit EPSWAMNFRE '25'
          edit EPSWAMNANG '12'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                edit EMOS_TYPE 'pf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 240
                  edit STEPS '216/222/228/234/240'
                  edit OVERLAP '1'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 360
                  edit STEPS '246/to/360/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '240/to/360'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
          endfamily
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            task cfml
              edit STEPS '216'
            family cf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family pf240
              edit STEPS '216/228/240'
              edit OVERLAP '1'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf300
              edit STEPS '246/to/300/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf360
              edit STEPS '306/to/360/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            task pfml
              edit STEPS '360'
            task accum
              edit STEPS '240'
              edit OVERLAP '1'
              label files ""
              label info ""
              event fdb
          endfamily
          family coupled
            edit EPSLEG '2'
            edit OCLEG '2'
            edit OCEPS '2'
            edit FCCHUNK '1'
            edit FCLENGTH '360'
            edit FCLENGTH0 '216'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '1'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
        endfamily
        family legC
          edit EPSLEG '3'
          edit FSFAMILY 'mars'
          edit OVERLAP '0'
          edit EMOS_TYPE 'an'
          edit EPSFCRES '255'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family cf
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            edit EMOS_TYPE 'pf'
            family cf408
              edit STEPS '366/to/408/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf480
              edit STEPS '414/to/480/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf600
              edit STEPS '486/to/600/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf768
              edit STEPS '606/to/768/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'ENWH'
                edit EMOS_STREAM 'ENWH'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family postproc
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            edit FCCHUNK '1'
            edit FCLENGTH '768'
            inlimit /emc_36r1/limits:postproc
            family dp
              edit RUNVARFC '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              edit EMOS_TYPE 'pf'
              edit STEPS '366/to/768/by/6'
              family cf
                edit EMOS_TYPE 'cf'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf
                task derivedp
                  label files ""
                  label info ""
                  event fdb
                task extra_arc
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit STEPS '360/to/768'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family extra
              trigger /emc_36r1/hc2/back/01/legBcpl/postproc/extra eq complete
              edit RUNVARFC '1'
              task wm_sfc_arc
                label files ""
                label info ""
                event fdb
              task wm_ua_arc
                label files ""
                label info ""
                event fdb
            endfamily
            family nino
              edit FCCHUNK '1'
              edit OCTOTAL '5'
              edit FCTOTAL '51'
              edit DAILY '0'
              edit RUNVARFC '1'
              edit FCLENGTH1 '360'
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit FCLENGTH '768'
              edit EPSCOUPLE '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              limit semaphore 10
              inlimit /limits:linux_cluster
              inlimit semaphore
              family 00
                edit OCGROUP 'an00'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '0'
                edit MEMBER '0'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 01
                edit OCGROUP 'an1'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '1'
                edit MEMBER '1'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 02
                edit OCGROUP 'an2'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '2'
                edit MEMBER '2'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 03
                edit OCGROUP 'an3'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '3'
                edit MEMBER '3'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
              family 04
                edit OCGROUP 'an4'
                edit FCGROUP 'fc0'
                edit EPSMEMBER '4'
                edit MEMBER '4'
                task nino_daily
                task tcyc
                  trigger nino_daily eq complete or nino_daily eq aborted
              endfamily
            endfamily
            task tcyctoecfs
              trigger nino eq complete
              edit QUEUE 'emos-e'
              edit QUEUE_EPILOG 'emos-exp'
              edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
              edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
              edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
              edit LOGDIR '/vol/emos_nc/output'
              edit SMSOUT '/vol/emos_nc/output'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit EPSCOUPLE '1'
              edit FCLENGTH '768'
              inlimit /limits:linux_cluster
          endfamily
        endfamily
        task hc_clean
          trigger legA eq complete and legB eq complete and legC eq complete
        task save_logs
          trigger legA eq complete and legB eq complete and legC eq complete and hc_clean eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1'
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
      endfamily
    endfamily
    family efi
      trigger ./back eq complete
      edit MAXDAY '15'
      edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/efi_hdc'
      inlimit /emc_36r1/limits:efih
      family stage
        edit RETRIEVE 'stage'
        edit RF_NAME 'all'
        task prep
      endfamily
      family op
        trigger ./stage eq complete
        family prep
          family stage
            edit QUEUE 'emos-e'
            edit QUEUE_EPILOG 'emos-exp'
            edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
            edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
            edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
            edit LOGDIR '/vol/emos_nc/output'
            edit SMSOUT '/vol/emos_nc/output'
            edit ACC '24'
            edit EFI_STEP1 '00'
            edit EFI_STEP2 '24'
            edit RETRIEVE 'stage'
            inlimit /limits:linux_cluster
            family all
              edit RF_NAME 'all'
              task prep
            endfamily
          endfamily
          family d0105_24
            edit EFI_STEP1 '000 012 024 036 048 060 072 084 096'
            edit EFI_STEP2 '024 036 048 060 072 084 096 108 120'
            edit ACC '24'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family d0610_24
            edit EFI_STEP1 '108 120 132 144 168 192 216'
            edit EFI_STEP2 '132 144 156 168 192 216 240'
            edit ACC '24'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
        endfamily
      endfamily
      family nop
        trigger ./stage eq complete
        family prep
          family stage
            edit QUEUE 'emos-e'
            edit QUEUE_EPILOG 'emos-exp'
            edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
            edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
            edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
            edit LOGDIR '/vol/emos_nc/output'
            edit SMSOUT '/vol/emos_nc/output'
            edit ACC '24'
            edit EFI_STEP1 '00'
            edit EFI_STEP2 '24'
            edit RETRIEVE 'stage'
            inlimit /limits:linux_cluster
            family all
              edit RF_NAME 'all'
              task prep
            endfamily
          endfamily
          family d0105_24
            edit EFI_STEP1 '000 012 024 036 048 060 072 084 096'
            edit EFI_STEP2 '024 036 048 060 072 084 096 108 120'
            edit ACC '24'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
          endfamily
          family d0610_24
            edit EFI_STEP1 '108 120 132 144 168 192 216'
            edit EFI_STEP2 '132 144 156 168 192 216 240'
            edit ACC '24'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
          endfamily
          family d1115_24
            edit EFI_STEP1 '240 264 288 312 336'
            edit EFI_STEP2 '264 288 312 336 360'
            edit ACC '24'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
          endfamily
          family d0102_06
            edit EFI_STEP1 '006 012 018 024 030 036 042 048'
            edit EFI_STEP2 '006 012 018 024 030 036 042 048'
            edit ACC '0'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
          endfamily
          family d0304_06
            edit EFI_STEP1 '054 060 066 072 078 084 090 096'
            edit EFI_STEP2 '054 060 066 072 078 084 090 096'
            edit ACC '0'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
          endfamily
          family d0506_06
            edit EFI_STEP1 '102 108 114 120 126 132 138 144'
            edit EFI_STEP2 '102 108 114 120 126 132 138 144'
            edit ACC '0'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
          endfamily
          family d0708_06
            edit EFI_STEP1 '150 156 162 168 174 180 186 192'
            edit EFI_STEP2 '150 156 162 168 174 180 186 192'
            edit ACC '0'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
          endfamily
          family d0910_06
            edit EFI_STEP1 '198 204 210 216 222 228 234 240'
            edit EFI_STEP2 '198 204 210 216 222 228 234 240'
            edit ACC '0'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
          endfamily
        endfamily
      endfamily
    endfamily
    family adm
      task hindsync
        trigger ../back eq complete and ../arch eq complete
        edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/efi_hdc'
        edit NH_WEEKS_KEPT '3'
        edit BACKUP_HCFDB '0'
      family fri
        complete ./fri/compl eq complete and ./fri/dummy eq queued
        edit SMSFILES '/home/ma/emos/def/emc_36r1/ref'
        task dummy
          cron -w 5 09:00
        task compl
          complete ./dummy eq active or ./dummy eq submitted
          trigger 1 eq 0
      endfamily
      family 1to2
        edit ORIG_FDB_ROOT 'undef'
        edit DEST_FDB_ROOT 'undef'
        edit SMSINCLUDE '/home/ma/emos/def/o/include'
        family chver
          repeat integer DELTA_YEAR -18 -1
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/efi_hdc'
          edit VERSION '9906'
          edit VERSION_TO '9936'
          edit CHVER_FDB_ROOT '/ma_fdb'
          edit GENERAT '1'
          edit CLEANER '0'
          family enfh
            edit CHVER_STREAM 'enfh'
            family 0
              edit CHVER_TYPE ':0000:[0:]:'
              task chver
                label info ""
            endfamily
            family 1
              edit CHVER_TYPE ':0000:1:'
              task chver
                label info ""
            endfamily
            family 2
              edit CHVER_TYPE ':0000:2:'
              task chver
                label info ""
            endfamily
            family 3
              edit CHVER_TYPE ':0000:3:'
              task chver
                label info ""
            endfamily
            family 4
              edit CHVER_TYPE ':0000:4:'
              task chver
                label info ""
            endfamily
          endfamily
          family enwh
            edit CHVER_STREAM 'enwh'
            task chver
              label info ""
          endfamily
          family efho
            edit CHVER_STREAM 'efho'
            task chver
              label info ""
          endfamily
          family ewho
            edit CHVER_STREAM 'ewho'
            task chver
              label info ""
          endfamily
          task save_logs
            trigger enfh eq complete and enwh eq complete and efho eq complete and ewho eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1'
            edit SCHOST 'localhost'
        endfamily
        family arch
          trigger ./chver:DELTA_YEAR gt ./arch:DELTA_YEAR or ./chver eq complete
          repeat integer DELTA_YEAR -18 -1
          edit EMOS_TYPE 'an'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
          edit FSFAMILY 'mars'
          edit OVERLAP '0'
          edit VERSION '9936'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family legA
            family cf
              edit EMOS_TYPE 'cf'
              edit NUMBER '0'
              family cf144
                edit STEPS '0/to/144/by/6'
                task pfpl
                  inlimit /emc_36r1/limits:pl
                  label files ""
                  label info ""
                  event fdb
                task pfsfc
                  trigger ../../postproc/dp eq complete
                  inlimit /emc_36r1/limits:sfc
                  label files ""
                  label info ""
                  event fdb
                task pfil
                  label files ""
                  label info ""
                  event fdb
                task pfwv
                  edit STREAM 'ENWH'
                  edit EMOS_STREAM 'ENWH'
                  edit STEPS '006/to/144/by/6'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf240
                edit STEPS '150/to/240/by/6'
                task pfpl
                  inlimit /emc_36r1/limits:pl
                  label files ""
                  label info ""
                  event fdb
                task pfsfc
                  trigger ../../postproc/dp eq complete
                  inlimit /emc_36r1/limits:sfc
                  label files ""
                  label info ""
                  event fdb
                task pfil
                  label files ""
                  label info ""
                  event fdb
                task pfwv
                  edit STREAM 'ENWH'
                  edit EMOS_STREAM 'ENWH'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family pf
              edit EMOS_TYPE 'pf'
              family pf072
                edit STEPS '0/to/072/by/6'
                task pfpl
                  inlimit /emc_36r1/limits:pl
                  label files ""
                  label info ""
                  event fdb
                task pfsfc
                  trigger ../../postproc/dp eq complete
                  inlimit /emc_36r1/limits:sfc
                  label files ""
                  label info ""
                  event fdb
                task pfil
                  label files ""
                  label info ""
                  event fdb
                task pfwv
                  edit STREAM 'ENWH'
                  edit EMOS_STREAM 'ENWH'
                  edit STEPS '006/to/072/by/6'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf144
                edit STEPS '078/to/144/by/6'
                task pfpl
                  inlimit /emc_36r1/limits:pl
                  label files ""
                  label info ""
                  event fdb
                task pfsfc
                  trigger ../../postproc/dp eq complete
                  inlimit /emc_36r1/limits:sfc
                  label files ""
                  label info ""
                  event fdb
                task pfil
                  label files ""
                  label info ""
                  event fdb
                task pfwv
                  edit STREAM 'ENWH'
                  edit EMOS_STREAM 'ENWH'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf192
                edit STEPS '150/to/192/by/6'
                task pfpl
                  inlimit /emc_36r1/limits:pl
                  label files ""
                  label info ""
                  event fdb
                task pfsfc
                  trigger ../../postproc/dp eq complete
                  inlimit /emc_36r1/limits:sfc
                  label files ""
                  label info ""
                  event fdb
                task pfil
                  label files ""
                  label info ""
                  event fdb
                task pfwv
                  edit STREAM 'ENWH'
                  edit EMOS_STREAM 'ENWH'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf240
                edit STEPS '198/to/240/by/6'
                task pfpl
                  inlimit /emc_36r1/limits:pl
                  label files ""
                  label info ""
                  event fdb
                task pfsfc
                  trigger ../../postproc/dp eq complete
                  inlimit /emc_36r1/limits:sfc
                  label files ""
                  label info ""
                  event fdb
                task pfil
                  label files ""
                  label info ""
                  event fdb
                task pfwv
                  edit STREAM 'ENWH'
                  edit EMOS_STREAM 'ENWH'
                  label files ""
                  label info ""
                  event fdb
                task pfml
                  edit STEPS '216'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family postproc
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit RUNVARFC '1'
              inlimit /emc_36r1/limits:postproc
              family dp
                edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
                family pf
                  edit EMOS_TYPE 'pf'
                  family 144
                    edit STEPS '006/to/144/by/6'
                    task derivedp
                      label files ""
                      label info ""
                      event fdb
                  endfamily
                  family 240
                    edit STEPS '150/to/240/by/6'
                    task derivedp
                      label files ""
                      label info ""
                      event fdb
                  endfamily
                  task extra_arc
                    edit STEPS '0/to/240'
                    edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                    edit RUNVARFC '1'
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family cf
                  edit EMOS_TYPE 'cf'
                  family 144
                    edit STEPS '006/to/144/by/6'
                    task derivedp
                      label files ""
                      label info ""
                      event fdb
                  endfamily
                  family 240
                    edit STEPS '150/to/240/by/6'
                    task derivedp
                      label files ""
                      label info ""
                      event fdb
                  endfamily
                  task extra_arc
                    edit STEPS '0/to/240'
                    edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                    edit RUNVARFC '1'
                    label files ""
                    label info ""
                    event fdb
                endfamily
              endfamily
            endfamily
            family coupled
              trigger ./cf eq complete and ./pf eq complete
              edit EPSLEG '1'
              edit OCLEG '1'
              edit OCEPS '1'
              edit FCCHUNK '1'
              edit FCLENGTH '240'
              edit FCLENGTH0 '0'
              edit OCEPSTOT '3'
              edit FCTOTAL '51'
              edit RUNVARFC '1'
              edit EPSCOUPLE '1'
              edit RUNCOUPLE '0'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              inlimit /emc_36r1/lag/ref/00:mutex
              task preparcoce
              task arcfc
                trigger preparcoce eq complete
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family legB
            edit EPSLEG '2'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            inlimit /emc_36r1/limits:mars
            inlimit /limits:mars
            family postproc
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              inlimit /emc_36r1/limits:postproc
              family dp
                edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
                family pf
                  edit EMOS_TYPE 'pf'
                  family 240
                    edit STEPS '216/222/228/234/240'
                    edit OVERLAP '1'
                    task derivedp
                      label files ""
                      label info ""
                      event fdb
                  endfamily
                  family 360
                    edit STEPS '246/to/360/by/6'
                    task derivedp
                      label files ""
                      label info ""
                      event fdb
                  endfamily
                  task extra_arc
                    edit STEPS '240/to/360'
                    edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                    edit RUNVARFC '1'
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family cf
                  edit EMOS_TYPE 'cf'
                  family 240
                    edit STEPS '216/222/228/234/240'
                    edit OVERLAP '1'
                    task derivedp
                      label files ""
                      label info ""
                      event fdb
                  endfamily
                  family 360
                    edit STEPS '246/to/360/by/6'
                    task derivedp
                      label files ""
                      label info ""
                      event fdb
                  endfamily
                  task extra_arc
                    edit STEPS '240/to/360'
                    edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                    edit RUNVARFC '1'
                    label files ""
                    label info ""
                    event fdb
                endfamily
              endfamily
            endfamily
            family cf
              edit EMOS_TYPE 'cf'
              edit NUMBER '0'
              task cfml
                edit STEPS '216'
              family cf240
                edit STEPS '216/228/240'
                edit OVERLAP '1'
                task pfpl
                  inlimit /emc_36r1/limits:pl
                  label files ""
                  label info ""
                  event fdb
                task pfsfc
                  trigger ../../postproc/dp eq complete
                  inlimit /emc_36r1/limits:sfc
                  label files ""
                  label info ""
                  event fdb
                task pfil
                  label files ""
                  label info ""
                  event fdb
                task pfwv
                  edit STREAM 'ENWH'
                  edit EMOS_STREAM 'ENWH'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf300
                edit STEPS '246/to/300/by/6'
                task pfpl
                  inlimit /emc_36r1/limits:pl
                  label files ""
                  label info ""
                  event fdb
                task pfsfc
                  trigger ../../postproc/dp eq complete
                  inlimit /emc_36r1/limits:sfc
                  label files ""
                  label info ""
                  event fdb
                task pfil
                  label files ""
                  label info ""
                  event fdb
                task pfwv
                  edit STREAM 'ENWH'
                  edit EMOS_STREAM 'ENWH'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf360
                edit STEPS '306/to/360/by/6'
                task pfpl
                  inlimit /emc_36r1/limits:pl
                  label files ""
                  label info ""
                  event fdb
                task pfsfc
                  trigger ../../postproc/dp eq complete
                  inlimit /emc_36r1/limits:sfc
                  label files ""
                  label info ""
                  event fdb
                task pfil
                  label files ""
                  label info ""
                  event fdb
                task pfwv
                  edit STREAM 'ENWH'
                  edit EMOS_STREAM 'ENWH'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              task accum
                edit STEPS '240'
                edit OVERLAP '1'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf
              edit EMOS_TYPE 'pf'
              family pf240
                edit STEPS '216/228/240'
                edit OVERLAP '1'
                task pfpl
                  inlimit /emc_36r1/limits:pl
                  label files ""
                  label info ""
                  event fdb
                task pfsfc
                  trigger ../../postproc/dp eq complete
                  inlimit /emc_36r1/limits:sfc
                  label files ""
                  label info ""
                  event fdb
                task pfil
                  label files ""
                  label info ""
                  event fdb
                task pfwv
                  edit STREAM 'ENWH'
                  edit EMOS_STREAM 'ENWH'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf300
                edit STEPS '246/to/300/by/6'
                task pfpl
                  inlimit /emc_36r1/limits:pl
                  label files ""
                  label info ""
                  event fdb
                task pfsfc
                  trigger ../../postproc/dp eq complete
                  inlimit /emc_36r1/limits:sfc
                  label files ""
                  label info ""
                  event fdb
                task pfil
                  label files ""
                  label info ""
                  event fdb
                task pfwv
                  edit STREAM 'ENWH'
                  edit EMOS_STREAM 'ENWH'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf360
                edit STEPS '306/to/360/by/6'
                task pfpl
                  inlimit /emc_36r1/limits:pl
                  label files ""
                  label info ""
                  event fdb
                task pfsfc
                  trigger ../../postproc/dp eq complete
                  inlimit /emc_36r1/limits:sfc
                  label files ""
                  label info ""
                  event fdb
                task pfil
                  label files ""
                  label info ""
                  event fdb
                task pfwv
                  edit STREAM 'ENWH'
                  edit EMOS_STREAM 'ENWH'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              task pfml
                edit STEPS '360'
              task accum
                edit STEPS '240'
                edit OVERLAP '1'
                label files ""
                label info ""
                event fdb
            endfamily
            family coupled
              edit EPSLEG '2'
              edit OCLEG '2'
              edit OCEPS '2'
              edit FCCHUNK '1'
              edit FCLENGTH '360'
              edit FCLENGTH0 '216'
              edit OCEPSTOT '3'
              edit FCTOTAL '51'
              edit RUNVARFC '1'
              edit EPSCOUPLE '1'
              edit RUNCOUPLE '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              inlimit /emc_36r1/lag/ref/00:mutex
              task preparcoce
              task arcfc
                trigger preparcoce eq complete
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family legC
            edit EPSLEG '3'
            edit FSFAMILY 'mars'
            edit OVERLAP '0'
            edit EMOS_TYPE 'an'
            edit EPSFCRES '255'
            inlimit /emc_36r1/limits:mars
            inlimit /limits:mars
            family cf
              edit EMOS_TYPE 'cf'
              edit NUMBER '0'
              family cf408
                edit STEPS '366/to/408/by/6'
                task pfpl
                  inlimit /emc_36r1/limits:pl
                  label files ""
                  label info ""
                  event fdb
                task pfsfc
                  trigger ../../postproc/dp eq complete
                  inlimit /emc_36r1/limits:sfc
                  label files ""
                  label info ""
                  event fdb
                task pfil
                  label files ""
                  label info ""
                  event fdb
                task pfwv
                  edit STREAM 'ENWH'
                  edit EMOS_STREAM 'ENWH'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf480
                edit STEPS '414/to/480/by/6'
                task pfpl
                  inlimit /emc_36r1/limits:pl
                  label files ""
                  label info ""
                  event fdb
                task pfsfc
                  trigger ../../postproc/dp eq complete
                  inlimit /emc_36r1/limits:sfc
                  label files ""
                  label info ""
                  event fdb
                task pfil
                  label files ""
                  label info ""
                  event fdb
                task pfwv
                  edit STREAM 'ENWH'
                  edit EMOS_STREAM 'ENWH'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf600
                edit STEPS '486/to/600/by/6'
                task pfpl
                  inlimit /emc_36r1/limits:pl
                  label files ""
                  label info ""
                  event fdb
                task pfsfc
                  trigger ../../postproc/dp eq complete
                  inlimit /emc_36r1/limits:sfc
                  label files ""
                  label info ""
                  event fdb
                task pfil
                  label files ""
                  label info ""
                  event fdb
                task pfwv
                  edit STREAM 'ENWH'
                  edit EMOS_STREAM 'ENWH'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf768
                edit STEPS '606/to/768/by/6'
                task pfpl
                  inlimit /emc_36r1/limits:pl
                  label files ""
                  label info ""
                  event fdb
                task pfsfc
                  trigger ../../postproc/dp eq complete
                  inlimit /emc_36r1/limits:sfc
                  label files ""
                  label info ""
                  event fdb
                task pfil
                  label files ""
                  label info ""
                  event fdb
                task pfwv
                  edit STREAM 'ENWH'
                  edit EMOS_STREAM 'ENWH'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family pf
              edit EMOS_TYPE 'pf'
              family cf408
                edit STEPS '366/to/408/by/6'
                task pfpl
                  inlimit /emc_36r1/limits:pl
                  label files ""
                  label info ""
                  event fdb
                task pfsfc
                  trigger ../../postproc/dp eq complete
                  inlimit /emc_36r1/limits:sfc
                  label files ""
                  label info ""
                  event fdb
                task pfil
                  label files ""
                  label info ""
                  event fdb
                task pfwv
                  edit STREAM 'ENWH'
                  edit EMOS_STREAM 'ENWH'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf480
                edit STEPS '414/to/480/by/6'
                task pfpl
                  inlimit /emc_36r1/limits:pl
                  label files ""
                  label info ""
                  event fdb
                task pfsfc
                  trigger ../../postproc/dp eq complete
                  inlimit /emc_36r1/limits:sfc
                  label files ""
                  label info ""
                  event fdb
                task pfil
                  label files ""
                  label info ""
                  event fdb
                task pfwv
                  edit STREAM 'ENWH'
                  edit EMOS_STREAM 'ENWH'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf600
                edit STEPS '486/to/600/by/6'
                task pfpl
                  inlimit /emc_36r1/limits:pl
                  label files ""
                  label info ""
                  event fdb
                task pfsfc
                  trigger ../../postproc/dp eq complete
                  inlimit /emc_36r1/limits:sfc
                  label files ""
                  label info ""
                  event fdb
                task pfil
                  label files ""
                  label info ""
                  event fdb
                task pfwv
                  edit STREAM 'ENWH'
                  edit EMOS_STREAM 'ENWH'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf768
                edit STEPS '606/to/768/by/6'
                task pfpl
                  inlimit /emc_36r1/limits:pl
                  label files ""
                  label info ""
                  event fdb
                task pfsfc
                  trigger ../../postproc/dp eq complete
                  inlimit /emc_36r1/limits:sfc
                  label files ""
                  label info ""
                  event fdb
                task pfil
                  label files ""
                  label info ""
                  event fdb
                task pfwv
                  edit STREAM 'ENWH'
                  edit EMOS_STREAM 'ENWH'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family postproc
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit FCCHUNK '1'
              edit FCLENGTH '768'
              inlimit /emc_36r1/limits:postproc
              family dp
                trigger /emc_36r1/00/legC/postproc/dp eq complete and /emc_36r1/00/legC/postproc/extra eq complete
                edit RUNVARFC '1'
                edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
                edit EMOS_TYPE 'pf'
                edit STEPS '366/to/768/by/6'
                family cf
                  edit EMOS_TYPE 'cf'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                  task extra_arc
                    edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                    edit STEPS '360/to/768'
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family pf
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                  task extra_arc
                    edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                    edit STEPS '360/to/768'
                    label files ""
                    label info ""
                    event fdb
                endfamily
              endfamily
              family extra
                trigger /emc_36r1/hc2/back/01/legBcpl/postproc/extra eq complete
                edit RUNVARFC '1'
                task wm_sfc_arc
                  label files ""
                  label info ""
                  event fdb
                task wm_ua_arc
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family nino
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit DAILY '0'
                edit RUNVARFC '1'
                edit FCLENGTH1 '360'
                edit QUEUE 'emos-e'
                edit QUEUE_EPILOG 'emos-exp'
                edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
                edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
                edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
                edit LOGDIR '/vol/emos_nc/output'
                edit SMSOUT '/vol/emos_nc/output'
                edit FCLENGTH '768'
                edit EPSCOUPLE '1'
                edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                limit semaphore 10
                inlimit /limits:linux_cluster
                inlimit semaphore
                family 00
                  edit OCGROUP 'an00'
                  edit FCGROUP 'fc0'
                  edit EPSMEMBER '0'
                  edit MEMBER '0'
                  task nino_daily
                  task tcyc
                    trigger nino_daily eq complete or nino_daily eq aborted
                endfamily
                family 01
                  edit OCGROUP 'an1'
                  edit FCGROUP 'fc0'
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  task nino_daily
                  task tcyc
                    trigger nino_daily eq complete or nino_daily eq aborted
                endfamily
                family 02
                  edit OCGROUP 'an2'
                  edit FCGROUP 'fc0'
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  task nino_daily
                  task tcyc
                    trigger nino_daily eq complete or nino_daily eq aborted
                endfamily
                family 03
                  edit OCGROUP 'an3'
                  edit FCGROUP 'fc0'
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  task nino_daily
                  task tcyc
                    trigger nino_daily eq complete or nino_daily eq aborted
                endfamily
                family 04
                  edit OCGROUP 'an4'
                  edit FCGROUP 'fc0'
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  task nino_daily
                  task tcyc
                    trigger nino_daily eq complete or nino_daily eq aborted
                endfamily
              endfamily
              task tcyctoecfs
                trigger nino eq complete
                edit QUEUE 'emos-e'
                edit QUEUE_EPILOG 'emos-exp'
                edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
                edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
                edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
                edit LOGDIR '/vol/emos_nc/output'
                edit SMSOUT '/vol/emos_nc/output'
                edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                edit EPSCOUPLE '1'
                edit FCLENGTH '768'
                inlimit /limits:linux_cluster
            endfamily
          endfamily
          task save_logs
            trigger legA eq complete and legB eq complete and legC eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1'
            edit SCHOST 'localhost'
        endfamily
        family clean
          task chver
            trigger ../arch eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/efi_hdc'
            edit GENERAT '0'
            edit CLEANER '1'
        endfamily
      endfamily
      family eto1
        trigger ./1to2 eq complete
        edit ORIG_FDB_ROOT 'undef'
        edit DEST_FDB_ROOT 'undef'
        family chver
          repeat integer DELTA_YEAR -18 -1
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/efi_hdc'
          edit VERSION '0046'
          edit VERSION_TO '9906'
          edit CHVER_FDB_ROOT '/ma_fdb'
          edit GENERAT '1'
          edit CLEANER '0'
          family enfh
            edit CHVER_STREAM 'enfh'
            family 0
              edit CHVER_TYPE ':0000:[0:]:'
              task chver
                label info ""
            endfamily
            family 1
              edit CHVER_TYPE ':0000:1:'
              task chver
                label info ""
            endfamily
            family 2
              edit CHVER_TYPE ':0000:2:'
              task chver
                label info ""
            endfamily
            family 3
              edit CHVER_TYPE ':0000:3:'
              task chver
                label info ""
            endfamily
            family 4
              edit CHVER_TYPE ':0000:4:'
              task chver
                label info ""
            endfamily
          endfamily
          family enwh
            edit CHVER_STREAM 'enwh'
            task chver
              label info ""
          endfamily
          family efho
            edit CHVER_STREAM 'efho'
            task chver
              label info ""
          endfamily
          family ewho
            edit CHVER_STREAM 'ewho'
            task chver
              label info ""
          endfamily
          task save_logs
            trigger enfh eq complete and enwh eq complete and efho eq complete and ewho eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1'
            edit SCHOST 'localhost'
        endfamily
        family arch
          trigger ./chver:DELTA_YEAR gt ./arch:DELTA_YEAR or ./chver eq complete
          repeat integer DELTA_YEAR -18 -1
          edit EMOS_TYPE 'an'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
          edit FSFAMILY 'mars'
          edit OVERLAP '0'
          edit VERSION '9906'
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family legA
            family cf
              edit EMOS_TYPE 'cf'
              edit NUMBER '0'
              family cf144
                edit STEPS '0/to/144/by/6'
                task pfpl
                  inlimit /emc_36r1/limits:pl
                  label files ""
                  label info ""
                  event fdb
                task pfsfc
                  trigger ../../postproc/dp eq complete
                  inlimit /emc_36r1/limits:sfc
                  label files ""
                  label info ""
                  event fdb
                task pfil
                  label files ""
                  label info ""
                  event fdb
                task pfwv
                  edit STREAM 'ENWH'
                  edit EMOS_STREAM 'ENWH'
                  edit STEPS '006/to/144/by/6'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf240
                edit STEPS '150/to/240/by/6'
                task pfpl
                  inlimit /emc_36r1/limits:pl
                  label files ""
                  label info ""
                  event fdb
                task pfsfc
                  trigger ../../postproc/dp eq complete
                  inlimit /emc_36r1/limits:sfc
                  label files ""
                  label info ""
                  event fdb
                task pfil
                  label files ""
                  label info ""
                  event fdb
                task pfwv
                  edit STREAM 'ENWH'
                  edit EMOS_STREAM 'ENWH'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family pf
              edit EMOS_TYPE 'pf'
              family pf072
                edit STEPS '0/to/072/by/6'
                task pfpl
                  inlimit /emc_36r1/limits:pl
                  label files ""
                  label info ""
                  event fdb
                task pfsfc
                  trigger ../../postproc/dp eq complete
                  inlimit /emc_36r1/limits:sfc
                  label files ""
                  label info ""
                  event fdb
                task pfil
                  label files ""
                  label info ""
                  event fdb
                task pfwv
                  edit STREAM 'ENWH'
                  edit EMOS_STREAM 'ENWH'
                  edit STEPS '006/to/072/by/6'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf144
                edit STEPS '078/to/144/by/6'
                task pfpl
                  inlimit /emc_36r1/limits:pl
                  label files ""
                  label info ""
                  event fdb
                task pfsfc
                  trigger ../../postproc/dp eq complete
                  inlimit /emc_36r1/limits:sfc
                  label files ""
                  label info ""
                  event fdb
                task pfil
                  label files ""
                  label info ""
                  event fdb
                task pfwv
                  edit STREAM 'ENWH'
                  edit EMOS_STREAM 'ENWH'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf192
                edit STEPS '150/to/192/by/6'
                task pfpl
                  inlimit /emc_36r1/limits:pl
                  label files ""
                  label info ""
                  event fdb
                task pfsfc
                  trigger ../../postproc/dp eq complete
                  inlimit /emc_36r1/limits:sfc
                  label files ""
                  label info ""
                  event fdb
                task pfil
                  label files ""
                  label info ""
                  event fdb
                task pfwv
                  edit STREAM 'ENWH'
                  edit EMOS_STREAM 'ENWH'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf240
                edit STEPS '198/to/240/by/6'
                task pfpl
                  inlimit /emc_36r1/limits:pl
                  label files ""
                  label info ""
                  event fdb
                task pfsfc
                  trigger ../../postproc/dp eq complete
                  inlimit /emc_36r1/limits:sfc
                  label files ""
                  label info ""
                  event fdb
                task pfil
                  label files ""
                  label info ""
                  event fdb
                task pfwv
                  edit STREAM 'ENWH'
                  edit EMOS_STREAM 'ENWH'
                  label files ""
                  label info ""
                  event fdb
                task pfml
                  edit STEPS '216'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family postproc
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit RUNVARFC '1'
              inlimit /emc_36r1/limits:postproc
              family dp
                edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
                family pf
                  edit EMOS_TYPE 'pf'
                  family 144
                    edit STEPS '006/to/144/by/6'
                    task derivedp
                      label files ""
                      label info ""
                      event fdb
                  endfamily
                  family 240
                    edit STEPS '150/to/240/by/6'
                    task derivedp
                      label files ""
                      label info ""
                      event fdb
                  endfamily
                  task extra_arc
                    edit STEPS '0/to/240'
                    edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                    edit RUNVARFC '1'
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family cf
                  edit EMOS_TYPE 'cf'
                  family 144
                    edit STEPS '006/to/144/by/6'
                    task derivedp
                      label files ""
                      label info ""
                      event fdb
                  endfamily
                  family 240
                    edit STEPS '150/to/240/by/6'
                    task derivedp
                      label files ""
                      label info ""
                      event fdb
                  endfamily
                  task extra_arc
                    edit STEPS '0/to/240'
                    edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                    edit RUNVARFC '1'
                    label files ""
                    label info ""
                    event fdb
                endfamily
              endfamily
            endfamily
            family coupled
              trigger ./cf eq complete and ./pf eq complete
              edit EPSLEG '1'
              edit OCLEG '1'
              edit OCEPS '1'
              edit FCCHUNK '1'
              edit FCLENGTH '240'
              edit FCLENGTH0 '0'
              edit OCEPSTOT '3'
              edit FCTOTAL '51'
              edit RUNVARFC '1'
              edit EPSCOUPLE '1'
              edit RUNCOUPLE '0'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              inlimit /emc_36r1/lag/ref/00:mutex
              task preparcoce
              task arcfc
                trigger preparcoce eq complete
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family legB
            edit EPSLEG '2'
            edit EPSWAMNFRE '25'
            edit EPSWAMNANG '12'
            inlimit /emc_36r1/limits:mars
            inlimit /limits:mars
            family postproc
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              inlimit /emc_36r1/limits:postproc
              family dp
                edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
                family pf
                  edit EMOS_TYPE 'pf'
                  family 240
                    edit STEPS '216/222/228/234/240'
                    edit OVERLAP '1'
                    task derivedp
                      label files ""
                      label info ""
                      event fdb
                  endfamily
                  family 360
                    edit STEPS '246/to/360/by/6'
                    task derivedp
                      label files ""
                      label info ""
                      event fdb
                  endfamily
                  task extra_arc
                    edit STEPS '240/to/360'
                    edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                    edit RUNVARFC '1'
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family cf
                  edit EMOS_TYPE 'cf'
                  family 240
                    edit STEPS '216/222/228/234/240'
                    edit OVERLAP '1'
                    task derivedp
                      label files ""
                      label info ""
                      event fdb
                  endfamily
                  family 360
                    edit STEPS '246/to/360/by/6'
                    task derivedp
                      label files ""
                      label info ""
                      event fdb
                  endfamily
                  task extra_arc
                    edit STEPS '240/to/360'
                    edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                    edit RUNVARFC '1'
                    label files ""
                    label info ""
                    event fdb
                endfamily
              endfamily
            endfamily
            family cf
              edit EMOS_TYPE 'cf'
              edit NUMBER '0'
              task cfml
                edit STEPS '216'
              family cf240
                edit STEPS '216/228/240'
                edit OVERLAP '1'
                task pfpl
                  inlimit /emc_36r1/limits:pl
                  label files ""
                  label info ""
                  event fdb
                task pfsfc
                  trigger ../../postproc/dp eq complete
                  inlimit /emc_36r1/limits:sfc
                  label files ""
                  label info ""
                  event fdb
                task pfil
                  label files ""
                  label info ""
                  event fdb
                task pfwv
                  edit STREAM 'ENWH'
                  edit EMOS_STREAM 'ENWH'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf300
                edit STEPS '246/to/300/by/6'
                task pfpl
                  inlimit /emc_36r1/limits:pl
                  label files ""
                  label info ""
                  event fdb
                task pfsfc
                  trigger ../../postproc/dp eq complete
                  inlimit /emc_36r1/limits:sfc
                  label files ""
                  label info ""
                  event fdb
                task pfil
                  label files ""
                  label info ""
                  event fdb
                task pfwv
                  edit STREAM 'ENWH'
                  edit EMOS_STREAM 'ENWH'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf360
                edit STEPS '306/to/360/by/6'
                task pfpl
                  inlimit /emc_36r1/limits:pl
                  label files ""
                  label info ""
                  event fdb
                task pfsfc
                  trigger ../../postproc/dp eq complete
                  inlimit /emc_36r1/limits:sfc
                  label files ""
                  label info ""
                  event fdb
                task pfil
                  label files ""
                  label info ""
                  event fdb
                task pfwv
                  edit STREAM 'ENWH'
                  edit EMOS_STREAM 'ENWH'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              task accum
                edit STEPS '240'
                edit OVERLAP '1'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf
              edit EMOS_TYPE 'pf'
              family pf240
                edit STEPS '216/228/240'
                edit OVERLAP '1'
                task pfpl
                  inlimit /emc_36r1/limits:pl
                  label files ""
                  label info ""
                  event fdb
                task pfsfc
                  trigger ../../postproc/dp eq complete
                  inlimit /emc_36r1/limits:sfc
                  label files ""
                  label info ""
                  event fdb
                task pfil
                  label files ""
                  label info ""
                  event fdb
                task pfwv
                  edit STREAM 'ENWH'
                  edit EMOS_STREAM 'ENWH'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf300
                edit STEPS '246/to/300/by/6'
                task pfpl
                  inlimit /emc_36r1/limits:pl
                  label files ""
                  label info ""
                  event fdb
                task pfsfc
                  trigger ../../postproc/dp eq complete
                  inlimit /emc_36r1/limits:sfc
                  label files ""
                  label info ""
                  event fdb
                task pfil
                  label files ""
                  label info ""
                  event fdb
                task pfwv
                  edit STREAM 'ENWH'
                  edit EMOS_STREAM 'ENWH'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf360
                edit STEPS '306/to/360/by/6'
                task pfpl
                  inlimit /emc_36r1/limits:pl
                  label files ""
                  label info ""
                  event fdb
                task pfsfc
                  trigger ../../postproc/dp eq complete
                  inlimit /emc_36r1/limits:sfc
                  label files ""
                  label info ""
                  event fdb
                task pfil
                  label files ""
                  label info ""
                  event fdb
                task pfwv
                  edit STREAM 'ENWH'
                  edit EMOS_STREAM 'ENWH'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              task pfml
                edit STEPS '360'
              task accum
                edit STEPS '240'
                edit OVERLAP '1'
                label files ""
                label info ""
                event fdb
            endfamily
            family coupled
              edit EPSLEG '2'
              edit OCLEG '2'
              edit OCEPS '2'
              edit FCCHUNK '1'
              edit FCLENGTH '360'
              edit FCLENGTH0 '216'
              edit OCEPSTOT '3'
              edit FCTOTAL '51'
              edit RUNVARFC '1'
              edit EPSCOUPLE '1'
              edit RUNCOUPLE '1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              inlimit /emc_36r1/lag/ref/00:mutex
              task preparcoce
              task arcfc
                trigger preparcoce eq complete
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family legC
            edit EPSLEG '3'
            edit FSFAMILY 'mars'
            edit OVERLAP '0'
            edit EMOS_TYPE 'an'
            edit EPSFCRES '255'
            inlimit /emc_36r1/limits:mars
            inlimit /limits:mars
            family cf
              edit EMOS_TYPE 'cf'
              edit NUMBER '0'
              family cf408
                edit STEPS '366/to/408/by/6'
                task pfpl
                  inlimit /emc_36r1/limits:pl
                  label files ""
                  label info ""
                  event fdb
                task pfsfc
                  trigger ../../postproc/dp eq complete
                  inlimit /emc_36r1/limits:sfc
                  label files ""
                  label info ""
                  event fdb
                task pfil
                  label files ""
                  label info ""
                  event fdb
                task pfwv
                  edit STREAM 'ENWH'
                  edit EMOS_STREAM 'ENWH'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf480
                edit STEPS '414/to/480/by/6'
                task pfpl
                  inlimit /emc_36r1/limits:pl
                  label files ""
                  label info ""
                  event fdb
                task pfsfc
                  trigger ../../postproc/dp eq complete
                  inlimit /emc_36r1/limits:sfc
                  label files ""
                  label info ""
                  event fdb
                task pfil
                  label files ""
                  label info ""
                  event fdb
                task pfwv
                  edit STREAM 'ENWH'
                  edit EMOS_STREAM 'ENWH'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf600
                edit STEPS '486/to/600/by/6'
                task pfpl
                  inlimit /emc_36r1/limits:pl
                  label files ""
                  label info ""
                  event fdb
                task pfsfc
                  trigger ../../postproc/dp eq complete
                  inlimit /emc_36r1/limits:sfc
                  label files ""
                  label info ""
                  event fdb
                task pfil
                  label files ""
                  label info ""
                  event fdb
                task pfwv
                  edit STREAM 'ENWH'
                  edit EMOS_STREAM 'ENWH'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf768
                edit STEPS '606/to/768/by/6'
                task pfpl
                  inlimit /emc_36r1/limits:pl
                  label files ""
                  label info ""
                  event fdb
                task pfsfc
                  trigger ../../postproc/dp eq complete
                  inlimit /emc_36r1/limits:sfc
                  label files ""
                  label info ""
                  event fdb
                task pfil
                  label files ""
                  label info ""
                  event fdb
                task pfwv
                  edit STREAM 'ENWH'
                  edit EMOS_STREAM 'ENWH'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family pf
              edit EMOS_TYPE 'pf'
              family cf408
                edit STEPS '366/to/408/by/6'
                task pfpl
                  inlimit /emc_36r1/limits:pl
                  label files ""
                  label info ""
                  event fdb
                task pfsfc
                  trigger ../../postproc/dp eq complete
                  inlimit /emc_36r1/limits:sfc
                  label files ""
                  label info ""
                  event fdb
                task pfil
                  label files ""
                  label info ""
                  event fdb
                task pfwv
                  edit STREAM 'ENWH'
                  edit EMOS_STREAM 'ENWH'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf480
                edit STEPS '414/to/480/by/6'
                task pfpl
                  inlimit /emc_36r1/limits:pl
                  label files ""
                  label info ""
                  event fdb
                task pfsfc
                  trigger ../../postproc/dp eq complete
                  inlimit /emc_36r1/limits:sfc
                  label files ""
                  label info ""
                  event fdb
                task pfil
                  label files ""
                  label info ""
                  event fdb
                task pfwv
                  edit STREAM 'ENWH'
                  edit EMOS_STREAM 'ENWH'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf600
                edit STEPS '486/to/600/by/6'
                task pfpl
                  inlimit /emc_36r1/limits:pl
                  label files ""
                  label info ""
                  event fdb
                task pfsfc
                  trigger ../../postproc/dp eq complete
                  inlimit /emc_36r1/limits:sfc
                  label files ""
                  label info ""
                  event fdb
                task pfil
                  label files ""
                  label info ""
                  event fdb
                task pfwv
                  edit STREAM 'ENWH'
                  edit EMOS_STREAM 'ENWH'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family pf768
                edit STEPS '606/to/768/by/6'
                task pfpl
                  inlimit /emc_36r1/limits:pl
                  label files ""
                  label info ""
                  event fdb
                task pfsfc
                  trigger ../../postproc/dp eq complete
                  inlimit /emc_36r1/limits:sfc
                  label files ""
                  label info ""
                  event fdb
                task pfil
                  label files ""
                  label info ""
                  event fdb
                task pfwv
                  edit STREAM 'ENWH'
                  edit EMOS_STREAM 'ENWH'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family postproc
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
              edit FCCHUNK '1'
              edit FCLENGTH '768'
              inlimit /emc_36r1/limits:postproc
              family dp
                trigger /emc_36r1/00/legC/postproc/dp eq complete and /emc_36r1/00/legC/postproc/extra eq complete
                edit RUNVARFC '1'
                edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
                edit EMOS_TYPE 'pf'
                edit STEPS '366/to/768/by/6'
                family cf
                  edit EMOS_TYPE 'cf'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                  task extra_arc
                    edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                    edit STEPS '360/to/768'
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family pf
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                  task extra_arc
                    edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                    edit STEPS '360/to/768'
                    label files ""
                    label info ""
                    event fdb
                endfamily
              endfamily
              family extra
                trigger /emc_36r1/hc2/back/01/legBcpl/postproc/extra eq complete
                edit RUNVARFC '1'
                task wm_sfc_arc
                  label files ""
                  label info ""
                  event fdb
                task wm_ua_arc
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family nino
                edit FCCHUNK '1'
                edit OCTOTAL '5'
                edit FCTOTAL '51'
                edit DAILY '0'
                edit RUNVARFC '1'
                edit FCLENGTH1 '360'
                edit QUEUE 'emos-e'
                edit QUEUE_EPILOG 'emos-exp'
                edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
                edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
                edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
                edit LOGDIR '/vol/emos_nc/output'
                edit SMSOUT '/vol/emos_nc/output'
                edit FCLENGTH '768'
                edit EPSCOUPLE '1'
                edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                limit semaphore 10
                inlimit /limits:linux_cluster
                inlimit semaphore
                family 00
                  edit OCGROUP 'an00'
                  edit FCGROUP 'fc0'
                  edit EPSMEMBER '0'
                  edit MEMBER '0'
                  task nino_daily
                  task tcyc
                    trigger nino_daily eq complete or nino_daily eq aborted
                endfamily
                family 01
                  edit OCGROUP 'an1'
                  edit FCGROUP 'fc0'
                  edit EPSMEMBER '1'
                  edit MEMBER '1'
                  task nino_daily
                  task tcyc
                    trigger nino_daily eq complete or nino_daily eq aborted
                endfamily
                family 02
                  edit OCGROUP 'an2'
                  edit FCGROUP 'fc0'
                  edit EPSMEMBER '2'
                  edit MEMBER '2'
                  task nino_daily
                  task tcyc
                    trigger nino_daily eq complete or nino_daily eq aborted
                endfamily
                family 03
                  edit OCGROUP 'an3'
                  edit FCGROUP 'fc0'
                  edit EPSMEMBER '3'
                  edit MEMBER '3'
                  task nino_daily
                  task tcyc
                    trigger nino_daily eq complete or nino_daily eq aborted
                endfamily
                family 04
                  edit OCGROUP 'an4'
                  edit FCGROUP 'fc0'
                  edit EPSMEMBER '4'
                  edit MEMBER '4'
                  task nino_daily
                  task tcyc
                    trigger nino_daily eq complete or nino_daily eq aborted
                endfamily
              endfamily
              task tcyctoecfs
                trigger nino eq complete
                edit QUEUE 'emos-e'
                edit QUEUE_EPILOG 'emos-exp'
                edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
                edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
                edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
                edit LOGDIR '/vol/emos_nc/output'
                edit SMSOUT '/vol/emos_nc/output'
                edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                edit EPSCOUPLE '1'
                edit FCLENGTH '768'
                inlimit /limits:linux_cluster
            endfamily
          endfamily
          task save_logs
            trigger legA eq complete and legB eq complete and legC eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1'
            edit SCHOST 'localhost'
        endfamily
        family clean
          task chver
            trigger ../arch eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/efi_hdc'
            edit GENERAT '0'
            edit CLEANER '1'
        endfamily
      endfamily
    endfamily
    task stage_pp
      edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
    task logfilestore
      edit QUEUE 'emos-e'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit KEEPLOGS 'no'
      edit LOGSTOECFS 'no'
      edit MOVE_TO_ECFS 'yes'
    task logfiles
      late -s +00:05 -c +00:25
      trigger logfilestore eq complete and back eq complete and arch eq complete and efi eq complete and adm eq complete
      edit QUEUE 'emos-e'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit MOVE_TO_ECFS 'no'
  endfamily
  family lag
    repeat date YMD 20091003 20121212 1
    edit EMOS_TYPE 'an'
    edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
    edit FSFAMILY 'mars'
    edit OVERLAP '0'
    edit LOGTASK 'lag'
    edit MARS_DB 'marsod'
    family ref
      edit SMSFILES '/home/ma/emos/def/emc_36r1/ref'
      family 12
        edit EMOS_BASE '12'
        limit mutex 1
        task events
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/ref'
          edit HINDCAST '1'
          edit ONLY_CF '0'
          event noefi
          event 240 
          event 360 
          event cpl
          event 768 
          event cf
      endfamily
      family 00
        edit DELTA_DAY '1'
        edit EMOS_BASE '00'
        limit mutex 1
        task events
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/ref'
          edit HINDCAST '1'
          edit ONLY_CF '0'
          event noefi
          event 240 
          event 360 
          event cpl
          event 768 
          event cf
      endfamily
      task smove
        trigger /emc_36r1/lag/12 eq complete
        edit SMSFILES '/home/ma/emos/def/emc_36r1/system'
      family loop
        task dummy
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit WSHOST 'localhost'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          time 12:45
      endfamily
      family warn
        trigger ./loop eq complete
        task dummy
          complete /emc_36r1/lag/logfiles eq complete
          time 16:00
      endfamily
      task checkarchive
        trigger /emc_36r1/main/make eq complete
        edit QUEUE 'emos-e'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        edit SMSFILES '/home/ma/emos/def/emc_36r1'
        edit FSFAMILY 'mars'
        edit EMOS_TYPE 'an'
        edit OSTREAM 'WEOV EFOV'
        edit HSTREAM 'EFHO EFHS ENFH ENWH EWHO'
        inlimit /limits:linux_cluster
        label info "none"
    endfamily
    family 12
      trigger ( /emc_36r1/main:YMD gt /emc_36r1/lag:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/lag:YMD)
      edit DELTA_DAY '0'
      edit EMOS_BASE '12'
      family legA
        trigger /emc_36r1/lag/ref/12/events eq complete
        edit EPSWAMNFRE '30'
        edit EPSWAMNANG '24'
        family init
          trigger ( /emc_36r1/main:YMD gt /emc_36r1/lag:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/lag:YMD and /emc_36r1/main/12/ref/an eq complete)
          edit FSFAMILY 'mc'
          task rmfdb
        endfamily
        family tc
          complete /emc_36r1/lag/ref/12/events:cf
          trigger ( /emc_36r1/main:YMD gt /emc_36r1/lag:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/lag:YMD and /emc_36r1/main/12/legA/postproc/tc eq complete)
          edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
          edit FSFAMILY 'mc'
          task clean
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/tc'
          task to_ecfs
            label files ""
            label info ""
            event fdb
          task to_mars
            trigger to_ecfs eq complete
        endfamily
        family archive
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family postproc
            trigger ( /emc_36r1/main:YMD gt /emc_36r1/lag:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/lag:YMD and /emc_36r1/main/12/legA/postproc eq complete) and /emc_36r1/lag/ref/12/events eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                complete /emc_36r1/lag/ref/12/events:cf
                edit EMOS_TYPE 'pf'
                family 144
                  edit STEPS '0/to/144/by/3'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 144
                  edit STEPS '0/to/144/by/3'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family cluster
              complete /emc_36r1/lag/ref/12/events:cf
              edit STEPS '72/84/96/108/120/132/144/156/168'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family a
                edit DOMAIN 'a'
                task cmpl
                  edit TYPE 'cm'
                  label files ""
                  label info ""
                  event fdb
                task cspl
                  edit TYPE 'cs'
                  label files ""
                  label info ""
                  event fdb
                task tube
                  edit STEPS '48/72/96/120/144/168/192/216/240'
                  edit REFERENCE '96/144/168/192/240'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family b
                trigger ./a eq complete
                edit DOMAIN 'b'
                task cmpl
                  edit TYPE 'cm'
                  label files ""
                  label info ""
                  event fdb
                task cspl
                  edit TYPE 'cs'
                  label files ""
                  label info ""
                  event fdb
                task tube
                  edit STEPS '48/72/96/120/144/168/192/216/240'
                  edit REFERENCE '96/144/168/192/240'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family c
                trigger ./b eq complete
                edit DOMAIN 'c'
                task cmpl
                  edit TYPE 'cm'
                  label files ""
                  label info ""
                  event fdb
                task cspl
                  edit TYPE 'cs'
                  label files ""
                  label info ""
                  event fdb
                task tube
                  edit STEPS '48/72/96/120/144/168/192/216/240'
                  edit REFERENCE '96/144/168/192/240'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family d
                trigger ./c eq complete
                edit DOMAIN 'd'
                task cmpl
                  edit TYPE 'cm'
                  label files ""
                  label info ""
                  event fdb
                task cspl
                  edit TYPE 'cs'
                  label files ""
                  label info ""
                  event fdb
                task tube
                  edit STEPS '48/72/96/120/144/168/192/216/240'
                  edit REFERENCE '96/144/168/192/240'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family e
                trigger ./d eq complete
                edit DOMAIN 'e'
                task cmpl
                  edit TYPE 'cm'
                  label files ""
                  label info ""
                  event fdb
                task cspl
                  edit TYPE 'cs'
                  label files ""
                  label info ""
                  event fdb
                task tube
                  edit STEPS '48/72/96/120/144/168/192/216/240'
                  edit REFERENCE '96/144/168/192/240'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family ensms
              complete /emc_36r1/lag/ref/12/events:cf
              edit STEPS '3/6/9/12/15/18/21/24/27/30/33/36/39/42/45/48/51/54/57/60/63/66/69/72/75/78/81/84/87/90/93/96/99/102/105/108/111/114/117/120/123/126/129/132/135/138/141/144/150/156/162/168/174/180/186/192/198/204/210/216/222/228/234/240'
              task mean
                edit TYPE 'em'
                label files ""
                label info ""
                event fdb
              task stdev
                edit TYPE 'es'
                label files ""
                label info ""
                event fdb
            endfamily
            family prob
              complete /emc_36r1/lag/ref/12/events:cf
              family 2t
                edit PARAM '2t'
                task arc_prob
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family swp_mwp
                edit PARAM 'swp_mwp'
                task arc_prob
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family 10fg
                edit PARAM '10fg'
                task arc_prob
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family ff
                edit PARAM 'ff'
                task arc_prob
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family rain
                edit PARAM 'rain'
                task arc_prob
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family t850
                edit PARAM 't850'
                task arc_prob
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family efi
              complete ( /emc_36r1/main/ref/12/events:cf) or ( /emc_36r1/main/ref/12/events:noefi)
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/lag:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/lag:YMD and /emc_36r1/main/12/legA/postproc eq complete)
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family 2t
                edit RF_NAME '2t'
                task efi
                  label info ""
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task efi
                  label info ""
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task efi
                  label info ""
              endfamily
              family tp
                edit RF_NAME 'tp'
                task efi
                  label info ""
              endfamily
            endfamily
          endfamily
          family cf
            trigger ( /emc_36r1/main:YMD gt /emc_36r1/lag:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/lag:YMD and /emc_36r1/main/12/legA/fc/cf eq complete)
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf036
              edit STEPS '0/to/36/by/3'
              task cfml
                label files ""
                label info ""
                event fdb
              task cfpl
                label files ""
                label info ""
                event fdb
              task cfsfc
                label files ""
                label info ""
                event fdb
              task cfil
                label files ""
                label info ""
                event fdb
              task cfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf072
              edit STEPS '39/to/72/by/3'
              task cfml
                label files ""
                label info ""
                event fdb
              task cfpl
                label files ""
                label info ""
                event fdb
              task cfsfc
                label files ""
                label info ""
                event fdb
              task cfil
                label files ""
                label info ""
                event fdb
              task cfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf108
              edit STEPS '075/to/108/by/3'
              task cfml
                label files ""
                label info ""
                event fdb
              task cfpl
                label files ""
                label info ""
                event fdb
              task cfsfc
                label files ""
                label info ""
                event fdb
              task cfil
                label files ""
                label info ""
                event fdb
              task cfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf144
              edit STEPS '111/to/144/by/3'
              task cfml
                label files ""
                label info ""
                event fdb
              task cfpl
                label files ""
                label info ""
                event fdb
              task cfsfc
                label files ""
                label info ""
                event fdb
              task cfil
                label files ""
                label info ""
                event fdb
              task cfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf180
              edit STEPS '150/to/180/by/6'
              task cfml
                label files ""
                label info ""
                event fdb
              task cfpl
                label files ""
                label info ""
                event fdb
              task cfsfc
                label files ""
                label info ""
                event fdb
              task cfil
                label files ""
                label info ""
                event fdb
              task cfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf240
              edit STEPS '186/to/240/by/6'
              task cfml
                label files ""
                label info ""
                event fdb
              task cfpl
                label files ""
                label info ""
                event fdb
              task cfsfc
                label files ""
                label info ""
                event fdb
              task cfil
                label files ""
                label info ""
                event fdb
              task cfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family cv
            trigger ( /emc_36r1/main:YMD gt /emc_36r1/lag:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/lag:YMD and /emc_36r1/main/12/legA/fc/cv eq complete)
            edit EMOS_TYPE 'cv'
            edit NUMBER '1/2'
            family cv036
              edit STEPS '0/to/036/by/3'
              task cfml
                label files ""
                label info ""
                event fdb
              task cfpl
                label files ""
                label info ""
                event fdb
              task cfsfc
                trigger ../../postproc/dp eq complete
                label files ""
                label info ""
                event fdb
              task cfil
                label files ""
                label info ""
                event fdb
              task cfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
              task derivedp
                label files ""
                label info ""
                event fdb
            endfamily
            family cv072
              edit STEPS '039/to/072/by/3'
              task cfml
                label files ""
                label info ""
                event fdb
              task cfpl
                label files ""
                label info ""
                event fdb
              task cfsfc
                trigger ../../postproc/dp eq complete
                label files ""
                label info ""
                event fdb
              task cfil
                label files ""
                label info ""
                event fdb
              task cfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
              task derivedp
                label files ""
                label info ""
                event fdb
            endfamily
            family cv108
              edit STEPS '075/to/108/by/3'
              task cfml
                label files ""
                label info ""
                event fdb
              task cfpl
                label files ""
                label info ""
                event fdb
              task cfsfc
                trigger ../../postproc/dp eq complete
                label files ""
                label info ""
                event fdb
              task cfil
                label files ""
                label info ""
                event fdb
              task cfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
              task derivedp
                label files ""
                label info ""
                event fdb
            endfamily
            family cv144
              edit STEPS '111/to/144/by/3'
              task cfml
                label files ""
                label info ""
                event fdb
              task cfpl
                label files ""
                label info ""
                event fdb
              task cfsfc
                trigger ../../postproc/dp eq complete
                label files ""
                label info ""
                event fdb
              task cfil
                label files ""
                label info ""
                event fdb
              task cfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
              task derivedp
                label files ""
                label info ""
                event fdb
            endfamily
            family cv180
              edit STEPS '150/to/180/by/6'
              task cfml
                label files ""
                label info ""
                event fdb
              task cfpl
                label files ""
                label info ""
                event fdb
              task cfsfc
                trigger ../../postproc/dp eq complete
                label files ""
                label info ""
                event fdb
              task cfil
                label files ""
                label info ""
                event fdb
              task cfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
              task derivedp
                label files ""
                label info ""
                event fdb
            endfamily
            family cv240
              edit STEPS '186/to/240/by/6'
              task cfml
                label files ""
                label info ""
                event fdb
              task cfpl
                label files ""
                label info ""
                event fdb
              task cfsfc
                trigger ../../postproc/dp eq complete
                label files ""
                label info ""
                event fdb
              task cfil
                label files ""
                label info ""
                event fdb
              task cfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
              task derivedp
                label files ""
                label info ""
                event fdb
            endfamily
            family cv300
              edit STEPS '246/to/300/by/6'
              task cfml
                label files ""
                label info ""
                event fdb
              task cfpl
                label files ""
                label info ""
                event fdb
              task cfsfc
                trigger ../../postproc/dp eq complete
                label files ""
                label info ""
                event fdb
              task cfil
                label files ""
                label info ""
                event fdb
              task cfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
              task derivedp
                label files ""
                label info ""
                event fdb
            endfamily
            family cv360
              edit STEPS '306/to/360/by/6'
              task cfml
                label files ""
                label info ""
                event fdb
              task cfpl
                label files ""
                label info ""
                event fdb
              task cfsfc
                trigger ../../postproc/dp eq complete
                label files ""
                label info ""
                event fdb
              task cfil
                label files ""
                label info ""
                event fdb
              task cfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
              task derivedp
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            complete /emc_36r1/lag/ref/12/events:cf
            trigger ( /emc_36r1/main:YMD gt /emc_36r1/lag:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/lag:YMD and /emc_36r1/main/12/legA/fc/pf eq complete and /emc_36r1/lag/ref/12/events eq complete)
            edit EMOS_TYPE 'pf'
            family pf024
              edit STEPS '0/to/024/by/3'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf048
              edit STEPS '027/to/048/by/3'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf072
              edit STEPS '051/to/072/by/3'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf096
              edit STEPS '075/to/096/by/3'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf120
              edit STEPS '099/to/120/by/3'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf144
              edit STEPS '123/to/144/by/3'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf192
              edit STEPS '150/to/192/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf240
              edit STEPS '198/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
              task pfml
                edit STEPS '216'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family icp
            complete /emc_36r1/lag/ref/12/events:cf
            trigger ( /emc_36r1/main:YMD gt /emc_36r1/lag:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/lag:YMD and /emc_36r1/main/12/legA/fc eq complete)
            edit ARCHTYPE 'icp'
            task ml
              edit FCLENGTH '240'
              label files ""
              label info ""
              event fdb
          endfamily
          task flush
            trigger ( /emc_36r1/main:YMD gt /emc_36r1/lag:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/lag:YMD and /emc_36r1/main/12/legA/fc eq complete)
            edit STREAM 'enfo'
            label files ""
            label info ""
            event fdb
        endfamily
        family ecfs
          trigger ( /emc_36r1/main:YMD gt /emc_36r1/lag:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/lag:YMD and /emc_36r1/main/12/sv eq complete and /emc_36r1/main/12/legA/fc eq complete)
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit FSFAMILY 'mc'
          task svsave
          task save
        endfamily
      endfamily
      family legB
        complete /emc_36r1/lag/ref/12/events:240
        edit EPSLEG '2'
        edit EPSWAMNFRE '25'
        edit EPSWAMNANG '12'
        inlimit /emc_36r1/limits:mars
        inlimit /limits:mars
        family postproc
          trigger ( /emc_36r1/main:YMD gt /emc_36r1/lag:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/lag:YMD and /emc_36r1/main/12/legB/postproc eq complete) and /emc_36r1/lag/ref/12/events eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
          inlimit /emc_36r1/limits:postproc
          family ensms
            complete /emc_36r1/lag/ref/12/events:cf
            family 360
              edit STEPS '246/to/360/by/6'
              task mean
                edit TYPE 'em'
                label files ""
                label info ""
                event fdb
              task stdev
                edit TYPE 'es'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family prob
            complete /emc_36r1/lag/ref/12/events:cf
            family 2t
              edit PARAM '2t'
              task arc_prob
                label files ""
                label info ""
                event fdb
            endfamily
            family swp_mwp
              edit PARAM 'swp_mwp'
              task arc_prob
                label files ""
                label info ""
                event fdb
            endfamily
            family 10fg
              edit PARAM '10fg'
              task arc_prob
                label files ""
                label info ""
                event fdb
            endfamily
            family ff
              edit PARAM 'ff'
              task arc_prob
                label files ""
                label info ""
                event fdb
            endfamily
            family rain
              edit PARAM 'rain'
              task arc_prob
                label files ""
                label info ""
                event fdb
            endfamily
            family t850
              edit PARAM 't850'
              task arc_prob
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family dp
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
            family pf
              complete /emc_36r1/lag/ref/12/events:cf
              edit EMOS_TYPE 'pf'
              family 240
                edit STEPS '216/222/228/234/240'
                edit OVERLAP '1'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family 360
                edit STEPS '246/to/360/by/6'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
              endfamily
              task extra_arc
                edit STEPS '240/to/360'
                edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                edit RUNVARFC '1'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf
              edit EMOS_TYPE 'cf'
              family 240
                edit STEPS '216/222/228/234/240'
                edit OVERLAP '1'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family 360
                edit STEPS '246/to/360/by/6'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
              endfamily
              task extra_arc
                edit STEPS '240/to/360'
                edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                edit RUNVARFC '1'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family efi
            complete ( /emc_36r1/main/ref/12/events:cf) or ( /emc_36r1/main/ref/12/events:noefi)
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
            family 2t
              edit RF_NAME '2t'
              task efi
                label info ""
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task efi
                label info ""
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task efi
                label info ""
            endfamily
            family tp
              edit RF_NAME 'tp'
              task efi
                label info ""
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task efi
                label info ""
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task efi
                label info ""
            endfamily
            family ts
              edit RF_NAME 'ts'
              task efi
                label info ""
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task efi
                label info ""
            endfamily
          endfamily
        endfamily
        family cf
          trigger ( /emc_36r1/main:YMD gt /emc_36r1/lag:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/lag:YMD and /emc_36r1/main/12/legB/fc/cf eq complete)
          edit EMOS_TYPE 'cf'
          edit NUMBER '0'
          family cf240
            edit STEPS '216/228/240'
            edit OVERLAP '1'
            task cfml
              label files ""
              label info ""
              event fdb
            task cfpl
              label files ""
              label info ""
              event fdb
            task cfsfc
              label files ""
              label info ""
              event fdb
            task cfil
              label files ""
              label info ""
              event fdb
            task cfwv
              edit STREAM 'WAEF'
              edit EMOS_STREAM 'WAEF'
              label files ""
              label info ""
              event fdb
          endfamily
          family cf300
            edit STEPS '246/to/300/by/6'
            task cfml
              label files ""
              label info ""
              event fdb
            task cfpl
              label files ""
              label info ""
              event fdb
            task cfsfc
              label files ""
              label info ""
              event fdb
            task cfil
              label files ""
              label info ""
              event fdb
            task cfwv
              edit STREAM 'WAEF'
              edit EMOS_STREAM 'WAEF'
              label files ""
              label info ""
              event fdb
          endfamily
          family cf360
            edit STEPS '306/to/360/by/6'
            task cfml
              label files ""
              label info ""
              event fdb
            task cfpl
              label files ""
              label info ""
              event fdb
            task cfsfc
              label files ""
              label info ""
              event fdb
            task cfil
              label files ""
              label info ""
              event fdb
            task cfwv
              edit STREAM 'WAEF'
              edit EMOS_STREAM 'WAEF'
              label files ""
              label info ""
              event fdb
          endfamily
          task accum
            edit STEPS '240'
            edit OVERLAP '1'
            label files ""
            label info ""
            event fdb
        endfamily
        family pf
          complete /emc_36r1/lag/ref/12/events:cf
          trigger ( /emc_36r1/main:YMD gt /emc_36r1/lag:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/lag:YMD and /emc_36r1/main/12/legB/fc/pf eq complete and /emc_36r1/lag/ref/12/events eq complete)
          edit EMOS_TYPE 'pf'
          family pf240
            edit STEPS '216/228/240'
            edit OVERLAP '1'
            task pfpl
              inlimit /emc_36r1/limits:pl
              label files ""
              label info ""
              event fdb
            task pfsfc
              trigger ../../postproc/dp eq complete
              inlimit /emc_36r1/limits:sfc
              label files ""
              label info ""
              event fdb
            task pfil
              label files ""
              label info ""
              event fdb
            task pfwv
              edit STREAM 'WAEF'
              edit EMOS_STREAM 'WAEF'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf300
            edit STEPS '246/to/300/by/6'
            task pfpl
              inlimit /emc_36r1/limits:pl
              label files ""
              label info ""
              event fdb
            task pfsfc
              trigger ../../postproc/dp eq complete
              inlimit /emc_36r1/limits:sfc
              label files ""
              label info ""
              event fdb
            task pfil
              label files ""
              label info ""
              event fdb
            task pfwv
              edit STREAM 'WAEF'
              edit EMOS_STREAM 'WAEF'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf360
            edit STEPS '306/to/360/by/6'
            task pfpl
              inlimit /emc_36r1/limits:pl
              label files ""
              label info ""
              event fdb
            task pfsfc
              trigger ../../postproc/dp eq complete
              inlimit /emc_36r1/limits:sfc
              label files ""
              label info ""
              event fdb
            task pfil
              label files ""
              label info ""
              event fdb
            task pfwv
              edit STREAM 'WAEF'
              edit EMOS_STREAM 'WAEF'
              label files ""
              label info ""
              event fdb
          endfamily
          task pfml
            edit STEPS '360'
          task accum
            edit STEPS '240'
            edit OVERLAP '1'
            label files ""
            label info ""
            event fdb
        endfamily
      endfamily
    endfamily
    family 00
      trigger ( /emc_36r1/main:YMD gt /emc_36r1/lag:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/lag:YMD)
      edit DELTA_DAY '1'
      edit EMOS_BASE '00'
      edit QUEUE 'os'
      family legA
        trigger /emc_36r1/lag/ref/00/events eq complete
        edit EPSWAMNFRE '30'
        edit EPSWAMNANG '24'
        family init
          trigger ( /emc_36r1/main:YMD gt /emc_36r1/lag:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/lag:YMD and /emc_36r1/main/00/ref/an eq complete)
          edit FSFAMILY 'mc'
          task rmfdb
        endfamily
        family tc
          complete /emc_36r1/lag/ref/00/events:cf
          trigger ( /emc_36r1/main:YMD gt /emc_36r1/lag:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/lag:YMD and /emc_36r1/main/00/legA/postproc/tc eq complete)
          edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
          edit FSFAMILY 'mc'
          task clean
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/tc'
          task to_ecfs
            label files ""
            label info ""
            event fdb
          task to_mars
            trigger to_ecfs eq complete
        endfamily
        family archive
          inlimit /emc_36r1/limits:mars
          inlimit /limits:mars
          family postproc
            trigger ( /emc_36r1/main:YMD gt /emc_36r1/lag:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/lag:YMD and /emc_36r1/main/00/legA/postproc eq complete) and /emc_36r1/lag/ref/00/events eq complete
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            family dp
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family pf
                complete /emc_36r1/lag/ref/00/events:cf
                edit EMOS_TYPE 'pf'
                family 144
                  edit STEPS '0/to/144/by/3'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family cf
                edit EMOS_TYPE 'cf'
                family 144
                  edit STEPS '0/to/144/by/3'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                family 240
                  edit STEPS '150/to/240/by/6'
                  task derivedp
                    label files ""
                    label info ""
                    event fdb
                endfamily
                task extra_arc
                  edit STEPS '0/to/240'
                  edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                  edit RUNVARFC '1'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family cluster
              complete /emc_36r1/lag/ref/00/events:cf
              edit STEPS '72/84/96/108/120/132/144/156/168'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family a
                edit DOMAIN 'a'
                task cmpl
                  edit TYPE 'cm'
                  label files ""
                  label info ""
                  event fdb
                task cspl
                  edit TYPE 'cs'
                  label files ""
                  label info ""
                  event fdb
                task tube
                  edit STEPS '48/72/96/120/144/168/192/216/240'
                  edit REFERENCE '96/144/168/192/240'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family b
                trigger ./a eq complete
                edit DOMAIN 'b'
                task cmpl
                  edit TYPE 'cm'
                  label files ""
                  label info ""
                  event fdb
                task cspl
                  edit TYPE 'cs'
                  label files ""
                  label info ""
                  event fdb
                task tube
                  edit STEPS '48/72/96/120/144/168/192/216/240'
                  edit REFERENCE '96/144/168/192/240'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family c
                trigger ./b eq complete
                edit DOMAIN 'c'
                task cmpl
                  edit TYPE 'cm'
                  label files ""
                  label info ""
                  event fdb
                task cspl
                  edit TYPE 'cs'
                  label files ""
                  label info ""
                  event fdb
                task tube
                  edit STEPS '48/72/96/120/144/168/192/216/240'
                  edit REFERENCE '96/144/168/192/240'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family d
                trigger ./c eq complete
                edit DOMAIN 'd'
                task cmpl
                  edit TYPE 'cm'
                  label files ""
                  label info ""
                  event fdb
                task cspl
                  edit TYPE 'cs'
                  label files ""
                  label info ""
                  event fdb
                task tube
                  edit STEPS '48/72/96/120/144/168/192/216/240'
                  edit REFERENCE '96/144/168/192/240'
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family e
                trigger ./d eq complete
                edit DOMAIN 'e'
                task cmpl
                  edit TYPE 'cm'
                  label files ""
                  label info ""
                  event fdb
                task cspl
                  edit TYPE 'cs'
                  label files ""
                  label info ""
                  event fdb
                task tube
                  edit STEPS '48/72/96/120/144/168/192/216/240'
                  edit REFERENCE '96/144/168/192/240'
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family reord
              complete /emc_36r1/lag/ref/00/events:cf
              edit EPSCOUPLE '1'
              edit EMOS_TYPE 'pf'
              edit FCLENGTH0 '0'
              edit FCLENGTH '240'
              edit EPSFCRES '639'
              edit FCCHUNK '1'
              edit STEP1 '0'
              edit STEP2 '240'
              edit STEPS '-1'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/leg'
              limit reord 1
              inlimit reord
              label info ""
              family 130
                edit GENERATE '1'
                edit ARCHIVE '0'
                edit PARAMS '130.128'
                task reord
                  label info ""
              endfamily
              family 228
                edit GENERATE '1'
                edit ARCHIVE '0'
                edit PARAMS '228.128'
                task reord
                  label info ""
              endfamily
              family 207
                edit GENERATE '1'
                edit ARCHIVE '0'
                edit PARAMS '207.128'
                task reord
                  label info ""
              endfamily
              family 167
                edit GENERATE '1'
                edit ARCHIVE '0'
                edit PARAMS '167.128'
                task reord
                  label info ""
              endfamily
              family 164
                edit GENERATE '1'
                edit ARCHIVE '0'
                edit PARAMS '164.128'
                task reord
                  label info ""
              endfamily
              family arch
                trigger ./130 eq complete and ./228 eq complete and ./207 eq complete and ./167 eq complete and ./164 eq complete
                edit GENERATE '0'
                edit ARCHIVE '1'
                inlimit /emc_36r1/lag/ref/00:mutex
                task reord
                  label info ""
              endfamily
            endfamily
            family ensms
              complete /emc_36r1/lag/ref/00/events:cf
              edit STEPS '3/6/9/12/15/18/21/24/27/30/33/36/39/42/45/48/51/54/57/60/63/66/69/72/75/78/81/84/87/90/93/96/99/102/105/108/111/114/117/120/123/126/129/132/135/138/141/144/150/156/162/168/174/180/186/192/198/204/210/216/222/228/234/240'
              task mean
                edit TYPE 'em'
                label files ""
                label info ""
                event fdb
              task stdev
                edit TYPE 'es'
                label files ""
                label info ""
                event fdb
            endfamily
            family prob
              complete /emc_36r1/lag/ref/00/events:cf
              family 2t
                edit PARAM '2t'
                task arc_prob
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family swp_mwp
                edit PARAM 'swp_mwp'
                task arc_prob
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family 10fg
                edit PARAM '10fg'
                task arc_prob
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family ff
                edit PARAM 'ff'
                task arc_prob
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family rain
                edit PARAM 'rain'
                task arc_prob
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family t850
                edit PARAM 't850'
                task arc_prob
                  label files ""
                  label info ""
                  event fdb
              endfamily
            endfamily
            family efi
              complete ( /emc_36r1/main/ref/00/events:cf) or ( /emc_36r1/main/ref/00/events:noefi)
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/lag:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/lag:YMD and /emc_36r1/main/00/legA/postproc eq complete)
              edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
              family 2t
                edit RF_NAME '2t'
                task efi
                  label info ""
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task efi
                  label info ""
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task efi
                  label info ""
              endfamily
              family tp
                edit RF_NAME 'tp'
                task efi
                  label info ""
              endfamily
            endfamily
          endfamily
          family coupled
            complete ( /emc_36r1/lag/ref/00/events:cf or /emc_36r1/lag/ref/00/events:360) or ( /emc_36r1/lag/ref/00/events eq complete and not /emc_36r1/lag/ref/00/events:cpl)
            trigger ./cf eq complete and ./pf eq complete
            edit EPSLEG '1'
            edit OCLEG '1'
            edit OCEPS '1'
            edit FCCHUNK '1'
            edit FCLENGTH '240'
            edit FCLENGTH0 '0'
            edit OCEPSTOT '3'
            edit FCTOTAL '51'
            edit RUNVARFC '1'
            edit EPSCOUPLE '1'
            edit RUNCOUPLE '0'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
            inlimit /emc_36r1/lag/ref/00:mutex
            task preparcoce
            task arcfc
              trigger preparcoce eq complete
              label files ""
              label info ""
              event fdb
          endfamily
          family cf
            trigger ( /emc_36r1/main:YMD gt /emc_36r1/lag:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/lag:YMD and /emc_36r1/main/00/legA/fc/cf eq complete)
            edit EMOS_TYPE 'cf'
            edit NUMBER '0'
            family cf036
              edit STEPS '0/to/36/by/3'
              task cfml
                label files ""
                label info ""
                event fdb
              task cfpl
                label files ""
                label info ""
                event fdb
              task cfsfc
                label files ""
                label info ""
                event fdb
              task cfil
                label files ""
                label info ""
                event fdb
              task cfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf072
              edit STEPS '39/to/72/by/3'
              task cfml
                label files ""
                label info ""
                event fdb
              task cfpl
                label files ""
                label info ""
                event fdb
              task cfsfc
                label files ""
                label info ""
                event fdb
              task cfil
                label files ""
                label info ""
                event fdb
              task cfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf108
              edit STEPS '075/to/108/by/3'
              task cfml
                label files ""
                label info ""
                event fdb
              task cfpl
                label files ""
                label info ""
                event fdb
              task cfsfc
                label files ""
                label info ""
                event fdb
              task cfil
                label files ""
                label info ""
                event fdb
              task cfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf144
              edit STEPS '111/to/144/by/3'
              task cfml
                label files ""
                label info ""
                event fdb
              task cfpl
                label files ""
                label info ""
                event fdb
              task cfsfc
                label files ""
                label info ""
                event fdb
              task cfil
                label files ""
                label info ""
                event fdb
              task cfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf180
              edit STEPS '150/to/180/by/6'
              task cfml
                label files ""
                label info ""
                event fdb
              task cfpl
                label files ""
                label info ""
                event fdb
              task cfsfc
                label files ""
                label info ""
                event fdb
              task cfil
                label files ""
                label info ""
                event fdb
              task cfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf240
              edit STEPS '186/to/240/by/6'
              task cfml
                label files ""
                label info ""
                event fdb
              task cfpl
                label files ""
                label info ""
                event fdb
              task cfsfc
                label files ""
                label info ""
                event fdb
              task cfil
                label files ""
                label info ""
                event fdb
              task cfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family cv
            trigger ( /emc_36r1/main:YMD gt /emc_36r1/lag:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/lag:YMD and /emc_36r1/main/00/legA/fc/cv eq complete)
            edit EMOS_TYPE 'cv'
            edit NUMBER '1/2'
            family cv036
              edit STEPS '0/to/036/by/3'
              task cfml
                label files ""
                label info ""
                event fdb
              task cfpl
                label files ""
                label info ""
                event fdb
              task cfsfc
                trigger ../../postproc/dp eq complete
                label files ""
                label info ""
                event fdb
              task cfil
                label files ""
                label info ""
                event fdb
              task cfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
              task derivedp
                label files ""
                label info ""
                event fdb
            endfamily
            family cv072
              edit STEPS '039/to/072/by/3'
              task cfml
                label files ""
                label info ""
                event fdb
              task cfpl
                label files ""
                label info ""
                event fdb
              task cfsfc
                trigger ../../postproc/dp eq complete
                label files ""
                label info ""
                event fdb
              task cfil
                label files ""
                label info ""
                event fdb
              task cfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
              task derivedp
                label files ""
                label info ""
                event fdb
            endfamily
            family cv108
              edit STEPS '075/to/108/by/3'
              task cfml
                label files ""
                label info ""
                event fdb
              task cfpl
                label files ""
                label info ""
                event fdb
              task cfsfc
                trigger ../../postproc/dp eq complete
                label files ""
                label info ""
                event fdb
              task cfil
                label files ""
                label info ""
                event fdb
              task cfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
              task derivedp
                label files ""
                label info ""
                event fdb
            endfamily
            family cv144
              edit STEPS '111/to/144/by/3'
              task cfml
                label files ""
                label info ""
                event fdb
              task cfpl
                label files ""
                label info ""
                event fdb
              task cfsfc
                trigger ../../postproc/dp eq complete
                label files ""
                label info ""
                event fdb
              task cfil
                label files ""
                label info ""
                event fdb
              task cfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
              task derivedp
                label files ""
                label info ""
                event fdb
            endfamily
            family cv180
              edit STEPS '150/to/180/by/6'
              task cfml
                label files ""
                label info ""
                event fdb
              task cfpl
                label files ""
                label info ""
                event fdb
              task cfsfc
                trigger ../../postproc/dp eq complete
                label files ""
                label info ""
                event fdb
              task cfil
                label files ""
                label info ""
                event fdb
              task cfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
              task derivedp
                label files ""
                label info ""
                event fdb
            endfamily
            family cv240
              edit STEPS '186/to/240/by/6'
              task cfml
                label files ""
                label info ""
                event fdb
              task cfpl
                label files ""
                label info ""
                event fdb
              task cfsfc
                trigger ../../postproc/dp eq complete
                label files ""
                label info ""
                event fdb
              task cfil
                label files ""
                label info ""
                event fdb
              task cfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
              task derivedp
                label files ""
                label info ""
                event fdb
            endfamily
            family cv300
              edit STEPS '246/to/300/by/6'
              task cfml
                label files ""
                label info ""
                event fdb
              task cfpl
                label files ""
                label info ""
                event fdb
              task cfsfc
                trigger ../../postproc/dp eq complete
                label files ""
                label info ""
                event fdb
              task cfil
                label files ""
                label info ""
                event fdb
              task cfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
              task derivedp
                label files ""
                label info ""
                event fdb
            endfamily
            family cv360
              edit STEPS '306/to/360/by/6'
              task cfml
                label files ""
                label info ""
                event fdb
              task cfpl
                label files ""
                label info ""
                event fdb
              task cfsfc
                trigger ../../postproc/dp eq complete
                label files ""
                label info ""
                event fdb
              task cfil
                label files ""
                label info ""
                event fdb
              task cfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
              task derivedp
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family pf
            complete /emc_36r1/lag/ref/00/events:cf
            trigger ( /emc_36r1/main:YMD gt /emc_36r1/lag:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/lag:YMD and /emc_36r1/main/00/legA/fc/pf eq complete and /emc_36r1/lag/ref/00/events eq complete)
            edit EMOS_TYPE 'pf'
            family pf024
              edit STEPS '0/to/024/by/3'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf048
              edit STEPS '027/to/048/by/3'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf072
              edit STEPS '051/to/072/by/3'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf096
              edit STEPS '075/to/096/by/3'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf120
              edit STEPS '099/to/120/by/3'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf144
              edit STEPS '123/to/144/by/3'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf192
              edit STEPS '150/to/192/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
            endfamily
            family pf240
              edit STEPS '198/to/240/by/6'
              task pfpl
                inlimit /emc_36r1/limits:pl
                label files ""
                label info ""
                event fdb
              task pfsfc
                trigger ../../postproc/dp eq complete
                inlimit /emc_36r1/limits:sfc
                label files ""
                label info ""
                event fdb
              task pfil
                label files ""
                label info ""
                event fdb
              task pfwv
                edit STREAM 'WAEF'
                edit EMOS_STREAM 'WAEF'
                label files ""
                label info ""
                event fdb
              task pfml
                edit STEPS '216'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family icp
            complete /emc_36r1/lag/ref/00/events:cf
            trigger ( /emc_36r1/main:YMD gt /emc_36r1/lag:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/lag:YMD and /emc_36r1/main/00/legA/fc eq complete)
            edit ARCHTYPE 'icp'
            task ml
              edit FCLENGTH '240'
              label files ""
              label info ""
              event fdb
          endfamily
          task flush
            trigger ( /emc_36r1/main:YMD gt /emc_36r1/lag:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/lag:YMD and /emc_36r1/main/00/legA/fc eq complete)
            edit STREAM 'enfo'
            label files ""
            label info ""
            event fdb
        endfamily
        family ecfs
          trigger ( /emc_36r1/main:YMD gt /emc_36r1/lag:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/lag:YMD and /emc_36r1/main/00/sv eq complete and /emc_36r1/main/00/legA/fc eq complete)
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/sv'
          edit FSFAMILY 'mc'
          task svsave
          task save
        endfamily
      endfamily
      family legB
        complete /emc_36r1/lag/ref/00/events:240
        edit EPSLEG '2'
        edit EPSWAMNFRE '25'
        edit EPSWAMNANG '12'
        inlimit /emc_36r1/limits:mars
        inlimit /limits:mars
        family postproc
          trigger ( /emc_36r1/main:YMD gt /emc_36r1/lag:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/lag:YMD and /emc_36r1/main/00/legB/postproc eq complete) and /emc_36r1/lag/ref/00/events eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
          inlimit /emc_36r1/limits:postproc
          family ensms
            complete /emc_36r1/lag/ref/00/events:cf
            family 360
              edit STEPS '246/to/360/by/6'
              task mean
                edit TYPE 'em'
                label files ""
                label info ""
                event fdb
              task stdev
                edit TYPE 'es'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family prob
            complete /emc_36r1/lag/ref/00/events:cf
            family 2t
              edit PARAM '2t'
              task arc_prob
                label files ""
                label info ""
                event fdb
            endfamily
            family swp_mwp
              edit PARAM 'swp_mwp'
              task arc_prob
                label files ""
                label info ""
                event fdb
            endfamily
            family 10fg
              edit PARAM '10fg'
              task arc_prob
                label files ""
                label info ""
                event fdb
            endfamily
            family ff
              edit PARAM 'ff'
              task arc_prob
                label files ""
                label info ""
                event fdb
            endfamily
            family rain
              edit PARAM 'rain'
              task arc_prob
                label files ""
                label info ""
                event fdb
            endfamily
            family t850
              edit PARAM 't850'
              task arc_prob
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family dp
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
            family pf
              complete /emc_36r1/lag/ref/00/events:cf
              edit EMOS_TYPE 'pf'
              family 240
                edit STEPS '216/222/228/234/240'
                edit OVERLAP '1'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family 360
                edit STEPS '246/to/360/by/6'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
              endfamily
              task extra_arc
                edit STEPS '240/to/360'
                edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                edit RUNVARFC '1'
                label files ""
                label info ""
                event fdb
            endfamily
            family cf
              edit EMOS_TYPE 'cf'
              family 240
                edit STEPS '216/222/228/234/240'
                edit OVERLAP '1'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
              endfamily
              family 360
                edit STEPS '246/to/360/by/6'
                task derivedp
                  label files ""
                  label info ""
                  event fdb
              endfamily
              task extra_arc
                edit STEPS '240/to/360'
                edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
                edit RUNVARFC '1'
                label files ""
                label info ""
                event fdb
            endfamily
          endfamily
          family efi
            complete ( /emc_36r1/main/ref/00/events:cf) or ( /emc_36r1/main/ref/00/events:noefi)
            edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
            family 2t
              edit RF_NAME '2t'
              task efi
                label info ""
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task efi
                label info ""
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task efi
                label info ""
            endfamily
            family tp
              edit RF_NAME 'tp'
              task efi
                label info ""
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task efi
                label info ""
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task efi
                label info ""
            endfamily
            family ts
              edit RF_NAME 'ts'
              task efi
                label info ""
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task efi
                label info ""
            endfamily
          endfamily
          family reord
            complete /emc_36r1/lag/ref/00/events:cf
            edit EPSCOUPLE '1'
            edit EMOS_TYPE 'pf'
            edit FCLENGTH0 '240'
            edit FCLENGTH '360'
            edit EPSFCRES '639'
            edit FCCHUNK '1'
            edit STEP1 '240'
            edit STEP2 '360'
            edit STEPS '-1'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/leg'
            limit reord 1
            inlimit reord
            label info ""
            family 130
              edit GENERATE '1'
              edit ARCHIVE '0'
              edit PARAMS '130.128'
              task reord
                label info ""
            endfamily
            family 228
              edit GENERATE '1'
              edit ARCHIVE '0'
              edit PARAMS '228.128'
              task reord
                label info ""
            endfamily
            family 207
              edit GENERATE '1'
              edit ARCHIVE '0'
              edit PARAMS '207.128'
              task reord
                label info ""
            endfamily
            family 167
              edit GENERATE '1'
              edit ARCHIVE '0'
              edit PARAMS '167.128'
              task reord
                label info ""
            endfamily
            family 164
              edit GENERATE '1'
              edit ARCHIVE '0'
              edit PARAMS '164.128'
              task reord
                label info ""
            endfamily
            family arch
              trigger ./130 eq complete and ./228 eq complete and ./207 eq complete and ./167 eq complete and ./164 eq complete
              edit GENERATE '0'
              edit ARCHIVE '1'
              inlimit /emc_36r1/lag/ref/00:mutex
              task reord
                label info ""
            endfamily
          endfamily
        endfamily
        family cf
          trigger ( /emc_36r1/main:YMD gt /emc_36r1/lag:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/lag:YMD and /emc_36r1/main/00/legB/fc/cf eq complete)
          edit EMOS_TYPE 'cf'
          edit NUMBER '0'
          family cf240
            edit STEPS '216/228/240'
            edit OVERLAP '1'
            task cfml
              label files ""
              label info ""
              event fdb
            task cfpl
              label files ""
              label info ""
              event fdb
            task cfsfc
              label files ""
              label info ""
              event fdb
            task cfil
              label files ""
              label info ""
              event fdb
            task cfwv
              edit STREAM 'WAEF'
              edit EMOS_STREAM 'WAEF'
              label files ""
              label info ""
              event fdb
          endfamily
          family cf300
            edit STEPS '246/to/300/by/6'
            task cfml
              label files ""
              label info ""
              event fdb
            task cfpl
              label files ""
              label info ""
              event fdb
            task cfsfc
              label files ""
              label info ""
              event fdb
            task cfil
              label files ""
              label info ""
              event fdb
            task cfwv
              edit STREAM 'WAEF'
              edit EMOS_STREAM 'WAEF'
              label files ""
              label info ""
              event fdb
          endfamily
          family cf360
            edit STEPS '306/to/360/by/6'
            task cfml
              label files ""
              label info ""
              event fdb
            task cfpl
              label files ""
              label info ""
              event fdb
            task cfsfc
              label files ""
              label info ""
              event fdb
            task cfil
              label files ""
              label info ""
              event fdb
            task cfwv
              edit STREAM 'WAEF'
              edit EMOS_STREAM 'WAEF'
              label files ""
              label info ""
              event fdb
          endfamily
          task accum
            edit STEPS '240'
            edit OVERLAP '1'
            label files ""
            label info ""
            event fdb
        endfamily
        family pf
          complete /emc_36r1/lag/ref/00/events:cf
          trigger ( /emc_36r1/main:YMD gt /emc_36r1/lag:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/lag:YMD and /emc_36r1/main/00/legB/fc/pf eq complete and /emc_36r1/lag/ref/00/events eq complete)
          edit EMOS_TYPE 'pf'
          family pf240
            edit STEPS '216/228/240'
            edit OVERLAP '1'
            task pfpl
              inlimit /emc_36r1/limits:pl
              label files ""
              label info ""
              event fdb
            task pfsfc
              trigger ../../postproc/dp eq complete
              inlimit /emc_36r1/limits:sfc
              label files ""
              label info ""
              event fdb
            task pfil
              label files ""
              label info ""
              event fdb
            task pfwv
              edit STREAM 'WAEF'
              edit EMOS_STREAM 'WAEF'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf300
            edit STEPS '246/to/300/by/6'
            task pfpl
              inlimit /emc_36r1/limits:pl
              label files ""
              label info ""
              event fdb
            task pfsfc
              trigger ../../postproc/dp eq complete
              inlimit /emc_36r1/limits:sfc
              label files ""
              label info ""
              event fdb
            task pfil
              label files ""
              label info ""
              event fdb
            task pfwv
              edit STREAM 'WAEF'
              edit EMOS_STREAM 'WAEF'
              label files ""
              label info ""
              event fdb
          endfamily
          family pf360
            edit STEPS '306/to/360/by/6'
            task pfpl
              inlimit /emc_36r1/limits:pl
              label files ""
              label info ""
              event fdb
            task pfsfc
              trigger ../../postproc/dp eq complete
              inlimit /emc_36r1/limits:sfc
              label files ""
              label info ""
              event fdb
            task pfil
              label files ""
              label info ""
              event fdb
            task pfwv
              edit STREAM 'WAEF'
              edit EMOS_STREAM 'WAEF'
              label files ""
              label info ""
              event fdb
          endfamily
          task pfml
            edit STEPS '360'
          task accum
            edit STEPS '240'
            edit OVERLAP '1'
            label files ""
            label info ""
            event fdb
        endfamily
        family coupled
          complete /emc_36r1/lag/ref/00/events:cf or /emc_36r1/lag/ref/00/events:360 or ( /emc_36r1/lag/ref/00/events eq complete and not /emc_36r1/lag/ref/00/events:cpl)
          trigger /emc_36r1/lag/00/legA/archive/coupled eq complete
          edit EPSLEG '2'
          edit OCLEG '2'
          edit OCEPS '2'
          edit FCCHUNK '1'
          edit FCLENGTH '360'
          edit FCLENGTH0 '216'
          edit OCEPSTOT '3'
          edit FCTOTAL '51'
          edit RUNVARFC '1'
          edit EPSCOUPLE '1'
          edit RUNCOUPLE '1'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
          inlimit /emc_36r1/lag/ref/00:mutex
          task preparcoce
          task arcfc
            trigger preparcoce eq complete
            label files ""
            label info ""
            event fdb
        endfamily
      endfamily
    endfamily
    task logfilestore
      edit QUEUE 'emos-e'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit KEEPLOGS 'no'
      edit LOGSTOECFS 'no'
      edit MOVE_TO_ECFS 'yes'
      edit LOGTASK 'lag'
      edit SMSFILES '/home/ma/emos/def/emc_36r1'
    task logfiles
      late -s +00:05 -c +00:35
      trigger /emc_36r1/main/make eq complete and ./12 eq complete and ./00 eq complete and logfilestore eq complete and ./ref/loop eq complete
      edit QUEUE 'emos-e'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit MOVE_TO_ECFS 'no'
      edit LOGTASK 'lag'
      edit SMSFILES '/home/ma/emos/def/emc_36r1'
  endfamily
  family pop
    repeat date YMD 20091003 20121212 1
    edit SMSURLBASE 'http://intra.ecmwf.int/plots/d/intra/'
    edit SMSURL 'sub=fixed/'
    edit VERSION '0046'
    edit STREAM 'EF'
    edit EMOS_STREAM 'EF'
    edit SMSFILES '/home/ma/emos/def/emc_36r1'
    edit LOGTASK 'pop'
    edit SUITE_START '20091124'
    family 12
      trigger ( /emc_36r1/main:YMD gt /emc_36r1/pop:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/pop:YMD and /emc_36r1/main/12/legA/fc/pf eq complete and /emc_36r1/main/12/ref/model eq complete)
      edit DELTA_DAY '0'
      edit EMOS_BASE '12'
      edit SMSFILES '/home/ma/emos/def/emc_36r1/pop'
      task events
        edit QUEUE 'emos-e'
        edit QUEUE_EPILOG 'emos-exp'
        edit WSHOST 'localhost'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        edit SMSFILES '/home/ma/emos/def/emc_36r1/ref'
        edit HINDCAST '1'
        edit ONLY_CF '0'
        event noefi
        event 240 
        event 360 
        event cpl
        event 768 
        event cf
      family data
        trigger events eq complete
        family metgram
          trigger ( /emc_36r1/main:YMD gt /emc_36r1/pop:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/pop:YMD and /emc_36r1/main/12/legA/fc/pf eq complete)
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit QUEUE 'emos'
          edit SMSTRIES '1'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/pop/data/epsgrams/grams'
          inlimit /limits:linux_cluster
          family 240
            trigger ( /emc_36r1/main:YMD gt /emc_36r1/pop:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/pop:YMD and /emc_36r1/main/12/legA/fc/pf eq complete and /emc_36r1/main/12/legA/fc/cf eq complete)
            edit KIND '10_days'
            edit RHOST 'epsgrams'
            edit RDIR '/webapps/epsgrams/data/'
            task gclean
              edit KEEP_DAYS '2'
              edit REMOTE_KEEP_DAYS '2'
            task gdb
              trigger gclean eq complete
              edit NEW_DAILY '1'
              label files ""
              label info ""
              label local "-"
              label remote "-"
              event fdb
              event 1 
            task glink
              trigger gdb eq complete
          endfamily
          family 360
            trigger ( /emc_36r1/main:YMD gt /emc_36r1/pop:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/pop:YMD and /emc_36r1/main/12/legB/postproc eq complete and /emc_36r1/main/12/legA/postproc eq complete)
            edit KIND '15_days'
            task gclean
              edit KEEP_DAYS '2'
              edit REMOTE_KEEP_DAYS '2'
            task gdb
              trigger gclean eq complete
              edit NEW_DAILY '1'
              label files ""
              label info ""
              label local "-"
              label remote "-"
              event fdb
              event 1 
            task glink
              trigger gdb eq complete
          endfamily
          family wave
            trigger ( /emc_36r1/main:YMD gt /emc_36r1/pop:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/pop:YMD and /emc_36r1/main/12/legA/fc/pf eq complete and /emc_36r1/main/12/legA/fc/cf eq complete)
            edit KIND '10_days_wave'
            edit RHOST 'epsgrams'
            edit RDIR '/webapps/epsgrams/data/'
            task gclean
              edit KEEP_DAYS '2'
              edit REMOTE_KEEP_DAYS '2'
            task gdb
              trigger gclean eq complete
              edit NEW_DAILY '1'
              label files ""
              label info ""
              label local "-"
              label remote "-"
              event fdb
              event 1 
            task glink
              trigger gdb eq complete
          endfamily
          family 15d
            trigger ( /emc_36r1/main:YMD gt /emc_36r1/pop:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/pop:YMD and /emc_36r1/main/12/legB/postproc eq complete and /emc_36r1/main/12/legA/postproc eq complete)
            edit KIND '15_days_pf'
            edit RHOST 'epsgrams'
            edit RDIR '/webapps/epsgrams/data/'
            task gclean
              edit KEEP_DAYS '2'
              edit REMOTE_KEEP_DAYS '2'
            task gdb
              trigger gclean eq complete
              edit NEW_DAILY '1'
              label files ""
              label info ""
              label local "-"
              label remote "-"
              event fdb
              event 1 
            task glink
              trigger gdb eq complete
          endfamily
          family efi
            trigger ( /emc_36r1/main:YMD gt /emc_36r1/pop:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/pop:YMD and /emc_36r1/main/12/legB/postproc eq complete and /emc_36r1/main/12/legA/postproc eq complete)
            edit KIND 'efi'
            edit RHOST 'epsgrams'
            edit RDIR '/webapps/epsgrams/data/'
            task gclean
              edit KEEP_DAYS '10'
              edit REMOTE_KEEP_DAYS '10'
            task gdb
              trigger gclean eq complete
              edit NEW_DAILY '1'
              label files ""
              label info ""
              label local "-"
              label remote "-"
              event fdb
              event 1 
            task glink
              trigger gdb eq complete
          endfamily
        endfamily
        family distribution
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/pop/data'
          edit DISTRIB_FDB '/leda/fdb1'
          inlimit /limits:linux_cluster
          task ff
            edit PARAMETER 'ff'
            edit STEP_LIST '84 108 132 156 180'
            edit RANK '9999'
          task 2t
            edit PARAMETER '2t'
            edit STEP_LIST '84 108 132 156 180'
            edit RANK '9998'
          task tp
            edit PARAMETER 'tp'
            edit STEP_LIST '102 126 150 174 198'
            edit RANK '9997'
          task tp5
            edit PARAMETER 'tp5'
            edit STEP_LIST '174'
            edit RANK '9996'
          task tp10
            edit PARAMETER 'tp10'
            edit STEP_LIST '240'
            edit RANK '9995'
          task cleanup
            trigger ff eq complete and ./2t eq complete and tp eq complete and tp5 eq complete and tp10 eq complete
        endfamily
        family fields
          trigger ( /emc_36r1/main:YMD gt /emc_36r1/pop:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/pop:YMD and /emc_36r1/main/12/legB/postproc eq complete and /emc_36r1/main/12/legA/postproc eq complete)
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit WSHOST 'linux_cluster'
          edit QUEUE 'plots-medium'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/pop/data/fields'
          inlimit /limits:linux_cluster
          inlimit /emc_36r1/limits:webplots
          task efi
            complete ( /emc_36r1/pop/12/events:cf) or ( /emc_36r1/pop/12/events:noefi)
            trigger /emc_36r1/pop/12/events eq complete
            event none
          task globalefi
            complete ( /emc_36r1/pop/12/events:cf) or ( /emc_36r1/pop/12/events:noefi)
            trigger /emc_36r1/pop/12/events eq complete
            event none
          family legA
            task proba
              complete /emc_36r1/pop/12/events:cf
              trigger /emc_36r1/pop/12/events eq complete
              edit RANGE 'legA'
              edit PLOT_STEPS_RANGE '00-24/12-36/24-48/36-60/48-72/60-84/72-96/84-108/96-120/108-132/120-144/132-156/144-168/156-180/168-192/180-204/192-216/204-228/216-240'
              edit PLOT_STEPS '12/to/240/by/12'
          endfamily
          family legB
            task proba
              complete /emc_36r1/pop/12/events:cf
              trigger /emc_36r1/pop/12/events eq complete
              edit RANGE 'legB'
              edit PLOT_STEPS '252/to/360/by/12'
              edit PLOT_STEPS_RANGE '228-252/240-264/252-276/264-288/276-300/288-312/300-324/312-336/324-348/336-360'
          endfamily
          task proba_mean
            complete /emc_36r1/pop/12/events:cf
            trigger /emc_36r1/pop/12/events eq complete
            edit PLOT_STEPS_MEAN '240-360'
          task lsmoro
            edit STEP_LIST '0/to/144/by/12'
        endfamily
        family tc
          complete /emc_36r1/pop/12/events:cf
          trigger ( /emc_36r1/pop/12/events eq complete) and ( ( /emc_36r1/lag:YMD gt /emc_36r1/pop:YMD) or ( /emc_36r1/lag:YMD eq /emc_36r1/pop:YMD and /emc_36r1/lag/12/legA/tc eq complete))
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/tc'
          inlimit /limits:linux_cluster
          task cleanup
          task check_data
          family post_process
            task strike_prob
              trigger ../check_data eq complete
          endfamily
          family diagnostics
            trigger post_process eq complete
            task det_errors
            task spreadskill
            task reliability
          endfamily
          family store
            trigger ( /emc_36r1/pop/12/data/tc/diagnostics eq complete)
            task to_ecfs_pop
          endfamily
        endfamily
        family tcyc
          complete /emc_36r1/pop/12/events:cf
          trigger ( /emc_36r1/main:YMD gt /emc_36r1/pop:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/pop:YMD and /emc_36r1/main/12/legB/fc eq complete)
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit FCCHUNK '1'
          edit OCTOTAL '5'
          edit FCTOTAL '51'
          edit FCLENGTH1 '216'
          edit DAILY '0'
          edit RUNVARFC '1'
          edit FCLENGTH '360'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
          edit EPSCOUPLE '1'
          edit RADIUS '500'
          limit semaphore 8
          inlimit /limits:linux_cluster
          inlimit semaphore
          task genes_stage
            edit DELTA_DAY '1'
          family genesis
            trigger genes_stage eq complete
            family da
              edit OCGROUP 'an0'
              edit FCGROUP 'fc0'
              edit EPSMEMBER 'da'
              edit MEMBER 'da'
              task tcyc
            endfamily
            family 00
              edit OCGROUP 'an00'
              edit FCGROUP 'fc0'
              edit EPSMEMBER '00'
              edit MEMBER '00'
              task tcyc
            endfamily
            family 01
              edit OCGROUP 'an1'
              edit FCGROUP 'fc0'
              edit EPSMEMBER '01'
              edit MEMBER '01'
              task tcyc
            endfamily
            family 02
              edit OCGROUP 'an2'
              edit FCGROUP 'fc0'
              edit EPSMEMBER '02'
              edit MEMBER '02'
              task tcyc
            endfamily
            family 03
              edit OCGROUP 'an3'
              edit FCGROUP 'fc0'
              edit EPSMEMBER '03'
              edit MEMBER '03'
              task tcyc
            endfamily
            family 04
              edit OCGROUP 'an4'
              edit FCGROUP 'fc0'
              edit EPSMEMBER '04'
              edit MEMBER '04'
              task tcyc
            endfamily
            family 05
              edit OCGROUP 'an0'
              edit FCGROUP 'fc1'
              edit EPSMEMBER '05'
              edit MEMBER '05'
              task tcyc
            endfamily
            family 06
              edit OCGROUP 'an1'
              edit FCGROUP 'fc1'
              edit EPSMEMBER '06'
              edit MEMBER '06'
              task tcyc
            endfamily
            family 07
              edit OCGROUP 'an2'
              edit FCGROUP 'fc1'
              edit EPSMEMBER '07'
              edit MEMBER '07'
              task tcyc
            endfamily
            family 08
              edit OCGROUP 'an3'
              edit FCGROUP 'fc1'
              edit EPSMEMBER '08'
              edit MEMBER '08'
              task tcyc
            endfamily
            family 09
              edit OCGROUP 'an4'
              edit FCGROUP 'fc1'
              edit EPSMEMBER '09'
              edit MEMBER '09'
              task tcyc
            endfamily
            family 10
              edit OCGROUP 'an0'
              edit FCGROUP 'fc2'
              edit EPSMEMBER '10'
              edit MEMBER '10'
              task tcyc
            endfamily
            family 11
              edit OCGROUP 'an1'
              edit FCGROUP 'fc2'
              edit EPSMEMBER '11'
              edit MEMBER '11'
              task tcyc
            endfamily
            family 12
              edit OCGROUP 'an2'
              edit FCGROUP 'fc2'
              edit EPSMEMBER '12'
              edit MEMBER '12'
              task tcyc
            endfamily
            family 13
              edit OCGROUP 'an3'
              edit FCGROUP 'fc2'
              edit EPSMEMBER '13'
              edit MEMBER '13'
              task tcyc
            endfamily
            family 14
              edit OCGROUP 'an4'
              edit FCGROUP 'fc2'
              edit EPSMEMBER '14'
              edit MEMBER '14'
              task tcyc
            endfamily
            family 15
              edit OCGROUP 'an0'
              edit FCGROUP 'fc3'
              edit EPSMEMBER '15'
              edit MEMBER '15'
              task tcyc
            endfamily
            family 16
              edit OCGROUP 'an1'
              edit FCGROUP 'fc3'
              edit EPSMEMBER '16'
              edit MEMBER '16'
              task tcyc
            endfamily
            family 17
              edit OCGROUP 'an2'
              edit FCGROUP 'fc3'
              edit EPSMEMBER '17'
              edit MEMBER '17'
              task tcyc
            endfamily
            family 18
              edit OCGROUP 'an3'
              edit FCGROUP 'fc3'
              edit EPSMEMBER '18'
              edit MEMBER '18'
              task tcyc
            endfamily
            family 19
              edit OCGROUP 'an4'
              edit FCGROUP 'fc3'
              edit EPSMEMBER '19'
              edit MEMBER '19'
              task tcyc
            endfamily
            family 20
              edit OCGROUP 'an0'
              edit FCGROUP 'fc4'
              edit EPSMEMBER '20'
              edit MEMBER '20'
              task tcyc
            endfamily
            family 21
              edit OCGROUP 'an1'
              edit FCGROUP 'fc4'
              edit EPSMEMBER '21'
              edit MEMBER '21'
              task tcyc
            endfamily
            family 22
              edit OCGROUP 'an2'
              edit FCGROUP 'fc4'
              edit EPSMEMBER '22'
              edit MEMBER '22'
              task tcyc
            endfamily
            family 23
              edit OCGROUP 'an3'
              edit FCGROUP 'fc4'
              edit EPSMEMBER '23'
              edit MEMBER '23'
              task tcyc
            endfamily
            family 24
              edit OCGROUP 'an4'
              edit FCGROUP 'fc4'
              edit EPSMEMBER '24'
              edit MEMBER '24'
              task tcyc
            endfamily
            family 25
              edit OCGROUP 'an0'
              edit FCGROUP 'fc5'
              edit EPSMEMBER '25'
              edit MEMBER '25'
              task tcyc
            endfamily
            family 26
              edit OCGROUP 'an1'
              edit FCGROUP 'fc5'
              edit EPSMEMBER '26'
              edit MEMBER '26'
              task tcyc
            endfamily
            family 27
              edit OCGROUP 'an2'
              edit FCGROUP 'fc5'
              edit EPSMEMBER '27'
              edit MEMBER '27'
              task tcyc
            endfamily
            family 28
              edit OCGROUP 'an3'
              edit FCGROUP 'fc5'
              edit EPSMEMBER '28'
              edit MEMBER '28'
              task tcyc
            endfamily
            family 29
              edit OCGROUP 'an4'
              edit FCGROUP 'fc5'
              edit EPSMEMBER '29'
              edit MEMBER '29'
              task tcyc
            endfamily
            family 30
              edit OCGROUP 'an0'
              edit FCGROUP 'fc6'
              edit EPSMEMBER '30'
              edit MEMBER '30'
              task tcyc
            endfamily
            family 31
              edit OCGROUP 'an1'
              edit FCGROUP 'fc6'
              edit EPSMEMBER '31'
              edit MEMBER '31'
              task tcyc
            endfamily
            family 32
              edit OCGROUP 'an2'
              edit FCGROUP 'fc6'
              edit EPSMEMBER '32'
              edit MEMBER '32'
              task tcyc
            endfamily
            family 33
              edit OCGROUP 'an3'
              edit FCGROUP 'fc6'
              edit EPSMEMBER '33'
              edit MEMBER '33'
              task tcyc
            endfamily
            family 34
              edit OCGROUP 'an4'
              edit FCGROUP 'fc6'
              edit EPSMEMBER '34'
              edit MEMBER '34'
              task tcyc
            endfamily
            family 35
              edit OCGROUP 'an0'
              edit FCGROUP 'fc7'
              edit EPSMEMBER '35'
              edit MEMBER '35'
              task tcyc
            endfamily
            family 36
              edit OCGROUP 'an1'
              edit FCGROUP 'fc7'
              edit EPSMEMBER '36'
              edit MEMBER '36'
              task tcyc
            endfamily
            family 37
              edit OCGROUP 'an2'
              edit FCGROUP 'fc7'
              edit EPSMEMBER '37'
              edit MEMBER '37'
              task tcyc
            endfamily
            family 38
              edit OCGROUP 'an3'
              edit FCGROUP 'fc7'
              edit EPSMEMBER '38'
              edit MEMBER '38'
              task tcyc
            endfamily
            family 39
              edit OCGROUP 'an4'
              edit FCGROUP 'fc7'
              edit EPSMEMBER '39'
              edit MEMBER '39'
              task tcyc
            endfamily
            family 40
              edit OCGROUP 'an0'
              edit FCGROUP 'fc8'
              edit EPSMEMBER '40'
              edit MEMBER '40'
              task tcyc
            endfamily
            family 41
              edit OCGROUP 'an1'
              edit FCGROUP 'fc8'
              edit EPSMEMBER '41'
              edit MEMBER '41'
              task tcyc
            endfamily
            family 42
              edit OCGROUP 'an2'
              edit FCGROUP 'fc8'
              edit EPSMEMBER '42'
              edit MEMBER '42'
              task tcyc
            endfamily
            family 43
              edit OCGROUP 'an3'
              edit FCGROUP 'fc8'
              edit EPSMEMBER '43'
              edit MEMBER '43'
              task tcyc
            endfamily
            family 44
              edit OCGROUP 'an4'
              edit FCGROUP 'fc8'
              edit EPSMEMBER '44'
              edit MEMBER '44'
              task tcyc
            endfamily
            family 45
              edit OCGROUP 'an0'
              edit FCGROUP 'fc9'
              edit EPSMEMBER '45'
              edit MEMBER '45'
              task tcyc
            endfamily
            family 46
              edit OCGROUP 'an1'
              edit FCGROUP 'fc9'
              edit EPSMEMBER '46'
              edit MEMBER '46'
              task tcyc
            endfamily
            family 47
              edit OCGROUP 'an2'
              edit FCGROUP 'fc9'
              edit EPSMEMBER '47'
              edit MEMBER '47'
              task tcyc
            endfamily
            family 48
              edit OCGROUP 'an3'
              edit FCGROUP 'fc9'
              edit EPSMEMBER '48'
              edit MEMBER '48'
              task tcyc
            endfamily
            family 49
              edit OCGROUP 'an4'
              edit FCGROUP 'fc9'
              edit EPSMEMBER '49'
              edit MEMBER '49'
              task tcyc
            endfamily
            family 50
              edit OCGROUP 'an0'
              edit FCGROUP 'fc10'
              edit EPSMEMBER '50'
              edit MEMBER '50'
              task tcyc
            endfamily
          endfamily
          task tcyctoecfs
            trigger genesis eq complete
        endfamily
      endfamily
      family web
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        edit WSHOST 'linux_cluster'
        edit QUEUE 'plots-medium'
        edit SMSFILES '/home/ma/emos/def/emc_36r1/pop/web'
        inlimit /limits:linux_cluster
        inlimit /emc_36r1/limits:webplots
        family pf
          complete /emc_36r1/pop/12/events:cf
          trigger /emc_36r1/pop/12/events eq complete
          task plumes
            trigger ( /emc_36r1/main:YMD gt /emc_36r1/pop:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/pop:YMD and /emc_36r1/main/12/legA/fc/cf eq complete and /emc_36r1/main/12/legA/fc/pf eq complete)
          family classic
            trigger ( /emc_36r1/main:YMD gt /emc_36r1/pop:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/pop:YMD and /emc_36r1/main/12/legA/postproc/clusters eq complete)
            edit ANALYSIS 'NO'
            task clplot
            task pltube
            task enplot
              edit STEPS '12 24 36 48 60 72 84 96 108 120 132 144 156 168 180 192'
              edit PRINTSTEPS '60 84 108 132 156 180'
          endfamily
          family efi
            complete /emc_36r1/pop/12/events:noefi
            trigger ../../data/fields eq complete
            edit QUEUE_EPILOG 'emos-exp'
            edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
            edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
            edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
            edit LOGDIR '/vol/emos_nc/output'
            edit SMSOUT '/vol/emos_nc/output'
            edit WSHOST 'linux_cluster'
            edit QUEUE 'plots-medium'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/pop/web'
            inlimit /limits:linux_cluster
            inlimit /emc_36r1/limits:webplots
            family efi_tpi
              task w_efi
                edit PARAMETER 'tpi'
            endfamily
            family efi_10wsi
              task w_efi
                edit PARAMETER '10wsi'
            endfamily
            family efi_2ti
              task w_efi
                edit PARAMETER '2ti'
            endfamily
            family efi_10fgi
              task w_efi
                edit PARAMETER '10fgi'
            endfamily
            task globalefi
              edit UPDATE_WEB_NOW 'yes'
          endfamily
          family proba
            trigger ../../data/fields eq complete
            edit QUEUE_EPILOG 'emos-exp'
            edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
            edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
            edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
            edit LOGDIR '/vol/emos_nc/output'
            edit SMSOUT '/vol/emos_nc/output'
            edit WSHOST 'linux_cluster'
            edit QUEUE 'plots-medium'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/pop/web'
            edit PLOT_AREAS 'Europe,North_America,Asia,South_Atlantic,South_Pacific,Australia,Africa,Global'
            inlimit /limits:linux_cluster
            inlimit /emc_36r1/limits:webplots
            family members
              family legA
                edit PLOT_STEPS '12/to/240/by/12'
                edit STEP_RANGE 'legA'
                task pb_2t
                task pb_ff
                task pb_tp
                  edit PLOT_STEPS '00-24/12-36/24-48/36-60/48-72/60-84/72-96/84-108/96-120/108-132/120-144/132-156/144-168/156-180/168-192/180-204/192-216/204-228/216-240'
                task w_proba_gusts
                  edit PLOT_STEPS '00-24/12-36/24-48/36-60/48-72/60-84/72-96/84-108/96-120/108-132/120-144/132-156/144-168/156-180/168-192/180-204/192-216/204-228/216-240'
                task pb_tpr
                task pb_mwp
                  edit PLOT_AREAS 'Global,Europe,Pacific,Indian,North_Atlantic,South_Atlantic'
                task pb_swh
                  edit PLOT_AREAS 'Global,Europe,Pacific,Indian,North_Atlantic,South_Atlantic'
              endfamily
              family legB
                trigger ( ./legA eq complete)
                edit PLOT_STEPS '252/to/360/by/12'
                edit STEP_RANGE 'legB'
                edit UPDATE_WEB_NOW 'yes'
                task pb_2t
                task pb_ff
                task pb_tp
                  edit PLOT_STEPS '228-252/240-264/252-276/264-288/276-300/288-312/300-324/312-336/324-348/336-360'
                task w_proba_gusts
                  edit PLOT_STEPS '228-252/240-264/252-276/264-288/276-300/288-312/300-324/312-336/324-348/336-360'
                task pb_mwp
                  edit PLOT_AREAS 'Global,Europe,Pacific,Indian,North_Atlantic,South_Atlantic'
                task pb_swh
                  edit PLOT_AREAS 'Global,Europe,Pacific,Indian,North_Atlantic,South_Atlantic'
              endfamily
              family means
                trigger ( ./legA eq complete)
                edit UPDATE_WEB_NOW 'yes'
                task long_pb_2t
                task long_pb_tpr
                task long_w_proba_gusts
                task long_pb_mwp
                  edit PLOT_AREAS 'Global,Europe,Pacific,Indian,North_Atlantic,South_Atlantic'
                task long_pb_swh
                  edit PLOT_AREAS 'Global,Europe,Pacific,Indian,North_Atlantic,South_Atlantic'
              endfamily
            endfamily
            family additional
              edit PLOT_STEPS '12/to/156/by/12'
              edit STEP_RANGE 'legA'
              task w_proba_rain
              family wmo_waves
                edit PLOT_STEPS '12/to/168/by/12'
                edit PLOT_AREAS 'Global'
                edit PLOT_GRID '2.5/2.5'
                task pb_swh
              endfamily
            endfamily
          endfamily
          family ensm
            trigger ( /emc_36r1/main:YMD gt /emc_36r1/pop:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/pop:YMD and /emc_36r1/main/12/legB/postproc/ensms eq complete and /emc_36r1/main/12/legA/postproc/ensms eq complete)
            task fields_ensm
              event 30 
            task plot_ensm_Z
              complete fields_ensm eq complete and not fields_ensm:30
              trigger fields_ensm eq complete and fields_ensm:30
            task plot_ensm_T
              complete fields_ensm eq complete and not fields_ensm:30
              trigger fields_ensm eq complete and fields_ensm:30
            task cleanup_ensm
              trigger plot_ensm_Z eq complete and plot_ensm_T eq complete
              edit SMSPASS 'stripped'
          endfamily
          family tc
            edit WSHOST 'linux_cluster'
            edit QUEUE 'emos-e'
            edit QUEUE_EPILOG 'emos-exp'
            edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
            edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
            edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
            edit LOGDIR '/vol/emos_nc/output'
            edit SMSOUT '/vol/emos_nc/output'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/pop/web/tc'
            inlimit /limits:linux_cluster
            family plotting
              trigger ( /emc_36r1/pop/12/data/tc/post_process eq complete)
              task spm
              task epsgram
              task tstorm
              task tstorm_web
                trigger tstorm eq complete
            endfamily
            task tc2web
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/pop:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/pop:YMD and ./plotting eq complete and /emc_36r1/pop/12/data/tc/diagnostics eq complete)
              edit WSHOST 'linux_cluster'
          endfamily
          family special
            family swdfp
              edit WSHOST 'linux_cluster'
              edit QUEUE 'serial'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/pop/web/swdfp'
              edit PLOTDOM '-45/-10/6/62'
              edit UPDATE_WEB_NOW 'yes'
              time 01:30
              family efi
                trigger /emc_36r1/pop/12/data/fields eq complete
                edit QUEUE 'emos-e'
                edit QUEUE_EPILOG 'emos-exp'
                edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
                edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
                edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
                edit LOGDIR '/vol/emos_nc/output'
                edit SMSOUT '/vol/emos_nc/output'
                inlimit /limits:linux_cluster
                family swdfp_tpi
                  task w_efi
                    edit PARAMETER 'tpi'
                endfamily
                family swdfp_10wsi
                  task w_efi
                    edit PARAMETER '10wsi'
                endfamily
                family swdfp_2ti
                  task w_efi
                    edit PARAMETER '2ti'
                endfamily
                family swdfp_10fgi
                  task w_efi
                    edit PARAMETER '10fgi'
                endfamily
              endfamily
              family proba
                trigger ( /emc_36r1/main:YMD gt /emc_36r1/pop:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/pop:YMD and /emc_36r1/main/12/legA/postproc/prob eq complete)
                edit PLOT_LOCATION '/webplots_e/data/swdfp/proba'
                edit STEPS '06/12/18/24/30/36/42/48/54/60/66/72/84/96/108/120/132/144'
                task swdfp_proba
                task swdfp_tp
                  trigger ./swdfp_proba eq complete
                task swdfp_ff
                  trigger ./swdfp_proba eq complete
                task swdfp_tp24
                  trigger ./swdfp_proba eq complete
                  edit STEPS '24/48/72/96/120'
                task swdfp_proba_clean
                  trigger ./swdfp_tp eq complete and ./swdfp_ff eq complete and ./swdfp_tp24 eq complete
              endfamily
              task swdfp_epsgram
                trigger /emc_36r1/pop/12/data/metgram eq complete
            endfamily
          endfamily
        endfamily
        family verification
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit DATEMASK '11.*.*'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/pop/web/verification'
          inlimit /limits:linux_cluster
          task w_probas
          task w_costloss
          task w_timeseries
        endfamily
        task web_update
          trigger pf/proba/members/legA eq complete and pf/proba/additional eq complete and pf/plumes eq complete and pf/classic eq complete
          edit WSHOST 'linux_cluster'
          edit QUEUE 'plots-medium'
          edit PACKAGES 'esuite'
          edit UPDATE_STREAM 'enfo'
          inlimit /emc_36r1/limits:webplots
      endfamily
      family paper
        trigger web/web_update eq complete
        edit QUEUE 'emos-e'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        edit SMSFILES '/home/ma/emos/def/emc_36r1/plot'
        inlimit /limits:linux_cluster
        task epsgram
          trigger ../data/metgram eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1/pop/paper'
        family enplot
          complete /emc_36r1/pop/12/events:cf
          trigger ( /emc_36r1/pop/12/events eq complete) and ( ( /emc_36r1/main:YMD gt /emc_36r1/pop:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/pop:YMD and /emc_36r1/main/12/legA/postproc/clusters eq complete))
          edit WSHOST 'linux_cluster'
          edit QUEUE 'plots-medium'
          edit ANALYSIS 'NO'
          task sthemi
        endfamily
        family tube
          trigger ( /emc_36r1/main:YMD gt /emc_36r1/pop:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/pop:YMD and /emc_36r1/main/12/legA/postproc/clusters eq complete)
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit WSHOST 'linux_cluster'
          edit QUEUE 'plots-medium'
          inlimit /limits:linux_cluster
          task ptubfr
          task tubfrfax
            trigger ptubfr eq complete
            edit QUEUE 'emos'
          task tubukfax
            trigger ptubuk eq complete
            edit QUEUE 'emos'
          task ptubuk
            trigger ptubfr eq complete
        endfamily
        family prob
          trigger ( /emc_36r1/main:YMD gt /emc_36r1/pop:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/pop:YMD and /emc_36r1/main/12/legA/postproc/prob eq complete)
          edit WSHOST 'linux_cluster'
          edit QUEUE 'plots-medium'
          edit ANALYSIS 'YES'
          inlimit /emc_36r1/limits:webplots
          task clplot
            trigger ../enplot eq complete
          task s144
            trigger clplot eq complete
            edit STEPS '144'
          task plumea
          task pb
        endfamily
      endfamily
      task cleanup
        trigger paper eq complete and web eq complete
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        edit WSHOST 'linux_cluster'
        edit QUEUE 'plots-medium'
        inlimit /limits:linux_cluster
        inlimit /emc_36r1/limits:webplots
    endfamily
    family 00
      trigger ( /emc_36r1/main:YMD gt /emc_36r1/pop:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/pop:YMD and /emc_36r1/main/00/legA/fc/pf eq complete and /emc_36r1/main/00/ref/model eq complete)
      edit DELTA_DAY '1'
      edit EMOS_BASE '00'
      edit SMSFILES '/home/ma/emos/def/emc_36r1/pop'
      task events
        edit QUEUE 'emos-e'
        edit QUEUE_EPILOG 'emos-exp'
        edit WSHOST 'localhost'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        edit SMSFILES '/home/ma/emos/def/emc_36r1/ref'
        edit HINDCAST '1'
        edit ONLY_CF '0'
        event noefi
        event 240 
        event 360 
        event cpl
        event 768 
        event cf
      family data
        trigger events eq complete
        family metgram
          trigger ( /emc_36r1/main:YMD gt /emc_36r1/pop:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/pop:YMD and /emc_36r1/main/00/legA/fc/pf eq complete)
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit QUEUE 'emos'
          edit SMSTRIES '1'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/pop/data/epsgrams/grams'
          inlimit /limits:linux_cluster
          family 240
            trigger ( /emc_36r1/main:YMD gt /emc_36r1/pop:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/pop:YMD and /emc_36r1/main/00/legA/fc/pf eq complete and /emc_36r1/main/00/legA/fc/cf eq complete)
            edit KIND '10_days'
            edit RHOST 'epsgrams'
            edit RDIR '/webapps/epsgrams/data/'
            task gclean
              edit KEEP_DAYS '2'
              edit REMOTE_KEEP_DAYS '2'
            task gdb
              trigger gclean eq complete
              edit NEW_DAILY '1'
              label files ""
              label info ""
              label local "-"
              label remote "-"
              event fdb
              event 1 
            task glink
              trigger gdb eq complete
          endfamily
          family 360
            trigger ( /emc_36r1/main:YMD gt /emc_36r1/pop:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/pop:YMD and /emc_36r1/main/00/legB/postproc eq complete and /emc_36r1/main/00/legA/postproc eq complete)
            edit KIND '15_days'
            task gclean
              edit KEEP_DAYS '2'
              edit REMOTE_KEEP_DAYS '2'
            task gdb
              trigger gclean eq complete
              edit NEW_DAILY '1'
              label files ""
              label info ""
              label local "-"
              label remote "-"
              event fdb
              event 1 
            task glink
              trigger gdb eq complete
          endfamily
          family wave
            trigger ( /emc_36r1/main:YMD gt /emc_36r1/pop:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/pop:YMD and /emc_36r1/main/00/legA/fc/pf eq complete and /emc_36r1/main/00/legA/fc/cf eq complete)
            edit KIND '10_days_wave'
            edit RHOST 'epsgrams'
            edit RDIR '/webapps/epsgrams/data/'
            task gclean
              edit KEEP_DAYS '2'
              edit REMOTE_KEEP_DAYS '2'
            task gdb
              trigger gclean eq complete
              edit NEW_DAILY '1'
              label files ""
              label info ""
              label local "-"
              label remote "-"
              event fdb
              event 1 
            task glink
              trigger gdb eq complete
          endfamily
          family 15d
            trigger ( /emc_36r1/main:YMD gt /emc_36r1/pop:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/pop:YMD and /emc_36r1/main/00/legB/postproc eq complete and /emc_36r1/main/00/legA/postproc eq complete)
            edit KIND '15_days_pf'
            edit RHOST 'epsgrams'
            edit RDIR '/webapps/epsgrams/data/'
            task gclean
              edit KEEP_DAYS '2'
              edit REMOTE_KEEP_DAYS '2'
            task gdb
              trigger gclean eq complete
              edit NEW_DAILY '1'
              label files ""
              label info ""
              label local "-"
              label remote "-"
              event fdb
              event 1 
            task glink
              trigger gdb eq complete
          endfamily
          family 15dclim
            trigger ( /emc_36r1/main:YMD gt /emc_36r1/pop:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/pop:YMD and /emc_36r1/main/00/legB/fc/pf eq complete and /emc_36r1/main/00/legB/fc/cf eq complete and /emc_36r1/00/efi/swarm/sync_climate eq complete)
            edit KIND 'clim_15_days'
            edit RHOST 'epsgrams'
            edit RDIR '/webapps/epsgrams/data/'
            task gclean
              edit KEEP_DAYS '2'
              edit REMOTE_KEEP_DAYS '2'
            task gdb
              trigger gclean eq complete
              edit NEW_DAILY '1'
              label files ""
              label info ""
              label local "-"
              label remote "-"
              event fdb
              event 1 
            task glink
              trigger gdb eq complete
          endfamily
          family efi
            trigger ( /emc_36r1/main:YMD gt /emc_36r1/pop:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/pop:YMD and /emc_36r1/main/00/legB/postproc eq complete and /emc_36r1/main/00/legA/postproc eq complete)
            edit KIND 'efi'
            edit RHOST 'epsgrams'
            edit RDIR '/webapps/epsgrams/data/'
            task gclean
              edit KEEP_DAYS '10'
              edit REMOTE_KEEP_DAYS '10'
            task gdb
              trigger gclean eq complete
              edit NEW_DAILY '1'
              label files ""
              label info ""
              label local "-"
              label remote "-"
              event fdb
              event 1 
            task glink
              trigger gdb eq complete
          endfamily
        endfamily
        family distribution
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/pop/data'
          edit DISTRIB_FDB '/leda/fdb1'
          inlimit /limits:linux_cluster
          task ff
            edit PARAMETER 'ff'
            edit STEP_LIST '84 108 132 156 180'
            edit RANK '9999'
          task 2t
            edit PARAMETER '2t'
            edit STEP_LIST '84 108 132 156 180'
            edit RANK '9998'
          task tp
            edit PARAMETER 'tp'
            edit STEP_LIST '102 126 150 174 198'
            edit RANK '9997'
          task tp5
            edit PARAMETER 'tp5'
            edit STEP_LIST '174'
            edit RANK '9996'
          task tp10
            edit PARAMETER 'tp10'
            edit STEP_LIST '240'
            edit RANK '9995'
          task cleanup
            trigger ff eq complete and ./2t eq complete and tp eq complete and tp5 eq complete and tp10 eq complete
        endfamily
        family fields
          trigger ( /emc_36r1/main:YMD gt /emc_36r1/pop:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/pop:YMD and /emc_36r1/main/00/legB/postproc eq complete and /emc_36r1/main/00/legA/postproc eq complete)
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit WSHOST 'linux_cluster'
          edit QUEUE 'plots-medium'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/pop/data/fields'
          inlimit /limits:linux_cluster
          inlimit /emc_36r1/limits:webplots
          task efi
            complete ( /emc_36r1/pop/00/events:cf) or ( /emc_36r1/pop/00/events:noefi)
            trigger /emc_36r1/pop/00/events eq complete
            event none
          task globalefi
            complete ( /emc_36r1/pop/00/events:cf) or ( /emc_36r1/pop/00/events:noefi)
            trigger /emc_36r1/pop/00/events eq complete
            event none
          family legA
            task proba
              complete /emc_36r1/pop/00/events:cf
              trigger /emc_36r1/pop/00/events eq complete
              edit RANGE 'legA'
              edit PLOT_STEPS_RANGE '00-24/12-36/24-48/36-60/48-72/60-84/72-96/84-108/96-120/108-132/120-144/132-156/144-168/156-180/168-192/180-204/192-216/204-228/216-240'
              edit PLOT_STEPS '12/to/240/by/12'
          endfamily
          family legB
            task proba
              complete /emc_36r1/pop/00/events:cf
              trigger /emc_36r1/pop/00/events eq complete
              edit RANGE 'legB'
              edit PLOT_STEPS '252/to/360/by/12'
              edit PLOT_STEPS_RANGE '228-252/240-264/252-276/264-288/276-300/288-312/300-324/312-336/324-348/336-360'
          endfamily
          task proba_mean
            complete /emc_36r1/pop/00/events:cf
            trigger /emc_36r1/pop/00/events eq complete
            edit PLOT_STEPS_MEAN '240-360'
          task lsmoro
            edit STEP_LIST '0/to/144/by/12'
        endfamily
        family tc
          complete /emc_36r1/pop/00/events:cf
          trigger ( /emc_36r1/pop/00/events eq complete) and ( ( /emc_36r1/lag:YMD gt /emc_36r1/pop:YMD) or ( /emc_36r1/lag:YMD eq /emc_36r1/pop:YMD and /emc_36r1/lag/00/legA/tc eq complete))
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/tc'
          inlimit /limits:linux_cluster
          task cleanup
          task check_data
            trigger /emc_36r1/pop/12/data/tc eq complete
          family post_process
            task strike_prob
              trigger ../check_data eq complete
          endfamily
          family diagnostics
            trigger post_process eq complete
            task det_errors
            task spreadskill
            task reliability
          endfamily
          family store
            trigger ( /emc_36r1/pop/00/data/tc/diagnostics eq complete)
            task to_ecfs_pop
          endfamily
        endfamily
        family tcyc
          complete /emc_36r1/pop/00/events:cf
          trigger ( /emc_36r1/main:YMD gt /emc_36r1/pop:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/pop:YMD and /emc_36r1/main/00/legB/fc eq complete)
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit FCCHUNK '1'
          edit OCTOTAL '5'
          edit FCTOTAL '51'
          edit FCLENGTH1 '216'
          edit DAILY '0'
          edit RUNVARFC '1'
          edit FCLENGTH '360'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/archive/leg'
          edit EPSCOUPLE '1'
          edit RADIUS '500'
          limit semaphore 5
          inlimit /limits:linux_cluster
          inlimit semaphore
          task genes_stage
            edit DELTA_DAY '2'
          family genesis
            trigger genes_stage eq complete
            family da
              edit OCGROUP 'an0'
              edit FCGROUP 'fc0'
              edit EPSMEMBER 'da'
              edit MEMBER 'da'
              task tcyc
            endfamily
            family 00
              edit OCGROUP 'an00'
              edit FCGROUP 'fc0'
              edit EPSMEMBER '00'
              edit MEMBER '00'
              task tcyc
            endfamily
            family 01
              edit OCGROUP 'an1'
              edit FCGROUP 'fc0'
              edit EPSMEMBER '01'
              edit MEMBER '01'
              task tcyc
            endfamily
            family 02
              edit OCGROUP 'an2'
              edit FCGROUP 'fc0'
              edit EPSMEMBER '02'
              edit MEMBER '02'
              task tcyc
            endfamily
            family 03
              edit OCGROUP 'an3'
              edit FCGROUP 'fc0'
              edit EPSMEMBER '03'
              edit MEMBER '03'
              task tcyc
            endfamily
            family 04
              edit OCGROUP 'an4'
              edit FCGROUP 'fc0'
              edit EPSMEMBER '04'
              edit MEMBER '04'
              task tcyc
            endfamily
            family 05
              edit OCGROUP 'an0'
              edit FCGROUP 'fc1'
              edit EPSMEMBER '05'
              edit MEMBER '05'
              task tcyc
            endfamily
            family 06
              edit OCGROUP 'an1'
              edit FCGROUP 'fc1'
              edit EPSMEMBER '06'
              edit MEMBER '06'
              task tcyc
            endfamily
            family 07
              edit OCGROUP 'an2'
              edit FCGROUP 'fc1'
              edit EPSMEMBER '07'
              edit MEMBER '07'
              task tcyc
            endfamily
            family 08
              edit OCGROUP 'an3'
              edit FCGROUP 'fc1'
              edit EPSMEMBER '08'
              edit MEMBER '08'
              task tcyc
            endfamily
            family 09
              edit OCGROUP 'an4'
              edit FCGROUP 'fc1'
              edit EPSMEMBER '09'
              edit MEMBER '09'
              task tcyc
            endfamily
            family 10
              edit OCGROUP 'an0'
              edit FCGROUP 'fc2'
              edit EPSMEMBER '10'
              edit MEMBER '10'
              task tcyc
            endfamily
            family 11
              edit OCGROUP 'an1'
              edit FCGROUP 'fc2'
              edit EPSMEMBER '11'
              edit MEMBER '11'
              task tcyc
            endfamily
            family 12
              edit OCGROUP 'an2'
              edit FCGROUP 'fc2'
              edit EPSMEMBER '12'
              edit MEMBER '12'
              task tcyc
            endfamily
            family 13
              edit OCGROUP 'an3'
              edit FCGROUP 'fc2'
              edit EPSMEMBER '13'
              edit MEMBER '13'
              task tcyc
            endfamily
            family 14
              edit OCGROUP 'an4'
              edit FCGROUP 'fc2'
              edit EPSMEMBER '14'
              edit MEMBER '14'
              task tcyc
            endfamily
            family 15
              edit OCGROUP 'an0'
              edit FCGROUP 'fc3'
              edit EPSMEMBER '15'
              edit MEMBER '15'
              task tcyc
            endfamily
            family 16
              edit OCGROUP 'an1'
              edit FCGROUP 'fc3'
              edit EPSMEMBER '16'
              edit MEMBER '16'
              task tcyc
            endfamily
            family 17
              edit OCGROUP 'an2'
              edit FCGROUP 'fc3'
              edit EPSMEMBER '17'
              edit MEMBER '17'
              task tcyc
            endfamily
            family 18
              edit OCGROUP 'an3'
              edit FCGROUP 'fc3'
              edit EPSMEMBER '18'
              edit MEMBER '18'
              task tcyc
            endfamily
            family 19
              edit OCGROUP 'an4'
              edit FCGROUP 'fc3'
              edit EPSMEMBER '19'
              edit MEMBER '19'
              task tcyc
            endfamily
            family 20
              edit OCGROUP 'an0'
              edit FCGROUP 'fc4'
              edit EPSMEMBER '20'
              edit MEMBER '20'
              task tcyc
            endfamily
            family 21
              edit OCGROUP 'an1'
              edit FCGROUP 'fc4'
              edit EPSMEMBER '21'
              edit MEMBER '21'
              task tcyc
            endfamily
            family 22
              edit OCGROUP 'an2'
              edit FCGROUP 'fc4'
              edit EPSMEMBER '22'
              edit MEMBER '22'
              task tcyc
            endfamily
            family 23
              edit OCGROUP 'an3'
              edit FCGROUP 'fc4'
              edit EPSMEMBER '23'
              edit MEMBER '23'
              task tcyc
            endfamily
            family 24
              edit OCGROUP 'an4'
              edit FCGROUP 'fc4'
              edit EPSMEMBER '24'
              edit MEMBER '24'
              task tcyc
            endfamily
            family 25
              edit OCGROUP 'an0'
              edit FCGROUP 'fc5'
              edit EPSMEMBER '25'
              edit MEMBER '25'
              task tcyc
            endfamily
            family 26
              edit OCGROUP 'an1'
              edit FCGROUP 'fc5'
              edit EPSMEMBER '26'
              edit MEMBER '26'
              task tcyc
            endfamily
            family 27
              edit OCGROUP 'an2'
              edit FCGROUP 'fc5'
              edit EPSMEMBER '27'
              edit MEMBER '27'
              task tcyc
            endfamily
            family 28
              edit OCGROUP 'an3'
              edit FCGROUP 'fc5'
              edit EPSMEMBER '28'
              edit MEMBER '28'
              task tcyc
            endfamily
            family 29
              edit OCGROUP 'an4'
              edit FCGROUP 'fc5'
              edit EPSMEMBER '29'
              edit MEMBER '29'
              task tcyc
            endfamily
            family 30
              edit OCGROUP 'an0'
              edit FCGROUP 'fc6'
              edit EPSMEMBER '30'
              edit MEMBER '30'
              task tcyc
            endfamily
            family 31
              edit OCGROUP 'an1'
              edit FCGROUP 'fc6'
              edit EPSMEMBER '31'
              edit MEMBER '31'
              task tcyc
            endfamily
            family 32
              edit OCGROUP 'an2'
              edit FCGROUP 'fc6'
              edit EPSMEMBER '32'
              edit MEMBER '32'
              task tcyc
            endfamily
            family 33
              edit OCGROUP 'an3'
              edit FCGROUP 'fc6'
              edit EPSMEMBER '33'
              edit MEMBER '33'
              task tcyc
            endfamily
            family 34
              edit OCGROUP 'an4'
              edit FCGROUP 'fc6'
              edit EPSMEMBER '34'
              edit MEMBER '34'
              task tcyc
            endfamily
            family 35
              edit OCGROUP 'an0'
              edit FCGROUP 'fc7'
              edit EPSMEMBER '35'
              edit MEMBER '35'
              task tcyc
            endfamily
            family 36
              edit OCGROUP 'an1'
              edit FCGROUP 'fc7'
              edit EPSMEMBER '36'
              edit MEMBER '36'
              task tcyc
            endfamily
            family 37
              edit OCGROUP 'an2'
              edit FCGROUP 'fc7'
              edit EPSMEMBER '37'
              edit MEMBER '37'
              task tcyc
            endfamily
            family 38
              edit OCGROUP 'an3'
              edit FCGROUP 'fc7'
              edit EPSMEMBER '38'
              edit MEMBER '38'
              task tcyc
            endfamily
            family 39
              edit OCGROUP 'an4'
              edit FCGROUP 'fc7'
              edit EPSMEMBER '39'
              edit MEMBER '39'
              task tcyc
            endfamily
            family 40
              edit OCGROUP 'an0'
              edit FCGROUP 'fc8'
              edit EPSMEMBER '40'
              edit MEMBER '40'
              task tcyc
            endfamily
            family 41
              edit OCGROUP 'an1'
              edit FCGROUP 'fc8'
              edit EPSMEMBER '41'
              edit MEMBER '41'
              task tcyc
            endfamily
            family 42
              edit OCGROUP 'an2'
              edit FCGROUP 'fc8'
              edit EPSMEMBER '42'
              edit MEMBER '42'
              task tcyc
            endfamily
            family 43
              edit OCGROUP 'an3'
              edit FCGROUP 'fc8'
              edit EPSMEMBER '43'
              edit MEMBER '43'
              task tcyc
            endfamily
            family 44
              edit OCGROUP 'an4'
              edit FCGROUP 'fc8'
              edit EPSMEMBER '44'
              edit MEMBER '44'
              task tcyc
            endfamily
            family 45
              edit OCGROUP 'an0'
              edit FCGROUP 'fc9'
              edit EPSMEMBER '45'
              edit MEMBER '45'
              task tcyc
            endfamily
            family 46
              edit OCGROUP 'an1'
              edit FCGROUP 'fc9'
              edit EPSMEMBER '46'
              edit MEMBER '46'
              task tcyc
            endfamily
            family 47
              edit OCGROUP 'an2'
              edit FCGROUP 'fc9'
              edit EPSMEMBER '47'
              edit MEMBER '47'
              task tcyc
            endfamily
            family 48
              edit OCGROUP 'an3'
              edit FCGROUP 'fc9'
              edit EPSMEMBER '48'
              edit MEMBER '48'
              task tcyc
            endfamily
            family 49
              edit OCGROUP 'an4'
              edit FCGROUP 'fc9'
              edit EPSMEMBER '49'
              edit MEMBER '49'
              task tcyc
            endfamily
            family 50
              edit OCGROUP 'an0'
              edit FCGROUP 'fc10'
              edit EPSMEMBER '50'
              edit MEMBER '50'
              task tcyc
            endfamily
          endfamily
          task tcyctoecfs
            trigger genesis eq complete
        endfamily
        family mjo
          complete /emc_36r1/lag/ref/00/events:cf
          trigger ( /emc_36r1/main:YMD gt /emc_36r1/pop:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/pop:YMD and /emc_36r1/main/00/legB/postproc eq complete)
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit FCLENGTH '360'
          edit ERA40END '2002033000'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/smsfiles/web'
          edit RUNVARFC '1'
          inlimit /limits:linux_cluster
          family stage
            edit RETRIEVE 'stage'
            edit DELTA_DAY '2'
            task chi200
            task olr
            task u200
            task u850
          endfamily
          family process
            task chi200
            task olr
            task u850
            task u200
          endfamily
          family ncep
            trigger ./process eq complete
            task olr_to_ncep
            task u200_to_ncep
            task u850_to_ncep
          endfamily
          task ncep_daily
            trigger ./ncep eq complete
          task ceof
            trigger process eq complete
          task mjo_migrate
            complete /emc_36r1/pop/00/events:768
            trigger process eq complete and ceof eq complete and ncep_daily eq complete
          task mjo_clean
            complete /emc_36r1/pop/00/events:768
            trigger process eq complete and ncep eq complete and ncep_daily eq complete and ceof eq complete and mjo_migrate eq complete
        endfamily
      endfamily
      family web
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        edit WSHOST 'linux_cluster'
        edit QUEUE 'plots-medium'
        edit SMSFILES '/home/ma/emos/def/emc_36r1/pop/web'
        inlimit /limits:linux_cluster
        inlimit /emc_36r1/limits:webplots
        family pf
          complete /emc_36r1/pop/00/events:cf
          trigger /emc_36r1/pop/00/events eq complete
          task plumes
            trigger ( /emc_36r1/main:YMD gt /emc_36r1/pop:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/pop:YMD and /emc_36r1/main/00/legA/fc/cf eq complete and /emc_36r1/main/00/legA/fc/pf eq complete)
          family classic
            trigger ( /emc_36r1/main:YMD gt /emc_36r1/pop:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/pop:YMD and /emc_36r1/main/00/legA/postproc/clusters eq complete)
            edit ANALYSIS 'NO'
            task clplot
            task pltube
            task enplot
              edit STEPS '12 24 36 48 60 72 84 96 108 120 132 144 156 168 180 192'
              edit PRINTSTEPS '60 84 108 132 156 180'
          endfamily
          family efi
            complete /emc_36r1/pop/00/events:noefi
            trigger ../../data/fields eq complete
            edit QUEUE_EPILOG 'emos-exp'
            edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
            edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
            edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
            edit LOGDIR '/vol/emos_nc/output'
            edit SMSOUT '/vol/emos_nc/output'
            edit WSHOST 'linux_cluster'
            edit QUEUE 'plots-medium'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/pop/web'
            inlimit /limits:linux_cluster
            inlimit /emc_36r1/limits:webplots
            family efi_tpi
              task w_efi
                edit PARAMETER 'tpi'
            endfamily
            family efi_10wsi
              task w_efi
                edit PARAMETER '10wsi'
            endfamily
            family efi_2ti
              task w_efi
                edit PARAMETER '2ti'
            endfamily
            family efi_10fgi
              task w_efi
                edit PARAMETER '10fgi'
            endfamily
            task globalefi
              edit UPDATE_WEB_NOW 'yes'
          endfamily
          family proba
            trigger ../../data/fields eq complete
            edit QUEUE_EPILOG 'emos-exp'
            edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
            edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
            edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
            edit LOGDIR '/vol/emos_nc/output'
            edit SMSOUT '/vol/emos_nc/output'
            edit WSHOST 'linux_cluster'
            edit QUEUE 'plots-medium'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/pop/web'
            edit PLOT_AREAS 'Europe,North_America,Asia,South_Atlantic,South_Pacific,Australia,Africa,Global'
            inlimit /limits:linux_cluster
            inlimit /emc_36r1/limits:webplots
            family members
              family legA
                edit PLOT_STEPS '12/to/240/by/12'
                edit STEP_RANGE 'legA'
                task pb_2t
                task pb_ff
                task pb_tp
                  edit PLOT_STEPS '00-24/12-36/24-48/36-60/48-72/60-84/72-96/84-108/96-120/108-132/120-144/132-156/144-168/156-180/168-192/180-204/192-216/204-228/216-240'
                task w_proba_gusts
                  edit PLOT_STEPS '00-24/12-36/24-48/36-60/48-72/60-84/72-96/84-108/96-120/108-132/120-144/132-156/144-168/156-180/168-192/180-204/192-216/204-228/216-240'
                task pb_tpr
                task pb_mwp
                  edit PLOT_AREAS 'Global,Europe,Pacific,Indian,North_Atlantic,South_Atlantic'
                task pb_swh
                  edit PLOT_AREAS 'Global,Europe,Pacific,Indian,North_Atlantic,South_Atlantic'
              endfamily
              family legB
                trigger ( ./legA eq complete)
                edit PLOT_STEPS '252/to/360/by/12'
                edit STEP_RANGE 'legB'
                edit UPDATE_WEB_NOW 'yes'
                task pb_2t
                task pb_ff
                task pb_tp
                  edit PLOT_STEPS '228-252/240-264/252-276/264-288/276-300/288-312/300-324/312-336/324-348/336-360'
                task w_proba_gusts
                  edit PLOT_STEPS '228-252/240-264/252-276/264-288/276-300/288-312/300-324/312-336/324-348/336-360'
                task pb_mwp
                  edit PLOT_AREAS 'Global,Europe,Pacific,Indian,North_Atlantic,South_Atlantic'
                task pb_swh
                  edit PLOT_AREAS 'Global,Europe,Pacific,Indian,North_Atlantic,South_Atlantic'
              endfamily
              family means
                trigger ( ./legA eq complete)
                edit UPDATE_WEB_NOW 'yes'
                task long_pb_2t
                task long_pb_tpr
                task long_w_proba_gusts
                task long_pb_mwp
                  edit PLOT_AREAS 'Global,Europe,Pacific,Indian,North_Atlantic,South_Atlantic'
                task long_pb_swh
                  edit PLOT_AREAS 'Global,Europe,Pacific,Indian,North_Atlantic,South_Atlantic'
              endfamily
            endfamily
            family additional
              edit PLOT_STEPS '12/to/156/by/12'
              edit STEP_RANGE 'legA'
              task w_proba_rain
              family wmo_waves
                edit PLOT_STEPS '12/to/168/by/12'
                edit PLOT_AREAS 'Global'
                edit PLOT_GRID '2.5/2.5'
                task pb_swh
              endfamily
            endfamily
          endfamily
          family ensm
            trigger ( /emc_36r1/main:YMD gt /emc_36r1/pop:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/pop:YMD and /emc_36r1/main/00/legB/postproc/ensms eq complete and /emc_36r1/main/00/legA/postproc/ensms eq complete)
            task fields_ensm
              event 30 
            task plot_ensm_Z
              complete fields_ensm eq complete and not fields_ensm:30
              trigger fields_ensm eq complete and fields_ensm:30
            task plot_ensm_T
              complete fields_ensm eq complete and not fields_ensm:30
              trigger fields_ensm eq complete and fields_ensm:30
            task cleanup_ensm
              trigger plot_ensm_Z eq complete and plot_ensm_T eq complete
          endfamily
          family tc
            edit WSHOST 'linux_cluster'
            edit QUEUE 'emos-e'
            edit QUEUE_EPILOG 'emos-exp'
            edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
            edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
            edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
            edit LOGDIR '/vol/emos_nc/output'
            edit SMSOUT '/vol/emos_nc/output'
            edit SMSFILES '/home/ma/emos/def/emc_36r1/pop/web/tc'
            inlimit /limits:linux_cluster
            family plotting
              trigger ( /emc_36r1/pop/00/data/tc/post_process eq complete)
              task spm
              task epsgram
              task tstorm
              task tstorm_web
                trigger tstorm eq complete
            endfamily
            task tc2web
              trigger ( /emc_36r1/main:YMD gt /emc_36r1/pop:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/pop:YMD and ./plotting eq complete and /emc_36r1/pop/00/data/tc/diagnostics eq complete)
              edit WSHOST 'linux_cluster'
          endfamily
          family special
            family swdfp
              edit WSHOST 'linux_cluster'
              edit QUEUE 'serial'
              edit SMSFILES '/home/ma/emos/def/emc_36r1/pop/web/swdfp'
              edit PLOTDOM '-45/-10/6/62'
              edit UPDATE_WEB_NOW 'yes'
              time 10:30
              family efi
                trigger /emc_36r1/pop/00/data/fields eq complete
                edit QUEUE 'emos-e'
                edit QUEUE_EPILOG 'emos-exp'
                edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
                edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
                edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
                edit LOGDIR '/vol/emos_nc/output'
                edit SMSOUT '/vol/emos_nc/output'
                inlimit /limits:linux_cluster
                family swdfp_tpi
                  task w_efi
                    edit PARAMETER 'tpi'
                endfamily
                family swdfp_10wsi
                  task w_efi
                    edit PARAMETER '10wsi'
                endfamily
                family swdfp_2ti
                  task w_efi
                    edit PARAMETER '2ti'
                endfamily
                family swdfp_10fgi
                  task w_efi
                    edit PARAMETER '10fgi'
                endfamily
              endfamily
              family proba
                trigger ( /emc_36r1/main:YMD gt /emc_36r1/pop:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/pop:YMD and /emc_36r1/main/00/legA/postproc/prob eq complete)
                edit PLOT_LOCATION '/webplots_e/data/swdfp/proba'
                edit STEPS '06/12/18/24/30/36/42/48/54/60/66/72/84/96/108/120/132/144'
                task swdfp_proba
                task swdfp_tp
                  trigger ./swdfp_proba eq complete
                task swdfp_ff
                  trigger ./swdfp_proba eq complete
                task swdfp_tp24
                  trigger ./swdfp_proba eq complete
                  edit STEPS '24/48/72/96/120'
                task swdfp_proba_clean
                  trigger ./swdfp_tp eq complete and ./swdfp_ff eq complete and ./swdfp_tp24 eq complete
              endfamily
              task swdfp_epsgram
                trigger /emc_36r1/pop/00/data/metgram eq complete
            endfamily
          endfamily
        endfamily
        family verification
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit DATEMASK '11.*.*'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/pop/web/verification'
          inlimit /limits:linux_cluster
          task w_probas
          task w_costloss
          task w_timeseries
        endfamily
        task web_update
          trigger pf/proba/members/legA eq complete and pf/proba/additional eq complete and pf/plumes eq complete and pf/classic eq complete
          edit WSHOST 'linux_cluster'
          edit QUEUE 'plots-medium'
          edit PACKAGES 'esuite'
          edit UPDATE_STREAM 'enfo'
          inlimit /emc_36r1/limits:webplots
      endfamily
      family paper
        trigger web/web_update eq complete
        edit QUEUE 'emos-e'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        edit SMSFILES '/home/ma/emos/def/emc_36r1/plot'
        inlimit /limits:linux_cluster
        task epsgram
          trigger ../data/metgram eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1/pop/paper'
        family enplot
          complete /emc_36r1/pop/00/events:cf
          trigger ( /emc_36r1/pop/00/events eq complete) and ( ( /emc_36r1/main:YMD gt /emc_36r1/pop:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/pop:YMD and /emc_36r1/main/00/legA/postproc/clusters eq complete))
          edit WSHOST 'linux_cluster'
          edit QUEUE 'plots-medium'
          edit ANALYSIS 'NO'
          task sthemi
        endfamily
        family tube
          trigger ( /emc_36r1/main:YMD gt /emc_36r1/pop:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/pop:YMD and /emc_36r1/main/00/legA/postproc/clusters eq complete)
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit WSHOST 'linux_cluster'
          edit QUEUE 'plots-medium'
          inlimit /limits:linux_cluster
          task ptubfr
          task tubfrfax
            trigger ptubfr eq complete
            edit QUEUE 'emos'
          task tubukfax
            trigger ptubuk eq complete
            edit QUEUE 'emos'
          task ptubuk
            trigger ptubfr eq complete
        endfamily
        family prob
          trigger ( /emc_36r1/main:YMD gt /emc_36r1/pop:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/pop:YMD and /emc_36r1/main/00/legA/postproc/prob eq complete)
          edit WSHOST 'linux_cluster'
          edit QUEUE 'plots-medium'
          edit ANALYSIS 'YES'
          inlimit /emc_36r1/limits:webplots
          task clplot
            trigger ../enplot eq complete
          task s144
            trigger clplot eq complete
            edit STEPS '144'
          task plumea
          task pb
        endfamily
      endfamily
      task cleanup
        trigger paper eq complete and web eq complete
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        edit WSHOST 'linux_cluster'
        edit QUEUE 'plots-medium'
        inlimit /limits:linux_cluster
        inlimit /emc_36r1/limits:webplots
    endfamily
    task logfilestore
      edit QUEUE 'emos-e'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit KEEPLOGS 'no'
      edit LOGSTOECFS 'no'
      edit MOVE_TO_ECFS 'yes'
    task logfiles
      trigger ../main/make eq complete and ./12 eq complete and ./00 eq complete and logfilestore eq complete
      edit QUEUE 'emos-e'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit MOVE_TO_ECFS 'no'
      edit SMSFILES '/home/ma/emos/def/emc_36r1'
  endfamily
  family msjobs
    complete ( /emc_36r1/msjobs:YMD le /emc_36r1/main:YMD) and /emc_36r1/main/logfiles eq active
    repeat date YMD 20091003 20121212 1
    edit QUEUE 'emos'
    edit QUEUE_EPILOG 'emos-exp'
    edit WSHOST 'ecgate'
    edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
    edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
    edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
    edit SMSOUT '/vol/emos_nc/output'
    edit LOGDIR '/vol/emos_nc/output'
    edit STREAM 'EF'
    edit EMOS_STREAM 'EF'
    edit SMSURLBASE 'http://ecgate.ecmwf.int:9080'
    edit SMSURL 'do/events'
    edit SMSFILES '/home/ma/emos/def/emc_36r1'
    edit DELTA_DAY '-1'
    edit LOGTASK 'msjobs'
    family 12
      trigger ( /emc_36r1/main:YMD gt /emc_36r1/msjobs:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/msjobs:YMD and /emc_36r1/main/12/legA/fc eq complete and /emc_36r1/main/12/legA/postproc/clusters eq complete and /emc_36r1/main/12/ref/model eq complete)
      edit EMOS_BASE '12'
      edit DELTA_DAY '-1'
      edit SMSFILES '/home/ma/emos/def/emc_36r1/msjobs'
      edit SMSTRIES '1'
      family ms036
        trigger ( /emc_36r1/main:YMD gt /emc_36r1/msjobs:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/msjobs:YMD and /emc_36r1/main/12/legA/fc eq complete)
        edit EMOS_TIME_STEP_H '036'
        edit MSJ_DIR 'ef12h036'
        task ms_ec
      endfamily
      family ms072
        trigger ( /emc_36r1/main:YMD gt /emc_36r1/msjobs:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/msjobs:YMD and /emc_36r1/main/12/legA/fc eq complete)
        edit EMOS_TIME_STEP_H '072'
        edit MSJ_DIR 'ef12h072'
        task ms_ec
      endfamily
      family ms120
        trigger ( /emc_36r1/main:YMD gt /emc_36r1/msjobs:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/msjobs:YMD and /emc_36r1/main/12/legA/fc eq complete)
        edit EMOS_TIME_STEP_H '120'
        edit MSJ_DIR 'ef12h120'
        task ms_ec
      endfamily
      family ms144
        trigger ( /emc_36r1/main:YMD gt /emc_36r1/msjobs:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/msjobs:YMD and /emc_36r1/main/12/legA/fc eq complete)
        edit EMOS_TIME_STEP_H '144'
        edit MSJ_DIR 'ef12h144'
        task ms_ec
      endfamily
      family ms240
        trigger ( /emc_36r1/main:YMD gt /emc_36r1/msjobs:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/msjobs:YMD and /emc_36r1/main/12/legA/fc eq complete)
        edit EMOS_TIME_STEP_H '240'
        edit MSJ_DIR 'ef12h240'
        task ms_ec
      endfamily
      family ms360
        trigger ( /emc_36r1/main:YMD gt /emc_36r1/msjobs:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/msjobs:YMD and /emc_36r1/main/12/legB/fc eq complete)
        edit EMOS_TIME_STEP_H '360'
        edit MSJ_DIR 'ef12h360'
        task ms_ec
      endfamily
      family ms240post
        edit EMOS_TIME_STEP_H '240'
        edit MSJ_DIR 'ef12h240post'
        task ms_ec
      endfamily
      family ms360post
        edit EMOS_TIME_STEP_H '360'
        edit MSJ_DIR 'ef12h360post'
        task ms_ec
      endfamily
      family ms240metgram
        trigger ( /emc_36r1/pop:YMD gt /emc_36r1/msjobs:YMD) or ( /emc_36r1/pop:YMD eq /emc_36r1/msjobs:YMD and /emc_36r1/pop/12/data/metgram/240/gdb eq complete)
        edit EMOS_TIME_STEP_H '240'
        edit MSJ_DIR 'ef12h240metgram'
        task ms_ec
      endfamily
      family ms360metgram
        trigger ( /emc_36r1/pop:YMD gt /emc_36r1/msjobs:YMD) or ( /emc_36r1/pop:YMD eq /emc_36r1/msjobs:YMD and /emc_36r1/pop/12/data/metgram/15d/gdb eq complete)
        edit EMOS_TIME_STEP_H '360'
        edit MSJ_DIR 'ef12h360metgram'
        task ms_ec
      endfamily
      family ems240
        trigger ( /emc_36r1/main:YMD gt /emc_36r1/msjobs:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/msjobs:YMD and /emc_36r1/main/12/legA/prod2diss eq complete)
        edit EMOS_TIME_STEP '240'
        edit MSJ_DIR 'eef12h240'
        task ms_ec
      endfamily
    endfamily
    family 00
      trigger ( /emc_36r1/main:YMD gt /emc_36r1/msjobs:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/msjobs:YMD and /emc_36r1/main/00/legA/fc eq complete and /emc_36r1/main/00/legA/postproc/clusters eq complete and /emc_36r1/main/00/ref/model eq complete)
      edit EMOS_BASE '00'
      edit DELTA_DAY '0'
      edit SMSFILES '/home/ma/emos/def/emc_36r1/msjobs'
      edit SMSTRIES '1'
      family ms036
        trigger ( /emc_36r1/main:YMD gt /emc_36r1/msjobs:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/msjobs:YMD and /emc_36r1/main/00/legA/fc eq complete)
        edit EMOS_TIME_STEP_H '036'
        edit MSJ_DIR 'ef00h036'
        task ms_ec
      endfamily
      family ms072
        trigger ( /emc_36r1/main:YMD gt /emc_36r1/msjobs:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/msjobs:YMD and /emc_36r1/main/00/legA/fc eq complete)
        edit EMOS_TIME_STEP_H '072'
        edit MSJ_DIR 'ef00h072'
        task ms_ec
      endfamily
      family ms120
        trigger ( /emc_36r1/main:YMD gt /emc_36r1/msjobs:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/msjobs:YMD and /emc_36r1/main/00/legA/fc eq complete)
        edit EMOS_TIME_STEP_H '120'
        edit MSJ_DIR 'ef00h120'
        task ms_ec
      endfamily
      family ms144
        trigger ( /emc_36r1/main:YMD gt /emc_36r1/msjobs:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/msjobs:YMD and /emc_36r1/main/00/legA/fc eq complete)
        edit EMOS_TIME_STEP_H '144'
        edit MSJ_DIR 'ef00h144'
        task ms_ec
      endfamily
      family ms240
        trigger ( /emc_36r1/main:YMD gt /emc_36r1/msjobs:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/msjobs:YMD and /emc_36r1/main/00/legA/fc eq complete)
        edit EMOS_TIME_STEP_H '240'
        edit MSJ_DIR 'ef00h240'
        task ms_ec
      endfamily
      family ms360
        trigger ( /emc_36r1/main:YMD gt /emc_36r1/msjobs:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/msjobs:YMD and /emc_36r1/main/00/legB/fc eq complete)
        edit EMOS_TIME_STEP_H '360'
        edit MSJ_DIR 'ef00h360'
        task ms_ec
      endfamily
      family ms240post
        edit EMOS_TIME_STEP_H '240'
        edit MSJ_DIR 'ef00h240post'
        task ms_ec
      endfamily
      family ms360post
        edit EMOS_TIME_STEP_H '360'
        edit MSJ_DIR 'ef00h360post'
        task ms_ec
      endfamily
      family ms240metgram
        trigger ( /emc_36r1/pop:YMD gt /emc_36r1/msjobs:YMD) or ( /emc_36r1/pop:YMD eq /emc_36r1/msjobs:YMD and /emc_36r1/pop/00/data/metgram/240/gdb eq complete)
        edit EMOS_TIME_STEP_H '240'
        edit MSJ_DIR 'ef00h240metgram'
        task ms_ec
      endfamily
      family ms360metgram
        trigger ( /emc_36r1/pop:YMD gt /emc_36r1/msjobs:YMD) or ( /emc_36r1/pop:YMD eq /emc_36r1/msjobs:YMD and /emc_36r1/pop/00/data/metgram/15d/gdb eq complete)
        edit EMOS_TIME_STEP_H '360'
        edit MSJ_DIR 'ef00h360metgram'
        task ms_ec
      endfamily
      family ems240
        trigger ( /emc_36r1/main:YMD gt /emc_36r1/msjobs:YMD) or ( /emc_36r1/main:YMD eq /emc_36r1/msjobs:YMD and /emc_36r1/main/00/legA/prod2diss eq complete)
        edit EMOS_TIME_STEP '240'
        edit MSJ_DIR 'eef00h240'
        task ms_ec
      endfamily
    endfamily
    task logfilestore
      edit QUEUE 'emos-e'
      edit QUEUE_EPILOG 'emos-exp'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit KEEPLOGS 'no'
      edit LOGSTOECFS 'no'
      edit MOVE_TO_ECFS 'yes'
      inlimit /limits:linux_cluster
    task logfiles
      trigger ./00 eq complete and ./12 eq complete and logfilestore eq complete
      edit QUEUE 'emos-e'
      edit QUEUE_EPILOG 'emos-exp'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit MOVE_TO_ECFS 'no'
      edit SMSFILES '/home/ma/emos/def/emc_36r1'
      inlimit /limits:linux_cluster
      time 13:15
  endfamily
  family verify
    trigger ./00:YMD gt ./verify:YMD or ( ./00:YMD eq ./verify:YMD and ./00/ref/tue eq complete)
    repeat date YMD 20091007 20121212 7
    edit ACCOUNT 'oeenfo'
    edit WSHOST 'linux_cluster'
    edit UPDATE_WEB_NOW 'yes'
    edit EMOS_BASE '00'
    edit QUEUE 'emos-e'
    edit QUEUE_EPILOG 'emos-exp'
    edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
    edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
    edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
    edit LOGDIR '/vol/emos_nc/output'
    edit SMSOUT '/vol/emos_nc/output'
    edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/leg/verify'
    edit FSFAMILY 'assim'
    edit FCTOTAL '51'
    edit DELTA_DAY '-27'
    edit RUNHINDCAST '1'
    edit RUNVARFC '1'
    edit EMOS_TYPE '00'
    edit LOGTASK '00'
    inlimit /limits:linux_cluster
    family init
      task list_verif
      task getana
        trigger list_verif eq complete
      task getmofc
        trigger getana eq complete
    endfamily
    family anomaly
      trigger init eq complete
      family temp
        edit VAR 'TEMP'
        task stats_verif
        task plot_verif
          trigger stats_verif eq complete
      endfamily
      family rain
        trigger temp eq complete
        edit VAR 'RAIN'
        task stats_verif
        task plot_verif
          trigger stats_verif eq complete
      endfamily
      family mslp
        trigger rain eq complete
        edit VAR 'MSLP'
        task stats_verif
        task plot_verif
          trigger stats_verif eq complete
      endfamily
      family 2mtm
        trigger mslp eq complete
        edit VAR '2MTM'
        task stats_verif
        task plot_verif
          trigger stats_verif eq complete
      endfamily
      family z500
        trigger ./2mtm eq complete
        task flowz500_verif
      endfamily
      task webvanomaly
        trigger ./z500 eq complete
        edit FSFAMILY 'mofc'
        edit QUEUE 'plots-medium'
    endfamily
    family scores
      trigger anomaly eq complete
      task mofcdaily_verif
      task mofcweekly_verif
        trigger mofcdaily_verif eq complete
      task alldaily
        trigger mofcweekly_verif eq complete
      task allweekly
        trigger alldaily eq complete
      task sctoweb
        trigger allweekly eq complete
        edit FSFAMILY 'mofc'
    endfamily
    family probability
      trigger scores eq complete
      family rain
        edit VAR 'RAIN'
        task table
        task stats
          trigger ./table eq complete
        task stats1
          trigger ./stats eq complete
        task plot_relia
          trigger ./stats1 eq complete
        task plot_roc
          trigger ./plot_relia eq complete
        task plot_rocmap
          trigger ./plot_roc eq complete
        task plot_rpssmap
          trigger ./plot_rocmap eq complete
        task plot_value
          trigger ./plot_rpssmap eq complete
      endfamily
      family temp
        trigger ./rain eq complete
        edit VAR 'TEMP'
        task table
        task stats
          trigger ./table eq complete
        task stats1
          trigger ./stats eq complete
        task plot_relia
          trigger ./stats1 eq complete
        task plot_roc
          trigger ./plot_relia eq complete
        task plot_rocmap
          trigger ./plot_roc eq complete
        task plot_rpssmap
          trigger ./plot_rocmap eq complete
        task plot_value
          trigger ./plot_rpssmap eq complete
      endfamily
      family mslp
        trigger ./temp eq complete
        edit VAR 'MSLP'
        task table
        task stats
          trigger ./table eq complete
        task stats1
          trigger ./stats eq complete
        task plot_relia
          trigger ./stats1 eq complete
        task plot_roc
          trigger ./plot_relia eq complete
        task plot_rocmap
          trigger ./plot_roc eq complete
        task plot_rpssmap
          trigger ./plot_rocmap eq complete
        task plot_value
          trigger ./plot_rpssmap eq complete
      endfamily
      family 2mtm
        trigger ./mslp eq complete
        edit VAR '2MTM'
        task table
        task stats
          trigger ./table eq complete
        task stats1
          trigger ./stats eq complete
        task plot_relia
          trigger ./stats1 eq complete
        task plot_roc
          trigger ./plot_relia eq complete
        task plot_rocmap
          trigger ./plot_roc eq complete
        task plot_rpssmap
          trigger ./plot_rocmap eq complete
        task plot_value
          trigger ./plot_rpssmap eq complete
      endfamily
      family pbtoweb
        trigger ./2mtm eq complete
        edit FSFAMILY 'mofc'
        task reliability
        task roc
          trigger ./reliability eq complete
        task rocmap
          trigger ./roc eq complete
        task rpssmap
          trigger ./rocmap eq complete
        task value
          trigger ./rpssmap eq complete
      endfamily
    endfamily
    task clean_verif
      trigger init eq complete and anomaly eq complete and scores eq complete and probability eq complete
      edit DELTA_DAY '-14'
    task logfilestore
      edit QUEUE 'emos-e'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit KEEPLOGS 'no'
      edit LOGSTOECFS 'no'
      edit MOVE_TO_ECFS 'yes'
      edit SMSFILES '/home/ma/emos/def/emc_36r1'
      edit FSFAMILY 'mc'
    task logfiles
      late -s +00:05 -c +00:35
      trigger logfilestore eq complete and init eq complete and anomaly eq complete and scores eq complete and probability eq complete and clean_verif eq complete
      edit QUEUE 'emos-e'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit FSFAMILY 'mc'
      edit SMSFILES '/home/ma/emos/def/emc_36r1'
      edit MOVE_TO_ECFS 'no'
  endfamily
  family efi_next
    edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/efi'
    edit EVERSION '0046'
    edit IMPL_DATE '2012121212'
    edit EMOS_BASE '00'
    edit MARS_DB 'marsod'
    edit EMOS_TYPE 'undef'
    edit EPS_VERSION '0046'
    edit VERSION '9939'
    edit SMSURLBASE 'http://diss.ecmwf.int/do/monitoring/summary'
    edit SMSURL 'GENFO/00'
    edit UPDATE_WEB_NOW 'yes'
    edit MARS_NEW '0'
    edit QUEUE 'ns'
    edit QUEUE_EPILOG 'ns'
    edit SMSCMD 'smssubmit %USER% %SCHOST% %SMSJOB% %SMSJOBOUT%'
    edit SMSKILL 'smskill %USER% %SCHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
    edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %SCHOST% %SMSRID% %SMSJOB%'
    edit LOGDIR '/s1b/emos_esuite/emos_data/9946/log'
    edit SMSOUT '/s1b/emos_esuite/emos_data/9946/log'
    limit ext 45
    inlimit /limits:c1b
    inlimit ext
    family mn
      trigger ./mn:YMD lt ./clim:YMD
      repeat date YMD 20080221 20121212 1
      family deps
        edit SSMFILES '/home/ma/emos/def/emc_36r1'
        family 00
          trigger ( ( ( /emc_36r1/main/00/legA/postproc eq complete) and ( /emc_36r1/main/00/legB/postproc eq complete)) and ( /emc_36r1/main:YMD eq ../../mn:YMD)) or ( /emc_36r1/main:YMD gt ../../mn:YMD)
          task dummy
            complete 1 eq 1
        endfamily
        family 12
          trigger ( ( ( /emc_36r1/main/12/legA/postproc eq complete) and ( /emc_36r1/main/12/legB/postproc eq complete)) and ( /emc_36r1/main:YMD eq ../../mn:YMD)) or ( /emc_36r1/main:YMD gt ../../mn:YMD)
          task dummy
            complete 1 eq 1
        endfamily
      endfamily
      family 12
        trigger ./deps/12 eq complete
        edit EMOS_BASE '12'
        family op
          edit EPSLEG '1'
          task stage
          family 024
            family 012_036
              edit ACC '024'
              edit EFI_STEP1 '012'
              edit EFI_STEP2 '036'
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 036_060
              edit ACC '024'
              edit EFI_STEP1 '036'
              edit EFI_STEP2 '060'
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 060_084
              edit ACC '024'
              edit EFI_STEP1 '060'
              edit EFI_STEP2 '084'
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 084_108
              edit ACC '024'
              edit EFI_STEP1 '084'
              edit EFI_STEP2 '108'
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 108_132
              edit ACC '024'
              edit EFI_STEP1 '108'
              edit EFI_STEP2 '132'
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 000_024
              edit ACC '024'
              edit EFI_STEP1 '000'
              edit EFI_STEP2 '024'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
            endfamily
            family 024_048
              edit ACC '024'
              edit EFI_STEP1 '024'
              edit EFI_STEP2 '048'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
            endfamily
            family 048_072
              edit ACC '024'
              edit EFI_STEP1 '048'
              edit EFI_STEP2 '072'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
            endfamily
            family 072_096
              edit ACC '024'
              edit EFI_STEP1 '072'
              edit EFI_STEP2 '096'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
            endfamily
            family 096_120
              edit ACC '024'
              edit EFI_STEP1 '096'
              edit EFI_STEP2 '120'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
            endfamily
            family 120_144
              edit ACC '024'
              edit EFI_STEP1 '120'
              edit EFI_STEP2 '144'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
            endfamily
            family 144_168
              edit ACC '024'
              edit EFI_STEP1 '144'
              edit EFI_STEP2 '168'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
            endfamily
            family 168_192
              edit ACC '024'
              edit EFI_STEP1 '168'
              edit EFI_STEP2 '192'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
            endfamily
            family 192_216
              edit ACC '024'
              edit EFI_STEP1 '192'
              edit EFI_STEP2 '216'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
            endfamily
            family 216_240
              edit ACC '024'
              edit EFI_STEP1 '216'
              edit EFI_STEP2 '240'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
            endfamily
          endfamily
          family 120
            trigger ./024 eq complete
            family 000_120
              edit ACC '120'
              edit EFI_STEP1 '000'
              edit EFI_STEP2 '120'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 024_144
              edit ACC '120'
              edit EFI_STEP1 '024'
              edit EFI_STEP2 '144'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
          endfamily
          family 240
            trigger ./024 eq complete
            family 000_240
              edit ACC '240'
              edit EFI_STEP1 '000'
              edit EFI_STEP2 '240'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
          endfamily
        endfamily
        family nop
          edit EPSLEG '2'
          task stage
          family 024
            family 012_036
              edit ACC '024'
              edit EFI_STEP1 '012'
              edit EFI_STEP2 '036'
              family 2tmin
                edit RF_NAME '2tmin'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 036_060
              edit ACC '024'
              edit EFI_STEP1 '036'
              edit EFI_STEP2 '060'
              family 2tmin
                edit RF_NAME '2tmin'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 060_084
              edit ACC '024'
              edit EFI_STEP1 '060'
              edit EFI_STEP2 '084'
              family 2tmin
                edit RF_NAME '2tmin'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 084_108
              edit ACC '024'
              edit EFI_STEP1 '084'
              edit EFI_STEP2 '108'
              family 2tmin
                edit RF_NAME '2tmin'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 108_132
              edit ACC '024'
              edit EFI_STEP1 '108'
              edit EFI_STEP2 '132'
              family 2tmin
                edit RF_NAME '2tmin'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 132_156
              edit ACC '024'
              edit EFI_STEP1 '132'
              edit EFI_STEP2 '156'
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 156_180
              edit ACC '024'
              edit EFI_STEP1 '156'
              edit EFI_STEP2 '180'
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 000_024
              edit ACC '024'
              edit EFI_STEP1 '000'
              edit EFI_STEP2 '024'
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 024_048
              edit ACC '024'
              edit EFI_STEP1 '024'
              edit EFI_STEP2 '048'
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 048_072
              edit ACC '024'
              edit EFI_STEP1 '048'
              edit EFI_STEP2 '072'
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 072_096
              edit ACC '024'
              edit EFI_STEP1 '072'
              edit EFI_STEP2 '096'
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 096_120
              edit ACC '024'
              edit EFI_STEP1 '096'
              edit EFI_STEP2 '120'
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 120_144
              edit ACC '024'
              edit EFI_STEP1 '120'
              edit EFI_STEP2 '144'
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 144_168
              edit ACC '024'
              edit EFI_STEP1 '144'
              edit EFI_STEP2 '168'
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 180_204
              edit ACC '024'
              edit EFI_STEP1 '180'
              edit EFI_STEP2 '204'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 204_228
              edit ACC '024'
              edit EFI_STEP1 '204'
              edit EFI_STEP2 '228'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 168_192
              edit ACC '24'
              edit EFI_STEP1 '168'
              edit EFI_STEP2 '192'
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 192_216
              edit ACC '24'
              edit EFI_STEP1 '192'
              edit EFI_STEP2 '216'
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 216_240
              edit ACC '24'
              edit EFI_STEP1 '216'
              edit EFI_STEP2 '240'
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 240_264
              edit ACC '24'
              edit EFI_STEP1 '240'
              edit EFI_STEP2 '264'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 264_288
              edit ACC '24'
              edit EFI_STEP1 '264'
              edit EFI_STEP2 '288'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 288_312
              edit ACC '24'
              edit EFI_STEP1 '288'
              edit EFI_STEP2 '312'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 312_336
              edit ACC '24'
              edit EFI_STEP1 '312'
              edit EFI_STEP2 '336'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 336_360
              edit ACC '24'
              edit EFI_STEP1 '336'
              edit EFI_STEP2 '360'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
          endfamily
          family 072
            family 012_084
              edit ACC '072'
              edit EFI_STEP1 '012'
              edit EFI_STEP2 '084'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 036_108
              edit ACC '072'
              edit EFI_STEP1 '036'
              edit EFI_STEP2 '108'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 060_132
              edit ACC '072'
              edit EFI_STEP1 '060'
              edit EFI_STEP2 '132'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 084_156
              edit ACC '072'
              edit EFI_STEP1 '084'
              edit EFI_STEP2 '156'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 108_180
              edit ACC '072'
              edit EFI_STEP1 '108'
              edit EFI_STEP2 '180'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 132_204
              edit ACC '072'
              edit EFI_STEP1 '132'
              edit EFI_STEP2 '204'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 156_228
              edit ACC '072'
              edit EFI_STEP1 '156'
              edit EFI_STEP2 '228'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
          endfamily
          family 120
            family 012_132
              edit ACC '120'
              edit EFI_STEP1 '012'
              edit EFI_STEP2 '132'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 036_156
              edit ACC '120'
              edit EFI_STEP1 '036'
              edit EFI_STEP2 '156'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 060_180
              edit ACC '120'
              edit EFI_STEP1 '060'
              edit EFI_STEP2 '180'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 084_204
              edit ACC '120'
              edit EFI_STEP1 '084'
              edit EFI_STEP2 '204'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 108_228
              edit ACC '120'
              edit EFI_STEP1 '108'
              edit EFI_STEP2 '228'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 240_360
              edit ACC '120'
              edit EFI_STEP1 '240'
              edit EFI_STEP2 '360'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
          endfamily
          family 240
            family 000_240
              edit ACC '240'
              edit EFI_STEP1 '000'
              edit EFI_STEP2 '240'
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
          endfamily
          family 360
            family 000_360
              edit ACC '360'
              edit EFI_STEP1 '000'
              edit EFI_STEP2 '360'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
          endfamily
        endfamily
        family arch_op
          trigger op eq complete
          edit EPSLEG '1'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
          task efi2fdb
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/efi_hdc'
            event none
          family arch
            trigger ./efi2fdb eq complete
            family 2t
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME '2t'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
            family 10ff
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME '10ff'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
            family 10fg
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME '10fg'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
            family tp
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME 'tp'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
          endfamily
        endfamily
        family arch_nop
          trigger nop eq complete
          edit EPSLEG '2'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
          task efi2fdb
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/efi_hdc'
            event none
          family arch
            trigger ./efi2fdb eq complete
            family 2t
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME '2t'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
            family 10ff
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME '10ff'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
            family 10fg
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME '10fg'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
            family tp
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME 'tp'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
            family 2tmin
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME '2tmin'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
            family 2tmax
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME '2tmax'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
            family ts
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME 'ts'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
            family hsttmax
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME 'hsstmax'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
          endfamily
        endfamily
        family data
          trigger arch_op eq complete and arch_nop eq complete
          edit WSHOST 'linux_cluster'
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/smsfiles/web'
          inlimit /limits:linux_cluster
          family 2t
            edit PARAM '2t'
            task efidata
          endfamily
          family 2tmin
            edit PARAM '2tmin'
            task efidata
          endfamily
          family 2tmax
            edit PARAM '2tmax'
            task efidata
          endfamily
          family tp
            edit PARAM 'tp'
            task efidata
          endfamily
          family 10ff
            edit PARAM '10ff'
            task efidata
          endfamily
          family 10fg
            edit PARAM '10fg'
            task efidata
          endfamily
          family sf
            edit PARAM 'sf'
            task efidata
          endfamily
          family hsstmax
            edit PARAM 'hsstmax'
            task efidata
              edit PARAM 'hsttmax'
          endfamily
        endfamily
        family plot
          trigger data eq complete
          edit WSHOST 'linux_cluster'
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit SMSURLBASE 'http://intra.ecmwf.int/plots'
          edit SMSURL 'd/inspect/esuite/esuite_33r1/medium/eps/efi/'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/smsfiles/web'
          edit METVIEW 'metview_new'
          inlimit /limits:linux_cluster
          family 2t
            edit PARAM '2t'
            task efiplot
          endfamily
          family 2tmin
            edit PARAM '2tmin'
            task efiplot
          endfamily
          family 2tmax
            edit PARAM '2tmax'
            task efiplot
          endfamily
          family tp
            edit PARAM 'tp'
            task efiplot
          endfamily
          family 10ff
            edit PARAM '10ff'
            task efiplot
          endfamily
          family 10fg
            edit PARAM '10fg'
            task efiplot
          endfamily
          family sf
            edit PARAM 'sf'
            task efiplot
          endfamily
          family hsstmax
            edit PARAM 'hsstmax'
            task efiplot
              edit PARAM 'hsstmax'
          endfamily
          family hsttmax
            edit PARAM 'hsttmax'
            task efiplot
          endfamily
        endfamily
      endfamily
      family 00
        trigger ( ./deps/00 eq complete or ./deps/00 eq unknown) and ( ./12 eq complete or ./12 eq unknown)
        edit EMOS_BASE '00'
        edit DELTA_DAY '1'
        edit STEP1 '-1'
        family op
          edit EPSLEG '1'
          task stage
          family 024
            family 000_024
              edit ACC '024'
              edit EFI_STEP1 '000'
              edit EFI_STEP2 '024'
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 024_048
              edit ACC '024'
              edit EFI_STEP1 '024'
              edit EFI_STEP2 '048'
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 048_072
              edit ACC '024'
              edit EFI_STEP1 '048'
              edit EFI_STEP2 '072'
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 072_096
              edit ACC '024'
              edit EFI_STEP1 '072'
              edit EFI_STEP2 '096'
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 096_120
              edit ACC '024'
              edit EFI_STEP1 '096'
              edit EFI_STEP2 '120'
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 120_144
              edit ACC '024'
              edit EFI_STEP1 '120'
              edit EFI_STEP2 '144'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
            endfamily
            family 144_168
              edit ACC '024'
              edit EFI_STEP1 '144'
              edit EFI_STEP2 '168'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
            endfamily
            family 168_192
              edit ACC '024'
              edit EFI_STEP1 '168'
              edit EFI_STEP2 '192'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
            endfamily
            family 192_216
              edit ACC '024'
              edit EFI_STEP1 '192'
              edit EFI_STEP2 '216'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
            endfamily
            family 216_240
              edit ACC '024'
              edit EFI_STEP1 '216'
              edit EFI_STEP2 '240'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
            endfamily
          endfamily
          family 120
            trigger ./024 eq complete
            family 000_120
              edit ACC '120'
              edit EFI_STEP1 '000'
              edit EFI_STEP2 '120'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 024_144
              edit ACC '120'
              edit EFI_STEP1 '024'
              edit EFI_STEP2 '144'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
          endfamily
          family 240
            trigger ./024 eq complete
            family 000_240
              edit ACC '240'
              edit EFI_STEP1 '000'
              edit EFI_STEP2 '240'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
          endfamily
        endfamily
        family nop
          edit EPSLEG '2'
          task stage
          family 024
            family 000_024
              edit ACC '024'
              edit EFI_STEP1 '000'
              edit EFI_STEP2 '024'
              family 2tmin
                edit RF_NAME '2tmin'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 024_048
              edit ACC '024'
              edit EFI_STEP1 '024'
              edit EFI_STEP2 '048'
              family 2tmin
                edit RF_NAME '2tmin'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 048_072
              edit ACC '024'
              edit EFI_STEP1 '048'
              edit EFI_STEP2 '072'
              family 2tmin
                edit RF_NAME '2tmin'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 072_096
              edit ACC '024'
              edit EFI_STEP1 '072'
              edit EFI_STEP2 '096'
              family 2tmin
                edit RF_NAME '2tmin'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 096_120
              edit ACC '024'
              edit EFI_STEP1 '096'
              edit EFI_STEP2 '120'
              family 2tmin
                edit RF_NAME '2tmin'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family reproc
              family 120_144
                edit ACC '024'
                edit EFI_STEP1 '120'
                edit EFI_STEP2 '144'
                family tp
                  edit RF_NAME 'tp'
                  task extindex
                endfamily
              endfamily
              family 144_168
                edit ACC '024'
                edit EFI_STEP1 '144'
                edit EFI_STEP2 '168'
                family tp
                  edit RF_NAME 'tp'
                  task extindex
                endfamily
              endfamily
            endfamily
            family 120_144
              edit ACC '024'
              edit EFI_STEP1 '120'
              edit EFI_STEP2 '144'
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 144_168
              edit ACC '024'
              edit EFI_STEP1 '144'
              edit EFI_STEP2 '168'
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10fg
                edit RF_NAME '10fg'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2tmin
                edit RF_NAME '2tmin'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2tmax
                edit RF_NAME '2tmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family hsttmax
                edit RF_NAME 'hsttmax'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family ts
                edit RF_NAME 'ts'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 168_192
              edit ACC '24'
              edit EFI_STEP1 '168'
              edit EFI_STEP2 '192'
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 192_216
              edit ACC '24'
              edit EFI_STEP1 '192'
              edit EFI_STEP2 '216'
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 216_240
              edit ACC '24'
              edit EFI_STEP1 '216'
              edit EFI_STEP2 '240'
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 240_264
              edit ACC '24'
              edit EFI_STEP1 '240'
              edit EFI_STEP2 '264'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 264_288
              edit ACC '24'
              edit EFI_STEP1 '264'
              edit EFI_STEP2 '288'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 288_312
              edit ACC '24'
              edit EFI_STEP1 '288'
              edit EFI_STEP2 '312'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 312_336
              edit ACC '24'
              edit EFI_STEP1 '312'
              edit EFI_STEP2 '336'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
            family 336_360
              edit ACC '24'
              edit EFI_STEP1 '336'
              edit EFI_STEP2 '360'
              family tp
                edit RF_NAME 'tp'
                task prepeps
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
              endfamily
            endfamily
          endfamily
          family 072
            family 000_072
              edit ACC '072'
              edit EFI_STEP1 '000'
              edit EFI_STEP2 '072'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 024_096
              edit ACC '072'
              edit EFI_STEP1 '024'
              edit EFI_STEP2 '096'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 048_120
              edit ACC '072'
              edit EFI_STEP1 '048'
              edit EFI_STEP2 '120'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 072_144
              edit ACC '072'
              edit EFI_STEP1 '072'
              edit EFI_STEP2 '144'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 096_168
              edit ACC '072'
              edit EFI_STEP1 '096'
              edit EFI_STEP2 '168'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 120_192
              edit ACC '072'
              edit EFI_STEP1 '120'
              edit EFI_STEP2 '192'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 144_216
              edit ACC '072'
              edit EFI_STEP1 '144'
              edit EFI_STEP2 '216'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
          endfamily
          family 120
            family 000_120
              edit ACC '120'
              edit EFI_STEP1 '000'
              edit EFI_STEP2 '120'
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 024_144
              edit ACC '120'
              edit EFI_STEP1 '024'
              edit EFI_STEP2 '144'
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 048_168
              edit ACC '120'
              edit EFI_STEP1 '048'
              edit EFI_STEP2 '168'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 072_192
              edit ACC '120'
              edit EFI_STEP1 '072'
              edit EFI_STEP2 '192'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 096_216
              edit ACC '120'
              edit EFI_STEP1 '096'
              edit EFI_STEP2 '216'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
            family 240_360
              edit ACC '120'
              edit EFI_STEP1 '240'
              edit EFI_STEP2 '360'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
          endfamily
          family 240
            family 000_240
              edit ACC '240'
              edit EFI_STEP1 '000'
              edit EFI_STEP2 '240'
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
          endfamily
          family 360
            family 000_360
              edit ACC '360'
              edit EFI_STEP1 '000'
              edit EFI_STEP2 '360'
              family tp
                edit RF_NAME 'tp'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 2t
                edit RF_NAME '2t'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
              family 10ff
                edit RF_NAME '10ff'
                task prepeps
                task extindex
                  trigger ./prepeps eq complete
              endfamily
            endfamily
          endfamily
        endfamily
        family arch_op
          trigger op eq complete
          edit EPSLEG '1'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
          task efi2fdb
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/efi_hdc'
            event none
          family arch
            trigger ./efi2fdb eq complete
            family 2t
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME '2t'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
            family 10ff
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME '10ff'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
            family 10fg
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME '10fg'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
            family tp
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME 'tp'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
          endfamily
        endfamily
        family arch_nop
          trigger nop eq complete
          edit EPSLEG '2'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
          task efi2fdb
            edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/efi_hdc'
            event none
          family arch
            trigger ./efi2fdb eq complete
            family 2t
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME '2t'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
            family 10ff
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME '10ff'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
            family 10fg
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME '10fg'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
            family tp
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME 'tp'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
            family 2tmin
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME '2tmin'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
            family 2tmax
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME '2tmax'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
            family ts
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME 'ts'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
            family hsttmax
              edit step1 '24'
              edit step2 '24'
              edit RF_NAME 'hsstmax'
              task efi
                label info ""
              family sot
                edit EFI_TYPE 'sot'
                task efi
              endfamily
            endfamily
          endfamily
        endfamily
        family data
          trigger arch_op eq complete and arch_nop eq complete
          edit WSHOST 'linux_cluster'
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/smsfiles/web'
          inlimit /limits:linux_cluster
          family 2t
            edit PARAM '2t'
            task efidata
          endfamily
          family 2tmin
            edit PARAM '2tmin'
            task efidata
          endfamily
          family 2tmax
            edit PARAM '2tmax'
            task efidata
          endfamily
          family tp
            edit PARAM 'tp'
            task efidata
          endfamily
          family 10ff
            edit PARAM '10ff'
            task efidata
          endfamily
          family 10fg
            edit PARAM '10fg'
            task efidata
          endfamily
          family sf
            edit PARAM 'sf'
            task efidata
          endfamily
          family hsstmax
            edit PARAM 'hsstmax'
            task efidata
          endfamily
        endfamily
        family plot
          trigger data eq complete
          edit WSHOST 'linux_cluster'
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit SMSURLBASE 'http://intra.ecmwf.int/plots'
          edit SMSURL 'd/inspect/esuite/esuite_33r1/medium/eps/efi/'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/smsfiles/web'
          edit METVIEW 'metview_new'
          inlimit /limits:linux_cluster
          family 2t
            edit PARAM '2t'
            task efiplot
          endfamily
          family 2tmin
            edit PARAM '2tmin'
            task efiplot
          endfamily
          family 2tmax
            edit PARAM '2tmax'
            task efiplot
          endfamily
          family tp
            edit PARAM 'tp'
            task efiplot
          endfamily
          family 10ff
            edit PARAM '10ff'
            task efiplot
          endfamily
          family 10fg
            edit PARAM '10fg'
            task efiplot
          endfamily
          family sf
            edit PARAM 'sf'
            task efiplot
          endfamily
          family hsstmax
            edit PARAM 'hsstmax'
            task efiplot
              edit PARAM 'hsttmax'
          endfamily
          family hsttmax
            edit PARAM 'hsttmax'
            task efiplot
              edit PARAM 'hsstmax'
          endfamily
        endfamily
      endfamily
      task efi_cleaner
        trigger ( ./12 eq complete or ./12 eq unknown) and ( ./00 eq complete or ./00 eq unknown)
    endfamily
    family hind
      repeat date YMD 20080717 20121212 7
      edit DELTA_DAY '0'
      family stage
        edit RF_NAME 'all'
        edit RETRIEVE 'stage'
        task prep
      endfamily
      family op
        trigger ./stage eq complete
        edit EPSLEG '1'
        family 024
          family 000_024
            edit ACC '024'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '024'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 012_036
            edit ACC '024'
            edit EFI_STEP1 '012'
            edit EFI_STEP2 '036'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 024_048
            edit ACC '024'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '048'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 036_060
            edit ACC '024'
            edit EFI_STEP1 '036'
            edit EFI_STEP2 '060'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 048_072
            edit ACC '024'
            edit EFI_STEP1 '048'
            edit EFI_STEP2 '072'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 060_084
            edit ACC '024'
            edit EFI_STEP1 '060'
            edit EFI_STEP2 '084'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 072_096
            edit ACC '024'
            edit EFI_STEP1 '072'
            edit EFI_STEP2 '096'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 084_108
            edit ACC '024'
            edit EFI_STEP1 '084'
            edit EFI_STEP2 '108'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 096_120
            edit ACC '024'
            edit EFI_STEP1 '096'
            edit EFI_STEP2 '120'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 108_132
            edit ACC '024'
            edit EFI_STEP1 '108'
            edit EFI_STEP2 '132'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 120_144
            edit ACC '024'
            edit EFI_STEP1 '120'
            edit EFI_STEP2 '144'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 144_168
            edit ACC '024'
            edit EFI_STEP1 '144'
            edit EFI_STEP2 '168'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 168_192
            edit ACC '024'
            edit EFI_STEP1 '168'
            edit EFI_STEP2 '192'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 192_216
            edit ACC '024'
            edit EFI_STEP1 '192'
            edit EFI_STEP2 '216'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 216_240
            edit ACC '024'
            edit EFI_STEP1 '216'
            edit EFI_STEP2 '240'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
        endfamily
        family 120
          trigger ./024 eq complete
          family 000_120
            edit ACC '120'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '120'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 024_144
            edit ACC '120'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '144'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
        endfamily
        family 240
          trigger ./024 eq complete
          family 000_240
            edit ACC '240'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '240'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
        endfamily
      endfamily
      family nop
        trigger ./stage eq complete
        edit EPSLEG '2'
        family 024
          family 000_024
            edit ACC '024'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '024'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
          endfamily
          family 012_036
            edit ACC '024'
            edit EFI_STEP1 '012'
            edit EFI_STEP2 '036'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
          endfamily
          family 024_048
            edit ACC '024'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '048'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
          endfamily
          family 036_060
            edit ACC '024'
            edit EFI_STEP1 '036'
            edit EFI_STEP2 '060'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
          endfamily
          family 048_072
            edit ACC '024'
            edit EFI_STEP1 '048'
            edit EFI_STEP2 '072'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
          endfamily
          family 060_084
            edit ACC '024'
            edit EFI_STEP1 '060'
            edit EFI_STEP2 '084'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
          endfamily
          family 072_096
            edit ACC '024'
            edit EFI_STEP1 '072'
            edit EFI_STEP2 '096'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
          endfamily
          family 084_108
            edit ACC '024'
            edit EFI_STEP1 '084'
            edit EFI_STEP2 '108'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
          endfamily
          family 096_120
            edit ACC '024'
            edit EFI_STEP1 '096'
            edit EFI_STEP2 '120'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
          endfamily
          family 108_132
            edit ACC '024'
            edit EFI_STEP1 '108'
            edit EFI_STEP2 '132'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
          endfamily
          family 120_144
            edit ACC '024'
            edit EFI_STEP1 '120'
            edit EFI_STEP2 '144'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
          endfamily
          family 144_168
            edit ACC '024'
            edit EFI_STEP1 '144'
            edit EFI_STEP2 '168'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
          endfamily
          family 168_192
            edit ACC '024'
            edit EFI_STEP1 '168'
            edit EFI_STEP2 '192'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
          endfamily
          family 192_216
            edit ACC '024'
            edit EFI_STEP1 '192'
            edit EFI_STEP2 '216'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
          endfamily
          family 216_240
            edit ACC '024'
            edit EFI_STEP1 '216'
            edit EFI_STEP2 '240'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
          endfamily
          family 240_264
            edit ACC '024'
            edit EFI_STEP1 '240'
            edit EFI_STEP2 '264'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
          endfamily
          family 264_288
            edit ACC '024'
            edit EFI_STEP1 '264'
            edit EFI_STEP2 '288'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
          endfamily
          family 288_312
            edit ACC '024'
            edit EFI_STEP1 '288'
            edit EFI_STEP2 '312'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
          endfamily
          family 312_336
            edit ACC '024'
            edit EFI_STEP1 '312'
            edit EFI_STEP2 '336'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
          endfamily
          family 336_360
            edit ACC '024'
            edit EFI_STEP1 '336'
            edit EFI_STEP2 '360'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
          endfamily
        endfamily
        family 000
          family 006_006
            edit ACC '000'
            edit EFI_STEP1 '006'
            edit EFI_STEP2 '006'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 012_012
            edit ACC '000'
            edit EFI_STEP1 '012'
            edit EFI_STEP2 '012'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 018_018
            edit ACC '000'
            edit EFI_STEP1 '018'
            edit EFI_STEP2 '018'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 024_024
            edit ACC '000'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '024'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 030_030
            edit ACC '000'
            edit EFI_STEP1 '030'
            edit EFI_STEP2 '030'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 036_036
            edit ACC '000'
            edit EFI_STEP1 '036'
            edit EFI_STEP2 '036'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 042_042
            edit ACC '000'
            edit EFI_STEP1 '042'
            edit EFI_STEP2 '042'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 048_048
            edit ACC '000'
            edit EFI_STEP1 '048'
            edit EFI_STEP2 '048'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 054_054
            edit ACC '000'
            edit EFI_STEP1 '054'
            edit EFI_STEP2 '054'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 060_060
            edit ACC '000'
            edit EFI_STEP1 '060'
            edit EFI_STEP2 '060'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 066_066
            edit ACC '000'
            edit EFI_STEP1 '066'
            edit EFI_STEP2 '066'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 072_072
            edit ACC '000'
            edit EFI_STEP1 '072'
            edit EFI_STEP2 '072'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 078_078
            edit ACC '000'
            edit EFI_STEP1 '078'
            edit EFI_STEP2 '078'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 084_084
            edit ACC '000'
            edit EFI_STEP1 '084'
            edit EFI_STEP2 '084'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 090_090
            edit ACC '000'
            edit EFI_STEP1 '090'
            edit EFI_STEP2 '090'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 096_096
            edit ACC '000'
            edit EFI_STEP1 '096'
            edit EFI_STEP2 '096'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 102_102
            edit ACC '000'
            edit EFI_STEP1 '102'
            edit EFI_STEP2 '102'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 108_108
            edit ACC '000'
            edit EFI_STEP1 '108'
            edit EFI_STEP2 '108'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 114_114
            edit ACC '000'
            edit EFI_STEP1 '114'
            edit EFI_STEP2 '114'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 120_120
            edit ACC '000'
            edit EFI_STEP1 '120'
            edit EFI_STEP2 '120'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 126_126
            edit ACC '000'
            edit EFI_STEP1 '126'
            edit EFI_STEP2 '126'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 132_132
            edit ACC '000'
            edit EFI_STEP1 '132'
            edit EFI_STEP2 '132'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 138_138
            edit ACC '000'
            edit EFI_STEP1 '138'
            edit EFI_STEP2 '138'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 144_144
            edit ACC '000'
            edit EFI_STEP1 '144'
            edit EFI_STEP2 '144'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 150_150
            edit ACC '000'
            edit EFI_STEP1 '150'
            edit EFI_STEP2 '150'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 156_156
            edit ACC '000'
            edit EFI_STEP1 '156'
            edit EFI_STEP2 '156'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 162_162
            edit ACC '000'
            edit EFI_STEP1 '162'
            edit EFI_STEP2 '162'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 168_168
            edit ACC '000'
            edit EFI_STEP1 '168'
            edit EFI_STEP2 '168'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 174_174
            edit ACC '000'
            edit EFI_STEP1 '174'
            edit EFI_STEP2 '174'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 180_180
            edit ACC '000'
            edit EFI_STEP1 '180'
            edit EFI_STEP2 '180'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 186_186
            edit ACC '000'
            edit EFI_STEP1 '186'
            edit EFI_STEP2 '186'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 192_192
            edit ACC '000'
            edit EFI_STEP1 '192'
            edit EFI_STEP2 '192'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 198_198
            edit ACC '000'
            edit EFI_STEP1 '198'
            edit EFI_STEP2 '198'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 204_204
            edit ACC '000'
            edit EFI_STEP1 '204'
            edit EFI_STEP2 '204'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 210_210
            edit ACC '000'
            edit EFI_STEP1 '210'
            edit EFI_STEP2 '210'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 216_216
            edit ACC '000'
            edit EFI_STEP1 '216'
            edit EFI_STEP2 '216'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 222_222
            edit ACC '000'
            edit EFI_STEP1 '222'
            edit EFI_STEP2 '222'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 228_228
            edit ACC '000'
            edit EFI_STEP1 '228'
            edit EFI_STEP2 '228'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 234_234
            edit ACC '000'
            edit EFI_STEP1 '234'
            edit EFI_STEP2 '234'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 240_240
            edit ACC '000'
            edit EFI_STEP1 '240'
            edit EFI_STEP2 '240'
            family 10dd
              edit RF_NAME '10dd'
              task prep
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 252_252
            edit ACC '000'
            edit EFI_STEP1 '252'
            edit EFI_STEP2 '252'
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 264_264
            edit ACC '000'
            edit EFI_STEP1 '264'
            edit EFI_STEP2 '264'
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 276_276
            edit ACC '000'
            edit EFI_STEP1 '276'
            edit EFI_STEP2 '276'
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 288_288
            edit ACC '000'
            edit EFI_STEP1 '288'
            edit EFI_STEP2 '288'
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 300_300
            edit ACC '000'
            edit EFI_STEP1 '300'
            edit EFI_STEP2 '300'
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 312_312
            edit ACC '000'
            edit EFI_STEP1 '312'
            edit EFI_STEP2 '312'
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 324_324
            edit ACC '000'
            edit EFI_STEP1 '324'
            edit EFI_STEP2 '324'
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 336_336
            edit ACC '000'
            edit EFI_STEP1 '336'
            edit EFI_STEP2 '336'
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 348_348
            edit ACC '000'
            edit EFI_STEP1 '348'
            edit EFI_STEP2 '348'
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
          family 360_360
            edit ACC '000'
            edit EFI_STEP1 '360'
            edit EFI_STEP2 '360'
            family 850t
              edit RF_NAME '850t'
              task prep
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep
            endfamily
          endfamily
        endfamily
        family 006
          family 000_006
            edit ACC '006'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '006'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 006_012
            edit ACC '006'
            edit EFI_STEP1 '006'
            edit EFI_STEP2 '012'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 012_018
            edit ACC '006'
            edit EFI_STEP1 '012'
            edit EFI_STEP2 '018'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 018_024
            edit ACC '006'
            edit EFI_STEP1 '018'
            edit EFI_STEP2 '024'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 024_030
            edit ACC '006'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '030'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 030_036
            edit ACC '006'
            edit EFI_STEP1 '030'
            edit EFI_STEP2 '036'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 036_042
            edit ACC '006'
            edit EFI_STEP1 '036'
            edit EFI_STEP2 '042'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 042_048
            edit ACC '006'
            edit EFI_STEP1 '042'
            edit EFI_STEP2 '048'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 048_054
            edit ACC '006'
            edit EFI_STEP1 '048'
            edit EFI_STEP2 '054'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 054_060
            edit ACC '006'
            edit EFI_STEP1 '054'
            edit EFI_STEP2 '060'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 060_066
            edit ACC '006'
            edit EFI_STEP1 '060'
            edit EFI_STEP2 '066'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 066_072
            edit ACC '006'
            edit EFI_STEP1 '066'
            edit EFI_STEP2 '072'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 072_078
            edit ACC '006'
            edit EFI_STEP1 '072'
            edit EFI_STEP2 '078'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 078_084
            edit ACC '006'
            edit EFI_STEP1 '078'
            edit EFI_STEP2 '084'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 084_090
            edit ACC '006'
            edit EFI_STEP1 '084'
            edit EFI_STEP2 '090'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 090_096
            edit ACC '006'
            edit EFI_STEP1 '090'
            edit EFI_STEP2 '096'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 096_102
            edit ACC '006'
            edit EFI_STEP1 '096'
            edit EFI_STEP2 '102'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 102_108
            edit ACC '006'
            edit EFI_STEP1 '102'
            edit EFI_STEP2 '108'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 108_114
            edit ACC '006'
            edit EFI_STEP1 '108'
            edit EFI_STEP2 '114'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 114_120
            edit ACC '006'
            edit EFI_STEP1 '114'
            edit EFI_STEP2 '120'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 120_126
            edit ACC '006'
            edit EFI_STEP1 '120'
            edit EFI_STEP2 '126'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 126_132
            edit ACC '006'
            edit EFI_STEP1 '126'
            edit EFI_STEP2 '132'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 132_138
            edit ACC '006'
            edit EFI_STEP1 '132'
            edit EFI_STEP2 '138'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 138_144
            edit ACC '006'
            edit EFI_STEP1 '138'
            edit EFI_STEP2 '144'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 144_150
            edit ACC '006'
            edit EFI_STEP1 '144'
            edit EFI_STEP2 '150'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 150_156
            edit ACC '006'
            edit EFI_STEP1 '150'
            edit EFI_STEP2 '156'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 156_162
            edit ACC '006'
            edit EFI_STEP1 '156'
            edit EFI_STEP2 '162'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 162_168
            edit ACC '006'
            edit EFI_STEP1 '162'
            edit EFI_STEP2 '168'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 168_174
            edit ACC '006'
            edit EFI_STEP1 '168'
            edit EFI_STEP2 '174'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 174_180
            edit ACC '006'
            edit EFI_STEP1 '174'
            edit EFI_STEP2 '180'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 180_186
            edit ACC '006'
            edit EFI_STEP1 '180'
            edit EFI_STEP2 '186'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 186_192
            edit ACC '006'
            edit EFI_STEP1 '186'
            edit EFI_STEP2 '192'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 192_198
            edit ACC '006'
            edit EFI_STEP1 '192'
            edit EFI_STEP2 '198'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 198_204
            edit ACC '006'
            edit EFI_STEP1 '198'
            edit EFI_STEP2 '204'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 204_210
            edit ACC '006'
            edit EFI_STEP1 '204'
            edit EFI_STEP2 '210'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 210_216
            edit ACC '006'
            edit EFI_STEP1 '210'
            edit EFI_STEP2 '216'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 216_222
            edit ACC '006'
            edit EFI_STEP1 '216'
            edit EFI_STEP2 '222'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 222_228
            edit ACC '006'
            edit EFI_STEP1 '222'
            edit EFI_STEP2 '228'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 228_234
            edit ACC '006'
            edit EFI_STEP1 '228'
            edit EFI_STEP2 '234'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
          family 234_240
            edit ACC '006'
            edit EFI_STEP1 '234'
            edit EFI_STEP2 '240'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep
            endfamily
          endfamily
        endfamily
        family 012
          family 000_012
            edit ACC '012'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '012'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 012_024
            edit ACC '012'
            edit EFI_STEP1 '012'
            edit EFI_STEP2 '024'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 024_036
            edit ACC '012'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '036'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 036_048
            edit ACC '012'
            edit EFI_STEP1 '036'
            edit EFI_STEP2 '048'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 048_060
            edit ACC '012'
            edit EFI_STEP1 '048'
            edit EFI_STEP2 '060'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 060_072
            edit ACC '012'
            edit EFI_STEP1 '060'
            edit EFI_STEP2 '072'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 072_084
            edit ACC '012'
            edit EFI_STEP1 '072'
            edit EFI_STEP2 '084'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 084_096
            edit ACC '012'
            edit EFI_STEP1 '084'
            edit EFI_STEP2 '096'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 096_108
            edit ACC '012'
            edit EFI_STEP1 '096'
            edit EFI_STEP2 '108'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 108_120
            edit ACC '012'
            edit EFI_STEP1 '108'
            edit EFI_STEP2 '120'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 120_132
            edit ACC '012'
            edit EFI_STEP1 '120'
            edit EFI_STEP2 '132'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 132_144
            edit ACC '012'
            edit EFI_STEP1 '132'
            edit EFI_STEP2 '144'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 144_156
            edit ACC '012'
            edit EFI_STEP1 '144'
            edit EFI_STEP2 '156'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 156_168
            edit ACC '012'
            edit EFI_STEP1 '156'
            edit EFI_STEP2 '168'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 168_180
            edit ACC '012'
            edit EFI_STEP1 '168'
            edit EFI_STEP2 '180'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 180_192
            edit ACC '012'
            edit EFI_STEP1 '180'
            edit EFI_STEP2 '192'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 192_204
            edit ACC '012'
            edit EFI_STEP1 '192'
            edit EFI_STEP2 '204'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 204_216
            edit ACC '012'
            edit EFI_STEP1 '204'
            edit EFI_STEP2 '216'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 216_228
            edit ACC '012'
            edit EFI_STEP1 '216'
            edit EFI_STEP2 '228'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 228_240
            edit ACC '012'
            edit EFI_STEP1 '228'
            edit EFI_STEP2 '240'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 240_252
            edit ACC '012'
            edit EFI_STEP1 '240'
            edit EFI_STEP2 '252'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 252_264
            edit ACC '012'
            edit EFI_STEP1 '252'
            edit EFI_STEP2 '264'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 264_276
            edit ACC '012'
            edit EFI_STEP1 '264'
            edit EFI_STEP2 '276'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 276_288
            edit ACC '012'
            edit EFI_STEP1 '276'
            edit EFI_STEP2 '288'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 288_300
            edit ACC '012'
            edit EFI_STEP1 '288'
            edit EFI_STEP2 '300'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 300_312
            edit ACC '012'
            edit EFI_STEP1 '300'
            edit EFI_STEP2 '312'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 312_324
            edit ACC '012'
            edit EFI_STEP1 '312'
            edit EFI_STEP2 '324'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 324_336
            edit ACC '012'
            edit EFI_STEP1 '324'
            edit EFI_STEP2 '336'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 336_348
            edit ACC '012'
            edit EFI_STEP1 '336'
            edit EFI_STEP2 '348'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
          family 348_360
            edit ACC '012'
            edit EFI_STEP1 '348'
            edit EFI_STEP2 '360'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
          endfamily
        endfamily
        family 072
          trigger ./024 eq complete
          family 000_072
            edit ACC '072'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '072'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 024_096
            edit ACC '072'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '096'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 048_120
            edit ACC '072'
            edit EFI_STEP1 '048'
            edit EFI_STEP2 '120'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 072_144
            edit ACC '072'
            edit EFI_STEP1 '072'
            edit EFI_STEP2 '144'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 096_168
            edit ACC '072'
            edit EFI_STEP1 '096'
            edit EFI_STEP2 '168'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 120_192
            edit ACC '072'
            edit EFI_STEP1 '120'
            edit EFI_STEP2 '192'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 144_216
            edit ACC '072'
            edit EFI_STEP1 '144'
            edit EFI_STEP2 '216'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 168_240
            edit ACC '072'
            edit EFI_STEP1 '168'
            edit EFI_STEP2 '240'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
        endfamily
        family 120
          trigger ./024 eq complete
          family 000_120
            edit ACC '120'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '120'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 024_144
            edit ACC '120'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '144'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 048_168
            edit ACC '120'
            edit EFI_STEP1 '048'
            edit EFI_STEP2 '168'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 072_192
            edit ACC '120'
            edit EFI_STEP1 '072'
            edit EFI_STEP2 '192'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 096_216
            edit ACC '120'
            edit EFI_STEP1 '096'
            edit EFI_STEP2 '216'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 120_240
            edit ACC '120'
            edit EFI_STEP1 '120'
            edit EFI_STEP2 '240'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 240_360
            edit ACC '120'
            edit EFI_STEP1 '240'
            edit EFI_STEP2 '360'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
        endfamily
        family 240
          trigger ./024 eq complete
          family 000_240
            edit ACC '240'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '240'
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
          family 120_360
            edit ACC '240'
            edit EFI_STEP1 '120'
            edit EFI_STEP2 '360'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
        endfamily
        family 360
          trigger ./024 eq complete
          family 000_360
            edit ACC '360'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '360'
            family tp
              edit RF_NAME 'tp'
              task prep
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep
            endfamily
          endfamily
        endfamily
      endfamily
      family check_nrt
        complete ./check_nrt/go eq complete
        edit REFPATH '/mofc/hc2'
        family catchup
          task check_nrt
            event ok
        endfamily
        family rt
          trigger catchup eq complete
          task check_nrt
            event ok
            cron -w 1 06:30
        endfamily
        task go
          complete ./rt/check_nrt:ok or ./catchup/check_nrt:ok
          trigger 0 eq 1
      endfamily
    endfamily
    family clim
      repeat date YMD 20080717 20121212 7
      edit DELTA_DAY '-7'
      family deps
        trigger ../hind:YMD gt ../clim:YMD
        task dummy
          complete 1 eq 1
      endfamily
      family op
        trigger deps eq complete
        edit EPSLEG '1'
        family 024
          family 000_024
            edit ACC '024'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '024'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 012_036
            edit ACC '024'
            edit EFI_STEP1 '012'
            edit EFI_STEP2 '036'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 024_048
            edit ACC '024'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '048'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 036_060
            edit ACC '024'
            edit EFI_STEP1 '036'
            edit EFI_STEP2 '060'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 048_072
            edit ACC '024'
            edit EFI_STEP1 '048'
            edit EFI_STEP2 '072'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 060_084
            edit ACC '024'
            edit EFI_STEP1 '060'
            edit EFI_STEP2 '084'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 072_096
            edit ACC '024'
            edit EFI_STEP1 '072'
            edit EFI_STEP2 '096'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 084_108
            edit ACC '024'
            edit EFI_STEP1 '084'
            edit EFI_STEP2 '108'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 096_120
            edit ACC '024'
            edit EFI_STEP1 '096'
            edit EFI_STEP2 '120'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 108_132
            edit ACC '024'
            edit EFI_STEP1 '108'
            edit EFI_STEP2 '132'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 120_144
            edit ACC '024'
            edit EFI_STEP1 '120'
            edit EFI_STEP2 '144'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 144_168
            edit ACC '024'
            edit EFI_STEP1 '144'
            edit EFI_STEP2 '168'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 168_192
            edit ACC '024'
            edit EFI_STEP1 '168'
            edit EFI_STEP2 '192'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 192_216
            edit ACC '024'
            edit EFI_STEP1 '192'
            edit EFI_STEP2 '216'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 216_240
            edit ACC '024'
            edit EFI_STEP1 '216'
            edit EFI_STEP2 '240'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
        endfamily
        family 120
          family 000_120
            edit ACC '120'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '120'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 024_144
            edit ACC '120'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '144'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
        endfamily
        family 240
          family 000_240
            edit ACC '240'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '240'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
        endfamily
      endfamily
      family op2fdb
        trigger op eq complete
        edit EPSLEG '1'
        task efi2fdb
          edit KIND 'clim1'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/efi_hdc'
      endfamily
      family nop
        trigger deps eq complete
        edit EPSLEG '2'
        family 024
          family 000_024
            edit ACC '024'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '024'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
          endfamily
          family 012_036
            edit ACC '024'
            edit EFI_STEP1 '012'
            edit EFI_STEP2 '036'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
          endfamily
          family 024_048
            edit ACC '024'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '048'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
          endfamily
          family 036_060
            edit ACC '024'
            edit EFI_STEP1 '036'
            edit EFI_STEP2 '060'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
          endfamily
          family 048_072
            edit ACC '024'
            edit EFI_STEP1 '048'
            edit EFI_STEP2 '072'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
          endfamily
          family 060_084
            edit ACC '024'
            edit EFI_STEP1 '060'
            edit EFI_STEP2 '084'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
          endfamily
          family 072_096
            edit ACC '024'
            edit EFI_STEP1 '072'
            edit EFI_STEP2 '096'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
          endfamily
          family 084_108
            edit ACC '024'
            edit EFI_STEP1 '084'
            edit EFI_STEP2 '108'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
          endfamily
          family 096_120
            edit ACC '024'
            edit EFI_STEP1 '096'
            edit EFI_STEP2 '120'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
          endfamily
          family 108_132
            edit ACC '024'
            edit EFI_STEP1 '108'
            edit EFI_STEP2 '132'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
          endfamily
          family 120_144
            edit ACC '024'
            edit EFI_STEP1 '120'
            edit EFI_STEP2 '144'
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
          endfamily
          family 144_168
            edit ACC '024'
            edit EFI_STEP1 '144'
            edit EFI_STEP2 '168'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
          endfamily
          family 168_192
            edit ACC '024'
            edit EFI_STEP1 '168'
            edit EFI_STEP2 '192'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
          endfamily
          family 192_216
            edit ACC '024'
            edit EFI_STEP1 '192'
            edit EFI_STEP2 '216'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
          endfamily
          family 216_240
            edit ACC '024'
            edit EFI_STEP1 '216'
            edit EFI_STEP2 '240'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
          endfamily
          family 240_264
            edit ACC '024'
            edit EFI_STEP1 '240'
            edit EFI_STEP2 '264'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
          endfamily
          family 264_288
            edit ACC '024'
            edit EFI_STEP1 '264'
            edit EFI_STEP2 '288'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
          endfamily
          family 288_312
            edit ACC '024'
            edit EFI_STEP1 '288'
            edit EFI_STEP2 '312'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
          endfamily
          family 312_336
            edit ACC '024'
            edit EFI_STEP1 '312'
            edit EFI_STEP2 '336'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
          endfamily
          family 336_360
            edit ACC '024'
            edit EFI_STEP1 '336'
            edit EFI_STEP2 '360'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
            family hsttmax
              edit RF_NAME 'hsttmax'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
          endfamily
        endfamily
        family 000
          family 006_006
            edit ACC '000'
            edit EFI_STEP1 '006'
            edit EFI_STEP2 '006'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 012_012
            edit ACC '000'
            edit EFI_STEP1 '012'
            edit EFI_STEP2 '012'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 018_018
            edit ACC '000'
            edit EFI_STEP1 '018'
            edit EFI_STEP2 '018'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 024_024
            edit ACC '000'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '024'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 030_030
            edit ACC '000'
            edit EFI_STEP1 '030'
            edit EFI_STEP2 '030'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 036_036
            edit ACC '000'
            edit EFI_STEP1 '036'
            edit EFI_STEP2 '036'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 042_042
            edit ACC '000'
            edit EFI_STEP1 '042'
            edit EFI_STEP2 '042'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 048_048
            edit ACC '000'
            edit EFI_STEP1 '048'
            edit EFI_STEP2 '048'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 054_054
            edit ACC '000'
            edit EFI_STEP1 '054'
            edit EFI_STEP2 '054'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 060_060
            edit ACC '000'
            edit EFI_STEP1 '060'
            edit EFI_STEP2 '060'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 066_066
            edit ACC '000'
            edit EFI_STEP1 '066'
            edit EFI_STEP2 '066'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 072_072
            edit ACC '000'
            edit EFI_STEP1 '072'
            edit EFI_STEP2 '072'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 078_078
            edit ACC '000'
            edit EFI_STEP1 '078'
            edit EFI_STEP2 '078'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 084_084
            edit ACC '000'
            edit EFI_STEP1 '084'
            edit EFI_STEP2 '084'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 090_090
            edit ACC '000'
            edit EFI_STEP1 '090'
            edit EFI_STEP2 '090'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 096_096
            edit ACC '000'
            edit EFI_STEP1 '096'
            edit EFI_STEP2 '096'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 102_102
            edit ACC '000'
            edit EFI_STEP1 '102'
            edit EFI_STEP2 '102'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 108_108
            edit ACC '000'
            edit EFI_STEP1 '108'
            edit EFI_STEP2 '108'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 114_114
            edit ACC '000'
            edit EFI_STEP1 '114'
            edit EFI_STEP2 '114'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 120_120
            edit ACC '000'
            edit EFI_STEP1 '120'
            edit EFI_STEP2 '120'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 126_126
            edit ACC '000'
            edit EFI_STEP1 '126'
            edit EFI_STEP2 '126'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 132_132
            edit ACC '000'
            edit EFI_STEP1 '132'
            edit EFI_STEP2 '132'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 138_138
            edit ACC '000'
            edit EFI_STEP1 '138'
            edit EFI_STEP2 '138'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 144_144
            edit ACC '000'
            edit EFI_STEP1 '144'
            edit EFI_STEP2 '144'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 150_150
            edit ACC '000'
            edit EFI_STEP1 '150'
            edit EFI_STEP2 '150'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 156_156
            edit ACC '000'
            edit EFI_STEP1 '156'
            edit EFI_STEP2 '156'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 162_162
            edit ACC '000'
            edit EFI_STEP1 '162'
            edit EFI_STEP2 '162'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 168_168
            edit ACC '000'
            edit EFI_STEP1 '168'
            edit EFI_STEP2 '168'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 174_174
            edit ACC '000'
            edit EFI_STEP1 '174'
            edit EFI_STEP2 '174'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 180_180
            edit ACC '000'
            edit EFI_STEP1 '180'
            edit EFI_STEP2 '180'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 186_186
            edit ACC '000'
            edit EFI_STEP1 '186'
            edit EFI_STEP2 '186'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 192_192
            edit ACC '000'
            edit EFI_STEP1 '192'
            edit EFI_STEP2 '192'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 198_198
            edit ACC '000'
            edit EFI_STEP1 '198'
            edit EFI_STEP2 '198'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 204_204
            edit ACC '000'
            edit EFI_STEP1 '204'
            edit EFI_STEP2 '204'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 210_210
            edit ACC '000'
            edit EFI_STEP1 '210'
            edit EFI_STEP2 '210'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 216_216
            edit ACC '000'
            edit EFI_STEP1 '216'
            edit EFI_STEP2 '216'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 222_222
            edit ACC '000'
            edit EFI_STEP1 '222'
            edit EFI_STEP2 '222'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 228_228
            edit ACC '000'
            edit EFI_STEP1 '228'
            edit EFI_STEP2 '228'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 234_234
            edit ACC '000'
            edit EFI_STEP1 '234'
            edit EFI_STEP2 '234'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 240_240
            edit ACC '000'
            edit EFI_STEP1 '240'
            edit EFI_STEP2 '240'
            family 10dd
              edit RF_NAME '10dd'
              task prep_sort
            endfamily
            family tcc
              edit RF_NAME 'tcc'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 252_252
            edit ACC '000'
            edit EFI_STEP1 '252'
            edit EFI_STEP2 '252'
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 264_264
            edit ACC '000'
            edit EFI_STEP1 '264'
            edit EFI_STEP2 '264'
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 276_276
            edit ACC '000'
            edit EFI_STEP1 '276'
            edit EFI_STEP2 '276'
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 288_288
            edit ACC '000'
            edit EFI_STEP1 '288'
            edit EFI_STEP2 '288'
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 300_300
            edit ACC '000'
            edit EFI_STEP1 '300'
            edit EFI_STEP2 '300'
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 312_312
            edit ACC '000'
            edit EFI_STEP1 '312'
            edit EFI_STEP2 '312'
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 324_324
            edit ACC '000'
            edit EFI_STEP1 '324'
            edit EFI_STEP2 '324'
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 336_336
            edit ACC '000'
            edit EFI_STEP1 '336'
            edit EFI_STEP2 '336'
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 348_348
            edit ACC '000'
            edit EFI_STEP1 '348'
            edit EFI_STEP2 '348'
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
          family 360_360
            edit ACC '000'
            edit EFI_STEP1 '360'
            edit EFI_STEP2 '360'
            family 850t
              edit RF_NAME '850t'
              task prep_sort
            endfamily
            family 500z
              edit RF_NAME '500z'
              task prep_sort
            endfamily
          endfamily
        endfamily
        family 006
          family 000_006
            edit ACC '006'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '006'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 006_012
            edit ACC '006'
            edit EFI_STEP1 '006'
            edit EFI_STEP2 '012'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 012_018
            edit ACC '006'
            edit EFI_STEP1 '012'
            edit EFI_STEP2 '018'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 018_024
            edit ACC '006'
            edit EFI_STEP1 '018'
            edit EFI_STEP2 '024'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 024_030
            edit ACC '006'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '030'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 030_036
            edit ACC '006'
            edit EFI_STEP1 '030'
            edit EFI_STEP2 '036'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 036_042
            edit ACC '006'
            edit EFI_STEP1 '036'
            edit EFI_STEP2 '042'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 042_048
            edit ACC '006'
            edit EFI_STEP1 '042'
            edit EFI_STEP2 '048'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 048_054
            edit ACC '006'
            edit EFI_STEP1 '048'
            edit EFI_STEP2 '054'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 054_060
            edit ACC '006'
            edit EFI_STEP1 '054'
            edit EFI_STEP2 '060'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 060_066
            edit ACC '006'
            edit EFI_STEP1 '060'
            edit EFI_STEP2 '066'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 066_072
            edit ACC '006'
            edit EFI_STEP1 '066'
            edit EFI_STEP2 '072'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 072_078
            edit ACC '006'
            edit EFI_STEP1 '072'
            edit EFI_STEP2 '078'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 078_084
            edit ACC '006'
            edit EFI_STEP1 '078'
            edit EFI_STEP2 '084'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 084_090
            edit ACC '006'
            edit EFI_STEP1 '084'
            edit EFI_STEP2 '090'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 090_096
            edit ACC '006'
            edit EFI_STEP1 '090'
            edit EFI_STEP2 '096'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 096_102
            edit ACC '006'
            edit EFI_STEP1 '096'
            edit EFI_STEP2 '102'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 102_108
            edit ACC '006'
            edit EFI_STEP1 '102'
            edit EFI_STEP2 '108'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 108_114
            edit ACC '006'
            edit EFI_STEP1 '108'
            edit EFI_STEP2 '114'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 114_120
            edit ACC '006'
            edit EFI_STEP1 '114'
            edit EFI_STEP2 '120'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 120_126
            edit ACC '006'
            edit EFI_STEP1 '120'
            edit EFI_STEP2 '126'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 126_132
            edit ACC '006'
            edit EFI_STEP1 '126'
            edit EFI_STEP2 '132'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 132_138
            edit ACC '006'
            edit EFI_STEP1 '132'
            edit EFI_STEP2 '138'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 138_144
            edit ACC '006'
            edit EFI_STEP1 '138'
            edit EFI_STEP2 '144'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 144_150
            edit ACC '006'
            edit EFI_STEP1 '144'
            edit EFI_STEP2 '150'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 150_156
            edit ACC '006'
            edit EFI_STEP1 '150'
            edit EFI_STEP2 '156'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 156_162
            edit ACC '006'
            edit EFI_STEP1 '156'
            edit EFI_STEP2 '162'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 162_168
            edit ACC '006'
            edit EFI_STEP1 '162'
            edit EFI_STEP2 '168'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 168_174
            edit ACC '006'
            edit EFI_STEP1 '168'
            edit EFI_STEP2 '174'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 174_180
            edit ACC '006'
            edit EFI_STEP1 '174'
            edit EFI_STEP2 '180'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 180_186
            edit ACC '006'
            edit EFI_STEP1 '180'
            edit EFI_STEP2 '186'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 186_192
            edit ACC '006'
            edit EFI_STEP1 '186'
            edit EFI_STEP2 '192'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 192_198
            edit ACC '006'
            edit EFI_STEP1 '192'
            edit EFI_STEP2 '198'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 198_204
            edit ACC '006'
            edit EFI_STEP1 '198'
            edit EFI_STEP2 '204'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 204_210
            edit ACC '006'
            edit EFI_STEP1 '204'
            edit EFI_STEP2 '210'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 210_216
            edit ACC '006'
            edit EFI_STEP1 '210'
            edit EFI_STEP2 '216'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 216_222
            edit ACC '006'
            edit EFI_STEP1 '216'
            edit EFI_STEP2 '222'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 222_228
            edit ACC '006'
            edit EFI_STEP1 '222'
            edit EFI_STEP2 '228'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 228_234
            edit ACC '006'
            edit EFI_STEP1 '228'
            edit EFI_STEP2 '234'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
          family 234_240
            edit ACC '006'
            edit EFI_STEP1 '234'
            edit EFI_STEP2 '240'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family ts
              edit RF_NAME 'ts'
              task prep_sort
            endfamily
            family 10fg
              edit RF_NAME '10fg'
              task prep_sort
            endfamily
            family 2tmin
              edit RF_NAME '2tmin'
              task prep_sort
            endfamily
            family 2tmax
              edit RF_NAME '2tmax'
              task prep_sort
            endfamily
          endfamily
        endfamily
        family 012
          family 000_012
            edit ACC '012'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '012'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 012_024
            edit ACC '012'
            edit EFI_STEP1 '012'
            edit EFI_STEP2 '024'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 024_036
            edit ACC '012'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '036'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 036_048
            edit ACC '012'
            edit EFI_STEP1 '036'
            edit EFI_STEP2 '048'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 048_060
            edit ACC '012'
            edit EFI_STEP1 '048'
            edit EFI_STEP2 '060'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 060_072
            edit ACC '012'
            edit EFI_STEP1 '060'
            edit EFI_STEP2 '072'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 072_084
            edit ACC '012'
            edit EFI_STEP1 '072'
            edit EFI_STEP2 '084'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 084_096
            edit ACC '012'
            edit EFI_STEP1 '084'
            edit EFI_STEP2 '096'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 096_108
            edit ACC '012'
            edit EFI_STEP1 '096'
            edit EFI_STEP2 '108'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 108_120
            edit ACC '012'
            edit EFI_STEP1 '108'
            edit EFI_STEP2 '120'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 120_132
            edit ACC '012'
            edit EFI_STEP1 '120'
            edit EFI_STEP2 '132'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 132_144
            edit ACC '012'
            edit EFI_STEP1 '132'
            edit EFI_STEP2 '144'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 144_156
            edit ACC '012'
            edit EFI_STEP1 '144'
            edit EFI_STEP2 '156'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 156_168
            edit ACC '012'
            edit EFI_STEP1 '156'
            edit EFI_STEP2 '168'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 168_180
            edit ACC '012'
            edit EFI_STEP1 '168'
            edit EFI_STEP2 '180'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 180_192
            edit ACC '012'
            edit EFI_STEP1 '180'
            edit EFI_STEP2 '192'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 192_204
            edit ACC '012'
            edit EFI_STEP1 '192'
            edit EFI_STEP2 '204'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 204_216
            edit ACC '012'
            edit EFI_STEP1 '204'
            edit EFI_STEP2 '216'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 216_228
            edit ACC '012'
            edit EFI_STEP1 '216'
            edit EFI_STEP2 '228'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 228_240
            edit ACC '012'
            edit EFI_STEP1 '228'
            edit EFI_STEP2 '240'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 240_252
            edit ACC '012'
            edit EFI_STEP1 '240'
            edit EFI_STEP2 '252'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 252_264
            edit ACC '012'
            edit EFI_STEP1 '252'
            edit EFI_STEP2 '264'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 264_276
            edit ACC '012'
            edit EFI_STEP1 '264'
            edit EFI_STEP2 '276'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 276_288
            edit ACC '012'
            edit EFI_STEP1 '276'
            edit EFI_STEP2 '288'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 288_300
            edit ACC '012'
            edit EFI_STEP1 '288'
            edit EFI_STEP2 '300'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 300_312
            edit ACC '012'
            edit EFI_STEP1 '300'
            edit EFI_STEP2 '312'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 312_324
            edit ACC '012'
            edit EFI_STEP1 '312'
            edit EFI_STEP2 '324'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 324_336
            edit ACC '012'
            edit EFI_STEP1 '324'
            edit EFI_STEP2 '336'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 336_348
            edit ACC '012'
            edit EFI_STEP1 '336'
            edit EFI_STEP2 '348'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
          family 348_360
            edit ACC '012'
            edit EFI_STEP1 '348'
            edit EFI_STEP2 '360'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
          endfamily
        endfamily
        family 072
          family 000_072
            edit ACC '072'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '072'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 024_096
            edit ACC '072'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '096'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 048_120
            edit ACC '072'
            edit EFI_STEP1 '048'
            edit EFI_STEP2 '120'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 072_144
            edit ACC '072'
            edit EFI_STEP1 '072'
            edit EFI_STEP2 '144'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 096_168
            edit ACC '072'
            edit EFI_STEP1 '096'
            edit EFI_STEP2 '168'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 120_192
            edit ACC '072'
            edit EFI_STEP1 '120'
            edit EFI_STEP2 '192'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 144_216
            edit ACC '072'
            edit EFI_STEP1 '144'
            edit EFI_STEP2 '216'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 168_240
            edit ACC '072'
            edit EFI_STEP1 '168'
            edit EFI_STEP2 '240'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
        endfamily
        family 120
          family 000_120
            edit ACC '120'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '120'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 024_144
            edit ACC '120'
            edit EFI_STEP1 '024'
            edit EFI_STEP2 '144'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 048_168
            edit ACC '120'
            edit EFI_STEP1 '048'
            edit EFI_STEP2 '168'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 072_192
            edit ACC '120'
            edit EFI_STEP1 '072'
            edit EFI_STEP2 '192'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 096_216
            edit ACC '120'
            edit EFI_STEP1 '096'
            edit EFI_STEP2 '216'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 120_240
            edit ACC '120'
            edit EFI_STEP1 '120'
            edit EFI_STEP2 '240'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 240_360
            edit ACC '120'
            edit EFI_STEP1 '240'
            edit EFI_STEP2 '360'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
        endfamily
        family 240
          family 000_240
            edit ACC '240'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '240'
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
          family 120_360
            edit ACC '240'
            edit EFI_STEP1 '120'
            edit EFI_STEP2 '360'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
        endfamily
        family 360
          family 000_360
            edit ACC '360'
            edit EFI_STEP1 '000'
            edit EFI_STEP2 '360'
            family tp
              edit RF_NAME 'tp'
              task prep_sort
            endfamily
            family 2t
              edit RF_NAME '2t'
              task prep_sort
            endfamily
            family 10ff
              edit RF_NAME '10ff'
              task prep_sort
            endfamily
          endfamily
        endfamily
      endfamily
      family nop2fdb
        trigger nop eq complete and ./op2fdb eq complete
        edit EPSLEG '2'
        task efi2fdb
          edit KIND 'clim2'
          edit SMSFILES '/home/ma/emos/def/emc_36r1/mc/efi_hdc'
      endfamily
      family arch_nop
        trigger nop2fdb eq complete
        edit EPSLEG '2'
        edit SMSFILES '/home/ma/emos/def/emc_36r1/archive'
        family 2t
          edit step1 '24'
          edit step2 '24'
          edit RF_NAME '2t'
          task eficlim
            label info ""
        endfamily
        family 10ff
          edit step1 '24'
          edit step2 '24'
          edit RF_NAME '10ff'
          task eficlim
            label info ""
        endfamily
        family 10fg
          edit step1 '24'
          edit step2 '24'
          edit RF_NAME '10fg'
          task eficlim
            label info ""
        endfamily
        family tp
          edit step1 '24'
          edit step2 '24'
          edit RF_NAME 'tp'
          task eficlim
            label info ""
        endfamily
        family 2tmin
          edit step1 '24'
          edit step2 '24'
          edit RF_NAME '2tmin'
          task eficlim
            label info ""
        endfamily
        family 2tmax
          edit step1 '24'
          edit step2 '24'
          edit RF_NAME '2tmax'
          task eficlim
            label info ""
        endfamily
        family ts
          edit step1 '24'
          edit step2 '24'
          edit RF_NAME 'ts'
          task eficlim
            label info ""
        endfamily
        family hsttmax
          edit step1 '24'
          edit step2 '24'
          edit RF_NAME 'hsttmax'
          task eficlim
            label info ""
        endfamily
        family 10dd
          edit step1 '24'
          edit step2 '24'
          edit RF_NAME '10dd'
          task eficlim
            label info ""
        endfamily
        family 850t
          edit step1 '24'
          edit step2 '24'
          edit RF_NAME '850t'
          task eficlim
            label info ""
        endfamily
        family 500z
          edit step1 '24'
          edit step2 '24'
          edit RF_NAME '500z'
          task eficlim
            label info ""
        endfamily
        family 2tmean
          edit step2 '24'
          edit RF_NAME '2t'
          edit step1 '0'
          task eficlim
            label info ""
        endfamily
        family 10ffmean
          edit step2 '24'
          edit RF_NAME '10ff'
          edit step1 '0'
          task eficlim
            label info ""
        endfamily
        family tcc
          edit step1 '24'
          edit step2 '24'
          edit RF_NAME 'tcc'
          task eficlim
            label info ""
        endfamily
        family tccmean
          edit step2 '24'
          edit RF_NAME 'tcc'
          edit step1 '0'
          task eficlim
            label info ""
        endfamily
      endfamily
      family gdb
        trigger ( op eq complete and nop eq complete) and ( arch_nop eq complete)
        edit RHOST 'epsgrams'
        edit RDIR '/webapps/epsgrams/data/'
        edit KIND 'clim_15_days'
        edit WSHOST 'linux_cluster'
        edit QUEUE 'emos-e'
        edit QUEUE_EPILOG 'emos-exp'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        inlimit /limits:linux_cluster
        task sync_climate
        task gclean
          trigger sync_climate eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1/pop/data/epsgrams/grams'
          edit KEEP_DAYS '14'
          edit REMOTE_KEEP_DAYS '14'
        task gdb
          trigger gclean eq complete and ( /emc_36r1/main:YMD ge /emc_36r1/efi_next/mn:YMD and /emc_36r1/main/00 eq complete)
          edit SMSFILES '/home/ma/emos/def/emc_36r1/pop/data/epsgrams/grams'
          label files "="
          label info "="
          label local "="
          label remote "="
          event fdb
          event 1 
        task glink
          trigger gdb eq complete
          edit SMSFILES '/home/ma/emos/def/emc_36r1/pop/data/epsgrams/grams'
      endfamily
      family check_nrt
        complete ./check_nrt/go eq complete
        edit REFPATH '/mofc/hc2'
        family catchup
          task check_nrt
            event ok
        endfamily
        family rt
          trigger catchup eq complete
          task check_nrt
            event ok
            cron -w 3 06:30
        endfamily
        task go
          complete ./rt/check_nrt:ok or ./catchup/check_nrt:ok
          trigger 0 eq 1
      endfamily
      task efi_cleaner
        trigger ( arch_nop eq complete or arch_nop eq unknown) and ( op eq complete or op eq unknown) and ( nop eq complete or nop eq unknown) and ( op2fdb eq complete or op2fdb eq unknown) and ( nop2fdb eq complete or nop2fdb eq unknown)
        edit DELTA_DAY '-43'
    endfamily
  endfamily
endsuite
suite elaw_36r1
  edit SUITE_TYPE 'law'
  edit USER 'emos'
  edit SCHOST 'c1b'
  edit STHOST '/s1b'
  edit STHOST_BKUP '/s1a'
  edit WSHOST 'linux_cluster'
  edit WSLOGDIR '/vol/emos_nc/output'
  edit SCLOGDIR '/s1b/emos_esuite/emos_data/0046/log'
  edit VERSION '0046'
  edit INIEXPVER '0001'
  edit WVINIEXPVER 'false'
  edit INIWDEXPVER '0046'
  edit INIWDCLASS 'od'
  edit FIRST_DAY '1'
  edit EPSTYPE 'fc'
  edit MEMBER '0'
  edit REINITIALIZE 'false'
  edit DATEMASK '*.*.*'
  edit YMD '20091016'
  edit USE_YMD 'true'
  edit FSFAMILY 'euroshelf'
  edit VERIF_WAIT '10'
  edit MARS_DB 'marsod'
  edit VERIFY 'verify'
  edit SUITE_START '2009101612'
  edit SMSTRIES '2'
  edit SMSHOME '/vol/emos_nc/output'
  edit SMSFILES '/home/ma/emos/def/elaw_36r1'
  edit SMSINCLUDE '/home/ma/emos/def/elaw_36r1/include'
  edit RESERVATION_ON '1'
  edit QUEUE 'os'
  edit QUEUE_EPILOG 'ns'
  edit SMSCMD 'smssubmit.36r1 %USER% %SCHOST% %SMSJOB% %SMSJOBOUT%'
  edit SMSKILL 'smskill %USER% %SCHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
  edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %SCHOST% %SMSRID% %SMSJOB%'
  edit LOGDIR '/s1b/emos_esuite/emos_data/0046/log'
  edit SMSOUT '/s1b/emos_esuite/emos_data/0046/log'
  edit ANTIME '12'
  edit DELTA_DAY '0'
  edit EMOS_BASE '12'
  edit EMOS_PROJECT '00'
  edit EMOS_TIME_STEP_H '0000'
  edit PGEN_BINDIR '/home/ma/emos/bin/diss'
  edit PGEN_OBJDIR '/home/ma/emos/bin/diss'
  edit DISS_REQDB_ROOT '/diss/requirements'
  edit DISS_BINDIR '/diss/bin/DISS'
  edit ANNPES '1'
  edit FCNPES '128'
  edit EMOS_DOMAIN 'M'
  edit PROD_TASK '1'
  edit EMOS_TYPE 'fc'
  edit DATA_COV 'no'
  edit EMOS_STREAM 'WV'
  edit STREAM 'DA'
  edit INISTREAM 'wave'
  edit ACCOUNT 'oewave'
  edit WAMLEN '10'
  edit WAM1ST 'no'
  edit GRID '0.25/0.25'
  edit RET_GRID '1.5/1.5'
  edit SMSLOGHOST 'c1b'
  edit SMSLOGPORT '9316'
  edit SLEEPTIME '1'
  inlimit /limits:c1b
  clock hybrid 
  family limits
    limit mars 2
    limit prod 5
    limit pdb 5
    limit q2diss 5
    limit prodcheck 1
  endfamily
  family main
    trigger /elaw_36r1/main:YMD le /e_36r1/main:YMD
    repeat date YMD 20091016 20121212 1
    edit LOGTASK 'main'
    family make
      task setup
      task wconst
        trigger setup eq complete
      task wavini
        trigger wconst eq complete
      family prod
        edit SMSFILES '/home/ma/emos/def/elaw_36r1/prod/make'
        edit EMOS_DOMAIN 'g'
        edit EMOS_TYPE 'fc'
        edit EMOS_STREAM 'OPER'
        task wparam
          edit EMOS_TYPE 'wp'
        task tf
          edit EMOS_TYPE 'tf'
        task sortreq
        task bitmap
        task clima
        task disssch
        task tools
        task oper
        task wave
          edit EMOS_STREAM 'WAVE'
      endfamily
    endfamily
    family 12
      trigger make eq complete
      edit EMOS_BASE '12'
      edit EMOS_PROJECT '12'
      edit DELTA_DAY '0'
      family ref
        edit QUEUE 'emos-e'
        edit QUEUE_EPILOG 'emos-exp'
        edit WSHOST 'localhost'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        task model
          trigger ( /e_36r1/main:YMD gt /elaw_36r1/main:YMD) or ( /e_36r1/main:YMD eq /elaw_36r1/main:YMD and /e_36r1/main/12/fc/model eq complete)
        task prod
          trigger ( /e_36r1/main:YMD gt /elaw_36r1/main:YMD) or ( /e_36r1/main:YMD eq /elaw_36r1/main:YMD and /e_36r1/main/12/getreq eq complete)
        task sst
          trigger ( /e_36r1/main:YMD gt /elaw_36r1/main:YMD) or ( /e_36r1/main:YMD eq /elaw_36r1/main:YMD and /e_36r1/main/12/an/surf_anal/sst eq complete)
      endfamily
      family euroshelf
        trigger ref/model eq complete
        edit EMOS_DOMAIN 'M'
        edit GRID '0.25/0.25'
        edit RET_GRID '0.25/0.25'
        edit WAMLEN '5'
        task wamobs
        task wamwind
        task wamwave
          trigger /elaw_36r1/main/12/ref/sst eq complete
        task wamrun
          trigger wamobs eq complete and wamwind eq complete and wamwave eq complete
          edit USER_PRIORITY '80'
          edit SMSTRIES '1'
          edit THREADS '2'
      endfamily
      family getreq
        trigger ref/prod eq complete
        edit EMOS_TYPE 'fc'
        edit EMOS_DOMAIN 'M'
        edit FSFAMILY 'ppdi'
        edit SMSFILES '/home/ma/emos/def/elaw_36r1/prod'
        edit SMSTRIES '1'
        edit EMOS_STREAM 'WAVE'
        task collectreq
          edit QUEUE 'diss'
          edit USER 'emos'
          edit WSHOST 'diss'
          edit QUEUE_EPILOG 'emos-exp'
          edit LOGDIR '/diss/log'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit SMSOUT '/vol/emos_nc/output'
        task req
          late -s +00:03 -c +00:10
          trigger collectreq eq complete
          label info ""
          event 1 
        task collectbitmap
          trigger req:001 or req eq complete
          edit QUEUE 'diss'
          edit USER 'emos'
          edit WSHOST 'diss'
          edit QUEUE_EPILOG 'emos-exp'
          edit LOGDIR '/diss/log'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit SMSOUT '/vol/emos_nc/output'
        task bitmap
          late -s +00:03 -c +00:10
          trigger collectbitmap eq complete
          label info ""
          event 1 
      endfamily
      family prod
        edit ACCOUNT 'oewavepg'
        edit SMSFILES '/home/ma/emos/def/elaw_36r1/prod'
        edit EMOS_TYPE 'fc'
        edit FSFAMILY 'ppdi'
        edit PROD_TASK '1'
        edit PGNPES '1'
        edit EMOS_TIME_STEP_H '0000'
        edit USER_PRIORITY '90'
        edit SMSURLBASE 'http://diss.ecmwf.int/do/monitoring/summary'
        edit SMSURL 'MWAVE/12'
        edit DELTA_DAY '-1'
        inlimit /limits:prod
        inlimit /elaw_36r1/limits:prod
        family 06
          edit SMSURL 'MWAVE/06'
          family an
            trigger /elaw_36r1/main/12/euroshelf/wamrun eq complete and ( /elaw_36r1/main/12/getreq/bitmap:001 or /elaw_36r1/main/12/getreq/bitmap eq complete) and ( /elaw_36r1/main/12/getreq/req:001 or /elaw_36r1/main/12/getreq/req eq complete)
            edit EMOS_TYPE 'an'
            edit EMOS_BASE '06'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
        endfamily
        family 12
          edit EMOS_TYPE 'fc'
          edit PROD_TASK '1'
          edit DELTA_DAY '0'
          family an
            trigger /elaw_36r1/main/12/euroshelf/wamrun eq complete and /elaw_36r1/main/12/prod/06/an/prodgen eq complete
            edit EMOS_TYPE 'an'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 003
            trigger ../12/an/prodgen eq complete
            edit EMOS_TIME_STEP_H '0003'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 006
            trigger ./003/prodgen eq complete
            edit EMOS_TIME_STEP_H '0006'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 009
            trigger ./006/prodgen eq complete
            edit EMOS_TIME_STEP_H '0009'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 012
            trigger ./009/prodgen eq complete
            edit EMOS_TIME_STEP_H '0012'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 015
            trigger ./012/prodgen eq complete
            edit EMOS_TIME_STEP_H '0015'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 018
            trigger ./015/prodgen eq complete
            edit EMOS_TIME_STEP_H '0018'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 021
            trigger ./018/prodgen eq complete
            edit EMOS_TIME_STEP_H '0021'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 024
            trigger ./021/prodgen eq complete
            edit EMOS_TIME_STEP_H '0024'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 027
            trigger ./024/prodgen eq complete
            edit EMOS_TIME_STEP_H '0027'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 030
            trigger ./027/prodgen eq complete
            edit EMOS_TIME_STEP_H '0030'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 033
            trigger ./030/prodgen eq complete
            edit EMOS_TIME_STEP_H '0033'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 036
            trigger ./033/prodgen eq complete
            edit EMOS_TIME_STEP_H '0036'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 039
            trigger ./036/prodgen eq complete
            edit EMOS_TIME_STEP_H '0039'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 042
            trigger ./039/prodgen eq complete
            edit EMOS_TIME_STEP_H '0042'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 045
            trigger ./042/prodgen eq complete
            edit EMOS_TIME_STEP_H '0045'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 048
            trigger ./045/prodgen eq complete
            edit EMOS_TIME_STEP_H '0048'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 051
            trigger ./048/prodgen eq complete
            edit EMOS_TIME_STEP_H '0051'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 054
            trigger ./051/prodgen eq complete
            edit EMOS_TIME_STEP_H '0054'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 057
            trigger ./054/prodgen eq complete
            edit EMOS_TIME_STEP_H '0057'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 060
            trigger ./057/prodgen eq complete
            edit EMOS_TIME_STEP_H '0060'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 063
            trigger ./060/prodgen eq complete
            edit EMOS_TIME_STEP_H '0063'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 066
            trigger ./063/prodgen eq complete
            edit EMOS_TIME_STEP_H '0066'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 069
            trigger ./066/prodgen eq complete
            edit EMOS_TIME_STEP_H '0069'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 072
            trigger ./069/prodgen eq complete
            edit EMOS_TIME_STEP_H '0072'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 075
            trigger ./072/prodgen eq complete
            edit EMOS_TIME_STEP_H '0075'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 078
            trigger ./075/prodgen eq complete
            edit EMOS_TIME_STEP_H '0078'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 081
            trigger ./078/prodgen eq complete
            edit EMOS_TIME_STEP_H '0081'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 084
            trigger ./081/prodgen eq complete
            edit EMOS_TIME_STEP_H '0084'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 087
            trigger ./084/prodgen eq complete
            edit EMOS_TIME_STEP_H '0087'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 090
            trigger ./087/prodgen eq complete
            edit EMOS_TIME_STEP_H '0090'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 093
            trigger ./090/prodgen eq complete
            edit EMOS_TIME_STEP_H '0093'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 096
            trigger ./093/prodgen eq complete
            edit EMOS_TIME_STEP_H '0096'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 099
            trigger ./096/prodgen eq complete
            edit EMOS_TIME_STEP_H '0099'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 102
            trigger ./099/prodgen eq complete
            edit EMOS_TIME_STEP_H '0102'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 105
            trigger ./102/prodgen eq complete
            edit EMOS_TIME_STEP_H '0105'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 108
            trigger ./105/prodgen eq complete
            edit EMOS_TIME_STEP_H '0108'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 111
            trigger ./108/prodgen eq complete
            edit EMOS_TIME_STEP_H '0111'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 114
            trigger ./111/prodgen eq complete
            edit EMOS_TIME_STEP_H '0114'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 117
            trigger ./114/prodgen eq complete
            edit EMOS_TIME_STEP_H '0117'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 120
            trigger ./117/prodgen eq complete
            edit EMOS_TIME_STEP_H '0120'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
        endfamily
      endfamily
      family prod2diss
        edit USER_PRIORITY '90'
        edit SMSURL 'MWAVE/12'
        edit SMSFILES '/home/ma/emos/def/elaw_36r1/prod'
        edit EMOS_TYPE 'fc'
        edit FSFAMILY 'ppdi'
        edit EMOS_TIME_STEP_H '0000'
        inlimit /elaw_36r1/limits:pdb
        inlimit /limits:pdb
        family 06
          edit EMOS_TYPE 'an'
          edit SMSURL 'MWAVE/06'
          family an
            trigger /elaw_36r1/main/12/prod/06/an/prodgen eq complete
            edit EMOS_TYPE 'an'
            edit EMOS_BASE '06'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
        endfamily
        family 12
          edit EMOS_BASE '12'
          edit EMOS_TYPE 'fc'
          family an
            trigger /elaw_36r1/main/12/prod/12/an/prodgen eq complete
            edit EMOS_TYPE 'an'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 003
            trigger /elaw_36r1/main/12/prod/12/003/prodgen eq complete and ( ./an/diss:pdb or ./an/diss eq complete)
            edit EMOS_TIME_STEP_H '0003'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 006
            trigger /elaw_36r1/main/12/prod/12/006/prodgen eq complete and ( ./003/diss:pdb or ./003/diss eq complete)
            edit EMOS_TIME_STEP_H '0006'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 009
            trigger /elaw_36r1/main/12/prod/12/009/prodgen eq complete and ( ./006/diss:pdb or ./006/diss eq complete)
            edit EMOS_TIME_STEP_H '0009'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 012
            trigger /elaw_36r1/main/12/prod/12/012/prodgen eq complete and ( ./009/diss:pdb or ./009/diss eq complete)
            edit EMOS_TIME_STEP_H '0012'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 015
            trigger /elaw_36r1/main/12/prod/12/015/prodgen eq complete and ( ./012/diss:pdb or ./012/diss eq complete)
            edit EMOS_TIME_STEP_H '0015'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 018
            trigger /elaw_36r1/main/12/prod/12/018/prodgen eq complete and ( ./015/diss:pdb or ./015/diss eq complete)
            edit EMOS_TIME_STEP_H '0018'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 021
            trigger /elaw_36r1/main/12/prod/12/021/prodgen eq complete and ( ./018/diss:pdb or ./018/diss eq complete)
            edit EMOS_TIME_STEP_H '0021'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 024
            trigger /elaw_36r1/main/12/prod/12/024/prodgen eq complete and ( ./021/diss:pdb or ./021/diss eq complete)
            edit EMOS_TIME_STEP_H '0024'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 027
            trigger /elaw_36r1/main/12/prod/12/027/prodgen eq complete and ( ./024/diss:pdb or ./024/diss eq complete)
            edit EMOS_TIME_STEP_H '0027'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 030
            trigger /elaw_36r1/main/12/prod/12/030/prodgen eq complete and ( ./027/diss:pdb or ./027/diss eq complete)
            edit EMOS_TIME_STEP_H '0030'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 033
            trigger /elaw_36r1/main/12/prod/12/033/prodgen eq complete and ( ./030/diss:pdb or ./030/diss eq complete)
            edit EMOS_TIME_STEP_H '0033'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 036
            trigger /elaw_36r1/main/12/prod/12/036/prodgen eq complete and ( ./033/diss:pdb or ./033/diss eq complete)
            edit EMOS_TIME_STEP_H '0036'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 039
            trigger /elaw_36r1/main/12/prod/12/039/prodgen eq complete and ( ./036/diss:pdb or ./036/diss eq complete)
            edit EMOS_TIME_STEP_H '0039'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 042
            trigger /elaw_36r1/main/12/prod/12/042/prodgen eq complete and ( ./039/diss:pdb or ./039/diss eq complete)
            edit EMOS_TIME_STEP_H '0042'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 045
            trigger /elaw_36r1/main/12/prod/12/045/prodgen eq complete and ( ./042/diss:pdb or ./042/diss eq complete)
            edit EMOS_TIME_STEP_H '0045'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 048
            trigger /elaw_36r1/main/12/prod/12/048/prodgen eq complete and ( ./045/diss:pdb or ./045/diss eq complete)
            edit EMOS_TIME_STEP_H '0048'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 051
            trigger /elaw_36r1/main/12/prod/12/051/prodgen eq complete and ( ./048/diss:pdb or ./048/diss eq complete)
            edit EMOS_TIME_STEP_H '0051'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 054
            trigger /elaw_36r1/main/12/prod/12/054/prodgen eq complete and ( ./051/diss:pdb or ./051/diss eq complete)
            edit EMOS_TIME_STEP_H '0054'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 057
            trigger /elaw_36r1/main/12/prod/12/057/prodgen eq complete and ( ./054/diss:pdb or ./054/diss eq complete)
            edit EMOS_TIME_STEP_H '0057'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 060
            trigger /elaw_36r1/main/12/prod/12/060/prodgen eq complete and ( ./057/diss:pdb or ./057/diss eq complete)
            edit EMOS_TIME_STEP_H '0060'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 063
            trigger /elaw_36r1/main/12/prod/12/063/prodgen eq complete and ( ./060/diss:pdb or ./060/diss eq complete)
            edit EMOS_TIME_STEP_H '0063'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 066
            trigger /elaw_36r1/main/12/prod/12/066/prodgen eq complete and ( ./063/diss:pdb or ./063/diss eq complete)
            edit EMOS_TIME_STEP_H '0066'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 069
            trigger /elaw_36r1/main/12/prod/12/069/prodgen eq complete and ( ./066/diss:pdb or ./066/diss eq complete)
            edit EMOS_TIME_STEP_H '0069'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 072
            trigger /elaw_36r1/main/12/prod/12/072/prodgen eq complete and ( ./069/diss:pdb or ./069/diss eq complete)
            edit EMOS_TIME_STEP_H '0072'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 075
            trigger /elaw_36r1/main/12/prod/12/075/prodgen eq complete and ( ./072/diss:pdb or ./072/diss eq complete)
            edit EMOS_TIME_STEP_H '0075'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 078
            trigger /elaw_36r1/main/12/prod/12/078/prodgen eq complete and ( ./075/diss:pdb or ./075/diss eq complete)
            edit EMOS_TIME_STEP_H '0078'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 081
            trigger /elaw_36r1/main/12/prod/12/081/prodgen eq complete and ( ./078/diss:pdb or ./078/diss eq complete)
            edit EMOS_TIME_STEP_H '0081'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 084
            trigger /elaw_36r1/main/12/prod/12/084/prodgen eq complete and ( ./081/diss:pdb or ./081/diss eq complete)
            edit EMOS_TIME_STEP_H '0084'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 087
            trigger /elaw_36r1/main/12/prod/12/087/prodgen eq complete and ( ./084/diss:pdb or ./084/diss eq complete)
            edit EMOS_TIME_STEP_H '0087'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 090
            trigger /elaw_36r1/main/12/prod/12/090/prodgen eq complete and ( ./087/diss:pdb or ./087/diss eq complete)
            edit EMOS_TIME_STEP_H '0090'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 093
            trigger /elaw_36r1/main/12/prod/12/093/prodgen eq complete and ( ./090/diss:pdb or ./090/diss eq complete)
            edit EMOS_TIME_STEP_H '0093'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 096
            trigger /elaw_36r1/main/12/prod/12/096/prodgen eq complete and ( ./093/diss:pdb or ./093/diss eq complete)
            edit EMOS_TIME_STEP_H '0096'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 099
            trigger /elaw_36r1/main/12/prod/12/099/prodgen eq complete and ( ./096/diss:pdb or ./096/diss eq complete)
            edit EMOS_TIME_STEP_H '0099'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 102
            trigger /elaw_36r1/main/12/prod/12/102/prodgen eq complete and ( ./099/diss:pdb or ./099/diss eq complete)
            edit EMOS_TIME_STEP_H '0102'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 105
            trigger /elaw_36r1/main/12/prod/12/105/prodgen eq complete and ( ./102/diss:pdb or ./102/diss eq complete)
            edit EMOS_TIME_STEP_H '0105'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 108
            trigger /elaw_36r1/main/12/prod/12/108/prodgen eq complete and ( ./105/diss:pdb or ./105/diss eq complete)
            edit EMOS_TIME_STEP_H '0108'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 111
            trigger /elaw_36r1/main/12/prod/12/111/prodgen eq complete and ( ./108/diss:pdb or ./108/diss eq complete)
            edit EMOS_TIME_STEP_H '0111'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 114
            trigger /elaw_36r1/main/12/prod/12/114/prodgen eq complete and ( ./111/diss:pdb or ./111/diss eq complete)
            edit EMOS_TIME_STEP_H '0114'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 117
            trigger /elaw_36r1/main/12/prod/12/117/prodgen eq complete and ( ./114/diss:pdb or ./114/diss eq complete)
            edit EMOS_TIME_STEP_H '0117'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 120
            trigger /elaw_36r1/main/12/prod/12/120/prodgen eq complete and ( ./117/diss:pdb or ./117/diss eq complete)
            edit EMOS_TIME_STEP_H '0120'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
        endfamily
        task check
          trigger ./12 eq complete
        family reservation
          trigger ./12 eq complete
          edit SMSFILES '/home/ma/emos/def/elaw_36r1/smsfiles'
          task loadl_cycle_limit
            edit LOADL_CYCLE '65'
            label info ""
        endfamily
      endfamily
    endfamily
    family 00
      trigger ./make eq complete and ./12 eq complete
      edit EMOS_BASE '00'
      edit EMOS_PROJECT '00'
      edit DELTA_DAY '1'
      edit RESERVATION_ON '1'
      edit QUEUE 'os'
      family ref
        edit QUEUE 'emos-e'
        edit QUEUE_EPILOG 'emos-exp'
        edit WSHOST 'localhost'
        edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
        edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
        edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
        edit LOGDIR '/vol/emos_nc/output'
        edit SMSOUT '/vol/emos_nc/output'
        task model
          trigger ( /e_36r1/main:YMD gt /elaw_36r1/main:YMD) or ( /e_36r1/main:YMD eq /elaw_36r1/main:YMD and /e_36r1/main/00/fc/model eq complete)
        task prod
          trigger ( /e_36r1/main:YMD gt /elaw_36r1/main:YMD) or ( /e_36r1/main:YMD eq /elaw_36r1/main:YMD and /e_36r1/main/00/getreq eq complete)
        task sst
          trigger ( /e_36r1/main:YMD gt /elaw_36r1/main:YMD) or ( /e_36r1/main:YMD eq /elaw_36r1/main:YMD and /e_36r1/main/00/an/surf_anal/sst eq complete)
      endfamily
      family euroshelf
        trigger ref/model eq complete
        edit EMOS_DOMAIN 'M'
        edit GRID '0.25/0.25'
        edit RET_GRID '0.25/0.25'
        edit WAMLEN '5'
        task wamobs
        task wamwind
        task wamwave
          trigger ( /elaw_36r1/main/00/ref/sst eq complete) and ( /elaw_36r1/main/12/euroshelf eq complete)
        task wamrun
          trigger wamobs eq complete and wamwind eq complete and wamwave eq complete
          edit USER_PRIORITY '80'
          edit SMSTRIES '1'
          edit THREADS '2'
      endfamily
      family getreq
        trigger ref/prod eq complete
        edit EMOS_TYPE 'fc'
        edit EMOS_DOMAIN 'M'
        edit FSFAMILY 'ppdi'
        edit SMSFILES '/home/ma/emos/def/elaw_36r1/prod'
        edit SMSTRIES '1'
        edit EMOS_STREAM 'WAVE'
        task collectreq
          edit QUEUE 'diss'
          edit USER 'emos'
          edit WSHOST 'diss'
          edit QUEUE_EPILOG 'emos-exp'
          edit LOGDIR '/diss/log'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit SMSOUT '/vol/emos_nc/output'
        task req
          late -s +00:03 -c +00:10
          trigger collectreq eq complete
          label info ""
          event 1 
        task collectbitmap
          trigger req:001 or req eq complete
          edit QUEUE 'diss'
          edit USER 'emos'
          edit WSHOST 'diss'
          edit QUEUE_EPILOG 'emos-exp'
          edit LOGDIR '/diss/log'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit SMSOUT '/vol/emos_nc/output'
        task bitmap
          late -s +00:03 -c +00:10
          trigger collectbitmap eq complete
          label info ""
          event 1 
      endfamily
      family prod
        edit ACCOUNT 'oewavepg'
        edit SMSFILES '/home/ma/emos/def/elaw_36r1/prod'
        edit EMOS_TYPE 'fc'
        edit FSFAMILY 'ppdi'
        edit PROD_TASK '1'
        edit PGNPES '1'
        edit EMOS_TIME_STEP_H '0000'
        edit USER_PRIORITY '90'
        edit SMSURLBASE 'http://diss.ecmwf.int/do/monitoring/summary'
        edit SMSURL 'MWAVE/00'
        inlimit /limits:prod
        inlimit /elaw_36r1/limits:prod
        family 18
          edit DELTA_DAY '0'
          edit SMSURL 'MWAVE/18'
          family an
            trigger /elaw_36r1/main/00/euroshelf/wamrun eq complete and ( /elaw_36r1/main/00/getreq/bitmap:001 or /elaw_36r1/main/00/getreq/bitmap eq complete) and ( /elaw_36r1/main/00/getreq/req:001 or /elaw_36r1/main/00/getreq/req eq complete)
            edit EMOS_TYPE 'an'
            edit EMOS_BASE '18'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
        endfamily
        family 00
          edit EMOS_TYPE 'fc'
          edit PROD_TASK '1'
          edit DELTA_DAY '1'
          family an
            trigger /elaw_36r1/main/00/euroshelf/wamrun eq complete and /elaw_36r1/main/00/prod/18/an/prodgen eq complete
            edit EMOS_TYPE 'an'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 003
            trigger ../00/an/prodgen eq complete
            edit EMOS_TIME_STEP_H '0003'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 006
            trigger ./003/prodgen eq complete
            edit EMOS_TIME_STEP_H '0006'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 009
            trigger ./006/prodgen eq complete
            edit EMOS_TIME_STEP_H '0009'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 012
            trigger ./009/prodgen eq complete
            edit EMOS_TIME_STEP_H '0012'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 015
            trigger ./012/prodgen eq complete
            edit EMOS_TIME_STEP_H '0015'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 018
            trigger ./015/prodgen eq complete
            edit EMOS_TIME_STEP_H '0018'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 021
            trigger ./018/prodgen eq complete
            edit EMOS_TIME_STEP_H '0021'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 024
            trigger ./021/prodgen eq complete
            edit EMOS_TIME_STEP_H '0024'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 027
            trigger ./024/prodgen eq complete
            edit EMOS_TIME_STEP_H '0027'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 030
            trigger ./027/prodgen eq complete
            edit EMOS_TIME_STEP_H '0030'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 033
            trigger ./030/prodgen eq complete
            edit EMOS_TIME_STEP_H '0033'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 036
            trigger ./033/prodgen eq complete
            edit EMOS_TIME_STEP_H '0036'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 039
            trigger ./036/prodgen eq complete
            edit EMOS_TIME_STEP_H '0039'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 042
            trigger ./039/prodgen eq complete
            edit EMOS_TIME_STEP_H '0042'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 045
            trigger ./042/prodgen eq complete
            edit EMOS_TIME_STEP_H '0045'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 048
            trigger ./045/prodgen eq complete
            edit EMOS_TIME_STEP_H '0048'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 051
            trigger ./048/prodgen eq complete
            edit EMOS_TIME_STEP_H '0051'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 054
            trigger ./051/prodgen eq complete
            edit EMOS_TIME_STEP_H '0054'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 057
            trigger ./054/prodgen eq complete
            edit EMOS_TIME_STEP_H '0057'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 060
            trigger ./057/prodgen eq complete
            edit EMOS_TIME_STEP_H '0060'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 063
            trigger ./060/prodgen eq complete
            edit EMOS_TIME_STEP_H '0063'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 066
            trigger ./063/prodgen eq complete
            edit EMOS_TIME_STEP_H '0066'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 069
            trigger ./066/prodgen eq complete
            edit EMOS_TIME_STEP_H '0069'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 072
            trigger ./069/prodgen eq complete
            edit EMOS_TIME_STEP_H '0072'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 075
            trigger ./072/prodgen eq complete
            edit EMOS_TIME_STEP_H '0075'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 078
            trigger ./075/prodgen eq complete
            edit EMOS_TIME_STEP_H '0078'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 081
            trigger ./078/prodgen eq complete
            edit EMOS_TIME_STEP_H '0081'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 084
            trigger ./081/prodgen eq complete
            edit EMOS_TIME_STEP_H '0084'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 087
            trigger ./084/prodgen eq complete
            edit EMOS_TIME_STEP_H '0087'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 090
            trigger ./087/prodgen eq complete
            edit EMOS_TIME_STEP_H '0090'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 093
            trigger ./090/prodgen eq complete
            edit EMOS_TIME_STEP_H '0093'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 096
            trigger ./093/prodgen eq complete
            edit EMOS_TIME_STEP_H '0096'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 099
            trigger ./096/prodgen eq complete
            edit EMOS_TIME_STEP_H '0099'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 102
            trigger ./099/prodgen eq complete
            edit EMOS_TIME_STEP_H '0102'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 105
            trigger ./102/prodgen eq complete
            edit EMOS_TIME_STEP_H '0105'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 108
            trigger ./105/prodgen eq complete
            edit EMOS_TIME_STEP_H '0108'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 111
            trigger ./108/prodgen eq complete
            edit EMOS_TIME_STEP_H '0111'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 114
            trigger ./111/prodgen eq complete
            edit EMOS_TIME_STEP_H '0114'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 117
            trigger ./114/prodgen eq complete
            edit EMOS_TIME_STEP_H '0117'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
          family 120
            trigger ./117/prodgen eq complete
            edit EMOS_TIME_STEP_H '0120'
            task prodgen
              late -s +00:05 -c +00:15
          endfamily
        endfamily
      endfamily
      family prod2diss
        edit USER_PRIORITY '90'
        edit SMSURL 'MWAVE/00'
        edit SMSFILES '/home/ma/emos/def/elaw_36r1/prod'
        edit EMOS_TYPE 'fc'
        edit FSFAMILY 'ppdi'
        edit EMOS_TIME_STEP_H '0000'
        inlimit /elaw_36r1/limits:pdb
        inlimit /limits:pdb
        family 18
          edit DELTA_DAY '0'
          edit EMOS_TYPE 'an'
          edit SMSURL 'MWAVE/18'
          family an
            trigger /elaw_36r1/main/00/prod/18/an/prodgen eq complete
            edit EMOS_TYPE 'an'
            edit EMOS_BASE '18'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
        endfamily
        family 00
          edit EMOS_BASE '00'
          edit EMOS_TYPE 'fc'
          family an
            trigger /elaw_36r1/main/00/prod/00/an/prodgen eq complete
            edit EMOS_TYPE 'an'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 003
            trigger /elaw_36r1/main/00/prod/00/003/prodgen eq complete and ( ./an/diss:pdb or ./an/diss eq complete)
            edit EMOS_TIME_STEP_H '0003'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 006
            trigger /elaw_36r1/main/00/prod/00/006/prodgen eq complete and ( ./003/diss:pdb or ./003/diss eq complete)
            edit EMOS_TIME_STEP_H '0006'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 009
            trigger /elaw_36r1/main/00/prod/00/009/prodgen eq complete and ( ./006/diss:pdb or ./006/diss eq complete)
            edit EMOS_TIME_STEP_H '0009'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 012
            trigger /elaw_36r1/main/00/prod/00/012/prodgen eq complete and ( ./009/diss:pdb or ./009/diss eq complete)
            edit EMOS_TIME_STEP_H '0012'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 015
            trigger /elaw_36r1/main/00/prod/00/015/prodgen eq complete and ( ./012/diss:pdb or ./012/diss eq complete)
            edit EMOS_TIME_STEP_H '0015'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 018
            trigger /elaw_36r1/main/00/prod/00/018/prodgen eq complete and ( ./015/diss:pdb or ./015/diss eq complete)
            edit EMOS_TIME_STEP_H '0018'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 021
            trigger /elaw_36r1/main/00/prod/00/021/prodgen eq complete and ( ./018/diss:pdb or ./018/diss eq complete)
            edit EMOS_TIME_STEP_H '0021'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 024
            trigger /elaw_36r1/main/00/prod/00/024/prodgen eq complete and ( ./021/diss:pdb or ./021/diss eq complete)
            edit EMOS_TIME_STEP_H '0024'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 027
            trigger /elaw_36r1/main/00/prod/00/027/prodgen eq complete and ( ./024/diss:pdb or ./024/diss eq complete)
            edit EMOS_TIME_STEP_H '0027'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 030
            trigger /elaw_36r1/main/00/prod/00/030/prodgen eq complete and ( ./027/diss:pdb or ./027/diss eq complete)
            edit EMOS_TIME_STEP_H '0030'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 033
            trigger /elaw_36r1/main/00/prod/00/033/prodgen eq complete and ( ./030/diss:pdb or ./030/diss eq complete)
            edit EMOS_TIME_STEP_H '0033'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 036
            trigger /elaw_36r1/main/00/prod/00/036/prodgen eq complete and ( ./033/diss:pdb or ./033/diss eq complete)
            edit EMOS_TIME_STEP_H '0036'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 039
            trigger /elaw_36r1/main/00/prod/00/039/prodgen eq complete and ( ./036/diss:pdb or ./036/diss eq complete)
            edit EMOS_TIME_STEP_H '0039'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 042
            trigger /elaw_36r1/main/00/prod/00/042/prodgen eq complete and ( ./039/diss:pdb or ./039/diss eq complete)
            edit EMOS_TIME_STEP_H '0042'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 045
            trigger /elaw_36r1/main/00/prod/00/045/prodgen eq complete and ( ./042/diss:pdb or ./042/diss eq complete)
            edit EMOS_TIME_STEP_H '0045'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 048
            trigger /elaw_36r1/main/00/prod/00/048/prodgen eq complete and ( ./045/diss:pdb or ./045/diss eq complete)
            edit EMOS_TIME_STEP_H '0048'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 051
            trigger /elaw_36r1/main/00/prod/00/051/prodgen eq complete and ( ./048/diss:pdb or ./048/diss eq complete)
            edit EMOS_TIME_STEP_H '0051'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 054
            trigger /elaw_36r1/main/00/prod/00/054/prodgen eq complete and ( ./051/diss:pdb or ./051/diss eq complete)
            edit EMOS_TIME_STEP_H '0054'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 057
            trigger /elaw_36r1/main/00/prod/00/057/prodgen eq complete and ( ./054/diss:pdb or ./054/diss eq complete)
            edit EMOS_TIME_STEP_H '0057'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 060
            trigger /elaw_36r1/main/00/prod/00/060/prodgen eq complete and ( ./057/diss:pdb or ./057/diss eq complete)
            edit EMOS_TIME_STEP_H '0060'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 063
            trigger /elaw_36r1/main/00/prod/00/063/prodgen eq complete and ( ./060/diss:pdb or ./060/diss eq complete)
            edit EMOS_TIME_STEP_H '0063'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 066
            trigger /elaw_36r1/main/00/prod/00/066/prodgen eq complete and ( ./063/diss:pdb or ./063/diss eq complete)
            edit EMOS_TIME_STEP_H '0066'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 069
            trigger /elaw_36r1/main/00/prod/00/069/prodgen eq complete and ( ./066/diss:pdb or ./066/diss eq complete)
            edit EMOS_TIME_STEP_H '0069'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 072
            trigger /elaw_36r1/main/00/prod/00/072/prodgen eq complete and ( ./069/diss:pdb or ./069/diss eq complete)
            edit EMOS_TIME_STEP_H '0072'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 075
            trigger /elaw_36r1/main/00/prod/00/075/prodgen eq complete and ( ./072/diss:pdb or ./072/diss eq complete)
            edit EMOS_TIME_STEP_H '0075'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 078
            trigger /elaw_36r1/main/00/prod/00/078/prodgen eq complete and ( ./075/diss:pdb or ./075/diss eq complete)
            edit EMOS_TIME_STEP_H '0078'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 081
            trigger /elaw_36r1/main/00/prod/00/081/prodgen eq complete and ( ./078/diss:pdb or ./078/diss eq complete)
            edit EMOS_TIME_STEP_H '0081'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 084
            trigger /elaw_36r1/main/00/prod/00/084/prodgen eq complete and ( ./081/diss:pdb or ./081/diss eq complete)
            edit EMOS_TIME_STEP_H '0084'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 087
            trigger /elaw_36r1/main/00/prod/00/087/prodgen eq complete and ( ./084/diss:pdb or ./084/diss eq complete)
            edit EMOS_TIME_STEP_H '0087'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 090
            trigger /elaw_36r1/main/00/prod/00/090/prodgen eq complete and ( ./087/diss:pdb or ./087/diss eq complete)
            edit EMOS_TIME_STEP_H '0090'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 093
            trigger /elaw_36r1/main/00/prod/00/093/prodgen eq complete and ( ./090/diss:pdb or ./090/diss eq complete)
            edit EMOS_TIME_STEP_H '0093'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 096
            trigger /elaw_36r1/main/00/prod/00/096/prodgen eq complete and ( ./093/diss:pdb or ./093/diss eq complete)
            edit EMOS_TIME_STEP_H '0096'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 099
            trigger /elaw_36r1/main/00/prod/00/099/prodgen eq complete and ( ./096/diss:pdb or ./096/diss eq complete)
            edit EMOS_TIME_STEP_H '0099'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 102
            trigger /elaw_36r1/main/00/prod/00/102/prodgen eq complete and ( ./099/diss:pdb or ./099/diss eq complete)
            edit EMOS_TIME_STEP_H '0102'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 105
            trigger /elaw_36r1/main/00/prod/00/105/prodgen eq complete and ( ./102/diss:pdb or ./102/diss eq complete)
            edit EMOS_TIME_STEP_H '0105'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 108
            trigger /elaw_36r1/main/00/prod/00/108/prodgen eq complete and ( ./105/diss:pdb or ./105/diss eq complete)
            edit EMOS_TIME_STEP_H '0108'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 111
            trigger /elaw_36r1/main/00/prod/00/111/prodgen eq complete and ( ./108/diss:pdb or ./108/diss eq complete)
            edit EMOS_TIME_STEP_H '0111'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 114
            trigger /elaw_36r1/main/00/prod/00/114/prodgen eq complete and ( ./111/diss:pdb or ./111/diss eq complete)
            edit EMOS_TIME_STEP_H '0114'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 117
            trigger /elaw_36r1/main/00/prod/00/117/prodgen eq complete and ( ./114/diss:pdb or ./114/diss eq complete)
            edit EMOS_TIME_STEP_H '0117'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
          family 120
            trigger /elaw_36r1/main/00/prod/00/120/prodgen eq complete and ( ./117/diss:pdb or ./117/diss eq complete)
            edit EMOS_TIME_STEP_H '0120'
            task diss
              late -s +00:05 -c +00:15
              label info ""
              event pdb
          endfamily
        endfamily
        task check
          trigger ./00 eq complete
        family reservation
          trigger ./00 eq complete
          edit SMSFILES '/home/ma/emos/def/elaw_36r1/smsfiles'
          task loadl_cycle_limit
            edit LOADL_CYCLE '65'
            label info ""
        endfamily
      endfamily
    endfamily
    task logfilestore
      edit QUEUE 'emos-e'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit KEEPLOGS 'no'
      edit LOGSTOECFS 'no'
      edit MOVE_TO_ECFS 'yes'
      edit LOGTASK 'main'
    task logfiles
      late -s +00:05 -c +00:15
      trigger make eq complete and ./12 eq complete and ./00 eq complete and logfilestore eq complete
      edit QUEUE 'emos-e'
      edit QUEUE_EPILOG 'emos-exp'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit MOVE_TO_ECFS 'no'
      edit LOGTASK 'main'
      inlimit /limits:linux_cluster
  endfamily
  family lag
    trigger /elaw_36r1/main/make eq complete
    repeat date YMD 20091016 20121212 1
    edit LOGTASK 'lag'
    edit MARS_DB 'marsod'
    family 12
      edit EMOS_BASE '12'
      edit EMOS_PROJECT '12'
      edit DELTA_DAY '0'
      task wavesave
        trigger ( /elaw_36r1/main:YMD gt /elaw_36r1/lag:YMD) or ( /elaw_36r1/main:YMD eq /elaw_36r1/lag:YMD and /elaw_36r1/main/12/euroshelf/wamrun eq complete)
        edit SMSFILES '/home/ma/emos/def/elaw_36r1/euroshelf'
      family euroshelf
        trigger ./wavesave eq complete
        task buoycol
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /limits:linux_cluster
        task monthly
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit DATEMASK '04.*.*'
          inlimit /limits:linux_cluster
        task uracol
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /limits:linux_cluster
        task fc_plot
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /limits:linux_cluster
        task uraplot
          trigger uracol eq complete
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /limits:linux_cluster
        task windplot
          trigger uracol eq complete
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /limits:linux_cluster
        task uraesf
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /limits:linux_cluster
        task urascat
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit DATEMASK '03.*.*'
          inlimit /limits:linux_cluster
      endfamily
      family archive
        trigger ( /elaw_36r1/main:YMD gt /elaw_36r1/lag:YMD) or ( /elaw_36r1/main:YMD eq /elaw_36r1/lag:YMD and /elaw_36r1/main/12/euroshelf/wamrun eq complete)
        edit FSFAMILY 'mars'
        inlimit /elaw_36r1/limits:mars
        inlimit /limits:mars
        task medite_06
          edit EMOS_BASE '06'
          label files ""
          label info ""
          event fdb
        task medite_12
          edit EMOS_BASE '12'
          label files ""
          label info ""
          event fdb
        task fc12m
          label files ""
          label info ""
          event fdb
        task flush
          edit STREAM 'wave'
          label files ""
          label info ""
          event fdb
      endfamily
      family clean
        trigger /elaw_36r1/main/make eq complete and euroshelf eq complete and archive eq complete
        task clold
        task clean_ws
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /limits:linux_cluster
      endfamily
    endfamily
    family 00
      edit EMOS_BASE '00'
      edit EMOS_PROJECT '00'
      edit DELTA_DAY '1'
      edit QUEUE 'os'
      task wavesave
        trigger ( /elaw_36r1/main:YMD gt /elaw_36r1/lag:YMD) or ( /elaw_36r1/main:YMD eq /elaw_36r1/lag:YMD and /elaw_36r1/main/00/euroshelf/wamrun eq complete)
        edit SMSFILES '/home/ma/emos/def/elaw_36r1/euroshelf'
      family euroshelf
        trigger ./wavesave eq complete
        task buoycol
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /limits:linux_cluster
        task monthly
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit DATEMASK '04.*.*'
          inlimit /limits:linux_cluster
        task uracol
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /limits:linux_cluster
        task fc_plot
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /limits:linux_cluster
        task uraplot
          trigger uracol eq complete
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /limits:linux_cluster
        task windplot
          trigger uracol eq complete
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /limits:linux_cluster
        task uraesf
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /limits:linux_cluster
        task urascat
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          edit DATEMASK '03.*.*'
          inlimit /limits:linux_cluster
      endfamily
      family archive
        trigger ( /elaw_36r1/main:YMD gt /elaw_36r1/lag:YMD) or ( /elaw_36r1/main:YMD eq /elaw_36r1/lag:YMD and /elaw_36r1/main/00/euroshelf/wamrun eq complete)
        edit FSFAMILY 'mars'
        inlimit /elaw_36r1/limits:mars
        inlimit /limits:mars
        task medite_18
          edit DELTA_DAY '0'
          edit EMOS_BASE '18'
          label files ""
          label info ""
          event fdb
        task medite_00
          edit EMOS_BASE '00'
          label files ""
          label info ""
          event fdb
        task fc12m
          label files ""
          label info ""
          event fdb
        task flush
          edit STREAM 'wave'
          label files ""
          label info ""
          event fdb
      endfamily
      family clean
        trigger /elaw_36r1/main/make eq complete and euroshelf eq complete and archive eq complete
        task clold
        task clean_ws
          edit QUEUE 'emos-e'
          edit QUEUE_EPILOG 'emos-exp'
          edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
          edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
          edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
          edit LOGDIR '/vol/emos_nc/output'
          edit SMSOUT '/vol/emos_nc/output'
          inlimit /limits:linux_cluster
      endfamily
    endfamily
    task checkarchive
      trigger /elaw_36r1/main/make eq complete and ./12 eq complete and ./00 eq complete
      edit QUEUE 'emos-e'
      edit QUEUE_EPILOG 'emos-exp'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit FSFAMILY 'mars'
      edit EMOS_TYPE 'an'
      inlimit /limits:linux_cluster
    task logfilestore
      edit QUEUE 'emos-e'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit KEEPLOGS 'no'
      edit LOGSTOECFS 'no'
      edit MOVE_TO_ECFS 'yes'
      edit LOGTASK 'lag'
    task logfiles
      late -s +00:05 -c +00:15
      trigger /elaw_36r1/main/make eq complete and ./12 eq complete and ./00 eq complete and checkarchive eq complete and logfilestore eq complete
      edit QUEUE 'emos-e'
      edit QUEUE_EPILOG 'emos-exp'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit MOVE_TO_ECFS 'no'
      edit LOGTASK 'lag'
      inlimit /limits:linux_cluster
  endfamily
endsuite
suite ema_36r1
  edit USER 'emos'
  edit SCHOST 'c1b'
  edit STHOST '/s1b'
  edit STHOST_BKUP '/s1a'
  edit SCHOST_BKP 'c1a'
  edit WSHOST 'linux_cluster'
  edit WSLOGDIR '/vol/emos_nc/output'
  edit SCLOGDIR '/s1b/emos_esuite/emos_data/0046/log'
  edit SMSLOGHOST 'c1b'
  edit SMSLOGPORT '9316'
  edit SMSURLBASE 'http://diss.ecmwf.int/do/monitoring/arrival'
  edit SMSURL 'USA/GMAED/12'
  edit VERSION '0046'
  edit FIRST_DAY '1'
  edit SMSTRIES '2'
  edit DELTA_DAY '0'
  edit EMOS_BASE '12'
  edit EMOS_TIME_STEP_H '0000'
  edit FSFAMILY ''
  edit EPSTYPE 'fc'
  edit MEMBER '0'
  edit INIMA ''
  edit INIMACONS '0'
  edit INILEVELS '91'
  edit VERIFY 'verify'
  edit PERIOD_4D '12'
  edit WINDOW_LENGTH_4D '12'
  edit WINDOW_OFFSET_4D '9'
  edit ANNPES '48'
  edit ANMEM '3584'
  edit ANSTACK '700'
  edit ANMRFS '160'
  edit FCNPES '92'
  edit FCMEM '3584'
  edit FCSTACK '690'
  edit FCMRFS '32'
  edit ININPES '48'
  edit REINITIALIZE 'false'
  edit DATEMASK '*.*.*'
  edit YMD 'dummy'
  edit SMSHOME '/vol/emos_nc/output'
  edit SMSFILES '/home/ma/emos/def/ema_36r1'
  edit SMSINCLUDE '/home/ma/emos/def/ema_36r1/include'
  edit ACCOUNT 'oemaed'
  edit STREAM 'MAED'
  edit EMOS_STREAM 'MAED'
  edit MARS_DB 'marsod'
  edit SUITE_START '20040814'
  edit QUEUE 'ns'
  edit QUEUE_EPILOG 'ns'
  edit SMSCMD 'smssubmit.36r1 %USER% %SCHOST% %SMSJOB% %SMSJOBOUT% '
  edit SMSKILL 'smskill %USER% %SCHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
  edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %SCHOST% %SMSRID% %SMSJOB%'
  edit LOGDIR '/s1b/emos_esuite/emos_data/0046/log'
  edit SMSOUT '/s1b/emos_esuite/emos_data/0046/log'
  edit INIEXPVER '0046'
  edit SLEEPTIME '1'
  inlimit /limits:c1b
  clock hybrid -86400
  family limits
    limit inidata 1
    limit mars 3
    limit prod 5
    limit pdb 5
    limit q2diss 5
  endfamily
  family make
    task setup
    task links
      trigger setup eq complete
      edit FSFAMILY 'fc'
      edit WSHOST 'bee01'
    task datalinks
      trigger setup eq complete and links eq complete
      edit FSFAMILY 'fc'
    task ifs
      trigger links eq complete and datalinks eq complete
      edit FSFAMILY 'fc'
    task wamabs
      trigger links eq complete and datalinks eq complete
    task wconst
      trigger wamabs eq complete
    task ma_bin
      trigger links eq complete and datalinks eq complete
    task ma_tools
      trigger links eq complete and datalinks eq complete
    task prepdata
      trigger links eq complete and datalinks eq complete
  endfamily
  family init
    trigger /pp1/main/extractions/ma eq complete and make eq complete and /ema_36r1:ECF_DATE le /o/main:YMD
    task ma_clean
      edit FSFAMILY 'gfc'
    task rmfdb
      trigger ma_clean eq complete
      edit FSFAMILY 'mars'
    task flush
      trigger ma_clean eq complete
      edit STREAM 'MAED'
      edit EMOS_TYPE 'an'
      edit FSFAMILY 'mars'
      edit SMSFILES '/home/ma/emos/def/ema_36r1/archive'
      label files ""
      label info ""
      event fdb
    task ma_stage
      trigger ma_clean eq complete
      edit DATEMASK '01.*.*'
      edit QUEUE 'emos-e'
      edit QUEUE_EPILOG 'emos-exp'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      inlimit /limits:linux_cluster
    task ma_generate
      trigger ma_stage eq complete and ma_clean eq complete
    task getini
      trigger ma_generate eq complete
      edit INIMA 'mfr ncep ukm dwd '
      edit FSFAMILY 'gfc'
    task wavini
      trigger ma_stage eq complete and ma_clean eq complete
      edit FSFAMILY 'gfc'
  endfamily
  family dwd
    trigger init/getini eq complete and init/wavini eq complete and /ema_36r1/make eq complete
    edit INIMA 'dwd'
    edit INIMACONS '0'
    family fc
      edit FCLENGTH '240'
      edit FCMEM '2000'
      edit FCSTACK '700'
      edit FCMRFS '0'
      edit FSFAMILY 'gfc'
      task inidata
        edit THREADS '4'
        inlimit /ema_36r1/limits:inidata
      task getfcdata
        trigger inidata eq complete
        edit FSFAMILY 'fc'
      task getpersSST
        trigger getfcdata eq complete
        edit FSFAMILY 'fc'
        edit SSTLEN '240'
        edit SSTINT '1'
        edit THREADS '2'
        edit NPES '4'
        edit MEM '1000'
        edit CPUTIME '960'
      task model
        trigger getpersSST eq complete
        edit PERSST 'true'
        edit FSFAMILY 'fc'
        edit FCLENGTH '240'
        meter step -1 240 240
    endfamily
    family archive
      trigger fc/inidata eq complete
      edit SMSFILES '/home/ma/emos/def/ema_36r1/archive'
      edit EMOS_TYPE 'an'
      edit FSFAMILY 'mars'
      task ma_anml
        inlimit /ema_36r1/limits:mars
        inlimit /limits:mars
        label files ""
        label info ""
        event fdb
      task sp2fdb
        trigger ../fc/model:step ge 0 or ../fc/model eq complete
        edit EMOS_TYPE 'fc'
        edit STEPS '0'
        label files ""
        label info ""
        event fdb
      family fc000
        trigger sp2fdb eq complete and ( ../fc/model:step ge 0 or ../fc/model eq complete)
        edit STEPS '0'
        task fcml
          label files ""
          label info ""
          event fdb
        task fcsfc
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcpl
          trigger fcsfc:fdb or fcsfc eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcil
          trigger fcpl:fdb or fcpl eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      family fc024
        trigger ../fc/model:step ge 24 or ../fc/model eq complete
        edit STEPS '3/6/9/12/15/18/21/24'
        task fcsfc
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcpl
          trigger fcsfc:fdb or fcsfc eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcil
          trigger fcpl:fdb or fcpl eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      family fc036
        trigger ../fc/model:step ge 36 or ../fc/model eq complete
        edit STEPS '27/30/33/36'
        task fcsfc
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcpl
          trigger fcsfc:fdb or fcsfc eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcil
          trigger fcpl:fdb or fcpl eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      family fc060
        trigger ../fc/model:step ge 60 or ../fc/model eq complete
        edit STEPS '39/42/45/48/51/54/57/60'
        task fcsfc
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcpl
          trigger fcsfc:fdb or fcsfc eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcil
          trigger fcpl:fdb or fcpl eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      family fc072
        trigger ../fc/model:step ge 72 or ../fc/model eq complete
        edit STEPS '63/66/69/72'
        task fcsfc
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcpl
          trigger fcsfc:fdb or fcsfc eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcil
          trigger fcpl:fdb or fcpl eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      family fc120
        trigger ../fc/model:step ge 120 or ../fc/model eq complete
        edit STEPS '78/84/90/96/102/108/114/120'
        task fcsfc
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcpl
          trigger fcsfc:fdb or fcsfc eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcil
          trigger fcpl:fdb or fcpl eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      family fc180
        trigger ../fc/model:step ge 180 or ../fc/model eq complete
        edit STEPS '126/132/138/144/150/156/162/168/174/180'
        task fcsfc
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcpl
          trigger fcsfc:fdb or fcsfc eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcil
          trigger fcpl:fdb or fcpl eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      family fc240
        trigger ../fc/model:step ge 240 or ../fc/model eq complete
        edit STEPS '186/192/198/204/210/216/222/228/234/240'
        task fcsfc
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcpl
          trigger fcsfc:fdb or fcsfc eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcil
          trigger fcpl:fdb or fcpl eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcextra
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      task fc12g
        trigger ../fc/model eq complete
        edit STREAM 'MAED'
        edit STEPS '0/3/6/9/12/15/18/21/24/27/30/33/36/39/42/45/48/51/54/57/60/63/66/69/72/78/84/90/96/102/108/114/120/126/132/138/144/150/156/162/168/174/180/186/192/198/204/210/216/222/228/234/240'
        edit STEPS_SPECTRAL ''
        inlimit /ema_36r1/limits:mars
        inlimit /limits:mars
        label files ""
        label info ""
        event fdb
    endfamily
  endfamily
  family ukm
    trigger init/getini eq complete and init/wavini eq complete and /ema_36r1/make eq complete
    edit INIMA 'ukm'
    edit INIMACONS '0'
    family fc
      edit FCLENGTH '240'
      edit FCMEM '2000'
      edit FCSTACK '700'
      edit FCMRFS '0'
      edit FSFAMILY 'gfc'
      task inidata
        edit THREADS '4'
        inlimit /ema_36r1/limits:inidata
      task getfcdata
        trigger inidata eq complete
        edit FSFAMILY 'fc'
      task getpersSST
        trigger getfcdata eq complete
        edit FSFAMILY 'fc'
        edit SSTLEN '240'
        edit SSTINT '1'
        edit THREADS '2'
        edit NPES '4'
        edit MEM '1000'
        edit CPUTIME '960'
      task model
        trigger getpersSST eq complete
        edit PERSST 'true'
        edit FSFAMILY 'fc'
        edit FCLENGTH '240'
        meter step -1 240 240
    endfamily
    family archive
      trigger fc/inidata eq complete
      edit SMSFILES '/home/ma/emos/def/ema_36r1/archive'
      edit EMOS_TYPE 'an'
      edit FSFAMILY 'mars'
      task ma_anml
        inlimit /ema_36r1/limits:mars
        inlimit /limits:mars
        label files ""
        label info ""
        event fdb
      task sp2fdb
        trigger ../fc/model:step ge 0 or ../fc/model eq complete
        edit EMOS_TYPE 'fc'
        edit STEPS '0'
        label files ""
        label info ""
        event fdb
      family fc000
        trigger sp2fdb eq complete and ( ../fc/model:step ge 0 or ../fc/model eq complete)
        edit STEPS '0'
        task fcml
          label files ""
          label info ""
          event fdb
        task fcsfc
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcpl
          trigger fcsfc:fdb or fcsfc eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcil
          trigger fcpl:fdb or fcpl eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      family fc024
        trigger ../fc/model:step ge 24 or ../fc/model eq complete
        edit STEPS '3/6/9/12/15/18/21/24'
        task fcsfc
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcpl
          trigger fcsfc:fdb or fcsfc eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcil
          trigger fcpl:fdb or fcpl eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      family fc036
        trigger ../fc/model:step ge 36 or ../fc/model eq complete
        edit STEPS '27/30/33/36'
        task fcsfc
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcpl
          trigger fcsfc:fdb or fcsfc eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcil
          trigger fcpl:fdb or fcpl eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      family fc060
        trigger ../fc/model:step ge 60 or ../fc/model eq complete
        edit STEPS '39/42/45/48/51/54/57/60'
        task fcsfc
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcpl
          trigger fcsfc:fdb or fcsfc eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcil
          trigger fcpl:fdb or fcpl eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      family fc072
        trigger ../fc/model:step ge 72 or ../fc/model eq complete
        edit STEPS '63/66/69/72'
        task fcsfc
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcpl
          trigger fcsfc:fdb or fcsfc eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcil
          trigger fcpl:fdb or fcpl eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      family fc120
        trigger ../fc/model:step ge 120 or ../fc/model eq complete
        edit STEPS '78/84/90/96/102/108/114/120'
        task fcsfc
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcpl
          trigger fcsfc:fdb or fcsfc eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcil
          trigger fcpl:fdb or fcpl eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      family fc180
        trigger ../fc/model:step ge 180 or ../fc/model eq complete
        edit STEPS '126/132/138/144/150/156/162/168/174/180'
        task fcsfc
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcpl
          trigger fcsfc:fdb or fcsfc eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcil
          trigger fcpl:fdb or fcpl eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      family fc240
        trigger ../fc/model:step ge 240 or ../fc/model eq complete
        edit STEPS '186/192/198/204/210/216/222/228/234/240'
        task fcsfc
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcpl
          trigger fcsfc:fdb or fcsfc eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcil
          trigger fcpl:fdb or fcpl eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcextra
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      task fc12g
        trigger ../fc/model eq complete
        edit STREAM 'MAED'
        edit STEPS '0/3/6/9/12/15/18/21/24/27/30/33/36/39/42/45/48/51/54/57/60/63/66/69/72/78/84/90/96/102/108/114/120/126/132/138/144/150/156/162/168/174/180/186/192/198/204/210/216/222/228/234/240'
        edit STEPS_SPECTRAL ''
        inlimit /ema_36r1/limits:mars
        inlimit /limits:mars
        label files ""
        label info ""
        event fdb
    endfamily
  endfamily
  family ncep
    trigger init/getini eq complete and init/wavini eq complete and /ema_36r1/make eq complete
    edit INIMA 'ncep'
    edit INIMACONS '0'
    family fc
      edit FCLENGTH '240'
      edit FCMEM '2000'
      edit FCSTACK '700'
      edit FCMRFS '0'
      edit FSFAMILY 'gfc'
      task inidata
        edit THREADS '4'
        inlimit /ema_36r1/limits:inidata
      task getfcdata
        trigger inidata eq complete
        edit FSFAMILY 'fc'
      task getpersSST
        trigger getfcdata eq complete
        edit FSFAMILY 'fc'
        edit SSTLEN '240'
        edit SSTINT '1'
        edit THREADS '2'
        edit NPES '4'
        edit MEM '1000'
        edit CPUTIME '960'
      task model
        trigger getpersSST eq complete
        edit PERSST 'true'
        edit FSFAMILY 'fc'
        edit FCLENGTH '240'
        meter step -1 240 240
      task derived_param
        trigger model eq complete
        edit MEMBER 'off'
        edit SMSFILES '/home/ma/emos/def/ema_36r1/mc/fc'
        edit STEPS '0/12/24/36/48/60/72/84/96/108/120/132/144/156/168/180/192/204/216/228/240'
        edit EMOS_TYPE 'fc'
    endfamily
    family archive
      trigger fc/inidata eq complete
      edit SMSFILES '/home/ma/emos/def/ema_36r1/archive'
      edit EMOS_TYPE 'an'
      edit FSFAMILY 'mars'
      task ma_anml
        inlimit /ema_36r1/limits:mars
        inlimit /limits:mars
        label files ""
        label info ""
        event fdb
      task sp2fdb
        trigger ../fc/model:step ge 0 or ../fc/model eq complete
        edit EMOS_TYPE 'fc'
        edit STEPS '0'
        label files ""
        label info ""
        event fdb
      family fc000
        trigger sp2fdb eq complete and ( ../fc/model:step ge 0 or ../fc/model eq complete)
        edit STEPS '0'
        task fcml
          label files ""
          label info ""
          event fdb
        task fcsfc
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcpl
          trigger fcsfc:fdb or fcsfc eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcil
          trigger fcpl:fdb or fcpl eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      family fc024
        trigger ../fc/model:step ge 24 or ../fc/model eq complete
        edit STEPS '3/6/9/12/15/18/21/24'
        task fcsfc
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcpl
          trigger fcsfc:fdb or fcsfc eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcil
          trigger fcpl:fdb or fcpl eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      family fc036
        trigger ../fc/model:step ge 36 or ../fc/model eq complete
        edit STEPS '27/30/33/36'
        task fcsfc
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcpl
          trigger fcsfc:fdb or fcsfc eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcil
          trigger fcpl:fdb or fcpl eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      family fc060
        trigger ../fc/model:step ge 60 or ../fc/model eq complete
        edit STEPS '39/42/45/48/51/54/57/60'
        task fcsfc
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcpl
          trigger fcsfc:fdb or fcsfc eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcil
          trigger fcpl:fdb or fcpl eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      family fc072
        trigger ../fc/model:step ge 72 or ../fc/model eq complete
        edit STEPS '63/66/69/72'
        task fcsfc
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcpl
          trigger fcsfc:fdb or fcsfc eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcil
          trigger fcpl:fdb or fcpl eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      family fc120
        trigger ../fc/model:step ge 120 or ../fc/model eq complete
        edit STEPS '78/84/90/96/102/108/114/120'
        task fcsfc
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcpl
          trigger fcsfc:fdb or fcsfc eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcil
          trigger fcpl:fdb or fcpl eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      family fc180
        trigger ../fc/model:step ge 180 or ../fc/model eq complete
        edit STEPS '126/132/138/144/150/156/162/168/174/180'
        task fcsfc
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcpl
          trigger fcsfc:fdb or fcsfc eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcil
          trigger fcpl:fdb or fcpl eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      family fc240
        trigger ../fc/model:step ge 240 or ../fc/model eq complete
        edit STEPS '186/192/198/204/210/216/222/228/234/240'
        task fcsfc
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcpl
          trigger fcsfc:fdb or fcsfc eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcil
          trigger fcpl:fdb or fcpl eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcextra
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      task fc12g
        trigger ../fc/model eq complete
        edit STREAM 'MAED'
        edit STEPS '0/3/6/9/12/15/18/21/24/27/30/33/36/39/42/45/48/51/54/57/60/63/66/69/72/78/84/90/96/102/108/114/120/126/132/138/144/150/156/162/168/174/180/186/192/198/204/210/216/222/228/234/240'
        edit STEPS_SPECTRAL ''
        inlimit /ema_36r1/limits:mars
        inlimit /limits:mars
        label files ""
        label info ""
        event fdb
    endfamily
  endfamily
  family mfr
    trigger init/getini eq complete and init/wavini eq complete and /ema_36r1/make eq complete
    edit INIMA 'mfr'
    edit INIMACONS '0'
    family fc
      edit FCLENGTH '240'
      edit FCMEM '2000'
      edit FCSTACK '700'
      edit FCMRFS '0'
      edit FSFAMILY 'gfc'
      task inidata
        edit THREADS '4'
        inlimit /ema_36r1/limits:inidata
      task getfcdata
        trigger inidata eq complete
        edit FSFAMILY 'fc'
      task getpersSST
        trigger getfcdata eq complete
        edit FSFAMILY 'fc'
        edit SSTLEN '240'
        edit SSTINT '1'
        edit THREADS '2'
        edit NPES '4'
        edit MEM '1000'
        edit CPUTIME '960'
      task model
        trigger getpersSST eq complete
        edit PERSST 'true'
        edit FSFAMILY 'fc'
        edit FCLENGTH '240'
        meter step -1 240 240
    endfamily
    family archive
      trigger fc/inidata eq complete
      edit SMSFILES '/home/ma/emos/def/ema_36r1/archive'
      edit EMOS_TYPE 'an'
      edit FSFAMILY 'mars'
      task ma_anml
        inlimit /ema_36r1/limits:mars
        inlimit /limits:mars
        label files ""
        label info ""
        event fdb
      task sp2fdb
        trigger ../fc/model:step ge 0 or ../fc/model eq complete
        edit EMOS_TYPE 'fc'
        edit STEPS '0'
        label files ""
        label info ""
        event fdb
      family fc000
        trigger sp2fdb eq complete and ( ../fc/model:step ge 0 or ../fc/model eq complete)
        edit STEPS '0'
        task fcml
          label files ""
          label info ""
          event fdb
        task fcsfc
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcpl
          trigger fcsfc:fdb or fcsfc eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcil
          trigger fcpl:fdb or fcpl eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      family fc024
        trigger ../fc/model:step ge 24 or ../fc/model eq complete
        edit STEPS '3/6/9/12/15/18/21/24'
        task fcsfc
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcpl
          trigger fcsfc:fdb or fcsfc eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcil
          trigger fcpl:fdb or fcpl eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      family fc036
        trigger ../fc/model:step ge 36 or ../fc/model eq complete
        edit STEPS '27/30/33/36'
        task fcsfc
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcpl
          trigger fcsfc:fdb or fcsfc eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcil
          trigger fcpl:fdb or fcpl eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      family fc060
        trigger ../fc/model:step ge 60 or ../fc/model eq complete
        edit STEPS '39/42/45/48/51/54/57/60'
        task fcsfc
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcpl
          trigger fcsfc:fdb or fcsfc eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcil
          trigger fcpl:fdb or fcpl eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      family fc072
        trigger ../fc/model:step ge 72 or ../fc/model eq complete
        edit STEPS '63/66/69/72'
        task fcsfc
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcpl
          trigger fcsfc:fdb or fcsfc eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcil
          trigger fcpl:fdb or fcpl eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      family fc120
        trigger ../fc/model:step ge 120 or ../fc/model eq complete
        edit STEPS '78/84/90/96/102/108/114/120'
        task fcsfc
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcpl
          trigger fcsfc:fdb or fcsfc eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcil
          trigger fcpl:fdb or fcpl eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      family fc180
        trigger ../fc/model:step ge 180 or ../fc/model eq complete
        edit STEPS '126/132/138/144/150/156/162/168/174/180'
        task fcsfc
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcpl
          trigger fcsfc:fdb or fcsfc eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcil
          trigger fcpl:fdb or fcpl eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      family fc240
        trigger ../fc/model:step ge 240 or ../fc/model eq complete
        edit STEPS '186/192/198/204/210/216/222/228/234/240'
        task fcsfc
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcpl
          trigger fcsfc:fdb or fcsfc eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcil
          trigger fcpl:fdb or fcpl eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcextra
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      task fc12g
        trigger ../fc/model eq complete
        edit STREAM 'MAED'
        edit STEPS '0/3/6/9/12/15/18/21/24/27/30/33/36/39/42/45/48/51/54/57/60/63/66/69/72/78/84/90/96/102/108/114/120/126/132/138/144/150/156/162/168/174/180/186/192/198/204/210/216/222/228/234/240'
        edit STEPS_SPECTRAL ''
        inlimit /ema_36r1/limits:mars
        inlimit /limits:mars
        label files ""
        label info ""
        event fdb
    endfamily
  endfamily
  family consensus
    trigger init/getini eq complete and init/wavini eq complete and /ema_36r1/make eq complete
    edit INIMA 'mfr ncep ukm dwd '
    edit INIMACONS '1'
    family fc
      edit FCLENGTH '240'
      edit FCMEM '2000'
      edit FCSTACK '700'
      edit FCMRFS '0'
      edit FSFAMILY 'gfc'
      task inidata
        edit THREADS '4'
        inlimit /ema_36r1/limits:inidata
      task getfcdata
        trigger inidata eq complete
        edit FSFAMILY 'fc'
      task getpersSST
        trigger getfcdata eq complete
        edit FSFAMILY 'fc'
        edit SSTLEN '240'
        edit SSTINT '1'
        edit THREADS '2'
        edit NPES '4'
        edit MEM '1000'
        edit CPUTIME '960'
      task model
        trigger getpersSST eq complete
        edit PERSST 'true'
        edit FSFAMILY 'fc'
        edit FCLENGTH '240'
        meter step -1 240 240
    endfamily
    family archive
      trigger fc/inidata eq complete
      edit SMSFILES '/home/ma/emos/def/ema_36r1/archive'
      edit EMOS_TYPE 'an'
      edit FSFAMILY 'mars'
      task ma_anml
        inlimit /ema_36r1/limits:mars
        inlimit /limits:mars
        label files ""
        label info ""
        event fdb
      task sp2fdb
        trigger ../fc/model:step ge 0 or ../fc/model eq complete
        edit EMOS_TYPE 'fc'
        edit STEPS '0'
        label files ""
        label info ""
        event fdb
      family fc000
        trigger sp2fdb eq complete and ( ../fc/model:step ge 0 or ../fc/model eq complete)
        edit STEPS '0'
        task fcml
          label files ""
          label info ""
          event fdb
        task fcsfc
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcpl
          trigger fcsfc:fdb or fcsfc eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcil
          trigger fcpl:fdb or fcpl eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      family fc024
        trigger ../fc/model:step ge 24 or ../fc/model eq complete
        edit STEPS '3/6/9/12/15/18/21/24'
        task fcsfc
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcpl
          trigger fcsfc:fdb or fcsfc eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcil
          trigger fcpl:fdb or fcpl eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      family fc036
        trigger ../fc/model:step ge 36 or ../fc/model eq complete
        edit STEPS '27/30/33/36'
        task fcsfc
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcpl
          trigger fcsfc:fdb or fcsfc eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcil
          trigger fcpl:fdb or fcpl eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      family fc060
        trigger ../fc/model:step ge 60 or ../fc/model eq complete
        edit STEPS '39/42/45/48/51/54/57/60'
        task fcsfc
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcpl
          trigger fcsfc:fdb or fcsfc eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcil
          trigger fcpl:fdb or fcpl eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      family fc072
        trigger ../fc/model:step ge 72 or ../fc/model eq complete
        edit STEPS '63/66/69/72'
        task fcsfc
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcpl
          trigger fcsfc:fdb or fcsfc eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcil
          trigger fcpl:fdb or fcpl eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      family fc120
        trigger ../fc/model:step ge 120 or ../fc/model eq complete
        edit STEPS '78/84/90/96/102/108/114/120'
        task fcsfc
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcpl
          trigger fcsfc:fdb or fcsfc eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcil
          trigger fcpl:fdb or fcpl eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      family fc180
        trigger ../fc/model:step ge 180 or ../fc/model eq complete
        edit STEPS '126/132/138/144/150/156/162/168/174/180'
        task fcsfc
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcpl
          trigger fcsfc:fdb or fcsfc eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcil
          trigger fcpl:fdb or fcpl eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      family fc240
        trigger ../fc/model:step ge 240 or ../fc/model eq complete
        edit STEPS '186/192/198/204/210/216/222/228/234/240'
        task fcsfc
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcpl
          trigger fcsfc:fdb or fcsfc eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcil
          trigger fcpl:fdb or fcpl eq complete
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
        task fcextra
          inlimit /ema_36r1/limits:mars
          inlimit /limits:mars
          label files ""
          label info ""
          event fdb
      endfamily
      task fc12g
        trigger ../fc/model eq complete
        edit STREAM 'MAED'
        edit STEPS '0/3/6/9/12/15/18/21/24/27/30/33/36/39/42/45/48/51/54/57/60/63/66/69/72/78/84/90/96/102/108/114/120/126/132/138/144/150/156/162/168/174/180/186/192/198/204/210/216/222/228/234/240'
        edit STEPS_SPECTRAL ''
        inlimit /ema_36r1/limits:mars
        inlimit /limits:mars
        label files ""
        label info ""
        event fdb
    endfamily
  endfamily
  family verify
    trigger dwd eq complete and ukm eq complete and ncep eq complete and mfr eq complete and consensus eq complete
    edit QUEUE 'emos-e'
    edit QUEUE_EPILOG 'emos-exp'
    edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
    edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
    edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
    edit LOGDIR '/vol/emos_nc/output'
    edit SMSOUT '/vol/emos_nc/output'
    edit SMSFILES '/home/ma/emos/def/ema_36r1/verif'
    inlimit /limits:linux_cluster
    task vmaprep
    family washington
      trigger vmaprep eq complete
      edit CENTRE 'washington'
      edit CENTRE_STREAM 'ma_washington_36r1'
      task vmacrd
        edit DATEMASK '28.*.*'
      task vmali1
      task vmali2
      task vmaup1
        trigger vmali1 eq complete
      task vmaup2
        trigger vmaup1 eq complete and vmali2 eq complete
      task vmacrmv
        edit DATEMASK '28.*.*'
      task vmamemv
        trigger vmaup2 eq complete
        edit DATEMASK '01.*.*'
      task vmaupmv
        trigger vmamemv eq complete
        edit DATEMASK '01.*.*'
      task vmacrmb
        edit DATEMASK '28.*.*'
      task vmamemb
        trigger vmaup2 eq complete
        edit DATEMASK '10.*.*'
      task vmaupmb
        trigger vmamemb eq complete
        edit DATEMASK '10.*.*'
    endfamily
    family bracknell
      trigger vmaprep eq complete
      edit CENTRE 'bracknell'
      edit CENTRE_STREAM 'ma_bracknell_36r1'
      task vmacrd
        edit DATEMASK '28.*.*'
      task vmali1
      task vmali2
      task vmaup1
        trigger vmali1 eq complete
      task vmaup2
        trigger vmaup1 eq complete and vmali2 eq complete
      task vmacrmv
        edit DATEMASK '28.*.*'
      task vmamemv
        trigger vmaup2 eq complete
        edit DATEMASK '01.*.*'
      task vmaupmv
        trigger vmamemv eq complete
        edit DATEMASK '01.*.*'
      task vmacrmb
        edit DATEMASK '28.*.*'
      task vmamemb
        trigger vmaup2 eq complete
        edit DATEMASK '10.*.*'
      task vmaupmb
        trigger vmamemb eq complete
        edit DATEMASK '10.*.*'
    endfamily
    family offenbach
      trigger vmaprep eq complete
      edit CENTRE 'offenbach'
      edit CENTRE_STREAM 'ma_offenbach_36r1'
      task vmacrd
        edit DATEMASK '28.*.*'
      task vmali1
      task vmali2
      task vmaup1
        trigger vmali1 eq complete
      task vmaup2
        trigger vmaup1 eq complete and vmali2 eq complete
      task vmacrmv
        edit DATEMASK '28.*.*'
      task vmamemv
        trigger vmaup2 eq complete
        edit DATEMASK '01.*.*'
      task vmaupmv
        trigger vmamemv eq complete
        edit DATEMASK '01.*.*'
      task vmacrmb
        edit DATEMASK '28.*.*'
      task vmamemb
        trigger vmaup2 eq complete
        edit DATEMASK '10.*.*'
      task vmaupmb
        trigger vmamemb eq complete
        edit DATEMASK '10.*.*'
    endfamily
    family toulouse
      trigger vmaprep eq complete
      edit CENTRE 'toulouse'
      edit CENTRE_STREAM 'ma_toulouse_36r1'
      task vmacrd
        edit DATEMASK '28.*.*'
      task vmali1
      task vmali2
      task vmaup1
        trigger vmali1 eq complete
      task vmaup2
        trigger vmaup1 eq complete and vmali2 eq complete
      task vmacrmv
        edit DATEMASK '28.*.*'
      task vmamemv
        trigger vmaup2 eq complete
        edit DATEMASK '01.*.*'
      task vmaupmv
        trigger vmamemv eq complete
        edit DATEMASK '01.*.*'
      task vmacrmb
        edit DATEMASK '28.*.*'
      task vmamemb
        trigger vmaup2 eq complete
        edit DATEMASK '10.*.*'
      task vmaupmb
        trigger vmamemb eq complete
        edit DATEMASK '10.*.*'
    endfamily
    family consensus
      trigger vmaprep eq complete
      edit CENTRE 'consensus'
      edit CENTRE_STREAM 'ma_consensus_36r1'
      task vmacrd
        edit DATEMASK '28.*.*'
      task vmali1
      task vmali2
      task vmaup1
        trigger vmali1 eq complete
      task vmaup2
        trigger vmaup1 eq complete and vmali2 eq complete
      task vmacrmv
        edit DATEMASK '28.*.*'
      task vmamemv
        trigger vmaup2 eq complete
        edit DATEMASK '01.*.*'
      task vmaupmv
        trigger vmamemv eq complete
        edit DATEMASK '01.*.*'
      task vmacrmb
        edit DATEMASK '28.*.*'
      task vmamemb
        trigger vmaup2 eq complete
        edit DATEMASK '10.*.*'
      task vmaupmb
        trigger vmamemb eq complete
        edit DATEMASK '10.*.*'
    endfamily
  endfamily
  family trans
    complete /ema_36r1/ncep eq complete and /ema_36r1/ncep/fc/model:step lt 240
    trigger /ema_36r1/ncep/fc eq complete and /ema_36r1/consensus/fc eq complete
    edit EMOS_TYPE 'fc'
    edit SMSFILES '/home/ma/emos/def/ema_36r1/archive'
    edit FSFAMILY 'ppdi'
    edit SMSTRIES '1'
    edit EMOS_STREAM 'MAED'
    edit SMSURLBASE 'http://diss.ecmwf.int/do/monitoring/arrival'
    edit SMSURL 'USA/GMAED/12'
    edit SCHEDULE_DISS_TIME '0'
    inlimit /ema_36r1/limits:prod
    family 000
      edit STEP '000'
      task trans_ma
    endfamily
    family 012
      trigger ./000 eq complete
      edit STEP '012'
      task trans_ma
    endfamily
    family 024
      trigger ./012 eq complete
      edit STEP '024'
      task trans_ma
    endfamily
    family 036
      trigger ./024 eq complete
      edit STEP '036'
      task trans_ma
    endfamily
    family 048
      trigger ./036 eq complete
      edit STEP '048'
      task trans_ma
    endfamily
    family 060
      trigger ./048 eq complete
      edit STEP '060'
      task trans_ma
    endfamily
    family 072
      trigger ./060 eq complete
      edit STEP '072'
      task trans_ma
    endfamily
    family 084
      trigger ./072 eq complete
      edit STEP '084'
      task trans_ma
    endfamily
    family 096
      trigger ./084 eq complete
      edit STEP '096'
      task trans_ma
    endfamily
    family 108
      trigger ./096 eq complete
      edit STEP '108'
      task trans_ma
    endfamily
    family 120
      trigger ./108 eq complete
      edit STEP '120'
      task trans_ma
    endfamily
    family 132
      trigger ./120 eq complete
      edit STEP '132'
      task trans_ma
    endfamily
    family 144
      trigger ./132 eq complete
      edit STEP '144'
      task trans_ma
    endfamily
    family 156
      trigger ./144 eq complete
      edit STEP '156'
      task trans_ma
    endfamily
    family 168
      trigger ./156 eq complete
      edit STEP '168'
      task trans_ma
    endfamily
    family 180
      trigger ./168 eq complete
      edit STEP '180'
      task trans_ma
    endfamily
    family 192
      trigger ./180 eq complete
      edit STEP '192'
      task trans_ma
    endfamily
    family 204
      trigger ./192 eq complete
      edit STEP '204'
      task trans_ma
    endfamily
    family 216
      trigger ./204 eq complete
      edit STEP '216'
      task trans_ma
    endfamily
    family 228
      trigger ./216 eq complete
      edit STEP '228'
      task trans_ma
    endfamily
    family 240
      trigger ./228 eq complete
      edit STEP '240'
      task trans_ma
    endfamily
  endfamily
  family pop
    edit QUEUE 'emos-e'
    edit QUEUE_EPILOG 'emos-exp'
    edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
    edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
    edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
    edit LOGDIR '/vol/emos_nc/output'
    edit SMSOUT '/vol/emos_nc/output'
    edit SMSFILES '/home/ma/emos/def/ema_36r1/plot'
    inlimit /limits:linux_cluster
    family 12
      trigger /ema_36r1/dwd eq complete and /ema_36r1/ukm eq complete and /ema_36r1/ncep eq complete and /ema_36r1/mfr eq complete and /ema_36r1/consensus eq complete
      edit EMOS_BASE '12'
      family paper
        task ma_plot
      endfamily
    endfamily
  endfamily
  family msjobs
    edit QUEUE 'emos'
    edit QUEUE_EPILOG 'emos-exp'
    edit WSHOST 'ecgate'
    edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
    edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
    edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
    edit SMSOUT '/vol/emos_nc/output'
    edit LOGDIR '/vol/emos_nc/output'
    edit VERSION '0046'
    edit STREAM 'MA'
    edit EMOS_STREAM 'MA'
    edit SMSTRYNO '1'
    edit SMSURLBASE 'http://ecgate.ecmwf.int:9080'
    edit SMSURL 'do/events'
    family 12
      trigger ../trans eq complete
      edit SMSFILES '/home/ma/emos/def/ema_36r1/msjobs'
      family msma
        edit EMOS_TIME_STEP_H '240'
        edit MSJ_DIR 'ma12h240'
        task ms_ec
      endfamily
    endfamily
  endfamily
  family check
    task checkarchive
      edit QUEUE 'emos-e'
      edit QUEUE_EPILOG 'emos-exp'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit FSFAMILY 'mars'
      edit EMOS_TYPE 'an'
      inlimit /limits:linux_cluster
      label info ""
  endfamily
  family save
    edit SMSFILES '/home/ma/emos/def/ema_36r1/restart'
    task logfilestore
      edit QUEUE 'emos-e'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit KEEPLOGS 'no'
      edit LOGSTOECFS 'no'
      edit MOVE_TO_ECFS 'yes'
  endfamily
  family log
    trigger make eq complete and init eq complete and dwd eq complete and ukm eq complete and ncep eq complete and mfr eq complete and consensus eq complete and verify eq complete and trans eq complete and pop eq complete and msjobs eq complete and check eq complete and save eq complete
    edit SMSFILES '/home/ma/emos/def/ema_36r1/restart'
    task logfiles
      late -s +00:05 -c +00:15
      edit QUEUE 'emos-e'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit MOVE_TO_ECFS 'no'
      time 14:00
  endfamily
endsuite
suite ejason2_val
  edit SMSHOME '/vol/emos_nc/output'
  edit SMSFILES '/home/ma/emos/def/ejason2_val/smsfiles'
  edit SMSINCLUDE '/home/ma/emos/def/ejason2_val/include'
  edit MEM '2500'
  edit USER 'emos'
  edit SCHOST 'c1b'
  edit SCHOST_BKUP 'c1a'
  edit WSHOST 'linux_cluster'
  edit ACCOUNT 'oesu'
  edit WSLOGDIR '/vol/emos_nc/output'
  edit SCLOGDIR '/s1b/emos_esuite/emos_data/9002/log'
  edit SMSLOGHOST 'c1b'
  edit SMSLOGPORT '9316'
  edit SMSURLBASE 'http://www.cnes.fr/web'
  edit SMSURL '1441-jason.php'
  edit QUEUE 'ns'
  edit QUEUE_EPILOG 'ns'
  edit SMSCMD 'smssubmit %USER% %SCHOST% %SMSJOB% %SMSJOBOUT%'
  edit SMSKILL 'smskill %USER% %SCHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
  edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %SCHOST% %SMSRID% %SMSJOB%'
  edit LOGDIR '/s1b/emos_esuite/emos_data/9002/log'
  edit SMSOUT '/s1b/emos_esuite/emos_data/9002/log'
  edit VERSION '5'
  edit EXPVER '9002'
  edit GRID '400'
  edit HOMEDIR '/s1b/emos_esuite/emos_data/jasonII'
  edit SLEEPTIME '1'
  inlimit /limits:c1b
  clock hybrid 
  family limits
    limit limiter 1
    limit c1b 7
    limit swarm 7
  endfamily
  family main
    repeat date YMD 20071231 20091231 1
    edit LOGTASK 'main'
    edit CALL_ECPDS 'false'
    edit DESTINATION 'EMI'
    edit ECPDS_STREAM 'JSN2'
    family make
      task common
      task local
        trigger common eq complete
      task purge
    endfamily
    task cleaner
      trigger make eq complete
      edit KEEP_DAYS '2'
    task stage
    family prod
      trigger make eq complete
      inlimit /ejason2_val/limits:limiter
      family 15P
        edit LABELS '15P'
        edit EMOS_CYCLE '12'
        task jason2_val
          trigger ( /ejason2_val/main:YMD lt /o/main:YMD) or ( /ejason2_val/main:YMD eq /o/main:YMD and /o/main/12/fc/model eq complete )
      endfamily
      family 18P
        edit LABELS '18P'
        edit EMOS_CYCLE '12'
        task jason2_val
          trigger ( /ejason2_val/main:YMD lt /o/main:YMD) or ( /ejason2_val/main:YMD eq /o/main:YMD and /o/main/12/fc/model eq complete )
      endfamily
      family 21P
        edit LABELS '21P'
        edit EMOS_CYCLE '12'
        task jason2_val
          trigger ( /ejason2_val/main:YMD lt /o/main:YMD) or ( /ejason2_val/main:YMD eq /o/main:YMD and /o/main/12/fc/model eq complete )
      endfamily
      family 24P
        edit LABELS '24P'
        edit EMOS_CYCLE '12'
        task jason2_val
          trigger ( /ejason2_val/main:YMD lt /o/main:YMD) or ( /ejason2_val/main:YMD eq /o/main:YMD and /o/main/12/fc/model eq complete )
      endfamily
      family 03P
        edit LABELS '03P'
        edit DELTA_DAY '1'
        edit EMOS_CYCLE '00'
        task jason2_val
          trigger ( /ejason2_val/main:YMD lt /o/main:YMD) or ( /ejason2_val/main:YMD eq /o/main:YMD and /o/main/00/fc/model eq complete )
      endfamily
      family 06P
        edit LABELS '06P'
        edit DELTA_DAY '1'
        edit EMOS_CYCLE '00'
        task jason2_val
          trigger ( /ejason2_val/main:YMD lt /o/main:YMD) or ( /ejason2_val/main:YMD eq /o/main:YMD and /o/main/00/fc/model eq complete )
      endfamily
      family 09P
        edit LABELS '09P'
        edit DELTA_DAY '1'
        edit EMOS_CYCLE '00'
        task jason2_val
          trigger ( /ejason2_val/main:YMD lt /o/main:YMD) or ( /ejason2_val/main:YMD eq /o/main:YMD and /o/main/00/fc/model eq complete )
      endfamily
      family 12P
        edit LABELS '12P'
        edit DELTA_DAY '1'
        edit EMOS_CYCLE '00'
        task jason2_val
          trigger ( /ejason2_val/main:YMD lt /o/main:YMD) or ( /ejason2_val/main:YMD eq /o/main:YMD and /o/main/00/fc/model eq complete )
        task arch_val
          trigger ( ./jason2_val eq complete) and ( ../03P eq complete) and ( ../06P eq complete) and ( ../09P eq complete) and ( ../24P eq complete) and ( ../15P eq complete) and ( ../18P eq complete) and ( ../21P eq complete)
          edit DELTA_DAY '0'
      endfamily
    endfamily
    task logfilestore
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit KEEPLOGS 'no'
      edit LOGSTOECFS 'no'
      edit MOVE_TO_ECFS 'yes'
    task logfiles
      trigger ( prod eq complete) and ( logfilestore eq complete)
      edit QUEUE 'emos'
      edit QUEUE_EPILOG 'emos-exp'
      edit WSHOST 'localhost'
      edit SMSCMD 'smssubmit %USER% %WSHOST% %SMSJOB% %SMSJOBOUT%'
      edit SMSKILL 'smskill %USER% %WSHOST% %SMSRID% %SMSJOB% %SMSJOBOUT%'
      edit SMSSTATUSCMD '/home/ma/emos/bin/smsstatus %USER% %WSHOST% %SMSRID% %SMSJOB%'
      edit LOGDIR '/vol/emos_nc/output'
      edit SMSOUT '/vol/emos_nc/output'
      edit MOVE_TO_ECFS 'no'
  endfamily
endsuite
